// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2011-2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2011-2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <malloc.h>
#include <algorithm>
#include <limits>
#include <string>
#include <set>
#include <clause_propagator.hpp>
#include <comparison_propagators.hpp>
#include <arithmetic_propagators.hpp>
#include <interval_propagators.hpp>
#include <neq_propagators.hpp>
#include <operator_definition_registry.hpp>
#include <AST_builder.hpp>

using namespace std;
using namespace hcsp;
using namespace CSPXMLParser;

void AST_builder::equate_variables(unsigned var_a, unsigned var_b)
{
    unsigned equiv_a= m_operator_connectivity.get_equivalent(var_a);
    unsigned equiv_b= m_operator_connectivity.get_equivalent(var_b);
    if (equiv_a == equiv_b)
        return;

    if (are_ne(equiv_a, equiv_b)) {
        mark_unsat();
        return;
    }
        
    const iset & result= intersect_variable(m_solver.var_domain(var_a), var_b);
    m_solver.set_var_domain(var_a, result);
    m_operator_connectivity.mark_equivalent(equiv_a, equiv_b);
    add_constraint<equal_propagator>(2) << var_a << var_b;

    // update any potential m_diff_set which pointed to the old equivalence
    // set representative to point to the new one.
    // This way are_ne() will always return a correct answer.
    unsigned result_equiv= m_operator_connectivity.get_equivalent(var_a);

    if (result_equiv == equiv_a)
        replace_var_in_diff_set(equiv_b, result_equiv);
    else
        replace_var_in_diff_set(equiv_a, result_equiv);
}

void AST_builder::replace_var_in_diff_set(unsigned from_var, unsigned to_var)
{
    if (m_diff_set.size() <= from_var || m_diff_set[from_var].empty())
        return;

    if (m_diff_set.size() <= to_var)
        m_diff_set.resize(to_var+1);

    std::set<unsigned>::iterator it= m_diff_set[from_var].begin();
    const std::set<unsigned>::iterator end= m_diff_set[from_var].end();

    while ( it!=end )  {
        unsigned diff_with= *it;
        
        HCSP_ASSERT(diff_with != to_var || is_marked_unsat());

        m_diff_set[diff_with].erase(from_var);
        m_diff_set[diff_with].insert(to_var);
        m_diff_set[from_var].erase(it++);
        m_diff_set[to_var].insert(diff_with);
    }
}

template <typename Iterator>
bool is_sorted(Iterator from, Iterator to)
{
    if (from == to)
        return true;
    Iterator next= from;
    ++next;
    for ( ; next != to ; ++next, ++from)
        if (*next < *to)
            return false;
    return true;
}

void AST_builder::instantiate_all_different(
    const vector<unsigned> &variables)
{
    if (variables.size() < 32) {
        
        for (unsigned i=1 ; i < variables.size() ; ++i)
            for (unsigned j=0 ; j != i ; ++j) {
                add_constraint<neq_propagator>(2)
                    << variables[i] << variables[j];
            }
        if (variables.size() <= 3)
            return;
    }
    constraint_connectivity_info & info=
        add_constraint<all_different_propagator>(variables.size());
    for (unsigned i=0 ; i != variables.size() ; ++i)
        info << variables[i];
}

void AST_builder::instantiate_weighted_sum(const vector<int> &coefficients,
                                           const vector<unsigned> &variables,
                                           bool is_ge, long long k)
{
    HCSP_ASSERT(coefficients.size() == variables.size());
    typedef linear_combination_ge_propagator::coefficient_type coeff_type;
    vector<coeff_type> actual_coefficients;
    actual_coefficients.reserve(coefficients.size());
    for (long coefficient : coefficients) {
        if (!is_ge)
            coefficient *= -1;
        if ( coefficient > numeric_limits<coeff_type>::max() ||
             coefficient < numeric_limits<coeff_type>::min()) {
            throw unsupported("coefficients of weightedsum are overflowing");
        }
        actual_coefficients.push_back(coefficient);
    }
    if (!is_ge)
        k *= -1;
    if ( k > numeric_limits<iset::value_type>::max() ||
         k < numeric_limits<iset::value_type>::min())
        throw unsupported("Limit of weightedsum is overflowing");
    constraint_connectivity_info & info=
        add_constraint<linear_combination_ge_propagator>(
            variables.size(), (iset::value_type)k, actual_coefficients);
    for (unsigned i=0 ; i != variables.size() ; ++i)
        info << variables[i]; 
}

void AST_builder::collect_or_vars(vector<unsigned> & or_variables,
                                  const AST & expr,
                                  const vector<unsigned> & variables,
                                  usage_type used_as)
{
    vector<const AST*> descendants;
    insert_descendants(expr, expr.getType(), descendants);
    for (unsigned i=0 ; i != descendants.size() ; ++i) {
        unsigned variable;
        variable= expression_to_bool_var(variables, *descendants[i], used_as);
        or_variables.push_back(variable);
    }
    for (unsigned i=0 ; i != or_variables.size() ; ) {
        unsigned variable= or_variables[i];
        variable= m_operator_connectivity.get_equivalent(variable);
        if ( !m_solver.var_domain(variable).contains(0) ) {
            or_variables.clear();
            or_variables.push_back(m_solver.singleton_var(1));
            return;
        }
        if ( m_solver.var_domain(variable).singleton() ) {
            // clearly it contains 0, which is the only element;
            or_variables[i]= or_variables.back();
            or_variables.pop_back();
        } else {
            or_variables[i]= variable;
            ++i;
        }
    }
    sort(or_variables.begin(), or_variables.end());
    or_variables.erase(unique(or_variables.begin(), or_variables.end()),
                       or_variables.end());
}

void AST_builder::collect_and_vars(vector<unsigned> & and_variables,
                                  const AST & expr,
                                  const vector<unsigned> & variables,
                                  usage_type used_as)
{
    vector<const AST*> descendants;
    insert_descendants(expr, expr.getType(), descendants);
    for (unsigned i=0 ; i != descendants.size() ; ++i) {
        unsigned variable;
        variable= expression_to_bool_var(variables, *descendants[i], used_as);
        and_variables.push_back(variable);
    }
    for (unsigned i=0 ; i != and_variables.size() ; ) {
        unsigned variable= and_variables[i];
        variable= m_operator_connectivity.get_equivalent(variable);
        if ( !m_solver.var_domain(variable).contains(0) ) {
            // Always true: erase element i, and fill element 'i' with another
            // one which will be used in the next iteration.
            and_variables[i]= and_variables.back();
            and_variables.pop_back();
            continue;
        }
        
        if ( m_solver.var_domain(variable).singleton() ) {
            // clearly it contains 0, i.e., variable's domain == {0}
            and_variables.clear();
            and_variables.push_back(m_solver.singleton_var(0));
            return;
        }
        and_variables[i]= variable;
        ++i;
    }
    sort(and_variables.begin(), and_variables.end());
    and_variables.erase(unique(and_variables.begin(), and_variables.end()),
                        and_variables.end());
}

void AST_builder::insert_descendants(const AST & expr, NodeType type,
                                     vector<const AST *>& descendants)
{
    if (expr.getType() == type) {
        for (int i=0 ; i != expr.nbArg(); ++i)
            insert_descendants(expr.getArg(i), type, descendants);
    } else {
        descendants.push_back(&expr);
    }
}

const iset &
AST_builder::intersect_variable(const iset & values, unsigned variable)
{
    const iset & domain= m_solver.var_domain(variable);
    if ( values.contains_all_of(domain) )
        return domain;
    iset out_domain = domain;
    out_domain.intersect(values);
    if (out_domain.empty())
        mark_unsat();
    m_solver.set_var_domain(variable, out_domain);
    return m_solver.var_domain(variable);
}

void AST_builder::connect_expr(const vector<unsigned> & vars,
                               unsigned out_var,
                               const AST & expr,
                               usage_type used_as)
{
    if (m_solver.var_domain(out_var).empty()) {
        mark_unsat();
        return;
    }
    const operator_definition *
        p_operator= operator_definition_registry::find(expr.getType());
    if (p_operator) {
        equate_variables(out_var, operator_to_variable(p_operator, vars, expr,
                                                       out_var, used_as));
        return;
    }
    switch (expr.getType()) {
      case VAR:   equate_variables(out_var, vars.at(expr.getVarId()));   break;
      case CST_BOOL: intersect_variable(iset(expr.getBoolean()),out_var);break;
      case CST_INT:  intersect_variable(iset(expr.getInteger()),out_var);break;
      case F_NEG: connect<neg_propagator>(vars, out_var, expr);          break;
      case F_POW: connect<pow_propagator>(vars, out_var, expr);          break;
      case F_IF:  connect_if(vars, out_var, expr);                       break;
      case F_MIN: connect<min_propagator>(vars, out_var, expr);          break;
      case F_MAX: connect<max_propagator>(vars, out_var, expr);          break;
      case F_EQ:  connect_eq(vars,  out_var, true,  expr, used_as);      break;
      case F_IFF: connect_eq(vars,  out_var, true,  expr, used_as);      break;
      case F_NE:  connect_eq(vars,  out_var, false, expr, used_as);      break;
      case F_XOR: connect_eq(vars,  out_var, false, expr, used_as);      break;
      case F_AND: connect_and(vars, out_var, expr, used_as);             break;
      case F_OR:  connect_or(vars,  out_var, expr, used_as);             break;
      case LIST:  connect_predicate(vars, out_var, expr);                break;
      default:
        throw unsupported("Unsupported expression type");
    }
}

void AST_builder::connect_and(const vector<unsigned> & variables,
                              unsigned result_var, const AST & expr,
                              usage_type used_as)
{
    const iset & out_val= intersect_variable(iset(make_pair(0,1)), result_var);
    bool is_false= (out_val.max() == 0);
    bool is_true= (out_val.min() == 1);
    if (is_true) {
        connect_expr(variables, m_solver.singleton_var(true), expr.getArg(0));
        connect_expr(variables, m_solver.singleton_var(true), expr.getArg(1));
        return;
    }

    usage_type next_used_as = is_false ? IMPLYING : used_as;
    vector<unsigned> clause_vars;
    vector<const iset*> clause_values;
    collect_and_vars(clause_vars, expr, variables, next_used_as);
    unsigned n_and_vars= clause_vars.size();
    iset false_set(0);
    clause_values.reserve(n_and_vars+1);
    clause_values.resize(n_and_vars, &false_set);

    iset true_set(1);
    clause_values.push_back(&true_set);
    clause_vars.push_back(result_var);
    if (used_as != BOUNDED_BELOW)
        // result=0 --> [ value(clause_vars[0])=0 || value(clause_vars[1])=0..)
        insert_clause(&clause_vars[0], &clause_values[0], clause_vars.size());
    if ( !is_false && used_as != BOUNDED_ABOVE) {
        // clause_vars[i]=0 --> result=0
        for (unsigned i = 0 ; i != n_and_vars ; ++i) 
            insert_clause(result_var, false_set, clause_vars[i], true_set);
    }
}

void AST_builder::connect_or(const vector<unsigned> & variables,
                             unsigned result_var, const AST & expr,
                             usage_type used_as)
{
    const iset & out_val= intersect_variable(iset(make_pair(0,1)), result_var);
    bool is_false= (out_val.max() == 0);
    bool is_true= (out_val.min() == 1);
    if (is_false) {
        if (used_as == BOUNDED_ABOVE)
            return;
        connect_expr(variables, m_solver.singleton_var(false), expr.getArg(0));
        connect_expr(variables, m_solver.singleton_var(false), expr.getArg(1));
        return;
    }
    if (is_true && used_as == BOUNDED_BELOW)
        return;
    
    usage_type next_used_as = is_true ? IMPLIED : used_as;
    vector<unsigned> clause_vars;
    vector<const iset*> clause_values;
    collect_or_vars(clause_vars, expr, variables, next_used_as);
    unsigned n_or_vars= clause_vars.size();
    iset true_set(1);
    clause_values.reserve(n_or_vars+1);
    clause_values.resize(n_or_vars, &true_set);

    iset false_set(0);
    clause_values.push_back(&false_set);
    clause_vars.push_back(result_var);
    if (used_as != BOUNDED_ABOVE)
        // result=1 --> [ value(clause_vars[0])=1 || value(clause_vars[1])=1..)
        insert_clause(&clause_vars[0], &clause_values[0], clause_vars.size());
    if ( !is_true && used_as != BOUNDED_BELOW) {
        // clause_vars[i]=1 --> result=1
        for (unsigned i = 0 ; i != n_or_vars ; ++i) 
            insert_clause(result_var, true_set, clause_vars[i], false_set);
    }
}

bool AST_builder::connect_if_simple_eq(const vector<unsigned> & variables,
                                       const AST & simple_expr,
                                       const AST & other_expr)
{
    unsigned variable;
    if (simple_expr.isVar())
        variable= variables.at(simple_expr.getVarId());
    else {
        iset::value_type value;
        if (simple_expr.isInteger())
            value= dynamic_cast<const ASTInteger&>(simple_expr).getVal();
        else if (simple_expr.isBoolean())
            value= dynamic_cast<const ASTBoolean&>(simple_expr).getVal();
        else
            return false;
        variable= m_solver.singleton_var(value);
    }

    variable= m_operator_connectivity.get_equivalent(variable);
    connect_expr(variables, variable, other_expr);
    return true;
}

bool AST_builder::connect_if_simple_ne(const vector<unsigned> & variables,
                                       const AST & simple_expr,
                                       const AST & other_expr)
{
    iset::value_type exclude_val;
    if (simple_expr.isVar()) {
        unsigned variable= variables.at(simple_expr.getVarId());
        variable= m_operator_connectivity.get_equivalent(variable);
        const iset & domain= m_solver.var_domain(variable);
        if ( !domain.singleton() )
            return false;
        exclude_val= domain.min();
    } else {
        if (simple_expr.isInteger())
            exclude_val= dynamic_cast<const ASTInteger&>(simple_expr).getVal();
        else if (simple_expr.isBoolean())
            exclude_val= dynamic_cast<const ASTBoolean&>(simple_expr).getVal();
        else
            return false;
    }
    if (is_boolean_result(other_expr.getType())) {
        if (exclude_val < 0 || exclude_val > 1)
            return true; // such exclude_val is different than any Boolean

        unsigned variable= m_solver.singleton_var(!exclude_val);
        variable= m_operator_connectivity.get_equivalent(variable);
        connect_expr(variables, variable, other_expr);
        return true;
    }
    if (other_expr.isVar()) {
        unsigned variable= variables.at(other_expr.getVarId());
        variable= m_operator_connectivity.get_equivalent(variable);
        iset next_domain= m_solver.var_domain(variable);
        next_domain.erase(exclude_val);
        m_solver.set_var_domain(variable, next_domain);
        return true;
    }
    return false;
}

void AST_builder::connect_eq(const vector<unsigned> & variables,
                             unsigned result_var, bool expect_val,
                             const AST & expr, usage_type used_as)
{
    const iset & out_val= intersect_variable(iset(make_pair(0,1)), result_var);
    bool force_eq= false, force_ne= false;
    if (out_val.singleton()) {
        if (out_val.min()==expect_val)
            force_eq= true;
        else
            force_ne= true;
    }
    // First attempt to optimize for trivial equality, try to avoid temporaries
    if (force_eq) {
        if (connect_if_simple_eq(variables, expr.getArg(0), expr.getArg(1)))
            return;
        if (connect_if_simple_eq(variables, expr.getArg(1), expr.getArg(0)))
            return;
    }
    if (force_ne) {
        if (connect_if_simple_ne(variables, expr.getArg(0), expr.getArg(1)))
            return;
        if (connect_if_simple_ne(variables, expr.getArg(1), expr.getArg(0)))
            return;
    }
    unsigned a= expression_to_variable(variables, expr.getArg(0));
    unsigned b= expression_to_variable(variables, expr.getArg(1),
                                       force_eq ? a : -1U);
    connect_eq_by_vars(result_var, expect_val, force_eq, force_ne, a, b,
                       used_as);

}

void AST_builder::connect_parity(const vector<unsigned> & vars, bool result)
{
    vector<unsigned> free_vars;
    free_vars.reserve(vars.size());
    bool result_with_const = result;
    for (unsigned var : vars) {
        const iset & domain = m_solver.var_domain(var);
        HCSP_ASSERT(0 <= domain.min() && domain.max() <= 1);
        if (domain.singleton())
            result_with_const ^= domain.min();
        else
            free_vars.push_back(var);
    }
    switch (free_vars.size()) {
      case 0:
        if (result_with_const) {
            mark_unsat();
            cout << m_solver.comment_prefix() << " UNSAT because XOR of ";
            for (unsigned var : vars)
                cout << var_printer(var) << ", ";
            cout << "does not give " << (int)result << endl;
        }
        return;
      case 1:
        intersect_variable(iset(result_with_const), free_vars[0]);
        return;
      case 2:
        if (result_with_const == 0) {
            equate_variables(free_vars[0], free_vars[1]);
            return;
        }
      default: { // fall-through
          constraint_connectivity_info & info=
              add_constraint<sum_mod_propagator>(
                  free_vars.size(), 2, result_with_const);
          for (unsigned var : free_vars)
              info << var;
      }
    }
}

void AST_builder::connect_eq_by_vars(unsigned result_var, bool expect_val,
                                     bool force_eq, bool force_ne,
                                     unsigned a, unsigned b, usage_type used_as)
{
    a= m_operator_connectivity.get_equivalent(a);
    b= m_operator_connectivity.get_equivalent(b);

    if (are_eq(a, b))
        equate_variables(m_solver.singleton_var(expect_val), result_var);
    else if (are_ne(a,b))
        equate_variables(m_solver.singleton_var(!expect_val), result_var);
    else if (force_eq)
        equate_variables(a, b);
    else if (force_ne)
        differentiate_variables(a,b);
    else if (used_as == EQUATED) {
        iset bool_domain(make_pair(0,1));
        if (bool_domain.contains_all_of(m_solver.var_domain(a)) &&
            bool_domain.contains_all_of(m_solver.var_domain(b))) {
            connect_parity(vector<unsigned>{result_var, a, b}, expect_val);
            return;
        }
        unsigned eq_result= expect_val ? result_var : var_not_var(result_var);
        add_constraint<is_equal_propagator>(3) << eq_result << a << b;
    } else {
        // IMPLIED  val --> (a == b)   ===>  !val || a==b
        // IMPLIED  val --> (a != b)   ===>  !val || a!=b
        // IMPLYING (a == b) --> val   ===>   val || a!=b
        // IMPLYING (a != b) --> val   ===>   val || a==b
        constraint_connectivity_info * p_info= NULL;
        HCSP_ASSERT(used_as == IMPLIED || used_as == IMPLYING);
        iset literal(used_as == IMPLYING);
        bool equality_atom= (used_as == IMPLIED) == (expect_val);
        if (equality_atom)
            p_info= &add_constraint<literal_or_equal_propagator>(3, literal);
        else
            p_info= &add_constraint<literal_or_neq_propagator>(3, literal);
        *p_info << result_var << a << b;
    }
}

void AST_builder::mark_unequal_variables(unsigned a, unsigned b)
{
    unsigned equiv_a= m_operator_connectivity.get_equivalent(a);
    unsigned equiv_b= m_operator_connectivity.get_equivalent(b);
    if (m_diff_set.size() <= max(equiv_a, equiv_b)) {
        if (m_diff_set.capacity() <= max(equiv_a, equiv_b))
            m_diff_set.reserve(max(equiv_a, equiv_b)*2+2);
            
        m_diff_set.resize(max(equiv_a,equiv_b)+1);
    }
    m_diff_set[equiv_a].insert(equiv_b);
    m_diff_set[equiv_b].insert(equiv_a);
    if (equiv_a == equiv_b)
        mark_unsat();
}

void AST_builder::differentiate_variables(unsigned a, unsigned b)
{
    add_constraint<neq_propagator>(2) << a << b;
    mark_unequal_variables(a,b);
}

void AST_builder::connect_overlap(unsigned var,
                                  const vector<unsigned> & arguments)
{
    if (var == -1U || m_solver.var_domain(var) != iset(0))
        throw unsupported("Only an unconditional !overlap(...) is supported");
    if (arguments.size() != 4)
        throw unsupported("Expecting exactly 4 arguments for "
                          "overlap(start1,len1,start2,len2)");
    const iset & length1_domain= m_solver.var_domain(arguments[1]);
    const iset & length2_domain= m_solver.var_domain(arguments[3]);
    if (!length1_domain.singleton() || !length2_domain.singleton())
        throw unsupported("Expecting singletons for the length in "
                          "overlap(start1,len1,start2,len2)");

    add_constraint<no_overlap_propagator>(2, length1_domain.min(),
                                          length2_domain.min())
        << arguments[0] << arguments[2];
}

void AST_builder::connect_sub_interval(unsigned var,
                                       const vector<unsigned> &arguments)
{
    if (var == -1U || !m_solver.var_domain(var).singleton())
        throw unsupported("Only an unconditional sub_interval(...) "
                          "is supported");
    bool is_positive = m_solver.var_domain(var).min();
    if (arguments.size() < 4 || (arguments.size() > 5 && !is_positive))
        throw unsupported("Expecting 4 or 5 arguments for "
                          "sub_interval(from1,to1,from2,to2[,k]), which means "
                          "[from1, to1-k] subseteq[from2,to2]");
    const iset & a_domain= m_solver.var_domain(arguments[2]);
    const iset & b_domain= m_solver.var_domain(arguments[3]);
    iset k_domain;
    if (arguments.size() > 4)
        k_domain= m_solver.var_domain(arguments[4]);
    if (!a_domain.singleton() || !b_domain.singleton() || k_domain.size()>1)
        throw unsupported("Need constant arguments const_from2,const_to2,k for "
                          "sub_interval(from1,to1,const_from2,const_to2[,k])");
    const iset::value_type a=a_domain.min(), b=b_domain.min();
    const iset::value_type k= k_domain.empty() ? 0 : k_domain.min();
    if (is_positive)
        add_constraint<sub_interval_propagator>(2, a, b)
            << arguments[0] << arguments[1];
    else
        add_constraint<no_sub_interval_propagator>(2, a, b, k)
            << arguments[0] << arguments[1];
}

void AST_builder::connect_predicate(const std::vector<unsigned> & variables,
                                    unsigned var, const AST & expr)
{
    const ASTList * pList= dynamic_cast<const ASTList*>(&expr);
    HCSP_ASSERT(pList);
    const ASTList & list= *pList;
    HCSP_ASSERT(list.nbArg() >= 2);
    HCSP_ASSERT(list[0].isVar());

    vector<unsigned> argument_vars(list.nbArg()-1);
    for (int i=1 ; i < list.nbArg() ; ++i)
        argument_vars[i-1]= expression_to_variable(variables, expr[i]);
    
    // FIXME: create an efficient dispatch table based on predicate names
    string name= list[0].getVarName();
    if (name == "overlap")
        connect_overlap(var, argument_vars);
    else if (name == "sub_interval")
        connect_sub_interval(var, argument_vars);
    else if (name == "all_diff")
        connect_all_diff(argument_vars, true);
    else
        throw unsupported(name+": unsupported predicate");
}

void AST_builder::connect_if(const vector<unsigned> & variables,
                             unsigned result_var,
                             const AST & expr)
{
    const AST & cond     = expr.getArg(0);
    const AST & if_true  = expr.getArg(1);
    const AST & if_false = expr.getArg(2);
    unsigned cond_var=  expression_to_variable(variables, cond);
    const iset & cond_val= intersect_variable(iset(make_pair(0,1)), cond_var);
    
    if (cond_val.min() == 1) { // condition is always true
        connect_expr(variables, result_var, if_true);
    } else if (cond_val.max() == 0) { // condition is always false
        connect_expr(variables, result_var, if_false);
    } else {
        unsigned then_var=  expression_to_variable(variables, if_true);
        unsigned else_var=  expression_to_variable(variables, if_false);
        add_constraint<if_then_else_propagator>(4)
            << result_var << cond_var << then_var << else_var;
    }
}

template <class Propagator>
void AST_builder::connect(const vector<unsigned> & variables,
                          unsigned out_var, const AST & expr)
{
    //Assert an efficiency property. If fails, should call non-template variant
    HCSP_ASSERT(operator_definition_registry::find(expr.getType()) == NULL);

    unsigned arity= expr.nbArg();
    // unfortunately, expression_to_variable may invalidate the connectivity
    // info. This means that first all inner constraints must be created,
    // and only then the outer constraint can be created.
    vector<unsigned> variable_of_arg(arity);
    set<unsigned> processed_vars;
    processed_vars.insert(out_var);
    for (unsigned i=0 ; i != arity ; ++i) {
        variable_of_arg[i]= expression_to_variable(variables, expr.getArg(i)); 
        const iset & domain= m_solver.var_domain(variable_of_arg[i]);
        if ( processed_vars.count(variable_of_arg[i]) > 0 &&
             !domain.singleton() ) {
            ostringstream name_stream;
            name_stream << "_equivalent_" << m_solver.count_variables()
                        << "_of_" << var_printer(variable_of_arg[i]);
            unsigned equivalent= add_auxiliary_var(name_stream.str(), domain);
            equate_variables(equivalent, variable_of_arg[i]);
            variable_of_arg[i]=equivalent;
        } 
        processed_vars.insert(variable_of_arg[i]);
    }

    constraint_connectivity_info&info= add_constraint<Propagator>(arity);
    info << out_var;
    for (unsigned i=0 ; i != arity ; ++i)
        info << variable_of_arg[i];
}

void AST_builder::insert_clause(unsigned var_a, const iset & literal_a,
                                unsigned var_b, const iset & literal_b,
                                unsigned var_c, const iset & literal_c)
{
    enum {LITERALS=3};
    unsigned vars[LITERALS] = {var_a, var_b, var_c};
    const iset *sets[LITERALS] = { &literal_a, &literal_b, &literal_c };

    insert_clause(vars, sets, LITERALS);
}

void AST_builder::insert_clause(unsigned var_a, const iset & literal_a,
                                unsigned var_b, const iset & literal_b)
{
    enum {LITERALS=2};
    unsigned vars[LITERALS] = {var_a, var_b};
    const iset *sets[LITERALS] = { &literal_a, &literal_b };

    insert_clause(vars, sets, LITERALS);
}

void AST_builder::insert_clause(const unsigned vars[], const iset *const sets[],
                                unsigned size) 
{
    unsigned nontrivial_literals= 0;
    unsigned last_nontrivial=-1U;
    for (unsigned i=0 ; i != size ; ++i) {
        if ( sets[i]->contains_all_of(m_solver.var_domain(vars[i])) )
            return; // clause is trivially satisfied.
        if ( sets[i]->contains_any_of(m_solver.var_domain(vars[i]) ) ) {
            ++nontrivial_literals;
            last_nontrivial=i;
        }
    }

    for (unsigned i=0 ; i+1 < size ; ++i)
        for (unsigned j=i+1 ; j < size ; ++j)
            HCSP_ASSERT(vars[i] != vars[j]);

    if (nontrivial_literals <= 1) {
        if (nontrivial_literals == 0)
            mark_unsat();
        else if (nontrivial_literals == 1)
            intersect_variable(*sets[last_nontrivial], vars[last_nontrivial]);
        return;
    }
    clause_propagator * p_clause= new clause_propagator;
    unique_ptr<constraint_propagator> propagator_passed(p_clause);
    clause_propagator::literals_type & literals= p_clause->update_literals();
    literals.reserve(nontrivial_literals);
    
    constraint_connectivity_info &
        info= m_solver.add_constraint(move(propagator_passed));
    info.reserve_arguments(nontrivial_literals);

    for (unsigned i=0 ; i != size ; ++i)
        if ( sets[i]->contains_any_of(m_solver.var_domain(vars[i]) ) ) {
            literals.push_back(*sets[i]);
            info << vars[i];
        }
    HCSP_ASSERT(literals.size() > 1); // smaller cases are handled above
}

unsigned AST_builder::evaluate_if_singletons(
    const operator_definition *p_operator,
    const unsigned *begin_vars, const unsigned *end_vars)
{
    const unsigned arg_count= end_vars-begin_vars;
    iset::value_type *values= static_cast<iset::value_type*>(
        alloca(sizeof(iset::value_type)*arg_count));
    for (unsigned i=0 ; i != arg_count ; ++i) {
        if ( begin_vars[i] == -1U)
            return -1U;
        const iset & domain= m_solver.var_domain(begin_vars[i]);
        if ( !domain.singleton())
            return -1U;
        values[i]= domain.min();
    }
    iset::value_type val= p_operator->eval(&values[0], &values[0]+arg_count);
    return m_solver.singleton_var(val);
}

unsigned AST_builder::known_var_of_arg(
    const operator_definition *p_operator,
    const vector<unsigned> &variables,
    unsigned out_var,
    const AST & argument_expr,
    unsigned argument_index)
{
    if (argument_expr.getType() == CST_INT)
        return m_solver.singleton_var(argument_expr.getInteger());
    if (argument_expr.getType() == VAR)
        return variables.at(argument_expr.getVarId());
    if (out_var == -1U)
        return -1U;

    const iset & out_domain= m_solver.var_domain(out_var);
    pair<iset::value_type, bool>
        reversed= p_operator->reverse_eval(out_domain, argument_index);
    if (reversed.second) 
        return m_solver.singleton_var(reversed.first);
    return -1U;
}

unsigned AST_builder::operator_to_variable(
    const operator_definition *p_operator,
    const vector<unsigned> &variables, const AST & expr, unsigned out_var,
    usage_type used_as)
{
    HCSP_ASSERT(p_operator != NULL);

    const unsigned arg_count= expr.nbArg();
    HCSP_ASSERT(arg_count > 0);
    // Using alloca rather than args[arg_count] for greater portability.
    unsigned *args= static_cast<unsigned*>(alloca(arg_count*sizeof(unsigned)));
    for (unsigned i=0 ; i != arg_count ; ++i)
        args[i]= known_var_of_arg(p_operator, variables, out_var,
                                  expr.getArg(i), i);
    
    for (unsigned i=0 ; i != arg_count ; ++i) {
        eval_type arg_type= p_operator->arg_eval_type(
            m_operator_connectivity, to_eval_type(used_as), i,
            &args[0], &args[0]+arg_count, out_var);
        usage_type arg_usage= EQUATED;
        if (arg_type == LOWER_BOUND_EVAL)
            arg_usage= BOUNDED_BELOW;
        else if (arg_type == UPPER_BOUND_EVAL)
            arg_usage= BOUNDED_ABOVE;
        args[i]= expression_to_variable(variables, expr.getArg(i),
                                        args[i], arg_usage);
    }

    unsigned result= evaluate_if_singletons( p_operator,
                                             &args[0], &args[0]+arg_count);
    if (result == -1U)
        result= p_operator->symbolic_eval(m_operator_connectivity,
                                          to_eval_type(used_as), &args[0],
                                          &args[0]+arg_count, out_var);
    return result;
}

unsigned AST_builder::expression_to_variable(
    const vector<unsigned> & variables, const AST & expr, unsigned default_var,
    usage_type used_as)
{
    const operator_definition *
        p_operator= operator_definition_registry::find(expr.getType());
    if (p_operator)
        return operator_to_variable(p_operator, variables, expr, default_var,
                                    used_as);
    
    switch (expr.getType()) {
      case VAR:
        return m_operator_connectivity.get_equivalent(
            variables.at(expr.getVarId()));
      case CST_BOOL:
        return m_solver.singleton_var(expr.getBoolean());
      case CST_INT:
        return m_solver.singleton_var(expr.getInteger());
      case F_NE:  return ne_to_var(variables, expr, default_var, used_as);
      default:{
          unsigned var= default_var;
          if (var == -1U)
              var= add_auxiliary_var("_internal_subexpression_",
                                     m_solver.count_variables());
          connect_expr(variables, var, expr, used_as); 
          return var;
      } 
    } 
} 
 
unsigned AST_builder::ne_to_var(const vector<unsigned> & variables,
                                const AST & expr, unsigned default_var,
                                usage_type used_as)
{
    if (!expr.getArg(0).isVar() || !expr.getArg(1).isVar()) {
        unsigned count= m_solver.count_variables();
        unsigned var= add_auxiliary_var("_internal_subexpression_", count);
        connect_expr(variables, var, expr, used_as);
        return var;
    } 
    
    unsigned a= variables.at(expr.getArg(0).getVarId());
    unsigned b= variables.at(expr.getArg(1).getVarId());
    if (a > b)
        swap(a,b);
    pair<usage_type, unsigned> & saved= m_is_eq_variables[make_pair(a,b)];
    unsigned & variable= saved.second;
    unsigned orig_variable= variable;
    if ( variable == 0 ) // Map entry was just constructed
        saved.first= used_as;
    else if ( saved.first == EQUATED || saved.first == used_as )
        return variable;
    else
        saved.first= EQUATED;
    variable= default_var;
    if (variable == -1U) {
        ostringstream name_stream;
        name_stream << "_internal_" << var_printer(a) << "_ne_"
                    << var_printer(b);
        variable= add_auxiliary_var(name_stream.str());
    } 
    connect_expr(variables, variable, expr, saved.first);
    if (orig_variable !=0) // Equals zero if map entry is new, no orig_variable
        equate_variables(orig_variable, variable);
    return variable;
}

unsigned AST_builder::add_auxiliary_var(const string & name,
                                        const iset & domain)
{
    return m_solver.add_var(domain, name.c_str());
}

unsigned AST_builder::add_auxiliary_var(
    const char *name_prefix, unsigned name_suffix, const iset & domain)
{
    ostringstream name_stream;
    name_stream << name_prefix << name_suffix;
    return add_auxiliary_var(name_stream.str(), domain);
}

namespace {
class solver_var_values: public VariableValuation
{
  public:
    solver_var_values(const solver & solver) : m_solver(solver) {}
    virtual FunctionValue getVarValue(unsigned int id) const
    {
        return m_solver.var_domain(id).min();
    }
  private:
    const solver & m_solver;
}; // class solver_var_values
} // namespace (anonymous)

iset::value_type AST_builder::evaluate(const AST &expression) const
{
    return expression.value(solver_var_values(m_solver));
}

unsigned AST_builder::var_not_var(unsigned variable, usage_type used_as)
{
    const operator_definition *
        p_operator= operator_definition_registry::find(F_NOT);

    unsigned args[1]= {variable};

    return p_operator->symbolic_eval(m_operator_connectivity,
                                     to_eval_type(used_as),
                                     &args[0], &args[0]+1);
}

void AST_builder::mark_unsat()
{
    m_solver.set_var_domain(0, iset());
}

bool AST_builder::is_marked_unsat() const
{
    return m_solver.count_variables() > 0 && m_solver.var_domain(0).empty();
}

void AST_builder::normalize_diff_set()
{
#ifndef NDEBUG
    unsigned size= m_diff_set.size();
    for (unsigned i=0 ; i != size; ++i) {
        if (m_diff_set[i].empty())
            continue;
        HCSP_ASSERT(m_operator_connectivity.get_equivalent(i) == i);
        set<unsigned>::iterator it;
        for (it= m_diff_set[i].begin() ; it != m_diff_set[i].end() ; ++it) {
            HCSP_ASSERT( *it== m_operator_connectivity.get_equivalent(*it) );
            if (*it == i)
                HCSP_ASSERT(is_marked_unsat());
        }
    }
#endif
}

/**
 * @todo
 *    -# Implement a known (and smarter) algorithm for fast clique detection.
 *       Look for an approximated algorithm that works on edges E with
 *       with O(|E|) or at most O(|E|*sqrt(|E|)).
 *
 *       In the current implementation we get bad complexity and poor results
 *
 *    -# Detect all_different only during conflict analysis when
 *       N different_propagator share M variables. The conflict analysis
 *       algorithm can then add these variables and constraints into
 *       a database edges and vertices relevant to all_different. Over time
 *       this database would grow and would be big enough to support a new or
 *       extend an existing all_different_propagator.
 */
void AST_builder::detect_all_diff()
{
    normalize_diff_set();
    const unsigned MIN_CLIQUE=3;
    unsigned all_diff_constraints=0;
    vector<bool> is_in_all_diff(m_diff_set.size(), false);
    for (unsigned lead=0 ; lead != m_diff_set.size() ; ++lead) {
        if (is_in_all_diff[lead] || m_diff_set[lead].size() < MIN_CLIQUE-1)
            continue;
        vector<unsigned> candidates;
        for ( set<unsigned>::iterator it= m_diff_set[lead].begin() ;
              it != m_diff_set[lead].end() ;
              ++it) {
            if (m_diff_set.at(*it).size() >= MIN_CLIQUE-1)
                candidates.push_back(*it);
        }
        if (candidates.size() < MIN_CLIQUE-1)
            continue;

        vector<unsigned> clique_vars;
        for (unsigned j=0 ; j < candidates.size() ; ++j) {
            if (is_diff_with_all(candidates[j], clique_vars))
                clique_vars.push_back(candidates[j]);
            
            unsigned var= candidates[j];
            bool connected_with_previous= true;
            for (unsigned k=0 ; k != clique_vars.size() ; ++k)
                if (m_diff_set[var].count(clique_vars[k])==0)
                    connected_with_previous= false;
            
            if (connected_with_previous)
                clique_vars.push_back(var);
                
        }
        clique_vars.push_back(lead);
        if (clique_vars.size() >= MIN_CLIQUE) {
            connect_all_diff(clique_vars, all_diff_constraints==0);
            for (unsigned i=0 ; i != clique_vars.size() ; ++i)
                is_in_all_diff[clique_vars[i]]= true;
            ++all_diff_constraints;
        }
    }
}

bool AST_builder::are_ne(unsigned var_a, unsigned var_b) const
{
    unsigned equiv_a= m_operator_connectivity.get_equivalent(var_a);
    unsigned equiv_b= m_operator_connectivity.get_equivalent(var_b);
    const iset & domain_a= m_solver.var_domain(equiv_a);
    const iset & domain_b= m_solver.var_domain(equiv_b);
    if (!domain_a.contains_any_of(domain_b))
        return true;
    if (equiv_a >= m_diff_set.size())
        return false;
    return m_diff_set[equiv_a].count(equiv_b) > 0;
}

bool AST_builder::are_eq(unsigned var_a, unsigned var_b) const
{
    unsigned equiv_a= m_operator_connectivity.get_equivalent(var_a);
    unsigned equiv_b= m_operator_connectivity.get_equivalent(var_b);
    if (equiv_a == equiv_b)
        return true;
    const iset & domain_a= m_solver.var_domain(equiv_a);
    if (!domain_a.singleton())
        return false;
    const iset & domain_b= m_solver.var_domain(equiv_b);
    return domain_a == domain_b;
}

bool AST_builder::is_diff_with_all(unsigned var,
                                   const vector<unsigned> & compared_vars)
{
    if (m_diff_set.at(var).size() < compared_vars.size())
        return false;
    for (unsigned i=0 ; i != compared_vars.size() ; ++i)
        if (m_diff_set.at(var).count(compared_vars[i]) == 0)
            return false;
    return true;
}

void AST_builder::connect_all_diff(const vector<unsigned> & all_diff_vars,
                                   bool is_first)
{
    HCSP_ASSERT(all_diff_vars.size() > 1);

    constraint_connectivity_info &
        info= add_constraint<all_different_propagator>(all_diff_vars.size());
    for (unsigned i=0 ; i != all_diff_vars.size() ; ++i)
        info << all_diff_vars[i];
    if (is_first)
        cout << "\n" << m_solver.comment_prefix();
    cout << "deduced all_different(";
    const char *prefix="";
    for (unsigned i=0 ; i != all_diff_vars.size() ; ++i) {
        cout << prefix << var_printer(all_diff_vars[i]);
        prefix= ",";
    }
    cout << ")\n" << m_solver.comment_prefix();
}

namespace {
template<typename T>
unique_ptr<AST> shallow_copy_type(const AST & expr)
{
    if (dynamic_cast<const T*>(&expr))
        return unique_ptr<AST>(new T(static_cast<const T&>(expr)));
    else
        return nullptr;
}
} // namespace

namespace hcsp {
unique_ptr<AST> deep_copy(const AST &expr)
{
    int nargs = expr.nbArg();
    unique_ptr<AST> ret;
    if ( !(ret = shallow_copy_type<ASTSymb>(expr)) &&
         !(ret = shallow_copy_type<ASTVar>(expr)) &&
         !(ret = shallow_copy_type<ASTInteger>(expr)) &&
         !(ret = shallow_copy_type<ASTBoolean>(expr)) &&
         !(ret = shallow_copy_type<ASTDictKey>(expr))) {
        if (dynamic_cast<const ASTAbstractFunction*>(&expr)) {
            auto &in_func= static_cast<const ASTAbstractFunction&>(expr);
            ret.reset(in_func.makeNode());
            for (int i=0 ; i != nargs ; ++i) {
                ret->setArg(i, deep_copy(in_func.getArg(i)).release());
            }
        }
        if (dynamic_cast<const ASTList*>(&expr)) {
            const ASTList & list = static_cast<const ASTList&>(expr);
            unique_ptr<ASTList> list_ret;
            for (int i=0 ; i != nargs ; ++i) {
                list_ret->push_back(deep_copy(list[i]).release());
            }
            ret= move(list_ret);
        }
        // ASTDict is not supported;
    }
    return move(ret);
}
}
