// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <algorithm>
#include <numeric>
#include <map>
#include <iset.hpp>
#include <small_iset.hpp>
#include <big_iset.hpp>
#include <variable_print_strategy.hpp>
#include <neq_propagators.hpp>

using namespace std;
using namespace hcsp;

/**
 * @brief A simple representation of a directed graph. The graph has
 * efficient methods for traversing all edges going out of a vertex.
 */
class all_different_propagator::graph
{
  private:
    typedef vector<unsigned> out_edges_type; //!< Vertex's outgoing edges.
  public:
    //! \brief An iterator that can go over all out edges of a vertex.
    typedef out_edges_type::const_iterator out_edge_iterator;

    /**
     * @brief Set the number of vertexes that this graph supports.
     *
     * If the new number of vertexes is increased then the old edges
     * and vertexes are kept intact.
     *
     * If the new number of vertexes is decreased then some edges and
     * vertexes may be lost. Decrease number of edges only after erasing
     * all edges.
     */
    void resize(unsigned nodes) { m_edges.resize(nodes); }
    
    /**
     * @brief Reserve space for the given vertex. This method's purpose is
     * only to improve performance.
     *
     * Note that edge iterators may be invalidated after this method is called.
     *
     * @param number_of_edges  The future number of out-edges to prepare for.
     * @param node             The vertex from which the edges go.
     */
    void reserve_edges_for_node(unsigned number_of_edges, unsigned node)
    { m_edges[node].reserve(number_of_edges); }

    //! \brief Add the directed edge (from, to).
    void add_edge(unsigned from, unsigned to)
    { m_edges[from].push_back(to); }

    //! \brief Remove all edges from all vertexes.
    void clear_edges()
    {
        for (unsigned i= 0 ; i != m_edges.size() ; ++i)
            m_edges[i].clear();
    }
    
    //! \brief Get the number of vertexes in the graph.
    unsigned node_count() const { return m_edges.size(); }

    //! \brief Get the number of edges going out from the given vertex.
    unsigned count_out_edges(unsigned from) const
    { return m_edges[from].size(); }

    //! \brief Get the begin() out-edge iterator going out from the given
    //!        vertex.
    out_edge_iterator out_edges_begin(unsigned node) const
    { return m_edges[node].begin(); }
    
    //! \brief Get the end() out-edge iterator going out from the given vertex.
    out_edge_iterator out_edges_end(unsigned node) const
    { return m_edges[node].end(); }
    
  private:
    
    vector<out_edges_type> m_edges; //!< One list of out-edges per vertex.
};

/**
 * A wrapper around vector<iset>. This wrapper lets the caller access
 * the vector by the means of the domain() and the arity() methods.
 * All other calls are either meaningless or forbidden.
 */
class all_different_propagator::reference_info final :
    public constraint_runtime_info, private var_watchers_base
{
  public:
    reference_info(const vector<iset> & domains) : m_domains(domains) {}
    
    virtual const iset & domain(unsigned param_index) const
    { return m_domains[param_index]; }
    
    virtual unsigned arity() const { return m_domains.size(); }
    

    //! These methods are meaningless for a reference_info.
    //! They are here only due to interface restriction.
    //! The user of reference_info uses only domain() and arity().
    //@{
    virtual bool reduced_between_activations(unsigned) const { return false; }
    virtual queue_type originating_queue() const { return VARIABLE_QUEUE; }
    
    virtual void put_in_queue(queue_type)
    { throw domain_error("Can't put_in_queue for reference_info"); }

    virtual void set_domain(unsigned, iset)
    { throw domain_error("Can't domain_update for reference_info"); }
    
    
    virtual void deallocate_watch(unsigned, watch_type, watch_handle_t) override
    { throw domain_error("Can't deallocate_watch reference_info"); }
    
    virtual watch_handle_t allocate_watch(unsigned, watch_type) override
    { throw domain_error("Can't allocate_watch reference_info"); }
    
    virtual void deallocate_value_watch(unsigned, iset::value_type, 
                                        watch_handle_t) override
    { throw domain_error("Can't deallocate_watch reference_info"); }
        
    virtual watch_handle_t allocate_value_watch(unsigned, iset::value_type) 
        override
    { throw domain_error("Can't allocate_watch reference_info"); }
    //@}
    
  private:
    const vector<iset> & m_domains;
};

/**
 * @brief The main implementation class for the all_different_propagator.  It
 *        contains all the code and data structure needed to implement the
 *        all-different propagator. This includes graph data structure and
 *        maximal matching algorithm.
 */
class all_different_propagator::impl
{
  public:
    /**
     * @brief Construct the bipartite-graph that represents the inputs,
     *        and value/node mappers.
     */
    impl(const constraint_runtime_info & info) { construct(info); }
    
    /**
     * @brief Construct the bipartite-graph that represents the inputs,
     *        and value/node mappers. Overloaded convenience constructor.
     */
    impl(const vector<iset> & domains);

    /**
     * @brief Propagate the bipartite graph onto the \p next_domains.
     *
     * See the long explanation on how \ref all_different_propagation works.
     *
     * @return false if no feasible solution exists.
     */
    bool propagate(vector<iset> & next_domains) const;

    /**
     * @brief Explain the propagation which the bipartite graph would make.
     *
     * See the long explanation on how \ref all_different_explanation works.
     *
     * @param parameter    The parameter whose propagation we have to explain.
     * @param removed_values_to_explain
     *                     The set of values which the propagation removed and
     *                     are required to be explained.
     *                     
     * @param explanation  The resulting explanation.
     */
    void explain_propagation(unsigned parameter,
                             const iset & removed_values_to_explain,
                             vector<iset> & explanation) const;

  private:
    class bfs_node_origin;
    class tarjan_node_info;
    typedef graph::out_edge_iterator out_edge_iterator;
    
    //! The type of the mapping of values to their value node counterparts.
    typedef map<iset::value_type, unsigned> value_to_node_type;

    //! See next_nodes() for more information
    
    enum completeness_type { ONLY_FIRST, ALL };
    
    //! Which edges to traverse next
    enum edges_used_type {
        UNMATCHING_EDGES,       //!< Traverse only edges which are not matching
        MATCHING_EDGES          //!< Traverse only matching edges
    };

    /**
     * @brief Construct this object.
     *
     * This is the actual code of the constructor.
     * Constructors are implemented by calling this method.
     */
    void construct(const constraint_runtime_info & info);
    
    /**
     * @brief Construct the translation table from domain values to graph
     *        node IDs.
     *
     * @return The total number of nodes in the graph
     */
    unsigned construct_value_to_node(const constraint_runtime_info & info);

    /**
     * @brief Get the list of value and variable nodes explaining the
     *        propagation towards a variable.
     *
     * @param[in]  directed_edges
     *             The directed graph with variable->value are matching
     *             edges from the original graph, and value->variable are
     *             non-matching edges of the original graph.
     * @param[in]  is_reachable_from_free
     *             For each vertex hold 0 iff it is not reachable from a
     *             free node (i.e. unmatched either variable or value node).
     * @param[in]  variable
     *             The variable which reduction this method has to explain.
     * @param[out] explaining_variable_nodes
     *             The variables that explain the reduction.
     * @param[out] explaining_value_nodes
     *             The value vertexes whose presence in explaining_value_nodes
     *             explain the reduction.
     */
    void get_explaining_nodes(
        const graph & directed_edges,
        const vector<unsigned char> & is_reachable_from_free,
        unsigned variable,
        vector<unsigned> & explaining_variable_nodes,
        vector<unsigned> & explaining_value_nodes) const;

    /**
     * @brief Generate the final explanation vector from a list of explaining
     *        nodes.
     */
    void explanation_from_explaining_nodes(
        const vector<unsigned> & explaining_value_nodes,
        const vector<unsigned> & explaining_variable_nodes,
        vector<iset> & explanation) const;

    /**
     * @brief Convert a set of values to a set of node indexes.
     *
     * @param[out] value_nodes  The resulting set of value nodes
     * @param[in]  values       The set of values to convert
     */
    void convert_values_to_nodes(vector<unsigned> & value_nodes,
                                 const iset & values) const;
    /**
     * @brief Find the maximal matching in the current bipartite graph.
     *
     * Use the Hopcroft Karp algorithm to find the maximal matching:
     *  http://en.wikipedia.org/wiki/Hopcroft%E2%80%93Karp_algorithm
     */
    int maximal_matching(vector<unsigned> & matching) const;
    
    /**
     * @brief Remove all element vertexes which are matched.
     *
     * @param[in,out] elements
     *                The element vertexes which we have to keep or remove.
     *                A vertex is removed from this list if and only if
     *                it has a matching vertex.
     *
     *                Note that the order of this vector is not preserved.
     * @param[in]     matching
     *                For each vertex either its matching vertex or -1U
     *                if the vertex is free.
     */
    void remove_if_matched(vector<unsigned> & elements,
                           const vector<unsigned> & matching) const;

    /**
      * @brief Check if the given free_value_nodes was the end node of an
      * augmenting path of bfs.
      * 
      * An augmenting path is a path that:
      *  -# Starts with a free value node (not connected to a matching edge).
      *  -# Alternates through not-matching and matching edges
      *  -# Ends with a free variable node.
      * If all edges of the path are toggled between matching and not-matching
      * number of matches will be increased by one.
      */
    bool is_augmenting_path( unsigned free_value_node,
                             const vector<bfs_node_origin> & bfs_nodes_info,
                             const vector<unsigned> & matching_table) const;
    
    /**
     * @brief Toggle matching/not-matching edges for an augmenting path
     * ending with free_value_node.
     *
     * If bfs defines an augmenting path that ends with 'free_value_node'
     *  (definition at is_augmenting_path() ) then toggle all edges on the
     * path between matching/not-matching. In this case, it is guaranteed to
     * increase matching by at least one edge.
     */
    void toggle_matching(
        unsigned free_value_node,
        const vector<bfs_node_origin> & bfs_nodes_info,
        vector<unsigned> & matching_table) const;

    /**
     * @brief Run BFS on the bipartite graph, from given variable roots, until
     *        a free value is found.
     *
     * The BFS starts with first_bfs_variables which are variable vertexes.
     * When it goes from a variable to a value vertex it uses non-matching
     * edges. When it goes back from a value to a variable vertex it uses
     * only matching edges.
     *
     * The BFS stops at a layer where a free value vertex is found, it means
     * that there is an augmenting path:
     *   - From one of the variable roots (which are free)
     *   - Through an alternating path of matching and non-matching edges.
     *   - To a free value.
     *
     * @param[in]  first_bfs_variables
     *             The first layer of BFS (the roots).
     * @param[in]  matching_table
     *             For each vertex V: matching_table[V] is the vertex
     *             that matches V, or -1U if V is free.
     * @param[out] bfs_nodes_info
     *             For each vertex its visitation history through BFS.
     * @param[out] free_value_nodes
     *             List of free value vertexes on the last layer of BFS.
     *             If no free value vertexes were found this list will be empty
     * @pre
     *    - bfs_nodes_info.size() == number of nodes
     *    - for each i: !bfs_nodes_info[i].visited()
     *    - for each variable in first_bfs_variables: variable is free
     */
    void bfs_until_free_values(const vector<unsigned> & first_bfs_variables,
                               const vector<unsigned> & matching_table,
                               vector<bfs_node_origin> & bfs_nodes_info,
                               vector<unsigned> & free_value_nodes) const;

    /**
     * @brief Make one step of BFS one layer to the next.
     *
     * At each step only part of the edges are considered according to the
     * the parameters passed.
     *
     * @param[in]  nodes
     *                The vertexes of the previous layer.
     * @param[out] next_nodes
     *                The vertexes of the next layer.
     * @param[in]  matching_table
     *                For each vertex matching_table[vertex] is the matching
     *                vertex or -1U if the vertex is free.
     * @param[in,out] bfs_nodes_info
     *                Information on BFS visitation history for each node.
     * @param[in]  edges_used
     *                Which edges to traverse (matching or others).
     *
     * @param[in]  completeness
     *                This is to support a misconceived hack which makes the
     *                algorithm deviate from the original Hopcroft Karp
     *                algorithm.
     *                - ALL: The algorithm behaves as expected.
     *                - ONLY_FIRST: The algorithm stops the moment next_nodes
     *                  is not empty.
     */
    void next_nodes(
        const vector<unsigned> & nodes, vector<unsigned> & next_nodes,
        const vector<unsigned> & matching_table,
        vector<bfs_node_origin> & bfs_nodes_info, edges_used_type edges_used,
        completeness_type completeness=ALL) const;


    /**
     * @brief Make a directed graph out of the original m_edges graph.
     *        Keep edges if any of the following conditions are met:
     *  - e=(variable->value)  and edge e is matched.
     *  - e=(value->variable)  and edge e is not-matched.
     */
    void get_directed_edges(const vector<unsigned> & matching_table,
                            graph & directed_edges) const;

    /**
     * @brief Use Tarjan's algorithm to find Strongly Connected Components
     *        (SCCs).
     * See http://en.wikipedia.org/wiki/Tarjan%27s_strongly_connected_components_algorithm
     *
     * The algorithm calls tarjan()  recursively. For that purpose several
     * support structures are passed to this method. This description 
     * is relevant both for external calls and the inner recursive calls.
     *
     * The algorithm is basically a DFS that records information as it goes.
     *
     * @param[in,out]  nodes_visit_info
     *                 For each vertex contains its visit information.
     *                 At the end, each vertex V will contain a vertex U
     *                 that represent the SCC that contains both U and V.
     * @param[in]      current_node
     *                 The current DFS node.
     * @param[in,out]  visit_count
     *                 The number of unique new vertexes visited so far.
     *
     *                 Initially, the caller should pass 0.
     * @param[in,out]  visited_unassociated_node_stack
     *                 A stack of vertexes, in the order they were first
     *                 discovered by DFS. Vertexes are removed from the
     *                 stack the moment they get associated with a SCC.
     *
     *                 Initially and finally it should be empty.
     */
    void tarjan(vector<tarjan_node_info> & nodes_visit_info,
                unsigned current_node, unsigned & visit_count,
                vector<unsigned> & visited_unassociated_node_stack,
                const graph & directed_edges) const;
    
    /**
     * @brief Calculate Strongly Connected Components (SCCs).
     *
     * @param[in]     directed_edges
     *                A set of directed edges of the graph
     * @param[out]    scc_associations
     *                scc_associations[node] gives the ID of the SCC to which
     *                this node is associated. The ID is an index of one node
     *                which belongs to this SCC.
     */
    void get_scc_associations(
        const graph & directed_edges,
        vector<unsigned> &scc_associations) const;

    /**
     * @brief Extract edges that connect nodes within each SCC.
     *
     * @param[in]     directed_edges
     *                A set of directed edges of the graph
     * @param[in,out] final_edges
     *                The set of edges (undirected) that should be present
     *                at the end of propagation
     *                No effort is made to avoid edge duplication.
     */
    void get_scc_edges(
        const graph & directed_edges,
        vector<pair<unsigned, unsigned> > & final_edges) const;

    /**
     * @brief Find edges reachable from \p node.
     *
     * Find all edges that can be reached from \p node when going
     * through the directed graph. This algorithm performs DFS starting
     * with \p node.
     *
     * @param[in]  node               The node to perform DFS from.
     * @param[in]  directed_edges     The directed graph to work on.
     * @param[in,out] is_reached_node Nodes already visited by the algorithm.
     * @param[in,out] final_edges     The set of edges to be updated.
     *                                Uniqueness is not guaranteed.
     */
    void add_reachable_edges(
        unsigned node,
        const graph & directed_edges,
        vector<unsigned char> & is_reached_node,
        vector<pair<unsigned, unsigned> > & final_edges) const;
    /**
     * @brief Find edges reachable from \p node.
     *
     * Unlike add_reachable_edges() only updates \p is_reached_node
     * to mark which nodes are reachable from \p node.
     */
    void mark_all_reachable(
        unsigned first_node, const graph & directed_edges,
        vector<unsigned char> & is_reached_node) const;

    /**
     * @brief Create a graph with its in-out edge direction reversed.
     */
    void reverse_edges(const graph & directed_edges,
                       graph & reversed_edges) const;
    
    unsigned                 m_number_of_variables;
    graph                    m_graph;
    vector<iset::value_type> m_node_to_value;
    value_to_node_type       m_value_to_node;
}; // class all_different_propagator::impl


/**
 * @brief Information on the origin that introduced this node to BFS.
 *
 * This object contains information whether the node was already visited by
 * BFS and, if so, why.
 */
class all_different_propagator::impl::bfs_node_origin
{
  public:
    bfs_node_origin() : m_origin(UNVISITED) {}
    //!\brief Was this node visited at all?
    bool visited() const { return m_origin != UNVISITED; };
    
    //!\brief Was this node visited by an external request?
    bool external() const { return m_origin == EXTERNAL; }

    //!\brief Mark this node as visited by an external request.
    void mark_external() { m_origin= EXTERNAL; }

    //!\brief Mark this node as visited when following an edge pointing from
    //!       the given node.
    void mark_origin(unsigned node) { m_origin= node; }

    //!\brief Get the node from which we got to this one.
    unsigned origin() const
    { HCSP_ASSERT(visited() &&!external()); return m_origin; }
  private:
    static const unsigned EXTERNAL = -1U;
    static const unsigned UNVISITED = -2U;
    unsigned m_origin;
}; // class all_different_propagator::impl::bfs_node_origin

/**
 * @brief Information that Tarjan's algorithm maintains on this vertex.
 *
 * See the attributes for more information.
 */
class all_different_propagator::impl::tarjan_node_info
{
  public:
    tarjan_node_info() : m_visited(false), m_on_stack(false) {}
    //! \brief Was this vertex visited by the algorithm?
    bool visited() const { return m_visited; }
    //! \brief Set the visit oder of this vertex.
    void set_visit_order(unsigned order)
    {
        m_visited= true;
        m_earliest_looping_visit= m_visit_order= order;
    }
    //! \brief Choose the earliest visit looping order of the two nodes.
    void set_earliest_looping_visit(const tarjan_node_info & other)
    {
        m_earliest_looping_visit= min(m_earliest_looping_visit,
                                      other.m_earliest_looping_visit);
    }
    //! \brief Is the visit order of this vertex also the earliest of any
    //!        other node participating in loops with this vertex?
    bool is_earliest_looping_visit() const
    { return m_visit_order == m_earliest_looping_visit; }

    //! \brief Mark this vertex is either on or off the stack.
    void mark_on_stack(bool flag) { m_on_stack= flag; }
    //! \brief Is this vertex on stack?
    bool on_stack() const { return m_on_stack; }
    //! \brief Return the vertex that represents this node's SCC.
    unsigned scc_association() const { return m_scc_association; }
    //! \brief Set the vertex ID that represents this node's SCC.
    void set_scc_association(unsigned id) { m_scc_association= id; }
    
  private:
    //! The DFS order at which this vertex was first visited.
    unsigned m_visit_order;

    /**
     * The earliest visit order of a vertex which participates in one loop
     * with this vertex. This value is updated only from time to tome.
     */
    unsigned m_earliest_looping_visit;

    //! The earliest vertex that belongs to the same SCC as this one.
    unsigned m_scc_association;

    //! Was this vertex visited by Tarjan's DFS.
    bool m_visited;
    
    /**
     * Is this node on Tarjan's stack? Whenever all elements of the SCC
     * are detected, then only then all its elements are popped of the stack.
     */
    bool m_on_stack;
}; // class all_different_propagator::impl::tarjan_node_info

//! Holder for the explanation rule used by all_different_propagator.
class all_different_propagator::all_diff_rule : public explanation_rule
{
  public:
    all_diff_rule(const vector<iset> & literals)
        : m_parameters(literals.size())
    {
        bool has_first_in_clique = false;
        unsigned first_in_clique= -1U;
        for (unsigned i=0 ; i != literals.size() && !has_first_in_clique; ++i) {
            if ( !literals[i].empty() ) {
                first_in_clique= i;
                has_first_in_clique= true;
            }
        }
        HCSP_ASSERT(has_first_in_clique);
        m_clique_values= literals[first_in_clique];
        m_clique_values.complement();
        HCSP_ASSERT(m_clique_values.size() <= literals.size());
        m_clique_parameters.reserve(m_clique_values.size()+1);
        
        for (unsigned i=0 ; i != literals.size() ; ++i) {
            if ( !literals[i].empty() ) {
                m_clique_parameters.push_back(i);
                HCSP_ASSERT(literals[i] == literals[first_in_clique]);
            }
        }
        // An optimal original explanation should come from something like:
        // (a in {x,y,z} && b in {x,y,z} && c in {x,y,z} ) ==> d not in {x,y,z}
        // Which is equivalent to 
        // (a !in {x,y,z} || b !in {x,y,z} || c !in {x,y,z} || d !in {x,y,z}).
        // Creating N+1 literals for N clique values.
        HCSP_ASSERT(m_clique_parameters.size() == m_clique_values.size()+1);
    }
    virtual explanation_rule * clone() const
    { return new all_diff_rule(*this); }
    
    virtual void get_explanation(std::vector<iset> & explanation) const
    {
        iset literal= m_clique_values;
        literal.complement();
        explanation.clear();
        explanation.resize(m_parameters);
        for (unsigned i=0 ; i != m_clique_parameters.size() ; ++i) {
            unsigned parameter= m_clique_parameters[i];
            explanation.at(parameter)= literal;
        }
    }
    virtual void print(
        std::ostream & out,
        const variable_print_strategy & strategy,
        const std::vector<unsigned> & constraint_variables) const
    {
        out << "ALL_DIFF({" << m_clique_values << "}";
        for (unsigned i=0 ; i != m_clique_parameters.size(); ++i) {
            unsigned parameter= m_clique_parameters[i];
            out << ", " << strategy.printer(constraint_variables.at(parameter));
        }
        out << ")";
    }
  private:
    unsigned m_parameters;
    iset m_clique_values;
    vector<unsigned> m_clique_parameters;
}; // class all_different_propagator::all_diff_rule

all_different_propagator::impl::impl(const vector<iset> & domains)
{
    HCSP_PROFILE_CURRENT;
    construct(reference_info(domains));
}

void all_different_propagator::impl::construct(
    const constraint_runtime_info &info)
{
    HCSP_PROFILE_CURRENT;
    m_number_of_variables= info.arity();
    unsigned node_count= construct_value_to_node(info);
    
    m_graph.resize(node_count);
    for (unsigned i=0 ; i != m_number_of_variables ; ++i)
        m_graph.reserve_edges_for_node(info.domain(i).size(), i);
    for (unsigned i=m_number_of_variables ; i!=m_graph.node_count() ; ++i)
        m_graph.reserve_edges_for_node(m_number_of_variables, i);
    
    // Connect edges between the nodes
    for (unsigned variable=0 ; variable != m_number_of_variables ; ++variable){
        const iset & domain= info.domain(variable);
        m_graph.reserve_edges_for_node(domain.size(), variable);
        for (iset::iterator it=domain.begin() ; it != domain.end() ; ++it) {
            unsigned value_node= m_value_to_node.find(*it)->second;
            m_graph.add_edge(variable, value_node);
            
            HCSP_ASSERT(variable != value_node);
            m_graph.reserve_edges_for_node(m_number_of_variables, value_node);
            m_graph.add_edge(value_node, variable);
        }
    }
}

/**
 * @page all_different_propagation All_different propagation algorithm
 *
 * This is an implementation of Regin's algorithm described in the paper
 * <em>
 * "A filtering algorithm for constraints of difference in CSPs"
 * In Proceedings of the Twelfth National Conference on Artificial
 * Intelligence (AAAI)
 * </em>
 *
 * @section all_different_maximal_matching Translation to the problem of maximal matching
 *
 * The question of whether there is a feasible assignment to variables
 * of the all_different constraint is easily translated to a maximal matching
 * problem. We simply generate a bipartite graph with two types of vertexes:
 *
 *  - On the left we have a vertex for each variable.
 *  - On the right we have a vertex for each value.
 *
 * For every variable that contains a value in its domain we mark an edge
 * between the variable and the value vertexes.
 *
 * If there is a feasible assignment then it can be represented by marking
 * each (variable,value) edge representing that feasible assignment. 
 * For such a feasible assignment all variable vertexes will have a marked
 * edge. In terms of matching, each variable vertex will be matched in a
 * feasible assignment. Such a matching is called a <em>perfect matching</em>.
 *
 * This means that a perfect matching in the bipartite graph exists if and
 * only if there is a feasible assignment to the corresponding all_different
 * constraint. There are efficient algorithms for finding a maximal 
 * matching in a bipartite graph, and such an algorithm will be used as
 * a basis for the propagation algorithm.
 *
 * @section all_different_extending_matching Extending a maximal matching to a propagation algorithm
 *
 * Given a maximal matching, it is possible to efficiently find out if an
 * edge in the bipartite graph participates in any possible maximal matching.
 * This capability is used to find out which values in a domain have support,
 * and which should be pruned away.
 *
 * First the algorithm defines an alternating path (or cycle) where every 
 * two adjacent edges one of them is a matching edge and the other
 * is not-matching.
 *
 * There are several properties that the algorithm uses:
 *
 *   - In an alternating cycle it is possible to toggle all matching and
 *     not-matching edges and still have a valid maximal matching. This is
 *     because a vertex that has a matching edge has exactly one matching
 *     edge (by the definition of matching). By toggling all the edges we are
 *     still guaranteed to have no more than one matching edge per vertex.
 *     We are also guaranteed to have the same number of matching edges.
 *
 *   - Consider a maximal alternating path which starts with a free vertex
 *     (without matching edges). All edges on this path can be toggled between
 *     matching and not-matching, just like for the cycle.
 *
 *   - An edge may not be part of a maximal matching if is neither part of
 *     a maximal matching nor part of any of the above two cases.
 *
 * These properties lead to the propagation algorithm.
 *
 * @section all_different_full_propagation The complete propagation algorithm.
 *
 *   -# Matching:= Find_Maximal_Matching.
 *   -# If not all variables are matched then <b>return CONFLICT</b>
 *   -# Mark all edges of the bipartite graph as directional such that
 *        - Matching edges point from variable to value nodes.
 *        - Not-matching edges point the other way (value to variable nodes)
 *   -# Supported := Matching
 *   -# Supported := Supported UNION (edges on directed paths which start with
 *                                   free vertexes)
 *   -# Supported := Supported UNION (edges on SCCs of the directed graph).
 *   -# Prune variable and value pairs not in \e Supported, from the
 *      corresponding domain.
 */
bool all_different_propagator::impl::propagate(vector<iset> & next_domains)
    const
{
    HCSP_PROFILE_CURRENT;
    
    vector<unsigned> matching;
    int failed_variable= maximal_matching(matching);
    if ( failed_variable >= 0 )
        return false;
    
    static graph directed_edges;
    get_directed_edges(matching, directed_edges);

    static vector<pair<unsigned, unsigned> > final_edges;
    final_edges.clear();
    final_edges.reserve(m_number_of_variables*m_number_of_variables);

    for (unsigned i=0 ; i != m_number_of_variables ; ++i)
        if (matching[i] != -1U)
            final_edges.push_back(make_pair(i, matching[i]));
        
    vector<unsigned char> was_node_visited;
    // Go through all free values and add all reachable edges from them.
    for (unsigned i=m_number_of_variables; i!=directed_edges.node_count(); ++i)
        if (matching[i] == -1U)
            add_reachable_edges(i, directed_edges, was_node_visited,
                                final_edges);
    
    get_scc_edges(directed_edges, final_edges);
    
    next_domains.clear();
    next_domains.resize(m_number_of_variables);
    for (unsigned i= 0; i!=final_edges.size() ; ++i) {
        unsigned variable=   min(final_edges[i].first, final_edges[i].second);
        unsigned value_node= max(final_edges[i].first, final_edges[i].second);
        next_domains[variable].insert(m_node_to_value[value_node]);
    }
    return true;
}

/**
 *@page all_different_explanation All_different explanation algorithm.
 *
 *  The method explanation of a single propagation stems from the way
 *  propagation \ref all_different_propagation "is performed".
 *
 *  The algorithm looks for a set of variables and values such that if
 *  any of these values were in the domain of all these variables then this
 *  propagation would not happen. This does not mean that no propagation would
 *  happen, but it means that exactly this one would be ruled out.
 *
 *@section all_different_explanation_differences The explanation algorithm is significantly different than propagation.
 *
 *  The \ref all_different_full_propagation "propagation algorithm" defines
 *  what stays in the domain after propagation, but here we want to explain the
 *  opposite. We want to explain why something was \e removed from a domain.
 *
 *  Another difference is in handling unsatisfiable states. The propagation
 *  algorithm simply exits once it finds that no perfect matching exists.
 *  Yet, the explanation algorithm has to continue and to show why no perfect
 *  matching can be found.
 *
 *  It is possible to transform a conflicting case into a regular propagation
 *  case. This is not really needed for the implementation of the explanation
 *  algorithm, but it can be used for a formal proof for its correctness.
 *
 *  The transformation is as follows: For each variable \f$ V_i \f$ that is not
 *  matched in the maximal matching, introduce a new value \f$ X_i \f$. Add
 *  this value to the original domain of \f$ V_i \f$. At this point the problem
 *  becomes satisfiable with the addition of the matching of \f$(V_i,X_i)\f$.
 *  For one of the unmatched variables, generate an explanation and remove all
 *  the \f$ X_k \f$ values from the explanation.
 *
 *
 *@section all_different_explanation_idea The idea behind the algorithm.
 *
 *  We want to find a minimal set of variables \f$ V \f$ and values \f$ X \f$
 *  such that all values \f$ x \in X \f$ meet the following condition.  If we
 *  insert \f$ x \f$ to the domains of all \f$ V \f$ variables then the
 *  reduction would not happen.
 *
 *  If we want to explain the reduction of some variable, say \f$ u \f$, then
 *  the explanation generated can be viewed in a different light. The set of
 *  variables that explain the propagation have an interesting property.
 *  The variables of \f$ U=V \setminus \{u\} \f$ take up all the values
 *  \f$ X \f$ to themselves. In other words, in order to have a solution then
 *  only \e U variables may be assigned with \e X values.
 *
 *  @note I believe that there was a paper with some aspects of it.
 *        It said something along the lines that an all_different propagator
 *        is not consistent if and only if there is such a set.
 *
 *  If the propagated variable \f$ u \f$ is not a member of this set \e U, then
 *  all of the \f$ X \f$ values will have to be removed from
 *  its domain since they are taken up by U. Assuming that
 *  \f$ V=\{u,v_1,v_2,\ldots,v_n\} \f$ this will create the following
 *  explanation: 
 *  \f[(   v_1 \not\in X \vee \cdots \vee v_n \not\in X \vee u \not\in X )\f]
 *
 *  To find the sets \e U and \e X the algorithm looks at all the edges going
 *  out of the explained variable \e u. Then it figures out which of the
 *  pointed values are taken up and by what set of variables.
 *
 *
 *@section all_different_explanation_reachable Vertexes reachable from a free vertex
 *
 *  First the algorithm uses the fact that any value vertex reachable from
 *  a free vertex is taken up by no variable. All such value vertexes
 *  are ignored, and the search for an explanation does not go through them.
 *  This aspect is explained later in this section.
 *
 *  @par Example
 *    The following example shows why no value vertex of an SCC is taken up
 *    if it is reachable from a free vertex. The SCC is {x,y,1,2}. \n
 *    @dot
 *     digraph reachable_free_vertexes {
 *        rankdir=BT;
 *        size="5.5,2";
 *        subgraph cluster0 { z y x   }
 *        subgraph cluster1 { 4 3 2 1 }
 *        x -> 1;
 *        1 -> y [style=dashed];
 *        y -> 2 [color=red];
 *        2 -> x [style=dashed];
 *        3 -> y [color=red,style=dashed];
 *        z -> 3 [color=red];
 *        4 -> z [color=red,style=dashed]
 *        4 [label="4\n free"];
 *     }
 *    @enddot
 *    \n
 *    Starting with this diagram, solid edges represent a matching between
 *    two vertexes. Dashed edges represent edges in the bipartite graph which
 *    are not a matching between the vertexes. \n
 *    \n
 *    Note that the value vertex 2 is reachable from the free vertex 4.
 *    After taking the path from 4 to 2 and switching edge direction, we get
 *    perfect matching without the participation of 2.
 *    This means that 2 is not taken up by any set of variables.
 *    @dot
 *     digraph reachable_free_vertexes_toggled {
 *        rankdir=BT;
 *        size="5.5,2";
 *        subgraph cluster0 { z y x   }
 *        subgraph cluster1 { 4 3 2 1 }
 *        x -> 1;
 *        1 -> y [style=dashed];
 *        2 -> x [style=dashed];
 *        2 -> y [color=red,style=dashed];
 *        y -> 3 [color=red];
 *        3 -> z [color=red,style=dashed];
 *        z -> 4 [color=red];
 *        2 [label="2\n free"];
 *     }
 *    @enddot
 *
 *  The algorithm walks from the explained variable in opposite
 *  direction of the edges. A variable vertex reachable from a free vertex
 *  will not be passed through the algorithm (unless this is the explained
 *  vertex). The reason for that is the fact that any variable vertex
 *  has only one outgoing edge, which by definition will point to a value
 *  reachable from a free vertex. The algorithm must go through that value
 *  vertex in order to reach the variable vertex.
 *
 *  This feature means that in general, when observing an edge, we can skip
 *  any target vertex that is reachable from a free vertex, regardless of
 *  its type.
 *
 *@section all_different_explanation_sccs SCCs common to the explained variable
 *
 *  Assume that \e u is the explained variable and \e U is its domain
 *  (i.e. set of values). All values of \e U which are in the same SCC as \e u
 *  are kept in the domain of \e u after the propagation (as explained
 *  below). This means that none of these values can cause or explain any
 *  propagation done towards \e u.
 *  
 *  The reason why the values of \e U which are in the SCC are kept in the
 *  domain of \e u because of the following two possibilities:
 *    - The outgoing edge of \e u must be in the SCC, otherwise there would
 *      be no path from \e u to the rest of the SCC. This edge is a part
 *      of a matching and of a feasible solution.
 *    - Any incoming edge <em> (x,u) </em> which is in the SCC can be
 *      made a matching edge. This is because \e x is also in the SCC and
 *      there must be a cycle from \e x through some vertexes to \e u and
 *      back to \e x.
 *      \n
 *      It is possible to toggle the direction of all edges of this
 *      cycle. After this, we get a new perfect matching of variables and
 *      values in which \e u and \e x are matched. 
 *      \n
 *      This implies that there is a feasible assignment in which
 *      <em>u=x</em>, hence the value \e x can't be removed from the domain of
 *      \e u.
 * 
 *  @par Example
 *    In the following example the domain of \e y is reduced, 1 is removed
 *    from its domain, but not 2 (which is part of the same SCC).
 *    Moreover, \e q and 4 which are connected to the SCC do not
 *    affect \e y.
 *    \n
 *    In this example the SCC is marked in gray.
 *    @dot
 *     digraph SCC_vertexes {
 *        rankdir=TB;
 *        size="5.5,2";
 *        subgraph cluster0 { q z y x }
 *        subgraph cluster1 { 4 3 2 1}
 *
 *        // The first SCC
 *        2 [style=filled];
 *        3 [style=filled];
 *        y [style=filled];
 *        z [style=filled];
 *        y -> 2;
 *        2 -> z [style=dashed];
 *        z -> 3 ;
 *        3 -> y [style=dashed];
 *
 *        // x is linked to this SCC
 *        x -> 1;
 *        1 -> y [style=dashed];
 *
 *        // q is linked to this SCC from another direction
 *        q -> 4;
 *        4 -> z [style=dashed];
 *     }
 *    @enddot
 *    \b Reminder: solid edges represent a matching between
 *    two vertexes. Dashed edges represent edges in the bipartite graph which
 *    are not a matching between the vertexes.
 *
 *  This property of <em>u</em>'s SCC lets the algorithm avoid cause and
 *  effect analysis of any path that goes through a value from this SCC.
 *  Any value vertex reachable from a free vertex does not constrain
 *  the problem, and is not explaining the propagation.
 *
 *@section all_different_explanation_cause Vertexes causing propagation.
 *
 *  Any vertex that there is a valid path from it to the explained variable
 *  is one of the vertexes explaining the propagation. A valid path
 *  must meet both of the following conditions:
 *   - It has no value vertex reachable from a free vertex.
 *   - It has no value vertex in the same SCC as the explained variable.
 *
 *  @par Example
 *    Assume the following domains:
 *       \f[
 *          \begin{array}{rl}
 *            a,b & \in  \{1,2\}   \\
 *            c   & \in  \{2,3,4\} \\
 *            d   & \in  \{3,4,6\} \\
 *            f   & \in  \{5,4\}   \\
 *            g,h & \in  \{6,7\}
 *          \end{array}
 *       \f]
 *    First we see that <em>f</em>'s value 4 cannot be removed unless
 *    both \e c, and \e d are reduced so that they take up this value.
 *    The following steps are the only way this can happen. \n
 *    \n
 *    It is easy to see that \e a and \e b take up the values 1 and 2.
 *    Similarly \e g and \e h take up 6 and 7. This causes a propagation
 *    resulting with \f$ c,d \in \{3,4\}\f$, which makes 3,4 to be taken
 *    up by \e c and \e d. Since without the propagation from <em>a,b,g,h</em>
 *    and their values <em>1,2,6,7</em> the propagation \f$ c,d \in \{3,4\}\f$
 *    would not happen, then we have
 *    to add all these variables and values to the explanation.\n
 *    \n
 *    This causes the removal of 4 from the domain of \e f. And the resulting
 *    explanation is:
 *     \f[( a\not\in \{1,2,3,4,6,7\} \vee
 *          b\not\in \{1,2,3,4,6,7\} \vee
 *          c \cdots \vee d \cdots \vee e \cdots \vee f \cdots \vee g \cdots
 *          \vee h \not\in \{1,2,3,4,6,7\}
 *        )\f]
 *    \n\n
 *    After this informal view let's look at the graph representation, Wee see
 *    that there are three SCCs, each marked with a different shade of gray.
 *    Edges from two SCCs ({1,2,a,b} and {6,7,g,h}) enter the third SCC
 *    ({3,4,c,d}).
 *    An edge from the third SCC enter to the explained variable \e f.
 *    @dot
 *     digraph SCC_vertexes {
 *        rankdir=TB;
 *        size="5.5,2";
 *        { node [style=filled,fillcolor=gray60]; h g 7 6; }
 *        { f 5; }
 *        { node [style=filled,fillcolor=gray90]; d c 4 3; }
 *        { node [style=filled,fillcolor=gray75]; b a 2 1; }
 *        subgraph cluster0_variables { h g f d c b a}
 *        subgraph cluster0_values    { 7 6 5 4 3 2 1}
 *        a -> 1 [weight=10];  // The first "leaf" SCC 
 *        1 -> b [style=dashed];
 *        b -> 2;
 *        2 -> a [style=dashed];
 *
 *        g -> 6;                // The second "leaf" SCC
 *        6 -> h [style=dashed];
 *        h -> 7;
 *        7 -> g [style=dashed];
 *
 *        c -> 3;                // The "inner" SCC
 *        3 -> d [style=dashed];
 *        d -> 4;
 *        4 -> c [style=dashed];
 *
 *        // The final nodes
 *        2 -> c [style=dashed];
 *        6 -> d [style=dashed];
 *        4 -> f [style=dashed];
 *        f -> 5;
 *     }
 *    @enddot
 *    \b Reminder: solid edges represent a matching between
 *    two vertexes. Dashed edges represent edges in the bipartite graph which
 *    are not a matching between the vertexes.\n
 *    \n
 *    None of the vertexes is reachable from a free vertex and the explained
 *    vertex is not in an SCC.
 *    This means that nodes <em>a,b,c,d,e,g,1,2,3,4,6,7</em> have
 *    valid paths to \e f. According to the rule from the beginning of this
 *    section, all  of these nodes, including \e f, are part of
 *    the explanation (this was seen also during the informal explanation).\n
 *    \n
 *    All these value vertexes explain the propagation because if we take one
 *    of these values and make them reachable from a free vertex, then
 *    the path to \e f, including 4, become reachable. As a result,
 *    4 no longer can be removed from the domain of \e f.\n
 *    \n
 *    Similarly, all these variables (except for the explained variable)
 *    explain the propagation. Take one of these variables and make it
 *    reachable from a free vertex. This variable has one outgoing edge,
 *    and it points to a value, which is on a path to \e f. This value becomes
 *    reachable from a free and, as we just have seen, 4 can no longer be
 *    removed from the domain of \e f. \n
 *    \n
 *    To summarize this, we can see that all vertexes in this example 
 *    explain the propagation.
 *
 *@section all_different_full_explanation The complete explanation algorithm
 *
 *    explain(variable \e u)
 *    -# Mark all edges of the bipartite graph as directional such that
 *        - Matching edges point from variable to value nodes.
 *        - Not-matching edges point the other way (value to variable nodes)
 *    -# EV := {u}
 *    -# Perform reverse DFS (going in revered edge direction) from \e u,
 *       and add all the visited vertexes to EV:
 *       - Ignore values reachable from a free vertex.
 *       - Ignore values that are in the same SCC as \e u.
 *    -# Create explanation \f$ (l_1 \vee l_2 \cdots l_n) \f$ such that
 *       \f$ l_i \equiv (v_i \not\in X) \f$, where \f$ v_i \f$ are all variable
 *       vertexes of EV, and X are all value vertexes of EV.
 */

void all_different_propagator::impl::explain_propagation(
    unsigned variable, const iset & removed_values_to_explain,
    vector<iset> & explanation) const
{
    HCSP_PROFILE_CURRENT;

    vector<unsigned> value_nodes_to_explain;
    convert_values_to_nodes(value_nodes_to_explain, removed_values_to_explain);
    vector<unsigned> matching;
    int failed_variable= maximal_matching(matching);
    if ( failed_variable >= 0 )
        variable= failed_variable;

    static graph directed_edges;
    get_directed_edges(matching, directed_edges);

    vector<unsigned char> is_reachable_from_free(m_graph.node_count(), false);
    for (unsigned node=0 ; node != directed_edges.node_count() ; ++node)
        if (matching[node] == -1U)
            mark_all_reachable(node, directed_edges, is_reachable_from_free);

    vector<unsigned> explaining_value_nodes;
    vector<unsigned> explaining_variable_nodes;
    if (failed_variable < 0)
        explaining_value_nodes= value_nodes_to_explain;
    get_explaining_nodes(directed_edges, is_reachable_from_free, variable,
                         explaining_variable_nodes, explaining_value_nodes);

    explanation_from_explaining_nodes(explaining_value_nodes,
                                      explaining_variable_nodes, explanation);
}

unsigned all_different_propagator::impl::construct_value_to_node(
    const constraint_runtime_info & info)
{
    HCSP_PROFILE_CURRENT;
    iset all_domain;
    for (unsigned i=0 ; i != m_number_of_variables ; ++i)
        all_domain.insert_all_of(info.domain(i));

    m_node_to_value.resize(m_number_of_variables + all_domain.size());
    copy(all_domain.begin(), all_domain.end(),
         m_node_to_value.begin()+m_number_of_variables);

    for (unsigned i= m_number_of_variables ; i != m_node_to_value.size() ; ++i)
        m_value_to_node.insert(make_pair(m_node_to_value[i], i));
 
    return m_node_to_value.size();
}

// See the documentation on impl::explain_propagation for more
// details on this algorithm. The documentation can be found in-line
// right before method definition (not method declaration).
// or in the Doxygen generated manual.
void all_different_propagator::impl::get_explaining_nodes(
    const graph & directed_edges,
    const vector<unsigned char> & is_reachable_from_free,
    unsigned variable,
    vector<unsigned> & explaining_variable_nodes,
    vector<unsigned> & explaining_value_nodes) const
{
    vector<unsigned> scc_associations;
    get_scc_associations(directed_edges, scc_associations);

    // reversed_edges:
    // (value,variable) matched edges;  (variable,value) unmatched edges
    static graph reversed_edges;
    reverse_edges(directed_edges, reversed_edges);
    vector<unsigned> nodes_to_handle= explaining_value_nodes;
    if ( nodes_to_handle.empty() )
        nodes_to_handle.push_back(variable);
    explaining_variable_nodes.push_back(variable);

    // A node reachable from a free node never causes propagation.
    vector<unsigned char> is_node_visited= is_reachable_from_free;
    while(!nodes_to_handle.empty() ) {
        unsigned affected_node= nodes_to_handle.back();
        nodes_to_handle.pop_back();
        is_node_visited[affected_node]= true;
        out_edge_iterator begin= reversed_edges.out_edges_begin(affected_node);
        out_edge_iterator end=   reversed_edges.out_edges_end(affected_node);
        for ( out_edge_iterator it= begin ; it != end; ++it) {
            unsigned causing_node= *it;
            if ( is_node_visited[causing_node] )
                continue;
            if (scc_associations[causing_node] == scc_associations[variable])
                continue; // This is tricky, see algorithm documentation.
            nodes_to_handle.push_back(causing_node);
            if (causing_node < affected_node)
                explaining_variable_nodes.push_back(causing_node);
            else
                explaining_value_nodes.push_back(causing_node);
        }
    }
}

void all_different_propagator::impl::explanation_from_explaining_nodes(
    const vector<unsigned> & explaining_value_nodes,
    const vector<unsigned> & explaining_variable_nodes,
    vector<iset> & explanation) const
{
    explanation.clear();
    explanation.resize(m_number_of_variables);

    iset explaining_values;
    for (unsigned i=0 ; i!= explaining_value_nodes.size() ; ++i)
        explaining_values.insert( m_node_to_value[ explaining_value_nodes[i]]);
    
    iset complement_explaining= explaining_values;
    complement_explaining.complement();

    for (unsigned i=0 ; i != explaining_variable_nodes.size() ; ++i)
        explanation[ explaining_variable_nodes[i] ]= complement_explaining;
}

void all_different_propagator::impl::convert_values_to_nodes(
    vector<unsigned> & value_nodes, const iset & values) const
{
    for (iset::iterator it=values.begin() ; it != values.end() ; ++it) {
        value_to_node_type::const_iterator
            conversion_it = m_value_to_node.find(*it);
        HCSP_ASSERT(conversion_it != m_value_to_node.end());
        value_nodes.push_back(conversion_it->second);
    }
}

/**
 *@page maximal_matching_algorithm Maximal matching algorithm.
 *
 *
 *@section maximal_matching_definitions Definitions required for the algorithm
 *
 *  The algorithm runs iteratively. Each iteration uses BFS to complete
 *  its task.
 *
 *  @par A free vertex
 *    Not connected to any matching edge.
 *
 *  @par Augmenting path
 *    An augmenting path is a path that:
 *     -# Starts with a free value vertex.
 *     -# Alternates through not-matching and matching edges
 *     -# Ends with a free variable vertex.
 *
 *@section maximal_matching_finding_augmenting_paths Finding augmenting paths
 *    -# BFS_layer := The set of free variables
 *    -# next BFS_layer := go from BFS_layer through new not-matching edges.
 *    -# If BFS_layer has free vertexes \b return the free vertexes.
 *    -# If BFS_layer is empty \b return {}.
 *    -# next BFS_layer := go from BFS_layer through new matching edges.
 *    -# Go to 2.
 *
 *   This algorithm has a little modification to make it more efficient
 *   considering my original misunderstanding of Hopcroft-Karp
 *   (see @ref maximal_matching_finding_augmenting_paths).
 *   The next BFS_layer (step 2 only) is done in a special way:
 *    -# For each vertex value v of BFS_layer:
 *       - Find <b>only the first</b> new not-matching edge, add its other end
 *         to the next BFS_layer.
 *    -# For each vertex value v of BFS_layer:
 *       - Find <b>all of the remaining</b> new not-matching edges, add their
 *         other end to the next BFS_layer.
 *
 *    This trick improves performance of the caller considerably, but
 *    seemingly not like the correct algorithm. This has to be fixed.
 *
 *@section maximal_matching_toggling_augmenting_paths Toggling augmenting paths
 *   This part does not match the original Hopcroft-Karp algorithm. It
 *   was due to incorrect or incomplete descriptions of the algorithm that
 *   I have read before implementing it. I have improved the incorrect
 *   algorithm, yet it seems to be not as good as the original.
 *
 *   To find the original algorithm read:\n
 *      http://en.wikipedia.org/wiki/Hopcroft%E2%80%93Karp_algorithm
 *
 *   My modified version of the algorithm:
 *    -# v := pop a vertex from the list of free values.
 *    -# If v = empty, return.
 *    -# Go backwards through the edges that discovered each vertex,
 *       until a free variable vertex is found.
 *    -# If the path has no shared vertexes with any of the previously
 *       found paths, toggle all edges along this path (each matching
 *       edge becomes not-matching and vice versa).
 *    -# Goto 1.
 *
 *@section maximal_matching_full The complete maximal matching algorithm on a bipartite graph
 *
 *   -# Mark all edges as unmatched, mark all vertexes as free.
 *   -# While has free variables
 *      -# Free_Vals :=
 *         @ref maximal_matching_finding_augmenting_paths "Perform BFS starting with all of the free variables"
 *      -# If Free_Vals = empty \b break
 *      -# @ref maximal_matching_toggling_augmenting_paths
 */
int all_different_propagator::impl::maximal_matching(
    vector<unsigned> & matching) const
{
    HCSP_PROFILE_CURRENT;
    
    vector<unsigned> unmatched_variables(m_number_of_variables);
    for (unsigned i=0 ; i != unmatched_variables.size() ; ++i)
        unmatched_variables[i]= i;
    matching.clear();
    matching.resize(m_graph.node_count(), -1U);

    // History of the last BFS run that searches for augmenting paths.
    vector<bfs_node_origin> bfs_nodes_info;
    
    // values vertexes which are not matched at the current iteration.
    static vector<unsigned> free_value_nodes; 
    while ( true ) {
        free_value_nodes.clear();
        bfs_nodes_info.clear();
        bfs_nodes_info.resize(m_graph.node_count());
        bfs_until_free_values(unmatched_variables, matching, bfs_nodes_info,
                              free_value_nodes);
        if (free_value_nodes.empty())
            break;

        // Augment the paths, making the matching set larger.
        for (unsigned i=0; i != free_value_nodes.size() ; ++i)
            toggle_matching(free_value_nodes[i], bfs_nodes_info, matching);

        // Remove matched variables from 'unmatched_variables'
        remove_if_matched(unmatched_variables, matching);
    }
    if ( unmatched_variables.empty() )
        return -1;
    else
        return unmatched_variables[0];
}

void all_different_propagator::impl::remove_if_matched(
    vector<unsigned> & elements, const vector<unsigned> & matching) const
{
    unsigned i=elements.size();
    while (i--) {
        unsigned element = elements[i];
        if (matching[element] != -1U) {
            // This is a matching element.
            elements[i]= elements.back();
            elements.pop_back();
        } 
    }
}

bool all_different_propagator::impl::is_augmenting_path(
    unsigned free_value_node, const vector<bfs_node_origin> & bfs_nodes_info,
    const vector<unsigned> & matching_table) const
{
    unsigned value_node= free_value_node;
    while (true) {
        HCSP_ASSERT(bfs_nodes_info[value_node].visited() &&
                   !bfs_nodes_info[value_node].external() );
    
        unsigned variable_node= bfs_nodes_info[value_node].origin();
        // This is a variable->value edge of the path, in an augmenting path it
        // should not be a matching edge.
        if ( matching_table[value_node] == variable_node )
            return false;

        //
        // Figure out what edge, if any, points to 'variable_node'
        //
        
        HCSP_ASSERT(bfs_nodes_info[variable_node].visited());
        if ( bfs_nodes_info[variable_node].external() )
            //  Here, 'variable_node' is the last node of the path.
            //  The path is augmenting iff 'variable_node' is also a free node.
            return matching_table[variable_node] == -1U;

        // An inner node in an alternating path can't be free (not matched)
        HCSP_ASSERT( matching_table[variable_node] != -1U);
        
        value_node= bfs_nodes_info[variable_node].origin();
        
        // This is a value->variable edge of the path, in an augmenting path it
        // should be a matching edge.
        if (matching_table[variable_node] != value_node)
            return false;
    }
}

void all_different_propagator::impl::toggle_matching(
    unsigned free_value_node, const vector<bfs_node_origin> & bfs_nodes_info,
    vector<unsigned> & matching_table) const
{
    // Make sure that the path is an augmenting path.
    // It is possible that it is not an augmenting path due to a previous
    // calls to toggle_matching() done for a different free node.
    if (!is_augmenting_path(free_value_node, bfs_nodes_info, matching_table))
        return;
 
    unsigned value_node= free_value_node;
    // This is an augmenting path: Toggle all matching/unmatching edges.
    // It starts with a free variable node and ends with a free value node.
    while (true) {
        unsigned variable_node= bfs_nodes_info[value_node].origin();
        // This is a variable->value edge of the path, in an augmenting path it
        // is not a matching edge.
        HCSP_ASSERT( matching_table[value_node] != variable_node );

        // Make it a matching edge.
        matching_table[value_node]= variable_node;
        matching_table[variable_node]= value_node;
        
        if ( bfs_nodes_info[variable_node].external() )
            // This is the last node, which used to be a free variable node.
            break;
        
        value_node= bfs_nodes_info[variable_node].origin();
        // This is a value->variable edge of the path, in an augmenting path it
        // is a matching edge.

        // Make it an not-matching edge.
        matching_table[value_node]= -1U;
        // No need to mark matching_table[variable_node] since it was set to
        // a different value (not to variable_node) by the first part of this
        // loop.
    }
}

void all_different_propagator::impl::bfs_until_free_values(
    const vector<unsigned> & first_bfs_variables,
    const vector<unsigned> & matching_table,
    vector<bfs_node_origin> & bfs_nodes_info,
    vector<unsigned> & free_value_nodes) const
{
    HCSP_PROFILE_CURRENT;
    free_value_nodes.clear();

    static vector<unsigned> next_variable_nodes, next_value_nodes;
    next_variable_nodes= first_bfs_variables;
    
    for (unsigned i=0 ; i != first_bfs_variables.size() ; ++i)
        bfs_nodes_info[ first_bfs_variables[i] ].mark_external();
    
    while (!next_variable_nodes.empty()) {
        next_value_nodes.clear();
        // I am splitting bfs next-value-nodes calculation into two phases.
        // This improves the chance to get more independent paths out of a
        // single bfs.
        // At the time, this was done as a way to overcome performance issues
        // caused by my misunderstanding of the Hopcroft-Karp algorithm.
        // Read the doxygen documentation of the maximal-matching algorithm.
        next_nodes(next_variable_nodes, next_value_nodes, matching_table,
                   bfs_nodes_info, UNMATCHING_EDGES, ONLY_FIRST);
        next_nodes(next_variable_nodes, next_value_nodes, matching_table,
                   bfs_nodes_info, UNMATCHING_EDGES);
        
        for (unsigned i=0 ; i != next_value_nodes.size() ; ++i) {
            unsigned value_node= next_value_nodes[i];
            if (matching_table[value_node] == -1U)
                free_value_nodes.push_back(value_node);
        }
        if ( !free_value_nodes.empty() )
            break;
        next_variable_nodes.clear();
        next_nodes(next_value_nodes, next_variable_nodes, matching_table,
                   bfs_nodes_info, MATCHING_EDGES);
        
    }
}

void all_different_propagator::impl::next_nodes(
    const vector<unsigned> & nodes, vector<unsigned> & next_nodes_vector,
    const vector<unsigned> & matching_table,
    vector<bfs_node_origin> & bfs_nodes_info, edges_used_type edges_used,
    completeness_type completeness) const
{
    if (edges_used == MATCHING_EDGES) {
        for (unsigned i=0 ; i != nodes.size() ; ++i) {
            const unsigned from=nodes[i];
            unsigned to= matching_table[from];
            if (bfs_nodes_info[to].visited() || to == -1U)
                continue;
            bfs_nodes_info[to].mark_origin(from);
            next_nodes_vector.push_back(to);
        }
    } else {
        // Ignore matching edges
        for (unsigned i=0 ; i != nodes.size() ; ++i) {
            const unsigned from=nodes[i];
            for ( out_edge_iterator it= m_graph.out_edges_begin(from);
                  it  != m_graph.out_edges_end(from) ;
                  ++it) {
                unsigned to= *it;
                if (bfs_nodes_info[to].visited() || matching_table[from] == to)
                    continue;
                bfs_nodes_info[to].mark_origin(from);
                next_nodes_vector.push_back(to);
                if (completeness == ONLY_FIRST)
                    break;
            }
        }
    }
}

void all_different_propagator::impl::get_directed_edges(
    const vector<unsigned> & matching_table,
    graph & directed_edges) const
{
    HCSP_PROFILE_CURRENT;
    directed_edges.clear_edges();
    directed_edges.resize(m_graph.node_count());
    for (unsigned i= m_number_of_variables ; i != m_graph.node_count() ; ++i)
        directed_edges.reserve_edges_for_node(m_graph.count_out_edges(i), i);

    for (unsigned from_node=0; from_node!=m_number_of_variables ; ++from_node){
        for ( out_edge_iterator it= m_graph.out_edges_begin(from_node);
              it != m_graph.out_edges_end(from_node) ;
              ++it) {
            unsigned to_node= *it;
            if ( from_node == matching_table[to_node])
                directed_edges.add_edge(from_node, to_node);
            else
                directed_edges.add_edge(to_node, from_node);
        }
    }
}

// http://en.wikipedia.org/wiki/Tarjan%27s_strongly_connected_components_algorithm
void all_different_propagator::impl::tarjan(
    vector<tarjan_node_info> & nodes_visit_info, unsigned current_node,
    unsigned & visit_count, vector<unsigned> & visited_unassociated_node_stack,
    const graph & directed_edges) const
{
    tarjan_node_info & current_node_info= nodes_visit_info[current_node];
    HCSP_ASSERT( !current_node_info.visited() );
    current_node_info.set_visit_order(visit_count++);
    
    visited_unassociated_node_stack.push_back(current_node);
    current_node_info.mark_on_stack(true);
    
    out_edge_iterator begin= directed_edges.out_edges_begin(current_node);
    out_edge_iterator end=   directed_edges.out_edges_end(current_node);
    for ( out_edge_iterator it= begin; it != end; ++it) {
        unsigned next_node= *it;
        const tarjan_node_info & next_node_info= nodes_visit_info[next_node];
        if (next_node_info.on_stack())
            // Detected a new loop, recalculate loop visit for current_node..
            current_node_info.set_earliest_looping_visit(next_node_info);
        else if (! next_node_info.visited() ) {
            // next_node_info is updated by this call to tarjan().
            tarjan(nodes_visit_info, next_node, visit_count,
                   visited_unassociated_node_stack, directed_edges);
            // If recursion found a loop earlier than current_node then
            // current_node is part of that loop - record it.
            current_node_info.set_earliest_looping_visit(next_node_info);
        }
    }
    if ( current_node_info.is_earliest_looping_visit()) {
        // This is the exit point from this SCC (the last time we see it).
        unsigned scc_node;
        do {
            HCSP_ASSERT(!visited_unassociated_node_stack.empty());
            scc_node= visited_unassociated_node_stack.back();
            
            visited_unassociated_node_stack.pop_back();
            nodes_visit_info[scc_node].mark_on_stack(false);
            nodes_visit_info[scc_node].set_scc_association(current_node);
        } while (scc_node != current_node);
    }
}

void all_different_propagator::impl::get_scc_associations(
    const graph & directed_edges,
    vector<unsigned> &scc_associations) const
{
    HCSP_PROFILE_CURRENT;
    
    vector<tarjan_node_info> nodes_visited(m_graph.node_count());
    unsigned                 visit_count=0;
    vector<unsigned>         visited_unassociated_node_stack;
    
    visited_unassociated_node_stack.reserve(m_graph.node_count());
    
    for (unsigned node=0 ; node != m_graph.node_count() ; ++node) {
        if (!nodes_visited[node].visited())
            tarjan(nodes_visited, node, visit_count,
                   visited_unassociated_node_stack, directed_edges);
    }

    scc_associations.clear();
    scc_associations.resize(nodes_visited.size());
    for (unsigned node=0; node != nodes_visited.size() ; ++node)
        scc_associations[node]= nodes_visited[node].scc_association();
}


void all_different_propagator::impl::get_scc_edges(
    const graph & directed_edges,
    vector<pair<unsigned, unsigned> > & final_edges) const
{
    vector<unsigned> scc_associations;
    get_scc_associations(directed_edges, scc_associations);
    
    for (unsigned i=0 ; i!=directed_edges.node_count(); ++i){
        unsigned from_node= i;

        out_edge_iterator end= directed_edges.out_edges_end(from_node);
        for ( out_edge_iterator it= directed_edges.out_edges_begin(from_node);
              it != end;
              ++it) {
            unsigned to_node= *it;
            if ( scc_associations[from_node] == scc_associations[to_node] )
                // The edge connects two nodes withing the same SCC
                final_edges.push_back(make_pair(from_node, to_node));
        }
    }
}

void all_different_propagator::impl::add_reachable_edges(
    unsigned node,
    const graph & directed_edges,
    vector<unsigned char> & is_reached_node,
    vector<pair<unsigned, unsigned> > & final_edges) const
{
    is_reached_node.resize(directed_edges.node_count(), 0); 
    if ( is_reached_node[node] )
        return;
    is_reached_node[node]= true;
    out_edge_iterator end= directed_edges.out_edges_end(node);
    for ( out_edge_iterator it= directed_edges.out_edges_begin(node);
          it != end;
          ++it) {
        unsigned to_node= *it;

        final_edges.push_back(make_pair(node, to_node));
        add_reachable_edges(to_node, directed_edges, is_reached_node,
                            final_edges);
    }
}

void all_different_propagator::impl::mark_all_reachable(
    unsigned node, const graph & directed_edges,
    vector<unsigned char> & is_reached_node) const
{
    if ( is_reached_node[node] )
        return;
    is_reached_node[node]= true;
    out_edge_iterator end= directed_edges.out_edges_end(node);
    for ( out_edge_iterator it= directed_edges.out_edges_begin(node);
          it != end;
          ++it) 
        mark_all_reachable(*it, directed_edges, is_reached_node);
}

void all_different_propagator::impl::reverse_edges(const graph &directed_edges,
                                                   graph &reversed_edges) const
{
    reversed_edges.clear_edges();
    reversed_edges.resize(directed_edges.node_count());
    for (unsigned from=0 ; from != directed_edges.node_count(); ++from) {
        for ( out_edge_iterator it= directed_edges.out_edges_begin(from);
              it != directed_edges.out_edges_end(from);
              ++it) {
            unsigned to= *it;
            reversed_edges.add_edge(to, from);
        }
    }
}


//virtual
void all_different_propagator::explain_propagation(
    unsigned propagated_param, const iset & values_to_explain,
    const iset & propagated_domain,
    const constraint_runtime_info & info_before_propagation,
    vector<iset> & explanation) const
{
    HCSP_PROFILE_CURRENT;
    
    iset removed_values_to_explain;
    if ( !propagated_domain.empty() ) {
        removed_values_to_explain=
            info_before_propagation.domain(propagated_param);
        removed_values_to_explain.erase_all_of(propagated_domain);
        removed_values_to_explain.intersect(values_to_explain);
        HCSP_ASSERT(!removed_values_to_explain.empty());
    }

    vector<unsigned char> big_parameters=
        find_big_parameters(info_before_propagation);
    if (big_parameters.empty()) {
        impl data_structure(info_before_propagation);
        data_structure.explain_propagation(propagated_param,
                                           removed_values_to_explain,
                                           explanation);
    } else {
        static vector<iset> domains;
        unsigned propagated_small_param=
            get_explanation_small_parameters(
                info_before_propagation, propagated_param,
                propagated_domain, big_parameters, domains);
        
        impl data_structure(domains);
        explanation.resize(domains.size());
        data_structure.explain_propagation(propagated_small_param,
                                           removed_values_to_explain,
                                           explanation);
        extend_explanation_for_big_domains(info_before_propagation.arity(),
                                           big_parameters, explanation);
    }
}

//virtual
std::unique_ptr<explanation_rule>
all_different_propagator::rule(const vector<iset> & literals) const
{
    return unique_ptr<explanation_rule>(new all_diff_rule(literals));
}

//virtual
void all_different_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << " all_different(";
        for (unsigned i=0 ; i != p_info->arity() ; ++i) {
            if (i != 0)
                out << ", ";
            p_info->print_param(out, i);
            out << "{" << p_info->domain(i) << "}";
        }            
        out << ")";
    } else {
        out << "all_different";
    }
}

//virtual
all_different_propagator::result_type
all_different_propagator::propagate( constraint_runtime_info & info)
{
    HCSP_PROFILE_CURRENT;
    if ( info.originating_queue() == info.VARIABLE_QUEUE) {
        info.put_in_queue(info.CONSTRAINT_QUEUE);
        return POSTPONE;
    }
    if (info.arity() > (unsigned)iset::MAX_DOMAIN) {
        HCSP_ASSERT(iset::MIN_DOMAIN == 0);
        return UNSAT;
    }
    bool important_modified= false;
    unsigned singletons=0;
    for (unsigned i=0 ; i!=info.arity() ; ++i)
        if (info.domain(i).singleton() && !info.reduced_between_activations(i))
            ++singletons;
    for (unsigned i=0 ; !important_modified && i!=info.arity() ; ++i) {
        if (info.reduced_between_activations(i) &&
            info.domain(i).size()+singletons < info.arity() )
            important_modified= true;
    }
    if (!important_modified)
        return CONSISTENT;

    vector<unsigned char> big_parameters= find_big_parameters(info);
    unsigned count_big= accumulate(big_parameters.begin(),big_parameters.end(),
                                   0U);
    if ( count_big == info.arity())
        return CONSISTENT;

    static vector<iset> next_domains;
    if (count_big == 0) {
        impl data_structure(info);
        if (!data_structure.propagate(next_domains))
            return UNSAT;
    } else {
        bool last_represents_big_params= false;
        static vector<iset> domains;
        if (!get_small_parameters(info, big_parameters,
                                  domains, last_represents_big_params))
            return UNSAT;
        impl data_structure(domains);
        if (!data_structure.propagate(next_domains))
            return UNSAT;
        extend_propagation_for_big_domains(info, domains, big_parameters,
                                           last_represents_big_params,
                                           next_domains);
    }

    return update_domains(info, next_domains);
}

//virtual
bool
all_different_propagator::conflicts(const constraint_runtime_info & info) const
{
    HCSP_PROFILE_CURRENT;
    iset singletons;
    for (unsigned i=0 ; i!=info.arity() ; ++i) {
        const iset & domain= info.domain(i);
        if (domain.singleton()) {
            iset::value_type value= domain.min();
            if (singletons.contains(value))
                return true;
            singletons.insert(value);
        }
    }
    vector<unsigned char> big_parameters= find_big_parameters(info);
    unsigned count_big= accumulate(big_parameters.begin(),big_parameters.end(),
                                   0U);
    if ( count_big + singletons.size() == info.arity())
        return false;

    static vector<iset> next_domains;
    if (count_big == 0) {
        impl data_structure(info);
        return !data_structure.propagate(next_domains);
    } else {
        bool last_represents_big_params= false;
        static vector<iset> domains;
        if (!get_small_parameters(info, big_parameters,
                                  domains, last_represents_big_params))
            return true;
        impl data_structure(domains);
        return !data_structure.propagate(next_domains);
    }
}

//virtual
bool all_different_propagator::constraint_check(
    const constraint_runtime_info & info) const
{
    iset values;
    unsigned size= info.arity();
    for (unsigned i=0 ; i != size ; ++i)
        values.insert(info.domain(i).min());
    return values.size() == size;
}

// static
unsigned all_different_propagator::count_unique_singletons(
    const constraint_runtime_info & info)
{
    HCSP_PROFILE_CURRENT;
    static vector<iset::value_type> singletons;
    singletons.clear();
    for (unsigned i=0 ; i != info.arity() ; ++i) {
        const iset & domain= info.domain(i);
        if (domain.singleton())
            singletons.push_back(domain.min());
    }
    if (singletons.empty())
        return 0;

    sort(singletons.begin(), singletons.end());
    singletons.erase( unique(singletons.begin(), singletons.end()),
                      singletons.end());

    for (unsigned i=0 ; i != info.arity() ; ++i) {
        const iset & domain= info.domain(i);
        if ( domain.singleton() )
            continue;
        iset::value_type min= domain.min();
        iset::value_type max= domain.max();
        for (unsigned j= singletons.size() ; j != 0 ; ) {
            --j;
            iset::value_type singleton = singletons[j];
            if ( singleton >= min && singleton <= max &&
                 domain.contains(singleton) ) {
                singletons[j] = singletons.back();
                singletons.pop_back();
            }
        }
    }
    return singletons.size();
}

// static
vector<unsigned char> all_different_propagator::find_big_parameters(
    const constraint_runtime_info & info)
{
    HCSP_PROFILE_CURRENT;
    vector<unsigned char> big_parameters(info.arity(), false);    
    bool stable= false;
    unsigned singletons= count_unique_singletons(info);
    iset::size_type max_small_elements= info.arity()-singletons;
    if (max_small_elements == 0)
        return big_parameters;
    if (max_small_elements > 1)
        --max_small_elements;
    iset::size_type next_max_small_elements= max_small_elements;
    while (!stable) {
        stable= true;
        for (unsigned i=0 ; i != info.arity() ; ++i) {
            if ( big_parameters[i] )
                continue;
            const iset & domain= info.domain(i);
            if ( domain.size() > max_small_elements ||
                 (domain.size() == 0 && domain.everything())) {
                stable= false;
                big_parameters[i] = true;
                max_small_elements= next_max_small_elements;
                --next_max_small_elements;
                if (max_small_elements == 0)
                    return big_parameters;
            }
        }
    }
    return big_parameters;
}

// static
vector<iset::value_type> all_different_propagator::get_small_domain_values(
    const constraint_runtime_info & info,
    const vector<unsigned char> & big_parameters)
{
    HCSP_PROFILE_CURRENT;
    vector<iset::value_type> values;
    values.reserve(info.arity()*4);
    for (unsigned i=0 ; i != info.arity() ; ++i) {
        if ( !big_parameters[i]) {
            const iset & domain= info.domain(i);
            values.insert(values.end(), domain.begin(), domain.end());
        }
    }
    sort(values.begin(), values.end());
    values.erase( unique(values.begin(), values.end()),
                  values.end());
    return values;
}

// static
iset all_different_propagator::get_big_domain(
    const constraint_runtime_info &info,
    const vector<unsigned char> & big_parameters,
    vector<iset::value_type> & values_to_check, const iset & small_domains)
{
    HCSP_PROFILE_CURRENT;
    iset big_domain;
    bool extra_value_required= false;
    for (unsigned i=0 ; i != info.arity() ; ++i)
        if ( big_parameters[i]) {
            const iset & domain= info.domain(i);
            if (!extra_value_required)
                extra_value_required= !small_domains.contains_all_of(domain);
            iset::value_type min= domain.min();
            iset::value_type max= domain.max();
            for (unsigned j= values_to_check.size() ; j != 0 ; ) {
                --j;
                iset::value_type value= values_to_check[j];
                if ( value >= min && value <= max && domain.contains(value)) {
                    big_domain.insert(value);
                    values_to_check[j]= values_to_check.back();
                    values_to_check.pop_back();
                }
            }
        }
    
    if ( extra_value_required && !big_domain.empty() &&
         big_domain.size() < info.arity()) {
        iset::value_type extra_value=
            small_domains.min() > iset::MIN_DOMAIN ?
            small_domains.min() - 1 :
            small_domains.begin().current_range().second+1;
        big_domain.insert(extra_value);
    }
    return big_domain;
}

bool all_different_propagator::get_small_parameters(
    const constraint_runtime_info & info,
    const vector<unsigned char> & big_parameters,
    vector<iset> & domains,
    bool & last_represents_big_params)
{
    HCSP_PROFILE_CURRENT;
    vector<iset::value_type> values_to_check=
        get_small_domain_values(info, big_parameters);

    iset small_domains;
    for (unsigned i=0 ; i!=values_to_check.size() ; ++i)
        small_domains.insert(values_to_check[i]);

    unsigned small_parameters_count=
        info.arity() -
        accumulate(big_parameters.begin(),big_parameters.end(),0U);
    
    domains.clear();
    domains.reserve(small_parameters_count+ 1);
    
    for (unsigned i=0 ; i!= info.arity() ; ++i)
        if (!big_parameters[i])
            domains.push_back(info.domain(i));
    domains.push_back(get_big_domain(info, big_parameters, values_to_check,
                                     small_domains));
    last_represents_big_params= !domains.back().empty();
    if (domains.back().empty())
        domains.pop_back();
        
    return small_domains.size() >= small_parameters_count;
}

// static
void all_different_propagator::extend_propagation_for_big_domains(
    const constraint_runtime_info & info, const vector<iset> & domains,
    const vector<unsigned char> & big_parameters,
    bool last_represents_big_params,
    vector<iset> & next_domains)
{
    HCSP_PROFILE_CURRENT;
    static vector<iset> small_domains;
    next_domains.swap(small_domains);
    next_domains.resize(info.arity());
    iset erased_from_big;
    if (last_represents_big_params) {
        erased_from_big= domains.back();
        erased_from_big.erase_all_of(small_domains.back());
    }
    for (unsigned i=0, small_parameter=0 ; i != info.arity() ; ++i)
        if (big_parameters[i]) {
            next_domains[i]= info.domain(i);
            next_domains[i].erase_all_of(erased_from_big);
        } else {
            next_domains[i]= small_domains[small_parameter++];
        }
}

// static
unsigned all_different_propagator::get_explanation_small_parameters(
    const constraint_runtime_info & info_before_propagation,
    unsigned propagated_param, const iset & propagated_domain,
    vector<unsigned char> & big_parameters, vector<iset> & domains)
{
    bool last_represents_big_params= false;
    get_small_parameters(info_before_propagation, big_parameters,
                         domains, last_represents_big_params);

    unsigned propagated_small_param=
        propagated_param-accumulate(big_parameters.begin(),
                                    big_parameters.begin()+propagated_param,
                                    0U);
    
    // Get rid of special handling for big domains.
    if ( last_represents_big_params ) {
        iset big_domain= domains.back();
        domains.pop_back();
        last_represents_big_params= false; // just for debug clarity
        // Check if the big domain is required for the explanation.
        if ( !propagated_domain.empty() &&
             big_parameters[propagated_param] ){
            // Need to explain a big parameter, but only the removal has to be
            // explained. So simply "make" the parameter small by considering
            // only parts that may be removed, i.e. mark the big explained
            // parameter as small and put the domains.back() to there.
            HCSP_ASSERT(big_parameters[propagated_param]);
            big_parameters[propagated_param]= false;
            domains.insert(domains.begin()+propagated_small_param, big_domain);
        }
    }
    return propagated_small_param;
}

// static
void all_different_propagator::extend_explanation_for_big_domains(
    unsigned arity, const vector<unsigned char> & big_parameters,
    vector<iset> & explanation)
{
    vector<iset> small_explanation= explanation;
    explanation.resize(arity);
    for (unsigned i=0, small_parameter=0 ; i != arity ; ++i) {
        if (big_parameters[i])
            explanation[i]= iset();
        else
            explanation[i]= small_explanation[small_parameter++];
    }
}


all_different_propagator::result_type
all_different_propagator::update_domains( constraint_runtime_info & info,
                                          const vector<iset> & next_domains)
{
    HCSP_PROFILE_CURRENT;
    for (unsigned i=0 ; i != info.arity(); ++i) {
        info.set_domain(i, next_domains[i]);
    }
    return CONSISTENT;
}

//  LocalWords:  bfs dfs scc tarjan Tarjan's
