// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <arithmetic_propagators.hpp>

#include <typeinfo>
#include <numeric>
#include <limits>
#include <iostream>
#include <algorithm>
#include <propagation.hpp>
#include <comparison_propagators.hpp>
#include <disjunction_specification.hpp>
#include <disjoint_propagators.hpp>
#include <sliced_runtime_info.hpp>
#include <int128.hpp>
using namespace std;
namespace hcsp {

typedef pair<long long, long long> interval_type;
static interval_type init_range()
{
    return interval_type(numeric_limits<long long>::max(),
                         numeric_limits<long long>::min());
}

class mul_truncate_propagator::algo
{
  public:
    typedef iset::iterator   iterator;
    typedef iset::value_type value_type;

    /** @brief Construct the algo object
     * @param domains  The array of domains to be propagated. The algorithms
     *                 works on this array and eliminate unsupported values.
     */
    algo(iset domains[3]) : m_domains(domains)
    {
        m_sizes[0]=m_sizes[1]=m_sizes[2]=0;
    }

    //! Do the actual work, update the domains array (passed to the constructor)
    void propagate_bounds();


    bool empty() const
    {
        return (m_domains[0].empty() ||
                m_domains[1].empty() ||
                m_domains[2].empty() );
    }
    void clear()
    {
        m_domains[0].clear();
        m_domains[1].clear();
        m_domains[2].clear();
    }
  private:
    /** @brief Update the inner record on the latest domain sizes
     *  @return True if the sizes have changes since last call.
     */
    bool update_sizes()
    {
        bool modified= false;
        for (unsigned i=0 ; i < 3 ; ++i) {
            iset::size_type current_size= m_domains[i].size();
            if (m_sizes[i] != current_size)
                modified= true;
            m_sizes[i]= current_size;
        }
        return modified;
    }
    /**@brief Propagate and update m_domains[0] which is the output result of
     *        mul.
     *
     * Performs weak bounds consistency, when only the bounds of multiplicand
     * and the multiplier are taken into account. Bad values of m_domains[0]
     * are not removed if they are within the bounds of the multiplication
     * result, even when they are on the bound of m_domains[0] and unsupported.
     */
    void propagate_out();
    
    /**@brief Propagate and update m_domains[multiplicand_index] from the
     *         result of mul and the multiplier.
     *
     * This will method will perform a propagation which is only slightly
     * stronger from weak bounds consistency. It propagates three different
     * zones, separately, and joins the results:
     *  - 0 value of the multiplicand and its support
     *  - strictly positive values of the multiplicand and their supports
     *    (performs bounds consistency on positive results).
     *  - negative values of the multiplicand then their supports
     *    (bounds consistency over negative results)
     *
     * @param multiplicand_index  Either 1 or 2: the parameter to update.
     */
    void propagate_to_multiplicand(unsigned multiplicand_index);

    /** @brief Propagate the case when the multiplicand is zero.
     *
     * Propagate the zone when the multiplicand is zero.
     * Read about zones in the documentation of propagate_to_multiplicand().
     * @param[in,out] multiplicand_bounds
     *                The values of the multiplicand which must be retained.
     *                This set is later intersected with the actual domain,
     *                meaning that it is allowed to be bigger than the domain,
     *                at this moment.
     * @param multiplicand_index
     *                Either 1 or 2: the parameter to update.
     */
    void insert_zero_supported_multiplicand(iset & multiplicand_bounds,
                                            unsigned multiplicand_index);
    
    /** @brief Propagate the case when the multiplicand is strictly positive
     *
     * Propagate the zone when the multiplicand is strictly positive.
     * Read about zones in the documentation of propagate_to_multiplicand().
     * @param[in,out] multiplicand_bounds
     *                The values of the multiplicand which must be retained.
     *                This set is later intersected with the actual domain,
     *                meaning that it is allowed to be bigger than the domain,
     *                at this moment.
     * @param multiplicand_index
     *                Either 1 or 2: the parameter to update.
     */
    void insert_positive_supported_multiplicand_bounds(
        iset & multiplicand_bounds, unsigned multiplicand_index);
    
    /** @brief Propagate the case when the multiplicand is negative
     *
     * Propagate the zone when the multiplicand is negative.
     * Read about zones in the documentation of propagate_to_multiplicand().
     * @param[in,out] multiplicand_bounds
     *                The values of the multiplicand which must be retained.
     *                This set is later intersected with the actual domain,
     *                meaning that it is allowed to be bigger than the domain,
     *                at this moment.
     * @param multiplicand_index
     *                Either 1 or 2: the parameter to update.
     */
    void insert_negative_supported_multiplicand_bounds(
        iset & multiplicand_bounds, unsigned multiplicand_index);

    //! The latest values of domains which are propagated.
    iset *const m_domains;
    iset::size_type m_sizes[3];
}; // class mul_truncate_propagator::algo

static inline void widen_range(interval_type & range, long long value)
{
    if (value < range.first)
        range.first= value;
    if (value > range.second)
        range.second= value;
}

static void intersect_with_interval(iset & domain,
                                    const interval_type & interval)
{
    interval_type fixed_interval= interval;
    if ( fixed_interval.first < iset::MIN_DOMAIN )
        fixed_interval.first= iset::MIN_DOMAIN;
    if ( fixed_interval.second > iset::MAX_DOMAIN)
        fixed_interval.second= iset::MAX_DOMAIN;
    if (fixed_interval.first > fixed_interval.second)
        domain.clear();
    else
        domain.intersect(fixed_interval);
}

static void insert_interval(iset & domain, const interval_type & interval)
{
    interval_type fixed_interval= interval;
    if (fixed_interval.first < iset::MIN_DOMAIN)
        fixed_interval.first= iset::MIN_DOMAIN;
    if (fixed_interval.second > iset::MAX_DOMAIN)
        fixed_interval.second= iset::MAX_DOMAIN;
    if (fixed_interval.first > fixed_interval.second)
        return;
    domain.insert(fixed_interval);
}
/** @brief Find the iterator representing the negative value which is closest
 *         to zero.
 *
 *  @return An iterator pointing to the required value or end() if no such
 *          value exists.
 */
static iset::iterator max_negative_it(const iset & set)
{
    if (set.empty() || set.min() >= 0)
        return set.end();
    else
        return --(set.lower_bound(0));
}
//virtual
constraint_propagator::result_type
neg_propagator::propagate( constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    const iset & in0= info.domain(0);
    const iset & in1= info.domain(1);

    iset out0=in0, out1=in1;
    hcsp::propagation::neg(in0, out1);
    hcsp::propagation::neg(in1, out0);
    if ( out0.empty() )
        return UNSAT;
    info.set_domain(0, move(out0));
    info.set_domain(1, move(out1));
    return CONSISTENT;
}

//virtual
bool neg_propagator::conflicts(const constraint_runtime_info & info) const
{
    HCSP_PROFILE_CURRENT;
    const iset & in0= info.domain(0);
    const iset & in1= info.domain(1);

    iset out0=in0;
    hcsp::propagation::neg(in1, out0);
    return out0.empty();
}

//virtual
void neg_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        p_info->print_param(out, 0);
        out << "== -";
        p_info->print_param(out, 1);
        out << ")";
    } else {
        out << "negate";
    }
}

/**
 * @brief Insert supports of nonnegative input values to abs.
 *
 * Insert supported values into out0, and out1. Insert values resulting from
 * nonnegative inputs to abs.
 *
 * @param[in] in0       For a=abs(b) this is the input domain of 'a'
 * @param[in] it1,end1  For a=abs(b) this is the range of input domains of 'b'
 *                      Usually, this range contains at least all nonnegative
 *                      values of 'b'.
 * @param[in,out] out0,out1   The sets to which supported values are inserted
 * @return   The last iterator used for support search.
 *           The search starts with \p it1 and may end with \p end1.
 */
// static
iset::iterator
abs_propagator::insert_nonnegative_supports(const iset & in0,
                                            iset::iterator it1,
                                            iset::iterator end1,
                                            iset &out0, iset &out1)
{
    iset::iterator it0= in0.lower_bound(0);
    while ( it1 != end1 && it0 != in0.end() ) {
        iset::value_range_type range0= it0.current_range();
        iset::value_range_type range1= it1.current_range();
        if ( range0.second >= range1.first &&
             range0.first <= range1.second) {
            // both iterators overlap.
            iset::value_range_type common(
                max(range0.first,  range1.first),
                min(range0.second, range1.second));
            out0.insert(common);
            out1.insert(common);
        }
        if (range0.second <= range1.second)
            it0.advance_range();
        else 
            it1.advance_range();
    }
    return it1;
}

/**
 * @brief Insert supports of negative input values to abs.
 *
 * Insert supported values into out0, and out1. Insert values resulting from
 * negative inputs to abs.
 *
 * @param[in] in0       For a=abs(b) this is the input domain of 'a'
 * @param[in] it1,end1  For a=abs(b) this is the range of input domains of 'b'
 *                      Usually, this range contains at least all negative
 *                      values of 'b'.
 * @param[in,out] out0,out1   The sets to which supported values are inserted
 * @return   The last iterator used for support search.
 *           The search starts with \p it1 and may end with \p end1.
 *           It is safe to start nonnegative support check from this value.
 */
// static
iset::iterator
abs_propagator::insert_negative_supports(const iset & in0,
                                         iset::iterator it1,
                                         iset::iterator end1,
                                         iset &out0, iset &out1)
{
    if (in0.empty() || *it1 >= 0)
        return it1;
    HCSP_ASSERT(numeric_limits<iset::value_type>::min() < 0);
    // the value of abs(MIN_INT) overflows, skip it.
    if (*it1 == numeric_limits<iset::value_type>::min())
        ++it1;
    iset::iterator it0= in0.end();
    it0.retract_range();
    while ( it0.current_range().second >= 0 && it1 != end1 ) {
        iset::value_range_type range0= it0.current_range();
        if (*it1 >=0 || range0.second < 0)
            break;
        range0.first= max(0, range0.first);
        
        iset::value_range_type range1= it1.current_range();
        bool finished1= (range1.second >= 0);
        if (finished1)
            range1.second= -1;
        
        iset::value_range_type intersection(max(range0.first,  -range1.second),
                                            min(range0.second, -range1.first));
        if ( intersection.first <= intersection.second) {
            // both iterators overlap.
            out0.insert(intersection);
            out1.insert(make_pair(-intersection.second, -intersection.first));
        }

        if ( -range1.second <= range0.first) {
            if (it0 == in0.begin())
                break;
            it0.retract_range();
        } else {
            if (finished1)
                break;
            it1.advance_range();
        } 
    }
    return it1;
}

//virtual
constraint_propagator::result_type
abs_propagator::propagate( constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    const iset & in0= info.domain(0);
    const iset & in1= info.domain(1);
    
    iset out0;
    iset out1;
    if (in1.min() >= 0) {
        // simply equate
        if (in0 == in1)
            return CONSISTENT;
        out0= in0;
        out0.intersect(in1);
        out1= out0;
    } else {
        iset::iterator it;
        it= insert_negative_supports(in0, in1.begin(), in1.end(), out0, out1);
        insert_nonnegative_supports(in0, it, in1.end(), out0, out1 );
    }
    if ( out0.empty() )
        return UNSAT;
    info.set_domain(0, move(out0));
    info.set_domain(1, move(out1));
    return CONSISTENT;
}

//virtual
bool abs_propagator::conflicts(const constraint_runtime_info & info) const
{
    HCSP_PROFILE_CURRENT;
    const iset & in0= info.domain(0);
    const iset & in1= info.domain(1);
    
    if (in1.min() >= 0) {
        return !in0.contains_any_of(in1);
    } else {
        iset out0;
        iset out1;
        iset::iterator it;
        it= insert_negative_supports(in0, in1.begin(), in1.end(), out0, out1);
        if (!out0.empty())
            return false;
        out0.clear();
        insert_nonnegative_supports(in0, it, in1.end(), out0, out1);
        return out0.empty();
    }
}

//virtual
void abs_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "";
        p_info->print_param(out, 0);
        out << "== abs(";
        p_info->print_param(out, 1);
        out << ")";
    } else {
        out << "abs";
    }
}

//virtual
constraint_propagator::result_type
less_equal_abs_propagator::propagate( constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    const iset & in0= info.domain(0);
    const iset & in1= info.domain(1);

    if (in0.empty() || in1.empty())
        return UNSAT;
    
    long long biggest_abs= max(value_abs(in1.min()), value_abs(in1.max()));
    iset::value_type min_in0= in0.min();

    if (in0.max() > biggest_abs) {
        iset out0= info.domain(0);
        out0.intersect(make_pair(in0.MIN_DOMAIN, biggest_abs));
        if (out0.empty())
            return UNSAT;
        info.set_domain(0, move(out0));
    }

    if (min_in0 <= 0)
        return CONSISTENT;
    pair<iset::value_type, iset::value_type> exclude_abs;
    exclude_abs.second= min_in0-1;
    exclude_abs.first=  max(iset::MIN_DOMAIN,-exclude_abs.second);
    if (in1.contains_any_of(iset(exclude_abs))) {
        iset out1= info.domain(1);
        out1.erase(exclude_abs);
        HCSP_ASSERT(!out1.empty()); // handled during out0's update
        info.set_domain(1, move(out1));
    }
    return CONSISTENT;
}

//virtual
bool less_equal_abs_propagator::conflicts(const constraint_runtime_info & info)
    const
{
    const iset & in0= info.domain(0);
    const iset & in1= info.domain(1);

    if (in0.empty() || in1.empty())
        return true;
    
    long long biggest_abs= max(value_abs(in1.min()), value_abs(in1.max()));
    return in0.min() > biggest_abs;
}

//virtual
void less_equal_abs_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "";
        p_info->print_param(out, 0);
        out << "<= abs(";
        p_info->print_param(out, 1);
        out << ")";
    } else {
        out << "less_equal_abs";
    }
}

//static
long long greater_equal_abs_propagator::min_abs(const iset & domain)
{
    iset::value_type domain_min= domain.min();
    if (domain_min>=0)
        return domain_min;
    
    iset::value_type max= domain.max();
    if (max <= 0)
        return -(long long)max;

    // Should pass due to the above couple of conditions.
    // Checking everything() to avoid overflow.
    HCSP_ASSERT( domain.everything() || domain.size() >= 2 ); 
    
    iset::iterator near_zero= domain.upper_bound(0);

    // This condition is impossible since otherwise it means that all values
    // are positive. Was ruled out by the "if (domain_min>=0)" condition.
    HCSP_ASSERT(near_zero != domain.begin());

    // This condition is impossible since otherwise it means that all values
    // are negative. This has been ruled out by the "if (max<=0)" condition.
    HCSP_ASSERT(near_zero != domain.end());
    
    long long positive= *near_zero;
    --near_zero;
    long long negative= *near_zero;
    HCSP_ASSERT(positive >= 0);
    HCSP_ASSERT(negative <= 0);
    return min(positive, -negative);
}

//virtual
constraint_propagator::result_type
greater_equal_abs_propagator::propagate( constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    const iset & in0= info.domain(0);
    const iset & in1= info.domain(1);

    if (in0.empty() || in1.empty())
        return UNSAT;
    
    long long smallest_abs= min_abs(in1);

    iset::value_type max0= in0.max();
    iset::value_type min0= in0.min();
    if (min0 < smallest_abs) {
        if (max0 < smallest_abs)
            return UNSAT; 
        iset out0= info.domain(0);
        out0.intersect(make_pair(smallest_abs, in0.MAX_DOMAIN));
        HCSP_ASSERT(!out0.empty());
        min0= out0.min();
        max0= out0.max();
        info.set_domain(0, move(out0));
    }

    if (in1.min() < -max0 || in1.max() > max0) {
        iset out1= info.domain(1);
        out1.intersect(make_pair(-max0, max0));
        HCSP_ASSERT(!out1.empty()); // handled during out0's update
        info.set_domain(1, move(out1));
    }
    return CONSISTENT;
}

//virtual
bool greater_equal_abs_propagator::conflicts(
    const constraint_runtime_info & info) const
{
    const iset & in0= info.domain(0);
    const iset & in1= info.domain(1);

    if (in0.empty() || in1.empty())
        return true;
    
    long long smallest_abs= min_abs(in1);
    return in0.max() < smallest_abs;
}

//virtual
void greater_equal_abs_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "";
        p_info->print_param(out, 0);
        out << ">= abs(";
        p_info->print_param(out, 1);
        out << ")";
    } else {
        out << "greater_equal_abs";
    }
}

//static
bool sum_wrap_propagator::propagate(constraint_runtime_info & info,
				    iset & limited_sum,
				    unsigned elements_to_sum)
{
    HCSP_ASSERT(elements_to_sum < info.arity());
    HCSP_ASSERT(elements_to_sum >= 2);
    
    static const iset unlimited=iset().complement();

    if (elements_to_sum == 2) {
        iset domain1= info.domain(1);
        iset domain2= info.domain(2);
        hcsp::propagation::plus(domain1, domain2, limited_sum);
        if (limited_sum.empty())
            return false;
        hcsp::propagation::minus(limited_sum, domain1, domain2);
        hcsp::propagation::minus(limited_sum, domain2, domain1);
        info.set_domain(1, move(domain1));
        info.set_domain(2, move(domain2));
        return true;
    }
    iset sum_tail= unlimited;
    
    hcsp::propagation::minus(limited_sum, info.domain(elements_to_sum),
                            sum_tail);
    
    if (!propagate(info, sum_tail, elements_to_sum-1))
        return false;
    iset domain= info.domain(elements_to_sum);
    hcsp::propagation::minus(limited_sum, sum_tail, domain);
    info.set_domain(elements_to_sum, move(domain));
    
    hcsp::propagation::plus(info.domain(elements_to_sum), sum_tail, 
                            limited_sum);
    return true;
}

//static
constraint_propagator::result_type
sum_wrap_propagator::propagate_zero( constraint_runtime_info & info )
{
    if ( !info.domain(0).contains(0))
        return UNSAT;
    else {
        info.set_domain(0, iset(0));
        return VALID;
    }
}

//static
constraint_propagator::result_type
sum_wrap_propagator::propagate_equality( constraint_runtime_info & info )
{
    const iset & in0= info.domain(0);
    const iset & in1= info.domain(1);
    iset out= in0;
    out.intersect(in1);
    if (out.empty())
        return UNSAT;
    info.set_domains({0,1}, move(out));
    return CONSISTENT;
}

//virtual
constraint_propagator::result_type
sum_wrap_propagator::propagate( constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    bool any_reduced = false;
    for (int i=info.arity()-1 ; i>=0; --i) {
        if (info.reduced_between_activations(i)) {
            any_reduced= true;
            break;
        }
    }
    if (!any_reduced)
        return CONSISTENT;
    
    const size_t arity= info.arity();
    switch (arity) {
      case 1: return propagate_zero(info);
      case 2: return propagate_equality(info);
      default: {
          iset sum= info.domain(0);
          if (! propagate(info, sum, arity-1) ) 
              return UNSAT;
          info.set_domain(0, move(sum));
      }
    }
    
    return CONSISTENT;
}

//virtual
bool sum_wrap_propagator::conflicts(const constraint_runtime_info & info) const
{
    HCSP_PROFILE_CURRENT;
    const size_t arity= info.arity();
    switch (arity) {
      case 1: return !info.domain(0).contains(0);
      case 2: return !info.domain(0).contains_any_of(info.domain(1));
      default: {
          iset sum= info.domain(0);
          sliced_runtime_info copy_info(info, 0, arity);
          return !propagate(copy_info, sum, arity-1);
      }
    }
    
    return CONSISTENT;
}

//virtual
void sum_wrap_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        p_info->print_param(out, 0);
        out << "== ";
        if (p_info->arity() == 1)
            out << "0";
        else {
            p_info->print_param(out, 1);
            for (unsigned i=2 ; i != p_info->arity() ; ++i) {
                out << " + ";
                p_info->print_param(out, i);
            }
        }
        out << ")";
    } else {
        out << "sum_wrap";
    }
}

constraint_propagator::result_type sum_truncate_propagator::propagate(
    constraint_runtime_info & info)
{
    iset a= info.domain(0), b= info.domain(1), c= info.domain(2);
    propagate(a, b, c);
    if (a.empty() || b.empty() || c.empty())
        return UNSAT;
    info.set_domain(0, move(a));
    info.set_domain(1, move(b));
    info.set_domain(2, move(c));
    
    return CONSISTENT;
}

bool sum_truncate_propagator::conflicts(const constraint_runtime_info & info)
    const
{
    iset a= info.domain(0), b= info.domain(1), c= info.domain(2);
    propagate(a, b, c);
    return a.empty() || b.empty() || c.empty();
}

void sum_truncate_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        p_info->print_param(out, 1);
        out << "+<trunc> ";
        p_info->print_param(out, 2);
        out << "==";
        p_info->print_param(out, 0);
        out << ")";
    } else {
        out << "sum_truncate";
    }
}

void sum_truncate_propagator::propagate(iset &a, iset & b, iset & c) const
{
    while (!a.empty() && !b.empty() && !c.empty()) {
        iset::size_type a_size= a.size(), b_size= b.size(), c_size= c.size();
        propagate_bounds(a, b, c);
        hcsp::propagation::plus(b, c, a);
        hcsp::propagation::minus(a, b, c);
        hcsp::propagation::minus(a, c, b);
        if (a.size() == a_size && b.size() == b_size && c.size() == c_size)
            return;
    }
}

void sum_truncate_propagator::propagate_bounds(iset &a, iset &b, iset &c) const
{
    pair<int64_t, int64_t> a_bound(c.min(), c.max());
    a_bound.first = max<int64_t>(b.min() + a_bound.first, a.min());
    a_bound.second = min<int64_t>(b.max() + a_bound.second, a.max());
    if (a_bound.first > a_bound.second) {
        c.clear(); b.clear(); a.clear();
        return;
    }
    a.intersect(a_bound);
    if (a.empty()) {
        c.clear(); b.clear(); a.clear();
        return;
    }
    pair<int64_t, int64_t> c_bound(a.min(), a.max());
    c_bound.first = max<int64_t>(c_bound.first - b.max(), c.min());
    c_bound.second = min<int64_t>(c_bound.second - b.min(), c.max());
    if (c_bound.first > c.max() && c_bound.second < c.min()) {
        c.clear(); b.clear(); a.clear();
        return;
    }
    c.intersect(c_bound);
    if (c.empty())
        return;
    
    pair<int64_t, int64_t> b_bound(a.min(), a.max());
    b_bound.first = max<int64_t>(b_bound.first - c.max(), b.min());
    b_bound.second = min<int64_t>(b_bound.second - c.min(), b.max());
    if (b_bound.first > b.max() && b_bound.second < b.min()) {
        c.clear(); b.clear(); a.clear();
        return;
    }
    b.intersect(b_bound);
}

// virtual
constraint_propagator::result_type sum_mod_propagator::propagate(
    constraint_runtime_info & info)
{
    if (!restrict_to_mod(info))
        return UNSAT;
    unsigned free_param= -1U;
    long long offset= m_result;
    
    for (unsigned i=0 ; i != info.arity() ; ++i) {
        const iset & domain = info.domain(i);
        if (domain.singleton())
            offset += m_mod - domain.min(); // add m_mod to keep offset positive
        else if (free_param == -1U)
            free_param= i;
        else
            // More than one unassigned parameter.  We assume that at most one
            // parameters may be reduced at a given time (an approximation).
            return CONSISTENT;
    }
    HCSP_ASSERT(offset >= 0);
    offset %= m_mod;
    if (free_param == -1U)
        return offset == 0 ? VALID : UNSAT;
    const iset & domain = info.domain(free_param);
    if (!domain.contains(offset))
        return UNSAT;
    if (domain.singleton())
        return VALID;
    info.set_domain(free_param, iset(offset));
    return VALID;
}

bool sum_mod_propagator::restrict_to_mod(constraint_runtime_info & info) const
{
    for (unsigned i=0 ; i != info.arity() ; ++i) {
        const iset & domain = info.domain(i);
        if (domain.empty())
            return false;
        if (domain.min() < 0 || domain.max() >= m_mod) {
            iset out_domain= info.domain(i);
            out_domain.intersect(make_pair(0, m_mod-1));
            if (out_domain.empty())
                return false;
            info.set_domain(i, move(out_domain));
        }
    }
    return true;
}
// virtual
bool sum_mod_propagator::conflicts(const constraint_runtime_info & info) const
{
    unsigned free_param= -1U;
    long long offset= m_result;
    unsigned count_free_params = 0;
    for (unsigned i=0 ; i != info.arity() ; ++i) {
        const iset & domain = info.domain(i);
        auto it = domain.lower_bound(0);
        if (it == domain.end() || *it >= m_mod)
            return true;
        auto next= it;
        ++next;
        if (next == domain.end() || *next >= m_mod)
            offset += m_mod - *it; // adding m_mod keeps offset positive
        else {
            free_param= i;
            ++count_free_params;
        }
    }
    if (count_free_params > 1)
        // More than one unassigned parameter.  We assume that at most one
        // parameters may be reduced at a given time (an approximation).
        return false;

    HCSP_ASSERT(offset >= 0);
    offset %= m_mod;

    if (free_param == -1U)
        return offset != 0;
    const iset & domain = info.domain(free_param);
    return !domain.contains(offset);
}

//virtual
void sum_mod_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        if (m_mod == 2) {
            out << "(xor(";
            for (unsigned i=0; i != p_info->arity(); ++i) {
                out << (i == 0 ? "" : ", ");
                p_info->print_param(out, i);
            }
            out << ") ";
        } else {
            out << "(";
            for (unsigned i=0; i != p_info->arity(); ++i) {
                out << (i == 0 ? "" : "+");
                p_info->print_param(out, i);
            }
            out << ") mod " << m_mod;
        }
        out << " == " << m_result << ") && (";
        for (unsigned i=0; i != p_info->arity(); ++i) {
            out << (i == 0 ? "" : ", ");
            p_info->print_param(out, i);
        }
        out << " in [0.." << (m_mod-1) << "]))";
    } else {
        if (m_mod == 2)
            out << "((xor vars) " << m_mod << " == " << m_result << ")";
        else
            out << "((sum vars) mod " << m_mod << " == " << m_result << ")";
    }
}

//virtual
constraint_propagator * sum_mod_propagator::intersect_with(
    const constraint_runtime_info & info,
    vector<unsigned> & copy_from_argument) const
{
    copy_from_argument.clear();
    long long result = m_result;
    for (unsigned i=0 ; i != info.arity(); ++i) {
        const iset & domain = info.domain(i);
        bool skip = false;
        if (domain.singleton()) {
            iset::value_type value = domain.min();
            if (0 <= value || value < m_mod) {
                skip = true;
                result += value;
            }
        }
        if (!skip)
            copy_from_argument.push_back(i);
    }
    result %= m_mod;
    if (result < 0)
        result += m_mod;
    if (copy_from_argument.size() == info.arity())
        return nullptr;
    else
        return new sum_mod_propagator(m_mod, result);
}

// virtual
bool sum_mod_propagator::may_combine(const constraint_propagator & other) const
{
    const auto * other_sum = dynamic_cast<const sum_mod_propagator*>(&other);
    if (other_sum)
        return other_sum->m_mod == m_mod;
    if (dynamic_cast<const equal_propagator*>(&other))
        return true;
    if (m_mod == m_result + 1 && 
        dynamic_cast<const linear_combination_ge_propagator*>(&other))
        return true;
    else
        return false;
}

// virtual
constraint_propagator *sum_mod_propagator::combine(
    const constraint_runtime_info & info,
    const constraint_propagator & other,
    const constraint_runtime_info & other_info,
    const vector<unsigned> & other_to_common_param,
    unsigned other_pivot_param,
    vector<unsigned> & copy_from_argument) const
{
    HCSP_ASSERT(may_combine(other));
    const auto * other_sum = dynamic_cast<const sum_mod_propagator*>(&other);
    HCSP_ASSERT(other_info.arity() == other_to_common_param.size());
    (void)other_info;
    if (other_sum)
        return combine(info, *other_sum, other_to_common_param,
                       copy_from_argument);
    if (dynamic_cast<const equal_propagator*>(&other))
        return combine_with_eq(info, other_to_common_param, copy_from_argument);
    if (dynamic_cast<const linear_combination_ge_propagator*>(&other))
        return combine_other_first(info, other, other_info, 
                                   other_to_common_param, other_pivot_param,
                                   copy_from_argument);
    
    return nullptr;
}

constraint_propagator *sum_mod_propagator::combine(
    const constraint_runtime_info & info,
    const sum_mod_propagator & other,
    const vector<unsigned> & other_to_common_param,
    vector<unsigned> & copy_from_argument) const
{
    constexpr unsigned int DELETED = -1U;
    if (other.m_mod != m_mod)
        return nullptr;
    for (unsigned i=0 ; i < info.arity() ; ++i)
        copy_from_argument.push_back(i);
    for (unsigned other_arg : other_to_common_param) {
        if (other_arg < info.arity())
            copy_from_argument[other_arg] = DELETED;
        else
            copy_from_argument.push_back(other_arg);
    }
    int64_t result = static_cast<int64_t>(m_result) + other.m_result;
    result %= m_mod;
    HCSP_ASSERT(result >= 0);

    copy_from_argument.erase(
        remove_if(copy_from_argument.begin(), copy_from_argument.end(),
                  [=](unsigned arg) { return arg == DELETED; }),
        copy_from_argument.end());
    if (copy_from_argument.empty() && result != 0) {
        return new clause_propagator{};
    }
    return new sum_mod_propagator(m_mod, result);
}

constraint_propagator *sum_mod_propagator::combine_with_eq(
    const constraint_runtime_info & info,
    const vector<unsigned> & other_to_common_param,
    vector<unsigned> & copy_from_argument) const
{
    HCSP_ASSERT(other_to_common_param.size() == 2);
    if ( other_to_common_param[0] < info.arity() &&
         other_to_common_param[1] < info.arity()) {
        if (m_mod != 2)
            return nullptr;
        for (unsigned i=0 ; i != info.arity(); ++i) {
            if (i != other_to_common_param[0] && i != other_to_common_param[1])
                copy_from_argument.push_back(i);
        }
    } else {
        for (unsigned i=0 ; i != info.arity(); ++i) {
            if (i == other_to_common_param[0])
                copy_from_argument.push_back(info.arity()+1);
            else if (i == other_to_common_param[1])
                copy_from_argument.push_back(info.arity());
            else
                copy_from_argument.push_back(i);
        }
    }
    return new sum_mod_propagator(m_mod, m_result);
}


//virtual
bool less_equal_sum_propagator::init( constraint_runtime_info & info )
{
    HCSP_ASSERT(count_params() < 0);
    info.allocate_watch(0, LOWER_BOUND_WATCH);
    for (unsigned i=1; i < info.arity()  ; ++i)
        info.allocate_watch(i, UPPER_BOUND_WATCH);
    return true;
}

// virtual
constraint_propagator * less_equal_sum_propagator::intersect_with(
    const constraint_runtime_info & info,
    std::vector<unsigned> & copy_from_argument) const
{
    unsigned arity= info.arity();
    for (unsigned i=0 ; i != arity ; ++i) {
        if (!info.domain(i).singleton())
            continue;
        // x0 <= x1 + x2 + x3 + ...
        // becomes
        // -x0 + x1 + x2 + x3 + ... >= 0
        vector<linear_combination_ge_propagator::coefficient_type> coefficients;
        coefficients.reserve(arity);
        coefficients.push_back(-1);
        coefficients.resize(arity, 1);
        linear_combination_ge_propagator combination(0, move(coefficients));
        return combination.intersect_with(info, copy_from_argument);
    }
    return NULL;
}

//virtual
constraint_propagator::result_type
less_equal_sum_propagator::propagate( constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;

    const unsigned arity= info.arity();
    if ( arity > 10 && info.originating_queue() == info.VARIABLE_QUEUE) {
        info.put_in_queue(info.CONSTRAINT_QUEUE);
        return POSTPONE;
    }
    long long max_sum= 0;
    if (info.domain(0).empty())
        return UNSAT;
    for (unsigned i=1 ; i < arity ; ++i)
        max_sum += info.domain(i).max();

    if (info.domain(0).max() > max_sum) {
        if (max_sum < info.domain(0).min())
            return UNSAT;
        iset out0 = info.domain(0);
        out0.intersect(make_pair(iset::MIN_DOMAIN, max_sum));
        if (out0.empty())
            return UNSAT;
        info.set_domain(0, move(out0));
    }

    iset::value_type min_sum= info.domain(0).min();
    for (unsigned i=1 ; i < arity ; ++i) {
        long long max_sum_other= max_sum- info.domain(i).max();
        long long min_value= min_sum-max_sum_other;
        if (info.domain(i).min() < min_value)
        {
            iset out_i = info.domain(i);
            out_i.intersect(make_pair(min_value, iset::MAX_DOMAIN));
            info.set_domain(i, move(out_i));
        }
    }
    
    return CONSISTENT;
}

//virtual
bool less_equal_sum_propagator::conflicts(const constraint_runtime_info & info)
    const
{
    const unsigned arity= info.arity();
    long long max_sum= 0;
    if (info.domain(0).empty())
        return true;
    for (unsigned i=1 ; i < arity ; ++i)
        max_sum += info.domain(i).max();

    return info.domain(0).min() > max_sum;
}

//virtual
void less_equal_sum_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        p_info->print_param(out, 0);
        out << "<= ";
        if (p_info->arity() == 1)
            out << "0";
        else {
            p_info->print_param(out, 1);
            for (unsigned i=2 ; i != p_info->arity() ; ++i) {
                out << " + ";
                p_info->print_param(out, i);
            }
        }
        out << ")";
    } else {
        out << "le_sum";
    }
}

//virtual
bool greater_equal_sum_propagator::init( constraint_runtime_info & info )
{
    HCSP_ASSERT(count_params() < 0);
    info.allocate_watch(0, UPPER_BOUND_WATCH);
    for (unsigned i=1; i < info.arity()  ; ++i)
        info.allocate_watch(i, LOWER_BOUND_WATCH);
    return true;
}

// virtual
constraint_propagator * greater_equal_sum_propagator::intersect_with(
    const constraint_runtime_info & info,
    std::vector<unsigned> & copy_from_argument) const
{
    unsigned arity= info.arity();
    for (unsigned i=0 ; i != arity ; ++i) {
        if (!info.domain(i).singleton())
            continue;
        // x0 >= x1 + x2 + x3 + ...
        // becomes
        // x0 - x1 - x2 - x3 - ... >= 0
        vector<linear_combination_ge_propagator::coefficient_type> coefficients;
        coefficients.reserve(arity);
        coefficients.push_back(1);
        coefficients.resize(arity, -1);
        
        linear_combination_ge_propagator combination(0, coefficients);
        return combination.intersect_with(info, copy_from_argument);
    }
    return NULL;
}

//virtual
constraint_propagator::result_type
greater_equal_sum_propagator::propagate( constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;

    const unsigned arity= info.arity();
    if ( arity > 10 && info.originating_queue() == info.VARIABLE_QUEUE) {
        info.put_in_queue(info.CONSTRAINT_QUEUE);
        return POSTPONE;
    }
    long long min_sum= 0;
    if (info.domain(0).empty())
        return UNSAT;
    for (unsigned i=1 ; i < arity ; ++i) {
        const iset & domain= info.domain(i);
        if (domain.empty())
            return UNSAT;
        min_sum += domain.min();
    }

    if (info.domain(0).min() < min_sum) {
        if (min_sum > info.domain(0).max())
            return UNSAT;
        iset out0 = info.domain(0);
        out0.intersect(make_pair(min_sum, iset::MAX_DOMAIN));
        if (out0.empty())
            return UNSAT;
        info.set_domain(0, move(out0));
    }
    
    iset::value_type max_sum= info.domain(0).max();
    for (unsigned i=1 ; i < arity ; ++i) {
        long long min_sum_other= min_sum- info.domain(i).min();
        long long max_value= max_sum-min_sum_other;
        if (info.domain(i).max() > max_value)
        {
            iset out_i = info.domain(i);
            out_i.intersect(make_pair(iset::MIN_DOMAIN, max_value));
            if (out_i.empty())
                return UNSAT;
            info.set_domain(i, move(out_i));
        }
    }
    
    return CONSISTENT;
}

//virtual
bool greater_equal_sum_propagator::conflicts(
    const constraint_runtime_info & info) const
{
    const unsigned arity= info.arity();
    long long min_sum= 0;
    if (info.domain(0).empty())
        return true;
    for (unsigned i=1 ; i < arity ; ++i) {
        const iset & domain= info.domain(i);
        if (domain.empty())
            return true;
        min_sum += domain.min();
    }

    if (info.domain(0).min() < min_sum) {
        if (min_sum > info.domain(0).max())
            return true;
        iset out_domain= info.domain(0);
        out_domain.intersect(make_pair(min_sum, iset::MAX_DOMAIN));
        return out_domain.empty();
    }
    return false;
}

//virtual
void greater_equal_sum_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        p_info->print_param(out, 0);
        out << ">= ";
        if (p_info->arity() == 1)
            out << "0";
        else {
            p_info->print_param(out, 1);
            for (unsigned i=2 ; i != p_info->arity() ; ++i) {
                out << " + ";
                p_info->print_param(out, i);
            }
        }
        out << ")";
    } else {
        out << "ge_sum";
    }
}


void linear_combination_ge_propagator::normalize_coefficients(
    unsigned next_arity, vector<coefficient_type> & new_coefficients,
    int64_t & new_k, vector<unsigned> & copy_from_argument) const
{
    if (next_arity == 0)
        return;
    const vector<coefficient_type> *p_coefficients;
    if (new_coefficients.empty())
        p_coefficients= &m_coefficients;
    else
        p_coefficients= &new_coefficients;

    HCSP_ASSERT(next_arity == p_coefficients->size());
    int64_t gcd_value= (*p_coefficients)[0];
    if (gcd_value < 0)
        gcd_value *= -1;
    for (unsigned i=1; i < next_arity && gcd_value > 1 ; ++i) {
        int64_t value= (*p_coefficients)[i];
        if (value < 0)
            value *= -1;
        gcd_value= gcd(gcd_value, value);
    }
    if (gcd_value == 1)
        return;
    if (new_coefficients.empty())
        new_coefficients= m_coefficients;
    for (auto & coeff: new_coefficients)
        coeff /= gcd_value;
    new_k= ceil_div(new_k, gcd_value);
    if (copy_from_argument.empty()) {
        for (unsigned i=0 ; i != next_arity; ++i)
            copy_from_argument.push_back(i);
    }
}

constraint_propagator *linear_combination_ge_propagator::new_empty_sum(
    int64_t new_k, vector<unsigned> & copy_from_argument) const
{
    if (new_k>0)
        // UNSAT since empty_sum >= new_k > 0 means 0 > 0.
        return new clause_propagator{}; 
    else {
        // VALID since empty_sum >= new_k for new_k <= is always true
        copy_from_argument.assign({0});
        return new clause_propagator{iset().complement()};
    }
}

constraint_propagator *linear_combination_ge_propagator::new_unary_sum(
    vector<coefficient_type> & new_coefficients, int64_t new_k,
    vector<unsigned> &copy_from_argument) const
{
    if (new_coefficients.empty()) {
        HCSP_ASSERT(copy_from_argument.empty());
        copy_from_argument.push_back(0);
        new_coefficients= m_coefficients;
    }
    HCSP_ASSERT(new_coefficients.size() == copy_from_argument.size());
    HCSP_ASSERT(new_coefficients.size() == 1);
    // The coefficient is normalized:
    HCSP_ASSERT(new_coefficients[0] == 1 || new_coefficients[0] == -1);

    if (new_coefficients[0] < 0) {
        HCSP_ASSERT(new_coefficients[0] == -1); // It's normalized
        new_k *= -1;
        // x0 <= new_k
        if (new_k < iset::MIN_DOMAIN)
            return new clause_propagator{iset()}; // UNSAT
        else if (new_k >= iset::MAX_DOMAIN)
            return new clause_propagator{iset().complement()}; // VALID
        else
            return new clause_propagator{iset(make_pair(iset::MIN_DOMAIN,
                                                        new_k))};
    } else {
        HCSP_ASSERT(new_coefficients[0] == 1); // It's normalized
        // x0 >= new_k
        if (new_k <= iset::MIN_DOMAIN)
            return new clause_propagator{iset().complement()}; // VALID
        else if (new_k > iset::MAX_DOMAIN)
            return new clause_propagator{iset()}; // UNSAT
        else
            return new clause_propagator{iset(make_pair(new_k,
                                                        iset::MAX_DOMAIN))};
    }
}

constraint_propagator *linear_combination_ge_propagator::new_binary_sum(
    vector<coefficient_type> & new_coefficients, int64_t new_k,
    vector<unsigned> &copy_from_argument) const
{
    if ( new_k < numeric_limits<iset::value_type>::min() ||
         new_k > numeric_limits<iset::value_type>::max())
        return nullptr;
    if (new_coefficients[0] == 1 && new_coefficients[1] == -1)
        return new difference_ge_propagator(new_k);
    if (new_coefficients[0] == -1 && new_coefficients[1] == 1) {
        if (copy_from_argument.empty())
            copy_from_argument.assign({1,0});
        else
            swap(copy_from_argument[0], copy_from_argument[1]);
        return new difference_ge_propagator(new_k);
    }
    return new_nary_sum(new_coefficients, new_k, copy_from_argument);
}

constraint_propagator *linear_combination_ge_propagator::new_nary_sum(
    vector<coefficient_type> & new_coefficients, int64_t new_k,
    vector<unsigned> &copy_from_argument) const
{
    if (copy_from_argument.empty())
        return nullptr;
    if ( new_k < numeric_limits<iset::value_type>::min() ||
         new_k > numeric_limits<iset::value_type>::max())
        return nullptr;
    HCSP_ASSERT(new_coefficients.size() == copy_from_argument.size());
    return new linear_combination_ge_propagator(new_k, new_coefficients);
}

unsigned linear_combination_ge_propagator::compact_coefficients(
    const constraint_runtime_info & info,
    vector<coefficient_type> & new_coefficients,
    int64_t & new_k,
    vector<unsigned> & copy_from_argument) const
{
    new_coefficients.clear();
    unsigned arity= info.arity();
    bool simplified= false;
    int128 next_bound(new_k);
    for (unsigned i=0; i != arity; ++i) {
        const iset & domain = info.domain(i);
        if (domain.singleton() || m_coefficients[i] == 0) {
            if (!simplified) {
                for (unsigned j=0 ; j < i ; ++j) {
                    new_coefficients.push_back(m_coefficients[j]);
                    copy_from_argument.push_back(j);
                }
                simplified= true;
            }
            next_bound -= int64_t(m_coefficients[i])*domain.min();
        } else if (simplified) {
            new_coefficients.push_back(m_coefficients[i]);
            copy_from_argument.push_back(i);
        }
    }
    if (next_bound.fits_int64()) {
        new_k= next_bound.as_int64();
    } else {
        // give up
        new_coefficients.clear();
        copy_from_argument.clear();
        simplified= false;
    }
    unsigned next_arity= simplified ? copy_from_argument.size() : arity;
    normalize_coefficients(next_arity, new_coefficients, new_k,
                           copy_from_argument);
    return next_arity;
}

// virtual
constraint_propagator *linear_combination_ge_propagator::intersect_with(
    const constraint_runtime_info & info,
    vector<unsigned> & copy_from_argument) const
{
    int64_t new_k= m_k;
    vector<coefficient_type> new_coefficients;
    unsigned next_arity= compact_coefficients(info, new_coefficients, new_k,
                                              copy_from_argument);
    if (new_coefficients.empty() && next_arity > 0)
        return nullptr;
    switch (next_arity) {
      case 0: return new_empty_sum(new_k, copy_from_argument);
      case 1: return new_unary_sum(new_coefficients, new_k, copy_from_argument);
      case 2: return new_binary_sum(new_coefficients, new_k,
                                    copy_from_argument);
      default:
        return new_nary_sum(new_coefficients, new_k, copy_from_argument);
    }
}

template <typename SlackType>
void linear_combination_ge_propagator::propagate_to_argument(
    constraint_runtime_info &info, unsigned argument, const SlackType &slack)
    const
{
    const iset & domain= info.domain(argument);
    if (domain.singleton())
        return;
    iset::value_type coefficient= m_coefficients[argument];
    if (coefficient > 0) {
        // subtract the contribution of this argument to the value of slack
        SlackType factor_slack= slack-int64_t(coefficient)*domain.max();
        int64_t element_worse= int64_t(coefficient)*domain.min();
        if (factor_slack+element_worse >= 0)  // all values are in range?
            return;
        SlackType min_allowable= -floor_div(factor_slack, coefficient);
        if (min_allowable <= iset::MIN_DOMAIN)
            return;
        HCSP_ASSERT(min_allowable <= iset::MAX_DOMAIN);
        iset next_domain= info.domain(argument);
        next_domain.intersect(make_pair(int64_t(min_allowable), domain.max()));
        info.set_domain(argument, move(next_domain));
    } else if (coefficient < 0) {
        // subtract the contribution of this argument to the value of slack
        SlackType factor_slack= slack-int64_t(coefficient)*domain.min();
        int64_t element_worse= int64_t(coefficient)*domain.max();
        if (factor_slack+element_worse >= 0)
            return;
        SlackType max_allowable= -ceil_div(factor_slack, coefficient);
        if (max_allowable >= iset::MAX_DOMAIN)
            return;
        HCSP_ASSERT(max_allowable >= iset::MIN_DOMAIN);
        
        iset next_domain= info.domain(argument);
        next_domain.intersect(make_pair(domain.min(), int64_t(max_allowable)));
        info.set_domain(argument, move(next_domain));
    }
    // No need to watch m_coefficients[i] == 0, as such elements have no
    // effect. 
}

// virtual
template <typename SlackType>
constraint_propagator::result_type
linear_combination_ge_propagator::const_propagate(
    constraint_runtime_info & info) const
{
    SlackType slack(-int64_t(m_k));
    for (int i=0; i < count_params()  ; ++i) {
        if (m_coefficients[i] > 0)
            slack += int64_t(m_coefficients[i])*info.domain(i).max();
        else if (m_coefficients[i] < 0)
            slack += int64_t(m_coefficients[i])*info.domain(i).min();
        // No need to watch m_coefficients[i] == 0, as such elements have no
        // effect. 
    }
    if (slack < 0)
        return UNSAT;
    for (int i=0; i < count_params()  ; ++i)
        propagate_to_argument(info, i, slack);
    return CONSISTENT;
}

// static
inline iset linear_combination_ge_propagator::make_range(int64_t first,
                                                         int64_t second)
{
    if (first > second)
        return iset();
    else
        return iset(iset::value_range_type(
                        max<int64_t>(first, iset::MIN_DOMAIN),
                        min<int64_t>(second, iset::MAX_DOMAIN)));
}

// virtual
bool linear_combination_ge_propagator::conflicts(
    const constraint_runtime_info & info) const
{
    HCSP_ASSERT((int)info.arity() == count_params());
    int128 max_sum;
    for (int i=0 ; i < count_params() ; ++i)
        if (m_coefficients[i] >= 0)
            max_sum += int64_t(m_coefficients[i])*info.domain(i).max();
        else
            max_sum += int64_t(m_coefficients[i])*info.domain(i).min();
    return max_sum < m_k;
}

bool linear_combination_ge_propagator::constraint_check(
    const constraint_runtime_info & info) const
{
    HCSP_ASSERT(info.arity() == 0U+count_params());
    int128 sum;
    for (int i=0 ; i < count_params() ; ++i)
        sum += 1LL * m_coefficients[i]*info.domain(i).min();
    return !(sum < m_k);
}


/**
 * Although the default tight_constraint_propagator::explain_propagation() is
 * good enough its O(N^2) complexity makes it too slow. To make things faster,
 * we have this faster, O(N) explanation generation.
 *
 * The basic concept of this is as follows. For
 *    a0*x0 + a1*x1 + a2*x2 + .... >= k
 *
 * Without loss of generality assume that we have to explain a propagation
 * towards x0.
 *
 * If D(x0) was modified then:
 *    a0*x0  >= k - (a1*x1 + a2*x2 + a3*x3 + ...)
 *
 *  Which means that x0 is reduced according to:
 *    a0*x0  >= min( k - (a1*x1 + a2*x2 + a3*x3 + ...) )
 *
 *  This is the same as:
 *    a0*x0  >= k - max (a1*x1 + a2*x2 + a3*x3 + ...)
 *
 *  Which gives:
 *    a0*x0  >= k - (max(a1*x1) + max(a2*x2) + max(a3*x3) + ...)
 *
 *  This constraint sets the lower bound of a0*x0. If the domain of x1 is
 *  modified such that max(a1*x1) is enlarged then x1 will have a weaker
 *  propagation. If a1 > 0 then max(a1*x1) == a1*max(D(x1)), in which case
 *  the literal of the explanation is (x1 > max(D(x1))). If a1 < 0 then the
 *  literal of x1 becomes (x1 < min(D(x1))). Same for x2,x3,... .
 *
 *  The literal of x0 depends on the value of
 *      slack= k - (max(a1*x1) + max(a2*x2) + max(a3*x3) + ...)
 *  If a0>0 then x0 >= slack/a0. Since x0 is integer an Q/a0 is real then the
 *  literal becomes x0 >= ceil(slack/a0). Similarly, if x0 < 0 then
 *  x0 <= slack/a0 which, when accounting for reals, becomes
 *  x0 <= floor(slack/a0).
 */
// virtual
void linear_combination_ge_propagator::explain_propagation(
    unsigned propagated_param, const iset & /*values_to_explain*/,
    const iset & /*propagated_domain*/, 
    const constraint_runtime_info & info_before_propagation,
    std::vector<iset> & explanation) const
{
    explanation.resize(m_coefficients.size());
    if (propagated_param >= m_coefficients.size())
        propagated_param= 0;
    int128 slack(m_k);
    for (unsigned i=0 ; i != m_coefficients.size(); ++i) {
        const iset & domain = info_before_propagation.domain(i);
        if (i == propagated_param)
            continue;
        explanation[i]= iset().complement();
        if (m_coefficients[i] > 0) {
            explanation[i].erase(make_pair(iset::MIN_DOMAIN, domain.max()));
            slack -= int64_t(m_coefficients[i])*domain.max();
        } else if (m_coefficients[i] < 0) {
            explanation[i].erase(make_pair(domain.min(), iset::MAX_DOMAIN));
            slack -= int64_t(m_coefficients[i])*domain.min();
        } else {
            explanation[i].clear();
        }
    }
    iset & literal= explanation[propagated_param];
    if (m_coefficients[propagated_param] > 0) {
        int128 bound= slack.ceil_div(m_coefficients[propagated_param]);
        if (bound > iset::MAX_DOMAIN)
            literal.clear();
        else if (bound >= iset::MIN_DOMAIN)
            literal= make_range(bound.as_int64(), iset::MAX_DOMAIN);
    } else if (m_coefficients[propagated_param] < 0) {
        int128 bound= slack.floor_div(m_coefficients[propagated_param]);
        if (bound < iset::MIN_DOMAIN)
            literal.clear();
        else if (bound <= iset::MAX_DOMAIN)
            literal= make_range(iset::MIN_DOMAIN, bound.as_int64());
    }
}

// virtual
bool linear_combination_ge_propagator::init(constraint_runtime_info & info)
{
    HCSP_ASSERT(count_params() == (int)info.arity());
    m_small_coefficients= true;
    int32_t max_coefficient= numeric_limits<int32_t>::max()/count_params();
    int32_t min_coefficient= numeric_limits<int32_t>::min()/count_params();
    for (int i=0; i < count_params()  ; ++i) {
        if (m_coefficients[i] > 0)
            info.allocate_watch(i, UPPER_BOUND_WATCH);
        else if (m_coefficients[i] < 0)
            info.allocate_watch(i, LOWER_BOUND_WATCH);
        // No need to watch m_coefficients[i] == 0, as such elements have no
        // effect. 
        if ( m_coefficients[i] < min_coefficient ||
             max_coefficient < m_coefficients[i]) {
            m_small_coefficients= false;
        }
    }
    return true;
}

// virtual
bool linear_combination_ge_propagator::get_disjunction_specification(
    disjunction_specification & disjunction) const
{
    disjunction.refer_to_constraint(*this, 0);
    return true;
}

// virtual
bool linear_combination_ge_propagator::disjoin_constraint(
    const disjunction_specification & this_constraint,
    const disjunction_specification & other_constraint,
    disjunction_specification & result_constraint) const
{
    if (!other_constraint.constraint())
        return false;

    //!@todo Avoiding all but linear_combination_ge_propagator seems restrictive
    if (!dynamic_cast<const linear_combination_ge_propagator*>(
            other_constraint.constraint()))
        return false;

    const linear_combination_ge_propagator & other=
        *static_cast<const linear_combination_ge_propagator*>(
            other_constraint.constraint());

    if (other.coefficients().size() != coefficients().size())
        return false;
    vector<coefficient_type> var_to_coefficient(
        max(m_coefficients.size(), other.coefficients().size()), 0);
    for (unsigned i= 0; i != this_constraint.constraint_arity(); ++i) {
        unsigned var= this_constraint.constraint_var(i);
        if (var >= var_to_coefficient.size())
            var_to_coefficient.resize(var+1, 0);
        var_to_coefficient[var]= m_coefficients[i];
    }
    for (unsigned i= 0; i != other_constraint.constraint_arity(); ++i) {
        unsigned var= other_constraint.constraint_var(i);
        if ( var >= var_to_coefficient.size() ||
             var_to_coefficient[var] != m_coefficients[i])
            return false;
    }

    result_constraint.reset_constraint(
        new linear_combination_ge_propagator(min(m_k, other.m_k),
                                             coefficients()),
        result_constraint.num_literals());
    return true;
}

// static
bool linear_combination_ge_propagator::accumulate_coefficients(
    vector<int64_t> & out_coefficients, int64_t & new_k,
    const disjunction_specification & constraint_spec, uint32_t factor)
{
    const difference_ge_propagator *diff_ge=
        dynamic_cast<const difference_ge_propagator*>(
            constraint_spec.constraint());
    int64_t this_k;
    if (diff_ge) {
        this_k= int64_t(diff_ge->get_min_difference()) * factor;
        unsigned external_arg0 = constraint_spec.constraint_var(0);
        unsigned external_arg1 = constraint_spec.constraint_var(1);
        if (max(external_arg0, external_arg1) >= out_coefficients.size())
            out_coefficients.resize(max(external_arg0,external_arg1)+1, 0);
        out_coefficients[external_arg0] += factor;
        out_coefficients[external_arg1] -= factor;
    } else {
        const linear_combination_ge_propagator & linear=
            dynamic_cast<const linear_combination_ge_propagator&>(
                *constraint_spec.constraint());
        this_k = static_cast<int64_t>(linear.m_k) * factor;
        const vector<coefficient_type> & in_coefficients= linear.coefficients();
        for (unsigned i=0 ; i != in_coefficients.size() ; ++i) {
            unsigned external_arg = constraint_spec.constraint_var(i);
            if (external_arg >= out_coefficients.size())
                out_coefficients.resize(external_arg+1, 0);
            out_coefficients[external_arg]+= int64_t(in_coefficients[i])*factor;
        }
    }
    // To avoid strict overflow, use unsigned addition and not signed.
    int64_t tested_k = static_cast<uint64_t>(new_k + this_k);
    if ((new_k >= 0) == (this_k >= 0)) {
        // input same sign, so the result should be same sign
        if ( (new_k >= 0) != (tested_k >= 0))
            return false; // different sign -> due to overflow
    }
    new_k= tested_k;
    return true;
}

// static
bool
linear_combination_ge_propagator::set_constraint_from_positional_coefficients(
    const vector<int64_t> & out_coefficients,
    int64_t new_k, disjunction_specification & result_constraint)
{
    if ( new_k < numeric_limits<iset::value_type>::min() ||
         new_k > numeric_limits<iset::value_type>::max())
        return false;
    auto constraint= new linear_combination_ge_propagator(new_k);
    result_constraint.reset_constraint(constraint,
                                       result_constraint.num_literals());
    for (unsigned i=0 ; i != out_coefficients.size() ; ++i) {
        if (out_coefficients[i] == 0)
            continue;
        if (out_coefficients[i] < numeric_limits<coefficient_type>::min())
            return false;
        if (out_coefficients[i] > numeric_limits<coefficient_type>::max())
            return false;
        constraint->m_coefficients.push_back(out_coefficients[i]);
        result_constraint.add_constraint_var(i);
    }
    if (result_constraint.constraint_arity() == 0) {
        if (new_k > 0) {
            // UNSAT: 0 >= new_k > 0
            result_constraint.reset_constraint(
                new clause_propagator, result_constraint.num_literals());
            return true;
        } else  {
            // We don't support sum >= k1 and -sum >= k2 where k1+k2 <= 0,
            // i.e., -k2 >= sum >= k1.
            return false;
        }
    }
    
    return true;
}

// static
iset::value_type linear_combination_ge_propagator::find_var_coefficient(
    unsigned var, const disjunction_specification & constraint_spec)
{
    unsigned arg= constraint_spec.constraint_index_of_var(var);
    const difference_ge_propagator *diff_ge=
        dynamic_cast<const difference_ge_propagator*>(
            constraint_spec.constraint());
    if (diff_ge)
        return (arg == 0) ? 1 : -1;
    const linear_combination_ge_propagator &linear=
        dynamic_cast<const linear_combination_ge_propagator&>(
            *constraint_spec.constraint());
    return linear.coefficients()[arg];
}

// virtual
bool linear_combination_ge_propagator::combine_base_constraints(
    disjunction_specification & this_constraint,
    disjunction_specification & other_constraint,
    unsigned pivot,
    disjunction_specification & result_constraint) const
{
    HCSP_ASSERT(this_constraint.constraint() == this);
    auto this_pivot_coefficient= find_var_coefficient(pivot, this_constraint);
    auto other_pivot_coefficient= find_var_coefficient(pivot, other_constraint);
    if ((this_pivot_coefficient < 0) == (other_pivot_coefficient < 0))
        return false; // can't eliminate same sign pivot
    uint32_t this_abs= this_pivot_coefficient;
    static_assert(sizeof(this_abs) == sizeof(this_pivot_coefficient),
                  "Mismatch type width or the signed and the abs values");
    if (this_pivot_coefficient < 0)
        this_abs = -this_abs;
    uint32_t other_abs= other_pivot_coefficient;
    static_assert(sizeof(other_abs) == sizeof(other_pivot_coefficient),
                  "Mismatch type width or the signed and the abs values");
    if (other_pivot_coefficient < 0)
        other_abs = -other_abs;
    
    uint32_t shared= gcd(this_abs, other_abs);
    uint32_t this_factor= other_abs/shared;
    uint32_t other_factor= this_abs/shared;
    
    int64_t new_k = 0;
    vector<int64_t> out_coefficients;
    if (!accumulate_coefficients(out_coefficients, new_k,
                                 this_constraint, this_factor))
        return false;
    if (!accumulate_coefficients(out_coefficients, new_k,
                                 other_constraint, other_factor))
        return false;
    return set_constraint_from_positional_coefficients(out_coefficients, new_k,
                                                       result_constraint);
}

// virtual
constraint_propagator *linear_combination_ge_propagator::new_propagator(
        disjunction_specification & disjunction,
        std::vector<unsigned> &copy_from_argument) const
{
    HCSP_ASSERT(disjunction.constraint() == this);
    for (unsigned i=0 ; i != disjunction.constraint_arity() ; ++i)
        copy_from_argument.push_back(disjunction.constraint_var(i));
    if (disjunction.num_literals() == 0) {
        return disjunction.release();
    }
    typedef disjoint_propagator<linear_combination_ge_propagator,
                                clause_propagation_algo> result_type;
    unique_ptr<result_type> result(new result_type(*this));
    vector<iset> & literals= result->algo().algo_b().update_literals();
    for (unsigned i=0 ; i != disjunction.num_literals() ; ++i) {
        copy_from_argument.push_back(disjunction.literal_var(i));
        literals.push_back(disjunction.literal_set(i));
    }
    return result.release();
}

// virtual
bool linear_combination_ge_propagator::may_combine(
    const constraint_propagator & other) const
{
    if ( dynamic_cast<const linear_combination_ge_propagator*>(&other) ||
         dynamic_cast<const difference_ge_propagator*>(&other))
        return true;
    if (dynamic_cast<const sum_mod_propagator*>(&other))
        return other.may_combine(*this);
    return false;
}

// virtual
constraint_propagator *linear_combination_ge_propagator::combine(
    const constraint_runtime_info & info,
    const constraint_propagator & other,
    const constraint_runtime_info & other_info,
    const std::vector<unsigned> & other_to_common_param,
    unsigned other_pivot_param,
    std::vector<unsigned> & copy_from_argument) const
{
    HCSP_ASSERT(may_combine(other));
    auto * sum_mod = dynamic_cast<const sum_mod_propagator*>(&other);
    if (sum_mod)
        return combine_sum_mod(
            info, *sum_mod, other_info, other_to_common_param, 
            other_pivot_param, copy_from_argument);

    disjunction_specification a(info, *this, vector<unsigned>());
    disjunction_specification b(other_info, other, other_to_common_param);
    unsigned pivot= other_to_common_param[other_pivot_param];
    return a.combine(b, pivot, copy_from_argument);
}

template <typename T1, typename T2>
//static 
constexpr bool linear_combination_ge_propagator::fits(const T2 &value)
{
    return (value >= numeric_limits<T1>::min() && 
            value <= numeric_limits<T1>::max());
}

// static
unsigned linear_combination_ge_propagator::find_non_pivot(
    const constraint_runtime_info & info, 
    const std::vector<unsigned>& other_to_common_param,
    unsigned pivot)
{
    HCSP_ASSERT(other_to_common_param.size() >= 1);
    if (other_to_common_param[0] != pivot) 
        return other_to_common_param[0];
    if (other_to_common_param.size() >= 2)
        return other_to_common_param[1];
    else
        return info.arity()+1;
}

constraint_propagator *linear_combination_ge_propagator::combine_sum_mod(
    const constraint_runtime_info & info, const sum_mod_propagator & other,
    const constraint_runtime_info & other_info,
    const vector<unsigned> & other_to_common_param, unsigned other_pivot_param,
    vector<unsigned> & copy_from_argument) const
{
    if (other_info.arity() != 2)
        return nullptr;
    HCSP_ASSERT(other.mod_value() == other.mod_result()+1);
    HCSP_ASSERT(other_pivot_param <= 1);
    copy_from_argument.resize(info.arity());
    iota(copy_from_argument.begin(), copy_from_argument.end(), 0);
    unsigned pivot = other_to_common_param.at(other_pivot_param);
    unsigned non_pivot = find_non_pivot(info, other_to_common_param, pivot);
    // a*pivot + rest >= K
    // a*(other.mod_result()-other) + rest >= K
    // (-a)*other + rest >= K - a*other.mod_result()
    vector<coefficient_type> new_coefficients = m_coefficients;
    auto next_k=  m_k - (long long)new_coefficients[pivot] * other.mod_result();
    auto next_coeff = -1L*new_coefficients[pivot];

    copy_from_argument[pivot] = non_pivot;
    if (non_pivot < info.arity())
        next_coeff += new_coefficients[non_pivot];
    
    if (!fits<iset::value_type>(next_coeff) || !fits<iset::value_type>(next_k))
        return nullptr;
    new_coefficients[pivot] = next_coeff;
    if (non_pivot < info.arity()) {
        new_coefficients.erase(new_coefficients.begin()+non_pivot);
        copy_from_argument.erase(copy_from_argument.begin()+non_pivot);
    }

    return new linear_combination_ge_propagator(next_k, move(new_coefficients));
}

// virtual
void linear_combination_ge_propagator::print(
    std::ostream & out, const constraint_runtime_info *p_info)
    const throw()
{
    if (p_info)
        p_info->print_prefix(out);
    out << "(";
    const char *positive_prefix= "";
    for (int i=0 ; i != count_params() ; ++i) {
        if (m_coefficients[i] == 0)
            continue;
        if (m_coefficients[i] > 0) {
            out << positive_prefix;
            if (m_coefficients[i] != 1) 
                out << m_coefficients[i] << "*";
        } else {
            if (m_coefficients[i] == -1)
                out << "-";
            else
                out << m_coefficients[i] << "*";
        }
        if (p_info)
            p_info->print_param(out, i);
        else
            out << "P" << i;
        positive_prefix= "+";
    }
    out << ">=" << m_k << ")";
}

template <class T>
bool approximated_arithmetic_propagator<T>::is_accurate(
    const iset & domain, unsigned effort, unsigned accurate_size_bound) const
{
    if (domain.everything())
        return false;
    else if (effort > 0)
        return ( domain.size() <= (accurate_size_bound+effort+4)*2);
    else
        return ( domain.size() <= accurate_size_bound);
}

template <class T>
bool approximated_arithmetic_propagator<T>::is_accurate(
    const iset & domain1, const iset & domain2,
    unsigned effort, unsigned accurate_size_pair_bound) const
{
    if (domain1.everything() || domain2.everything())
        return false;
    unsigned long long square_domain= domain1.size();
    square_domain= square_domain * domain2.size();
    return square_domain <= (accurate_size_pair_bound+10*effort);
}


//virtual
template <class T>
constraint_propagator::result_type
approximated_arithmetic_propagator<T>::propagate(constraint_runtime_info &info)
{
    HCSP_PROFILE_CURRENT;
    HCSP_ASSERT(info.arity() == 3);
    bool any_reduced= false;
    for (unsigned i=0 ; i!=3 && !any_reduced ; ++i)
        any_reduced= info.reduced_between_activations(i);
    
    if (!any_reduced)
        return CONSISTENT;

    iset domains[3]= { info.domain(0),  info.domain(1),  info.domain(2) };
    const unsigned effort= info.effort();
    do {
        if ( is_accurate(domains[1], domains[2], effort,
                         info.accurate_size_pair_bound())) {
            propagate_accurately(domains[0], domains[1], domains[2]);

            if (domains[0].empty())
                return UNSAT;
            copy_domains_or_bounds_to_info(domains, info);
            return CONSISTENT;
        }

        propagate_bounds(domains, info);
        if (domains[0].empty())
            return UNSAT;
    } while (copy_domains_to_info(domains, info));
    return CONSISTENT;
}

template <class T>
bool approximated_arithmetic_propagator<T>::conflicts(
    const constraint_runtime_info &info) const
{
    HCSP_PROFILE_CURRENT;
    HCSP_ASSERT(info.arity() == 3);
    for (unsigned i=0 ; i!=3 ; ++i)
        if (info.domain(i).empty())
            return true;

    iset domains[3]= { info.domain(0),  info.domain(1),  info.domain(2) };
    const unsigned effort= info.effort();
    bool changed= true;
    while (changed) {
        if ( is_accurate(domains[1], domains[2], effort,
                         info.accurate_size_pair_bound())) {
            propagate_accurately(domains[0], domains[1], domains[2]);

            return domains[0].empty();
        }
        iset orig_domains[3]= {domains[0], domains[1], domains[2]};
        propagate_bounds(domains, info);
        if (domains[0].empty())
            return true;
        changed= false;
        for (unsigned i=0 ; i != 3 && !changed; ++i)
            if (domains[i] != orig_domains[i])
                changed= true;
    } 
    return false;
}

//virtual
template <class Propagator>
void approximated_arithmetic_propagator<Propagator>::print(
    std::ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        p_info->print_param(out, 0);
        out << "== ";
        p_info->print_param(out, 1);
        out << short_name();
        p_info->print_param(out, 2);
        out << ")";
    } else {
        out << long_name();
    }
}

//virtual
template <class Propagator>
void approximated_arithmetic_propagator<Propagator>::explain_propagation(
    unsigned propagated_param, const iset & values_to_explain,
    const iset & propagated_domain, 
    const constraint_runtime_info & info_before_propagation,
    vector<iset> & explanation) const
{
    if (info_before_propagation.effort() > 0) {
        // Assume that Propagator is the final object
        HCSP_ASSERT(typeid(Propagator) == typeid(*this));
        Propagator this_copy(dynamic_cast<const Propagator&>(*this));
        explanation_generators::explain_propagation_tightly(
            this_copy, propagated_param, values_to_explain, propagated_domain,
            info_before_propagation, explanation);
    }
    else {
        explanation_generators::explain_propagation(
            propagated_param, values_to_explain, propagated_domain,
            info_before_propagation, explanation);
    }
}

// static
iset::value_type arithmetic_propagation_types::wrap(long long value)
{
    const long long modulo = 1LL+iset::MAX_DOMAIN-iset::MIN_DOMAIN;
    HCSP_STATIC_ASSERT(iset::MAX_DOMAIN > iset::MIN_DOMAIN);
    return value % modulo;
}

// static
iset::value_type arithmetic_propagation_types::saturate(long long value)
{
    HCSP_STATIC_ASSERT(iset::MAX_DOMAIN >= iset::MIN_DOMAIN);
    if (value < iset::MIN_DOMAIN)
        return iset::MIN_DOMAIN;
    else if (value > iset::MAX_DOMAIN)
        return iset::MAX_DOMAIN;
    else
        return value;
}

// virtual
template <class Propagator>
void approximated_arithmetic_propagator<Propagator>::propagate_accurately(
    iset & result, iset & left, iset & right) const
{
    vector<iset::value_type> right_values(right.size());
    copy(right.begin(), right.end(), right_values.begin());
    iset result_supported, right_supported, left_supported;
    for (iset::iterator left_it=left.begin(); left_it!=left.end() ; ++left_it){
        for (unsigned i=0 ; i != right_values.size(); ++i) {
            long long result_value;
            result_value= Propagator::evaluate(*left_it, right_values[i]);
            if ( result_value > iset::MAX_DOMAIN ||
                 result_value < iset::MIN_DOMAIN ||
                 !result.contains(result_value) )
                continue;
            result_supported.insert(result_value);
            right_supported.insert(right_values[i]);
            left_supported.insert(*left_it);
        }
    }
    result= result_supported;
    left= left_supported;
    right= right_supported;
}

template <class Propagator>
void approximated_arithmetic_propagator<Propagator>::propagate_bounds(
    iset domains[3], const constraint_runtime_info & info) const
{
    for (unsigned i=0 ; i != 3 ; ++i) {
        if (domains[i].min() < 0 && domains[i].max() > 0) {
            iset positive[3]= {domains[0], domains[1], domains[2]};
            positive[i].intersect(make_pair(1, iset::MAX_DOMAIN));
            if (positive[i].empty())
                positive[0]= positive[1]= positive[2]= iset();
            else
                propagate_bounds(positive, info);

            iset negative[3]= {domains[0], domains[1], domains[2]};
            negative[i].erase(make_pair(1, iset::MAX_DOMAIN));
            if (negative[i].empty())
                negative[0]= negative[1]= negative[2]= iset();
            else
                propagate_bounds(negative, info);

            for (unsigned j=0 ; j != 3 ; ++j) {
                domains[j]= positive[j];
                domains[j].insert_all_of(negative[j]);
            }
            return;
        }
    }
    propagate_unisigned_bounds(domains, info);
    if (domains[0].empty() || domains[1].empty() || domains[2].empty())
        domains[0]= domains[1]= domains[2]= iset();
}

//! Return the value closest to zero, assuming a uni-signed set
static iset::value_type inner(const iset & domain)
{
    // -x < y
    // For integer types, this is the same as 
    // -x+1 <= y
    // Which is normalized into
    // 1-x <= y
    
    if (domain.min() < 0)
        return domain.max();
    else
        return domain.min();
}

//! Return the value farthest from zero, assuming a uni-signed set
static iset::value_type outer(const iset & domain)
{
    // -x < y
    // For integer types, this is the same as 
    // -x+1 <= y
    // Which is normalized into
    // 1-x <= y
    
    if (domain.min() < 0)
        return domain.min();
    else
        return domain.max();
}


//virtual
void mul_wrap_propagator::propagate_unisigned_bounds(
    iset domains[3], const constraint_runtime_info &) const
{
    HCSP_PROFILE_CURRENT;
    
    for (unsigned i=0 ; i != 3 ; ++i)
        HCSP_ASSERT(domains[i].min() == 0 || domains[i].max() == 0 ||
                   (domains[i].min() > 0) == (domains[i].max() > 0) );

    long long small= (long long)inner(domains[1])*inner(domains[2]);
    long long big= (long long)outer(domains[1])*outer(domains[2]);
    if (small > big)
        swap(small,big);
    ops::intersect(domains[0], small, big);
    if (domains[0].empty())
        return;

    if (!domains[0].contains(0)) {
        domains[1].erase(0);
        domains[2].erase(0);
        if (domains[1].empty() || domains[2].empty())
            return;
    }
    long long biggest1= max(1LL*domains[1].max(), -1LL*domains[1].min());
    long long biggest2= max(1LL*domains[2].max(), -1LL*domains[2].min());
    if (biggest1 * biggest2 > iset::MAX_DOMAIN)
        return;
        
    // If domains[1].contains(0) then any value of domains[2] is supported
    if ( !domains[1].contains(0) ) {
        small= (long long)inner(domains[0])/outer(domains[1]);
        big=   (long long)outer(domains[0])/inner(domains[1]);
        if (small > big)
            swap(small, big);
        ops::intersect(domains[2], small, big);
        if (domains[2].empty())
            return;
    }
    // If domains[2].contains(0) then any value of domains[1] is supported
    if ( !domains[2].contains(0)) {
        small= (long long)inner(domains[0])/outer(domains[2]);
        big=   (long long)outer(domains[0])/inner(domains[2]);
        if (small > big)
            swap(small, big);
        ops::intersect(domains[1], small, big);
    }
}

void mul_truncate_propagator::algo::propagate_out()
{
    if (empty())
        return;
    interval_type extremes(init_range());
    widen_range(extremes, (long long)m_domains[1].min()*m_domains[2].min());
    widen_range(extremes, (long long)m_domains[1].min()*m_domains[2].max());
    widen_range(extremes, (long long)m_domains[1].max()*m_domains[2].min());
    widen_range(extremes, (long long)m_domains[1].max()*m_domains[2].max());
    intersect_with_interval(m_domains[0], extremes);
    if (extremes.first >= 0 || extremes.second <= 0)
        return;
    
    interval_type positive(init_range());
    interval_type negative(init_range());
    positive.second= extremes.second;
    negative.first= extremes.first;
    iterator it1= m_domains[1].lower_bound(0);
    iterator it2= m_domains[2].lower_bound(0);
    iterator negative_it1= it1, negative_it2=it2;
    if (negative_it1 == m_domains[1].begin())
        negative_it1= m_domains[1].end();
    else
        --negative_it1;
    if (negative_it2 == m_domains[2].begin())
        negative_it2= m_domains[2].end();
    else
        --negative_it2;
    if (it1 != m_domains[1].end() && it2 != m_domains[2].end())
        widen_range(positive, (long long)*it1 *(*it2));
    if (negative_it1!=m_domains[1].end() && negative_it2!=m_domains[2].end())
        widen_range(positive, (long long)*negative_it1 *(*negative_it2));
    if (negative_it1 != m_domains[1].end() && it2 != m_domains[2].end())
        widen_range(negative, (long long)*negative_it1 *(*it2));
    if (it1!=m_domains[1].end() && negative_it2!=m_domains[2].end())
        widen_range(negative, (long long)*it1 *(*negative_it2));

    interval_type hole(negative.second+1, positive.first-1);
    if (hole.first > hole.second)
        return;
    m_domains[0].erase(hole);
}

void mul_truncate_propagator::algo::propagate_to_multiplicand(
    unsigned multiplicand_index)
{
    if (empty())
        return;
    const iset & multiplier= m_domains[3-multiplicand_index];
    const iset & out = m_domains[0];
    if (out.contains(0) && multiplier.contains(0))
        return;
    iset multiplicand_bounds;
    insert_zero_supported_multiplicand(multiplicand_bounds,
                                       multiplicand_index);
    insert_positive_supported_multiplicand_bounds(multiplicand_bounds,
                                                  multiplicand_index);
    insert_negative_supported_multiplicand_bounds(multiplicand_bounds,
                                                  multiplicand_index);

    m_domains[multiplicand_index].intersect(multiplicand_bounds);
}

void mul_truncate_propagator::algo::insert_zero_supported_multiplicand(
    iset & bounds, unsigned multiplicand_index)
{
    const iset & multiplicand= m_domains[multiplicand_index];
    const iset & out = m_domains[0];

    if (out.contains(0) && multiplicand.contains(0))
        bounds.insert(0);
}

void
mul_truncate_propagator::algo::insert_positive_supported_multiplicand_bounds(
    iset & bounds, unsigned multiplicand_index)
{
    const iset & multiplier= m_domains[3-multiplicand_index];
    const iset & out = m_domains[0];

    iterator min_positive_multiplier= multiplier.upper_bound(0);
    if (out.max()>0 && min_positive_multiplier != multiplier.end()) {
        iterator near0_out= out.upper_bound(0);
        value_type max_multiplier= multiplier.max();
        interval_type positive_interval(
            (*near0_out-1LL+max_multiplier)/max_multiplier, // round up
            out.max()/(*min_positive_multiplier));          // round down
        
        insert_interval(bounds, positive_interval);
    }
    // The negative multiplier which is closest to zero:
    iterator max_negative_multiplier_it= max_negative_it(multiplier);
    iterator max_negative_out_it=        max_negative_it(out);
    if ( max_negative_out_it!=out.end() &&
         max_negative_multiplier_it!=multiplier.end() ){
        interval_type positive_interval(
            // round up, away from zero
            (*max_negative_out_it+1LL+multiplier.min())/multiplier.min(),
            out.min()/(long long)*max_negative_multiplier_it); // round down
        
        insert_interval(bounds, positive_interval);
    }
}

void
mul_truncate_propagator::algo::insert_negative_supported_multiplicand_bounds(
    iset & bounds, unsigned multiplicand_index)
{
    const iset & multiplier= m_domains[3-multiplicand_index];
    const iset & out = m_domains[0];

    // The negative multiplier which is closest to zero:
    iterator max_negative_multiplier_it= max_negative_it(multiplier);
    if (out.max()>0 && max_negative_multiplier_it != multiplier.end()) {
        iterator min_pos_out= out.upper_bound(0);
        interval_type negative_interval(
            // round up towards 0
            out.max()/(long long)*max_negative_multiplier_it, 
            // round down (away from zero)
            (*min_pos_out-1LL-multiplier.min())/multiplier.min());
        
        insert_interval(bounds, negative_interval);
    }
    iterator max_negative_out_it=        max_negative_it(out);
    iterator min_positive_multiplier_it= multiplier.upper_bound(0);
    if ( multiplier.max()>0  && max_negative_out_it != out.end() ) {
        interval_type negative_interval(
            // round up towards 0
            out.min()/(long long)*min_positive_multiplier_it, 
            // round down (away from zero)
            (*max_negative_out_it+1LL-multiplier.max())/multiplier.max());
        
        insert_interval(bounds, negative_interval);
    }
}

constraint_propagator::result_type
mul_truncate_propagator::propagate(constraint_runtime_info &info)
{
    HCSP_PROFILE_CURRENT;
    HCSP_ASSERT(info.arity() == 3);
    bool any_reduced= false;
    for (unsigned i=0 ; i!=3 && !any_reduced ; ++i)
        any_reduced= info.reduced_between_activations(i);
    
    if (!any_reduced)
        return CONSISTENT;
    
    iset domains[3]= { info.domain(0),  info.domain(1),  info.domain(2) };
    const unsigned effort= info.effort();
    bool first_iteration= true;
    while (true) {
        if ( is_accurate(domains[1], domains[2], effort,
                         info.accurate_size_pair_bound())) {
            propagate_accurately(domains[0], domains[1], domains[2]);

            if (domains[0].empty())
                return UNSAT;
            copy_domains_or_bounds_to_info(domains, info);
            return CONSISTENT;
        }
        if (!first_iteration)
            break;

        algo(domains).propagate_bounds();
        if (domains[0].empty())
            return UNSAT;
        first_iteration= false;
    }
    copy_domains_to_info(domains, info);
    return CONSISTENT;
}

bool mul_truncate_propagator::conflicts(const constraint_runtime_info &info)
    const
{
    HCSP_PROFILE_CURRENT;
    HCSP_ASSERT(info.arity() == 3);
    for (unsigned i=0 ; i!=3 ; ++i)
        if (info.domain(i).empty())
            return true;
    
    iset domains[3]= { info.domain(0),  info.domain(1),  info.domain(2) };
    const unsigned effort= info.effort();
    bool first_iteration= true;
    while (true) {
        if ( is_accurate(domains[1], domains[2], effort,
                         info.accurate_size_pair_bound())) {
            propagate_accurately(domains[0], domains[1], domains[2]);

            return domains[0].empty();
        }
        if (!first_iteration)
            break;

        algo(domains).propagate_bounds();
        if (domains[0].empty()) {
            return true;
        }
        first_iteration= false;
    }
    return false;
}

void mul_truncate_propagator::algo::propagate_bounds()
{
    unsigned last_modified= -1U;
    update_sizes();
    while (last_modified >= 4) {
        if (last_modified == -1U) {
            propagate_out();
            if (update_sizes())
                last_modified=1;
            else
                last_modified= 0;
        } else
            last_modified= 0;
            
        propagate_to_multiplicand(1);
        if (update_sizes())
            last_modified=2;
        propagate_to_multiplicand(2);
        if (update_sizes()) {
            last_modified=3;
            propagate_to_multiplicand(1);
            if (update_sizes())
                last_modified=4;
        }
        
        if (last_modified > 1) {
            propagate_out();
            if (update_sizes())
                last_modified=5;
        }
    }
    if (empty())
        clear();
}
    
//! Watch the removal of values closest to zero. Assume that the polarity
//! of the domain is one sided, either positive or negative.
static void allocate_inner_watch(constraint_runtime_info & info, unsigned param)
{
    if (info.domain(param).min() >= 0)
        info.allocate_watch(param, watch_types::LOWER_BOUND_WATCH);
    else
        info.allocate_watch(param, watch_types::UPPER_BOUND_WATCH);
}

//! Watch the removal of values farthest from zero. Assume that the polarity
//! of the domain is one sided, either positive or negative.
static void allocate_outer_watch(constraint_runtime_info & info, unsigned param)
{
    if (info.domain(param).min() >= 0)
        info.allocate_watch(param, watch_types::UPPER_BOUND_WATCH);
    else
        info.allocate_watch(param, watch_types::LOWER_BOUND_WATCH);
}

//virtual
bool less_equal_mul_propagator::init( constraint_runtime_info & info )
{
    HCSP_ASSERT(info.arity() == 3);
    HCSP_ASSERT(info.arity() == (unsigned)count_params());

    info.allocate_watch(0, LOWER_BOUND_WATCH);
    if ( (info.domain(1).min() >= 0 && info.domain(2).min() >= 0) ||
         (info.domain(1).max() <= 0 && info.domain(2).max() <= 0) ) {
        // In this case the multiplication always gives nonnegative values
        allocate_outer_watch(info, 1);
        allocate_outer_watch(info, 2);
    } else if ( (info.domain(1).max() <= 0 && info.domain(2).min() >= 0) ||
                (info.domain(1).min() >= 0 && info.domain(2).max() <= 0) ) {
        // In this case the multiplication always gives non-positive values
        allocate_inner_watch(info, 1);
        allocate_inner_watch(info, 2);
    } else if (info.domain(0).max() <= 0 && 
               ( info.domain(1).max() <= 0 || info.domain(2).min() >= 0 ) ) {
        // In this case the output is restricted to negative values and it
        // is possible to constrain multiplication inputs to ensure that.
        info.allocate_watch(1, UPPER_BOUND_WATCH);
        info.allocate_watch(2, LOWER_BOUND_WATCH);
    } else if (info.domain(0).max() <= 0 && 
               ( info.domain(2).max() <= 0 || info.domain(1).min() >= 0 ) ) {
        // In this case the output is restricted to negative values and it
        // is possible to constrain multiplication inputs to ensure that.
        info.allocate_watch(1, LOWER_BOUND_WATCH);
        info.allocate_watch(2, UPPER_BOUND_WATCH);
    } else {
        // We can't guarantee anything in multiplication inputs
        info.allocate_watch(1, MODIFICATION_WATCH);
        info.allocate_watch(2, MODIFICATION_WATCH);
    }
    return true;
}

//virtual
constraint_propagator::result_type
less_equal_mul_propagator::propagate( constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;

    HCSP_ASSERT(info.arity() == 3);
    const iset & a= info.domain(1);
    const iset & b= info.domain(2);
    long long max_mul= max(max(1LL*a.min()*b.min(), 1LL*a.min()*b.max()),
                           max(1LL*a.max()*b.min(), 1LL*a.max()*b.max()));
    iset::value_type min_bound= info.domain(0).min();
    if (min_bound > max_mul)
        return UNSAT;
    if (info.domain(0).max() > max_mul)
    {
        iset out0= info.domain(0);
        out0.intersect(make_pair(min_bound, max_mul));
        info.set_domain(0, move(out0));
    }
    propagate_to_multiplicand(1, info);
    propagate_to_multiplicand(2, info);

    return CONSISTENT;
}

//virtual
bool less_equal_mul_propagator::conflicts(
    const constraint_runtime_info & info) const
{
    HCSP_ASSERT(info.arity() == 3);
    const iset & a= info.domain(1);
    const iset & b= info.domain(2);
    long long max_mul= max(max(1LL*a.min()*b.min(), 1LL*a.min()*b.max()),
                           max(1LL*a.max()*b.min(), 1LL*a.max()*b.max()));
    iset::value_type min_bound= info.domain(0).min();
    return min_bound > max_mul;
}

/*
   Constraint: L <= a*b
   
            |        L>0               | L <= 0
   ---------+--------------------------+----------------
            |                          |
   max(a)>0 | L/max(a) <= b            | L/min_{+}(a) <= b      
            | ceil(L/max(a)) <= b      | ceil(L/min_{+}(a)) <= b
            | positive <= b            | non-positive <= b    
   ---------+--------------------------+----------------
            |                          |
   min(a)<0 | L/min(a) >= b            | L/max_{-}(a) >= b    
            | floor(L/min(a)) >= b     | floor(L/max{-}(a)) >= b
            |                          |
            |  negative >= b           | nonnegative >= b
   ---------+--------------------------+----------------
   a=0      | illegal                  | b=unlimited
   ---------+--------------------------+----------------
   Result   | b \not \in (             | if max(a)>=0 && min(a)<=0
            |          floor(L/min(a)),| then b=unlimited         
            |          ceil(L/max(a))) | Otherwise act by case    

 */
void less_equal_mul_propagator::propagate_to_multiplicand(
    unsigned multiplicand_index, constraint_runtime_info & info )
{
    HCSP_ASSERT(multiplicand_index <= 2);
    unsigned multiplier_index= 3-multiplicand_index;
    const iset & multiplier=   info.domain(multiplier_index);
    iset::value_type lower_bound= info.domain(0).min();

    pair<long long, long long> exclude;
    exclude.first= numeric_limits<long long>::min();
    exclude.second= numeric_limits<long long>::max();
    if (lower_bound > 0) {
        if ( multiplier.min() < 0)
            exclude.first= floor_div(lower_bound ,multiplier.min())+1;
        if ( multiplier.max() > 0)
            exclude.second= ceil_div(lower_bound ,multiplier.max())-1;
    } else {
        if (multiplier.min() <= 0 && multiplier.max() >=0)
            return;
        if (multiplier.max() > 0) {
            iset::iterator it= multiplier.upper_bound(0);
            exclude.second= ceil_div(lower_bound, *it)-1;
        }
        if (multiplier.min() < 0) {
            iset::iterator it= multiplier.lower_bound(0);
            if (it == multiplier.end() || *it == 0)
                --it;
            exclude.first= floor_div(lower_bound, *it)+1;
        }
    }
    HCSP_ASSERT(exclude.first <= exclude.second);
    if (exclude.first > iset::MAX_DOMAIN || exclude.second < iset::MIN_DOMAIN)
        return;
    exclude.first=  max<long long>(exclude.first,  iset::MIN_DOMAIN);
    exclude.second= min<long long>(exclude.second, iset::MAX_DOMAIN);
    if (info.domain(multiplicand_index).contains_any_of(iset(exclude)))
    {
        iset next_domain = info.domain(multiplicand_index);
        next_domain.erase(exclude);
        info.set_domain(multiplicand_index, move(next_domain));
    }
}

// virtual
constraint_propagator *less_equal_mul_propagator::intersect_with(
    const constraint_runtime_info & info,
    std::vector<unsigned> & copy_from_argument) const
{
    const iset & out= info.domain(0);
    for (unsigned i=1 ; i != 3 ; ++i) {
        unsigned other= 3-i;
        if (!info.domain(i).singleton())
            continue;
        iset::value_type coefficient= info.domain(i).min();
        typedef linear_combination_ge_propagator::coefficient_type coeff_type;
        if (coefficient > numeric_limits<coeff_type>::max() ||
            coefficient < numeric_limits<coeff_type>::min())
            continue;
        coeff_type typed_coefficient= coefficient;
        if (out.singleton()) {
            copy_from_argument.assign({other});
            iset::value_type k= out.min();
            return new linear_combination_ge_propagator(
                k, initializer_list<coeff_type>{typed_coefficient});
        }
        copy_from_argument.assign({other, 0});
        if (coefficient == 1)
            return new difference_ge_propagator(0);
                
        return new linear_combination_ge_propagator{
            0, initializer_list<coeff_type>{typed_coefficient, -1}};
    }
    return nullptr;
}


//virtual
void less_equal_mul_propagator::print(ostream & out,
                                      const constraint_runtime_info *p_info)
    const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        p_info->print_param(out, 0);
        out << "<= ";
        p_info->print_param(out, 1);
        out << "*";
        p_info->print_param(out, 2);
        out << ")";
    } else {
        out << "le_mul";
    }
}

//virtual
bool greater_equal_mul_propagator::init( constraint_runtime_info & info )
{
    HCSP_ASSERT(info.arity() == (unsigned)count_params());
    info.allocate_watch(0, UPPER_BOUND_WATCH);
    if ( (info.domain(1).min() >= 0 && info.domain(2).min() >= 0) ||
         (info.domain(1).max() <= 0 && info.domain(2).max() <= 0) ) {
        // In this case the multiplication always gives nonegative values
        allocate_inner_watch(info, 1);
        allocate_inner_watch(info, 2);
    } else if ( (info.domain(1).max() <= 0 && info.domain(2).min() >= 0) ||
                (info.domain(1).min() >= 0 && info.domain(2).max() <= 0) ) {
        // In this case the multiplication always gives non-positive values
        allocate_outer_watch(info, 1);
        allocate_outer_watch(info, 2);
    } else if (info.domain(0).min() >= 0 && 
               ( info.domain(1).min() >= 0 || info.domain(2).min() >= 0 ) ) {
        // In this case the output is restricted to nonnegative values and it
        // is possible to constrain multiplication inputs to ensure that.
        info.allocate_watch(1, LOWER_BOUND_WATCH);
        info.allocate_watch(2, LOWER_BOUND_WATCH);
    } else if (info.domain(0).min() >= 0 && 
               ( info.domain(2).max() <= 0 || info.domain(1).max() <= 0 ) ) {
        // In this case the output is restricted to nonnegative values and it
        // is possible to constrain multiplication inputs to ensure that.
        info.allocate_watch(1, UPPER_BOUND_WATCH);
        info.allocate_watch(2, UPPER_BOUND_WATCH);
    } else {
        // We can't guarantee anything in multiplication inputs
        info.allocate_watch(1, MODIFICATION_WATCH);
        info.allocate_watch(2, MODIFICATION_WATCH);
    }
    return true;
}

//virtual
constraint_propagator::result_type
greater_equal_mul_propagator::propagate( constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;

    HCSP_ASSERT(info.arity() == 3);
    const iset & a= info.domain(1);
    const iset & b= info.domain(2);
    long long min_mul= min(min(1LL*a.min()*b.min(), 1LL*a.min()*b.max()),
                           min(1LL*a.max()*b.min(), 1LL*a.max()*b.max()));
    iset::value_type max_bound= info.domain(0).max();
    if (max_bound < min_mul)
        return UNSAT;
    if (info.domain(0).min() < min_mul)
    {
        iset out0= info.domain(0);
        out0.intersect(make_pair(min_mul, max_bound));
        info.set_domain(0, move(out0));
    }

    propagate_to_multiplicand(1, info);
    propagate_to_multiplicand(2, info);

    return CONSISTENT;
}

//virtual
bool greater_equal_mul_propagator::conflicts(
    const constraint_runtime_info & info) const
{
    HCSP_PROFILE_CURRENT;

    HCSP_ASSERT(info.arity() == 3);
    const iset & a= info.domain(1);
    const iset & b= info.domain(2);
    long long min_mul= min(min(1LL*a.min()*b.min(), 1LL*a.min()*b.max()),
                           min(1LL*a.max()*b.min(), 1LL*a.max()*b.max()));
    return info.domain(0).max() < min_mul;
}

/*
   Constraint: L >= a*b
   
            |        L>=0              | L < 0
   ---------+--------------------------+----------------
            |                          |
   max(a)>0 | L/min_{+}(a) >= b        | L/max(a) >= b      
            | floor(L/min{+}(a)) >= b  | floor(L/max(a)) >= b
            |  nonnegative >= b        | negative >= b    
   ---------+--------------------------+----------------
            |                          |
   min(a)<0 | L/max_{-}(a) <= b        | L/min(a) <= b    
            | ceil(L/max_{-}(a))<=b    | ceil(L/min(a))<=b
            |                          |
            |  non-positive <= b       | positive <= b
   ---------+--------------------------+----------------
   a=0      | b= unlimited             | illegal
   ---------+--------------------------+----------------
   Result   | if max(a)>=0 && min(a)<=0| b \not\in (floor(L/max(a)),
            | then b=unlimited         |            ceil(L/min(a)))
            | Otherwise act by case    |

 */
void greater_equal_mul_propagator::propagate_to_multiplicand(
    unsigned multiplicand_index, constraint_runtime_info & info )
{
    HCSP_ASSERT(multiplicand_index <= 2);
    unsigned multiplier_index= 3-multiplicand_index;
    const iset & multiplier=   info.domain(multiplier_index);
    iset::value_type upper_bound= info.domain(0).max();

    pair<long long, long long> exclude;
    exclude.first= numeric_limits<long long>::min();
    exclude.second= numeric_limits<long long>::max();
    if (upper_bound >=0) {
        if (multiplier.min() <= 0 && multiplier.max() >= 0)
            return;
        if (multiplier.max() > 0) {
            iset::iterator it= multiplier.upper_bound(0);
            exclude.first= floor_div(upper_bound, *it)+1;
        }
        if (multiplier.min() < 0) {
            iset::iterator it= multiplier.lower_bound(0);
            if (it == multiplier.end() || *it == 0)
                --it;
            exclude.second= ceil_div(upper_bound, *it)-1;
        }
    } else {
        if (multiplier.max() > 0)
            exclude.first= floor_div(upper_bound, multiplier.max())+1;
        if (multiplier.min() < 0)
            exclude.second= ceil_div(upper_bound, multiplier.min())-1;
    }
    HCSP_ASSERT(exclude.first <= exclude.second);
    if (exclude.first > iset::MAX_DOMAIN || exclude.second < iset::MIN_DOMAIN)
        return;
    exclude.first=  max<long long>(exclude.first,  iset::MIN_DOMAIN);
    exclude.second= min<long long>(exclude.second, iset::MAX_DOMAIN);
    if (info.domain(multiplicand_index).contains_any_of(iset(exclude)))
    {
        iset next_domain= info.domain(multiplicand_index);
        next_domain.erase(exclude);
        info.set_domain(multiplicand_index, move(next_domain));
    }
}

// virtual
constraint_propagator *greater_equal_mul_propagator::intersect_with(
    const constraint_runtime_info & info,
    std::vector<unsigned> & copy_from_argument) const
{
    const iset & out= info.domain(0);
    for (unsigned i=1 ; i != 3 ; ++i) {
        unsigned other= 3-i;
        if (!info.domain(i).singleton())
            continue;
        long long coefficient= -1LL*info.domain(i).min();
        typedef linear_combination_ge_propagator::coefficient_type coeff_type;
        if (coefficient > numeric_limits<coeff_type>::max() ||
            coefficient < numeric_limits<coeff_type>::min())
            continue;
        coeff_type typed_coefficient= coefficient;
        if ( out.singleton() &&
             out.min() != numeric_limits<iset::value_type>::min() ) {
            // out >= (-coefficient)*other ---> coefficient*other >= -out
            copy_from_argument.assign({other});
            iset::value_type k= -out.min();
            return new linear_combination_ge_propagator(
                k, initializer_list<coeff_type>{typed_coefficient});
        }
        // out >= (-coefficient)*other ---> out+coefficient*other >= 0
        copy_from_argument.assign({0, other});
        if (coefficient == -1)
            return new difference_ge_propagator(0);

        return new linear_combination_ge_propagator{
            0, initializer_list<coeff_type>{1, typed_coefficient}};
    }
    return nullptr;
}

//virtual
void greater_equal_mul_propagator::print(ostream & out,
                                      const constraint_runtime_info *p_info)
    const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        p_info->print_param(out, 0);
        out << ">= ";
        p_info->print_param(out, 1);
        out << "*";
        p_info->print_param(out, 2);
        out << ")";
    } else {
        out << "le_mul";
    }
}

//static
void arithmetic_propagation_types::intersect_truncated(
    iset & out, long long first, long long second)
{
    typedef iset::value_range_type range_type;
    HCSP_ASSERT(first <= second);
    HCSP_STATIC_ASSERT(iset::MAX_DOMAIN > iset::MIN_DOMAIN);
    first= max<long long>(first, iset::MIN_DOMAIN);
    second= min<long long>(second, iset::MAX_DOMAIN);
    if (first > second)
        return;
    
    out.intersect(range_type(first, second));
}

//static
void arithmetic_propagation_types::intersect_wrapped(
    iset & out, long long first, long long second)
{
    typedef iset::value_range_type range_type;
    HCSP_ASSERT(first <= second);
    HCSP_STATIC_ASSERT(iset::MAX_DOMAIN > iset::MIN_DOMAIN);
    const unsigned long long modulo = 1LL+iset::MAX_DOMAIN-iset::MIN_DOMAIN;
    if ( (unsigned long long)second-first >= modulo-1 )
        return;
    
    // Normalize range.first;
    if ( first > iset::MAX_DOMAIN || first < iset::MIN_DOMAIN ) {
        long long new_first= first % modulo;
        if (new_first > iset::MAX_DOMAIN)
            new_first -= modulo;
        else if (new_first < iset::MIN_DOMAIN)
            new_first += modulo;
        long long diff= new_first - first;
        first= new_first;
        second += diff;
    }
    HCSP_ASSERT(iset::MIN_DOMAIN <= first && first <= iset::MAX_DOMAIN);
    if ( second > iset::MAX_DOMAIN ) {
        iset domain(range_type(first, iset::MAX_DOMAIN));
        long long new_second = second-modulo;
        // This case of double wrap around has been handled at the beginning
        HCSP_ASSERT(new_second < iset::MAX_DOMAIN);
        domain.insert(range_type(iset::MIN_DOMAIN, new_second));
        out.intersect(domain);
    } else {
        out.intersect(range_type(first, second));
    }
    
}

template <class Propagator>
bool
approximated_arithmetic_propagator<Propagator>::copy_domains_to_info(
    iset domains[3], constraint_runtime_info & info)
{
    HCSP_ASSERT(info.arity() == 3);
    bool any_modified= false;
    for (unsigned i=0 ; i != 3 ; ++i) {
        if (domains[i] != info.domain(i)) {
            HCSP_ASSERT(info.domain(i).contains_all_of(domains[i]));
            info.set_domain(i, domains[i]);
            any_modified= true;
        }
    }
    return any_modified;
}

template <class Propagator>
bool
approximated_arithmetic_propagator<Propagator>::copy_domains_or_bounds_to_info(
    const iset domains[3], constraint_runtime_info & info)
{
    HCSP_ASSERT(info.arity() == 3);
    bool any_modified= false;
    for (unsigned i=0 ; i != 3 ; ++i) {
        if (domains[i] != info.domain(i)) {
            HCSP_ASSERT(info.domain(i).contains_all_of(domains[i]));
            unsigned max_ranges= info.max_accurate_ranges()+info.effort();
            if ( domains[i].range_count() <= max_ranges){
                info.set_domain(i, domains[i]);
                any_modified= true;
            } else if (domains[i].min() > info.domain(i).min() ||
                       domains[i].max() < info.domain(i).max()) {
                iset out_i = info.domain(i);
                out_i.intersect(make_pair(domains[i].min(), domains[i].max()));
                info.set_domain(i, move(out_i));
            }
        }
    }
    return any_modified;
}

void mod_propagator::propagate_when_zero_out(iset domains[3]) const
{
    iset::value_type min_divider= inner(domains[2]);
    if (min_divider < 0)
        min_divider= -min_divider;
    if (min_divider > 1) {
        domains[1].erase(make_pair(1,min_divider-1));
        domains[1].erase(make_pair(1-min_divider,-1));
        if (domains[1].empty()) {
            domains[0].clear();
            domains[2].clear();
            return;
        }
    }
    iset::value_type max_abs_divided= outer(domains[1]);
    if (max_abs_divided < 0)
        max_abs_divided = -max_abs_divided;
    domains[2].intersect(make_pair(-max_abs_divided, max_abs_divided));
}

//virtual
void mod_propagator::propagate_unisigned_bounds_base(
    iset domains[3], const constraint_runtime_info &) const
{
    for (unsigned i=0 ; i != 3 ; ++i)
        HCSP_ASSERT(domains[i].min() == 0 || domains[i].max() == 0 ||
                   (domains[i].min() > 0) == (domains[i].max() > 0) );

    if (!domains[0].contains(0)) {
        domains[1].erase(0);
        domains[2].erase(0);
        if (domains[1].empty() || domains[2].empty())
            return;
    }
    iset next_domains[3];
    bool is_neg[3];
    for (unsigned i=0 ; i != 3 ; ++i) {
        is_neg[i]= (!domains[i].empty() && outer(domains[i]) < 0);
        if (is_neg[i]) {
            next_domains[i].assign(make_pair(iset::MIN_DOMAIN,
                                             iset::MAX_DOMAIN));
            propagation::neg(domains[i], next_domains[i]);
            // We should have handled MIN_INT earlier, and for any other x<0
            // we have -x >=0
            HCSP_ASSERT(next_domains[i].min() >= 0);
        } else 
            next_domains[i]= domains[i];
    }
    if (is_neg[0] != is_neg[1]) {
        next_domains[0].intersect(iset(0));
        if ( !next_domains[0].empty() )
            propagate_when_zero_out(next_domains);
    }
    next_domains[2].erase(0);
    if ( !next_domains[0].empty() && !next_domains[1].empty() &&
         !next_domains[2].empty())
        propagate_positive_bounds(next_domains);
    for (unsigned i=0 ; i != 3 ; ++i) {
        if (is_neg[i])
            propagation::neg(next_domains[i], domains[i]);
        else
            domains[i]= next_domains[i];
    }
}

//virtual
void div_propagator::propagate_unisigned_bounds(
    iset domains[3], const constraint_runtime_info &) const
{
    for (unsigned i=0 ; i != 3 ; ++i)
        HCSP_ASSERT(domains[i].min() == 0 || domains[i].max() == 0 ||
                   (domains[i].min() > 0) == (domains[i].max() > 0) );
    bool dividend_is_free= (domains[0].contains(0) && domains[2].contains(0));

    if (!domains[0].contains(0)) {
        domains[1].erase(0);
        domains[2].erase(0);
        if (domains[1].empty() || domains[2].empty())
            return;
    }
    domains[2].erase(0);

    bool empty= domains[2].empty();
    if (!empty) {
        propagate_unisigned_bounds_division(domains);
        empty= domains[0].empty();
    }
    if (!empty && !dividend_is_free) {
        propagate_unisigned_bounds_dividend(domains);
        empty= domains[1].empty();
    }
    if (!empty)
        propagate_unisigned_bounds_divisor(domains);

    if (empty)
        domains[0]= domains[2]= iset();
    if (empty && !dividend_is_free)
        domains[1].clear();
    
    if (dividend_is_free) {
        domains[0].insert(0);
        domains[2].insert(0);
    }
}

void div_propagator::propagate_unisigned_bounds_division(iset domains[3]) const
{
    long long inner_result= (long long)inner(domains[1])/outer(domains[2]);
    long long outer_result= (long long)outer(domains[1])/inner(domains[2]);
    if (outer_result <0)
        intersect_truncated(domains[0], outer_result, inner_result);
    else
        intersect_truncated(domains[0], inner_result, outer_result);
}

void div_propagator::propagate_unisigned_bounds_dividend(iset domains[3]) const
{
    long long inner_result= (long long)inner(domains[0])*inner(domains[2]);
    long long outer_result= (long long)outer(domains[0])*outer(domains[2]);

    long long slack;
    if (domains[2].max() > 0)
        slack= domains[2].max()-1LL;
    else
        slack= -1LL-domains[2].min();

    if (outer_result >0)
        intersect_truncated(domains[1], inner_result, outer_result+ slack);
    else if (outer_result <0)
        intersect_truncated(domains[1], outer_result-slack, inner_result);
    else
        intersect_truncated(domains[1], -slack, +slack);
}

void div_propagator::propagate_unisigned_bounds_divisor(iset domains[3]) const
{
    iset gives0;
    if (inner(domains[0])==0) {
        long long inner_divisor= inner(domains[1]);
        long long min_abs_divisor=inner_divisor>0?inner_divisor:-inner_divisor;
        gives0= domains[2];
        if (outer(domains[2]) > 0)
            intersect_truncated(gives0, min_abs_divisor+1,
                                numeric_limits<long long>::max());
        else
            intersect_truncated(gives0, numeric_limits<long long>::min(),
                                -min_abs_divisor-1);
        
    }
    iset nonzero_division= domains[0];
    nonzero_division.erase(0);

    if (nonzero_division.empty())
        domains[2]= gives0;
    else {
        // a +(k/c)=b/c  where 0<=k<c
        // c*(a+k/c)=b
        // c=b/(a+k/c) =>  b/(a+1)<c<=b/a ==> floor(b/(a+1))+1<=c<=b/a
        iset::value_type a = inner(nonzero_division);
        if (a == 0) // dead code to tell gcc that this case is impossible
            return;
        int64_t outer_divisor= 1LL*outer(domains[1])/a;
        long long outer_division = outer(nonzero_division);
        if (outer_division == 0)
            return;  // dead code to tell gcc that this case is impossible
        if (outer_division > 0)
            ++outer_division;
        else
            --outer_division;
        
        long long inner_divisor= 1LL*inner(domains[1])/outer_division;
        if (inner_divisor > 0)
            ++inner_divisor;
        else
            --inner_divisor;
        if (inner_divisor <= outer_divisor && outer_divisor >= 0)
            intersect_truncated(domains[2], inner_divisor, outer_divisor);
        else if (inner_divisor >= outer_divisor && outer_divisor < 0)
            intersect_truncated(domains[2], outer_divisor, inner_divisor);
        domains[2].insert_all_of(gives0);
    }
}

bool mod_propagator::propagate_min_int_divided(
    const iset domains[3], iset next_domains[3],
    const constraint_runtime_info &info) const
{
    const iset::value_type min_int= numeric_limits<iset::value_type>::min();
    if (domains[0].empty() || domains[0].min() > 0)
        return false;

    unsigned effort= 10; //! \todo Need to pass this down from propagate()
    if (is_accurate(domains[2], effort, info.accurate_size_bound())) {
        bool keep_min_int= false;
        for ( iset::iterator it= domains[2].begin() ;
              it != domains[2].end();
              ++it) {
            iset::value_type divider= *it;
            if (divider==0)
                continue;
            if (divider < 0)
                divider *= -1;
            iset::value_type keep= min_int % divider;
            if (!domains[0].contains(keep))
                continue;
            next_domains[0].insert(keep);
            next_domains[2].insert(*it);
            keep_min_int= true;
        }
        return keep_min_int;
    }
    next_domains[2]= domains[2];
    iset::value_type extreme= outer(domains[2]);
    iset keep= domains[0];
    if (extreme < 0)
        keep.intersect(make_pair(extreme+1, -1-extreme));
    else
        keep.intersect(make_pair(1-extreme, extreme-1));
    next_domains[0].insert_all_of(keep);
    return !keep.empty();
}

void
mod_propagator::propagate_min_int(iset domains[3], bool has_min_int[3],
                                  const constraint_runtime_info &info) const
{
    bool keep_min_int[3]={false,false,false};
    iset next_domains[3]={domains[0], domains[1], domains[2]};
    if (domains[0].empty() || domains[1].empty() || domains[2].empty())
        next_domains[0]= next_domains[1]= next_domains[2]= iset();
    else
        propagate_unisigned_bounds_base(next_domains, info);
    if (has_min_int[1] && has_min_int[2]) {
        if (domains[0].contains(0)) {
            next_domains[0].insert(0);
            keep_min_int[1]= true;
            keep_min_int[2]= true;
        }
    }
    if (has_min_int[1] && propagate_min_int_divided(domains, next_domains,
                                                    info))
        keep_min_int[1]= true;
    
    if (has_min_int[2] && domains[0].contains_any_of(domains[1])) {
        iset keep= domains[0];
        keep.intersect(domains[1]);
        next_domains[0].insert_all_of(keep);
        next_domains[1].insert_all_of(keep);
        keep_min_int[2]= true;
    }

    
    for (unsigned i=0 ; i != 3 ; ++i) {
        domains[i]= next_domains[i];
        if (keep_min_int[i])
            domains[i].insert(numeric_limits<iset::value_type>::min());
    }
}
//virtual
void mod_propagator::propagate_unisigned_bounds(
    iset domains[3], const constraint_runtime_info &info) const
{
    if (domains[2].singleton() &&  domains[1].size() < 100) {
        propagate_accurately(domains[0], domains[1], domains[2]);
        return;
    }
    const iset::value_type min_int= numeric_limits<iset::value_type>::min();
        
    iset domain1=domains[1];
    bool domain1_is_free= (domains[0].contains(0) && domains[2].contains(0));
    bool has_min_int[3];
    bool any_empty= false;
    for (unsigned i=0 ; i != 3 ; ++i) {
        has_min_int[i] = domains[i].contains(min_int);
        if (has_min_int[i]) {
            domains[i].erase(min_int);
            any_empty= any_empty || domains[i].empty();
        }
    }
    
    if (has_min_int[1] || has_min_int[2]) {
        propagate_min_int(domains, has_min_int, info);
    } else {
        if (any_empty)
            domains[0]= domains[1]= domains[2]= iset();
        else
            propagate_unisigned_bounds_base(domains, info);
    }
    if (domain1_is_free) {
        domains[0].insert(0);
        domains[1]= domain1;
        domains[2].insert(0);
    }
}

//virtual
void mod_propagator::propagate_positive_bounds(iset domains[3]) const
{
    if (domains[1].max() < domains[2].min()) {
        domains[0].intersect(domains[1]);
        domains[1].intersect(domains[0]);
        return;
    }
    intersect_truncated(domains[0], 0, domains[2].max()-1);
    if (domains[2].singleton() && !domains[1].everything() &&
        domains[1].size() < (unsigned) domains[2].min()) {
        iset supported0, supported1;
        for ( iset::iterator it= domains[1].begin() ;
              it != domains[1].end();
              it.advance_range() ) {
            iset::value_range_type range1= it.current_range();
            iset::value_range_type range0( range1.first  % domains[2].min(),
                                           range1.second % domains[2].min()) ;

            if (range0.first > range0.second) {
                iset::value_range_type first1(range1.first,
                                              range1.second-range0.second-1);
                iset::value_range_type second1(range1.second-range0.second,
                                               range1.second);
                
                accumulate_positive_bounds(supported0, supported1, domains[0],
                                           first1, domains[2].min());
                accumulate_positive_bounds(supported0, supported1, domains[0],
                                           second1, domains[2].min());
            } else {
                accumulate_positive_bounds(supported0, supported1, domains[0],
                                           range1, domains[2].min());
            }
        }
        domains[0].intersect(supported0);
        domains[1].intersect(supported1);
    }
}

void mod_propagator::accumulate_positive_bounds(iset & supported0,
                                                iset & supported1,
                                                iset domain0,
                                                iset::value_range_type range1,
                                                iset::value_type mod) const
{
    iset::value_range_type range0(range1.first % mod, range1.second % mod);
    iset::value_type diff = range1.first- range0.first;
    
    HCSP_ASSERT(range1.first <= range1.second);
    HCSP_ASSERT(range0.first <= range0.second);
    
    domain0.intersect(range0);
    if ( domain0.empty() )
        return;
    
    supported0.insert_all_of(domain0);
    iset domain1(make_pair(iset::MIN_DOMAIN, iset::MAX_DOMAIN));
    propagation::plus(domain0, iset(diff), domain1);
    supported1.insert_all_of(domain1);
}

static iset even_values()
{
    iset result;
    iset::value_type max_even= iset::MAX_DOMAIN;
    if (max_even %2 != 0)
        --max_even;
    result.insert(make_pair(0, max_even), 2);
    return result;
}

//virtual
void pow_propagator::propagate_unisigned_bounds(
    iset domains[3], const constraint_runtime_info &) const
{
    // @todo It may be possible to reduce domains[2] for more cases.
    if (iset(make_pair(0,1)).contains_all_of(domains[1])) {
        domains[0].intersect(domains[1]);
        domains[1]= domains[0];
    } else if(iset(make_pair(iset::MIN_DOMAIN,1)).contains_all_of(domains[2])){
        iset support0, support1, support2;
        if (domains[2].min() < -1 && domains[0].contains(0)) {
            support0.insert(0);
            support1.assign(make_pair(iset::MIN_DOMAIN, iset::MAX_DOMAIN));
            (support2= domains[2]).erase(make_pair(-1, iset::MAX_DOMAIN));
        }
        if (domains[2].contains(-1)) {
            if (domains[0].contains(0)) {
                support0.insert(0);
                support1.insert_all_of(iset(make_pair(-1, 1)).complement());
                support2.insert(-1);
            }
            iset small_values= iset(domains[1]).intersect(make_pair(-1,1));
            small_values.intersect(domains[0]);
            if (!small_values.empty()) {
                support0.insert_all_of(small_values);
                support1.insert_all_of(small_values);
                support2.insert(-1);
            }
        }
        if ( domains[2].contains(0) && domains[1] != iset(0) &&
             domains[0].contains(1)) {
            support0.insert(1);
            support1.insert_all_of(iset(0).complement());
            support2.insert(0);
        }
        if (domains[2].contains(1) && domains[0].contains_any_of(domains[1])) {
            support0.insert_all_of(domains[1]);
            support1.insert_all_of(domains[0]);
            support2.insert(1);
        }
        if (domains[1].contains(0) && domains[0].contains(0)) {
            support0.insert(0);
            support1.insert(0);
            support2.insert(make_pair(iset::MIN_DOMAIN, iset::MAX_DOMAIN));
        }
        domains[0].intersect(support0);
        domains[1].intersect(support1);
        domains[2].intersect(support2);
    }
    else if(even_values().contains_all_of(domains[2])){
        domains[0].intersect(make_pair(0, iset::MAX_DOMAIN));
    }
}

template class approximated_arithmetic_propagator<mul_wrap_propagator>;
template class approximated_arithmetic_propagator<mul_truncate_propagator>;
template class approximated_arithmetic_propagator<div_propagator>;
template class approximated_arithmetic_propagator<mod_propagator>;
template class approximated_arithmetic_propagator<pow_propagator>;
} // namespace hcsp
