// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <limits>
#include <algorithm>
#include <numeric>
#include <string>
#include <sstream>
#include <assignment_stack.hpp>

using namespace std;
using namespace hcsp;

assignment_stack::assignment_stack() :
    m_var_print_strategy(*this)
{}

assignment_stack::~assignment_stack()
{ }


void assignment_stack::throw_out_of_range(const char * location_str) const 
    throw(std::out_of_range)
{
    if ( location_str )
        throw out_of_range(string("Out of range while calling ")+location_str);
    else
        throw out_of_range("Out of range");
}

assignment_stack::change_index_t
assignment_stack::var_modification_entry_before(
    unsigned variable_index, change_index_t constraint_modification) const
{
    change_index_t var_modification;
    var_modification= m_var_states[variable_index].m_latest_modification_entry;
    if ( constraint_modification == NO_PREVIOUS_MODIFICATION )
        return NO_PREVIOUS_MODIFICATION;

    while (!is_earlier_modification(var_modification, constraint_modification))
        var_modification=
            m_modifications[var_modification].m_previous_modification_entry;

    return var_modification;
}

void assignment_stack::complete_var_update(unsigned variable_index)
{
    const iset & last_updated_domain= var_domain(variable_index);
    if ( last_updated_domain.singleton() )
        m_previously_assigned_value[variable_index]=
            make_pair(true, last_updated_domain.min());
}

void assignment_stack::undo_level(unsigned level_to_forget)
{
    assert_level(level_to_forget, "undo_level(level_to_forget)");
    const unsigned forget_entry=
        m_decisions[level_to_forget-1].m_modification_entry;
    for ( change_index_t modification_entry= forget_entry ;
          modification_entry != m_modifications.size() ;
          ++modification_entry) {
        unsigned var_index= m_modifications[modification_entry].m_variable;
        variable_state & var_state= m_var_states[var_index];
        if ( is_earlier_modification(var_state.m_latest_modification_entry,
                                     modification_entry) )
            continue; // variable already updated.

        bool was_singleton = var_state.m_domain.singleton();
        if (was_singleton)
            m_previously_assigned_value[var_index]=
                make_pair(true, var_state.m_domain.min());

        var_state.m_domain=
            m_modifications[modification_entry].m_previous_domain;
        if (was_singleton)
            lower_var_heap_cost(var_index);
        else
            raise_var_heap_cost(var_index);
        var_state.m_latest_modification_entry=
            m_modifications[modification_entry].m_previous_modification_entry;
    }
    m_modifications.resize(forget_entry);
    m_decisions.resize(level_to_forget-1);
}

assignment_stack::time_type
assignment_stack::var_modification_time(unsigned var_index) const
{
    change_index_t modification;
    modification= m_var_states.at(var_index).m_latest_modification_entry;
    if (modification == NO_PREVIOUS_MODIFICATION)
        return 1;
    HCSP_ASSERT( modification < m_modifications.size() );
    return m_modifications[modification].m_time;
}

assignment_stack::change_index_t assignment_stack::nth_modification(
    unsigned n, const std::vector<implication_node> & nodes) const
{
    switch ( nodes.size()-n-1 ) {
      case 0: { // Find last modification
          change_index_t last= NO_PREVIOUS_MODIFICATION;
          for ( unsigned i=0 ; i != nodes.size() ; ++i)
              last= latest_modification(nodes[i].m_modification_entry, last);
          return last;
      }
      case 1: { // Find before last modification
          // Sorted from smaller to bigger:
          change_index_t latest_modifications[2] = {NO_PREVIOUS_MODIFICATION,
                                                    NO_PREVIOUS_MODIFICATION};
          for ( unsigned i=0 ; i != nodes.size() ; ++i) {
              change_index_t entry= nodes[i].m_modification_entry;
              if ( is_earlier_modification(latest_modifications[1], entry) ) {
                  latest_modifications[0]= latest_modifications[1];
                  latest_modifications[1]= entry;
              } else {
                  latest_modifications[0]=
                      latest_modification(entry, latest_modifications[0]);
              }
          }
          return latest_modifications[0];
      }
      default:
        ; // Nothing, work on it outside of the switch
    }
    static vector<change_index_t> times;
    times.clear();
    for (unsigned i=0 ; i != nodes.size() ; ++i)
        times.push_back( nodes[i].m_modification_entry );
    sort(times.begin(), times.end(), &is_earlier_modification);
    return times.at(n);
}

unsigned assignment_stack::asserting_level_of_nth(
    unsigned n, const vector<implication_node> &nodes,
    const vector<iset> & literals) const
{
    if ( nodes.empty() )
        throw_out_of_range("Can't pass an empty vector to "
                           "asserting_level(variables)");
    if ( nodes.size() == 1 )
        return 0;
    change_index_t latest_non_false= nth_modification(n, nodes);
    // before_latest_modification: The latest modification entry which is
    //     earlier than 'latest_non_false', but is not strictly satisfied.
    change_index_t before_latest_modification= NO_PREVIOUS_MODIFICATION;
    for (unsigned i=0 ; i != nodes.size() ; ++i) {
        change_index_t node_modification= nodes[i].m_modification_entry;
        if ( !is_earlier_modification(node_modification, latest_non_false) )
            continue;
        if ( !is_earlier_modification(before_latest_modification,
                                      node_modification) )
            continue;
        before_latest_modification=
            latest_modification(before_latest_modification,
                                relax(node_modification, literals.at(i)));
    }
    for ( unsigned tested_level=m_decisions.size() ;
          tested_level != 0 ;
          --tested_level) {
        const single_decision_data & decision = m_decisions[tested_level-1];
        if ( !is_earlier_modification(before_latest_modification,
                                      decision.m_modification_entry) )
            return tested_level;
    }
    return 0;
}

assignment_stack::change_index_t
assignment_stack::go_back_to(change_index_t modification_entry,
                             change_index_t target_entry,
                             unsigned max_skips) const
{
    change_index_t best= modification_entry;
    change_index_t step= modification_entry;
    unsigned skips= 0;
    while (skips < max_skips && !is_earlier_modification(step, target_entry)) {
        best= step;
        step= m_modifications[step].m_previous_modification_entry;
        ++skips;
    } 
    return best;
}

assignment_stack::change_index_t
assignment_stack::relax(change_index_t late_modification,
                        const iset &literal) const
{
    change_index_t modification_entry= late_modification;
    change_index_t upper_bound_entry= late_modification;
    change_index_t lower_bound_entry= NO_PREVIOUS_MODIFICATION;
    unsigned max_additional_skips= 0;
    // Lower calls to expensive iset::contains_any_of() by performing a kind
    // of binary search. With time skip_size increases and we are getting
    // closer and closer to true binary search.
    while (is_earlier_modification(lower_bound_entry ,upper_bound_entry)) {
        const variable_modification_entry &
            entry_data= m_modifications[modification_entry];
        const iset & prev_set= entry_data.m_previous_domain;
        if (prev_set.contains_any_of(literal))
            // prev_set is no longer unsat
            lower_bound_entry= modification_entry;
        else
            upper_bound_entry= entry_data.m_previous_modification_entry;
        if (upper_bound_entry == lower_bound_entry)
            return lower_bound_entry;
        // Notice that when NO_PREVIOUS_MODIFICATION we want unsigned overflow:
        change_index_t target_entry= (upper_bound_entry+lower_bound_entry+1)/2;
        modification_entry= upper_bound_entry;
        modification_entry= go_back_to(modification_entry, target_entry,
                                       max_additional_skips);
        if (max_additional_skips < numeric_limits<unsigned>::max()/2)
            max_additional_skips= max_additional_skips*2+1;
    }
    return lower_bound_entry;
}

assignment_stack::change_index_t
assignment_stack::relax_before(change_index_t reference_modification,
                               const iset & literal,
                               change_index_t goal_modification) const
{
    change_index_t modification_entry= reference_modification;
    while (modification_entry != NO_PREVIOUS_MODIFICATION) {
        const variable_modification_entry &
            entry_data= m_modifications[modification_entry];
        const iset & prev_set= entry_data.m_previous_domain;
        if (prev_set.contains_any_of(literal))
            return reference_modification; // prev_set is no longer unsat
        modification_entry= entry_data.m_previous_modification_entry;
        if (is_earlier_modification(modification_entry, goal_modification))
            return modification_entry;
    }
    return reference_modification;
}

vector<unsigned>
assignment_stack::get_busy_constraints(unsigned first_asked_clause) const
{
    vector<unsigned> clauses;
    for (change_index_t mod=0 ; mod != m_modifications.size(); ++mod) {
        int cause= m_modifications[mod].m_modifying_constraint;
        if (cause >=0 && (unsigned)cause >= first_asked_clause)
            clauses.push_back(cause);
    }
    return clauses;
}

inline assignment_stack::change_index_t
assignment_stack::last_modification_in(const vector<assignment> &clause) const
{
    if ( clause.empty() )
        return NO_PREVIOUS_MODIFICATION;
    change_index_t last= clause[0].m_modification_entry;
    for (unsigned i=1 ; i < clause.size() ; ++i) {
        change_index_t modification= clause[i].m_modification_entry;
        if ( is_earlier_modification(last, modification) )
            last= modification;
    }
    return last;
}

unsigned
assignment_stack::decision_level_of(change_index_t tested_modification) const
{
    if (tested_modification == NO_PREVIOUS_MODIFICATION)
        return 0;
    
    for ( unsigned decision_level=m_decisions.size() ;
          decision_level != 0 ;
          --decision_level ) {
        change_index_t modification_of_decision=
            m_decisions[decision_level-1].m_modification_entry;
        
        if ( !is_earlier_modification(tested_modification,
                                      modification_of_decision ) )
            return decision_level;
    }
    return 0;
}

const iset & assignment_stack::var_domain_end_level(
    unsigned variable_index, int level_to_find) const
{
    if (level_to_find > (int)level())
        throw_out_of_range("var_domain_end_level(variable, level)");

    if (level_to_find == -1)
        return var_domain_after_modification(
            variable_index, assignment_stack::NO_PREVIOUS_MODIFICATION);
    if ( (int)level() == level_to_find)
        return var_domain(variable_index);

    if (level_to_find == 0 && variable_index < m_level0_domains.size())
        return m_level0_domains[variable_index];
    
    change_index_t target_modification;
    target_modification= decision_modification(level_to_find+1)-1;

    return var_domain_after_modification(variable_index, target_modification);
}

const iset & assignment_stack::var_domain_after_modification(
    unsigned var_index, change_index_t target_modification) const
{
    const iset *p_domain= &m_var_states.at(var_index).m_domain;
    change_index_t tested_modification=
        m_var_states.at(var_index).m_latest_modification_entry;
    
    while (is_earlier_modification(target_modification, tested_modification)) {
        const variable_modification_entry & entry=
            m_modifications[tested_modification];
        p_domain= & entry.m_previous_domain;
        tested_modification=  entry.m_previous_modification_entry;
    }
    return *p_domain;
}

void assignment_stack::construct_level0_domains()
{
    m_level0_domains.reserve(m_var_states.size());
    for (unsigned i= m_level0_domains.size(); i != m_var_states.size(); ++i)
        m_level0_domains.push_back(m_var_states[i].m_domain);
}

void assignment_stack::set_var_domain(unsigned  var_index,
                                      iset new_domain,
                                      time_type current_time,
                                      int       cause)
{
    variable_state & updated_var_state = m_var_states.at(var_index);
    if (level() == 0 && !m_level0_domains.empty())
        m_level0_domains[var_index] = new_domain;

    updated_var_state.m_prev_size= updated_var_state.m_domain.size();
    if (m_modifications.size()+1 == NO_PREVIOUS_MODIFICATION)
        throw std::overflow_error("Too many modification entries");
    if (new_domain.singleton()) 
        updated_var_state.m_last_singleton_time= current_time;

    //! \todo Avoid redundant m_modifications entries for multiple
    //!       updates a propagator does for a single variable.
    m_modifications.push_back( variable_modification_entry() );
    variable_modification_entry & entry= m_modifications.back();
    entry.m_variable= var_index;

    //
    // Update the new modification entry
    //
    entry.m_previous_domain= move(updated_var_state.m_domain);
    entry.m_previous_modification_entry=
        updated_var_state.m_latest_modification_entry;
    entry.m_time= current_time;
    entry.m_modifying_constraint= cause;

    //
    // Update the variable state
    //
        
    updated_var_state.m_latest_modification_entry=
        m_modifications.size()-1;
    updated_var_state.m_domain= move(new_domain);
    if (updated_var_state.m_domain.singleton())
        raise_var_heap_cost(var_index);
    else
        lower_var_heap_cost(var_index);
    complete_var_update(var_index);
}

void assignment_stack::set_var_initial_domain(unsigned var_index, iset domain)
{
    HCSP_ASSERT(level() == 0);
    HCSP_ASSERT(m_modifications.empty());
    
    m_var_states.at(var_index).m_domain= move(domain);
    complete_var_update(var_index);
}

void assignment_stack::add_var(iset domain, const char *name,
                               bool update_singletons)
{
    unsigned variable= m_var_states.size();
    if ( domain.singleton() ) {
        m_previously_assigned_value.push_back(make_pair(true, domain.min()));
        if (update_singletons)
            m_singleton_variable_ids.insert(make_pair(domain.min(), variable));
    } else {
        m_previously_assigned_value.push_back(make_pair(false,0));
    }
    variable_state state;
    state.m_domain = move(domain);
    state.m_latest_modification_entry= NO_PREVIOUS_MODIFICATION;
    
    if (name == NULL)
        m_names.push_back(string());
    else
        m_names.push_back(string(name));
    m_var_states.push_back(std::move(state));
    m_labels.push_back(NO_LABEL);
    complete_var_update(variable);
}

unsigned assignment_stack::singleton_var(iset::value_type value)
{
    map<iset::value_type,unsigned>::const_iterator it;
    it = m_singleton_variable_ids.find(value);
    if (it != m_singleton_variable_ids.end())
        return it->second;
    unsigned variable= m_var_states.size();
    ostringstream name;
    name << "the_constant_" << value;
    add_var(iset(value), name.str().c_str(), true);
    HCSP_ASSERT(variable == m_var_states.size()-1);
    return variable;
}
    
bool assignment_stack::validate() const
{
    if (m_previously_assigned_value.size() != m_var_states.size())
        throw runtime_error("inconsistent m_previously_assigned_value.size()");
        
    for (unsigned i=0 ; i != m_var_states.size() ; ++i) {
        validate_var_state(i);
    }

    return true;
}

void assignment_stack::validate_var_state(unsigned var_index) const
{
    const iset & domain = m_var_states[var_index].m_domain;
    ostringstream message;
    message << "(var_index=" << var_index << ") "
            << var_printer(var_index) << "==" << domain;

    if ( domain.singleton() ) {
        if ( !m_previously_assigned_value[var_index].first) {
            message << " is singleton, but its assignment was not recorded";
            throw runtime_error(message.str());
        }
        // NOTE: It is impossible to perform the test below because
        //       the interface permits the following scenario:
        //           set V1 := 1; make a decision ; V1 := 2 ; undo decision
        //       In this case, after the undo, the last assigned value is 2
        //       but the domain is 1. There is no way to overcome this.
        // if ( m_previously_assigned_value[var_index].second != domain.max()) {
        //     message << " but the assignment recorded is to "
        //             << m_previously_assigned_value[var_index].second;
        //     throw runtime_error(message.str());
        // }
    }
}

void assignment_stack::clear()
{
    m_level0_domains.clear();
    m_previously_assigned_value.clear();
    m_previously_assigned_value.resize(m_var_states.size(),
                                       make_pair(false,0));
    for (unsigned var_index=0 ; var_index!=m_var_states.size() ; ++var_index) {
        variable_state & var_state= m_var_states[var_index];
        change_index_t modification_entry=var_state.m_latest_modification_entry;
        if ( modification_entry == NO_PREVIOUS_MODIFICATION)
            continue;
        unsigned best_entry;
        do {
            HCSP_ASSERT(modification_entry < m_modifications.size());
            
            best_entry= modification_entry;
            modification_entry=
                m_modifications[best_entry].m_previous_modification_entry;

        } while ( modification_entry != NO_PREVIOUS_MODIFICATION);
        var_state.m_domain= m_modifications[best_entry].m_previous_domain;
        var_state.m_latest_modification_entry= modification_entry;
    }
    for (unsigned var_index=0 ; var_index!=m_var_states.size() ; ++var_index) {
        if ( var_domain(var_index).singleton() ) {
            m_previously_assigned_value[var_index]=
                make_pair(true, var_domain(var_index).min());
        }
    }
    m_modifications.clear();
    m_decisions.clear();
    HCSP_ASSERT(validate());
}

void assignment_stack::erase_conflicting_modifications(
    unsigned conflicting_constraint)
{
    while (true) {
        if (m_modifications.empty())
            return;
        if ( m_modifications.back().m_modifying_constraint < 0 )
            return;
        unsigned cause= m_modifications.back().m_modifying_constraint;
        if (cause != conflicting_constraint)
            return;
        
        unsigned variable= m_modifications.back().m_variable;

        m_var_states[variable].m_latest_modification_entry=
            m_modifications.back().m_previous_modification_entry;
        
        bool was_singleton= m_var_states[variable].m_domain.singleton();
        m_var_states[variable].m_domain=
            m_modifications.back().m_previous_domain;
        
        if (was_singleton)
            lower_var_heap_cost(variable);
        else
            raise_var_heap_cost(variable);

        m_modifications.pop_back();
    }
}

assignment_stack::change_index_t assignment_stack::before_propagation(
    change_index_t propagated_at) const
{
    if (propagated_at == NO_PREVIOUS_MODIFICATION)
        return NO_PREVIOUS_MODIFICATION;
    if (is_earlier_modification(latest_modification(), propagated_at))
        return latest_modification();
    int cause= modifying_constraint(propagated_at);
    if (cause < 0)
        return propagated_at-1;
    change_index_t before;
    for (before= propagated_at-1; NO_PREVIOUS_MODIFICATION != before; --before){
        if (cause != modifying_constraint(before))
            break;
    }
    return before;
}

bool assignment_stack::is_better_variable(int first_variable,
                                          int second_variable,
                                          int factor) const
{
    if (first_variable < 0)
        return false;
    if (second_variable < 0)
        return true;
    const iset & first_domain=  var_domain(first_variable);
    const iset & second_domain= var_domain(second_variable);
    unsigned long long first_size=  first_domain.size();
    unsigned long long second_size= second_domain.size();
    if (first_size == 0&& first_domain.everything())
        first_size= 1LL+iset::MAX_DOMAIN-iset::MIN_DOMAIN;
    if (second_size == 0&& second_domain.everything())
        second_size= 1LL+iset::MAX_DOMAIN-iset::MIN_DOMAIN;
    if ( first_size == 1 )
        return false;
    else if ( second_size == 1 )
        return true;
    
    double first_cost=  m_var_states[first_variable].m_cost;
    double second_cost= m_var_states[second_variable].m_cost;

    // Converted domain/cost comparisons into domain*other_cost to
    // avoid costly divisions.
    double left = round_to_double(first_size*second_cost*factor);
    double right = round_to_double(second_size*first_cost);
    if ( left != right)
        return left < right;
    else if (m_var_states[first_variable].m_last_singleton_time !=
             m_var_states[second_variable].m_last_singleton_time)
        return (m_var_states[first_variable].m_last_singleton_time >
                m_var_states[second_variable].m_last_singleton_time);
    else
        return first_variable < second_variable;
}

void assignment_stack::lower_var_costs()
{
    m_cost_additive *= m_cost_baseline_multiplier;

    static const unsigned REAL_LOWERING_RATIO=1024*1024*1024;
    static const double LOWERING_MULTIPLIER = 1.0/REAL_LOWERING_RATIO;
    if (m_cost_additive > REAL_LOWERING_RATIO) {
        m_cost_additive *= LOWERING_MULTIPLIER;
        for (unsigned i=0 ; i != m_var_states.size() ; ++i)
            m_var_states[i].m_cost *= LOWERING_MULTIPLIER;
    }
}

void assignment_stack::reset_var_costs(double analysis_cost_factor)
{
    m_cost_additive= analysis_cost_factor;
    for (auto & elem : m_var_states) {
        elem.m_cost = 0.0;
        elem.m_last_singleton_time = 0;
    }
}

void assignment_stack::disable_var_costs_heap()
{
    m_var_costs_heap.clear();
    m_var_in_heap.clear();
}

void assignment_stack::enable_var_costs_heap()
{
    using namespace std::placeholders;
    if (!m_var_costs_heap.empty())
        return;
    m_var_costs_heap.resize(m_var_states.size());
    iota(m_var_costs_heap.begin(), m_var_costs_heap.end(), 0);
    make_heap(m_var_costs_heap.begin(), m_var_costs_heap.end(),
              [=](unsigned a, unsigned b)
              {
                  return is_better_variable(b,a);
              });
    m_var_in_heap.resize(m_var_states.size());
    for (unsigned i=0 ; i != m_var_costs_heap.size() ; ++i)
    {
        unsigned var = m_var_costs_heap[i];
        m_var_in_heap[var] = i;
    }
}

void assignment_stack::add_var_cost(unsigned var, double cost)
{ 
    m_var_states[var].m_cost += cost; 
    raise_var_heap_cost(var);
}

void assignment_stack::raise_var_heap_cost(unsigned var)
{
    if (m_var_in_heap.empty())
        return;
    unsigned heap_entry= m_var_in_heap[var];
    while (true)
    {
        // May need to push the element towards the heap leafs.
        unsigned child_entry1 = heap_entry * 2 + 1;
        unsigned child_entry2 = child_entry1 + 1;
        if (child_entry1 >= m_var_costs_heap.size())
            break;
        int child_var1 = m_var_costs_heap[child_entry1];
        int child_var2 = -1;
        if (child_entry2 < m_var_costs_heap.size())
            child_var2 = m_var_costs_heap[child_entry2];
        
        bool candidate1 = is_better_variable(child_var1, var);
        bool candidate2 = is_better_variable(child_var2, var);
        if (!candidate1 && !candidate2)
            break;
        bool choose1 = candidate1 && (
            !candidate2 || is_better_variable(child_var1, child_var2));
        if (choose1)
        {
            swap_heap_entries(child_entry1, heap_entry);
            heap_entry= child_entry1;
        }
        else
        {
            swap_heap_entries(child_entry2, heap_entry);
            heap_entry= child_entry2;
        }
    }
}

void assignment_stack::lower_var_heap_cost(unsigned var)
{
    if (m_var_in_heap.empty())
        return;
    unsigned heap_entry= m_var_in_heap[var];
    while (heap_entry > 0)
    {
        // May need to push the element towards the heap leafs.
        unsigned parent_entry = (heap_entry-1) >> 1;
        unsigned parent_var = m_var_costs_heap[parent_entry];
        if (!is_better_variable(var, parent_var))
            break;
        swap_heap_entries(parent_entry, heap_entry);
        heap_entry= parent_entry;
    }
}

bool assignment_stack::print_singleton_variables(ostream & out) const
{
    if (m_singleton_variable_ids.empty())
        return false;
    
    const char *prefix= "Singletons: ";
    map<iset::value_type, unsigned>::const_iterator it;
    for ( it= m_singleton_variable_ids.begin() ;
          it != m_singleton_variable_ids.end();
          ++it) {
        out << prefix << var_printer(it->second) << "=" << it->first;
        prefix= ", ";
    }
    return true;
}

void assignment_stack::print_var_name(ostream & out, unsigned var_index) const
{
    if (m_names[var_index].empty())
        out << "V" << var_index;
    else
        out << m_names[var_index];
}
