// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <sstream>
#include <big_iset.hpp>
#include "iset_helpers.hpp"


using namespace std;
using namespace hcsp;
using namespace hcsp::iset_helpers;

const big_iset::value_type big_iset::MIN_DOMAIN;
const big_iset::value_type big_iset::MAX_DOMAIN;
const bool big_iset::iterator::CHECKS_BOUNDS;
const bool big_iset::iterator::INVALIDATED_BY_ERASE;

// explicit
big_iset::big_iset(const std::string & text)
{
    assign_set(*this, text);
}

big_iset & big_iset::insert(const value_range_type & inserted_range)
{
    HCSP_PROFILE_CURRENT;
    if ( inserted_range.first == inserted_range.second &&
         in_stride(inserted_range.first) ) {
        size_type prev_size= size();
        m_impl.insert(impl_value(inserted_range.first));
        compact_attempt(prev_size);
        return *this;
    }
    if (m_skip_size == 1) {
        m_impl.insert(inserted_range);
    } else {
        HCSP_PROFILE_LINE;
        flatten();
        m_impl.insert(inserted_range);
        compact();
    } 
    return *this;
}

big_iset & big_iset::insert(const value_range_type & range, unsigned stride)
{
    HCSP_PROFILE_CURRENT;
    if (everything())
        return *this;
    if ( (range.second-range.first)%stride != 0)
        throw runtime_error("Inserting range with a wrong stride");
    if (stride == 1 || range.first == range.second)
        return insert(range);
    // Avoid creating huge strides at this moment, they just confuse matters
    if ( stride > MAX_DOMAIN/2U &&
         (size_type)range.first+stride-range.second==0 ) {
        return insert(range.first).insert(range.second);
    }
    if (size() == 1 && stride != m_skip_size) {
        value_type value= min();
        size_type next_offset= floor_mod(value, stride);
        if (next_offset == (size_type)floor_mod(range.first,stride)) {
            clear();
            m_skip_size= stride;
            m_offset= next_offset;
            insert(value); // put it back, under the new stride info.
        }
    } else if ( empty() ) {
        m_skip_size= stride;
        m_offset= floor_mod(range.first, stride);
    }

    if ( stride == m_skip_size &&
         external_value(impl_value(range.first)) == range.first) {
        size_type prev_size= size();
        m_impl.insert(make_pair(impl_value(range.first),
                                impl_value(range.second)));
        compact_attempt(prev_size);
    } else {
        value_type val=range.first;
        for (; val != range.second; val+=stride)
            insert(val);
        insert(val);
    }
    return *this;
}


big_iset & big_iset::erase(const value_range_type & erased_range)
{
    value_type first=  upper_impl_value(erased_range.first);
    value_type second= lower_impl_value(erased_range.second);
    if (first <= second) {
        size_type prev_size= size();
        m_impl.erase(value_range_type(first, second));
        compact_attempt(prev_size);
    }
    return *this;
}

big_iset::iterator big_iset::erase(iterator erased_iterator)
{
    return iterator(m_impl.erase(erased_iterator.m_iter_impl),
                    this);
}

big_iset & big_iset::intersect(const value_range_type & range)
{
    value_type first=  upper_impl_value(range.first);
    value_type second= lower_impl_value(range.second);
    if (first <= second) {
        size_type prev_size= size();
        m_impl.intersect(value_range_type(first, second));
        compact_attempt(prev_size);
    } else
        clear();
    return *this;
}

big_iset & big_iset::intersect(const value_range_type & range, size_type skip)
{
    if (skip == 1 || range.first >= range.second || empty() )
        return intersect(range);
    if ( skip == 0 || (range.second-range.first)%skip != 0)
        throw runtime_error("Intersecting with range with a wrong stride");
    if ( everything() ) {
        big_iset tmp;
        tmp.insert(range, skip);
        swap(tmp);
        return *this;
    }
    
    intersect(range);
    if (empty())
        return *this;
    size_type offset;
    if (range.first < 0) {
        offset= (-(size_type)range.first)%skip;
        if (offset > 0)
            offset= skip-offset;
    } else {
        offset= range.first%skip;
    }
    
    if (m_skip_size % skip == 0) {
        if (m_offset % skip  != offset)
            clear();
        return *this; // already intersected by 'intersect(range)'
    }
    if (skip % m_skip_size == 0 && m_offset != offset % m_skip_size) {
        clear();
        return *this;
    }
    intersect_different_strides(range, skip);
    return *this;
}

big_iset & big_iset::intersect(const pair<long long, long long> & range,
                               size_type skip)
{
    // 'fixed' must not be pair<long long,long long> in order to avoid
    // infinite recursion when intersect(fixed, skip) is called.
    value_range_type fixed; 
    if ( range.first < iset::MIN_DOMAIN ) {
        long long first= range.first;
        long long diff = iset::MIN_DOMAIN - first;
        // Use unsigned type to avoid possible overflow during round-up.
        // After this division no overflow is possible, so go back to signed.
        long long steps= ((unsigned long long)diff+(skip-1))/skip;
        HCSP_ASSERT(first+steps*skip >= iset::MIN_DOMAIN);
        HCSP_ASSERT(first+(steps-1)*skip <= iset::MIN_DOMAIN);
        fixed.first = first+steps*skip;
    } else {
        fixed.first= range.first;
    }
    
    if ( range.second > iset::MAX_DOMAIN) {
        long long second= range.second;
        long long diff = second - iset::MAX_DOMAIN;
        // Use unsigned type to avoid possible overflow during round-up.
        // After this division no overflow is possible, so go back to signed.
        long long steps= ((unsigned long long)diff+(skip-1))/skip;
        HCSP_ASSERT(second-steps*skip <= iset::MAX_DOMAIN);
        HCSP_ASSERT(second-(steps-1)*skip > iset::MAX_DOMAIN);
        fixed.second = second - steps*skip;
    } else {
        fixed.second= range.second;
    }
    return intersect(fixed, skip);
}

void big_iset::insert_intersection_same_start(const value_range_type & range1,
                                              size_type step1, 
                                              const value_range_type & range2,
                                              size_type step2,
                                              size_type step_gcd)
{
    HCSP_ASSERT(range1.first == range2.first);
    HCSP_ASSERT(((unsigned)range1.second-range1.first) % step1 == 0);
    HCSP_ASSERT(((unsigned)range2.second-range2.first) % step2 == 0);
    value_range_type shared(range1.first,
                            std::min(range1.second, range2.second));
    if (range1.first > shared.second) {
        return;
    }
    if (range1.first == shared.second) {
        insert(range1.first);
        return;
    }
    if (step_gcd == -1U)
        step_gcd= gcd(step1, step2);
    unsigned long long shared_step= step1/step_gcd;
    shared_step *= step2;
    
    unsigned diff= (unsigned)shared.second-range1.first;

    unsigned steps_forward= diff/shared_step;
    if (steps_forward == 0) {
        insert(range1.first);
        return;
    }
    shared.second= steps_forward*shared_step+range1.first;
    insert(shared, shared_step);
}

// static
long long big_iset::num_of_step1_needed(const value_range_type & range1,
                                        size_type step1, 
                                        const value_range_type & range2,
                                        size_type step2,
                                        unsigned & step_gcd)
{
    // To find a range which is the intersection of range1 and range2, with
    // the given step sizes, we need to find smallest x1,x2>=0 that solve the
    // equation:
    //     range1.first+x1*step1 == range2.first+x2*step2
    // or, after canonization:
    //     x1*step1 + -x2*step2 == (range2.first-range1.first)
    long long x1, x2;
    
    unsigned result_step1, result_step2;
    int sign= (range2.first>=range1.first)*2-1;
    unsigned abs_rhs = sign*((unsigned)range2.first-range1.first);
    if (!linear_diophantine_equation(step1, step2, abs_rhs, 
                                     x1, result_step1, x2, result_step2,
                                     &step_gcd))
        return -1;
    HCSP_ASSERT(x1 <= -1U); // guaranteed by linear_diophantine_equation()
    x1 *= sign; x2 *= sign;
    
    if (x1 <0) {
	x1 += result_step1;
        x2 += result_step2;
    }
    
    if (x2 < 0) {
        long long k= -floor_div(x2, result_step2);
        x1 += result_step1*k;
        x2 += result_step2*k; // will not be used... line kept for consistency
    }
    return x1;
}

//static
void big_iset::remove_smaller_elements(value_range_type &range, size_type step,
                                       value_type element)
{
    if (element <= range.first)
        return;
    unsigned diff= (unsigned)element-range.first;
    
    // need to round up, but since (diff+step-1)/step may overflow we don't try
    unsigned steps= diff/step;
    range.first += steps*step;
    if (range.first < element) // may happen since we rounded down
        range.first += step;
}

//static
void big_iset::remove_bigger_elements(value_range_type &range, size_type step,
                                      value_type element)
{
    if (range.second <= element)
        return;
    unsigned diff= (unsigned)range.second-element;
    
    // need to round up, but since (diff+step-1)/step may overflow we don't try
    unsigned steps= diff/step; 
    range.second -= steps*step;
    if (element < range.second) // may happen since we rounded down
        range.second -= step;
}    

//static
big_iset::size_type big_iset::count_elements(const value_range_type &range,
                                             size_type step)
{
    if (range.first > range.second)
        return 0;
    size_type diff= (unsigned)range.second-range.first;
    return diff/step+1;
}

//static
bool big_iset::is_in_range(const value_range_type & range, size_type step,
                           value_type element)
{
    if (element < range.first || range.second < element)
        return false;
    size_type diff= (unsigned)element-range.first;
    return diff % step == 0;
}

bool big_iset::insert_if_simple_intersection(value_range_type & range1,
                                             size_type step1, 
                                             value_range_type & range2,
                                             size_type step2)
{
    remove_smaller_elements(range1, step1, range2.first); 
    remove_bigger_elements(range1, step1, range2.second);
    
    remove_smaller_elements(range2, step2, range1.first); 
    remove_bigger_elements(range2, step2, range1.second);

    if ( range2.first > range2.second || range1.first > range1.second)
        return true; // no overlap.

    if (step1 == 1) {
        insert(range2, step2);
        return true;
    }
    if (step2 == 1) {
        insert(range1, step1);
        return true;
    }
    iset::size_type count1= count_elements(range1, step1);
    if (count1 <= MAX_ELEMENTS_IN_SMALL_RANGE) {
        HCSP_ASSERT(value_type((count1-1)*step1+range1.first) == range1.second);
        iset::value_type element= range1.first;
        for (unsigned i=0 ; i < count1 ; ++i, element += step1)
            if (is_in_range(range2, step2, element))
                insert(element);
        return true;
    }
    iset::size_type count2= count_elements(range2, step2);
    if (count2 <= MAX_ELEMENTS_IN_SMALL_RANGE)
        return insert_if_simple_intersection(range2, step2, range1, step1);
    return false;
}

void big_iset::insert_intersection(value_range_type range1, size_type step1, 
                                   value_range_type  range2, size_type step2)
{
    // May remove trivially irrelevant elements from range1 and range2
    if ( insert_if_simple_intersection(range1, step1, range2, step2) )
        return;
        
    HCSP_ASSERT(step1 > 0 && step2 > 0);
    if (range1.first == range2.first) {
        insert_intersection_same_start(range1, step1, range2, step2);
        return;
    }

    unsigned step_gcd;
    long long num_steps;
    num_steps= num_of_step1_needed(range1, step1, range2, step2, step_gcd);
    if (num_steps < 0) 
        return;                 // no solution found
    
    // It is possible that num_steps*step1 will overflow a signed long long,
    // the following test makes sure that it does not overflow.
    if ( num_steps >= -1U || (unsigned long long)num_steps*step1 > -1U)
        return;
    long long first= range1.first+num_steps*step1;
    if (first > std::min(range1.second,range2.second))
	return;
    
    HCSP_ASSERT(first >= range1.first && first >= range2.first);
    HCSP_ASSERT( (first-range1.first)%step1 == 0);
    HCSP_ASSERT( (first-range2.first)%step2 == 0);
    iset::value_range_type tight_range1=range1, tight_range2= range2;
    tight_range1.first= tight_range2.first= first;

    insert_intersection_same_start(tight_range1, step1, tight_range2, step2,
                                   step_gcd);
}

   
void big_iset::intersect_different_strides(const big_iset & other)
{
    big_iset ret;
    iterator other_it= other.begin();
    iterator it= lower_bound(*other_it);
    if (it != end() && *other_it < *it)
        other_it= other.lower_bound(*it);
    
    value_range_type range, other_range;
    size_type step_size, other_step_size;
    bool advance= true, advance_other= true;
    while ( (!advance || it != end()) &&
            (!advance_other || other_it != other.end())) {
        if (advance)
            it= it.find_stride(end(), range, step_size);
        if (advance_other)
            other_it= other_it.find_stride(other.end(), other_range,
                                           other_step_size);
        ret.insert_intersection(range, step_size, other_range, other_step_size);
        advance_other= range.second >= other_range.second;
        advance=       other_range.second >= range.second;
    }
    swap(ret);
}

void big_iset::intersect_different_strides(const value_range_type & other_range,
                                           size_type other_step_size)
{
    big_iset ret;
    iterator it= lower_bound(other_range.first);
    
    while ( it != end() ) {
        value_range_type range;
        size_type step_size;
        it= it.find_stride(end(), range, step_size);
        ret.insert_intersection(range, step_size, other_range, other_step_size);
    }
    swap(ret);
}

big_iset & big_iset::intersect(const big_iset & other)
{
    if ( other.empty() || empty())
        return clear();
    if ( other.everything() )
        return *this;
    if ( everything() )
        return *this= other;

    size_type prev_size= size();
    intersect(value_range_type(other.min(), other.max()));
    if (empty())
        return *this;
    if ( same_stride(*this, other) ) {
        m_impl.intersect(other.m_impl);
        compact_attempt(prev_size);
        return *this;
    }
    if ( interleaved_stride(*this, other) ) {
        clear();
        return *this;
    }

    intersect_different_strides(other);
    compact_attempt(prev_size);
    
    return *this;
}

big_iset & big_iset::erase_all_of(const big_iset & other)
{
    if ( other.everything() || empty() )
        return clear();
    if ( other.empty() )
        return *this;
    if (interleaved_stride(*this, other))
        return *this;
    size_type prev_size= size();
    if (same_stride(*this, other)) {
        m_impl.erase_all_of(other.m_impl);
        compact_attempt(prev_size);
        return *this;
    }        

    value_type this_max= max();
    iterator other_it;
    for ( other_it= other.lower_bound(min());
          other_it!=other.end() && *other_it<=this_max;
          other_it.advance_range()) {
        erase(other_it.current_range());
    }
    compact_attempt(prev_size);
    return *this;
}

big_iset & big_iset::insert_all_of(const big_iset & other)
{
    if ( other.empty() || everything() )
        return *this;
    if ( other.everything() || empty() )
        return *this= other;
    if (same_stride(*this, other)) {
        size_type prev_size= size();
        m_impl.insert_all_of(other.m_impl);
        compact_attempt(prev_size);
        return *this;
    }
    for ( iterator other_it= other.begin() ;
          other_it != other.end() ;
          other_it.advance_range()) {
        insert(other_it.current_range());
    }
    return *this;
}
big_iset & big_iset::complement()
{
    flatten();
    m_impl.complement();
    compact();
    return *this;
}

bool big_iset::contains_all_of(const value_range_type & range,
                               unsigned skip_size) const
{
    big_iset_impl::iterator it= m_impl.lower_bound(impl_value(range.first));
    if ( range.first != external_value(*it))
        return false;
    if ( range.first == range.second )
        return true;
    if (skip_size % m_skip_size > 0)
        return false;
    value_type prev_covered= range.first;
    for ( ;
          it != m_impl.end() && prev_covered < range.second ;
          it.advance_range() ) {
        value_type next_to_cover= prev_covered+skip_size;
        value_range_type this_range= external_range(it.current_range());
        if (this_range.first > next_to_cover)
            return false;
        if (this_range.second < next_to_cover)
            continue;
        unsigned diff= (external_value(lower_impl_value(this_range.second))-
                        (unsigned)prev_covered)/ skip_size * skip_size;
        prev_covered += diff;
    }
    return prev_covered >= range.second;
}

bool big_iset::contains_all_of(const big_iset & other) const
{
    if ( same_stride(*this, other) )
        return m_impl.contains_all_of(other.m_impl);
    if (other.empty() || everything() )
        return true;
    if (empty() || other.everything())
        return false;
    if (other.size() > size())
        return false;
    if (other.min() < min() || other.max() > max())
        return false;
    
    big_iset_impl::iterator other_it;
    for ( other_it =other.m_impl.begin();
          other_it != other.m_impl.end();
          other_it.advance_range() ) {
        if (!contains_all_of(other.external_range(other_it.current_range()),
                             other.m_skip_size))
            return false;
    }
    return true;
}

bool big_iset::contains_any_of(const big_iset & other) const
{
    if (empty() || other.empty())
        return false;
    if (other.everything() || everything() )
        return true;
    if (other.min() > max() || other.max() < min())
        return false;

    if (same_stride(*this, other))
        return m_impl.contains_any_of(other.m_impl);
    if (interleaved_stride(*this, other))
        return false;
    iterator it=begin(), other_it=other.begin();
    while (true) {
        if (it == end() || other_it == other.end())
            return false;
        if (*it < *other_it) {
            it= lower_bound(*other_it);
            if (it != end() && *it < *other_it)
                ++it;
        }
        else if (*it > *other_it) {
            other_it= other.lower_bound(*it);
            if (other_it != other.end() && *it > *other_it)
                ++other_it;
        }
        else
            return true;
    }
}

bool big_iset::contains(value_type value) const
{
    return in_stride(value) && m_impl.contains(impl_value(value));
}

big_iset::iterator big_iset::iterator::find_stride(iterator end,
                                                   value_range_type &range,
                                                   size_type & step_size) const
{
    big_iset_impl::iterator it;
    it= m_iter_impl.find_stride(end.m_iter_impl, range, step_size);
    if (range.first != range.second)
        step_size *= m_set->m_skip_size;
    range= m_set->external_range(range);
    return iterator(it, m_set);
}

big_iset::iterator big_iset::iterator::find_back_stride(
    iterator begin, value_range_type &range, size_type & step_size) const
{
    big_iset_impl::iterator it;
    it= m_iter_impl.find_back_stride(begin.m_iter_impl, range, step_size);
    if (range.first != range.second)
        step_size *= m_set->m_skip_size;
    else
        step_size = 1;
    range= m_set->external_range(range);
    return iterator(it, m_set);
}

big_iset::iterator big_iset::find_contained_tail(const big_iset & other,
                                                 iterator last) const
{
    return iset_helpers::find_contained_tail(*this, last, other);
}

big_iset::value_type big_iset::rand() const
{
    return external_value(m_impl.rand());
}

void big_iset::validate() const
{
    m_impl.validate();
    if (m_skip_size < m_offset)
        throw runtime_error("big::iset::validate: "
                            "skip size is smaller than offset");
    if (!empty()) {
        if ((long long)m_skip_size * m_impl.min()+(long long)m_offset < MIN_DOMAIN)
            throw runtime_error("big::iset::validate: "
                                "min() is smaller than MIN_DOMAIN");
        if ((long long)m_skip_size * m_impl.max()+(long long)m_offset > MAX_DOMAIN)
            throw runtime_error("big::iset::validate: "
                                "min() is smaller than MIN_DOMAIN");
    }
}

namespace hcsp {
std::ostream & operator<<(std::ostream & out, const big_iset & domain)
{
    print_set(out, domain);
    return out;
}
}

namespace hcsp {
std::istream & operator>>(std::istream & in, big_iset & domain)
{
    read_set(in, domain);
    return in;
}
}

void big_iset::show_self()
{
    show_set(*this);
}

bool big_iset::same_elements(const big_iset &a, const big_iset &b)
{
    if (same_stride(a,b))
        return a.m_impl==b.m_impl;
    for (iterator it_a=a.begin(), it_b=b.begin() ; ; ++it_a, ++it_b) {
        if (it_a == a.end())
            return it_b == b.end();
        if (it_b == b.end())
            return false;
        if (*it_a != *it_b)
            return false;
    }
}

void big_iset::flatten()
{
    if (m_skip_size == 1)
        return;
    big_iset_impl flat;
    for ( iterator it= begin() ; it != end() ; ++it)
        flat.insert(*it);
    m_impl.swap(flat);
    m_offset=0;
    m_skip_size=1;
}

void big_iset::compact()
{
    HCSP_PROFILE_CURRENT;
    m_compactions_skipped= 0;
    // If size() != range_count() then there are at least two values with
    // distance of m_skip_size. It means that m_skip_size can't be lowered.
    if (everything() || size() <= 1 || m_impl.size() != m_impl.range_count())
        return;
    big_iset_impl::iterator it= m_impl.begin();
    value_type prev= external_value(*it);
    ++it;
    value_type current= external_value(*it);
    size_type skip_size= 0U+ current - prev;
    if (skip_size == m_skip_size)
        return;
    if(it != m_impl.end()) {
        for (++it, prev=current; it != m_impl.end() ; ++it, prev=current ){
            current= external_value(*it);
            size_type distance= 0U+ current - prev;
            if (skip_size != distance)
                skip_size= gcd(skip_size, distance);
            if (skip_size == m_skip_size)
                return;
        }
    }
    big_iset dense;
    dense.m_offset= floor_mod(min(), skip_size);
    dense.m_skip_size= skip_size;
    for (it=m_impl.begin(); it != m_impl.end() ; ++it){
        value_range_type range= external_range(it.current_range());
        dense.m_impl.insert(make_pair(dense.impl_value(range.first),
                                      dense.impl_value(range.second)));
    }
    swap(dense);
}
