// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <sstream>
#include <big_iset_impl.hpp>


using namespace std;
using namespace hcsp;

const big_iset_impl::value_type big_iset_impl::MIN_DOMAIN;
const big_iset_impl::value_type big_iset_impl::MAX_DOMAIN;
const bool big_iset_impl::iterator::CHECKS_BOUNDS;
const bool big_iset_impl::iterator::INVALIDATED_BY_ERASE;

//
// Possible results of calls to m_ranges.lower_range_bound(value), are in
// the following interval:
//
//                                         m_ranges.lower_bound(value)
//                                                   |                   
//                                                   v         
// (or MIN_DOMAIN)  [l1...key1...u1]          [l2...key2...u2] (or end() )
//                        ``````````^^^^^^^^^^~~~~~~
//                       possible locations for value
//                       (this is what we look for)
//
// There are three interesting cases marked with ````, ^^^ and ~~~
//
//
big_iset_impl::ranges_set_type::iterator
big_iset_impl::lower_range_bound(value_type value)
{
    ranges_set_type::iterator it= m_ranges.lower_bound(value);
    if (it != m_ranges.end() && it->second.first <= value) {
        // This is the third case (marked with ~~~)
        // Nothing to do. 'it' already points to an interesting range.
    } else {
        // This is either the first case (marked with ````) or the second
        // case (marked with ^^^).
        if (it == m_ranges.begin())
            return it;          // The second case (marked with ^^^).
        ranges_set_type::iterator previous_it= it;
        --previous_it;
        if (previous_it->second.second < value)
            return it;          // The second case (marked with ^^^).
        
        // This is the first case (marked with ````).
        it= previous_it;
        // Fall through
    }
    HCSP_ASSERT(it->second.first <= value && value <= it->second.second);
    return it;
}

big_iset_impl &
big_iset_impl::insert_independent(const value_range_type & inserted_range,
                                  ranges_set_type::iterator where)
{
    value_type middle= middle_of(inserted_range);
    m_ranges.insert(where, ranges_set_type::value_type(middle,inserted_range));

    // It is only now that it is safe to modify size without the danger of
    // exceptions.
    m_size+= distance_between(inserted_range.first, inserted_range.second)+1;

    return *this;
}

// static
inline bool big_iset_impl::ranges_touch(const value_range_type & a,
                                const value_range_type &b)
{
    // Check if ranges barely touch, or not overlap at all.
    if ( a.second < b.first )
        return a.second+1 == b.first;
    if ( b.second < a.first )
        return b.second+1 == a.first;

    return true;
}

bool big_iset_impl::range_touches_iterator(const value_range_type &range,
                                   const ranges_set_type::iterator & it) const
{
    if (it == m_ranges.end())
        return false;
    else
        return ranges_touch(range, it->second);
}


bool big_iset_impl::range_overlaps_iterator(const value_range_type &range,
                                    ranges_set_type::const_iterator it) const
{
    if (it == m_ranges.end())
        return false;
    else
        return ranges_overlap(range, it->second);
}


big_iset_impl &
big_iset_impl::merge_at_iterator(const value_range_type & inserted_range,
                                 ranges_set_type::iterator where)
{
    HCSP_ASSERT(range_touches_iterator(inserted_range, where));
    
    m_size -= range_size(where->second);
    value_range_type & result_range = where->second;
    result_range.first=  std::min(result_range.first, inserted_range.first);
    result_range.second= std::max(result_range.second, inserted_range.second);
    ranges_set_type::iterator it= where;
    ranges_set_type::iterator next_it= ++it;
    
    for ( ; range_touches_iterator(result_range, it) ; ++it ) {
        result_range.first=  std::min(result_range.first, it->second.first);
        result_range.second= std::max(result_range.second, it->second.second);
        m_size -= range_size(it->second);
    }
    m_ranges.erase( next_it, it);

    it= where;
    while ( it!=m_ranges.begin() ) {
        --it;
        if ( !ranges_touch(it->second, result_range) ) {
            ++it;
            break;
        }
        result_range.first=  std::min(result_range.first, it->second.first);
        result_range.second= std::max(result_range.second, it->second.second);
        m_size -= range_size(it->second);
    }
    m_ranges.erase( it, where );
            
    m_size += range_size(result_range);

    return *this;
}

big_iset_impl & big_iset_impl::insert(const value_range_type & inserted_range)
{
    HCSP_PROFILE_CURRENT;
    assert_valid_range(inserted_range, "big_iset_impl::insert");

    ranges_set_type::iterator it= m_ranges.lower_bound(inserted_range.first);
    // 'it' can be either on 'inserted_range' or after, but never before
    if (it != m_ranges.end()) {
        // the above statement is asserted here:
        HCSP_ASSERT(! (it->second.second < inserted_range.first) ); 
        if ( range_subsumed(inserted_range, it->second) )
            return *this;
        
        if ( ranges_touch(inserted_range, it->second) )
            return merge_at_iterator(inserted_range, it);
    }
    if ( it != m_ranges.begin()) {
        --it;
        if ( range_subsumed(inserted_range, it->second) )
            return *this;
        if ( ranges_touch(inserted_range, it->second) )
            return merge_at_iterator(inserted_range, it);
    }
    return insert_independent(inserted_range, it /* hint */ );
}

big_iset_impl &
big_iset_impl::erase_from_single_iterator(ranges_set_type::iterator where,
                                          const value_range_type erased_range)
{
    HCSP_ASSERT(range_overlaps_iterator(erased_range, where));
    // Should the upper part of 'where' be kept intact?
    value_range_type where_bounds = where->second;
    if (erased_range.second < where_bounds.second) {
        where= recenter_iterator_around(where, erased_range.second+1,
                                        where_bounds.second);
        size_type orig_iterators_size= range_size(where->second);

        // Should the lower part of 'where' be kept intact?

        auto a = where->second.first;
        auto b = erased_range.first;
        if (a < b && b != numeric_limits<decltype(b)>::min()) {
            value_range_type reinserted_range(a, b-1);
            value_type middle= middle_of(reinserted_range);
            // If the insertion throws now, then no visible change happened.
            m_ranges.insert(
                where, ranges_set_type::value_type(middle,reinserted_range));
            // Note that at this point, the INVARIANT IS BROKEN.
            m_size+= range_size(reinserted_range); // unsigned, OK if overflows
        }
        where->second.first= erased_range.second+1;
        where_bounds = where->second;
        HCSP_ASSERT(where->second.first <= where->first);
        m_size -= orig_iterators_size;
        m_size += range_size(where->second);
    }
    // Should the lower part of 'where' be kept intact?
    else if (where->second.first < erased_range.first) {
        // At this moment we know that the upper part should be removed.
        where= recenter_iterator_around(where, where_bounds.first,
                                        erased_range.first-1);
        m_size -= range_size(erased_range.first, where->second.second);
        where->second.second= erased_range.first-1;
        where_bounds = where->second;
    } else {
        // Nothing is kept intact. The whole iterator  should be removed.
        HCSP_ASSERT(range_subsumes_iterator(erased_range, where));
        m_size -= range_size(where->second);
        m_ranges.erase(where);
    }
    return *this;
}

    
big_iset_impl & big_iset_impl::erase(const value_range_type & erased_range)
{
    assert_valid_range(erased_range, "big_iset_impl::erase");

    ranges_set_type::iterator from = lower_range_bound(erased_range.first);
    ranges_set_type::iterator to = from;
    while ( to != m_ranges.end() && range_overlaps_iterator(erased_range, to) )
        ++to;

    if (from == to)
        return *this;
    ranges_set_type::iterator back_it= to;
    --back_it;
    if ( from == back_it )
        return erase_from_single_iterator(from, erased_range);
    
    from=   recenter_iterator_below(from,   erased_range.first);
    back_it= recenter_iterator_above(back_it, erased_range.second);

    if ( !range_subsumes_iterator(erased_range, from) ) {
        m_size -= range_size(erased_range.first, from->second.second);
        from->second.second= erased_range.first-1;
        ++from;
    }

    if ( !range_subsumes_iterator(erased_range, back_it) ) {
        m_size -= range_size(back_it->second.first, erased_range.second);
        back_it->second.first= erased_range.second+1;
        --to;
    }
    ranges_set_type::iterator it= from;
    while (it != to) {
        m_size -= range_size(it->second);
        m_ranges.erase(it++);
    }
    return *this;
}

big_iset_impl::iterator big_iset_impl::erase(iterator erased_iterator)
{
    if (erased_iterator == end())
        return erased_iterator;
    value_type value= *erased_iterator;

    ranges_set_type::iterator where= erased_iterator.m_range_iterator;
    ranges_set_type::iterator next= where;
    ++next;
    if (value < where->second.second ) {
        where= recenter_iterator_above(where, value);
        size_type orig_iterators_size= range_size(where->second);

        // Should the lower part of 'where' be kept intact?
        if (where->second.first < value) {
            value_range_type reinserted_range(where->second.first, value-1);
            value_type middle= middle_of(reinserted_range);
            // If the insertion throws now, then no visible change happened.
            m_ranges.insert(
                where, ranges_set_type::value_type(middle,reinserted_range));
            // Note that at this point, the INVARIANT IS BROKEN.
            m_size+= range_size(reinserted_range); // unsigned, OK if overflows
        }
        where->second.first= value+1;
        HCSP_ASSERT(where->second.first <= where->first);
        m_size -= orig_iterators_size;
        m_size += range_size(where->second);
        return iterator(where, value+1);
    }
    // Should the lower part of 'where' be kept intact?
    else if (where->second.first < value) {
        // At this moment we know that the upper part should be removed.
        where= recenter_iterator_below(where, value);
        m_size -= range_size(value, where->second.second);
        where->second.second= value-1;
    } else {
        // Nothing is kept intact. The whole iterator  should be removed.
        m_size -= range_size(where->second);
        m_ranges.erase(where);
    }
    return iterator(next, MIN_DOMAIN);
}

big_iset_impl & big_iset_impl::intersect(const value_range_type & range)
{
    ranges_set_type::iterator from= lower_range_bound(range.first);
    ranges_set_type::iterator to=   upper_range_bound(range.second);

    if ( from == to ) {
        clear();
        return *this;
    }
    ranges_set_type::iterator back_it= to;
    --back_it;
    if ( from == back_it ) {
        from= recenter_iterator_around(from, range.first, range.second);
        back_it= from;
    } else {
        from=    recenter_iterator_around(from,    range.first, range.second);
        back_it= recenter_iterator_around(back_it, range.first, range.second);
    }

    m_size -= range_size(back_it->second);
    back_it->second.second= ::min(back_it->second.second, range.second); 
    m_size += range_size(back_it->second);

    m_size -= range_size(from->second);
    from->second.first= :: max(from->second.first, range.first);
    m_size += range_size(from->second);
    
    ranges_set_type::iterator it= m_ranges.begin();
    while ( it != from){
        m_size -= range_size(it->second);
        m_ranges.erase(it++);
    }

    it= to;
    while (it != m_ranges.end()) {
        m_size -= range_size(it->second);
        m_ranges.erase(it++);
    }
    return *this;
}

big_iset_impl & big_iset_impl::intersect(const big_iset_impl & other)
{
    if ( other.empty() || empty())
        return clear();
    if ( other.everything() )
        return *this;
    if ( everything() )
        return *this= other;
    
    intersect(value_range_type(other.min(), other.max()));
    ranges_set_type::const_iterator other_it= other.m_ranges.begin();
    value_type last_in= other_it->second.second;
    while (++other_it != other.m_ranges.end()) {
        erase(value_range_type(last_in+1, other_it->second.first-1));
        last_in= other_it->second.second;
    }
    
    return *this;
}

big_iset_impl & big_iset_impl::erase_all_of(const big_iset_impl & other)
{
    if ( other.everything() || empty() )
        return clear();
    if ( other.empty() )
        return *this;

    value_type this_max= max();
    ranges_set_type::iterator other_it;
    for ( other_it= const_cast<big_iset_impl&>(other).lower_range_bound(min());
          other_it!=other.m_ranges.end() && other_it->second.first<=this_max;
          ++other_it) {
        erase(other_it->second);
    }
    return *this;
}

big_iset_impl & big_iset_impl::insert_all_of(const big_iset_impl & other)
{
    if ( other.empty() || everything() )
        return *this;
    if ( other.everything() || empty() )
        return *this= other;

    ranges_set_type::const_iterator other_it;
    for ( other_it= other.m_ranges.begin() ;
          other_it!=other.m_ranges.end() ;
          ++other_it) {
        insert(other_it->second);
    }
    return *this;
}

big_iset_impl & big_iset_impl::complement()
{
    if (empty())
        return insert(value_range_type(MIN_DOMAIN, MAX_DOMAIN));
    if (everything())
        return clear();
    
    ranges_set_type result_ranges;
    size_type result_size=0;
    
    ranges_set_type::const_iterator it= m_ranges.begin();
    if (it->second.first != MIN_DOMAIN) {
        value_range_type range(MIN_DOMAIN, it->second.first-1);
        result_ranges[middle_of(range)] = range;
        result_size += range_size(MIN_DOMAIN, it->second.first-1);
    }
    value_type last_range= it->second.second;
    
    while (++it != m_ranges.end()) {
        value_range_type range(last_range+1, it->second.first-1);
        result_ranges[middle_of(range)]= range;
        result_size += range_size(last_range+1, it->second.first-1);
        last_range= it->second.second;        
    }
    if (last_range != MAX_DOMAIN) {
        value_range_type range(last_range+1, MAX_DOMAIN);
        result_ranges[middle_of(range)]= range;
        result_size += range_size(last_range+1, MAX_DOMAIN);
    }
    m_ranges.swap(result_ranges);
    ::swap(result_size, m_size);
    return *this;
}

bool big_iset_impl::contains_all_of(const big_iset_impl & other) const
{
    if (other.empty() || everything() )
        return true;
    if (empty() || other.everything())
        return false;
    if (other.size() > size())
        return false;
    if (other.min() < min() || other.max() > max())
        return false;
    ranges_set_type::const_iterator other_it= other.m_ranges.begin();
    for ( ; other_it != other.m_ranges.end() ; ++other_it) {
        big_iset_impl * mutable_this= const_cast<big_iset_impl*>(this);
        ranges_set_type::const_iterator it=
            mutable_this->lower_range_bound(other_it->second.first);
        if (!range_subsumed_by_iterator(other_it->second, it))
            return false;
    }
    return true;
}

bool big_iset_impl::contains_any_of(const big_iset_impl & other) const
{
    if (empty() || other.empty())
        return false;
    if (other.everything() || everything() )
        return true;
    if (other.min() > max() || other.max() < min())
        return false;
    ranges_set_type::const_iterator other_it= other.m_ranges.begin();
    for ( ; other_it != other.m_ranges.end() ; ++other_it) {
        big_iset_impl * mutable_this= const_cast<big_iset_impl*>(this);
        ranges_set_type::const_iterator it
            = mutable_this->lower_range_bound(other_it->second.first);
        if (range_overlaps_iterator(other_it->second, it))
            return true;
    }
    return false;
}

bool big_iset_impl::contains(value_type value) const
{
    ranges_set_type::iterator it=
        const_cast<big_iset_impl&>(*this).lower_range_bound(value);
    if ( it == m_ranges.end() )
        return false;
    return (value >= it->second.first);
}

big_iset_impl::iterator big_iset_impl::iterator::find_stride(iterator end,
                                                   value_range_type &range,
                                                   size_type & step_size) const
{
    step_size= 1;
    if ( m_range_iterator == end.m_range_iterator ){
        if (m_current_value < end.m_current_value) {
            HCSP_ASSERT(end.middle_of_range());
            range.first= **this;
            range.second= end.m_current_value-1;
            return end;
        } else
            throw runtime_error("An empty stride was requested");
    }
    range= current_range();
    iterator next= *this;
    next.advance_range();
    if (range.first != range.second)
        return next;

    if (next == end)
        return next;

    range.second= *next;
    step_size= range_diff(range);

    do {
        range.second= *next;
        ++next;
    } while (next != end && range_diff(range.second, *next) == step_size);
    
    return next;
}

big_iset_impl::iterator big_iset_impl::iterator::find_back_stride(
    iterator begin, value_range_type &range, size_type & step_size) const
{
    if ( *this == begin )
        throw runtime_error("An Empty stride was requested");

    iterator first= *this;
    first.retract_range();
    range= first.current_range();
    if (middle_of_range())
        range.second= **this-1;
    if ( first.m_range_iterator == begin.m_range_iterator &&
         begin.middle_of_range()) {
        range.first= *begin;
        first= begin;
    }
    step_size= 1;
    if (range.first != range.second || first == begin)
        return first;

    --first;
    range.first= *first;
    step_size= range_diff(range);
    while (first != begin) {
        --first;
        if (*first + step_size != (size_type) range.first) {
            ++first;
            return first;
        }
        range.first= *first;
    }
    return first;
}

const big_iset_impl::iterator &
big_iset_impl::iterator::advance(size_type count)
{
    if (count == 0)
        return *this;
    while (count > 0) {
        const value_range_type & range= m_range_iterator->second;
        if ( range.first == MIN_DOMAIN && range.second == MAX_DOMAIN &&
             m_current_value == MIN_DOMAIN)
            break;
        size_type size;
        size= range_size(**this, range.second);
        if (count < size)
            break;
        count -= size;
        advance_range();
    }
    if (count > 0)
        m_current_value= **this + count;
    return *this;
}

big_iset_impl::value_type big_iset_impl::rand() const
{
    if (everything())
        return hcsp::random::range(MIN_DOMAIN, MAX_DOMAIN);
    if (empty())
        throw std::runtime_error(
            "Can't hcsp::big_iset::rand() for an empty set");
    iterator it= begin();
    it.advance(hcsp::random::gen(size()));
    return *it;
}

big_iset_impl::ranges_set_type::iterator
big_iset_impl::recenter_iterator_around(ranges_set_type::iterator where,
                                value_type from, value_type to)
{
    if ( !range_overlaps_iterator(value_range_type(from,to), where) )
        return where;
    if (from <= where->first && where->first <= to)
        return where;           // no adjustment is needed, already in range
    from= ::max(where->second.first,  from);
    to=   ::min(where->second.second, to);

    ranges_set_type::iterator result;
    value_type middle= middle_of(value_range_type(from, to));
    result= m_ranges.insert(ranges_set_type::value_type(middle,
                                                        where->second)).first;
    m_ranges.erase(where);
    return result;
}

void big_iset_impl::validate() const
{
    size_type accumulated=0;
    ranges_set_type::const_iterator it= m_ranges.begin();
    ranges_set_type::const_iterator slow_it= it;
    value_type last_end=0;
    for (  unsigned i=0 ; it != m_ranges.end() ; ++i, ++it) {
        if (i != 0 && it == slow_it)
            throw runtime_error("big_iset::validate: found cycle in m_ranges");
        assert_valid_range(it->second, "big_iset_impl::validate");
        if (it->first < it->second.first || it->second.second < it->first)
            throw runtime_error("big_iset::validate: key not in the middle");
        if ( i !=0 )
            if (last_end >= it->second.first || last_end+1 == it->second.first)
                throw runtime_error("big_iset::validate: ranges touch");
        last_end= it->second.second;
                
        accumulated+= range_size(it->second);
        if (i %2 == 1)
            ++slow_it;
    }
    if (accumulated != m_size)
        throw runtime_error(
            "big_iset::validate: m_size does not match m_ranges");
}

bool
big_iset_impl::same_ranges(const ranges_set_type &a, const ranges_set_type &b)
{
    if (a.size() != b.size())
        return false;
    for ( ranges_set_type::const_iterator a_it= a.begin(), b_it= b.begin();
          a_it != a.end() ;
          ++b_it, ++a_it ) {
        HCSP_ASSERT(b_it != b.end());
        if (a_it->second != b_it->second)
            return false;
    }
    return true;
}

void
big_iset_impl::throw_out_of_range(const char *text) const throw(out_of_range)
{
    ostringstream out;
    out << "iset out of range " << text;
    throw out_of_range(out.str());
}

