// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <algorithm>
#include <iset.hpp>
#include <clause_data.hpp>

using namespace std;
using namespace hcsp;

// static
void literal_data::print_right_hand_side(ostream & out, const iset & set)
{
    if (set.size() == 1) {
        out << "=" << set;
        return;
    }
    if (set.empty()) {
        out << ":{}";
        return;
    }

    if (set.begin().advance_range() == set.end()) {
        // This is a single range
        iset::value_range_type range= set.begin().current_range();
        if (range.first == iset::MIN_DOMAIN)
            out << "<=" << range.second;
        else if (range.second == iset::MAX_DOMAIN)
            out << ">=" << range.first;
        else
            out << ":{" << set << "}";
        return;
    }

    // Check if the set is a simple inequality, i.e., it has the form:
    // [MIN_DOMAIN..value-1],[value+1..MAX_DOIMAIN]
    iset::iterator it= set.begin();
    iset::value_range_type first_range= it.current_range();
    if ( first_range.first == iset::MIN_DOMAIN &&
         it.advance_range() != set.end()) {
        iset::value_range_type second_range= it.current_range();
        if ( second_range.second == iset::MAX_DOMAIN &&
             first_range.second+1 == second_range.first-1) {
            out << "!=" << first_range.second+1;
            return;
        }
    }
    out << ":{" << set << "}";
}

// static
void literal_data::print(ostream &out, const variable_printer &printer,
                         const iset & set)
{
    out << printer;
    print_right_hand_side(out, set);
}

bool clause_data::equivalent(const clause_data & other) const
{
    if (size() < other.size())
        return other.equivalent(*this);
    if (other.size() <= 4) {
        // make it faster for trivial cases of difference
        for (unsigned i=0 ; i != size() ; ++i) {
            if ((*this)[i].m_set.empty())
                continue;
            bool var_found= false;
            for (unsigned j=0 ; j != other.size() && !var_found ; ++j)
                if ( (*this)[i].m_variable == other[j].m_variable)
                    var_found= true;
            if (!var_found)
                return false;
        }
    }
    
    clause_data this_canonical= *this;
    this_canonical.canonicalize();
    
    clause_data other_canonical= other;
    other_canonical.canonicalize();

    return this_canonical == other_canonical;
}

void clause_data::canonicalize()
{
    sort();
    // Merge adjacent literals that refer to a common variable.
    // The set of the first literal is emptied.
    for (unsigned i= 1 ; i < size() ; ++i) {
        if ((*this)[i-1].m_variable != (*this)[i].m_variable)
            continue;
        (*this)[i].m_set.insert_all_of((*this)[i-1].m_set);
        (*this)[i-1].m_set.clear();
    }
    // Erase all literals with empty sets, preserving order.
    unsigned write, read;
    for (read= 0, write=0 ; read != size() ; ++read) {
        if ( !(*this)[read].m_set.empty() )
            (*this)[write++]= (*this)[read];
    }
    erase(begin()+write, end());
}

void clause_data::sort()
{
    std::sort(begin(), end());
}

void clause_data::print(std::ostream & out,
                        const variable_print_strategy & strategy) const
{
    out << print_prefix();
    for (unsigned i=0 ; i!=size() ; ++i) {
        if (i != 0)
            out << print_separator();
        const literal_data & literal= (*this)[i];
        literal_data::print(out, strategy.printer(literal.m_variable),
                            literal.m_set);
    }
    out << print_suffix();
}
