// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <clause_propagator.hpp>
#include <clause_data.hpp>
#include <disjunction_specification.hpp>
using namespace std;
using namespace hcsp;

clause_propagator::result_type
clause_propagation_algo::propagate( constraint_runtime_info & info)
{
    HCSP_PROFILE_CURRENT;
    unsigned known_marks= (m_watch[0].is_valid_mark() +
                           m_watch[1].is_valid_mark());
    if ( known_marks == 0 )
        return UNSAT;
    if ( known_marks == 1 ) {
        HCSP_ASSERT(m_watch[0].is_valid_mark());
        if (info.querying_domain())
            return refresh_watch(info, 0);
        else
            return unit_propagate(info, m_watch[0].m_param);
    }
    const iset * p_domain[NUM_OF_WATCHES];
    const iset * p_literal[NUM_OF_WATCHES]={0,0};
    for (unsigned i=0 ; i != NUM_OF_WATCHES ; ++i) {
        HCSP_ASSERT(m_watch[i].is_valid_mark());
        unsigned param_index= m_watch[i].m_param;
        p_literal[i]= &m_literals[param_index];
        p_domain[i]= &info.domain(param_index);
        if ( p_literal[i]->contains_all_of(*p_domain[i]) )
            return VALID;
    }
    bool is_valid= false;
    for (unsigned i=0 ; i != NUM_OF_WATCHES ; ++i) {
        if ( p_literal[i] && !p_domain[i]->contains(m_watch[i].m_value) ) {
            result_type refresh_result= refresh_watch(info, i);
                if (refresh_result == UNSAT)
                    return refresh_result;
                if (refresh_result == VALID) {
                    if (!info.querying_domain())
                        return VALID;
                    is_valid= true;
                }
        }
    } 
    return is_valid ? VALID: CONSISTENT;
}

bool
clause_propagation_algo::conflicts(const constraint_runtime_info & info) const
{
    HCSP_ASSERT(info.arity() == m_literals.size());
    for (unsigned i=0 ; i != NUM_OF_WATCHES ; ++i) {
        if (!m_watch[i].is_valid_mark())
            continue;
        unsigned param_index= m_watch[i].m_param;
        if (m_literals[param_index].contains_any_of(info.domain(param_index)))
            return false;
    }
    for (unsigned i=0 ; i != m_literals.size() ; ++i) {
        if (m_literals[i].contains_any_of(info.domain(i)))
            return false;
    }
    return true;
}

bool clause_propagation_algo::constraint_check(
    const constraint_runtime_info & info) const
{
    HCSP_ASSERT(m_literals.size() == info.arity());
    for (unsigned i=0 ; i != m_literals.size() ; ++i) {
        if (m_literals[i].contains_any_of(info.domain(i)))
            return true;
    }
    return false;
}

void clause_propagation_algo::print(std::ostream & out,
                                    const constraint_runtime_info * p_info)
    const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        cout << " ";
    }
    out << clause_data::print_prefix();
    for (unsigned i=0 ; i != m_literals.size() ; ++i) {
        if (i != 0)
            out << clause_data::print_separator();
        const value_watch_info_type * entry= NULL;
        for (unsigned j= 0 ; j != NUM_OF_WATCHES ; ++j)
            if ( m_watch[j].is_valid_mark() &&
                 m_watch[j].m_param+0U == i ) {
                entry= &m_watch[j];
            }
        if (entry) {
            out << "*";
            if (entry->is_valid_watch())
                out << entry->m_value << "?";
        }
        if (p_info)
            p_info->print_param(out, i);
        else
            out << "P" << i;
        literal_data::print_right_hand_side(out, m_literals[i]);
    }
    out << clause_data::print_suffix();
}

clause_propagator::result_type
clause_propagation_algo::refresh_watch(constraint_runtime_info & info,
                                       unsigned watch_index)
{
    HCSP_PROFILE_CURRENT;
    HCSP_ASSERT(watch_index <= 1);
    HCSP_ASSERT(m_watch[watch_index].is_valid_mark());
    
    unsigned requested_watched= m_watch[watch_index].m_param;

    if ( m_literals[requested_watched].contains_any_of(
             info.domain(requested_watched)) ) {
        change_watch(info, watch_index, requested_watched);
        if (!m_watch[1-watch_index].is_valid_mark())
            return unit_propagate(info, requested_watched);
        return CONSISTENT;
    }
    unsigned other_watched= m_watch[1-watch_index].m_param;
    for ( unsigned i= (requested_watched+1)%m_literals.size() ;
          i != requested_watched ;
          i = (i+1 == m_literals.size()) ? 0 : i+1 ) {
        if (i == other_watched)
            continue;
        const iset & domain  = info.domain(i);
        const iset & literal = m_literals[i];
        if ( literal.contains_any_of(domain) ) {
            if ( literal.contains_all_of(domain) ) {
                return VALID;
            }
            change_watch(info, watch_index, i);
            return CONSISTENT;
        }
    }
    if (!m_watch[1-watch_index].is_valid_mark())
        return UNSAT;
    // Can't find a valid place for this watch. We are left only with one
    // watch (which could also be invalid).
    result_type result= unit_propagate(info, other_watched);
    if (result == UNSAT)
        return result;
    if ( info.querying_domain() &&
         !info.domain(other_watched).contains( m_watch[1-watch_index].m_value))
        change_watch(info, 1-watch_index, other_watched);
    return VALID;
}

bool clause_propagation_algo::init( constraint_runtime_info & info)
{
    HCSP_PROFILE_CURRENT;
    HCSP_ASSERT(info.arity() == (unsigned)count_params());
    m_watch[0].invalidate_mark();
    m_watch[1].invalidate_mark();
    if (info.arity() != m_literals.size())
        throw runtime_error("Number of variables and literals differ");
    if ( m_literals.empty() )
        return false;
    int first_overlap= -1;
    int second_overlap= -1;
    if (!info.querying_domain()) {
        for (unsigned i=0 ; i != m_literals.size() ; ++i) {
            if ( m_literals[i].contains_all_of(info.domain(i)) ) {
                m_watch[0].m_param= i;
                return true;
            }
        }
    }
    for (unsigned i=0 ; i != m_literals.size() && second_overlap < 0 ; ++i) {
        const iset & domain = info.domain(i);
        if ( !m_literals[i].contains_any_of(domain) )
            continue;
        if ( first_overlap < 0)
            first_overlap=  i;
        else
            second_overlap= i;
    }
    if (first_overlap < 0)
        return false;
    if (first_overlap >= 0 && second_overlap >= 0) {
        add_watch(info, first_overlap);
        add_watch(info, second_overlap);
    } else if (first_overlap >= 0) {
        if (info.querying_domain())  {
            add_watch(info, first_overlap);
            for (unsigned i=0 ; i != m_literals.size(); ++i) {
                if (i == (unsigned)first_overlap || m_literals[i].empty())
                    continue;
                iset::value_type value= m_literals[i].min();
                m_watch[1].m_watch_handle= info.allocate_value_watch(i, value);
                m_watch[1].m_param= i;
                m_watch[1].m_value= value;
                break;
            }
        } else
            m_watch[0].m_param= first_overlap;
    } else
        HCSP_ASSERT(second_overlap < 0);
    return true;
}
    
clause_propagator::result_type
clause_propagation_algo::unit_propagate(constraint_runtime_info & info,
                                  unsigned literal_index)
{
    HCSP_PROFILE_CURRENT;
    const iset & literal= m_literals[literal_index];
    const iset & domain= info.domain(literal_index);
    if (!literal.contains_all_of(domain)) {
        iset next_domain = domain;
        next_domain.intersect(literal);
        if (next_domain.empty())
            return UNSAT;
            
        info.set_domain(literal_index, move(next_domain));
    }
    return VALID;
}

iset::value_type clause_propagation_algo::get_best_watch_value(
    const constraint_runtime_info & info, unsigned param_index,
    iset::value_type hint) const
{
    const iset & domain= info.domain(param_index);
    const iset & literal= m_literals[param_index];
    // Try to avoid running intersect() at first, it may be expensive when
    // the domains are large.
    iset::iterator it= literal.lower_bound(hint);
    if (it != literal.end() && domain.contains(*it))
        return *it;
    it= domain.lower_bound(hint);
    if (it != domain.end() && literal.contains(*it))
        return *it;
    
    iset candidate_values= literal;
    candidate_values.intersect( domain );
    HCSP_ASSERT( !candidate_values.empty() );
    iset::iterator above_hint= candidate_values.lower_bound(hint);
    if (above_hint == candidate_values.end())
        return candidate_values.min();
    else
        return *above_hint;
}

void clause_propagation_algo::change_watch(constraint_runtime_info & info,
                                           unsigned watch_index,
                                           unsigned new_param_index)
{
    HCSP_ASSERT(m_watch[watch_index].is_valid_watch());
    iset::value_type hint= m_watch[watch_index].m_value;
    remove_watch(info, watch_index);
    add_watch(info, new_param_index, hint);
}

void clause_propagation_algo::remove_watch(constraint_runtime_info & info,
                                           unsigned watch_index)
{
    HCSP_ASSERT(m_watch[watch_index].is_valid_watch());
    HCSP_ASSERT( m_watch[watch_index].m_param >=0 );
    info.deallocate_value_watch(m_watch[watch_index].m_param,
                                m_watch[watch_index].m_value,
                                m_watch[watch_index].m_watch_handle);
    m_watch[watch_index].m_param= -1;
}

void clause_propagation_algo::add_watch(constraint_runtime_info & info,
                                        unsigned param_index,
                                        iset::value_type hint)
{
    HCSP_PROFILE_CURRENT;
    unsigned next_watch= 0;
    if ( m_watch[0].m_param >= 0 )
        next_watch= 1;
    
    HCSP_ASSERT( m_watch[next_watch].m_param < 0 ); // Must have free watches

    iset::value_type value= get_best_watch_value(info, param_index, hint);
    m_watch[next_watch].m_watch_handle=
        info.allocate_value_watch(param_index, value);
    m_watch[next_watch].m_param= param_index;
    m_watch[next_watch].m_value= value;
}
    
void clause_propagation_algo::throw_modifying_clause_at_runtime()
    throw(runtime_error)
{
    throw runtime_error("Modifying the clause at runtime");
}

// virtual
clause_propagator * clause_propagator::intersect_with(
    const constraint_runtime_info & info,
    vector<unsigned> & copy_from_argument) const
{
    HCSP_ASSERT(info.arity() == literals().size());
    const literals_type & in_literals= literals();
    unsigned i;
    for (i=0; i != in_literals.size() ; ++i) {
        if (!info.domain(i).contains_all_of(in_literals[i]))
            break;
    }
    if (i == in_literals.size())
        return NULL;
    copy_from_argument.clear();
    unique_ptr<clause_propagator> ret(clone());
    ret->forget_watches();
    literals_type & out_literals= ret->update_literals();
    HCSP_ASSERT(out_literals == in_literals);
    bool has_empty= -1U;
    for ( ; i != out_literals.size() ; ++i) {
        out_literals[i].intersect(info.domain(i));
        if (has_empty == false && out_literals[i].empty())
            has_empty= true;;
    }
    if (!has_empty)
        return ret.release();
    for (i=0; i != out_literals.size() ; ++i) {
        if (out_literals[i].empty())
            continue;
        out_literals[i].swap(out_literals[copy_from_argument.size()]);
        copy_from_argument.push_back(i);
    }
    out_literals.erase(out_literals.begin()+copy_from_argument.size(),
                       out_literals.end());
    return ret.release();
}

// virtual
bool clause_propagator::potentially_falsifying(
    const constraint_runtime_info & info, unsigned param) const
{
    HCSP_ASSERT(info.arity() == (unsigned)count_params());
    const iset & domain = info.domain(param);
    const iset & literal= literals()[param];
    return literal.contains_any_of(domain) && !literal.contains_all_of(domain);
}

// virtual
bool clause_propagator::get_disjunction_specification(
    disjunction_specification& disjunction) const
{
    disjunction.set_literals(&literals()[0], 0, literals().size());
    disjunction.reset_constraint(NULL, 0);
    return true;
}

// Perform combine the other way around. Conceptually, calling
// clause.combine(other) attempts to call other.combine(clause).
// For that it has to change the order of values in other_to_common_param,
// other_pivot_param, and copy_from_argument accordingly.

// virtual
constraint_propagator *clause_propagator::combine(
    const constraint_runtime_info & info,
    const constraint_propagator & other,
    const constraint_runtime_info & other_info,
    const std::vector<unsigned> & other_to_common_param,
    unsigned other_pivot_param,
    std::vector<unsigned> & copy_from_argument) const
{
    HCSP_ASSERT(!dynamic_cast<const clause_propagator*>(&other));
    return combine_other_first(info, other, other_info, other_to_common_param,
                               other_pivot_param, copy_from_argument);
}
