// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2012,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2012,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <comparison_propagators.hpp>

#include <iostream>
#include <limits>

#include <constraint_watchless_info.hpp>
#include <constraint_explicit_info.hpp>
#include <sliced_runtime_info.hpp>
#include <clause_propagator.hpp>
#include <arithmetic_propagators.hpp>
#include <interval_propagators.hpp>
#include <disjoint_propagators.hpp>
#include <controlled_ptr.hpp>
#include <disjunction_specification.hpp>

// Include the bodies of the literal_or_constraint_propagator methods.
#include <literal_constraint_propagators.inl>
using namespace std;
using namespace hcsp;

/**
 * The DIFF_GE_k(N) rule, for a-b >= k comes from: 
 * \f[
 *    a-b \ge k \Rightarrow ( a \le  N \Rightarrow b \le N-k)
 * \f]
 * Which is equivalent with
 * \f[
 *    a-b \ge k \Rightarrow ( a \ge N+1 \vee b \le N-k)
 * \f]
 */
class difference_ge_propagator::diff_ge_rule : public explanation_rule
{
  public:
    // Assuming unlimited domains, the explanation should be:
    // literals[0]= [N+1, iset::MAX_DOMAIN]
    // literals[1]= [iset::MIN_DOMAIN,N-k]
    //
    // For limited domains:
    // If N >= iset::MAX_DOMAIN, then literals[0].empty();
    // If N-k+1 <= iset::MIN_DOMAIN, then literals[1].empty();
    //
    // Note that if k > iset::MAX_DOMAIN-iset::MIN_DOMAIN then
    // both literals are empty
    diff_ge_rule(const vector<iset> & literals, int k) : m_k(k)
    {
        HCSP_ASSERT(literals.size() == 2);
        if (literals[0].empty())
            m_N= iset::MAX_DOMAIN;
        else
            m_N= literals[0].min()-1;
        
        HCSP_ASSERT(m_N <= iset::MAX_DOMAIN);
        if ( m_N == iset::MAX_DOMAIN )
            HCSP_ASSERT( literals[0].empty() );
        else
            HCSP_ASSERT( m_N+1LL == literals[0].min());

        if ( 1LL+m_N-k <= iset::MIN_DOMAIN )
            HCSP_ASSERT( literals[1].empty() );
        else
            HCSP_ASSERT( 0LL+m_N-k == literals[1].max());

    }
    virtual diff_ge_rule * clone() const { return new diff_ge_rule(*this); }
    
    // literals[0]= [N+1, iset::MAX_DOMAIN]
    // literals[1]= [iset::MIN_DOMAIN,N-k]
    virtual void get_explanation(std::vector<iset> & literals) const
    {
        literals.clear();
        literals.resize(2);
        if ( m_N < iset::MAX_DOMAIN )
            literals[0].assign(make_pair(m_N+1, iset::MAX_DOMAIN));

        if ( 0LL+m_N-m_k >= iset::MIN_DOMAIN )
            literals[1].assign(make_pair(iset::MIN_DOMAIN, m_N-m_k));
    }
    virtual void print(std::ostream & out, const variable_print_strategy &,
                       const std::vector<unsigned> &) const
    {
        out << "DIFF_GE_";
        if (m_k>=0)
            out << m_k;
        else
            out << "minus_" << -static_cast<iset::size_type>(m_k);
        out << "(" << m_N << ")";
    }
  private:
    iset::value_type m_k;
    iset::value_type m_N;
}; // difference_ge_propagator::diff_ge_rule

//virtual
constraint_propagator::result_type
equal_propagator::propagate( constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    const iset & in0= info.domain(0);
    const iset & in1= info.domain(1);
    bool need_to_update1= !in0.contains_all_of(in1);
    bool need_to_update0= !in1.contains_all_of(in0);
    if (need_to_update0) {
        iset out0 = in0;
        out0.intersect(in1);
        if (out0.empty())
            return UNSAT;
        info.set_domains({0,1}, move(out0));
        return CONSISTENT;
    }
    if (need_to_update1) {
        iset out1= info.domain(1);
        out1.intersect(in0);
        if (out1.empty())
            return UNSAT;
        info.set_domain(1, move(out1));
    }
    return CONSISTENT;
}

// virtual
constraint_propagator * equal_propagator::combine(
    const constraint_runtime_info & info,
    const constraint_propagator & other_propagator,
    const constraint_runtime_info & other_info,
    const std::vector<unsigned> & other_to_common_param,
    unsigned other_pivot_param,
    std::vector<unsigned> & copy_from_argument) const
{
    HCSP_ASSERT(info.arity() == 2);
    unsigned pivot= other_to_common_param[other_pivot_param];
    unsigned replacement_argument= 1-pivot;
    
    copy_from_argument.resize(other_info.arity());
    
    sliced_runtime_info copy_info(other_info, 0, other_info.arity());
    constraint_watchless_info watchless_info(copy_info);
    watchless_info.fake_put_in_queue(info.CONSTRAINT_QUEUE);
    for (unsigned i=0 ; i != other_to_common_param.size(); ++i) {
        if (other_to_common_param[i] == replacement_argument)
            return NULL;
        if (i == other_pivot_param) {
            copy_from_argument[i]= replacement_argument;
            watchless_info.set_domain(i, info.domain(replacement_argument));
        } else
            copy_from_argument[i]= other_to_common_param[i];
    }
    unique_ptr<constraint_propagator> constraint(other_propagator.clone());
    if ( constraint->init(watchless_info) &&
         constraint->propagate(watchless_info) != UNSAT)
        return NULL;
    constraint->forget_watches();
    return constraint.release();
}

constraint_propagator * equal_propagator::new_propagator(
    disjunction_specification & disjunction,
    vector<unsigned> &copy_from_argument) const
{
    copy_from_argument.clear();
    copy_from_argument.reserve(disjunction.arity());
    if (disjunction.num_literals() == 0) {
        copy_from_argument.push_back(disjunction.left_var());
        copy_from_argument.push_back(disjunction.right_var());
        return disjunction.release();
    }
    if (disjunction.num_literals() == 1)  {
        copy_from_argument.push_back(disjunction.literal_var(0));
        copy_from_argument.push_back(disjunction.left_var());
        copy_from_argument.push_back(disjunction.right_var());
        const iset & literal= disjunction.literal_set(0);
        return new literal_or_equal_propagator(literal);
    }
    typedef disjoint_propagator<equal_propagator, clause_propagation_algo
                                > equal_or_clause_propagator;
    unique_ptr<equal_or_clause_propagator> result;
    result.reset(new equal_or_clause_propagator());
    vector<iset> & literals= result->algo().algo_b().update_literals();
    literals.reserve(disjunction.num_literals());
    copy_from_argument.push_back(disjunction.left_var());
    copy_from_argument.push_back(disjunction.right_var());
    for (unsigned i=0 ; i != disjunction.num_literals(); ++i) {
        copy_from_argument.push_back(disjunction.literal_var(i));
        literals.push_back(disjunction.literal_set(i));
    }
    return result.release();
}

//virtual
bool equal_propagator::conflicts(const constraint_runtime_info & info) const
{
    const iset & in0= info.domain(0);
    const iset & in1= info.domain(1);
    return !in0.contains_any_of(in1);
}

//virtual
void equal_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        p_info->print_param(out, 0);
        out << "==";
        p_info->print_param(out, 1);
        out << ")";
    } else {
        out << "equal";
    }
}

//virtual
bool difference_ge_propagator::init(constraint_runtime_info & info)
{
    HCSP_ASSERT(info.arity() == (unsigned)count_params());
    info.allocate_watch(1, LOWER_BOUND_WATCH);
    info.allocate_watch(0, UPPER_BOUND_WATCH);
    return true;
}

// virtual
constraint_propagator::result_type difference_ge_propagator::propagate(
    constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    const iset & first= info.domain(0);
    const iset & second= info.domain(1);
    if ( first.empty() || second.empty() )
        return UNSAT;
    iset::value_type first_min= first.min();
    iset::value_type first_max= first.max();
    iset::value_type second_min= second.min();
    iset::value_type second_max= second.max();
    // Because first-second >= m_k:
    if ((long long)first_max-second_min < m_k)
        return UNSAT;

    // Because first-second >= m_k
    // We limit second from above: second <= first-m_k
    if (second_max > 0LL+first_max-m_k)
    {
        iset second_out = second;
        second_out.intersect(make_pair(iset::MIN_DOMAIN, first_max-m_k));
        if (second_out.empty())
            return UNSAT;
        info.set_domain(1, move(second_out));
    }
    // Because first-second >= m_k
    // We limit first from below: first >= second+m_k
    if (first_min < 0LL+second_min+m_k)
    {
        iset first_out = info.domain(0);
        first_out.intersect(make_pair(second_min+m_k, iset::MAX_DOMAIN));
        if (first_out.empty())
            return UNSAT;
        info.set_domain(0, move(first_out));
    }
    if (0LL+info.domain(0).min()-info.domain(1).max() >= m_k)
        return VALID;
    else
        return CONSISTENT;
}

// virtual
constraint_propagator * difference_ge_propagator::intersect_with(
    const constraint_runtime_info & info,
    std::vector<unsigned> & copy_from_argument) const
{
    if (!info.domain(0).singleton() && !info.domain(1).singleton())
        return NULL;
    vector<linear_combination_ge_propagator::coefficient_type> coefficients{
        1, -1};
    linear_combination_ge_propagator combination(m_k, move(coefficients));
    return combination.intersect_with(info, copy_from_argument);
}

// virtual
bool difference_ge_propagator::get_disjunction_specification(
    disjunction_specification& disjunction) const
{
    disjunction.set_literals(NULL, 0, 0);
    disjunction.refer_to_constraint(*this, 0);
    return true;
}

// static
bool difference_ge_propagator::may_combine_diff_ge_propagator(
    const constraint_propagator &tested)
{
    return (dynamic_cast<const difference_ge_propagator*>(&tested) ||
            dynamic_cast<const literal_or_difference_ge_propagator*>(&tested) ||
            dynamic_cast<const difference_ge_or_clause_propagator*>(&tested) ||
            dynamic_cast<const clause_propagator*>(&tested));
}

// virtual
//! @todo move to disjoint_propagator<....,clause_propagation_algo>
bool difference_ge_or_clause_propagator::get_disjunction_specification(
    disjunction_specification& disjunction) const
{
    const vector<iset> &lits= algo().algo_b().literals();
    disjunction.set_literals(&lits[0], 2, lits.size());
    iset::value_type k= algo().algo_a().get_min_difference();
    disjunction.reset_constraint(new difference_ge_propagator(k), 0);
    return true;
}

constraint_propagator * difference_ge_propagator::new_propagator(
    disjunction_specification & disjunction,
    vector<unsigned> &copy_from_argument) const
{
    copy_from_argument.clear();
    copy_from_argument.reserve(disjunction.arity());
    if (disjunction.num_literals() == 0) {
        copy_from_argument.push_back(disjunction.left_var());
        copy_from_argument.push_back(disjunction.right_var());
        return disjunction.release();
    }
    if (disjunction.num_literals() == 1)  {
        copy_from_argument.push_back(disjunction.literal_var(0));
        copy_from_argument.push_back(disjunction.left_var());
        copy_from_argument.push_back(disjunction.right_var());
        const iset & literal= disjunction.literal_set(0);
        return new literal_or_difference_ge_propagator(literal, m_k);
    }
    unique_ptr<difference_ge_or_clause_propagator> result;
    result.reset(new difference_ge_or_clause_propagator(m_k));
    vector<iset> & literals= result->algo().algo_b().update_literals();
    literals.reserve(disjunction.num_literals());
    copy_from_argument.push_back(disjunction.left_var());
    copy_from_argument.push_back(disjunction.right_var());
    for (unsigned i=0 ; i != disjunction.num_literals(); ++i) {
        copy_from_argument.push_back(disjunction.literal_var(i));
        literals.push_back(disjunction.literal_set(i));
    }
    return result.release();
}

bool difference_ge_propagator::disjoin_incompatible_constraint(
    const disjunction_specification & this_constraint,
    const disjunction_specification & other_constraint,
    disjunction_specification & result) const
{
    HCSP_ASSERT(this_constraint.constraint() == this);
    for (unsigned i=0 ; i != this_constraint.arity() ; ++i) {
        unsigned this_constraint_var= this_constraint.constraint_var(i);
        if (other_constraint.is_constraint_var(this_constraint_var) ||
            other_constraint.is_literal(this_constraint_var))
            return false;
    }
    for (unsigned i=0 ; i != other_constraint.arity() ; ++i) {
        unsigned other_constraint_var= other_constraint.constraint_var(i);
        if (this_constraint.is_constraint_var(other_constraint_var) ||
            this_constraint.is_literal(other_constraint_var))
            return false;
    }

    constraint_explicit_info other_info(other_constraint.constraint_arity());
    for (unsigned i=0 ; i != other_info.arity() ; ++i)
        other_info.set_domain(i, other_constraint.constraint_domain(i));
    if (!other_constraint.constraint()->conflicts(other_info))
        return false;
    vector<iset> explanation;
    other_constraint.constraint()->explain_propagation(
        0, other_info.domain(0), iset(), other_info, explanation);
    
    for (unsigned i=0 ; i != other_info.arity() ; ++i) {
        unsigned other_constraint_var= other_constraint.constraint_var(i);
        result.add_literal(explanation[i], other_constraint_var);
    }
    result.append_constraint(this_constraint);
    return true;
}

// virtual
bool difference_ge_propagator::disjoin_constraint(
    const disjunction_specification & this_constraint,
    const disjunction_specification & other_constraint,
    disjunction_specification & result) const
{
    HCSP_ASSERT(this == this_constraint.constraint());
    const constraint_propagator *p_other= other_constraint.constraint();
    if (!p_other)
        return false;
    const difference_ge_propagator *
        p_other_child= dynamic_cast<const difference_ge_propagator*>(p_other);
    if (!p_other_child)
        return false;
    for (unsigned i=0 ; i != this_constraint.arity() ; ++i) {
        unsigned this_constraint_var= this_constraint.constraint_var(i);
        unsigned other_constraint_var= other_constraint.constraint_var(i);
        if (this_constraint_var != other_constraint_var)
            return disjoin_incompatible_constraint(this_constraint,
                                                   other_constraint, result);
    }
    if (p_other_child->get_min_difference() < get_min_difference())
        result.append_constraint(other_constraint);
    else
        result.append_constraint(this_constraint);
    return true;
}



// Assuming the vector of variables v[0]...v[n-1], and the input variables:
//    [i, v[0], ...,v[n-1], d]
// propagate v[i-m_first_index] == d.
constraint_propagator::result_type element_propagator::propagate(
    constraint_runtime_info & info)
{
    if (info.arity() <= 2)
        return UNSAT;
    const iset & result_domain = info.domain(info.arity()-1);
    int array_size= info.arity()-2;
    iset accumulated;
    iset index_set = info.domain(0);
    for (int i= 0 ; i < array_size ; ++i) {
        const iset & element_domain = info.domain(i+1);
        if ( index_set.contains(m_first_index+i)) {
            if (result_domain.contains_any_of(element_domain))
                accumulated.insert_all_of(element_domain);
            else
                index_set.erase(m_first_index + i);
        }
    }
    int last_index= m_first_index + array_size - 1;
    index_set.intersect(make_pair(m_first_index, last_index));
    if (index_set.empty())
        return UNSAT;
    bool result_domain_changed= !accumulated.contains_all_of(result_domain);
    bool result_restricts_elem= (index_set.singleton() && 
                                 !result_domain.contains_all_of(accumulated));
    info.set_domain(0, move(index_set));
    if (result_domain_changed)
    {
        accumulated.intersect(info.domain(info.arity()-1));
        info.set_domain(info.arity()-1, move(accumulated));
    }
    if (result_restricts_elem) {
        int element_arg = info.domain(0).min() - m_first_index + 1;
        iset next_elem_domain = info.domain(info.arity()-1);
        next_elem_domain.intersect(info.domain(element_arg));
        info.set_domain(element_arg, move(next_elem_domain));
    }
    return CONSISTENT;
}

void element_propagator::explain_equality_propagation(
    unsigned propagated_param, const iset & values_to_explain,
    const iset & propagated_domain, const constraint_runtime_info & info,
    vector<iset> & explanation) const
{
    explanation.resize(info.arity());

    iset::value_type first_index= max(iset::MIN_DOMAIN, m_first_index);
    iset::value_type last_index = min(iset::MAX_DOMAIN,
                                      m_first_index + (int)info.arity()-3);
    
    iset index = iset(info.domain(0)).intersect(make_pair(first_index, 
                                                          last_index));
    iset::value_type propagated_index = propagated_param-1+m_first_index;
    explanation[0].assign(make_pair(first_index, last_index));
    explanation[0].erase(propagated_param-1+m_first_index);
    iset removed = values_to_explain;
    removed.intersect(info.domain(propagated_param));
    removed.erase_all_of(propagated_domain);
    explanation[0].erase_all_of(index);
    for (auto i: index)
    {
        if (i != propagated_index)
            removed.insert_all_of(info.domain(i-m_first_index+1));
    }
    if (removed.empty())
    {
        (explanation.back()= propagated_domain).complement();
        explanation[propagated_param] = propagated_domain;
    }
    else
    {
        explanation.back()= removed;
        (explanation[propagated_param]= removed).complement();
    }
    for (auto i: index)
    {
        if (i != propagated_index)
            explanation[i-m_first_index+1]= explanation[propagated_param];
    }
}

void element_propagator::explain_index_propagation(
    const iset & values_to_explain, const iset & propagated_domain, 
    const constraint_runtime_info & info, vector<iset> & explanation) const
{
    explanation.resize(info.arity());

    auto index_range =make_pair(max(m_first_index, iset::MIN_DOMAIN),
                                min(m_first_index+(int)info.arity()-3,
                                    iset::MAX_DOMAIN));
    iset index_bound(index_range);
    
    iset interesting_index_erased= info.domain(0);
    interesting_index_erased.erase_all_of(propagated_domain);
    if (values_to_explain.contains_any_of(interesting_index_erased))
        interesting_index_erased.intersect(values_to_explain);
    if (!interesting_index_erased.contains_any_of(index_bound))
    {
        explanation[0]= index_bound;
        return;
    }
    interesting_index_erased.intersect(index_bound);
    const iset & value_domain = info.domain(info.arity()-1);
    for (auto index: interesting_index_erased)
    {
        explanation[index - m_first_index + 1]= value_domain;
    }
    (explanation.back()= value_domain).complement();
    
    (explanation[0] = interesting_index_erased)
        .complement()
        .intersect(index_bound);
}

void element_propagator::explain_result_propagation(
    const iset & values_to_explain, const iset & propagated_domain, 
    const constraint_runtime_info & info, vector<iset> & explanation) const
{
    explanation.resize(info.arity());

    iset interseting_removed_result= info.domain(info.arity()-1);
    interseting_removed_result.intersect(values_to_explain);
    interseting_removed_result.erase_all_of(propagated_domain);
    if (interseting_removed_result.empty())
    {
        interseting_removed_result= info.domain(info.arity()-1);
        interseting_removed_result.erase_all_of(propagated_domain);
        interseting_removed_result.assign(interseting_removed_result.min());
    }
    (explanation.back()= interseting_removed_result).complement();

    auto index_range = make_pair(max(m_first_index, iset::MIN_DOMAIN),
                                 min(m_first_index+(int)info.arity()-3, 
                                     iset::MAX_DOMAIN));
    iset interesting_index(index_range);
    interesting_index.intersect(info.domain(0));
    for (auto index: interesting_index)
    {
        explanation[index - m_first_index + 1]= interseting_removed_result;
    }
    (explanation[0] = interesting_index).complement().intersect(index_range);
}

void element_propagator::explain_conflict(
    const constraint_runtime_info & info, vector<iset> & explanation) const
{
    explanation.resize(info.arity());

    auto index_range = make_pair(max(iset::MIN_DOMAIN, m_first_index), 
                                 min(m_first_index+(int)info.arity()-3, 
                                     iset::MAX_DOMAIN));
    if (index_range.first > index_range.second)
        return;
    iset interesting_index(index_range);
    interesting_index.intersect(info.domain(0));
    if (interesting_index.empty())
    {
        explanation[0].assign(index_range);
        return;
    }
    const iset & value_domain = info.domain(info.arity()-1);
    for (auto index: interesting_index)
    {
        explanation[index - m_first_index + 1]= value_domain;
    }
    (explanation.back()= value_domain).complement();
    
    (explanation[0] = interesting_index).complement().intersect(index_range);
}


void element_propagator::explain_propagation(
    unsigned propagated_param, const iset & values_to_explain,
    const iset & propagated_domain, const constraint_runtime_info & info,
    vector<iset> & explanation) const
{
    explanation.clear();
    if (propagated_domain.empty())
    {
        explain_conflict(info, explanation);
    }
    else if (propagated_param == 0)
    {
        explain_index_propagation(values_to_explain, propagated_domain,
                                  info, explanation);
    }
    else if (propagated_param == info.arity()-1)
    {
        explain_result_propagation(values_to_explain, propagated_domain,
                                   info, explanation);
    }
    else
    {
        explain_equality_propagation(propagated_param, values_to_explain,
                                     propagated_domain, info, explanation);
    }
}

bool element_propagator::conflicts(const constraint_runtime_info & info) const
{
    if (info.arity() <= 2)
        return true;
    const iset & index_set = info.domain(0);
    const iset & result_domain = info.domain(info.arity()-1);
    int array_size= info.arity()-2;
    for (int i= 0 ; i < array_size ; ++i) {
        if ( index_set.contains(m_first_index + i) &&
             result_domain.contains_any_of(info.domain(i+1))) {
            return false;
        }
    }
    return true;
}

void element_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (!p_info) {
        out << "element";
        return;
    }
    p_info->print_prefix(out);
    out << "({";
    for (unsigned i=1 ; i < p_info->arity()-1 ; ++i) {
        if (i > 1)
            out << ", ";
        p_info->print_param(out, i);
    }
    out << "}[";
    p_info->print_param(out, 0);
    if (m_first_index < 0)
        out << "+" << -m_first_index;
    else if (m_first_index > 0)
        out << "-" << m_first_index;
    out << "] == ";
    p_info->print_param(out, p_info->arity()-1);
    out << ")";
}

constraint_propagator *element_propagator::intersect_with(
    const constraint_runtime_info & info,
    std::vector<unsigned> & copy_from_argument) const
{
    if (info.arity() <= 2)
        return new clause_propagator{};
    const iset & index_set = info.domain(0);
    const iset & result_domain = info.domain(info.arity()-1);
    unsigned array_size= info.arity()-2;
    unsigned first_pos;
    for (first_pos= 0 ; first_pos < array_size ; ++first_pos) {
        if ( index_set.contains(m_first_index + first_pos) &&
             result_domain.contains_any_of(info.domain(first_pos+1))) {
            break;
        }
    }
    unsigned end_pos;
    for (end_pos = array_size ; first_pos < end_pos; --end_pos) {
        if ( index_set.contains(m_first_index + end_pos - 1) &&
             result_domain.contains_any_of(info.domain(end_pos))) {
            break;
        }
    }
    if (first_pos == 0 && end_pos == array_size)
        return nullptr;
    if (first_pos == end_pos-1 && index_set.singleton()) {
        copy_from_argument.push_back(first_pos+1);
        copy_from_argument.push_back(info.arity()-1);
        return new equal_propagator;
    }
    copy_from_argument.push_back(0);
    for (unsigned i= first_pos ; i < end_pos ; ++i)
        copy_from_argument.push_back(i+1);
    copy_from_argument.push_back(info.arity()-1);
    return new element_propagator(first_pos + m_first_index);
}

// Combine the case where:
//    this:    (y-x >= k1)
//    other:   (x-y >= k2)
// If k1+k2 >0 we can sum 'this' and 'other':
//    0 >= k1+k2 > 0  - which conflicts, hence we learned the empty clause
// Otherwise
//         y-x >= k1
//         y-x <= -k2
//    -k2 >= y-x >= k1
// i.e.   y-x \in [k1,-k2]
// If k1==k2==0 then y==x

bool difference_ge_propagator::combine_circular_pivot(
    disjunction_specification &a, disjunction_specification &b,
    iset::value_type b_k, disjunction_specification & result) const
{
    (void)b;
    HCSP_ASSERT(a.left_var() == b.right_var());
    HCSP_ASSERT(b.left_var() == a.right_var());
    HCSP_ASSERT(a.constraint() == this);
    if (0LL+get_min_difference()+b_k > 0) // 0 >= sum >0
        return true; // the inner constraint is UNSAT, i.e., non-existent.
    // Here: -k2 >= y-x >= k1, i.e., y-x \in [k1, -k2]
    // And we have made sure that -k2 >= k1 in the condition above.
    // @todo when k1==-k2 then there should be a way to propagate y==x+k1  
    if (get_min_difference() == 0 && b_k == 0) {
        // 0 >= y-x >= 0 , i.e., x==y
        result.reset_constraint(new equal_propagator, result.num_literals());
        result.add_constraint_var(a.left_var());
        result.add_constraint_var(a.right_var());
        return true;
    }
    return false;
}

// combine:
// left:  y1-x >= k1
// right: x-y2 >= k2
// where x is the pivot.
// Combination result is:
//
//    x \in [y2.min()+k2, y1.max()-k1]
//       ||
//    [y2, y1-k1-k2] \not\subseteq[y2.min(), y1.max()-k1-k2]
// In many cases y2.max() > y1.min()-k1-k2, when the literal of x becomes empty
// where this == y2, other == y1.
bool difference_ge_propagator::combine_middle_pivot(
    disjunction_specification &a, disjunction_specification &b,
    iset::value_type b_k, disjunction_specification & result) const
{
    HCSP_ASSERT(this == a.constraint());
    HCSP_ASSERT(a.right_var() == b.left_var());
    HCSP_ASSERT(a.left_var() != b.right_var());

    long long a_k= get_min_difference();
    long long new_k= a_k+b_k;
    if (new_k > numeric_limits<iset::value_type>::max())
        return false;

    if (b.right_domain().min() > 0LL+a.left_domain().max()-new_k) {
        result.reset_constraint(new difference_ge_propagator(new_k),
                                result.num_literals());
        result.add_constraint_var(a.left_var());
        result.add_constraint_var(b.right_var());
        return true;
    }
    // We add the literal for   x \in [y2.min()+k2, y1.max()-k1]
    long long lower= max(0LL+iset::MIN_DOMAIN, 0LL+b.right_domain().min()+b_k);
    long long upper= min(0LL+iset::MAX_DOMAIN, a.left_domain().max()-a_k);
    if (lower <= iset::MAX_DOMAIN && upper >= iset::MIN_DOMAIN)
        result.add_literal(iset(make_pair(lower, upper)), a.right_var());

    // We add the interval part:
    //    [y2, y1-k1-k2] \not\subseteq[y2.min(), y1.max()-k1-k2]
    long long y1_max = a.left_domain().max();
    long long y2_min = b.right_domain().min();
    if (y1_max - new_k > iset::MAX_DOMAIN || y1_max - new_k < iset::MIN_DOMAIN)
        return false;

    result.reset_constraint(new no_sub_interval_propagator(
                                y2_min, y1_max-new_k, new_k),
                            result.num_literals());
    result.add_constraint_var(b.right_var());
    result.add_constraint_var(a.left_var());
    return true;
}

// virtual
bool difference_ge_propagator::combine_base_constraints(
    disjunction_specification &a, disjunction_specification &b, unsigned pivot,
    disjunction_specification &result) const
{
    HCSP_ASSERT(a.constraint() == this);
    result.clear();
    const difference_ge_propagator *p_b_constraint;
    p_b_constraint= dynamic_cast<const difference_ge_propagator*>(
        b.constraint());
    if (!p_b_constraint)
        return false;
    iset::value_type b_k= p_b_constraint->get_min_difference();
    iset::value_type a_k= get_min_difference();

    if ( a.left_var() == b.right_var() && a.right_var() == b.left_var())
        return combine_circular_pivot(a, b, b_k, result);
    else if (a.right_var() == pivot && pivot == b.left_var())
        return combine_middle_pivot(a, b, b_k, result);
    else if (b.right_var() == pivot && pivot == a.left_var())
        return p_b_constraint->combine_middle_pivot(b, a, a_k, result);
    return false;
}

// static
constraint_propagator *difference_ge_propagator::combine(
    const constraint_propagator & propagator,
    const constraint_runtime_info & info,
    const constraint_propagator & other_propagator,
    const constraint_runtime_info & other_info,
    const std::vector<unsigned> & other_to_common_param,
    unsigned other_pivot_param,
    std::vector<unsigned> & copy_from_argument)
{
    HCSP_ASSERT(propagator.may_combine(other_propagator));
    disjunction_specification a(info, propagator, vector<unsigned>());
    disjunction_specification b(other_info, other_propagator,
                                other_to_common_param);
    unsigned pivot= other_to_common_param[other_pivot_param];
    return a.combine(b, pivot, copy_from_argument);
}

// virtual
bool
difference_ge_propagator::conflicts(const constraint_runtime_info & info) const
{
    HCSP_PROFILE_CURRENT;
    const iset & first= info.domain(0);
    const iset & second= info.domain(1);
    if ( first.empty() || second.empty() )
        return true;
    return (long long)first.max()- second.min() < m_k;
}

// virtual
unique_ptr<explanation_rule>
difference_ge_propagator::rule(const std::vector<iset> & literals) const
{
    return unique_ptr<explanation_rule>(new diff_ge_rule(literals, m_k));
}

//virtual
void difference_ge_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        p_info->print_param(out, 0);
        out << "-";
        p_info->print_param(out, 1);
        out << ">=" << m_k << ")";
    } else {
        out << "diff_ge_";
        if (m_k >= 0)
            out << m_k;
        else
            out << "minus_" << -static_cast<iset::size_type>(m_k);
    }
}

constraint_propagator::result_type if_then_else_propagator::propagate_then(
    constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    HCSP_ASSERT(info.domain(1) == iset(1));
    const iset & in_result= info.domain(0);
    const iset & in_then= info.domain(2);
    iset common= in_then;
    common.intersect(in_result);
    if (common.empty())
        return UNSAT;
    info.set_domain(2, common);
    info.set_domain(0, common);

    return common.singleton() ? VALID : CONSISTENT;
}

constraint_propagator::result_type if_then_else_propagator::propagate_else(
    constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    HCSP_ASSERT(info.domain(1) == iset(0));
    const iset & in_result= info.domain(0);
    const iset & in_else= info.domain(3);
    iset common= in_else;
    common.intersect(in_result);
    if (common.empty())
        return UNSAT;
    info.set_domain(3, common);
    info.set_domain(0, common);

    return common.singleton() ? VALID : CONSISTENT;
}
    
// virtual
constraint_propagator::result_type if_then_else_propagator::propagate(
    constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    const iset & in_result= info.domain(0);
    const iset & in_cond= info.domain(1);
    const iset & in_then= info.domain(2);
    const iset & in_else= info.domain(3);
    iset out_cond= in_cond;
    out_cond.intersect(make_pair(0,1));
    if ( !out_cond.singleton() ) {
        if ( !in_then.contains_any_of(in_result) )
            out_cond.intersect(make_pair(0,0));
        if ( !in_else.contains_any_of(in_result) )
            out_cond.intersect(make_pair(1,1));
    }
    if ( out_cond.empty() )
        return UNSAT;
    info.set_domain(1, out_cond);
    // The steps above could have made out_cond a singleton
    if ( out_cond.singleton() ) { 
        if (out_cond.contains(1))
            return propagate_then(info);
        else
            return propagate_else(info);
    }

    // Condition is unknown yet. Join domains of else and then.
    iset out_result= in_then;
    out_result.insert_all_of(in_else);
    out_result.intersect(in_result);
    if (out_result.empty())
        return UNSAT;           // Should not happen in current code.
    info.set_domain(0, move(out_result));
    return CONSISTENT;
}

// virtual
bool
if_then_else_propagator::conflicts(const constraint_runtime_info & info) const
{
    HCSP_PROFILE_CURRENT;
    const iset & in_result= info.domain(0);
    const iset & in_cond= info.domain(1);
    const iset & in_then= info.domain(2);
    const iset & in_else= info.domain(3);
    bool may_be_else= in_cond.contains(0);
    bool may_be_then= in_cond.contains(1);
    if (may_be_then && in_then.contains_any_of(in_result))
        return false;
    if (may_be_else && in_else.contains_any_of(in_result))
        return false;
    return true;
}

//virtual
void if_then_else_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        p_info->print_param(out, 0);
        out << "=(";
        p_info->print_param(out, 1);
        out << " ? ";
        p_info->print_param(out, 2);
        out << " : ";
        p_info->print_param(out, 3);
        out << ")";
    } else {
        out << "if_then_else";
    }
}

constraint_propagator::result_type is_equal_propagator::propagate_diff(
    constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    HCSP_ASSERT(info.domain(0)==iset(0));
    const iset & in_first = info.domain(1);
    const iset & in_second = info.domain(2);
    if (in_first.singleton()) {
        iset out_second= in_second;
        out_second.erase_all_of(in_first);
        if (out_second.empty())
            return UNSAT;
        info.set_domain(2, move(out_second));
        return VALID;
    }

    if (in_second.singleton()) {
        iset out_first = in_first;
        out_first.erase_all_of(in_second);
        if (out_first.empty())
            return UNSAT;
        info.set_domain(1, move(out_first));
        return VALID;
    }
    return CONSISTENT;
}

constraint_propagator::result_type is_equal_propagator::propagate_equal(
    constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    HCSP_ASSERT(info.domain(0)==iset(1));
    const iset & in_first = info.domain(1);
    const iset & in_second = info.domain(2);
    if (in_first == in_second)
        return CONSISTENT;
    
    iset out = in_first;
    out.intersect(in_second);
    if (out.empty())
        return UNSAT;
    info.set_domains({1,2}, move(out));
    return CONSISTENT;
}

// virtual
constraint_propagator::result_type is_equal_propagator::propagate(
    constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    const iset zero(0);
    const iset one(1);
    const iset & in_cond = info.domain(0);
    const iset & first = info.domain(1);
    const iset & second = info.domain(2);
    iset out_cond= in_cond;
    out_cond.intersect(make_pair(0,1));
    bool can_be_equal= first.contains_any_of(second);
    bool can_be_different= !(first.singleton() && first == second);
    if ( !can_be_equal )
        out_cond.intersect(make_pair(0,0));
    if ( !can_be_different )
        out_cond.intersect(make_pair(1,1));
    if (out_cond.empty())
        return UNSAT;
    info.set_domain(0, move(out_cond));
    if (info.domain(0) == zero)
        return propagate_diff(info);
    else if (info.domain(0) == one)
        return propagate_equal(info);
    else
        return CONSISTENT;
}

// virtual
bool is_equal_propagator::conflicts(const constraint_runtime_info & info) const
{
    const iset & in_cond = info.domain(0);
    const iset & a = info.domain(1);
    const iset & b = info.domain(2);
    if (in_cond.contains(0)) {
        if (in_cond.contains(1))
            return false;
        // require to be different
        if (!a.singleton() || !b.singleton())
            return false;
        return a == b; 
    }
    // require to be equal
    return !in_cond.contains(1) || !a.contains_any_of(b);
}

//virtual
void is_equal_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        p_info->print_param(out, 0);
        out << " iff ";
        p_info->print_param(out, 1);
        out << " = ";
        p_info->print_param(out, 2);
        out << ")";
    } else {
        out << "is_equal";
    }
}

// virtual
constraint_propagator::result_type
literal_or_equal_propagator::propagate_non_literal_part(
    constraint_runtime_info & info)
{
    const iset & in_first= info.domain(1);
    const iset & in_second= info.domain(2);

    // We assume that trivial cases have been handled by the propagate()
    // algorithm before propagate_non_literal_part() was called.
    HCSP_ASSERT(state_of_non_literal_part(info) == CONSISTENT);
    
    iset out= in_first;
    out.intersect(in_second);
    HCSP_ASSERT(!out.empty()); // handled prior to calling current method

    info.set_domains({1,2}, move(out));
    
    return CONSISTENT;
}

// virtual
constraint_propagator::result_type
literal_or_equal_propagator::state_of_non_literal_part(
    const constraint_runtime_info & info) const
{
    const iset & in_first= info.domain(1);
    const iset & in_second= info.domain(2);
    if ( in_first.empty() || in_second.empty() )
        return UNSAT;

    if ( !in_first.contains_any_of(in_second) )
        return UNSAT;
    if (in_first.singleton() && in_second.singleton())
        return VALID;
    else
        return CONSISTENT;
}

void literal_or_equal_propagator::print_non_literal_part(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_param(out, 1);
        out << "==";
        p_info->print_param(out, 2);
    } else {
        out << "equal_propagator";
    }
}

// virtual
bool literal_iff_less_than_propagator::init_non_literal_part(
    constraint_runtime_info & info)
{
    const iset & in_first = info.domain(1);
    const iset & in_second = info.domain(2);
    if (in_first.min() >= in_second.max())
        return false;
    
    info.allocate_watch(1, LOWER_BOUND_WATCH);
    info.allocate_watch(2, UPPER_BOUND_WATCH);
    return true;
}

// virtual
bool literal_iff_less_than_propagator::init_neg_non_literal_part(
    constraint_runtime_info & info)
{
    // This is the case that checks V1 >= V2
    const iset & in_first = info.domain(1);
    const iset & in_second = info.domain(2);
    if (in_first.max() < in_second.min())
        return false;
    
    info.allocate_watch(1, UPPER_BOUND_WATCH);
    info.allocate_watch(2, LOWER_BOUND_WATCH);
    return true;
}

// virtual
constraint_propagator::result_type
literal_iff_less_than_propagator::propagate_non_literal_part(
    constraint_runtime_info & info)
{
    const iset & in_first = info.domain(1);
    const iset & in_second = info.domain(2);

    // These assertions should pass if state_of_non_literal_part() returns
    // correct values.
    HCSP_ASSERT(!in_first.empty() && !in_second.empty());
    HCSP_ASSERT(in_first.min() < in_second.max());
    if (in_first.max() >= in_second.max()) { // update V1: V1 < max_domain(V2)
        iset out_first= info.domain(1);
        out_first.erase(make_pair(in_second.max(), iset::MAX_DOMAIN));
        HCSP_ASSERT(!out_first.empty());
        info.set_domain(1, move(out_first));
    }
    if (in_first.min() >= in_second.min()) { // update V2: min_domain(V1) < V2
        iset out_second= info.domain(2);
        out_second.erase(make_pair(iset::MIN_DOMAIN, in_first.min()));
        HCSP_ASSERT(!out_second.empty());
        info.set_domain(2, move(out_second));
    }
    if (info.domain(1).max() < info.domain(2).min())
        return VALID;
    else
        return CONSISTENT;
}

// This methods propagates V1 >= V2
// virtual
constraint_propagator::result_type
literal_iff_less_than_propagator::propagate_neg_non_literal_part(
    constraint_runtime_info & info)
{
    const iset & in_first = info.domain(1);
    const iset & in_second = info.domain(2);

    // These assertions should pass if state_of_non_literal_part() returns
    // correct values.
    HCSP_ASSERT(!in_first.empty() && !in_second.empty());
    HCSP_ASSERT(in_first.max() >= in_second.min());
    if (in_first.min() < in_second.min()) { // update V1: V1 >= min_domain(V2)
        iset out_first= info.domain(1);
        out_first.intersect(make_pair(in_second.min(), iset::MAX_DOMAIN));
        HCSP_ASSERT(!out_first.empty());
        info.set_domain(1, move(out_first));
    }
    if (in_first.max() < in_second.max()) { // update V2: max_domain(V1) >= V2
        iset out_second= info.domain(2);
        out_second.intersect(make_pair(iset::MIN_DOMAIN, in_first.max()));
        HCSP_ASSERT(!out_second.empty());
        info.set_domain(2, move(out_second));
    }
    if (info.domain(1).min() >= info.domain(2).max())
        return VALID;
    else
        return CONSISTENT;
}

// virtual
constraint_propagator::result_type
literal_iff_less_than_propagator::state_of_non_literal_part(
        const constraint_runtime_info & info) const
{
    const iset & in_first = info.domain(1);
    const iset & in_second = info.domain(2);
    if (in_first.empty() || in_second.empty())
        return UNSAT;
    if (in_first.max() < in_second.min())
        return VALID;
    if (in_first.min() >= in_second.max())
        return UNSAT;

    return CONSISTENT;
}

// virtual
void literal_iff_less_than_propagator::print_non_literal_part(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_param(out, 1);
        out << "<";
        p_info->print_param(out, 2);
    } else {
        out << "less_than_propagator";
    }
}

// virtual
constraint_propagator *literal_or_difference_ge_propagator::intersect_with(
    const constraint_runtime_info & info,
    vector<unsigned> & copy_from_argument) const
{
    const iset & literal_domain= info.domain(0);
    if (!guard_set().empty() && guard_set().contains_all_of(literal_domain)) {
        copy_from_argument.assign({0});
        return new clause_propagator({iset().complement()});
    }
    unique_ptr<constraint_propagator> result;
    sliced_runtime_info inner_info(info, 1, info.arity());
    result.reset(difference_ge_propagator(m_k)
                 .intersect_with(inner_info, copy_from_argument));
    if (!guard_set().contains_any_of(literal_domain)) {
        if (result.get()) {
            for (auto & arg : copy_from_argument)
                arg = arg+1;
        } else {
            result.reset(new difference_ge_propagator(m_k));
            copy_from_argument.assign({1,2});
        }
        return result.release();
    }
    if (result.get() && dynamic_cast<clause_propagator*>(result.get())) {
        clause_propagator *as_clause;
        as_clause= static_cast<clause_propagator*>(result.get());
        vector<iset> & literals= as_clause->update_literals();
        literals.insert(literals.begin(), guard_set());
        for (auto & arg : copy_from_argument)
            arg = arg+1;
        copy_from_argument.insert(copy_from_argument.begin(), 0);
        return result.release();
    }
    return nullptr;
}

// virtual
bool literal_or_difference_ge_propagator::get_disjunction_specification(
    disjunction_specification& disjunction) const
{
    disjunction.set_literals(&guard_set(), 0, 1);
    iset::value_type k= get_min_difference();
    disjunction.reset_constraint(new difference_ge_propagator(k), 1);
    return true;
} 

//virtual
bool literal_or_difference_ge_propagator::init_non_literal_part(
    constraint_runtime_info & info)
{
    if (0LL+info.domain(1).max() - info.domain(2).min() < m_k)
        return false;
    
    info.allocate_watch(2, LOWER_BOUND_WATCH);
    info.allocate_watch(1, UPPER_BOUND_WATCH);
    return true;
}

// virtual
constraint_propagator::result_type
literal_or_difference_ge_propagator::propagate_non_literal_part(
    constraint_runtime_info & info )
{
    const iset & in_first= info.domain(1);
    const iset & in_second= info.domain(2);

    // We assume that trivial cases have been handled by the propagate()
    // algorithm before propagate_non_literal_part() was called.
    HCSP_ASSERT(state_of_non_literal_part(info) == CONSISTENT);
    // Because first-second >= m_k:
    if ((long long)in_first.max()-in_second.min() < m_k)
        return UNSAT;

    // Because first-second >= m_k
    // We limit second from above: second <= first-m_k
    if (in_second.max() > 0LL+in_first.max() - m_k) {
        iset out_second= info.domain(2);
        out_second.intersect(make_pair(iset::MIN_DOMAIN, in_first.max()-m_k));
        info.set_domain(2, move(out_second));
    }
    // Because first-second >= m_k
    // We limit first from below: first >= second+m_k
    if (in_first.min() < 0LL+in_second.min()+m_k) {
        iset out_first = info.domain(1);
        out_first.intersect(make_pair(in_second.min()+m_k, iset::MAX_DOMAIN));
        info.set_domain(1, move(out_first));
    }
    if (info.domain(1).empty() || info.domain(2).empty())
        return UNSAT;
    else if (0LL+info.domain(1).min()-info.domain(2).max() >= m_k)
        return VALID;
    else
        return CONSISTENT;
}

// virtual
constraint_propagator::result_type
literal_or_difference_ge_propagator::state_of_non_literal_part(
    const constraint_runtime_info & info) const
{
    const iset & in_first= info.domain(1);
    const iset & in_second= info.domain(2);
    if ( in_first.empty() || in_second.empty() )
        return UNSAT;

    if ((long long)in_first.max()-in_second.min() < m_k)
        return UNSAT;
    else if (0LL+in_first.min()-in_second.max() >= m_k)
        return VALID;
    else
        return CONSISTENT;
}

void literal_or_difference_ge_propagator::print_non_literal_part(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_param(out, 1);
        out << "-";
        p_info->print_param(out, 2);
        out << ">=" << m_k;
    } else {
        out << "diff_ge_";
        if (m_k >= 0)
            out << m_k;
        else
            out << "minus_" << -static_cast<iset::size_type>(m_k);
    }
}

// static
void min_max_propagator::get_result_supports(const vector<iset> & domains,
                                             vector<iset> & supports)
{
    supports.resize(domains.size());
    for (unsigned i=1 ; i != domains.size() ; ++i) {
        supports[i]= domains[i];
        supports[i].intersect(domains[0]);
    }
}

//static
bool min_max_propagator::propagate_to_max_result(vector<iset> & domains)
{
    HCSP_PROFILE_CURRENT;
    const iset in_result= domains[0];
    iset & out_result= domains[0];
    out_result.clear();

    // max_of_smallest=max(Dx.min(), Dy.min(), Dz.min()), for a=max(x,y,z), 
    // [where Dx,Dy and Dz are the domains of the x,y and z variables]
    iset::value_type max_of_smallest= iset::MIN_DOMAIN;
    for (unsigned i=1 ; i != domains.size() ; ++i) {
        max_of_smallest= max(domains[i].min(), max_of_smallest);
        out_result.insert_all_of(domains[i]);
    }
    // max(x,y,z) can't be less than max_of_smallest.
    out_result.intersect(make_pair(max_of_smallest, iset::MAX_DOMAIN));
    out_result.intersect(in_result);
    // If max_of_smallest is not contained in out_result, then any value 'q'
    // above it in out_result can be extended to a feasible assignment.
    // Extending to a feasible assignment: 
    // This 'q' is contained in the original in_result, in one of max(...)
    // arguments' domains. Also, it is possible to assign all other
    // arguments to their minimal value (which is not bigger than out_result).
    // This assignment satisfies a=max(x,y,z)

    if (out_result.empty())
        return false;
    return true;
}

// static
bool min_max_propagator::erase_values_above( vector<iset> & domains,
                                             iset::value_type max_legal)
{
    for (unsigned i=0 ; i != domains.size() ; ++i) {
        if (domains[i].empty())
            return false;
        if (domains[i].max() > max_legal) {
            iset & updated_domain= domains[i];
            updated_domain.intersect(make_pair(iset::MIN_DOMAIN, max_legal));
            if ( updated_domain.empty() )
                return false;
        }
    }
    return true;
}

//static
unsigned min_max_propagator::find_index_of_max(const vector<iset> & values)
{
    iset::value_type max_value= iset::MIN_DOMAIN;
    unsigned index_of_max= values.size();
    for (unsigned i=0 ; i != values.size() ; ++i) {
        if ( values[i].empty() )
            continue;
        iset::value_type local_max= values[i].max();
        if (local_max >= max_value) {
            max_value= local_max;
            index_of_max= i;
        }
    }
    return index_of_max;
}

// static
iset::value_type
min_max_propagator::find_max_lower_bound(const vector<iset> & domains)
{
    iset::value_type max_lower_bound= iset::MIN_DOMAIN;
    for (unsigned i=1 ; i != domains.size() ; ++i) {
        if ( !domains[i].empty() )
            max_lower_bound= max(domains[i].min(), max_lower_bound);
    }
    return max_lower_bound;
}

/**
 * @page max_propagation max propagation algorithm
 *
 * This algorithm implements an accurate propagation of 
 *    \f[ v_0= \max(v_1, v_2, v_3, ... , v_n) \f]
 *
 * For this explanation, \f$ D_i \f$ stands for the domain of \f$ v_i \f$.
 *
 * \par Propagating to D0:
 *
 *  Value \f$ x \in D_0 \f$ is supported if and only if:
 * \f[
 *    \exists i\in \{1\ldots n\} .
 *            (x \in D_i \wedge x \ge \min(D_1) \wedge x \ge \min(D_2) \wedge
 *             \cdots \wedge x \ge \min(D_n) )
 * \f]
 *
 * This fact leads to the following pair of rules:
 *
 *   \f[ M = \max( \min(D_1), \min(D_2), \ldots) \f]
 *   \f[ x \ge M \wedge
 *       \exists i\in \{1\ldots n\} . x \in D_i  \f]
 *
 * The propagation to D0 then becomes:
 *   \f[
 *      D_0' \gets D_0 \cap [m, \infty) \cap (D_1 \cup D_2 \cdots)
 *   \f]
 *
 * \n
 * <b>Propagating to \f$D_i\f$:</b>\n
 *
 *  Value \f$ x \in D_k \f$ is supported if and only if either of the following
 *  conditions are met:
 *  -# There are other variables with bigger values that support D0.
 *     These variables give the freedom to assign lower or equal values
 *     to the k-th variable.
 *     \f[ \exists i\in \{1\ldots n\}. i\ne k\wedge
 *                                     \max(D_i \cap D_0) \ge x
 *     \f]
 *     Or in other words,
 *     \f[ \max(\max(D_1 \cap D_0), \ldots, \max(D_{k-1} \cap D_0),
 *         \max(D_{k+1} \cap D_0),  \ldots, \max(D_n \cap D_0)) \ge x
 *     \f]
 *     The propagator looks for \e t for which the above equation can be
 *     replaced with \f$ t \ge x \f$.
 *     - find \e j such that
 *       \f$ \max(D_j \cap D_0) =
 *           \max(\max(D_1 \cap D_0), \max(D_2 \cap D_0), \ldots)
 *       \f$
 *     - if \f$ j \ne k \f$ then
 *       \f$ t \gets \max(D_j \cap D_0) \f$
 *     - otherwise \e j = \e k  and then
 *      \f$ t \gets \max_{i\ne k}(max(D_i \cap D_0))
 *      \f$ 
 *  -# Other variables support the case when x is the biggest value:
 *     \f[ x \in D_0 \wedge
 *         \forall i\in \{1\ldots n\}. \min(D_i \cap D_0) \le x
 *     \f]
 *     The propagator looks for \e s for which the above equation can be
 *     replaced with \f$ s \le x \f$  \n
 *     \f$ s \gets \max(\min(D_1 \cap D_0), \min(D_1 \cap D_0),\ldots)
 *     \f$
 *
 * The propagation can now be performed by:
 *  \f[ D_k'\gets D_k \cap ( (-\infty,t] \cup ([s, \infty) \cap D_0) ) ) \f]
 */
// static
bool min_max_propagator::propagate_max(vector<iset> & domains)
{
    HCSP_PROFILE_CURRENT;
    if (!propagate_to_max_result(domains))
        return false;
    if (!erase_values_above(domains, domains[0].max()))
        return false;

    static vector<iset> supporting_max;
    get_result_supports(domains, supporting_max) ;
    unsigned index_of_biggest_max_support= find_index_of_max(supporting_max);
    iset::value_type max_lower_bound= find_max_lower_bound(supporting_max);
    iset default_below_max(
        make_pair(iset::MIN_DOMAIN,
                  supporting_max[index_of_biggest_max_support].max()));

    supporting_max[index_of_biggest_max_support].clear();
    unsigned index_of_next_max_support= find_index_of_max(supporting_max);
    
    iset next_below_max;
    if ( index_of_next_max_support < domains.size() ) { // found such index?
        iset::value_type max= supporting_max[index_of_next_max_support].max();
        next_below_max.assign(make_pair(iset::MIN_DOMAIN, max));
    }

    iset causing_max = domains[0];
    causing_max.intersect(make_pair(max_lower_bound, iset::MAX_DOMAIN) );
    for (unsigned i=1 ; i != domains.size() ; ++i) {
        iset next_domain;
        if (i != index_of_biggest_max_support) 
            next_domain= default_below_max;
        else if (!next_below_max.empty())
            next_domain= next_below_max;
        next_domain.insert_all_of(causing_max);
        domains[i].intersect(next_domain);
    }
    
    return CONSISTENT;
}

//virtual
constraint_propagator::result_type
max_propagator::propagate( constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    static vector<iset> domains;
    domains.resize(info.arity());
    for (unsigned i=0 ; i != domains.size() ; ++i)
        domains[i]= info.domain(i);

    if (!propagate_max(domains))
        return UNSAT;
    
    for (unsigned i=0 ; i != domains.size() ; ++i)
        info.set_domain(i, domains[i]);

    return CONSISTENT;
}

//virtual
bool max_propagator::conflicts(const constraint_runtime_info & info) const
{
    static vector<iset> domains;
    domains.resize(info.arity());
    for (unsigned i=0 ; i != domains.size() ; ++i)
        domains[i]= info.domain(i);

    return !propagate_max(domains);
}

//virtual
void max_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        p_info->print_param(out, 0);
        out << "=max(";
        p_info->print_param(out, 1);
        for (unsigned i=2 ; i < p_info->arity() ; ++i) {
            out << ", ";
            p_info->print_param(out, i);
        }
        out << ")";
    } else {
        out << "max";
    }
}

//virtual
constraint_propagator::result_type
min_propagator::propagate( constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    static vector<iset> domains;
    domains.resize(info.arity());
    for (unsigned i=0 ; i != domains.size() ; ++i)
        mirror(domains[i]= info.domain(i));

    if (!propagate_max(domains))
        return UNSAT;
    
    for (unsigned i=0 ; i != domains.size() ; ++i) {
        mirror(domains[i]);
        info.set_domain(i, domains[i]);
    }

    return CONSISTENT;
}

//virtual
bool min_propagator::conflicts(const constraint_runtime_info & info) const
{
    static vector<iset> domains;
    domains.resize(info.arity());
    for (unsigned i=0 ; i != domains.size() ; ++i)
        mirror(domains[i]= info.domain(i));

    return !propagate_max(domains);
}

// static
void min_propagator::mirror(iset & elements)
{
    iset result;
    for ( iset::iterator it= elements.begin() ;
          it != elements.end() ;
          it.advance_range()) {
        pair<iset::value_type, iset::value_type> range= it.current_range();
        iset::value_type offset= iset::MAX_DOMAIN + iset::MIN_DOMAIN;
        result.insert(make_pair(offset-range.second, offset-range.first));
    }
    elements= result;
}

//virtual
void min_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        p_info->print_param(out, 0);
        out << "=min(";
        p_info->print_param(out, 1);
        for (unsigned i=2 ; i < p_info->arity() ; ++i) {
            out << ", ";
            p_info->print_param(out, i);
        }
        out << ")";
    } else {
        out << "min";
    }
}

