// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <conflict_constraint_dependency.hpp>

#include <set>
#include <algorithm>
#include <utilities.hpp>
#include <pool_alloc.hpp>

using namespace std;
using namespace hcsp;

const unsigned conflict_constraint_dependency::ERASED;



void conflict_constraint_dependency::mark_used(unsigned constraint,
                                               unsigned current_backtrack_count)
{
    HCSP_ASSERT(constraint >= m_first_conflict_constraint);
    unsigned internal_index= constraint - m_first_conflict_constraint;
    
    if (internal_index >= m_usage_backtrack_count.size()) {
        m_usage_backtrack_count.resize(internal_index+1, 0);
        m_prerequisites.resize(internal_index+1);
    }
    
    HCSP_ASSERT(m_usage_backtrack_count.at(internal_index) != ERASED);
    
    m_usage_backtrack_count[internal_index]= current_backtrack_count;
}

void conflict_constraint_dependency::mark_dependency(unsigned prerequisite,
                                                     unsigned consequent)
{
    HCSP_ASSERT(prerequisite >= m_first_conflict_constraint);
    HCSP_ASSERT(m_usage_backtrack_count.at(prerequisite-
                                          m_first_conflict_constraint)
               != ERASED);

    HCSP_ASSERT(consequent >= m_first_conflict_constraint);
    unsigned consequent_internal_index= consequent-m_first_conflict_constraint;
    HCSP_ASSERT(m_usage_backtrack_count[consequent_internal_index] != ERASED);

    if ( consequent_internal_index >= m_prerequisites.size() )
        m_prerequisites.resize(consequent_internal_index+1);
    m_prerequisites[consequent_internal_index].push_back(prerequisite);
}

void conflict_constraint_dependency::compact_prerequisites(unsigned consequent)
{
    HCSP_ASSERT(consequent >= m_first_conflict_constraint);
    unsigned consequent_internal_index= consequent-m_first_conflict_constraint;
    
    HCSP_ASSERT(m_usage_backtrack_count[consequent_internal_index] != ERASED);
    HCSP_ASSERT( consequent_internal_index < m_prerequisites.size() );

    size_t capacity= m_prerequisites[consequent_internal_index].capacity();
    size_t size= m_prerequisites[consequent_internal_index].size();

    if (capacity == size)
        return;

    constraint_prerequisites new_value;
    new_value.reserve(m_prerequisites[consequent_internal_index].size());
    new_value= m_prerequisites[consequent_internal_index];
    m_prerequisites[consequent_internal_index].swap(new_value);

    HCSP_ASSERT(size == m_prerequisites[consequent_internal_index].capacity());
}

template <class Touched_Container>
void conflict_constraint_dependency::add_prerequisites(
    vector<unsigned> & prerequisites,
    Touched_Container & touched, unsigned consequent)const
{
    HCSP_ASSERT(consequent >= m_first_conflict_constraint);
    
    unsigned internal_constraint_index= consequent-m_first_conflict_constraint;
    HCSP_ASSERT(internal_constraint_index < m_prerequisites.size());

    // This case is possible when the prerequisite was removed earlier,
    // but the consequent remained.
    if ( m_usage_backtrack_count[internal_constraint_index] == ERASED )
        return;

    const constraint_prerequisites & this_constraint_prerequisites
        = m_prerequisites[internal_constraint_index];
    for (unsigned i=0 ; i != this_constraint_prerequisites.size() ; ++i) {
        unsigned prerequisite= this_constraint_prerequisites[i];
        if (touched.find(prerequisite) != touched.end())
            continue;
        touched.insert(prerequisite); // avoid reentry

        unsigned internal_prerequisite_index;
        internal_prerequisite_index= prerequisite-m_first_conflict_constraint;
        HCSP_ASSERT(internal_prerequisite_index < m_prerequisites.size());
        // If the prerequisite was removed earlier
        if (m_usage_backtrack_count[internal_prerequisite_index] == ERASED)
            continue;
        prerequisites.push_back(prerequisite);
        
        add_prerequisites(prerequisites, touched, prerequisite);
    }
}

vector<unsigned> conflict_constraint_dependency::active_constraints(
    unsigned backtrack_count) const
{
    vector<unsigned> active_constraints_vector;
    for ( unsigned i=0; i != m_usage_backtrack_count.size() ; ++i) {
        if ( m_usage_backtrack_count[i] == ERASED )
            continue;
        if ( m_usage_backtrack_count[i] >= backtrack_count ) {
            unsigned conflict_constraint= i+m_first_conflict_constraint;
            active_constraints_vector.push_back(conflict_constraint);
        }
    }
    return active_constraints_vector;
}

vector<unsigned> conflict_constraint_dependency::other_constraints(
    const vector<unsigned> &constraints) const
{
    vector<unsigned> sorted_constraints= constraints;
    sort(sorted_constraints.begin(), sorted_constraints.end());
    
    vector<unsigned> complement_container;
    for (unsigned i=0 ; i != m_usage_backtrack_count.size() ; ++i) {
        if ( m_usage_backtrack_count[i] == ERASED )
            continue;
        unsigned conflict_constraint= i+m_first_conflict_constraint;
        if ( !binary_search(sorted_constraints.begin(),
                            sorted_constraints.end(),
                            conflict_constraint) )
            complement_container.push_back(conflict_constraint);
    }
    return complement_container;
}


vector<unsigned> conflict_constraint_dependency::prerequisite_closure(
    const vector<unsigned> & consequents) const
{
    set<unsigned, less<unsigned>,pool_alloc<unsigned> > touched_constraints;

    vector<unsigned> prerequisites;

    for (unsigned i=0 ; i != consequents.size() ; ++i)
        add_prerequisites(prerequisites, touched_constraints, consequents[i]);

    return prerequisites;
}



void conflict_constraint_dependency::erase_constraint(
    unsigned conflict_constraint)
{
    HCSP_ASSERT(conflict_constraint >= m_first_conflict_constraint);
    unsigned constraint_internal_index=
        conflict_constraint-m_first_conflict_constraint;
    
    HCSP_ASSERT(m_usage_backtrack_count[constraint_internal_index] != ERASED);
    HCSP_ASSERT( constraint_internal_index < m_prerequisites.size() );

    HCSP_ASSERT( m_prerequisites.size() == m_usage_backtrack_count.size() );

    constraint_prerequisites().swap(m_prerequisites[constraint_internal_index]);
    m_usage_backtrack_count[constraint_internal_index]= ERASED;
}

