// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <iostream>
#include <stdexcept>
#include <constraint_propagator.hpp>
#include <comparison_propagators.hpp>
using namespace std;
using namespace hcsp;

//virtual
bool constraint_propagator::init( constraint_runtime_info & info )
{
    HCSP_ASSERT(count_params() < 0 || info.arity() == (unsigned)count_params());
    const unsigned size= info.arity();
    for (unsigned i=0 ; i != size ; ++i)
        if ( !info.domain(i).singleton() )
            info.allocate_watch(i, MODIFICATION_WATCH);
    return true;
}

// virtual
unsigned constraint_propagator::watched_param_index(unsigned) const
{
    HCSP_RUNTIME_ERROR("constraint_propagator::watched_param_index\n"
                      "\t\t\t\t has no dynamic watches to work on");
}

// virtual
bool constraint_propagator::get_disjunction_specification(
    disjunction_specification &) const
{
    return false;
}

// virtual
bool constraint_propagator::disjoin_constraint(
    const disjunction_specification &, const disjunction_specification &, 
    disjunction_specification &) const
{
    return false;
}

// virtual
bool constraint_propagator::combine_base_constraints(
    disjunction_specification &, disjunction_specification &,
    unsigned, disjunction_specification &) const
{
    return false;
}

// virtual
constraint_propagator *constraint_propagator::new_propagator(
        disjunction_specification&, vector<unsigned> &) const
{
    return NULL;
}

//virtual
bool constraint_propagator::may_combine(
    const constraint_propagator &other) const
{
    return dynamic_cast<const equal_propagator*>(&other);
}

constraint_propagator *constraint_propagator::combine_other_first(
    const constraint_runtime_info & info,
    const constraint_propagator & other,
    const constraint_runtime_info & other_info,
    const std::vector<unsigned> & other_to_common_param,
    unsigned other_pivot_param,
    std::vector<unsigned> & copy_from_argument) const
{
    unsigned arity= info.arity();
    unsigned other_arity= other_info.arity();
    
    // We swap this and other. The vector contains a map from the original
    // parameters of 'this' to the swapped combined parameters.
    vector<unsigned> this_to_new_common_param(arity);
    for (unsigned i=0 ; i != arity ; ++i)
        this_to_new_common_param[i]= other_arity+i;
    for (unsigned i=0 ; i != other_arity ; ++i) {
        if (other_to_common_param[i] < arity)
            this_to_new_common_param[other_to_common_param[i]]= i;
    }
    unsigned pivot= other_to_common_param[other_pivot_param];
    constraint_propagator * result= other.combine(
        other_info, *this, info, this_to_new_common_param, pivot,
        copy_from_argument);
    if (!result)
        return NULL;
    // We swapped this and other, and got copy_from_argument[i] which refers to
    // parameters. Its lower values refer to 'other' and higher values to
    // 'this'. We have to change that, so lower values will refer to parameters
    // of `'this', and higher to 'other'.
    for (unsigned i=0; i != copy_from_argument.size() ; ++i) {
        unsigned from_argument= copy_from_argument[i];
        if (from_argument < other_arity)
            copy_from_argument[i]= other_to_common_param[from_argument];
        else 
            copy_from_argument[i]= from_argument-other_arity;
    }
    return result;
}

//virtual
constraint_propagator *constraint_propagator::combine(
    const constraint_runtime_info & info,
    const constraint_propagator & other,
    const constraint_runtime_info & other_info,
    const std::vector<unsigned> & other_to_common_param,
    unsigned other_pivot_param,
    std::vector<unsigned> & copy_from_argument) const
{
    if (dynamic_cast<const equal_propagator*>(&other))
        return combine_other_first(info, other, other_info,
                                   other_to_common_param, other_pivot_param,
                                   copy_from_argument);
    HCSP_ASSERT(!may_combine(other));
    return NULL;
}

ostream & hcsp::operator<<(ostream& out,
                          propagation_types::result_type result)
{
    switch (result) {
      case constraint_propagator::UNSAT:      out << "UNSAT" ;      break;
      case constraint_propagator::VALID:      out << "VALID" ;      break;
      case constraint_propagator::CONSISTENT: out << "CONSISTENT" ; break;
      case constraint_propagator::POSTPONE:   out << "POSTPONE" ;   break;
      default:
        out << "unknown result_type(" << static_cast<unsigned>(result);
    }
    return out;
}

