// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <disjoint_propagators.hpp>

#include <sliced_runtime_info.hpp>

using namespace hcsp;
using namespace std;

multi_disjoint_propagator&
multi_disjoint_propagator::operator=(const multi_disjoint_propagator& other)
{
    m_propagators.clear();
    m_clause= other.m_clause;
    m_first_propagator_argument= other.m_first_propagator_argument;
    m_active_propagators[0] = other.m_active_propagators[0];
    m_active_propagators[1] = other.m_active_propagators[1];
    m_propagators.reserve(other.m_propagators.size());
    for (unsigned i=0 ; i != other.m_propagators.size(); ++i) {
        m_propagators.emplace_back(other.m_propagators[i]->clone());
    }
    return *this;
}

unsigned multi_disjoint_propagator::append_constraint(
    std::unique_ptr<constraint_propagator> propagator, unsigned arity)
{
    m_propagators.push_back(move(propagator));
    if (m_first_propagator_argument.empty())
        m_first_propagator_argument.push_back(0);
    unsigned post_last_argument= m_first_propagator_argument.back()+arity;
    m_first_propagator_argument.push_back(post_last_argument);
    return m_propagators.size()-1;
}

// virtual
bool multi_disjoint_propagator::init(constraint_runtime_info & info)
{
    m_active_propagators[0]= m_active_propagators[1]= -1U;
    unsigned num_undecided= 0;
    HCSP_ASSERT(m_propagators.size()+1 == m_first_propagator_argument.size());
    for (unsigned i=0 ; i != m_propagators.size() ; ++i) {
        const unsigned begin_args= m_first_propagator_argument[i];
        const unsigned end_args= m_first_propagator_argument[i+1];
        sliced_runtime_info propagator_info(info, begin_args, end_args);
        if (m_propagators[i]->init(propagator_info) &&
            num_undecided < 2) {
            m_active_propagators[num_undecided]= i;
            ++num_undecided;
        }
                
    }
    if (!m_clause.literals().empty()) {
        const unsigned first_literal= m_first_propagator_argument.back();
        sliced_runtime_info clause_info(info, first_literal, count_params());
        if (m_clause.init(clause_info) &&
            num_undecided < 2) {
            m_active_propagators[num_undecided]= m_propagators.size();
            ++num_undecided;
        }
    }
    HCSP_ASSERT(m_active_propagators[0] == -1U ||
               m_active_propagators[0] != m_active_propagators[1]);
    return num_undecided;
}

bool multi_disjoint_propagator::advance_active_propagators(
    unsigned i, const unsigned orig_active_propagators[2])
{
    HCSP_ASSERT(m_active_propagators[0] == -1U ||
               m_active_propagators[0] != m_active_propagators[1]);
    unsigned candidate= m_active_propagators[i];
    while (true) {
        ++candidate;
        if (candidate == m_propagators.size()+1)
            candidate= 0;
        if (candidate == orig_active_propagators[i]) {
            // completed a whole circle, nothing to find here.
            return false;
        }
        if (candidate == orig_active_propagators[1-i] ||
            candidate == m_active_propagators[1-i] )
            continue;
        m_active_propagators[i]= candidate;
        HCSP_ASSERT(m_active_propagators[0] != m_active_propagators[1]);
        return true;
    }
}

// virtual
constraint_propagator::result_type
multi_disjoint_propagator::propagate(constraint_runtime_info & info)
{
    unsigned orig_active_propagators[2]= {m_active_propagators[0],
                                          m_active_propagators[1]};
    if (m_active_propagators[0] != -1U)
        HCSP_ASSERT(m_active_propagators[0] != m_active_propagators[1]);
    while (true) {
        unsigned propagator_idx= m_active_propagators[0];
        if (propagator_idx == -1U)
            return UNSAT;
        const unsigned begin_args= begin_disjunct_argument(propagator_idx);
        const unsigned end_args= end_disjunct_argument(propagator_idx);
        sliced_runtime_info propagator_info(info, begin_args, end_args);
        result_type result= propagate_nth(propagator_idx, propagator_info);
        if ( (result == VALID && !propagator_info.pending_commit()) ||
             result == POSTPONE)
            return result;
        if (result == UNSAT &&
            advance_active_propagators(0, orig_active_propagators))
            continue;
        result_type result2= propagate_for_active_propagators_2nd(
            info, orig_active_propagators, result);
        if (result == UNSAT)
            return result2;
        if (result2 == UNSAT) {
            propagator_info.commit();
            return result;
        }
        return CONSISTENT;
    }
}

constraint_propagator::result_type
multi_disjoint_propagator::propagate_nth(unsigned disjunct_index,
                                         constraint_runtime_info & info)
{
    if (disjunct_index < m_propagators.size())
        return m_propagators[disjunct_index]->propagate(info);
    else
        return m_clause.propagate(info);
}

constraint_propagator::result_type
multi_disjoint_propagator::propagate_for_active_propagators_2nd(
    constraint_runtime_info & info, const unsigned orig_active_propagators[2],
    result_type caller_result)
{
    while (true) {
        unsigned propagator_idx= m_active_propagators[1];
        if (propagator_idx == -1U)
            return UNSAT;
        const unsigned begin_args= begin_disjunct_argument(propagator_idx);
        const unsigned end_args= end_disjunct_argument(propagator_idx);
        sliced_runtime_info propagator_info(info, begin_args, end_args);
        result_type result= propagate_nth(propagator_idx, propagator_info);
        if (caller_result == UNSAT) {
            if (result != UNSAT)
                propagator_info.commit();
            return result;
        }
        if (result == UNSAT &&
            !advance_active_propagators(1, orig_active_propagators))
            return UNSAT;
        else if ( (result == VALID && !propagator_info.pending_commit()) ||
                  result == POSTPONE)
            return result;
        else if (result != UNSAT)
            return CONSISTENT;
    }
}

bool multi_disjoint_propagator::conflicts_nth(
    const constraint_runtime_info & info, unsigned disjunct_index) const
{
    const unsigned begin_args= begin_disjunct_argument(disjunct_index);
    const unsigned end_args= end_disjunct_argument(disjunct_index);
    sliced_runtime_info propagator_info(info, begin_args, end_args);
    if (disjunct_index < m_propagators.size())
        return m_propagators[disjunct_index]->conflicts(propagator_info);
    else
        return m_clause.conflicts(propagator_info);
   
}

// virtual
bool multi_disjoint_propagator::conflicts(const constraint_runtime_info & info)
    const
{
    for (unsigned i=0 ; i != 2; ++i) {
        if (m_active_propagators[i] == -1U)
            continue;
        if (!conflicts_nth(info, m_active_propagators[i]))
            return false;
    }
    for (unsigned i=0 ; i != m_propagators.size()+1; ++i) {
        if (i != m_active_propagators[0] && i != m_active_propagators[1] &&
            !conflicts_nth(info, i))
            return false;
    }
    return true;
}

// virtual
bool multi_disjoint_propagator::constraint_check(
    const constraint_runtime_info & info) const
{
    for (unsigned i=0 ; i != m_propagators.size(); ++i) {
        const unsigned begin_args= m_first_propagator_argument[i];
        const unsigned end_args= m_first_propagator_argument[i+1];
        sliced_runtime_info propagator_info(info, begin_args, end_args);
        if (m_propagators[i]->constraint_check(propagator_info))
            return true;
    }
    
    sliced_runtime_info clause_info(info, m_first_propagator_argument.back(),
                                    count_params());
    return m_clause.constraint_check(clause_info);
}

// virtual
void multi_disjoint_propagator::print(
    std::ostream &out, const constraint_runtime_info *p_info) const throw()
{
    out << "( ";
    for (unsigned i=0 ; i != m_propagators.size(); ++i) {
        if (i != 0)
            out << " || ";
        if (p_info) {
            sliced_runtime_info disjunct_info(*p_info,
                                              begin_disjunct_argument(i),
                                              end_disjunct_argument(i));
            m_propagators[i]->print(out, &disjunct_info);
        }
    }
    if (!m_clause.literals().empty()) {
        sliced_runtime_info disjunct_info(*p_info,
                                          m_first_propagator_argument.back(),
                                          count_params());
        if (!m_propagators.empty())
            out << " || ";
        m_clause.print(out, &disjunct_info);
    }
    out << ")";
}

// virtual
unsigned multi_disjoint_propagator::watched_param_index(unsigned watch_index)
    const
{
    unsigned watches_so_far= 0;
    for (unsigned i=0 ; i != m_propagators.size(); ++i) {
        /**
         * @todo this algorithm has a poor complexity, there alternatives:
         *   -#  Use a watch sandbox so that watch_index refers only to one of
         *       the two propagators in m_active_propagators.
         *   -#  Maintain a counter for all propagators and every propagation
         *       update the counts of the propagated constraints only.
        */
        unsigned propagator_watches= m_propagators[i]->count_watches();
        unsigned local_watch_index= watch_index-watches_so_far;
        if (local_watch_index < propagator_watches) {
            unsigned local_param= m_propagators[i]->watched_param_index(
                local_watch_index);
            return local_param+m_first_propagator_argument[i];
        }
        watches_so_far += propagator_watches;
    }
    unsigned clause_watches= m_clause.count_watches();
    unsigned clause_watch_index= watch_index-watches_so_far;
    if (clause_watch_index < clause_watches) {
        unsigned clause_param= m_clause.watched_param_index(clause_watch_index);
        return clause_param+m_first_propagator_argument.back();
    } else
        return -1U;  // should not happen.
}

void multi_disjoint_propagator::explain_nth_propagation(
    unsigned disjunct_index, unsigned propagated_param,
    const iset & values_to_explain, const iset & propagated_domain, 
    const constraint_runtime_info & info_before_propagation,
    std::vector<iset> & disjunct_explanation) const
{
    if (disjunct_index == m_propagators.size()) {
        disjunct_explanation= m_clause.literals();
        return;
    }
    const unsigned begin_args= begin_disjunct_argument(disjunct_index);
    const unsigned end_args= end_disjunct_argument(disjunct_index);
    sliced_runtime_info propagator_info(info_before_propagation,
                                        begin_args, end_args);
    if (begin_args <= propagated_param && propagated_param < end_args) {
        unsigned disjunct_param= propagated_param-begin_args;
        m_propagators[disjunct_index]
            ->explain_propagation(disjunct_param, values_to_explain,
                                  propagated_domain, propagator_info,
                                  disjunct_explanation);
    } else {
        m_propagators[disjunct_index]
            ->explain_propagation(0, iset().complement(), iset(),
                                  propagator_info, disjunct_explanation);
    }
}

// virtual
void multi_disjoint_propagator::explain_propagation(
    unsigned propagated_param, const iset & values_to_explain,
    const iset & propagated_domain, 
    const constraint_runtime_info & info_before_propagation,
    vector<iset> & explanation) const
{
    explanation.clear();
    explanation.reserve(info_before_propagation.arity());
    vector<iset> disjunct_explanation;
    disjunct_explanation.reserve(info_before_propagation.arity());
    for (unsigned i=0 ; i != m_propagators.size()+1 ; ++i) {
        disjunct_explanation.clear();
        explain_nth_propagation(i, propagated_param, values_to_explain,
                                propagated_domain, info_before_propagation,
                                disjunct_explanation);
        explanation.insert(explanation.end(), disjunct_explanation.begin(),
                           disjunct_explanation.end());
    }
}

void multi_disjoint_propagator::possibly_crippling_values_nth(
    unsigned disjunct_index, unsigned propagated_param,
    const iset & propagated_domain,
    const constraint_runtime_info & info_before_propagation,
    std::vector<iset> & disjunct_values) const
{
    const unsigned begin_args= begin_disjunct_argument(disjunct_index);
    const unsigned end_args= end_disjunct_argument(disjunct_index);
    sliced_runtime_info propagator_info(info_before_propagation,
                                        begin_args, end_args);
    if (begin_args <= propagated_param && propagated_param < end_args) {
        unsigned disjunct_param= propagated_param-begin_args;
        m_propagators[disjunct_index]
            ->possibly_crippling_values(disjunct_param, propagated_domain,
                                        propagator_info, disjunct_values);
    } else {
        disjunct_values= m_clause.literals();
    }
}

// virtual
void multi_disjoint_propagator::possibly_crippling_values(
    unsigned propagated_param, const iset & propagated_domain,
    const constraint_runtime_info & info_before_propagation,
    std::vector<iset> & values) const
{
    values.clear();
    values.reserve(info_before_propagation.arity());
    vector<iset> disjunct_values;
    disjunct_values.reserve(info_before_propagation.arity());
    for (unsigned i=0 ; i != m_propagators.size()+1 ; ++i) {
        disjunct_values.clear();
        possibly_crippling_values_nth(i, propagated_param, propagated_domain,
                                      info_before_propagation, disjunct_values);
        values.insert(values.end(), disjunct_values.begin(),
                      disjunct_values.end());
    }
}

void multi_disjoint_propagator::copy_disjuncts_until(
    unsigned disjunct_index, vector<unsigned> & copy_from_argument,
    multi_disjoint_propagator & result) const
{
    for (unsigned i=0 ; i != disjunct_index ; ++i) {
        unique_ptr<constraint_propagator> propagator(m_propagators[i]->clone());
        result.append_constraint(move(propagator), disjunct_arity(i));
    }
    for (unsigned i=0 ; i != end_disjunct_argument(disjunct_index) ; ++i)
        copy_from_argument.push_back(i);
}

void multi_disjoint_propagator::intersect_with_nth(
    unsigned disjunct_index, const constraint_runtime_info & info,
    std::vector<unsigned> & copy_from_argument, bool & is_lazy,
    multi_disjoint_propagator & result) const
{
    const unsigned begin_args= begin_disjunct_argument(disjunct_index);
    const unsigned end_args= end_disjunct_argument(disjunct_index);
    sliced_runtime_info disjunct_info(info, begin_args, end_args);
    if (m_propagators[disjunct_index]->conflicts(disjunct_info)) {
        if (is_lazy)
            copy_disjuncts_until(disjunct_index, copy_from_argument, result);
        is_lazy= false;
        return;
    }
    vector<unsigned> result_disjunct_arguments;
    unique_ptr<constraint_propagator> disjunct(
        m_propagators[disjunct_index]->intersect_with(
            disjunct_info, result_disjunct_arguments));
    if (disjunct.get()) {
        if (is_lazy)
            copy_disjuncts_until(disjunct_index, copy_from_argument, result);
        is_lazy= false;
        result.append_constraint(move(disjunct),
                                 disjunct_arity(disjunct_index));
        unsigned offset= m_first_propagator_argument[disjunct_index];
        for (unsigned i=0 ; i != result_disjunct_arguments.size(); ++i)
            copy_from_argument.push_back(result_disjunct_arguments[i]+offset);
    } else if (!is_lazy) {
        disjunct.reset(m_propagators[disjunct_index]->clone());
        result.append_constraint(move(disjunct),
                                 disjunct_arity(disjunct_index));
        for (unsigned i= begin_disjunct_argument(disjunct_index);
             i != end_disjunct_argument(disjunct_index);
             ++i)
            copy_from_argument.push_back(i);
    }
}

void multi_disjoint_propagator::intersect_with_clause(
    const constraint_runtime_info & info,
    std::vector<unsigned> & copy_from_argument, bool & updated,
    multi_disjoint_propagator & result) const
{
    updated= false;
    if (m_clause.literals().empty())
        return;
    const unsigned begin_args= m_first_propagator_argument.back();
    const unsigned end_args= count_params();
    sliced_runtime_info disjunct_info(info, begin_args, end_args);
    vector<iset>& out_literals= result.m_clause.update_literals();
    for (unsigned i=0 ; i != literals().size() ; ++i) {
        if (literals()[i].contains_any_of(disjunct_info.domain(i))) {
            if (updated) {
                out_literals.push_back(literals()[i]);
                copy_from_argument.push_back(begin_args+i);
            }
        } else if (!updated) {
            copy_disjuncts_until(m_propagators.size(), copy_from_argument,
                                 result);
            for (unsigned j=0; j != i; ++j) {
                out_literals.push_back(literals()[j]);
                copy_from_argument.push_back(begin_args+j);
            }
            updated= true;
        }
    }
}

// virtual
constraint_propagator * multi_disjoint_propagator::intersect_with(
    const constraint_runtime_info & info,
    std::vector<unsigned> & copy_from_argument) const
{
    unique_ptr<multi_disjoint_propagator> result(new multi_disjoint_propagator);
    bool updated= false;
    for (unsigned i=0 ; i!= m_propagators.size() ; ++i)
        intersect_with_nth(i, info, copy_from_argument, updated, *result);
    intersect_with_clause(info, copy_from_argument, updated, *result);
    if (updated)
        return result.release();
    else
        return NULL;
}

