// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <disjunction_specification.hpp>

#include <limits>
#include <clause_propagator.hpp>

using namespace std;
using namespace hcsp;

disjunction_specification::disjunction_specification(
    const constraint_runtime_info & info,
    const constraint_propagator & propagator,
    const vector<unsigned> & local_to_external_param)
    : m_info(&info), m_local_to_external_param(local_to_external_param),
      m_constraint_arity(0),
      m_num_literals(0), m_first_literal_index(0), m_first_literal_ptr(NULL),
      m_first_constraint_index(0)
{
    propagator.get_disjunction_specification(*this);
    if (propagator.count_params() >= 0)
        set_constraint_arity(propagator.count_params()-num_literals());
}

disjunction_specification::disjunction_specification()
    : m_info(NULL), m_constraint_arity(0),
      m_num_literals(0), m_first_literal_index(0), m_first_literal_ptr(NULL),
      m_first_constraint_index(0)
{
}

void disjunction_specification::clear()
{
    m_info= NULL;
    m_local_to_external_param.clear();
    m_constraint_arity= 0;
    m_num_literals= 0;
    m_first_literal_index= 0;
    m_first_literal_ptr= NULL;;
    m_literal_storage.clear();
    m_first_constraint_index= 0;
    m_constraint.reset();
}

void disjunction_specification::resolve_literals(
    const disjunction_specification &a, const disjunction_specification &b,
    unsigned pivot)
{
    m_local_to_external_param.clear();
    m_literal_storage.clear();
    for (unsigned i=0 ; i != a.num_literals(); ++i) {
        m_local_to_external_param.push_back(a.literal_var(i));
        HCSP_ASSERT(!a.literal_set(i).empty());
        m_literal_storage.push_back(a.literal_set(i));
    }
    bool has_empty= false;
    for (unsigned i=0 ; i != b.num_literals(); ++i) {
        HCSP_ASSERT(!b.literal_set(i).empty());
        unsigned this_index= a.literal_index_of_var(b.literal_var(i));
        if (this_index < a.num_literals()) {
            HCSP_ASSERT(m_local_to_external_param[this_index]==b.literal_var(i));
            if (pivot == m_local_to_external_param[this_index]) {
                m_literal_storage[this_index].intersect(b.literal_set(i));
                has_empty= has_empty || m_literal_storage[this_index].empty();
            } else 
                m_literal_storage[this_index].insert_all_of(b.literal_set(i));
        } else {
            m_local_to_external_param.push_back(b.literal_var(i));
            m_literal_storage.push_back(b.literal_set(i));
        }
    }
    m_constraint_arity= 0;
    m_first_literal_ptr= &m_literal_storage[0];
    m_first_literal_index= 0;
    m_num_literals= m_literal_storage.size();
    if (!has_empty)
        return;
    unsigned write= 0;
    for (unsigned i=0 ; i != m_literal_storage.size(); ++i) {
        if (m_literal_storage[i].empty())
            continue;
        if (i != write) {
            m_literal_storage[write].swap(m_literal_storage[i]);
            m_local_to_external_param[write]= m_local_to_external_param[i];
        }
        ++write;
    }
    m_literal_storage.erase(m_literal_storage.begin()+write,
                            m_literal_storage.end());
    m_local_to_external_param.erase(m_local_to_external_param.begin()+write,
                            m_local_to_external_param.end());
    m_num_literals= m_literal_storage.size();
}

constraint_propagator * disjunction_specification::new_clause(
    vector<unsigned> &copy_from_argument) const
{
    copy_from_argument.clear();
    unique_ptr<clause_propagator> clause(new clause_propagator);
    vector<iset> & literals= clause->update_literals();
    copy_from_argument.resize(num_literals());
    literals.resize(num_literals());
    for (unsigned i=0 ; i != num_literals(); ++i) {
        copy_from_argument[i]= literal_var(i);
        literals[i]= literal_set(i);
    }
    return clause.release();
}

constraint_propagator *disjunction_specification::make_constraint(
    vector<unsigned> & copy_from_argument)
{
    for (unsigned i=0 ; i != num_literals() ; ++i)
        HCSP_ASSERT(!literal_set(i).empty());
    
    if (has_constraint())
        return constraint()->new_propagator(*this, copy_from_argument);
    else
        return new_clause(copy_from_argument);
}

void disjunction_specification::append_literals(
    const disjunction_specification & in_constraint)
{
    for (unsigned i=0 ; i != in_constraint.num_literals(); ++i)
        add_literal(in_constraint.literal_set(i), in_constraint.literal_var(i));
}

void disjunction_specification::append_constraint(
    const disjunction_specification & in_constraint)
{
    HCSP_ASSERT(!has_constraint());
    HCSP_ASSERT(in_constraint.has_constraint());
    
    reset_constraint(in_constraint.constraint()->clone(), num_literals());
    for (unsigned i=0 ; i!= in_constraint.constraint_arity() ; ++i)
        add_constraint_var(in_constraint.constraint_var(i));
}

void disjunction_specification::steal_constraint(
    disjunction_specification & in_constraint)
{
    HCSP_ASSERT(!has_constraint());
    reset_constraint(in_constraint.release(), num_literals());
    for (unsigned i=0 ; i!= in_constraint.constraint_arity() ; ++i)
        add_constraint_var(in_constraint.constraint_var(i));
}

constraint_propagator *disjunction_specification::combine_literal_pivot(
    const disjunction_specification & other,
    unsigned pivot, std::vector<unsigned> & copy_from_argument) const
{
    if (!is_literal(pivot) || !other.is_literal(pivot))
        return NULL;

    disjunction_specification result;
    if (constraint() && other.constraint()) {
        disjunction_specification both_constraints;
        if (!constraint()->disjoin_constraint(*this, other, both_constraints))
            return NULL;
        disjunction_specification other_literals, this_literals;
        this_literals.append_literals(*this);
        other_literals.append_literals(other);
        result.resolve_literals(this_literals, other_literals, pivot);
        result.append_literals(both_constraints);
        result.steal_constraint(both_constraints);
    } else if (other.constraint()) {
        result.resolve_literals(*this, other, pivot);
        result.append_constraint(other);
    } else if (constraint()) {
        result.resolve_literals(*this, other, pivot);
        result.append_constraint(*this);
    }
    return result.make_constraint(copy_from_argument);
}

// static
constraint_propagator *disjunction_specification::combine_constraint_pivot(
    disjunction_specification &other, unsigned pivot,
    vector<unsigned> & copy_from_argument)
{
    if (!constraint() || !other.constraint())
        return NULL;
    if (is_literal(pivot) || other.is_literal(pivot))
        return NULL;

    disjunction_specification inner_result;
    if (!constraint()->combine_base_constraints(*this, other, pivot,
                                                inner_result))
        return NULL;

    disjunction_specification result;
    //!@todo remove this if, it is only for stability vs. previous version
    if (left_var() == pivot)
        result.resolve_literals(other, *this, pivot);
    else
        result.resolve_literals(*this, other, pivot);
    HCSP_ASSERT(result.m_local_to_external_param.size()==result.num_literals());
    result.append_literals(inner_result);
    result.steal_constraint(inner_result);
    return result.make_constraint(copy_from_argument);
}

constraint_propagator *disjunction_specification::combine(
    disjunction_specification & other,
    unsigned pivot, std::vector<unsigned> & copy_from_argument)
{
    copy_from_argument.clear();
    if (has_constraint()) {
        for (unsigned i=0 ; i != constraint_arity(); ++i)
            if (other.is_literal(constraint_var(i)))
                return NULL;
    }
    if (other.has_constraint()) {
        for (unsigned i=0 ; i != other.constraint_arity(); ++i)
            if (is_literal(other.constraint_var(i)))
                return NULL;
    }

    bool this_is_literal= is_literal(pivot);
    bool other_is_literal= other.is_literal(pivot);
    
    if (this_is_literal && other_is_literal)
        return combine_literal_pivot(other, pivot, copy_from_argument);
    else if (!this_is_literal && !other_is_literal)
        return combine_constraint_pivot(other, pivot, copy_from_argument);
    else
        return NULL;
}

