// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <explanation_generators.hpp>

#include <iostream>
#include <stdexcept>
#include <constraint_propagator.hpp>
using namespace std;
using namespace hcsp;

/**
 * @brief Manages constraint data needed for maximizing conflicting domains.
 *
 * This class contains functionality to help find the biggest domains which
 * still cause conflicts for a constraint.
 *
 * This class also generates an explanation clause upon request.
 */
class explanation_generators::conflicting_info final
    : public constraint_runtime_info, private var_watchers_base
{
  public:
    /**
     * This constructor gets the domains that conflict with the analyzed
     * constraint propagator. Some of the domains are in the form of a runtime
     * info, and one is in the form of a separate parameter.
     *
     * @param info  All these domains, except for \p conflicting_param which
     *              conflict the analyzed constraint propagator.
     * @param conflicting_param
     *              If conflicting_param < info.arity() then the domains of
     *              this parameter should be taken from \p conflicting_values
     *              rather than from \p info.
     * @param conflicting_values
     *              If conflicting_param < info.arity() then this value of
     *              \p conflicting_param and the other parameters of \p info
     *              conflict with the analyzed constraint propagator.
     */
    conflicting_info(const constraint_runtime_info & info,
                     unsigned conflicting_param,
                     const iset & conflicting_values);
    
    virtual bool reduced_between_activations(unsigned) const { return true; }
    virtual queue_type originating_queue() const { return CONSTRAINT_QUEUE; }
    virtual void put_in_queue(queue_type)
    { HCSP_RUNTIME_ERROR("In constraint_queue, other queues are unsupported"); }

    virtual const iset & domain(unsigned param_index) const
    { return m_current_domains.at(param_index); }

    /**
     * @brief Set the domain
     *
     * The parameter is marked dirty, so that reread_dirty_domains()
     * will reread its domain from m_conflicting_domains.
     */
    virtual void set_domain(unsigned param_index, iset new_domain) override
    {
        const iset & orig = domain(param_index);
        if (orig.size() == new_domain.size() && 
            orig.empty() == new_domain.empty())
            return;
        if ( !m_is_dirty.at(param_index) ) {
            m_is_dirty[param_index]= true;
            m_dirty_vars.push_back(param_index);
        }
        m_current_domains.at(param_index) = move(new_domain);
    }

    virtual unsigned arity() const { return m_current_domains.size(); }
    
    virtual void deallocate_watch(unsigned, watch_type, 
                                  watch_handle_t watch_handle) override
    { deallocate_value_watch(0,0, watch_handle); }
    
    virtual watch_handle_t allocate_watch(unsigned, watch_type) override
    { return allocate_value_watch(0,0); }

    virtual void deallocate_value_watch(unsigned, iset::value_type, 
                                        watch_handle_t watch_handle)
    { m_allocated_watch_list.erase(watch_handle.get_it()); }
    
    virtual watch_handle_t allocate_value_watch(unsigned, iset::value_type)
    { 
        m_allocated_watch_list.push_back(watcher_info_type{0});
        return --m_allocated_watch_list.end();
    }

    virtual unsigned effort() const
    { return m_effort; }
    /**
     * @brief Reread all dirty domains from m_conflicting_domains.
     *
     * Any domain which was touched by domain_update is marked as dirty,
     * and is copied from m_conflicting_domains. After this, all parameters
     * are marked clean.
     */
    void reread_dirty_domains();

    /**
     * @brief Mark all removed values as conflicting for param_index.
     *
     * This method assumes that a value which is not in the current domain
     * of param_index has been \b removed by the previous propagation.
     * All such values are added to m_conflicting_domains.
     *
     * @param param_index  The parameter which has been propagated from
     *                     the state of 'everything'.
     */
    void removed_values_conflict(unsigned param_index)
    {
        const iset & param_domain= m_current_domains[param_index];
        iset & next_conflicting= m_conflicting_domains[param_index];
        if ( !param_domain.contains_all_of(next_conflicting) )
            next_conflicting.insert_all_of(iset(param_domain).complement());
    }


    /**
     * @brief Get the complemented sets of the conflicting assignment.
     *
     * This method is used for generating an explanation clause for the
     * conflicting assignment, but it is not limited to this use.
     * It takes all the conflicting domains and puts their complement in
     * the parameter 'explanation'
     *
     * @param[out] explanation  The output, i.e., the set of complemented
     *                          conflicting domains.
     */
    void negate_conflicting_assignment(vector<iset> & explanation) const
    {
        explanation.resize(m_conflicting_domains.size());
        for (unsigned i=0 ; i != explanation.size(); ++i)
            (explanation[i]= m_conflicting_domains[i]).complement();
    }
    const iset & conflicting_domain(unsigned param_index) const
    { return m_conflicting_domains[param_index]; }
    
  private:
    //! The current conflicting domains. This class tries to enlarge these sets
    vector<iset> m_conflicting_domains;

    //! The current domains, i.e., the sets returned by this->domain(i).
    vector<iset> m_current_domains;

    //! For each variable m_is_dirty[variable] != 0 iff the variable is dirty.
    vector<unsigned char> m_is_dirty;

    //! The set of dirty variables.
    vector<unsigned> m_dirty_vars;
    watch_list_t m_allocated_watch_list;
    unsigned     m_effort;
};

explanation_generators::conflicting_info::conflicting_info(
    const constraint_runtime_info & info, unsigned conflicting_param,
    const iset & conflicting_values)
    : m_conflicting_domains(info.arity()),
      m_is_dirty(info.arity(), false), 
      m_effort(info.effort()) 
{
    for (unsigned i=0 ; i != m_conflicting_domains.size() ; ++i)
        m_conflicting_domains[i]= info.domain(i);
    if (conflicting_param < m_conflicting_domains.size() )
        m_conflicting_domains[conflicting_param]= conflicting_values;
    
    m_current_domains= m_conflicting_domains;
}

void explanation_generators::conflicting_info::reread_dirty_domains()
{
    while (!m_dirty_vars.empty()) {
        unsigned var= m_dirty_vars.back();
        m_dirty_vars.pop_back();
        if ( m_is_dirty.at(var) ) {
            m_current_domains.at(var)= m_conflicting_domains.at(var);
            m_is_dirty.at(var)= false;
        }
    }
}

// static
void explanation_generators::explain_propagation(
    unsigned propagated_param, const iset &, const iset & propagated_domain,
    const constraint_runtime_info & info_before_propagation,
    vector<iset> & explanation)
{
    // Implementation rationale:
    // Without loss of generality, assume that the propagated parameter is v1.
    // The input domains imply the resulting propagation:
    // (v1 \in D1 AND v2 \in D2 AND .... AND v_n \in D_n) -> v1 \in D1'
    // Converting this to a CNF clause produces:
    // (v1 \in complement(D1) OR .... OR v_n \in complement(D_n)) OR v1 \in D1'
    // and normalizing gives the explanation clause.:
    //    (v1 \in (complement(D1)\union D1') OR
    //     v2 \in complement(D2) OR .... OR v_n \in complement(D_n) )
    const unsigned size= info_before_propagation.arity();
    HCSP_ASSERT(propagated_param < size);
    explanation.resize(size);
    for (unsigned i=0 ; i != size ; ++i) {
        explanation[i]= info_before_propagation.domain(i);
        explanation[i].complement();
    }
    explanation[propagated_param].insert_all_of(propagated_domain);
}

// static
void explanation_generators::explain_propagation_tightly(
    constraint_propagator & propagator,
    unsigned propagated_param, const iset &values_to_explain,
    const iset & propagated_domain,
    const constraint_runtime_info & info_before_propagation,
    vector<iset> & explanation)
{
    const unsigned size= info_before_propagation.arity();

    static vector<unsigned> refining_order; // order of literals to minimize
    refining_order.clear();
    refining_order.reserve(size);
    int effect= propagated_domain.empty() ? -1 : (int)propagated_param;
    int cause= info_before_propagation.latest_modified_param(effect);

    if ( cause >= 0 )
        refining_order.push_back(cause);
    int other_cause= -1;
    if ( effect < 0 && cause >= 0) // UNSAT has two causes
        other_cause= info_before_propagation.latest_modified_param(cause);
    if (other_cause >= 0)
        refining_order.insert(refining_order.begin(), other_cause);
    
    // Minimizing the 'effect' literal barely helps since much of it
    // is going to be eliminated by resolution, anyway. 
    // Minimizing the 'cause' literals helps most because this affects next
    // propagations.
    for (unsigned i=0 ; i != size ; ++i)
        if ( (int)i != effect && (int)i != cause && (int)i != other_cause)
            refining_order.push_back(i);

    if ( effect >= 0 )
        refining_order.push_back(effect);

    const unsigned *begin=&refining_order[0], *end=begin+refining_order.size();
    refined_explanation(propagator, propagated_param, values_to_explain,
                        propagated_domain,
                        begin, end, info_before_propagation, explanation);
}

// static
void explanation_generators::refined_explanation(
    constraint_propagator & propagator,
    unsigned propagated_param, const iset & values_to_explain,
    const iset & propagated_domain,
    const unsigned *refine_order_begin, const unsigned *refine_order_end,
    const constraint_runtime_info & info_before_propagation,
    vector<iset> & explanation)
{
    const unsigned size= info_before_propagation.arity();
    iset conflicting_values;
    unsigned conflicting_param;
    if (propagated_domain.empty()) {
        conflicting_param= size;
    } else {
        conflicting_param= propagated_param;
        conflicting_values= info_before_propagation.domain(propagated_param);
        conflicting_values.erase_all_of(propagated_domain);
        if ( conflicting_values.contains_any_of(values_to_explain) )
            conflicting_values.intersect(values_to_explain);
    }
    
    conflicting_info info(info_before_propagation, conflicting_param,
                          conflicting_values);
    bool handled_conflicting_param= false;
    for ( const unsigned *p_param=refine_order_begin ;
          p_param != refine_order_end ;
          ++p_param) {
        unsigned param_index= *p_param;
        handled_conflicting_param |= (param_index == conflicting_param);
        if ( info.conflicting_domain(param_index).everything() )
            continue;
        info.reread_dirty_domains();
        
        info.set_domain(param_index, move(iset().complement()));
        if (propagator.propagate(info) == UNSAT)
            info.set_domain(param_index, iset());
        info.removed_values_conflict(param_index);
    }
    HCSP_ASSERT(handled_conflicting_param || conflicting_param >=size);
    info.negate_conflicting_assignment(explanation);
}

//static
void explanation_generators::possibly_crippling_values(
    constraint_propagator & propagator,
    unsigned propagated_param, const iset & propagated_domain, 
    const constraint_runtime_info & info_before_propagation,
    std::vector<iset> & values)
{
    iset conflicting_values= info_before_propagation.domain(propagated_param);
    conflicting_values.erase_all_of(propagated_domain);
    conflicting_info info(info_before_propagation, propagated_param,
                          conflicting_values);
    const iset everything(make_pair(iset::MIN_DOMAIN, iset::MAX_DOMAIN));
    for (unsigned i=0 ; i != info_before_propagation.arity() ; ++i) {
        if ( info.conflicting_domain(i).everything() )
            continue;
        info.reread_dirty_domains();
        
        info.set_domain(i, move(iset().complement()));
        iset values_kept;
        if (propagator.init(info) && propagator.propagate(info) != UNSAT)
            values_kept= info.domain(i);
        if (values_kept.everything())
            continue;
        if (i> values.size())
            values.resize(i-1, everything);
        values.push_back(values_kept);
    }
    if (propagated_param < values.size()) {
        values[propagated_param].erase_all_of(propagated_domain);
    }
}
