// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <cmath>
#include <cstdlib>
#include <unistd.h>
#include <typeinfo>
#include <fstream>
#include <flags.hpp>
using namespace std;
using namespace hcsp;

bool flags::parse(int argc, char *argv[])
{
    string legal_args_list= legal_args();
    const char *legal_args_list_str= legal_args_list.c_str();
    while (true) {
        int option= getopt(argc, argv, legal_args_list_str);
        if (option < 0)
            break;
        if (option == '?') {
            usage(cout, argv[0]);
            return false;
        }
        if (!parse_option(option)) {
            cerr << "Bad option -" << static_cast<char>(option) << "\n";
            usage(cout, argv[0]);
            return false;
        }
    }
    if (help)
        usage(cout, argv[0]);
    else
        validate_consistent();
    return true;
}

void flags::usage(ostream & out, const char * argv0)
{
    out << "Usage: " << argv0 << " ";
    usage_short(out);
    out << "\n";
    usage_long(out);
}

void flags::usage_short(ostream & out)
{
    out << "[-h]";
}

bool flags::parse_option(char option)
{
    if (option == 'h') {
        help= true;
        return true;
    } else {
        return false;
    }
}

void flags::usage_long(ostream & out)
{
    out << "  -h  Gives this help message\n";
}

void hcsp_flags::validate_consistent() const
{
    if (intersect_with_ground && has_proof_option())
        throw runtime_error("Can't work with proofs when intersect_with_ground "
                            "is in effect, try the -G flag.\n"
                            "Note: it will make a different runs.");
}

//virtual
void hcsp_flags::apply(solver & problem) const
{
    if (!quiet_start)
        cout << "c Seed is " << hcsp::random::set_seed(seed) << endl;
    problem.report_level(report_level);
    problem.set_validate_unsat(validate_unsat);
    problem.set_produce_proof(produce_proof);
    if ( max_conflict_clause+1 != 0)
        problem.max_conflict_clause(max_conflict_clause);
    if ( opportunistic_bj > 0 )
        problem.opportunistic_back_jumping_threshold(opportunistic_bj);
    if ( restart_threshold >= 0 )
        problem.restart_threshold(restart_threshold);
    if ( backtrack_limit >= 0)
        problem.backtrack_limit(backtrack_limit);
    if (cost_lowering_factor != 0)
        problem.set_cost_lowering_factor(cost_lowering_factor);
    if (analysis_cost_factor > 0)
        problem.set_analysis_cost_factor(analysis_cost_factor);
    if (conflict_cost_factor > 0)
        problem.set_conflict_cost_factor(conflict_cost_factor);
    if (only_optimal)
        problem.set_only_optimal();
    if (all_results)
        problem.set_all_results();
    if ( effort_boost >= 0)
        problem.conflicts_before_boosting_effort(effort_boost);
    if ( keep_constraints_from_conflicts >= 1 )
        problem.keep_constraints_from_n_conflicts(
            keep_constraints_from_conflicts);
    if ( erase_after_backtracks >= 1)
        problem.erase_constraints_every_n_conflicts(erase_after_backtracks);
    problem.produce_proof(has_proof_option());
    if ( intersect_with_ground >= 0)
        problem.should_intersect_with_ground(intersect_with_ground);
    int relax= min<unsigned>(solver::COMPLETE_RELAXATION, relaxation);
    problem.relaxation(static_cast<solver::relaxation_types>(relax));
    problem.cause_relaxation(cause_relaxation);
    problem.constraints_combination(constraints_combination);
    problem.should_explainer_check_age(explainer_check_age);
    problem.should_augment_explanation(augment_explanation);
}


void hcsp_flags::usage_short(ostream & out)
{
    flags::usage_short(out);
    out << " [-s seed] [-l report_level] [-o opportunistic_threshold]\n"
        << "       [-c max_conflict_clause] [-b backtrack_limit] "
        <<         "[-r restart_threshold]\n"
        << "       [-C keep_constraints_from_conflicts] "
        <<         "[-E erase_after_backtracks]\n"
        << "       [-e conflicts_before_boosting_effort] "
        <<         "[-f cost_lowering_factor]\n"
        << "       [-i analysis_cost_factor] [-k conflict_cost_factor] "
        <<         "[-a|-p|-y|-A|-g|-J|-m]\n"
        << "       [file1 [file2[...]]]";
}

void hcsp_flags::usage_long(ostream & out)
{
    flags::usage_long(out);
    out << "  -p  Toggle whether a proof is to be produced (default no)\n"
        << "  -V  Toggle whether to validate UNSAT using the proof mechanism "
        <<            "(default no)\n"
        << "  -G  Toggle whether to intersect conflict clause domains with "
        <<            "ground zero\n"
        << "      domains\n"
        << "  -g  Toggle whether to let explainer check domain ages "
        << "(default yes)\n"
        << "  -A  Toggle whether to augment explanations (default yes)\n"
        << "      Augmented explanations are weaker clauses than strict "
        <<        "explanations; they\n"
        << "      add freedom to the pivot if the resolution will hide it\n"
        << "  -a  Toggle whether to use global all_different or not\n"
        << "  -l N Set the report level to 0<=N<=4 (default 0)\n"
        << "  -j N Set the relaxation mode for literals:\n"
        << "       0=none, 1=level, 2=complete (default 2)\n"
        << "       Where 1=level means that a literal is relaxed only if "
        << "it moves\n"
        << "       a literal to an earlier level.\n"
        << "  -J   Toggle whether to relax the cause node of a conflicting"
        <<       " constraint (default no)\n"
        << "  -m   Toggle whether to combine general constraints into a new "
        << "constraint.\n"
        << "       Attempt to avoid explanation clauses (default yes)\n"
        << "  -c N  Will set the maximum size of conflict clause which is "
        << "still guaranteed\n"
        << "        to be inserted\n"
        << "  -o opportunistic_threshold\n"
        << "     The minimum number of decisions to be undone below which "
        << "opportunistic\n"
        << "     back-jumping is not performed.\n"
        << "  -b N  Maximum backtracks allowed\n"
        << "  -r N>0  Number of backtracks performed before the first "
        << "restart takes place\n"
        << "  -r 0    Disable restarts\n"
        << "  -e N    Conflicts per constraint before boosting its effort\n"
        << "  -f N    Upon backtrack multiply all costs by (N-1)/N\n"
        << "  -i FLOAT By how much variables in conflicts have higher costs\n"
        << "           than initial cost\n"
        << "  -k FLOAT In decision prefer latest conflict variables by this "
        <<            "much\n"
        << "  -C N    Clauses used in the last N conflicts are not erased "
        << "during\n"
        << "         old clause erasure\n"
        << "  -E N   Perform clause erasure every N conflicts\n";
}

static solver::reporting_levels parse_report_level(const char* text)
{
    return static_cast<solver::reporting_levels>(atoi(text));
}


bool hcsp_flags::parse_option(char option)
{
    if (flags::parse_option(option))
        return true;
    switch (option) {
      case 'V': validate_unsat=!validate_unsat;                  break;
      case 'p': produce_proof=!produce_proof;                    break;
      case 'd': all_results = !all_results;                      break;
      case 'D': only_optimal= !only_optimal;                     break;
      case 'G': intersect_with_ground=!intersect_with_ground;    break;
      case 'A': augment_explanation=!augment_explanation;        break;
      case 'g': explainer_check_age=!explainer_check_age;        break;
      case 'a': use_all_diff= (use_all_diff <= 0);               break;
      case 'J': cause_relaxation= !cause_relaxation;             break;
      case 'm': constraints_combination= !constraints_combination;break;
      case 'l': report_level= parse_report_level(optarg);        break;
      case 'f': cost_lowering_factor=  strtoul(optarg, NULL,0);  break;
      case 'i': analysis_cost_factor=  strtod(optarg, NULL);     break;
      case 'k': conflict_cost_factor=  strtod(optarg, NULL);     break;
      case 'j': relaxation=            strtoul(optarg, NULL, 0); break;
      case 'o': opportunistic_bj=      strtoul(optarg, NULL, 0); break;
      case 'c': max_conflict_clause=   strtoul(optarg, NULL, 0); break;
      case 'b': backtrack_limit=       strtoul(optarg, NULL, 0); break;
      case 'r': restart_threshold=     strtoul(optarg, NULL, 0); break;
      case 'e': effort_boost=          strtoul(optarg, NULL, 0); break;
      case 'C': keep_constraints_from_conflicts= strtoul(optarg, NULL, 0);break;
      case 'E': erase_after_backtracks=strtoul(optarg, NULL, 0); break;
      case 's': seed= strtoul(optarg, NULL, 0);                  break;
      default:
        return false;
    }
    return true;
}
