// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <clause_propagator.hpp>
#include <solver.hpp>
#include <algorithm>
#include "formula_impl.hpp"
using namespace hcsp;
using namespace std;

pool_alloc<formula_impl> formula_impl::m_pool_base;

pool_alloc<not_formula_impl> not_formula_impl::m_pool_not;

pool_alloc<binary_operator_formula_impl>
binary_operator_formula_impl::m_pool_binary_op;

pool_alloc<clause_formula_impl> clause_formula_impl::m_pool_clause;

// static
void formula_impl::add_bool_clause(solver & solver, int literal1, int literal2)
{
    int literals[]={literal1, literal2};
    add_bool_clause(solver, literals, 2);
}

// static
void formula_impl::add_bool_clause(solver & solver,
                                   int literal1, int literal2, int literal3)
{
    int literals[]={literal1, literal2, literal3};
    add_bool_clause(solver, literals, 3);
}

// static
void formula_impl::add_bool_clause(solver & solver, int literals[],
                                   unsigned size)
{
    clause_propagator * p_clause= new clause_propagator;
    unique_ptr<constraint_propagator> p_constraint(p_clause);
    clause_propagator::literals_type & sets = p_clause->update_literals();
    sets.reserve(size);
    constraint_connectivity_info &
        info= solver.add_constraint(move(p_constraint));
    for (unsigned i=0 ; i != size; ++i) {
        if (literals[i] < 0) { // a negated literal?
            info << neg(literals[i]); // negate again to get variable index
            sets.push_back(iset(0));
        } else {
            // literals[i] >= 0
            info << literals[i];
            sets.push_back(iset(1));
        }
    }
}

//virtual
formula_impl* formula_impl::and_with_clause(clause_formula_impl * clause)
{
    // The right hand side is not a clause, fall-back to a simple and_with
    return clause->and_with(this, false);
}

//virtual
formula_impl* formula_impl::or_with_clause(clause_formula_impl * clause)
{
    // The right hand side is not a clause, fall-back to a simple or_with
    return clause->or_with(this, false);
}

//static
bool formula_impl::is_true_var(const solver & solver, unsigned variable)
{
    const iset & domain= solver.var_domain(variable);
    return domain.singleton() && domain.contains(1);
}

// static
bool formula_impl::is_false_var(const solver & solver, unsigned variable)
{
    const iset & domain= solver.var_domain(variable);
    return domain.singleton() && domain.contains(0);
}

//static 
true_formula_impl* true_formula_impl::get()
{
    static true_formula_impl singleton;
    return &singleton;
}

//virtual
unsigned true_formula_impl::reify(solver & solver) const
{
    return solver.singleton_var(1);
}

//virtual
void true_formula_impl::print(std::ostream & out,
                              const variable_print_strategy &,
                              op_priority) const
{
    out << "true";
}

//static
false_formula_impl* false_formula_impl::get()
{
    static false_formula_impl singleton;
    return &singleton;
}

//virtual
void false_formula_impl::constrain_solver(solver & solver) const
{
    int literals[]={(int)solver.singleton_var(0)};
    add_bool_clause(solver, literals, 1); // effectively forcing a conflict
}

//virtual
unsigned false_formula_impl::reify(solver & solver) const
{
    return solver.singleton_var(0);
}

//virtual
void false_formula_impl::print(std::ostream & out,
                               const variable_print_strategy &,
                               op_priority) const
{
    out << "false";
}

not_formula_impl::not_formula_impl(formula_impl * in_operand)
    : m_operand(in_operand)
{
    if (in_operand)
        in_operand->link();
}

// virtual
not_formula_impl::~not_formula_impl()
{
    if (m_operand)
        m_operand->unlink();
    m_operand= NULL;
}

// virtual
void not_formula_impl::constrain_solver(solver & solver) const
{
    int inner_variable= operand()->reify(solver);
    const iset & inner_domain = solver.var_domain(inner_variable);
    if (inner_domain.contains(0)) {
        if (!inner_domain.singleton())
            solver.set_var_domain(inner_variable, iset(0));
        return;
    }

    int literals[]={neg(inner_variable)};
    add_bool_clause(solver, literals, 1); // effectively forcing a conflict
}

// virtual
unsigned not_formula_impl::reify(solver & solver) const
{
    int inner_variable= operand()->reify(solver);
    if (is_true_var(solver, inner_variable))
        return solver.singleton_var(0);
    if (is_false_var(solver, inner_variable))
        return solver.singleton_var(1);
    
    int result_variable= solver.add_var(iset(iset::value_range_type(0,1)));

    add_bool_clause(solver, result_variable, inner_variable);
    add_bool_clause(solver, neg(result_variable), neg(inner_variable));
    
    return result_variable;
}

// virtual
formula_impl* not_formula_impl::and_with(formula_impl* other, bool initial)
{
    if (this == other)
        return this;
    if (operand() == other) // A and not(A)
        return false_formula_impl::get();
    if (initial)
        // Maybe other is 'false' or 'true' and it can optimize this operation
        return other->and_with(this, false);
    return new binary_and_formula_impl(this, other);
}

// virtual
formula_impl* not_formula_impl::or_with(formula_impl* other, bool initial)
{
    if (this == other)
        return this;
    if (operand() == other) // A or not(A)
        return true_formula_impl::get();
    if (initial)
        // Maybe other is 'false' or 'true' and it can optimize this operation
        return other->or_with(this, false);
    return new binary_or_formula_impl(this, other);
}

// virtual
void not_formula_impl::print(std::ostream & out,
                             const variable_print_strategy &strategy,
                             op_priority) const
{
    out << "not(" << operand()->printer(strategy) << ")";
}

binary_operator_formula_impl::binary_operator_formula_impl(
    formula_impl *first_formula_impl, formula_impl *second_formula_impl)
    : m_first(first_formula_impl), m_second(second_formula_impl)
{
    if (first_formula_impl)
        first_formula_impl->link();
    if (second_formula_impl)
        second_formula_impl->link();
}

binary_operator_formula_impl::~binary_operator_formula_impl()
{
    if (m_first)
        m_first->unlink();
    m_first= NULL;
    if (m_second)
        m_second->unlink();
    m_second= NULL;
}

//virtual
formula_impl* binary_operator_formula_impl::and_with(formula_impl *other,
                                                     bool initial)
{
    if (other == this)
        return this;
    if (initial)
        // Maybe other is 'false' or 'true' and it can optimize this operation
        return other->and_with(this, false);
    return new binary_and_formula_impl(this, other);
}

//virtual
formula_impl* binary_operator_formula_impl::or_with(formula_impl *other,
                                                    bool initial)
{
    if (other == this)
        return this;
    if (initial)
        // Maybe other is 'false' or 'true' and it can optimize this operation
        return other->or_with(this, false);
    return new binary_or_formula_impl(this, other);
}

//virtual
void binary_and_formula_impl::constrain_solver(solver & solver) const
{
    first()->constrain_solver(solver);
    second()->constrain_solver(solver);
}

//virtual
unsigned binary_and_formula_impl::reify(solver & solver) const
{
    int first_variable= first()->reify(solver);
    if ( is_false_var(solver,first_variable) )
        return solver.singleton_var(0);
    
    int second_variable= second()->reify(solver);
    if (is_false_var(solver,second_variable) )
        return solver.singleton_var(0);

    if (is_true_var(solver,first_variable))
        return second_variable;
    if (is_true_var(solver,second_variable))
        return first_variable;
    int result_variable= solver.add_var(iset(iset::value_range_type(0,1)));

    // There are three rules connecting first,second and result:
    // 1. result  --> (first && second)
    add_bool_clause(solver, neg(result_variable), first_variable);
    add_bool_clause(solver, neg(result_variable), second_variable);

    // 2. (first && second) --> result
    add_bool_clause(solver,
                    neg(first_variable), neg(second_variable), result_variable);

    return result_variable;
}

//virtual
void
binary_and_formula_impl::print(std::ostream & out,
                               const variable_print_strategy &strategy,
                               op_priority priority) const
{
    bool need_parentheses= ((int)priority > (int)AND_OP);
    if (need_parentheses)
        out << "(";
    first()->print(out, strategy, AND_OP);
    out << "&&";
    second()->print(out, strategy, AND_OP);
    if (need_parentheses)
        out << ")";
}

//virtual
void binary_or_formula_impl::constrain_solver(solver & solver) const
{
    int first_variable= first()->reify(solver);
    if ( is_true_var(solver,first_variable) )
        return;
    if (is_false_var(solver,first_variable) ) {
         second()->constrain_solver(solver);
         return;
    }
        
    int second_variable= second()->reify(solver);
    if ( is_true_var(solver,second_variable) )
        return;
    add_bool_clause(solver, first_variable, second_variable);
}

//virtual
unsigned binary_or_formula_impl::reify(solver & solver) const
{
    int first_variable= first()->reify(solver);
    if (is_true_var(solver, first_variable))
        return first_variable;
    int second_variable= second()->reify(solver);
    if ( is_false_var(solver, first_variable) ||
         is_true_var(solver,second_variable))
        return second_variable;
    if (is_false_var(solver, second_variable))
        return first_variable;

    int result_variable= solver.add_var(iset(iset::value_range_type(0,1)));

    // There are three rules connecting first,second and result:
    // 1. result  --> (first || second)
    add_bool_clause(solver,
                    neg(result_variable), first_variable, second_variable);

    // 2. first --> result
    add_bool_clause(solver, neg(first_variable), result_variable);
    
    // 3. second --> result
    add_bool_clause(solver, neg(second_variable), result_variable);

    return result_variable;
}

//virtual
void
binary_or_formula_impl::print(std::ostream & out,
                              const variable_print_strategy & strategy,
                              op_priority priority) const
{
    bool need_parentheses= ((int)priority > (int)OR_OP);
    if (need_parentheses)
        out << "(";
    first()->print(out, strategy, OR_OP);
    out << "||";
    second()->print(out, strategy, OR_OP);
    if (need_parentheses)
        out << ")";
}

clause_formula_impl::clause_formula_impl(unsigned variable, const iset & set)
{
    m_clause.reserve(1); // clause is optimized for the single entry case
    m_clause.push_back(clause_data::value_type(variable, set));
}

//virtual
void clause_formula_impl::constrain_solver(solver & solver) const
{
    if (m_clause.empty())
        throw std::runtime_error("clause_formula_impl::constrain_solver empty"
                                 "clauses should be handled upstream");
    clause_propagator * p_clause= new clause_propagator;
    unique_ptr<constraint_propagator> p_constraint(p_clause);
    clause_propagator::literals_type & sets = p_clause->update_literals();
    sets.reserve(m_clause.size());
    constraint_connectivity_info &
        info= solver.add_constraint(move(p_constraint));
    for (unsigned i=0 ; i != m_clause.size(); ++i) {
        sets.push_back(m_clause[i].m_set);
        info << m_clause[i].m_variable;
    }
}

//virtual
unsigned clause_formula_impl::reify(solver & solver) const
{
    unsigned result_variable= solver.add_var(iset(iset::value_range_type(0,1)));
    
    variable_implies_clause(solver, result_variable);
    clause_implies_variable(solver, result_variable);

    return result_variable;
}

//virtual
bool clause_formula_impl::evaluate(const solver & solver) const
{
    for (unsigned i=0 ; i != m_clause.size() ; ++i) {
        const iset & domain= solver.var_domain(m_clause[i].m_variable);
        if ( m_clause[i].m_set.contains_all_of(domain) )
            return true;
    }
    return false;
}

void clause_formula_impl::variable_implies_clause(solver & solver,
                                                  unsigned variable) const
{
    clause_propagator * p_clause= new clause_propagator;
    unique_ptr<constraint_propagator> p_constraint(p_clause);
    clause_propagator::literals_type & sets = p_clause->update_literals();
    sets.reserve(m_clause.size()+1);
    constraint_connectivity_info &
        info= solver.add_constraint(move(p_constraint));
    info << variable;
    sets.push_back(iset(0));
    for (unsigned i=0 ; i != m_clause.size(); ++i) {
        sets.push_back(m_clause[i].m_set);
        info << m_clause[i].m_variable;
    }
    
}

void clause_formula_impl::clause_implies_variable(solver & solver,
                                                  unsigned variable) const
{
    // Need to show that every literal implies the variable
    for (unsigned i=0 ; i != m_clause.size(); ++i) {
        clause_propagator * p_clause= new clause_propagator;
        unique_ptr<constraint_propagator> p_constraint(p_clause);
        clause_propagator::literals_type & sets = p_clause->update_literals();
        sets.reserve(2);
        constraint_connectivity_info &info= solver.add_constraint(
            move(p_constraint));
        info << m_clause[i].m_variable << variable;
        sets.push_back(m_clause[i].m_set);
        sets.back().complement();
        sets.push_back(iset(1));
    }
}

//virtual
formula_impl* clause_formula_impl::make_not()
{
    formula_impl *result= NULL;
    if (m_clause.empty())
        throw std::runtime_error("clause_formula_impl::make_not empty "
                                 "clauses should be handled upstream");
    
    for (unsigned i= 0 ; i < m_clause.size() ; ++i) {
        formula_impl *next= NULL;
        next= new clause_formula_impl(m_clause[i].m_variable,
                                      iset(m_clause[i].m_set).complement());
        if (result)
            result= new binary_and_formula_impl(result, next);
        else
            result= next;
    }
    return result;
}

//virtual
formula_impl* clause_formula_impl::and_with(formula_impl *other, bool initial)
{
    if (other == this)
        return this;

    if (initial)
        return other->and_with_clause(this);
    else
         return other->and_with(this, false);
}

//virtual
formula_impl* clause_formula_impl::or_with(formula_impl *other, bool initial)
{
    if (other == this)
        return this;
    if (initial)
        return other->or_with_clause(this);
    else
        return other->or_with(this, false);
}

//virtual
formula_impl* clause_formula_impl::and_with_clause(clause_formula_impl *other)
{
    if ( m_clause.size() == 1 && other->m_clause.size() == 1 &&
         m_clause[0].m_variable == other->m_clause[0].m_variable) {
        iset intersected= m_clause[0].m_set;
        
        intersected.intersect(other->m_clause[0].m_set);
        
        if (intersected.empty() )
            return false_formula_impl::get();
        if (intersected == m_clause[0].m_set)
            return this;
        if (intersected == other->m_clause[0].m_set)
            return other;
        return new clause_formula_impl(m_clause[0].m_variable, intersected);
    }
    return new binary_and_formula_impl(other, this);
}

//virtual
formula_impl* clause_formula_impl::or_with_clause(clause_formula_impl *other)
{
    if (m_clause.size() < other->m_clause.size())
        return other->or_with_clause(this);
    clause_data result;
    result.reserve(m_clause.size()+other->m_clause.size());
    result= other->m_clause;
    sort(result.begin(), result.end());

    unsigned sorted_elements= result.size();
    for (unsigned i=0 ; i != m_clause.size() ; ++i) {
        const iset & set = m_clause[i].m_set;
        unsigned variable = m_clause[i].m_variable;
        clause_data::iterator begin= result.begin();
        clause_data::iterator end= result.begin()+sorted_elements;
        
        clause_data::iterator it= lower_bound(begin, end, variable);
        if (it != end && it->m_variable == variable) {
            it->m_set.insert_all_of(set);
            if (it->m_set.everything())
                return true_formula_impl::get();
        } else {
            result.push_back(literal_data(variable, set));
        }
    }
    if (m_clause.equivalent(result))
        return this;
    else if (other->m_clause.equivalent(result))
        return other;
    else
        return new clause_formula_impl(result);
}

//virtual
void clause_formula_impl::print(std::ostream & out,
                                const variable_print_strategy &strategy,
                                op_priority) const
{
    m_clause.print(out, strategy);
}

