// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

/**
 * @file formula_impl.hpp
 * @brief This file defines formula
 *
 * The formulas can be used to define a CSP problem in a more natural way,
 * or can result from an automatic process such as interpolant generation
 */

#ifndef HCSP_FORMULA_IMPL_HPP_
#define HCSP_FORMULA_IMPL_HPP_

#include <iosfwd>
#include <clause_data.hpp>
#include <pool_alloc.hpp>

namespace hcsp {
class solver;
class clause_formula_impl;
/**
 * @brief A base class for all formula implementation classes (not for users)
 *
 * This is an interface for concrete implementation classes of formulae.
 * A formula will point to a concrete formula_impl object.
 *
 * These objects are reference counted; the user of the object has to
 * mark when it is used or not by calling link() or unlink() methods.
 * This means that the object should never be deleted directly.
 */
class formula_impl: private no_copy
{
  protected:
    //! The priority of the operators, lower numbers mean lower priority.
    enum op_priority { EXTERNAL, OR_OP, AND_OP };
  public:
    /**
     * @brief An object that uses variable_print_strategy for printing a
     *        formula_impl object.
     *
     * This object prints a formula, passed to it during construction,
     * using the given strategy to print variable names. For example
     * @code
     *   out << printer_type(my_strategy, first_formula) << " "
     *       << printer_type(my_strategy, second_formula, OR_OP) << "\n";
     * @endcode
     * See formula_impl::print() for more details.
     */
    class printer_type
    {
      public:
        /**
         * @brief Construct a printer object
         *
         * @param strategy    The strategy used for printing variable names
         * param formula_impl The formula to print
         * @param priority    The priority of the calling environment.
         *                    see formula_impl::print() for more details
         */
        printer_type(const variable_print_strategy &strategy,
                     const formula_impl &formula_impl,
                     op_priority priority=EXTERNAL)
            : m_strategy(strategy), m_formula_impl(formula_impl),
              m_priority(priority)
        {}
        
        //! Using this printer, print the formula
        friend std::ostream & operator<<(std::ostream & out,
                                         const printer_type &printer)
        {
            printer.m_formula_impl.print(out, printer.m_strategy,
                                         printer.m_priority);
            return out;
        }
      private:
        const variable_print_strategy &m_strategy; //!< The strategy to use  
        const formula_impl & m_formula_impl;       //!< The formula to print
        const op_priority m_priority;              //!< Print with this priority
    }; // class formula_impl::printer_type
    formula_impl() : m_references(0) {}

    //Make sure we are using pool_alloc's free list and not malloc/free.
    void* operator new(std::size_t size, const std::nothrow_t&) throw()
    { return size == sizeof(formula_impl) ? m_pool_base.allocate(1) : NULL; }
    
    void* operator new(size_t size) throw (std::bad_alloc)
    {
        void * ret= operator new(size, std::nothrow);
        if (ret)
            return ret;
        else
            throw std::bad_alloc();
    }

    void operator delete(void* ptr)
    { m_pool_base.deallocate(reinterpret_cast<formula_impl*>(ptr), 1); }
    
    //! Reference counting: mark that this object has one more use.
    void link() { ++m_references; }
    //! Reference counting: mark that this object has one less use, free it if
    //! there are no more uses.
    void unlink()
    {
        if (m_references && --m_references == 0)
            delete this;
    }
    
    /**
     * @brief Add this formula as a constraint to the solver.
     *
     * Force any satisfying assignment to evaluate the formula as 'true'.
     *
     * @param sovler  The solver object to be constrained by this formula.
     * @post The formula constrains the solver
     */
    virtual void constrain_solver(solver & solver) const= 0;
    
    /**
     * @brief Adds a variable which will hold the evaluation of this formula.
     *
     * Returns an index of a variable. Force any satisfying assignment to
     * evaluate this variable to the same value as the formula.
     *
     * @param sovler  The solver object to be constrained.
     *
     * @return The index of the variable to get the same value as the formula.
     *
     * @post The returned variable is constraint to the evaluation of the
     *       formula, in the solver.
     */
    virtual unsigned reify(solver & solver) const = 0;

    /**
     * @brief Check if the formula is true in the current complete assignment
     *
     * @param sovler  The solver whose variables are consulted.
     *
     * @return True if the formula is satisfied.
     *
     * @pre The solver contains a complete assignment to all variables 
     *      participating in the formula
     */
    virtual bool evaluate(const solver & solver) const = 0;

    /**
     * @brief Return a formula which is a negation of this formula
     *
     * @return A formula which, when evaluated, returns the Boolean negation
     *         of the input formula.
     */
    virtual formula_impl* make_not()=0;
    
    /**
     * @brief Return a formula resulting from this AND other.
     *
     * @param other   A pointer to the other formula to AND
     * @param initial In case of recursion initial=true only for the upper
     *                stage of the recursion deeper calls get initial=false
     * @return A formula which can be old or a new one. The caller is
     *         responsible for increasing the reference count at least once.
     */
    virtual formula_impl* and_with(formula_impl *other, bool initial=true) = 0;
    
    /**
     * @brief Return a formula resulting from this OR other.
     *
     * @param other   A pointer to the other formula to OR
     * @param initial In case of recursion initial=true only for the upper
     *                stage of the recursion deeper calls get initial=false
     * @return A formula which can be old or a new one. The caller is
     *         responsible for increasing the reference count at least once.
     */
    virtual formula_impl* or_with(formula_impl * other, bool initial=true) = 0;

    //!AND this formula with the given clause formula
    virtual formula_impl* and_with_clause(clause_formula_impl * clause);
    //!OR this formula with the given clause formula
    virtual formula_impl* or_with_clause(clause_formula_impl * clause);
    /**
     * @brief Print the formula in a human readable form.
     *
     * @param out       Print to this stream.
     * @param strategy  The strategy to use for printing variable names
     * @param priority  The priority of the calling environment
     *                  if the printed operator has a lower priority
     *                  than the environment then parentheses may required.
     *                  for example in "B && (C || D) && E"
     *                  the "||" operator has a lower priority than its
     *                  surrounding environment and parentheses were applied.
     */
    virtual void print(std::ostream & out,
                       const variable_print_strategy & strategy,
                       op_priority priority=EXTERNAL) const = 0;

    printer_type printer(const variable_print_strategy &strategy,
                         op_priority priority=EXTERNAL) const
    { return printer_type(strategy, *this, priority); }
  protected:
    virtual ~formula_impl() {}
    /**
     * @brief For a literal ID, create a negated literal ID.
     *
     * An non-negated literal has the ID of a Boolean variable, which is
     * nonnegative. A negated literal has a negative value; negating it gives
     * an ID of a Boolean variable associated with this literal.  Negating
     * twice gives the original literal.
     *
     * @param literal   The original literal ID
     * @return The ID of the negated literal.
     */
    static int neg(int literal) { return -1-literal; }
    //! Put a Boolean clause with three literals into the solver
    static void add_bool_clause(solver & solver,
                                int literal1, int literal2, int literal3);
    //! Put a Boolean clause with two literals into the solver
    static void add_bool_clause(solver & solver, int literal1, int literal2);
    /**
     * @brief Put a Boolean clause with several literals into the solver
     *
     * @param solver    The solver that accepts the new clause
     * @param literals,size  An array and its size containing a list of
     *                       literals to be added.
     */
    static void add_bool_clause(solver & solver, int literals[], unsigned size);
    //! Is this variable constantly true
    static bool is_true_var(const solver & solver, unsigned variable);
    //! Is this variable constantly false
    static bool is_false_var(const solver & solver, unsigned variable);
  private:
    //! The reference count of the object, when it reaches zero the object
    //! is deleted.
    unsigned m_references;
    static pool_alloc<formula_impl> m_pool_base;
}; // class formula_impl

/**
 * @brief A trivial formula which is always true
 *
 * This is a singleton.
 */
class true_formula_impl: public formula_impl
{
  public:
    //! Get a singleton object
    static true_formula_impl * get();
    virtual void constrain_solver(solver & ) const {}
    virtual unsigned reify(solver & solver) const;
    virtual bool evaluate(const solver &) const { return true; }
    virtual formula_impl* make_not();
    virtual formula_impl* and_with(formula_impl * other, bool =true)
    { return other; }
    virtual formula_impl* or_with(formula_impl *, bool =true) { return this; }
    virtual void print(std::ostream & out,
                       const variable_print_strategy & strategy,
                       op_priority priority) const;
  private:
    true_formula_impl()
    {
        //it is a singleton, since its construction it is always linked 
        link();
    }
}; // class true_formula_impl


/**
 * @brief A trivial formula which is always false
 *
 * This is a singleton.
 */
class false_formula_impl: public formula_impl
{
  public:
    static false_formula_impl * get();
    virtual void constrain_solver(solver & solver) const;
    virtual unsigned reify(solver & solver) const;
    virtual bool evaluate(const solver &) const { return false; }
    virtual formula_impl* make_not() { return true_formula_impl::get(); }
    virtual formula_impl* and_with(formula_impl* , bool =true) { return this; }
    virtual formula_impl* or_with(formula_impl* other, bool =true)
    { return other; }
    
    virtual void print(std::ostream & out,
                       const variable_print_strategy & strategy,
                       op_priority priority) const;
  private:
    false_formula_impl()
    {
        //it is a singleton, since its construction it is always linked 
        link();
    }
}; // class false_formula_impl

/**
 * @brief A negation formula
 *
 */
class not_formula_impl: public formula_impl
{
  public:
    not_formula_impl(formula_impl * operand);
    
    //Make sure we are using pool_alloc's free list and not malloc/free.
    void* operator new(std::size_t size, const std::nothrow_t&) throw()
    {
        if (size == sizeof(not_formula_impl))
            return m_pool_not.allocate(1);
        else
            return NULL;
    }
    
    void* operator new(size_t size) throw (std::bad_alloc)
    {
        void * ret= operator new(size, std::nothrow);
        if (ret)
            return ret;
        else
            throw std::bad_alloc();
    }
    void operator delete(void* ptr)
    {
        m_pool_not.deallocate(reinterpret_cast<not_formula_impl*>(ptr), 1);
    }
    
    
    virtual void constrain_solver(solver & solver) const;
    virtual unsigned reify(solver & solver) const;
    virtual bool evaluate(const solver & solver) const
    { return !m_operand->evaluate(solver); }
    virtual formula_impl* make_not()
    {
        return m_operand;
    }
    virtual formula_impl* and_with(formula_impl* , bool initial =true);
    virtual formula_impl* or_with(formula_impl* other, bool initial=true);

    const formula_impl * operand() const { return  m_operand; }
    virtual void print(std::ostream & out,
                       const variable_print_strategy & strategy,
                       op_priority priority) const;
  protected:
    virtual ~not_formula_impl();
    
  private:
    formula_impl *m_operand;
    static pool_alloc<not_formula_impl> m_pool_not;
}; // class not_formula_impl

/**
 * @brief An abstract binary formula
 *
 * This class servers as a base class for binary operator formulae such as OR
 * and AND.
 */
class binary_operator_formula_impl: public formula_impl
{
  public:
    binary_operator_formula_impl(formula_impl *first, formula_impl * second);
    //Make sure we are using pool_alloc's free list and not malloc/free.
    void* operator new(std::size_t size, const std::nothrow_t&) throw()
    {
        if (size == sizeof(binary_operator_formula_impl))
            return m_pool_binary_op.allocate(1);
        else
            return NULL;
    }
    
    void* operator new(size_t size) throw (std::bad_alloc)
    {
        void * ret= operator new(size, std::nothrow);
        if (ret)
            return ret;
        else
            throw std::bad_alloc();
    }
    void operator delete(void* ptr)
    {
        m_pool_binary_op.deallocate(
            reinterpret_cast<binary_operator_formula_impl*>(ptr), 1);
    }
    
    virtual formula_impl* make_not()
    { return new not_formula_impl(this); }
    virtual formula_impl* and_with(formula_impl * other, bool initial=true);
    virtual formula_impl* or_with(formula_impl * other, bool initial= true);
  protected:
    virtual ~binary_operator_formula_impl();
    //! Get the first operand of the operator
    const formula_impl *first() const { return m_first; }
    //! Get the second operand of the operator 
    const formula_impl *second() const { return m_second; }
  private:
    //! The two operands of the operator
    formula_impl *m_first, *m_second;
    static pool_alloc<binary_operator_formula_impl> m_pool_binary_op;
}; // class binary_operator_formula_impl

/**
 * @brief A simple binary OR formula.
 */
class binary_or_formula_impl: public binary_operator_formula_impl
{
  public:
    binary_or_formula_impl(formula_impl *first_formula,
                           formula_impl *second_formula)
        : binary_operator_formula_impl(first_formula, second_formula) {}
    virtual void constrain_solver(solver & solver) const;
    virtual unsigned reify(solver & solver) const;
    virtual bool evaluate(const solver & solver) const
    { return first()->evaluate(solver) || second()->evaluate(solver); }
    virtual void print(std::ostream & out,
                       const variable_print_strategy & strategy,
                       op_priority priority) const;
}; // class binary_or_formula_impl

/**
 * @brief A simple binary AND formula.
 */
class binary_and_formula_impl: public binary_operator_formula_impl
{
  public:
    binary_and_formula_impl(formula_impl *first_formula,
                            formula_impl *second_formula)
        : binary_operator_formula_impl(first_formula, second_formula) {}
    virtual void constrain_solver(solver & solver) const;
    virtual unsigned reify(solver & solver) const;
    virtual bool evaluate(const solver & solver) const
    { return first()->evaluate(solver) && second()->evaluate(solver); }
    virtual void print(std::ostream & out,
                       const variable_print_strategy & strategy,
                       op_priority priority) const;
};

/**
 * @brief A signed-clause formula
 */
class clause_formula_impl: public formula_impl
{
  public:
    clause_formula_impl(const clause_data & clause) : m_clause(clause) {}
    clause_formula_impl(unsigned variable, const iset & set);
    //Make sure we are using pool_alloc's free list and not malloc/free.
    void* operator new(std::size_t size, const std::nothrow_t&) throw()
    {
        if (size == sizeof(clause_formula_impl))
            return m_pool_clause.allocate(1);
        else
            return NULL;
    }
    
    void* operator new(size_t size) throw (std::bad_alloc)
    {
        void * ret= operator new(size, std::nothrow);
        if (ret)
            return ret;
        else
            throw std::bad_alloc();
    }
    void operator delete(void* ptr)
    { m_pool_clause.deallocate(reinterpret_cast<clause_formula_impl*>(ptr),1); }

    virtual void constrain_solver(solver & solver) const;
    virtual unsigned reify(solver & solver) const;
    virtual bool evaluate(const solver & solver) const;
    virtual formula_impl* make_not();
    virtual formula_impl* and_with(formula_impl * other, bool initial=true);
    virtual formula_impl* or_with(formula_impl * other, bool initial= true);
    virtual formula_impl *and_with_clause(clause_formula_impl *other);
    virtual formula_impl *or_with_clause(clause_formula_impl *other);
    virtual void print(std::ostream & out,
                       const variable_print_strategy & strategy,
                       op_priority priority) const;
  private:
    /**
     * @brief Constrain the implication variable -> m_clause
     * @param solver The CSP to constrain
     * @param variable If this Boolean variable is true then m_clause must be
     *                 satisfied.
     */
    void variable_implies_clause(solver & solver, unsigned variable) const;
    /**
     * @brief Constrain the implication m_clause -> variable
     * @param solver The CSP to constrain
     * @param variable If m_clause is satisfied then this Boolean variable
     *                 should be true.
     */
    void clause_implies_variable(solver & solver, unsigned variable) const;

    clause_data m_clause;
    static pool_alloc<clause_formula_impl> m_pool_clause;
}; // class clause_formula_impl

// virtual
inline formula_impl* true_formula_impl::make_not()
{ return false_formula_impl::get(); }

} // namespace hcsp
#endif // HCSP_FORMULA_IMPL_HPP_
