// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2014, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2014, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <unistd.h>
#include <typeinfo>
#include <fstream>
#ifdef HCSP_VALGRIND
#include <valgrind/valgrind.h>
#endif
#include <fzn_solver.hpp>
#include <flags.hpp>
using namespace std;
using namespace hcsp;

class fzn_flags : public hcsp_flags
{
  public:
    fzn_flags() : all_results(false)
    { seed = 1; }
    bool parse_option(char option) override;
  protected:
    void usage_short(std::ostream & out) override
    {
        flags::usage_short(out);
        out << " [-a] [-f] [-p N]";
    }
    void usage_long(std::ostream & out) override
    {
        flags::usage_long(out);
        out << "  -a    Print all solutions for SAT problems or intermediate\n"
            << "        solutions for optimization problems.\n"
            << "  -f    (unsupported) Should ignore search options.\n"
            << "  -p N  (unsupported) Set number of processes/threads\n";
    }
    
    virtual std::string legal_args() const override
    { return flags::legal_args() + "afp:"; }
  public:
    bool all_results;
}; // class fzn_flags

bool fzn_flags::parse_option(char option)
{
    if (flags::parse_option(option))
        return true;
    switch (option) {
      case 'a': all_results= true; break;
      case 'f': /* ignored */; break;
      case 'p': /* ignored */; break;
      default:
        return false;
    }
    return true;
}

void solve_file(istream & in, const char *file_name,
                const fzn_flags & flags,
                table_propagator_base::memory_manager & manager,
                bool need_to_free)
{
    unique_ptr<fzn_solver> p_problem;
    try {
        cout << "% Constructing " << file_name << "\n";
        p_problem.reset(new fzn_solver(in, string(file_name), flags, manager));
        cout << "% All results = " << flags.all_results << endl;
        p_problem->set_only_optimal(!flags.all_results);
        p_problem->set_all_results(flags.all_results);
        cout << "% Has " << p_problem->count_constraints() << " constraints\n";
    } catch (std::bad_alloc) {
        cout << "% Out of memory while constructing the problem\n"
             << "=====UNKNOWN=====\n";
    } catch (AST_builder::unsupported e) {
        cout << "\n" "% UNSUPPORTED\n"
             << "% " << e.what() << "\n"
             << "=====UNKNOWN=====" << endl;
    }
    if (!p_problem.get())
        return;
    p_problem->solve_and_print(cout);
    if (!need_to_free)
        p_problem.release();
}

int main_impl(int argc, char *argv[])
{
    HCSP_PROFILE_CURRENT;
    bool free_all= false;
#   ifdef HCSP_VALGRIND
    free_all= RUNNING_ON_VALGRIND;
#   endif
    fzn_flags flags;
    flags.backtrack_limit=1000000000;
    if (!flags.parse(argc, argv))
        return 1;
    if (flags.all_done())
        return 0;
    table_propagator_base::memory_manager *memory_manager=
        new table_propagator_base::memory_manager;
    unique_ptr<table_propagator_base::memory_manager> auto_free;

    // memory manager never freed, to speed up exit, unless under valgrind
    if (free_all)
        auto_free.reset(memory_manager);
    
    if ( argv[optind] ) {
        ifstream in(argv[optind]);
        solve_file(in, argv[optind], flags, *memory_manager, free_all);
    } else {
        solve_file(cin, "<stdin>", flags, *memory_manager, free_all);
    }

    return 0;
}

int main(int argc, char *argv[])
{
    try {
        return main_impl(argc, argv);
    } catch (std::exception & e) {
        cout << "% Exception(" << typeid(e).name() << "):\n" << e.what()
             << endl;
        cout << "=====UNKNOWN=====\n";
    } catch (...) {
        cout << "% unknown exception" << std::endl;
        cout << "=====UNKNOWN=====\n";
    }
    return 1;                                               
}
