// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2012,2014, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2012,2014, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <fzn_solver.hpp>

#include <sstream>
#include <AST.hh>
#include <parser_context.hpp>
#include <table_propagators.hpp>
#include <comparison_propagators.hpp>
#include <interval_propagators.hpp>
#include <fzn_parser.tab.hpp>
#include <flags.hpp>

using namespace std;
using namespace hcsp;

namespace hcsp {
namespace fzn_constraint {
typedef parser_context::ast_argument ast_argument;
typedef parser_context::ast_array_argument ast_array_argument;
typedef parser_context::array_argument array_argument;
typedef parser_context::var_argument var_argument;
typedef CSPXMLParser::FunctionAbs FunctionAbs;
typedef CSPXMLParser::FunctionAdd FunctionAdd;
typedef CSPXMLParser::FunctionAnd FunctionAnd;
typedef CSPXMLParser::FunctionDiv FunctionDiv;
typedef CSPXMLParser::FunctionIf FunctionIf;
typedef CSPXMLParser::FunctionEQ FunctionEQ;
typedef CSPXMLParser::FunctionLT FunctionLT;
typedef CSPXMLParser::FunctionLE FunctionLE;
typedef CSPXMLParser::FunctionMax FunctionMax;
typedef CSPXMLParser::FunctionMin FunctionMin;
typedef CSPXMLParser::FunctionMod FunctionMod;
typedef CSPXMLParser::FunctionMul FunctionMul;
typedef CSPXMLParser::FunctionNE FunctionNE;
typedef CSPXMLParser::FunctionNot FunctionNot;
typedef CSPXMLParser::FunctionOr FunctionOr;
typedef CSPXMLParser::FunctionSub FunctionSub;
typedef CSPXMLParser::FunctionXor FunctionXor;


template <typename T>
inline unique_ptr<T> new_unique() { return unique_ptr<T>(new T); }

unique_ptr<var_argument> to_var_argument(const ast_argument& in)
{
    if (!in.m_ast->isVar())
        return nullptr;
    unique_ptr<var_argument> ret(new var_argument);
    ret->m_variable.m_var_id= in.m_ast->getVarId();
    return move(ret);
}

unique_ptr<array_argument> to_array_argument(const ast_array_argument& in)
{
    unique_ptr<array_argument> ret(new array_argument);
    for (const auto & ast_ptr: in.m_array) {
        if (!ast_ptr->isVar())
            return nullptr;
        parser_context::var_or_value var;
        var.m_var_id = ast_ptr->getVarId();
        ret->m_array.push_back(var);
    }
    return move(ret);
}
class bool_clause: public parser_context::constraint_type
{
  public:
    virtual bool connect_constraint(
        ostream & debug_stream, const position_point & point,
        parser_context & context, const argument_list & arguments,
        parser_context::symbol_info *defines_var,
        unique_ptr<AST> & ast) override;
    void connect_clause(parser_context & context,
                        const var_or_value_list & positive,
                        const var_or_value_list & negative);
}; // bool_clause

class table_constraint: public parser_context::constraint_type
{
  public:
    virtual bool connect_constraint(
        ostream & debug_stream, const position_point & point,
        parser_context & context, const argument_list & arguments,
        parser_context::symbol_info *defines_var,
        unique_ptr<AST> & ast) override;
}; // table_constraint


/**
 * @brief A base class for all binary constraints under the parser_context.
 *
 * This class overrides the pure virtual connect_constraint(). A concrete class
 * inheriting from binary_constraint_type must supply connect2() which defines
 * how to connect the constraint to the two input arguments.
 */
class binary_constraint_type : public parser_context::constraint_type
{
  public:
    virtual bool connect_constraint(
        ostream & debug_stream, const position_point & point,
        parser_context & context, const argument_list & arguments,
        parser_context::symbol_info *defines_var,
        unique_ptr<AST> & ast) override;
    virtual void connect2(parser_context & context,
                          unique_ptr<AST> left, unique_ptr<AST> right,
                          parser_context::symbol_info *defines_var,
                          unique_ptr<AST> & out) = 0;
}; // class binary_constraint_type

class bool_le: public binary_constraint_type
{
  public:
    virtual void connect2(parser_context & context,
                          unique_ptr<AST> left, unique_ptr<AST> right,
                          parser_context::symbol_info *,
                          unique_ptr<AST> &) override
    {
        context.connect_expr(context.singleton_var(true),
                             *make<FunctionOr>(
                                 make<FunctionNot>(move(left)),
                                 move(right)));
    }
}; // class bool_le

class bool_lt: public binary_constraint_type
{
  public:
    virtual void connect2(parser_context & context,
                          unique_ptr<AST> left, unique_ptr<AST> right,
                          parser_context::symbol_info *,
                          unique_ptr<AST> &) override
    {
        context.connect_expr(context.singleton_var(true),
                             *make<FunctionAnd>(
                                 make<FunctionNot>(move(left)),
                                 move(right)));
    }
}; // class bool_lt

template <typename OP>
//! Defines the FlatZinc int_OP (where OP is the template parameter) constraint.
class int_OP : public binary_constraint_type
{
  public:
    virtual void connect2(parser_context & context,
                          unique_ptr<AST> left, unique_ptr<AST> right,
                          parser_context::symbol_info *,
                          unique_ptr<AST> &) override
    {
        context.connect_expr(context.singleton_var(true),
                             *make<OP>(move(left), move(right)));
    }
}; // class int_OP

class no_overlap_constraint : public parser_context::constraint_type
{
  public:
    bool connect_constraint(
        ostream & debug_stream, const position_point & point,
        parser_context & context, const argument_list & arguments,
        parser_context::symbol_info *defines_var,
        unique_ptr<AST> & ast) override;
};  // class no_overlap_constraint

class sub_interval_constraint : public parser_context::constraint_type
{
  public:
    sub_interval_constraint(bool is_positive,
                            unsigned num_args=4)
        : m_is_positive(is_positive),
          m_num_args(num_args) {}
    bool connect_constraint(
        ostream & debug_stream, const position_point & point,
        parser_context & context, const argument_list & arguments,
        parser_context::symbol_info *defines_var,
        unique_ptr<AST> & ast) override;
  private:
    bool m_is_positive;
    unsigned m_num_args;
};  // class sub_interval_constraint


template <typename OP>
unique_ptr<parser_context::constraint_type>
inline make_OP() { return new_unique<int_OP<OP> >(); }

/**
 * @brief A base class for constraints acting as functions under the
 *        parser_context. 
 *
 * This class overrides the pure virtual connect_constraint(). A concrete class
 * inheriting from binary_function_constraint_type must supply connect3() which
 * defines how to connect the constraint to the arguments
 */
class binary_function_constraint_type : public parser_context::constraint_type
{
  public:
    typedef parser_context::symbol_info symbol_info;
    virtual bool connect_constraint(
        ostream & debug_stream, const position_point & point,
        parser_context & context, const argument_list & arguments,
        symbol_info *defines_var, unique_ptr<AST> & ast) override;
    virtual void connect3(parser_context & context, unique_ptr<AST> left,
                          unique_ptr<AST> right, unique_ptr<AST> out,
                          symbol_info * defines_var, unique_ptr<AST> & result)=0 ;
}; // class binary_function_constraint_type

class bool_le_reif: public binary_function_constraint_type
{
  public:
    virtual void connect3(parser_context & context,
                          unique_ptr<AST> left, unique_ptr<AST> right,
                          unique_ptr<AST> out, symbol_info * defines_var,
                          unique_ptr<AST> &result) override
    {
        unique_ptr<AST> expr_ast = make<FunctionOr>(
            make<FunctionNot>(move(left)),
            move(right));
        if (out->isVar() && defines_var && !defines_var->m_is_array &&
            0U+out->getVarId() == defines_var->m_var_or_set_list[0].m_var_id) {
            result = move(expr_ast);
        } else {
            context.connect_expr(context.singleton_var(true),
                                 *make<FunctionEQ>(move(expr_ast), move(out)));
        }
    }
}; // class bool_le_reif

class bool_lt_reif: public binary_function_constraint_type
{
  public:
    virtual void connect3(parser_context & context,
                          unique_ptr<AST> left, unique_ptr<AST> right,
                          unique_ptr<AST> out, symbol_info * defines_var,
                          unique_ptr<AST> &result) override
    {
        unique_ptr<AST> expr_ast = make<FunctionAnd>(
            make<FunctionNot>(move(left)),
            move(right));
        if (out->isVar() && defines_var && !defines_var->m_is_array &&
            0U+out->getVarId() == defines_var->m_var_or_set_list[0].m_var_id) {
            result = move(expr_ast);
        } else {
            context.connect_expr(context.singleton_var(true),
                                 *make<FunctionEQ>(move(expr_ast), move(out)));
        }
    }
}; // class bool_lt_reif

/**
 * @brief Defines the FlatZinc constraint that acts as a function.
 *
 * The constraint is defined as a function according to an AST node of type
 * FUNC.
 */
template <typename FUNC>
class int_FUNC : public binary_function_constraint_type
{
  public:
    virtual void connect3(parser_context & context, unique_ptr<AST> left,
                          unique_ptr<AST> right, unique_ptr<AST> out,
                          symbol_info * defines_var,
                          unique_ptr<AST> &result)  override
    {
        unique_ptr<AST> expr_ast = make<FUNC>(move(left), move(right));
        if (out->isVar() && defines_var && !defines_var->m_is_array &&
            0U+out->getVarId() == defines_var->m_var_or_set_list[0].m_var_id) {
            result = move(expr_ast);
        } else {
            context.connect_expr(context.singleton_var(true),
                                 *make<FunctionEQ>(move(expr_ast), move(out)));
        }
    }
}; // class int_OP

template <typename OP>
inline unique_ptr<parser_context::constraint_type>
make_FUNC() { return new_unique<int_FUNC<OP> >(); }


class plus_constraint_type : public binary_function_constraint_type
{
  public:
    virtual void connect3(parser_context & context, unique_ptr<AST> left,
                          unique_ptr<AST> right, unique_ptr<AST> out,
                          symbol_info * defines_var,
                          unique_ptr<AST> &result)  override
    {
        if (defines_var && !defines_var->m_is_array) {
            unsigned defined_id= defines_var->m_var_or_set_list[0].m_var_id;
            if (out->isVar() && 0U+out->getVarId() == defined_id)
                result = make<FunctionAdd>(move(left), move(right));
            else if (left->isVar() && 0U+left->getVarId() == defined_id)
                result = make<FunctionSub>(move(out), move(right));
            else if (right->isVar() && 0U+right->getVarId() == defined_id)
                result = make<FunctionSub>(move(out), move(left));
        }
        if (!result) {
            context.connect_expr(context.singleton_var(true),
                                 *make<FunctionEQ>(make<FunctionAdd>(
                                                       move(left), move(right)),
                                                   move(out)));
        }
    }
}; // class int_OP

class array_constraint_base : public parser_context::constraint_type
{
  public:
    virtual bool connect_constraint(
        ostream & debug_stream, const position_point & point,
        parser_context & context, const argument_list & arguments,
        parser_context::symbol_info *defines_var,
        unique_ptr<AST> & ast) override;
    virtual unique_ptr<AST> connect_FUNC(unique_ptr<AST> left,
                                         unique_ptr<AST> right) const = 0;
    virtual int initial() const = 0;
}; // class array_constraint_base

template <typename FUNC, int Initial>
class array_FUNC : public array_constraint_base
{
  public:
    virtual unique_ptr<AST> connect_FUNC(unique_ptr<AST> left,
                                         unique_ptr<AST> right) const
    {
        unique_ptr<FUNC> ret(new FUNC);
        ret->setArg(0, left.release());
        ret->setArg(1, right.release());
        return move(ret);
    }
    virtual int initial() const { return Initial; }
}; // class int_OP

class array_xor : public parser_context::constraint_type
{
  public:
    virtual bool connect_constraint(
        ostream & debug_stream, const position_point & point,
        parser_context & context, const argument_list & arguments,
        parser_context::symbol_info *defines_var,
        unique_ptr<AST> & ast) override;
}; // class array_constraint_base

template <typename FUNC, int Initial>
inline unique_ptr<parser_context::constraint_type>
make_array_FUNC(){ return make_unique<array_FUNC<FUNC, Initial> >(); }

class array_element : public parser_context::constraint_type
{
  public:
    virtual bool connect_constraint(
        ostream & debug_stream, const position_point & point,
        parser_context & context, const argument_list & arguments,
        parser_context::symbol_info *defines_var,
        unique_ptr<AST> & ast) override;

  private:
    bool connect_constraint(
        ostream & debug_stream, const position_point & point,
        parser_context & context, const var_or_value & index,
        const var_or_value_list & array, const var_or_value & result,
        unique_ptr<AST> & ast);
    void intersect_result_domain(parser_context & context, iset & index_domain,
                                 const var_or_value_list & array,
                                 iset & result_domain);
};

class all_different_type : public parser_context::constraint_type
{
  public:
    virtual bool connect_constraint(
        ostream & debug_stream, const position_point & point,
        parser_context & context, const argument_list & arguments,
        parser_context::symbol_info *defines_var,
        unique_ptr<AST> & ast) override;
}; // class all_different_type


class lin_constraint_type : public parser_context::constraint_type
{
  public:
    enum direction_type {
        EQ,
        LE,
        NE
    };
    
    lin_constraint_type(direction_type direction) : m_direction(direction) {}
    virtual bool connect_constraint(
        ostream & debug_stream, const position_point & point,
        parser_context & context, const argument_list & arguments,
        parser_context::symbol_info *defines_var,
        unique_ptr<AST> & ast) override;
  private:
    virtual bool is_reif() const { return false; }
    void make_unique_vars(
        parser_context & context, vector<int> & coefficient_values,
        vector<unsigned> & var_ids) const;
    
    bool connect_constraint(
        ostream & debug_stream, const position_point & point,
        parser_context & context,
        const var_or_value_list & coefficients, const var_or_value_list & vars,
        const var_or_value & result, const var_or_value * reif) const;

    argument_list from_ast(parser_context & context,
                           const argument_list & arguments) const;

    bool instantiate_weighted_sum(ostream & debug_stream,
                                  const position_point & point,
                                  parser_context & context,
                                  const var_or_value_list & coefficients,
                                  const var_or_value_list & vars,
                                  const var_or_value & result) const;
    
    unique_ptr<AST> eval_lin(const var_or_value_list &a,
                             const var_or_value_list &b) const;
    direction_type m_direction;
}; // class lin_constraint_type

class lin_reif_constraint_type : public lin_constraint_type
{
  public:
    lin_reif_constraint_type(direction_type direction)
        : lin_constraint_type(direction)
    {}
    virtual bool is_reif() const { return true; }
};
    

class abs_constraint_type : public binary_constraint_type
{
  public:
    virtual void connect2(parser_context & context,
                          unique_ptr<AST> in, unique_ptr<AST> out,
                          parser_context::symbol_info *defines_var,
                          unique_ptr<AST> & result) override
    {
        unique_ptr<AST> expr_ast = make<FunctionAbs>(move(in));
        if (out->isVar() && defines_var && !defines_var->m_is_array &&
            0U+out->getVarId() == defines_var->m_var_or_set_list[0].m_var_id) {
            result = move(expr_ast);
        } else {
            context.connect_expr(context.singleton_var(true),
                                 *make<FunctionEQ>(move(expr_ast), move(out)));
        }
    }
}; // class abs_constraint_type

// virtual
bool binary_constraint_type::connect_constraint(
    ostream & debug_stream, const position_point & point,
    parser_context & context, const argument_list & arguments,
    parser_context::symbol_info* defines_var,
    unique_ptr<AST> & ast)
{
    if (arguments.size() != 2) {
        debug_stream << point << " got " << arguments.size()
                     << " arguments when expecting 2\n";
        return false;
    }
    const parser_context::var_argument *var_arguments[2]= {
        dynamic_cast<var_argument*>(arguments[0].get()),
        dynamic_cast<var_argument*>(arguments[1].get())
    };
    const parser_context::ast_argument *ast_arguments[2]= {
        dynamic_cast<ast_argument*>(arguments[0].get()),
        dynamic_cast<ast_argument*>(arguments[1].get())
    };
    if ((!var_arguments[0] && !ast_arguments[0]) ||
        (!var_arguments[1] && !ast_arguments[1])) {
        debug_stream << point << " expecting variable/constant argument\n";
        return false;
    }
    unique_ptr<AST> left_var;
    if (var_arguments[0])
        left_var = to_ast(var_arguments[0]->m_variable);
    else
        left_var = deep_copy(*ast_arguments[0]->m_ast);
    
    unique_ptr<AST> right_var;
    if (var_arguments[1])
        right_var= (to_ast(var_arguments[1]->m_variable));
    else
        right_var = deep_copy(*ast_arguments[1]->m_ast);
    connect2(context, move(left_var), move(right_var), defines_var, ast);
    return true;
}

// virtual
bool binary_function_constraint_type::connect_constraint(
    ostream & debug_stream, const position_point & point,
    parser_context & context, const argument_list & arguments,
    parser_context::symbol_info* defines_var, unique_ptr<AST> & out)
{
    if (arguments.size() != 3) {
        debug_stream << point << " got " << arguments.size()
                     << " arguments when expecting 3\n";
        return false;
    }
    bool all_ast = true;
    for (const auto & ptr_arg : arguments) {
        if (!dynamic_cast<const ast_argument*>(ptr_arg.get())) {
            all_ast = false;
            break;
        }
    }
    if (all_ast) {
        connect3(context,
                 deep_copy(*static_cast<ast_argument&>(*arguments[0]).m_ast),
                 deep_copy(*static_cast<ast_argument&>(*arguments[1]).m_ast),
                 deep_copy(*static_cast<ast_argument&>(*arguments[2]).m_ast),
                 defines_var, out);
        return true;
    }
    const parser_context::var_argument *var_arguments[3]= {
        dynamic_cast<var_argument*>(arguments[0].get()),
        dynamic_cast<var_argument*>(arguments[1].get()),
        dynamic_cast<var_argument*>(arguments[2].get())
    };
    if (!var_arguments[0] || !var_arguments[1] || !var_arguments[2]) {
        debug_stream << point << " expecting variable/constant argument\n";
        return false;
    }
    unique_ptr<AST> left_var(to_ast(var_arguments[0]->m_variable));
    unique_ptr<AST> right_var(to_ast(var_arguments[1]->m_variable));
    unique_ptr<AST> result_var(to_ast(var_arguments[2]->m_variable));
    connect3(context, move(left_var), move(right_var), move(result_var),
             defines_var, out);
    return true;
}

// virtual
bool no_overlap_constraint::connect_constraint(
    ostream & debug_stream, const position_point & point,
    parser_context & context, const argument_list & orig_arguments,
    parser_context::symbol_info*, unique_ptr<AST> &)
{
    argument_list arguments = context.from_ast(orig_arguments);
    if (arguments.size() != 4) {
        debug_stream << point << " got " << arguments.size()
                     << " arguments when expecting 4\n";
        return false;
    }
    vector<const parser_context::var_or_value*> var_arguments;
    for (const auto & entry : arguments) {
        const auto * as_var_argument =
            dynamic_cast<const var_argument* >(entry.get());
        if (!as_var_argument) {
            debug_stream << point << " expecting only variables/constants\n";
            return false;
        }
        var_arguments.push_back(& as_var_argument->m_variable);
    }
    if (var_arguments[1]->m_var_id != -1U || var_arguments[3]->m_var_id != -1U) {
        debug_stream << point << " expecting second and forth const arguments\n";
        return false;
    }
    context.connect_constraint(
        vector<unsigned>{var_arguments[0]->to_id(context),
                         var_arguments[2]->to_id(context)},
        make_unique<no_overlap_propagator>(var_arguments[1]->m_value,
                                           var_arguments[3]->m_value));
    return true;
}

// virtual
bool sub_interval_constraint::connect_constraint(
    ostream & debug_stream, const position_point & point,
    parser_context & context, const argument_list & orig_arguments,
    parser_context::symbol_info*, unique_ptr<AST> &)
{
    argument_list arguments = context.from_ast(orig_arguments);
    if (arguments.size() != m_num_args) {
        debug_stream << point << " got " << arguments.size()
                     << " arguments when expecting " << m_num_args << "\n";
        return false;
    }
    vector<const parser_context::var_or_value*> var_arguments;
    for (const auto & entry : arguments) {
        const auto * as_var_argument =
            dynamic_cast<const var_argument* >(entry.get());
        if (!as_var_argument) {
            debug_stream << point << " expecting only variables/constants\n";
            return false;
        }
        var_arguments.push_back(& as_var_argument->m_variable);
    }
    for (unsigned i= 2 ; i != var_arguments.size() ; ++i) {
        if (var_arguments[i]->m_var_id != -1U) {
            debug_stream << point << " expecting argument " << i
                         << " to be a constraint\n";
            return false;
        }
    }
    vector<unsigned> vars{var_arguments[0]->to_id(context),
                          var_arguments[1]->to_id(context)};
    if (m_is_positive) {
        context.connect_constraint(
            vars, make_unique<sub_interval_propagator>(
                var_arguments[2]->m_value, var_arguments[3]->m_value));
    } else {
        int k= m_num_args == 5 ? var_arguments[4]->m_value : 0;
        context.connect_constraint(
            vars, make_unique<no_sub_interval_propagator>(
                var_arguments[2]->m_value, var_arguments[3]->m_value, k));
    }
    return true;
}

// virtual
bool array_element::connect_constraint(
    ostream & debug_stream, const position_point & point,
    parser_context & context, const argument_list & orig_arguments,
    parser_context::symbol_info*, unique_ptr<AST> & ast_result)
{
    argument_list arguments = context.from_ast(orig_arguments);
    if (arguments.size() != 3) {
        debug_stream << point << " got " << arguments.size()
                     << " arguments when expecting 3\n";
        return false;
    }
    const parser_context::var_argument *index, *result;
    const parser_context::array_argument *array;
    index= dynamic_cast<var_argument*>(arguments[0].get());
    array= dynamic_cast<array_argument*>(arguments[1].get());
    result= dynamic_cast<var_argument*>(arguments[2].get());
    
    if (!index || !result) {
        debug_stream << point << " the first and last arguments should be "
            "variable/constant\n";
        return false;
    }
    if (!array) {
        debug_stream << point << " the second argument should be an array\n";
        return false;
    }
    return connect_constraint(debug_stream, point, context, index->m_variable,
                              array->m_array, result->m_variable, ast_result);
}

// virtual
bool all_different_type::connect_constraint(
    ostream & debug_stream, const position_point & point,
    parser_context & context, const argument_list & orig_arguments,
    parser_context::symbol_info*, unique_ptr<AST>&)
{
    argument_list arguments = context.from_ast(orig_arguments);
    if (arguments.size() != 1) {
        debug_stream << point << " got " << arguments.size()
                     << " arguments when expecting 1 array argument\n";
        return false;
    }
    array_argument * array= dynamic_cast<array_argument*>(arguments[0].get());
    if (!array) {
        debug_stream << point << " expecting array argument\n";
        return false;
    }
    vector<unsigned> variables;
    iset values;
    for (const var_or_value & entry: array->m_array) {
        if (entry.m_var_id != -1U)
            variables.push_back(entry.m_var_id);
        else if (!values.contains(entry.m_value))
            values.insert(entry.m_value);
        else {
            debug_stream << "% " << point
                         << ": all_different is trivially UNSAT since value "
                         << entry.m_value << " repeats\n";
            context.mark_unsat();
            return true;
        }
    }
    if (!values.empty()) {
        for (unsigned var: variables) {
            iset var_domain= context.domain(var);
            var_domain.erase_all_of(values);
            if (var_domain.empty()) {
                debug_stream << "% " << point
                             << ": all_different is trivially UNSAT since value "
                             << context.var_name(var)
                             << " conflicts values {" << values << "}\n";
                context.mark_unsat();
                return true;
            }
            context.set_domain(var, move(var_domain));
        }
    }
    context.instantiate_all_different(variables);
    return true;
}

// virtual
bool array_constraint_base::connect_constraint(
    ostream & debug_stream, const parser_context::position_point & point,
    parser_context & context, const argument_list & arguments,
    parser_context::symbol_info* result_var, unique_ptr<AST> & ast_out)
{
    if (arguments.size() != 2) {
        debug_stream << point << " got " << arguments.size()
                     << " arguments when expecting 2\n";
        return false;
    }
    array_argument * array= dynamic_cast<array_argument*>(arguments[0].get());
    auto * ast_array= dynamic_cast<ast_array_argument*>(arguments[0].get());
    if (!array && !ast_array) {
        debug_stream << point << " argument 1 should be an array\n";
        return false;
    }
    var_argument * var= dynamic_cast<var_argument*>(arguments[1].get());
    auto ast_arg= dynamic_cast<ast_argument*>(arguments[1].get());
    if (!var && !ast_arg) {
        debug_stream << point
                     << " argument 2 should be either a variable or a value\n";
        return false;
    }
    unique_ptr<AST> accumulated;
    if (array ? array->m_array.empty() : ast_array->m_array.empty()) {
        accumulated.reset(new ASTInteger(initial()));
    } else {
        if (array) {
            accumulated= context.to_ast(array->m_array[0]);
            for (unsigned i=1 ; i < array->m_array.size(); ++i)
                accumulated= connect_FUNC(move(accumulated),
                                          context.to_ast(array->m_array[i]));
        } else {
            accumulated= deep_copy(*ast_array->m_array[0]);
            for (unsigned i=1 ; i < ast_array->m_array.size(); ++i)
                accumulated= connect_FUNC(move(accumulated),
                                          deep_copy(*ast_array->m_array[i]));
        }
    }
    unique_ptr<AST> result;
    if (var)
        result= context.to_ast(var->m_variable);
    else
        result= deep_copy(*ast_arg->m_ast);

    if (result->isVar() && result_var && !result_var->m_is_array) {
        if (0U+result->getVarId() == result_var->m_var_or_set_list[0].m_var_id) {
            ast_out = move(accumulated);
            return true;
        }
    }
    context.connect_expr(context.singleton_var(true),
                         *make<FunctionEQ>(move(accumulated), move(result)));
    return true;
}

// virtual
bool array_xor::connect_constraint(
    ostream & debug_stream, const parser_context::position_point & point,
    parser_context & context, const argument_list & orig_arguments,
    parser_context::symbol_info*, unique_ptr<AST> &)
{
    argument_list arguments = context.from_ast(orig_arguments);
    if (arguments.size() != 1) {
        debug_stream << point << " got " << arguments.size()
                     << " arguments when expecting 1\n";
        return false;
    }
    array_argument * array= dynamic_cast<array_argument*>(arguments[0].get());
    if (!array) {
        debug_stream << point << " argument 1 should be an array\n";
        return false;
    }
    vector<unsigned> vars;
    bool result = true;
    for (const var_or_value & elem : array->m_array) {
        if (elem.m_var_id == -1U)
            result ^= (elem.m_value != 0);
        else
            vars.push_back(elem.m_var_id);
    }
    context.connect_parity(vars, result);
    return true;
}
bool array_element::connect_constraint(
    ostream & debug_stream, const position_point & point,
    parser_context & context, const var_or_value & index,
    const var_or_value_list & array, const var_or_value & result,
    unique_ptr<AST> &)
{
    unsigned result_var = result.m_var_id;
    if (result_var == -1U)
        result_var= context.singleton_var(result.m_value);
    unsigned index_var = index.m_var_id;
    if (index_var == -1U)
        index_var= context.singleton_var(index.m_value);
    iset result_domain= context.domain(result);
    iset index_domain= context.domain(index);
    intersect_result_domain(context, index_domain, array, result_domain);
    context.set_domain(result_var, move(result_domain));
    if (index_domain.empty()) {
        debug_stream << "% " << point
                     << ": the constraint is initially falsified\n"; 
        context.mark_unsat();
        return true;
    }
    context.set_domain(index_var, move(index_domain));
    context.instantiate_element_constraint(debug_stream, point,
                                           index_var, 1, array, result_var);
    return true;
}

void array_element::intersect_result_domain(parser_context & context,
                                            iset & index_domain,
                                            const var_or_value_list & array,
                                            iset & result_domain)
{
    if (array.size() == 0) {
        index_domain.clear();
        return;
    }
    index_domain.intersect(make_pair(1,array.size()));
    auto it= index_domain.begin();
    iset accumulated_result;
    while (it != index_domain.end()) {
        const iset & element_domain= context.domain(array[*it-1]);
        if (!element_domain.contains_any_of(result_domain)) {
            it= index_domain.erase(it);
            continue;
        }
        iset intersected= element_domain;
        intersected.intersect(result_domain);
        accumulated_result.insert_all_of(intersected);
        ++it;
    }
    result_domain= accumulated_result;
}

//!\todo split function
parser_context::argument_list lin_constraint_type::from_ast(
    parser_context & context, const argument_list & arguments) const
{
    auto a= dynamic_cast<array_argument*>(arguments[0].get());
    auto b= dynamic_cast<ast_array_argument*>(arguments[1].get());

    if (!b || is_reif() || m_direction != LE)
        return context.from_ast(arguments);
    parser_context::argument_list ret(3);
    auto new_a = new array_argument;
    auto c= dynamic_cast<var_argument*>(arguments[2].get());
    if (c)
        ret[2].reset(new var_argument(*c));
    else
        ret[2].reset(new var_argument{
                context.from_ast(
                    *dynamic_cast<ast_argument&>(*arguments[2]).m_ast)});
    ret[0].reset(new_a);
    if (a) {
        *new_a= *a;
    } else {
        auto ast_a = dynamic_cast<ast_array_argument*>(arguments[0].get());
        for (auto const & ast_ptr : ast_a->m_array)
            new_a->m_array.push_back(context.from_ast(*ast_ptr));
    }
    a = new_a;
    auto new_b = new array_argument;
    ret[1].reset(new_b);
    for (unsigned i=0; i != a->m_array.size(); ++i) {
        HCSP_ASSERT(a->m_array[i].m_var_id == -1U);
        if (!b->m_array[i]->isOperator()) {
            new_b->m_array.push_back(context.from_ast(*b->m_array[i]));
        } else if (a->m_array[i].m_value > 0) {
            var_or_value id;
            id.m_var_id= context.expression_to_variable(
                *b->m_array[i], -1U, AST_builder::BOUNDED_ABOVE);
            new_b->m_array.push_back(id);
        } else {
            var_or_value id;
            id.m_var_id= context.expression_to_variable(
                *b->m_array[i], -1U, AST_builder::BOUNDED_BELOW);
            new_b->m_array.push_back(id);
        }
    }
    return move(ret);
}

// virtual
bool lin_constraint_type::connect_constraint(
    ostream & debug_stream, const position_point & point,
    parser_context & context, const argument_list & orig_arguments,
    parser_context::symbol_info*, unique_ptr<AST> &)
{
    unsigned expected_args = 3 + is_reif();
    if (orig_arguments.size() != expected_args) {
        debug_stream << point << " got " << orig_arguments.size()
                     << " arguments when expecting " << expected_args << endl;
        return false;
    }
    
    argument_list arguments = from_ast(context, orig_arguments);
    
    const array_argument *a= dynamic_cast<array_argument*>(arguments[0].get());
    const array_argument *b= dynamic_cast<array_argument*>(arguments[1].get());
    const var_argument *result= dynamic_cast<var_argument*>(arguments[2].get());
    const var_argument *reif= nullptr;
    if (is_reif())
        reif= dynamic_cast<var_argument*>(arguments[3].get());
        
    if (a == nullptr || b == nullptr) {
        debug_stream << point << ": The first two arguments should be arrays\n";
        return false;
    }
    if (result == nullptr) {
        debug_stream << point << ": The third arguments should be a scalar\n";
        return false;
    }
    if (is_reif() && !reif) {
        debug_stream << point << ": The last arguments should be a scalar\n";
        return false;
    }
    const var_or_value * reif_var = nullptr;
    if (reif)
        reif_var= &reif->m_variable;
    return connect_constraint(debug_stream, point, context, a->m_array,
                              b->m_array, result->m_variable, reif_var);
}

void lin_constraint_type::make_unique_vars(
    parser_context & context, vector<int> & coefficient_values,
    vector<unsigned> & var_ids) const
{
    vector<pair<unsigned, int64_t> > zipped;
    zipped.resize(var_ids.size());
    for (unsigned i=0 ; i != var_ids.size() ; ++i) {
        zipped[i].first= var_ids[i];
        zipped[i].second= coefficient_values[i];
    }
    sort(zipped.begin(), zipped.end());
    for (unsigned i=1 ; i < zipped.size(); ++i) {
        if (zipped[i-1].first == zipped[i].first) {
            zipped[i].second += zipped[i-1].second;
            zipped[i-1].second= 0;
        }
    }
    zipped.erase(remove_if(zipped.begin(), zipped.end(),
                           [](const pair<unsigned, int64_t> & t)
                           { return t.second == 0; }),
                 zipped.end());
    for (unsigned i=0 ; i != zipped.size(); ++i) {
        if (zipped[i].second <= numeric_limits<int>::max())
            continue;
        zipped.emplace_back(context.make_aliased_var(zipped[i].first),
                            zipped[i].second - numeric_limits<int>::max());
        zipped[i].second = numeric_limits<int>::max();
    }
    if (zipped.size() == var_ids.size())
        return;
    var_ids.clear();
    coefficient_values.clear();
    for (const auto & elem: zipped) {
        var_ids.push_back(elem.first);
        coefficient_values.push_back(elem.second);
    }
}

bool lin_constraint_type::instantiate_weighted_sum(
    ostream & debug_stream, const position_point & point,
    parser_context & context, const var_or_value_list & coefficients,
    const var_or_value_list & vars, const var_or_value & result) const
{
    vector<int> coefficient_values;
    coefficient_values.reserve(coefficients.size());
    for (const var_or_value & coeff : coefficients) {
        if (coeff.m_var_id != -1U) {
            debug_stream << point
                         << ": first argument should be a list of int but got "
                         << context.var_name(coeff.m_var_id) << endl;
            return false;
        }
        coefficient_values.push_back(coeff.m_value);
    }
    vector<unsigned> var_ids;
    var_ids.reserve(coefficients.size());
    for (const var_or_value & var : vars) {
        if (var.m_var_id == -1U)
            var_ids.push_back(context.singleton_var(var.m_value));
        else
            var_ids.push_back(var.m_var_id);
    }
    if (result.m_var_id != -1U) {
        debug_stream << point
                     << ": last argument should be an int but got "
                     << context.var_name(result.m_var_id) << endl;
            return false;
    }
    make_unique_vars(context, coefficient_values, var_ids);
    context.instantiate_weighted_sum(coefficient_values, var_ids, false,
                                     result.m_value);
    if (m_direction == EQ)
        context.instantiate_weighted_sum(coefficient_values, var_ids, true,
                                         result.m_value);
    HCSP_ASSERT(m_direction != NE);
    return true;
}

bool lin_constraint_type::connect_constraint(
    ostream & debug_stream, const position_point & point,
    parser_context & context,
    const var_or_value_list & coefficients, const var_or_value_list & vars,
    const var_or_value & result,
    const var_or_value * reif) const
{
    if (coefficients.size() != vars.size()) {
        debug_stream << point << ": The first two arguments should be arrays "
            "of same size\n";
        return false;
    }
    
    if (!reif && m_direction != NE)
        return instantiate_weighted_sum(debug_stream, point, context, 
                                        coefficients, vars, result);
    unique_ptr<AST> lin = eval_lin(coefficients, vars);
    unique_ptr<AST> expr;
    switch (m_direction) {
      case EQ:
        expr= make<FunctionEQ>(move(lin), to_ast(result));
        break;
      case LE:
        expr= make<FunctionLE>(move(lin), to_ast(result));
        break;
      case NE:
        expr= make<FunctionLE>(move(lin), to_ast(result));
        break;
      default:
        debug_stream << point << ": unsupported direction\n";
        return false;
    }
    if (reif)
        expr= make<FunctionEQ>(move(expr), to_ast(*reif));
    context.connect_expr(context.singleton_var(true), *expr);
    return true;
}

unique_ptr<CSPXMLParser::AST> lin_constraint_type::eval_lin(
    const var_or_value_list &a, const var_or_value_list &b) const
{
    if (a.empty())
        return make_unique<ASTInteger>(0);
    unique_ptr<AST> accumulated;
    HCSP_ASSERT(a.size() == b.size());
    for (unsigned i=0 ; i!=a.size() ; ++i) {
        unique_ptr<AST> mul= make<FunctionMul>(to_ast(a[i]), to_ast(b[i]));
        if (accumulated.get())
            accumulated= make<FunctionAdd>(move(accumulated), move(mul));
        else
            accumulated= move(mul);
    }
    return move(accumulated);
}


// virtual
bool bool_clause::connect_constraint(
    ostream & debug_stream, const position_point & point,
    parser_context & context, const argument_list & orig_arguments,
    parser_context::symbol_info*, unique_ptr<AST> &)
{
    argument_list arguments = context.from_ast(orig_arguments);
    if (arguments.size() != 2) {
        debug_stream << point << " got " << arguments.size()
                     << " arguments when expecting 2\n";
        return false;
    }
    const array_argument *a= dynamic_cast<array_argument*>(arguments[0].get());
    const array_argument *b= dynamic_cast<array_argument*>(arguments[1].get());
    if (a == nullptr || b == nullptr) {
        debug_stream << point << ": The first two arguments should be arrays\n";
        return false;
    }
    connect_clause(context, a->m_array, b->m_array);
    return true;
}

void bool_clause::connect_clause(parser_context & context,
                                 const var_or_value_list & positive,
                                 const var_or_value_list & negative)
{
    unique_ptr<AST> accumulated;
    for (const var_or_value & entry : positive) {
        unique_ptr<AST> literal(to_ast(entry));
        if (accumulated.get() == nullptr)
            accumulated= move(literal);
        else
            accumulated= make<FunctionOr>(move(accumulated), move(literal));
    }
    for (const var_or_value & entry : negative) {
        unique_ptr<AST> literal(make<FunctionNot>(to_ast(entry)));
        if (accumulated.get() == nullptr)
            accumulated= move(literal);
        else
            accumulated= make<FunctionOr>(move(accumulated), move(literal));
    }
    if (!accumulated.get())
        accumulated.reset(new ASTInteger(0));
    context.connect_expr(context.singleton_var(true), *accumulated);
}

// virtual
bool table_constraint::connect_constraint(
    ostream & debug_stream, const position_point & point,
    parser_context & context, const argument_list & orig_arguments,
    parser_context::symbol_info*, unique_ptr<AST> &)
{
    argument_list arguments = context.from_ast(orig_arguments);
    if (arguments.size() != 2) {
        debug_stream << point << " got " << arguments.size()
                     << " arguments when expecting 2\n";
        return false;
    }
    
    const array_argument *var_array, *table_spec;
    var_array= dynamic_cast<array_argument*>(arguments[0].get());
    table_spec= dynamic_cast<array_argument*>(arguments[1].get());
    
    if (!var_array || !table_spec) {
        debug_stream << point << " expecting two array arguments\n";
        return false;
    }
    unsigned arity = var_array->m_array.size();
    if (arity == 0) {
        debug_stream << point << " expecting non-empty array of variables\n";
        return false;
    }
    if (table_spec->m_array.size() % arity != 0) {
        debug_stream << point << " table-spec array must be dividable by arity "
                     << "(" << arity << ")\n"; 
        return false;
    }
    vector<vector<int> > tuples;
    for (const var_or_value & value : table_spec->m_array) {
        if (tuples.empty() || tuples.back().size() == arity) {
            tuples.emplace_back();
        }
        if (value.m_var_id != -1U) {
            debug_stream << point << " second array must contain constants\n";
            return false;
        }
        tuples.back().push_back(value.m_value);
    }
    vector<unsigned> vars;
    for (const var_or_value & var : var_array->m_array) {
        if (var.m_var_id == -1U)
            vars.push_back(context.singleton_var(var.m_value));
        else
            vars.push_back(var.m_var_id);
    }
    context.connect_table_propagator(vars, tuples);
    return true;
}

} } // namespace hcsp::fzn_constraint

using namespace hcsp::fzn_constraint;

fzn_solver::fzn_solver(std::istream & in,
                       const std::string & file_name,
                       const hcsp_flags & flags,                
                       hcsp::table_propagator_base::memory_manager& manager)
    : m_context(*this, fzn_scanner_methods::get(), manager, &in, file_name)
{
    m_comment = "%";
    m_summary_prefix = "% s";
    m_objective_prefix = "% previous objective";
    m_context.detect_all_diff_enabled(flags.use_all_diff);
    register_constraint_types();

    yy::fzn_parser parser(m_context);
    flags.apply(*this);
    parser.set_debug_stream(cout);
    if (parser.parse() != 0)
        throw std::runtime_error(file_name+": Parsing failed");
    if (!m_context.create_constraints(cout))
        throw std::runtime_error(file_name+": Parsing failed");
}

// virtual
void fzn_solver::print_solution(ostream & out) const
{
    m_context.print_solution(out);
    out << "----------" << endl;
}

void fzn_solver::solve_and_print(std::ostream & out)
{
    out << "% Solving " << flush;
    result_type result= solve(&out);
    out << "% " << backtracks_done() << " backtracks";
    if (restarts_done())
        out << " " << restarts_done() << " restarts";
    out << "\n";
    out << "% " << non_fallback_combines() << " non-fallback combines "
        << "out of " << total_combines() << " total\n";
    out << "% learned " << backtracks_without_fallback_combines()
        << " non-fallback constraints\n";
    switch( result ) {
      case SAT:
        print_solution(out);
        break;
      case OPTIMAL:
        out << "==========\n";
        break;
      case SAT_DONE:
        break;
      case UNSAT:
        out << "=====UNSATISFIABLE=====\n";
        break;
      case BACKTRACK_LIMIT:
        out << "% Too many backtracks\n";
        out << "=====UNKNOWN=====\n";
        break;
      case TIMEOUT:
        out << "% Timeout\n";
        out << "=====UNKNOWN=====\n";
        break;
      case MEMOUT:
        out << "% Out of memory\n";
        out << "=====UNKNOWN=====\n";
        break;
    }
    out << "% ASSIGNMENTS " << decisions_done() << "\n";
}

void fzn_solver::register_constraint_types()
{
    constexpr auto lin_EQ = lin_constraint_type::EQ;
    constexpr auto lin_LE = lin_constraint_type::LE;
    constexpr auto lin_NE = lin_constraint_type::NE;
    pair<string, std::unique_ptr<parser_context::constraint_type> > entries[]= {
        {"all_different_int",      new_unique<all_different_type>()},
        {"array_bool_and",         make_array_FUNC<FunctionAnd, 1>()},
        {"array_bool_element",     new_unique<array_element>()},
        {"array_bool_or",          make_array_FUNC<FunctionOr, 0>()},
        {"array_bool_xor",         new_unique<array_xor>()},
        {"array_int_element",      new_unique<array_element>()},
        {"array_var_bool_element", new_unique<array_element>()},
        {"array_var_int_element",  new_unique<array_element>()},
        {"bool2int",               make_OP<FunctionEQ>()},
        {"bool_and",               make_FUNC<FunctionAnd>()},
        {"bool_clause",            new_unique<bool_clause>()},
        {"bool_eq",                make_OP<FunctionEQ>()},
        {"bool_eq_reif",           make_FUNC<FunctionEQ>()},
        {"bool_le",                new_unique<bool_le>()},
        {"bool_le_reif",           new_unique<bool_le_reif>()},
        {"bool_lin_eq",            make_unique<lin_constraint_type>(lin_EQ)},
        {"bool_lin_le",            make_unique<lin_constraint_type>(lin_LE)},
        {"bool_lt",                new_unique<bool_lt>()},
        {"bool_lt_reif",           new_unique<bool_lt_reif>()},
        {"bool_not",               make_OP<FunctionNE>()},
        {"bool_or",                make_FUNC<FunctionOr>()},
        {"bool_xor",               make_FUNC<FunctionXor>()},
        {"hcsp_no_overlap",        make_unique<no_overlap_constraint>()},
        {"hcsp_sub_interval",      make_unique<sub_interval_constraint>(true)},
        {"hcsp_no_sub_interval",   make_unique<sub_interval_constraint>(false)},
        {"hcsp_no_sub_interval_k",
                                make_unique<sub_interval_constraint>(false,5)},
        {"hcsp_table_bool",        make_unique<table_constraint>()},
        {"hcsp_table_int",         make_unique<table_constraint>()},
        {"int_abs",                new_unique<abs_constraint_type>()},
        {"int_div",                make_FUNC<FunctionDiv>()},
        {"int_eq",                 make_OP<FunctionEQ>()},
        {"int_eq_reif",            make_FUNC<FunctionEQ>()},
        {"int_le",                 make_OP<FunctionLE>()},
        {"int_le_reif",            make_FUNC<FunctionLE>()},
        {"int_lin_eq",          make_unique<lin_constraint_type>(lin_EQ)},
        {"int_lin_eq_reif",     make_unique<lin_reif_constraint_type>(lin_EQ)},
        {"int_lin_le",          make_unique<lin_constraint_type>(lin_LE)},
        {"int_lin_le_reif",     make_unique<lin_reif_constraint_type>(lin_LE)},
        {"int_lin_ne",          make_unique<lin_constraint_type>(lin_NE)},
        {"int_lin_ne_reif",     make_unique<lin_reif_constraint_type>(lin_NE)},
        {"int_lt",                 make_OP<FunctionLT>()},
        {"int_lt_reif",            make_FUNC<FunctionLT>()},
        {"int_max",                make_FUNC<FunctionMax>()},
        {"int_min",                make_FUNC<FunctionMin>()},
        {"int_mod",                make_FUNC<FunctionMod>()},
        {"int_ne",                 make_OP<FunctionNE>()},
        {"int_ne_reif",            make_FUNC<FunctionNE>()},
        {"int_plus",               new_unique<plus_constraint_type>()},
        {"int_times",              make_FUNC<FunctionMul>()},
    };
    for (auto & entry : entries)
        m_context.register_constraint(entry.first, move(entry.second));
    
}
