// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009-2012,2014, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009-2012,2014, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <unistd.h>
#include <typeinfo>
#include <fstream>
#ifdef HCSP_VALGRIND
#include <valgrind/valgrind.h>
#endif
#include <xcsp_solver.hpp>
#include <c_like_solver.hpp>
#include <fzn_solver.hpp>
#include <flags.hpp>
#include <AST_builder.hpp>
#include <opb_context.hpp>
#include <opb_parser.tab.hpp>
using namespace std;
using namespace hcsp;

class per_format_flags : public hcsp_flags
{
  public:
    enum format_type {
        XCSP_FORMAT, //!< The xml format used in the CSP competition
        OPB_FORMAT,  //!< The OPB format used by Pseudo Boolean solvers
        FZN_FORMAT,  //!< The FlatZinc format
        C_FORMAT     //!< A format with C expression syntax
    };
    per_format_flags() : input_format(XCSP_FORMAT) {}
    virtual bool parse_option(char option);
  protected:
    virtual void usage_short(std::ostream & out)
    {
        hcsp_flags::usage_short(out);
        out << " [-F formal]";
    }
    virtual void usage_long(std::ostream & out)
    {
        hcsp_flags::usage_long(out);
        out << "  -F T   "
            << "The input file format (xcsp, c, or opb) [default: xcsp]\n";
    }
    
    virtual std::string legal_args() const
    { return hcsp_flags::legal_args() + "F:"; }
  public:
    format_type input_format;
}; // class per_format_flags

class opb_solver : public solver
{
  public:
    opb_solver(istream&in, const string & file_name,
               const hcsp_flags & flags,
               hcsp::table_propagator_base::memory_manager&):
        m_context(*this, &in, file_name)
    {
        yy::opb_parser parser(m_context);
        flags.apply(*this);
        if (m_context.trace_parsing() )
            parser.set_debug_level(1);
        parser.set_debug_stream(cout);
        if (parser.parse() != 0)
            throw std::runtime_error(file_name+": Parsing failed");
        if (!m_context.build_solver())
            set_var_domain(0, iset());
    }
    virtual void print_solution(ostream & out) const
    {
        m_context.print_solution(out);
    }
    virtual iset::value_type evaluate_objective() const
    {
        return m_context.evaluate_objective();
    }
  private:
    
    opb::context m_context;
}; // class opb_solver

bool per_format_flags::parse_option(char option)
{
    if (hcsp_flags::parse_option(option))
        return true;
    
    switch (option) {
      case 'F':
        if (strcasecmp(optarg, "xcsp") == 0)
            input_format= XCSP_FORMAT;
        else if (strcasecmp(optarg, "opb") == 0)
            input_format= OPB_FORMAT;
        else if (strcasecmp(optarg, "c") == 0)
            input_format= C_FORMAT;
        else if (strcasecmp(optarg, "fzn") == 0) {
            input_format= FZN_FORMAT;
        } else {
            cerr << "Unknown input format (i.e. -F) '" << optarg << "'. "
                 << "Must be one of xcsp, c, fzn, or opb\n";
            return false;
        }
        return true;
      default:
        return false;
    }
    return true;
}

template <class Solver>
Solver *construct(istream & in, const char *file_name, const string & comment,
                  const per_format_flags & flags,
                  table_propagator_base::memory_manager & manager)
{
    cout << comment << " Constructing " << file_name << " " << flush;
    unique_ptr<Solver> p_problem;
    p_problem.reset(new Solver(in, string(file_name), flags, manager));
    Solver & problem= *p_problem;
    cout << comment << " with " << problem.count_constraints()
         << " constraints\n";
    return p_problem.release();
}

void solve(solver & problem)
{
    cout << "c Solving " << flush;
    solver::result_type result= problem.solve(&cout);
    cout << problem.backtracks_done() << " backtracks";
    if (problem.restarts_done())
        cout << " " << problem.restarts_done() << " restarts";
    cout << "\n";
    cout << "c " << problem.non_fallback_combines() << " non-fallback combines "
         << "out of " << problem.total_combines() << " total\n";
    cout << "c learned " << problem.backtracks_without_fallback_combines()
         << " non-fallback constraints\n";
    switch( result ) {
      case solver::SAT:
        cout << "s SATISFIABLE\n";
        problem.print_solution(cout);
        break;
      case solver::UNSAT:
        cout << "s UNSATISFIABLE\n";
        break;
      case solver::BACKTRACK_LIMIT:
        cout << "s UNKNOWN\n";
        cout << "c Too many backtracks\n";
        break;
      case solver::TIMEOUT:
        cout << "s UNKNOWN\n";
        cout << "c Timeout\n";
        break;
      case solver::MEMOUT:
        cout << "s UNKNOWN\n";
        cout << "c Out of memory\n";
        break;
      case solver::OPTIMAL:
      case solver::SAT_DONE: // fall through
        break;
    }
    cout << "d ASSIGNMENTS " << problem.decisions_done() << "\n";
}

template <class Solver>
void solve_file(istream & in, const char *file_name,
                const per_format_flags & flags,
                table_propagator_base::memory_manager & manager,
                bool need_to_free)
{
    unique_ptr<Solver> p_problem;
    try {
        p_problem.reset(construct<Solver>(in, file_name, "c", flags, manager));
    } catch (std::bad_alloc) {
        cout << "s UNKNOWN\n"
             << "c Out of memory while constructing the problem\n";
        
    } catch (AST_builder::unsupported e) {
        cout << "\n" "s UNSUPPORTED\n"
             << "c " << e.what() << endl;
    }
    if (!p_problem.get())
        return;
    solve(*p_problem);
    if (!need_to_free)
        p_problem.release();
}

void solve_fzn_file(istream & in, const char *file_name,
                    const per_format_flags & flags,
                    table_propagator_base::memory_manager & manager,
                    bool need_to_free)
{
    unique_ptr<fzn_solver> p_problem;
    try {
        p_problem.reset(construct<fzn_solver>(in, file_name, "%", flags,
                                              manager));
    } catch (std::bad_alloc) {
        cout << "% Out of memory while constructing the problem\n"
             << "=====UNKNOWN=====\n";
        
    } catch (AST_builder::unsupported e) {
        cout << "\n" "% UNSUPPORTED\n"
             << "% " << e.what() << "\n"
             << "=====UNKNOWN=====" << endl;
    }
    if (!p_problem.get())
        return;
    p_problem->solve_and_print(cout);
    if (!need_to_free)
        p_problem.release();
}

void solve_file(istream & in, const char *file_name,
                const per_format_flags & flags,
                table_propagator_base::memory_manager & manager,
                bool need_to_free)
{
    switch (flags.input_format) {
      case per_format_flags::XCSP_FORMAT:
        solve_file<xcsp_solver>(in, file_name, flags, manager, need_to_free);
        break;
      case per_format_flags::OPB_FORMAT:
        solve_file<opb_solver>(in, file_name, flags, manager, need_to_free);
        break;
      case per_format_flags::C_FORMAT:
        solve_file<c_like_solver>(in, file_name, flags, manager, need_to_free);
        break;
      case per_format_flags::FZN_FORMAT:
        solve_fzn_file(in, file_name, flags, manager, need_to_free);
        break;
      default:
        cerr << "Unknown input format " << flags.input_format << "\n";
    }
}

int main_impl(int argc, char *argv[])
{
    HCSP_PROFILE_CURRENT;
    bool free_all= false;
#   ifdef HCSP_VALGRIND
    free_all= RUNNING_ON_VALGRIND;
#   endif
    per_format_flags flags;
    flags.backtrack_limit=1000000000;
    if (!flags.parse(argc, argv))
        return 1;
    if (flags.all_done())
        return 0;
    table_propagator_base::memory_manager *memory_manager=
        new table_propagator_base::memory_manager;
    unique_ptr<table_propagator_base::memory_manager> auto_free;

    // memory manager never freed, to speed up exit, unless under valgrind
    if (free_all)
        auto_free.reset(memory_manager);
    
    if ( argv[optind] ) {
        for (unsigned i=optind ; argv[i] ; ++i) {
            ifstream file(argv[i]);
            bool has_next= argv[i+1] != NULL;
            solve_file(file, argv[i], flags, *memory_manager,
                       has_next||free_all);
        }
    } else {
        solve_file(cin, "<stdin>", flags, *memory_manager, free_all);
    }

    return 0;
}

int main(int argc, char *argv[])
{
    try {
        return main_impl(argc, argv);
    } catch (std::exception & e) {
        cout << "Exception(" << typeid(e).name() << "):\n" << e.what() << endl;
    } catch (...) {
        cout << "unknown exception" << std::endl;
    }
    return 1;                                               
}
