// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <interval_propagators.hpp>

#include <iostream>
#include <limits>
#include <disjunction_specification.hpp>
#include <disjoint_propagators.hpp>
#include <clause_propagator.hpp>

using namespace std;
using namespace hcsp;
 
// virtual
constraint_propagator::result_type no_overlap_propagator::propagate(
    constraint_runtime_info & info )
{
    const iset & in_v0= info.domain(0);
    const iset & in_v1= info.domain(1);
    
    iset case1_v0= in_v0;
    iset case1_v1= in_v1;
    propagate(case1_v0, m_length0, case1_v1);
    
    iset case2_v0= in_v0;
    iset case2_v1= in_v1;
    propagate(case2_v1, m_length1, case2_v0);

    iset out_v0= case1_v0;
    out_v0.insert_all_of(case2_v0);

    iset out_v1= case1_v1;
    out_v1.insert_all_of(case2_v1);

    if (out_v0.empty())
        return UNSAT;
    
    info.set_domain(0, move(out_v0));
    info.set_domain(1, move(out_v1));
    
    return CONSISTENT;
}

// virtual
bool
no_overlap_propagator::conflicts(const constraint_runtime_info & info) const
{
    const iset & in_v0= info.domain(0);
    const iset & in_v1= info.domain(1);

    iset case1_v0= in_v0;
    iset case1_v1= in_v1;
    propagate(case1_v0, m_length0, case1_v1);
    
    iset case2_v0= in_v0;
    iset case2_v1= in_v1;
    propagate(case2_v1, m_length1, case2_v0);

    iset out_v0= case1_v0;
    out_v0.insert_all_of(case2_v0);

    iset out_v1= case1_v1;
    out_v1.insert_all_of(case2_v1);

    return out_v0.empty();
}


void no_overlap_propagator::propagate(iset & left, unsigned length,
                                      iset & right) const
{
    long long min_right= (long long)left.min() + (long long)length;
    long long max_left=  (long long)right.max()- (long long)length;
    if ( min_right > (long long)iset::MAX_DOMAIN ||
         max_left  < (long long)iset::MIN_DOMAIN ) {
        left.clear();
        right.clear();
        return;
    }
    
    left.intersect(make_pair(iset::MIN_DOMAIN, max_left));
    right.intersect(make_pair(min_right, iset::MAX_DOMAIN));
}

void no_overlap_propagator::extend_unsat_towards_equality(iset & v0, iset &v1)
    const
{
    long long min_overlapping_v0= v1.max()-(long long)m_length0+1;
    min_overlapping_v0= max<long long>(min_overlapping_v0, v1.min());
    if (min_overlapping_v0 < v0.min())
        v0.insert(make_pair(min_overlapping_v0, v0.min()));

    long long min_overlapping_v1= v0.max()-(long long)m_length1+1;
    min_overlapping_v1= max<long long>(min_overlapping_v1, v0.min());
    if (min_overlapping_v1 < v1.min())
        v1.insert(make_pair(min_overlapping_v1, v1.min()));

    long long max_overlapping_v0= v1.min()+(long long)m_length1-1;
    max_overlapping_v0= min<long long>(max_overlapping_v0, v1.max());
    if (max_overlapping_v0 > v0.max())
        v0.insert(make_pair(v0.max(), max_overlapping_v0));
    
    long long max_overlapping_v1= v0.min()+(long long)m_length0-1;
    max_overlapping_v1= min<long long>(max_overlapping_v1, v0.max());
    if (max_overlapping_v1 > v1.max())
        v1.insert(make_pair(v1.max(), max_overlapping_v1));
}

void no_overlap_propagator::explain_empty(iset &v0, iset & v1,
                                          vector<iset> & explanation) const
{
    extend_first_min_for_reduction(v0, m_length0, v1);
    extend_first_min_for_reduction(v1, m_length1, v0);
    extend_first_max_for_reduction(v0, v1, m_length1);
    extend_first_max_for_reduction(v1, v0, m_length0);
    
    (explanation[0]= v0).complement();
    (explanation[1]= v1).complement();
}


void no_overlap_propagator::extend_first_min_for_reduction(
    iset & first, unsigned first_length, const iset & second) const
{
    long long min_overlapping= second.max()-(long long)first_length+1;
    min_overlapping= max<long long>(min_overlapping, iset::MIN_DOMAIN);
    first.insert(make_pair(min_overlapping, first.min()));
}
    
void no_overlap_propagator::extend_first_max_for_reduction(
    iset & first, const iset & second, unsigned second_length) const
{
    long long max_overlapping= second.min()+(long long)second_length-1;
    max_overlapping= min<long long>(max_overlapping, iset::MAX_DOMAIN);
    first.insert(make_pair(first.max(), max_overlapping));
}

//virtual
void no_overlap_propagator::explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain,
        const constraint_runtime_info & info,
        vector<iset> & explanation) const
{
    (void)values_to_explain;
    explanation.resize(2);
    iset vars[2]= {info.domain(0), info.domain(1) };
    
    if ( propagated_domain.empty() ) {
        vars[0].assign(make_pair(vars[0].min(), vars[0].max()));
        vars[1].assign(make_pair(vars[1].min(), vars[1].max()));
        extend_unsat_towards_equality(vars[0], vars[1]);
        explain_empty(vars[0], vars[1], explanation);
    }
    else {
        iset &propagated_var= vars[propagated_param];
        propagated_var.erase_all_of(propagated_domain);
        if (propagated_var.contains_any_of(values_to_explain))
            propagated_var.intersect(values_to_explain);
        else
            propagated_var.assign(propagated_var.min());

        vars[0].assign(make_pair(vars[0].min(), vars[0].max()));
        vars[1].assign(make_pair(vars[1].min(), vars[1].max()));
        if (propagated_param == 0)
            extend_first_for_reduction(vars[1], m_length1, vars[0], m_length0);
        else
            extend_first_for_reduction(vars[0], m_length0, vars[1], m_length1);
        explain_empty(vars[0], vars[1], explanation);
    }
}

// virtual
void no_overlap_propagator::print(
    std::ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        p_info->print_param(out, 0);
        out << "+" << m_length0 << "<=";
        p_info->print_param(out, 1);
        out << " || ";
        p_info->print_param(out, 1);
        out << "+" << m_length1 << "<=";
        p_info->print_param(out, 0);
        out << ")";
    } else {
        out << "no_overlap(L0=" << m_length0 << ", L1=" << m_length1 << ")";
    }
}

// virtual
bool sub_interval_propagator::init( constraint_runtime_info & info )
{
    HCSP_ASSERT((unsigned)count_params() == info.arity());
    if (!can_be_satisfied(info.domain(0), info.domain(1)))
        return false;
    if (can_be_falsified(info.domain(0), info.domain(1))) {
        info.allocate_watch(0, UPPER_BOUND_WATCH);
        info.allocate_watch(1, LOWER_BOUND_WATCH);
    }
    return true;
}

constraint_propagator::result_type 
sub_interval_propagator::update_domains( constraint_runtime_info & info,
                                         iset::value_type lowest_x,
                                         iset::value_type highest_y)
{
    const iset & y= info.domain(1);
    HCSP_ASSERT(y.min() <= highest_y && highest_y <= y.max());
    if (y.max() > highest_y) {
        iset out1= info.domain(1);
        out1.intersect(make_pair(y.MIN_DOMAIN, highest_y));
        HCSP_ASSERT(!out1.empty());
        info.set_domain(1, move(out1));
    }

    const iset & x= info.domain(0);
    HCSP_ASSERT(x.min() <= lowest_x && lowest_x <= x.max());
    if (x.min() < lowest_x) {
        iset out0 = info.domain(0);
        out0.intersect(make_pair(lowest_x, x.MAX_DOMAIN));
        HCSP_ASSERT(!out0.empty());
        info.set_domain(0, move(out0));
    }
    if (can_be_falsified(info.domain(0), info.domain(1)))
        return CONSISTENT;
    else
        return VALID;
}

// virtual 
constraint_propagator::result_type 
sub_interval_propagator::propagate( constraint_runtime_info & info )
{
    const iset & x= info.domain(0);
    const iset & y= info.domain(1);
    if (!can_be_satisfied(x,y))
        return UNSAT;

    iset::value_type lowest_x=  x.max(); // may be lowered further
    iset::value_type highest_y= y.min(); // may be risen further
    // reminder: constraint is x>y || (m_a <= x && y <= m_b)
    if (m_a <= x.max() && y.min() <= m_b) {
        // The case when m_a <= x && y <= m_b is possible, handle it.
        lowest_x=  m_a;
        highest_y= m_b;
    }

    // The constraint may be satisfied due to x>y.
    // Soften the restrictions of lowest_x accordingly.
    if (lowest_x > y.min()) {
        // lowest_x may be lowered, softening the restriction.
        // x>y (i.e., x>=y+1) gives more freedom in this case.
        HCSP_ASSERT(y.min() != numeric_limits<iset::value_type>::max());
        lowest_x= y.min()+1;
    }
    lowest_x = max(lowest_x, x.min()); // lowest_x can't be lower than x.min()
    
    // The constraint may be satisfied due to x>y.
    // Soften the restrictions of highest_y accordingly.
    if (x.max() > highest_y) {
        // highest_y may be risen, softening the restriction.
        // x>y (i.e., x-1>=y) gives more freedom in this case.
        HCSP_ASSERT(x.max() != numeric_limits<iset::value_type>::min());
        highest_y= x.max()-1;
    }
    highest_y= min(highest_y, y.max()); // highest_y never higher than y.max()

    return update_domains(info, lowest_x, highest_y);
}

// virtual 
bool sub_interval_propagator::conflicts(const constraint_runtime_info & info)
    const
{
    const iset & x= info.domain(0);
    const iset & y= info.domain(1);
    return (!can_be_satisfied(x,y));
}

//virtual 
void sub_interval_propagator::print(
    std::ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "([";
        p_info->print_param(out, 0);
        out << ",";
        p_info->print_param(out, 1);
        out << "] sub_interval_of [" << m_a << "," << m_b << "])";
    } else {
        out << "sub_interval";
    }
}

// virtual
bool no_sub_interval_propagator::init( constraint_runtime_info & info )
{
    HCSP_ASSERT((unsigned)count_params() == info.arity());
    if (!can_be_satisfied(info.domain(0), info.domain(1)))
        return false;
    if (can_be_falsified(info.domain(0), info.domain(1))) {
        info.allocate_watch(0, LOWER_BOUND_WATCH);
        info.allocate_watch(1, UPPER_BOUND_WATCH);
    }
    return true;
}

// virtual 
constraint_propagator::result_type 
no_sub_interval_propagator::propagate( constraint_runtime_info & info )
{
    const iset & x= info.domain(0);
    const iset & y= info.domain(1);
    if (!can_be_satisfied(x,y))
        return UNSAT;
    long long highest_x= min<long long>(x.max(), 0LL+y.max()-m_k);
    long long lowest_y= max<long long>(0LL+x.min()+m_k, y.min());
    if (x.min() >= m_a && lowest_y <= 0LL+m_b+m_k) { 
        // x<m_a is false, require y>m_b+m_k
        lowest_y= 0LL+m_b+m_k+1;
    }
    if (y.max() <= 0LL+m_b+m_k && highest_x >= m_a) {
        // y>m_b+m_k is false, require x<m_a
        highest_x= m_a-1LL;
    }
    if (y.min() != lowest_y) {
        HCSP_ASSERT(y.max() >= lowest_y && lowest_y > y.min());
        iset out1 = info.domain(1);
        out1.intersect(make_pair(lowest_y, y.MAX_DOMAIN));
        info.set_domain(1, move(out1));
    }
    if (x.max() != highest_x) {
        HCSP_ASSERT(x.min() <= highest_x && highest_x <= x.max());
        iset out0 = info.domain(0);
        out0.intersect(make_pair(x.MIN_DOMAIN, highest_x));
        info.set_domain(0, move(out0));
    }
    if (can_be_falsified(info.domain(0),info.domain(1)))
        return CONSISTENT;
    else
        return VALID;
}

// virtual
constraint_propagator * no_sub_interval_propagator::new_propagator(
    disjunction_specification & disjunction,
    vector<unsigned> &copy_from_argument) const
{
    HCSP_ASSERT(this == disjunction.constraint());
    copy_from_argument.clear();
    copy_from_argument.reserve(disjunction.arity());
    copy_from_argument.push_back(disjunction.left_var());
    copy_from_argument.push_back(disjunction.right_var());
    if (disjunction.num_literals() == 0)
        return disjunction.release();

    typedef disjoint_propagator<no_sub_interval_propagator,
                                clause_propagation_algo
                                > no_sub_interval_or_clause_propagator;
    unique_ptr<no_sub_interval_or_clause_propagator> propagator;
    propagator.reset(new no_sub_interval_or_clause_propagator(
                         m_a, m_b, m_k));
    
    vector<iset> & lits= propagator->algo().algo_b().update_literals();
    lits.resize(disjunction.num_literals());
    
    for (unsigned i=0 ; i != disjunction.num_literals() ; ++i) {
        copy_from_argument.push_back(disjunction.literal_var(i));
        lits[i]= disjunction.literal_set(i);
    }
    return propagator.release();
}

// virtual 
bool no_sub_interval_propagator::conflicts(const constraint_runtime_info & info)
    const
{
    const iset & x= info.domain(0);
    const iset & y= info.domain(1);
    return !can_be_satisfied(x,y);
}

//virtual 
void no_sub_interval_propagator::print(
    std::ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "([";
        p_info->print_param(out, 0);
        out << ",";
        p_info->print_param(out, 1);
        if (m_k > 0)
            out << "-" << m_k;
        else if (m_k < 0)
            out << "+" << -(iset::size_type)m_k;
            
        out << "] no_sub_interval_of [" << m_a << "," << m_b << "])";
    } else {
        out << "(";
        if (m_k > 0)
            out << "[x,y-" << m_k << "]";
        else if (m_k < 0)
            out << "[x,y+" << -(iset::size_type)m_k << "]";
        out << "no_sub_interval_of [" << m_a << "," << m_b << "])";
    }
}
