// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#ifndef HCSP_ISET_HELPERS_HPP_
#define HCSP_ISET_HELPERS_HPP_

#include <unistd.h>
#include <iostream>
#include <sstream>
#include <limits>
#include <iset.hpp>
#include <utilities.hpp>

namespace hcsp {
namespace iset_helpers {

// explicit
template<class Set>
void assign_set(Set & out, const std::string & text)
{
    using namespace std;
    istringstream s(text);
    s >> out;
    char ch= s.get();
    if (s) {
        ostringstream error;
        error << "unexpected '" << ch << "' while parsing an iset of '"
              << text << "\n";
        throw parse_error(error.str());
    }
}

template<class Set>
void print_set(std::ostream &out, const Set & data)
{
    bool is_first= true;
    typedef typename Set::value_type value_type;
    for (typename Set::iterator it=data.begin() ; it!=data.end() ; ){
        if (!is_first)
            out << ",";
        iset::size_type step_size;
        typename Set::value_range_type range;
        it= it.find_stride(data.end(), range, step_size);
        out << range.first;
        if (range.first == range.second)
            ; // do nothing, everything was printed
        else if (value_type(range.first+step_size) == range.second)
            out << "," << range.second;
        else if (step_size == 1)
            out << ":" << range.second;
        else if (value_type(range.first+2*step_size)==range.second)
            out << "," << value_type(range.first+step_size)
                << "," << range.second;
        else 
            out << ":" << step_size << ":" << range.second;
        is_first= false;
    }
}

template <class Value>
void validate_interval(const std::pair<Value,Value> & range, unsigned step)
{
   if (range.first > range.second) {
       std::ostringstream s;
       s << "Input range is unordered " << range.first << ":"
         << range.second;
       throw parse_error(s.str());
    }

   if ( (0U+range.second-range.first)%step != 0) {
       std::ostringstream s;
       s << "Input interval size does not match skip size of " << step
         << "bounds are " << range.first << ":" << range.second;
       throw parse_error(s.str());
   }
 }

template <class Value>
char get_interval_end(std::istream &in, std::pair<Value,Value> &range,
                      unsigned & step)
{
    char ch;
    in >> ch;
    bool is_nonpositive=(!!in && ch == '-');
    in.unget();
    if (is_nonpositive) {
        in >> range.second;
    } else {
        in >> step;
        range.second= step;
        if (step == 0)
            is_nonpositive= true;
    }
    if (!in) {
        std::ostringstream s;
        s << "2nd part of the range is missing " << range.first << ":";
        throw parse_error(s.str());
    }
    in >> ch;
    if (!!in && ch == ':') {
        if (is_nonpositive) {
            std::ostringstream s;
            s << "Not expecting ':' after the range "
              << range.first << ":" << range.second;
            throw parse_error(s.str());
        }
        in >> range.second;
        if (!in) {
            std::ostringstream s;
            s << "3nd part of the stride is missing "
              << range.first << ":" << step << ":";
            throw parse_error(s.str());
        }
        in >> ch;
    } else
        step= 1;
    return ch;
}

template <class Set>
void read_set(std::istream & in, Set & data)
{
    data.clear();
    ios_state_restore ios_state_guard(in);
    in.unsetf(std::ios_base::skipws);
    bool done= false;
    while(!done && in) {
        char ch;
        typename Set::value_range_type range;
        unsigned step=1;
        in >> range.first;
        range.second= range.first;
        if (!in) break;
        in >> ch;
        if (!in)
            done= true;
        else if (ch == ':') {
            ch= get_interval_end(in, range, step);
            validate_interval(range, step);
        }

        if (step == 1)
            data.insert(range);
        else {
            for ( typename Set::value_type i= range.first ;
                  i!=range.second ;
                  i+=step)
                data.insert(i);
            data.insert(range.second);
        }
        if (!done && ch != ',') {
            in.putback(ch);
            done= true;
        }
    }
}

template <class Set>
std::pair<bool, typename Set::value_type> find_contained_tail(
    const typename Set::value_range_type & small_range,
    const typename Set::size_type small_stride,
    const Set & big)
{
    typedef typename Set::size_type  size_type;
    typedef typename Set::value_type value_type;
    typedef typename Set::iterator iterator;
    typedef typename Set::value_range_type value_range_type;
    
    std::pair<bool, value_type> found(false,0);
    value_type next=small_range.second;

    iterator big_begin= big.lower_bound(small_range.first);
    while (true) {
        iterator big_it= big.lower_bound(next);
        if (big_it == big.end() || *big_it != next)
            return found;
        found.first= true;
        found.second= next;
        if (next == small_range.first)
            return found;
        ++big_it;
        value_range_type big_range;
        size_type big_stride;
        big_it= big_it.find_back_stride(big_begin, big_range, big_stride);
        next -= small_stride;
        if (big_range.first > next)
            continue;
        if (small_stride % big_stride != 0)
            return found;
        size_type diff= (size_type)big_range.first - small_range.first;
        if (diff == 0) {
            found.second= big_range.first;
            return found;
        }
        size_type skew= diff % small_stride;
        if (skew > 0)
            big_range.first += small_stride-skew;
        found.second= big_range.first;
        next= big_range.first - small_stride;
    }
}

template <class Set>
typename Set::iterator find_contained_tail(const Set & small,
                                           typename Set::iterator last_small_it,
                                           const Set & big)
{
    HCSP_PROFILE_CURRENT;
    typedef typename Set::iterator iterator;
    typedef typename Set::value_range_type value_range_type;
    typedef typename Set::value_type value_type;
    typedef typename Set::size_type size_type;
    if (big.empty() || small.begin() == last_small_it)
        return last_small_it;

    for ( iterator small_it= last_small_it, next_small_it;
          small_it != small.begin() ;
          small_it= next_small_it) {
        value_range_type small_range;
        size_type        small_stride;
        next_small_it=
            small_it.find_back_stride(small.begin(), small_range, small_stride);
        std::pair<bool, value_type> big_compared=
            find_contained_tail(small_range, small_stride, big);
        if (!big_compared.first)
            return small_it;
        if (big_compared.second != small_range.first)
            return small.lower_bound(big_compared.second);
    }
    return small.begin();
}

template <class Set>
void show_set(const Set & domain)
{
    // Don't use cout or cerr on purpose. To avoid the possibility that
    // the stream is in bad state.
    static volatile size_t written=0;
    std::ostringstream o;
    o << domain;
    written= write(1, o.str().c_str(), o.str().size());
    written += write(1, "\n", 1);
    
}

} // namespace iset_helpers
} // namespace hcsp

#endif // HCSP_ISET_HELPERS_HPP_
