// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include "learned_type.hpp"

#include <clause_data.hpp>
#include <constraint_connectivity_info.hpp>
#include <constraint_explicit_info.hpp>
#include <constraint_aged_info.hpp>
#include <clause_propagator.hpp>
#include <solver.hpp>

using namespace std;
using namespace hcsp;

const solver_types::change_index_t learned_type::NO_PREVIOUS_MODIFICATION;
const solver_types::change_index_t learned_type::POST_LAST_MODIFICATION;

/**
 * @brief The state of the learned variables either on the front nodes or
 *        right after a given modification.
 */
class learned_type::info_type final : public constraint_runtime_info,
                                      private var_watchers_base
{
  public:
    /**
     * @brief Construct an info object relative to the nodes of the given
     *        learned_type object.
     *
     * @param learned_constraint
     *        The learned constraint whose variables and nodes we want to access
     * @param latest_modification
     *        If given, i.e., not POST_LAST_MODIFICATION, and is earlier
     *        than learned_constraint.modification(i) then domain(i) will return
     *        the domain right after 'latest_modification'.
     *        Otherwise, either latest_modification==POST_LAST_MODIFICATION
     *        or latest_modification is later than modification(i) then
     *        domain(i) will return the domain right after modification(i)
     */
    info_type(const learned_type & learned_constraint,
              change_index_t latest_modification= POST_LAST_MODIFICATION)
        : m_learned(learned_constraint),
          m_latest_modification(latest_modification),
          m_effort(learned_constraint.effort()),
          m_was_propagated(false), m_allow_latest_modification_query(true)
    {}
    virtual bool reduced_between_activations(unsigned param) const
    { return !m_was_propagated || (
            param < m_is_modified.size() && m_is_modified[param]); }
    virtual queue_type originating_queue() const { return VARIABLE_QUEUE; }
    virtual void put_in_queue(queue_type) { }
    void undo_changes() { m_is_modified.clear(); }
    virtual const iset & domain(unsigned param_index) const
    {
        if (param_index < m_is_modified.size() && m_is_modified[param_index])
            return m_new_domains[param_index];
        change_index_t modification= m_learned.modification(param_index);
        if ( assignment_stack::
             is_earlier_modification(m_latest_modification, modification)
             &&
             m_latest_modification != POST_LAST_MODIFICATION) {
            modification= m_latest_modification;
        }
        return m_learned.m_stack.var_domain_after_modification(
            m_learned.variable(param_index), modification);
    }
    virtual void set_domain(unsigned param_index, iset new_domain) override
    {
        const iset & orig = domain(param_index);
        if (orig.size() == new_domain.size() &&
            orig.empty() == new_domain.empty())
            return;
        if (param_index >= m_is_modified.size()) {
            m_is_modified.resize(arity(), false);
            m_new_domains.reserve(arity());
        }
        HCSP_ASSERT(param_index < m_is_modified.size());

        if (m_is_modified[param_index]) {
            m_new_domains[param_index] = move(new_domain);
        } 
        else
        {
            if (m_new_domains.size() <= param_index)
                m_new_domains.resize(param_index+1);
            m_new_domains[param_index]= move(new_domain);
            m_is_modified[param_index]= true;
        }
    }
    // Number of variables connected to this constraint
    virtual unsigned     arity() const
    { return m_learned.m_implication_nodes.size(); }
    virtual void deallocate_watch(unsigned, watch_type, 
                                  watch_handle_t watch_handle) override
    { deallocate_value_watch(0,0, watch_handle); }
    virtual watch_handle_t allocate_watch(unsigned, watch_type) override
    { return allocate_value_watch(0,0);}
    void deallocate_value_watch(unsigned, iset::value_type, 
                                watch_handle_t watch_handle) override
    { m_allocated_watch_list.erase(watch_handle.get_it()); }
    watch_handle_t allocate_value_watch(unsigned, 
                                        iset::value_type) override
    { 
        m_allocated_watch_list.push_back(watcher_info_type{0});
        return --m_allocated_watch_list.end(); 
    }
    virtual void print_param(std::ostream &out, unsigned param) const throw()
    { m_learned.m_stack.print_var_name(out, m_learned.variable(param)); }
    virtual void print(std::ostream & out) const throw()
    {
        for (unsigned i=0 ; i!= arity() ; ++i) {
            if (i!=0)
                out << ", ";
            print_param(out, i);
            out << "={" << domain(i) << "}";
        }
    }
    virtual int latest_modified_param(unsigned except) const
    {
        if (m_allow_latest_modification_query)
            return m_learned.latest_node_index(except);
        else
            return -1;
    }
    unsigned to_variable(unsigned param) const
    { return m_learned.variable(param); }
    void set_effort(unsigned in_effort) { m_effort= in_effort; }
    virtual unsigned effort() const { return m_effort; }
    void mark_propagated() { m_was_propagated= true; }
     //! Enable latest_modified_param() or disable it, making it return -1
    void allow_latest_modification_query(bool flag= true)
    { m_allow_latest_modification_query= flag; }
    
  private:
    const learned_type & m_learned;
    /**
     * @brief What historical value to fetch for variables.
     *        - If not POST_LAST_MODIFICATION, then this->domain(i) will get
     *          the domains right after
     *          earliest(m_latest_modification, m_learned.modification(i));
     *        - If POST_LAST_MODIFICATION, then access domain(i) right after
     *          m_learned.modification(i)
     */
    change_index_t m_latest_modification;
    unsigned m_effort;
    bool m_was_propagated;
    bool m_allow_latest_modification_query;
    vector<bool> m_is_modified;
    vector<iset> m_new_domains;
    watch_list_t m_allocated_watch_list;
}; // class learned_type::info_type

vector<learned_type::implication_node> learned_type::m_implication_nodes_cache;
vector<iset> learned_type::m_literals_cache;
vector<unsigned> learned_type::m_variable_to_literal_index_cache;

learned_type::learned_type(const assignment_stack & stack,
                           const solver & solver)
    : m_stack(stack), m_solver(solver),
      m_explanation_for_these_mods(false), m_effort(0)
{
    // using static to save costly allocs. Speed up around 1%
    m_variable_to_literal_index_cache.swap(m_variable_to_literal_index);
    m_literals_cache.swap(m_literals);
    m_implication_nodes_cache.swap(m_implication_nodes);
    
    m_implication_nodes.clear();
    m_literals.clear();
    m_variable_to_literal_index.resize(m_stack.count_variables());
    m_latest_level= m_stack.level();
    m_nodes_at_latest_level= 0;
}

learned_type::~learned_type()
{
    swap_if_first_bigger(m_variable_to_literal_index,
                         m_variable_to_literal_index_cache);
    swap_if_first_bigger(m_implication_nodes, m_implication_nodes_cache);
    swap_if_first_bigger(m_literals, m_literals_cache);
}

void learned_type::intersect_with_ground()
{
    if (constraint()) {
        intersect_general_constraint_with_ground();
        return;
    }
    unsigned read=0, write=0;
    for (read=0 ; read != size() ; ++read) {
        literal(read).intersect(m_stack.var_domain_end_level(variable(read),0));
        if (m_literals[read].empty())
            continue;
        if (read != write) {
            m_literals[write]= m_literals[read];
            m_implication_nodes[write]= m_implication_nodes[read];
            unsigned node_variable= m_implication_nodes[write].m_variable;
            m_variable_to_literal_index[node_variable]= write;
        }
        ++write;
    }
    if (read != write) {
        m_literals.erase(m_literals.begin()+write, m_literals.end());
        m_implication_nodes.erase(m_implication_nodes.begin()+write,
                                  m_implication_nodes.end());
    }
}

void learned_type::intersect_general_constraint_with_ground()
{
    HCSP_ASSERT(constraint());
    
    info_type info(*this);
    constraint_aged_info ground_info(info, m_stack,
                                     m_stack.decision_modification(1));
    vector<unsigned> copy_from_argument;
    unique_ptr<constraint_propagator> new_constraint;
    new_constraint.reset(m_constraint->intersect_with(ground_info,
                                                      copy_from_argument));
    if (!new_constraint.get())
        return;
    update_constraint(new_constraint);
    if (copy_from_argument.empty())
        return;
    if (copy_from_argument.size() > m_implication_nodes.size())
        throw runtime_error("constraint::intersect_with returns too many "
                            "arguments");
    vector<implication_node> next_implications(copy_from_argument.size());
    for (unsigned i=0 ; i != copy_from_argument.size() ; ++i) {
        next_implications[i]= m_implication_nodes.at(copy_from_argument[i]);
        m_variable_to_literal_index[next_implications[i].m_variable]= i;
    }
    m_implication_nodes.swap(next_implications);
    for (unsigned i=0 ; i != copy_from_argument.size() ; ++i) {
        HCSP_ASSERT(variable_to_pos(variable(i))== i);
    }
}

int learned_type::latest_node_index(unsigned except) const
{
    int last_index= -1;
    change_index_t last_time= NO_PREVIOUS_MODIFICATION;
    for (unsigned i=0 ; i != size() ; ++i) {
        if (i == except)
            continue;
        change_index_t time = modification(i);
        if ( last_index < 0 ||
             assignment_stack::is_earlier_modification(last_time, time) ) {
            last_index= i;
            last_time=  time;
        }
    }
    return last_index;
}

void learned_type::construct_nodes_and_explanation(
    const constraint_connectivity_info & info,
    const vector<iset> & explanation,
    change_index_t end_modification)
{
    m_implication_nodes.resize(info.arity());
    for (unsigned i=0; i != m_implication_nodes.size() ; ++i)
        m_implication_nodes[i].m_variable= info.to_variable(i);
    construct_from_variables(explanation, end_modification);
    if (m_solver.cause_relaxation())
        relax_cause();
    relax_if_required();
}

void learned_type::construct_from_variables(const vector<iset> & explanation,
                                            change_index_t end_modification)
{
    m_literals= explanation;
    m_explanation_for_these_mods= !explanation.empty();
    for (unsigned i=0; i != m_implication_nodes.size() ; ++i) {
        unsigned variable_index= m_implication_nodes[i].m_variable;
        change_index_t mod= m_stack.var_modification_entry(variable_index);
        if (!m_stack.is_earlier_modification(mod, end_modification)) {
            mod= m_stack.var_modification_entry_before(
                variable_index, end_modification);
        }
        m_variable_to_literal_index[variable_index]= i;
        unconditional_set_modification(i, mod);
    }
    if (explanation.empty())
        update_explanation();
    for (unsigned i=0 ; i != m_literals.size(); ++i) {
        if ( m_literals[i].empty() )
            set_modification(i, NO_PREVIOUS_MODIFICATION);
    }
    compute_level();
}

void learned_type::construct_from(
    const constraint_connectivity_info & info,
    const constraint_propagator & conflicting_constraint,
    const vector<iset> & explanation,
    change_index_t end_modification)
{
    m_effort= m_solver.get_effort(info);
    m_literals.clear();
    set_original_constraint(conflicting_constraint);
    construct_nodes_and_explanation(info, explanation, end_modification);
}

void learned_type::print(ostream & out) const
{
    if (constraint()) {
        info_type info(*this);
        constraint()->print(out, &info);
    } else {
        out << clause_data::print_prefix();
        for (unsigned j=0 ; j != size() ;  ++j) {
            if (j != 0)
                out << clause_data::print_separator();;
            literal_data::print(
                out, m_stack.var_printer(variable(j)),
                m_literals[j]);
        }
        out << clause_data::print_suffix();
    }
}

void learned_type::show_self() const
{
    print(cout);
}

vector<iset> learned_type::possibly_crippling_values() const
{
    if (!constraint())
        return m_literals;
    
    info_type info(*this);
    vector<iset> values;
    constraint()->possibly_crippling_values(0, iset(), info, values);
    if (values.size() < info.arity()) {
        values.resize(info.arity(), iset());
    }
    return values;
}

int learned_type::asserting_level() const
{
    if (unsat())
        return -1;
    if (!constraint())
        return m_stack.asserting_level_of_nth(size()-1, m_implication_nodes,
                                              m_literals);
    int level= m_stack.asserting_level_of_nth(size()-1, m_implication_nodes,
                                              possibly_crippling_values());
    if (level <= 0)
        return level;
    unique_ptr<constraint_propagator> constraint_clone(constraint()->clone());
    for ( ; level > 0 ; --level) {
        change_index_t pre_level= m_stack.decision_modification(level)-1;
        info_type info(*this, pre_level);
        if (constraint_clone->init(info) &&
            constraint_clone->propagate(info) != constraint_propagator::UNSAT)
            break;
    }
    return level;
}

unsigned learned_type::opportunistic_level() const
{
    if (constraint())
        return m_stack.asserting_level_of_nth(size()-2, m_implication_nodes,
                                              possibly_crippling_values());
    else
        return m_stack.asserting_level_of_nth(size()-2, m_implication_nodes,
                                              m_literals);
}

bool learned_type::relax(unsigned parameter, const iset & in_literal)
{
    change_index_t orig_change;
    orig_change= modification(parameter);
    if ( is_level_modification(orig_change) )
        --m_nodes_at_latest_level;
    change_index_t change= m_stack.relax(orig_change, in_literal);
    set_modification(parameter, change);
    if ( is_level_modification(change) )
        ++m_nodes_at_latest_level;
    if (m_nodes_at_latest_level == 0)
        compute_level();
    if (change == orig_change) {
        return false;
    } else {
        m_explanation_for_these_mods= false;
        return true;
    }
}

bool learned_type::relax_level(unsigned parameter, const iset &in_literal)
{
    if ( m_latest_level == 0)
        return false;
    change_index_t orig_mod= modification(parameter);
    if ( !is_level_modification(orig_mod) )
        return false;
    change_index_t next_mod= m_stack.relax_before(orig_mod, in_literal,
                                                  level_modification());
    set_modification(parameter, next_mod);
    if ( !is_level_modification(next_mod) ) {
        --m_nodes_at_latest_level;
    
	if (m_nodes_at_latest_level == 0)
	    compute_level();
    }
    if (orig_mod == next_mod) {
        return false;
    } else {
        m_explanation_for_these_mods= false;
        return true;
    }
}

bool learned_type::relax_all()
{
    HCSP_ASSERT(conflicts());
    if (m_nodes_at_latest_level==0)
        return false;
    bool stable= false;
    update_explanation();
    bool relaxed= false;
    while (!stable) {
        stable= true;
        for (unsigned i=0 ; i!=size() && m_nodes_at_latest_level!=0 ; ++i) {
            if (relax(i, m_literals[i])) {
                relaxed= true;
                stable= !constraint();
            }
               
        }
        if (!stable) 
            update_explanation();
    }
    HCSP_ASSERT(conflicts());
    HCSP_ASSERT(m_literals.empty() || explanation_conflicts());
    return relaxed;
}

bool learned_type::relax_cause_for_clause(const vector<iset> & clause)
{
    bool relaxed= false;
    unsigned prev_param= -1U;
    unsigned param= latest_node_index();
    while (param != prev_param) {
        prev_param= param;
        if (relax(param, clause[param])) {
            relaxed= true;
            param = latest_node_index();
        }
    }
    HCSP_ASSERT(!(relaxed && m_explanation_for_these_mods));
    return relaxed;
}

bool learned_type::relax_cause_for_constraint(
    constraint_propagator &cloned_constraint, info_type & info,
    const vector<const iset*> & ground_domains)
{
    bool relaxed= false;
    unsigned prev_param= -1U;
    while (m_latest_level > 0) {
        unsigned param= latest_node_index();
        if (prev_param == param)
            return true;
        prev_param= param;
        info.undo_changes();
        HCSP_ASSERT(ground_domains[param] != NULL);
        info.set_domain(param, *ground_domains[param]);
        if (cloned_constraint.propagate(info) == cloned_constraint.UNSAT) {
            change_index_t first_decision_mod= m_stack.decision_modification(1);
            HCSP_ASSERT(!m_stack.is_earlier_modification(modification(param),
                                                        first_decision_mod));
            change_index_t ground_mod= m_stack.var_modification_entry_before(
                variable(param), first_decision_mod);
            set_modification(param, ground_mod);
            --m_nodes_at_latest_level;
            if (m_nodes_at_latest_level == 0)
                compute_level();
        } else {
            const iset & supported_values= info.domain(param);
            if (!relax(param, supported_values))
                return relaxed;
        }
        relaxed= true;
        HCSP_ASSERT(conflicts());
    }
    return relaxed;
}

bool learned_type::relax_cause()
{
    HCSP_ASSERT(conflicts());
    bool relaxed= false;
    if (m_nodes_at_latest_level==0)
        return false;

    if (!constraint())
        return relax_cause_for_clause(m_literals);
    if (!m_literals.empty() && explanation_conflicts())
        relaxed= relax_cause_for_clause(m_literals);
    if (m_latest_level == 0)
        return relaxed;
    unique_ptr<constraint_propagator> cloned(constraint()->clone());
    info_type info(*this);
    change_index_t last_ground_mod= m_stack.decision_modification(1)-1;
    info_type ground_info(*this, last_ground_mod);
    vector<const iset*> ground_domains(size(), NULL);
    for (unsigned i=0 ; i != size() ; ++i) {
        if (m_stack.is_earlier_modification(last_ground_mod, modification(i))) {
            ground_domains[i]= &ground_info.domain(i);
            info.set_domain(i, *ground_domains[i]);
        }
    }
    if (!cloned->init(info))
        return relaxed;
    
    if (relax_cause_for_constraint(*cloned, info, ground_domains)) {
        if (!m_literals.empty() && !explanation_conflicts())
            m_literals.clear();
        relaxed= true;
    }
    
    return relaxed;
}

void learned_type::update_explanation(const constraint_runtime_info &info)
{
    if (!m_literals.empty() && m_explanation_for_these_mods)
        return;
    if (!constraint())
        return;
    
    bool is_clause= dynamic_cast<const clause_propagator*>(constraint());
    if (!m_literals.empty() && is_clause)
        // The explanation of clauses never change, no need to update.
        return;
    bool need_report= (m_solver.report_level() >= m_solver.PROPAGATION_MODS &&
                       !is_clause);
        
    vector<iset> prev_explanation;
    if (need_report && !m_literals.empty() )
        prev_explanation.swap(m_literals);
    else
        m_literals.clear();
    constraint()->explain_propagation(0, iset().complement(), iset(), info,
                                      m_literals);
    m_explanation_for_these_mods= true;
    if ( need_report && prev_explanation != m_literals ) {
        m_solver.report_explanation(cout, info, m_literals);
    }
}

void learned_type::update_explanation()
{
    if (!m_literals.empty() && m_explanation_for_these_mods)
        return;
    if (!constraint())
        return;
    info_type info(*this);
    info.allow_latest_modification_query(m_solver.should_explainer_check_age());
    update_explanation(info);
}

bool learned_type::explanation_conflicts() const
{
    info_type info(*this);
    for (unsigned i=0 ; i != m_literals.size() ; ++i) {
        if (m_literals[i].contains_any_of(info.domain(i)))
            return false;
    }
    return true;
}

bool learned_type::conflicts() const
{
    info_type info(*this);
    if (constraint()) {
        return constraint()->conflicts(info);
    } else {
        HCSP_ASSERT(m_literals.size() == info.arity());
        for (unsigned i=0 ; i != m_literals.size() ; ++i) {
            if (m_literals[i].contains_any_of(info.domain(i)))
                return false;
        }
        return true;
    }
}

bool learned_type::relax_level_all(const vector<iset> & literals)
{
    bool stable= false;
    bool relaxed= false;
    while (!stable) {
        stable= true;
        unsigned orig_latest_level= m_latest_level;
        for (unsigned i=0 ; i!=size() && m_nodes_at_latest_level!=0 ; ++i) {
            relaxed= relax_level(i, literals[i]) || relaxed;
        }
        if (orig_latest_level != m_latest_level)
            stable= false;
    }
    HCSP_ASSERT(conflicts());
    return relaxed;
}

bool learned_type::relax_level_all()
{
    if (m_nodes_at_latest_level==0)
        return false;
    update_explanation();
    return relax_level_all(m_literals);
}

bool learned_type::relax_if_required()
{
    bool relaxed= false;
    // It is possible that if constraint() != nullptr that
    // conflict_analysis_complete() will erroneously return 'true', due to
    // the way relaxation relaxes some nodes more than others. Overcome this
    // by recalculating the nodes which otherwise would make
    // conflict_analysis_complete() return true.
    if (m_latest_level != 0 && m_nodes_at_latest_level == 1 && constraint())
        relaxed= relax_cause() || relaxed;
    while (m_latest_level != 0 && m_nodes_at_latest_level == 1 && constraint()){
        info_type connectivity(*this);
        change_index_t mod= m_stack.decision_modification(m_latest_level)-1;
        constraint_aged_info info(connectivity, m_stack, mod, 0);
        if (!constraint()->conflicts(info))
            break;
        construct_from_variables(vector<iset>(), mod);
        if (m_solver.cause_relaxation() || (
                m_latest_level != 0 && m_nodes_at_latest_level == 1))
            relaxed= relax_cause() || relaxed;
    }
    
    switch (m_solver.relaxation()) {
      case solver::NO_RELAXATION:
        break;
      case solver::LEVEL_RELAXATION:
        relaxed = relax_level_all() || relaxed;
        break;
      case solver::COMPLETE_RELAXATION:
        relaxed = relax_all() || relaxed;
        break;
    }
    return relaxed;
}

bool learned_type::conflict_analysis_complete()
{
    if (m_latest_level != 0)
        return m_nodes_at_latest_level == 1;
    else if (unsat())
        return true;

    HCSP_ASSERT(m_latest_level ==0 && !unsat());
    
    change_index_t last_modified= modification(latest_node_index());
    
    // Does the learned_clause conflicts only initial state?
    if (last_modified == NO_PREVIOUS_MODIFICATION) {
        // Considering initial state, this is really an empty clause.
        // Mark it empty for the callers.
        if (m_solver.should_intersect_with_ground())
            clear();
            
        return true;
    }
    else
        return false;
}

unsigned learned_type::get_this_on_right_common_param(
    const constraint_connectivity_info & other_connectivity,
    unsigned other_pivot_index,
    vector<unsigned> & to_common_param) const
{
    to_common_param.resize(size());
    unsigned other_arity= other_connectivity.arity();
    for (unsigned i=0 ; i != size() ; ++i)
        to_common_param[i]= other_arity+i;

    unsigned this_pivot_index= -1U;
    for (unsigned i=0 ; i != other_arity ; ++i) {
        unsigned var= other_connectivity.to_variable(i);
        if (var < m_variable_to_literal_index.size()) {
            unsigned this_pos= variable_to_pos(var);
            if (this_pos < size() && variable(this_pos) == var)
                to_common_param[this_pos]= i;
            if (i == other_pivot_index)
                this_pivot_index= this_pos;
        }
    }
    return this_pivot_index;
}

void learned_type::reconstruct_from_combined(
    const constraint_connectivity_info &other_connectivity,
    unique_ptr<constraint_propagator> & result,
    const vector<unsigned> & copy_from_argument,
    change_index_t end_modification)
{
    // Using a temporary 'variable' rather than using 'm_implication_nodes'
    // because we need to read the old m_implication_nodes in the loop, through
    // variable().
    vector<unsigned> variables(copy_from_argument.size());
    for (unsigned i=0 ; i != copy_from_argument.size() ; ++i) {
        if (copy_from_argument[i] < size())
            variables[i]= variable(copy_from_argument[i]);
        else {
            unsigned other_param= copy_from_argument[i]-size();
            variables[i]= other_connectivity.to_variable(other_param);
        }
    }

    m_implication_nodes.resize(variables.size());
    for (unsigned i=0 ; i != variables.size() ; ++i)
        m_implication_nodes[i].m_variable= variables[i];
    
    update_constraint(result.release());
    m_effort= max<unsigned>(m_effort, other_connectivity.effort());
    construct_from_variables(vector<iset>(), end_modification);
}

bool learned_type::combined_conflicts(
    const constraint_connectivity_info & other_connectivity,
    const constraint_propagator & combined_constraint,
    const vector<unsigned> & combined_arguments,
    change_index_t pivot_mod) const
{
    info_type prev_info(*this);
    constraint_aged_info other_info(other_connectivity, m_stack,
                                    pivot_mod,
                                    m_effort);
    constraint_explicit_info info(combined_arguments.size());
    for (unsigned i=0 ; i != combined_arguments.size(); ++i) {
        unsigned arg= combined_arguments[i];
        
        // prefer other_info when a variable is shared since the combined
        // constraint will get the earlier time of other_info.
        if (arg < size()) {
            unsigned var= variable(arg);
            for (unsigned j=0 ; j != other_connectivity.arity(); ++j) {
                if (other_connectivity.to_variable(j) == var) {
                    arg= j+size();
                    break;
                }
            }
        }
        if (arg < size())
            info.set_domain(i, prev_info.domain(arg));
        else {
            unsigned other_arg= arg-size();
            info.set_domain(i, other_info.domain(other_arg));
        }
    }
    return combined_constraint.conflicts(info);
}

bool learned_type::combine(
    const constraint_propagator &other_propagator,
    const constraint_connectivity_info & other_connectivity,
    unsigned other_pivot_index, change_index_t pivot_mod)
{
    unique_ptr<constraint_propagator> result;
    vector<unsigned> copy_from_argument;
    if (constraint())
        result.reset(combine_as_constraint(other_propagator, other_connectivity,
                                           other_pivot_index, pivot_mod,
                                           copy_from_argument));
    if (!result.get())
        result.reset(combine_as_clause(other_propagator, other_connectivity,
                                       other_pivot_index, pivot_mod,
                                       copy_from_argument));

    if (!result.get())
        return false;
    if (!combined_conflicts(other_connectivity, *result, copy_from_argument,
                            pivot_mod)) {
        if (m_solver.report_level() >= solver::PROPAGATION_MODS)
            cout << "combined doesn't meet the requirements, "
                 << " falling back to the explanations\n";
        return false;
    }
    reconstruct_from_combined(other_connectivity, result, copy_from_argument,
                              pivot_mod);
    if (m_solver.report_level() >= solver::PROPAGATION_MODS) {
        info_type info(*this);
        cout << "after combine: ";
        constraint()->print(cout, &info);
        cout << "\n\n";

    }
    if(!conflicts()) {
        info_type info(*this);
        cout << "Failing assumption for domains " << info << "\n";
        throw runtime_error("learned_type::combine(): Assumption violated");
    }
    if (m_solver.cause_relaxation() || (
            m_latest_level != 0 && m_nodes_at_latest_level > 1))
        relax_cause();
    relax_if_required();
    
    return true;
}

constraint_propagator *learned_type::combine_as_constraint(
    const constraint_propagator &other_propagator,
    const constraint_connectivity_info & other_connectivity,
    unsigned other_pivot_index, change_index_t pivot_mod,
    vector<unsigned> & copy_from_argument)
{
    if (!constraint() || !constraint()->may_combine(other_propagator))
        return NULL;
    info_type info(*this);
    constraint_aged_info other_info(other_connectivity, m_stack, pivot_mod,
                                    other_connectivity.effort());
    vector<unsigned> other_to_common_param(other_connectivity.arity());
    for (unsigned i=0 ; i != other_to_common_param.size() ; ++i) {
        unsigned var= other_connectivity.to_variable(i);
        unsigned this_param= variable_to_pos(var);
        bool is_shared= (this_param < size() && variable(this_param) == var);
        if (is_shared)
            other_to_common_param[i]= this_param;
        else
            other_to_common_param[i]= size()+i;
    }
    return constraint()->combine(
                     info, other_propagator, other_info, other_to_common_param,
                     other_pivot_index, copy_from_argument);
}

constraint_propagator *learned_type::combine_as_clause(
    const constraint_propagator &other_propagator,
    const constraint_connectivity_info & other_connectivity,
    unsigned other_pivot_index, change_index_t pivot_mod,
    vector<unsigned> &copy_from_argument)
{
    update_explanation();
    clause_propagator as_clause;

    as_clause.update_literals().swap(m_literals);
    swap_on_exit<vector<iset> >swapper(as_clause.update_literals(), m_literals);
    
    if (!other_propagator.may_combine(as_clause))
        return NULL;

    info_type connectivity(*this);
    constraint_aged_info info(connectivity, m_stack, pivot_mod, m_effort);
    constraint_aged_info other_info(other_connectivity, m_stack, pivot_mod,
                                    other_connectivity.effort());
    vector<unsigned> this_to_common_param;
    unsigned pivot;
    pivot= get_this_on_right_common_param(other_connectivity, other_pivot_index,
                                          this_to_common_param);

    unique_ptr<constraint_propagator> result(
        other_propagator.combine(
            other_info, as_clause, info, this_to_common_param,
            pivot, copy_from_argument));
    if (result.get()) {
        // copy_from_argument[i] assumes that 'other' is on the left and 'this'
        // is on the right. We need to fix it to be the other way around.
        unsigned other_arity= other_info.arity();
        for (unsigned i=0; i != copy_from_argument.size() ; ++i) {
            copy_from_argument[i]=
                (copy_from_argument[i]+size())%(other_arity+size());
        }
    }
    return result.release();
}

void learned_type::add_vars_potentially_falsifying(vector<unsigned> & vars)
    const
{
    if (constraint()) {
        info_type info(*this);
        constraint_aged_info ground_info(info, m_stack,
                                         m_stack.decision_modification(1));
        for (unsigned i=0 ; i != size() ; ++i) {
            if (constraint()->potentially_falsifying(ground_info,i))
                vars.push_back(variable(i));
        }
    } else {
        for (unsigned i=0 ; i != size() ; ++i) {
            const iset &
                ground_domain= m_stack.var_domain_end_level(variable(i),0);
            if ( ground_domain.contains_any_of(m_literals[i]) )
                vars.push_back(variable(i));
        }
    }
}

// Erase the literal pointed by literal_index from the clause. Update
// variable_to_literal_index, such that all variables used in the clause
// point back correctly into the clause.
void learned_type::erase_literal(unsigned literal_index)
{
    unsigned erased_variable= variable(literal_index);
    if ( is_level_modification(modification(literal_index)) )
        --m_nodes_at_latest_level;
    
    // Erase the literal_index element from clause as follows:
    // Put the last element of the clause instead of the erased element,
    // then erase the last element.
   
    // Make sure that the reverse mapping of the erased variable is correct.
    HCSP_ASSERT( variable_to_pos(erased_variable)==literal_index );

    if (literal_index+1 != m_implication_nodes.size()) {
        unsigned moved_variable= m_implication_nodes.back().m_variable;
        m_variable_to_literal_index[moved_variable]= literal_index;
        m_implication_nodes[literal_index]= m_implication_nodes.back();
        m_literals[literal_index].swap(m_literals.back());
    }

    // This assignment is not strictly required. This assignment will cause
    // a crash of any accidental use of this index.
    m_variable_to_literal_index[erased_variable]= -1U;
    
    m_implication_nodes.pop_back();
    m_literals.pop_back();
}


// Insert a new literal into the clause, effectively creating a new
// disjunction of the rest of the clause with this literal. Update
// variable_to_literal_index, such that all variables used in the clause
// point back correctly into the clause.
void learned_type::insert_literal(const iset & in_set,
                                  unsigned in_variable,
                                  change_index_t literal_modification)
{
    HCSP_ASSERT(m_literals.size() == m_implication_nodes.size());
    bool is_new_variable= true;
    unsigned variable_pos_in_clause = variable_to_pos(in_variable);
    if ( variable_pos_in_clause < m_implication_nodes.size() &&
         variable(variable_pos_in_clause)==in_variable) {
        is_new_variable= false;
    }
    if ( is_new_variable ) {
        variable_pos_in_clause= m_literals.size();
        m_variable_to_literal_index[in_variable]= variable_pos_in_clause;
        m_literals.push_back(in_set);
        m_implication_nodes.push_back( implication_node() );
        m_implication_nodes.back().m_variable= in_variable;
        unconditional_set_modification(variable_pos_in_clause,
                                       literal_modification);
    } else {
        change_index_t current_mod= modification(variable_pos_in_clause);
        m_literals[variable_pos_in_clause].insert_all_of(in_set);

        if ( is_level_modification(current_mod) )
            --m_nodes_at_latest_level;
            
        literal_modification= assignment_stack::latest_modification(
            literal_modification, current_mod);
        set_modification(variable_pos_in_clause, literal_modification);
    }
    if ( is_level_modification(literal_modification) )
        ++m_nodes_at_latest_level;
}

bool learned_type::intersect_literal(const iset &in_literal,
                                     unsigned in_variable,
                                     change_index_t in_modification)
{
    HCSP_ASSERT(m_literals.size() == m_implication_nodes.size());
    if ( in_variable >= m_variable_to_literal_index.size() ||
         variable(variable_to_pos(in_variable)) != in_variable )
        // variable_to_pos(in_variable) is inconsistent because in_variable is
        // not in this learned_type.
        return false;
    unsigned position= variable_to_pos(in_variable);
    change_index_t orig_shared_mod= modification(position);
    m_literals[position].intersect(in_literal);
    if (m_literals[position].empty()) {
        erase_literal(position);
        return false;
    }
    if ( is_level_modification(orig_shared_mod) )
        --m_nodes_at_latest_level;
    set_modification(position, in_modification);
    if ( is_level_modification(in_modification) )
        ++m_nodes_at_latest_level;
    return true;
}

void learned_type::validate() const
{
    if (constraint()) {
        int count_params= constraint()->count_params();
        if ( count_params >= 0 &&
             m_implication_nodes.size() != (unsigned)count_params)
            throw runtime_error("validate(): num of constraint params and "
                                "implication nodes mismatch");
    }
    if (!m_literals.empty() || !constraint())
        if (m_literals.size() != m_implication_nodes.size())
            throw runtime_error(
                "validate(): literals,implication_nodes diff size");
    for (unsigned i=0 ; i != m_implication_nodes.size() ; ++i) {
        if (variable_to_pos(variable(i)) != i) 
            throw runtime_error("validate(): inconsistent variable() and"
                                " variable_to_pos()");
    }
}

void learned_type::convert_to_constraint()
{
    if (constraint())
        throw runtime_error("convert_to_constraint(): already a constraint");
    HCSP_ASSERT((validate(),1));
    clause_propagator *new_clause= new clause_propagator;
    vector<iset> saved_literals;
    saved_literals.swap(m_literals);
    update_constraint(new_clause);
    m_explanation_for_these_mods= false;
    saved_literals.swap(m_literals);
    new_clause->update_literals()= m_literals;
    HCSP_ASSERT((validate(),1));
}

void learned_type::convert_to_clause()
{
    HCSP_ASSERT(conflicts());
    if (!constraint())
        return;
    if (m_literals.empty())
        update_explanation();
    HCSP_ASSERT(explanation_conflicts());
    m_constraint.reset();
    // remove empty literals
    unsigned write=0;
    for (unsigned i=0; i != size(); ++i) {
        if (m_literals[i].empty())
            continue;
        if (i != write) {
            m_literals[i].swap(m_literals[write]);
            swap(m_implication_nodes[i], m_implication_nodes[write]);
            m_variable_to_literal_index[variable(write)]= write;
        }
        ++write;
    }
    m_literals.erase(m_literals.begin()+write, m_literals.end());
    m_implication_nodes.erase(m_implication_nodes.begin()+write,
                              m_implication_nodes.end());
    compute_level();
}
