// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#ifndef LEARNED_TYPE_HPP_
#define LEARNED_TYPE_HPP_

#include <vector>
#include <iostream>
#include <assignment_stack.hpp>
#include <solver_types.hpp>
#include <controlled_ptr.hpp>

namespace hcsp {
class constraint_runtime_info;
class constraint_connectivity_info;
class constraint_propagator;
class clause_propagator;
class solver;
/**
 * @brief Information on the currently built conflict constraint.
 *
 * It efficiently maintains queries and operations on literals
 * such as: finding a literal by variable index, the variable of a given
 * literal and the set of values restricting it.
 */
class learned_type : private no_copy // to allow unique_ptr<T> for incomplete T
{
  private:
    static const solver_types::change_index_t NO_PREVIOUS_MODIFICATION=
        assignment_stack::NO_PREVIOUS_MODIFICATION;
    static const solver_types::change_index_t POST_LAST_MODIFICATION=
        assignment_stack::POST_LAST_MODIFICATION;
  public:
    typedef solver_types::change_index_t change_index_t;
    typedef assignment_stack::implication_node implication_node;
    class info_type;
    friend class info_type;
    /**
     * @brief constructor for the learned type
     *
     * @param stack   The current assignment stack with its history information
     *                on variable domains.
     * @param solver  The solver on which we learn this constraint.
     */
    learned_type(const assignment_stack & stack, const solver & solver);
    ~learned_type();

    //! This is used only for testing. Must be removed once we start learning
    //! constraints.
    void convert_to_constraint();
    //! Convert the constraint to a clause. This is to be used when the
    //! conflict-analysis algorithm can't figure out how to work with the
    //! constraint directly and it can proceed only with a clause.
    void convert_to_clause();

    //! Update m_literals for the current constraint() with info_type(*this)
    void update_explanation();
    //! Update m_literals for the current constraint() with the given info
    void update_explanation(const constraint_runtime_info & info);

    void clear()
    {
        m_constraint.reset();
        m_implication_nodes.clear();
        m_literals.clear();
        m_explanation_for_these_mods= false;
        m_nodes_at_latest_level=0;
    }

    /**
     * @brief Construct the learned_type for the given conflicting constraint.
     *
     * @param info        The connectivity of the conflicting constraint.
     * @param constraint  The propagator of the conflicting constraint.
     * @param explanation If given, i.e., not empty, then this is the
     *                    explanation clause of the constraint.
     * @param before_mod  Construct the learned_type with modification entries
     *                    before this one. The code assumes that the constraint
     *                    conflicts at that point.
     */
    void construct_from(const constraint_connectivity_info & info,
                        const constraint_propagator & constraint,
                        const std::vector<iset> & explanation,
                        change_index_t before_mod= POST_LAST_MODIFICATION);

    //! Insert a new literal, possibly overlapping an existing literal.
    void insert_literal(const iset & literal_set, unsigned literal_variable,
                        change_index_t literal_modification);

    //! Intersect literal domains with ground level domains for their variable,
    //! erasing empty literals. (Works both for clauses and constraints).
    void intersect_with_ground();
    
    /**
     * Intersect a literal of the clause with the given domain.
     * If the literal becomes empty then it is removed from the clause.
     * The modification entry of the literal is updated such that 
     * latest uninteresting reductions will be ignored.
     *
     * @param[in] literal_set           The new constraining set of the literal
     * @param[in] literal_variable      The variable of the literal to work on.
     * @param[in] literal_modification  The modification ID of the operation
     * @return true if the result contains a literal for this variable
     */
    bool intersect_literal(const iset & literal_set, unsigned literal_variable,
                           change_index_t literal_modification);
    bool       unsat() const      { return m_implication_nodes.empty(); }
    size_t     size() const       { return m_implication_nodes.size(); }
    iset      &literal(unsigned i)      { return m_literals.at(i); }
    const iset&literal(unsigned i) const{ return m_literals.at(i); }
    unsigned   variable(unsigned i) const
    { return m_implication_nodes[i].m_variable; }
    /**
     * @brief Find the latest parameter which is not equal to \p except.
     *
     * @param except Find the latest parameter which is not equal to this index.
     *               The default parameter makes the method return the latest
     *               parameter, without any exception.
     * @return The latest parameter if one can be found, otherwise return a
     *         negative value. A parameter can't be found if this is an empty
     *         clause or if this is a single literal and except==0.
     */
    int        latest_node_index(unsigned except= -1U) const;
    void       print(std::ostream & out) const;

    change_index_t modification(unsigned parameter) const
    { return m_implication_nodes[parameter].m_modification_entry; }
    
    size_t     variable_to_pos(unsigned variable_index) const
    { return m_variable_to_literal_index[variable_index]; }
    bool       has_variable(unsigned variable_index) const
    {
        return variable_index < m_variable_to_literal_index.size() &&
               variable(variable_to_pos(variable_index)) == variable_index;
    }

    //! A level at which the conflict constraint is not consistent, before
    //! which it is consistent. If unsat() then return -1.
    int        asserting_level() const;
    unsigned   opportunistic_level() const;

    /**
     * @brief Move the node back in time as much as possible without violating
     *        the invariant that learned_type must be conflicting.
     *
     * @param parameter  The parameter to be affected
     * @return True if the node was actually moved
     */
    bool       relax(unsigned parameter)
    { return relax(parameter, m_literals.at(parameter)); }
    /**
     * @brief Move the node back in time without violating the invariant that
     *        learned_type must be conflicting.
     *
     * This method moves the node back in time by the minimum amount required to
     * cross the decision level boundary, if that does not violate the
     * invariant.
     *
     * @param parameter  The parameter to be affected
     * @return True if the node was actually moved
     */
    bool       relax_level(unsigned parameter)
    { return relax_level(parameter, m_literals.at(parameter)); }
    //! Run relax() for all parameters.
    //! @return True if any node has been moved backwards.
    bool       relax_all();
    /**
     * @brief Run relax_level() for all parameters.
     *
     * If this operation changed m_latest_level then it is rerun until
     * m_latest_level is stable.
     *
     * @return True if any node has been moved backwards.
     */
    bool       relax_level_all();
    
    /**
     * @brief If the solver requires it then relax the constraint.
     *
     * According to the relaxation level of the constraint call one of the
     * relaxation methods, or none.
     */
    bool       relax_if_required();

    /**
     * @brief Minimal relaxation to ensure that the latest node does cause
     *        the propagation of the learned_type.
     *
     * This method makes sure that the latest node in learned_type is causing
     * the conflict of learned_type. If the latest node is a cause for the
     * conflict then we will move to an earlier node and repeat the process.
     *
     * @return True if any node has been moved backwards.
     */
    bool       relax_cause();
    /**
     * @brief Is this clause sufficient result of conflict analysis.
     *
     * There are two basic cases:
     *  - The latest literals are from nonzero level\n
     *    return true if and only if there is exactly one literal in the
     *    latest decision level.
     *  - All literals are from level zero\n
     *    If m_solver.should_intersect_with_ground() (we would intersect the
     *    literals with ground level domains) then return true, since the
     *    problem is unsatisfiable. If proof is required then return true only
     *    if this is the empty clause.
     *
     * @param[in] need_proof  Are we running in the proof mode?
     * @return True if there is no need to continue conflict analysis with
     *         this clause.
     */
    bool conflict_analysis_complete();

    /**
     * @brief Attempt to perform generalized resolution over two constraints.
     *
     * If available, activate a rule that combines this->constraint() with
     * other_constraint into a new constraint. If no specialized code is
     * available to perform a generalized resolution (a.k.a. combine) then
     * return false, which tells the caller to fall-back to explanation-based
     * resolution.
     *
     * @param other_constraint   The constraint to combine with this
     * @param other_connectivity The connectivity object of other_constraint
     * @param other_pivot_index  The parameter of other_constraint which is the
     *                           pivot of the combination rule.
     * @param pivot_mod          The modification entry of the pivot.
     *
     * @pre (1) The variable other_connectivity.to_variable(other_pivot_index)
     *      is present in this; (2) other_constraint and other_connectivity are
     *      referring to the same constraint and have the same arity; (3)
     *      Propagating this after pivot_mod reaches a conflict; (4) right
     *      before pivot_mod, if we propagate other_constraint then the domains
     *      will be strong enough to conflict this.
     * @post If returning true, then this should conflict the domains right
     *      before pivot_mod.
     */
    bool combine(const constraint_propagator &other_constraint,
                 const constraint_connectivity_info & other_connectivity,
                 unsigned other_pivot_index, change_index_t pivot_mod);
    /**
     * @brief Append a list of variables that may falsify the learned_type
     *        constraint in some cases.
     *
     * The satisfiability is checked with regards to the ground domain, i.e, the
     * domain which does not depend on decisions. If an assignment to a variable
     * can falsify the constraint for some partial assignment of the ground
     * domain, then this variable is appended to vars. The variable is appended
     * even if this value already existed in the vector.
     *
     * @param vars[in,out] Regardless to the initial contents of vars, append to
     *                     it the variables that control the satisfiability of
     *                     this learned_type.
     */
    void add_vars_potentially_falsifying(std::vector<unsigned> & vars) const;
    
    friend
    std::ostream & operator<<(std::ostream & out, const learned_type & clause)
    {
        clause.print(out);
        return out;
    }
    const constraint_propagator *constraint() const
    { return m_constraint.get(); }
    constraint_propagator *mutable_constraint()
    { return m_constraint.get_mutable(); }

    //! True if the learned_type conflicts the domain at the nodes.
    //! During conflict analysis a learned_type must always conflict
    bool conflicts() const;
    //! True if all literals in m_literals conflict the nodes, or if
    //! this is a clause and all literals in m_literals conflict the nodes.
    bool explanation_conflicts() const;
    unsigned effort() const { return m_effort; }

    //! Is constraint() populated already in the current CSP?
    bool is_existing_constraint() const
    { return constraint() && !m_constraint.is_mutable(); }

    bool has_explanation() const { return !m_literals.empty(); }
    const std::vector<iset> & get_explanation() const { return m_literals; }
    
    //! Validate the invariant of this class.
    void validate() const;
    void show_self() const;
  private:
    bool combined_conflicts(
        const constraint_connectivity_info & other_info,
        const constraint_propagator & combined_constraint,
        const std::vector<unsigned> & combined_arguments,
        change_index_t pivot_mod) const;

    /**
     * @brief Try to combine 'this' and 'other' treating 'this' as a constraint
     *
     * @param[in] other_propagator    The other propagator to combine
     * @param[in] other_connectivity  The connectivity of the other propagator
     * @param[in] other_pivot_index   Pivot's parameter position in the other
     *                                constraint
     * @param[in] pivot_mod   The modification time when the pivot was reduced
     * @param[out] copy_from_argument
     *             For the produced propagator, this is its list of arguments.
     *             The values are index in the concatenated parameters
     *             of 'this' and 'other' propagator, where 'this' parameters
     *             are placed on the left (index 0 and above) and other's are on
     *             the right (highest values).
     * @return If could combine the two constraints, return the combined
     *         propagator (the caller must free). Otherwise return NULL.
     */
    constraint_propagator *combine_as_constraint(
        const constraint_propagator &other_propagator,
        const constraint_connectivity_info & other_connectivity,
        unsigned other_pivot_index, change_index_t pivot_mod,
        std::vector<unsigned> &copy_from_argument);
    
    /**
     * @brief Try to combine 'this' and 'other' treating 'this' as a clause.
     *
     * If this is a regular constraint, and not a clause, then the method takes
     * constraint's explanation for a clause.
     *
     * @param[in] other_propagator    The other propagator to combine
     * @param[in] other_connectivity  The connectivity of the other propagator
     * @param[in] other_pivot_index   Pivot's parameter position in the other
     *                                constraint
     * @param[in] pivot_mod   The modification time when the pivot was reduced
     * @param[out] copy_from_argument
     *             For the produced propagator, this is its list of arguments.
     *             The values are index in the concatenated parameters
     *             of 'this' and 'other' propagator, where 'this' parameters
     *             are placed on the left (index 0 and above) and other's are on
     *             the right (highest values).
     * @return If could combine the two constraints, return the combined
     *         propagator (the caller must free). Otherwise return NULL.
     */
    constraint_propagator *combine_as_clause(
        const constraint_propagator &other_propagator,
        const constraint_connectivity_info & other_connectivity,
        unsigned other_pivot_index, change_index_t pivot_mod,
        std::vector<unsigned> &copy_from_argument);
    
    /**
     * @brief Construct the explanation and complete the implication nodes based
     *        on the variables' part in the implication nodes.
     *
     * This method initializes implication nodes according to the variables
     * stored in m_implication_nodes and according to the explanation.
     *
     * @param explanation
     *        If provided, is the initial explanation for the propagation that
     *        this learned_type describes.
     *
     * @param end_modification
     *        The upper limit for all modifications in the implication
     *        nodes. All nodes should be earlier than this value. If
     *        POST_LAST_MODIFICATION is passed then there is no limit and any
     *        modification is accepted.
     *
     * @pre set_original_constraint has been called, and m_implication_nodes
     *      contain  the variables
     * @post class invariants are maintained.
     */
    void construct_from_variables(const std::vector<iset> & explanation,
                                  change_index_t end_modification);
    
    /**
     * @brief Rebuild this from its previous state and products of constraint
     *        combination. 
     *        
     * This method re-initializes implication nodes according to the parameters
     * of the product of constraint combination, which are passed by method
     * arguments. Some of the variables, of the constructed nodes, come from the
     * previous state of 'this', and some come from 'other_connectivity'.
     * The variable ID is decided according to 'copy_from_argument'.
     *
     * @param other_connectivity The connectivity of the other combined
     *                           constraint .
     * @param result             The outcome of constraint combination
     * @param copy_from_argument Defines the variable of m_implication_nodes[i],
     *                           if it's < size() then take old variable(i),
     *                           otherwise take param (i-size()) from other.
     * @param end_modification
     *        The upper limit for all modifications in the implication
     *        nodes. All nodes should be earlier than this value.
     *
     * @post class invariants are maintained.
     */
    void reconstruct_from_combined(
        const constraint_connectivity_info &other_connectivity,
        std::unique_ptr<constraint_propagator> & result,
        const std::vector<unsigned> & copy_from_argument,
        change_index_t end_modification);
    /**
     * @brief Construct implication nodes and explanations for this learned_type
     *
     * This method initializes implication nodes according to the info and
     * according to the explanation.
     *
     * @param info         The run-time connectivity info used as the basis for
     *                     implication nodes construction.
     * @param explanation  If provided, is the initial explanation for the
     *                     propagation that this learned_type describes.
     * @param end_modification
     *                     The upper limit for all modifications in the
     *                     implication nodes. All nodes should be earlier than
     *                     this value. If POST_LAST_MODIFICATION is passed then
     *                     there is no limit and any modification is accepted.
     * @pre set_original_constraint has been called.
     * @post class invariants are maintained.
     */
    void construct_nodes_and_explanation(
        const constraint_connectivity_info & info,
        const std::vector<iset> & explanation,
        change_index_t latest_modification);

    /**
     * @brief Get for each parameter of 'this' calculate its index in the
     *        combined of 'this' and 'other' parameters.
     *
     * We think of parameters to the two constraints as if they were
     * concatenated. When both propagators share a variable, they both will
     * use the same index: they will share the index of 'other'. The indices of
     * 'other' start on the left, i.e., with index 0, and the indices of the
     * 'right' constraint are to the right, i.e., end with the highest parameter
     * index.
     *
     * @param[in] other_connectivity  The connectivity of the other constraint
     * @param[in] other_pivot_index   The parameter of the pivot relative
     *                                to 'other'.
     * @param[out] to_common_param    The translation of 'this' parameters to
     *                                their index in the common parameters
     * @return The parameter of the pivot variable in 'this'.
     */
    unsigned get_this_on_right_common_param(
        const constraint_connectivity_info & other_connectivity,
        unsigned other_pivot_index,
        std::vector<unsigned> & to_common_param) const;
    
    template <class T>
    void swap_if_first_bigger(std::vector<T> & first, std::vector<T> & second)
    {
        if (first.capacity() > second.capacity())
            std::swap(first, second);
    }
    /**
     * @brief Minimal relaxation to ensure that the latest node causes a
     *        conflict of the given clause
     *
     * This is like relax_cause but it works only on the provided clause.
     *
     * @param clause  A clause to check on the implication nodes. It is most
     *                likely that the clause is an explanation of the
     *                constraint, or the original clause.
     * @return True if anything has changed on the learned_type.
     */
    bool relax_cause_for_clause(const std::vector<iset> & clause);
    
    /**
     * @brief Minimal relaxation to ensure that the latest node causes a
     *        conflict of the given constraint
     *
     * This is like relax_cause but it works only on the provided
     * constraint.
     *
     * @param[in,out] cloned_constraint
     *                A copy of learned_type's constraint which is safe to
     *                propagate. The cloned_constraint is initialized with
     *                the provided \p info.
     * @param[in,out] info.
     *                An info with unknown state not earlier than the ground
     *                level. Invoking info.undo_changes() will bring it back to
     *                the state defined by the current learned_type object.
     * param[in]      ground_domains
     *                For each parameter \i for which modification(i) is not a
     *                ground level modification, ground_domains[i] points to the
     *                latest value of variable(i) at ground level.
     * @return True if anything has changed on the learned_type.
     */
    bool relax_cause_for_constraint(
        constraint_propagator &cloned_constraint, info_type & info,
        const std::vector<const iset*> & ground_domains);
    bool relax(unsigned parameter, const iset & literal);
    bool relax_level(unsigned parameter, const iset & literal);
    bool relax_level_all(const std::vector<iset> & literals);
    void set_modification(unsigned parameter, change_index_t mod)
    {
        if (mod != modification(parameter))
            unconditional_set_modification(parameter, mod);
    }
    void unconditional_set_modification(unsigned parameter, change_index_t mod)
    {
        m_explanation_for_these_mods= false;
        m_implication_nodes[parameter].m_modification_entry= mod;
    }
    
    //! A special case of intersect_with_ground() which works only when
    //! m_constraint.get() != NULL. It is used in the implementation of
    //! intersect_with_ground().
    void intersect_general_constraint_with_ground();
    /**
     * @brief For each parameter find values that if added to the domain
     *        will possibly make the propagation not happen.
     *
     * A value for a given variable is checked assuming all other domains stay
     * intact. The result may be over-approximated, if a value make the
     * propagation not happen then it must be in the returned set. On the other
     * hand if a value does not affects propagation, it is allowed (but not
     * required, and preferably not) to be in the set.
     */
    std::vector<iset> possibly_crippling_values() const;
    void compute_level()
    {
	change_index_t latest_modification= NO_PREVIOUS_MODIFICATION;
	for (unsigned i=0 ; i != m_implication_nodes.size() ; ++i) {
	    latest_modification= assignment_stack::latest_modification(
                latest_modification, modification(i));
	}
	m_latest_level= m_stack.decision_level_of(latest_modification);
	m_nodes_at_latest_level= 0;
	for (unsigned i=0 ; i != size() ; ++i)
	    if ( is_level_modification( modification(i) ) )
		++m_nodes_at_latest_level;
    }
    void       erase_literal(unsigned literal_index);
    change_index_t level_modification() const
    { return m_stack.decision_modification(m_latest_level); }

    bool       is_level_modification(change_index_t mod) const
    { return !m_stack.is_earlier_modification( mod, level_modification()); }

    //! Set the current constraint() to point to and own the given
    //! new_constraint. Mark it as a new constraint.
    void update_constraint(
        std::unique_ptr<constraint_propagator> &new_constraint)
    {
        m_constraint= move(new_constraint);
        m_explanation_for_these_mods= false;
        m_literals.clear();
    }
    //! Set the current constraint() to point to and own the given
    //! new_constraint. Mark it as a new constraint.
    void update_constraint(constraint_propagator * new_constraint)
    {
        m_constraint.reset(new_constraint);
        m_explanation_for_these_mods= false;
        m_literals.clear();
    }
    //! Set the current constraint() to point to constraint_ref, but not own it.
    //! Also, mark that this is an existing constraint.
    void set_original_constraint(const constraint_propagator & constraint_ref)
    {
        m_constraint.refer_to(constraint_ref);
        m_explanation_for_these_mods= false;
        m_literals.clear();
    }
    
    const assignment_stack & m_stack;
    const solver & m_solver;
    //! True if m_literals was created for the nodes and modification times
    //! of m_implication_nodes. If nodes are relaxed then the explanation
    //! is still valid but m_explanation_for_these_mods==false.
    bool m_explanation_for_these_mods;
    std::vector<implication_node> m_implication_nodes;
    /**
     * @brief Either the literals of the current clause or an explanation for
     *        the constraint().
     *
     * If we have learned a constraint then m_constraint should point to it
     * and m_literals should be either empty or an explanation to the
     * constraint.
     */
    std::vector<iset> m_literals;

    //! If we have learned a constraint then m_constraint should point to it
    //! and m_literals is either empty or is an explanation for the constraint.
    controlled_ptr<constraint_propagator> m_constraint;
    //! Reverse map from variable index back to a relevant parameter in the
    //! clause.
    std::vector<unsigned> m_variable_to_literal_index;
    unsigned         m_latest_level;
    unsigned         m_nodes_at_latest_level;
    unsigned         m_effort;

    static std::vector<implication_node> m_implication_nodes_cache;
    static std::vector<iset> m_literals_cache;
    static std::vector<unsigned> m_variable_to_literal_index_cache;
}; // class learned_type
} // namespace hcsp
#endif // #LEARNED_TYPE_HPP_
