// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <neq_propagators.hpp>

#include <iostream>
#include <literal_constraint_propagators.inl>
#include <arithmetic_propagators.hpp>

using namespace std;
using namespace hcsp;

class neq_propagator::neq_rule : public explanation_rule
{
  public:
    neq_rule(const vector<iset> & literals)
    {
        HCSP_ASSERT(literals.size() == 2);
        HCSP_ASSERT(literals[0] == literals[1]);
        iset conflicting= literals[0];
        conflicting.complement();
        HCSP_ASSERT(conflicting.singleton());
        m_value = conflicting.max();
    }
    virtual neq_rule * clone() const { return new neq_rule(*this); }
    
    virtual void get_explanation(std::vector<iset> & explanation) const
    {
        explanation.clear();
        explanation.resize(2);
        explanation[0].assign(m_value);
        explanation[0].complement();
        explanation[1]= explanation[0];
    }
    virtual void print(std::ostream & out, const variable_print_strategy &,
                       const std::vector<unsigned> &) const
    { out << "DIFF(" << m_value << ")"; }
  private:
    iset::value_type m_value;
}; // neq_propagator::neq_rule

class literal_or_neq_propagator::literal_or_neq_rule: public explanation_rule
{
  public:
    literal_or_neq_rule(const vector<iset> & literals, const iset & guard_set)
        : m_guard_set(guard_set)
    {
        HCSP_ASSERT(literals.size() == 3);
        HCSP_ASSERT(literals[1] == literals[2]);
        m_bad_precondition= literals[1].empty();
        if (m_bad_precondition)
            return;
        iset conflicting= literals[1];
        conflicting.complement();
        HCSP_ASSERT(conflicting.singleton());
        m_value = conflicting.max();
    }
    virtual literal_or_neq_rule * clone() const
    { return new literal_or_neq_rule(*this); }
    
    virtual void get_explanation(std::vector<iset> & explanation) const
    {
        explanation.clear();
        explanation.resize(3);
        if (m_bad_precondition) {
            explanation[0].assign(make_pair(iset::MIN_DOMAIN,iset::MAX_DOMAIN));
        } else {
            explanation[0]= m_guard_set;
            explanation[1].assign(m_value);
            explanation[1].complement();
            explanation[2]= explanation[1];
        }
    }
    virtual void print(std::ostream & out, const variable_print_strategy &,
                       const std::vector<unsigned> &) const
    {
        if (m_bad_precondition)
            out << "LIT_OR_NEQ_DOM";
        else
            out << "LIT_OR_NEQ({" << m_guard_set << "}," <<  m_value << ")";
    }
  private:
    iset::value_type m_value;
    iset m_guard_set;
    bool m_bad_precondition;
}; // literal_or_neq_propagator::literal_or_neq_rule


// virtual
bool neq_propagator::init( constraint_runtime_info & info )
{
    HCSP_ASSERT(info.arity() == (unsigned)count_params());
    const unsigned size= info.arity();
    for (unsigned i=0 ; i != size ; ++i)
        info.allocate_watch(i, FINALIZATION_WATCH);
    return true;
}

// virtual
constraint_propagator * neq_propagator::intersect_with(
    const constraint_runtime_info & info,
    vector<unsigned> & copy_from_argument) const
{
    HCSP_ASSERT(info.arity() == 2);
    static const iset zero_one(make_pair(0,1));
    if (info.domain(0) == zero_one && info.domain(1) == zero_one) {
        copy_from_argument = vector<unsigned>{0,1};
        return new sum_mod_propagator(/* mod */ 2, /* result */ 1);
    } else {
        return nullptr;
    }
}

// virtual
neq_propagator::result_type
neq_propagator::propagate( constraint_runtime_info & info)
{
    HCSP_PROFILE_CURRENT;
    const iset & in0= info.domain(0);
    const iset & in1= info.domain(1);
    if ( !in0.contains_any_of(in1) )
        return VALID;
    if (in0.singleton()) {
        iset out1 = in1;
        out1.erase_all_of(in0);
        if (out1.empty())
            return UNSAT;
        info.set_domain(1, move(out1));
        return VALID;
    }
    if (in1.singleton()) {
        iset out0 = in0;
        out0.erase_all_of(in1);
        info.set_domain(0, move(out0));
        return VALID;
    }
    return CONSISTENT;
} 

// virtual
bool neq_propagator::conflicts(const constraint_runtime_info & info) const
{
    return info.domain(0).singleton() && info.domain(0)==info.domain(1);
}

// virtual
bool neq_propagator::constraint_check(const constraint_runtime_info &info) const
{
    HCSP_ASSERT(info.arity() == 2);
    return info.domain(0) != info.domain(1);
}

// virtual
std::unique_ptr<explanation_rule> neq_propagator::rule(
        const std::vector<iset> & literals) const
{
    return unique_ptr<explanation_rule>(new neq_rule(literals));
}

//virtual
void neq_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        p_info->print_param(out, 0);
        out << "!=";
        p_info->print_param(out, 1);
        out << ")";
    } else {
        out << "neq";
    }
}

// virtual
bool difference_neq_constant_propagator::init( constraint_runtime_info & info )
{
    HCSP_ASSERT((unsigned)count_params() == info.arity());
    const unsigned size= info.arity();
    for (unsigned i=0 ; i != size ; ++i)
        info.allocate_watch(i, FINALIZATION_WATCH);
    return true;
}


//virtual
difference_neq_constant_propagator::result_type
difference_neq_constant_propagator::propagate( constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    
    const iset & in0= info.domain(0);
    const iset & in1= info.domain(1);
    if (in0.singleton())
    {
        long long forbid1= in0.max() - (long long)m_constant;
        iset out1 = in1;
        if (forbid1 >= iset::MIN_DOMAIN && forbid1 <= iset::MAX_DOMAIN)
            out1.erase(forbid1);
        if (out1.empty())
            return UNSAT;
        info.set_domain(1, move(out1));
        return VALID;
    }
    if (in1.singleton())
    {
        long long forbid0= in1.max() + (long long)m_constant;
        iset out0= in0;
        if (forbid0 >= iset::MIN_DOMAIN && forbid0 <= iset::MAX_DOMAIN)
            out0.erase(forbid0);
        if (out0.empty())
            return UNSAT;
        info.set_domain(0, move(out0));
        return VALID;
    }
    return CONSISTENT;
} 

//virtual
bool difference_neq_constant_propagator::conflicts(
    const constraint_runtime_info & info) const
{
    const iset & in0= info.domain(0);
    const iset & in1= info.domain(1);
    if (in0.singleton() && in1.singleton())
        return !constraint_check(info);
    else
        return false;
}

bool difference_neq_constant_propagator::constraint_check(
    const constraint_runtime_info & info ) const
{
    long long left= info.domain(0).min(), right= info.domain(1).min();
    return left-right != m_constant;
}

//virtual
void difference_neq_constant_propagator::print(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        p_info->print_param(out, 0);
        out << "-";
        p_info->print_param(out, 1);
        out << "!=" << m_constant << ")";
    } else {
        out << "difference_neq_constant(" << m_constant << ")";
    }
}

//virtual
bool literal_or_neq_propagator::init_non_literal_part(
    constraint_runtime_info & info)
{
    const iset & domain1= info.domain(1);
    const iset & domain2= info.domain(2);
    if ( ! domain1.contains_any_of(domain2) ) // always different
        return true;

    if (domain1.singleton() && domain1 == domain2)
        return false;

    info.allocate_watch(1, FINALIZATION_WATCH);
    info.allocate_watch(2, FINALIZATION_WATCH);
    
    return true;
}

// virtual
constraint_propagator::result_type
literal_or_neq_propagator::propagate_non_literal_part(
    constraint_runtime_info & info )
{
    const iset & in_first= info.domain(1);
    const iset & in_second= info.domain(2);

    // We assume that trivial cases have been handled by the propagate()
    // algorithm before propagate_non_literal_part() was called.
    HCSP_ASSERT(state_of_non_literal_part(info) == CONSISTENT);

    if ( in_second.singleton() ) {
        iset out_first = in_first;
        out_first.erase_all_of(in_second);
        HCSP_ASSERT( !out_first.empty() );
        info.set_domain(1, move(out_first));
        return VALID;
    }
    if ( in_first.singleton() ) {
        iset out_second = in_second;
        out_second.erase_all_of(in_first);
        HCSP_ASSERT( !out_second.empty() );
        info.set_domain(2, move(out_second));
        return VALID;
    }
    if ( in_first.contains_any_of(in_second) )
        return CONSISTENT;
    else
        return VALID;
}

// virtual
constraint_propagator::result_type
literal_or_neq_propagator::state_of_non_literal_part(
    const constraint_runtime_info & info) const
{
    const iset & in_first= info.domain(1);
    const iset & in_second= info.domain(2);
    if ( in_first.empty() || in_second.empty() )
        return UNSAT;

    if (!in_first.contains_any_of(in_second))
        return VALID;
    if (in_first.singleton() && in_first == in_second)
        return UNSAT;
    return CONSISTENT;
}

// virtual
std::unique_ptr<explanation_rule> literal_or_neq_propagator::rule(
        const std::vector<iset> & literals) const
{
    return unique_ptr<explanation_rule>(new literal_or_neq_rule(literals,
                                                              guard_set()));
}

void literal_or_neq_propagator::print_non_literal_part(
    ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_param(out, 1);
        out << "!=";
        p_info->print_param(out, 2);
    } else {
        out << "neq_propagator";
    }
}

