// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2011, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2011, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <algorithm>
#include <iset.hpp>
#include <clause_propagator.hpp>
#include <solver.hpp>
#include <constraint_connectivity_info.hpp>
#include <opb_conjunction.hpp>
#include <opb_variable.hpp>
#include <opb_context.hpp>

using namespace std;
using namespace hcsp;
using namespace hcsp::opb;


//! A function object for comparing literals, starting with variable name.
class conjunction::literal_variable_comparator
{
  public:
    /**
     * @brief Compare the two literals
     *
     * Start with variables, if they are different then return true if
     * the first is lexicographically smaller than the second. If variables
     * are the same then return true if the first is negative and the second
     * is positive.
     */
    bool operator()(const literal_type &a, const literal_type &b)
    {
        if (a.second != b.second)
            return strcmp(a.second->name(), b.second->name())<0;
        // put negated literals before positive literals
        if (a.first < 0 && b.first > 0)
            return true;
        else
            return false;
    }
}; // class hcsp::opb::conjunction::literal_variable_comparator

conjunction::conjunction(int polarity, const variable & variable)
    : m_is_false(false)
{
    append_literal(polarity, variable);
}

conjunction::~conjunction()
{
}

void conjunction::append_literal(int polarity, const variable &variable)
{
    m_literals.push_back(literal_type(polarity, &variable));
}


unsigned conjunction::get_atom_variable(context & context,
                                        int coefficient,
                                        bool apply_not,
                                        eval_type bound,
                                        const char *atom_name)
{
    if (m_is_false || coefficient == 0)
        return context.get_solver().singleton_var(apply_not);

    if (m_literals.size() == 1) {
        HCSP_ASSERT(m_literals[0].first > 0);
        // first make sure that the variable itself is the main literal
        m_main_values.set(1);
        m_main_connectivity.set_equals_variable(m_literals[0].second->id());
        m_variables.insert(make_pair(m_main_values, m_main_connectivity));
    }
    
    atom_connectivity_t connectivity(coefficient, apply_not, bound);
    atom_values_t values(coefficient, apply_not);
    unsigned variable;
    variable=variable_for_coefficient(context, values, connectivity, atom_name);
    if ( !m_main_connectivity.is_as_strong_as(connectivity) )
        strengthen_conjunction_result(context, connectivity);
    
    return variable;
}

unsigned conjunction::variable_for_coefficient(
    context&context, const atom_values_t &values,
    const atom_connectivity_t &need_connectivity, const char *atom_name)
{
    HCSP_ASSERT(values.has_values());
    typedef variable_collection_t::iterator iterator;
    
    pair<iterator, bool> cached_value_info;

    const atom_connectivity_t unbound;     // Keep the variable unconstrained
    // The variable may exist, in that case its bound is unchanged
    cached_value_info= m_variables.insert(make_pair(values, unbound));
    
    iterator cached_value_it= cached_value_info.first;
    atom_connectivity_t & connectivity= cached_value_it->second;
    if (!cached_value_info.second) {
        // This is a new variable, constrain it appropriately
        strengthen_variable(context, values, need_connectivity, connectivity);
        return connectivity.m_variable;
    }
    
    iset domain(values.m_on_one);
    domain.insert(values.m_on_zero);
    unsigned variable= context.get_solver().add_var(domain, atom_name);
    connectivity.m_variable= variable;

    if (m_main_values.has_values()) {
        strengthen_variable(context, values, need_connectivity, connectivity);
    } else {
        m_main_values= values;
        m_main_connectivity.m_variable= variable;
    }
    return variable; 
} 

void conjunction::strengthen_variable(
    context & context,
    const atom_values_t &values,
    const atom_connectivity_t &required_connectivity,
    atom_connectivity_t & connectivity)
{
    unsigned main_variable= m_main_connectivity.m_variable;

    if (main_variable == connectivity.m_variable)
        return;

    //
    // link the main variable with the current variable
    //

    if ( required_connectivity.m_one_affects_atom &&
         !  connectivity.m_one_affects_atom ) {
        insert_clause(context.get_solver(),
                      main_variable, iset(m_main_values.m_on_one).complement(),
                      connectivity.m_variable, iset(values.m_on_one));
        connectivity.m_one_affects_atom= true;
    }
 
    if ( required_connectivity.m_zero_affects_atom &&
         ! connectivity.m_zero_affects_atom ) {
        insert_clause(context.get_solver(),
                      main_variable, iset(m_main_values.m_on_zero).complement(),
                      connectivity.m_variable, iset(values.m_on_zero));
        connectivity.m_zero_affects_atom= true;
    } 
}

void conjunction::strengthen_conjunction_result(
    context & context, const atom_connectivity_t &connectivity)
{
    unsigned main_variable= m_main_connectivity.m_variable;

    if ( connectivity.m_zero_affects_atom &&
         !m_main_connectivity.m_zero_affects_atom ) {
        a_false_literal_implies_zero_atom(context.get_solver(), main_variable);
        m_main_connectivity.m_zero_affects_atom= true;
    }

    if ( connectivity.m_one_affects_atom &&
         !m_main_connectivity.m_one_affects_atom ) {
        true_literals_imply_conjunction(context.get_solver(), main_variable);
        m_main_connectivity.m_one_affects_atom= true;
    } 
}

void conjunction::normalize()
{
    for (unsigned i=0 ; i != m_literals.size() ; ++i) {
        if (m_literals[i].first > 1)
            m_literals[i].first=1;
        else if (m_literals[i].first < -1)
            m_literals[i].first=-1;
    }
    sort(m_literals.begin(), m_literals.end(), literal_variable_comparator());
    m_literals.erase(unique(m_literals.begin(), m_literals.end()),
                     m_literals.end());
    
    for (unsigned i=1 ; i < m_literals.size() ; ++i) {
        if (m_literals[i-1].second == m_literals[i].second) {
            HCSP_ASSERT(m_literals[i-1].first < 0);
            HCSP_ASSERT(m_literals[i].first > 0);
            m_literals.clear();
            m_is_false= true;
            break;
        }
    }
}

conjunction *conjunction::get_not(context & context) const
{
    if (context.trace_construction())
        cout << *this << ".get_not()->";
    if (m_literals.size() != 1) {
        if (context.trace_construction())
            cout << "NULL\n";
        return NULL;
    }
    conjunction *negated;
    negated= new conjunction(-m_literals[0].first, *m_literals[0].second);
    if (context.trace_construction())
        cout << *negated << "\n";
    return &context.record_unique_conjunction(negated);
}

size_t conjunction::hash() const
{
    if (m_is_false)
        return 0;
    size_t accumulated=1;
    for (unsigned i=0 ; i != m_literals.size() ; ++i) {
        accumulated *= 98321; // a prime number bigger than 2**(sizeof(char)*2)
        accumulated += m_literals[i].first * m_literals[i].second->hash();
    }
    return accumulated;
}

void conjunction::a_false_literal_implies_zero_atom(solver & solver,
                                                    unsigned atom_variable)const
{
    for (unsigned i=0 ; i != m_literals.size() ; ++i) {
        insert_clause(solver,
                      m_literals[i].second->id(), iset(m_literals[i].first > 0),
                      atom_variable, iset(m_main_values.m_on_zero));
    }
}

void conjunction::insert_clause(solver & solver,
                                unsigned variable_a, const iset & set_a,
                                unsigned variable_b, const iset & set_b) const
{
    HCSP_ASSERT(variable_a != variable_b);
    clause_propagator* clause= new clause_propagator;
    unique_ptr<constraint_propagator> constraint(clause);
    clause_propagator::literals_type & literals= clause->update_literals();
    clause= NULL;
    literals.resize(2);
    literals[0]= set_a;
    literals[1]= set_b;
    solver.add_constraint(move(constraint)).reserve_arguments(2)
        << variable_a << variable_b;
}

void conjunction::true_literals_imply_conjunction(solver & solver,
                                                  unsigned atom_variable) const
{
    clause_propagator* clause= new clause_propagator;
    unique_ptr<constraint_propagator> constraint(clause);
    clause_propagator::literals_type & literals= clause->update_literals();
    clause= NULL;
    literals.resize(m_literals.size()+1);
    constraint_connectivity_info &
        info = solver.add_constraint(move(constraint));
    for (unsigned i=0 ; i != m_literals.size() ; ++i) {
        literals[i].assign(m_literals[i].first < 0);
        info << m_literals[i].second->id();
    }
    literals.back().assign(m_main_values.m_on_one);
    info << atom_variable;
}

bool conjunction::evaluate(const context & context) const
{
    for (unsigned i=0 ; i != m_literals.size() ; ++i) {
        unsigned var_id= m_literals[i].second->id();
        const iset & domain= context.get_solver().var_domain(var_id);
        HCSP_ASSERT(domain.singleton());
        bool positive_literal=  domain.min();
        if (m_literals[i].first < 0)
            positive_literal= !positive_literal;
        if (!positive_literal)
            return false;
    }
    return true;
}

namespace hcsp { namespace opb {
bool operator<(const conjunction &a, const conjunction &b)
{
    if (a.m_is_false || b.m_is_false)
        // put the false conjuncts before all the rest
        return a.m_is_false > b.m_is_false;

    size_t min_size= min(a.m_literals.size(), b.m_literals.size());
    for (unsigned i=0 ; i != min_size ; ++i) {
        // compare the variable and only then its polarity in the literal
        if (a.m_literals[i].second != b.m_literals[i].second) {
            const char * a_name= a.m_literals[i].second->name();
            const char * b_name= b.m_literals[i].second->name();
            HCSP_ASSERT(strcmp(a_name, b_name) != 0);
            return strcmp(a_name, b_name)<0;
        }
        if (a.m_literals[i].first != b.m_literals[i].first)
            return a.m_literals[i].first < b.m_literals[i].first;
    }
    return false;
}

std::ostream& operator<<(std::ostream&out, const conjunction& conjunct)
{
    for (unsigned i=0 ; i != conjunct.m_literals.size() ; ++i) {
        if (i != 0)
            out << " ";
        if (conjunct.m_literals[i].first < 0)
            out << "~";
        out << conjunct.m_literals[i].second->name();
        unsigned id= conjunct.m_literals[i].second->id();

        if (id != -1U)
            out << "[" << id << "]";
        
    }
    if (conjunct.m_literals.empty())
        out << "empty";
    return out;
}
}} // namespace hcsp::opb

conjunction::atom_connectivity_t::atom_connectivity_t(int coefficient,
                                                      bool apply_not,
                                                      eval_type bound)
    : m_variable(-1U),
      m_zero_affects_atom(false),
      m_one_affects_atom(false)
{
    if ( (bound & LOWER_BOUND_EVAL) == LOWER_BOUND_EVAL)
        // atom >= min_value
        m_zero_affects_atom= true;
    
    if ( (bound & UPPER_BOUND_EVAL) == UPPER_BOUND_EVAL)
        // atom <= max_value
        m_one_affects_atom= true;
    
    if (coefficient < 0)
        // -abs(coefficient)*conjunction >= value
        //    means abs(coefficient)*conjunction <= -value (all reversed)
        // same for -abs(coefficient)*conjunction <= value
        std::swap(m_zero_affects_atom, m_one_affects_atom);

    if (apply_not)
        std::swap(m_zero_affects_atom, m_one_affects_atom);
}

void conjunction::atom_connectivity_t::set_equals_variable(unsigned variable)
{
    m_variable= variable;
    m_zero_affects_atom= true;
    m_one_affects_atom= true;
}

bool conjunction::atom_connectivity_t::is_as_strong_as(
    const atom_connectivity_t & other) const
{
    if (m_zero_affects_atom < other.m_zero_affects_atom)
        return false;
    if (m_one_affects_atom < other.m_one_affects_atom)
        return false;
    return true;
}
