// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2011, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2011, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <string>
#include <solver.hpp>
#include <arithmetic_propagators.hpp>
#include <opb_context.hpp>
#include <opb_constraint.hpp>
#include <opb_expression.hpp>
#include <opb_variable.hpp>
#include <opb_conjunction.hpp>

using namespace std;
using namespace hcsp;
using namespace hcsp::opb;

/**
 * @brief A hash entry for a conjunction.
 *
 * It contains both the calculated hash and the conjunction itself.
 * Comparisons are first made on the hash value and only then on the
 * conjunction.
 *
 * The ownership of the conjunction object is inconclusive. It is up to the
 * caller to decide if the conjunction object should be deleted by calling the
 * delete_conjunction_object() method, or not.
 */
class context::conjunction_hash_entry
{
  public:
    //! Construct an entry, including calculating and recording hash value.
    conjunction_hash_entry(conjunction * conjunction_in)
        : m_hash(conjunction_in->hash()), m_conjunction(conjunction_in)
    {}

    //! Delete contained the conjunction object
    void delete_conjunction_object() const
    {
        delete m_conjunction;
    }

    //! Get the current conjunction object
    conjunction & get_conjunction() const
    {
        return *m_conjunction;
    }
    
    //! Compare the hash and, if equal, compare the conjunction
    friend bool operator==(const conjunction_hash_entry & a,
                           const conjunction_hash_entry & b)
    { return a.m_hash==b.m_hash && *a.m_conjunction == *b.m_conjunction; }
    
    //! Compare the hash and, if equal, compare the conjunction
    friend bool operator<(const conjunction_hash_entry & a,
                          const conjunction_hash_entry & b)
    {
        return (a.m_hash<b.m_hash ||
                (a.m_hash==b.m_hash && *a.m_conjunction < *b.m_conjunction));
    }
  private:
    size_t       m_hash;
    
    /**
     * @brief A potentially owned conjunction.
     *
     * For more information on the ownership model, read the description of
     * the class. These comments are at the beginning of the class.
     */
    conjunction *m_conjunction;
};  // class opb::context::conjunction_hash_entry

context::context(solver & solver,
                 std::istream* is, const std::string & filename)
    : m_scanner(0),
      m_is(is),
      m_filename(filename),
      m_direction(NO_OBJECTIVE),
      m_objective(NULL),
      m_objective_var(-1U),
      m_solver(solver)
{
    *is >> std::noskipws;
    init_scanner();
}

//virtual
context::~context()
{
    destroy_scanner();
    delete m_objective;
    m_objective= NULL;

    while ( !m_constraints.empty() ) {
        delete m_constraints.back();
        m_constraints.back()= NULL;
        m_constraints.pop_back();
    }
    while ( !m_conjunctions.empty() ) {
        // Since the context owns the conjunctions it has to delete them
        // in the destructor.
        m_conjunctions.begin()->delete_conjunction_object();
        m_conjunctions.erase(m_conjunctions.begin());
    }
}


bool context::trace_parsing() const
{
    return m_solver.report_level() >= 3;
}

bool context::trace_construction() const
{
    return m_solver.report_level() >= 2;
}

bool context::trace_failure() const
{
    return m_solver.report_level() >= 1;
}

void context::add_constraint(constraint *p_constraint)
{
    unique_ptr<constraint> auto_free(p_constraint);
    p_constraint->normalize(*this);
    m_constraints.push_back(p_constraint);
    auto_free.release();
}

void context::set_objective(expression *objective, direction bound)
{
    delete m_objective;
    m_objective= objective;
    m_direction= bound;
    if (trace_construction())
        cout << "objective[" << *m_objective << "].normalize()\n";
    //int rhs=0;
    //! @todo Normalizing the objective is disabled because it seems to hurt
    //!       performance. Fix the issue and re-enable normalization;
    //!       
    //    m_objective->normalize(*this, rhs);
    //m_objective_offset= -rhs;
    //if (trace_construction())
    //    cout << "[disabled] normalize()->objective[" << *m_objective << "]+"
    //         << m_objective_offset << "\n";
}

iset::value_type context::evaluate_objective() const
{
    return m_objective->evaluate(*this);
}

const variable& context::get_variable(char *name)
{
    return *m_variables.insert(name).first;
}

conjunction& context::record_unique_conjunction(conjunction*p_conjunction)
{
    unique_ptr<conjunction> auto_free(p_conjunction);
    p_conjunction->normalize();
    conjunction_hash_entry entry(p_conjunction);
    pair<set<conjunction_hash_entry>::iterator,bool>
        insert_result= m_conjunctions.insert(entry);
    if ( insert_result.second )
        // Successfully inserted, should not be deleted.
        // (otherwise 'conjunction' has to be deleted upon return)
        auto_free.release();

    return insert_result.first->get_conjunction();
}

bool context::build_solver()
{
    set<variable>::iterator it;
    for (it= m_variables.begin(); it != m_variables.end() ; ++it) {
        variable & variable= const_cast<opb::variable&>(*it);
        HCSP_ASSERT(variable.id() == -1U);
        variable.associate(m_solver);
    }
    for (unsigned i=0 ; i != m_constraints.size() ; ++i)
        if (!m_constraints[i]->apply(*this))
            return false;
    if (m_objective) {
        if (m_direction == MIN_OBJECTIVE)
            return connect_minimization();
        else
            return connect_maximization();
    }
    return true;
}

bool context::connect_maximization()
{
    HCSP_ASSERT(m_direction== MAX_OBJECTIVE);

    // Here we maximize the objective function max(objective)= max::objective
    // such that objective >= max::objective is satisfiable while
    // objective >= max::objective+1 is not.

    expression::constraint_type bound= expression::CONSTRAIN_AT_LEAST;
    m_objective_var=  m_objective->get_evaluating_variable(*this, bound,
                                                           "max::objective");
    m_solver.objective_variable(m_objective_var);
    m_solver.objective_is_minimizing(false);
    return true;
}

bool context::connect_minimization()
{
    HCSP_ASSERT(m_direction== MIN_OBJECTIVE);
    
    // Here we minimize the objective function min(objective)= min::objective
    // such that objective <= min::objective is satisfiable while
    // objective <= min::objective-1 is not.
    //
    // Since opb::expression does not work with 'expression <= variable'
    // constraints then we have to negate everything. Negating the
    // expressions in the previous paragraph gives:
    //
    // Here we maximize the negated objective function
    // max(-objective)= max::neg::objective such that
    // -objective >= max::neg::objective is satisfiable while
    // -objective >= max::neg::objective+1 is not.
    expression::constraint_type bound= expression::CONSTRAIN_AT_LEAST;
    unsigned neg_objective;
    m_objective->negate_coefficients();
    neg_objective=  m_objective->get_evaluating_variable(*this, bound,
                                                         "max::neg::objective");
    m_objective->negate_coefficients(); // reverse the effect of negation
    const iset & neg_objective_domain= m_solver.var_domain(neg_objective);
    
    const iset minimized_domain(make_pair(-neg_objective_domain.max(),
                                          -neg_objective_domain.min()));
    
    m_objective_var= m_solver.add_var(minimized_domain, "min::objective");
    unique_ptr<constraint_propagator> neg(new neg_propagator);
    m_solver.add_constraint(move(neg)) << neg_objective << m_objective_var;
    m_solver.objective_variable(m_objective_var);
    m_solver.objective_is_minimizing(true);
    return true;

}

void context::print_solution(ostream & out) const
{
    cout << "v ";
    set<variable>::iterator it;
    for ( it= m_variables.begin(); it != m_variables.end() ; ++it) {
        HCSP_ASSERT(it->id() != -1U);
        const iset & domain= m_solver.var_domain(it->id());
        HCSP_ASSERT(domain.singleton());
        out << ' ';
        if (domain.min() == 0)
            out << '-';
        out << it->name();
    }
    out << endl;
}

namespace hcsp { namespace opb {
ostream & operator<<(ostream &out, const context & data)
{
    for (unsigned i=0 ; i != data.m_constraints.size() ; ++i) {
        cout << "constraint[" << i << "]: "
             << *data.m_constraints[i] << "\n";
    }
    return out;
}
}} // namespace hcsp::opb
