// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2011, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2011, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <memory>
#include <algorithm>
#include <limits>
#include <iset.hpp>
#include <solver.hpp>
#include <clause_propagator.hpp>
#include <arithmetic_propagators.hpp>
#include <opb_expression.hpp>
#include <opb_conjunction.hpp>
#include <opb_context.hpp>

using namespace std;
using namespace hcsp;
using namespace hcsp::opb;

/**
 * @brief An OPB atom, combining a conjunction of literals and a coefficient.
 *
 * The value of the atom is calculated as follows:
 *  - m_coefficient * (m_apply_not \b xor evaluate(m_conjunction))
 */
struct expression::atom_type
{
    //! Read field documentation for more information on the constructor
    atom_type(int coefficient,
              bool apply_not,
              conjunction* conjunction)
        : m_coefficient(coefficient),
          m_apply_not(apply_not),
          m_conjunction(conjunction)
    {}
    atom_type() {}
    int          m_coefficient;//!< The integer coefficient to multiply at end
    bool         m_apply_not;  //!< Apply a 'not' over the conjunction
    conjunction *m_conjunction; //!< The conjunction of literals
};

/**
 * @brief Compare two atoms according to the conjunction they have
 */
class expression::compare_conjunctions
{
  public:
    bool operator()(const atom_type &a, const atom_type &b)
    {
        if ( *a.m_conjunction == *b.m_conjunction )
            return a.m_apply_not < b.m_apply_not;
        else
            return *a.m_conjunction < *b.m_conjunction;
    }
};

expression::expression(int coefficient, conjunction &conjunction)
{
    append_atom(coefficient, conjunction);
}

expression::~expression()
{
    while (!m_atoms.empty()) {
        // Don't delete the conjunction because it is owned by opb::context
        // not the expression.
        m_atoms.back().m_conjunction= NULL;
        m_atoms.pop_back();
    }
}

unsigned expression::get_evaluating_variable(context & context,
                                             constraint_type bound,
                                             const char *expression_name) const
{
    unsigned expression_variable= -1U;
    iset sum_domain(0);
    eval_type conjunction_bound= BOTH_BOUNDS_EVAL;
    if (bound == CONSTRAIN_AT_LEAST)
        conjunction_bound= LOWER_BOUND_EVAL;
    for (unsigned i=0 ; i != m_atoms.size() ; ++i) {
        int coefficient= m_atoms[i].m_coefficient;
        bool apply_not= m_atoms[i].m_apply_not;
        unsigned atom_variable;
        const char *sub_exp_name= NULL;
        if (m_atoms.size() == 1)
            sub_exp_name= expression_name;
        atom_variable= m_atoms[i].m_conjunction->get_atom_variable(
            context, coefficient, apply_not, conjunction_bound, sub_exp_name);
        if (i == 0) {
            expression_variable= atom_variable;
            sum_domain.insert(coefficient);
        } else {
            sum_domain.assign(make_pair(sum_domain.min()+min(0, coefficient),
                                        sum_domain.max()+max(0, coefficient)));
            unsigned sum_result;
            if (i+1==m_atoms.size())
                sub_exp_name= expression_name;
            sum_result= context.get_solver().add_var(sum_domain, sub_exp_name);
            unique_ptr<constraint_propagator> constraint;
            if (bound == CONSTRAIN_EQUAL){
                constraint.reset(new sum_wrap_propagator);
            } else {
                HCSP_ASSERT(bound == CONSTRAIN_AT_LEAST);
                constraint.reset(new less_equal_sum_propagator);
            }
            context.get_solver().add_constraint(move(constraint))
                << sum_result << expression_variable << atom_variable;
            expression_variable= sum_result;
        }
    }
    return expression_variable;
}

void expression::append_atom(int coefficient, conjunction& conjunction)
{
    m_atoms.push_back(atom_type(coefficient, false, &conjunction));
}

void expression::negate_coefficients()
{
    for (unsigned i=0 ; i != m_atoms.size() ; ++i)
        m_atoms[i].m_coefficient= -m_atoms[i].m_coefficient;
}

void expression::make_atoms_positive(int & rhs)
{
    for (unsigned i=0 ; i != m_atoms.size() ; ++i) {
        if (m_atoms[i].m_coefficient < 0) {
            // Use the rule of (a*x)==(a*(1-!x))== -a*(!x)+a
            m_atoms[i].m_apply_not= !m_atoms[i].m_apply_not;
            rhs -= m_atoms[i].m_coefficient;
            m_atoms[i].m_coefficient *= -1;
        }
    }
}

void expression::merge_similar_conjunctions()
{
    sort(m_atoms.begin(), m_atoms.end(), compare_conjunctions());
    for (unsigned i = 1 ; i < m_atoms.size() ; ++i) {
        if (m_atoms[i].m_conjunction == m_atoms[i-1].m_conjunction) {
            m_atoms[i].m_coefficient += m_atoms[i-1].m_coefficient;
            m_atoms[i-1].m_coefficient= 0; // mark the other atom for deletion.
        }
    }
    unsigned write=0;
    for (unsigned i = 0 ; i < m_atoms.size() ; ++i) {
        if (m_atoms[i].m_coefficient==0 || m_atoms[i].m_conjunction->is_false())
            continue; // skip the atom, effectively deleting it.
        m_atoms[write++]= m_atoms[i];
    }
    if (write != m_atoms.size())
        m_atoms.erase(m_atoms.begin()+write, m_atoms.end());
}

void expression::normalize(context & context, int & rhs)
{
    if (m_atoms.empty())
        return;
    // For conjunctions which have only one literal that happens to be
    // negated, remove the logical not and move it to the whole atom.
    for (unsigned i=0 ; i != m_atoms.size() ; ++i) {
        if (m_atoms[i].m_conjunction->is_negated_literal()) {
            conjunction *p_not= m_atoms[i].m_conjunction->get_not(context);
            HCSP_ASSERT(p_not != NULL);
            m_atoms[i].m_conjunction= p_not;
            m_atoms[i].m_apply_not= !m_atoms[i].m_apply_not;
        }
    }
    
    // Make it easier to merge a sum of literals such as (2*!x) + (3*x)
    // by making all atoms without logical not. This aids the following phases
    // to detect such literals.
    for (unsigned i=0 ; i != m_atoms.size() ; ++i) {
        if (m_atoms[i].m_apply_not) {
            // Use the rule of (a*x)==(a*(1-!x))== -a*(!x)+a
            m_atoms[i].m_apply_not= false;
            rhs -= m_atoms[i].m_coefficient;
            m_atoms[i].m_coefficient *= -1;
        }
    }
    merge_similar_conjunctions();
    make_atoms_positive(rhs);
    
    sort(m_atoms.begin(), m_atoms.end(), compare_conjunctions());
}

bool expression::all_coefficients_positive() const
{
    for (unsigned i=0 ; i != m_atoms.size() ; ++i)
        if (m_atoms[i].m_coefficient <= 0)
            return false;
    return true;
}

void expression::add_or_between_conjunctions(context & context) const
{
    if (context.trace_construction())
        cout <<"[" << *this << "].add_or_between_conjunctions()\n";
    vector<unsigned > variables(m_atoms.size());
    for (unsigned i=0 ; i != m_atoms.size() ; ++i) {
        variables[i]= m_atoms[i].m_conjunction->get_atom_variable(
            context, 1, false, LOWER_BOUND_EVAL);        
    }
    clause_propagator *p_clause= new clause_propagator;
    unique_ptr<constraint_propagator> constraint(p_clause);
    clause_propagator::literals_type & literals= p_clause->update_literals();
    literals.resize(m_atoms.size(), iset());
    p_clause= NULL; // just in case, to avoid accidental use of this pointer
    constraint_connectivity_info &
        info= context.get_solver().add_constraint(move(constraint));
    info.reserve_arguments(m_atoms.size());
    for (unsigned i=0 ; i != m_atoms.size() ; ++i) {
        info << variables[i];
        if (m_atoms[i].m_apply_not)
            literals[i].assign(0);
        else
            literals[i].assign(m_atoms[i].m_coefficient);
    }
}

size_t expression::size() const
{
    return m_atoms.size();
}

int expression::atom_coefficient(unsigned index) const
{
    return m_atoms.at(index).m_coefficient;
}

bool expression::atom_is_logically_negated(unsigned index) const
{
    return m_atoms.at(index).m_apply_not;
}

const conjunction & expression::atom_conjunction(unsigned index) const
{
    return *m_atoms.at(index).m_conjunction;
}

iset::value_type expression::evaluate(const context & context) const
{
    // Calculating the return value using long long; assuming that
    // iset::value_type is significantly smaller than long long. The return
    // value is not long long because we can't be certain that the type of
    // iset::value_type won't be long long in the future. In which case the
    // implementation of this method will have to be modified; it has to be
    // able to manage overflows and underflows gracefully.
    
    long long sum=0;
    HCSP_ASSERT(sizeof(sum) > sizeof(iset::value_type));
    
    for (unsigned i=0 ; i != m_atoms.size() ; ++i) {
        bool is_positive= m_atoms[i].m_conjunction->evaluate(context);
        if (m_atoms[i].m_apply_not)
            is_positive= !is_positive;
        if (is_positive && m_atoms[i].m_coefficient != 0)
            sum += m_atoms[i].m_coefficient;
    }
    sum= min<long long>(sum, numeric_limits<iset::value_type>::max());
    sum= max<long long>(sum, numeric_limits<iset::value_type>::min());

    return sum;
}

namespace hcsp { namespace opb {
ostream& operator<<(ostream&out, const expression &data)
{
    if (data.m_atoms.empty())
        return out << "empty_sum";
    for (unsigned i=0 ; i != data.m_atoms.size() ; ++i) {
        if (i>0)
            out << " ";
        int coefficient= data.m_atoms[i].m_coefficient;
        if (coefficient >= 0)
            out << "+";
        out << coefficient << " ";
        const char *prefix="", *suffix="";
        if (data.m_atoms[i].m_apply_not) {
            if (data.m_atoms[i].m_conjunction->size() <= 1) {
                prefix= "~";
            } else {
                prefix= "~(";
                suffix= ")";
            }
        }
        out << prefix << *data.m_atoms[i].m_conjunction << suffix;
    }
    return out;
}
}} // namespace hcsp::opb
