// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010,2011 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010,2011 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <config.hpp>
#ifdef HAS_ALLOCA_H
#include <alloca.h>
#endif
#include <malloc.h>
#include <cstring>
#include <sstream>
#include <vector>
#include <map>
#include <solver.hpp>
#include "operator_definition_registry.hpp"
#include "operator_connectivity.hpp"

using namespace std;
using namespace hcsp;

/**
 * @brief 
 */
struct operator_connectivity::operator_entry_key
{
    operator_entry_key(const operator_definition & definition,
                       const unsigned *begin_vars, const unsigned *end_vars)
        : m_operator(&definition), m_variables(begin_vars, end_vars)
    {}
    
    friend bool operator<(const operator_entry_key & a,
                          const operator_entry_key & b)
    {
        // NOTE: To avoid random fluctuations in behavior we don't want to
        //       compare the map elements by m_operator pointers. Otherwise
        //       even the slightest optimization change may cause different
        //       operator order in the map, resulting with different runs.
        if (a.m_operator->tightness() != b.m_operator->tightness())
            // start with the strongest operators
            return a.m_operator->tightness() >  b.m_operator->tightness();
        if (a.m_variables == b.m_variables) {
            int comparison= strcmp(a.m_operator->name(), b.m_operator->name());
            if (comparison == 0)
                HCSP_ASSERT(a.m_operator == b.m_operator);
            return comparison < 0;
        } else {
            return a.m_variables < b.m_variables;
        }
    }
    friend bool operator==(const operator_entry_key & a,
                           const operator_entry_key & b)
    { return a.m_operator == b.m_operator && a.m_variables == b.m_variables; }
    
    friend bool operator!=(const operator_entry_key & a,
                           const operator_entry_key & b)
    { return !(a==b); }
    //! A pointer to the singleton operator definition. Not to be deleted.
    const operator_definition *m_operator;
    vector<unsigned> m_variables;
};

struct operator_connectivity::operator_entry_data
{
    //! The output variable associated with a given operator
    unsigned    m_variable;

    //! The bounds over the operator (one of eval_type).
    eval_type   m_bound;

    //! Flags to be passed to the operator when creating the constraints.
    //! These flags are created by the operator and their semantics is defined
    //! for each operator separately.
    unsigned    m_constraint_flags;
};

struct operator_connectivity::impl
    : public map<operator_entry_key, operator_entry_data>
{};

operator_connectivity::operator_connectivity(solver & the_solver)
    : m_solver(the_solver), m_p_impl(new impl)
{}

operator_connectivity::~operator_connectivity()
{}


unsigned operator_connectivity::record_operator(
  const operator_definition & recorded_operator, eval_type bound,
  const unsigned *begin_vars, const unsigned *end_vars,
  unsigned default_var,
  const iset & domain, unsigned constraint_flags)
{
    const unsigned arity=end_vars-begin_vars;
    // Using alloca and not vars[arity] because it is more portable
    unsigned * vars= static_cast<unsigned*>(alloca(arity*sizeof(unsigned)));
    memcpy(&vars[0], begin_vars, arity*sizeof(unsigned));
    recorded_operator.canonize_order(*this, &vars[0], &vars[0]+arity);
    const operator_entry_key key(recorded_operator, &vars[0], &vars[0]+arity);
    impl::iterator it= m_p_impl->lower_bound(key);
    if (it == m_p_impl->end() || it->first != key) {
        operator_entry_data data;
        if (default_var == -1U) {
            ostringstream name_build;
            name_build << recorded_operator.name() << m_p_impl->size();
            data.m_variable= m_solver.add_var(domain, name_build.str().c_str());
        } else
            data.m_variable= default_var;
        data.m_bound= bound;
        data.m_constraint_flags= constraint_flags;
        it= m_p_impl->insert(it, impl::value_type(key, data));
    } else {
        it->second.m_bound |= bound;
        it->second.m_constraint_flags|= constraint_flags;
    }
    return it->second.m_variable;
}

void operator_connectivity::create_all_constraints()
{
    for ( impl::const_iterator it= m_p_impl->begin();
          it != m_p_impl->end() ;
          ++it) {
        it->first.m_operator->create_constraints(
            m_solver, it->second.m_bound, it->second.m_constraint_flags,
            it->second.m_variable, &it->first.m_variables[0],
            &it->first.m_variables[0]+it->first.m_variables.size());
    }
}

const iset & operator_connectivity::var_domain(unsigned variable) const
{
    return m_solver.var_domain(variable);
}
