// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010,2011,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010,2011,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <memory>
#include <algorithm>
#include <limits>
#include <constraint_connectivity_info.hpp>
#include <clause_propagator.hpp>
#include <arithmetic_propagators.hpp>
#include <comparison_propagators.hpp>
#include <neq_propagators.hpp>
#include <solver.hpp>
#include <XMLParser_constants.h>
#include "operator_connectivity.hpp"
#include "operator_definition_registry.hpp"

using namespace std;
using namespace hcsp;

namespace {
void link(solver & the_solver, constraint_propagator* constraint,
          unsigned out_variable,
          const unsigned *begin_vars, const unsigned *end_vars)
{
    constraint_connectivity_info & info=
        the_solver.add_constraint(
            unique_ptr<constraint_propagator>(constraint));
    info.reserve_arguments((end_vars-begin_vars)+1);
    info << out_variable;
    for (const unsigned *it= begin_vars; it!=end_vars; ++it)
        info << *it;
}

template <class Constraint_Propagator, typename... Args>
void create(solver & the_solver, unsigned out_variable,
            const unsigned *begin_vars, const unsigned *end_vars,
            const Args &... args)
{
    link(the_solver, new Constraint_Propagator(args...), out_variable,
         begin_vars, end_vars);
}

template <class Constraint_Propagator>
void create(solver & the_solver, unsigned out_variable, unsigned in_var)
{
    const unsigned vars[1]={in_var};
    create<Constraint_Propagator>(the_solver, out_variable, &vars[0], &vars[1]);
}
template <class Constraint_Propagator, class T1>
void create(solver & the_solver, unsigned out_variable, unsigned in_var,
            const T1 &a1)
{
    const unsigned vars[1]={in_var};
    create<Constraint_Propagator>(the_solver, out_variable, &vars[0], &vars[1],
                                  a1);
}

} // anonymous namespace

namespace {
//! An auxiliary class used as a base for a class representing an operator
//! which is differs from another operator only by argument order.
class reversed_definition : public operator_definition
{
  public:
    virtual iset::value_type eval(const iset::value_type *begin_val,
                                  const iset::value_type *end_val) const
    {
        ASSERT_SIZE(begin_val,end_val,2);
        const iset::value_type reverse[2]={begin_val[1], begin_val[0]};
        return reversed_op().eval(&reverse[0], &reverse[2]);
    }
    virtual void create_constraints(solver & , eval_type,
                                    unsigned,
                                    unsigned ,
                                    const unsigned *,
                                    const unsigned *) const
    {
        throw runtime_error(string(name())+"_definition::create_constraints "
                            "should not be called, use the reversed operator "
                            "instead ");
    }

    virtual unsigned symbolic_eval(operator_connectivity & state,
                                   eval_type type,
                                   const unsigned *begin_vars,
                                   const unsigned *end_vars,
                                   unsigned default_var) const
    {
        ASSERT_SIZE(begin_vars,end_vars,2);
        const unsigned reverse[2]={begin_vars[1], begin_vars[0]};
        return reversed_op().symbolic_eval(state, type, &reverse[0],
                                           &reverse[2], default_var);
    }
    virtual eval_type arg_eval_type(operator_connectivity & state,
                                    eval_type type,
                                    unsigned argument,
                                    const unsigned *begin_vars,
                                    const unsigned *end_vars,
                                    unsigned out_variable) const
    {
        ASSERT_SIZE(begin_vars,end_vars,2);
        const unsigned vars[2]={begin_vars[1], begin_vars[0]};
        return reversed_op().arg_eval_type(state, type, 1-argument, &vars[0],
                                           &vars[2], out_variable);
    }
    virtual double tightness() const { return reversed_op().tightness(); }
  protected:
    virtual const operator_definition & reversed_op() const = 0;
}; // class reversed_definition
} // end of anonymous namespace

namespace {
template <typename Constraint_Propagator>
class direct_operator_definition : public operator_definition
{
  public:
    virtual void create_constraints(solver & the_solver, eval_type,
                                    unsigned,
                                    unsigned out_variable,
                                    const unsigned *begin_vars,
                                    const unsigned *end_vars) const
    {
        unique_ptr<constraint_propagator> constraint(new Constraint_Propagator);
        constraint_connectivity_info &
            info= the_solver.add_constraint(move(constraint));
        info.reserve_arguments((end_vars-begin_vars)+1);
        info << out_variable;
        for (const unsigned *it= begin_vars; it!=end_vars; ++it)
            info << *it;
    }
};
} // end of anonymous namespace

namespace {
class div_definition : public direct_operator_definition<div_propagator>
{
  public:
    virtual iset::value_type eval(const iset::value_type *begin_val,
                                  const iset::value_type *end_val) const
    {
        ASSERT_SIZE(begin_val,end_val,2);
        return begin_val[1]==0 ? 0 : begin_val[0]/begin_val[1];
    }
    virtual unsigned symbolic_eval(operator_connectivity & state,
                                   eval_type,
                                   const unsigned *begin_vars,
                                   const unsigned *end_vars,
                                   unsigned default_var) const
    {
        ASSERT_SIZE(begin_vars,end_vars,2);
        return state.record_operator(*this, STRICT_EVAL,
                                     begin_vars, end_vars, default_var);
    }

    /*
    // This will have to be uncommented when non strict eval types
    // are implemented in symbolic_eval
    virtual eval_type arg_eval_type(operator_connectivity &,
                                    eval_type type,
                                    unsigned argument,
                                    const unsigned *, const unsigned *,
                                    unsigned) const
    {
        if (argument == 0)
            return type;
        HCSP_ASSERT(argument == 1);
        switch (type) {
          case BOTH_BOUNDS_EVAL: return BOTH_BOUNDS_EVAL;
          case UPPER_BOUND_EVAL: return LOWER_BOUND_EVAL;
          default:               return type;
        }
    }
    */
    virtual double tightness() const { return 0.995025; }
    virtual const char *name() const { return "div"; }
};
} // end of anonymous namespace

namespace {

class mod_definition : public direct_operator_definition<mod_propagator>
{
  public:
    virtual iset::value_type eval(const iset::value_type *begin_val,
                                  const iset::value_type *end_val) const
    {
        ASSERT_SIZE(begin_val,end_val,2);
        return begin_val[1]==0 ? 0 : begin_val[0] % begin_val[1];
    }
    virtual unsigned symbolic_eval(operator_connectivity & state,
                                   eval_type,
                                   const unsigned *begin_vars,
                                   const unsigned *end_vars,
                                   unsigned default_var) const
    {
        ASSERT_SIZE(begin_vars,end_vars,2);
        return state.record_operator(*this, STRICT_EVAL,
                                     begin_vars, end_vars, default_var);
    }
    virtual double tightness() const { return 0.995025; }
    virtual const char *name() const { return "mod"; }
}; // class mod_definition
} // end of anonymous namespace

namespace {
class sum_definition : public operator_definition
{
  public:
    virtual iset::value_type eval(const iset::value_type *begin_val,
                                  const iset::value_type *end_val) const
    {
        ASSERT_SIZE(begin_val,end_val,2);
        long long result= 1LL* begin_val[0] + begin_val[1];
        result= min<long long>(result, iset::MAX_DOMAIN);
        result= max<long long>(result, iset::MIN_DOMAIN);
        return result;
    }
    virtual void create_constraints(solver & the_solver, eval_type type,
                                    unsigned constraint_flags,
                                    unsigned out_variable,
                                    const unsigned *begin_vars,
                                    const unsigned *end_vars) const;
    
    virtual unsigned symbolic_eval(operator_connectivity & state,
                                   eval_type type,
                                   const unsigned *begin_vars,
                                   const unsigned *end_vars,
                                   unsigned default_var) const;

    void canonize_order(const operator_connectivity & ,
                        unsigned *begin_vars, unsigned *end_vars) const
    {
        ASSERT_SIZE(begin_vars,end_vars,2);
        if (begin_vars[0] > begin_vars[1])
            swap(begin_vars[0], begin_vars[1]);
    }
    
    virtual eval_type arg_eval_type(operator_connectivity &,
                                    eval_type type,
                                    unsigned,
                                    const unsigned *,
                                    const unsigned *,
                                    unsigned) const
    { return type; }
    virtual double tightness() const { return 0.996269; }
    virtual const char *name() const { return "sum"; }
  private:
    void create_sum_with_constant(solver & the_solver,
                                  eval_type type,
                                  unsigned out_variable,
                                  unsigned in_variable,
                                  iset::value_type in_value) const;
    
    static const operator_definition & get_sub()
    { return *operator_definition_registry::find(F_SUB); }
    static const operator_definition & get_mul()
    { return *operator_definition_registry::find(F_MUL); }
}; // class sum_definition
} // end of anonymous namespace

namespace {
class sub_definition : public operator_definition
{
  public:
    virtual iset::value_type eval(const iset::value_type *begin_val,
                                  const iset::value_type *end_val) const
    {
        ASSERT_SIZE(begin_val,end_val,2);
        long long result= 1LL* begin_val[0] - begin_val[1];
        result= min<long long>(result, iset::MAX_DOMAIN);
        result= max<long long>(result, iset::MIN_DOMAIN);
        return result;
    }
    virtual void create_constraints(solver & the_solver, eval_type type,
                                    unsigned constraint_flags,
                                    unsigned out_variable,
                                    const unsigned *begin_vars,
                                    const unsigned *end_vars) const;
    
    virtual unsigned symbolic_eval(operator_connectivity & state,
                                   eval_type type,
                                   const unsigned *begin_vars,
                                   const unsigned *end_vars,
                                   unsigned default_var) const
    {
        ASSERT_SIZE(begin_vars,end_vars,2);
        const iset & domain= state.var_domain(begin_vars[1]);
        if (domain.singleton() && domain.min() == 0)
            return begin_vars[0];
        if (begin_vars[0] == begin_vars[1])
            return state.get_const_variable(0);
        unsigned out_var= state.record_operator(*this, type, begin_vars,
                                                end_vars, default_var);
        
        // out <= x0 - x1 implies(same type) x1 <= x0 - out
        unsigned reverse[2]= { begin_vars[0], out_var};
        unsigned second_var= state.record_operator(
            *this, type, &reverse[0], &reverse[2], begin_vars[1]);
        HCSP_ASSERT(second_var == begin_vars[1]);
        (void)second_var;
        const operator_definition &
            add_operator = *operator_definition_registry::find(F_ADD);
        
        // out <= x0 - x1 implies(reverse type) x0 >= x1 + out
        unsigned add[2]= { out_var, begin_vars[1] };
        unsigned add_var= state.record_operator(
            add_operator, negate_type(type), &add[0], &add[2], begin_vars[0]);
        HCSP_ASSERT(add_var == begin_vars[0]);
        (void)add_var;
        return out_var;
    }
    virtual eval_type arg_eval_type(operator_connectivity &,
                                    eval_type type,
                                    unsigned argument,
                                    const unsigned *,
                                    const unsigned *,
                                    unsigned) const
    {
        if (argument == 0)
            return type;
        else 
            return negate_type(type);
    }
    virtual double tightness() const { return 0.996269; }
    virtual const char *name() const { return "sub"; }
}; // class sub_definition
} // end of anonymous namespace

namespace {
class mul_definition : public operator_definition
{
  public:
    virtual iset::value_type eval(const iset::value_type *begin_val,
                                  const iset::value_type *end_val) const
    {
        ASSERT_SIZE(begin_val,end_val,2);
        long long result= 1LL* begin_val[0] * begin_val[1];
        result= min<long long>(result, iset::MAX_DOMAIN);
        result= max<long long>(result, iset::MIN_DOMAIN);
        return result;
    }
    virtual void create_constraints(solver & the_solver, eval_type type,
                                    unsigned constraint_flags,
                                    unsigned out_variable,
                                    const unsigned *begin_vars,
                                    const unsigned *end_vars) const;
    
    virtual unsigned symbolic_eval(operator_connectivity & state,
                                   eval_type type,
                                   const unsigned *begin_vars,
                                   const unsigned *end_vars,
                                   unsigned default_var) const
    {
        ASSERT_SIZE(begin_vars,end_vars,2);
        for (unsigned i=0 ; i != 2 ; ++i) {
            const iset & domain= state.var_domain(begin_vars[i]);
            if (domain.singleton() && domain.min() == 0)
                return begin_vars[i];
        }
        for (unsigned i=0 ; i != 2 ; ++i) {
            const iset & domain= state.var_domain(begin_vars[i]);
            if (domain.singleton() && domain.min() == 1)
                return begin_vars[1-i];
        }
        return state.record_operator(*this, type,
                                     begin_vars, end_vars, default_var);
    }
    void canonize_order(const operator_connectivity & ,
                        unsigned *begin_vars, unsigned *end_vars) const
    {
        ASSERT_SIZE(begin_vars,end_vars,2);
        if (begin_vars[0] > begin_vars[1])
            swap(begin_vars[0], begin_vars[1]);
    }
    
    virtual eval_type arg_eval_type(operator_connectivity & state,
                                    eval_type type,
                                    unsigned argument,
                                    const unsigned *begin_vars,
                                    const unsigned *end_vars,
                                    unsigned) const
    {
        if (begin_vars == NULL)
            return STRICT_EVAL;
        ASSERT_SIZE(begin_vars,end_vars,2);
        HCSP_ASSERT(argument <= 1);
        if (begin_vars[1-argument] == -1U)
            return STRICT_EVAL;
        const iset & domain= state.var_domain(begin_vars[1-argument]);
        if (domain.min() >= 0)
            return type;
        else if (domain.max() > 0)
            return STRICT_EVAL;
        switch (type) {
          case LOWER_BOUND_EVAL: return UPPER_BOUND_EVAL;
          case UPPER_BOUND_EVAL: return LOWER_BOUND_EVAL;
          default: return type;
        }
    }
    virtual double tightness() const { return 0.999713; }
    virtual const char *name() const { return "mul"; }
}; // class mul_definition
} // anonymous namespace

namespace {
class abs_definition : public operator_definition
{
  public:
    virtual iset::value_type eval(const iset::value_type *begin_val,
                                  const iset::value_type *end_val) const
    {
        ASSERT_SIZE(begin_val,end_val,1);
        long long result= begin_val[0];
        if (result < 0)
            result *= -1;
        result= min<long long>(result, iset::MAX_DOMAIN);
        return result;
    }
    virtual void create_constraints(solver & the_solver, eval_type type,
                                    unsigned constraint_flags,
                                    unsigned out_variable,
                                    const unsigned *begin_vars,
                                    const unsigned *end_vars) const;
    
    virtual unsigned symbolic_eval(operator_connectivity & state,
                                   eval_type type,
                                   const unsigned *begin_vars,
                                   const unsigned *end_vars,
                                   unsigned default_var) const
    {
        ASSERT_SIZE(begin_vars,end_vars,1);
        const iset & domain= state.var_domain(begin_vars[0]);
        if (domain.min() >= 0)
            return begin_vars[0];
        else {
            return state.record_operator(
                *this, type, begin_vars, end_vars, default_var,
                iset(make_pair(0, iset::MAX_DOMAIN)));
        }
    }
    virtual eval_type arg_eval_type(operator_connectivity &,
                                    eval_type type,
                                    unsigned,
                                    const unsigned *,
                                    const unsigned *,
                                    unsigned) const
    {
        if (type == UPPER_BOUND_EVAL)
            return BOTH_BOUNDS_EVAL;
        else
            return STRICT_EVAL;
    }
    virtual double tightness() const { return 0.990099; }
    virtual const char *name() const { return "abs"; }
}; // class abs_definition

} // end of anonymous namespace

namespace {
class difference_ge_definition : public operator_definition
{
  public:
    virtual iset::value_type eval(const iset::value_type *begin_val,
                                  const iset::value_type *end_val) const
    {
        ASSERT_SIZE(begin_val,end_val,3);
        return 0LL+begin_val[0] - begin_val[1] >= begin_val[2] ;
    }
    virtual void create_constraints(solver & the_solver, eval_type type,
                                    unsigned constraint_flags,
                                    unsigned out_variable,
                                    const unsigned *begin_vars,
                                    const unsigned *end_vars) const;
    
    virtual unsigned symbolic_eval(operator_connectivity & state,
                                   eval_type type,
                                   const unsigned *begin_vars,
                                   const unsigned *end_vars,
                                   unsigned default_var) const
    {
        ASSERT_SIZE(begin_vars,end_vars,3);
        static const iset bool_domain(iset::value_range_type(0,1));
        const iset & k_domain= state.var_domain(begin_vars[2]);
        HCSP_ASSERT(k_domain.singleton());
        const iset::value_type k= k_domain.min();
        
        if (begin_vars[0] == begin_vars[1])
            return state.get_const_variable(0 >= k);

        const iset & domain0= state.var_domain(begin_vars[0]);
        const iset & domain1= state.var_domain(begin_vars[1]);
        if ( domain0.empty() || domain1.empty() ||
             0LL+domain0.min()-domain1.max() >= k)
            return state.get_const_variable(1);

        if ( 0LL+domain0.max()-domain1.min() < k )
            return state.get_const_variable(0);

        return state.record_operator(
            *this, type, begin_vars, end_vars, default_var, bool_domain);
    }

    virtual eval_type arg_eval_type(operator_connectivity & state,
                                    eval_type type,
                                    unsigned argument,
                                    const unsigned *,
                                    const unsigned *,
                                    unsigned out_var) const
    {
        // The operator sets the output to either 0 or 1, any other value
        // will be removed and hence has to be ignored.
        bool is_false= out_var != -1U && !state.var_domain(out_var).contains(1);
        bool is_true=  out_var != -1U && !state.var_domain(out_var).contains(0);
        // LOWER_BOUND_EVAL: out_var -> (x-y >= k) -->
        //                   !out_var || x >= k+y ;
        //                   !out_var || y <= x-k
        // UPPER_BOUND_EVAL: !out_var -> !(x-y >= k) -->
        //                   out_var || x-y < k -->
        //                   out_var || x < y+k;
        //                   out_var || y > x-k;
        if (type == UPPER_BOUND_EVAL || type == LOWER_BOUND_EVAL)
            return argument==0 ? type : negate_type(type);
        else if (is_false)
            // This is !(x-y>=k)  which is (x-y<k)
            return argument==0? UPPER_BOUND_EVAL : LOWER_BOUND_EVAL; 
        else if (is_true)
            // This is (x-y>=k)
            return argument==0? LOWER_BOUND_EVAL : UPPER_BOUND_EVAL;
        else
            return BOTH_BOUNDS_EVAL;
    }
    
    virtual double tightness() const { return 0.5; }
    virtual const char *name() const { return "diff_ge"; }
  private:
    // If the output variable is const then create the right constraint for
    // this case. If nothing else needs to be done return true.
    bool create_const_out_constraints(
        solver & the_solver, eval_type type, unsigned out_variable,
        unsigned x, unsigned y, iset::value_type k) const;
    
}; // class lt_definition
} // end of anonymous namespace

namespace {
class lt_definition : public operator_definition
{
  public:
    virtual iset::value_type eval(const iset::value_type *begin_val,
                                  const iset::value_type *end_val) const
    {
        ASSERT_SIZE(begin_val,end_val,2);
        return begin_val[0] < begin_val[1];
    }
    virtual void create_constraints(solver &, eval_type, unsigned, unsigned,
                                    const unsigned *, const unsigned *) const
    {
        throw runtime_error("lt_definition::create_constraints should not "
                            "be called, used difference_ge_definition instead");
    }        
    
    virtual unsigned symbolic_eval(operator_connectivity & state,
                                   eval_type type,
                                   const unsigned *begin_vars,
                                   const unsigned *end_vars,
                                   unsigned default_var) const
    {
        ASSERT_SIZE(begin_vars,end_vars,2);
        // x<y  --> y-x>0 --> y-x>=1
        unsigned vars[3]= {begin_vars[1], begin_vars[0],
                           state.get_const_variable(1)};
        return get_diff_ge().symbolic_eval(state, type, &vars[0], &vars[3],
                                           default_var);
    }

    virtual eval_type arg_eval_type(operator_connectivity & state,
                                    eval_type type,
                                    unsigned argument,
                                    const unsigned *begin_vars,
                                    const unsigned *end_vars,
                                    unsigned out_var) const
    {
        ASSERT_SIZE(begin_vars,end_vars,2);
        // x<y  --> y-x>0 --> y-x>=1
        unsigned vars[3]= {begin_vars[1], begin_vars[0],
                           state.get_const_variable(1)};
        unsigned reverse_argument= 1-argument;
        return get_diff_ge().arg_eval_type(state, type, reverse_argument,
                                           &vars[0], &vars[3], out_var);
    }
    
    virtual double tightness() const { return 0.5; }
    virtual const char *name() const { return "lt"; }
    static const operator_definition & get_diff_ge()
    { return *operator_definition_registry::find_predicate("difference_ge"); }

}; // class lt_definition
} // end of anonymous namespace

namespace {
class gt_definition : public reversed_definition
{
  public:
    virtual const char *name() const { return "gt"; }
  private:
    virtual const operator_definition & reversed_op() const
    { return *operator_definition_registry::find(F_LT); }
}; // class gt_definition
} // end of anonymous namespace

namespace {
class ge_definition : public operator_definition
{
  public:
    virtual iset::value_type eval(const iset::value_type *begin_val,
                                  const iset::value_type *end_val) const
    {
        ASSERT_SIZE(begin_val,end_val,2);
        return begin_val[0] >= begin_val[1];
    }    
    virtual void create_constraints(solver &, eval_type,
                                    unsigned,
                                    unsigned,
                                    const unsigned *,
                                    const unsigned *) const
    {
        throw runtime_error("ge_definition::create_constraints should not "
                            "be called, use difference_ge instead");
    }
    
    virtual unsigned symbolic_eval(operator_connectivity & state,
                                   eval_type type,
                                   const unsigned *begin_vars,
                                   const unsigned *end_vars,
                                   unsigned default_var) const
    {
        ASSERT_SIZE(begin_vars,end_vars,2);
        // x >= y --> x-y >= 0
        unsigned vars[3]= {begin_vars[0], begin_vars[1],
                           state.get_const_variable(0)};
        return get_diff_ge().symbolic_eval(state, type, &vars[0], &vars[3],
                                           default_var);
    }
    virtual eval_type arg_eval_type(operator_connectivity & state,
                                    eval_type type,
                                    unsigned argument,
                                    const unsigned *begin_vars,
                                    const unsigned *end_vars,
                                    unsigned out_var) const
    {
        ASSERT_SIZE(begin_vars,end_vars,2);
        unsigned vars[3]= {begin_vars[0], begin_vars[1],
                           state.get_const_variable(0)};
        return get_diff_ge().arg_eval_type(state, type, argument,
                                           &vars[0], &vars[3], out_var);
    }
    virtual double tightness() const { return get_diff_ge().tightness(); }
    virtual const char *name() const { return "ge"; }
  private:
    static const operator_definition & get_diff_ge()
    { return *operator_definition_registry::find_predicate("difference_ge"); }
}; // class ge_definition
} // end of anonymous namespace

namespace {
class le_definition : public reversed_definition
{
  public:
    virtual const char *name() const { return "le"; }
  private:
    virtual const operator_definition & reversed_op() const
    { return *operator_definition_registry::find(F_GE); }
}; // class le_definition
} // end of anonymous namespace

namespace {
class not_definition : public operator_definition
{
  private:
    enum { ORIGINAL_CONSTRAINT= 1 };
  public:
    virtual iset::value_type eval(const iset::value_type *begin_val,
                                  const iset::value_type *end_val) const
    {
        ASSERT_SIZE(begin_val,end_val,1);
        return !begin_val[0];
    }
    virtual void create_constraints(solver & the_solver, eval_type type,
                                    unsigned constraint_flags,
                                    unsigned out_variable,
                                    const unsigned *begin_vars,
                                    const unsigned *end_vars) const;
    
    virtual unsigned symbolic_eval(operator_connectivity & state,
                                   eval_type type,
                                   const unsigned *begin_vars,
                                   const unsigned *end_vars,
                                   unsigned default_var) const
    {
        ASSERT_SIZE(begin_vars,end_vars,1);
            
        const iset & domain= state.var_domain(begin_vars[0]);

        if ( !domain.contains(0) )
            return state.get_const_variable(0);
        
        if ( domain == iset(0) )
            return state.get_const_variable(1);

        // Consider the case when there are two constraints a=not(b) ;
        // a=not(c) If the pair a=not(b) , b=not(a) is recorded first then
        // a=not(c) will be recorded but c=not(a) will not because 'not(a)' is
        // already set to 'b'.
        //
        // To avoid this issue, we mark a=not(c) as ORIGINAL_CONSTRAINT
        static const iset all= iset().complement();
        unsigned result_var= state.record_operator(
            *this, type, begin_vars, end_vars, default_var, all,
            ORIGINAL_CONSTRAINT);

        unsigned reverse[1]={result_var};
            
        state.record_operator(
            *this, negate_type(type), &reverse[0], &reverse[1], begin_vars[0],
            iset(iset::value_range_type(0,1)));

        return result_var;
    }
    //! See operator_definition::reverse_eval
    virtual pair<iset::value_type, bool>
    reverse_eval(const iset & out_domain, unsigned /*argument*/) const
    {
        if (out_domain == iset(0))
            return make_pair(1, true);
        if (out_domain == iset(1))
            return make_pair(0, true);
        return make_pair(0, false);
    }
    virtual eval_type arg_eval_type(operator_connectivity &,
                                    eval_type type,
                                    unsigned,
                                    const unsigned *,
                                    const unsigned *,
                                    unsigned) const
    {
        return negate_type(type);
    }
    virtual double tightness() const { return 0.50; }
    virtual const char *name() const { return "not"; }
}; // class not_definition
} // end of anonymous namespace

// virtual
eval_type operator_definition::arg_eval_type(
    operator_connectivity &, eval_type, unsigned,
    const unsigned *, const unsigned *, unsigned) const
{
    return STRICT_EVAL;
}

// virtual
void operator_definition::canonize_order(const operator_connectivity & ,
                                         unsigned *, unsigned *) const
{
}

// static
const operator_definition *operator_definition_registry::find(unsigned type_id)
{
    const container_t & definitions= get().m_definitions;
    container_t::const_iterator it;
    it= lower_bound(definitions.begin(), definitions.end(),
                    make_pair(type_id, (operator_definition*)NULL) );
    if (it == definitions.end() || it->first != type_id)
        return NULL;
    else
        return it->second;
};

// static
const operator_definition_registry & operator_definition_registry::get()
{
    static const operator_definition_registry operators;
    return operators;
}

// static
const operator_definition *operator_definition_registry::find_predicate(
    const string & name)
{
    const predicate_container_t & definitions= get().m_predicate_definitions;
    predicate_container_t::const_iterator it;
    it= lower_bound(definitions.begin(), definitions.end(),
                    make_pair(name, (operator_definition*)NULL) );
    if (it == definitions.end() || it->first != name)
        return NULL;
    else
        return it->second;
};

operator_definition_registry::operator_definition_registry()
{
    m_definitions.push_back(make_pair(F_NOT, new not_definition));
    m_definitions.push_back(make_pair(F_DIV, new div_definition));
    m_definitions.push_back(make_pair(F_MOD, new mod_definition));
    m_definitions.push_back(make_pair(F_MUL, new mul_definition));
    m_definitions.push_back(make_pair(F_ADD, new sum_definition));
    m_definitions.push_back(make_pair(F_SUB, new sub_definition));
    m_definitions.push_back(make_pair(F_ABS, new abs_definition));
    m_definitions.push_back(make_pair(F_LE,  new le_definition));
    m_definitions.push_back(make_pair(F_GE,  new ge_definition));
    m_definitions.push_back(make_pair(F_LT,  new lt_definition));
    m_definitions.push_back(make_pair(F_GT,  new gt_definition));
    sort(m_definitions.begin(), m_definitions.end());

    m_predicate_definitions.push_back(make_pair(string("difference_ge"),
                                                new difference_ge_definition));
    sort(m_predicate_definitions.begin(), m_predicate_definitions.end());
}


operator_definition_registry::~operator_definition_registry()
{
    while (!m_definitions.empty()) {
        delete m_definitions.back().second;
        m_definitions.back().second= NULL;
        m_definitions.pop_back();
    }
    while (!m_predicate_definitions.empty()) {
        delete m_predicate_definitions.back().second;
        m_predicate_definitions.back().second= NULL;
        m_predicate_definitions.pop_back();
    }
}

// virtual
void mul_definition::create_constraints(solver & the_solver, eval_type type,
                                        unsigned ,
                                        unsigned out_variable,
                                        const unsigned *begin_vars,
                                        const unsigned *end_vars) const
{
    ASSERT_SIZE(begin_vars, end_vars, 2);
    if (begin_vars[0] == begin_vars[1]) {
        unsigned pow_vars[2] = {
            begin_vars[0],
            the_solver.singleton_var(2)};
        // FIXME: Need to register pow for hashing purposes.
        create<pow_propagator>(the_solver, out_variable, &pow_vars[0],
                               &pow_vars[0]+2);
        return;
    }
            
    if (type == STRICT_EVAL)
        create<mul_truncate_propagator>(the_solver, out_variable,
                                        begin_vars, end_vars);
    else {
        if ( (type & LOWER_BOUND_EVAL) == LOWER_BOUND_EVAL)
            create<less_equal_mul_propagator>(the_solver, out_variable,
                                              begin_vars, end_vars);
            
        if ( (type & UPPER_BOUND_EVAL) == UPPER_BOUND_EVAL)
            create<greater_equal_mul_propagator>(the_solver, out_variable,
                                                 begin_vars, end_vars);
    }
}

// virtual
unsigned sum_definition::symbolic_eval(operator_connectivity & state,
                                       eval_type type,
                                       const unsigned *begin_vars,
                                       const unsigned *end_vars,
                                       unsigned default_var) const
{
    ASSERT_SIZE(begin_vars,end_vars,2);
    for (unsigned i=0 ; i != 2 ; ++i) {
        const iset & domain= state.var_domain(begin_vars[i]);
        if (domain.singleton() && domain.min() == 0)
            return begin_vars[1-i];
    }
        
    const iset & domain0= state.var_domain(begin_vars[0]);
    const iset & domain1= state.var_domain(begin_vars[1]);
    if (domain0.singleton() && domain1.singleton()) {
        long long result= 0LL+domain1.min()+domain0.min();
        if (iset::MIN_DOMAIN >= result && result <= iset::MAX_DOMAIN)
            return state.get_const_variable(result);
    }

    if (begin_vars[0] == begin_vars[1]) {
        unsigned mul_vars[]= { begin_vars[0], state.get_const_variable(2)};
        return state.record_operator(get_mul(), type,
                                     &mul_vars[0], &mul_vars[2], default_var);
    }
    unsigned out_var= state.record_operator(*this, type, begin_vars,
                                            end_vars, default_var);
    unsigned sub[2]= { out_var, begin_vars[0]};
    unsigned sub_var= state.record_operator(
        get_sub(), type, &sub[0], &sub[2], begin_vars[1]);
    HCSP_ASSERT(sub_var == begin_vars[1]);
    sub[1]= begin_vars[1]; 
    sub_var= state.record_operator(get_sub(), type, &sub[0], &sub[2],
                                   begin_vars[0]);
    HCSP_ASSERT(sub_var == begin_vars[0]);
    (void)sub_var;
    return out_var;
}

void sum_definition::create_sum_with_constant(solver & the_solver,
                                              eval_type type,
                                              unsigned out_variable,
                                              unsigned in_variable,
                                              iset::value_type in_value) const
{
    if (in_value == 0) {
        create<equal_propagator>(the_solver, out_variable, in_variable);
        return;
    }

    if (type == STRICT_EVAL) {
        unsigned vars[2]= {in_variable, the_solver.singleton_var(in_value) };
        create<sum_truncate_propagator>(the_solver, out_variable,
                                        &vars[0], &vars[2]);
        return;
    }
    if ( (type & LOWER_BOUND_EVAL) == LOWER_BOUND_EVAL) {
        if (in_value == iset::MIN_DOMAIN) {
            unsigned vars[2]= {in_variable, the_solver.singleton_var(in_value)};
            create<less_equal_sum_propagator>(the_solver, out_variable,
                                              &vars[0], &vars[2]);
        } else {
            // out <= in + k --> in-out >= -k
            unsigned vars[1]= {out_variable};
            create<difference_ge_propagator>(the_solver, in_variable,
                                             &vars[0], &vars[1], -in_value);
        }
    }
    if ( (type & UPPER_BOUND_EVAL) == UPPER_BOUND_EVAL) {
        // out >= in + k --> out-in >= k
        create<difference_ge_propagator>(the_solver, out_variable,
                                         in_variable, in_value);
    }
}

// virtual
void sum_definition::create_constraints(solver & the_solver, eval_type type,
                                        unsigned,
                                        unsigned out_variable,
                                        const unsigned *begin_vars,
                                        const unsigned *end_vars) const
{
    ASSERT_SIZE(begin_vars, end_vars, 2);
    if (begin_vars[0] == begin_vars[1]) {
        unsigned mul_vars[2] = {
            the_solver.singleton_var(2),
            begin_vars[0]
        };
        // FIXME: Need to register mul for hashing purposes.
        create<mul_truncate_propagator>(the_solver, out_variable, &mul_vars[0],
                                        &mul_vars[0]+2);
        return;
    }

    typedef linear_combination_ge_propagator::coefficient_type coeff_type;
    const iset & domain0= the_solver.var_domain(begin_vars[0]);
    const iset & domain1= the_solver.var_domain(begin_vars[1]);
    if (domain0.singleton())
        create_sum_with_constant(the_solver, type, out_variable,
                                 begin_vars[1], domain0.min());
    else if ( domain1.singleton()) 
        create_sum_with_constant(the_solver, type, out_variable,
                                 begin_vars[0], domain1.min());
    else if (type == STRICT_EVAL)
        create<sum_truncate_propagator>(the_solver, out_variable, begin_vars,
                                        end_vars);
    else {
        if ( (type & LOWER_BOUND_EVAL) == LOWER_BOUND_EVAL) {
            if (domain1.singleton() && domain1.min() != iset::MIN_DOMAIN) {
                // out <= in + k --> in-out >= -k
                create<difference_ge_propagator>(the_solver, begin_vars[0],
                                                 out_variable,
                                                 -domain1.min());
            }
            else // 0 <= -v0+v1+v2
                create<linear_combination_ge_propagator>(
                    the_solver, out_variable, begin_vars, end_vars,
                    0, std::initializer_list<coeff_type>{-1,1,1});
        }
            
        if ( (type & UPPER_BOUND_EVAL) == UPPER_BOUND_EVAL) {
            if (domain1.singleton()) {
                // out >= in + k --> out-in >= k
                create<difference_ge_propagator>(the_solver, out_variable,
                                                 begin_vars[0], domain1.min());
            } else // v0-v1-v2 >= 0
                create<linear_combination_ge_propagator>(
                    the_solver, out_variable, begin_vars, end_vars,
                    0, std::initializer_list<coeff_type>{1,-1,-1});
        }
    }
}

// virtual
void sub_definition::create_constraints(solver & , eval_type,
                                        unsigned,
                                        unsigned ,
                                        const unsigned *,
                                        const unsigned *) const
{
    // no need to insert, the sum_definition will do it.
    return;
}

// virtual
void abs_definition::create_constraints(solver & the_solver, eval_type type,
                                        unsigned,
                                        unsigned out_variable,
                                        const unsigned *begin_vars,
                                        const unsigned *end_vars) const
{
    ASSERT_SIZE(begin_vars, end_vars, 1);
    if ( out_variable == begin_vars[0] ) {
        iset domain= the_solver.var_domain(out_variable);
        domain.intersect(make_pair(0, iset::MAX_DOMAIN));
        the_solver.set_var_domain(out_variable, move(domain));
        return;
    }

    if (type == STRICT_EVAL)
        create<abs_propagator>(the_solver, out_variable, begin_vars, end_vars);
    else {
        if ( (type & LOWER_BOUND_EVAL) == LOWER_BOUND_EVAL)
            create<less_equal_abs_propagator>(the_solver, out_variable,
                                              begin_vars, end_vars);
            
        if ( (type & UPPER_BOUND_EVAL) == UPPER_BOUND_EVAL)
            create<greater_equal_abs_propagator>(the_solver, out_variable,
                                                 begin_vars, end_vars);
    }
}

bool difference_ge_definition::create_const_out_constraints(
    solver & the_solver, eval_type type, unsigned out_variable,
    unsigned x, unsigned y, iset::value_type k) const
{
    const iset & out_domain= the_solver.var_domain(out_variable);
    if (!out_domain.contains(1) && out_domain.contains(0)) {
        if (type != IMPLIED_EVAL) { // otherwise: "0 --> x-y >= k" (vacuous)
            // x-y < k --> y-x >= -k
            if (k == numeric_limits<iset::value_type>::min())
                throw runtime_error("Unsupported: x-y < k when k is INT_MIN");
            create<difference_ge_propagator>(the_solver, y, x, -k);
        }
        return true;
    }

    if (!out_domain.contains(0) && out_domain.contains(1)) {
        if (type != IMPLIES_EVAL)  // otherwise: "1 <-- x-y >= k" (vacuous)
            create<difference_ge_propagator>(the_solver, x, y, k);
        return true;
    }
    return false;
}

// virtual
void difference_ge_definition::create_constraints(
    solver & the_solver, eval_type type, unsigned,
    unsigned out_variable, const unsigned *begin_vars,
    const unsigned *end_vars) const
{
    ASSERT_SIZE(begin_vars, end_vars, 3);
    unsigned x= begin_vars[0], y=begin_vars[1];
    const iset & domain0= the_solver.var_domain(x);
    const iset & domain1= the_solver.var_domain(y);
    const iset & domain_k= the_solver.var_domain(begin_vars[2]);
    if (domain_k.empty() || domain0.empty() || domain1.empty())
        return;
    HCSP_ASSERT(domain_k.singleton());
    iset::value_type k= domain_k.min();
    if ( 0LL+domain0.min()-domain1.max() >= k )   { // Case: always true
        if (type != IMPLIED_EVAL) {// otherwise: "out --> true" (vacuous).
            iset out_domain = the_solver.var_domain(out_variable);
            out_domain.intersect(iset(1));
            the_solver.set_var_domain(out_variable, move(out_domain));
        }
        return;
    }

    if ( (begin_vars[0] == begin_vars[1] && k>0) ||
         0LL+domain0.max()-domain1.min() < k) { // Case: always false
        if (type != IMPLIES_EVAL) { // otherwise: "out <-- false" (vacuous).
            iset out_domain = the_solver.var_domain(out_variable);
            out_domain.intersect(iset(0));
            the_solver.set_var_domain(out_variable, move(out_domain));
        }
        return;
    }
    if (create_const_out_constraints(the_solver, type, out_variable, x, y, k))
        return;
    if ((type & IMPLIED_EVAL) == IMPLIED_EVAL)
        create<literal_or_difference_ge_propagator>(
            the_solver, out_variable, &begin_vars[0], &begin_vars[2], iset(0),
            k);
    if ((type & IMPLIES_EVAL) == IMPLIES_EVAL) {
        // This means: out <--- (x-y>=k), i.e., out || (x-y<k),
        // or (out || y-x > -k) and finally (out || y-x >= 1-k).
        unsigned vars[2] = {y,x};
        create<literal_or_difference_ge_propagator>(
            the_solver, out_variable, &vars[0], &vars[2], iset(1), 1-k);
    }
}

//virtual
void not_definition::create_constraints(solver & the_solver, eval_type type,
                                        unsigned constraint_flags,
                                        unsigned out_variable,
                                        const unsigned *begin_vars,
                                        const unsigned *end_vars) const
{
    ASSERT_SIZE(begin_vars,end_vars,1);
    unsigned in_variable= begin_vars[0];
    // Because the operator is added twice to the registry,
    // add only the constraint that generates this order.
    if (constraint_flags != ORIGINAL_CONSTRAINT)
        return;
    if (type == UPPER_BOUND_EVAL) { // out >= not(v0), i.e., out || v0
        // add: not(out) || not(v0)
        clause_propagator *p_clause= new clause_propagator;
        unique_ptr<constraint_propagator> constraint(p_clause);
        vector<iset> & literals= p_clause->update_literals();
        literals.resize(2);
        literals[0].assign(1);
        literals[1].assign(1);
        the_solver.add_constraint(move(constraint))
            << out_variable <<in_variable;
    } else if (type==LOWER_BOUND_EVAL) {// out <= not(v0), i.e., out->not v0
        clause_propagator *p_clause= new clause_propagator;
        unique_ptr<constraint_propagator> constraint(p_clause);
        vector<iset> & literals= p_clause->update_literals();
        literals.resize(2);
        literals[0].assign(0);
        literals[1].assign(0);
        the_solver.add_constraint(move(constraint))
            << out_variable << in_variable;
    } else
        create<neq_propagator>(the_solver, out_variable, begin_vars, end_vars);
}
