// -*- mode:C++;c-basic-offset:4 -*- 
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2012,2014, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2012,2014, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <typeinfo>
#include <algorithm>
#include <memory>
#include <parser_context.hpp>
#include <clause_propagator.hpp>
#include <comparison_propagators.hpp>
#include <solver.hpp>

using namespace std;
using namespace CSPXMLParser;

namespace hcsp {

unsigned parser_context::var_or_value::to_id(parser_context & context) const
{
    if (m_var_id == -1U)
        return context.singleton_var(m_value);
    else
        return m_var_id;
}
parser_context::parser_context(
    solver & solver, const scanner_methods & methods,
    hcsp::table_propagator_base::memory_manager & manager,
    std::istream* is, const std::string & filename)
    : m_scanner(NULL),
      m_is(is),
      m_filename(filename),
      m_direction(NO_OBJECTIVE),
      m_objective_var(-1U),
      m_AST_builder(solver),
      m_solver(solver),
      m_methods(methods),
      m_manager(manager)
{
    *is >> std::noskipws;
    m_methods.start_scanner(*this);
}

parser_context::~parser_context()
{
    m_methods.destroy_scanner(*this);
}

ostream & operator<<(ostream &out, const parser_context::position_point &point)
{
    out << *point.m_file_name << ":" << point.m_line_number << ":"
        << point.m_column_number;
    return out;
}

void parser_context::connect_table_propagator(
    const vector<unsigned> & vars,
    const vector<vector<iset::value_type> > & tuples)
{
    connect_constraint(
        vars, make_unique<support_table_propagator>(m_manager, tuples));
}
    

void parser_context::instantiate_element_constraint(
    std::ostream & debug_stream, const position_point & point,
    unsigned index, iset::value_type first_pos, const var_or_value_list & array,
    unsigned result)
{
    iset index_domain= domain(index);
    index_domain.intersect(make_pair(first_pos, first_pos+array.size()-1));
    vector<unsigned> vars{index};
    for (int i = index_domain.min() ; i <= index_domain.max(); ++i) {
        unsigned elem_id= array[i-first_pos].m_var_id;
        if (elem_id == -1U)
            elem_id= singleton_var(array[i-first_pos].m_value);
        vars.push_back(elem_id);
    }
    vars.push_back(result);
    if (index_domain.singleton()) {
        if (report_domain_deduction())
            debug_stream << "% " << point
                         << ": the constraint is equivalent with equality\n";
        equate_variables(vars[1],vars[2]);
        return;
    }
    make_unique_vars(vars);
    connect_constraint(
        vars, make_unique<element_propagator>(index_domain.min()));
}

void parser_context::make_unique_vars(vector<unsigned> & vars) 
{
    iset singleton_vars;
    for (unsigned & var : vars) {
        const iset & var_domain = domain(var);
        if (var_domain.singleton()) {
            if (singleton_vars.contains(var))
                var = singleton_var(var_domain.min(), true);
            else
                singleton_vars.insert(var);
        }
    }
}

unsigned parser_context::singleton_var(iset::value_type value, bool force_new)
{
    if (!force_new)
        return m_solver.singleton_var(value);
    ostringstream name;
    name << "V" << m_solver.count_variables() << "_constant_";
    if (value >= 0)
        name << value;
    else
        name << "minus_" << -int64_t(value);
    return m_solver.add_var(iset(value), name.str().c_str());
}

bool parser_context::define_variables(ostream & debug_stream,
                                      const iset & in_domain,
                                      const parse_var_list & var_list,
                                      const var_or_value * literal,
                                      const annotation_list * annotations,
                                      element_type type)
{
    for (auto & elem: var_list)
        if (!define_variable(debug_stream, in_domain, elem, literal,
                             annotations, type))
            return false;
    return true;
}

void parser_context::intersect_annotations(symbol_info &a, symbol_info &b) const
{
    a.m_is_intermediate= a.m_is_intermediate && b.m_is_intermediate;
    b.m_is_intermediate= a.m_is_intermediate;

    auto a_found= find_if(a.m_annotations.begin(), a.m_annotations.end(),
                          [](const annotation_type & a_annotation) {
                              return (a_annotation.m_name == "is_defined_var");
                          });
    auto b_found= find_if(b.m_annotations.begin(), b.m_annotations.end(),
                          [](const annotation_type & b_annotation) {
                              return (b_annotation.m_name ==
                                      "is_defined_var");
                          });

    if ( a_found != a.m_annotations.end() && b_found == b.m_annotations.end())
        a.m_annotations.erase(
            remove_if(a.m_annotations.begin(), a.m_annotations.end(),
                      [](const annotation_type &a_annotation) {
                          return a_annotation.m_name == "is_defined_var";
                      }),
            a.m_annotations.end());

    if ( b_found != b.m_annotations.end() && a_found == a.m_annotations.end())
        b.m_annotations.erase(
            remove_if(b.m_annotations.begin(), b.m_annotations.end(),
                      [](const annotation_type &b_annotation) {
                          return b_annotation.m_name == "is_defined_var";
                      }),
            b.m_annotations.end());
}

bool parser_context::define_variable(ostream & debug_stream,
                                     iset in_domain,
                                     const parse_var & var,
                                     const var_or_value* literal,
                                     const annotation_list * annotations,
                                     element_type type)
{
    //! \todo This function is way to long, it requires splitting.
    if (type & SET_TYPE) {
        debug_stream << var.second << ": set variables are unsupported.\n";
        return false;
    }
    auto array_it= m_name_to_symbol.find(var.first);
    if (array_it != m_name_to_symbol.end()) {
        debug_stream << var.second << ": redefinition of " << var.first << endl;
        debug_stream << array_it->second.m_location << ": ";
        array_it->second.print(debug_stream, var.first);
        debug_stream << " was previously defined here" << endl;
        return false;
    }
    
    auto it= m_name_to_symbol.insert(make_pair(var.first, symbol_info())).first;
    var_or_value id;
    symbol_info * literal_symbol= nullptr;
    if (literal) {
        id= *literal;
        if (id.m_var_id != -1U) {
            in_domain.intersect(m_solver.var_domain(id.m_var_id));
            m_solver.set_var_domain(id.m_var_id, move(in_domain));
            auto literal_symbol_it = m_name_to_symbol.end();
            if (id.m_var_id < m_id_to_name.size())
                literal_symbol_it=
                    m_name_to_symbol.find(m_id_to_name[id.m_var_id]);
            if (literal_symbol_it != m_name_to_symbol.end())
                literal_symbol = &literal_symbol_it->second;
        }
    } else {
        id.m_var_id= m_solver.add_var(move(in_domain), var.first.c_str());
        if (m_id_to_name.size() <= id.m_var_id) {
            m_id_to_name.resize(id.m_var_id*2+2);
        }
        m_id_to_name[id.m_var_id]= var.first;
        id.m_value = 0;  // To overcome gcc's "uninitialized" warning
    }
    it->second.m_var_or_set_list.push_back(id);
    it->second.m_is_array= false;
    bool is_defined_var= false, var_is_introduced = false;
    if (annotations) {
        for (const auto & annotation: *annotations) {
            if (annotation.m_name == "is_defined_var")
                is_defined_var= true;
            else if (annotation.m_name == "var_is_introduced")
                var_is_introduced = true;
        }
    }

    it->second.m_is_intermediate = is_defined_var && var_is_introduced;
    it->second.m_location = var.second;
    it->second.m_type = type;
    if (annotations)
        it->second.m_annotations= *annotations;
    if (literal_symbol)
        intersect_annotations(it->second, *literal_symbol);
    return true;
}

bool parser_context::define_arrays(ostream & debug_stream,
                                   const iset & in_domain,
                                   unsigned dimension,
                                   const parse_var_list & var_list,
                                   const var_or_value_list* array_literal,
                                   const annotation_list * annotations,
                                   element_type type)
{
    for (auto & elem: var_list)
        if (!define_array(debug_stream, in_domain, dimension, elem,
                          array_literal, annotations, type))
            return false;
    return true;
}

void parser_context::symbol_info::print(ostream & out, const string &name) const
{
    out << name;
    if (m_is_array)
        out << "[1.." << m_var_or_set_list.size() << "]";
}

bool parser_context::define_array(ostream & debug_stream,
                                  iset in_domain,
                                  unsigned dimension,
                                  const parse_var & var,
                                  const var_or_value_list * array_literal,
                                  const annotation_list * annotations,
                                  element_type type)
{
    if (type & SET_TYPE) {
        debug_stream << var.second << ": set variables are unsupported.\n";
        return false;
    }
    auto var_it = m_name_to_symbol.find(var.first);
    if (var_it != m_name_to_symbol.end()) {
        debug_stream << var.second << ": redefinition of " << var.first
                     << "[1.." << dimension << "]\n";
        debug_stream << var_it->second.m_location << ": ";
        var_it->second.print(debug_stream, var.first);
        debug_stream << " was previously defined here" << endl;
        return false;
    }
    symbol_info array;
    array.m_is_array= true;
    array.m_is_intermediate = false;
    array.m_location = var.second;
    array.m_type = type;
    if (annotations)
        array.m_annotations = *annotations;
    auto array_it = m_name_to_symbol.insert(make_pair(var.first, array)).first;
    for (unsigned i=0 ; i < dimension ; ++i) {
        ostringstream name;
        name << var.first << "[" << i << "]";
        var_or_value id;
        if (array_literal && i < array_literal->size()) {
            id= (*array_literal)[i];
            if (id.m_var_id != -1U) {
                iset var_domain = in_domain;
                var_domain.intersect(m_solver.var_domain(id.m_var_id));
                m_solver.set_var_domain(id.m_var_id, move(var_domain));
            }
        } else {
            id.m_var_id= m_solver.add_var(in_domain, name.str().c_str());
            id.m_value = 0;  // To overcome gcc's "uninitialized" warning
        }
        array_it->second.m_var_or_set_list.push_back(id);
    }
    
    return true;
}

bool parser_context::define_scalar_constant(ostream & debug_stream,
                                            const parse_var & var,
                                            int32_t value)
{
    auto array_it= m_name_to_symbol.find(var.first);
    if (array_it != m_name_to_symbol.end()) {
        debug_stream << var.second << ": redefinition of " << var.first << endl;
        debug_stream << array_it->second.m_location << ": ";
        array_it->second.print(debug_stream, var.first);
        debug_stream << " was previously defined here" << endl;
        return false;
    }
    
    auto it= m_name_to_symbol.insert(make_pair(var.first, symbol_info())).first;
    it->second.m_is_array= false;
    it->second.m_is_intermediate = false;
    it->second.m_location = var.second;
    it->second.m_type = INT_TYPE;
    var_or_value var_value;
    var_value.m_value= value;
    var_value.m_var_id = -1U;
    it->second.m_var_or_set_list.push_back(var_value);
    return true;
}

bool parser_context::define_array_constant(ostream & debug_stream,
                                           const parse_var & var,
                                           const vector<int32_t> & values)
{
    auto var_it = m_name_to_symbol.find(var.first);
    if (var_it != m_name_to_symbol.end()) {
        debug_stream << var.second << ": redefinition of " << var.first
                     << "[1.." << values.size() << "]\n";
        debug_stream << var_it->second.m_location << ": ";
        var_it->second.print(debug_stream, var.first);
        debug_stream << " was previously defined here" << endl;
        return false;
    }
    auto it= m_name_to_symbol.insert(make_pair(var.first, symbol_info())).first;
    symbol_info & array= it->second;
    array.m_is_array= true;
    array.m_is_intermediate= false;
    array.m_location = var.second;
    array.m_type = INT_TYPE;
    for (auto in_entry : values) {
        var_or_value out_entry;
        out_entry.m_value= in_entry;
        out_entry.m_var_id= -1U;
        array.m_var_or_set_list.push_back(out_entry);
    }

    return true;
}

bool parser_context::define_array_set_constant(ostream & debug_stream,
                                               const parse_var & var,
                                               const vector<iset> & sets)
{
    auto var_it = m_name_to_symbol.find(var.first);
    if (var_it != m_name_to_symbol.end()) {
        debug_stream << var.second << ": redefinition of " << var.first
                     << "[1.." << sets.size() << "]\n";
        debug_stream << var_it->second.m_location << ": ";
        var_it->second.print(debug_stream, var.first);
        debug_stream << " was previously defined here" << endl;
        return false;
    }
    auto it= m_name_to_symbol.insert(make_pair(var.first, symbol_info())).first;
    symbol_info & array= it->second;
    array.m_is_array= true;
    array.m_is_intermediate = false;
    array.m_location = var.second;
    array.m_type = INT_SET_TYPE;
    for (const auto & in_entry : sets) {
        var_or_set out_entry;
        out_entry.m_set= in_entry;
        out_entry.m_var_id= -1U;
        array.m_var_or_set_list.push_back(out_entry);
    }

    return true;
}

void parser_context::update_variable_identity_function()
{
    unsigned num_variables= m_solver.count_variables();
    if ( m_variable_identity_function.size() < num_variables) {
        unsigned i= m_variable_identity_function.size();
        m_variable_identity_function.resize(num_variables);
        for ( ; i < m_variable_identity_function.size() ; ++i)
            m_variable_identity_function[i]=i;
    }

}

void parser_context::connect_expr(unsigned out_variable, const AST &formula,
                                  usage_type used_as)
{
    update_variable_identity_function();
    m_AST_builder.connect_expr(m_variable_identity_function, out_variable,
                               formula, used_as);
}
unsigned parser_context::make_aliased_var(unsigned var)
{
    stringstream name;
    name << m_solver.var_printer(var)
         << ".alias" << m_solver.count_variables();
    unsigned aliased=
        m_solver.add_var(m_solver.var_domain(var), name.str().c_str());
    m_solver.add_constraint(make_unique<equal_propagator>())
        << aliased << var;
    return aliased;
}

void parser_context::connect_constraint(
    const std::vector<unsigned> & vars,
    std::unique_ptr<constraint_propagator> constraint)
{
    vector<pair<unsigned, unsigned> > var_order;
    var_order.reserve(vars.size());
    for (unsigned i=0 ; i != vars.size() ; ++i)
        var_order.push_back(make_pair(vars[i], i));
    sort(var_order.begin(), var_order.end());
    vector<unsigned> unique_vars(vars);
    for (unsigned i=1 ; i < var_order.size() ; ++i) {
        unsigned var = var_order[i].first;
        if (var == var_order[i-1].first) {
            unique_vars[var_order[i].second]= make_aliased_var(var);
        }            
    }
    constraint_connectivity_info &
        info= m_solver.add_constraint(move(constraint));
    info.reserve_arguments(vars.size());
    for (unsigned var : unique_vars)
        info << var;
}

parser_context::symbol_info *
parser_context::symbol_from_annotation(ostream & debug_stream,
                                       const position_point & point,
                                       const argument_type & arg)
{
    const annotation_argument * annotation_arg=
        dynamic_cast<const annotation_argument*>(&arg);
    if (annotation_arg) {
        const string & name = annotation_arg->m_annotation.m_name;
        auto it = m_name_to_symbol.find(name);
        if (m_name_to_symbol.end() != it)
            return &it->second;
        debug_stream << "% " << point << ": unknown defines_var("
                     << name << ")\n";
    }
    return nullptr;
}

vector<unsigned> parser_context::collect_vars(const argument_list & arguments)
{
    vector<unsigned> ret;
    for (const auto & p_arg: arguments) {
        if (!p_arg)
            continue;
        auto array = dynamic_cast<const array_argument*>(p_arg.get());
        if (array) {
            for (const var_or_value & base : array->m_array) {
                ret.push_back(base.to_id(*this));
            }
        } else {
            auto var = dynamic_cast<const var_argument*>(p_arg.get());
            if (var)
                ret.push_back(var->m_variable.to_id(*this));
        }
    }
    return ret;
}

bool parser_context::has_defined_var(const argument_list & arguments) const
{
    for (const auto & ptr_argument : arguments) {
        auto var_arg = dynamic_cast<const var_argument*>(ptr_argument.get());
        auto array_arg = dynamic_cast<const array_argument*>(ptr_argument.get());
        if (var_arg && is_intermediate_var(var_arg->m_variable.m_var_id))
            return true;
        if (array_arg) {
            for (const auto & entry : array_arg->m_array) {
                if (is_intermediate_var(entry.m_var_id))
                    return true;
            }
        }
    }
    return false;
}

bool parser_context::is_intermediate_var(unsigned id) const
{
    if (id == -1U || m_id_to_name.size() <= id)
        return false;
    const string & name= m_id_to_name[id];
    if (name.empty())
        return false;
    auto it= m_name_to_symbol.find(name);
    if (it == m_name_to_symbol.end())
        return false;
    return it->second.m_is_intermediate;
}

bool parser_context::connect_constraint(ostream & debug_stream,
                                        const position_point & point,
                                        const std::string & name,
                                        const argument_list & arguments,
                                        const annotation_list *annotations)
{
    auto it= m_constraint_types.find(name);
    if (it == m_constraint_types.end()) {
        debug_stream << point << " constraint " << name << " is unknown\n";
        return false;
    }
    symbol_info * defines_var = nullptr;
    if (annotations) {
        for (const annotation_type & annotation:  *annotations) {
            if (annotation.m_name != "defines_var")
                continue;
            if (!annotation.m_arguments && annotation.m_arguments->size() != 1)
                continue;
            annotation.m_arguments->at(0).get();
            const argument_type *arg = annotation.m_arguments->at(0).get();
            if (arg)
                defines_var= symbol_from_annotation(debug_stream, point, *arg);
        }
    }
    if (defines_var && defines_var->m_is_intermediate) {
        defines_var->m_defined_by = make_shared<postponed_constraint>(
            postponed_constraint{point, it->second.get(), 
                                 clone_all(arguments), 0});
        return true;
    }
    if (has_defined_var(arguments)) {
        m_postponed_constraints.emplace_back(
            make_unique<postponed_constraint>(
                postponed_constraint{point, it->second.get(),
                                     clone_all(arguments), 0}));
        return true;
    }
    unique_ptr<AST> ast;
    if (!it->second->connect_constraint(debug_stream, point, *this,
                                        arguments, defines_var,ast))
        return false;
    if (ast)
        connect_expr(defines_var->m_var_or_set_list[0].m_var_id, *ast,
                     AST_builder::EQUATED);
    return true;
}

void parser_context::print_var_solution(
    ostream &out, const name_to_symbol_map::value_type & elem) const
{
    const auto & result = elem.second.m_var_or_set_list.at(0);
    out << elem.first << " = ";

    iset values;
    if (result.m_var_id == -1U)
        values= result.m_set;
    else
        values = m_solver.var_domain(result.m_var_id);

    print_values(out, elem.second.m_type, values);
    out << ";\n";
}

void parser_context::print_values(ostream & out, element_type type,
                                  const iset & values) const
{
    switch (type) {
      case BOOL_TYPE:
        out << (values.min() ? "true" : "false");
        break;
      case INT_TYPE:
        out << values;
        break;
      case BOOL_SET_TYPE: {
          out << "{";
          if (values.contains(0))
              out << "false";
          if (values.contains(1))
              out << (values.contains(0) ? ", true" : ",true");
          out << "}";
          break;
      }
      case INT_SET_TYPE: {
        out << "{";
        const char *prefix= "";
        for (auto it= values.begin(); it!=values.end(); it= it.advance_range()){
            iset::value_range_type range= it.current_range();
            out << prefix << range.first;
            if (range.first != range.second)
                out << ".." << range.second;
            prefix=", ";
        }
        break;
      }
      case UNKNOWN_TYPE:
        out << "UNKNOWN_TYPE{" << values << "}";
    }
}

void parser_context::print_array_solution(
    ostream &out, const name_to_symbol_map::value_type & elem,
    const annotation_type & annotation) const
{
    out << elem.first << " = ";

    if (annotation.m_arguments.get() ==nullptr ||
        annotation.m_arguments->empty() ||
        annotation.m_arguments->at(0) == nullptr) {
        out << "[];\n";
        return;
    }

    const argument_type *argument = annotation.m_arguments->at(0).get();
    auto dimensions = dynamic_cast<const array_dimension_argument*>(argument);
    if (dimensions == nullptr) {
        out << "[];\n";
        return;
    }

    out << "array" << dimensions->m_dimensions.size() << "d(";
    for (const auto & dim : dimensions->m_dimensions)
        out << dim.first << ".." << dim.second << ", ";
    out << "[";
    bool first = true;
    for (const auto & var : elem.second.m_var_or_set_list) {
        if (!first)
            out << ", ";
        iset values;
        if (var.m_var_id == -1U)
            values= var.m_set;
        else
            values= m_solver.var_domain(var.m_var_id);
        print_values(out, elem.second.m_type, values);
        first= false;
    }
    out << "]);\n";
}

void parser_context::print_solution(ostream & out) const
{
    for (auto const & elem : m_name_to_symbol) {
        for (const annotation_type & annotation : elem.second.m_annotations) {
            if (annotation.m_name == "output_var") {
                print_var_solution(out, elem);
                break;
            }
            if (annotation.m_name == "output_array") {
                print_array_solution(out, elem, annotation);
                break;
            }
        }
    }
}
void parser_context::set_objective(ostream & debug_stream, 
                                   AST *objective, direction dir,
                                   unsigned objective_var)
{
    if (!objective && objective_var != -1U)
        objective = ast_var(objective_var).release();
    m_objective_var = objective_var;
    if (is_intermediate_var(objective_var)) {
        unique_ptr<AST> tmp_ast;
        instantiate_var_define(debug_stream, objective_var, tmp_ast);
    }
    m_objective.reset(objective);
    m_direction= dir;
    if (dir == MIN_OBJECTIVE)
        connect_minimization();
    else
        connect_maximization();
}

void parser_context::connect_maximization()
{
    HCSP_ASSERT(m_direction== MAX_OBJECTIVE);

    // Here we maximize the objective function max(objective)= max::objective
    // such that objective >= max::objective is satisfiable while
    // objective >= max::objective+1 is not.


    unsigned objective_bound;
    objective_bound= m_solver.add_var(iset().complement(), "max::objective");
    connect_expr(objective_bound, *m_objective, AST_builder::BOUNDED_BELOW);
    if (m_objective_var == -1U)
        m_solver.objective_variable(objective_bound);
    else
        m_solver.objective_variable(m_objective_var);
    m_solver.objective_is_minimizing(false);
}

void parser_context::connect_minimization()
{
    HCSP_ASSERT(m_direction== MIN_OBJECTIVE);
    
    // Here we minimize the objective function min(objective)= min::objective
    // such that objective <= min::objective is satisfiable while
    // objective <= min::objective-1 is not.
    //
    unsigned objective_bound;
    objective_bound= m_solver.add_var(iset().complement(), "min::objective");
    connect_expr(objective_bound, *m_objective, AST_builder::BOUNDED_ABOVE);
    if (m_objective_var == -1U)
        m_solver.objective_variable(objective_bound);
    else
        m_solver.objective_variable(m_objective_var);

    m_solver.objective_is_minimizing(true);
}

unsigned parser_context::expression_to_variable(
    const AST & expr, unsigned default_var, usage_type used_as)
{
    update_variable_identity_function();
    return m_AST_builder.expression_to_variable(
        m_variable_identity_function, expr, default_var, used_as);
}

unique_ptr<AST> parser_context::constraint_type::to_ast(
    const var_or_value & literal) const
{
    if (literal.m_var_id == -1U)
        return ast_value(literal.m_value);
    else
        return ast_var(literal.m_var_id);
}

unique_ptr<AST> parser_context::to_ast(const var_or_value & literal) const
{
    if (literal.m_var_id == -1U)
        return ast_value(literal.m_value);
    else
        return ast_var(literal.m_var_id);
}

parser_context::var_or_value parser_context::from_ast(const AST& ast)
{
    update_variable_identity_function();
    var_or_value ret;
    ret.m_value = 0;
    ret.m_var_id = -1U;
    if (dynamic_cast<const ASTInteger*>(&ast)) {
        ret.m_value = static_cast<const ASTInteger&>(ast).getVal();
    } else if (dynamic_cast<const ASTBoolean*>(&ast)) {
        ret.m_value = static_cast<const ASTBoolean&>(ast).getVal();
    } else {
        ret.m_var_id= m_AST_builder.expression_to_variable(
            m_variable_identity_function, ast);
    }
    return ret;
}

auto parser_context::from_ast(const argument_list &arguments) -> argument_list
{
    argument_list ret;
    ret.reserve(arguments.size());
    for (const auto & ptr_arg : arguments) {
        auto as_ast_arg = dynamic_cast<const ast_argument*>(ptr_arg.get());
        if (as_ast_arg) {
            ret.emplace_back(new var_argument(from_ast(*as_ast_arg->m_ast)));
            continue;
        }
        auto as_ast_array_arg
            = dynamic_cast<const ast_array_argument*>(ptr_arg.get());
        if (as_ast_array_arg) {
            unique_ptr<array_argument> array(new array_argument);
            for (auto const & ast_ptr : as_ast_array_arg->m_array) {
                array->m_array.emplace_back(from_ast(*ast_ptr));
            }
            ret.push_back(move(array));
            continue;
        }
        ret.emplace_back(ptr_arg->clone());
    }
    return ret;
}

unique_ptr<AST> parser_context::constraint_type::ast_var(unsigned var) const
{
    stringstream name;
    name << "$" << var;
    return std::unique_ptr<AST>(new ASTVar(name.str(), var));
}

unique_ptr<AST> parser_context::ast_var(unsigned var) const
{
    return std::unique_ptr<AST>(new ASTVar(var_name(var), var));
}

auto parser_context::to_ast_array_argument(
    const var_or_value_list & args) const -> unique_ptr<ast_array_argument> 
{
    unique_ptr<ast_array_argument> ast_array(new ast_array_argument);
    auto & vec= ast_array->m_array;
    for (const auto & elem: args) {
        unique_ptr<AST> ast_elem = to_ast(elem);
        if (!ast_elem)
            return nullptr;
        vec.emplace_back(move(ast_elem));
    }
    return move(ast_array);
}

string parser_context::var_name(unsigned var) const
{
    stringstream name;
    name << m_solver.var_printer(var);
    return name.str();
}

// static
unique_ptr<AST> parser_context::ast_value(int32_t value)
{ return std::unique_ptr<AST>(new ASTInteger(value)); }


bool parser_context::instantiate_var_define(ostream & debug_stream,
                                            unsigned id,
                                            unique_ptr<AST> & ast)
{
    if (id == -1U || id >= m_id_to_name.size())
        return true;
    const string & name = m_id_to_name[id];
    auto it = m_name_to_symbol.find(name);
    if (it == m_name_to_symbol.end())
        return true;
    if (it->second.m_defined_by_ast) {
        ast= deep_copy(*it->second.m_defined_by_ast);
        return true;
    }
    postponed_constraint * defined_by = it->second.m_defined_by.get();
    if (!defined_by)
        return true;
    if (defined_by->m_count_instantiated != 0)
        return true;
    if (instantiate_postponed(debug_stream, *defined_by, &it->second, ast)) {
        if (ast)
            it->second.m_defined_by_ast= deep_copy(*ast);
        return true;
    }
    return false;
}

bool parser_context::instantiate_postponed(
    ostream & debug_stream,
    postponed_constraint & postponed,
    symbol_info* defined_var,
    unique_ptr<AST> & ast_out)
{
    ++postponed.m_count_instantiated;
    argument_list children;
    for (const auto & ptr_arg : postponed.m_arguments) {
        children.emplace_back(ptr_arg->clone());
        auto var_arg = dynamic_cast<var_argument*>(ptr_arg.get());
        if (var_arg) {
            unique_ptr<AST> ast;
            if (!instantiate_var_define(debug_stream,
                                        var_arg->m_variable.m_var_id, ast)) {
                return false;
            }
            if (!ast)
                ast= to_ast(var_arg->m_variable);
            children.back().reset(new ast_argument{move(ast)});
            continue;
        }
        auto array_arg = dynamic_cast<array_argument*>(ptr_arg.get());
        if (array_arg) {
            unique_ptr<ast_array_argument> ast_array(new ast_array_argument);
            for (const auto & var_entry: array_arg->m_array) {
                unique_ptr<AST> ast;
                if (var_entry.m_var_id == -1U)
                    ast.reset(new ASTInteger(var_entry.m_value));
                else if (!instantiate_var_define(debug_stream,
                                                 var_entry.m_var_id, ast))
                    return false;
                if (!ast)
                    ast= to_ast(var_entry);
                ast_array->m_array.push_back(move(ast));
            }
            children.back()= move(ast_array);
        }
    }
    return postponed.m_constraint->connect_constraint(
        debug_stream, postponed.m_point, *this, children, defined_var, ast_out);
}

bool parser_context::create_constraints(ostream & debug_stream)
{
    for (auto & ptr_postponed : m_postponed_constraints) {
        unique_ptr<AST> out;
        if (!instantiate_postponed(debug_stream, *ptr_postponed, nullptr, out))
            return false;
        if (out) {
            connect_expr(singleton_var(true), *out);
        }
    }
    m_AST_builder.create_constraints();
    return true;
}


} // namespace hcsp
