// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <stdexcept>
#include <iostream>
#include <solver.hpp>
#include <proof_collector.hpp>
#include <proof_print_visitor.hpp>
#include <variable_print_strategy.hpp>

using namespace std;
using namespace hcsp;

/**
 * @brief Data collected for one explanation of a propagator run.
 */
class proof_collector::explanation_data
{
  public:
    //! Default constructor
    explanation_data() : m_p_rule(NULL) {}
    
    //! Copy constructor
    explanation_data(const explanation_data & other) : m_p_rule(NULL)
    { *this= other; }
    
    //! Destructor
    ~explanation_data() { delete m_p_rule; m_p_rule= NULL; }
    
    //! Assignment
    explanation_data & operator=(const explanation_data & other)
    {
        delete m_p_rule;
        m_p_rule= NULL; // if clone() throws, this will avoid double free

        if (other.m_p_rule)
            m_p_rule=     other.m_p_rule->clone();
        m_constraint= other.m_constraint;
        m_literals=   other.m_literals;

        return *this;
    }
    //! Get the ID of the explained constraint.
    unsigned constraint() const { return m_constraint; }
    //! Set the ID of the explained constraint, return old ID
    unsigned constraint(unsigned next)
    {
        unsigned orig= m_constraint;
        m_constraint= next;
        return orig;
    }
    
    //! Set the explanation rule of the propagator run.
    void rule(std::unique_ptr<explanation_rule> rule_in)
    { m_p_rule= rule_in.release(); }

    //! Tell whether an explanation rule is associated with this propagation.
    bool has_rule() const
    { return m_p_rule != NULL; }

    /**
     * @brief Get the explanation rule associated with this propagation
     *
     * @return A reference to the explanation associated with this propagation
     * @pre There is a rule associated with this propagation (has_rule()==true)
     */
    const explanation_rule & rule() const { return *m_p_rule; }

    //! Set the literals associated with this propagation
    void literals(const std::vector<iset> & literals_in)
    { m_literals= literals_in; }

    //! Get the literals associated with this propagation
    const std::vector<iset> & literals() const
    { return m_literals; }    
  private:
    unsigned m_constraint;
    explanation_rule * m_p_rule;
    std::vector<iset> m_literals;
}; // class proof_collector::explanation_data

// virtual
void proof_collector_dummy::print_proof(ostream & ) const
{
    throw runtime_error("solver::generate_proofs() was not called, "
                          "no proof available");
}

proof_collector::proof_collector(const solver & the_solver) :
    m_solver(the_solver),
    m_latest_resolution_start(-1U),
    m_first_conflict_clause(-1U)
{}

// virtual
proof_collector::~proof_collector()
{}

//virtual
void proof_collector::print_proof(ostream &out) const
{
    proof_print_visitor visitor(out, *this);
    
    traverse_proof(visitor);
}

//virtual
void proof_collector::traverse_proof(proof_visitor & visitor) const
{
    if ( m_resolve_history.empty() )
        throw runtime_error("No unsatisfiability record available");
    unsigned clause= last_conflict_clause();
    
    vector<bool> visited(m_first_conflict_clause, true);
    visited.resize(clause+1, false);

    state st(*this);
    traverse_proof(visitor, st, clause, visited);
}

void proof_collector::traverse_proof(proof_visitor & visitor,
                                     state & st,
                                     unsigned clause,
                                     vector<bool> & visited) const
{
    if ( visited[clause] )
        return;
    visited[clause]= true;
    
    unsigned resolve_begin= clause_to_resolve_index(clause);
    // first, find resolve end
    unsigned resolve_end= resolve_begin;
    do {
        ++resolve_end;
    } while ( is_inner_resolve_entry(resolve_end) );

    // Next, recurse into all constraints leading to 'clause'.
    for (unsigned i= resolve_begin ; i!=resolve_end ; ++i) {
        int resolved_clause= m_resolve_history[i].m_clause;
        if ( !is_explanation_clause(resolved_clause) )
            traverse_proof(visitor, st, resolved_clause, visited);
    }
    // Finally, print these resolutions
    int front_clause_id= m_resolve_history[resolve_begin].m_clause;
    clause_data front_clause;
    get_clause(front_clause_id, front_clause);
    if ( is_explanation_clause(front_clause_id) )
        visitor.visit_explanation(front_clause_id);
        
    for (unsigned i= resolve_begin+1 ; i<resolve_end ; ++i)
        process_resolution(visitor, st,
                           front_clause_id, front_clause, clause, i);
        
    if ( resolve_end == m_resolve_history.size() && !front_clause.empty() )
        process_conflict_with_initial(visitor, st,
                                      front_clause_id, front_clause);
}

unsigned proof_collector::explanation_count() const
{
    return m_explanations.size();
}

void proof_collector::process_resolution(
    proof_visitor & visitor, state & st,
    int & work_clause_id, clause_data & work_clause,
    unsigned final_clause, unsigned resolution_entry) const
{
    clause_data resolvent, consequent;
                        
    const resolve_data & this_resolve= m_resolve_history[resolution_entry];
    int consequent_id;
    if (is_inner_resolve_entry(resolution_entry+1)) {
        consequent_id= st.intermediate_clause_id();
    } else {
        consequent_id= final_clause;
    }
    int resolvent_id= this_resolve.m_clause;
    if ( is_explanation_clause(resolvent_id) )
        visitor.visit_explanation(resolvent_id);
    get_clause(resolvent_id, resolvent);

    resolve(work_clause, resolvent, this_resolve.m_pivot, consequent);

    visitor.visit_resolution(this_resolve.m_pivot,
                             work_clause_id, work_clause,
                             resolvent_id, resolvent,
                             consequent_id, consequent);

    work_clause_id= consequent_id;
    work_clause= consequent;
}

inline const proof_collector::explanation_data &
proof_collector::get_explanation_data(int clause_id) const
{
    return m_explanations.at(explanation_index_of_clause(clause_id));
}

bool proof_collector::is_explanation_clause(int clause_id) const
{
    return explanation_index_of_clause(clause_id) < m_explanations.size();
}

void proof_collector::get_clause(int clause_id, clause_data & clause) const
{
    if (clause_id >= 0)
        get_solver_clause( clause_id, clause);
    else if (is_explanation_clause(clause_id))
        get_explanation_clause( clause_id, clause);
    else
        throw runtime_error("proof_collector::get_clause(): non-explanation "
                            "intermediate constraints are not supported");
}

unsigned proof_collector::get_constraint_explained_by(int clause_id) const
{
    HCSP_ASSERT(clause_id < 0);
    return get_explanation_data(clause_id).constraint();
}

void proof_collector::print_resolution(
    ostream & out, unsigned pivot,
    int resolvent1_id, const clause_data & resolvent1,
    int resolvent2_id, const clause_data & resolvent2,
    int consequent_id, const clause_data & consequent) const
{
    out << "resolve[" << var_printer(pivot) << "] ";
    print_clause(out, resolvent1_id, resolvent1);
    out << " with ";
    print_clause(out, resolvent2_id, resolvent2);
    out << "\n" "consequent: ";
    print_clause(out, consequent_id, consequent);
    out << "\n" << endl;
}

void proof_collector::append_resolution_entry(
    unsigned constraint_id, const vector<iset> & literals, int pivot_variable,
    unique_ptr<explanation_rule> rule)
{
    resolve_data this_resolve;
    this_resolve.m_pivot= pivot_variable;

    if ( m_solver.is_clause(constraint_id) )
        this_resolve.m_clause= constraint_id;
    else {
        m_explanations.push_back(explanation_data());
        m_explanations.back().constraint(constraint_id);
        if (rule.get() == NULL)
            m_explanations.back().literals(literals);
        else
            m_explanations.back().rule(move(rule));
        this_resolve.m_clause= -(int)m_explanations.size();
    }
    m_resolve_history.push_back(this_resolve);
}

void proof_collector::pop_resolution_entry()
{
    int clause= m_resolve_history.back().m_clause;
    if (clause < 0 && (unsigned)-clause == m_explanations.size())
        m_explanations.pop_back();
    m_resolve_history.pop_back();
}

void proof_collector::print_explanation(ostream & out, int clause_id) const
{
    const explanation_data & explanation= get_explanation_data(clause_id);
    if (explanation.has_rule()) {
        out << "rule ";
        vector<unsigned> constraint_variables;
        m_solver.get_constraint_variables(explanation.constraint(),
                                          constraint_variables);
        explanation.rule().print(out, m_solver.get_variable_print_strategy(),
                                 constraint_variables);
        out << ": ";
    }
    m_solver.print_constraint(out, explanation.constraint() );
    out << "\n" "consequent: ";
    clause_data clause;
    get_clause( clause_id, clause);;
    print_clause(out, clause_id, clause);
    out << "\n\n";
}

void proof_collector::process_conflict_with_initial(
    proof_visitor & visitor, state & st,
    int & clause_id, clause_data & clause) const
{
    while (clause.size()) {
        int consequent_id= st.intermediate_clause_id();
        unsigned variable= clause.back().m_variable;

        clause_data consequent= clause;
        consequent.pop_back();
        const iset & initial = m_solver.var_initial_domain(variable);
        visitor.visit_initial_resolution(
            variable, initial,
            clause_id, clause, consequent_id, consequent);

        HCSP_ASSERT(!clause.back().m_set.contains_any_of(initial));
        clause_id= consequent_id;
        clause= consequent;
    }
}

void proof_collector::print_initial_resolution(
    ostream & out, unsigned variable, const iset & initial,
    int resolvent_id, const clause_data & resolvent,
    int consequent_id, const clause_data & consequent) const
{
    out << "resolve[" << var_printer(variable) << "] ";
    print_clause(out, resolvent_id, resolvent);
    out << " with initial domain (" << var_printer(variable);
    if ( initial.singleton() )
        out << "=" << initial;
    else
        out << ":{" << initial << "}";
    out << ")\n" << "consequent: ";
    
    print_clause(out, consequent_id, consequent);
    out << "\n\n";
}

void proof_collector::get_constraint_variables(unsigned constraint_id,
                                               clause_data & clause) const
{
    clause.clear();
    if (m_solver.has_constraint(constraint_id)) {
        vector<unsigned> constraint_variables;
        m_solver.get_constraint_variables(constraint_id, constraint_variables);
        clause.resize(constraint_variables.size());
        for (unsigned i=0 ; i != clause.size() ; ++i)
            clause[i].m_variable= constraint_variables[i];
    } else {
        saved_clause_set::const_iterator it= find_saved_clause(constraint_id);
        if (it == m_saved_clauses.end())
            throw runtime_error("Can't find this constraint_it");
        for (unsigned i=0 ; i != clause.size() ; ++i)
            clause[i].m_variable= it->second[i].m_variable;
    }
}
    
void proof_collector::get_solver_clause(unsigned clause_id,
                                        clause_data & clause) const
{
    if (m_solver.has_constraint(clause_id)) {
        if ( !m_solver.is_clause(clause_id) )
            runtime_error("proof_collector::get_solver_clause(): not a clause");
        get_constraint_variables(clause_id, clause);
        const vector<iset> & resolvent= m_solver.clause_literals(clause_id);
        HCSP_ASSERT(resolvent.size() == clause.size());
        for (unsigned i=0 ; i != clause.size() ; ++i)
            clause[i].m_set= resolvent[i];
    } else {
        saved_clause_set::const_iterator it= find_saved_clause(clause_id);
        if (it == m_saved_clauses.end())
            throw runtime_error("Can't find this constraint_it");
        clause= it->second;
    }
}

void proof_collector::get_explanation_clause(int clause_id,
                                             clause_data & clause) const
{        
    const explanation_data & explanation= get_explanation_data(clause_id);

    get_constraint_variables(explanation.constraint(), clause);
    if ( explanation.has_rule()) {
        vector<iset> literals;
        explanation.rule().get_explanation(literals);
        HCSP_ASSERT(literals.size() == clause.size());
        for (unsigned i= 0 ; i != clause.size() ; ++i )
            clause[i].m_set= literals[i];
    } else {
        const vector<iset> &literals= explanation.literals();
        HCSP_ASSERT(literals.size() == clause.size());
        for (unsigned i= 0 ; i != clause.size() ; ++i )
            clause[i].m_set= literals[i];
    }
    
    // keep only nonempty literals
    for (unsigned i= clause.size() ; i != 0; ) {
        --i;
        if (clause[i].m_set.empty()) {
            clause[i]= clause.back();
            clause.pop_back();
        }
    }
}

//static
void proof_collector::resolve(const clause_data & first,
                              const clause_data & second,
                              unsigned pivot, clause_data & consequent)
{
    consequent.clear();
    consequent.reserve(first.size()+second.size()-1);
    iset first_pivot_set;
    for (unsigned i=0 ; i != first.size() ; ++i) {
        if (first[i].m_variable == pivot)
            first_pivot_set.insert_all_of(first[i].m_set);
        else
            consequent.push_back(first[i]);
    }
    iset second_pivot_set;
    for (unsigned i=0 ; i != second.size() ; ++i) {
        if (second[i].m_variable == pivot)
            second_pivot_set.insert_all_of(second[i].m_set);
        else
            consequent.push_back(second[i]);
    }
    consequent.canonicalize();
    iset pivot_set= first_pivot_set;
    pivot_set.intersect(second_pivot_set);
    if (!pivot_set.empty())
        consequent.push_back(literal_data(pivot, pivot_set));
}

//static
void proof_collector::print_clause(ostream & out, int id,
                                   const clause_data & clause) const
{
    out << "[" << id << "]" << " ";
    clause.print(out, m_solver.get_variable_print_strategy());
}
