// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <stdexcept>
#include <iostream>
#include <map>
#include <solver.hpp>
#include <proof_interpolant_visitor.hpp>

using namespace std;
using namespace hcsp;

proof_interpolant_visitor::proof_interpolant_visitor(
    const proof_collector &collector, const solver & the_solver)
    : m_collector(collector),
      m_solver(the_solver),
      m_variables_touched_by_labels(the_solver.variables_touched_by_labels()),
      m_empty_clause_id(-1U),
      m_need_trace(the_solver.report_level() > solver::SEARCH)
    {}

// virtual
proof_interpolant_visitor::~proof_interpolant_visitor()
{
}

// virtual
void proof_interpolant_visitor::visit_explanation(int explanation_clause_id)
{
    HCSP_ASSERT(m_formulas.find(explanation_clause_id) == m_formulas.end());
    unsigned constraint;
    constraint= m_collector.get_constraint_explained_by(explanation_clause_id);
    if ( m_solver.constraint_labels(constraint) & solver::A ) {
        HCSP_ASSERT(explanation_clause_id < 0);
        if (need_trace())
            cout << "HypC-A";
        introduce_clause_in_A(explanation_clause_id);
    } else {
        if (need_trace())
            cout << "HypC-B";
        m_formulas[explanation_clause_id]= formula(true);
    }
    if (need_trace())
        cout << " (constraint[" << constraint
             << "] explanation["<< explanation_clause_id << "])\n---> " 
             << printer(m_formulas[explanation_clause_id]) << "\n";
}

void proof_interpolant_visitor::introduce_clause_in_A(int clause_id)
{
    clause_data clause;
    m_collector.get_clause(clause_id, clause);
    for (unsigned i=clause.size() ; i > 0 ; ) {
        --i;
        if ( (m_variables_touched_by_labels.at(clause[i].m_variable) & B) == 0){
            clause[i].swap(clause.back());
            clause.pop_back();
        }
    }
    m_formulas[clause_id]= formula(clause);
}

//virtual
void proof_interpolant_visitor::visit_resolution(
    unsigned pivot,
    int resolvent1_id, const clause_data &,
    int resolvent2_id, const clause_data &,
    int consequent_id, const clause_data & consequent)
{
    const formula & first=  get_formula(resolvent1_id);
    const formula & second= get_formula(resolvent2_id);
    formula result;
    if (m_variables_touched_by_labels.at(pivot) & B) {
        if (need_trace())
            cout << "Res-B[";
        // Rule Res-B
        result= (first && second);
    } else {
        // Rule Res-A
        if (need_trace())
            cout << "Res-A[";
        result= (first || second);
    }
    if (need_trace())
        cout << printer(pivot) << "] { " << resolvent1_id << ": "
             << printer(first)
             << "}\n           { " << resolvent2_id << ": " << printer(second)
             << "}\n---> " << consequent_id << ": " << printer(result) << "\n";
    HCSP_ASSERT(m_formulas.find(consequent_id) == m_formulas.end());
    m_formulas[consequent_id]= result;

    if (consequent.empty())
        m_empty_clause_id= consequent_id;
}

//virtual
void proof_interpolant_visitor::visit_initial_resolution(
    unsigned variable, const iset & ,
    int resolvent_id, const clause_data & ,
    int consequent_id, const clause_data & consequent)
{
    const formula & first=  get_formula(resolvent_id);
    formula result;
    if (m_variables_touched_by_labels.at(variable) & B) {
        if (need_trace())
            cout << "HypC-B + Res-B";
        // We arbitrarily put the unary domain constraint into B if the variable
        // is used by any constraint in B.
        // The HypC-B rule produces a true_formula, and then Res-B gets rid
        // of it, by anding it with first.
        result= first;
    } else {
        if (need_trace())
            cout << "HypC-A + Res-A";
        // We introduce the unary domain constraint through HypC-A:
        // (A,B) |- <literal>[ <literal \downarrow B> ]
        // but since the variable is not in B, then <literal \downarrow B>
        // is false.
        result= formula(false);
    }
    if (need_trace())
        cout << " (initial domain of " << printer(variable) << " with clause["
             << resolvent_id << "])\n---> " << printer(result) << "\n";
    HCSP_ASSERT(m_formulas.find(consequent_id) == m_formulas.end());
    m_formulas[consequent_id]= result;
    if (consequent.empty())
        m_empty_clause_id= consequent_id;
}

formula proof_interpolant_visitor::get_interpolant() const
{
    HCSP_ASSERT(m_formulas.find(m_empty_clause_id) != m_formulas.end());
    return m_formulas.find(m_empty_clause_id)->second;
}

const formula & proof_interpolant_visitor::get_formula(int clause_id)
{
    if (m_formulas.find(clause_id) == m_formulas.end()) {
        if ( m_solver.constraint_labels(clause_id) & A ) {
            if (need_trace())
                cout << "HypC-A";
            introduce_clause_in_A(clause_id);
        } else {
            if (need_trace())
                cout << "HypC-B";
            m_formulas[clause_id]= formula(true);
        }
        if (need_trace())
            cout << " (clause[" << clause_id << "])\n---> "
                 << printer(m_formulas[clause_id]) << "\n";
    } 
    return m_formulas[clause_id];
}

