// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <limits>
#include <propagation.hpp>

using namespace std;
using namespace hcsp;
using namespace hcsp::propagation;

static void normalize_and_insert_expand_stride(
    pair<long long, long long> range,
    iset & accumulated_result,
    unsigned stride)
{
    static const iset::size_type max_diff=
        (iset::size_type)iset::MAX_DOMAIN-iset::MIN_DOMAIN;
    const long long modulo = 1LL+max_diff;
    HCSP_ASSERT( (range.second - range.first) % (long long)stride == 0);
    long long value= range.first;
    while ( value < iset::MIN_DOMAIN) {
        accumulated_result.insert(value+modulo);
        if (value == range.second)
            return;
        value+=stride;
    }
    while ( value <= iset::MAX_DOMAIN) {
        accumulated_result.insert(value);
        if (value == range.second)
            return;
        value+=stride;
    }
    while ( true ) {
        accumulated_result.insert(value-modulo);
        if (value >= range.second)
            return;
        value+=stride;
    }
}

static void normalize_and_insert(pair<long long, long long> range,
                                 iset & accumulated_result,
                                 unsigned stride=1)
{
    typedef iset::value_range_type range_type;
    HCSP_ASSERT(range.first <= range.second);
    HCSP_STATIC_ASSERT(iset::MAX_DOMAIN > iset::MIN_DOMAIN);
    static const iset::size_type max_diff=
        (iset::size_type)iset::MAX_DOMAIN-iset::MIN_DOMAIN;
    // Make sure that the size of iset's domain is a power of two
    HCSP_STATIC_ASSERT( (max_diff & (max_diff+1)) == 0);
    const long long modulo = 1LL+max_diff;
    if ( (unsigned long long)range.second-range.first >= max_diff) {
        if (stride == 1 || (range.second-range.first)/stride> 100 )
            accumulated_result.assign(range_type(iset::MIN_DOMAIN,
                                                 iset::MAX_DOMAIN));
        else
            normalize_and_insert_expand_stride(range, accumulated_result,
                                               stride);
        return;
    }
    
    // Normalize range.first;
    if ( range.first > iset::MAX_DOMAIN || range.first < iset::MIN_DOMAIN ) {
        long long new_first= range.first % modulo;
        if (new_first > iset::MAX_DOMAIN)
            new_first -= modulo;
        else if (new_first < iset::MIN_DOMAIN)
            new_first += modulo;
        long long diff= new_first - range.first;
        range.first= new_first;
        range.second += diff;
    }
    HCSP_ASSERT(iset::MIN_DOMAIN <= range.first &&
               range.first <= iset::MAX_DOMAIN);
    if ( range.second > iset::MAX_DOMAIN ) {
        iset::value_type mid= iset::MAX_DOMAIN;
        // round down to the stride's size
        mid -= ((iset::size_type)iset::MAX_DOMAIN-range.first)%stride;
        accumulated_result.insert(range_type(range.first, mid), stride);
        HCSP_ASSERT((long long)mid+stride > iset::MAX_DOMAIN);
        mid = (long long)mid+stride-modulo;
        long long new_second = range.second-modulo;
        // This case of double wrap around has been handled at the beginning
        HCSP_ASSERT(new_second < iset::MAX_DOMAIN);
        accumulated_result.insert(range_type(mid, new_second), stride);
    } else {
        accumulated_result.insert(range_type(range), stride);
    }
}


static void plus_ranges(const iset::value_range_type & a,
                        const iset::value_range_type & b, unsigned stride,
                        iset & accumulated_result)
{
    HCSP_STATIC_ASSERT(sizeof(iset::value_type)*2 <= sizeof(long long));

    HCSP_ASSERT((0LL+ a.second - a.first+0ULL) % stride == 0);
    HCSP_ASSERT((0LL+ b.second - b.first+0ULL) % stride == 0);
    
    normalize_and_insert(make_pair(0LL+a.first +b.first,
                                   0LL+a.second+b.second),
                         accumulated_result, stride);
}


static void plus_ranges(const iset::value_range_type & a, unsigned a_stride,
                        const iset::value_range_type & b, unsigned b_stride,
                        iset & accumulated_result)
{
    HCSP_ASSERT((0LL+ a.second - a.first+0ULL) % a_stride == 0);
    HCSP_ASSERT((0LL+ b.second - b.first+0ULL) % b_stride == 0);

    if (a_stride > b_stride)
        return plus_ranges(b, b_stride, a, a_stride, accumulated_result);

    if (a_stride== b_stride || (a_stride == 1 && a.first == a.second))
        return plus_ranges(a, b, b_stride, accumulated_result);
    if (b_stride % a_stride == 0 && 0LL+a.second - a.first >= 0LL+b_stride)
        return plus_ranges(a, b, a_stride, accumulated_result);
    for (iset::value_type i=b.first ; i != b.second ; i+= b_stride)
        plus_ranges(a, make_pair(i,i), a_stride, accumulated_result);
    plus_ranges(a, make_pair(b.second,b.second), a_stride, accumulated_result);
}

static void minus_ranges(const iset::value_range_type & a, unsigned a_stride,
                         const iset::value_range_type & b, unsigned b_stride,
                         iset & accumulated_result)
{
    iset::value_range_type minus_b;
    if (b.first == numeric_limits<iset::value_type>::min() &&
        b.first == -1 - numeric_limits<iset::value_type>::max()) {
        HCSP_PROFILE_LINE;
        // Avoid a -MIN_INT overflow
        plus_ranges(a, a_stride, make_pair(b.first, b.first), 1,
                    accumulated_result);
        if (b.second == b.first)
            return;
        HCSP_PROFILE_LINE;
        minus_b= make_pair(-b.second, -b_stride-b.first);
    } else {
        minus_b= make_pair(-b.second, -b.first);
    }

    plus_ranges(a, a_stride, minus_b, b_stride, accumulated_result);
}


void hcsp::propagation::plus(const iset & a, const iset & b, iset & plus_result)
{
    HCSP_PROFILE_CURRENT;
    if (a.empty() || b.empty() || plus_result.empty()) {
        plus_result.clear();
        return;
    }
    if (a.everything() || b.everything())
        return;
    iset temp_result;

    iset::iterator b_last= b.end();
    --b_last;
    // do first and last ranges of b before doing the middle so that
    // fixed-point will be detected earlier
    for (iset::iterator i= a.begin() ; i != a.end() ; ) {
        iset::value_range_type a_range;
        unsigned a_stride;
        i= i.find_stride(a.end(), a_range, a_stride);
        plus_ranges(a_range, b_last.current_range(), a_stride, temp_result);
    }

    
    for (iset::iterator i= a.begin() ; i != a.end() ; ) {
        if (temp_result.contains_all_of(plus_result))
            return;
        iset::value_range_type a_range;
        unsigned a_stride;
        i= i.find_stride(a.end(), a_range, a_stride);
        
        for (iset::iterator j= b.begin() ; j != b_last && j != b.end(); ) {
            iset::value_range_type b_range;
            unsigned b_stride;
            j= j.find_stride(b.end(), b_range, b_stride);
            plus_ranges(a_range, a_stride, b_range, b_stride, temp_result);
        }
    }
    plus_result.intersect(temp_result);
}

void hcsp::propagation::minus(const iset & sum, const iset & subtrahend,
                             iset & minus_result)
{
    HCSP_PROFILE_CURRENT;
    if (sum.empty() || subtrahend.empty() || minus_result.empty()) {
        minus_result.clear();
        return;
    }
    if (sum.everything() || subtrahend.everything())
        return;
    iset temp_result;
    iset::iterator subtrahend_last= subtrahend.end();
    --subtrahend_last;
    // do first and last ranges of subtrahend before doing the middle so that
    // fixed-point will be detected earlier
    for (iset::iterator i= sum.begin() ; i != sum.end() ; ) {
        iset::value_range_type sum_range;
        unsigned sum_stride;
        i= i.find_stride(sum.end(), sum_range, sum_stride);
        minus_ranges(sum_range, sum_stride,
                     subtrahend_last.current_range(), 1, temp_result);
    }
    
    for (iset::iterator i= sum.begin() ; i != sum.end() ; ) {
        if ( temp_result.contains_all_of(minus_result) )
            return;
        iset::value_range_type sum_range;
        unsigned sum_stride;
        i= i.find_stride(sum.end(), sum_range, sum_stride);
        for (  iset::iterator j= subtrahend.begin() ;
               j != subtrahend_last && j != subtrahend.end(); ) {
            iset::value_range_type subtrahend_range;
            unsigned subtrahend_stride;
            j= j.find_stride(subtrahend.end(), subtrahend_range,
                             subtrahend_stride);
            minus_ranges(sum_range, sum_stride,
                         subtrahend_range, subtrahend_stride, temp_result);
        }
    }
    minus_result.intersect(temp_result);
}

void hcsp::propagation::neg(const iset & original, iset & negated_result)
{
    HCSP_PROFILE_CURRENT;
    iset temp_result;
    for ( iset::iterator it= original.begin() ;
          it != original.end();
          it.advance_range() ) {
        iset::value_range_type range= it.current_range();
        normalize_and_insert(make_pair(0LL- range.second, 0LL-range.first),
                             temp_result);
    }
    negated_result.intersect(temp_result);
}
