// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2013 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2013 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <sliced_runtime_info.hpp>
#include <constraint_runtime_info.hpp>

using namespace std;
using namespace hcsp;

void sliced_runtime_info::construct()
{
    m_querying_domain= true;
    m_is_modified.resize(m_end-m_begin, false);
}

bool sliced_runtime_info::reduced_between_activations(unsigned param_index)
    const
{
    return (m_const_info.reduced_between_activations(param_index+m_begin) ||
            m_is_modified[param_index]);
}

void sliced_runtime_info::set_domain(unsigned param_index, iset new_domain)
{
    if (domain(param_index).size() == new_domain.size() &&
        domain(param_index).empty() == new_domain.empty())
        return;
    if (!querying_domain()) {
        HCSP_ASSERT(m_p_info);
        HCSP_ASSERT(!m_is_modified[param_index]);
        m_p_info->set_domain(param_index+m_begin, new_domain);
    }
        
    if (!m_is_modified[param_index]) {
        m_modified_domain.reserve(arity());
        if (m_modified_domain.size() <= param_index)
            m_modified_domain.resize(param_index+1);
        m_modified_domain.at(param_index) = move(new_domain);
        m_is_modified[param_index]= true;
    }
}

void sliced_runtime_info::deallocate_watch(unsigned param_index,
                                           watch_type type, 
                                           watch_handle_t watch_handle)
{
    m_p_info->deallocate_watch(param_index+m_begin, type, watch_handle);
}

watch_handle_t sliced_runtime_info::allocate_watch(unsigned param_index,
                                                   watch_type type)
{
    return m_p_info->allocate_watch(param_index+m_begin, type);
}

void sliced_runtime_info::deallocate_value_watch(
    unsigned param_index, iset::value_type value, watch_handle_t watch_handle)
{
    m_p_info->deallocate_value_watch(param_index+m_begin, value, watch_handle);
}

watch_handle_t sliced_runtime_info::allocate_value_watch(unsigned param_index,
                                                         iset::value_type value)
{
    return m_p_info->allocate_value_watch(param_index+m_begin, value);
}

int sliced_runtime_info::latest_modified_param(unsigned except) const
{
    unsigned external_except= -1U;
    if (except < m_end-m_begin)
        external_except= except+m_begin;
    int latest= m_const_info.latest_modified_param(external_except);
    if (latest >= (int)m_begin && latest< (int)m_end)
        return latest-m_begin;
    if (latest < 0)
        // did not find any possible value
        return -1;
    
    // If we got here then latest points to a value out of m_begin, m_end range.
    // Try to find a value different than both 'latest' and 'except'
    
    latest= m_const_info.latest_modified_param(latest);
    if (latest < 0)
        return -1;
    
    if ((unsigned)latest == except)
        // Maybe there is a value, but we can't instruct m_const_info to search
        // for the latest modification excluding the two values simultaneously.
        return -1;
    if (latest >= (int)m_begin && latest< (int)m_end)
        return latest-m_begin;
    else
        return -1;
}

void sliced_runtime_info::set_querying_domain(bool state)
{
    if (!state && m_querying_domain) {
        commit();
        m_modified_domain.clear();
    }
    m_querying_domain= state;
}

void sliced_runtime_info::print_param(ostream & out, unsigned param) const
    throw()
{
    m_const_info.print_param(out, param+m_begin);
}

void sliced_runtime_info::print(ostream & out) const
    throw()
{
    print_prefix(out);
    for (unsigned i=0 ; i != arity() ; ++i) {
        if ( i!=0 )
            out << ",";
        print_param(out, i);
        out << "={" << domain(i) << "}";
    }
    out << ")";
}


void sliced_runtime_info::commit()
{
    for (unsigned i=0 ; i != m_modified_domain.size() ; ++i) {
        if (m_is_modified[i]) {
            m_p_info->set_domain(i+m_begin, move(m_modified_domain[i]));
            m_is_modified[i]= false;
        }
    }
}
