// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <cassert>
#include <unistd.h>
#include <iostream>
#include <sstream>
#include <limits>
#include <utilities.hpp>
#include <small_iset.hpp>
#include "iset_helpers.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::iset_helpers;

// static const
const small_iset::value_type small_iset::MIN_DOMAIN;
// static const
const small_iset::value_type small_iset::MAX_DOMAIN;

//static const
const bool small_iset::iterator::CHECKS_BOUNDS;

//static const
const bool small_iset::iterator::INVALIDATED_BY_ERASE;

// explicit
small_iset::small_iset(const std::string & text)
{
    assign_set(*this, text);
}

void small_iset::throw_out_of_range(const char *text) const throw(out_of_range)
{
    ostringstream out;
    out << "small_iset out of range " << text;
    throw out_of_range(out.str());
}

small_iset::value_type small_iset::min() const
{
    if (empty())
        throw domain_error("min() over empty an set");
    return find_lsb_index(m_bits);
}

small_iset::value_type small_iset::max() const
{
    if (empty())
        throw domain_error("max() over empty an set");
    return find_msb_index(m_bits);
}

small_iset::value_type small_iset::rand() const
{
    if (everything())
        return hcsp::random::range(MIN_DOMAIN, MAX_DOMAIN);
    if (empty())
        throw runtime_error("Can't hcsp::small_iset::rand() for an empty set");
    size_type element_index= hcsp::random::gen(size());
    
    for (iterator it= begin() ; ; it.advance_range()) {
        HCSP_ASSERT(it != end());

        value_range_type current_range= it.current_range();
        size_type current_range_size= range_size(current_range);
        
        if (element_index >= current_range_size)
            element_index -= current_range_size;
        else
            return current_range.first+element_index;
    }
}

small_iset & small_iset::insert(const value_range_type & range)
{
    assert_valid_range(range, "while inserting a range");
    Bits zeros_below_range= unsigned_ones_on_and_above<Bits>(range.first);
    Bits zeros_above_range= unsigned_ones_on_and_below<Bits>(range.second);
    
    m_bits |= (zeros_below_range & zeros_above_range);
    return *this;
}

small_iset & small_iset::insert(const value_range_type & range,
                                unsigned stride)
{
    if (stride == 1)
        insert(range);
    else {
        if ( (range.second-range.first)%stride != 0)
            throw_out_of_range("While inserting range with a wrong stride");
        value_type val;
        for (val= range.first; val != range.second; val+=stride)
            insert(val);
        insert(val);
    }
    return *this;
}


small_iset & small_iset::erase(const value_range_type & range)
{
    if (range.first > range.second)
        throw_out_of_range(": illegel unordered values in pair");
    if (range.first > MAX_DOMAIN || range.second < MIN_DOMAIN)
        return *this;
    value_type first=  std::max(range.first,  MIN_DOMAIN);
    value_type second= std::min(range.second, MAX_DOMAIN);
    assert(first <= second);
    Bits ones_below_range= unsigned_ones_below<Bits>(first);
    Bits ones_above_range= unsigned_ones_above<Bits>(second);
    m_bits &= (ones_below_range | ones_above_range);
    
    return *this;
}

small_iset::iterator small_iset::erase(iterator erased_iterator)
{
    HCSP_ASSERT(erased_iterator != end());
    HCSP_ASSERT(erased_iterator.m_pIset == this);
    value_type erased_value= *erased_iterator;
    erase(erased_value);
    ++erased_iterator;
    return erased_iterator;
}

small_iset & small_iset::intersect(const value_range_type & range)
{
    if ( range.first > range.second )
        throw_out_of_range("while inserting a range");
    value_range_type fixed(std::max(range.first,  MIN_DOMAIN),
                           std::min(range.second, MAX_DOMAIN));
    if (fixed.first > fixed.second)
        return clear(); // intersecting with range above MAX or below MIN
    assert_valid_range(fixed, "while intersecting a range");

    Bits zeros_below_range= unsigned_ones_on_and_above<Bits>(fixed.first);
    Bits zeros_above_range= unsigned_ones_on_and_below<Bits>(fixed.second);
    
    m_bits &= (zeros_below_range & zeros_above_range);
    
    return *this;
}

small_iset & small_iset::intersect(const value_range_type & range,
                                   size_type skip_size)
{
    if ( range.first == range.second )
        return intersect(range);
    if ( skip_size == 0)
        throw_out_of_range("while inserting a range");

    value_range_type fixed(range);
    
    if ( fixed.first < MIN_DOMAIN) {
        size_type diff= (size_type)MIN_DOMAIN-fixed.first;
        fixed.first += diff/skip_size*skip_size;
        if (fixed.first < MIN_DOMAIN)
            fixed.first += skip_size;
    }
    if ( fixed.second > MAX_DOMAIN) {
        size_type diff= (size_type)fixed.second-MAX_DOMAIN;
        fixed.second -= diff/skip_size*skip_size;
        if (fixed.second > MAX_DOMAIN)
            fixed.second -= skip_size;
    }
    if ( fixed.first > fixed.second ) {
        clear();
        return *this;
    }
    if (fixed.first > fixed.second)
        return clear(); // intersecting with range above MAX or below MIN
    assert_valid_range(fixed, "while intersecting a range");

    small_iset tmp;
    tmp.insert(fixed, skip_size);
    return intersect(tmp);
}

small_iset & small_iset::intersect(const pair<long long, long long> & range,
                                   size_type skip)
{
    // 'fixed' must not be pair<long long,long long> in order to avoid
    // infinite recursion when intersect(fixed, skip) is called.
    value_range_type fixed; 
    if ( range.first < iset::MIN_DOMAIN ) {
        long long first= range.first;
        long long diff = iset::MIN_DOMAIN - first;
        // Use unsigned type to avoid possible overflow during round-up.
        // After this division no overflow is possible, so go back to signed.
        long long steps= ((unsigned long long)diff+(skip-1))/skip;
        HCSP_ASSERT(first+steps*skip >= iset::MIN_DOMAIN);
        HCSP_ASSERT(first+(steps-1)*skip <= iset::MIN_DOMAIN);
        fixed.first = first+steps*skip;
    } else {
        fixed.first= range.first;
    }
    
    if ( range.second > iset::MAX_DOMAIN) {
        long long second= range.second;
        long long diff = second - iset::MAX_DOMAIN;
        // Use unsigned type to avoid possible overflow during round-up.
        // After this division no overflow is possible, so go back to signed.
        long long steps= ((unsigned long long)diff+(skip-1))/skip;
        HCSP_ASSERT(second-steps*skip <= iset::MAX_DOMAIN);
        HCSP_ASSERT(second-(steps-1)*skip > iset::MAX_DOMAIN);
        fixed.second = second - steps*skip;
    } else {
        fixed.second= range.second;
    }
    return intersect(fixed, skip);
}

small_iset::iterator small_iset::find_contained_tail(const small_iset & other,
                                                     iterator last) const
{
    return iset_helpers::find_contained_tail(*this, last, other);
}

namespace hcsp {
std::ostream & operator<<(std::ostream & out, const small_iset & domain)
{
    print_set(out, domain);
    return out;
}
}

namespace hcsp {
std::istream & operator>>(std::istream & in, small_iset & domain)
{
    read_set(in, domain);
    return in;
}
}

void small_iset::show_self()
{
    show_set(*this);
}

void small_iset::iterator::throw_runtime_error(const char * text) const
    throw(runtime_error)
{
    throw runtime_error(text);
}

small_iset::value_range_type small_iset::iterator::current_range() const
{
    if (m_pIset == NULL)
        throw runtime_error("small_iset::iterator::current_range "
                             "dereferencing NULL iterator");
    if (m_current_bit >= sizeof(Bits)*8)
        throw runtime_error("small_iset::iterator::current_range "
                            "dereferencing empty iterator");
    Bits future= unsigned_ones_above<Bits>(m_current_bit);
    Bits futureHoles= future & ~m_pIset->m_bits;
    const Bits highestBit= Bits(1) << (sizeof(Bits)*8-1);
    Bits lastNonHole = (futureHoles >> 1) | highestBit;
    return make_pair(m_current_bit, find_lsb_index(lastNonHole));
}

const small_iset::iterator & small_iset::iterator::advance_range()
{
    if (m_pIset == NULL)
        throw runtime_error("small_iset::iterator::advance_range "
                             "dereferencing NULL iterator");
    if (m_current_bit >= sizeof(Bits)*8)
        throw runtime_error("small_iset::iterator::advance_range "
                            "dereferencing empty iterator");
    m_current_bit= current_range().second;
    return ++(*this);
}

const small_iset::iterator & small_iset::iterator::retract_range()
{
    if (m_pIset == NULL)
        throw runtime_error("small_iset::iterator::retract_range "
                             "dereferencing NULL iterator");
    if (!middle_of_range())
        --(*this);
    Bits extended_range= unsigned_ones_on_and_above<Bits>(m_current_bit);
    extended_range |= m_pIset->m_bits;
    if (~extended_range == 0)
        m_current_bit= 0;
    else
        m_current_bit= find_msb_index(~extended_range)+1;
    return *this;
}

small_iset::iterator small_iset::iterator::find_stride(
    iterator end, value_range_type &range, size_type & step_size) const
{
    if (m_pIset == NULL)
        throw runtime_error("small_iset::iterator::find_stride "
                            "dereferencing NULL iterator");
    if (m_pIset != end.m_pIset)
        throw runtime_error("small_iset::iterator::find_stride "
                            "refers to different small_iset objects");
        
    if ( m_current_bit >= end.m_current_bit)
        throw runtime_error("Empty an stride was requested");
    step_size= 1;
    range= current_range();
    iterator next= *this;
    if (range.second >= (small_iset::value_type) end.m_current_bit) {
        range.second= end.m_current_bit-1;
        next= end;
        return next;
    }
    iterator prev= *this;
    next.advance_range();
    if (range.first != range.second)
        return next;

    if (next == end)
        return next;

    range.second= *next;
    step_size= range_size(range)-1;
    if (step_size > 0U+ MAX_DOMAIN)
        return next;

    do {
        prev= next;
        range.second= *prev;
        ++next;
    } while (next != end && range_size(range.second, *next)-1 == step_size);
    
    return next;
}

small_iset::iterator small_iset::iterator::find_back_stride(
    iterator begin, value_range_type &range, size_type & step_size) const
{
    if ( *this == begin )
        throw runtime_error("An Empty stride was requested");

    iterator first= *this;
    first.retract_range();
    range= first.current_range();
    if (middle_of_range())
        range.second= m_current_bit-1;
    if (range.first < *begin && range.second >= *begin) {
        range.first= *begin;
        first= begin;
    }
    step_size= 1;
    if (range.first != range.second || first == begin)
        return first;

    --first;
    range.first= *first;
    step_size= range_size(range)-1;
    while (first != begin) {
        --first;
        if (*first + step_size != (size_type) range.first) {
            ++first;
            return first;
        }
        range.first= *first;
    }
    return first;
}


const small_iset::iterator & small_iset::iterator::advance(size_type count)
{
    if (m_pIset == NULL)
        throw runtime_error("small_iset::iterator::advance_range "
                             "dereferencing NULL iterator");
    if (m_current_bit >= sizeof(Bits)*8)
        throw runtime_error("small_iset::iterator::advance_range "
                            "dereferencing empty iterator");

    while (count > 0 && m_current_bit < sizeof(Bits)*8) {
        value_range_type range= current_range();
        size_type diff;
        if (range.second < 0)
            diff= range.second-range.first;
        else
            diff= static_cast<size_type>(range.second)-range.first;
        if (diff < count) {
            count -= diff;
            --count;
            advance_range();
        } else {
            m_current_bit= range.first + count;
            count=0;
        }
    }
    return *this;
}
