// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <solver.hpp>

#include <clause_propagator.hpp>
#include <conflict_constraint_dependency.hpp>
#include <constraint_aged_info.hpp>
#include <var_watchers.hpp>
#include <watch_types.hpp>
#include <constraint_connectivity_info.inl>
#include <constraint_watchless_info.hpp>
#include <proof_collector.hpp>
#include <proof_interpolant_visitor.hpp>
#include <proof_checker_visitor.hpp>
#include <clause_data.hpp>
#include <timer.hpp>
#include "learned_type.hpp"

using namespace std;
using namespace hcsp;

const unsigned solver::UNINITIALIZED_LEVEL;

class solver::timeout_exception : public exception
{
  public:
    timeout_exception() {}
    virtual const char * what() const throw()
    { return "timeout exception"; }
};

//! \brief This functor propagates one constraint.
class solver::propagate_one_constraint_functor
{
  public:
    propagate_one_constraint_functor(solver & the_solver)
        : m_solver(the_solver)
    {}
    
    /**
     * This is a simple wrapper that calls solver::propagate_one_constraint()
     *
     * @return false if propagator returned constraint_propagator::UNSAT,
     *         true otherwise.
     */
    bool operator()(unsigned constraint_index)
    {
        constraint_propagator::result_type result;
        result= m_solver.propagate_one_constraint( constraint_index );
        return result != constraint_propagator::UNSAT;
    }
  private:
    solver & m_solver;

};

//! Upon destruction restore the original number of clauses.
//! This class will remove all new conflict clauses added after the
//! initial state of the solver.
class solver::initial_state_restorer
{
  public:
    initial_state_restorer(solver & the_solver) : m_solver(the_solver) {}
    ~initial_state_restorer()
    {
        m_solver.forget_all_watches();
    }
  private:
    solver & m_solver; 
};

class is_removed_constraint_functor
{
  public:
    /**
     * @param queue   This functor will insert elements into this queue.
     * @param except  If except>=0 then any constraint_index==except will be
     *                ignored.
     */
    is_removed_constraint_functor(
        const vector<constraint_propagator*> & propagators)
        : m_propagators( propagators )
    {}
    bool operator()(unsigned constraint) const
    {
        return (constraint >= m_propagators.size() ||
                m_propagators[constraint] == NULL);
    }
  private:
    const vector<constraint_propagator*> & m_propagators;
}; // is_removed_constraint_functor

solver::solver()
    : m_current_time(2), m_use_variable_queue(true),
      m_report_level(NONE), m_now_solving(false),
      m_backtrack_limit(1000000000), m_backtracks_done(0),
      m_backtracks_for_iteration(0), m_backtracks_since_restart(0),
      m_first_restart_threshold(16), m_restarts_done(0), m_decisions_done(0),
      m_total_combines(0), m_non_fallback_combines(0),
      m_backtracks_without_fallback_combines(0),
      m_current_conflict_analysis_has_fallback(false),      
      m_explainer_check_age(true),
      m_intersect_with_ground(true),
      m_relaxation(COMPLETE_RELAXATION), m_cause_relaxation(false),
      m_constraints_combination(true), m_augment_explanation(true),
      m_only_optimal(false), m_all_results(false), m_produce_proof(false),
      m_validate_unsat(false), m_proof_collector(new proof_collector_dummy),
      m_analysis_cost_factor(2.0), m_conflict_cost_factor(8.0),
      m_opportunistic_back_jumping_threshold(1000),
      m_conflicts_before_boosting_effort(4), m_max_conflict_clause(100),
      m_erase_constraints_every_n_conflicts(256),
      m_keep_constraints_from_n_conflicts(512),
      m_variable_event_comparator(m_assignment_stack),
      m_variable_queue(m_variable_event_comparator),
      m_last_chosen_var(-1), m_user_variables(-1U),
      m_objective_variable(-1), m_is_minimizing(true),
      m_p_out(&cout),
      m_comment("c"), m_summary_prefix("s"), m_objective_prefix("o")
{
    m_assignment_stack.set_cost_lowering_factor(24);
}

/**
 * @brief Possible views that can represent a given propagation explanation.
 *
 * An explanation clause, represented by literals, is always part of these
 * objects. Explanation rules, however, are optional.
 */
struct solver::explanation_views
{
    explanation_views(const std::vector<iset> & literals,
                      std::unique_ptr<explanation_rule> & rule)
        : m_literals(literals), m_rule(rule.release()) {}
    explanation_views(const std::vector<iset> & literals)
        : m_literals(literals) {}

    bool consistent() const
    {
        if (m_rule.get() == NULL)
            return true;
        vector<iset> literals;
        m_rule->get_explanation(literals);
        return literals == m_literals;
    }
        
    const std::vector<iset> & m_literals;
    std::unique_ptr<explanation_rule> m_rule;
};

solver::~solver() throw()
{
    // Offensive programming: Make the collection NULL to catch accidental
    // referencing.
    free_container(m_constraint_info_collection);
    while ( !m_propagator_pool.empty() ) {
        delete m_propagator_pool.back();

        // Offensive programming to catch access after destruction.
        m_propagator_pool.back()= NULL;
        m_propagator_pool.pop_back();
    }
}

void solver::produce_proof(bool val)
{
    if (val)
        m_proof_collector.reset(new proof_collector(*this));
    else
        m_proof_collector.reset(new proof_collector_dummy);
}

constraint_connectivity_info &
solver::add_constraint(std::unique_ptr<constraint_propagator> p_propagator)
{
    HCSP_PROFILE_CURRENT;
    HCSP_ASSERT(m_constraint_info_collection.size()==m_propagator_pool.size());
    m_constraint_info_collection.push_back(
        constraint_connectivity_info(*this, 
                                     m_constraint_info_collection.size()));
    try {
        m_propagator_pool.push_back(NULL);
    } catch(...) {
        m_constraint_info_collection.pop_back();
    }
    m_propagator_pool.back()= p_propagator.release();
    return m_constraint_info_collection.back();
}

unsigned solver::add_var(iset domain, const char * name, bool update_singletons)
{
    m_assignment_stack.add_var(move(domain), name, update_singletons);
    return m_assignment_stack.count_variables()-1;
}

unsigned solver::singleton_var(iset::value_type value)
{
    return m_assignment_stack.singleton_var(value);
}

bool solver::propagate_constraint_in_constraint_queue()
{
    HCSP_PROFILE_CURRENT;
    while ( !m_constraint_queue.empty() ) {
        m_current_queue= constraint_runtime_info::CONSTRAINT_QUEUE;
        unsigned constraint_index= m_constraint_queue.pop();
        constraint_propagator::result_type result;
        result = propagate_one_constraint(constraint_index);
        if ( result == constraint_propagator::UNSAT ) 
            return false; 
        if ( !m_variable_queue.empty() ) // propagation may have modified vars
            return true;
    } 
    return true; 
} 

iset solver::removed_values(unsigned var_index,
                            change_index_t modification_index) const
{
    iset removed= m_assignment_stack.value_before_modification(
        modification_index);
    removed.erase_all_of(m_assignment_stack.var_domain(var_index));
    return removed;
}

bool solver::propagate_constraints_in_variable_queue()
{
    HCSP_PROFILE_CURRENT;
    m_current_queue= constraint_runtime_info::VARIABLE_QUEUE;
    while ( !m_variable_queue.empty() ) {
        const auto & var_info = m_variable_queue.top();
        const unsigned variable= var_info.m_variable;
        change_index_t modification= var_info.m_modification;
        m_variable_queue_handles[variable].second = false;

        m_variable_queue.pop();
        if (!propagate_constraints_touching_var(variable, modification))
            return false; 
    }
    return true;
}
 
bool solver::propagate_constraints_touching_var(unsigned variable,
                                                change_index_t modification)
{
    auto & value_watch_data= m_var_value_watchers[variable];
    if ( !value_watch_data.empty() ) {
        if (! value_watch_data.for_each_until_false(
                propagate_one_constraint_functor(*this), 
                [=](iset::value_type key) {
                    const iset & before= 
                        m_assignment_stack.value_before_modification(
                            modification);
                    const iset & after= m_assignment_stack.var_domain(variable);
                    return before.contains(key) && !after.contains(key);
                }))
            return false; 
    }
    
    auto & watch_data= m_var_watchers[variable];
    if ( watch_data.empty() )
        return true;
    small_iset types_to_watch;
    types_to_watch.insert(MODIFICATION_WATCH);
    if ( var_domain(variable).singleton() )
        types_to_watch.insert(FINALIZATION_WATCH);
    if ( watch_data.watches(LOWER_BOUND_WATCH) &&
         m_assignment_stack.lower_bound_changed(variable, modification) )
        types_to_watch.insert(LOWER_BOUND_WATCH);
    if ( watch_data.watches(UPPER_BOUND_WATCH) &&
         m_assignment_stack.upper_bound_changed(variable, modification) )
        types_to_watch.insert(UPPER_BOUND_WATCH);
        
    return watch_data.for_each_until_false(
        propagate_one_constraint_functor(*this), 
        [=](iset::value_type key) { return types_to_watch.contains(key); });
}


bool solver::propagate_dirty_learned_constraint(
    constraint_run_level & constraint_relevance)
{
    // We will act as if the constraints are activated by the
    // CONSTRAINT_QUEUE which does not depend on the state of the watches.
    m_current_queue= constraint_runtime_info::CONSTRAINT_QUEUE;
    
    unsigned constraint_id= constraint_relevance.m_constraint;
    
    bool initial_report= (report_level() >= SEARCH);
    unsigned level= m_assignment_stack.level();
    
    if (constraint_relevance.m_run_level == UNINITIALIZED_LEVEL) {
        change_index_t mod= m_assignment_stack.latest_modification();
        if (level != 0)
            // In order for a clause to get correct watches we go even more
            // backwards than the current location.
            mod= m_assignment_stack.decision_modification(level)-1;
        if (!init_one_constraint(constraint_id, mod, true))
            return false;
    }
    constraint_propagator::result_type result;
    result= propagate_one_constraint(constraint_id, initial_report);
    constraint_relevance.m_run_level= level;
    return result != constraint_propagator::UNSAT;
}

bool solver::propagate_constraints()
{
    unsigned level= m_assignment_stack.level();
    unsigned first= m_learned_rerun_level.size();
    bool success= true;
    while (success && first > 0 &&
           level < m_learned_rerun_level[first-1].m_run_level) {
        --first;
        if (!propagate_dirty_learned_constraint(m_learned_rerun_level[first]))
            success= false;
    }
    // Erasing:
    if (level == 0)
        m_learned_rerun_level.clear();
    unsigned last= first;
    for (unsigned i= first ; i != m_learned_rerun_level.size() ; ++i) {
        unsigned constraint_id= m_learned_rerun_level[i].m_constraint;
        unsigned arity= m_constraint_info_collection[constraint_id].arity();
        if (!is_clause(constraint_id) && arity <= m_max_conflict_clause) {
            m_learned_rerun_level[last]= m_learned_rerun_level[i];
            ++last;
        }
    }
    m_learned_rerun_level.erase(m_learned_rerun_level.begin()+last,
                                m_learned_rerun_level.end());
    
    while (!m_constraint_queue.empty() || !m_variable_queue.empty()) {
        if ( !propagate_constraints_in_variable_queue() )
            return false;
        if ( !propagate_constraint_in_constraint_queue() )
            return false;
    }
    return true;
}

bool solver::init_one_constraint(unsigned constraint_index,
                                 change_index_t modification_of_reference_state,
                                 bool is_learned)
{
    constraint_propagator & propagator= *m_propagator_pool[constraint_index];

    constraint_connectivity_info & current_info=
        m_constraint_info_collection[constraint_index];

    HCSP_ASSERT( propagator.count_params() < 0 ||
                (unsigned)propagator.count_params() == current_info.arity() );
    bool init_result= false;
    if ( is_learned && current_info.arity() > m_max_conflict_clause ) {
        constraint_watchless_info watchless(current_info);
        init_result= propagator.init( watchless );
        propagator.forget_watches();
    } else {
        constraint_aged_info aged_info(current_info, m_assignment_stack,
                                       modification_of_reference_state+1);
        init_result= propagator.init( aged_info );
    }
    current_info.reset();
    m_last_constraint= constraint_index; // Required by proof generation

    if (init_result)
        return true;
    else {
        if ( report_level() >= SEARCH ) {
            cout << m_current_time << " UNSAT ";
            propagator.print(cout, &current_info);
            cout << "\n";
        }
        return false;
    }
}

void solver::initial_intersect_with_ground(unsigned constraint_index)
{
    if (!should_intersect_with_ground())
        return;
    constraint_propagator & propagator= *m_propagator_pool[constraint_index];

    constraint_connectivity_info &
        connectivity= m_constraint_info_collection[constraint_index];

    vector<unsigned> arg_mapping;
    unique_ptr<constraint_propagator> intersected(
        propagator.intersect_with(connectivity, arg_mapping));
    if (!intersected.get())
        return;
    if (report_level() >= PROPAGATION_MODS) {
        cout << "Simplifying ";
        propagator.print(cout, &connectivity);
        cout << " using initial domains\n";
        for (unsigned i=0 ; i != connectivity.arity() ; ++i) {
            if (i != 0)
                cout << ", ";
            cout << var_printer(connectivity.to_variable(i))
                 << "={" << connectivity.domain(i) << "}";
        }
        cout << " ";
    }
    bool same_args= (arg_mapping.size() == connectivity.arity());
    for (unsigned i=0; same_args && i < arg_mapping.size() ; ++i) {
        same_args= (arg_mapping[i] == i);
    }
    if (!same_args) {
        constraint_connectivity_info new_connectivity(*this, constraint_index);
        for (auto arg : arg_mapping)
            new_connectivity << connectivity.to_variable(arg);
        connectivity= move(new_connectivity);
    }
    delete m_propagator_pool[constraint_index];
    m_propagator_pool[constraint_index]= intersected.release();
    if (report_level() >= PROPAGATION_MODS) {
        cout << "to ";
        m_propagator_pool[constraint_index]->print(cout, &connectivity);
        cout << "\n";
    }
}

bool solver::first_time_propagate_constraints()
{
    HCSP_PROFILE_CURRENT;
    unique_queue<> first_constraint_queue;

    m_var_watchers.clear();
    m_var_watchers.resize(m_assignment_stack.count_variables());
    m_var_value_watchers.clear();
    m_var_value_watchers.resize(m_assignment_stack.count_variables());
    m_assignment_stack.reset_var_costs(m_analysis_cost_factor);
    m_assignment_stack.disable_var_costs_heap();
    m_current_queue= constraint_runtime_info::CONSTRAINT_QUEUE;
    m_original_constraint_count= m_propagator_pool.size();
    m_conflict_constraint_dependencies.reset(
        new conflict_constraint_dependency(m_original_constraint_count));
    produce_proof(m_proof_collector->want_proof()); // forget stored constraints

    for ( unsigned i=0 ; i < m_propagator_pool.size() ; ++i) {
        if ( !m_propagator_pool[i] ) // propagator was erased
            continue;
        initial_intersect_with_ground(i);
        if ( !init_one_constraint(i, m_assignment_stack.latest_modification()))
            return false;

        constraint_propagator::result_type result= propagate_one_constraint(i);
        if ( result == constraint_propagator::UNSAT )
            return false;
        while ( !m_constraint_queue.empty() ) {
            HCSP_ASSERT( constraint_queue_enabled() );
            unsigned constraint_index= m_constraint_queue.pop();
            if (constraint_index <= i) {
                first_constraint_queue.
                    set_max_elements(m_constraint_queue.get_max_elements());
                first_constraint_queue.insert(constraint_index);
            }
        }
    }
    if ( !first_constraint_queue.empty() )
        m_constraint_queue.swap(first_constraint_queue);
    
    if ( !propagate_constraints() )
        return false;

    for (unsigned i=0 ; i != m_constraint_info_collection.size() ; ++i)
        add_constraint_cost_to_variables(i, 1.0);

    m_assignment_stack.enable_var_costs_heap();

    return true;
}

unsigned solver::count_irredundant_literals(const std::vector<iset> &literals,
                                            constraint_connectivity_info &info)
    const
{
    unsigned count=0;
    for (unsigned i=0 ; i!= literals.size() ; ++i) {
        unsigned variable= info.to_variable(i);
        const iset & ground_domain= var_ground_level_domain(variable);
        if ( literals[i].contains_any_of(ground_domain))
            ++count;
    }
    return count;
}

void solver::add_constraint_cost_to_variables(
    unsigned constraint_index, double base,
    const std::vector<iset> * p_literals)
{
    constraint_connectivity_info &
        info= m_constraint_info_collection[constraint_index];

    unsigned arity= info.arity();
    double cost;
    if (!p_literals && is_clause(constraint_index)) {
        p_literals= &clause_literals(constraint_index);
    }
    cost= base;
    if (p_literals && p_literals->size() < arity)
        arity= p_literals->size();
    
    for (unsigned i = arity; i-- != 0 ;) {
        unsigned variable= info.to_variable(i);
        if (p_literals) {
            const iset & ground_domain= var_ground_level_domain(variable);
            if (! (*p_literals)[i].contains_any_of(ground_domain))
                continue;
        }
        m_assignment_stack.add_var_cost(variable, cost);
    }
}

void solver::report_last_propagation(constraint_propagator::result_type result,
                                     bool report_removal) const
{
    if ( report_level() < SEARCH && !report_removal)
        return;
    bool unsat= (result == constraint_propagator::UNSAT);
    if ( report_level() <= SEARCH && !unsat && !report_removal)
        return;
    const constraint_connectivity_info & info=
        m_constraint_info_collection[m_last_constraint];

    const constraint_propagator & propagator=
        *m_propagator_pool[m_last_constraint];

    unique_ptr<constraint_aged_info> before_propagation;
    if (info.last_activation_reduced())
        before_propagation.reset(
            new constraint_aged_info(info, m_assignment_stack,
                                     m_assignment_stack.latest_modification()));
    if ( report_level() >= PROPAGATED_CONSTRAINTS || unsat) {
        cout << m_current_time << " " << result << " ";
        if (before_propagation.get())
            propagator.print(cout, before_propagation.get());
        else
            propagator.print(cout, &info);
            
        cout << "\n";
    }
    if ( (report_level() >= PROPAGATION_MODS &&
          info.last_activation_reduced()) || report_removal) {
        for ( unsigned i=0 ; i != info.arity() ; ++i) {
            if ( !info.reduced_by_last_activation(i) )
                continue;
            cout << m_current_time << " constraint[" << m_last_constraint 
                 << "] changed param[" << i << "] "
                 << var_printer(info.to_variable(i))
                 << " := {" << info.domain(i) << "}";
            if (report_removal && before_propagation.get()) {
                iset removed= before_propagation->domain(i);
                removed.erase_all_of(info.domain(i));
                cout << " removed " << "{" << removed << "}";
                
            }
            cout << "\n";
        }
    }
}

constraint_propagator::result_type
solver::propagate_one_constraint(unsigned constraint_index, bool report_removal)
{
    HCSP_PROFILE_CURRENT;
    constraint_connectivity_info & info=
        m_constraint_info_collection[constraint_index];
    info.prepare_next_activation();

    // The use of volatile target_time lets the debuggers modify target_time
    // and then break on the "cout <<...." line. This is much faster
    // than the regular conditional breakpoints of GDB.
    static volatile unsigned target_time=-1U;
    if (m_current_time == target_time)
        cout << m_current_time << " reached target time\n";
    constraint_propagator & propagator= *m_propagator_pool[constraint_index];
    constraint_propagator::result_type result= propagator.propagate(info);
    if ( result == constraint_propagator::VALID &&
         // make sure that we use constraint queue and not variable queue
         constraint_index < m_constraint_queue.get_max_elements() ) 
        m_constraint_queue.prohibit_insertion(constraint_index);
    
    if (result != constraint_propagator::POSTPONE)
        info.finalize_prev_activation(m_current_time);
    m_last_constraint= constraint_index;
    if ( report_level() >= SEARCH || report_removal)
        report_last_propagation(result, report_removal);
    if ( result != constraint_propagator::POSTPONE)
        advance_current_time();
    if (info.last_activation_reduced())
        mark_used(constraint_index);
    
    return result;
}

void solver::mark_used(unsigned constraint_index)
{
    if (constraint_index < m_original_constraint_count)
        // can't mark m_backtrack_count_of_cc_use, since not a conflict
        // constraint
        return;
    m_conflict_constraint_dependencies->mark_used(constraint_index,
                                                  m_backtracks_for_iteration);
}

void solver::mark_dependency(unsigned prerequisite_constraint,
                             unsigned consequent_constraint)
{
    if (prerequisite_constraint < m_original_constraint_count)
        return;
    m_conflict_constraint_dependencies->mark_dependency(prerequisite_constraint,
                                                        consequent_constraint);
}


void solver::forget_conflict_constraint(unsigned constraint_index)
{
    HCSP_ASSERT(constraint_index >=m_original_constraint_count);
    constraint_propagator *p_propagator= m_propagator_pool[constraint_index];
    if (report_level() >= PROPAGATION_MODS)
        cout << "removing rare conflict constraint[" << constraint_index
             << "]\n";
    m_propagator_pool[constraint_index]= NULL;
    m_constraint_info_collection[constraint_index].free();
    delete p_propagator;

    m_conflict_constraint_dependencies->erase_constraint(constraint_index);
}

void solver::put_in_queue(unsigned constraint_index,
                          constraint_runtime_info::queue_type queue)
{
    if ( queue != constraint_runtime_info::CONSTRAINT_QUEUE)
        throw runtime_error("solver::put_in_queue() "
                            "works only for CONSTRAINT_QUEUE");
    if ( m_constraint_queue.get_max_elements() < m_propagator_pool.size() )
        m_constraint_queue.set_max_elements(m_propagator_pool.size());
    m_constraint_queue.insert(constraint_index);
}

void solver::print_recent_proof(std::ostream  & out) const
{
    m_proof_collector->print_proof(out);
}

formula solver::get_interpolant() const
{
    if (! m_proof_collector->want_proof() )
        throw runtime_error("solver::print_interpolant() "
                            "works only when proof mechanism is activated");

    // this should not throw std::bad_cast because want_proof is true iff
    // this is a proof_collector object.
    proof_collector &
        collector= dynamic_cast<proof_collector&>(*m_proof_collector);
    
    proof_interpolant_visitor visitor(collector, *this);
    m_proof_collector->traverse_proof(visitor);

    return visitor.get_interpolant();
}

bool solver::is_correctly_unsat() const
{
    //! \todo Need to account for trivially UNSAT csp which has an empty domain
    //!       and it exits before any constraint is derived.
    
    if (! m_proof_collector->want_proof() )
        throw runtime_error("solver::is_correctly_unsat() "
                            "works only when proof mechanism is activated");

    // this should not throw std::bad_cast because want_proof is true iff
    // this is a proof_collector object.
    proof_collector &
        collector= dynamic_cast<proof_collector&>(*m_proof_collector);
    
    proof_checker_visitor visitor(collector, *m_p_out);
    m_proof_collector->traverse_proof(visitor);

    return visitor.is_correct();
}

int solver::choose_free_variable()
{
    HCSP_PROFILE_CURRENT;
    if ( !m_assignment_stack.has_free_vars())
        return -1;
    int best_variable= m_last_chosen_var;
    unsigned write_pos=0;
    for ( unsigned read_pos=0;
          read_pos != m_analyzed_variables.size();
          ++read_pos ) {
        unsigned variable= m_analyzed_variables[read_pos];
        if (m_assignment_stack.var_domain(variable).singleton())
            continue;
        if (write_pos != read_pos)
            m_analyzed_variables[write_pos] = m_analyzed_variables[read_pos];
        ++write_pos;
        if ( m_assignment_stack.is_better_variable( variable, best_variable))
            best_variable= variable;
    }
    m_analyzed_variables.erase(m_analyzed_variables.begin()+write_pos,
                               m_analyzed_variables.end());
    int best_rest_variable = m_assignment_stack.find_best_free_variable();
    if (m_assignment_stack.is_better_variable(best_rest_variable, 
                                              best_variable,
                                              m_conflict_cost_factor))
        best_variable= best_rest_variable;
    m_last_chosen_var= best_variable;
    return best_variable;
}

void solver::restrict_variable(unsigned var_index, bool fastest)
{
    ++m_decisions_done;
    if (var_index == 154943)
    {
        for (unsigned i = 1 ; 
             i < m_assignment_stack.m_var_costs_heap.size();
             ++i)
            HCSP_ASSERT(m_assignment_stack.is_better_variable(
                            var_index, m_assignment_stack.m_var_costs_heap[i]));
    }
        
    pair<bool, iset::value_type> prev=
        m_assignment_stack.previously_assigned_var_value(var_index);
    const iset & current= m_assignment_stack.var_domain(var_index);
    HCSP_ASSERT(!current.empty());
    iset::value_type decision_value= iset::MIN_DOMAIN; // shut up GCC warnings

    if ( (!fastest || !prev.first) && m_objective_variable >=0 &&
         var_index == (unsigned)m_objective_variable ) {
        if (m_is_minimizing)
            decision_value= current.min();
        else
            decision_value= current.max();
    } else if ( prev.first) {
        if (current.contains(prev.second))
            decision_value= prev.second;
        else {
            // Find the closest bound
            iset::value_type first_third= (current.min()*2LL+current.max())/3;
            iset::value_type two_thirds= (current.max()*2LL+current.min())/3;
            if (prev.second <= first_third)
                decision_value= current.min();
            else if (prev.second > two_thirds)
                decision_value= current.max();
            else
                decision_value= random::gen(2) ? current.min() : current.max();
        }
    } else {
        decision_value= current.min();
    } 
    
    if (report_level() >= SEARCH)
        cout << m_current_time << " restrict:: " << var_printer(var_index)
             << " := " << decision_value <<"\n";
    m_assignment_stack.var_decision(var_index, m_current_time,
                                    iset(decision_value));
    HCSP_ASSERT(m_backtracks_starting_level.size()==m_assignment_stack.level());
    m_backtracks_starting_level.push_back(m_backtracks_since_restart);

    notify_modification(var_index);
    advance_current_time();
}

void solver::complement_decision(unsigned decision_variable,
                                 const iset & decision_values)
{
    if (report_level() >= SEARCH)
        cout << m_current_time << " Remove bad values :: "
             << var_printer(decision_variable)
             << " in {" << decision_values <<"}\n";
    iset next_domain = var_domain(decision_variable);
    next_domain.erase_all_of(decision_values);
    set_var_domain(decision_variable, move(next_domain));
    notify_modification(decision_variable);
    advance_current_time();
    
    if ( !var_domain(decision_variable).singleton() )
        restrict_variable(decision_variable, false);
}

void solver::deallocate_watch(unsigned variable_index, watch_type type,
                              watch_handle_t watch_handle)
{
    m_var_watchers.at(variable_index).deallocate_watch(type, watch_handle);
}

void solver::deallocate_value_watch(unsigned variable_index,
                                    iset::value_type value,
                                    watch_handle_t watch_handle)
{
    auto & watchers= m_var_value_watchers.at(variable_index);
    watchers.deallocate_watch(value, watch_handle);
}

watch_handle_t solver::allocate_value_watch(unsigned var_index, 
                                            iset::value_type value,
                                            unsigned constraint_index)
{
    auto & watchers= m_var_value_watchers.at(var_index);
    return watchers.allocate_watch( value, constraint_index );
}

watch_handle_t solver::allocate_watch(unsigned variable_index, watch_type type,
                                      unsigned constraint_index)
{
    auto & watchers= m_var_watchers.at(variable_index);
    return watchers.allocate_watch(type, constraint_index);
}

void solver::validate_flags_consistency() const
{
    if (should_intersect_with_ground() && m_proof_collector->want_proof())
        //! \todo Fix proofs to work with should_intersect_with_ground()
        throw runtime_error("Proofs are invalid with "
                            "should_intersect_with_ground()");
}

void solver::advance_iteration_time(timer& iteration_time)
{
    timer current_time;
    if (!m_only_optimal)
        cout << m_comment << " === iteration time "
             << current_time.cpu_diff_str(iteration_time) << endl;
    iteration_time= current_time;
}

bool solver::improve_solution(timer & iteration_time,
                              solver::result_type & result)
{
    save_solution_as_best();
    reset();
    iset objective_dom= var_domain(objective_variable());
    if (objective_is_minimizing())
        objective_dom.erase(make_pair(best_objective(), iset::MAX_DOMAIN));
    else
        objective_dom.erase(make_pair(iset::MIN_DOMAIN, best_objective()));
    if (objective_dom.empty()) {
        cout << m_comment << " ==== no more values to try ====" << endl;
        result= SAT;
        return false;
    }
    iset::value_type best_possible= objective_dom.max();
    const char *beyond= " above ", *what = "maximize ";
    if (objective_is_minimizing()) {
        best_possible= objective_dom.min();
        beyond= " below ";
        what= "minimize ";
    }
    set_var_domain(objective_variable(), move(objective_dom));
    if (!m_only_optimal)
        cout << m_comment << " ==== trying to " << what
             << var_printer(objective_variable())
             << beyond <<  best_objective() << ", but not" << beyond
             << best_possible << " ==" << endl;
    result= solve_current_constraints(false);
    advance_iteration_time(iteration_time);
    return result == SAT;
}

void solver::manage_unsat()
{
    if (m_produce_proof)
        print_recent_proof(*m_p_out);
    if (m_validate_unsat && !is_correctly_unsat())
        throw std::runtime_error("Bad UNSAT result");
}

solver::result_type solver::solve(ostream * p_out)
{
    m_backtracks_done= 0;
    m_restarts_done= 0;
    m_decisions_done= 0;
    m_total_combines= 0;
    m_non_fallback_combines= 0;
    m_backtracks_without_fallback_combines= 0;
    m_current_conflict_analysis_has_fallback= false;
    m_p_out= p_out ? p_out : &cout;
    if (has_empty_domains())
        return UNSAT;
    if (m_objective_variable >= 0)
        return optimize();

    result_type result= solve_current_constraints(false);
    if (result == UNSAT)
        manage_unsat();
    if (result != SAT || !m_all_results)
        return result;
    
    *p_out << "\n";
    while (result == SAT) {
        print_solution(*p_out);
        unique_ptr<clause_propagator> clause(new clause_propagator);
        vector<iset> & anti_solution = clause->update_literals();
        for (unsigned i=0 ; i != count_variables() ; ++i) {
            anti_solution.push_back(iset(var_domain(i)).complement());
        }
        constraint_connectivity_info & info= add_constraint(move(clause));
        for (unsigned i=0 ; i != anti_solution.size() ; ++i)
            info << i;
        reset();
        result= solve_current_constraints(false);
    }
    reset();
    return SAT_DONE;
}

solver::result_type solver::optimize()
{
    if (!m_only_optimal)
        cout << m_comment << " ==== Solving SAT problem at first ====" << endl;
    timer iteration_time;

    result_type result= solve_current_constraints(true);
    advance_iteration_time(iteration_time);
    if (result != SAT) {
        reset(); 
        return result;
    }
    cout << m_summary_prefix << " SATISFIABLE\n";
    while (improve_solution(iteration_time, result))
        ;
    if (result != SAT) {
        if (result == UNSAT)
            cout << m_comment << " ==== went past best value ====" << endl
                 << m_summary_prefix << " OPTIMUM FOUND\n";
        else
            cout << m_comment
                 << " ==== run did not succeed, can't guarantee optimality "
                 << "====" << endl;
        reset();
    }
    restore_best_solution();
    if (m_only_optimal)
        print_solution(cout);
    return OPTIMAL;
}

solver::result_type solver::solve_current_constraints(bool fastest)
{
    validate_flags_consistency();
    if (has_empty_domains())
        return UNSAT;
    m_current_time= 2;
    try {
        if (report_level() >= SEARCH) {
            m_assignment_stack.print_singleton_variables(cout);
            cout << endl;
        }
        result_type result= solve_impl(fastest);
        if (result != SAT)
            m_proof_collector->prepare_to_print();
        return result;
    } catch (timeout_exception & e) {
        return TIMEOUT;
    } catch (std::bad_alloc & e) {
        return MEMOUT;
    }
}

void solver::prepare_for_solve()
{
    m_next_restart_threshold= m_first_restart_threshold;
    m_last_chosen_var= -1;
    m_analyzed_variables.clear();
    m_backtracks_starting_level.clear();
    m_backtracks_starting_level.push_back(0);
    
    if ( constraint_queue_enabled() )
        m_constraint_queue.set_max_elements(m_propagator_pool.size());
    m_variable_queue.clear();
    m_variable_queue_handles.clear();
    m_variable_queue_handles.resize(count_variables());
    for (auto & elem : m_variable_queue_handles)
    {
        elem.second = false;
    }
    m_constraint_queue.clear();
    
    m_backtracks_for_iteration=0;
    m_backtracks_since_restart=0;
}

solver::result_type solver::solve_impl(bool fastest)
{
    HCSP_PROFILE_CURRENT;
    restore_on_exit<bool> restorer(m_now_solving);
    m_now_solving= true;
    prepare_for_solve();
    if (!first_time_propagate_constraints()) {
        if ( m_proof_collector->want_proof() ) {
            int inserted_conflict_constraint= -1;
            analyze_conflict(inserted_conflict_constraint);
        }
        return UNSAT;
    }

    initial_state_restorer initial_state(*this);
    while (true) {
        if ( !propagate_constraints() ) {
            int inserted_conflict_constraint= -1;
            int backtrack_level= analyze_conflict(inserted_conflict_constraint);
            if (backtrack_level < 0)
                return UNSAT;
            result_type result= backtrack(backtrack_level,
                                          inserted_conflict_constraint);
            if ( result != SAT )
                return result;
            if (inserted_conflict_constraint >= 0)
                m_learned_rerun_level.push_back(
                    constraint_run_level{UNINITIALIZED_LEVEL,
                            (unsigned)inserted_conflict_constraint});
            continue;
        }
        int var= choose_free_variable(); 
        if (var < 0) 
            return SAT; 
        restrict_variable(var, fastest);
        if ( report_level() >= SEARCH ) 
            cout << "choice depth = " << m_assignment_stack.level() << "\n"; 
    } 
}

bool solver::has_empty_domains() const
{
    for (unsigned i=0 ; i != count_variables() ; ++i) {
        if ( var_domain(i).empty() )
            return true;
    }
    return false;
}

// FIXME: enhance to work with:
//           Rules other than resolution 
//           For example: 
//                 a!=b   b!=c   c!=a
//                 -------------------
//                 all_different(a,b,c)
int solver::analyze_conflict(int & inserted_constraint_index)
{
    HCSP_PROFILE_CURRENT;
    inserted_constraint_index= -1;
    if ( m_assignment_stack.level() == 0 && !m_proof_collector->want_proof())
        return -1;
    const int default_level = static_cast<int>( m_assignment_stack.level() )-1;
    m_assignment_stack.erase_conflicting_modifications(m_last_constraint);

    learned_type learned_constraint(m_assignment_stack, *this); 
    if (!generate_conflict_constraint(learned_constraint))
        return default_level;
    
    int asserting_level= learned_constraint.asserting_level();
    
    if (asserting_level >= 0 && should_intersect_with_ground() )
        learned_constraint.intersect_with_ground();

    inserted_constraint_index= insert_constraint(learned_constraint);
    m_proof_collector->name_conflict_clause(inserted_constraint_index);

    return find_backtrack_level(learned_constraint, asserting_level);
}

int solver::find_backtrack_level(const learned_type & learned_constraint,
                                 int asserting_level) const
{
    if (learned_constraint.unsat())
        return -1;
    
    // If the clause is too big it will run only once, but it makes no sense to
    // do the single run at the alternative level (which doesn't propagate)
    if ( learned_constraint.size() > m_max_conflict_clause)
        return asserting_level;
    
    if ( asserting_level+0U < m_opportunistic_back_jumping_threshold)
        return asserting_level;
    
    int alternative_level= learned_constraint.opportunistic_level();
    unsigned distance= asserting_level-alternative_level;
    if (distance >= m_opportunistic_back_jumping_threshold)
        return alternative_level;
    else
        return asserting_level;
}

bool solver::generate_conflict_constraint(learned_type &learned_constraint)
{
    const vector<iset> *p_literals= NULL;
    if (learned_constraint.has_explanation())
        p_literals= & learned_constraint.get_explanation();
    add_constraint_cost_to_variables(m_last_constraint, 
                                     m_assignment_stack.cost_additive(),
                                     p_literals);
    learned_constraint.clear();
    m_analyzed_variables.clear();
    m_current_conflict_analysis_has_fallback= false;
    
    mark_used(m_last_constraint);
    mark_used(m_propagator_pool.size());
    mark_dependency(m_last_constraint, m_propagator_pool.size());
    m_constraint_info_collection[m_last_constraint].mark_conflicting();
    
    explanation_views conflicting_explanation= get_constraint_explanation(
        m_last_constraint, -1U, iset(),
        conflicting_modification_entry());
    collect_conflict_nodes(learned_constraint,
                           conflicting_explanation, m_last_constraint);
    bool success= resolve_conflicting_implications(learned_constraint);
    if (!m_current_conflict_analysis_has_fallback)
        ++m_backtracks_without_fallback_combines;
    
    if (!success)
        return false;
    if (learned_constraint.is_existing_constraint()) {
        // The generated explanation we used in the code above was stronger than
        // the original constraint. That's why we are trying to learn an
        // explanation of existing constraint.
        learned_constraint.convert_to_clause();
        learned_constraint.relax_if_required();
    }
    return true;
}

bool solver::resolve_conflicting_implications(learned_type & learned_constraint)
{
    // It is possible that learned_constraint is already a UIP when the
    // explanation is more accurate than the propagation, and the conflicting
    // constraint should have conflicted earlier.
    change_index_t prev_last_modified= -2U;
    while ( !learned_constraint.conflict_analysis_complete() ) {
        unsigned node= learned_constraint.latest_node_index();
        change_index_t last_modified= learned_constraint.modification(node);
        int cause= m_assignment_stack.modifying_constraint(last_modified);
        if ( cause < 0 ) {
            if ( report_level() >= PROPAGATION_MODS )
                cout << "No cause for "
                     << var_printer(learned_constraint.variable(node)) << "\n";
            return false;    // Probably optimized away (not inserted)
        }
        mark_used(cause);
        mark_dependency(cause, m_propagator_pool.size());

        if (prev_last_modified == last_modified)
            throw runtime_error("The conflict analysis loops infinitely");
        prev_last_modified= last_modified;
        resolve(learned_constraint, learned_constraint.variable(node), cause);
        HCSP_ASSERT(learned_constraint.conflicts());
#ifndef NDEBUG
        learned_constraint.validate();
#endif
    }
    learned_constraint.add_vars_potentially_falsifying(m_analyzed_variables);
    
    m_conflict_constraint_dependencies->
        compact_prerequisites(m_propagator_pool.size());
    return true;
}

bool solver::compute_restart(unsigned level)
{
    if (level == 0)
        return false;
    if ( m_next_restart_threshold == 0 )
        return false;
    unsigned level_backtracks= (m_backtracks_since_restart -
                                m_backtracks_starting_level.at(level));
    if ( level_backtracks < m_next_restart_threshold )
        return false;
    m_next_restart_threshold= (m_next_restart_threshold*3+1)/2;
    if (report_level() >= SEARCH)
        cout << "Next restart after "  << m_next_restart_threshold
             << " backtracks\n";
    return true;
}

void solver::restart()
{
    for (unsigned i= 1 ; i <= m_assignment_stack.level() ; ++i) {
        unsigned decision_variable= m_assignment_stack.decision_variable(i);
        m_analyzed_variables.push_back(decision_variable);
    }
    m_last_chosen_var= -1;
    m_assignment_stack.undo_level(1);
    m_assignment_stack.forget_previously_assigned_var_values();
    ++m_restarts_done;
    m_backtracks_since_restart= 0;
    if ( report_level() >= SEARCH )
        cout << m_current_time << " restart " << m_restarts_done << "\n";
    
    if ( report_level() >= PROPAGATION_MODS )
        show_self();
    if ( report_level() >= SEARCH )
        cout << "choice depth = 0\n";
    m_backtracks_starting_level.clear();
    m_backtracks_starting_level.push_back(0);
}

solver::result_type solver::backtrack(unsigned level,
                                      int inserted_conflict_constraint) 
{
    for (auto elem : m_variable_queue)
        m_variable_queue_handles[elem.m_variable].second = false;
    m_variable_queue.clear();
    m_constraint_queue.clear();

    ++m_backtracks_done;
    ++m_backtracks_for_iteration;
    ++m_backtracks_since_restart;

    if (m_backtracks_for_iteration > m_backtrack_limit)
        return BACKTRACK_LIMIT;
    
    m_assignment_stack.lower_var_costs();

    remove_unused_conflict_constraints_maybe();

    if ( compute_restart(level) ) {
        restart();
        return SAT;
    }
    
    unsigned var= m_assignment_stack.decision_variable(level+1);
    iset::value_type value= m_assignment_stack.decision_value(level+1);

    m_assignment_stack.undo_level(level+1);
    m_backtracks_starting_level.erase(
        m_backtracks_starting_level.begin()+level+1,
        m_backtracks_starting_level.end());
    if ( report_level() >= SEARCH )
        cout << m_current_time << " backtrack " << m_backtracks_done
             << " undo choice " << var_printer(var) << " = " << value << "\n";
    
    if (inserted_conflict_constraint < 0)
        complement_decision(var, iset(value));
    if ( report_level() >= PROPAGATION_MODS )
        show_self();
    if ( report_level() >= SEARCH )
        cout << "choice depth = " << m_assignment_stack.level() << "\n";
    return SAT;
}

void solver::save_erased_prerequisites_for_proof(
    const vector<unsigned> &clauses_to_keep)
{
    HCSP_ASSERT( m_conflict_constraint_dependencies.get() );
    if ( !m_proof_collector->want_proof() )
        return;
    conflict_constraint_dependency &
        dependencies= *m_conflict_constraint_dependencies;

    // Must record *all* prerequisites because we don't know which of them
    // will be erased in the future. When one of these is erased later then
    // we don't have a way to know that a clause is a prerequisite to a stored
    // clause.
    vector<unsigned>
        prerequisites= dependencies.prerequisite_closure(clauses_to_keep);

    for (unsigned i=0 ; i != prerequisites.size() ; ++i) {
        const constraint_connectivity_info &
            info= m_constraint_info_collection.at(prerequisites[i]);
        
        const clause_propagator::literals_type &
            literals= clause_literals(prerequisites[i]);

        HCSP_ASSERT(literals.size() == info.arity());
        clause_data clause(literals.size());
        for (unsigned j=0 ; j != literals.size() ; ++j) {
            clause[j].m_variable= info.to_variable(j);
            clause[j].m_set= literals[j];
        }
        m_proof_collector->save_clause(prerequisites[i], clause);
    }
}

void solver::remove_unused_conflict_constraints_maybe()
{
    if ( m_backtracks_for_iteration%erase_constraints_every_n_conflicts() != 0)
        return;
    if ( m_backtracks_for_iteration < keep_constraints_from_n_conflicts() )
        return;
    
    // Get constraints whose propagation is recorded in the assignment stack
    unsigned first_conflict= m_original_constraint_count;
    vector<unsigned> busy_constraints=
        m_assignment_stack.get_busy_constraints(first_conflict);

    HCSP_ASSERT( m_conflict_constraint_dependencies.get() );
    conflict_constraint_dependency &
        dependencies= *m_conflict_constraint_dependencies;
    
    // Get constraints which were active in recent backtracks
    unsigned min_active_backtrack=
        m_backtracks_for_iteration - keep_constraints_from_n_conflicts();
    
    vector<unsigned>
        to_keep= dependencies.active_constraints(min_active_backtrack);
    
    to_keep.insert(to_keep.end(),
                   busy_constraints.begin(), busy_constraints.end());
    for (auto v: m_learned_rerun_level)
        to_keep.push_back(v.m_constraint);
    
    vector<unsigned> to_erase= dependencies.other_constraints(to_keep);
    
    save_erased_prerequisites_for_proof(to_keep);

    if (!to_erase.empty()) {
        vector<bool> is_touch_var(count_variables(), false);
        for (unsigned i=0 ; i != to_erase.size(); ++i) {
            const constraint_connectivity_info & info= 
                m_constraint_info_collection[to_erase[i]];
            const unsigned arity= info.arity();
            for (unsigned j=0 ; j != arity ; ++j)
                is_touch_var[info.to_variable(j)]= true;
        }
        for (unsigned i=0 ; i != to_erase.size(); ++i)
            forget_conflict_constraint(to_erase[i]);
        
        is_removed_constraint_functor removal_functor(m_propagator_pool);
        for (unsigned var=0 ; var != m_var_watchers.size() ; ++var) {
            if (!is_touch_var[var])
                continue;
            m_var_value_watchers[var].deallocate_watches(removal_functor);
            m_var_watchers[var].deallocate_watches(removal_functor);
        }
    }
}

void solver::reset()
{
    m_assignment_stack.clear();
}

void solver::forget_learned_constraints()
{
    if (m_original_constraint_count==0)
        return;
    HCSP_ASSERT(m_original_constraint_count <=
               m_constraint_info_collection.size());
    
    m_constraint_info_collection.erase(
        m_constraint_info_collection.begin()+m_original_constraint_count,
        m_constraint_info_collection.end());

    while (m_propagator_pool.size() > m_original_constraint_count) {
        delete m_propagator_pool.back();
         // safety measure to detect reuse of freed entries
        m_propagator_pool.back()= NULL;
        m_propagator_pool.pop_back();
    }
}

void solver::set_var_domain(unsigned var_index, iset new_domain,
                            int updating_constraint)
{
    if (m_now_solving)
        m_assignment_stack.set_var_domain(var_index, move(new_domain),
                                          m_current_time, updating_constraint);
    else
        m_assignment_stack.set_var_initial_domain(var_index, move(new_domain));
}

/**
 * @brief A functor that inserts its parameters into a queue.
 *
 * It will insert all of its parameters into the queue, as long as they are
 * not equal to 'except' (when -1 means that there is no such limitation.
 */
class solver::inserter_except
{
  public:
    /**
     * @param queue   This functor will insert elements into this queue.
     * @param except  If except>=0 then any constraint_index==except will be
     *                ignored.
     */
    inserter_except(unique_queue<> & queue, int except)
        : m_queue(queue), m_except(except)
    {}
    void operator()(unsigned constraint) const
    {
        if (m_except < 0 || constraint != m_except + 0U)
            m_queue.insert(constraint);
    }
  private:
    unique_queue<> & m_queue;
    int              m_except;

};

bool solver::variable_event::comparator::operator()(
    const variable_event & a, const variable_event & b) const
{ 
    unsigned size_a = m_stack.var_domain(a.m_variable).size();
    unsigned size_b = m_stack.var_domain(b.m_variable).size();
    double diff_a = a.m_size_before - size_a;
    double diff_b = b.m_size_before - size_b;
    auto decision_mod = m_stack.decision_modification(m_stack.level());
    auto mod_a = m_stack.var_modification_entry(a.m_variable) - decision_mod;
    auto mod_b = m_stack.var_modification_entry(b.m_variable) - decision_mod;
    double diff_ratio_a = size_b * diff_a * mod_b;
    double diff_ratio_b = size_a * diff_b * mod_a;
    // Prefer variables with biggest change that did not change recently.
    return diff_ratio_a < diff_ratio_b;
}

void solver::notify_modification(unsigned var_index, int except_constraint)
{
    HCSP_ASSERT( var_index < m_var_watchers.size() );
    change_index_t modification=
        m_assignment_stack.var_modification_entry(var_index);
    if ( !constraint_queue_enabled()) {
        auto & handle= m_variable_queue_handles[var_index];
        if (handle.second)
            return;
        handle.second = true;
        variable_event event;
        event.m_variable = var_index;
        event.m_size_before = m_assignment_stack.var_prev_size(var_index);
        event.m_modification = modification;
        handle.first= m_variable_queue.push(event);
        return;
    }
    if ( !m_var_value_watchers[var_index].empty() ) {
        iset removed= removed_values(var_index, modification);
        m_var_value_watchers[var_index].for_each(
            inserter_except(m_constraint_queue, except_constraint), removed);
    }
    small_iset types_to_watch;
    types_to_watch.insert(MODIFICATION_WATCH);
    if ( var_domain(var_index).singleton() )
        types_to_watch.insert(FINALIZATION_WATCH);
    m_var_watchers[var_index].for_each(
        inserter_except(m_constraint_queue, except_constraint),types_to_watch);
}

solver::explanation_views solver::get_constraint_explanation(
    unsigned constraint_index, unsigned propagated_param,
    const iset & values_to_explain, change_index_t modification) const
{
    constraint_propagator *propagator= m_propagator_pool.at(constraint_index);
    HCSP_ASSERT(propagator);
    clause_propagator * p_clause= dynamic_cast<clause_propagator*>(propagator);
    if (p_clause)
        return explanation_views(p_clause->literals());

    const constraint_connectivity_info & info= 
        m_constraint_info_collection.at(constraint_index);
    
    const bool no_param= (propagated_param+1 == 0);
    iset propagated_domain;
    if (!no_param)
        propagated_domain= m_assignment_stack.var_domain_after_modification(
            info.to_variable(propagated_param), modification);
    
    static vector<iset> explanation;
    unsigned effort= get_effort(info);
    constraint_aged_info before_propagation(info, m_assignment_stack,
                                            modification, effort);
    before_propagation.allow_latest_modification_query(
        should_explainer_check_age());
    propagator->explain_propagation(
        no_param ? 0 : propagated_param,
        m_augment_explanation ? values_to_explain : iset(iset().complement()),
        propagated_domain, before_propagation, explanation);
    explanation_views returned_rule(explanation);

    if (m_proof_collector->want_proof())
        returned_rule.m_rule= propagator->rule(explanation);
    if ( report_level() >= PROPAGATION_MODS )
        report_explanation(cout, info, explanation);
    HCSP_ASSERT(returned_rule.consistent());
    return returned_rule;
}

void solver::report_explanation(ostream & out,
                                const constraint_runtime_info & info,
                                const vector<iset> &explanation)
    const
{
    out << "Explanation:";
    bool first= true;
    out << clause_data::print_prefix();
    for (unsigned i=0 ; i != explanation.size() ;  ++i) {
        if (explanation[i].empty())
            continue;
        if (!first)
            out << clause_data::print_separator();
        first= false;
        info.print_param(out, i);
        literal_data::print_right_hand_side(out,explanation[i]);
    }
    out << clause_data::print_suffix() << "\n";
}

solver::change_index_t solver::conflicting_modification_entry() const
{
    change_index_t entry= m_assignment_stack.latest_modification();
    if (entry == assignment_stack::NO_PREVIOUS_MODIFICATION)
        return 0;
    while (true) {
        int cause= m_assignment_stack.modifying_constraint(entry);
        if (cause <0 || cause +0U != m_last_constraint)
            // 'entry' is not due to the analyzed conflict, take the next one.
            return entry+1;
        if (entry == 0)
            break;
        --entry;
    }
    return 0;
}

void solver::collect_conflict_nodes(learned_type & conflict_constraint,
                                    explanation_views & explanation,
                                    unsigned constraint_index,
                                    change_index_t before_modification)
{
    const vector<iset> & literals= explanation.m_literals;
    //TODO(clause->constraint) Recording explanation works correctly only for
    // when a conflict analysis uses only clause resolution. Otherwise the
    // explanation is irrelevant for generalized resolution. The following line
    // has to be extended to generalized resolution.
    m_proof_collector->record_conflict(
        constraint_index, literals, move(explanation.m_rule));
    const constraint_connectivity_info & info= 
        m_constraint_info_collection.at(constraint_index);
    record_explanation_variables(info, literals);
    
    conflict_constraint.construct_from(info,
                                       *m_propagator_pool[constraint_index],
                                       explanation.m_literals,
                                       before_modification);
}

void solver::relax(learned_type & clause, unsigned literal_pos) const
{
    switch (m_relaxation) {
      case NO_RELAXATION:
        break;
      case LEVEL_RELAXATION:
        clause.relax_level(literal_pos);
        break;
      case COMPLETE_RELAXATION:
        clause.relax(literal_pos);
        break;
    }
}

unsigned solver::insert_constraint(const learned_type &conflict_constraint)
{
    unique_ptr<constraint_propagator> constraint;
    if (conflict_constraint.constraint()) {
        constraint.reset(conflict_constraint.constraint()->clone());
    } else {
        clause_propagator *p_clause= new clause_propagator;
        constraint.reset(p_clause);
        p_clause->update_literals()= conflict_constraint.get_explanation();

    }    
    constraint_connectivity_info & info= add_constraint(move(constraint));
    info.reserve_arguments( conflict_constraint.size() );
    for (unsigned i=0 ; i != conflict_constraint.size() ; ++i) {
        info << conflict_constraint.variable(i);
    }
    unsigned constraint_index= m_propagator_pool.size()-1;
    if ( constraint_queue_enabled() )
        m_constraint_queue.set_max_elements(m_propagator_pool.size());
    if (!m_now_solving)
        return constraint_index;

    if (m_report_level >= SEARCH) {
        cout << "Inserting ";
        m_propagator_pool[constraint_index]->print(cout, &info);
        cout << "\n";
    }
    return constraint_index;
}

bool solver::prepare_to_resolve(learned_type        &constraint,
                                unsigned             pivot_variable,
                                unsigned             other_constraint,
                                unsigned            &other_pivot_index,
                                const vector<iset>* &p_other_literals)
{
    unsigned constraint_pivot_index= constraint.variable_to_pos(pivot_variable);
    constraint_connectivity_info & other_connectivity =
        m_constraint_info_collection[other_constraint];
    other_connectivity.mark_conflicting();
    change_index_t pivot_mod= constraint.modification(constraint_pivot_index);
    if ( report_level() >= PROPAGATION_MODS )
        report_resolution_start(cout, constraint, other_constraint, pivot_mod);
    other_pivot_index=other_connectivity.find_param_of_variable(pivot_variable);
    constraint.update_explanation();
    const iset & pivot_literal= constraint.literal(constraint_pivot_index);
    explanation_views explanation= get_constraint_explanation(
        other_constraint, other_pivot_index, pivot_literal, pivot_mod);
    add_constraint_cost_to_variables(other_constraint, 
                                     m_assignment_stack.cost_additive(),
                                     &explanation.m_literals);
    p_other_literals= &explanation.m_literals;
    const iset &pivot_domain= m_assignment_stack.var_domain_after_modification(
        pivot_variable, pivot_mod);
    if (!(*p_other_literals)[other_pivot_index].contains_any_of(pivot_domain))
        return false;
    HCSP_ASSERT(
        !pivot_literal.contains_all_of((*p_other_literals)[other_pivot_index]));
    m_proof_collector->record_resolution(
        other_constraint, *p_other_literals, pivot_variable,
        move(explanation.m_rule));

    if ( report_level() >= PROPAGATION_MODS )
        cout << "with pivot " << var_printer(pivot_variable) << "\n";

    return true;
}

void solver::resolve(learned_type      &constraint,
                     unsigned           pivot_variable,
                     unsigned           other_constraint)
{
    HCSP_PROFILE_CURRENT;
    const vector<iset> *p_other_literals = NULL;
    unsigned other_pivot_index;
    if (!prepare_to_resolve(constraint, pivot_variable, other_constraint,
                            other_pivot_index, p_other_literals))
        return;

    ++m_total_combines;
    constraint_connectivity_info & other_connectivity =
        m_constraint_info_collection[other_constraint];
    unsigned constraint_pivot_index= constraint.variable_to_pos(pivot_variable);
    change_index_t pivot_mod= constraint.modification(constraint_pivot_index);
    constraint_propagator & other_constraint_propagator =
        *m_propagator_pool[other_constraint];
    if ( constraints_combination() &&
         constraint.combine(other_constraint_propagator, other_connectivity,
                            other_pivot_index, pivot_mod)) {
        ++m_non_fallback_combines;
        return;
    }
    m_current_conflict_analysis_has_fallback= true;
    //! @todo move the following two lines to learned_type
    constraint.convert_to_clause();
    resolve_clause(constraint, other_connectivity, *p_other_literals,
                   other_pivot_index, pivot_mod);

    record_explanation_variables(other_connectivity, *p_other_literals);
}

void solver::restart_learning(learned_type & learned_constraint,
                              explanation_views & explanation,
                              unsigned constraint_index)
{
    if ( report_level() >= PROPAGATION_MODS )
        cout << "Explanation is conflicting, restarting learning\n";
    m_analyzed_variables.clear();
    unsigned param = learned_constraint.latest_node_index();
    collect_conflict_nodes(learned_constraint, explanation, constraint_index,
                           learned_constraint.modification(param) );
}

void solver::record_explanation_variables(
    const constraint_connectivity_info & connectivity,
    const vector<iset> & literals)
{
    for (unsigned i=0 ; i != literals.size() ; ++i) {
         if ( literals[i].empty() )
             continue;
         unsigned variable= connectivity.to_variable(i);
         if ( var_ground_level_domain(variable).contains_any_of(literals[i]) )
             m_analyzed_variables.push_back(variable);
    }
}

void solver::resolve_clause(
    learned_type &clause, 
    const constraint_connectivity_info & other_connectivity,
    const vector<iset> & other_literals,
    unsigned           other_pivot_index,
    change_index_t     other_constraint_activation) const
{
    HCSP_ASSERT( other_connectivity.arity() == other_literals.size() );
    HCSP_ASSERT( other_pivot_index < other_literals.size() );
    unsigned pivot_variable= other_connectivity.to_variable(other_pivot_index);
    unsigned clause_pivot_index= clause.variable_to_pos(pivot_variable);
    const iset & other_literal= other_literals[other_pivot_index];
    if (m_relaxation == COMPLETE_RELAXATION)
        HCSP_ASSERT(!other_literal.contains_all_of(
                       clause.literal(clause_pivot_index)));
    
    for (unsigned i=0 ; i != other_literals.size() ; ++i) {
        if ( i == other_pivot_index || other_literals[i].empty() )
            continue;
        unsigned variable= other_connectivity.to_variable(i);
        change_index_t modification;
        modification= m_assignment_stack.var_modification_entry_before(
            variable, other_constraint_activation); 
        clause.insert_literal(other_literals[i], variable, modification);
        if (m_relaxation != NO_RELAXATION)
            relax(clause, clause.variable_to_pos(variable));
    }
    
    // Make sure that the same constraint will not be used for the
    // remaining conflict analysis.
    change_index_t past_modification;
    past_modification= m_assignment_stack.var_modification_entry_before(
        pivot_variable, clause.modification(clause_pivot_index));
    bool has_literal=clause.intersect_literal(other_literal, pivot_variable,
                                              past_modification);    
    if (cause_relaxation())
        clause.relax_cause();
    if (has_literal && m_relaxation != NO_RELAXATION)
        relax(clause, clause.variable_to_pos(pivot_variable));
        
    if ( report_level() >= PROPAGATION_MODS )
        cout << "consequent: " << clause << "\n\n" ;
    HCSP_ASSERT(clause.conflicts());
}

void solver::report_resolution_start(ostream & out,
                                     const learned_type &constraint,
                                     unsigned other_constraint,
                                     change_index_t modification) const
{
    const constraint_connectivity_info & other_connectivity =
        m_constraint_info_collection[other_constraint];
    
    out << "resolving constraint " << constraint << "\n" "with ";
    constraint_aged_info before_propagation(other_connectivity,
                                            m_assignment_stack, modification);
    
    m_propagator_pool[other_constraint]->print(out, &before_propagation);
    out << "\n" ;
}

void solver::forget_all_watches()
{
    for (unsigned i=0 ; i != m_propagator_pool.size(); ++i)
        if (m_propagator_pool[i])
            m_propagator_pool[i]->forget_watches();

    m_var_watchers.clear();
    m_var_value_watchers.clear();
}

void solver::throw_timeout_exception()
{
    throw timeout_exception();
}

void solver::show_self() 
{
    for (unsigned i=0 ; i != m_assignment_stack.count_variables() ; ++i) {
        cout << i << ": " << var_printer(i) << "=" << var_domain(i) << "\n";
    }
}

bool solver::is_clause(unsigned constraint_index) const
{
    constraint_propagator *p_constraint=m_propagator_pool.at(constraint_index);
    if ( dynamic_cast<clause_propagator*>(p_constraint) )
        return true;
    else
        return false;
}

void solver::print_constraint(ostream & out, unsigned constraint_index) const
{
    const constraint_connectivity_info & constraint_info=
        m_constraint_info_collection.at(constraint_index);
    
    m_propagator_pool.at(constraint_index)->print(out, &constraint_info);
}

const vector<iset> & solver::clause_literals(unsigned clause_id) const
{
    const constraint_propagator *p_constraint= m_propagator_pool.at(clause_id);
    return dynamic_cast<const clause_propagator&>(*p_constraint).literals();
}

solver_types::label_type solver::constraint_labels(unsigned constraint) const
{
    return m_constraint_info_collection.at(constraint).labels();
}

solver_types::label_type
solver::constraint_labels(unsigned constraint, label_type label)
{
    constraint_connectivity_info &
        info= m_constraint_info_collection.at(constraint);
    label_type orig= info.labels();
    info.labels(label);
    return orig;
}

solver_types::label_type solver::variable_labels(unsigned variable) const
{
    return m_assignment_stack.variable_labels(variable);
}

solver_types::label_type
solver::variable_labels(unsigned variable, label_type label)
{
    return m_assignment_stack.variable_labels(variable, label);
}

vector<solver::label_type> solver::variables_touched_by_labels() const
{
    vector<label_type> labels(count_variables(), NO_LABEL);
    for (unsigned i=0 ; i != m_constraint_info_collection.size() ; ++i) {
        label_type current_label= constraint_labels(i);
        if ( current_label == NO_LABEL )
            continue;
        const constraint_connectivity_info &
            info= m_constraint_info_collection[i];
        for (unsigned j= info.arity(); j > 0 ; ) {
            --j;
            labels[info.to_variable(j)] |= current_label;
        }
    }
    for (unsigned i=0 ; i != labels.size(); ++i)
        labels[i] |= variable_labels(i);
    return labels;
}

void solver::get_constraint_variables(unsigned constraint_index,
                                      vector<unsigned> & variables) const
{
    const constraint_connectivity_info & info=
        m_constraint_info_collection.at(constraint_index);
    variables.resize(info.arity());
    for (unsigned i=0 ; i != variables.size() ; ++i) 
        variables[i]= info.to_variable(i);
}

//virtual
void solver::print_solution(ostream & out) const
{
    cout << m_comment << " ";
    for (unsigned i=0 ; i != count_variables() && i != m_user_variables; ++i)
        out << var_printer(i) << "=" << var_domain(i) << ", ";
    out << "\n" "v ";
    for (unsigned i=0 ; i != count_variables()  && i != m_user_variables; ++i)
        out << var_domain(i) << " ";
    out << endl;
}

void solver::save_solution_as_best()
{
    m_best_objective= evaluate_objective();
    if (!m_only_optimal) {
        cout << m_comment << " best result so far:\n";
        cout << m_objective_prefix << " " << m_best_objective << endl;
        print_solution(cout);
    }
    if (m_is_minimizing)
        HCSP_ASSERT(m_best_objective <= var_domain(m_objective_variable).min());
    else
        HCSP_ASSERT(m_best_objective >= var_domain(m_objective_variable).min());

    m_best_solution.resize(count_variables());
    for ( unsigned i=0 ; i != m_best_solution.size() ; ++i) {
        m_best_solution[i]= var_domain(i).min();
    }
}

void solver::restore_best_solution()
{
    HCSP_ASSERT(m_best_solution.size() == count_variables());
    for ( unsigned i=0 ; i != m_best_solution.size() ; ++i) {
        set_var_domain(i, iset(m_best_solution[i]));
    }
    set_var_domain(m_objective_variable, iset(m_best_objective));
}

// virtual
iset::value_type solver::evaluate_objective() const
{
    HCSP_ASSERT( var_domain(m_objective_variable).singleton() );
    return var_domain(m_objective_variable).max();
}

ostream & hcsp::operator<<(ostream& out, solver::result_type result)
{
    switch (result) {
      case solver::OPTIMAL:          out << "OPTIMAL";          break;
      case solver::SAT:              out << "SAT";              break;
      case solver::SAT_DONE:         out << "SAT_DONE";         break;
      case solver::UNSAT:            out << "UNSAT";            break;
      case solver::BACKTRACK_LIMIT:  out << "BACKTRACK_LIMIT";  break;
      case solver::TIMEOUT:          out << "TIMEOUT";          break;
      case solver::MEMOUT:           out << "MEMOUT";           break;
    }
    return out;
}
