// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <cassert>
#include <hash_map.hpp>
#include <pool_alloc.hpp>
#include <algorithm>
#include <map>
#include <iterator>
#include <table_propagators.hpp>
#include <unistd.h>
using namespace std;
using namespace hcsp;

/**
 * @brief A sub-tree of a decision tree for storing tuples.
 */
class table_propagator_base::decision_sub_tree
{
  private:
    typedef pair<iset::value_type, decision_sub_tree*> value_type;

  public:
    ~decision_sub_tree()
    {
        while (!m_children.empty())
            {
                delete m_children.back().second;
                m_children.back().second= NULL;
                m_children.pop_back();
            }
    }
    bool empty() const { return m_children.empty(); }
    size_t size() const { return m_children.size(); }
    /**
     * @brief Insert a single tuple into the decision tree.
     *
     * Insert a tuple, or part of it, into the given sub-tree.
     * Only if the parameter is the first parameter, then the whole tuple
     * is inserted, otherwise only part of the tuple is inserted into the
     * sub-tee.
     * @param first_parameter
     *        The first parameter of the tuple to be inserted.
     * @param parameter
     *        The current parameter of the tuple to be inserted, the next
     *        parameter is calculated by
     *        table_propagator_base::get_next_parameter().
     */
    void insert(unsigned first_parameter, unsigned parameter,
                const tuple_value_type & a_tuple);
    //! Find a given child of this node
    decision_sub_tree *child(unsigned i) { return m_children[i].second; }
    //! Find the value of given child edge of this node
    iset::value_type value(unsigned i) const { return m_children[i].first; }
  private:
    
    vector<value_type> m_children;
}; // class table_propagator_base::decision_sub_tree

/**
 * @brief A decision node used for checking if a tuple is in the DAG.
 *
 * Like with BDDs the construction process of these decision nodes makes sure
 * that for any possible decision combination sub-tree there is at most one
 * copy. This is done by a hash table which stores references to all nodes and
 * hashes by the children of each node. When someone wants to get a pointer to
 * a sub-tree either an existing node is returned, if appropriate node exists
 * in the hash, or a new node is created.
 *
 * An object of type table_propagator_base::decision_node_manger is
 * responsible for allocating unique nodes and for collecting unused garbage
 * nodes.
 */
class table_propagator_base::decision_node
{
  public:
    struct hash_edges;
    typedef pair<iset::value_type, const decision_node*>  value_type;
    friend class table_propagator_base::memory_manager;

    /**
     * @brief Check if the decision nodes are equivalent
     *
     * Since the decision nodes are unique, equal children edges means the
     * nodes should be equal. This seems not very efficient, but it is
     * needed when non-unique nodes are being detected and not inserted
     * into the data structure.
     */
    friend bool operator==(const decision_node &a, const decision_node &b)
    {
        return a.m_edges == b.m_edges;
    }

    /**
     * @brief Find an edge going out of this node and labeled with \p value
     *
     * @param value   The value of the edge to look for
     * @return If found, the edge labeled by a \p value edge, NULL if not found.
     */
    const value_type * find_edge(iset::value_type value) const
    {
        for (unsigned i=0 ; i != m_edges.size() ; ++i)
            if (m_edges[i].first == value)
                return &m_edges[i];
        return NULL;
    }
    //! Return the set of values on labels of edges going out of this node.
    iset child_values() const
    {
        iset ret;
        for (unsigned i=0 ; i != m_edges.size() ; ++i)
            ret.insert(m_edges[i].first);
        return ret;
    }

    /**
     * @brief Check if the node supports the specified values.
     *
     * This method looks at all parameters starting with \p parameter,
     * and advancing with table_propagator_base::get_next_parameter().
     *
     * Look for a sub-tuple, defined by these parameters and from their
     * respective domains in \p info, such that there is a path on the
     * sub-tree labeled with these values.
     *
     * @param[in] first_parameter
     *                The first parameter of the tree, i.e., the parameter to
     *                which the root node refers to.   
     * @param[in] parameter
     *                The parameter which the current node refers to.
     * @param[in] info
     *                The current domains to look in.
     * @param[in,out] accumulated_values
     *                For each parameter, this method will add values which
     *                are found to support this sub-tree
     * @return  If some values are supported by this sub-tree.
     */
    inline bool supports(unsigned first_parameter, unsigned parameter,
                         const constraint_runtime_info & info,
                         vector<iset> & accumulated_values) const;

    /**
     * @brief Check if the node has support for 'parameter'..end
     *
     * This method eeks for a path of edges whose values support the domains
     * starting with the given parameter.
     *
     * @param[in] parameter
     *                The parameter which the current node refers to.
     * @param[in] info
     *                The current domains to look in.
     * @param[in,out] visited
     *                Nodes which we the algorithm visited. It is assumed that
     *                a second visit to the same node means that the first
     *                visit did not find support, and hence it will not find
     *                support the following visit.
     * @return  True if some values are supported by this sub-tree.
     */
    bool has_support(unsigned parameter, const constraint_runtime_info & info,
                     HASH_SET<const decision_node*> & visited) const;

    /**
     * @brief For each supported parameter value, accumulate supporting tuples
     *
     * For each value, of the parameter, look for a DAG path that supports it,
     * accumulating at least one such path per-value in \p accumulated_values.
     *
     * @param[in] parameter
     *                The parameter which the DAG starts with, next parameters
     *                are found by repeatedly calling
     *                table_propagator_base::get_next_parameter().
     * @param[in] info
     *                The current domains to look in.
     * @param[in,out] accumulated_values
     *                For each parameter, this method will add values which
     *                are found to support this sub-tree
     */
    inline void collect_supports(unsigned param,
                                 const constraint_runtime_info & info,
                                 vector<iset> & accumulated_values) const;

    //! For debugging, print the node and its children as a table.
    void show_self() const
    {
        vector<iset::value_type> prefix;
        recuse_show_self(prefix);
    }
    
  private:
    /**
     * @brief Seek for a tuple that supports the given parameter value.
     *
     * This method looks for a supported tuple under the specified DAG,
     * and blow a given edge of the root node, i.e.,
     * If a supported tuple is found, then its values are inserted into
     * the \p accumulated_values vector.
     *
     * @param edge_node  A pointer to the child to check, may be NULL
     *                   if edge is a leaf edge,
     * @param edge_value The value of the edge pointing to this child.
     * @param param      The parameter of the root of this DAG.
     * @param info       The state, i.e. domain, of the variables checked for
     *                   support in this table.
     * @param accumulated_values
     *                   The set of supported values we accumulate.
     */
    inline void accumulate_edge_support(const decision_node * edge_node,
                                        iset::value_type edge_value,
                                        unsigned param,
                                        const constraint_runtime_info & info,
                                        vector<iset> &accumulated_values)const;

    /**
     * @brief For each supported parameter value, accumulate supporting tuples,
     *        iterating DAG children and info domain values simultaneously.
     *
     * This method is exactly the same as collect_supports(), in fact
     * it is intended to be called by collect_supports() when
     * simultaneously iterating DAG children and info domain values is the
     * most efficient.
     *
     * For more, read the documentation of collect_supports().
     */
    void collect_supports_iterating_children_and_values(
        unsigned param, const constraint_runtime_info & info,
        vector<iset> & accumulated_values) const;
    
    /**
     * @brief For each supported parameter value, accumulate supporting tuples,
     *        iterating DAG children.
     *
     * This method is exactly the same as collect_supports(), in fact
     * it is intended to be called by collect_supports() when
     * iterating DAG children and looking up values of info domains is the
     * most efficient.
     *
     * For more, read the documentation of collect_supports().
     */
    void collect_supports_iterating_children(
        unsigned param, const constraint_runtime_info & info,
        vector<iset> & accumulated_values) const;
    
    /**
     * @brief Check if the node supports the specified values.  iterating DAG
     *        children and info domain values simultaneously.
     *
     * This method is exactly the same as supports(), in fact it is intended
     * to be called by supports() when simultaneously iterating DAG children
     * and info domain values is the most efficient.
     *
     * For more, read the documentation of supports().
     */
    bool supports_iterating_children_and_values(
        unsigned first_parameter, unsigned parameter,
        const constraint_runtime_info & info,
        vector<iset> & accumulated_values) const;
    
    /**
     * @brief Check if the node supports the specified values, iterating DAG
     *        children.
     *
     * This method is exactly the same as supports(), in fact it is intended
     * to be called by supports() when iterating DAG children and looking up
     * values of info domains is the most efficient.
     *
     * For more, read the documentation of supports().
     */
    bool supports_iterating_children(unsigned first_parameter,
                                     unsigned parameter,
                                     const constraint_runtime_info & info,
                                     vector<iset> & accumulated_values) const;

    /**
     * @brief Recursive implementation of show_self()
     * @param prefix The prefix to print for each traversed tuple
     */
    void recuse_show_self(vector<iset::value_type> & prefix) const;

    //! The edges, labeled with values, pointing to DAG children
    vector<value_type> m_edges;
}; // class table_propagator_base::decision_node

/**
 * @brief A function object which gives a hash function of DAG node's outgoing
 *        edges.
 *
 * The hash function looks only at the edges outgoing from the node. Edges
 * define each node uniquely, letting this hash function to be used as a
 * search tool for previously allocated nodes.
 */
struct table_propagator_base::decision_node::hash_edges
{
    //! The hash function.
    size_t operator()(const decision_node &node) const
    {
        size_t hash=0;
        
        for (unsigned i=0 ; i != node.m_edges.size() ; ++i)
            {
                hash^= reinterpret_cast<size_t>(node.m_edges[i].second);
                hash^= 47*node.m_edges[i].first;
                hash= hash * 98317;
                hash= (hash >> 16) | (hash << 16);
            }
        return hash;
    }
};

//! The implementation class of the memory_manager
class table_propagator_base::memory_manager::impl
{
  public:
    impl() : m_erasures_since_gc(0), m_enabled_gc(true) {}
    /**
     * @brief Get a DAG node that represents the same tuples as the parameter
     *
     * Gives a unique DAG node for the given decision sub tree.
     *
     * @param p_node A pointer to the requested decision sub tree.
     * @return A node, existing if possible, that represents the sub-tree.
     */
    const decision_node * get_node(decision_sub_tree * p_node);

    //! Implementation of
    //! table_propagator_base::memory_manager::disable_garbage_collection()
    void disable_gc() { m_enabled_gc= false; }

    //! Mark the specified DAG as used, so that it will not be freed.
    void mark_dag_as_used(const decision_node & dag)
    { m_used_dags.insert(&dag); }
    
    //! Mark the specified DAG as free, so that it can be freed.
    void mark_dag_as_free(const decision_node & dag)
    {
        m_used_dags.erase(&dag);
        ++m_erasures_since_gc;
        if (m_erasures_since_gc*2 > m_used_dags.size())
            garbage_collect();
    }
    //! Implementation of
    //! table_propagator_base::memory_manager::garbage_collect()
    void garbage_collect();
  private:
    /**
     * @brief Mark nodes which are accessible from the specified node.
     *
     * @param accessible_nodes The set which holds the found nodes
     * @param p_node           The DAG node to start the search from
     */
    static void mark_accessible_nodes(
        HASH_SET<const decision_node*> & accessible_nodes,
        const decision_node * p_node);

    //! The set of all allocated nodes, hashed by this hash function.
    HASH_SET<decision_node, decision_node::hash_edges> m_nodes_hash;

    //! The set of DAGs currently in used, which should not be freed.
    HASH_SET<const decision_node*> m_used_dags;

    //! Number of DAGs marked as free since last garbage collection run.
    unsigned m_erasures_since_gc;

    //! If false, garbage_collect() will do nothing.
    bool     m_enabled_gc;
}; // class table_propagator_base::memory_manager::impl

//! The actual tuples that define the propagator.
class table_propagator_base::tuples : private no_copy
{
  public:
    tuples(memory_manager::impl &manager,
           const vector<tuple_value_type> &tuples_in);
    ~tuples()
    {
        for (unsigned i=0 ; i != m_dags.size() ; ++i)
            m_p_manager->mark_dag_as_free(*m_dags[i]);
    }
    int count_columns() const
    { return m_domains.empty() ? -1 : (int) m_domains.size(); }
    //! Mark that this object is referenced from a new location.
    void add_reference() { ++m_reference_count; }

    /**
     * @brief Mark that this object is dereferenced from a location.
     *
     * If the reference count reaches zero, the object is delete.
     */
    void remove_reference()
    {
        if (--m_reference_count == 0)
            delete this;
    }
    
    //! Returns true if the single tuple of \p info is in this set of tuples.
    bool find_tuple( const constraint_runtime_info & info ) const;

    //! Collect all values of info.domain(param) supported by the tuples
    void collect_supports(unsigned param, 
                          const constraint_runtime_info & info,
                          vector<iset> & accumulated_values) const
    { m_dags[param]->collect_supports(param, info, accumulated_values); }

    bool has_support(const constraint_runtime_info &info) const
    {
        HASH_SET<const decision_node*> visited;
        return m_dags[0]->has_support(0, info, visited);
    }
    
    //! Get the set of values allowed by the tuples for one parameter
    const iset & domain(unsigned param) const { return m_domains[param]; }

    //! Is this set of tuples empty.
    bool empty() const { return m_domains.empty(); }

    //! To be called from a debugger, to print the set of tuples.
    void show_self() const;
  private:

    //! The number of references to this tuple object.
    unsigned m_reference_count;

    //! A decision DAG for each parameter
    vector<const decision_node*> m_dags;

    //! The merger of all tuples into a flat vector of sets.
    vector<iset> m_domains;

    //! The memory manager responsible for allocating and freeing the DAGs.
    memory_manager::impl *m_p_manager;
}; // class support_table_propagator::tuples

const table_propagator_base::decision_node*
table_propagator_base::memory_manager::impl::get_node(
    decision_sub_tree * p_node)
{
    if (p_node == NULL || p_node->empty())
        return NULL;
    
    decision_node node;
    node.m_edges.resize(p_node->size());
    for (unsigned i=0 ; i != p_node->size() ; ++i)
        {
            node.m_edges[i].second= get_node(p_node->child(i));
            node.m_edges[i].first= p_node->value(i);
        }
    
    return &(*m_nodes_hash.insert(node).first);
}

void table_propagator_base::memory_manager::impl::garbage_collect()
{
    if (! m_enabled_gc )
        return;
    
    // A simple mark and seep algorithm
    HASH_SET<const decision_node*> active_nodes;
    for ( HASH_SET<const decision_node*>::iterator it= m_used_dags.begin();
          it != m_used_dags.end();
          ++it)
        {
            mark_accessible_nodes(active_nodes, *it);
        }
    HASH_SET<decision_node, decision_node::hash_edges>::iterator it;
    it = m_nodes_hash.begin();
    while (it != m_nodes_hash.end())
        {
            if (active_nodes.find(&*it) == active_nodes.end())
                m_nodes_hash.erase(it++);
            else
                ++it;
        }
    m_erasures_since_gc= 0;
}

void table_propagator_base::memory_manager::impl::mark_accessible_nodes(
    HASH_SET<const decision_node*> & accessible_nodes,
    const decision_node * p_node)
{
    if (p_node == NULL)
        return;
    if (accessible_nodes.insert(p_node).second == false)
        {
            // the node is already inside the set, meaning that all its
            // children have also been previously inserted.
            return;
        }
    for (unsigned i= 0 ; i != p_node->m_edges.size() ; ++i)
        {
            mark_accessible_nodes(accessible_nodes,
                                  p_node->m_edges[i].second);
        }
}

void table_propagator_base::decision_sub_tree::insert(
    unsigned first_parameter, unsigned parameter,
    const tuple_value_type & a_tuple)
{
    unsigned i;
    for (i=0 ; i != m_children.size() ; ++i)
        if (m_children[i].first >= a_tuple[parameter])
            break;

    if (i >= m_children.size() || m_children[i].first > a_tuple[parameter])
        m_children.insert(m_children.begin()+i,
                          value_type(a_tuple[parameter], 0));
    parameter= get_next_parameter(first_parameter, parameter);
    if (parameter < a_tuple.size())
        {
            if (!m_children[i].second)
                m_children[i].second= new decision_sub_tree;
            m_children[i].second->insert(first_parameter, parameter, a_tuple);
        }
}

inline bool table_propagator_base::decision_node::supports(
    unsigned first_param, unsigned param, const constraint_runtime_info & info,
    vector<iset> & accumulated_values) const
{
#ifdef USE_BIG_ISET
    return supports_iterating_children_and_values(first_param, param, info,
                                                  accumulated_values);
#else
    return supports_iterating_children(first_param, param, info,
                                       accumulated_values);
#endif
}

bool table_propagator_base::decision_node::supports_iterating_children(
    unsigned first_param, unsigned param, const constraint_runtime_info & info,
    vector<iset> & accumulated_values) const
{
    const iset & seek_values= info.domain(param);
    unsigned next_param = get_next_parameter(first_param, param);
    for (unsigned i=0; i != m_edges.size() ; ++i)
        {
            iset::value_type value= m_edges[i].first;
            if ( !seek_values.contains(value))
                continue;
            if ( !m_edges[i].second || // this is a leaf
                 m_edges[i].second->supports(
                     first_param, next_param, info, accumulated_values) )
                {
                    accumulated_values[param].insert(value);
                    return true;
                }
        }
    return false;
}

bool
table_propagator_base::decision_node::supports_iterating_children_and_values(
    unsigned first_param, unsigned param, const constraint_runtime_info & info,
    vector<iset> & accumulated_values) const
{
    const iset & seek_values= info.domain(param);
    unsigned next_param = get_next_parameter(first_param, param);
    unsigned i=0;
    iset::iterator it=seek_values.begin();
    const iset::iterator end=seek_values.end();
    while (it != end && i != m_edges.size())
        {
            iset::value_type value= *it;
            iset::value_type child_value= m_edges[i].first;
            if (child_value == value)
                {
                    if ( !m_edges[i].second || // this is a leaf
                         m_edges[i].second->supports(
                             first_param, next_param, info, accumulated_values))
                        {
                            accumulated_values[param].insert(value);
                            return true;
                        }
                }
            if (child_value <= value) 
                ++i;
            if (child_value >= value)
                {
                    HCSP_ASSERT(it != end);
                    HCSP_ASSERT(end == info.domain(param).end());
                    ++it;
                }
        }
    return false;
}

void table_propagator_base::decision_node::collect_supports(
    unsigned param, const constraint_runtime_info & info,
    vector<iset> & accumulated_values) const
{
#ifdef USE_BIG_ISET
    return collect_supports_iterating_children_and_values(
        param, info, accumulated_values);
#else
    return collect_supports_iterating_children(param, info, accumulated_values);
#endif
}

void table_propagator_base::decision_node::accumulate_edge_support(
    const decision_node * edge_node, iset::value_type edge_value,
    unsigned param,
    const constraint_runtime_info & info,
    vector<iset> & accumulated_values) const
{
    if (!edge_node)
        {
            unsigned next_param= get_next_parameter(param, param);
            
            // Check if edge_node is pointed from a leaf.
            if (next_param == info.arity())
                accumulated_values[param].insert(edge_value);
            return;    
        }
    // Only for the first parameter, check if edge_value is already
    // contained in previously found set of values.
    if ( accumulated_values[param].contains(edge_value) )
        return;

    unsigned next_param= get_next_parameter(param, param);
    if (edge_node->supports(param, next_param, info, accumulated_values))
        accumulated_values[param].insert(edge_value);
}

void
table_propagator_base::decision_node::collect_supports_iterating_children(
    unsigned first_param, const constraint_runtime_info & info,
    vector<iset> & accumulated_values) const
{
    const iset & seek_values= info.domain(first_param);
    if (seek_values.size() == accumulated_values[first_param].size())
        return;
    for (unsigned i=0 ; i != m_edges.size() ; ++i)
        {
            iset::value_type value= m_edges[i].first;
            if ( seek_values.contains(value) )
                accumulate_edge_support(m_edges[i].second, value,
                                        first_param,
                                        info, accumulated_values);
        }
}

void
table_propagator_base::
decision_node::collect_supports_iterating_children_and_values(
    unsigned first_param, const constraint_runtime_info & info,
    vector<iset> & accumulated_values) const
{
    const iset & seek_values= info.domain(first_param);
    if (seek_values.size() == accumulated_values[first_param].size())
        return;
    unsigned i=0;
    iset::iterator it=seek_values.begin();
    const iset::iterator end=seek_values.end();
    while (it != end && i != m_edges.size())
        {
            iset::value_type value= *it;
            iset::value_type child_value= m_edges[i].first;
            if ( child_value== value )
                accumulate_edge_support(m_edges[i].second, value, first_param,
                                        info, accumulated_values);
            if (child_value <= value)
                ++i;
            if (child_value >= value)
                {
                    HCSP_ASSERT(it != end);
                    HCSP_ASSERT(end == info.domain(first_param).end());
                    ++it;
                } 
        }
}

bool table_propagator_base::decision_node::has_support(
    unsigned parameter,
    const constraint_runtime_info & info,
    HASH_SET<const decision_node*> & visited) const
{
    if (!visited.insert(this).second)
        // Already visited, but the algorithm did not stop then. This means
        // that this node does not lead to a supporting tuple.
        return false;
    const iset  & domain= info.domain(parameter);
    for (unsigned i=0 ; i != m_edges.size() ; ++i) {
        if (!domain.contains(m_edges[i].first))
            continue;
        if (!m_edges[i].second /* A leaf. All values are supported below */||
            m_edges[i].second->has_support(parameter+1, info, visited))
            return true;
    }
    return false;
}

void table_propagator_base::decision_node::recuse_show_self(
    vector<iset::value_type> & prefix) const
{
    prefix.push_back(0);
    for (unsigned i=0; i != m_edges.size() ; ++i)
        {
            prefix.back()= m_edges[i].first;
            if (m_edges[i].second)
                {
                    m_edges[i].second->recuse_show_self(prefix);
                }
            else
                {
                    const char *padding="";
                    for (unsigned j=0 ; j != prefix.size() ; ++j)
                        {
                            cout << padding << prefix[j];
                            padding= " ";
                        }
                    cout << "\n";
                }
        }
    prefix.pop_back();
}

table_propagator_base::memory_manager::memory_manager()
{
    m_p_impl= new impl;
}

table_propagator_base::memory_manager::~memory_manager()
{
    m_p_impl->disable_gc();
    delete m_p_impl;
}

void table_propagator_base::memory_manager::garbage_collect()
{
    m_p_impl->garbage_collect();
}

void table_propagator_base::memory_manager::disable_garbage_collection()
{
    m_p_impl->disable_gc();
}

inline const table_propagator_base::decision_node *
table_propagator_base::memory_manager::get_node(decision_sub_tree & tree)
{
    return m_p_impl->get_node(&tree);
}

table_propagator_base::tuples::tuples(
    memory_manager::impl & manager,
    const vector<tuple_value_type> & tuples_in) : m_reference_count(1)
{
    HCSP_PROFILE_CURRENT;
    m_p_manager= &manager;
    if (tuples_in.empty())
        return; // should this ever happen?
    m_dags.resize(tuples_in[0].size(), NULL);

    for (unsigned i=0 ; i != m_dags.size(); ++i)
        {
            decision_sub_tree sub_tree;
            
            for (unsigned j=0 ; j != tuples_in.size() ; ++j)
                sub_tree.insert(i, i, tuples_in[j]);
            m_dags[i]= manager.get_node(&sub_tree);
            manager.mark_dag_as_used(*m_dags[i]);
        }

    m_domains.resize(tuples_in[0].size());
    for (unsigned i=0 ; i != m_dags.size(); ++i)
        m_domains[i]= m_dags[i]->child_values();
}

bool table_propagator_base::tuples::find_tuple(
    const constraint_runtime_info & info) const
{
    if (m_dags.empty())
        return false;
    const decision_node *node= m_dags[0];
    if (info.arity() == 0)
        return 0;
    for (unsigned i=0 ; i != info.arity() ; ++i)
        {
            const decision_node::value_type *
                element= node->find_edge(info.domain(i).min());
            if (!element)
                return false;
            node= element->second;
        }
    return true;
}

void table_propagator_base::tuples::show_self() const
{
    if (!m_dags.empty())
        m_dags[0]->show_self();
}
    
//virtual
constraint_propagator::result_type
support_table_propagator::propagate( constraint_runtime_info & info )
{
    bool any_modified= false;
    for (unsigned i=0 ; !any_modified && i!=info.arity() ; ++i)
        any_modified= info.reduced_between_activations(i);
    if (!any_modified)
        return CONSISTENT;

    if ( get_tuples().empty() )
        return UNSAT;
    
    for (unsigned i=0 ; i!=info.arity() ; ++i)
    {
        const iset & in_domain = info.domain(i);
        if (in_domain.empty())
            return UNSAT;
        const iset & tuple_domain = get_tuples().domain(i);
        if (tuple_domain.contains_all_of(in_domain))
            continue;
        iset out_domain= in_domain;
        out_domain.intersect(tuple_domain);
        info.set_domain(i, move(out_domain));
    }
    
    vector<iset> supported_values(info.arity());
    for (unsigned i=0 ; i!= info.arity() ; ++i)
    {
        get_tuples().collect_supports(i, info, supported_values);
        if (supported_values[i].empty())
            return UNSAT;
        HCSP_ASSERT(info.domain(i).contains_all_of(supported_values[i]));
        info.set_domain(i, move(supported_values[i]));
    }
    return CONSISTENT;
}

//virtual
bool support_table_propagator::conflicts(const constraint_runtime_info & info)
    const
{
    if ( get_tuples().empty() )
        return info.arity() > 0;
    
    for (unsigned i=0 ; i!=info.arity() ; ++i)
        if (!get_tuples().domain(i).contains_any_of(info.domain(i)))
            return true;
    return !get_tuples().has_support(info);
}

void table_propagator_base::show_self() const
{
    return get_tuples().show_self();
}

table_propagator_base::table_propagator_base(
    memory_manager & manager, const std::vector<tuple_value_type> &tuples_in)
{
    m_p_tuples = new tuples(*manager.m_p_impl, tuples_in);
}

//virtual
table_propagator_base::~table_propagator_base()
{
    m_p_tuples->remove_reference();
}

table_propagator_base::table_propagator_base(
    const table_propagator_base & other) : constraint_propagator(other)
{
    other.m_p_tuples->add_reference();
    m_p_tuples= other.m_p_tuples;
}

table_propagator_base & table_propagator_base::operator=(
    const table_propagator_base & other)
{
    if (other.m_p_tuples== m_p_tuples)
        return *this;
    
    m_p_tuples->remove_reference();
    m_p_tuples= NULL;
    other.m_p_tuples->add_reference();
    m_p_tuples= other.m_p_tuples;
    return *this;
}

int table_propagator_base::count_params() const
{
    return m_p_tuples->count_columns();
}

bool table_propagator_base::find_tuple(const constraint_runtime_info & info)
    const
{
    return m_p_tuples ? m_p_tuples->find_tuple(info) : false;
}
