// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <numeric>
#include <algorithm>
#include <iostream>
#include <iomanip>
#include <string>
#include <sstream>
#include <iset.hpp>
#include <tuple_set.hpp>

using namespace std;
using namespace hcsp;

class tuple_set::order_ignoring_column
{
  public:
    order_ignoring_column(unsigned column) : m_column(column) {}
    bool operator()(const vector<iset> & a, const vector<iset> & b)
    {
        if (a.size() != b.size())
            return a.size() > b.size(); // Prefer the small sizes at the end
        for (unsigned i=0 ; i != a.size() ; ++i) {
            if (i == m_column || a[i] == b[i])
                continue;
            iset unique_in_a= a[i];
            unique_in_a.erase_all_of(b[i]);
            if (unique_in_a.empty())
                return false;
            iset unique_in_b= b[i];
            unique_in_b.erase_all_of(a[i]);
            if (unique_in_b.empty())
                return true;
            return unique_in_a.min() < unique_in_b.min();
        }
        return false;
    }
  private:
    unsigned m_column;
};

void tuple_set::insert(const vector<vector<int> > & base_tuples)
{
    if (base_tuples.empty())
        return;
    const size_t arity= (m_tuples.empty() ?
                         base_tuples[0].size() :
                         m_tuples[0].size());
    if (arity == 0) {
        m_tuples.push_back(vector<iset>());
        return;
    }
    for (unsigned i=0; i != base_tuples.size() ; ++i) {
        HCSP_ASSERT(arity == base_tuples[i].size());
        if ( !m_tuples.empty() && arity > 0  &&
             is_equal_prefix(arity-1, m_tuples.back(), base_tuples[i]) )
            m_tuples.back().back().insert(base_tuples[i].back());
        else
            push_back(base_tuples[i]);
    }
    merge_tuples();
}

//static
bool tuple_set::is_equal_prefix(
    unsigned elements_to_compare, const vector<iset> & set_tuple,
    const vector<int> & element_tuple)
{
    HCSP_ASSERT(elements_to_compare <= set_tuple.size());
    HCSP_ASSERT(elements_to_compare <= element_tuple.size());
    for (unsigned i=0 ; i != elements_to_compare ; ++i)
        if ( !set_tuple[i].singleton() ||
             !set_tuple[i].contains(element_tuple[i]) )
            return false;
    return true;
}


void tuple_set::merge_tuples()
{
    if (m_tuples.empty())
        return;
    const unsigned arity= m_tuples[0].size();
    for (unsigned i=0 ; i != arity ; ++i) {
        if (m_tuples.size() == 1)
            return;
        sort(m_tuples.begin(), m_tuples.end(), order_ignoring_column(i));
        for (unsigned j=m_tuples.size()-1 ; j != 0 ; --j) {
            if ( is_equal_ignoring_column(i, m_tuples[j-1], m_tuples[j])) {
                m_tuples[j-1][i].insert_all_of(m_tuples[j][i]);
                m_tuples[j].swap(m_tuples.back());
                m_tuples.pop_back();
            }
        }
    }
}

// static
bool tuple_set::is_equal_ignoring_column(unsigned column,
                                         const std::vector<iset> & a,
                                         const std::vector<iset> & b)
{
    if (a.size() != b.size())
        return false;
    for (unsigned i=0 ; i != a.size() ; ++i)
        if (i != column && a[i] != b[i])
            return false;
    return true;
}

void tuple_set::push_back(const vector<int> & tuple)
{
    m_tuples.push_back(vector<iset>());
    vector<iset> & new_tuple= m_tuples.back();
    new_tuple.reserve(tuple.size());
    
    for (unsigned i=0 ; i != tuple.size() ; ++i)
        new_tuple.push_back(iset(tuple[i]));
}

namespace hcsp {
ostream & operator<<(ostream & out, const tuple_set & the_set)
{
    if (the_set.m_tuples.empty())
        return out << "[]";
    ostringstream temp_out;
    const string empty;
    vector<size_t> sizes;
    for (unsigned i=0; i != the_set.m_tuples.size() ; ++i) {
        const vector<iset> & tuple= the_set.m_tuples[i];
        for (unsigned j=0 ; j != tuple.size() ; ++j) {
            temp_out.str(empty);
            temp_out << tuple[j];
            if ( j < sizes.size())
                sizes[j]= max(sizes[j], temp_out.str().size());
            else
                sizes.push_back(temp_out.str().size());
        }
                
    }
    size_t total_size= accumulate(sizes.begin(), sizes.end(), 0);
    out << "[";
    bool is_wide= total_size + sizes.size()*2> 78;
    for (unsigned i=0; i != the_set.m_tuples.size() ; ++i) {
        if (i !=0)
            out << " ";
        out << "(";
        const vector<iset> & tuple= the_set.m_tuples[i];
        for (unsigned j=0 ; j != tuple.size() ; ++j) {
            if (is_wide) {
                out << tuple[j];
            } else {
                temp_out.str(empty);
                temp_out << tuple[j];
                out << setw(sizes[j]) << temp_out.str();
            }

            if (j+1 != tuple.size() +1) {
                if (is_wide)
                    out << ";\n   ";
                else
                    out << "; ";
            }
        }
        if (is_wide)
            out << "\n ) \n";
        else
            out << ") \n";
    }
    out << "]";
    return out;
}
} // namespace hcsp

void tuple_set::show_self() const
{
    cout << *this << endl;
}
