// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <config.hpp>
#include <cassert>
#include <time.h>
#include <sys/time.h>
#include <signal.h>
#include <fstream>
#include <iomanip>
#include <stdexcept>
#include <cstdlib>
#include <limits>
#include <cstring>
// Under visual C++ 2005 there is no such header file.
// Fortunately it is not needed, since clock() has exactly the milliseconds
// granularity we need.
#ifdef HAS_SYS_RESOURCE_H
#  include <sys/resource.h>
#endif
#ifdef HAS_WINDOWS_H
#  include <windows.h>
#endif
#include <time.h>               // Needed for clock() and CLOCKS_PER_SEC
#include "utilities.hpp"
using namespace std;
using namespace hcsp;

#ifdef HAS_SIGPROF
static struct sigaction s_prev_action;
#endif

// Defined here, so that the virtual table will go to only to utilities.o
parse_error::~parse_error() throw()
{
}

ios_state_restore::ios_state_restore(basic_ios<char>& ios)
    : m_saved_flags(ios.flags()),
      m_saved_state(ios.rdstate()),
      m_stream(ios)
{
}

ios_state_restore::~ios_state_restore()
{
    m_stream.flags(m_saved_flags);
    m_stream.clear(m_saved_state);
}


// The lower three bits map from a byte to the index of its LSB.
// Bits 3,4,5  map from a byte to the index of its MSB.
// NOTE that this table has been autogenerated using the following perl code
/*
#!/usr/bin/perl
for $x(0..255) {
  for ($lsb=0 ; $lsb < 7 && (($x >> $lsb) & 1) == 0 ; ++$lsb) { }
  for ($msb=7 ; $msb > 0 && ($x >> $msb)       == 0 ; --$msb) { }
  print "0$msb$lsb";
  print ($x < 255 ? "," : " ");
  printf(" // %X\n",$x/16) if ($x %16 == 15);
}
*/
//   0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
static const unsigned char g_indexesForByte[256] = {
    007,000,011,010,022,020,021,020,033,030,031,030,032,030,031,030, // 0
    044,040,041,040,042,040,041,040,043,040,041,040,042,040,041,040, // 1
    055,050,051,050,052,050,051,050,053,050,051,050,052,050,051,050, // 2
    054,050,051,050,052,050,051,050,053,050,051,050,052,050,051,050, // 3
    066,060,061,060,062,060,061,060,063,060,061,060,062,060,061,060, // 4
    064,060,061,060,062,060,061,060,063,060,061,060,062,060,061,060, // 5
    065,060,061,060,062,060,061,060,063,060,061,060,062,060,061,060, // 6
    064,060,061,060,062,060,061,060,063,060,061,060,062,060,061,060, // 7
    077,070,071,070,072,070,071,070,073,070,071,070,072,070,071,070, // 8
    074,070,071,070,072,070,071,070,073,070,071,070,072,070,071,070, // 9
    075,070,071,070,072,070,071,070,073,070,071,070,072,070,071,070, // A
    074,070,071,070,072,070,071,070,073,070,071,070,072,070,071,070, // B
    076,070,071,070,072,070,071,070,073,070,071,070,072,070,071,070, // C
    074,070,071,070,072,070,071,070,073,070,071,070,072,070,071,070, // D
    075,070,071,070,072,070,071,070,073,070,071,070,072,070,071,070, // E
    074,070,071,070,072,070,071,070,073,070,071,070,072,070,071,070  // F
};

unsigned hcsp::find_lsb_index(uint32_t bits)
{
    if (bits & 0x0000FFFF) {
        if (bits & 0x000000FF)
            return 0 +  (g_indexesForByte[ bits      & 0xFF] & 07);
        else // 0x0000FF00
            return 8 +  (g_indexesForByte[(bits>> 8) & 0xFF] & 07);
    } else { // 0xFFFF0000
        if (bits & 0x00FF0000)
            return 16 + (g_indexesForByte[(bits>>16) & 0xFF] & 07);
        else // 0 or 0xFF000000
            return 24 + (g_indexesForByte[ bits>>24        ] & 07);
    }
}

unsigned hcsp::find_msb_index(uint32_t bits)
{
    if (bits & 0xFFFF0000) {
        if (bits & 0xFF000000  )
            return 24 + (g_indexesForByte[ bits>>24        ] >> 3);
        else // 0x00FF0000
            return 16 + (g_indexesForByte[(bits>>16) & 0xFF] >> 3);
    } else { // 0x0000FFFF
        if (bits & 0x0000FF00)
            return 8  + (g_indexesForByte[(bits>> 8) & 0xFF] >> 3);
        else // 0 or 0x000000FF
            return 0  + (g_indexesForByte[ bits      & 0xFF] >> 3);
    }
}

// NOTE that this table has been autogenerated using the following perl code
/*
#!/usr/bin/perl
sub count {
    my $count=0;
    foreach (0..7) { ++$count if ($_[0] & (1<<$_));}
    return $count;
}
foreach (0..127) {
    printf ("0x%x%x", count($_*2+1), count($_*2));
    print ($_ < 127 ? "," : " ");
    printf(" // %X\n",$_/8) if ($_ % 8== 7);
}
*/
unsigned hcsp::count_ones(uint32_t bits)
{
    unsigned count= 0;
    static const unsigned char count_map[128]= {
        //10   32   54   76   98   BA   DC   FE (the LSB of the int)
        0x10,0x21,0x21,0x32,0x21,0x32,0x32,0x43, // 0
        0x21,0x32,0x32,0x43,0x32,0x43,0x43,0x54, // 1
        0x21,0x32,0x32,0x43,0x32,0x43,0x43,0x54, // 2
        0x32,0x43,0x43,0x54,0x43,0x54,0x54,0x65, // 3
        0x21,0x32,0x32,0x43,0x32,0x43,0x43,0x54, // 4
        0x32,0x43,0x43,0x54,0x43,0x54,0x54,0x65, // 5
        0x32,0x43,0x43,0x54,0x43,0x54,0x54,0x65, // 6
        0x43,0x54,0x54,0x65,0x54,0x65,0x65,0x76, // 7
        0x21,0x32,0x32,0x43,0x32,0x43,0x43,0x54, // 8
        0x32,0x43,0x43,0x54,0x43,0x54,0x54,0x65, // 9
        0x32,0x43,0x43,0x54,0x43,0x54,0x54,0x65, // A
        0x43,0x54,0x54,0x65,0x54,0x65,0x65,0x76, // B
        0x32,0x43,0x43,0x54,0x43,0x54,0x54,0x65, // C
        0x43,0x54,0x54,0x65,0x54,0x65,0x65,0x76, // D
        0x43,0x54,0x54,0x65,0x54,0x65,0x65,0x76, // E
        0x54,0x65,0x65,0x76,0x65,0x76,0x76,0x87  // F
    };
    
    for (unsigned i=0 ; i != sizeof(bits) ; ++i) {
        unsigned char a_byte= bits >> 8*i;
        unsigned byte_data= count_map[a_byte >> 1];
        unsigned nibble_pos= (a_byte & 1) * 4;
        count += (byte_data >> nibble_pos) & 0x0F;
    }
    return count;
}

// NOTE that this table has been autogenerated using the following perl code
/*
#!/usr/bin/perl
sub ranges {
  my ($v, $x, $strip)=(0, @_);
  while ($strip) { $strip= ($x >= 0x80); $x= ($x << 1) & 0xFF; }
  while ($x) {
    ++$v;
    while ( ($x & 1) == 0) { $x= $x >> 1; }
    while ( ($x & 1) == 1) { $x= $x >> 1; }
  }
  return $v;
}
foreach $x (0..255) {
  printf "%-2d", 16*ranges($x, 1)+ ranges($x, 0);
  print ($x < 255 ? ", " : "  ");
  printf(" // %X\n",$x/16) if ($x %16 == 15);
}
*/
unsigned hcsp::count_ranges_of_one(uint32_t bits)
{
    unsigned count= 0;
    // An entry i in this array, holds information on the number of 1's ranges
    // in the bit representation of 'i'. Each entry contains two parts:
    // - The lower 4 bits is the count of ranges in i's bits
    // - The upper 4 bits ignores the range ending with bit 7 (byte's MSB)
    //   when counting i's ranges.
    static const unsigned char count_map[256]= {
        0 , 17, 17, 17, 17, 34, 17, 17, 17, 34, 34, 34, 17, 34, 17, 17,  // 0
        17, 34, 34, 34, 34, 51, 34, 34, 17, 34, 34, 34, 17, 34, 17, 17,  // 1
        17, 34, 34, 34, 34, 51, 34, 34, 34, 51, 51, 51, 34, 51, 34, 34,  // 2
        17, 34, 34, 34, 34, 51, 34, 34, 17, 34, 34, 34, 17, 34, 17, 17,  // 3
        17, 34, 34, 34, 34, 51, 34, 34, 34, 51, 51, 51, 34, 51, 34, 34,  // 4
        34, 51, 51, 51, 51, 68, 51, 51, 34, 51, 51, 51, 34, 51, 34, 34,  // 5
        17, 34, 34, 34, 34, 51, 34, 34, 34, 51, 51, 51, 34, 51, 34, 34,  // 6
        17, 34, 34, 34, 34, 51, 34, 34, 17, 34, 34, 34, 17, 34, 17, 17,  // 7
        1 , 18, 18, 18, 18, 35, 18, 18, 18, 35, 35, 35, 18, 35, 18, 18,  // 8
        18, 35, 35, 35, 35, 52, 35, 35, 18, 35, 35, 35, 18, 35, 18, 18,  // 9
        18, 35, 35, 35, 35, 52, 35, 35, 35, 52, 52, 52, 35, 52, 35, 35,  // A
        18, 35, 35, 35, 35, 52, 35, 35, 18, 35, 35, 35, 18, 35, 18, 18,  // B
        1 , 18, 18, 18, 18, 35, 18, 18, 18, 35, 35, 35, 18, 35, 18, 18,  // C
        18, 35, 35, 35, 35, 52, 35, 35, 18, 35, 35, 35, 18, 35, 18, 18,  // D
        1 , 18, 18, 18, 18, 35, 18, 18, 18, 35, 35, 35, 18, 35, 18, 18,  // E
        1 , 18, 18, 18, 18, 35, 18, 18, 1 , 18, 18, 18, 1 , 18, 1 , 1    // F
    };
    
    for (unsigned i=0 ; i != sizeof(bits) ; ++i) {
        unsigned char a_byte= bits >> 8*i;
        unsigned bit_above= (bits >> 8*i >> 8) & 1;
        unsigned byte_data= count_map[ a_byte ];
        unsigned oct_pos= bit_above * 4;
        count += (byte_data >> oct_pos) & 7;
    }
    return count;
}

// For algorithm description (extended Euclidean) look at:
//     http://en.wikipedia.org/wiki/Extended_Euclidean_algorithm
pair<long long, long long> hcsp::bezout_coefficients(unsigned a, unsigned b,
                                                    unsigned * p_gcd)
{
    // r_i = a*x_i + b*y_i
    // r_1 = a, x_1=1, y_1=0
    // r_2 = b, x_2=0, y_2=1
    // q_i = floor(r_{i-2}/r{i-1})
    // r_i = r_{i-2} - q_i*r{i-1}
    // r_i = a*(x_{i-2}-q_i*x_{i-1}) + b*(y_{i-2}-q_i*y_{i-1})
    unsigned r_i_minus_2 = a, r_i_minus_1 = b;
    long long x_i_minus_2= 1, x_i_minus_1= 0;
    long long y_i_minus_2= 0, y_i_minus_1= 1;
    while (r_i_minus_1 > 0) {
        long long q_i = r_i_minus_2/r_i_minus_1;
        unsigned r_i = r_i_minus_2%r_i_minus_1;
        long long x_i= x_i_minus_2-q_i*x_i_minus_1;
        long long y_i= y_i_minus_2-q_i*y_i_minus_1;
        
        x_i_minus_2 = x_i_minus_1;
        x_i_minus_1 = x_i;

        y_i_minus_2= y_i_minus_1;
        y_i_minus_1= y_i;
        r_i_minus_2= r_i_minus_1;
        r_i_minus_1= r_i;
    }
    pair<long long, long long> ret(x_i_minus_2, y_i_minus_2);
    unsigned gcd= r_i_minus_2;
    if (ret.first<0) {
        ret.first+=(long long)b/gcd;
        ret.second-=(long long)a/gcd;
    }
    if (p_gcd)
        *p_gcd= gcd;
    return ret;
}

// Try to find a solution to a*x+b*x== rhs, when a,b or rhs are simple
// (all zero cases guaranteed to be covered).
// If a solution is found then has_solution== true.
//
// return true if a trivial case was handled and all output parameters were set
static bool trivial_linear_diophantine_equation(unsigned a, unsigned b,
                                                unsigned rhs,
                                                long long & x, unsigned &step_x,
                                                long long & y, unsigned &step_y,
                                                bool & has_solution)
{
    has_solution= false;
    if (rhs == 0) {
        x= y= 0;
        if (a == 0 || b == 0) {
            step_x= (a==0);
            step_y= (b==0);
        } else {
            unsigned g=gcd(a,b);
            step_x= b/g;
            step_y= a/g;
        }
        has_solution= true;
        return true;
    }
    if (a == 0) {
        if (b == 0 || rhs % b != 0)
            return true;
        y= -(long long)(rhs /b);
        x= 0;
        step_x=1;
        step_y=0;
        has_solution= true;
        return true;
    }
    if (b == 0) {
        if (rhs % a != 0)
            return true;
        x= (rhs /a);
        y= 0;
        step_x=0;
        step_y=1;
        has_solution= true;
        return true;
    }
    return false;
}

bool hcsp::linear_diophantine_equation(unsigned a, unsigned b, unsigned rhs,
                                      long long & x, unsigned & step_x,
                                      long long & y, unsigned & step_y,
				      unsigned *p_step_gcd)
{
    bool ret;
    if (trivial_linear_diophantine_equation(a,b,rhs, x, step_x, y, step_y,ret))
        return ret;
    unsigned gcd;
    pair<long long, long long> bezout= bezout_coefficients(a, b, & gcd);
    if (p_step_gcd)
	*p_step_gcd= gcd;
    if (rhs % gcd != 0)
        return false;
    a /= gcd; b /= gcd; rhs /= gcd;

    step_x= b;
    step_y =a;

    // Note that a*bezout.first+b*bezout.second == 1.
    // This means that for x'= rhs*bezout.first and y'=rhs*bezout.second:
    //         a*x' +b*y' == rhs
    // I.e., x' and y' are a solution to this linear Diophantine equation.
    // But x' or y' may be too big for 64 bit variables. We need to fix it.
    //
    // First we find an x such that 0 <= x < step_x which solves the equation.
    // In order to find this x we calculate (rhs*bezout.first) % step_x.
    // (we simply go back (rhs*bezout.first)/step_x steps of size step_x).
    // This is equivalent with (rhs*(bezout.first% step_x)) % step_x.
    x= ((unsigned long long)rhs*(bezout.first % step_x)) % step_x;

    // y =(a*x -rhs)/b 
    // x could be UINT_MAX-1, and a could be UINT_MAX, so a*x could be
    // nearly ULLONG_MAX. This means that it does not fit into long long.
    unsigned long long x_mult_a = (unsigned long long)x*a;
    if (x_mult_a <= rhs) { // subtraction is negative
        // The value of x_mult_a is not too big to fit a signed long long,
        // since rhs is smaller than pow(2,32).
        long long remains= (long long)x_mult_a - rhs;
        HCSP_ASSERT(remains % b == 0);
        y = remains/b;
    } else {  // subtraction is positive or zero
        unsigned long long remains= x_mult_a - rhs;
        HCSP_ASSERT(remains % b == 0);
        // Make sure the MSB is zero, otherwise it will be interpreted as a
        // sign bit (when converted to long long).
        // This should not happen since the following have to hold:
        //   - b == 1
        //   - but then, since, a*x+b*y=rhs  a solution would be
        //     x=0, y=-rhs.
        //   - Since the code has found the smallest x>=0 then it must have
        //     found x=0.
        //   - This implies that y=-rhs <0, which implies that 'remains' must
        //     be negative (contrary to the assumption).
        HCSP_ASSERT( ((remains / b) >> 63) == 0 );
        y = remains/b;
    }
    return true;
}


// static
volatile profile_data::ticks_type profile_data::m_global_clock_ticks= 0;
// static
bool profile_data::m_clock_active= false;

//static
void profile_data::Impl_sigprof_handler()
{
    ++profile_data::m_global_clock_ticks;
}

#if defined(HAS_SIGPROF)
static void sigprof_handler(int n,  siginfo_t *p_info, void *p_rest )
{
   profile_data::Impl_sigprof_handler();
   if ( s_prev_action.sa_handler != SIG_IGN &&
        s_prev_action.sa_handler != SIG_DFL)
       s_prev_action.sa_sigaction(n,p_info,p_rest); 
}

// static
void profile_data::activate_clock()
{
    if (m_clock_active)
        return;
    
    struct sigaction action;
    memset(&action, 0x00, sizeof(action));
    action.sa_sigaction= &sigprof_handler;
    action.sa_flags= SA_SIGINFO;
    sigaction(SIGPROF, &action, &s_prev_action);
    
    struct itimerval timer;
    timer.it_interval.tv_usec= 1000*1000/TICKS_PER_SECOND;
    timer.it_interval.tv_sec= 0;
    timer.it_value= timer.it_interval;
    
    setitimer(ITIMER_PROF, &timer, NULL);
    
    m_clock_active= true;
}
#elif defined(HAS_CreateThread)
static DWORD WINAPI time_counter( LPVOID p_args)
{
    pair<unsigned, profile_data::ticks_type*> args=
        *(pair<unsigned, profile_data::ticks_type*> *) p_args;
    static const int interval= 1000*1000/args.first;
    microseconds_t target= user_time_microseconds()+interval;
    while (true) {
        microseconds_t current= user_time_microseconds();
        if (current < target && (target-current)>=1000) {
            Sleep( (target-current) / 1000);
            continue;
        }
        ++*args.second;
        target += interval;
    }
    return 0;
}
void profile_data::activate_clock()
{
    if (m_clock_active)
        return;

    static pair<unsigned, volatile ticks_type*> args(TICKS_PER_SECOND+0,
                                                     &m_global_clock_ticks);

    DWORD threadId;
    CreateThread(NULL,              // default security attributes
                 0,                 // use default stack size  
                 &time_counter,     // thread function name
                 &args,             // argument to thread function 
                 0,                 // use default creation flags 
                 &threadId);        // returns the thread identifier 
    
    m_clock_active= true;
}
#else
#   error "Don't know how to set-up a profile timer on this system"
#endif

profile_data::profile_data(const char * label)
    : m_p_label(NULL), m_total_ticks_in_scope(0), m_scope_exits(0)
{
    activate_clock();
    size_t length= strlen(label)+1;
    m_p_label= new char[ length ];
    memcpy(m_p_label, label, length);
}

profile_data::~profile_data()
{
    ios_state_restore restorer(cout);
    cout << "At " << m_p_label << " total time " << setprecision(5)
         << static_cast<double>(m_total_ticks_in_scope)/TICKS_PER_SECOND
         << " total calls " << m_scope_exits << "\n";
    delete[] m_p_label;
    m_p_label= NULL;
}

static unsigned long clock_based_microseconds()
{
    // Although this compiles on visual c++ 2005, clock() has a slightly
    // different meaning on Windows. It measures the wall clock time rather
    // than the CPU time.
    unsigned clocks_per_usec= CLOCKS_PER_SEC/(1000*1000);
    if (clocks_per_usec >= 1)
        return clock()/clocks_per_usec;
    else
        return clock()*(1000*1000/CLOCKS_PER_SEC);
}

bool hcsp::trace_active()
{
    static bool active = getenv("ACTIVATE_HCSP_TRACE");
    return active;
}

microseconds_t hcsp::user_time_microseconds()
{
#if defined(HAS_RUSAGE)
    struct rusage usage;
    if (getrusage(RUSAGE_SELF, &usage) != 0) {
        return clock_based_microseconds();
    }
    microseconds_t ret= usage.ru_utime.tv_sec;
    ret *= 1000*1000;

    ret += usage.ru_utime.tv_usec;

    return ret;
#elif defined(HAS_GetProcessTimes)
    (void)clock_based_microseconds;
    FILETIME creationTime, exitTime, kernelTime, user_time;
    GetProcessTimes(GetCurrentProcess(),
                    &creationTime, &exitTime, &kernelTime, &user_time);
    static const int bits_in_low= sizeof(user_time.dwLowDateTime)*8;
    microseconds_t combined=
        (user_time.dwLowDateTime +
         ((microseconds_t)user_time.dwHighDateTime << bits_in_low));
    // combined is in units of 100 nano seconds, convert to microseconds
    // round to nearest;
    return (combined+5)/10;
#else
#   error "Don't know how to measure CPU time on this system"
#endif
}

unsigned hcsp::random::set_seed(int argc, char *argv[])
{
    unsigned seed=0;
    if (argc> 1)
        seed= strtoul(argv[1], NULL, 0);
    return set_seed(seed);
}

unsigned hcsp::random::set_seed(unsigned seed)
{
    if (seed == 0) {
        // This is not portable. On other systems should look into funny stuff
        // such as clock milli-seconds, process ID, activation count,
        // network load, CPU load, etc.
        ifstream randomStream("/dev/urandom");
        if (randomStream)
            randomStream.read(reinterpret_cast<char*>(&seed), sizeof(seed));
        if (!randomStream)
            // A poor random seed. FIXME: need to add other stuff here.
            seed= time(0);
    }
    srand(seed);
    return seed;
}

unsigned hcsp::random::gen()
{
    HCSP_TRACE("gen");
    const unsigned skip = (numeric_limits<unsigned>::digits+2)/3;
    // Not very portable. Assume that unsigned contains at most three RAND_MAX;
    assert(RAND_MAX >= (1U << skip)-1);
    return (0U+rand()) ^ ((0U+rand()) << skip) ^ ((0U+rand()) << 2*skip);
}

unsigned hcsp::random::gen(unsigned domain)
{
    HCSP_TRACE("gen");
    if (domain == 0)
        throw invalid_argument("random::gen(0) is not allowed");
    if (domain-1 < RAND_MAX)
        return rand() % domain;
    else
        return gen() % domain;
}

unsigned hcsp::random::range(unsigned from, unsigned to)
{
    if (from > to)
        throw invalid_argument("random::range: wrong parameter order");
    if (from == to)
        return from;
    else if (to-from+1 == 0)
        return gen();
    else
        return from+gen(to-from+1);
}

int hcsp::random::range(int from, int to)
{
    unsigned offset= numeric_limits<int>::min();
    return static_cast<int>(range(from-offset, to-offset)+offset);
}

void hcsp::throw_runtime_error(const char *msg)
{
    throw runtime_error(msg);
}
