// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2011,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2011,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <algorithm>
#include <string>
#include <sstream>
#include <set>
#include <cmath>
#include <flags.hpp>
#include <solver.hpp>
#include <clause_propagator.hpp>
#include <table_propagators.hpp>
#include <tuple_set.hpp>
#include <xcsp_solver.hpp>
#include <AST_builder.hpp>
// This is a hack to avoid errors due to issues in the CSP-XML-parser
#pragma GCC diagnostic ignored "-Wshadow"
#pragma GCC diagnostic ignored "-Wunused-parameter"
#pragma GCC diagnostic ignored "-Wsign-compare"
#pragma GCC diagnostic warning "-Wuninitialized"
#include <XMLParser_libxml2.hh>
using namespace std;
using namespace hcsp;
using namespace CSPXMLParser;


class xcsp_callbacks : public CSPParserCallback
{
  private:
    struct relation_info_type;
    class predicate_info_type;
    typedef AST_builder::unsupported unsupported;
    enum global_constraint_ids {
        ALLDIFFERENT=0,
        WEIGHTEDSUM=5
    };
    
  public:
    xcsp_callbacks(xcsp_solver & the_solver,
                   table_propagator_base::memory_manager & manager)
        : m_solver(the_solver),
          m_AST_builder(the_solver),
          m_current_domain_id(-1),
          m_current_relation_id(-1),
          m_current_constraint_reference_id(-1),
          m_memory_manager(manager)
    {}
    //! \brief Start a new problem instance with the given \p name
    virtual void beginInstance(const string & name)
    {
        m_instance_name= name;
        m_domains.clear();
    }

    //! \brief Get the number of domains.
    virtual void beginDomainsSection(int num_domains)
    {
        m_domains.resize(num_domains);
    }
    virtual void beginDomain(const string & domain_name,
                             int domain_id,
                             int num_values)
    {
        if (m_current_domain_id >=0)
            throw runtime_error( "did not finish previous domain" );

        const string & old_name=  m_domains.at(domain_id).first;
        if (old_name != "") {
            string error_text= domain_name+": overwriting existing "+old_name;
            throw runtime_error( error_text.c_str() );
        }
            
        m_domains[domain_id].first= domain_name;
        (void) num_values;
        m_current_domain_id= domain_id;
    }

    virtual void addDomainValue(int value)
    {
        m_domains.at(m_current_domain_id).second.insert(value);
    }

    virtual void addDomainValue(int from, int to)
    {
        m_domains.at(m_current_domain_id).second.insert(make_pair(from, to));
    }

    virtual void endDomain()
    {
        m_current_domain_id= -1;
    }

    virtual void endDomainsSection() {}

    virtual void beginVariablesSection(int num_variables)
    {
        m_variables.resize(num_variables);
        m_AST_builder.reserve_variables(num_variables);
    }
  
    virtual void addVariable(const string & variable_name, int variable_id, 
                             const string & domain_name, int domain_id)
    {
        const string & old_name= m_variables.at(variable_id).first;
        if (old_name != "") {
            string error_text= domain_name+": overwriting existing "+old_name;
            throw runtime_error( error_text.c_str() );
        }
        m_variables[variable_id].first=  variable_name;
        if ( m_domains.at(domain_id).first != domain_name)
            throw runtime_error( "mismatch domain name");
        const iset & domain= m_domains[domain_id].second;
        m_variables[variable_id].second= domain;
    }

    unsigned add_auxiliary_var(
        const string & name,
        const iset & domain= iset(make_pair(iset::MIN_DOMAIN,
                                            iset::MAX_DOMAIN)));
    unsigned add_auxiliary_var(
        const char *name_prefix, unsigned name_suffix,
        const iset & domain= iset(make_pair(iset::MIN_DOMAIN,
                                            iset::MAX_DOMAIN)));
    
    virtual void endVariablesSection()
    {
        for (unsigned i=0 ; i != m_variables.size() ; ++i) {
            if (m_variables[i].first == "")
                m_solver.add_var(iset(0), "a_dummy_0_var", true);
            else
                m_solver.add_var(m_variables[i].second,
                                 m_variables[i].first.c_str(), true);
        }
        m_variables.clear();
        m_solver.no_more_user_variables();
    }
    virtual void beginRelationsSection(int num_relations);  
    virtual void beginRelation(const string & relation_name, int relation_id,
                               int arity, int num_tuples,
                               RelType relation_kind);

    virtual void addRelationTuple(int arity, int tuple[]);

    virtual void addRelationTuple(int arity, int tuple[], int cost)
    {
        (void)arity;
        (void)tuple;
        (void)cost;
        throw unsupported("Soft tuples are not supported");
    }

    virtual void endRelation();
    virtual void endRelationsSection() {}

    virtual void beginPredicatesSection(int num_predicates)
    {
        (void)num_predicates;
    }
  
    virtual void beginPredicate(const string&predicate_name, int predicate_id);

    //! Add a formal parameter to the current predicate
    virtual void addFormalParameter(int parameter_index,
                                    const string & parameter_name,
                                    const string & parameter_type)
    {
        (void)parameter_index;
        (void)parameter_name;
        (void)parameter_type;
    }

    //! Assign an expression to the current predicate, using an AST.
    virtual void predicateExpression(AST *tree);

    //! Assign an expression to the current predicate, using the original text.
    virtual void predicateExpression(const string &)
    {
        throw unsupported("predicateExpression unsupported");
    }
        
    
    virtual void endPredicate() {}
    virtual void endPredicatesSection() {}

    /**
     * @brief Begin the section at which constraints are linked to variables.
     */
    virtual void beginConstraintsSection(int num_constraints)
    {
        (void) num_constraints; // we don't really need this.
    }
  
    /**
     * @brief Begin the declaration of a constraint instance and its linkage.
     *
     * @param constraint_name The name of this constraint instance.
     * @param constraint_id   The numeric ID of this constraint instance.
     * @param arity           The arity of this constraint.
     * @param reference_name  The name constraint's definition. It may be
     *                        any refer to any definition type that
     *                        reference_kind supports.
     * @param reference_kind RelationType,PredicateType or GlobalConstraintType
     * @param reference_id    The ID associated with reference_name.
     * @param scope           The parameters as seen from the rest of the CSP
     */
    virtual void beginConstraint(const string & constraint_name,
                                 int constraint_id,
                                 int arity, 
                                 const string & reference_name,
                                 CSPDefinitionType reference_kind,
                                 int reference_id,
                                 const ASTList &scope);


    /**
     * @brief If the weightedsum is an objective function, instantiate it
     *
     * If constraint name starts with Optimize:maximize or Optimize:minimize and
     * the sole parameter has a zero coefficient then set the variable as the
     * objective variable. The objective variable will be minimized or maximized
     * depending on the name of the constraint, which starts with either
     * Optimize:minimize or Optimize:maximize.
     *
     * @param[in] constraint_name  Either Optimize:minimize, Optimize:maximize,
     *                             otherwise not an objective function.
     * @param[in] coefficients     If not exactly one coefficient or its not
     *                             zero then this is not an objective function
     * @param[in] variables        If an objective function, this is the
     *                             variable to optimize.
     * @return True if objective function was instantiated.
     *                        
     * @pre Current constraint is global:weightedsum
     */
    bool maybe_set_objective(const string & constraint_name,
                             const vector<int> & coefficients,
                             const vector<unsigned> & variables);
    
    /**
     * @brief Instantiate global:weightedsum or objective
     *
     * If constraint name starts with Optimize:maximize or Optimize:minimize and
     * the sole parameter has a zero coefficient then set the variable as the
     * objective variable. The objective variable will be minimized or maximized
     * depending on the name of the constraint, which starts with either
     * Optimize:minimize or Optimize:maximize.
     *
     * @param[in] parameters  The parameters of weightedsum as defined in
     *                        XCSP2_1
     * @pre Current constraint is global:weightedsum
     */
    void instantiate_weighted_sum_or_objective(const ASTList &parameters);
    /**
     * @brief Set the list of parameters for this constraint.
     *
     * @param The list of parameters as seen from the constraint.
     *        A parameter may either refer to a CSP variable, may be
     *        an integer, an expression (unsupported) or a tag (unsupported).
     */
    virtual void constraintParameters(const ASTList &parameters);
    virtual void endConstraint() {}
    virtual void endConstraintsSection()
    {
        m_AST_builder.create_constraints();
    }
    virtual void endInstance() {}
    void detect_all_diff_enabled(bool flag)
    { m_AST_builder.detect_all_diff_enabled(flag); }
  private:

    /**
     * @brief Assign the expression to the given variable
     *
     * @param variables   The variables used by the whole expression
     * @param out_variable The variable containing the evaluated expression
     * @param expr        The subexpression to process.
     * Note: this may invalidate constraint_runtime_info references.
     *       Complete construction of parameters before creating the
     *       constraint, otherwise the creation of the parameters may
     *       invalidate the constraint_runtime_info before the parameters
     *       are added to it.
     */
    void connect_expr(const vector<unsigned> & variables,
                      unsigned out_variable, const AST & expr);
    
    void parameters_to_variables(const ASTList & parameters,
                                 vector<unsigned> &variables);
    void mark_unsat(); //! Mark the problem as UNSAT
    bool is_bool_variable(unsigned variable) const
    {
        const iset & domain= m_solver.var_domain(variable);
        return (domain.min() >= 0 && domain.max() <= 1) ;
    }

    typedef pair<string, iset> variable_info_type;
    
    xcsp_solver & m_solver;
    AST_builder m_AST_builder;
    string m_instance_name;
    int m_current_domain_id;
    vector<pair<string, iset> > m_domains;
    vector<variable_info_type>  m_variables;
    int m_current_relation_id;
    string m_current_constraint_name;
    vector<relation_info_type>  m_relations;
    vector<predicate_info_type> m_predicates;
    int m_current_constraint_reference_id;
    CSPDefinitionType m_current_constraint_reference_kind;
    table_propagator_base::memory_manager & m_memory_manager;
}; // class xcsp_callbacks

struct xcsp_callbacks::relation_info_type
{
    //! Different things that can be implied from this relation.
    enum special_relation_type {
        REGULAR,                //!< Nothing is inferred from this relation
        IMPLY_DIFF,             //!< Besides other semantics diff is implied
        DIFF,                   //!< Equivalent with a!=b
        EQUAL                   //!< Equivalent with a=b
    };
    relation_info_type() : m_special(REGULAR) {}
    relation_info_type(const relation_info_type & other) : m_special(REGULAR)
    { *this= other; }
    
    relation_info_type & operator=(const relation_info_type & other) 
    {
        if (other.m_constraint.get() == m_constraint.get())
            return *this;
        m_constraint.reset();
        if (other.m_constraint.get())
            m_constraint.reset(other.m_constraint->clone());
        m_name=            other.m_name;
        m_arity=           other.m_arity;
        m_is_conflict_set= other.m_is_conflict_set;
        m_tuples=          other.m_tuples;
        m_compact_tuples=  other.m_compact_tuples;
        m_special=         other.m_special;
        m_special_precondition_domains= other.m_special_precondition_domains;
        m_domains=         other.m_domains;
        
        return *this;
    }
    
    void build(table_propagator_base::memory_manager & manager);

    /**
     * @brief Check if the set of tuples has (a,a) tuple for every value
     *        in the tuple set.
     * A value \p a is in the tuple set if there is either (a,b) or (b,a)
     * tuple in the set.
     * @return true if every value has an equality tuple.
     * @pre Tuples are sorted and unique, otherwise false may be returned.
     */
    bool tuples_have_all_equal_pairs() const;
    //! Check if every tuple (a,b) satisfies a!=b/
    bool tuples_are_all_different() const;

    /**
     * @brief Check if the tuples represent diff.
     * @return true if represent a!=b
     * @pre The tuples are sorted and unique, otherwise it may return false
     */
    bool tuples_equivalent_with_diff() const;

    /**
     * @brief Check if tuples imply a!=b
     * @return true if every supported assign (a,b) satisfies a!=b
     * @pre The tuples are sorted and unique, otherwise it may return false
     */
    bool tuples_imply_diff() const;
    
    /**
     * @brief Check if the tuples represent equality
     * @return true if represent a==b
     * @pre The tuples are sorted and unique, otherwise it may return false
     */
    bool tuples_equivalent_with_equality() const;

    //! Check if all tuples are of the form (a,a)
    bool tuples_are_equal_pairs() const;

    /**
     * @brief Check if tuples represent a!=b.
     *
     * Consider the domain D which is the join of all values in the tuples.
     * Check that the tuple list is \f$T=\{a,b \in D | a \ne b \} \f$
     *
     * @pre The tuple list is sorted and unique. Otherwise, the method may
     *      return wrong 'false'
     */
    bool tuples_are_complementary_equal_pairs() const;

    /**
     * @brief Test if variable's domains are contained in the precondition.
     *
     * For each relation variable check if m_special_precondition_domains
     * contains all values of variable's domain.
     *
     * @param[in] solver             The solver (containing the domains)
     * @param[in] relation_variables Variable indices accessed by this relation
     */
    bool within_precondition_domains(
        const solver & solver,
        const vector<unsigned> & relation_variables) const;

    /**
     * @brief Collect the domain of one column of the tuples.
     *
     * Collect all values of a given position in all tuples.
     *
     * @param[in] parameter The position in each tuple to process.
     * @param[out] domain The collected values.
     */
    void collect_tuple_domain(unsigned parameter, iset & domain) const;
    //! Like collect_tuple_domain, but for all tuple positions.
    void collect_tuple_domains(vector<iset> & domains) const;

    void instantiate(xcsp_callbacks & callbacks,
                     const vector<unsigned> & relation_variables);
    void constrain_by_domains(xcsp_solver & the_solver,
                              const vector<unsigned> & relation_variables,
                              const vector<iset> & domains);
    
    void instantiate_conflict_set_relation(
        xcsp_solver & the_solver,
        const vector<unsigned> & relation_variables);
    void instantiate_support_set_relation(
        xcsp_solver & the_solver,
        const vector<unsigned> & relation_variables);

    string m_name;
    unsigned m_arity;
    bool m_is_conflict_set;
    vector< vector<int> > m_tuples;
    tuple_set m_compact_tuples;
    unique_ptr<constraint_propagator> m_constraint;

    special_relation_type m_special;
    //! The precondition to allow m_special processing. If vector is
    //! not empty, then all variable domains must be contained in
    //! m_special_precondition_domains.
    vector<iset> m_special_precondition_domains;
    //! An additional restriction over the domain of parameters.
    vector<iset> m_domains;
};

class xcsp_callbacks::predicate_info_type
{
  public:
    predicate_info_type(string name="", AST * tree= NULL)
        : m_tree(tree), m_name(name) { }
    const string & get_name() const { return m_name; }
    const AST * get() const { return m_tree; }
    void set(AST *tree) { m_tree = tree ; }
  private:
    AST * m_tree;
    string m_name;
};

template <typename Iterator>
bool is_sorted(Iterator from, Iterator to)
{
    if (from == to)
        return true;
    Iterator next= from;
    ++next;
    for ( ; next != to ; ++next, ++from)
        if (*next < *to)
            return false;
    return true;
}
void xcsp_callbacks::relation_info_type::collect_tuple_domains(
    vector<iset> & domains) const
{
    domains.clear();
    if (m_tuples.empty())
        return;
    domains.resize(m_tuples[0].size());

    for (unsigned i=0 ; i != domains.size() ; ++i)
        collect_tuple_domain(i, domains[i]);
}

void xcsp_callbacks::relation_info_type::collect_tuple_domain(
    unsigned parameter, iset & domain) const
{
    iset::value_type min_value= iset::MAX_DOMAIN, max_value= iset::MIN_DOMAIN;
    for (unsigned i=0 ; i != m_tuples.size() ; ++i) {
        min_value= min(min_value, m_tuples[i][parameter]);
        max_value= max(max_value, m_tuples[i][parameter]);
    }
    domain.clear();
    if ( (1LL+max_value-min_value)/sizeof(int) <= m_tuples.size()) {
        // prefer radix sort, when it takes not too much memory.
        vector<unsigned char> is_in_domain(1LL+max_value-min_value, false);
        for (unsigned i=0 ; i != m_tuples.size() ; ++i)
            is_in_domain.at(m_tuples[i][parameter]-min_value)= 1;
        for (unsigned i=0 ; i != is_in_domain.size() ; ++i)
            if (is_in_domain[i])
                domain.insert(i+min_value);
    } else {
        vector<int> values;
        values.reserve(m_tuples.size());
        for (unsigned i=0 ; i != m_tuples.size() ; ++i)
            values.push_back(m_tuples[i][parameter]);
        sort(values.begin(), values.end());
        values.erase(unique(values.begin(), values.end()), values.end());
        for (unsigned i=0 ; i != values.size() ; ++i)
            domain.insert(values[i]);
    }
}

bool xcsp_callbacks::relation_info_type::tuples_are_equal_pairs() const
{
    for (unsigned i=0 ; i != m_tuples.size() ; ++i)
        if (m_tuples[i][0] != m_tuples[i][1])
            return false;
    return true;
}

bool
xcsp_callbacks::relation_info_type::tuples_are_complementary_equal_pairs()const
{
    // If all tuples are unique, then  x != y for which D(x)=D(y)=N
    // gives M=N^2-N tuples. Check if M (i.e. number of tuples) can represent
    // x != y, (i.e. there is an integer N such that M=N^2-N).
    // This is solved by the quadratic equation N^2-N-M = 0
    // N=(1 +- sqrt(1+4M))/2, since N>0 and integer we get N=(1+sqrt(1+4M))/2
    unsigned domain_size=  (unsigned int)((sqrt(m_tuples.size()*4.0+1)+1.1)/2);
    if (domain_size*(domain_size-1) != m_tuples.size())
        return false;
    // No equal values
    for (unsigned i=0 ; i != m_tuples.size() ; ++i)
        if (m_tuples[i][0] == m_tuples[i][1])
            return false;
    // Tuples are sorted and unique
    for (unsigned i=1 ; i != m_tuples.size() ; ++i) {
        if (m_tuples[i-1][0] > m_tuples[i][0])
            return false;
        else if (m_tuples[i-1][0] == m_tuples[i][0] &&
                 // either unsorted or non-unique
                 m_tuples[i-1][1] >= m_tuples[i][1] ) 
            return false;
    }

    vector<iset> domains;
    collect_tuple_domains(domains);
    if ( domains[0] != domains[1] || domains[0].size() != domain_size)
        return false;
    else
        return true;
}

bool xcsp_callbacks::relation_info_type::tuples_equivalent_with_diff() const
{
    if (m_tuples.empty() || m_tuples[0].size() != 2)
        return false;
    if (m_is_conflict_set)
        return tuples_are_equal_pairs();
    else
        return tuples_are_complementary_equal_pairs();
}

bool xcsp_callbacks::relation_info_type::tuples_have_all_equal_pairs() const
{
    if (m_tuples.empty() || m_tuples[0].size() != 2)
        return false;
    iset::value_type last_equal_value= iset::MAX_DOMAIN;
    if (last_equal_value == m_tuples[0][0])
        last_equal_value= iset::MIN_DOMAIN;
    iset::value_type last_tested_value= last_equal_value;
    for (unsigned i=0 ; i != m_tuples.size(); ++i) {
        // Finished all tuples starting with last_tested_value?
        if (last_tested_value != m_tuples[i][0]) {
            if (last_tested_value != last_equal_value)
                return false;
            last_tested_value= m_tuples[i][0];
        }
        if (last_tested_value == m_tuples[i][1])
            last_equal_value= last_tested_value;
    }
    if (last_tested_value != last_equal_value)
        return false;
    vector<iset> domains;
    collect_tuple_domains(domains);
    return domains[0] == domains[1];
}

bool xcsp_callbacks::relation_info_type::tuples_are_all_different() const
{
    if (m_tuples.empty() || m_tuples[0].size() != 2)
        return false;
    for (unsigned i=0 ; i != m_tuples.size(); ++i) {
        if (m_tuples[i][0] == m_tuples[i][1])
            return false;
    }
    return true;
}


bool xcsp_callbacks::relation_info_type::tuples_imply_diff() const
{
    static bool disable_imply= getenv("DISABLE_IMPLY");
    if (disable_imply)
        return false;
    if (m_is_conflict_set)
        return tuples_have_all_equal_pairs();
    else
        return tuples_are_all_different();
}


bool xcsp_callbacks::relation_info_type::tuples_equivalent_with_equality()const
{
    if (m_tuples.empty() || m_tuples[0].size() != 2)
        return false;
    if (m_is_conflict_set)
        return tuples_are_complementary_equal_pairs();
    else
        return tuples_are_equal_pairs();
}

void xcsp_callbacks::relation_info_type::build(
    table_propagator_base::memory_manager & manager)
{
    bool is_sorted= true;
    for (unsigned i=1 ; i < m_tuples.size() && is_sorted ; ++i) {
        for (unsigned j=0 ; j != m_tuples[0].size() && is_sorted ; ++j) {
            if (m_tuples[i-1][j] < m_tuples[i][j])
                break;
            if (m_tuples[i-1][j] > m_tuples[i][j])
                is_sorted= false;
        }
    }
    if (!is_sorted)
        sort(m_tuples.begin(), m_tuples.end());
    m_tuples.erase(unique(m_tuples.begin(), m_tuples.end()),
                   m_tuples.end());
    if ( tuples_equivalent_with_diff() )
        m_special= DIFF;
    else if ( tuples_equivalent_with_equality() )
        m_special= EQUAL;
    else if ( tuples_imply_diff())
        m_special= IMPLY_DIFF;
    if (m_special != REGULAR && m_is_conflict_set)
        collect_tuple_domains(m_special_precondition_domains);
    if ( m_is_conflict_set ) {
        m_compact_tuples.insert(m_tuples);
    } else {
        m_constraint.reset(new support_table_propagator(manager, m_tuples));
        vector<vector<int> >().swap(m_tuples);
    }
}

//virtual
void xcsp_callbacks::beginRelationsSection(int num_relations)
{
    m_relations.resize(num_relations);
}

//virtual
void xcsp_callbacks::beginRelation(
    const string & relation_name, int relation_id, int arity, int num_tuples,
    RelType relation_kind)
{
    if (m_current_relation_id >=0)
        throw runtime_error( "did not finish previous relation" );
    relation_info_type & relation_data= m_relations.at(relation_id);
    string & name= relation_data.m_name;
    if (name != "") {
        string error_text= relation_name+": overwriting existing "+name;
        throw runtime_error( error_text.c_str() );
    }
    name=  relation_name;
    relation_data.m_arity= arity;
    switch (relation_kind) {
      case REL_SOFT:
        throw unsupported("REL_SOFT: soft relations are not supported");
      case REL_CONFLICT:
        relation_data.m_is_conflict_set= true;
        break;
      case REL_SUPPORT:
        relation_data.m_is_conflict_set= false;
        break;
    }
    relation_data.m_tuples.reserve(num_tuples);

    m_current_relation_id= relation_id;
}

//virtual
void xcsp_callbacks::addRelationTuple(int arity, int tuple[])
{
    relation_info_type &
        relation_data= m_relations.at(m_current_relation_id);
    if ( relation_data.m_arity != arity+0U )
        throw runtime_error("Mismatch arity");
    relation_data.m_tuples.resize(relation_data.m_tuples.size()+1);
    relation_data.m_tuples.back().reserve(arity);
    for (int i=0 ; i != arity ; ++i)
        relation_data.m_tuples.back().push_back(tuple[i]);
}

// virtual
void xcsp_callbacks::endRelation()
{
    m_relations.at(m_current_relation_id).build(m_memory_manager);
    m_current_relation_id= -1;
}

// virtual
void xcsp_callbacks::beginPredicate(const string& predicate_name,
                                    int predicate_id)
{
    if ( predicate_id < 0 || m_predicates.size() != (unsigned)predicate_id)
        throw runtime_error("Expecting a different predicate ID");
    m_predicates.push_back(predicate_info_type(predicate_name));
}

// virtual
void xcsp_callbacks::predicateExpression(AST *tree)
{
    if ( m_predicates.back().get() )
        throw runtime_error("Updating an expression for an old predicate");
    m_predicates.back().set(tree);
}

// virtual
void xcsp_callbacks::beginConstraint(const string & constraint_name,
                                     int constraint_id,
                                     int arity, 
                                     const string & reference_name,
                                     CSPDefinitionType reference_kind,
                                     int reference_id,
                                     const ASTList &scope)
{
    switch (reference_kind) {
      case PredicateType:
        if ( m_predicates.at(reference_id).get_name() != reference_name )
            throw runtime_error( "reference name mismatch" );
        break;
      case GlobalConstraintType:
        switch (reference_id) {
          case ALLDIFFERENT:
            HCSP_ASSERT(reference_name == "global:alldifferent");
            break;
          case WEIGHTEDSUM:
            HCSP_ASSERT(reference_name == "global:weightedsum");
            break;
          default:
            throw unsupported("GlobalConstraintType unsupported");
        }
        break;
      case RelationType:
        if (m_relations.at(reference_id).m_name != reference_name)
            throw runtime_error( "relation name mismatch" );
            
        break;
      default:
        throw unsupported("Unsupported reference type");
    }
    m_current_constraint_name= constraint_name;
    (void)constraint_id;
    (void)arity;
    (void)scope;
    m_current_constraint_reference_id= reference_id;
    m_current_constraint_reference_kind= reference_kind;
}

bool xcsp_callbacks::maybe_set_objective(
    const string & constraint_name,
    const vector<int> & coefficients,
    const vector<unsigned> & variables)
{
    HCSP_ASSERT(coefficients.size() == variables.size());
    if (coefficients.size() == 1 && coefficients[0] == 0) {
        if (constraint_name.find("Optimize:maximize") == 0) {
            m_solver.objective_is_minimizing(false);
            m_solver.objective_variable(variables.at(0));
            return true;
        } else if (constraint_name.find("Optimize:minimize") == 0) {
            m_solver.objective_is_minimizing(true);
            m_solver.objective_variable(variables.at(0));
            return true;
        }
    }
    return false;
}

void xcsp_callbacks::instantiate_weighted_sum_or_objective(
    const ASTList &parameters)
{
    HCSP_ASSERT(m_current_constraint_reference_kind == GlobalConstraintType);
    HCSP_ASSERT(m_current_constraint_reference_id == WEIGHTEDSUM);
    HCSP_ASSERT(parameters.size() == 3);
    long long k= parameters.getArg(2).getInteger();
    bool is_ge= false, is_le= false;
    switch (parameters.getArg(1).getType()) {
      case SYMB_GE: is_ge= true;      break;
      case SYMB_GT: is_ge= true; ++k; break;
      case SYMB_LE: is_le= true;      break;
      case SYMB_LT: is_le= true; --k; break;
      case SYMB_EQ: is_le= is_ge= true; break;
      default:
        stringstream err;
        err << "global:weightedsum does not support '";
        parameters.getArg(1).infixExpression(err);
        err << "'";
        throw unsupported(err.str().c_str());
    }
    vector<int> coefficients;
    vector<unsigned> variables;
    for (int i=0 ; i != parameters.getArg(0).size(); ++i) {
        int coeff= parameters.getArg(0).getArg(i)["coef"].getInteger();
        unsigned var= parameters.getArg(0).getArg(i)["var"].getVarId();
        coefficients.push_back(coeff);
        variables.push_back(var);
    }
    if (maybe_set_objective(m_current_constraint_name, coefficients, variables))
        return;
    if (is_ge)
        m_AST_builder.instantiate_weighted_sum(coefficients, variables, true,
                                               k);
    if (is_le)
        m_AST_builder.instantiate_weighted_sum(coefficients, variables, false,
                                               k);
}

// virtual
void xcsp_callbacks::constraintParameters(const ASTList &parameters)
{
    if ( m_current_constraint_reference_kind == GlobalConstraintType &&
         m_current_constraint_reference_id == WEIGHTEDSUM) {
        instantiate_weighted_sum_or_objective(parameters);
        return;
    }
    vector<unsigned> variables;
    parameters_to_variables(parameters, variables);
    switch (m_current_constraint_reference_kind) {
      case PredicateType: {
          unsigned predicate_id= m_current_constraint_reference_id;
          predicate_info_type & predicate_info = m_predicates.at(predicate_id);
              
          connect_expr(variables, m_solver.singleton_var(true),
                       *predicate_info.get());
              
          break;
      }
      case GlobalConstraintType: {
          if (m_current_constraint_reference_id != ALLDIFFERENT)
              throw unsupported("GlobalConstraintType unsupported");
          m_AST_builder.instantiate_all_different(variables);
          break;
      }
      case RelationType: {
          unsigned relation_id= m_current_constraint_reference_id;
          m_relations.at(relation_id).instantiate( *this, variables );
          break;
      }
      default:
        throw unsupported("Unsupported reference type");
    }
        
}

void xcsp_callbacks::parameters_to_variables(const ASTList & parameters,
                                             vector<unsigned> &variables)
{
    variables.reserve(parameters.nbArg());
    for (int i=0 ; i != parameters.nbArg(); ++i) {
        switch (parameters[i].getType()) {
          case CST_INT:
            variables.push_back(
                m_solver.singleton_var(parameters[i].getInteger()));
            break;
          case CST_BOOL:
            variables.push_back(
                m_solver.singleton_var(parameters[i].getBoolean()));
            break;
          case VAR:
            variables.push_back(parameters[i].getVarId());
            break;
          case LIST:
            parameters_to_variables(parameters[i].list(), variables);
            break;
          default: {
              ostringstream error_text;
              error_text << "Unsupported expression in parameter " << i << ":";
              parameters[i].infixExpression(error_text);
              throw unsupported(error_text.str().c_str());
          }
        }
    }
}

void xcsp_callbacks::connect_expr(const vector<unsigned> & vars,
                                  unsigned out_var,
                                  const AST & expr)
{
    m_AST_builder.connect_expr(vars, out_var, expr);
}
            
bool xcsp_callbacks::relation_info_type::within_precondition_domains(
    const solver & solver,
    const vector<unsigned> & relation_variables) const
{
    for (unsigned i=0 ; i != m_special_precondition_domains.size() ; ++i) {
        unsigned variable= relation_variables[i];
        const iset & domain= solver.var_domain(variable);
        if (!m_special_precondition_domains[i].contains_all_of(domain))
            return false;
    }
    return true;
}

void xcsp_callbacks::relation_info_type::instantiate(
    xcsp_callbacks & callbacks,
    const vector<unsigned> & relation_variables)
{
    bool done= false;
    if ( m_special != REGULAR &&
         within_precondition_domains(callbacks.m_solver, relation_variables)) {
        HCSP_ASSERT(relation_variables.size() == 2);
        unsigned a= relation_variables[0];
        unsigned b= relation_variables[1];
        done= true;
        if (m_special == DIFF)
            callbacks.m_AST_builder.differentiate_variables(a, b);
        else if (m_special == EQUAL)
            callbacks.m_AST_builder.equate_variables(a, b);
        else {
            HCSP_ASSERT(m_special == IMPLY_DIFF);
            done= false;
            callbacks.m_AST_builder.mark_unequal_variables(a,b);
        }
    }
    xcsp_solver & the_solver= callbacks.m_solver;
    if (!done && m_is_conflict_set)
        instantiate_conflict_set_relation(the_solver, relation_variables);
    else if (!done)
        instantiate_support_set_relation(the_solver, relation_variables);

    constrain_by_domains(the_solver, relation_variables, m_domains);
}

void xcsp_callbacks::relation_info_type::constrain_by_domains(
    xcsp_solver & the_solver,
    const vector<unsigned> & relation_variables,
    const vector<iset> & domains)
{
    // constrain the variables according to the domains.
    for (unsigned i=0 ; i != domains.size() ; ++i) {
        unsigned var_id= relation_variables.at(i);
        const iset & current_domain= the_solver.var_domain(var_id);
        
        if (!domains[i].contains_all_of(current_domain)) {
            iset next_domain = the_solver.var_domain(var_id);
            next_domain.intersect(domains[i]);
            the_solver.set_var_domain(var_id, move(next_domain));
        }
    }
}

void xcsp_callbacks::relation_info_type::instantiate_support_set_relation(
    xcsp_solver & the_solver,
    const vector<unsigned> & relation_variables)
{    
    unique_ptr<constraint_propagator> constraint(m_constraint->clone());
    constraint_connectivity_info &
        info= the_solver.add_constraint(move(constraint));
    for (unsigned i=0 ; i != relation_variables.size() ; ++i)
        info << relation_variables[i];
}

void xcsp_callbacks::relation_info_type::instantiate_conflict_set_relation(
    xcsp_solver & the_solver,
    const vector<unsigned> & relation_variables)
{
    const vector<vector<iset> > & iset_tuples= m_compact_tuples.tuples();
    for (unsigned i=0 ; i != iset_tuples.size(); ++i) {
        const vector<iset> & tuple= iset_tuples[i];
        clause_propagator * p_clause= new clause_propagator;
        clause_propagator::literals_type &lits=p_clause->update_literals();
        unique_ptr<constraint_propagator> auto_free_constraint(p_clause);
        constraint_connectivity_info &
            info= the_solver.add_constraint(move(auto_free_constraint));
        info.reserve_arguments(relation_variables.size());
        lits.reserve(relation_variables.size());
        for (unsigned j=0 ; j != tuple.size() ; ++j) {
            lits.push_back(tuple[j]);
            lits[j].complement();
            info << relation_variables[j];
        }
    }
}

xcsp_solver::xcsp_solver(istream & in,
                         const std::string &,
                         const hcsp_flags & flags,
                         table_propagator_base::memory_manager & manager)
{
    xcsp_callbacks callbacks(*this, manager);
    callbacks.detect_all_diff_enabled(flags.use_all_diff);
    XMLParser_libxml2<> parser(callbacks);
    parser.parse(in);
    flags.apply(*this);
}
