// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2011,2012,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2011,2012,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#ifndef AST_BUILER_HPP_
#define AST_BUILER_HPP_

#include <memory>
#include <vector>
#include <set>
#include <map>
#include <AST.hh>
#include <solver.hpp>
#include <operator_connectivity.hpp>
#include <variable_print_strategy.hpp>

namespace hcsp {
/**
 * @invariant
 *    For each i,j:
 *      - If m_diff_set[i].count(j)>0 then m_diff_set[j].count(i)>0
 *      - If m_diff_set[i].count(i)>0 then is_marked_unsat().
 *      - If !m_diff_set[i].empty() then \n
 *        i==m_operator_connectivity.get_equivalent(i);
 */
class AST_builder {
  public:
    typedef CSPXMLParser::AST AST;
    typedef CSPXMLParser::ASTList ASTList;
    enum usage_type {
        EQUATED,                //!< The expression evaluation should be exact
        BOUNDED_BELOW,          //!< The expression is only bounded from below.
        // var --> formula, is the same as (int)var <= (int)formula
        IMPLIED= BOUNDED_BELOW, //!< The expression is implied by the variable
        BOUNDED_ABOVE,          //!< The expression is only bounded from above.
        // formula --> var, is the same as (int)var >= (int)formula
        IMPLYING=BOUNDED_ABOVE  //!< The expression must imply the variable
    };

    class predicate_info_type;
    class unsupported : public std::exception
    {
      public:
        unsupported(const char * text) : m_text(text) {}
        unsupported(const std::string & text) : m_text(text) {}
        virtual const char * what() const throw() { return m_text.c_str(); }
        virtual ~unsupported() throw() {}
      private:
        std::string m_text;
    };
    
  public:
    AST_builder(solver & the_solver)
        : m_solver(the_solver),
          m_operator_connectivity(the_solver),
          m_detect_all_diff_enabled(true)
    {}

    //! Mark if all_different should be automatically detected and constructed
    void detect_all_diff_enabled(bool state) { m_detect_all_diff_enabled=state;}

    hcsp::variable_printer var_printer(unsigned var) const
    { return m_solver.var_printer(var); }

    unsigned add_auxiliary_var(
        const std::string & name,
        const iset & domain= iset(std::make_pair(iset::MIN_DOMAIN,
                                                 iset::MAX_DOMAIN)));
    unsigned add_auxiliary_var(
        const char *name_prefix, unsigned name_suffix,
        const iset & domain= iset(std::make_pair(iset::MIN_DOMAIN,
                                                 iset::MAX_DOMAIN)));
    
    void reserve_variables(unsigned num_variables)
    {
        m_operator_connectivity.vars_reserve(num_variables);
    }

    iset::value_type evaluate(const AST& expression) const;

    void create_constraints()
    {
        if (m_detect_all_diff_enabled)
            detect_all_diff();
        m_operator_connectivity.create_all_constraints();
    }
    void instantiate_all_different(const std::vector<unsigned> &variables);
    /**
     * @brief Instantiate global:weightedsum
     *
     * Instantiate the constraint a_1*x_1 + a_2*x_2 + ... OP k
     * where OP is either '>=' or '<=', i.e., add it to the solver.
     *
     * @param[in] coefficients  The integer coefficients a_1,a_2,a_3,...
     * @param[in] variables     The variables x_1,x_2,x_3, .. which are
     *                          multiplied by coefficients a_1,a_2,a_3,...
     * @param[in] is_ge         True if the operator OP is '>='.
     * @param[in] k             The integer bound of the sum
     */
    void instantiate_weighted_sum(const std::vector<int> &coefficients,
                                  const std::vector<unsigned> &variables,
                                  bool is_ge, long long k);
    
    //! Intersect the domain of the variable with the values provided.
    const iset & intersect_variable(const iset & values, unsigned variable);

    //! Make the two variables equal.
    void equate_variables(unsigned var_a, unsigned var_b);
    
    /**
     * @brief Mark these variables as different for later processing.
     *
     * The marking is used by xcsp_solver infer the all_different constraint.
     * @param[in] a,b  The two variable indexes of variables to be marked
     *                 different.
     * @pre a!=b
     */
    void mark_unequal_variables(unsigned a, unsigned b);
    
    /**
     * @brief Add different constraint.
     *
     * Add diff_propagators and mark the variables as different for
     * further processing
     * @param[in] a,b The two variable
     * indexes of variables to be marked different.
     * @pre a!=b
     */
    void differentiate_variables(unsigned a, unsigned b);

    /**
     * @brief Assign the expression to the given variable
     *
     * @param variables   The variables used by the whole expression
     * @param out_variable The variable containing the evaluated expression
     * @param expr        The subexpression to process.
     * Note: this may invalidate constraint_runtime_info references.
     *       Complete construction of parameters before creating the
     *       constraint, otherwise the creation of the parameters may
     *       invalidate the constraint_runtime_info before the parameters
     *       are added to it.
     */
    void connect_expr(const std::vector<unsigned> & variables,
                      unsigned out_variable, const AST & expr,
                      usage_type used_as= EQUATED);
    void connect_parity(const std::vector<unsigned> & vars, bool result);
    void mark_unsat();             //!< Mark the problem as UNSAT
    bool is_marked_unsat() const;  //!< Was the problem marked as UNSAT?
    /**
     * @brief Find a variable that represents this expression
     *
     * @param variables   The variables used by the whole expression
     * @param expr        The subexpression to process.
     * @param default_var Unless it is -1U try to connect this variable
     * @param used_as     The way the variable will be used by other
     *                    expressions. BOUNDED_ABOVE and BOUNDED_BELOW mean
     *                    that the variable is bounded by an external entity
     *                    only from above or below. EQUATED means that the
     *                    value must be evaluated exactly.
     * Note: this may invalidate constraint_runtime_info references.
     */
    unsigned expression_to_variable(
        const std::vector<unsigned> & variables, const AST & expr,
        unsigned default_var= -1U,
        usage_type used_as= EQUATED);
  private:
    
    //! \deprecated In the future connect_expr will only use
    //!             operator_definition_registry
    template <class Propagator>
    void connect(const std::vector<unsigned> & variables, unsigned out_var,
                 const AST & expr);
    //! Connect eq(simple_expr, other_expr) only if 'simple_expr' is a constant
    //! or a variable, otherwise return 'false'
    bool connect_if_simple_eq(const std::vector<unsigned> & variables,
                              const AST & simple_expr, const AST & other_expr);
    //! Connect neq(simple_expr, other_expr) only if 'simple_expr' is a constant
    //! or a singleton variable, otherwise return 'false'
    bool connect_if_simple_ne(const std::vector<unsigned> & variables,
                              const AST & simple_expr, const AST & other_expr);

    //! Do nodes of this type evaluate as a Boolean?
    static bool is_boolean_result(NodeType type)
    {
        switch (type) {
          case CST_BOOL: case F_EQ: case F_NE:  case F_GE:  case F_GT:
          case F_LE:     case F_LT: case F_NOT: case F_AND: case F_OR:
          case F_XOR:    case F_IFF:
            return true;
          default:
            return false;
        }
    }
    
    //! Connect V[var] <=> ( V[expr.getVar(0)] AND V[expr.getVar(1)])
    void connect_and(const std::vector<unsigned> & variables, unsigned var,
                     const AST & expr, usage_type used_as=EQUATED);
    //! Connect V[var] <=> ( V[expr.getVar(0)] OR V[expr.getVar(1)])
    void connect_or(const std::vector<unsigned> & variables, unsigned var,
                    const AST & expr, usage_type used_as=EQUATED);
    //! Connect (V[var]==expect_val) <=> (V[expr.getVar(0)]==V[expr.getVar(1)])
    void connect_eq(const std::vector<unsigned> & variables,
                    unsigned var, bool expect_val,
                    const AST & expr, usage_type used_as);
    /**
     * @brief Connect two variables which are connected by an equality operator
     *
     * @param result_var   The variable holding the result of comparison
     * @param expect_val   The required value for result_var. If expect_val==1
     *                     then we connect equality, otherwise we connect neq.
     * @param force_eq     Add a constraint that forces Var(a)==Var(b)
     * @param force_eq     Add a constraint that forces Var(a)!=Var(b)
     * @param a,b          Variable indices to connect
     * @param used_as      The way the caller intends to restrict result_var
     */
    void connect_eq_by_vars(unsigned result_var, bool expect_val,
                            bool force_eq, bool force_ne,
                            unsigned a, unsigned b, usage_type used_as);
    
    void connect_overlap(unsigned var,
                         const std::vector<unsigned> &argument_vars);
    void connect_sub_interval(unsigned var,
                              const std::vector<unsigned> &argument_vars);
    void connect_predicate(const std::vector<unsigned> & variables,
                           unsigned var, const AST & expr);
    //! Connect V[var] == ( V[expr.getVar(0)] ? V[expr.getVar(1)]
    //!                                       : V[expr.getVar(2)])
    void connect_if(const std::vector<unsigned> & variables,
                    unsigned var, const AST & expr);
    
    template <class Propagator>
    constraint_connectivity_info & add_constraint(unsigned arity)
    {
        std::unique_ptr<constraint_propagator> constraint(new Propagator);
        return m_solver.add_constraint(move(constraint)).reserve_arguments(
            arity);
    }
    template <class Propagator, class Param>
    constraint_connectivity_info & add_constraint(unsigned arity,
                                                  const Param & param)
    {
        std::unique_ptr<constraint_propagator> constraint;
        constraint.reset(new Propagator(param));
        return m_solver.add_constraint(move(constraint)).reserve_arguments(
            arity);
    }
    template <class Propagator, class Param1, class Param2>
    constraint_connectivity_info & add_constraint(unsigned arity,
                                                  const Param1 & param1,
                                                  const Param2 & param2)
    {
        std::unique_ptr<constraint_propagator> constraint;
        constraint.reset(new Propagator(param1, param2));
        return m_solver.add_constraint(move(constraint)).reserve_arguments(
            arity);
    }
    template <class Propagator, class Param1, class Param2, class Param3>
    constraint_connectivity_info & add_constraint(unsigned arity,
                                                  const Param1 & param1,
                                                  const Param2 & param2,
                                                  const Param3 & param3)
    {
        std::unique_ptr<constraint_propagator> constraint;
        constraint.reset(new Propagator(param1, param2, param3));
        return m_solver.add_constraint(move(constraint)).reserve_arguments(
            arity);
    }

    friend eval_type to_eval_type(usage_type usage)
    {
        switch(usage) {
          case EQUATED:       return STRICT_EVAL;
          case BOUNDED_ABOVE: return UPPER_BOUND_EVAL;
          case BOUNDED_BELOW: return LOWER_BOUND_EVAL;
        }
        return UNKNOWN_EVAL;
    }
    unsigned evaluate_if_singletons(
        const operator_definition *p_operator,
        const unsigned *begin_vars, const unsigned *end_vars);

    /**
     * @brief If already set, return the variable for a given argument.
     *
     * If the argument is a constant number then create and return a variable
     * for this argument.
     *
     * @param p_operator     The operator to work with
     * @param variables      The translation from AST's variable ID to global
     *                       variable indices.
     * @param out_var        The evaluation of the operator will be written to
     *                       this variable. The value -1U means that no
     *                       variable has been selected yet.
     * @param argument_expr  The expression representing this argument
     * @param argument_index The index of the queried argument
     */ 
    unsigned known_var_of_arg(const operator_definition *p_operator,
                              const std::vector<unsigned> &variables,
                              unsigned out_var, const AST & argument_expr,
                              unsigned argument_index);
    
    unsigned operator_to_variable(const operator_definition *p_operator,
                                  const std::vector<unsigned> &variables,
                                  const AST & expr, unsigned out_var,
                                  usage_type used_as);
    
    /**
     * @brief Find a variable that is the logical negation of this variable.
     *
     * Note: this may invalidate constraint_runtime_info references.
     */
    unsigned var_not_var(unsigned variable, usage_type usage = EQUATED);

    /**
     * @brief Find a variable that represents this boolean expression.
     *
     * The difference from expression_to_variable is that the domain of
     * the returned variable is intersected with {0,1}.
     * @param variables   The CSP variables used by the complete expression
     * @param expr        The subexpression to process
     * @param enabled_by  Only if any of the values are in out_variable
     *                    must this expression affect result
     *
     * Note: this may invalidate constraint_runtime_info references.
     */
    unsigned expression_to_bool_var(const std::vector<unsigned> & variables,
                                    const AST & expr,
                                    usage_type used_as)
    {
        using namespace std;
        unsigned out_var= expression_to_variable(variables, expr, -1U, used_as);
        const iset & out_values= m_solver.var_domain(out_var);
        static const iset bool_domain(std::make_pair(0,1));
        if (!out_values.contains_any_of(bool_domain)) {
            if (expr.getType() == CST_INT)
                throw runtime_error("putting a non 0,1 constant in a boolean "
                                    "context");
            else
                throw runtime_error("putting a non 0,1 expression in a boolean "
                                    "context");
        }
        intersect_variable(iset(std::make_pair(0,1)), out_var);
        return out_var;
    }

    /**
     * @brief Find a variable that represent the disequality 
     *        requested by this expression.
     * @param variables   Each entry translate expression variables
     *                    to global variables.
     * @param expr        The expression V[expr.getArg(0)]+ V[expr.getArg(1)]
     * @param default_var If possible, use this variable as return type.
     * @param enabled_by  Only if any of the values are in out_variable
     *                    must this expression affect result
     *
     * Note: this may invalidate constraint_runtime_info references.
     */
    unsigned ne_to_var(const std::vector<unsigned> &variables, const AST &expr,
                       unsigned default_var, usage_type used_as);
    
    void insert_clause(unsigned var_a, const iset & literal_a,
                       unsigned var_b, const iset & literal_b);
    void insert_clause(unsigned var_a, const iset & literal_a,
                       unsigned var_b, const iset & literal_b,
                       unsigned var_c, const iset & literal_c);
    /**
     * @brief Insert a clause defined by a list of variables and sets.
     *
     * @pre for each 0<=i<j<size: vars[i] != vars[j].
     *
     * @param vars,sets   The variables and sets that participate in the clause,
     *                    such that the i-th literal is vars[i] \in sets[i].
     * @param size        The size of the input arrays vars[] and sets[].
     *
     */
    void insert_clause(const unsigned vars[], const iset * const sets[],
                       unsigned size);

    /**
     * @brief Collect variable arguments of a global OR constraint.
     *
     * For expressions like or(or(or....)) collect all the arguments which
     * are leafs to the OR expressions. As an optimization, take into
     * account which output values (of the expression) affect global result.
     * Sometimes, when it helps, don't propagate values that would not affect
     * the whole CSP, as marked by 'enabled_by'
     * @param[out] or_variables
     *                    The variables which should be OR-ed. An empty list
     *                    marks a false OR expression.
     * @param expr        The OR expression.
     * @param variables   Each entry translate expression variables
     *                    to global variables.
     */
    void collect_or_vars(std::vector<unsigned> & or_variables,
                         const AST & expr,
                         const std::vector<unsigned> & variables,
                         usage_type used_as);
    
    /**
     * @brief Collect variable arguments of a global AND constraint.
     *
     * For expressions like and(and(and....)) collect all the arguments which
     * are leafs to the AND expressions. As an optimization, take into
     * account which output values (of the expression) affect global result.
     * Sometimes, when it helps, don't propagate values that would not affect
     * the whole CSP, as marked by 'used_as'
     * @param[out] and_variables
     *                    The variables which should be AND-ed. An empty list
     *                    marks a false AND expression.
     * @param expr        The AND expression.
     * @param variables   Each entry translate expression variables
     *                    to global variables.
     */
    void collect_and_vars(std::vector<unsigned> & and_variables,
                         const AST & expr,
                         const std::vector<unsigned> & variables,
                         usage_type used_as);
    /**
     * @brief Collect the first descendants which are different from this type
     *
     * This method collects descendants, left to right. It collects the
     * front-line of elements which have different type than \p type.
     * It pushes these descendants at the end of \p variables without
     * erasing older elements.
     *
     * @param expr        The sub-expression to start looking for nodes
     *                    different than \p type.
     * @param type        Collect the children of this type
     * @param descendants The collected descendants.
     */
    void insert_descendants(const AST & expr, NodeType type,
                            std::vector<const AST*> &descendants);
    bool is_bool_variable(unsigned variable) const
    {
        const iset & domain= m_solver.var_domain(variable);
        return (domain.min() >= 0 && domain.max() <= 1) ;
    }

    //! Normalize the m_diff_set structure according to the equivalence classes
    //! of variables.
    void normalize_diff_set();
    //! Extract 'all_different' constraints out of cliques of i!=j constraints
    void detect_all_diff();
    bool is_diff_with_all(unsigned var,
                          const std::vector<unsigned> &compared_vars);
    /**
     * Replace all occurrences of from_var with to_var
     *
     * If in the process we detect that a variable is different from itself
     * then mark_unsat() is called.
     *
     * @todo Move m_diff_set and methods to the same location as
     *       operator_connectivity::get_equivalent().
     *
     * @pre For each i,j:
     *      - If m_diff_set[i].count(j)>0 then m_diff_set[j].count(i)>0
     *      - If m_diff_set[i].count(i)>0 then is_marked_unsat().
     * @post For each i,j:
     *      - If m_diff_set[i].count(j)>0 then m_diff_set[j].count(i)>0.
     *      - If m_diff_set[i].count(i)>0 then is_marked_unsat().
     *      - if (i==from_var) m_diff_set[i].empty().
     *      - m_diff_set[i].count(from_var)==0.
     *      - foreach i in prev(m_diff_set[from_var]) : \n
     *        m_diff_set[to_var].count(i)>0.
     */
    void replace_var_in_diff_set(unsigned from_var, unsigned to_var);
    
    /**
     * @brief Tell if the two variables are known to be different.
     *
     * This method checks whether there is a not_equal constraint between the
     * equivalence sets of var_a and var_b. Or if their domains do not
     * share elements. This method will always return the
     * correct result, assuming that equivalent variables were marked by
     * equate_variables() and different variables by mark_unequal_variables().
     *
     * When var_a and var_b belong to the same equivalence set this method may
     * still return true. It means that there is a conflict in that equivalence
     * set, when variables are required to be, simultaneously, equal and
     * different. Clearly, this case means that the problem is unsatisfiable.
     * When this happens, then is_marked_unsat() is guaranteed to be \p true.
     *
     * @pre m_diff_set points only to representatives of variable equivalence
     *      sets. 
     *
     * @param var_a,var_b   The variables to check
     * @return True exactly when the variables belong to equivalence
     *         sets which are constrained to be different.
     *
     */
    bool are_ne(unsigned var_a, unsigned var_b) const;

    /**
     * @brief Tell if the two variables are known to be equal.
     *
     * This method checks whether the variables belong to the same equivalence
     * set or if their domains are equal and singletons.
     *
     * When var_a and var_b were marked different then this method may
     * still return true. It means that there is a conflict in that equivalence
     * set, when variables are required to be, simultaneously, equal and
     * different. Clearly, this case means that the problem is unsatisfiable.
     * When this happens, then is_marked_unsat() is guaranteed to be \p true.
     *
     * @return True exactly when the variables belong to the same equivalence
     *         set or have the same singleton domains.
     *
     */
    bool are_eq(unsigned var_a, unsigned var_b) const;
    void connect_all_diff(const std::vector<unsigned> &all_diff_vars,
                          bool is_first);
    
    solver & m_solver;
    operator_connectivity m_operator_connectivity;
    /**
     * @brief Map pair of variables to the cached evaluation of their
     *        (dis)equality.
     *
     * For two variables V1 and V2, m_is_eq_variables[make_pair(V1,V2)],
     * if defined contains a boolean variable X.second which forces V1 == V2 or
     * V1 != V2.
     *
     * X.first defines the exact relation between X.second and the V1,V2 pair.
     * It is assumed that X is boolean (i.e. its domain is {0,1}).
     * -# If X.first.contains(1) then X=1 implies V1=V2.
     * -# If X.first.contains(0) then X=0 implies V1!=V2
     * -# From the first two rules it is possible to conclude that
     *    if X.first.contains(iset(make_pair(0,1,))) then X=1 iff V1=V2.
     */
    std::map<std::pair<unsigned,unsigned>,
             std::pair<usage_type,unsigned> > m_is_eq_variables;
    std::vector<std::set<unsigned> > m_diff_set;
    //! Should the builder automatically detect and construct all_different.
    bool m_detect_all_diff_enabled;
};

std::unique_ptr<CSPXMLParser::AST> deep_copy(const CSPXMLParser::AST &expr);

class AST_builder::predicate_info_type
{
  public:
    predicate_info_type(std::string name="", AST * tree= NULL)
        : m_tree(tree), m_name(name) { }
    const std::string & get_name() const { return m_name; }
    const AST * get() const { return m_tree; }
    void set(AST *tree) { m_tree = tree ; }
  private:
    AST * m_tree;
    std::string m_name;
};

} // namespace hcsp
#endif // AST_BUILER_HPP_
