// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_ARITHMETIC_PROPAGATORS_HPP_
#define HCSP_ARITHMETIC_PROPAGATORS_HPP_

#include <iset.hpp>
#include <constraint_propagator.hpp>

namespace hcsp {
class int128;
struct arithmetic_propagation_types
{
    enum overflow_behavior_t { WRAP, TRUNCATE };
    static iset::value_type wrap(long long value);
    static iset::value_type saturate(long long value);

    
    static void intersect_wrapped(iset & out,
                                  long long first, long long second);
    static void intersect_truncated(iset & out,
                                    long long first, long long second);
    template <int on_overflow>
    struct ops;
};

template<>
struct arithmetic_propagation_types::ops<arithmetic_propagation_types::WRAP>
{
    static iset::value_type to_iset(long long value)
    { return wrap(value); }
    static void intersect(iset & out, long long first, long long second)
    { intersect_wrapped(out, first, second); }
};

template<>
struct arithmetic_propagation_types::ops<arithmetic_propagation_types::TRUNCATE>
{
    static long long to_iset(long long value)
    { return value; }
    static void intersect(iset & out, long long first, long long second)
    { intersect_truncated(out, first, second); }
};

/**
 * @brief Propagate the semantics of: v1 == -v2.
 *
 * The propagator implements a wrap-around logic. This means that the
 * equality is modulo domain_size(iset). If -v2 is below iset::MIN_DOMAIN
 * then the comparison becomes v1 == domain_size -v2
 */
class neg_propagator : public constraint_propagator
{
  public:
    virtual int count_params() const { return 2; }
    virtual result_type propagate(constraint_runtime_info & info);
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual neg_propagator* clone() const
    { return new neg_propagator(*this); }
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        iset::value_type neg_right= -info.domain(1).min();
        if (neg_right < iset::MIN_DOMAIN)
            neg_right += 1LL + iset::MAX_DOMAIN - iset::MIN_DOMAIN;
        return info.domain(0).min() == neg_right;
    }
}; // class neg_propagator

//! Propagate the semantics of: v1 == abs(v2).
class abs_propagator : public tight1_constraint_propagator
{
  public:
    virtual int count_params() const { return 2; }
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual abs_propagator* clone() const
    { return new abs_propagator(*this); }
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        iset::value_type result= info.domain(0).min(), in=info.domain(1).min();
        if (in <0 && -(long long)in > iset::MAX_DOMAIN)
            return false;
        return result == (in < 0 ? -in : in) ;
    }
  private:
    // Comments in the source file
    static
    iset::iterator insert_negative_supports(const iset & in0,
                                            iset::iterator it1,
                                            iset::iterator end1,
                                            iset &out0, iset &out1);
    // Comments in the source file
    static
    iset::iterator insert_nonnegative_supports(const iset & in0,
                                               iset::iterator it1,
                                               iset::iterator end1,
                                               iset &out0, iset &out1);
}; // class abs_propagator

//! Propagate the semantics of: v0 <= abs(v1).
class less_equal_abs_propagator : public tight_constraint_propagator
{
  public:
    virtual int count_params() const { return 2; }
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual less_equal_abs_propagator* clone() const
    { return new less_equal_abs_propagator(*this); }
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        iset::value_type result= info.domain(0).min(), in=info.domain(1).min();
        if (in < 0)
            return result <= 0LL-in; // avoid overflow of -MIN_INT by long long
        else
            return result <= in;
    }
  private:
    static long long value_abs(long long value)
    { return value<0 ? -value : value; }
}; // class less_equal_abs_propagator

//! Propagate the semantics of: v0 >= abs(v1).
class greater_equal_abs_propagator : public tight_constraint_propagator
{
  public:
    virtual int count_params() const { return 2; }
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual greater_equal_abs_propagator* clone() const
    { return new greater_equal_abs_propagator(*this); }
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        iset::value_type result= info.domain(0).min(), in=info.domain(1).min();
        if (in < 0)
            return result >= 0LL-in; // avoid overflow of -MIN_INT by long long
        else
            return result >= in;
    }
  private:
    static long long min_abs(const iset & domain);
}; // class greater_equal_abs_propagator
/**
 * @brief Propagate the semantics of: \f$ v_0 = v_1 + v_2 + ... + v_n \f$
 *
 * The propagator implements a wrap-around logic. This means that the
 * equality is modulo domain_size(iset).
 */
class sum_wrap_propagator : public tight1_constraint_propagator,
                            protected arithmetic_propagation_types
{
  private:
    typedef arithmetic_propagation_types::ops<WRAP> ops;
  public:
    virtual int count_params() const { return -1; }
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual sum_wrap_propagator* clone() const
    { return new sum_wrap_propagator(*this); }
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        long long sum=0 ;
        for (unsigned i=1 ; i != info.arity() ; ++i)
            sum += info.domain(i).min();
        return ops::to_iset(sum) == info.domain(0).min();
    }
  private:
    /**
     * @brief Propagate \f$ \mathrm{limited\_sum}=
     *                       v_1+v_2+ \ldots v_{\mathrm{elements\_to\_sum} }\f$
     *
     * This will propagate the sum of info.domain(1) up until
     * info.domain(elements_to_sum) from, and back to, limited_sum.
     * The domains of parameters 1..elements_to_sum are updated, but
     * info.domain(0) is ignored - neither consulted nor updated.
     *
     * @param[in,out] info             The propagated domains
     * @param[in,out] limited_sum      The domain of the resulting sum
     * @param[in]     elements_to_sum  Number of elements to sum up, starting
     *                                 with info.domain(1).
     * @return False if and only if the domains conflict with this constraint.
     */
    static bool propagate(constraint_runtime_info & info, iset & limited_sum,
                          unsigned elements_to_sum);
    
    //! for arity==2 propagate v0=v1
    static result_type propagate_equality( constraint_runtime_info & info );

    //! for arity==1 propagate v0=0
    static result_type propagate_zero( constraint_runtime_info & info );
}; // class sum_wrap_propagator

/**
 * @brief Propagate the semantics of: \f$ v_0 == v_1 + v_2\f$
 *
 * The propagator implements a truncate logic. This means that the
 * the constraint considers overflow values are unsupported.
 */
class sum_truncate_propagator : public tight1_constraint_propagator,
                                protected arithmetic_propagation_types
{
  public:
    int count_params() const override { return 3; }
    result_type propagate(constraint_runtime_info & info) override;
    bool conflicts(const constraint_runtime_info & info) const override;
    void print(std::ostream & out,
               const constraint_runtime_info *p_info) const throw() override;
    sum_truncate_propagator* clone() const override
    { return new sum_truncate_propagator(*this); }
    bool constraint_check(const constraint_runtime_info & info) const
    {
        long long sum= (long long)info.domain(1).min()+info.domain(2).min();
        return sum == info.domain(0).min();
    }
  private:
    void propagate(iset &a, iset & b, iset & result) const;
    void propagate_bounds(iset &a, iset &b, iset &c) const;
}; // sum_truncate_propagator

/**
 * @brief Propagate the semantics of:
 *
 * \f[
 *   m\_result=(v_0 + v_1 + ... + v_n) \mbox{mod} m\_mod \land
 *   (v_0,v_1,...,v_n \in [0,m\_mod-1])
 * \f]
 
 * If m_mod==2 then this constraint is equivalent with:
 *  XOR(v_0, v_1, ..., v_n)= m_result && (v_0,v_1,..., \in {0,1})
 *
 * @invariant  m_mod > 0
 */
class sum_mod_propagator : public constraint_propagator
{
  public:
    sum_mod_propagator(iset::value_type mod, iset::value_type result)
        : m_mod(mod), m_result(result)
    {
        if (mod <= 0)
            throw std::out_of_range("sum_mod_propagator(mod, result): mod "
                                    "should be positive");
    }
    
    int count_params() const override { return -1; }
    result_type propagate(constraint_runtime_info & info) override;
    bool conflicts(const constraint_runtime_info & info) const override;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info)
        const throw() override;
    sum_mod_propagator* clone() const override
    { return new sum_mod_propagator(*this); }
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        long long sum=0 ;
        for (unsigned i=0 ; i != info.arity() ; ++i) {
            const iset::value_type value = info.domain(i).min();
            if  (value < 0 || value >= m_mod)
                return false;
            sum += value;
        }
        return sum % m_mod == m_result;
    }
    constraint_propagator * intersect_with(
        const constraint_runtime_info & info,
        std::vector<unsigned> & copy_from_argument) const override;
    constraint_propagator *combine(
        const constraint_runtime_info & info,
        const constraint_propagator & other,
        const constraint_runtime_info & other_info,
        const std::vector<unsigned> & other_to_common_param,
        unsigned other_pivot_param,
        std::vector<unsigned> & copy_from_argument) const override;
    bool may_combine(const constraint_propagator & other) const override;
    iset::value_type mod_value() const { return m_mod; }
    iset::value_type mod_result() const { return m_result; }
  private:
    bool restrict_to_mod(constraint_runtime_info & info) const;

    constraint_propagator *combine(
        const constraint_runtime_info & info,
        const sum_mod_propagator & other,
        const std::vector<unsigned> & other_to_common_param,
        std::vector<unsigned> & copy_from_argument) const;
    constraint_propagator *combine_with_eq(
        const constraint_runtime_info & info,
        const std::vector<unsigned> & other_to_common_param,
        std::vector<unsigned> & copy_from_argument) const;
    
    iset::value_type m_mod;
    iset::value_type m_result;
}; // class sum_mod_propagator

/**
 * @brief Propagate the semantics of: v0 >= v1 + v2 + ...
 *
 * The propagator implements infinite arithmetic.
 */
class greater_equal_sum_propagator : public tight_constraint_propagator
{
  public:
    virtual int count_params() const { return -1; }
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual bool init( constraint_runtime_info & info );
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual greater_equal_sum_propagator* clone() const
    { return new greater_equal_sum_propagator(*this); }
    virtual constraint_propagator * intersect_with(
        const constraint_runtime_info & info,
        std::vector<unsigned> & copy_from_argument) const;
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        long long sum=0 ;
        for (unsigned i=1 ; i != info.arity() ; ++i)
            sum += info.domain(i).min();
        return info.domain(0).min() >= sum;
    }
}; // class greater_equal_sum_propagator

/**
 * @brief Propagate the semantics of: a0*v0+a1*v1+...+a_n*v_n >= k.
 *
 * The propagator implements infinite arithmetic.
 */
class linear_combination_ge_propagator : public constraint_propagator
{
  public:
    typedef int32_t coefficient_type;
    linear_combination_ge_propagator(iset::value_type combination_lower_bound)
        : m_k(combination_lower_bound)
    {}
    template <typename... Coefficients>
    linear_combination_ge_propagator(
        iset::value_type combination_lower_bound,
        const Coefficients &... coefficients_in)
        : m_k(combination_lower_bound),
          m_coefficients(coefficients_in...)
    {}
    template <typename... Coefficients>
    linear_combination_ge_propagator(
        iset::value_type combination_lower_bound,
        Coefficients &&... coefficients_in)
        : m_k(combination_lower_bound),
          m_coefficients(coefficients_in...)
    {}
    const std::vector<coefficient_type> & coefficients() const
    { return m_coefficients; }
    iset::value_type lower_bound() const { return m_k; }

    virtual int count_params() const { return m_coefficients.size(); }
    virtual result_type propagate(constraint_runtime_info & info)
    {
        if (m_small_coefficients)
            return const_propagate<int64_t>(info);
        else
            return const_propagate<int128>(info);
    }
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation) const;
    virtual bool init(constraint_runtime_info & info);
    virtual bool get_disjunction_specification(
        disjunction_specification & disjunction) const;
    virtual bool disjoin_constraint(
        const disjunction_specification & this_constraint,
        const disjunction_specification & other_constraint,
        disjunction_specification & result_constraint) const;
    virtual bool combine_base_constraints(
        disjunction_specification & this_constraint,
        disjunction_specification & other_constraint,
        unsigned pivot,
        disjunction_specification & result_constraint) const;
    virtual constraint_propagator * new_propagator(
        disjunction_specification & disjunction,
        std::vector<unsigned> &copy_from_argument) const;
    virtual bool may_combine(const constraint_propagator & other) const;
    virtual constraint_propagator * combine(
        const constraint_runtime_info & info,
        const constraint_propagator & other,
        const constraint_runtime_info & other_info,
        const std::vector<unsigned> & other_to_common_param,
        unsigned other_pivot_param,
        std::vector<unsigned> & copy_from_argument) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    constraint_propagator *intersect_with(
        const constraint_runtime_info & info,
        std::vector<unsigned> & copy_from_argument) const;    
    virtual linear_combination_ge_propagator* clone() const
    { return new linear_combination_ge_propagator(*this); }
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const;
  private:
    /**
     * @brief Create simplified coefficients based on the info domains.
     *
     * Remove all parameters with singleton domains and normalize the
     * coefficients.
     *
     * @param[in] info
     *        The domains in which to look for singletons.
     * @param[out] new_coefficients
     *        The updated coefficients if any update is required, otherwise
     *        empty.
     * @param[in,out] new_k
     *        The k value of the limit (a0*x0 + a1*x1+ a2*x2 + ... >= k). This
     *        value is read and then updated.
     * @param[in,out] copy_from_argument
     *        If empty and the coefficients need to be updated, then this
     *        vector is filled with 0,1,2,... , otherwise it is kept unchanged.
     * @return The new arity.
     * 
     * @pre There are no zero coefficients
     * @post If info.arity() != return then new_coefficients.size() ==  return.
     *       Otherwise new_coefficients may be empty.
     */
    unsigned compact_coefficients(
        const constraint_runtime_info & info,
        std::vector<coefficient_type> & new_coefficients,
        int64_t & new_k,
        std::vector<unsigned> & copy_from_argument) const;
    
    /**
     * @brief Divide all coefficients by their abs() gcd.
     *
     * @param[in,out] new_coefficients
     *        If not empty use these coefficients, otherwise use
     *        m_coefficients. Upon return, if any change is required, these are
     *        the updated coefficients.
     * @param[in,out] new_k
     *        The k value of the limit (a0*x0 + a1*x1+ a2*x2 + ... >= k). This
     *        value is read and then updated.
     * @param[in,out] copy_from_argument
     *        If empty and the coefficients need to be updated, then this
     *        vector is filled with 0,1,2,... , otherwise it is kept unchanged.
     * 
     * @pre There are no zero coefficients
     */
    void normalize_coefficients(
        unsigned next_arity,
        std::vector<coefficient_type> & new_coefficients, int64_t & new_k,
        std::vector<unsigned> & copy_from_argument) const;
    /**
     * @brief Create a constraint for which no variables are required.
     *
     * This is called in cases when the constraint may look like a*x0 >= k
     * and the domain is singleton D(x0)={1}. The constraint is either a
     * tautology or always false.
     *
     * @param[in]     new_k   The value of 'k' in the constraint 0>=k.
     * @param[in,out] copy_from_argument
     *                The arguments of the original constraint which will be
     *                used in the new constraint. This is needed since HaifaCSP 
     *                requires at least one variable for a tautology constraint.
     * @return The resulting constraint, or nullptr if there are implementation
     *         difficulties.
     */
    constraint_propagator *new_empty_sum(
        int64_t new_k, std::vector<unsigned> & copy_from_argument) const;
    /**
     * @brief Create a linear combination with one variable.
     *
     * Create a constraint of the form a0*x0 >= new_k.
     * Most likely it will create a simple literal rather than a constraint.
     *
     * @param[in]     new_k   The value of 'k' in the constraint 0>=k.
     * @param[in,out] new_coefficients
     *                If non-empty on entry then use these coefficients,
     *                otherwise use the coefficients m_coefficients.
     *                The contents of this vector is destroyed.
     * @param[in,out] copy_from_argument
     *                The arguments of the original constraint which will be
     *                used in the new constraint. 
     * @return The resulting constraint, or nullptr if there are implementation
     *         difficulties.
     *
     * @pre new_coefficients.size() == 1 || (
     *      new_coefficients.empty() && m_coefficients.size() == 1);
     */
    constraint_propagator *new_unary_sum(
        std::vector<coefficient_type> & new_coefficients, int64_t new_k,
        std::vector<unsigned> &copy_from_argument) const;
    /**
     * @brief Create a linear combination with two variables.
     *
     * Create a constraint of the form a0*x0 + a1*x1 >= new_k.
     * If can be written as x_i-x_j >= new_k creates difference_ge_propagator.
     * Otherwise it will create a linear_combination_ge_propagator.
     *
     * @param[in]     new_k   The lower bound of the linear combination.
     * @param[in,out] new_coefficients
     *                If non-empty on entry then use these coefficients,
     *                otherwise use the coefficients m_coefficients.
     *                The contents of this vector is destroyed.
     * @param[in,out] copy_from_argument
     *                The arguments of the original constraint which will be
     *                used in the new constraint. 
     * @return The resulting constraint, or nullptr if there are implementation
     *         difficulties.
     *
     * @pre new_coefficients.size() == 2 || (
     *      new_coefficients.empty() && m_coefficients.size() == 2);
     */
    constraint_propagator *new_binary_sum(
        std::vector<coefficient_type> & new_coefficients, int64_t new_k,
        std::vector<unsigned> &copy_from_argument) const;    
    /**
     * @brief Create a linear combination with unlimited number of variables.
     *
     * If always creates difference_ge_propagator.
     *
     * @param[in]     new_k   The lower bound of the linear combination
     * @param[in,out] new_coefficients
     *                If non-empty on entry then use these coefficients,
     *                otherwise use the coefficients m_coefficients.
     *                The contents of this vector is destroyed.
     * @param[in,out] copy_from_argument
     *                The arguments of the original constraint which will be
     *                used in the new constraint. 
     * @return The resulting constraint, or nullptr if there are implementation
     *         difficulties.
     */
    constraint_propagator *new_nary_sum(
        std::vector<coefficient_type> & new_coefficients, int64_t new_k,
        std::vector<unsigned> &copy_from_argument) const;

    template <typename SlackType>
    result_type const_propagate(constraint_runtime_info & info) const;
    template<typename SlackType>
    void propagate_to_argument(constraint_runtime_info & info,
                               unsigned argument, const SlackType &slack) const;
    /**
     * @brief Add the coefficients of constraint_spec to out_coefficients.
     *
     * For each variable in the linear combination or equivalent, which is
     * provided by constraint_spec, multiply its coefficient by 'factor' and add
     * the result to out_coefficients.
     *
     * @param[in,out] out_coefficients
     *       The coefficients for each variable. out_coefficients[i] refers to
     *       constraint_spec.constraint_var(j) == i, for some value of j.
     * @param[in,out] new_k
     *       The sum of the values of k*factor in a0*x0 + a1*x1 +... >=k
     * @param[in] constraint_spec
     *       The constraint to add to out_coefficients and new_k
     * @param[in] factor
     *       By how much to multiply the coefficients and 'k' before adding to
     *       the output.
     * @return True if all is fine.
     */
    static bool accumulate_coefficients(
        std::vector<int64_t> & out_coefficients, int64_t & new_k,
        const disjunction_specification & constraint_spec,
        uint32_t factor);
    /**
     * @brief Create a result constraint, without a disjunction
     *
     * Taking out_coefficients and new_k create a constraint and put it into
     * result_constraint. This will not interfere with any literal, if exists in
     * result_constraint. The resulting constraint may be of a type other than
     * linear_combination_ge_propagator, e.g., difference_ge_propagator.
     *
     * @param[in] out_coefficients,new_k
     *            The defining data for the constraint that we need to create.
     * @param[in,out] result_constraint
     *            The constraint specification to populate
     * @return True if all is fine.
     */
    static bool set_constraint_from_positional_coefficients(
        const std::vector<int64_t> & out_coefficients,
        int64_t new_k, disjunction_specification & result_constraint);
    
    constraint_propagator *combine_sum_mod(
        const constraint_runtime_info & info,
        const sum_mod_propagator & other,
        const constraint_runtime_info & other_info,
        const std::vector<unsigned> & other_to_common_param,
        unsigned other_pivot_param,
        std::vector<unsigned> & copy_from_argument) const;
    template <typename T1, typename T2>
    static constexpr bool fits(const T2 & value);

    static unsigned find_non_pivot(
        const constraint_runtime_info & info, 
        const std::vector<unsigned>& other_to_common_param,
        unsigned pivot);

    /**
     * @brief Find the coefficient of the variable in the input constraint.
     *
     * The constraint may be of many types, including difference_ge_propagator
     * and linear_combination_ge_propagator.
     *
     * @param[in] var             The variable to query
     * @param[in] constraint_spec The constraint to check
     * @return The coefficient.
     */
    static iset::value_type find_var_coefficient(
        unsigned var, const disjunction_specification & constraint_spec);
    static inline iset make_range(int64_t first, int64_t second);
    iset::value_type m_k;
    std::vector<coefficient_type> m_coefficients;
    bool m_small_coefficients = false;
}; // class linear_combination_ge_propagator

/**
 * @brief Propagate the semantics of: v0 <= v1 + v2 + ...
 *
 * The propagator implements infinite arithmetic.
 */
class less_equal_sum_propagator : public tight_constraint_propagator
{
  public:
    virtual int count_params() const { return -1; }
    virtual bool init( constraint_runtime_info & info );
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual less_equal_sum_propagator* clone() const
    { return new less_equal_sum_propagator(*this); }
    virtual constraint_propagator * intersect_with(
        const constraint_runtime_info & info,
        std::vector<unsigned> & copy_from_argument) const;
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        long long sum=0 ;
        for (unsigned i=1 ; i != info.arity() ; ++i)
            sum += info.domain(i).min();
        return info.domain(0).min() <= sum;
    }
}; // class less_equal_sum_propagator

//! A base class for constraint propagators which may sometimes retain
//! unsupported values in their domains.
template <class Propagator>
class approximated_arithmetic_propagator: public constraint_propagator,
                                          protected arithmetic_propagation_types
{
    typedef constraint_propagator parent_type;
  public:
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();

    virtual void explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation) const;
    
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        return info.domain(0).min() ==
            Propagator::evaluate(info.domain(1).min(),
                                 info.domain(2).min());
    }
  protected:
    virtual void propagate_accurately(iset &result, iset &left, iset &right)
        const;
    void propagate_bounds(iset domains[3],
                          const constraint_runtime_info & info) const;
    bool copy_domains_to_info(iset domains[3], constraint_runtime_info & info);
    /**
     * @brief Depending on their sizes copy either domains or their bounds to
     *        the target info.
     *
     * For each of the 3 parameters, and if domain[i] has more than
     * info.max_accurate_ranges() then propagate only bounds to info. Otherwise
     * propagate the whole domain.
     *
     * @param[in]     domains The domains to copy to the target info.
     * @param[in,out] info    The info to be updated.
     */
    bool copy_domains_or_bounds_to_info(const iset domains[3],
                                        constraint_runtime_info & info);
    
    virtual void propagate_unisigned_bounds(
        iset domains[3], const constraint_runtime_info & info) const = 0;
    virtual const char *short_name() const=0;
    virtual const char *long_name() const=0;
    
  protected:
    //! Does the domain, by itself, small enough to allow accurate propagation?
    bool is_accurate(const iset & domain, unsigned effort,
                     unsigned accurate_size_bound) const;
    //! Is the pair of domains small enough to allow accurate propagation?
    bool is_accurate(const iset & domain1, const iset & domain2,
                     unsigned effort, unsigned accurate_size_pair_bound) const;
};

//! Propagate the semantics of: v1 == (v2 * v3)
//! Overflowing values of v2*v3 are wrapped modulo the size of the domain.
class mul_wrap_propagator:
    public approximated_arithmetic_propagator<mul_wrap_propagator>
{
  public:
    virtual int count_params() const { return 3; }
    mul_wrap_propagator *clone() const
    { return new mul_wrap_propagator(*this); }
    // accessed from approximated_arithmetic_propagator<mul_wrap_propagator>
    static iset::value_type evaluate(iset::value_type left,
                                     iset::value_type right)
    { return ops::to_iset((long long)left * right); }
  private:
    typedef arithmetic_propagation_types::ops<WRAP> ops;
    
    virtual const char *short_name() const { return "*";   }
    virtual const char *long_name() const  { return "mul"; }
    virtual void propagate_unisigned_bounds(
        iset domains[3], const constraint_runtime_info & info) const;
}; // class mul_wrap_propagator

//! Propagate the semantics of: v1 == (v2 * v3)
//! Overflowing values of v2*v3 are removed.
class mul_truncate_propagator:
    public approximated_arithmetic_propagator<mul_truncate_propagator>
{
  public:
    class algo;
    virtual int count_params() const { return 3; }
    mul_truncate_propagator* clone() const
    { return new mul_truncate_propagator(*this); }
    
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    
    bool constraint_check(const constraint_runtime_info & info) const
    {
        return info.domain(0).min() == ((long long) info.domain(1).min()*
                                        info.domain(2).min());
    }
    
    static long long evaluate(iset::value_type a, iset::value_type b)
    { return (long long)a*b; }
  private:
    virtual const char *short_name() const { return "*";   }
    virtual const char *long_name() const  { return "mul"; }
    virtual void propagate_unisigned_bounds(
        iset[3], const constraint_runtime_info &) const
    {
        throw std::runtime_error(
            "mul_truncate_propagator::propagate_unisigned_bounds() "
            "should never be called");
    }
}; // class mul_truncate_propagator

 


//! Propagate the semantics of: v0 <= v1 * v2
class less_equal_mul_propagator: public tight_constraint_propagator
{
  public:
    virtual int count_params() const { return 3; }
    virtual bool init( constraint_runtime_info & info );
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    less_equal_mul_propagator *clone() const
    { return new less_equal_mul_propagator(*this); }
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        long long res= 1LL*info.domain(1).min() * info.domain(2).min();
        return info.domain(0).min() <= res;
    }
    virtual constraint_propagator * intersect_with(
        const constraint_runtime_info & info,
        std::vector<unsigned> & copy_from_argument) const;
  private:
    void propagate_to_multiplicand(unsigned multiplicand_index,
                                   constraint_runtime_info & info);
    
}; // class less_equal_mul_propagator


//! Propagate the semantics of: v0 >= v1 * v2
class greater_equal_mul_propagator: public tight_constraint_propagator
{
  public:
    virtual int count_params() const { return 3; }
    virtual bool init( constraint_runtime_info & info );
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    greater_equal_mul_propagator *clone() const
    { return new greater_equal_mul_propagator(*this); }
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        long long res= 1LL*info.domain(1).min()*info.domain(2).min();
        return info.domain(0).min() >=  res;
    }
    virtual constraint_propagator * intersect_with(
        const constraint_runtime_info & info,
        std::vector<unsigned> & copy_from_argument) const;
  private:
    void propagate_to_multiplicand(unsigned multiplicand_index,
                                   constraint_runtime_info & info);
}; // class greater_equal_mul_propagator


//! Propagate the semantics of: v1 == v2 / v3
class div_propagator: public approximated_arithmetic_propagator<div_propagator>
{
  public:
    friend class approximated_arithmetic_propagator<div_propagator>;
    virtual int count_params() const { return 3; }
    div_propagator *clone() const { return new div_propagator(*this); }
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        if (info.domain(2).min() == 0)
            return info.domain(0).min() == 0;
        else
            return (info.domain(0).min() ==
                    1LL*info.domain(1).min()/info.domain(2).min());
    }
    virtual void explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation) const
    {
        // Assume that propagate() was not overridden
        div_propagator div_copy(*this);
        explanation_generators::explain_propagation_tightly(
            div_copy, propagated_param, values_to_explain, propagated_domain,
            info_before_propagation, explanation);
    }
    
  private:
    typedef approximated_arithmetic_propagator<div_propagator> base_type;
    virtual void propagate_unisigned_bounds(
        iset domains[3], const constraint_runtime_info & info) const;
    static long long evaluate(iset::value_type left, iset::value_type right)
    {
        if (right == 0)
            return 0;
        else
            // cast to long long will hopefully deal with MIN_INT/-1
            return (long long)left/right;
    }
    virtual const char *short_name() const { return "/";   }
    virtual const char *long_name() const  { return "div"; }
    //! Propagate to domains[2], i.e. for a=b/c propagate to c.
    void propagate_unisigned_bounds_divisor(iset domains[3]) const;
    //! Propagate to domains[0], i.e. for a=b/c propagate to a.
    void propagate_unisigned_bounds_division(iset domains[3]) const;
    //! Propagate to domains[1], i.e. for a=b/c propagate to b.
    void propagate_unisigned_bounds_dividend(iset domains[3]) const;

}; // class div_propagator

//! Propagate the semantics of: v1 == v2 % v3
class mod_propagator: public approximated_arithmetic_propagator<mod_propagator>
{
  public:
    friend class approximated_arithmetic_propagator<mod_propagator>;
    virtual int count_params() const { return 3; }
    mod_propagator *clone() const { return new mod_propagator(*this); }
    virtual void explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation) const
    {
        // Assume that propagate() was not overridden
        mod_propagator mod_copy(*this);
        explanation_generators::explain_propagation_tightly(
            mod_copy, propagated_param, values_to_explain, propagated_domain,
            info_before_propagation, explanation);
    }
    
  private:
    typedef approximated_arithmetic_propagator<mod_propagator> base_type;
    virtual void propagate_unisigned_bounds(
        iset domains[3], const constraint_runtime_info & info) const;
    void propagate_unisigned_bounds_base(
        iset domains[3], const constraint_runtime_info & info) const;
    void propagate_when_zero_out(iset domains[3]) const;
    bool propagate_min_int_divided(const iset domains[3], iset next_domains[3],
                                   const constraint_runtime_info & info) const;
    void propagate_min_int(iset domains[3], bool has_min_int[3],
                           const constraint_runtime_info & info) const;
    static iset::value_type evaluate(iset::value_type left,
                                     iset::value_type right)
    {
        if (right == 0)
            return 0;
        else
            // cast to long long will hopefully deal with MIN_INT%-1
            return (long long)left%right;
    }
    virtual const char *short_name() const { return "%";   }
    virtual const char *long_name() const  { return "mod"; }
    void propagate_positive_bounds(iset domains[3]) const;
    void accumulate_positive_bounds( iset & supported0,
                                     iset & supported1,
                                     iset domain0,
                                     iset::value_range_type range1,
                                     iset::value_type mod) const;
}; // class mod_propagator


//! Propagate the semantics of: v1 == pow(v2, v3) (i.e. v1 == v2^{v3})
class pow_propagator: public approximated_arithmetic_propagator<pow_propagator>
{
  public:
    friend class approximated_arithmetic_propagator<pow_propagator>;
    virtual int count_params() const { return 3; }
    pow_propagator *clone() const { return new pow_propagator(*this); }
    
  private:
    typedef arithmetic_propagation_types::ops<TRUNCATE> ops;
    virtual const char *short_name() const { return "**";   }
    virtual const char *long_name() const  { return "pow"; }
    static long long evaluate(iset::value_type left, iset::value_type right)
    {
        if (left >= -1 && left <= 1)
            return left;
        else if (right < 0)
            return 0;
        else if (left == 2)
            return 1 << right;
        else if (left == -2)
            return (1 << right)*(right%2? 1 : -1);
        else 
            return nontrivial_pow(left,right);
    }
    
    static long long nontrivial_pow(iset::value_type a, iset::value_type b)
    {
        long long value;
        if (a > iset::MAX_DOMAIN || a < iset::MIN_DOMAIN)
            return a;
        if (b <= 4) {
            value= 1;
            for (iset::value_type i=0 ; i != b ; ++i) {
                value = value*a;
                if (value > iset::MAX_DOMAIN || value < iset::MIN_DOMAIN)
                    break;
            }
            return value;
        }
        iset::value_type half_b= b/2;
        value= nontrivial_pow(a, half_b);
        if (value > iset::MAX_DOMAIN || value < iset::MIN_DOMAIN)
            return value;
        value= value*value;
        if (value > iset::MAX_DOMAIN || value < iset::MIN_DOMAIN)
            return value;
        if (half_b *2 > b)
            value= value*a;
        return value;
    }
    virtual void propagate_unisigned_bounds(
        iset domains[3], const constraint_runtime_info & info) const;
}; // class pow_propagator


} // namespace hcsp

#endif // HCSP_ARITHMETIC_PROPAGATORS_HPP_
