// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef ASSIGNMENT_STACK_HPP_
#define ASSIGNMENT_STACK_HPP_

#include <iosfwd>
#include <stdexcept>
#include <vector>
#include <map>
#include <string>
#include <iset.hpp>
#include <solver_types.hpp>
#include <variable_print_strategy.hpp>

namespace hcsp
{

/**
 * @brief Manage decisions, assignments to CSP variables and backtracking.
 *
 * This class maintains a time-stamp for every domain reduction event,
 * and assignment. It also memorizes the cause for reduction (index of
 * reducing constraint if any).
 */
class assignment_stack : public solver_types
{
  private:                     // Private types
    struct variable_modification_entry
    {
        change_index_t m_previous_modification_entry;
        iset           m_previous_domain;
        time_type      m_time;
        unsigned       m_variable;
        int            m_modifying_constraint;
    };
    
    struct variable_state
    {
        iset           m_domain;
        change_index_t m_latest_modification_entry;
        double         m_cost;
        time_type      m_last_singleton_time;
        iset::size_type m_prev_size;
    };
    struct single_decision_data
    {
        single_decision_data(unsigned variable,
                             change_index_t modification_entry)
            : m_variable(variable), m_modification_entry(modification_entry) {}
        single_decision_data() {} // Should not be called, needed for resize
        // Strictly speaking, m_variable is redundant and can be derived
        // from m_var_states+m_modifications, but that would be quite
        // inefficient.
        unsigned       m_variable;
        change_index_t m_modification_entry;
    };

  public:                       // public types and methods
    struct assignment
    {
        change_index_t m_modification_entry;
        unsigned       m_variable;
        iset           m_literal;
    };
    
    //! Information on a node of the implication graph
    struct implication_node
    {
        change_index_t m_modification_entry;
        unsigned       m_variable;
    }; // struct front_node

    //! A specific variable printing strategy that takes names from the stack
    class assignment_variables_print_strategy : public variable_print_strategy
    {
      public:
        assignment_variables_print_strategy(const assignment_stack & stack)
            : m_stack(stack)
        {}
        virtual void print(std::ostream & out, unsigned var_index) const
        { m_stack.print_var_name(out, var_index); }
      private:
        const assignment_stack & m_stack;
    };
    
    static const change_index_t NO_PREVIOUS_MODIFICATION= ~(change_index_t)0;
    static const change_index_t POST_LAST_MODIFICATION= ~(change_index_t)0-1;
    assignment_stack();
    ~assignment_stack();

    //! \brief Forget all history, return to initial state.
    void clear();

    void print_var_name(std::ostream & out, unsigned var_index) const;
    void var_decision(unsigned variable, time_type time, iset domain)
    {
        if (m_level0_domains.size() <= variable && level() == 0)
            construct_level0_domains();
        m_decisions.push_back(
            single_decision_data(variable, m_modifications.size()));
        set_var_domain(variable, std::move(domain), time);
    }

    //! Erase the latest modification entries if they are due to
    //! the conflicting constraint
    void erase_conflicting_modifications(unsigned conflicting_constraint);

    //! Get the latest modification entry currently in the assignment_stack.
    change_index_t latest_modification() const
    {
        change_index_t size= m_modifications.size();
        if (size == 0)
            return NO_PREVIOUS_MODIFICATION;
        else
            return size-1;
    }
        
    //!\brief Get the index of the variable set at the given decision level
    unsigned decision_variable(unsigned requested_level) const
        throw(std::out_of_range)
    {
        assert_level(requested_level, "decision_variable(level)");
        return m_decisions[requested_level-1].m_variable;
    }
    //!\brief Get the index of the variable set at the last decision level
    unsigned decision_variable() const throw(std::out_of_range)
    { return decision_variable(level()); }

    //! Get the modification entry of the requested decision.
    change_index_t decision_modification(unsigned requested_level) const
        throw(std::out_of_range)
    {
        if (requested_level == 0)
            return NO_PREVIOUS_MODIFICATION;
        assert_level(requested_level, "decision_modification(level)");
        return m_decisions[requested_level-1].m_modification_entry;
    }

    //!\brief Get the value of the variable set at the requested decision level
    iset::value_type decision_value(unsigned requested_level) const
        throw(std::out_of_range)
    {
        assert_level(requested_level, "decision_value()");
        unsigned variable = decision_variable(requested_level);
        const iset * p_domain = &m_var_states[ variable ].m_domain;
        if (p_domain->empty()) {
            change_index_t modification= decision_modification(requested_level);
            p_domain = &var_domain_after_modification(variable, modification);
        }
        HCSP_ASSERT(p_domain->singleton());
        return p_domain->max();
    }
    //!\brief Get the value of the variable set at the last decision level.
    iset::value_type decision_value() const  throw(std::out_of_range)
    { return decision_value(level()); }

    /**
     * @brief Return the current decision level.
     *
     * If decision level is 0 (zero) then there is no decision on the
     * assignment stack.
     */
    unsigned level() const { return m_decisions.size(); }
    void undo_level(unsigned level_to_forget);

    change_index_t var_modification_entry(unsigned variable_index) const
    { return m_var_states[variable_index].m_latest_modification_entry; }

    change_index_t var_modification_entry_before(
        unsigned variable_index, change_index_t constraint_modification) const;

    //! Get the modification time of this variable. The initial state of the
    //! variable (without any modification) has the time 1.
    time_type var_modification_time(unsigned var_index) const;
    
    static change_index_t latest_modification(change_index_t a,
                                              change_index_t b)
    {
        if ( a == NO_PREVIOUS_MODIFICATION )
            return b;
        else if ( b == NO_PREVIOUS_MODIFICATION )
            return a;
        else
            return std::max(a,b);
    }
    static bool is_earlier_modification(change_index_t first,
                                        change_index_t second)
    {
        if (second == NO_PREVIOUS_MODIFICATION)
            return false;
        else if (first == NO_PREVIOUS_MODIFICATION)
            return true;
        else
            return first < second;
    }
    
    int modifying_constraint(change_index_t modification_entry) const
    { return m_modifications.at(modification_entry).m_modifying_constraint; }
    /**
     * @brief Find the time before which the constraint which creates this
     *        modification entry is propagated.
     *
     * Go back in time before the activation that caused 'propagated_at'
     * modification.  If the constraint modified only one constraint then return
     * propagated_at-1 and otherwise return propagated_at-number_of_changes.
     *
     * @param propagated_at  The modification of the constraint before which we
     *                       want to return.
     * @param return The first modification time before the propagation that
     *               caused 'propagated_at' modification.
     */
    change_index_t before_propagation(change_index_t propagated_at) const;

    const variable_print_strategy & get_variable_print_strategy() const
    { return m_var_print_strategy; }
    variable_printer var_printer(unsigned index) const
    { return variable_printer(get_variable_print_strategy(), index); }
    const iset & var_domain(unsigned var_index) const
    { return m_var_states.at(var_index).m_domain; }
    iset::size_type var_prev_size(unsigned var_index) const
    { return m_var_states.at(var_index).m_prev_size; }
    
    /**
     * @brief Get the domain of the variable at the end of \p level
     *
     * The first access to this domain at the next level will give this domain
     *
     * @param variable_index  The variable to look for.
     * @param level_to_find   Find the domain of the variable at the end of
     *                        this level. If level_to_find=-1 then look at
     *                        the initial domain.
     */
    const iset & var_domain_end_level(unsigned variable_index,
                                      int level_to_find) const;
    /**
     * Get the domain of the given variable after the requested modification.
     */
    const iset & var_domain_after_modification(unsigned variable_index,
                                               change_index_t modification)
        const;
    void set_var_domain(unsigned var_index, iset new_domain, 
                        time_type current_time, int updating_constraint=-1);
    void set_var_initial_domain(unsigned var_index, iset new_domain);

    /**
     * @brief Tell whether the lower bound of the variable changed
     *
     * The lower bound of the variable is the minimal value of its domain.
     * Tell if changed on the given modification or on a later one.
     *
     * @param var_index     The variable to check.
     * @param modification  Variable's modification entry to analyze
     * @return True if and only if the lower bound chance on modification or
     *         after it.
     */
    bool lower_bound_changed(unsigned var_index, change_index_t modification)
        const
    {
        return (var_domain(var_index).min() !=
                value_before_modification(modification).min());
    }
    
    /**
     * @brief Tell whether the upper bound of the variable changed
     *
     * The upper bound of the variable is the maximal value of its domain.
     * Tell if changed on the given modification or on a later one.
     *
     * @param var_index     The variable to check.
     * @param modification  Variable's modification entry to analyze
     * @return True if and only if the upper bound chance on modification or
     *         after it.
     */
    bool upper_bound_changed(unsigned var_index, change_index_t modification)
        const
    {
        return (var_domain(var_index).max() !=
                value_before_modification(modification).max());
    }

    std::pair<bool, iset::value_type>
    previously_assigned_var_value(unsigned var_index) const
    { return m_previously_assigned_value.at(var_index); }
    void forget_previously_assigned_var_values()
    {
        for (unsigned i=0 ; i != m_previously_assigned_value.size() ; ++i) {
            const iset & domain =var_domain(i);
            if ( domain.singleton() ) {
                m_previously_assigned_value[i].first= true;
                m_previously_assigned_value[i].second= domain.min();
            } else {
                m_previously_assigned_value[i].first= false;
            }
        }
    }
    
    bool has_free_vars() const { return find_best_free_variable() >= 0; }
    unsigned count_variables() const { return m_var_states.size(); }
    /**
     * @brief Add a new CSP variable.
     *
     * @param domain            The domain of the new variable
     * @param name              The name of the variable, or NULL if none
     * @param update_singletons If the domain is a singleton, and this is there
     *                          is no singleton with this value should it be
     *                          be recorded as a singleton?
     */
    void add_var(iset domain, const char *name= NULL,
                 bool update_singletons= false);
    
    //! Get a variable with this singleton domain, possibly creating one.
    unsigned singleton_var(iset::value_type value);
    /**
     * @brief Return the "asserting decision level" of a learned clause
     * described by implication nodes and literals.
     *
     * @return the lowest level on which the solver would propagate
     *         the following mutated clause.
     *         To create the mutated clause, remove all elements later
     *         than the n-th latest assignment. \n
     *         For example:
     *          -# If n==clause.size()-1 then the no mutation takes place,
     *             and we expect this clause to be a unit clause.
     *          -# If n==clause.size()-2 then we remove only the latest
     *             assignment, and the mutated clause is a unit clause.
     */
    unsigned asserting_level_of_nth(
        unsigned n, const std::vector<implication_node> & nodes,
        const std::vector<iset> & literals) const;
    /**
     * @brief Return the n-th modification time of all literals in the
     * clause.
     *
     * For example n=0 will give the earliest literal, and clause.size()-1
     * would return the latest literal.
     */
    change_index_t nth_modification(
        unsigned n, const std::vector<implication_node> &nodes) const;
    /**
     * @brief Get the value of a variable right before this modification entry.
     *
     * Each modification entry describes one change done to the domain of a
     * single variable. This method returns the domain prior to the given
     * modification entry.
     *
     * @param modification  This method uses this modification entry as a
     *                      reference point. It will return the domain of this
     *                      variable right before the modification took place.
     * @return The domain of the variable right before the modification
     *         took place.
     */
    const iset & value_before_modification(change_index_t modification) const
        throw(std::out_of_range)
    {
        return m_modifications.at(modification).m_previous_domain;
    }

    //! Return a modification entry, such that it describes a move from
    //! an unresolved literal to an unsatisfied literal.
    //! If the literal is satisfied now (and earlier), then return the latest
    //! entry.
    //!   -# unresolved:  We did not decide if it is true or false, yet.
    //!   -# unsatisfied: The literal is false - does not overlap the variable.
    change_index_t relax(change_index_t late_modification,
                         const iset & literal) const;
    /**
     * @brief Get a modification entry which is before the boundary
     *        modification, as long as it still falsifies the literal.
     *
     * @param reference_modification
     *        The modification entry of the variable, which we want to relax
     * @param literal
     *        The set which gets intersected with the domain of the variable
     *        at reference_modification.
     * @param goal_modification
     *        The modification entry before which to try to relax the literal.
     *
     * @return a modification entry which falsifies the literal and, if
     *         possible, is right before the goal_modification. If we can't
     *         find a correct entry before goal_modification, then return
     *         the reference_modification.
     */
    change_index_t relax_before(change_index_t reference_modification,
                                const iset & literal,
                                change_index_t goal_modification)const;
    
    /**
     * @brief Get a set of constraints that assignment_stack's modification
     *        list references.
     *
     * Get all constraints with IDs \p smallest_to_check or higher which
     * are pointed by active modification entries.
     */
    std::vector<unsigned> get_busy_constraints(unsigned smallest_to_check)
        const;
    unsigned decision_level_of(change_index_t tested_modification) const;
    /**
     * @brief Print the list of singleton variables, if one exists.
     *
     * This method does not print a newline, and it prints "Singletons: "
     * only if there is at least one singleton.
     *
     * @param out   The stream to print the list into.
     * @return true  Was anything printed?
     */
    bool print_singleton_variables(std::ostream & out) const;
    
    /**
     * @brief Get the labels of the variable
     *
     * A label on a variable can be viewed as a label on the unary constraint
     * that the domain induces on the variable.  The problem modeler may label
     * each variable differently. He can combine several labels by applying a
     * bit-wise \p OR.
     *
     * @param variable  The index of the variable to query
     * @return A combination of labels of this variable
     */
    label_type variable_labels(unsigned variable) const
    { return m_labels.at(variable); }
    
    /**
     * @brief Set the labels of the variable, overwriting the old value.
     *
     * A label on a variable can be viewed as a label on the unary constraint
     * that the domain induces on the variable.  The problem modeler may
     * label, through this method, each variable differently. He can combine
     * several labels by applying a bit-wise \p OR.
     *
     * @param variable  The index of the variable to set
     * @param label     The updated set of labels
     * @return The original labels of this constraint
     */
    label_type variable_labels(unsigned variable, label_type label)
    {
        label_type old= m_labels.at(variable);
        m_labels[variable]= label;
        return old;
    }
    
    friend class solver;

    /**@brief Start with zero cost for all variables.
     *
     * @param analysis_cost_factor  Initially, this is the base cost added
     *                              to variables.
     */
    void reset_var_costs(double analysis_cost_factor);
    void disable_var_costs_heap();
    void enable_var_costs_heap();

    double cost_additive() const { return m_cost_additive; }
    /** @brief Lower the weight of the existing m_var_costs values.
     *
     * Cost lowering may be effective and not real. This may be achieved by
     * increasing m_cost_additive rather than lowering all m_var_costs values.
     * The effect will be the same since the next cost updates will be bigger
     * relative to the previous costs, and this is equivalent with lowering
     * variable costs without touching m_cost_additive.
     */
    void lower_var_costs();

    void add_var_cost(unsigned var, double cost);

    /**@brief Find the best variable for the next decision.
     * If no free variable exists, return -1.
     */
    int find_best_free_variable() const
    {
        if ( m_var_costs_heap.empty() || 
             var_domain(m_var_costs_heap[0]).singleton())
            return -1;
        else
            return m_var_costs_heap[0];
    }

    /* @brief Does the first variable have better characteristics according to
     * the current variable ordering heuristics?
     *
     * @param first_variable   Is this variable better than the other one
     * @param second_variable  Is this variable worse than the first one?
     * @param factor           By how many times \p first_variable has to be
     *                         better than \p second_variable in order to
     *                         return true?
     * @return True if the first variable is better than the second variable
     *         by the given factor.
     */
    bool is_better_variable(int first_variable, int second_variable,
                            int factor=1) const;

    /** @brief Set by how much the baseline cost additive of m_var_costs should
     *         be boosted after each backtrack, effectively lowering all
     *         m_var_costs by 1/(1+1.0/(factor-1))= (factor-1)/factor.
     */
    void set_cost_lowering_factor(unsigned factor)
    { m_cost_baseline_multiplier= 1+1.0/(factor-1); }

    /** @brief validate the consistency of the object
     *
     * This method makes sure that invariants are satisfied in this object.
     * If an invariant is violated then runtime_error is throws.
     *
     * @throws std::runtime_error if the object is inconsistent
     * @return Always true
     */
    bool validate() const;
    
  private:
    void construct_level0_domains();

    void lower_var_heap_cost(unsigned variable);
    void raise_var_heap_cost(unsigned variable);
    void swap_heap_entries(unsigned entry1, unsigned entry2)
    {                 
        unsigned var1= m_var_costs_heap[entry1];
        unsigned var2= m_var_costs_heap[entry2];
        m_var_costs_heap[entry2] = var1;
        m_var_costs_heap[entry1] = var2;
        m_var_in_heap[var1]= entry2;
        m_var_in_heap[var2]= entry1;
    }
    
    //! Make sure that the invariants of this variable are upheld.
    void validate_var_state(unsigned var_index) const;
    
    change_index_t last_modification_in(const std::vector<assignment> &clause)
        const;
    /**
     * @brief Traverse the modifications from the starting point, but not before
     *        target_entry.
     *
     *  If target_entry >= starting_point then return starting_point. Perform at
     *  most max_skips steps from starting_point towards target_entry. If the
     *  algorithm reaches target_entry, it returns it if it went to an earlier
     *  point before target_entry then return the entry of the previous step.
     *
     * @param starting_point The modification we start our search with.
     * @param target_entry   The modification entry we don't want to go past it.
     * @param max_skips      The maximal number of steps to go before returning.
     * @return Either starting_point, or an entry of at most max_skips from
     *         starting_point which is not earlier than target_entry.
     */
    change_index_t go_back_to(change_index_t starting_point,
                              change_index_t target_entry,
                              unsigned max_skips) const;
    
    /** @brief Complete all the little missing operations of the last update
     */
    void complete_var_update(unsigned var);
    void assert_level(unsigned tested_level, const char *location_str=0) 
        const throw(std::out_of_range)
    {
        if ( tested_level == 0 || tested_level > level() )
            throw_out_of_range(location_str);
    }
    void throw_out_of_range(const char * location_str) const 
        throw(std::out_of_range);
    
    std::vector<variable_state>              m_var_states;

    //! If nonempty then it defines and is in-sync with all domains at level 0
    std::vector<iset>                m_level0_domains;
    
    //! A min heap containing variable indices minimizing 
    //! var_domain(variable).size() / m_var_states[variable].m_cost
    std::vector<unsigned>                    m_var_costs_heap;
    std::vector<unsigned>                    m_var_in_heap;
    //! \brief The base cost added to m_var_states[variable].m_cost
    double                                   m_cost_additive;
    
    //! \brief Effectively lower the costs by multiplying the m_cost_additive by
    //!        m_cost_baseline_multiplier
    double                                   m_cost_baseline_multiplier;

    std::vector<std::pair<bool,iset::value_type> > m_previously_assigned_value;
    std::vector<single_decision_data>        m_decisions;
    std::vector<variable_modification_entry> m_modifications;

    //! A set of labels for each variable
    std::vector<label_type>                  m_labels;
    assignment_variables_print_strategy      m_var_print_strategy;
    std::vector<std::string>                 m_names;
    //! A map from a value to a variable which initial domain is of this value
    std::map<iset::value_type, unsigned>     m_singleton_variable_ids;
}; // assignment_stack

} // namespace hcsp

#endif // ASSIGNMENT_STACK_HPP_
