// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_BIG_ISET_HPP_
#define HCSP_BIG_ISET_HPP_

#include <string>
#include <iterator>
#include <climits>
#include <stdexcept>
#include <iosfwd>
#include <map>
#include <utilities.hpp>
#include <pool_alloc.hpp>
#include <big_iset_impl.hpp>

namespace hcsp {
/**
 * @brief A class efficiently representing a big set of integers,
 *        using a set of value ranges.
 *
 *
 * Exceptions may be thrown during element allocation in many of the
 * methods. Unless otherwise stated, when exception is thrown, the
 * original state of the object is retained. In all cases, invariants
 * are kept intact even if the state of the object is modified during
 * exception throwing.
 *
 * @invariant On top of invariants guaranteed by std::map we have:
 *   - m_size should be synchronized with m_ranges. (i.e. summing up
 *     sizes of all ranges in m_ranges gives m_size (modulo domain size of
 *     m_size)
 *   - Each map element has consistent key and value such that
 *     value.first <= key <= value.second
 *   - Two neighboring ranges in m_ranges can't be merged. Consider two
 *     consecutive elements in the range A and B (A.key < B.key):
 *       - They do not overlap: A.value.second < B.value.first
 *       - There is at least one element missing in their gap:
 *         B.value.first-A.value.second > 1
 */
class big_iset {
  public:
    class iterator;
    friend class iterator;
    typedef iterator                               const_iterator;
    typedef std::reverse_iterator<const_iterator>  const_reverse_iterator;
    typedef std::reverse_iterator<iterator>        reverse_iterator;
    typedef big_iset_impl::size_type               size_type;
    typedef big_iset_impl::value_type              value_type;
    typedef big_iset_impl::value_range_type        value_range_type;
    static const value_type MAX_DOMAIN= big_iset_impl::MAX_DOMAIN;
    static const value_type MIN_DOMAIN= big_iset_impl::MIN_DOMAIN;

    //
    // Methods:
    //

    //! Create an big_iset with an empty domain
    big_iset() : m_skip_size(1), m_offset(0), m_compactions_skipped(0) { }
    big_iset(const value_range_type & range) : m_skip_size(1), m_offset(0),
                                               m_compactions_skipped(0)
    {
        insert(range);
    }
    big_iset(const big_iset & other) { *this= other;}
    big_iset(big_iset && other)= default;
    big_iset(const std::string & text);
    big_iset & operator=(const big_iset & other);
    big_iset & operator=(big_iset && other) = default;
    size_type size() const { return m_impl.size(); }
    //! Return the number of disconnected ranges
    size_type range_count() const { return m_impl.range_count(); }
    iterator begin() const;
    iterator end() const;
    reverse_iterator rbegin() const;
    reverse_iterator rend() const;
    iterator upper_bound(value_type element) const;
    iterator lower_bound(value_type element) const;

    /**
     * @brief Find the tail of elements contained in other.
     *
     * Find the first iterator after which all elements, until 'last',
     * are included in 'other'. The 'last' iterator is the point before which
     * the tail ends.
     *
     * @param other The set in which we look for the contained tail.
     * @param last  An iterator marking the end of the tail. The last element
     *              of the tail is right before 'last'.
     * @return An iterator of the first element in the tail which is contained
     *         in 'other'.
     */
    iterator find_contained_tail(const big_iset & other, iterator last) const;
    
    value_type min() const;
    value_type max() const;
    value_type rand() const;
    big_iset & clear()
    {
        m_impl.clear();
        m_skip_size=1;
        m_offset= 0;
        m_compactions_skipped= 0;
        return *this;
    }

    big_iset & insert(const value_range_type & range);
    
    /**
     * @brief Insert a stride potentially generating holes.
     *
     * @param range     Minimum and maximum values to insert.
     * @param skip_size Distance between consecutive pairs of values
     * @return *this
     * @pre range.first <= range.second  and range.second-range.first
     *      is dividable by skip_size
     * @post this contains range.first, range.first+skip_size, ... range.second
     */
    big_iset & insert(const value_range_type & range, unsigned stride);
    big_iset & insert(value_type element)
    {
        insert(value_range_type(element, element));
        return *this;
    }
    big_iset & assign(const value_range_type & range)
    {
        clear();
        insert(range);
        return *this;
    }
    big_iset & assign(value_type element)
    {
        clear();
        insert(element);
        return *this;
    }
    
    big_iset & erase(const value_range_type & range);
    //! erase one element from the big_iset.
    big_iset & erase(value_type element)
    {
        erase(value_range_type(element, element));
        return *this;
    }
    iterator erase(iterator it);
        
    /**
     * @brief Erase all elements of the parameter from \p this.
     *
     * Unlike other methods, it is possible that due to an allocation exception
     * that the intersect() operation will be partial. This means that
     * some elements will be removed, and some will not.
     *
     * Still, the object retains all invariants regardless of potential
     * exceptions.
     */
    big_iset & erase_all_of(const big_iset & erased);
    /**
     * @brief Insert all elements of the parameter from \p this.
     *
     * Unlike other methods, it is possible that due to an allocation exception
     * that the intersect() operation will be partial. This means that
     * some elements will be removed, and some will not.
     *
     * Still, the object retains all invariants regardless of potential
     * exceptions.
     */
    big_iset & insert_all_of(const big_iset & inserted);
    
    /**
     * @brief Intersect the two sets, keeping only shared elements.
     *
     * Unlike other methods, it is possible that due to an allocation exception
     * that the intersect() operation will be partial. This means that
     * some elements will be removed, and some will not.
     *
     * Still, the object retains all invariants regardless of potential
     * exceptions.
     */
    big_iset & intersect(const big_iset & other);
    /**
     * @brief Intersect the set with the given range, keeping only shared
     *        elements.
     *
     * Unlike intersect(const big_iset&), the set will be unmodified when an
     * exception is throw.
     * @param range  The range of values to keep. If range.first>range.second
     *               then the range is considered empty.
     */
    big_iset & intersect(const value_range_type & range);

    /**
     * @brief Intersect the set with the given range, and the given skip size,
     *        keeping only shared elements.
     *
     * Like intersect(const value_range_type &), the set will be unmodified
     * when an exception is throw.
     * @param range  The range of values to keep. If range.first>range.second
     *               then the range is considered empty.
     * @param skip   Consider only elements which can be calculated as
     *               range.first+k*skip (for an integer k)
     * @pre (range.second-range.first)%skip == 0.
     */
    big_iset & intersect(const value_range_type & range, size_type skip);

    /**
     * @brief Intersect the set with the given range, and the given skip size,
     *        keeping only shared elements.
     *
     * Like intersect(const value_range_type &, size_type), with one difference.
     * It accepts values which are much bigger or smaller than otherwise
     * legal with iset. This method ignores values which are too big or to
     * small in a given range and keeps the legal values which are in the
     * intersection.
     * @param range  The range of values to keep. If range.first>range.second
     *               then the range is considered empty.
     * @param skip   Consider only elements which can be calculated as
     *               range.first+k*skip (for an integer k)
     * @pre (range.second-range.first)%skip == 0.
     */
    big_iset & intersect(const std::pair<long long, long long> & range,
                         size_type skip);
    
    big_iset & complement();
        
    bool empty() const { return m_impl.empty(); }
    bool everything() const
    { return ( m_skip_size == 1 && m_impl.everything()); }
    
    bool singleton() const { return m_impl.singleton(); }
    bool contains(value_type element) const;
    //! \brief Checks set inclusion.
    //! True if for every element e in other, this->contains(e). \n
    //! Note that this is vacuously if other.empty();
    bool contains_all_of(const big_iset & other) const;
    //! \brief Checks set overlapping.
    //! True exists an element e in other such that this->contains(e). \n
    //! Note that if other.empty(), then the result is false.
    bool contains_any_of(const big_iset & other) const;
    //! \brief Return \p true if and only if set contents is the same.
    friend bool operator==(const big_iset &a, const big_iset &b)
    { return (a.size() == b.size() && same_elements(a,b)); }

    //! \brief Return \p true if and only if set contents is different.
    friend bool operator!=(const big_iset &a, const big_iset &b)
    { return !(a==b); }

    //! \brief Swap the contents of the two objects at O(1).
    inline void swap(big_iset & other);
    //! Make sure that all invariants are met
    void validate() const;
    
    friend std::ostream & operator<<(std::ostream & out,
                                     const big_iset & domain);
    friend std::istream & operator>>(std::istream & in,
                                     big_iset & domain);
    //
    //
    //
  protected:
    static size_type range_size(value_type from, value_type to)
    {
        // Be very careful to avoid integer overflows
        if (to >= 0)
            return static_cast<size_type>(to)- from + 1;
        else 
            return static_cast<size_type>(to -from) + 1;
    }
    static size_type range_size(const value_range_type & range)
    { return range_size(range.first, range.second); }
    
    //
    //
    //
  private:
    //! How many elements a range may have to be considered a small range.
    //! Simpler algorithms are used with small ranges. These algorithms
    //! have a bad time complexity, but work well with small ranges.
    enum { MAX_ELEMENTS_IN_SMALL_RANGE= 9 };
    //! Print the current state of the iset to cout.
    //! This function is intended to be called from a debugger only.
    void show_self();
    
    void flatten();
    void compact();
    void compact_attempt(unsigned prev_size)
    {
        if (prev_size == size())
            return;
        if(m_compactions_skipped > 5)
            compact();
        else
            ++m_compactions_skipped;
    }

    //! Intersect \p this and \p other assuming they have different skip sizes
    void intersect_different_strides(const big_iset & other);

    //! Intersect \p this and \p range assuming they have different skip sizes
    void intersect_different_strides(const value_range_type & range,
                                     size_type skip);

    /**@ brief Find how many steps (size step1) from range1.first are needed
     *         to find an element shared by range1 and range2.
     * 
     * Find how many steps of size step1, from range1.first, are needed to
     * reach to an element which is also reachable from range2.first
     * using >=0 steps of size step2. If no such value exists return -1.
     *
     * Note that the return value may require more steps than what is available
     * in either range1 or range2. This possibility has to be checked by the
     * caller.
     *
     * @param[in] range1,step1   The first interval and skip size
     * @param[in] range2,step2   The second interval and skip size
     * @param[out] step_gcd      The value of gcd(step1,step2) which is the
     *                           side-effect of this function.
     * @return If no solution possible then -1, otherwise the number of steps
     *         of size step1 (from range1.first) needed to reach an element
     *         which is of the right distance from range1.first and
     *         range2.first.
     */
    long long num_of_step1_needed(const value_range_type & range1,
                                  size_type step1, 
                                  const value_range_type & range2,
                                  size_type step2,
                                  unsigned & step_gcd);

    /** @brief Try simple algorithms to insert the intersection of range1,range2
     *
     * If it is possible to insert the intersection using trivial algorithms,
     * then do so and return true. Otherwise return false.
     *
     * \<b>Side effect</b>: range1 and range2 will be trimmed, removing
     *                      some of the irrelevant values.
     *
     * @param[in,out] range1,range2  The ranges (will be trimmed)
     * @param[in]     step1,step2    The size of the skips inside the ranges
     * @return True if and only if the method inserted everything to be inserted
     *         (it is possible that nothing has been inserted, if no elements
     *          are common to both ranges).
     */
    bool insert_if_simple_intersection(value_range_type & range1,
                                       size_type step1, 
                                       value_range_type & range2,
                                       size_type step2);
    
    /** @brief Insert the intersection of the two ranges with, potentially,
     *         different skip sizes, but which start at the same point.
     *
     * This is a special case of insert_intersection() which assumes that
     * range1.first == range2.first. For more information read the documentation
     * of insert_intersection().
     *
     * @param range1,step1  The specification of the first interval
     * @param range2,step2  The specification of the second interval
     * @param step_gcd      gcd(step1,step2)
     *
     * @pre If step_gcd is provided then step_gcd==gcd(step1, step2)
     */
    void insert_intersection_same_start(const value_range_type & range1,
                                        size_type step1, 
                                        const value_range_type & range2,
                                        size_type step2,
                                        size_type step_gcd=-1U);
    
    /** @brief Insert the intersection of the two ranges with, potentially,
     *         different skip sizes.
     *
     * set1 := {range1.first:step1:range1.second}\n
     * set2 := {range2.first:step2:range2.second}\n
     * this->insert(set1 intersect set 2)\n
     *
     * Note that the ranges are passed by value (copied), to allow the
     * function to modify the values of the parameters without affecting
     * the originals.
     *
     * @param range1,step1  The specification of the first interval
     * @param range2,step2  The specification of the second interval
     *
     * @post Every element which is both in set1 and set2 (as defined above)
     *       is also in *this. An element is in *this only if it was there
     *       before or if it is contained both in set1 and set2. No element
     *       has been removed from *this.
     */
    void insert_intersection(value_range_type range1, size_type step1, 
                             value_range_type range2, size_type step2);

    //! update range bounds (with steps) so that it is not below \p element.
    static void remove_smaller_elements(value_range_type &range, size_type step,
                                        value_type element);
    //! update range bounds (with steps) so that it is not above \p element.
    static void remove_bigger_elements(value_range_type &range, size_type step,
                                       value_type element);
    
    static bool is_in_range(const value_range_type & range,
                            size_type step, value_type element);
                
    //! count the number of elements in the interval (with steps).
    static size_type count_elements(const value_range_type &range,
                                    size_type step);
    
    
    bool contains_all_of(const value_range_type & range,
                         unsigned skip_size) const;
    static bool same_elements(const big_iset & a, const big_iset & b);
    static bool same_stride(const big_iset & a, const big_iset & b)
    { return  a.m_skip_size == b.m_skip_size && a.m_offset== b.m_offset;}
    /**
     * @brief Tell if the strides of 'a' and 'b' are aligned in a way that
     * guarantees no overlap.
     *
     * For example: 0:3:99 and 1:3:100, or 0:6:96 and 1:4:97. It does not check
     * the range so that other causes of !a.contains_any_of(b) are not
     * detected. For example: 0:2:50  and 52:2:100 do not overlap but this
     * function will not detect this and will return false.
     *
     * @param a,b  The two sets to check
     * @return True if the stride step and offset guarantee
     *         !a.contains_any_of(b) for any possible contents of m_impl.
     */
    static bool interleaved_stride(const big_iset & a, const big_iset & b)
    {
        if (a.m_skip_size <= 1 || b.m_skip_size <= 1)
            return false;
        unsigned skip_gcd= gcd(a.m_skip_size, b.m_skip_size);
        if (skip_gcd <= 1)
            return false;
        else
            return a.m_offset % skip_gcd != b.m_offset % skip_gcd;
    }
    static value_type floor_div(value_type value, size_type divider)
    {
        if (value >= 0)
            return value/divider;
        size_type absolute= -value;
        if ( absolute % divider == 0)
            return -value_type(absolute / divider);
        else
            return -value_type(absolute / divider)-1;
    }
    static value_type floor_mod(value_type value, size_type divider)
    {
        if (value >= 0)
            return value%divider;
        size_type absolute= -(size_type)value;
        if ( absolute % divider == 0)
            return 0;
        else
            return divider-(absolute % divider);
    }
    bool in_stride(value_type value) const
    {
        return (size_type) floor_mod(value, m_skip_size) == m_offset;
    }
    value_type external_value(value_type value) const
    { return value*m_skip_size + m_offset; }
    value_range_type external_range(value_range_type range) const
    { return value_range_type( external_value(range.first),
                               external_value(range.second) );}
    value_type upper_impl_value(value_type value) const
    {
        if (value > 0)
            return floor_div(value-m_offset-1, m_skip_size)+1;
        else
            return floor_div(value+m_skip_size-m_offset-1, m_skip_size);
    }
    value_type lower_impl_value(value_type value) const
    {
        if (value >= 0)
            return floor_div(value-m_offset, m_skip_size);
        else
            return floor_div(value+m_skip_size-m_offset, m_skip_size)-1;
    }
    value_type impl_value(value_type value) const
    { return lower_impl_value(value);}
    
    big_iset_impl m_impl;
    size_type     m_skip_size;
    size_type     m_offset;
    //! Count the number of compactions skipped since the last true compaction.
    unsigned char m_compactions_skipped;
}; // class big_iset

class big_iset::iterator {
  public:
    typedef std::bidirectional_iterator_tag iterator_category;
    typedef big_iset::value_type          value_type;
    typedef std::ptrdiff_t             difference_type;
    typedef const value_type*          pointer;
    typedef value_type                 reference;
    struct Impl_range_accessor : value_range_type {
        Impl_range_accessor(const value_range_type & other)
            : value_range_type(other) {}
        const value_range_type * operator->() const {return this; }
    };
    static const bool CHECKS_BOUNDS= false;
    static const bool INVALIDATED_BY_ERASE= true;
  private:
    iterator(big_iset_impl::iterator iter_impl, const big_iset *p_set)
        : m_iter_impl(iter_impl), m_set(p_set) {}
    
  public:
    friend class big_iset;
    iterator() : m_set(NULL) {}

    value_type operator*() const
    { return m_set->external_value(*m_iter_impl);}
    value_range_type current_range() const
    {
        if (m_set->m_skip_size == 1) {
            HCSP_ASSERT(m_set->m_offset == 0);
            return m_iter_impl.current_range();
        } else {
            value_type val= **this;
            return value_range_type(val, val);
        }
    }
    const iterator & advance_range()
    {
        if (m_set->m_skip_size == 1)
            m_iter_impl.advance_range();
        else
            ++*this;
        return *this;
    }

    /**
     * @brief Go to the previous range.
     *
     * If the current position of the iterator is in the middle of a range,
     * then the iterator will go to the beginning of the range. However,
     * if the iterator is at the beginning of a range then the iterator
     * will go to the beginning of the previous range.
     *
     * This means that operator--(), operator++() and retract_range() do not
     * mix well.
     */
    const iterator & retract_range()
    {
        if (m_set->m_skip_size == 1)
            m_iter_impl.retract_range();
        else
            --*this;
        return *this;
    }
    
    /**
     * @brief Find a stride of elements.
     *
     * A stride is a sequence with a constant step size between consecutive
     * elements. Find a stride of at least \p min_count elements.
     *
     * @param[in]  end        The search stops before end is dereferenced.
     * @param[out] range      The first and last element of the stride.
     * @param[out] step_size  The step size of the detected stride
     * @return One iterator after range.second
     */
    iterator find_stride(iterator end,
                         value_range_type &range,
                         size_type & step_size) const;

    /**
     * @brief Find a stride of elements, backwards.
     *
     * A stride is a sequence with a constant step size between consecutive
     * elements. Find a stride of at least \p min_count elements.
     * The first element of the stride is not before begin, and the
     * last element of the stride starts one element before 'this'.
     *
     * @param[in]  begin      The search stops at begin or after it.
     * @param[out] range      The first and last element of the stride.
     * @param[out] step_size  The step size of the detected stride
     * @return Iterator pointing to range.first
     */
    iterator find_back_stride(iterator begin,
                              value_range_type &range,
                              size_type & step_size) const;
    
    const iterator & advance(size_type count);
    const iterator & operator++();
    iterator operator++(int);
    const iterator & operator--();
    iterator operator--(int);
    Impl_range_accessor operator->() const { return current_range(); }
    inline bool operator==(const iterator & other) const;
    inline bool operator!=(const iterator & other) const;
  private:
    big_iset_impl::iterator m_iter_impl;
    const big_iset         *m_set;
};

inline big_iset & big_iset::operator=(const big_iset & other)
{
    m_skip_size= other.m_skip_size;
    m_offset= other.m_offset;
    m_impl= other.m_impl;
    m_compactions_skipped= 0;
    compact();
    return *this;
}

inline big_iset::iterator big_iset::begin() const
{
    return iterator(m_impl.begin(), this);
}


inline big_iset::iterator big_iset::end() const
{
    return iterator(m_impl.end(), this);
}

inline big_iset::reverse_iterator big_iset::rbegin() const
{
    return reverse_iterator(end());
}

inline big_iset::reverse_iterator big_iset::rend() const
{
    return reverse_iterator(begin());
}

inline big_iset::iterator big_iset::lower_bound(value_type element) const
{
    return iterator(m_impl.lower_bound(upper_impl_value(element)), this);
}


inline big_iset::iterator big_iset::upper_bound(value_type element) const
{
    return iterator(m_impl.upper_bound(lower_impl_value(element)), this);
}

inline big_iset::value_type big_iset::min() const
{
    return external_value(m_impl.min());
}

inline big_iset::value_type big_iset::max() const
{
    return external_value(m_impl.max());
}

inline void big_iset::swap(big_iset & other)
{
    m_impl.swap(other.m_impl);
    std::swap(m_skip_size, other.m_skip_size);
    std::swap(m_offset, other.m_offset);
}

inline const big_iset::iterator & big_iset::iterator::advance(size_type n)
{
    m_iter_impl.advance(n);
    return *this;
}

inline const big_iset::iterator & big_iset::iterator::operator++()
{
    ++m_iter_impl;
    return *this;
}

inline big_iset::iterator big_iset::iterator::operator++(int)
{
    iterator ret=*this;
    ++(*this);
    return ret;
}

inline const big_iset::iterator & big_iset::iterator::operator--()
{
    --m_iter_impl;
    return *this;
}

inline bool big_iset::iterator::operator==(const iterator & other) const
{
    return (m_set == other.m_set && m_iter_impl == other.m_iter_impl);
}

inline bool big_iset::iterator::operator!=(const iterator & other) const
{
    return !(*this == other);
}

} // namespace hcsp

#endif // HCSP_BIG_ISET_HPP_
