// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_BIG_ISET_IMPL_HPP_
#define HCSP_BIG_ISET_IMPL_HPP_

#include <limits>
#include <string>
#include <iterator>
#include <climits>
#include <stdexcept>
#include <iosfwd>
#include <map>
#include <utilities.hpp>
#include <pool_alloc.hpp>

namespace hcsp {
/**
 * @brief A class efficiently representing a big set of integers,
 *        using a set of value ranges.
 *
 *
 * Exceptions may be thrown during element allocation in many of the
 * methods. Unless otherwise stated, when exception is thrown, the
 * original state of the object is retained. In all cases, invariants
 * are kept intact even if the state of the object is modified during
 * exception throwing.
 *
 * @invariant On top of invariants guaranteed by std::map we have:
 *   - m_size should be synchronized with m_ranges. (i.e. summing up
 *     sizes of all ranges in m_ranges gives m_size (modulo domain size of
 *     m_size)
 *   - Each map element has consistent key and value such that
 *     value.first <= key <= value.second
 *   - Two neighboring ranges in m_ranges can't be merged. Consider two
 *     consecutive elements in the range A and B (A.key < B.key):
 *       - They do not overlap: A.value.second < B.value.first
 *       - There is at least one element missing in their gap:
 *         B.value.first-A.value.second > 1
 */
class big_iset_impl {
  private:
    static const unsigned SIGNIFICANT_BITS= std::numeric_limits<int>::digits;
    //! A value with all bits set, except the most significant value bit.
    static const int HALF_MAX_DOMAIN=(1 << (SIGNIFICANT_BITS-1))-1;
    
  public:
    class iterator;
    friend class iterator;
    typedef iterator                               const_iterator;
    typedef std::reverse_iterator<const_iterator>  const_reverse_iterator;
    typedef std::reverse_iterator<iterator>        reverse_iterator;
    typedef unsigned                               size_type;
    typedef int                                    value_type;
    typedef std::pair<value_type, value_type>      value_range_type;
    typedef value_type                             const_reference;
    static const value_type MAX_DOMAIN= (HALF_MAX_DOMAIN<<1) + (value_type)1;
    static const value_type MIN_DOMAIN= -MAX_DOMAIN          - (value_type)1;

    //
    // Methods:
    //

    //! Create an big_iset_impl with an empty domain
    big_iset_impl() : m_size(0) { }
    big_iset_impl(const value_range_type & range) : m_size(0)
    {
        insert(range);
    }
    big_iset_impl(big_iset_impl && other) noexcept
        : m_ranges(std::move(other.m_ranges)), m_size(other.m_size)
    {
        other.m_size = 0;
    }
    big_iset_impl &operator=(const big_iset_impl & other) = default;
    big_iset_impl &operator=(big_iset_impl && other) noexcept
    {
        m_ranges = std::move(other.m_ranges);
        m_size = other.m_size;
        other.m_size = 0;
        return *this;
    }
    size_type size() const { return m_size; }
    //! Return the number of disconnected ranges
    size_type range_count() const { return m_ranges.size(); }
    iterator begin() const;
    iterator end() const;
    reverse_iterator rbegin() const;
    reverse_iterator rend() const;
    iterator upper_bound(value_type element) const;
    iterator lower_bound(value_type element) const;
    value_type min() const;
    value_type max() const;
    value_type rand() const;
    big_iset_impl & clear()
    {
        m_ranges.clear();
        m_size= 0;
        return *this;
    }

    big_iset_impl & insert(const value_range_type & range);
    big_iset_impl & insert(value_type element)
    {
        insert(value_range_type(element, element));
        return *this;
    }
    big_iset_impl & assign(const value_range_type & range)
    {
        clear();
        insert(range);
        return *this;
    }
    big_iset_impl & assign(value_type element)
    {
        clear();
        insert(element);
        return *this;
    }
    
    big_iset_impl & erase(const value_range_type & range);
    //! erase one element from the big_iset_impl.
    void erase(value_type element)
    {
        erase(value_range_type(element, element));
    }
    iterator erase(iterator it);
        
    /**
     * @brief Erase all elements of the parameter from \p this.
     *
     * Unlike other methods, it is possible that due to an allocation exception
     * that the intersect() operation will be partial. This means that
     * some elements will be removed, and some will not.
     *
     * Still, the object retains all invariants regardless of potential
     * exceptions.
     */
    big_iset_impl & erase_all_of(const big_iset_impl & erased);
    /**
     * @brief Insert all elements of the parameter from \p this.
     *
     * Unlike other methods, it is possible that due to an allocation exception
     * that the intersect() operation will be partial. This means that
     * some elements will be removed, and some will not.
     *
     * Still, the object retains all invariants regardless of potential
     * exceptions.
     */
    big_iset_impl & insert_all_of(const big_iset_impl & inserted);
    
    /**
     * @brief Intersect the two sets, keeping only shared elements.
     *
     * Unlike other methods, it is possible that due to an allocation exception
     * that the intersect() operation will be partial. This means that
     * some elements will be removed, and some will not.
     *
     * Still, the object retains all invariants regardless of potential
     * exceptions.
     */
    big_iset_impl & intersect(const big_iset_impl & other);
    /**
     * @brief Intersect the set with the given range, keeping only shared
     *        elements.
     *
     * Unlike intersect(const big_iset_impl&), the set will be unmodified when an
     * exception is throw.
     */
    big_iset_impl & intersect(const value_range_type & range);
    big_iset_impl & complement();
        
    bool empty() const { return m_ranges.empty(); }
    bool everything() const
    {
        return ( (m_size == (size_type) MAX_DOMAIN - MIN_DOMAIN + 1)
                 &&
                 !m_ranges.empty());
    }
    
    bool singleton() const { return m_size == 1; }
    bool contains(value_type element) const;
    //! \brief Checks set inclusion.
    //! True if for every element e in other, this->contains(e). \n
    //! Note that this is vacuously if other.empty();
    bool contains_all_of(const big_iset_impl & other) const;
    //! \brief Checks set overlapping.
    //! True exists an element e in other such that this->contains(e). \n
    //! Note that if other.empty(), then the result is false.
    bool contains_any_of(const big_iset_impl & other) const;
    //! \brief Return \p true if and only if set contents is the same.
    friend bool operator==(const big_iset_impl &a, const big_iset_impl &b)
    { return (a.m_size == b.m_size && same_ranges(a.m_ranges, b.m_ranges)); }

    //! \brief Return \p true if and only if set contents is different.
    friend bool operator!=(const big_iset_impl &a, const big_iset_impl &b)
    { return !(a==b); }

    //! \brief Swap the contents of the two objects at O(1).
    inline void swap(big_iset_impl & other);
    //! Make sure that all invariants are met
    void validate() const;
    
    //
    //
    //
  protected:
    static size_type range_diff(value_type from, value_type to)
    { return static_cast<size_type>(to)- from; }
    static size_type range_diff(const value_range_type & range)
    { return range_diff(range.first, range.second); }

    static size_type range_size(value_type from, value_type to)
    { return range_diff(from,to) + 1; }
    static size_type range_size(const value_range_type & range)
    { return range_size(range.first, range.second); }
    
    //
    //
    //
  private:
    typedef std::map<value_type,
                     value_range_type,
                     std::less<value_type>,
                     pool_alloc<std::pair<const value_type, value_range_type> >
    > ranges_set_type;
    
    static bool same_ranges(const ranges_set_type &a,
                            const ranges_set_type &b);
    
    void assert_valid_range(const value_range_type & range,
                            const char *text) const
    {
        if ( range.first > range.second)
            throw_out_of_range(text);
    }

    void throw_out_of_range(const char *text) const throw(std::out_of_range);
    ranges_set_type::iterator lower_range_bound(value_type value);
    ranges_set_type::iterator upper_range_bound(value_type value)
    {
        ranges_set_type::iterator result= lower_range_bound(value);
        if (result != m_ranges.end() && result->second.first <= value)
            ++result;
        return result;
    }

    big_iset_impl & insert_independent(const value_range_type & inserted_range,
                               ranges_set_type::iterator where);
    static inline bool ranges_touch(const value_range_type &a,
                                    const value_range_type &b);
    bool range_touches_iterator(const value_range_type &range,
                                const ranges_set_type::iterator & it) const;

    static bool ranges_overlap(const value_range_type &a,
                               const value_range_type &b)
    { return a.first <= b.second && b.first <= a.second; }
    bool range_overlaps_iterator(const value_range_type &range,
                                 ranges_set_type::const_iterator it) const;

    bool range_subsumes_iterator(const value_range_type &range,
                                 const ranges_set_type::const_iterator&it)const
    {
        return (it == m_ranges.end() ||
                (range.first       <= it->second.first &&
                 it->second.second <= range.second));
    }
        
    bool range_subsumed(const value_range_type &small,
                        const value_range_type &big)
        const
    {
        return (big.first <= small.first && small.second <= big.second);
    }

    bool range_subsumed_by_iterator(const value_range_type &range,
                                    const ranges_set_type::const_iterator & it)
        const
    {
        return (it != m_ranges.end() && range_subsumed(range, it->second));
    }

    /**
     * @brief Merge the range with the elements pointed by the iterator.
     *
     * Neighbors that touch the range are also merged in.
     *
     * @pre The inserted range should either overlap the pointed element,
     *      or touch one of its edges.
     *
     * @return *this
     */
    big_iset_impl & merge_at_iterator(const value_range_type & inserted_range,
                              ranges_set_type::iterator where);

    /**
     * @brief If possible, change the key of the iterator around the given
     *        values.
     *
     * If \p min_value and/or max_value are contained in the range of \p where,
     * then make sure that the key of the \p where iterator is in the range
     * [\p min_value, \p max_value].
     *
     * This can be done by possibly reinserting the element of m_ranges with
     * a different key.
     *
     * @param[in] where      The iterator which we want to meet this constraint
     * @param[in] min_value  The lower constraint of the key
     * @param[in] max_value  The upper constraint of the key
     *
     * @return Either \p where, or an iterator pointing to a reinserted item.
     */
    ranges_set_type::iterator
    recenter_iterator_around(ranges_set_type::iterator where,
                             value_type min_value, value_type max_value);
    
    /**
     * @brief If possible, change the key of the iterator to above this value.
     *
     * If \p value is contained in the range of \p where,
     *  then make sure that the key of the \p where iterator is above \p value.
     *
     * This can be done by possibly reinserting the element of m_ranges with
     * a different key.
     *
     * @param[in] where   The iterator which we want to meet this constraint.
     * @param[in] value   The value that defines the constraint over the key.
     *
     * @return Either \p where, or an iterator pointing to a reinserted item.
     */
    ranges_set_type::iterator
    recenter_iterator_above(ranges_set_type::iterator where, value_type value)
    {
        
        if ( where == m_ranges.end() || value == MAX_DOMAIN )
            return where;
        else
            return recenter_iterator_around(where,
                                            value+1, where->second.second);
    }

    /**
     * @brief If possible, change the key of the iterator to below this value.
     *
     * If \p value is contained in the range of \p where,
     * then make sure that the key of the \p where iterator is below \p value.
     *
     * This can be done by possibly reinserting the element of m_ranges with
     * a different key.
     *
     * @param[in] where   The iterator which we want to meet this constraint.
     * @param[in] value   The value that defines the constraint over the key.
     *
     * @return Either \p where, or an iterator pointing to a reinserted item.
     */
    ranges_set_type::iterator
    recenter_iterator_below(ranges_set_type::iterator where, value_type value)
    {
        
        if ( where == m_ranges.end() || value == MIN_DOMAIN)
            return where;
        else
            return recenter_iterator_around(where,
                                            where->second.first, value-1);
    }
    
    /**
     * @brief Erase the range from the elements pointed by the iterator.
     *
     * Note that other elements of m_ranges are kept intact even if they
     * overlap erased_range.
     *
     * @pre
     *    -# The \p erased_range should overlap the range pointed by \p where.
     *    -# The invariants are met.
     * @post
     *    -# The invariants are retained.
     *    -# Only parts of \p where which do not overlap \p erased_range,
     *       if any, are kept in \p m_ranges.
     * @return *this
     */
    big_iset_impl & erase_from_single_iterator(ranges_set_type::iterator where,
                                       const value_range_type erased_range);

    inline static size_type distance_between(value_type first,
                                             value_type second)
    {
        if (second < 0 && first < 0)
            return second-first;
        else
            return (size_type)second-first;
    }
    /**
     * @brief Find the value in the middle of the given range.
     *
     * This method takes care to avoid overflows in a signed type.
     * In C++ overflows in signed type are undefined and forbidden.
     */
    inline static value_type middle_of(const value_range_type & range)
    {
        size_type distance= distance_between(range.first, range.second);
        return range.first + (value_type)(distance/2);
    }
    
    //
    //
    //
    
    ranges_set_type m_ranges;   //!< The actual ranges of values big_iset_impl holds
    
    /**
     * @brief This attribute holds the number of value_type elements that
     *        big_iset_impl holds.
     *
     *  For systems where sizeof(size_type) == size_type(value_type),
     *  in case the big_iset_impl represents everything() then m_size == 0.
     */
    size_type          m_size;
}; // class big_iset_impl

class big_iset_impl::iterator {
  public:
    typedef std::bidirectional_iterator_tag iterator_category;
    typedef big_iset_impl::value_type          value_type;
    typedef std::ptrdiff_t             difference_type;
    typedef const value_type*          pointer;
    typedef value_type                 reference;
    struct Impl_range_accessor : value_range_type {
        Impl_range_accessor(const value_range_type & other)
            : value_range_type(other) {}
        const value_range_type * operator->() const {return this; }
    };
    static const bool CHECKS_BOUNDS= false;
    static const bool INVALIDATED_BY_ERASE= true;
  private:
    iterator(ranges_set_type::iterator range_iterator, value_type value)
        : m_range_iterator(range_iterator), m_current_value(value) {}
  public:
    friend class big_iset_impl;
    iterator() {}

    value_type operator*() const
    { return std::max(m_current_value, m_range_iterator->second.first); }
    value_range_type current_range() const
    { return value_range_type(operator*(), m_range_iterator->second.second); }
    const iterator & advance_range()
    {
        ++m_range_iterator;
        m_current_value= MIN_DOMAIN;
        return *this;
    }

    /**
     * @brief Go to the previous range.
     *
     * If the current position of the iterator is in the middle of a range,
     * then the iterator will go to the beginning of the range. However,
     * if the iterator is at the beginning of a range then the iterator
     * will go to the beginning of the previous range.
     *
     * This means that operator--(), operator++() and retract_range() do not
     * mix well.
     */
    const iterator & retract_range()
    {
        if (!middle_of_range())
            --m_range_iterator;
        m_current_value= MIN_DOMAIN;
        return *this;
    }
    
    /**
     * @brief Find a stride of elements.
     *
     * A stride is a sequence with a constant step size between consecutive
     * elements. Find a stride of at least \p min_count elements.
     *
     * @param[in]  end        The search stops before end is dereferenced.
     * @param[out] range      The first and last element of the stride.
     * @param[out] step_size  The step size of the detected stride
     * @return One iterator after range.second
     */
    iterator find_stride(iterator end,
                         value_range_type &range,
                         size_type & step_size) const;
    
    /**
     * @brief Find a stride of elements, backwards.
     *
     * A stride is a sequence with a constant step size between consecutive
     * elements. Find a stride of at least \p min_count elements.
     * The first element of the stride is not before begin, and the
     * last element of the stride starts one element before 'this'.
     *
     * @param[in]  begin      The search stops at begin or after it.
     * @param[out] range      The first and last element of the stride.
     * @param[out] step_size  The step size of the detected stride
     * @return Iterator pointing to range.first
     */
    iterator find_back_stride(iterator begin,
                              value_range_type &range,
                              size_type & step_size) const;
    
    const iterator & advance(size_type count);
    const iterator & operator++();
    iterator operator++(int);
    const iterator & operator--();
    iterator operator--(int);
    Impl_range_accessor operator->() const { return current_range(); }
    inline bool operator==(const iterator & other) const;
    inline bool operator!=(const iterator & other) const;
  private:
    bool middle_of_range() const
    {
        if (m_current_value == MIN_DOMAIN)
            return false;

        // Can't be the end() iterator because the end() iterator has
        // m_current_value == MIN_DOMAIN
        return  m_current_value > m_range_iterator->second.first;
    }
    ranges_set_type::iterator m_range_iterator;
    /**
     * If inside the range of m_range_iterator, then it is the current
     * value of \p this iterator. If before the range, then the current
     * value is the first lowest value of m_range_iterator.
     */
    value_type               m_current_value;
};

inline big_iset_impl::iterator big_iset_impl::begin() const
{
    ranges_set_type & mutable_ranges= const_cast<ranges_set_type&>(m_ranges);
    ranges_set_type::iterator it= mutable_ranges.begin();
    value_type first= m_ranges.empty() ? MIN_DOMAIN : it->second.first;
    return iterator(it, first);
}


inline big_iset_impl::iterator big_iset_impl::end() const
{
    ranges_set_type & mutable_ranges= const_cast<ranges_set_type&>(m_ranges);
    return iterator(mutable_ranges.end(), MIN_DOMAIN);
}

inline big_iset_impl::reverse_iterator big_iset_impl::rbegin() const
{
    return reverse_iterator(end());
}

inline big_iset_impl::reverse_iterator big_iset_impl::rend() const
{
    return reverse_iterator(begin());
}

inline big_iset_impl::iterator big_iset_impl::lower_bound(value_type element) const
{
    big_iset_impl & mutable_this= const_cast<big_iset_impl&>(*this);
    ranges_set_type::iterator range_it=mutable_this.lower_range_bound(element);
    if (range_it == m_ranges.end() || range_it->second.first > element)
        return iterator(range_it, MIN_DOMAIN);
    else
        return iterator(range_it, element);
}


inline big_iset_impl::iterator big_iset_impl::upper_bound(value_type element) const
{
    iterator result= lower_bound(element);
    if (result != end() && *result == element)
        ++result;
    return result;
}

inline big_iset_impl::value_type big_iset_impl::min() const
{
    return m_ranges.begin()->second.first;
}

inline big_iset_impl::value_type big_iset_impl::max() const
{
    ranges_set_type::const_iterator back_iterator= m_ranges.end();
    --back_iterator;
    return back_iterator->second.second;
}

inline void big_iset_impl::swap(big_iset_impl & other)
{
    m_ranges.swap(other.m_ranges);
    size_type orig_size= m_size;
    m_size= other.m_size;
    other.m_size= orig_size;
}

inline const big_iset_impl::iterator & big_iset_impl::iterator::operator++()
{
    if ( m_current_value < m_range_iterator->second.first )
        m_current_value= m_range_iterator->second.first;

    const value_range_type & range= m_range_iterator->second;
    if ( range.second == m_current_value || range.second == range.first )
        advance_range();
    else
        ++m_current_value;
    return *this;
}

inline big_iset_impl::iterator big_iset_impl::iterator::operator++(int)
{
    iterator ret=*this;
    ++(*this);
    return ret;
}

inline const big_iset_impl::iterator & big_iset_impl::iterator::operator--()
{
    if ( m_current_value == MIN_DOMAIN ||
         m_current_value == m_range_iterator->second.first ) {
        --m_range_iterator;
        m_current_value= m_range_iterator->second.second;
    }
    else
        --m_current_value;
    return *this;
}

inline bool big_iset_impl::iterator::operator==(const iterator & other) const
{
    if (m_range_iterator != other.m_range_iterator)
        return false;
    if (m_current_value == other.m_current_value)
        return true;
    return **this == *other;
}

inline bool big_iset_impl::iterator::operator!=(const iterator & other) const
{
    return !(*this == other);
}

} // namespace hcsp

#endif // HCSP_BIG_ISET_IMPL_HPP_
