// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_CLAUSE_DATA_HPP_
#define HCSP_CLAUSE_DATA_HPP_

#include <vector>
#include <iosfwd>
#include <pool_alloc.hpp>
#include <iset.hpp>
#include <variable_print_strategy.hpp>

namespace hcsp {

/** 
 * @brief Information on one clause literal
 */ 
struct literal_data
{ 
    literal_data(unsigned variable, const iset & set)
        : m_variable(variable), m_set(set) {}
    literal_data() {}
    //! swap this literal with another literal.
    void swap(literal_data & other)
    {
        std::swap(m_variable, other.m_variable);
        m_set.swap(other.m_set);
    }
    friend bool operator<(const literal_data & left,
                          const literal_data & right)
    { return left.m_variable < right.m_variable; }
    
    friend bool operator<(const literal_data & left, unsigned variable_index)
    { return left.m_variable < variable_index; }

    friend bool operator==(const literal_data &a, const literal_data &b)
    { return a.m_variable == b.m_variable && a.m_set == b.m_set; }
    
    /**
     * @brief Print the literal in a human readable form
     *
     * Use >=, <=, =, != whenever possible, otherwise use : to indicate
     * element inclusion in a set.
     */
    static void print(std::ostream & out, 
                      const variable_printer & variable_printer,
                      const iset & set);

    /**
     * @brief Print the right hand side of the literal, all except the variable.
     *
     * Use >=, <=, =, != whenever possible, otherwise use : to indicate
     * element inclusion in a set.
     */
    static void print_right_hand_side(std::ostream & out, const iset & set);

    unsigned m_variable;
    iset     m_set;
}; // struct literal_data
 
//! A clause as a set of literals
class clause_data : public std::vector<literal_data, pool_alloc<literal_data> >
{
  private:
    typedef std::vector<literal_data, pool_alloc<literal_data> > parent_t;
  public:
    clause_data() {}
    clause_data(unsigned initial_size) : parent_t(initial_size) {}

    /**
     * @brief Are the two clauses equivalent?
     *
     * Compare semantically, differences in representation do not affect the
     * result of this function. It returns true if the canonical
     * representations of the clauses are equal.
     *
     * @param[in] a  The object to compare.
     * @return true if the clauses are semantically equivalent.
     */
    bool equivalent(const clause_data & other) const;
    
    /**
     * @brief Transform the clause to a canonical representation.
     *
     * For any two logically equivalent clauses, this method will transform
     * them to have exactly the same representation. A canonical representation
     * has no empty literals, each literal has a unique variable and
     * the literals are sorted in an ascending order of variable indexes.
     */
    void canonicalize();

    void print(std::ostream &out, const variable_print_strategy &strategy)const;

    //! Get the string which the solver prints at the beginning of clauses
    static const char* print_prefix()    { return "["; }
    //! Get the string which the solver prints to separate literals of a clause
    static const char* print_separator() { return ", "; }
    //! Get the string which the solver prints at the end of clauses
    static const char* print_suffix()    { return "]"; }
    //! Sort the literals in an ascending order of variable indexes.
    void sort();
}; // class clause_data
}  // namespace hcsp

#endif // HCSP_CLAUSE_DATA_HPP_
