// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_CLAUSE_PROPAGATOR_HPP_
#define HCSP_CLAUSE_PROPAGATOR_HPP_

#include <initializer_list>

#include <iset.hpp>
#include <constraint_propagator.hpp>
#include <propagation_algo.hpp>

namespace hcsp {
/**
 * @class clause_propagation_algo
 * @brief This class propagates a signed-clause.
 *
 * From its construction, for each literal of the signed-clause it holds the
 * set part of the literal, but not the variable name. Variable information
 * is passed to propagate() by a constraint_runtime_info object.
 *
 * @par Example:
 *      For \f$ (x_1 \in \{1,2\} \vee x_2 \in \{3,4\}) \f$
 *      the clause_propagation_algo will hold \f$\{1,2\}\f$ for literal 0,
 *      and \f$\{3,4\}\f$ for literal 1.
 *      The constraint_runtime_info object will pass state of the variables
 *      (rather than their name).
 *
 * Like in boolean SAT, the clause propagator will maintain two watches.
 * The algorithm that maintains the watches has to consider several corner
 * cases not present in boolean SAT.
 *
 * For more information on watches of signed-clauses read
 * @ref signed-clause-watches.
 */

/**
 * @page signed-clause-watches Using and maintaining watches on signed-clauses.
 *
 * @section signed-clause-watch-usage
 *
 * The solver considers a clause for propagation only when the propagator has
 * a chance to reduce the domain of a variable. The mechanism of watches
 * allows the solver to avoid propagation in many cases when the propagation
 * will not yield reduction.
 *
 * The clause propagator uses value watches for this efficiency goal. The
 * clause maintains two value watches (var1, val1) and (var2, val2).
 * A constraint is propagated if any of the following conditions are met:
 *  - It was never propagated before.
 *  - One of the watches if triggered.
 * 
 * The watch (var, val) is triggered when val is removed from the domain of
 * var.
 * After every time a clause is propagated it has to update the watches, such
 * that any interesting domain reduction would trigger propagation.
 *
 * This view is way too simplistic for a backtracking solver. For
 * backtracking a more accurate requirement for watches has to be formulated:
 *  - If the solver backtracks to some older state then, even without
 *    recalculating its watches, interesting domain reductions will still
 *    trigger the watches.
 *
 *    If prior to backtracking the clause had two watches (containing two
 *    different supporting values) then after backtracking these two watches
 *    still describe support values of the clause. This means that in this
 *    case the watches will be still waiting for correct events even after
 *    backtracking.
 *
 *  - When a clause becomes VALID after it was propagated as a unit clause,
 *    then we can simply leave the watches untouched. This scheme would work
 *    correctly only if we backtrack to the point at which the clause was
 *    consistent.
 *
 *    <b>CRITICAL NOTE:</b>\n
 *    This above assumption means that this scheme will not work if and when
 *    the solver would backtrack to an intermediate state (which is not
 *    arc consistent). An example for such a state is when the algorithm
 *    explicitly complements bad decision values (generating a modification
 *    entry) and immediately makes a decision (a second modification entry).
 *    In this example, backtracking to the state between the two modifications
 *    would be buggy as any dependency on the complemented value has not
 *    been addressed. A watch may still be waiting for the, now defunct,
 *    removed decision value.
 *
 *  - When a clause becomes VALID due to an external event then the watches
 *    would behave the same way as described by the previous bullet (i.e. unit
 *    clause).
 *  - When a clause is VALID since it was inserted as a conflict clause and
 *    is immediately propagated as a unit clause, then the solver must be
 *    more clever with it. The watches will have to be:
 *     -# A watch that contain one of the support values of the \e true
 *        literal. (i.e. a value of the propagated variable - after
 *        propagation). 
 *     -# Find the earliest modification after which this clause becomes
 *        a unit clause. Add a watch that would be triggered by this
 *        modification.
 *
 *        If the solver backtracks to a position after this, but before
 *        the previous propagation we would still like to propagate this
 *        clause (since the propagation has been undone).
 *        However, since we backtrack to the earliest level possible at
 *        which the inserted clause is still unit (or even earlier than that),
 *        then this situation will not happen.
 *
 *        <b>CRITICAL NOTE:</b>\n
 *        This code assumes that backtracking is performed to the earliest
 *        level at which the newly inserted clause is a unit clause (or
 *        earlier). If backtracking is performed to a later level, then
 *        there will be nothing to trigger this clause's propagation.
 *        No watch insertion can guarantee that, and a different mechanism
 *        will have to be implemented.
 */
class clause_propagation_algo : private propagation_types, watch_types
{
  private:
    enum { NUM_OF_WATCHES=2 };
  public:
    typedef std::vector<iset> literals_type;
    
    clause_propagation_algo() 
    {
        for (unsigned i=0 ; i != NUM_OF_WATCHES ; ++i)
            m_watch[i].invalidate_mark();
    }

    int count_params() const { return m_literals.size(); }
    
    literals_type & update_literals() throw(std::runtime_error)
    {
        if ( m_watch[0].is_valid_watch() || m_watch[1].is_valid_watch() )
            throw_modifying_clause_at_runtime();
        return m_literals;
    }
    const literals_type & literals() const { return m_literals; }

    bool init(constraint_runtime_info & info);
    result_type propagate(constraint_runtime_info & info);
    bool conflicts(const constraint_runtime_info & info) const;
    bool constraint_check(const constraint_runtime_info & info) const;

    void print(std::ostream & out,
               const constraint_runtime_info * p_info) const throw();
    /**
     * @brief Count the number of watches that are managed by this propagator.
     *
     * @return The number of watches which are registered to the outside
     *         world, i.e., m_watch[i].is_valid_watch() == true.
     */
    unsigned count_watches() const
    { return m_watch[0].is_valid_watch() + m_watch[1].is_valid_watch(); }

    unsigned watched_param_index(unsigned watch_index) const
    { return m_watch[watch_index].m_param; }

    void forget_watches()
    {
        for (unsigned i=0 ; i != NUM_OF_WATCHES ; ++i)
            m_watch[i].invalidate_watch();
    }
    void explain_propagation(
        unsigned /*propagated_param*/, const iset & /*values_to_explain*/,
        const iset & /*propagated_domain*/, 
        const constraint_runtime_info & /*info_before_propagation*/,
        std::vector<iset> & explanation) const
    { explanation= m_literals; }

  private:
    //! \brief Decide on which value we should put a watch.
    iset::value_type get_best_watch_value(const constraint_runtime_info & info,
                                          unsigned param_index,
                                          iset::value_type hint) const;
    
    void remove_watch(constraint_runtime_info & info, unsigned watch_index);
    /**
     * @brief  Add a watch on the given parameter.
     *
     * @pre At most one watch is already allocated (m_param >= 0).
     *
     * @param[in,out] info         The proxy that allocates the watch.
     * @param[in]     param_index  The parameter to watch.
     * @param[i]      hint         Prefer to watch values on or above the hint.
     *
     */
    void add_watch(constraint_runtime_info & info, unsigned param_index,
                   iset::value_type hint=iset::MIN_DOMAIN);
    void throw_modifying_clause_at_runtime() throw(std::runtime_error);

    /**
     * Refresh the watch stored at m_watch[watch_index], potentially
     * invoking unit propagation. 
     *
     * Try to find a new value and/or variable to watch. The new watch must
     * not interfere with the second watch used by this clause.
     *
     * @pre  The current watch is inconsistent, i.e. the watched value is not
     *       in the current domain of the variable.
     *
     * @param[in,out] info         The proxy that allocates the watch
     * @param[in]     watch_index  The index of the watch in m_watch.
     * @return The conclusions from this operation. If unit-propagation was
     *         triggered, then return the result of the propagation.
     *         If can't find an alternative watch, return UNSAT. Otherwise
     *         return CONSISTENT.
     */
    result_type refresh_watch(constraint_runtime_info &info,
                              unsigned watch_index);

    result_type unit_propagate(constraint_runtime_info & info,
                               unsigned literal_index);
    
    void change_watch(constraint_runtime_info & info,
                      unsigned watch_index,
                      unsigned new_param_index);
    
    literals_type   m_literals;
    /**
     * @brief Watches and marks maintained by this propagator.
     *
     * The solver knows which clause watches a given variable. A clause may
     * watch one, two variables or no variables at all. A watch marks one
     * of the clause as interesting, and uses these marks to speed-up
     * propagation. It is possible to set a mark without letting the solver
     * maintain a watch for it, that's why a value_watch_info_type has
     * is_valid_mark() and a separate is_valid_watch().
     *
     * As a special case, if this is a unit clause at decision level 0,
     * then m_watch array may potentially contain only one valid entry.
     * This entry marks the variable that has to be reduced, but it is not
     * really a watch - in a sense that the solver does not know about it.
     */
    value_watch_info_type m_watch[NUM_OF_WATCHES];
}; // class clause_propagation_algo

class clause_propagator :
    public constraint_propagator_from_algo<clause_propagation_algo>
{
  public:
    typedef clause_propagation_algo algo;
    typedef algo::literals_type literals_type;

    clause_propagator() {}
    clause_propagator(std::initializer_list<iset> in_literals)
    { m_algo.update_literals()= in_literals; }
    clause_propagator *clone() const { return new clause_propagator(*this); }
    literals_type & update_literals() throw(std::runtime_error)
    { return m_algo.update_literals(); }
    
    const literals_type & literals() const { return m_algo.literals(); }
    
    virtual clause_propagator * intersect_with(
        const constraint_runtime_info & info,
        std::vector<unsigned> & copy_from_argument) const;
    virtual bool potentially_falsifying(const constraint_runtime_info & info,
                                        unsigned param) const;
    virtual bool get_disjunction_specification(
        disjunction_specification & disjunction) const;
    virtual bool may_combine(const constraint_propagator & other) const
    {
        if (dynamic_cast<const clause_propagator*>(&other))
            return false;
        else
            return other.may_combine(*this);
    }
    /**
     * @brief Combine the clause with another propagator.
     *
     * The method returns NULL if the other constraint is also a clause, which
     * means that the caller has to fall-back to applying clause-resolution.
     * Otherwise this method calls to the combine() method of the other
     * constraint.
     */
    virtual constraint_propagator * combine(
        const constraint_runtime_info & info,
        const constraint_propagator & other_propagator,
        const constraint_runtime_info & other_info,
        const std::vector<unsigned> & other_to_common_param,
        unsigned other_pivot_param,
        std::vector<unsigned> & copy_from_argument) const;
}; // class clause_propagator

} // namespace hcsp
#endif // HCSP_CLAUSE_PROPAGATOR_HPP_
