// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_COMPARISON_PROPAGATORS_HPP_
#define HCSP_COMPARISON_PROPAGATORS_HPP_


#include <iset.hpp>
#include <constraint_propagator.hpp>
#include <literal_constraint_propagators.hpp>
#include <disjoint_propagators.hpp>
#include <clause_propagator.hpp>

namespace hcsp {
class literal_or_difference_ge_propagator;
class difference_ge_or_clause_propagator;

//! Propagate the semantics of: v1 == v2.
class equal_propagator : public tight1_constraint_propagator
{
  public:
    virtual int count_params() const { return 2; }
    virtual result_type propagate( constraint_runtime_info & info );
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual bool conflicts(const constraint_runtime_info & info) const;
    constraint_propagator * new_propagator(
        disjunction_specification & disjunction,
        std::vector<unsigned> &copy_from_argument) const;
    
    virtual bool may_combine(const constraint_propagator &) const
    { return true; }
    
    virtual constraint_propagator * combine(
        const constraint_runtime_info & info,
        const constraint_propagator & other_propagator,
        const constraint_runtime_info & other_info,
        const std::vector<unsigned> & other_to_common_param,
        unsigned other_pivot_param,
        std::vector<unsigned> & copy_from_argument) const;
    
    virtual equal_propagator* clone() const
    { return new equal_propagator(*this); }
    
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    { return info.domain(0).min() == info.domain(1).min(); }
}; // class equal_propagator

// Assuming the vector of variables v[0]...v[n-1], and the input variables:
//    [i, v[0], ...,v[n-1], d]
// propagate v[i-m_first_index] == d.
class element_propagator : public constraint_propagator
{
  public:
    element_propagator(iset::value_type first_index)
        : m_first_index(first_index)
    {}
    result_type propagate( constraint_runtime_info & info ) override;
    void print(std::ostream & out,
               const constraint_runtime_info *p_info) const throw() override;
    bool conflicts(const constraint_runtime_info & info) const override;
    constraint_propagator * intersect_with(
        const constraint_runtime_info & info,
        std::vector<unsigned> & copy_from_argument) const override;
    
    element_propagator* clone() const override
    { return new element_propagator(*this); }

    void explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation) const override;
    
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        int64_t element = info.domain(0).min();
        unsigned last = info.arity() - 1;
        // Remember that info.domain(0) is the index, and info.domain(last) is
        // the result of the array.
        unsigned num_elements = info.arity()-2;
        int64_t index = element - m_first_index;
        if (index <0 || index >= num_elements)
            return false;
        return info.domain(index+1).min() == info.domain(last).min();
    }
  private:

    void explain_equality_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain, const constraint_runtime_info & info,
        std::vector<iset> & explanation) const;

    void explain_index_propagation(
        const iset & values_to_explain, const iset & propagated_domain, 
        const constraint_runtime_info & info, 
        std::vector<iset> & explanation) const;

    void explain_result_propagation(
        const iset & values_to_explain, const iset & propagated_domain, 
        const constraint_runtime_info & info, 
        std::vector<iset> & explanation) const;

    void explain_conflict(const constraint_runtime_info & info, 
                          std::vector<iset> & explanation) const;

    // The array of variables starts at this index.
    iset::value_type m_first_index;
}; // class element_propagator
/**
 * @brief Propagate the semantics of: v1 - v0 >= k
 * @todo It seems that it is possible to utilize watches to get
 *       a faster constraint propagation.
 */
class difference_ge_propagator : public tight_constraint_propagator
{
  public:
    difference_ge_propagator(int k) : m_k(k) {}
    virtual int count_params() const { return 2; }
    virtual bool init( constraint_runtime_info & info );
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual std::unique_ptr<explanation_rule> rule(
        const std::vector<iset> & literals) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual difference_ge_propagator* clone() const
    { return new difference_ge_propagator(*this); }
    virtual constraint_propagator * intersect_with(
        const constraint_runtime_info & info,
        std::vector<unsigned> & copy_from_argument) const;

    virtual bool disjoin_constraint(
        const disjunction_specification & this_constraint,
        const disjunction_specification & other_constraint,
        disjunction_specification & result_constraint) const;

    virtual bool combine_base_constraints(
        disjunction_specification & this_constraint,
        disjunction_specification & other_constraint,
        unsigned pivot,
        disjunction_specification & result_constraint) const;
    
    virtual constraint_propagator * new_propagator(
        disjunction_specification & disjunction,
        std::vector<unsigned> &copy_from_argument) const;
    
    virtual bool get_disjunction_specification(
        disjunction_specification & disjunction) const;
    virtual bool may_combine(const constraint_propagator &other_propagator)
        const
    { return may_combine_diff_ge_propagator(other_propagator); }
    virtual constraint_propagator * combine(
        const constraint_runtime_info & info,
        const constraint_propagator & other_propagator,
        const constraint_runtime_info & other_info,
        const std::vector<unsigned> & other_to_common_param,
        unsigned other_pivot_param,
        std::vector<unsigned> & copy_from_argument) const
    {
        return combine(*this, info, other_propagator, other_info,
                       other_to_common_param, other_pivot_param,
                       copy_from_argument);
    }
    
    //! The lower bound of difference v0-v1.
    iset::value_type get_min_difference() const { return m_k; }
    
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    { return 0LL+info.domain(0).min() - info.domain(1).min() >= m_k; }
  private:
    /**
     * @brief Disjoin two constraints which can't be merged into one x-y>=k
     *        constraint.
     *
     * Transfer one of the constraints into a clause, if possible. If the method
     * can't create a constraint+literals it will return NULL. This may happen
     * when the two constraints have some common variables.
     */
    bool disjoin_incompatible_constraint(
        const disjunction_specification & this_constraint,
        const disjunction_specification & other_constraint,
        disjunction_specification & result_constraint) const;

    /**
     * @brief Can the given propagator be combined with other propagators
     *        of type x-y >= x (possibly with a disjunction).
     *
     * @param propagator  The propagator to query.
     */
    static bool may_combine_diff_ge_propagator(const constraint_propagator &
                                               propagator);
    bool combine_middle_pivot(
        disjunction_specification &a, disjunction_specification &b,
        iset::value_type b_k,
        disjunction_specification & result_constraint) const;
    bool combine_circular_pivot(
        disjunction_specification &a, disjunction_specification &b,
        iset::value_type b_k,
        disjunction_specification & result_constraint) const;
    
    class diff_ge_rule;
    friend class literal_or_difference_ge_propagator;
    friend class difference_ge_or_clause_propagator;
    static constraint_propagator * combine(
        const constraint_propagator & propagator,
        const constraint_runtime_info & info,
        const constraint_propagator & other_propagator,
        const constraint_runtime_info & other_info,
        const std::vector<unsigned> & other_to_common_param,
        unsigned other_pivot_param,
        std::vector<unsigned> & copy_from_argument);
    
    iset::value_type m_k;
}; // class difference_ge_propagator

//! Propagate the semantics of: if (v2) then v1=v3 ; else v1 = v4;
class if_then_else_propagator : public tight_constraint_propagator
{
  public:
    virtual int count_params() const { return 4; }
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual if_then_else_propagator* clone() const
    { return new if_then_else_propagator(*this); }
    
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        iset::value_type result=     info.domain(0).min();
        iset::value_type cond=       info.domain(1).min();
        iset::value_type then_value= info.domain(2).min();
        iset::value_type else_value= info.domain(3).min();
        return (cond == 0 || cond == 1) &&
            result == (cond ? then_value : else_value);
    }
  private:
    //! Propagate the domains assuming that condition=1
    result_type propagate_then(constraint_runtime_info & info );
    
    //! Propagate the domains assuming that condition=0
    result_type propagate_else(constraint_runtime_info & info );
}; // class if_then_else_propagator

//! Propagate the semantics of: v1 <=> (v2 == v3)
class is_equal_propagator : public tight_constraint_propagator
{
  public:
    virtual int count_params() const { return 3; }
    // FIXME: smart watches may speed it up considerably.
    // virtual bool init( constraint_runtime_info & info );
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual is_equal_propagator* clone() const
    { return new is_equal_propagator(*this); }
    
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    { return info.domain(0).min() == (
        info.domain(1).min() == info.domain(2).min()); }
  private:
    //! Propgate asserting inequality between var1 and var2 assuming var0=0
    result_type propagate_diff( constraint_runtime_info & info );
    //! Propgate asserting equality between var1 and var2 assuming var0=1
    result_type propagate_equal( constraint_runtime_info & info );
}; // class is_equal_propagator

//! Propagate the semantics of: v0 \in L || (v1 == v2)
class literal_or_equal_propagator
    : public literal_or_constraint_propagator<literal_or_equal_propagator,
                                              tight_constraint_propagator>
{
    typedef literal_or_constraint_propagator<literal_or_equal_propagator,
                                             tight_constraint_propagator>parent;
    friend class literal_or_constraint_propagator<literal_or_equal_propagator,
                                                  tight_constraint_propagator>;
  public:
    literal_or_equal_propagator(const iset &literal) : parent(literal) {}
    virtual int count_params() const { return 3; }
    literal_or_equal_propagator* clone() const
    { return new literal_or_equal_propagator(*this); }

  private:
    virtual
    result_type propagate_non_literal_part(constraint_runtime_info & info);

    virtual result_type state_of_non_literal_part(
        const constraint_runtime_info & info) const;

    
    virtual void print_non_literal_part(
        std::ostream & out,
        const constraint_runtime_info *p_info) const throw();
    
    //! For testing
    bool
    constraint_check_non_literal_part(const constraint_runtime_info &info)const
    { return info.domain(1) == info.domain(2); }
}; // class literal_or_equal_propagator

//! Propagate the semantics of: (v1 in L) <=> (v2 < v3)
class literal_iff_less_than_propagator:
    public literal_iff_constraint_propagator<literal_iff_less_than_propagator,
                                             tight_constraint_propagator>
{
    typedef literal_iff_constraint_propagator<literal_iff_less_than_propagator,
                                              tight_constraint_propagator>parent;
    friend class literal_iff_constraint_propagator<
        literal_iff_less_than_propagator, tight_constraint_propagator>;
  public:
    literal_iff_less_than_propagator(const iset & literal) : parent(literal) {}
    virtual literal_iff_less_than_propagator* clone() const
    { return new literal_iff_less_than_propagator(*this); }
    virtual int count_params() const { return 3; }
    
  private:
    virtual bool init_non_literal_part(constraint_runtime_info & info);
    virtual bool init_neg_non_literal_part(constraint_runtime_info & info);
    virtual
    result_type propagate_non_literal_part(constraint_runtime_info & info);
    virtual
    result_type propagate_neg_non_literal_part(constraint_runtime_info &info);

    virtual result_type state_of_non_literal_part(
        const constraint_runtime_info & info) const;
    
    virtual void print_non_literal_part(
        std::ostream & out,
        const constraint_runtime_info *p_info) const throw();
    virtual bool constraint_check_non_literal_part(
        const constraint_runtime_info & info) const
    { return info.domain(1).min() < info.domain(2).min(); }
}; // class literal_iff_less_than_propagator

//! Propagate the semantics of: v0 \in L || (v2-v1 >= k)
class literal_or_difference_ge_propagator
    : public literal_or_constraint_propagator<
    literal_or_difference_ge_propagator, tight_constraint_propagator>
{
    typedef literal_or_difference_ge_propagator self;
    typedef literal_or_constraint_propagator<self,
                                             tight_constraint_propagator>parent;
    friend class literal_or_constraint_propagator<
        self, tight_constraint_propagator>;
  public:
    literal_or_difference_ge_propagator(const iset &literal, iset::value_type k)
        : parent(literal), m_k(k) {}
    virtual self* clone() const { return new self(*this); }
    virtual constraint_propagator * intersect_with(
        const constraint_runtime_info & info,
        std::vector<unsigned> & copy_from_argument) const;
    virtual int count_params() const { return 3; }

    virtual bool get_disjunction_specification(
        disjunction_specification & disjunction) const;
    virtual bool may_combine(const constraint_propagator & other) const
    { return difference_ge_propagator::may_combine_diff_ge_propagator(other); }
    virtual constraint_propagator * combine(
        const constraint_runtime_info & info,
        const constraint_propagator & other_propagator,
        const constraint_runtime_info & other_info,
        const std::vector<unsigned> & other_to_common_param,
        unsigned other_pivot_param,
        std::vector<unsigned> & copy_from_argument) const
    {
        return difference_ge_propagator::combine(
            *this, info, other_propagator, other_info,
            other_to_common_param, other_pivot_param, copy_from_argument);
    }
    //! The lower bound of difference v0-v1.
    iset::value_type get_min_difference() const { return m_k; }
    
  private:
    virtual bool init_non_literal_part( constraint_runtime_info & info);
    virtual
    result_type propagate_non_literal_part(constraint_runtime_info & info);

    virtual result_type state_of_non_literal_part(
        const constraint_runtime_info & info) const;

    
    virtual void print_non_literal_part(
        std::ostream & out,
        const constraint_runtime_info *p_info) const throw();
    
    //! For testing
    bool
    constraint_check_non_literal_part(const constraint_runtime_info &info)const
    { return 0LL+info.domain(1).min() - info.domain(2).min() >= m_k; }
    iset::value_type m_k;
}; // class literal_or_difference_ge_propagator

class difference_ge_or_clause_propagator :
        public disjoint_propagator<difference_ge_propagator,
                                   clause_propagation_algo>
{
  private:
    typedef disjoint_propagator<difference_ge_propagator,
                                clause_propagation_algo> parent;
  public:
    difference_ge_or_clause_propagator(iset::value_type k) : parent(k) {}
    virtual bool get_disjunction_specification(
        disjunction_specification & disjunction) const;
    virtual bool may_combine(const constraint_propagator &other) const
    { return difference_ge_propagator::may_combine_diff_ge_propagator(other); }
    virtual constraint_propagator * combine(
        const constraint_runtime_info & info,
        const constraint_propagator & other_propagator,
        const constraint_runtime_info & other_info,
        const std::vector<unsigned> & other_to_common_param,
        unsigned other_pivot_param,
        std::vector<unsigned> & copy_from_argument) const
    {
        return difference_ge_propagator::combine(
            *this, info, other_propagator, other_info,
            other_to_common_param, other_pivot_param, copy_from_argument);
    }
}; // class difference_ge_or_clause_propagator


/**
 * @brief An internal helper class
 *
 * This internal helper class contains code shared by min_propagator and
 * max_propagator.
 */
class min_max_propagator : public tight_constraint_propagator
{
  protected:
    /**
     * @brief Propagate the semantics of: v0 == max(v1, v2, ...)
     *
     * Propagate the semantics of 'max' modifying \p domains.
     * @param[in,out] The domains to propagate. These domains are reduced
     *                by this mehtod.
     * @return returns false if and only if the domains conflict
     */
    static bool propagate_max(std::vector<iset> & domains);
  private:
    /**
     * @brief For \f$ v_0= \max(v_1, v_2, \ldots) \f$,
     * get the values of \f$ v_1,v_2, \ldots \f$
     * that support v0.
     *
     * @param[in]  info     The domains of the variables.
     * @param[out] supports Vector of supports.
     *                      When supports[0] is empty and supports[1]
     *                      refers to the first non-result argument of 'max'.
     */
    static void get_result_supports(const std::vector<iset> & domains,
                                    std::vector<iset> & supports);
     //! Find the smallest value of domains[i].max().
    static iset::value_type
    find_max_lower_bound(const std::vector<iset> & domains);

    /**
     * @brief Propagate the 'min' propagator to its result domain.
     *
     * For v0=max(v1,v2,...) propagate from domains[1], domains[2], ...
     * to domains[0].
     *
     * @param[in,out] domains  The domains of the variables
     * @return true if and only if conflicting.
     */
    static bool propagate_to_max_result(std::vector<iset> & domains);

    /**
     * @brief  For each element domains[i] erase all values above max_legal.
     * @param[in,out] domains    The domains to operate on.
     * @param[in]     max_legal  Erase all values bigger than max_legal.
     * @return  True if, at the end, none of the domains are empty
     */
    static bool erase_values_above( std::vector<iset> & domains,
                                    iset::value_type max_legal);
    /**
     * @brief Find the domain with the biggest max() value.
     *
     * Find the index of a domain which has the biggest element.
     *
     * @param[in] domains  The domains to work on. Some may be empty.
     * @return If all domains are empty return domains.size(), otherwise
     *         return an index of domains[i] which contains the biggest
     *         value among the values of \p domains.
     */
    static unsigned find_index_of_max(const std::vector<iset> & domains);
    
};

/**
 * @brief Propagate the semantics of: v1 == max(v2, v3)
 *
 * @todo
 *    -# Explanations can improve conflict analysis's quality.
 *    -# It seems that it is possible to utilize watches to get
 *       a faster constraint propagation.
 */
class max_propagator : public min_max_propagator
{
  public:
    virtual int count_params() const { return -1; }
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual max_propagator* clone() const
    { return new max_propagator(*this); }

    
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        iset::value_type result=iset::MIN_DOMAIN ;
        for (unsigned i=1 ; i != info.arity() ; ++i)
            if ( result < info.domain(i).min() )
                result= info.domain(i).min();
        return result == info.domain(0).min();
    }
}; // class max_propagator


/**
 * @brief Propagate the semantics of: v1 == min(v2, v3)
 *
 * @todo
 *    -# Explanations can improve conflict analysis's quality.
 *    -# It seems that it is possible to utilize watches to get
 *       a faster constraint propagation.
 */
class min_propagator : public min_max_propagator
{
  public:
    virtual int count_params() const { return -1; }
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual min_propagator* clone() const
    { return new min_propagator(*this); }
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        iset::value_type result=iset::MAX_DOMAIN ;
        for (unsigned i=1 ; i != info.arity() ; ++i)
            if ( result > info.domain(i).min() )
                result= info.domain(i).min();
        return result == info.domain(0).min();
    }
  private:
    //! Mirror the values. Values originally big become small and vice versa.
    static void mirror(iset & elements);
}; // class min_propagator


} // namespace hcsp

#endif // HCSP_COMPARISON_PROPAGATORS_HPP_
