// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010,2013 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010,2013 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_CONFLICT_CONSTRAINT_DEPENDENCY_HPP_
#define HCSP_CONFLICT_CONSTRAINT_DEPENDENCY_HPP_

#include <vector>
namespace hcsp {

/**
 * @brief Record the resolution dependency between conflict constraints
 *
 * This class records, for each conflict constraint, the set of conflict
 * constraints that were used during one call to solver::analyze_conflict().
 * It also records the usage frequency of conflict constraints, and is able to
 * answer queries on these constraints.
 */
class conflict_constraint_dependency
{
  public:
    conflict_constraint_dependency(unsigned first_conflict_constraint)
        : m_first_conflict_constraint(first_conflict_constraint)
    {}

    /**
     * @brief Mark the constraint as being accessed after a given total
     *        backtracks count is reached.
     *
     * @param constraint  The conflict constraint which we mark as being used.
     * @param current_backtrack_count
     *                    The current counter of backtracks encountered,
     *                    so far.
     *
     * @pre \p constraint is a conflict constraint and is not erased, i.e,
     *      \p constraint >= m_first_conflict_constraint and
     *      m_usage_backtrack_count[constraint] != ERASED
     */
    void mark_used(unsigned constraint, unsigned current_backtrack_count);

    /**
     * @brief Mark the dependency between these two conflict constraints
     *
     * @param prerequisite One of the prerequisite conflict constraints used to
     *                     derive \p consequent
     * @param consequent   The resulting conflict constraint.
     *
     * @pre \p prerequisite and \p consequent are conflict constraint which were
     *       not erased, i.e, \p prerequisite >= m_first_conflict_constraint and
     *      \p consequent >= m_first_conflict_constraint  and
     *       m_usage_backtrack_count[\p prerequisite] != ERASED and
     *       m_usage_backtrack_count[\p consequent] != ERASED
     */
    void mark_dependency(unsigned prerequisite, unsigned consequent);

    /**
     * @brief Compact the list of prerequisites of the said \p consequent.
     *
     * The only effect of this method is to optimize memory use of the list
     * of prerequisites which are stored for one \p consequent.
     *
     * @param consequent  The consequent conflict constraint for which data is
     *                    to be compacted
     * @pre \p \p consequent is a conflict constraint which is not erased, i.e,
     *      \p consequent >= m_first_conflict_constraint and
     *       m_usage_backtrack_count[\p consequent] != ERASED
     */
    void compact_prerequisites(unsigned consequent);

    /**
     * @brief Get all the conflict constraints that have been active on the
     *        requested backtrack or later.
     *
     * Activity of a constraints is updated when mark_used is called for it.
     *
     * @return The set of active conflict constraint.
     * @post For each \p i: \p return[i] is a conflict constraint which is not
     *      erased, i.e., \p return[i] >= m_first_conflict_constraint  and
     *      m_usage_backtrack_count[\p return[i]] != ERASED
     */
    std::vector<unsigned> active_constraints(unsigned backtrack_count) const;


    /**
     * @brief Return all existing constraints which are not in the input
     * container
     *
     * @param constraints  The set of constraint we don't expect in the output.
     * @return The complementary set of existing constraints
     */
    std::vector<unsigned>
    other_constraints(const std::vector<unsigned> &constraints) const;
    
    /**
     * @brief Get all the conflict constraints which are prerequisites to the
     *        input set, even if they are on the input set
     *
     * Start with the set of \p consequents. Perform DFS from each constraint to
     * its prerequisite, if it is not in the returned list
     *
     * @param consequents      The set of consequents to start DFS from.
     * @return The set of prerequisites.
     * @pre For each \p i \p prerequisites[i] is a conflict constraint which is
     *      not erased, i.e., \p prerequisites[i] >= m_first_conflict_constraint
     *      and m_usage_backtrack_count[\p prerequisites[i]] != ERASED
     */
    std::vector<unsigned>
    prerequisite_closure(const std::vector<unsigned> & consequents) const;

    /**
     * @brief Remove the conflict constraint and all related information.
     * @pre conflict_constraint is a conflict constraint which is not erased,
     *      i.e., \p conflict_constraint >= m_first_conflict_constraint and
     *      m_usage_backtrack_count[conflict_constraint] != ERASED
     */
    void erase_constraint(unsigned conflict_constraint);
    
  private:
    /**
     * @brief The set of prerequisite conflict constraints used for deriving one
     *        consequent conflict constraint.
     *
     * Even though a conflict constraint is usually derived from many
     * constraints, not only from conflict constraints, we store only conflict
     * constraints. These constraints can't be removed unless the current
     * constraint is also removed.
     */
    typedef std::vector<unsigned> constraint_prerequisites;

    //! Recursively, add all prerequisites of the constraint in the container.
    //! Use templates to avoid header file dependency on any given container
    //! type.
    template <class Touched_Container>
    void add_prerequisites(std::vector<unsigned> & prerequisites,
                           Touched_Container & touched,
                           unsigned consequent_constraint) const;

    
    static const unsigned ERASED = -1U;
    
    unsigned m_first_conflict_constraint;
    /**
     * @brief A set of prerequisite conflict constraint for each conflict
     *        constraint.
     *
     * To save space, m_prerequisites[i] refers to the i-th conflict constraint
     * not the first constraint of the system. So that the true conflict
     * constraint of entry \p i is m_first_conflict_constraint+\p i.
     *
     * The data stored in the vector contains the original constraint indexes,
     * which are exactly the same as those in hcsp::solver.
     */
    std::vector<constraint_prerequisites> m_prerequisites;
    //! To save space, m_usage_backtrack_count[0] refers to the first conflict
    //! constraint
    std::vector<unsigned>             m_usage_backtrack_count;
}; // class conflict_constraint_dependency

} // namespace hcsp

#endif // HCSP_CONFLICT_CONSTRAINT_DEPENDENCY_HPP_
