// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_CONSTRAINT_AGED_INFO_HPP_
#define HCSP_CONSTRAINT_AGED_INFO_HPP_

#include <constraint_runtime_info.hpp>

namespace hcsp {

class constraint_connectivity_info;
class assignment_stack;

/**
 * This class maintains the run time information for constraint activation,
 * as was available right after the given modification entry.
 */
class constraint_aged_info final: public constraint_runtime_info
{
  private:
    template <typename T> static T* mutable_or_null(T* ptr)   { return ptr; }
    template <typename T> static T* mutable_or_null(const T*) { return 0;   }

    //! A an abstract class used to store object references with to_variable()
    //! methods into constraint_aged_info.
    class abstract_param_to_var
    {
      public:
        virtual ~abstract_param_to_var() {}
        virtual unsigned to_variable(unsigned param) const = 0;
    }; // class constraint_aged_info::abstract_param_to_var

    //! A concrete class that stored a reference to a concrete object that has
    //! a to_variable() method.
    template <typename Param_To_Var>
    class concrete_param_to_var: public abstract_param_to_var
    {
      public:
        concrete_param_to_var(const Param_To_Var & param_to_var)
            : m_param_to_var_ref(param_to_var) {}
        virtual unsigned to_variable(unsigned param) const
        { return m_param_to_var_ref.to_variable(param); }
        
      private:
        const Param_To_Var & m_param_to_var_ref;
    }; // class constraint_aged_info::concrete_param_to_var
    
  public:
    /**
     * Construct the aged runtime constraint information.
     *
     * @param base_info
     *           An info object that has a method to_variable(i) which gives the
     *           variable ID for parameter i. It also holds the current state of
     *           the constraint runtime. The \p base_info object may be modified
     *           by methods of constraint_aged_info and is kept up-to-date by
     *           external means.
     *           Note that if param_to_var is const then the constructed object
     *           will not be allowed to modify it.
     *
     * @param assignments
     *           The assignment stack which contains the set of assignments
     *           and decisions which lead to the current domains of variables.
     *           This \e assignments is kept up to date by external means.
     *
     * @param propagation_modification
     *           The reference modification entry. The constraint_aged_info
     *           object refers to domains right before the said constraint
     *           propagation.
     *
     * @param effort_in
     *           The effort to be returned by this object's effort() method
     */
    template<typename Info_To_Var>
    constraint_aged_info(Info_To_Var & info_to_var,
                         const assignment_stack & assignments,
                         change_index_t propagation_modification,
                         unsigned effort_in=-1U)
        : m_param_to_var(new concrete_param_to_var<Info_To_Var>(info_to_var)),
          m_const_base_info(info_to_var),
          m_p_base_info(mutable_or_null(&info_to_var)),
          m_assignment_stack(assignments),
          m_after_modification(assignments.before_propagation(
                                   propagation_modification)),
          m_effort(effort_in != -1U ? effort_in : info_to_var.effort()),
          m_allow_latest_modification_query(true)
    {}
    
    virtual ~constraint_aged_info() {}
    constraint_aged_info(constraint_aged_info &&) = default;
    constraint_aged_info & operator=(constraint_aged_info &&) = default;

    virtual bool reduced_between_activations(unsigned) const
    { return false; }

    virtual queue_type originating_queue() const
    { return m_const_base_info.originating_queue(); }

    //! This method may not be called for constraint_aged_info since
    //! it modifies solver's state.
    virtual void put_in_queue(queue_type queue);
    
    //! Return the value of the indexed variable right after the said
    //! modification.
    //! \param param_index  The index of the parameter in constraint's
    //!                     parameter list.
    virtual const iset & domain(unsigned param_index) const;

    //! This should not be called, as the domains are intended only to be read.
    virtual void set_domain(unsigned param_index, iset new_domain) override;
    // Number of variables connected to this constraint
    virtual unsigned     arity() const;
    virtual void deallocate_watch(unsigned param_index, watch_type type,
                                  watch_handle_t watch_handle) override;
    virtual watch_handle_t allocate_watch(unsigned param_index, 
                                          watch_type type) override;
    virtual void deallocate_value_watch(unsigned param_index,
                                        iset::value_type value,
                                        watch_handle_t watch_handle) override;
    virtual watch_handle_t allocate_value_watch(unsigned param_index,
                                                iset::value_type value)override;
    virtual int latest_modified_param(unsigned except) const;

    virtual unsigned effort() const { return m_effort; }

    virtual void print(std::ostream & ) const throw();
    //! This will print the header of the possibly with constraint name and
    //! index
    virtual void print_prefix(std::ostream & ) const throw();

    // FIXME: this should go to a new cpp file, with a reasonable default
    //! Print the identifying information on this parameter (e.g. name)
    virtual void print_param(std::ostream &, unsigned ) const throw();

    //! Enable latest_modified_param() or disable it, making it return -1
    void allow_latest_modification_query(bool flag)
    { m_allow_latest_modification_query= flag; }

  private:
    unsigned to_variable(unsigned param) const
    { return m_param_to_var->to_variable(param); }

    //! Owns a pointer to object with to_variable() method
    std::unique_ptr<abstract_param_to_var> m_param_to_var;
    
    //! The constraint_runtime_info with which this object was constructed.
    const constraint_runtime_info       &m_const_base_info;
    /**
     * Point to the mutable constraint_runtime_info with which this object was
     * constructed. If it was constructed with a const constraint_runtime_info,
     * The this will pointer will be NULL.
     */
    constraint_runtime_info             *m_p_base_info;
    const assignment_stack &m_assignment_stack;
    change_index_t          m_after_modification;
    unsigned                m_effort;
    bool                    m_allow_latest_modification_query;
};

} // namespace hcsp
#endif // HCSP_CONSTRAINT_AGED_INFO_HPP_
