// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_CONSTRAINT_CONNECTIVITY_INFO_HPP_
#define HCSP_CONSTRAINT_CONNECTIVITY_INFO_HPP_

#include <vector>
#include <iosfwd>
#include <constraint_runtime_info.hpp>
#include <constraint_propagator.hpp>
#include <iset.hpp>

namespace hcsp {

class solver;

//! This class maintains all constraint information, including
//! runtime, connectivity without the constraint propagator itself.
//!
//! FIXME: suboptimal implementation:
//!     -#  m_arguments is a vector, which will cause an extra dereferencing.
//!         Also this data is static, so it should be probably put together
//!         with other static data.
//!     -#  m_constraint_index is redundant, and can be passed by arguments.
//!     -#  m_p_solver is redundant. There should be a better way to pass
//!         a reference to the solver without wasting precious "hot" memory.
//!        
class constraint_connectivity_info final : public constraint_runtime_info
{
  public:
    
    constraint_connectivity_info() 
        : m_p_solver(NULL), m_conflict_rate(0), m_last_activation_time(0),
          m_labels(NO_LABEL)
    {}
    constraint_connectivity_info(solver & solver, unsigned constraint_index)
        : m_p_solver(&solver), m_constraint_index(constraint_index),
          m_conflict_rate(0),
          m_last_activation_time(0), m_last_activation_reduced(false),
          m_labels(NO_LABEL)
    {}
    constraint_connectivity_info(constraint_connectivity_info &&) = default;
    constraint_connectivity_info & operator=(constraint_connectivity_info &&)
    = default;

    constraint_connectivity_info & reserve_arguments(unsigned count)
    {
        m_arguments.reserve(count);
        return *this;
    }
    void reset()
    {
        m_last_activation_time= 0;
        m_last_activation_reduced= false;
    }
    //! Set the labels associated with this constraint, return this
    constraint_connectivity_info & labels(label_type label)
    {
        m_labels= label;
        return *this;
    }
    //! Get the labels associated with this constraint
    label_type labels() const { return m_labels; }
    //! Add a new variable to the constraint. This link is bidirectional,
    //! and both constraint and variable know about each other.
    constraint_connectivity_info & operator<<(unsigned variable_index);
    virtual ~constraint_connectivity_info();
    virtual bool reduced_between_activations(unsigned param_index) const;
    virtual queue_type   originating_queue() const;
    virtual void         put_in_queue(queue_type queue);
    virtual const iset & domain(unsigned param_index) const;
    virtual void set_domain(unsigned param_index, iset domain) override;
    virtual unsigned     arity() const   { return m_arguments.size(); }
    bool last_activation_reduced() const { return m_last_activation_reduced; }
    inline bool reduced_by_last_activation(unsigned param) const;
    void prepare_next_activation()       { m_last_activation_reduced= false; }
    void finalize_prev_activation(time_type current_time);
    unsigned to_variable(unsigned param_index) const
    { return m_arguments.at(param_index); }
    //! Find the parameter which points to the requested variable.
    unsigned find_param_of_variable(unsigned variable_index) const
    {
        for (unsigned i=0 ; i != m_arguments.size() ; ++i) {
            if ( m_arguments[i] == variable_index )
                return i;
        }
        return ~static_cast<unsigned>(0);
    }

    virtual void deallocate_watch(unsigned param_index, watch_type type,
                                  watch_handle_t watch_handle);
    virtual watch_handle_t allocate_watch(unsigned param_index, 
                                          watch_type type);
    virtual void deallocate_value_watch(unsigned param_index,
                                        iset::value_type value,
                                        watch_handle_t watch_handle);
    virtual watch_handle_t allocate_value_watch(unsigned param_index,
                                                iset::value_type value);
    virtual void print(std::ostream & out) const throw();
    //! This will print the header of the possibly with constraint name and
    //! index
    virtual void print_prefix(std::ostream & out) const throw();
    //! Print the identifying information on this parameter (e.g. name)
    virtual void print_param(std::ostream &, unsigned ) const throw();
    //! Free the memory pointed from this entry, effectively destroying the
    //! contents of this entry.
    void free()
    {
        if (m_arguments.capacity() != 0)
            std::vector<unsigned>().swap(m_arguments);
    }
    friend std::ostream &operator<<(std::ostream & out, 
                                    const constraint_connectivity_info & info)
    { info.print(out); return out; }
    unsigned conflict_rate() const { return m_conflict_rate; }
    void mark_conflicting() { ++m_conflict_rate; }
  private:
    
    solver               *m_p_solver; // deallocation not managed in this class
    std::vector<unsigned> m_arguments;
    unsigned              m_constraint_index;

    //! A heuristic counter for the effect of this constraint on conflicts
    unsigned              m_conflict_rate;
    time_type             m_last_activation_time;
    bool                  m_last_activation_reduced;

    //! The set of labels set to this variable and its domain
    label_type            m_labels;
};   // constraint_connectivity_info

} // namespace hcsp
#endif // HCSP_CONSTRAINT_CONNECTIVITY_INFO_HPP_
