// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_CONSTRAINT_EXPLICIT_INFO_HPP_
#define HCSP_CONSTRAINT_EXPLICIT_INFO_HPP_

#include <constraint_runtime_info.hpp>
#include <vector>
#include <iset.hpp>

namespace hcsp {
class variable_printer;
/**
 * @brief A constraint_runtime_info which contains variable domains and does
 *        not take them from external sources.
 *
 * @todo It is advisable to merge the definitions of constraint_explicit_info
 *       with hcsp::trivial_runtime_info (in test-impl/trivial_runtime_info.hpp).
 *       This has not been done yet since trivial_runtime_info has several
 *       costs related to testing which we may not want to pay everywhere else. 
 */
class constraint_explicit_info final : public constraint_runtime_info,
                                       private var_watchers_base
{
  public:
    constraint_explicit_info(unsigned in_arity);
    constraint_explicit_info(const constraint_explicit_info & info);
    constraint_explicit_info& operator=(const constraint_explicit_info & info);
    constraint_explicit_info(constraint_explicit_info && info)= default;
    constraint_explicit_info& operator=(constraint_explicit_info &&) = default;
    virtual ~constraint_explicit_info();
    virtual bool reduced_between_activations(unsigned) const { return true; }

    virtual queue_type originating_queue() const { return CONSTRAINT_QUEUE; }
    virtual void put_in_queue(queue_type) {}

    /**
     * @brief Sets the variable printer for an argument
     *
     * This methods makes it possible to print arguments using different
     * strategies. If called, then the default strategy for the argument is
     * overwritten and the argument will be printed using this printer.
     *
     * @param arg     The argument to effect
     * @param printer The variable printer to use for printing the argument
     */
    void set_var_printer(unsigned arg, const variable_printer & printer);

    virtual const iset & domain(unsigned param_index) const
    { return m_domains[param_index]; }
    virtual void set_domain(unsigned param_index, iset new_domain) override
    { 
        if (m_domains[param_index].size() != new_domain.size() ||
            m_domains[param_index].empty() != new_domain.empty())
            m_domains[param_index]= std::move(new_domain); 
    }
    virtual unsigned arity() const
    { return m_domains.size(); }
    virtual void deallocate_watch(unsigned, watch_type, 
                                  watch_handle_t watch_handle) override
    { deallocate_value_watch(0,0, watch_handle); }
    virtual watch_handle_t allocate_watch(unsigned, watch_type) override
    { return allocate_value_watch(0,0); }
    virtual void deallocate_value_watch(unsigned, iset::value_type, 
                                        watch_handle_t watch_handle) override
    { m_watch_list.erase(watch_handle.get_it()); }
    virtual watch_handle_t allocate_value_watch(unsigned, iset::value_type)
        override
    { 
        m_watch_list.push_back(watcher_info_type{0});
        return --m_watch_list.end();
    }
    virtual void print(std::ostream & ) const throw();
    virtual void print_prefix(std::ostream & ) const throw();
    virtual void print_param(std::ostream &, unsigned ) const throw();
    
  private:
    std::vector<iset> m_domains;
    std::vector<variable_printer> m_var_printers;
    watch_list_t m_watch_list;
}; // class constraint_explicit_info

} // namespace hcsp
#endif // HCSP_CONSTRAINT_EXPLICIT_INFO_HPP_
