// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_CONSTRAINT_PROPAGATOR_HPP_
#define HCSP_CONSTRAINT_PROPAGATOR_HPP_

#include <vector>
#include <iosfwd>
#include <memory>
#include <propagation_algo.hpp>
#include <constraint_runtime_info.hpp>
#include <explanation_rule.hpp>
#include <explanation_generators.hpp>
#include <iset.hpp>

namespace hcsp {
class disjunction_specification;
template <typename T> class controlled_ptr;
/**
 * @brief The basic constraint propagation interface.
 */
class constraint_propagator : protected watch_types, public propagation_types 
{
  public:
    //! enum result_type is inherited from propagation_types

    virtual ~constraint_propagator() {}

    /**
     * @brief Get the number of parameters this constraint expects to get.
     *
     * In some rare cases this property is defined only after init(). When
     * this value is unknown then a negative count is returned.
     *
     * @return Either negative or the number of parameters of this constraint.
     */
    virtual int count_params() const { return -1; }
    /**
     * @brief Initialize the constraint and the 'info' before the first
     * propagation of the constraint.
     *
     * The default implementation puts modification watches on all parameters.
     * @return false if initially unsatisfiable
     */
    virtual bool init( constraint_runtime_info & info );
    /**
     * @brief Propagate the constraint onto the 'info' object.
     *
     * This call may change the watches associated with the constraint,
     * and reduce domains. It may not widen domains of variables.
     */
    virtual result_type propagate( constraint_runtime_info & info ) = 0;

    /**
     * @brief Tell if propagation would lead to UNSAT
     *
     * @return True if and only if running propagate() would return UNSAT.
     */
    virtual bool conflicts(const constraint_runtime_info & info) const = 0;

    virtual bool constraint_check(const constraint_runtime_info & info) const=0;

    //! \brief A simplistic printing method.
    friend std::ostream & operator<<(std::ostream & out, 
                                     const constraint_propagator & propagator)
    {
        propagator.print(out);
        return out;
    }

    /**
     * @brief A fine grained printing method.
     *
     * When this method accepts a constraint_runtime_info, it may use the
     * information in the p_info object to print names of variables in
     * correct semantic locations.
     */
    virtual void print(std::ostream &out,
                       const constraint_runtime_info *p_info=NULL)const throw()
    {
        if (p_info)
            p_info->print(out);
    }

    /**
     * @brief Find out on what parameter this watch_index waits.
     *
     * @param watch_index  The index of this watch in this propagators.
     *                     This index points to an internal array of watches.
     *
     * @pre watch_index < count_watches(). See count_watches() for implications
     *
     * @return  The index of the parameter that the n-th watch waits for.
     */
    virtual unsigned watched_param_index(unsigned watch_index) const;
    /**
     * @brief Give the number of watches that belong to this propagator.
     *
     * It is critical to make the behavior of all of these methods consistent:
     *   - count_watches()
     *   - watched_param_index()
     *
     * The propagator does not control watches with index >= count_watches().
     * For such an index, the above methods cannot be called.
     *
     * @return  Number of watches.
     */
    virtual unsigned count_watches() const                 { return 0; }
    
    /**
     * Forget all watches managed by this constraint.
     *
     * If a watch is managed by this constraint, then even though it
     * is no longer managed, it is still allocated.
     *
     * If some watches are not managed by this constraint, then
     * they will not be affected by this call.
     *
     * The default implementation of this method does nothing.
     */
    virtual void forget_watches() { }
    
    /**
     * @brief Explain a past domain reduction done by this propagator.
     *
     * Generate a clause which explains the queried reduction (see the
     * postcondition for more details).
     *
     * The default behavior is fast, but it generates a very poor clause.
     * In the default explanation, each literal uses the biggest possible
     * set of values, which make the clause propagate poorly.
     * For a better, but somewhat slower, behavior you can use
     * tight_constraint_propagator::explain_propagation().
     *
     * @param[in]  propagated_param
     *             The index of the reduced parameter which we need to explain.
     * @param[in]  values_to_explain
     *             The set of values which, if removed, the caller wants to
     *             explain.
     *             Only the values of \p values_to_explain have to be
     *             \e correct w.r.t. the domains in the \p explanation.\n
     *             \n
     *             The caller of this method does not care if other values
     *             are added or omitted from the explanation.
     *             This is basically a refinement from propagated_param 
     *             to specific values. This freedom lets the explanation
     *             algorithm have potentially conflicting choices:
     *              - Fewer literals
     *              - Literals with smaller sets.
     *             These requirements may conflict and getting the best
     *             explanation becomes a matter of heuristics for the
     *             propagator.
     * @param[in]  propagated_domain
     *             The domain resulting from the propagation which we need
     *             to explain.
     * @param[in]  info_before_propagation
     *             The state of the solver before the propagation which we need
     *             to explain.
     * @param[out] explanation
     *             The literals of the explanation clause. The clause
     *             is \f$ ( v_0 \in \mbox{explanation}[0]\vee \cdots \vee
     *                     v_1 \in \mbox{explanation}[N-1]) \f$
     *
     * @pre  The domain reduction was performed by this propagator for the
     *       given \b info_before_propagation
     * @post The resulting clause must:
     *     - Reduce the propagated parameter to \b propagated_domain
     *       for the same input domains.
     *     - For an arbitrary input domains, it should not remove domain values
     *       which are still supported by the inputs to this constraint.
     */
    virtual void explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation) const
    {
        explanation_generators::explain_propagation(
            propagated_param, values_to_explain, propagated_domain,
            info_before_propagation, explanation);
    }

    /**
     * @brief For each parameter get the values that adding them may cripple
     * propagation.
     *
     * See explanation_generators::possibly_crippling_values() for more
     * details.
     *
     * Approximation: The constraint is allowed to over-approximate the
     * result. 
     * 
     * @param[in]  propagated_param,propagated_domain
     *             The propagation that may be ruined by the values
     * @param[in]  info_before_propagation
     *             The state of the solver before the propagation which we need
     *             to operate on
     * @param[out] values
     *             For each parameter if values.size() is big enough then
     *             values[parameter] is the set of values which will cripple
     *             propagation.
     */
    virtual void possibly_crippling_values(
        unsigned propagated_param, const iset & propagated_domain,
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & values) const
    {
        std::unique_ptr<constraint_propagator> this_clone(clone());
        explanation_generators::possibly_crippling_values(
            *this_clone, propagated_param, propagated_domain,
            info_before_propagation, values);
    }
    
    /**
     * @brief Return the rule for this explanation, if possible.
     * @param[in] literals The explanation literals;
     * @return The reduction rule that creates the explanation from
     *         this constraint. It is up to the implementation to either
     *         return NULL or an object.
     */
    virtual std::unique_ptr<explanation_rule> rule(
        const std::vector<iset> & literals) const
    { (void)literals; return std::unique_ptr<explanation_rule>(); }

    /**
     * Create an identical propagator to this one.
     * The newly created propagator will be able to follow the semantics of
     * the old propagator:
     *  - If the propagator has members that define its semantics then
     *    these members have to be copied. For example, if this is a clause
     *    then all the literals have to be copied.
     *  - If the propagator has members that control propagation/explanation
     *    heuristics (in case of inaccurate propagation) then they have
     *    to be copied.
     *  - If the propagator contains propagation history of some sort
     *    (which may affect heuristic outcomes) then this may or may not
     *    be copied.
     */
    virtual constraint_propagator * clone() const= 0;

    /**
     * @brief Intersect the constraint with the given domains.
     *
     * If the intersection can be represented as a stronger constraint then
     * allocate and return the new constraint. If the new constraint requires
     * fewer or differently ordered arguments then write the new order in
     * \p copy_from_argument.
     *
     * @param[in]  info
     *             The domains to intersect with
     * @param[out] copy_from_argument
     *             If not empty then this is the new order of variables in the
     *             new constraint. Argument i of the new constraint will be the
     *             same as copy_from_argument[i] of the old constraint.
     * @return If not NULL, the new and stronger constraint. If NULL then
     *         no stronger constraint is provided. Note that it is possible that
     *         theoretically a stronger constraint can be created but NULL
     *         is nevertheless returned.
     */
    virtual constraint_propagator * intersect_with(
        const constraint_runtime_info & info,
        std::vector<unsigned> & copy_from_argument) const
    {
        (void)copy_from_argument;
        (void)info;
        return NULL;
    }

    /**
     * @param Tell if an assignment to the parameter can falsify the constraint
     *        for a sub-domain of info.
     *
     * This method is allowed to return false when it is to difficult to compute
     * the result, but if it returns true then it must be correct.
     *
     * @return True if the there an assignment to all other parameters other
     *         than param, from the domain of info, for which param controls the
     *         satisfiability of. That is, it can find two different values for
     *         param, out of info.domain(param), one of them satisfies the
     *         constraint and the other falsifies it.
     */
    virtual bool potentially_falsifying(const constraint_runtime_info & info,
                                        unsigned param) const
    {
        (void)info;
        (void)param;
        return false;
    }
    /**
     * @brief Split a disjunction of clause (or literal) with a constraint
     *        into literals and constraint.
     *
     * If the constraint is comprised from (clause || constraint) or from
     * (literal || constraint) then split the constraint into a constraint and a
     * set of literals. If it is a simple clause then return the literals and a
     * NULL constraint. If it has no literal part then return empty set of
     * literals.
     *
     * This method is optional and if the deriving class implements it then it
     * will update the fields and return true. Otherwise it will return false
     * and the value of the disjunction_specification is unspecified.
     *
     * @param[out] disjunction The constraint disjunction split out into a
     *             constraint and a set of literals.
     *
     * @return True if the arguments were updated and got their correct values.
     */
    virtual bool get_disjunction_specification(
        disjunction_specification & disjunction) const;

    /**
     * @brief Create a disjunction of two constraints: 'this' and 'other'
     *
     * If it is not possible to represent the disjoint constraint, or this
     * method is not implemented, then return false.
     *
     * @param[in] this_constraint
     *            The disjunction that contains 'this' constraint, with
     *            the relevant argument mapping and domains.
     *
     * @param[in] other_constraint
     *            The disjunction that contains a constraint which we want do
     *            disjoin with the relevant argument mapping and domains.
     *
     * @param[out] result_constraint
     *            The result from performing the disjunction. It may contain
     *            new literals or, when there will be support, both constraints.
     *
     * @return true if the method is implemented and if it could produce a valid
     *         disjunction. If it returns false then the output parameter of
     *         this method is unspecified. 
     */
    virtual bool disjoin_constraint(
        const disjunction_specification & this_constraint,
        const disjunction_specification & other_constraint,
        disjunction_specification & result_constraint) const;
    /**
     * @brief Create an over-approximated conjunction of the constraint part of
     * this_constraint and other_constraint.
     *
     * If it is not possible to represent the combined result, or this
     * method is not implemented, then return false.
     *
     * @param[in,out] this_constraint,other_constraint
     *            The disjunction specifications that hold the two constraints.
     *            this_constraint.constraint() is must be equal to 'this'
     *            The connectivity of the constraints is defined here, including
     *            the domains.
     *            The method may perform destructive operations on these
     *            arguments.
     *
     * @param[in] pivot
     *            An argument of this_constraint which shares a variable with
     *            other_constraint. If the method returns an over-approximation
     *            then it prefers to strengthen this pivot to the point when
     *            the resulting constraint conflicts the domains of
     *            this_constraint and other_constraint.
     * @param[out] result_constraint
     *            A constraint specification of the combined constraint. This
     *            type allows the value to either point to an existing
     *            constraint or to allocate a new one. Upon its destruction it
     *            may delete the pointed constraint, depending on actual
     *            ownership. A constraint may add literals.
     *            Initially this object is empty.
     * @param[out] copy_from_argument
     *            Each of the arguments of the inner_constraint has an entry in
     *            this vector. If it is smaller than this->arity() then it
     *            refers to a variable that 'this' propagates, and otherwise it
     *            refers to a variable that 'other' propagates. The values of
     *            this vector must not repeat and should be below
     *            this->arity()+other.arity().
     *
     * @return true if the method is implemented and if it could produce a
     *         single constraint that represents the combination. If it returns
     *         false then other parameters of this method are unspecified.
     *
     */
    virtual bool combine_base_constraints(
        disjunction_specification & this_constraint,
        disjunction_specification & other_constraint,
        unsigned pivot,
        disjunction_specification & result_constraint) const;
    
    /**
     * @brief Create a constraint based on the disjunction specification.
     *
     * This method returns a constraint according to the provided
     * specification. During its operation it destroys 'disjunction', and
     * updates copy_from_argument to match constraint_var(..) and
     * literal_var(..) values of the disjunction. If 'this' is owned by the
     * disjunction specification then the object in 'disjunction' may be
     * returned, or embedded in the returned propagator.
     *
     * @param[in,out] disjunction
     *                The disjunction specification that defines the constraint
     *                to be returned. The method may clear the disjunction after
     *                the call, but is not required to.
     * @param[out]    copy_from_argument
     *                The final mapping of the new constraint from some other
     *                arguments as defined by disjunction.constraint_var(...)
     *                and disjunction.literal_var(...)
     *
     * @return NULL if method is not defined, otherwise a constraint that
     *         matches the specification. It is legal to return 'this' or embed
     *         it in the result if it is owned by 'disjunction'. In which case
     *         the method must release the propagator from 'disjunction'.
     *
     * @pre disjunction.constraint() == this
     */
    virtual constraint_propagator * new_propagator(
        disjunction_specification & disjunction,
        std::vector<unsigned> &copy_from_argument) const;
    /**
     * @brief Is there any chance that 'this' and 'other' may be combined using
     *        a specialized combination rule.
     *
     * By default return false, unless other is an equal_propagator.
     */
    virtual bool may_combine(const constraint_propagator & other) const;
    /**
     * @brief Attempt to combine 'this' and 'other' into an new constraint
     *        with the pivot exist-out to a separate literal.
     *
     * The 'combine' operation is the extension of clause resolution into
     * general constraints. If the resulting constraint refers to the pivot
     * variable it will be in a separate literal which will be in a disjunciton
     * with other parts of the constraints.
     * The resulting constraint must conflict the domains at the time when other
     * is propagated.
     *
     * If combine can't complete its work without violating one of the
     * requirements it returns NULL.
     *
     * @param[in] info  The domains of 'this' before it is propagated.
     *
     * @param[in] other,other_info
     *            The other constraint and domains to combine with.
     *
     * @param[in] other_to_common_param
     *            Consider val=other_to_common_param[i]. If val < info.arity()
     *            then other_info.domain(i) and info.domain(val) refer to the
     *            same variable. Otherwise val < info.arity()+other_info.arity()
     *            and is otherwise unspecified. All values of
     *            other_to_common_param must be different from each other.
     *
     * @param[in] other_pivot_param
     *            The pivot, defined by its index in other. Note that
     *            copy_from_argument[other_pivot_param] < info.arity(), which
     *            means that the pivot variable is shared between the
     *            constraints.
     * @param[out] copy_from_argument
     *            Defines the arguments in the resulting constraint. Assume that
     *            copy_from_argument[i]==val, then val<info.arity() means that
     *            argument i of the result is the same as 'val' argument of
     *            'this'. Otherwise, i.e., val>=info.arity(), then the result
     *            takes its argument from val-info.arity() argument of 'other'.
     *
     * @return If not NULL, then this is a constraint implied from 'this' and
     *         'other' for the given input parameters, and strong enough to
     *         reach a conflict with the domains of info and other_info.
     * @pre The preconditions are:
     *      -# other_to_common_param has no repetitions.
     *      -# other_to_common_param.size() == other_info.arity().
     *      -# Forall other_to_common_param[i]<info.arity()+other_info.arity().
     *      -# other_info's arity() and domains match 'other'.
     *      -# info's arity() and domains match 'this'.
     *      -# Propagating both 'this' and 'other' on the earlier of 'info' and
     *         'other_info' reaches a conflict.
     *      -# other_pivot_param is a valid param of other_info, and
     *         other_to_common_param[other_pivot_param] is a valid param of
     *         'info'.
     * @post The post conditions, if the result is not NULL, are:
     *      -# There are no two copy_from_argument elements that point to the
     *         same variable, either directly or according to the mapping of
     *         other_to_common_param.
     *      -# The result conflicts with the earliest of the two input domains
     *         of info and other_info.
     */
    virtual constraint_propagator * combine(
        const constraint_runtime_info & info,
        const constraint_propagator & other,
        const constraint_runtime_info & other_info,
        const std::vector<unsigned> & other_to_common_param,
        unsigned other_pivot_param,
        std::vector<unsigned> & copy_from_argument) const;

    /**
     * @brief Calls other.combine(...,*this,...) and fixes arguments
     *        accordingly.
     *
     * @param[in] info  The domains of 'this' before it is propagated.
     *
     * @param[in] other,other_info
     *            The other constraint and domains to combine with.
     *
     * @param[in] other_to_common_param
     *            The conversion from other's argument indices to the common
     *            indices.
     *
     * @param[in] other_pivot_param
     *            The pivot, defined by its index in other.
     *
     * @param[out] copy_from_argument
     *            Defines the arguments in the resulting constraint. 

     * @return The result of other.combine(...,*this,...)
     */
    constraint_propagator *combine_other_first(
        const constraint_runtime_info & info,
        const constraint_propagator & other,
        const constraint_runtime_info & other_info,
        const std::vector<unsigned> & other_to_common_param,
        unsigned other_pivot_param,
        std::vector<unsigned> & copy_from_argument) const;
    
  private:
    static void test_concept()
    { (void)&propagation_algo_concept<constraint_propagator>::test_concept; }
}; // class constraint_propagator

/**
 * @brief A constraint propagation with tight explanations - by default.
 *
 * This is exactly like constraint_propagator, only with a tighter
 * explain_propagation() which calls
 * constraint_propagator::explain_propagation_tightly().
 */
class tight_constraint_propagator : public constraint_propagator
{
  public:
    virtual void explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation) const
    {
        std::unique_ptr<constraint_propagator> propagator(clone());
        explanation_generators::explain_propagation_tightly(
            *propagator, propagated_param, values_to_explain, propagated_domain,
            info_before_propagation, explanation);
    }
    
};

/**
 * @brief A constraint propagation with tight explanations for one parameter.
 *
 * This is exactly like constraint_propagator, only with a tighter
 * explain_propagation(). Tightening happens only on one of constraint
 * parameters. For many constraints such tightening is sufficient to get
 * an optimal explanation (at much lower cost).
 */
class tight1_constraint_propagator : public constraint_propagator
{
  public:
    virtual void explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation) const
    {
        std::unique_ptr<constraint_propagator> propagator(clone());
        explanation_generators::explain_propagation_tightly1(
            *propagator, propagated_param, values_to_explain, propagated_domain,
            info_before_propagation, explanation);
    }
    
};

template <typename Propagation_Algo>
class constraint_propagator_from_algo : public constraint_propagator
{
  protected:
    typedef Propagation_Algo algo_t;
  public:
    constraint_propagator_from_algo() {}
    template <typename T1>
    constraint_propagator_from_algo<algo_t>(const T1 &v1)
        : m_algo(v1)
    {}
    template <typename T1,typename T2>
    constraint_propagator_from_algo<algo_t>(const T1 &v1, T2 &v2)
        : m_algo(v1,v2)
    {}
    template <typename T1,typename T2,typename T3>
    constraint_propagator_from_algo<algo_t>(const T1 &v1, const T2 &v2,
                                            const T3 &v3)
        : m_algo(v1,v2, v3)
    {}
    virtual constraint_propagator_from_algo<algo_t>* clone() const
    {
        (void)&propagation_algo_concept<algo_t>::test_concept;
        return new constraint_propagator_from_algo<algo_t>(*this);
    }

    //! Return the underlying constraint propagation algorithm
    Propagation_Algo       & algo()       { return m_algo; }
    
    //! Return the underlying constraint propagation algorithm
    const Propagation_Algo & algo() const { return m_algo; }
    virtual int count_params() const { return m_algo.count_params(); }
    virtual bool init(constraint_runtime_info & info)
    {
        HCSP_ASSERT( count_params() < 0 ||
                    (unsigned)count_params() == info.arity() );
        return m_algo.init(info);
    }
    virtual bool constraint_check(const constraint_runtime_info & info) const
    { return m_algo.constraint_check(info); }
    virtual result_type propagate(constraint_runtime_info & info)
    { return m_algo.propagate(info); }
    virtual bool conflicts(const constraint_runtime_info & info) const
    { return m_algo.conflicts(info); }

    void print(std::ostream & out,
               const constraint_runtime_info * p_info=NULL) const throw()
    { m_algo.print(out, p_info); }
    unsigned count_watches() const
    { return m_algo.count_watches(); }

    unsigned watched_param_index(unsigned watch_index) const
    { return m_algo.watched_param_index(watch_index); }

    void forget_watches()
    { m_algo.forget_watches(); }

    void explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation) const
    {
        m_algo.explain_propagation(
            propagated_param, values_to_explain, propagated_domain,
            info_before_propagation, explanation);
    }

  protected:
    algo_t m_algo;
}; // constraint_propagator_from_algo

template <class Propagator_Algo>
class propagator_printer
{
  public:
    propagator_printer(const Propagator_Algo & algo,
                       const constraint_runtime_info * info)
        : m_algo(algo), m_info(info) { }
    friend std::ostream & operator<<(
        std::ostream & out, const propagator_printer<Propagator_Algo> & printer)
    {
        printer.m_algo.print(out, printer.m_info);
        return out;
    }
  private:
    const Propagator_Algo & m_algo;
    const constraint_runtime_info * m_info;
}; // propagator_printer<Propagator_Algo>

template <class Propagator_Algo>
propagator_printer<Propagator_Algo> make_printer(Propagator_Algo &algo,
                                                 constraint_runtime_info *info)
{
    return propagator_printer<Propagator_Algo>(algo, info);
}

std::ostream & operator<<(std::ostream&, constraint_propagator::result_type);
} // namespace hcsp
#endif // HCSP_CONSTRAINT_PROPAGATOR_HPP_
