// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_CONSTRAINT_RUNTIME_INFO_HPP_
#define HCSP_CONSTRAINT_RUNTIME_INFO_HPP_

#include <vector>
#include <initializer_list>
#include <watch_types.hpp>
#include <solver_types.hpp>
#include <iset.hpp>

namespace hcsp {

class watch_handle_t;
//! This class maintains the run time information for constraint activation.
//! It is an abstract class, so that there can be different concrete
//! data structures. For example, testing code may use a lean data structure.
//! On the other hand, the solver may use a complex data structure with
//! references to itself from this data structure.
class constraint_runtime_info : protected watch_types,
                                protected solver_types
{
  public:
    //! \brief The type of the queue that causes constraint activation.
    enum queue_type {
        //! A queue of modified variables that are processed. Check watches
        //! of one at a time, and activate affected constraints.
        //! A variable is present only once in the queue.
        //!
        //! When more than one variable is modified, it is possible that
        //! a constraint will be activated more than once, each time due
        //! to a different watch, without a change between activations.
        VARIABLE_QUEUE,

        //! A queue of constraints that have to be activated.
        //! A constraint is present only once in this queue.
        CONSTRAINT_QUEUE
    };
    virtual ~constraint_runtime_info() {}
    constraint_runtime_info() = default;
    constraint_runtime_info(constraint_runtime_info&&) = default;
    constraint_runtime_info & operator=(constraint_runtime_info&&) = default;
    constraint_runtime_info(const constraint_runtime_info&) = default;
    constraint_runtime_info & operator=(const constraint_runtime_info&)=default;

    //! Did this variable change since last activation of the
    //! constraint propagator.
    //! This is intended to work correctly only from inside the constraint
    //! propagator.
    //! \param param_index  The index of the parameter in constraint's
    //!                     parameter list.
    virtual bool reduced_between_activations(unsigned param_index) const=0;

    //! \brief Tell what queue triggered the activation of this constraint.
    virtual queue_type originating_queue() const = 0;
    
    //! \brief Put this constraint on the specified queue.
    //! VARIABLE_QUEUE is not supported.
    virtual void put_in_queue(queue_type queue) = 0;

    //! Return the latest value of the indexed variable.
    //! \param param_index  The index of the parameter in constraint's
    //!                     parameter list.
    virtual const iset & domain(unsigned param_index) const =0;

    //! Let the caller update the latest value of the indexed variable.
    //! This class may notify some other data structure of this update.
    //! In a given invocation of the constraint propagator, the updated domain
    //! should be a subset of the original domain (i.e. may be equal).
    //! \param param_index  The index of the parameter in constraint's
    //!                     parameter list.
    //! \param new_domain   The new value of the domain.
    virtual void set_domain(unsigned param_index, iset new_domain)=0;
    void set_domains(std::initializer_list<unsigned> params, iset new_domain)
    {
        unsigned moved_to_param = -1U;
        auto it = params.begin();
        for ( ; ; ++it) {
            if (it == params.end())
                return;
            if (new_domain.size() != domain(*it).size() || new_domain.empty()) {
                moved_to_param= *it;
                set_domain(*it, std::move(new_domain));
                break;
            }
        }
        for (++it; it != params.end(); ++it)
            set_domain(*it, domain(moved_to_param));
    }
    // Number of variables connected to this constraint
    virtual unsigned     arity() const                      =0;
    virtual void deallocate_watch(unsigned param_index, watch_type type,
                                  watch_handle_t watch_handle)=0;
    /**@brief Allocate a watch of the given type on the parameter
     * @param param_index  Add a watch to the variable associated with this
     *                     parameter.
     * @param type         The type of the watch to allocate
     * @return An ID of the watch. Every pair of watches that watch the same
     *         variable and type will have a different ID.
     *         ID -1U is never valid. See var_watchers_impl::allocate_watch().
     */
    virtual watch_handle_t allocate_watch(unsigned param_index, 
                                          watch_type type)=0;
    virtual void deallocate_value_watch(unsigned param_index,
                                        iset::value_type value,
                                        watch_handle_t watch_handle)= 0;
    /**@brief Allocate a watch of the given type on the parameter
     * @param param_index  Add a watch to the variable associated with this
     *                     parameter.
     * @param value        The value of the watch to allocate
     * @return An ID of the watch. Every pair of watches that watch the same
     *         variable and value will have a different ID.
     *         ID -1U is never valid. See var_watchers_impl::allocate_watch().
     */
    virtual watch_handle_t allocate_value_watch(unsigned param_index,
                                                iset::value_type value)=0;
    /**
     * @brief If implemented, return the latest modified parameter affecting
     *        this propagation, other than the \p except parameter.
     *
     * @param except  The parameter to ignore. Consider all other parameters
     *                when looking for the latest modification.
     *                If no parameter is modified except for this one, then
     *                return -1.
     * @return latest modified input parameter, if a valid answer is known.
     *                If no applicable answer known or method is unimplemented
     *                then -1 is returned. 
     */
    virtual int latest_modified_param(unsigned except) const
    { (void) except; return -1; }

    /**
     * @brief How much effort the propagator should invest in its operations.
     *
     * The returned value controls the trade-off between accuracy and speed of
     * inaccurate propagators. Affected operation nay include propagation and
     * explanation generation, according to the way the propagator decides to
     * interpret the value.
     *
     * @return 0 for minimal effort, the higher the value the more accurate
     *         and slower the propagator may decide to get.
     */
    virtual unsigned effort() const { return 0; }
    /**
     * @brief Do we only query the constraint as-if it is propagated.
     *
     * For normal propagation false is returned. When we call
     * constraint::propagate() only in order to find what would happen if it
     * is propagated, without actually affecting domains, then this method
     * should return true. When querying_domain()==true subsequent runs
     * may "resurrect" values which constraint::propagate() has removed from a
     * domain.
     *
     * If the constraint holds no state and does not define special watches then
     * it should not care if querying_domain() is true or false. However if it
     * sets watches it may need to use a different policy in the different run
     * modes.
     *
     * ASSUMPTIONS regarding to changes in querying_domain() over time:
     *    - The value of querying_domain() is set before the first propagation
     *      of the constraint either to true or false.
     *    - During successive propagation it may change from true to false,
     *      but not back to 'true'
     *    - Right after backtrack any change is allowed.
     */
    virtual bool querying_domain() const { return false; }
    virtual void print(std::ostream & ) const throw();
    //! This will print the header of the possibly with constraint name and
    //! index
    virtual void print_prefix(std::ostream & ) const throw() {}
    //! When looking at one domain's size, in an approximated propagator,
    //! propagate approximately when above this limit.
    virtual unsigned accurate_size_bound() const;
    //! When looking at a product of two domains' size, in an approximated
    //! propagator, propagate approximately when above this limit.
    virtual unsigned accurate_size_pair_bound() const;
    //! After an approximated propagator performed an accurate propagation, pass
    //! it on only if its size is below this limit.
    virtual unsigned max_accurate_ranges() const;

    // FIXME: this should go to a new cpp file, with a reasonable default
    //! Print the identifying information on this parameter (e.g. name)
    virtual void print_param(std::ostream &, unsigned ) const throw() {}
  private:
    //! For debugging only, to be called from within the debugger
    void show_self() const;
};
inline std::ostream & operator<<(std::ostream &out,
                                 const constraint_runtime_info & info)
{
    info.print(out);
    return out;
}

} // namespace hcsp
#endif // HCSP_CONSTRAINT_RUNTIME_INFO_HPP_
