// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2013,2014, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2013,2014, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_CONSTRAINT_WATCHLESS_INFO_HPP_
#define HCSP_CONSTRAINT_WATCHLESS_INFO_HPP_

#include <vector>
#include <watch_types.hpp>
#include <solver_types.hpp>
#include <iset.hpp>
#include <constraint_runtime_info.hpp>
namespace hcsp {

/**
 * @brief A wrapper for constraint_runtime_info which records no watches
 *
 * This class is useful when we want to run a constraint only once and not
 * bother with its watches. It passes all calls, except for watch management,
 * to the wrapped constraint_runtime_info object.
 */
class constraint_watchless_info final : public constraint_runtime_info,
                                        private var_watchers_base
{
  public:
    /**
     * @brief Constructor
     *
     * @param orig  The original constraint_connectivity_info object to be
     *              wrapped.
     */
    constraint_watchless_info(constraint_runtime_info & orig)
        : m_info(orig), m_is_faked_queue(false)
    {}

    virtual bool reduced_between_activations(unsigned param_index) const
    { return m_info.reduced_between_activations(param_index); }

    virtual queue_type originating_queue() const
    { 
        if (m_is_faked_queue)
            return m_faked_queue;
        return m_info.originating_queue(); 
    }
    
    virtual void put_in_queue(queue_type queue)
    {
        m_is_faked_queue = false;
        m_info.put_in_queue(queue); 
    }

    //! Make it look as if the constraint is in a different queue, although
    //! the original 'info' variable is not affected.
    void fake_put_in_queue(queue_type queue)
    {
        m_is_faked_queue = true;
        m_faked_queue = queue;
    }

    virtual const iset & domain(unsigned param_index) const
    { return m_info.domain(param_index); }

    virtual void  set_domain(unsigned param_index, iset new_domain)
    { return m_info.set_domain(param_index, std::move(new_domain)); }

    virtual unsigned     arity() const
    { return m_info.arity(); }
    
    virtual void deallocate_watch(unsigned , watch_type, 
                                  watch_handle_t handle) override
    { m_watches.erase(handle.get_it()); }
    virtual watch_handle_t allocate_watch(unsigned, watch_type) override
    { 
        m_watches.emplace_back(watcher_info_type{0});
        return --m_watches.end();
    }
    virtual void deallocate_value_watch(unsigned, iset::value_type, 
                                        watch_handle_t watch_handle) override
    { m_watches.erase(watch_handle.get_it()); }
    virtual watch_handle_t allocate_value_watch(unsigned,
                                                iset::value_type) override
    { return allocate_watch(0, watch_type(0)); }
    virtual int latest_modified_param(unsigned except) const
    { return m_info.latest_modified_param(except); }
    virtual unsigned effort() const
    { return m_info.effort(); }
    virtual void print(std::ostream & out ) const throw()
    { m_info.print(out); }
    virtual void print_prefix(std::ostream &out ) const throw()
    { m_info.print_prefix(out); }
    virtual void print_param(std::ostream &out, unsigned param ) const throw()
    { m_info.print_param(out, param); }
  private:
    constraint_runtime_info & m_info;
    watch_list_t m_watches;
    //! If true then the queue information is read locally from m_faked_queue
    bool m_is_faked_queue;

    //! A locally modified value of the current queue.
    queue_type m_faked_queue;
}; // class constraint_watchless_info 

} // namespace hcsp
#endif // HCSP_CONSTRAINT_WATCHLESS_INFO_HPP_
