// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2013, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2013, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_CONTROLLED_PTR_HPP_
#define HCSP_CONTROLLED_PTR_HPP_
#include <cstddef>
#include <memory>
namespace hcsp {
/**
 * @brief Like unique_ptr<T> but with the ability to control the destruction.
 *
 * Unlike unique_ptr<T> it is possible to have a pointer that upon destruction
 * will be either deleted or retained.
 */
template<typename T>
class controlled_ptr
{
  public:
    typedef       T* pointer;
    typedef const T* const_pointer;
    typedef const T& const_reference;
    typedef       T  element_type;
    controlled_ptr() : m_ptr(0), m_auto_free(0) {}
    controlled_ptr(pointer ptr) : m_ptr(ptr), m_auto_free(ptr)  {}
    controlled_ptr(controlled_ptr && other) noexcept
        : m_ptr(other.m_ptr), m_auto_free(other.m_auto_free)
    {
        other.m_auto_free= 0;
    }
    controlled_ptr & operator=(controlled_ptr && other) noexcept
    {
        if (&other == this)
            return *this;
        if (other.m_ptr != m_ptr)
            delete m_auto_free;
        
        m_auto_free= other.m_auto_free;
        other.m_auto_free= 0;
        m_ptr= other.m_ptr;
        return *this;
    }
    controlled_ptr & operator=(std::unique_ptr<T> other)
    {
        delete m_auto_free;
        m_ptr= m_auto_free= other.release();
        return *this;
    }
    ~controlled_ptr()
    {
        delete m_auto_free;
        m_ptr= m_auto_free= 0;
    }
    //! Make this refer to the given pointer, but do not make it own it.
    //! If the pointer is owned already then do nothing.
    void refer_to(const_reference ref)
    {
        if (m_auto_free == &ref)
            return;
        delete m_auto_free;
        m_auto_free= 0;
        m_ptr = &ref;
    }
    const_pointer operator->() const { return m_ptr; }
    const_pointer get() const { return m_ptr; }
    pointer get_mutable()  { return m_auto_free; }
    bool is_mutable() const { return m_auto_free != NULL; }
    const_reference operator*() const { return *m_ptr; }
    pointer release()
    {
        pointer saved= m_auto_free;
        m_auto_free= NULL;
        return saved;
    }
    pointer release_or_clone()
    {
        if (m_auto_free) 
            return release();
        else if (m_ptr)
            return m_ptr->clone();
        else
            return NULL;
    }
    void reset(pointer ptr= 0)
    {
        if (ptr != m_auto_free)
            delete m_auto_free;
        m_ptr= m_auto_free= ptr;
    }
  private:
    const_pointer m_ptr;
    pointer m_auto_free;
}; // class controlled_ptr<T>

} // namespace hcsp

#endif // HCSP_CONTROLLED_PTR_HPP_
