// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2013 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2013 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_DISJOINT_PROPAGATION_ALGO_HPP_
#define HCSP_DISJOINT_PROPAGATION_ALGO_HPP_

#include <vector>
#include <iosfwd>
#include <propagation_algo.hpp>
#include <constraint_runtime_info.hpp>
#include <sliced_runtime_info.hpp>

namespace hcsp {
/**
 * @brief An internal class
 *
 * This is a base-class of disjoint_propagation_algo<Algo1,Algo2> class
 * and is used in its implementation.
 */
class disjoint_propagation_algo_base : protected propagation_types
{
  public:
    class print_buffer : private no_copy
    {
      public:
        print_buffer();
        ~print_buffer();

        //! Get the stream buffer used for printing algo_a
        std::ostream & stream_a();
        
        //! Get the stream buffer used for printing algo_b
        std::ostream & stream_b();

        //! Flush all buffered output to the output stream.
        void flush_to(std::ostream & out);
        
      private:
        std::unique_ptr<std::ostringstream> m_stream_a;
        std::unique_ptr<std::ostringstream> m_stream_b;
    }; // class disjoint_propagation_algo_base::print_buffer

}; // class disjoint_propagation_algo_base

/**
 * @brief A constraint propagation algorithm resulting from a disjunction of
 *        two propagation algorithms.
 *
 * We have two constraints A(x1,x2,...,x_n) and B(y1,y2,...,y_n) which can be
 * propagated by the two propagation algorithms Algo_A and Algo_B. This is a
 * constraint propagation algorithm for the constraint:
 *  - A(x1,x2,...,x_n) || B(y1,y2,...,y_n).
 *
 * @tparam Algo_A, Algo_B  The two propagation algorithms from which this class
 *                         makes a disjunction.
 */
template <typename Algo_A, typename Algo_B>
class disjoint_propagation_algo  : private disjoint_propagation_algo_base
{
  private:
    typedef disjoint_propagation_algo<Algo_A, Algo_B> self_type;

    //! The common part of all non-copy constructors
    void construct()
    {
        (void)&self_type::test_concept;
        m_a_count_params= 0;
        int a_count= m_algo_a.count_params();
        if (a_count > 0)
            m_a_count_params= a_count;
    }
  public:
    disjoint_propagation_algo() { construct(); }
    
    template<typename T1>
    disjoint_propagation_algo(const T1 &v1) : m_algo_a(v1) { construct(); }

    template<typename T1, typename T2>
    disjoint_propagation_algo(const T1 &v1, const T2 &v2) : m_algo_a(v1, v2)
    { construct(); }
    
    template<typename T1, typename T2, typename T3>
    disjoint_propagation_algo(const T1 &v1, const T2 &v2, const T3 &v3)
        : m_algo_a(v1, v2, v3)
    { construct(); }

    //! Count the total number of parameters.
    //! \return The sum, for both algo objects, of the counts of parameters.
    //!         Return a negative number (-1) if number of parameters is not
    //!         know for any of the algo objects 
    int count_params() const
    {
        if (m_algo_a.count_params() < 0 || m_algo_b.count_params() < 0)
            return -1;
        unsigned sum = 0U+ m_algo_a.count_params() + m_algo_b.count_params();
        return sum;
    }
    //! An accessor to the Algo_A object.
    const Algo_A & algo_a() const { return m_algo_a; }
    //! An accessor to the Algo_A object. 
    Algo_A       & algo_a()       { return m_algo_a; }
    //! An accessor to the Algo_B object.
    const Algo_B & algo_b() const { return m_algo_b; }
    //! An accessor to the Algo_B object.
    Algo_B       & algo_b()       { return m_algo_b; }
    bool init(constraint_runtime_info & info);
    result_type propagate(constraint_runtime_info & info);
    bool conflicts(const constraint_runtime_info & info) const;
    bool constraint_check(const constraint_runtime_info & info) const;
    void print(std::ostream & out,
               const constraint_runtime_info * p_info) const throw();
    unsigned count_watches() const
    { return (m_algo_a.count_watches() + m_algo_b.count_watches()); }
    unsigned watched_param_index(unsigned watch_index) const;
    void forget_watches()
    {
        m_algo_a.forget_watches();
        m_algo_b.forget_watches();
    }
    void explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation) const;
    
  private:
    /**
     * @brief Get the number of parameters that will be passed to the Algo_A
     *        object.
     *
     * @param info  The context of the constraint. Sometimes the count is
     *              implied from the calling context of the propagator.
     *
     * @return The number of parameters to be passed to Algo_A. If the count
     *         can't be inferred then a std::runtime_error exception is thrown.
     */
    unsigned a_count_params(const constraint_runtime_info &info) const
    {
        int a_count= m_algo_a.count_params();
        if (a_count >= 0)
            return a_count;
        int b_count= m_algo_b.count_params();
        if (b_count < 0)
            throw_runtime_error("Can't operate on disjoined propagation "
                                "algorithms without knowing their parameter "
                                "counts");
        a_count= info.arity()-b_count;
        if (a_count < 0)
            throw_runtime_error("The second propagation algorithm reports more"
                                " arguments than the total");

        return a_count;
    }

    //! (compile time) Make sure that all propagation_algo objects adhere to
    //! the propagation_algo_concept.
    static void test_concept()
    {
        (void)&propagation_algo_concept<Algo_A>::test_concept;
        (void)&propagation_algo_concept<Algo_B>::test_concept;
        (void)&propagation_algo_concept<self_type>::test_concept;
    }
    Algo_A m_algo_a;
    Algo_B m_algo_b;
    //! Number of params in the first algo object, or zero if unknown.
    unsigned m_a_count_params;
};

template <typename Algo_A, typename Algo_B>
bool disjoint_propagation_algo<Algo_A, Algo_B>::init(
    constraint_runtime_info & info)
{
    m_a_count_params= a_count_params(info);
    sliced_runtime_info a_info(info, 0, m_a_count_params);
    sliced_runtime_info b_info(info, m_a_count_params, info.arity());
    bool a_ok= m_algo_a.init(a_info);
    bool b_ok= m_algo_b.init(b_info);
    return a_ok || b_ok;
}

template <typename Algo_A, typename Algo_B>
bool disjoint_propagation_algo<Algo_A, Algo_B>::constraint_check(
    const constraint_runtime_info & info) const
{
    unsigned a_count= a_count_params(info);
    sliced_runtime_info a_info(info, 0U, a_count);
    sliced_runtime_info b_info(info, a_count, info.arity());
    return m_algo_a.constraint_check(a_info)||m_algo_b.constraint_check(b_info);
}

template <typename Algo_A, typename Algo_B>
propagation_types::result_type
disjoint_propagation_algo<Algo_A, Algo_B>::propagate(
    constraint_runtime_info & info)
{
    HCSP_ASSERT(a_count_params(info) == m_a_count_params);
    sliced_runtime_info a_info(info, 0, m_a_count_params);
    sliced_runtime_info b_info(info, m_a_count_params, info.arity());
    
    result_type a_result= m_algo_a.propagate(a_info);
    if (a_result == UNSAT) {
        b_info.set_querying_domain(false);
        return m_algo_b.propagate(b_info);
    }
    if (a_result == VALID && !a_info.pending_commit())
        return VALID;
    if (!a_info.pending_commit())
        // No need to check m_algo_b, not even update its watches, since
        // even if m_algo_b will UNSAT this will not affect the outside world.
        // As for watches, even if m_algo_b would to become UNSAT it should not
        // affect anything until a m_algo_a has something to propagate. And
        // if m_algo_a will have anything to propagate, it will be activated
        // by its own watches
        return CONSISTENT;
    result_type b_result= m_algo_b.propagate(b_info);
    if (b_result == UNSAT) {
        a_info.commit();
        return a_result;
    }
    if (b_result == VALID && !b_info.pending_commit())
        return VALID;
    return CONSISTENT;
}

template <typename Algo_A, typename Algo_B>
bool disjoint_propagation_algo<Algo_A, Algo_B>::conflicts(
    const constraint_runtime_info & info) const
{
    HCSP_ASSERT(a_count_params(info) == m_a_count_params);
    sliced_runtime_info a_info(info, 0, m_a_count_params);
    if (!m_algo_a.conflicts(a_info))
        return false;
    sliced_runtime_info b_info(info, m_a_count_params, info.arity());
    
    return m_algo_b.conflicts(b_info);
}

template <typename Algo_A, typename Algo_B>
void disjoint_propagation_algo<Algo_A, Algo_B>::print(
    std::ostream & out,
    const constraint_runtime_info * p_info) const throw()
{
    print_buffer buffer;
    std::unique_ptr<constraint_runtime_info> a_info, b_info;
    if (p_info) {
        p_info->print_prefix(out);
        unsigned a_count= a_count_params(*p_info);
        unsigned total_count= p_info->arity();
        a_info.reset(new sliced_runtime_info(*p_info, 0, a_count));
        b_info.reset(new sliced_runtime_info(*p_info, a_count, total_count));
    }
    m_algo_a.print(buffer.stream_a(), a_info.get());
    m_algo_b.print(buffer.stream_b(), b_info.get());
    buffer.flush_to(out);
}

template <typename Algo_A, typename Algo_B>
unsigned disjoint_propagation_algo<Algo_A, Algo_B>::watched_param_index(
    unsigned watch_index) const
{
    unsigned a_watches= m_algo_a.count_watches();
    if (watch_index < a_watches)
        return m_algo_a.watched_param_index(watch_index);
    else {
        unsigned b_watch_index= watch_index-a_watches;
        unsigned b_param_index= m_algo_b.watched_param_index(b_watch_index);
        return b_param_index+m_a_count_params;
    }
}
        
template <typename Algo_A, typename Algo_B>
void disjoint_propagation_algo<Algo_A, Algo_B>::explain_propagation(
    unsigned propagated_param, const iset & values_to_explain,
    const iset & propagated_domain, 
    const constraint_runtime_info & info_before_propagation,
    std::vector<iset> & explanation) const
{
    explanation.reserve(info_before_propagation.arity());

    unsigned a_count= a_count_params(info_before_propagation);
    sliced_runtime_info a_info(info_before_propagation, 0, a_count);
    sliced_runtime_info b_info(info_before_propagation,
                               a_count, info_before_propagation.arity());
    
    std::vector<iset> & a_explanation= explanation;
    std::vector<iset> b_explanation;
    if (propagated_param < a_count && !propagated_domain.empty()) {
        m_algo_a.explain_propagation(propagated_param, values_to_explain,
                                     propagated_domain, a_info, a_explanation);
    } else {
        // Need to explain why m_algo_a was falsified 
        m_algo_a.explain_propagation(0, iset().complement(),
                                     iset(), a_info, a_explanation);
    }
    if (propagated_param >= a_count && !propagated_domain.empty()) {
        m_algo_b.explain_propagation(propagated_param-a_count,
                                     values_to_explain,
                                     propagated_domain, b_info, b_explanation);
    } else {
        // Need to explain why m_algo_b was falsified 
        m_algo_b.explain_propagation(0,
                                     iset().complement(),
                                     iset(), b_info, b_explanation);
    }
    HCSP_ASSERT(&explanation == &a_explanation);
    explanation.insert(explanation.end(),
                       b_explanation.begin(), b_explanation.end());
}

} // namespace hcsp
#endif // HCSP_DISJOINT_PROPAGATION_ALGO_HPP_
