// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_DISJOINT_PROPAGATORS_HPP_
#define HCSP_DISJOINT_PROPAGATORS_HPP_

#include <disjoint_propagation_algo.hpp>
#include <clause_propagator.hpp>
#include <constraint_propagator.hpp>

namespace hcsp {

/**
 * @brief A propagator which is a disjunction of two other constraints.
 *
 * The two inner constraints are set during template instantiation, i.e.,
 * during compilation. This limitation is supposed to make it faster than the
 * more general constraint of multi_disjoint_propagator.
 */
template <typename Algo1, typename Algo2>
class disjoint_propagator : public constraint_propagator_from_algo<
    disjoint_propagation_algo<Algo1, Algo2> >
{
  private:
    typedef constraint_propagator_from_algo<
      disjoint_propagation_algo<Algo1, Algo2> > base_type;
  public:
    disjoint_propagator() {}
    
    template <typename A1>
    disjoint_propagator(const A1 &p1) : base_type(p1) {}
    
    template <typename A1, typename A2>
    disjoint_propagator(const A1 &p1, const A2 &p2) : base_type(p1, p2) {}
    
    template <typename A1, typename A2, typename A3>
    disjoint_propagator(const A1 &p1, const A2 &p2, const A3 &p3)
        : base_type(p1, p2, p3) {}
}; // class disjoint_propagator<Algo1,Algo2>

/**
 * @brief A propagator which is a disjunction of several other constraints
 *        and literals.
 *
 * The set of literals and inner constraints are set slightly after
 * construction, before the first propagation.
 *
 * @todo The handling of watches is inefficient, we should put the watches in a
 *       sandbox and connect only the watches of two propagators to the
 *       solver. This way, when a variable of a propagator which is not one of
 *       m_active_propagators is modified the multi_disjoint_propagator will not
 *       be activated by it.
 */
class multi_disjoint_propagator : public constraint_propagator
{
  public:
    multi_disjoint_propagator()
    {
        m_active_propagators[0]= m_active_propagators[1]= -1U;
        m_first_propagator_argument.push_back(0);
    }
    multi_disjoint_propagator(const multi_disjoint_propagator & other)
    { *this= other; }
    multi_disjoint_propagator& operator=(const multi_disjoint_propagator&other);
    std::vector<iset> & update_literals() { return m_clause.update_literals(); }
    const std::vector<iset> & literals() const { return m_clause.literals(); }
    /**
     * @brief Add a constraint to the disjunction.
     *
     * The added constraint must not share variables with any other propagator
     * or literal in the disjunction. This limitation is not substantial, but
     * the current implementation assumes this.
     *
     * @param[in] propagator   A pointer to the propagator which to be added to
     *                         the disjoint constraint propagator. 
     * @param[in] arity        The arity of the newly added propagator.
     * @return The index of the constraint among other disjoint constraints in
     *         this propagator.
     */
    unsigned append_constraint(
        std::unique_ptr<constraint_propagator> propagator, unsigned arity);
    //! Get a mutable reference of the constraint according to its position.
    constraint_propagator & disjoint_constraint(unsigned constraint_index)
    { return *m_propagators.at(constraint_index); }
    //! Get a const reference of the constraint according to its position.
    const constraint_propagator & disjoint_constraint(unsigned constraint_index)
        const
    { return *m_propagators.at(constraint_index); }
        
    virtual int count_params() const
    { return m_first_propagator_argument.back()+m_clause.count_params(); }
    virtual bool init(constraint_runtime_info & info);
    virtual result_type propagate(constraint_runtime_info & info);
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual bool constraint_check(const constraint_runtime_info & info) const;
    virtual void print(std::ostream &out,
                       const constraint_runtime_info *p_info=NULL)const throw();
    virtual unsigned watched_param_index(unsigned watch_index) const;
    virtual unsigned count_watches() const
    {
        unsigned count= m_clause.count_watches();
        for (unsigned i=0 ; i != m_propagators.size() ; ++i)
            count += m_propagators[i]->count_watches();
        return count;
    }
    virtual void forget_watches()
    {
        m_clause.forget_watches();
        for (unsigned i=0 ; i != m_propagators.size() ; ++i)
            m_propagators[i]->forget_watches();
    }
        
    void explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation) const;
    
    virtual void possibly_crippling_values(
        unsigned propagated_param, const iset & propagated_domain,
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & values) const;    
    virtual multi_disjoint_propagator * clone() const
    { return new multi_disjoint_propagator(*this); }
    virtual constraint_propagator * intersect_with(
        const constraint_runtime_info & info,
        std::vector<unsigned> & copy_from_argument) const;
    
  private:
    /** 
     * @brief Advance m_active_propagators[i].
     *
     * The result should neither overlap orig_active_propagators[2] nor
     * the other entry of m_active_propagators[]. Other than that, the
     * propagator is neither propagated or consulted. The algorithm stops when
     * orig_active_propagators[i] == m_active_propagators[i], which means that
     * the we have come full circle.
     *
     * @param activity_index
     *        Tells to advance m_active_propagators[activity_index]
     * @param orig_active_propagators
     *        Tells advance_active_propagators() to skip these indices
     * @return True iff could update m_active_propagators[activity_index]
     *        according to the requirements.
     *
     */
    bool advance_active_propagators(unsigned activity_index,
                                    const unsigned orig_active_propagators[2]);
    result_type propagate_for_active_propagators_2nd(
        constraint_runtime_info & info,
        const unsigned orig_active_propagators[2],
        result_type caller_result);

    /**
     * @brief propagate the nth disjunct.
     *
     * Simply activates constraint_propagator::propagate(info) for the requested
     * disjunct. If disjunct_index < m_propagators.size() then a propagator is
     * propagated, otherwise the literals are propagated.
     */
    result_type propagate_nth(unsigned disjunct_index,
                              constraint_runtime_info & info);
    /**
     * @brief Check if the nth disjunct conflicts.
     *
     * Simply activates constraint_propagator::conflicts(info) for the requested
     * disjunct. If disjunct_index < m_propagators.size() then a propagator is
     * queried, otherwise the literals are queried.
     */
    bool conflicts_nth(const constraint_runtime_info & info,
                       unsigned propagator_index) const;
    void explain_nth_propagation(
        unsigned disjunct_index, unsigned propagated_param,
        const iset & values_to_explain, const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & disjunct_explanation) const;
    void possibly_crippling_values_nth(
        unsigned disjunct_index, unsigned propagated_param,
        const iset & propagated_domain,
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & values) const;
    /**
     * @brief Intersect this propagator with a clause.
     *
     * Intersect the propagator with a clause, returning the result in
     * 'result'. If can't produce a better constrain that 'this', even if it is
     * theoretically possible, then 'result' will be empty, and updated= false.
     *
     * @param[in] info  The domains that represent the literals of the clause
     *                  to intersect with.
     * @param[out] copy_from_argument
     *                  The mapping of the arguments of 'result' to the
     *                  arguments of 'this'.
     * @param[out] updated
     *                  False if 'result' should be viewed as if it is a copy
     *                  of 'this', even if it is not. True if it is up to date.
     * @param[out] result
     *                  The resulting multi_disjoint_propagator.
     */
    void intersect_with_clause(
        const constraint_runtime_info & info,
        std::vector<unsigned> & copy_from_argument, bool & updated,
        multi_disjoint_propagator & result) const;
    /**
     * @brief Intersect the n-th disjunct with the relevant sub-clause.
     *
     * Intersect the disjunct with a clause, appending the result to
     * 'result'. If updated==true then the disjunct of 'result' is updated
     * even if it is equal to the disjunct of 'this'. If updated=false then the
     * disjunct, and all proceeding disjuncts, are added if the resulting
     * disjunct is stronger than that of 'this'
     *
     * @param[in] disjunct_index
     *            The index of the disjunct, which can refer to the literals
     *            as a whole, but not to a single literal. 
     * @param[in] info
     *            The domains that represent the literals of the clause to
     *            intersect with. 
     * @param[in,out] copy_from_argument
     *            The mapping of the arguments of 'result' to the arguments of
     *            'this'. If disjuncts are added to 'result' then this vector is
     *            updated accordingly. 
     * @param[in,out] updated
     *            False if 'result' should be viewed as if it is a copy of
     *            'this', even if it is not. True if it is up to date. 
     * @param[in,out] result
     *             The resulting multi_disjoint_propagator.
     */
    void intersect_with_nth(
        unsigned disjunct_index, const constraint_runtime_info & info,
        std::vector<unsigned> & copy_from_argument, bool & updated,
        multi_disjoint_propagator & result) const;
    /**
     * @brief Get the number of variables that this disjunct affects.
     *
     * @param[in] disjunct_index
     *            If disjunct_index<m_propagators.size() then this is index of
     *            the propagator, otherwise it refers to the literals.
     * @return If disjunct_index<m_propagators.size() return the arity of the
     *            propagator otherwise return the number of literals.
     */
    unsigned disjunct_arity(unsigned disjunct_index) const
    {
        return (end_disjunct_argument(disjunct_index)
                -
                begin_disjunct_argument(disjunct_index));
    }
    void copy_disjuncts_until(unsigned disjunct_index,
                              std::vector<unsigned> & copy_from_argument,
                              multi_disjoint_propagator & result) const;
    
    unsigned begin_disjunct_argument(unsigned propagator_idx) const
    {
        return m_first_propagator_argument[propagator_idx];
    }
    unsigned end_disjunct_argument(unsigned propagator_idx) const
    {
        if (propagator_idx < m_propagators.size())
            return m_first_propagator_argument[propagator_idx+1];
        else
            return count_params();
    }
    
    
    /**
     * @brief Zero or more literals of this constraint.
     *
     * The i-th literal is argument number m_first_propagator_argument.back()+i
     * of the multi_disjoint_propagator.
     */
    clause_propagation_algo m_clause;
    /**
     * @brief The constraint propagators that participate in the disjunction.
     * The propagators are owned by the multi_disjoint_propagator and it will
     * delete them in the end.
     */
    std::vector<std::unique_ptr<constraint_propagator> > m_propagators;
    /** @brief For the i-th propagator its first argument is mapped to the
     *         m_first_propagator_argument[i] of the multi_disjoint_propagator
     *
     * @invariant m_first_propagator_argument[0] == 0;
     *            m_first_propagator_argument.size()== m_propagators.size()+1;
     *            m_first_propagator_argument.back()+m_clause.count_params()==
     *            count_params();
     */
    std::vector<unsigned> m_first_propagator_argument;
    //! The propagators which need to be checked first, or m_clause if
    //! the value == m_propagators.size();
    unsigned m_active_propagators[2];
}; // class multi_disjoint_propagator

} // namespace hcsp
#endif // HCSP_DISJOINT_PROPAGATORS_HPP_
