// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#ifndef HCSP_DISJUNCTION_SPECIFICATION_HPP_
#define HCSP_DISJUNCTION_SPECIFICATION_HPP_


#include <vector>

#include <iset.hpp>
#include <constraint_propagator.hpp>
#include <controlled_ptr.hpp>

namespace hcsp {
/**
 * @brief A class to hold data from similar propagator and their connectivity.
 *
 * Currently this class supports literal_or_difference_ge_propagator and
 * difference_ge_propagator, but it can be easily extended to other
 * propagators.
 *
 * The constraint has a literal part, where there is zero or more literals
 * of the form var \in Set, and a non-literal part. The non literal part,
 * currently, is of the form left_var-right_var >= m_k.
 *
 */
class disjunction_specification
{
  public:
    disjunction_specification(
        const constraint_runtime_info & info,
        const constraint_propagator & propagator,
        const std::vector<unsigned> & local_to_external_param);
    disjunction_specification();
    
    void set_literals(const iset literals[],
                      unsigned first_literal_index, unsigned in_num_literals)
    {
        m_first_literal_ptr= literals;
        m_first_literal_index= first_literal_index;
        m_num_literals= in_num_literals;
    }
    void add_constraint_var(unsigned var)
    {
        m_local_to_external_param.push_back(var);
        ++m_constraint_arity;
    }
    void add_literal(const iset & lit, unsigned variable)
    {
        HCSP_ASSERT(constraint_arity() == 0);
        m_local_to_external_param.push_back(variable);
        m_literal_storage.push_back(lit);
        m_first_literal_ptr= &m_literal_storage[0];
        ++m_num_literals;
    }

    //! @todo get rid of the first_literal_index argument
    void reset_constraint(constraint_propagator* in_constraint,
                          unsigned first_constraint_index)
    {
        m_constraint.reset(in_constraint);
        m_constraint_arity= 0;
        m_first_constraint_index= first_constraint_index;
    }
    void refer_to_constraint(const constraint_propagator & in_constraint,
                             unsigned first_constraint_index)
    {
        m_constraint.refer_to(in_constraint);
        m_first_constraint_index= first_constraint_index;
    }
    
    /**
     * @brief Resolve the literals of the two objects into 'this'
     *
     * Resolve the literals of 'a' and 'b' with the given pivot. Put the result
     * into 'this'. If both 'a' and 'b' do not have the pivot then the literals
     * are simply merged. If only one of 'a' or 'b' contain the pivot then the
     * result is unspecified.
     *
     * @param[in] a,b   The two disjunctions which we want to resolve
     * @param[in] pivot The pivot of the resolution.
     *
     * @post The result of the resolution is in 'this'
     */
    void resolve_literals(
        const disjunction_specification &a, const disjunction_specification &b,
        unsigned pivot);

    constraint_propagator *combine(
        disjunction_specification & other,
        unsigned pivot, std::vector<unsigned> & copy_from_argument);
    
    constraint_propagator *combine_literal_pivot(
        const disjunction_specification & other,
        unsigned pivot, std::vector<unsigned> & copy_from_argument) const;

    constraint_propagator *combine_constraint_pivot(
        disjunction_specification &other,
        unsigned pivot, std::vector<unsigned> & copy_from_argument);
    
    /**
     * @brief The set that defines the literal whose index is literal_num
     *
     * @param literal_num  Is the number of the literal to query (start with 0).
     *                     This value counts the literals and ignore the part of
     *                     the non-clausal constraint.
     *                     
     * @return The set that restricts the variable, i.e., the set that defines
     *         the literal.
     * @pre literal_num < num_literals();
     */
    const iset & literal_set(unsigned literal_num) const
    {
        HCSP_ASSERT(literal_num < num_literals());
        return *(m_first_literal_ptr+literal_num);
    }

    /**
     * @brief The variable that defines the literal whose index is literal_num
     *
     * @param literal_num  Is the number of the literal to query (start with 0).
     *                     This value counts the literals and ignore the part of
     *                     the non-clausal constraint.
     *                     
     * @return The variable that is restricted by the literal.
     * @pre literal_num < num_literals();
     */
    unsigned literal_var(unsigned literal_num) const
    {
        unsigned in_param= m_first_literal_index+literal_num;
        if (in_param < m_local_to_external_param.size())
            return m_local_to_external_param[in_param];
        else
            return in_param;
    }

    unsigned literal_index_of_var(unsigned var) const
    {
        for (unsigned i=0 ; i != num_literals() ; ++i)
            if (literal_var(i) == var)
                return i;
        return -1U;
    }
    unsigned constraint_index_of_var(unsigned var) const
    {
        for (unsigned i=0 ; i != constraint_arity() ; ++i)
            if (constraint_var(i) == var)
                return i;
        return -1U;
    }
    /**
     * @brief The domain of the given literal part of the disjunction
     *
     * Note that the disjunction may be comprised of a disjunction between
     * literals and an inner constraint. It may also contain only an inner
     * constraint with no literals, or only literals with no inner constraint.
     *
     * @param literal_num  Is the number of the literal to query (start with 0).
     *                     This value counts the literals and ignore the part of
     *                     the non-clausal constraint.
     *                     
     * @return The variable that is restricted by the literal.
     * @pre literal_num < num_literals();
     */
    const iset & literal_domain(unsigned literal_num) const
    { return m_info->domain(m_first_literal_index+literal_num); }

    //! The number of literals in this constraint
    unsigned num_literals() const { return m_num_literals; }
    
    const constraint_propagator *constraint() const
    { return m_constraint.get(); }

    constraint_propagator *release()
    {
        constraint_propagator * result= m_constraint.release_or_clone();
        return result;
    }
    //! The total number of parameters for this constraint
    unsigned arity() const { return m_constraint_arity+num_literals(); }

    //! Reset everything in the object, as if constructed with default ctor.
    void clear();

    //! Does the variable belong to the literal part of the constraint?
    //! 'variable' is not parameter index of this object, but rather the
    //! globally shared index used in m_local_to_external_param.
    bool is_literal(unsigned variable) const
    { return literal_index_of_var(variable) < num_literals(); }
    bool is_constraint_var(unsigned variable) const
    { return constraint_index_of_var(variable) < num_literals(); }
    //! Get the i-th variable of the non-literal part of the constraint
    unsigned constraint_var(unsigned in_constraint_index) const
    {
        HCSP_ASSERT(has_constraint());
        unsigned in_param= m_first_constraint_index+in_constraint_index;
        if (in_param < m_local_to_external_param.size())
            return m_local_to_external_param[in_param];
        else
            return in_param;
    }
    //! The first variable of the non-literal part of the constraint
    unsigned left_var() const  { return constraint_var(0); }
    //! The second variable of the non-literal part of the constraint
    unsigned right_var() const { return constraint_var(1); }
    bool has_constraint() const { return m_constraint_arity; }
    unsigned constraint_arity() const { return m_constraint_arity; }
    void set_constraint_arity(unsigned in_arity)
    { m_constraint_arity= in_arity; }

    //! The domain of the i-th parameter of this constraint.
    const iset & constraint_domain(unsigned i) const
    {
        HCSP_ASSERT(has_constraint());
        return m_info->domain(m_first_constraint_index+i);
    }
        
    //! The domain of the first variable of the non-literal part.
    const iset& left_domain() const  { return constraint_domain(0); }
    //! The domain of the second variable of the non-literal part.
    const iset& right_domain() const { return constraint_domain(1); }

    /**
     * @brief Create a clause out of this disjunction_specification
     *
     * @param[out] copy_from_argument
     *             Mapping from the arguments of the newly created clause to the
     *             arguments of the original constraints.
     * @return     The clause defined by 'this'
     * @pre The constraint part is empty.
     */
    constraint_propagator* new_clause(
        std::vector<unsigned>& copy_from_argument) const;

    /**
     * @brief Create a general constraint out of this disjunction_specification.
     *
     * Create a constraint that may or may not be a disjunction with literals,
     * but regardless there is a non-empty constraint part.
     *
     * @param[out] copy_from_argument
     *             Mapping from the arguments of the newly created constraint to
     *             the arguments of the original constraints.
     * @return     The constraint defined by 'this'
     * @pre The constraint part is not empty.
     */
    constraint_propagator *make_constraint(
        std::vector<unsigned> & copy_from_argument);

    //! Append in_constraint's literals to 'this'
    void append_literals(const disjunction_specification & in_literals);
    //! Append in_constraint's inner constraint to 'this'.
    void append_constraint(const disjunction_specification & in_constraint);

    //! Unlike append_constraint it calls with_constraint.release()
    void steal_constraint(disjunction_specification & in_constraint);
    
  private:
    const constraint_runtime_info * m_info;
    /**
     * @brief Mapping from this disjunction_specification's argument to global
     *        argument indices, a.k.a., variables.
     * 
     * For each m_local_to_external_param[i]==value, if argument i does not
     * exist then value == -1U. Otherwise value is this is the global index
     * of info and other_info (as defined in combine()).
     * If m_local_to_external_param.size() < arity() then for each
     * index >= m_local_to_external_param.size() we assume the external argument
     * to be == index.
     */
    std::vector<unsigned> m_local_to_external_param;
    //! The arity of the constraint part of the disjunction
    unsigned m_constraint_arity;
    //! The number of literals in the disjunction
    unsigned m_num_literals;
    unsigned m_first_literal_index;
    const iset * m_first_literal_ptr;
    std::vector<iset> m_literal_storage;
    unsigned m_first_constraint_index;
    controlled_ptr<constraint_propagator> m_constraint;
}; // class disjunction_specification
} // namespace hcsp

#endif // DISJUNCTION_SPECIFICATION
