// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_EQUIVALENCE_CLASSES_HPP_
#define HCSP_EQUIVALENCE_CLASSES_HPP_

#include <vector>
#include <utilities.hpp>

namespace hcsp {

/**
 * @brief Manage equivalence classes of integer indexes.
 *
 * This class maintains equivalence classes of non-sparse integers.  At each
 * step the user of this structure marks pair of integers as equivalent
 * uniting their equivalence classes. Then the user can ask for any given
 * integer to which equivalent class it belongs, to which the reply will be an
 * element of the equivalence class which represents all other elements in the
 * same equivalence class.
 *
 * The implementation is simply the union-find data structure whose
 * operations' complexity is the amortized reverse of Ackerman function, which
 * is O(1) for any practical purpose.
 */

class equivalence_classes
{
  public:
    /**
     * @brief Merge the equivalence classes of two elements.
     *
     * @param element1, element2
     *        The elements which should belong to equivalence one class
     * @post find(element1) == find(element2)
     */
    void unite(unsigned element1, unsigned element2);
    /**
     * @brief Find the equivalence class of one element.
     *
     * @param element   The element to look for.
     * @return  A member of the equivalence class of 'element' which represents
     *          all the members of this class.
     */
    unsigned find(unsigned element) const;
    /**
     * @brief Allocate the memory of the structure.
     *
     * This method has no functional effect, it affects only performance.
     *
     * @param size  The number of elements to reserve the memory for.
     */
    void reserve(unsigned size) { m_parents.reserve(size); }
  private:
    enum { NO_ELEMENT =  -1U };
    bool has_parent(unsigned element) const
    { return element<initialized_elements() && m_parents[element]!=NO_ELEMENT;}

    unsigned parent(unsigned element) const
    {
        HCSP_ASSERT(element < initialized_elements());
        return m_parents[element];
    }

    void parent(unsigned element, unsigned next_parent) const
    { m_parents[element]= next_parent; }

    unsigned initialized_elements() const { return m_parents.size(); }
    // The data is mutable because in union-find data structure, the find()
    // call compacts the representation along the searched path.
    mutable std::vector<unsigned> m_parents;
};

} // namespace hcsp

#endif // HCSP_EQUIVALENCE_CLASSES_HPP_
