// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_EXPLANATION_GENERATORS_HPP_
#define HCSP_EXPLANATION_GENERATORS_HPP_

#include <vector>
#include <watch_types.hpp>
#include <propagation_algo.hpp>
#include <constraint_runtime_info.hpp>
#include <iset.hpp>

namespace hcsp {
class constraint_propagator;
/**
 * @brief A placeholder for constraint explanation generation algorithms
 */
class explanation_generators : protected watch_types, public propagation_types 
{
  public:
    /**
     * @brief Explain a past domain reduction done by the propagator or
     *        propagation algo.
     *
     * Generate a clause which explains the queried reduction (see the
     * postcondition for more details).
     *
     * The default behavior is fast, but it generates a very poor clause.
     * In the default explanation, each literal uses the biggest possible
     * set of values, which make the clause propagate poorly.
     * For a better, but somewhat slower, behavior you can use
     * explanation_generators::explain_propagation_tightly().
     *
     * @param[in]  propagated_param
     *             The index of the reduced parameter which we need to explain.
     * @param[in]  values_to_explain
     *             The set of values which, if removed, the caller wants to
     *             explain.
     *             Only the values of \p values_to_explain have to be
     *             \e correct in the \p explanation.\n
     *             \n
     *             The caller of this method does not care if other values
     *             are added or omitted from the explanation.
     *             This is basically a refinement from propagated_param 
     *             to specific values. This freedom lets the explanation
     *             algorithm have potentially conflicting choices:
     *              - Fewer literals
     *              - Literals with smaller sets.
     *             These requirements may conflict and getting the best
     *             explanation becomes a matter of heuristics for the
     *             propagator or propagation algo.
     * @param[in]  propagated_domain
     *             The domain resulting from the propagation which we need
     *             to explain.
     * @param[in]  info_before_propagation
     *             The state of the solver before the propagation which we need
     *             to explain.
     * @param[out] explanation
     *             The literals of the explanation clause. The clause
     *             is \f$ ( v_0 \in \mbox{explanation}[0]\vee \cdots \vee
     *                     v_1 \in \mbox{explanation}[N-1]) \f$
     *
     * @pre  The domain reduction was performed by the propagator or the
     *       propagation algo for the given \b info_before_propagation
     * @post The resulting clause must:
     *     - Reduce the propagated parameter to \b propagated_domain
     *       for the same input domains.
     *     - For an arbitrary input domains, it should not remove domain values
     *       which are still supported by the inputs to the constraint.
     */
    static void explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation);
    
    /**
     * @brief Generate a tight explanation for a past domain reduction.
     *
     * Unlike explain_propagation(), this method generates the smallest
     * possible explanation for a given constraint propagation/conflict.
     * This means that removing any value or literal from the explanation
     * produces a clause that cannot serve as an explanation. This tightness
     * comes at the cost of performance.
     *
     * This method calls constraint_propagator::propagate() N times
     * (N is the arity of the constraint). This has a severe effect on overall
     * complexity, and should not be used when arity is big.
     * Instead, the implementor of the constraint is advised to implement
     * a specialized method for the propagator or the propagation_algo.
     *
     * This method should be called from a specific explain_propagation
     * implementation. The interface is exactly the same as that of
     * explain_propagation().
     *
     * @param[in]  propagator
     *             A copy of the original propagator or propagation algo.
     *             This must be a copy not the original propagator.
     * @param[in]  propagated_param
     *             The index of the reduced parameter which we need to explain.
     * @param[in]  values_to_explain
     *             The set of values which, if removed, the caller wants to
     *             explain.
     * @param[in]  propagated_domain
     *             The domain resulting from the propagation which we need
     *             to explain.
     * @param[in]  info_before_propagation
     *             The state of the solver before the propagation which we need
     *             to explain.
     * @param[out] explanation
     *             The literals of the explanation clause.
     *
     * @pre  The domain reduction was performed by this propagator for the
     *       given \b info_before_propagation. The method modifies the
     *       state of the propagator, so either it should be stateless or
     *       this should be an independent propagator object.
     * @post The resulting clause must:
     *     - Reduce the propagated parameter to \b propagated_domain
     *       for the same input domains.
     *     - For an arbitrary input domains, it should not remove domain values
     *       which are still supported by the inputs to this constraint.
     */
    static void explain_propagation_tightly(
        constraint_propagator & propagator,
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation);

    /**
     * @brief Generate a refined explanation for a past domain reduction.
     *
     * This method can be used to implement explain_propagation_tightly, or
     * explain_propagation for any propagator. It refines explanation for some
     * of the parameters.  This method uses propagator::propagate()
     * to refine the explanation, so the propagator should be either stateless
     * or to be a copy of the original propagator or propagation algo.
     *
     * The interface of this method is an extension to that of
     * explain_propagation().
     *
     * @param[in]  propagator
     *             A copy of the original propagator or propagation algo.
     *             This must be a copy not the original propagator.
     * @param[in]  propagated_param
     *             The index of the reduced parameter which we need to explain.
     * @param[in]  values_to_explain
     *             The set of values which, if removed, the caller wants to
     *             explain.
     * @param[in]  propagated_domain
     *             The domain resulting from the propagation which we need
     *             to explain.
     * @param[in]  refine_order_begin
     *             Pointer to the array containing parameter indexes
     *             that need to be refined, in their order.
     * @param[in]  refine_order_end
     *             Pointer past the last element of the array pointed by
     *             refine_order_begin. These pointers act like begin() end().
     * @param[in]  info_before_propagation
     *             The state of the solver before the propagation which we need
     *             to explain.
     * @param[out] explanation
     *             The literals of the explanation clause.
     *
     * @pre  The domain reduction was performed by this propagator for the
     *       given \b info_before_propagation. The method modifies the
     *       state of the propagator, so either it should be stateless or
     *       this should be an independent propagator object.
     * @post The resulting clause must:
     *     - Reduce the propagated parameter to \b propagated_domain
     *       for the same input domains.
     *     - For an arbitrary input domains, it should not remove domain values
     *       which are still supported by the inputs to this constraint.
     */
    static void refined_explanation(
        constraint_propagator & propagator,
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain,
        const unsigned *refine_order_begin, const unsigned *refine_order_end,
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation);

    /**
     * @brief tight explanations for one parameter.
     *
     * This is similar to explain_propagation_tightly().
     *
     * Tightening happens only on one of constraint
     * parameters. For many constraints such tightening is sufficient to get
     * an optimal explanation (at much lower cost).
     */
    static void explain_propagation_tightly1(
        constraint_propagator & propagator,
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation)
    {

        unsigned order[2];
        unsigned size=0;
        if (propagated_domain.empty()) {
            // minimizing cause1 barely helps as it is going to be eliminated
            // by the first resolution anyway.
            int cause1= info_before_propagation.latest_modified_param(-1U);
            int cause2= info_before_propagation.latest_modified_param(cause1);
            if (cause2 >= 0)
                order[size++]= cause2;
            if (cause1 >= 0)
                order[size++]= cause1;
            if (size == 0)
                order[size++]= 0;
        } else {
            // Minimizing the propagated literal barely helps since much of it
            // is going to be eliminated by resolution, anyway.
            //
            // Minimizing the 'cause' literal helps considerably because
            // it will most likely participate in the next resolution,
            // and minimizing also minimizes that next explanation.
            int effect= propagated_param;
            int cause= info_before_propagation.latest_modified_param(effect);
            if (cause >= 0)
                order[size++]= cause;
            order[size++]= effect;
        }
        
        refined_explanation(propagator, propagated_param, values_to_explain,
                            propagated_domain, &order[0], &order[size],
                            info_before_propagation, explanation);
    }
    /**
     * @brief For each parameter get all values that if assigned to the
     *        variable will cause the constraint not to propagate.
     *
     * Unlike explain_propagation() this method looks at each parameter
     * individually and see what values if set to info.domain() there will be no
     * propagation to propagated_param. This means that the result can be weaker
     * than an explanation and not fit the requirements of explanations. If the
     * constraint is given with \p values as input domain it is possible that
     * it will not propagate the required domain.
     *
     * Approximation: The constraint is allowed to over-approximate the
     * result. It may also return a smaller vector, indicating that the missing
     * values are not to be modified.
     *
     * @param[in]  propagator
     *             The propagator to work on.
     * @param[in]  propagated_param,propagated_domain
     *             The reduced parameter and its domain whose propagation is
     *             inhibited by returned values.
     * @param[in]  info_before_propagation
     *             The state of the solver before the propagation which we need
     *             to query.
     * @param[out] values
     *             values.size() <= info_before_propagation.arity().
     *             For element index i in values:
     *               - Adding any value x from values[i] to info.domain(i)
     *                 will not be able to propagate propagated_domain, it
     *                 will propagate something weaker.
     */
    static void possibly_crippling_values(
        constraint_propagator & propagator,
        unsigned propagated_param, const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & values);
    
  private:
    class conflicting_info;
}; // class explanation_generators

} // namespace hcsp
#endif // HCSP_EXPLANATION_GENERATORS_HPP_
