// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_EXPLANATION_RULE_HPP_
#define HCSP_EXPLANATION_RULE_HPP_

#include <vector>
#include <iosfwd>
#include <iset.hpp>

namespace hcsp {
class variable_print_strategy;

/**
 * @brief Interface for retaining formal rules that create propagation
 *        explanations.
 *
 * A concrete child of this abstract class contains an explanation rule.  An
 * explanation rule is a rule that derives the explanation clause from the
 * constraint.
 *
 * \b Example: \n
 * Consider a<=b and initial domains D(a)={1..100}, D(b)={1..5}.   \n
 * Propagation reduces the domain of \e a such that D'(a)={1..5}.  \n
 * Explanation: \f$ ( a \in [MIN_INT,5] \vee b \in [6,MAX_INT] \f$ \n
 * Rule: LE_6
 */
class explanation_rule {
  public:
    //! Destructor
    virtual ~explanation_rule() {}

    //! Deep copy of this object
    virtual explanation_rule * clone() const=0;

    /**
     * @brief Get the explanation implied by this rule.
     *
     * @param[out] explanation
     *             The explanation literals. They should be identical to
     *             the return of hcsp::constraint_propagator::get_explanation()
     */
    virtual void get_explanation(std::vector<iset> & explanation) const=0;

    //! Print the explanation rule: name and parameters.
    virtual void print(
        std::ostream & out,
        const variable_print_strategy & var_printer,
        const std::vector<unsigned> & constraint_variables) const=0;
}; // class explanation_rule


} // namespace hcsp
#endif // HCSP_EXPLANATION_RULE_HPP_
