// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_FLAGS_HPP_
#define HCSP_FLAGS_HPP_
#include <string>
#include <iosfwd>
#include <solver.hpp>
/**
 * @file flags.hpp
 * @brief This file defines run-time flags for programs based on HaifaCSP
 */
namespace hcsp {

/**
 * @brief A top-level class used for parsing and defining command line options
 */
class flags
{
  public:
    flags() : help(false) {}
    virtual ~flags() {}
    /**
     * @brief Parse the command line and populate the fields of this object
     *
     * @note A user who wishes to add flags should override the virtual
     *       functions below, and let flags::parse() do the job for him.
     *       Read the following methods to find what you have to override.
     *
     * @param argc,argv  The list of arguments copied from main().
     * @return True if the parsing succeeded, false if there was a failure.
     *         If help was requested, then true is returned and all_done() will
     *         tell the caller that the program should be exited successfully.
     */
    bool parse(int argc, char *argv[]);
    //! Return true if nothing more to do in the program, probably as a result
    //! of the -h flag.
    bool all_done() const { return help; }
    /**
     * @brief Apply the flags on the solver and on the global state
     *
     * Set the parameters of the solver according to these flags. Also set
     * global state such as random seed.
     *
     * @param solver The solver to update
     */
    virtual void apply(solver & solver) const { (void)solver; }
  protected:
    /**
     * @brief Get the set of legal args in getopt format
     *
     * The class that overrides this one should concatenate his string with
     * this one.
     */
    virtual std::string legal_args() const { return "h"; }
    /**
     * @brief Parse a single option
     *
     * @param option A single option, as returned by getopt();
     * @return True if the option was recognized by this class and its parents
     */
    virtual bool parse_option(char option);

    /**
     * @brief Print a short usage message containing the list of options without
     *        descriptions.
     *
     * For example [-h] [-s <integer>].
     *
     * If you want to add more options, you inherit from this class and
     * override this method. Then you call parent's method and later
     * concatenate your text to the parent's.
     */
    virtual void usage_short(std::ostream & out);
    /**
     * @brief Print a long usage message containing a full description of the
     *        flags.
     *
     * For example "-h  print this help message"
     *
     * If you want to add more options, you inherit from this class and
     * override this method. Then you call parent's method and later
     * concatenate your text to the parent's.
     */
    virtual void usage_long(std::ostream & out);
    //! validate that all the flags are consistent with each other
    virtual void validate_consistent() const {}
  private:
    //! print a help message by concatenating usage_short() and usage_long()
    void usage(std::ostream & out, const char * argv0);
    
    bool help; //!< should we only print a help message
}; // class flags

/**
 * @brief A class to handle all of HaifaCSP's flags
 */
class hcsp_flags: public flags
{
  public:
    hcsp_flags() : report_level(solver::NONE),
                  seed(0),
                  opportunistic_bj(0),
                  max_conflict_clause(-1U),
                  cost_lowering_factor(0),
                  analysis_cost_factor(-1),  // Negative marks defaults
                  conflict_cost_factor(-1),
                  all_results(false),
                  only_optimal(false),
                  backtrack_limit(-1), 
                  restart_threshold(-1),
                  use_all_diff(-1),
                  effort_boost(-1),
                  keep_constraints_from_conflicts(-1),
                  erase_after_backtracks(-1),
                  validate_unsat(false),
                  produce_proof(false),
                  intersect_with_ground(-1),
                  relaxation(solver::COMPLETE_RELAXATION),
                  cause_relaxation(false),
                  constraints_combination(true),
                  explainer_check_age(true),
                  augment_explanation(true),
                  quiet_start(false)
    {}
    
    //! Returns true if there is at least one option that uses the proof
    //! recording mechanism.
    bool has_proof_option() const { return validate_unsat || produce_proof; }
    
    virtual void apply(solver & solver) const;
    
  protected:
    virtual bool parse_option(char option);

    virtual void usage_short(std::ostream & out);
    virtual void usage_long(std::ostream & out);
    
    //! Make sure the flags do not conflict each other otherwise throw.
    void validate_consistent() const;
    virtual std::string legal_args() const
    { return flags::legal_args() + "DdmVGgApaJj:s:l:o:c:b:r:e:C:E:L:f:i:k:"; }
    
  public:
    solver::reporting_levels report_level;
    unsigned seed;
    unsigned opportunistic_bj;
    unsigned max_conflict_clause;
    unsigned cost_lowering_factor; // -f
    double analysis_cost_factor;   // -i
    double conflict_cost_factor;   // -k
    bool all_results; // -d Finds all results
    bool only_optimal; // -D prints only the optimal result without intermediate
    int backtrack_limit;
    int restart_threshold;
    int use_all_diff;
    int effort_boost;
    int keep_constraints_from_conflicts;
    int erase_after_backtracks;
    bool validate_unsat;
    bool produce_proof;
    int intersect_with_ground; // -1= default, 0=false, 1=true
    unsigned relaxation; // -j=N
    bool cause_relaxation; // -J
    bool constraints_combination; // -m  toggle on/off where initial value=true
    bool explainer_check_age;
    bool augment_explanation;
    bool quiet_start;
}; // class hcsp_flags

} // namespace hcsp

#endif // HCSP_FLAGS_HPP_
