// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010,2011 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010,2011 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#ifndef HCSP_FORMULA_HPP_
#define HCSP_FORMULA_HPP_

#include <iosfwd>
#include <iset.hpp>

namespace hcsp {

class solver;
class formula_impl;
struct literal_data;
class clause_data;
class variable_print_strategy;

/**
 * @brief A formula.
 *
 * This is class manages a formula which evaluates to false or true in a given
 * variable context. It can be constructed from other formulae or from
 * basic structures such as clauses or Boolean values.
 *
 * The formulae can be used to define a CSP problem in a more natural way,
 * or can result from an automatic process such as interpolant generation.
 *
 * This class formula shares a lot of concepts with operator_definition
 * with the main difference that operator_definition is used only by
 * xcsp_solver and formula is intended for wider use.
 *
 * @todo This class will have to replace or encapsulate operator_definition
 *       when class formula gets enough capabilities.
 */
class formula
{
  public:
    /**
     * @brief An object used for printing a formula
     *
     * A formula refers to variables which names we have to print. In order to
     * do this correctly, we have to use variable_print_strategy. For
     * convenience, the printer object holds the formula and the strategy and
     * prints them. For example:
     *
     * @code
     *   cout << printer(my_strategy, my_formula);
     * @endcode
     */
    class printer
    {
      public:
        /**
         * @brief Construct a formula printer
         * @param strategy  The strategy used to print variables' name
         * @param data      The formula to print
         */
        printer(const variable_print_strategy & strategy, const formula & data)
            : m_strategy(strategy), m_formula(data) {}
        
        //! Using this printer, print the formula
        friend
        std::ostream & operator<<(std::ostream & out, const printer & printer)
        {
            printer.m_formula.print(out, printer.m_strategy);
            return out;
        }
      private:
        const variable_print_strategy & m_strategy; //!< The strategy to use  
        const formula & m_formula;                  //!< The formula to print
    }; // formula::printer
    
    //! A default NULL formula which can't be evaluated or put into the solver
    formula() : m_formula(NULL) {}

    //! A constant formula that evaluates according with the Boolean value
    explicit formula(bool value);

    //! A formula which is true if and only if the literal is satisfied.
    formula(const literal_data &literal)  { construct(literal); }

    //! A formula which is true if and only if the literal is satisfied.
    formula(unsigned variable, const iset & set);
    
    //! A formula which is true if and only if the clause is satisfied.
    formula(const clause_data &clause);

    formula(const formula & other)    { link(other.m_formula); }

    //! For internal use only: link with the given formula implementation object
    formula(formula_impl *p_formula)  { link(p_formula); }
    ~formula() { unlink(); }

    //! Trivial assignment
    formula & operator=(const formula & other)
    {
        if (m_formula == other.m_formula)
            return *this;
        unlink();
        link(other.m_formula);
        return *this;
    }

    /**
     * @brief Force this formula to true in the given solver.
     * @param solver   The solver to constrain
     * @post \p solver is constrained with this formula
     */
    void constrain_solver(solver & solver) const;

    /**
     * @brief Check if the formula is true in the current complete assignment
     *
     * @param sovler  The solver whose variables are consulted.
     *
     * @return True if the formula is satisfied.
     *
     * @pre The solver contains a complete assignment to all variables 
     *      participating in the formula
     */
    bool evaluate(const solver & solver) const;

    void print(std::ostream &out, const variable_print_strategy &strategy)const;
    //! Create a formula by applying AND on two formulas
    friend formula operator&&(const formula &a, const formula &b);
    //! Create a formula by applying OR on two formulas
    friend formula operator||(const formula &a, const formula &b);
    //! Create a formula by applying NOT on a formula
    friend formula operator!(const formula &);
  private:
    //! construct the object from this literal, assume all current data is junk
    void construct(const literal_data &literal);
    
    /**
     * @brief Link formula_impl with this formula.
     *
     * Link the input formula_imp object by assigning it to the internal
     * data structures of this object and advancing its reference counter.
     *
     * @pre  This object is not currently linked with any formula
     * @post This object is linked with p_formula and the reference counter
     *       of the formula is increased by one.
     * @param p_formula The formula to link to, if not NULL.
     */
    void link(formula_impl * p_formula)
    {
        m_formula= p_formula;
        if (m_formula)
            mark_linked(*p_formula);
    }
    //! Mark the given formula_impl as linked, i.e., advance its reference count
    static void mark_linked(formula_impl & formula);
    /**
     * @brief Unlink the current formula_impl object from this.
     *
     * @pre Either m_formula is NULL or it is validly linked to an formula_impl
     * @post This formula is not linked to any object, and if it used to be
     *       the old object reference count is lowered.
     */
    void unlink()
    {
        if (m_formula)
            mark_unlinked(*m_formula);
        m_formula= NULL;
    }

    //! Mark the given formula_impl as unlinked, i.e., decrease reference count
    static void mark_unlinked(formula_impl & formula);
    
    //! The formula that this formula is linked to, if not NULL
    formula_impl *m_formula;
}; // class formula

} // namespace hcsp

#endif // HCSP_FORMULA_HPP_

