// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_INT128_HPP_
#define HCSP_INT128_HPP_
#include <stdexcept>
#include <cstdint>

namespace hcsp {
//! Support basic integer operations on 128 bits
class int128
{
  private:
    enum rounding_mode { ROUND_UP, ROUND_DOWN, ROUND_TO_ZERO};
  public:
    int128()
    {
    }
    explicit int128(int64_t value)
    {
        m_val[0]= value;
        // >> performs sign extension, repeating the sign bit.
        m_val[1]= value >> 63;
    }
    int64_t as_int64() const
    {
        if (!fits_int64())
            throw std::out_of_range("int128::as_int64() out of bounds");
        // Even if negative, two's complement will do the right thing
        return m_val[0];
    }
    explicit operator int64_t() const { return as_int64(); }
    bool fits_int64() const
    {
        if (is_negative64(m_val[0]))
            return m_val[1]+1 == 0;
        else
            return m_val[1] == 0;
    }
    int128 operator+(const int128 & in) const
    {
        int128 ret(*this);
        // m_val[0] becomes smaller iff and only if there is an unsigned
        // overflow.
        ret.m_val[0] += in.m_val[0];
        if (ret.m_val[0] < m_val[0])
            ++ret.m_val[1];
        ret.m_val[1] += in.m_val[1];
        
        return ret;
    }
    int128 operator+(int64_t in) const     { return *this + int128(in); }
    int128 & operator+=(int64_t in)        { return *this= *this + in; }
    int128 & operator++()                  { return *this += 1; }
    int128 & operator+=(const int128 & in) { return *this= *this + in; }
    int128 operator-(const int128 & in) const
    {
        int128 ret(*this);
        // m_val[0] becomes bigger iff and only if there is an unsigned
        // underflow.
        ret.m_val[0] -= in.m_val[0];
        if (ret.m_val[0] > m_val[0])
            --ret.m_val[1];
        ret.m_val[1] -= in.m_val[1];
        
        return ret;
    }
    int128 operator-(int64_t in) const     { return *this - int128(in); }
    int128 operator-() const
    {
        int128 ret;
        ret.m_val[0]= ~m_val[0];
        ret.m_val[1]= ~m_val[1];
        ++ret;
        return ret;
    }
    int128 & operator-=(int64_t in)        { return *this= *this - in; }
    int128 & operator--()                  { return *this -= 1; }
    int128 & operator-=(const int128 & in) { return *this= *this - in; }
    bool operator==(const int128 & other) const
    { return (other.m_val[0] == m_val[0] && other.m_val[1] == m_val[1]); }
    int128 operator>>(unsigned n) const
    {
        if (n >= 128)
            return is_negative()? int128(-1) : int128();
        int128 ret;
        if (n >= 64) {
            ret.m_val[1]= is_negative() ? uint64_t(0)-1 : uint64_t(0);
            ret.m_val[0]= static_cast<int64_t>(m_val[1]) >> (n-64);
        } else {
            ret.m_val[1]= static_cast<int64_t>(m_val[1]) >> n;
            // Do the shift twice, to avoid the case where we shift "<< 64".
            // Avoid it since it is undefined.
            uint64_t lost_n_bits= m_val[1] << (63-n) << 1;
            ret.m_val[0] = (m_val[0] >> n)|lost_n_bits;
        }
        return ret;
    }
    bool operator==(int64_t other) const { return *this == int128(other); }
    bool operator!=(const int128 & other) const { return !(*this == other); }
    bool operator!=(int64_t other) const { return !(*this == other); }
    bool operator<(const int128 & right) const
    {
        if (m_val[1] == right.m_val[1])
            return m_val[0] < right.m_val[0];
        else
            return (static_cast<int64_t>(m_val[1])
                    <
                    static_cast<int64_t>(right.m_val[1])
                    );
    }
    bool operator<(int64_t other) const { return *this < int128(other); }
    bool operator>(int64_t other) const { return int128(other) < *this; }
    bool operator>(const int128 &other) const { return other < *this; }
    bool operator>=(int64_t other) const { return !(*this < other); }
    bool operator>=(const int128 & other) const { return !(*this < other); }
    bool operator<=(int64_t other) const { return !(*this > other); }
    bool operator<=(const int128 & other) const { return !(*this > other); }
    int128 operator/(int32_t in) const { return divide(in, ROUND_TO_ZERO); }
    // avoid accidental truncation from 64 bit to 32
    void operator/(uint64_t)const = delete;
    int128 ceil_div(int32_t in) const { return divide(in, ROUND_UP); }
    int128 floor_div(int32_t in) const { return divide(in, ROUND_DOWN); }
  private:
    int128 divide(int32_t in, rounding_mode mode) const;
    static inline uint64_t divide_u64(uint64_t numerator, uint64_t denominator,
                                      rounding_mode mode);
    static inline int64_t divide_s64(int64_t numerator, int64_t denominator,
                                     rounding_mode mode);

    bool is_negative() const { return is_negative64(m_val[1]); }
    static bool is_negative64(int64_t val) { return val < 0; }
    
    uint64_t m_val[2]= {0,0};
}; // class int128
inline int128 ceil_div(const int128 &numerator, int32_t denominator)
{ return numerator.ceil_div(denominator); }

inline int128 floor_div(const int128 &numerator, int32_t denominator)
{ return numerator.floor_div(denominator); }

} // namespace hcsp

#endif // HCSP_INT128_HPP_
