// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_INTERVAL_PROPAGATORS_HPP_
#define HCSP_INTERVAL_PROPAGATORS_HPP_


#include <iset.hpp>
#include <constraint_propagator.hpp>

namespace hcsp {

//! Propagate the semantics of: v0+L0 <= v1 || v1+L1 <= v0
//! An interval that starts on point x and has length L is defined as
//! [x,x+L),i.e., all values t such that x<=t<x+L.
class no_overlap_propagator : public constraint_propagator
{
  public:
    no_overlap_propagator(unsigned length0, unsigned length1)
        : m_length0(length0), m_length1(length1) {}
    
    virtual int count_params() const { return 2; }
    // FIXME: Need to add lower and upper bound watches.
    //virtual bool init( constraint_runtime_info & info );
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain,
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual no_overlap_propagator* clone() const
    { return new no_overlap_propagator(*this); }
    
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        iset::value_type v0= info.domain(0).min();
        iset::value_type v1= info.domain(1).min();
        if (v0 < v1) 
            return constraint_check(v0, m_length0, v1);
        else
            return constraint_check(v1, m_length1, v0);
    }
  private:
    //! Check if none of the points of 'left' overlap the first corner of
    //! 'right'.
    bool constraint_check(iset::value_type left,  unsigned left_width,
                          iset::value_type right) const
    { return (long long)left+(long long)left_width <= right; }

    //! Propagate the domains of 'left' and 'right' taking only the width
    //! of 'left' into account.
    void propagate(iset & left, unsigned left_length, iset & right) const;

    /**
     * @brief Generate an explanation why v0 and v1 conflict.
     *
     * v0 and v1 are mutated by the algorithm. They may gain additional values
     * as long as those values still conflict.
     */
    void explain_empty(iset &v0, iset &v1, std::vector<iset>&explanation)const;

    /**
     * @brief Add elements to v0 and v1 as long as they still conflict.
     *
     * Add elements to v0 and v1 such that:
     *  -# No satisfying assignments are introduced.
     *  -# The v0 and v1 sets are made as equal as possible.
     */
    void extend_unsat_towards_equality(iset & v0, iset &v1) const;
    
    /**
     * @brief Add small elements to 'first' without affecting propagation of
     *        'second'
     *
     * Add elements to 'first' such that:
     *  -# 'second' will be updated the same way after this addition.
     *  -# The added elements form the longest valid contiguous set of
     *     smallest possible elements, ending with the original first.min().
     */
    void extend_first_min_for_reduction( iset & first, unsigned first_length,
                                         const iset & second) const;
    
    /**
     * @brief Add big elements to 'first' without affecting propagation of
     *        'second'
     *
     * Add elements to 'first' such that:
     *  -# 'second' will be updated the same way after this addition.
     *  -# The added elements form the longest valid contiguous set of
     *     biggest possible elements, starting with the original first.max().
     */
    void extend_first_max_for_reduction(
        iset & first, const iset & second, unsigned second_length) const;
    
    /**
     * @brief Add elements to 'first' without affecting propagation of 'second'
     *
     * Add elements to 'first' such that:
     *  -# 'second' will be updated the same way after this addition.
     *  -# No elements can be added without affecting 'second'
     */
    void extend_first_for_reduction(
        iset & first, unsigned first_length,
        const iset & second, unsigned second_length) const
    {
        extend_first_min_for_reduction(first, first_length, second);
        extend_first_max_for_reduction(first, second, second_length);
    }

    const unsigned m_length0;   //!< The length of the first interval
    const unsigned m_length1;   //!< The length of the second interval
}; // class no_overlap_propagator

/**
 * @brief Propagate the semantics of: [v1,v2] \subsetof [a,b]
 * More formally, the constraint requires:
 * v1 > v2 || (a <= v1 && v2 <= b)
 * Claim: If a>b then the constraint is equivalent with v1 > v2.
 * Proof: Assume a>b. 
 *      - Direction 1: assume that the constraint is satisfied.
 *                     The case when the constraint is satisfied due to v1>v2
 *                     is trivial. If v1<=v2 then a<=v1<=v2<=b, which 
 *                     is contrary to our initial assumption that a>b, and this
 *                     case is impossible. This means that if the constraint
 *                     is satisfied then v1>v2 is also falsified.
 *      - Direction 2: If the constraint is falsified then v1>v2 is also 
 *                     falsified.
 *
 *      We have seen that when a>b the constraint is equivalent with v1>v2.
 * The conclusion is that for an interval [x,y] when x>y we can view it as
 * an empty set.
 */
class sub_interval_propagator : public tight_constraint_propagator
{
  public:
    //! Create a reference interval with [a,b].
    sub_interval_propagator(iset::value_type a, iset::value_type b)
        : m_a(a), m_b(b) {}
    virtual int count_params() const { return 2; }
    virtual bool init( constraint_runtime_info & info );
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual sub_interval_propagator* clone() const
    { return new sub_interval_propagator(*this); }
    
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    { 
        bool result= can_be_satisfied(info.domain(0), info.domain(1));
        HCSP_ASSERT(result != can_be_falsified(info.domain(0), info.domain(1)));
        return result;
    }
  private:
    bool can_be_satisfied(const iset & x, const iset & y) const
    {
        return x.max() > y.min() || (m_a <= x.max() && y.min() <= m_b);
    }
    bool can_be_falsified(const iset & x, const iset & y) const
    {
        return x.min() <= y.max() && (x.min() < m_a || m_b < y.max());
    }
    result_type update_domains( constraint_runtime_info & info,
                                iset::value_type lowest_x,
                                iset::value_type highest_y);

    const iset::value_type m_a, m_b;

}; // sub_interval_propagator

/**
 * @brief Propagate the semantics of: [v1,v2-k] \not\subsetof [a,b]
 * More formally, the constraint requires:
 * v2-v1 >= k && (v1 <a || v2-k > b)
 * Claim: If a>b then the constraint is equivalent with v2-v1 >= k.
 * Proof: Assume a>b. 
 *      - Direction 1: assume that the constraint is falsified, i.e.,
 *                     v1>v2-k || (v1 >=a && v2-k <= b). If this expression is
 *                     true when !(v1>v2-k) then (v1 >=a && v2-k <= b) and
 *                     v1<=v2-k are true. This means that a<=v1<=v2-k<=b and
 *                     is contrary to our assumption that a>b. This means that
 *                     the expression may be true only if v1>v2-k, in which case
 *                     v2-v1 >= k is falsified.
 *      - Direction 2: If the constraint is satisfied then v2-v1>=k
 *
 * We have seen that when a>b, the constraint is equivalent with v2-v1 >= k.
 * The conclusion is that for an interval [x,y] when x>y we can view it as
 * an empty set.
 */
class no_sub_interval_propagator : public tight_constraint_propagator
{
  public:
    //! Create a reference interval with [a,b] and remember k.
    no_sub_interval_propagator(iset::value_type a, iset::value_type b,
                               iset::value_type k=0)
        : m_a(a), m_b(b), m_k(k) {}
    virtual int count_params() const { return 2; }
    virtual bool init( constraint_runtime_info & info );
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual no_sub_interval_propagator* clone() const
    { return new no_sub_interval_propagator(*this); }
    virtual constraint_propagator *new_propagator(
        disjunction_specification & disjunction,
        std::vector<unsigned> & copy_from_argument) const;
    
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    { 
        bool result= can_be_satisfied(info.domain(0), info.domain(1));
        HCSP_ASSERT(result != can_be_falsified(info.domain(0), info.domain(1)));
        return result;
    }
  private:
    bool can_be_satisfied(const iset & x, const iset & y) const
    {
        return (x.min() <= 0LL+y.max()-m_k && (
                    x.min() < m_a || m_b < 0LL+y.max()-m_k));
    }
    bool can_be_falsified(const iset & x, const iset & y) const
    {
        return (x.max() > 0LL+y.min()-m_k || (
                    m_a <= x.max() && 0LL+y.min()-m_k <= m_b));
    }

    iset::value_type m_a, m_b, m_k;

}; // no_sub_interval_propagator


} // namespace hcsp

#endif // HCSP_INTERVAL_PROPAGATORS_HPP_
