// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2012, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2012, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_LITERAL_CONSTRAINT_PROPAGATORS_HPP_
#define HCSP_LITERAL_CONSTRAINT_PROPAGATORS_HPP_

#include <constraint_propagator.hpp>

namespace hcsp {
/**
 * @brief A skeleton class for defining (literal || constraint) propagators.
 *
 * This class defines most of the things needed for:
 *    - (V0 \in A) || constraint(V1,...,Vn)
 *
 * Note that literal_or_constraint_propagator.hpp, which contains the
 * declaration of this class, does not contain the body of several
 * methods. The missing methods are defined in
 * literal_or_constraint_propagator.inl. The
 * literal_or_constraint_propagator.inl file must be included from the
 * translation unit in which you define your constraint, say my_propagator,
 * otherwise you will get 'undefined symbol' during link-time.
 *
 * @tparam Child  literal_or_constraint_propagator uses the Child parameter to
 *                access the definition of the inner 'constraint'. It uses
 *                Child directly rather than calling virtual functions in
 *                order to avoid the penalty of virtual function call and to
 *                make it easier for the compiler to inline the callee.
 *
 * @tparam Parent The parent of the class. This lets us choose the right
 *                parent in each case. The usual possibilities are
 *                constraint_propagator and tight_constraint_propagator.
 * 
 */
template <class Child, class Parent= constraint_propagator>
class literal_or_constraint_propagator : public Parent
{
  public:
    typedef typename Parent::result_type result_type;
    using Parent::UNSAT;
    using Parent::VALID;
    using Parent::CONSISTENT;
    using Parent::MODIFICATION_WATCH;

    /** @brief Construct the constraint using a literal
     * @param literal The set for which if V0 \in literal then
     *                the literal_or_constraint_propagator is satisfied.
     */
    literal_or_constraint_propagator(const iset &literal) :
        m_literal(literal) {}

    // The body of the method is in literal_constraint_propagators.inl
    virtual bool init( constraint_runtime_info & info );
    
    // The body of the method is in literal_constraint_propagators.inl
    virtual result_type propagate( constraint_runtime_info & info );

    // The body of the method is in literal_constraint_propagators.inl
    virtual bool conflicts(const constraint_runtime_info & info) const;
    
    // The body of the method is in literal_constraint_propagators.inl
    void print(std::ostream & out, const constraint_runtime_info *p_info)
        const throw();
    
    virtual unsigned count_watches() const override
    { return m_watch_handle.empty() ? 0 : 1; }
    
    virtual unsigned watched_param_index(unsigned watch_index) const
    {
        (void)watch_index;
        HCSP_ASSERT(watch_index == 0);
        return 0;
    }

    virtual void forget_watches() { m_watch_handle.clear(); }

    //! The set that defines the literal, which is: V0 in guard_set()
    const iset & guard_set() const { return m_literal; }
    
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        return (m_literal.contains_all_of(info.domain(0)) ||
                this_child().Child::constraint_check_non_literal_part(info)); }
  protected:    
    virtual bool init_non_literal_part( constraint_runtime_info & info)
    {
        const unsigned size= info.arity();
        for (unsigned i=1 ; i < size ; ++i)
            if ( !info.domain(i).singleton() )
                info.allocate_watch(i, MODIFICATION_WATCH);
        return true;
    }

    virtual
    result_type propagate_non_literal_part(constraint_runtime_info & info)=0;

    virtual result_type state_of_non_literal_part(
        const constraint_runtime_info & info) const = 0;

    
    virtual void print_non_literal_part(
        std::ostream & out,
        const constraint_runtime_info *p_info) const throw() = 0;

    virtual bool constraint_check_non_literal_part(
        const constraint_runtime_info & info) const = 0;
    
  private:
    Child &       this_child()       { return *static_cast<Child*>(this); }
    const Child & this_child() const { return *static_cast<const Child*>(this);}
    
    iset m_literal;
    watch_handle_t m_watch_handle;
    iset::value_type m_watched_value;
    
}; // class literal_or_constraint_propagator<Child, Parent>

/**
 * @brief A skeleton class for defining (literal iff constraint) propagators.
 *
 * This class defines most of the things needed for:
 *    - (V0 \in A) iff constraint(V1,...,Vn)
 *
 * Note that literal_or_constraint_propagator.hpp, which contains the
 * declaration of this class, does not contain the body of several
 * methods. The missing methods are defined in
 * literal_or_constraint_propagator.inl. The
 * literal_or_constraint_propagator.inl file must be included from the
 * translation unit in which you define your constraint, say my_propagator,
 * otherwise you will get 'undefined symbol' during link-time.
 *
 * @tparam Child  literal_or_constraint_propagator uses the Child parameter to
 *                access the definition of the inner 'constraint'. It uses
 *                Child directly rather than calling virtual functions in
 *                order to avoid the penalty of virtual function call and to
 *                make it easier for the compiler to inline the callee.
 *
 * @tparam Parent The parent of the class. This lets us choose the right
 *                parent in each case. The usual possibilities are
 *                constraint_propagator and tight_constraint_propagator.
 *
 * @invariant !m_positive_watch_handle.empty() iff 
 *            !m_negated_watch_handle.empty().
 *            I.e. both literal watches are valid, or none.
 */
template <class Child, class Parent>
class literal_iff_constraint_propagator : public Parent
{
  public:
    typedef typename Parent::result_type result_type;
    using Parent::UNSAT;
    using Parent::VALID;
    using Parent::CONSISTENT;
    using Parent::MODIFICATION_WATCH;

    /** @brief Construct the constraint using a literal
     * @param literal The set for which if V0 \in literal then
     *                the literal_or_constraint_propagator is satisfied.
     */
    literal_iff_constraint_propagator(const iset &literal) :
        m_literal(literal) {}

    // The body of the method is in literal_constraint_propagators.inl
    virtual bool init(constraint_runtime_info & info );
    
    // The body of the method is in literal_constraint_propagators.inl
    virtual result_type propagate(constraint_runtime_info & info );

    // The body of the method is in literal_constraint_propagators.inl
    virtual bool conflicts(const constraint_runtime_info & info) const;
    
    // The body of the method is in literal_constraint_propagators.inl
    void print(std::ostream & out, const constraint_runtime_info *p_info)
        const throw();
    
    virtual unsigned count_watches() const
    {
        // see invariant
        HCSP_ASSERT( m_positive_watch_handle.empty()
                     == 
                     m_negated_watch_handle.empty());
        return (!m_positive_watch_handle.empty())*2; 
    }
    
    virtual unsigned watched_param_index(unsigned watch_index) const
    {
        (void)watch_index;
        HCSP_ASSERT(watch_index == 0 || watch_index == 1);
        return 0;
    }

    virtual void forget_watches()
    {
        m_positive_watch_handle.clear();
        m_negated_watch_handle.clear();
    }
    
    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    {
        return (m_literal.contains_all_of(info.domain(0)) ==
                this_child().Child::constraint_check_non_literal_part(info)); }
  protected:
    //! The set that defines the literal, which is: V0 in guard_set()
    const iset & guard_set() const { return m_literal; }

    /**
     * @brief Initialize the case when non_literal_part is satisfied.
     *
     * There are two possibilities for the non_literal_part which may,
     * depending on the literal, satisfy the whole constraint. This method
     * initializes the case when non_literal_part may be positive.
     *
     * @param info  The current context of the whole constraint
     * @return false if non_literal_part can't be satisfied
     */
    virtual bool init_non_literal_part(constraint_runtime_info & info)
    {
        const unsigned size= info.arity();
        for (unsigned i=1 ; i < size ; ++i)
            if ( !info.domain(i).singleton() )
                info.allocate_watch(i, MODIFICATION_WATCH);
        return true;
    }

    /**
     * @brief Initialize the case when non_literal_part is falsified.
     *
     * There are two possibilities for the non_literal_part which may,
     * depending on the literal, satisfy the whole constraint. This method
     * initializes the case when non_literal_part may be negative.
     *
     * @param info  The current context of the whole constraint
     * @return false if non_literal_part can't be falsified
     */
    virtual bool init_neg_non_literal_part(constraint_runtime_info & info)
    {
        (void) info; // suppress 'unused variable' warnings.
        return true;
    }
    
    /**
     * @brief Propagate the non_literal_part making it satisfied.
     *
     * There are two possibilities for the non_literal_part which may,
     * depending on the literal, satisfy the whole constraint. This method
     * is for the case when non_literal_part must be satisfied
     *
     * @param info  The current context of the whole constraint
     * @return the result type relevant only to the non_literal_part.
     */
    virtual
    result_type propagate_non_literal_part(constraint_runtime_info & info)=0;

    /**
     * @brief Propagate (!non_literal_part) falsifying the non_literal_part.
     *
     * There are two possibilities for the non_literal_part which may,
     * depending on the literal, satisfy the whole constraint. This method
     * is for the case when non_literal_part must be falsified.
     *
     * @param info  The current context of the whole constraint
     * @return the result type relevant only to (!non_literal_part).
     */
    virtual
    result_type propagate_neg_non_literal_part(constraint_runtime_info &info)=0;

    virtual result_type state_of_non_literal_part(
        const constraint_runtime_info & info) const = 0;

    virtual void print_non_literal_part(
        std::ostream & out,
        const constraint_runtime_info *p_info) const throw() = 0;

    virtual bool constraint_check_non_literal_part(
        const constraint_runtime_info & info) const = 0;
    
  private:
    Child &       this_child()       { return *static_cast<Child*>(this); }
    const Child & this_child() const { return *static_cast<const Child*>(this);}

    //! Propagate the literal, assuming its current value is in_cond.
    //! Assume that the non_literal_part is VALID.
    result_type propagate_literal(constraint_runtime_info & info,
                                  const iset & in_cond);
    
    //! Propagate the negation of the literal, assuming its current value is
    //! in_cond. Assume that the non_literal_part is UNSAT.
    result_type propagate_neg_literal(constraint_runtime_info & info,
                                      const iset & in_cond);

    //! Allocate a positive watch from \p cond_domain values.
    //! Assume that no positive watch is currently allocated
    void allocate_positive_watch(constraint_runtime_info & info,
                                 const iset & cond_domain);
    //! Allocate a negative watch which is not one of \p cond_domain values.
    //! Assume that no positive watch is currently allocated
    void allocate_negative_watch(constraint_runtime_info & info,
                                 const iset & cond_domain);

    
    iset m_literal;

    //! This is the handle of the watch that guards the possibility of
    //! the literal being satisfied. empty() if not watched.
    watch_handle_t m_positive_watch_handle;
    
    //! This is a value of the domain for which
    //! m_positive_watched_value \in m_literal, if such value exists.
    iset::value_type m_positive_watched_value;
    
    //! This is the handle of the watch that guards the possibility of
    //! the literal being falsified. empty() if not watched.
    watch_handle_t m_negated_watch_handle;
    
    //! This is a value of the domain for which
    //! m_negated_watched_value \not\in m_literal, if such value exists.
    iset::value_type m_negated_watched_value;
}; // class literal_iff_constraint_propagator<Child, Parent>

} // end of namespace hcsp

#endif // HCSP_LITERAL_CONSTRAINT_PROPAGATORS_HPP_
