// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2012, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2012, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_LITERAL_CONSTRAINT_PROPAGATORS_INL_
#define HCSP_LITERAL_CONSTRAINT_PROPAGATORS_INL_

#include <literal_constraint_propagators.hpp>

namespace hcsp {

//virtual 
template <class Child, class Parent>
bool literal_or_constraint_propagator<Child, Parent>::init(
    constraint_runtime_info & info )
{
    const iset & cond_domain= info.domain(0);
    m_watch_handle.clear();
    if ( m_literal.contains_all_of(cond_domain) )
        return true; // trivially satisfied, no init to non_literal_part
    bool literal_may_be_true= m_literal.contains_any_of(cond_domain);
    if ( literal_may_be_true ) {
        m_watched_value= iset(m_literal).intersect(cond_domain).min();
        m_watch_handle= info.allocate_value_watch(0, m_watched_value);
    }
    bool non_literal_part_may_be_true= init_non_literal_part(info);
    return literal_may_be_true || non_literal_part_may_be_true;
}

//virtual
template <class Child, class Parent>
typename Parent::result_type
literal_or_constraint_propagator<Child, Parent>::propagate(
    constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    Child &child= this_child();
    const iset & in_cond= info.domain(0);
    if (in_cond.empty())
        return UNSAT;
    if (m_literal.contains_all_of(in_cond))
        return VALID;
        
    // At this point we are free to satisfy or falsify the literal.
    result_type non_literal_state;
    non_literal_state= child.Child::state_of_non_literal_part(info);
    if (non_literal_state == VALID)
        return VALID;
        
    if (!m_literal.contains_any_of(in_cond)) { // [1] (see below)
        if (non_literal_state == UNSAT)
            return UNSAT;
        return child.Child::propagate_non_literal_part(info);
    }


    iset out_cond= in_cond;
    out_cond.intersect(m_literal);
    HCSP_ASSERT(!out_cond.empty()); // Avoided due to [1]
        
    if ( !m_watch_handle.empty() &&
         !in_cond.contains(m_watched_value) &&
         in_cond.contains_any_of(m_literal) ) {
        info.deallocate_value_watch(0, m_watched_value, m_watch_handle);
        m_watched_value= out_cond.min();
        m_watch_handle= info.allocate_value_watch(0, m_watched_value);
    }
    if (non_literal_state != UNSAT)
        return CONSISTENT; // we still don't know which part is satisfied
    info.set_domain(0, std::move(out_cond));
    return VALID;
}

//virtual
template <class Child, class Parent>
bool literal_or_constraint_propagator<Child, Parent>::conflicts(
    const constraint_runtime_info & info) const
{
    HCSP_PROFILE_CURRENT;
    const iset & in_cond= info.domain(0);
    if (in_cond.empty())
        return true;
    if (m_literal.contains_any_of(in_cond))
        return false;
    
    return this_child().Child::state_of_non_literal_part(info) == UNSAT;
}

//virtual
template <class Child, class Parent>
void literal_or_constraint_propagator<Child, Parent>::print(
    std::ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        p_info->print_param(out, 0);
        if (m_literal.singleton())
            out << "=" << m_literal;
        else
            out << ":{" << m_literal << "} ";
        out << "|| ";
        print_non_literal_part(out, p_info);
        out << ")";
    } else {
        out << "literal_or_";
        print_non_literal_part(out, p_info);
    }
}    

template <class Child, class Parent>
void literal_iff_constraint_propagator<Child, Parent>::allocate_positive_watch(
    constraint_runtime_info & info, const iset & cond_domain)
{
    iset satisfying_values= cond_domain;
    satisfying_values.intersect(m_literal);
    m_positive_watched_value= satisfying_values.min();
    m_positive_watch_handle= info.allocate_value_watch(
        0, m_positive_watched_value);
}

template <class Child, class Parent>
void literal_iff_constraint_propagator<Child, Parent>::allocate_negative_watch(
    constraint_runtime_info & info, const iset & cond_domain)
{
    iset falsifying_values= cond_domain;
    falsifying_values.erase_all_of(m_literal);
    m_negated_watched_value= falsifying_values.min();
    m_negated_watch_handle= 
        info.allocate_value_watch(0, m_negated_watched_value);
}

//virtual 
template <class Child, class Parent>
bool literal_iff_constraint_propagator<Child, Parent>::init(
    constraint_runtime_info & info )
{
    Child &child= this_child();
    const iset & cond_domain= info.domain(0);
    m_positive_watch_handle.clear();
    m_negated_watch_handle.clear();
    bool literal_may_be_true=  m_literal.contains_any_of(cond_domain);
    bool literal_may_be_false= !m_literal.contains_all_of(cond_domain);

    if (literal_may_be_true)
        literal_may_be_true= child.Child::init_non_literal_part(info);
    
    if (literal_may_be_false)
        literal_may_be_false= child.Child::init_neg_non_literal_part(info);
    
    // If we know the value of the literal (positive or negative) at this
    // time, then there is no need to watch and wait for its change.
    if ( literal_may_be_true && literal_may_be_false) {
        allocate_positive_watch(info, cond_domain);
        allocate_negative_watch(info, cond_domain);
        
    }
    return literal_may_be_false || literal_may_be_true;
}

template <class Child, class Parent>
typename Parent::result_type
literal_iff_constraint_propagator<Child, Parent>::propagate_literal(
    constraint_runtime_info & info, const iset & in_cond)
{
    if (m_literal.contains_all_of(in_cond))
        return VALID;
    if (!m_literal.contains_any_of(in_cond))
        return UNSAT;
    info.set_domain(0, std::move(iset(m_literal).intersect(info.domain(0))));
    return VALID;
}

template <class Child, class Parent>
typename Parent::result_type
literal_iff_constraint_propagator<Child, Parent>::propagate_neg_literal(
    constraint_runtime_info & info, const iset & in_cond)
{
    if (!m_literal.contains_any_of(in_cond))
        return VALID;
    if (m_literal.contains_all_of(in_cond))
        return UNSAT;
    iset next_domain = info.domain(0);
    next_domain.erase_all_of(m_literal);
    info.set_domain(0, std::move(next_domain));
    return VALID;
}

//virtual
template <class Child, class Parent>
typename Parent::result_type
literal_iff_constraint_propagator<Child, Parent>::propagate(
    constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    Child &child= this_child();
    const iset & in_cond= info.domain(0);
    if (in_cond.empty())
        return UNSAT;
    
    result_type non_literal_state;
    non_literal_state= child.Child::state_of_non_literal_part(info);
    
    if (non_literal_state == VALID)
        return propagate_literal(info, in_cond);
    if (non_literal_state == UNSAT)
        return propagate_neg_literal(info, in_cond);
    
    if (m_literal.contains_all_of(in_cond))
        return child.Child::propagate_non_literal_part(info);
    if (!m_literal.contains_any_of(in_cond))
        return child.Child::propagate_neg_non_literal_part(info);

    if ( !info.reduced_between_activations(0))
        return CONSISTENT;
    HCSP_ASSERT(!m_negated_watch_handle.empty());
    HCSP_ASSERT(!m_positive_watch_handle.empty());
    if ( !in_cond.contains(m_positive_watched_value) ) {
        info.deallocate_value_watch(0, m_positive_watched_value,
                                    m_positive_watch_handle);
        allocate_positive_watch(info, in_cond);
    }
    if ( !in_cond.contains(m_negated_watched_value) ) {
        info.deallocate_value_watch(0, m_negated_watched_value,
                                    m_negated_watch_handle);
        allocate_negative_watch(info, in_cond);
    }
    return CONSISTENT;
}

//virtual
template <class Child, class Parent>
bool literal_iff_constraint_propagator<Child, Parent>::conflicts(
    const constraint_runtime_info & info) const
{
    HCSP_PROFILE_CURRENT;
    const iset & in_cond= info.domain(0);
    if (m_literal.contains_all_of(in_cond)) {
        return (state_of_non_literal_part(info) == UNSAT);
    } else if (!m_literal.contains_any_of(in_cond)) {
        return (state_of_non_literal_part(info) == VALID);
    } else {
        return false;
    }
}

//virtual
template <class Child, class Parent>
void literal_iff_constraint_propagator<Child, Parent>::print(
    std::ostream & out, const constraint_runtime_info *p_info) const throw()
{
    if (p_info) {
        p_info->print_prefix(out);
        out << "(";
        p_info->print_param(out, 0);
        if (m_literal.singleton())
            out << "=" << m_literal;
        else
            out << ":{" << m_literal << "}";
        out << " iff ";
        print_non_literal_part(out, p_info);
        out << ")";
    } else {
        out << "literal_iff_";
        print_non_literal_part(out, p_info);
    }
}    

} // end of namespace hcsp

#endif // HCSP_LITERAL_CONSTRAINT_PROPAGATORS_INL_
