// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#ifndef MID_ACCESS_PRIORITY_QUEUE_HPP_
#define MID_ACCESS_PRIORITY_QUEUE_HPP_

#include <algorithm>
#include <functional>
#include <stdexcept>
#include <vector>

namespace hcsp {
/**
 * @brief A priority queue that can read and modify any element
 *
 * Unlike the standard priority queue, this queue provides methods for
 * accessing and modifying any entry. To access an element one has to use a
 * 'key' which points to the element. The key is a sequential index of the
 * element, such that for a queue with N elements a key may be any value
 * between 0 and N-1, inclusive.
 */
template <typename Element, typename Compare=std::less<Element> >
class mid_access_priority_queue
{
  public:
    class mutable_reference;
    typedef mutable_reference reference;
    typedef Element const_reference;
    typedef unsigned size_type;
    typedef size_type key_type;
    typedef Element data_type;
    
    mid_access_priority_queue() {}

    template <typename Iterator>
    mid_access_priority_queue(Iterator begin, Iterator end)
    { assign(begin, end); }
    template <typename Iterator>
    void assign(Iterator begin, Iterator end);

    /**
     * @brief Insert a new element with a new key.
     *
     * Insert an element into the queue, with key = original(size()).
     * It is allowable repeatedly insert the same element, and to expect them
     * all enter the queue.
     *
     * @param element   The element to be inserted 
     */
    inline void insert(Element element);
    inline const Element & max() const;
    inline key_type key_of_max() const;
    inline size_type size() const { return m_heap.size(); }
    inline const_reference operator[](key_type key) const;
    inline reference operator[](key_type key);
    inline void swap_keys(key_type key1, key_type key2);
    inline void pop_last_key();
    
  private:
    struct heap_entry
    {
        friend bool operator<(const heap_entry & a, const heap_entry & b)
        { return Compare()(a.m_element, b.m_element); }
        Element    m_element;
        size_type m_key;
    }; // struct mid_update_heap::heap_entry
    void fix_heap_position(size_type heap_pos);
    
    static size_type parent(size_type child_key)
    { return (child_key-1)/2; }
    static size_type child(size_type parent_key, bool is_right)
    { return parent_key*2+1+is_right; }
    void swap_heap_entries(size_type pos1, size_type pos2);
    
    std::vector<heap_entry> m_heap;
    std::vector<size_type> m_key_to_heap_pos;
}; // class mid_access_priority_queue

template <typename Element, typename Compare>
class mid_access_priority_queue<Element, Compare>::mutable_reference
{
  private:
    friend class mid_access_priority_queue<Element, Compare>;
    mutable_reference(mid_access_priority_queue &queue, size_type heap_pos)
        : m_queue(queue), m_heap_pos(heap_pos) {}
    
  public:
    operator Element() const { return m_queue.m_heap[m_heap_pos].m_element; }
    mutable_reference & operator=(Element element)
    {
        m_queue.m_heap[m_heap_pos].m_element= element;
        m_queue.fix_heap_position(m_heap_pos);
        return *this;
    }
    mutable_reference & operator=(const mutable_reference & other)
    { return *this= static_cast<Element>(other); }

    bool operator==(const mutable_reference & other) const
    { return static_cast<Element>(*this) == static_cast<Element>(other); }

    bool operator<(const mutable_reference & other) const
    { return static_cast<Element>(*this) < static_cast<Element>(other); }
    
  private:
    void fix_heap_position(size_type pos);
    
    mid_access_priority_queue & m_queue;
    size_type m_heap_pos;
}; // class mid_access_priority_queue::mutable_reference

template <typename Element, typename Compare>
template <typename Iterator>
void mid_access_priority_queue<Element, Compare>::assign(Iterator begin,
                                                         Iterator end)
{
    std::vector<heap_entry> heap;
    std::vector<size_type> element_index_to_heap_entry;
    heap.swap(m_heap);
    element_index_to_heap_entry.swap(m_key_to_heap_pos);
    
    heap.resize(std::distance(begin, end));
    unsigned i= 0;
    for (Iterator it= begin ; it != end ; ++it) {
        heap[i].m_element= *it;
        heap[i].m_key= i;
        ++i;
    }
    
    std::make_heap(heap.begin(), heap.end());

    element_index_to_heap_entry.resize(heap.size());
    for (i=0 ; i != heap.size() ; ++i)
        element_index_to_heap_entry[heap[i].m_key]= i;
        
    heap.swap(m_heap);
    element_index_to_heap_entry.swap(m_key_to_heap_pos);        
}

template <typename Element, typename Compare>
const Element & mid_access_priority_queue<Element, Compare>::max() const
{ return m_heap[0].m_element; }

template <typename Element, typename Compare>
typename mid_access_priority_queue<Element, Compare>::key_type
mid_access_priority_queue<Element, Compare>::key_of_max() const
{ return m_heap[0].m_key; }

template <typename Element, typename Compare>
void mid_access_priority_queue<Element, Compare>::insert(Element element)
{
    unsigned position = m_heap.size();
    m_heap.push_back(heap_entry());
    m_heap.back().m_element= element;
    m_heap.back().m_key= position;
    try {
        m_key_to_heap_pos.push_back(position);
    } catch (...) {
        m_heap.pop_back();
        throw;
    }
    fix_heap_position(position);
}

template <typename Element, typename Compare>
void mid_access_priority_queue<Element, Compare>::swap_heap_entries(
    size_type pos1, size_type pos2)
{
    std::swap(m_heap[pos1], m_heap[pos2]);
    m_key_to_heap_pos[m_heap[pos1].m_key]= pos1;
    m_key_to_heap_pos[m_heap[pos2].m_key]= pos2;
}

template <typename Element, typename Compare>
void mid_access_priority_queue<Element, Compare>::fix_heap_position(
    size_type pos)
{
    bool updated= false;
    while ( pos != 0 && m_heap[parent(pos)] < m_heap[pos]) {
        swap_heap_entries(pos, parent(pos));
        pos= parent(pos);
        updated= true;
    }
    if (updated)
        return;
    while (child(pos, true) < m_heap.size()) {
        size_type left= child(pos, false), right= child(pos, true);
        if (m_heap[pos] < m_heap[left] || m_heap[pos] < m_heap[right]) {
            size_type next_pos= m_heap[left] < m_heap[right] ? right : left;
            swap_heap_entries(next_pos, pos);
            pos= next_pos;
        } else {
            return;
        }
    }
    size_type left= child(pos, false);
    if (left< m_heap.size() && m_heap[pos] < m_heap[left]) {
        swap_heap_entries(pos, left);
    }
}

template <typename Element, typename Compare>
void mid_access_priority_queue<Element, Compare>::swap_keys(
    key_type key1, key_type key2)
{
    if (key1 == key2)
        return;
    if (key1 >= m_key_to_heap_pos.size() || key2 >= m_key_to_heap_pos.size()) {
        throw std::out_of_range("mid_access_priority_queue::swap_keys out of "
                                "range");
    }
    size_type pos1= m_key_to_heap_pos[key1];
    size_type pos2= m_key_to_heap_pos[key2];

    m_heap[pos1].m_key= key2;
    m_heap[pos2].m_key= key1;
    m_key_to_heap_pos[key1]= pos2;
    m_key_to_heap_pos[key2]= pos1;
}


template <typename Element, typename Compare>
void mid_access_priority_queue<Element, Compare>::pop_last_key()
{
    size_type pos= m_key_to_heap_pos.back();
    size_type last_pos= m_heap.size()-1;
    if (pos < last_pos) {
        swap_heap_entries(pos, last_pos);
        m_heap.pop_back();
        m_key_to_heap_pos.pop_back();
        fix_heap_position(pos);
    } else {
        m_heap.pop_back();
        m_key_to_heap_pos.pop_back();
    }
}

template <typename Element, typename Compare>
typename mid_access_priority_queue<Element, Compare>::const_reference
mid_access_priority_queue<Element, Compare>::operator[](key_type key) const
{
    size_type pos= m_key_to_heap_pos.at(key);
    return m_heap[pos].m_element;
}

template <typename Element, typename Compare>
typename mid_access_priority_queue<Element, Compare>::mutable_reference
mid_access_priority_queue<Element, Compare>::operator[](key_type key)
{
    return mutable_reference(*this, m_key_to_heap_pos[key]);
}

} // namespace hcsp
#endif // MID_ACCESS_PRIORITY_QUEUE_HPP_
