// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_DIFF_PROPAGATORS_HPP_
#define HCSP_DIFF_PROPAGATORS_HPP_


#include <iset.hpp>
#include <constraint_propagator.hpp>
#include <literal_constraint_propagators.hpp>

namespace hcsp {

//! Propagate the semantics of: v1 != v2.
class neq_propagator : public constraint_propagator
{
  public:
    virtual int count_params() const { return 2;}
    virtual bool init( constraint_runtime_info & info );
    virtual constraint_propagator * intersect_with(
        const constraint_runtime_info & info,
        std::vector<unsigned> & copy_from_argument) const override;
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual bool constraint_check( const constraint_runtime_info & info ) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual neq_propagator* clone() const { return new neq_propagator(*this); }
    virtual std::unique_ptr<explanation_rule> rule(
        const std::vector<iset> & literals) const;
  private:
    class neq_rule;
}; // neq_propagator

//! Propagate the semantics of: forall i!=j. [v_i != v_j]
class all_different_propagator : public constraint_propagator
{
  public:
    all_different_propagator(int in_count_params=-1)
        : m_count_params(in_count_params)
    {}
    virtual int count_params() const { return m_count_params;} 

    virtual result_type propagate( constraint_runtime_info & info );
    
    virtual bool conflicts(const constraint_runtime_info & info) const;

    virtual bool constraint_check( const constraint_runtime_info & info ) const;
    
    virtual void explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain,
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation) const;
    
    virtual std::unique_ptr<explanation_rule> rule(
        const std::vector<iset> & literals) const;
    
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual all_different_propagator* clone() const
    { return new all_different_propagator(*this); }
  private:
    class all_diff_rule;
    //! \brief Update the domains of \p info after propagation
    result_type update_domains( constraint_runtime_info & info,
                                const std::vector<iset> & next_domains);

    static std::vector<iset::value_type> get_small_domain_values(
        const constraint_runtime_info & info,
        const std::vector<unsigned char> & big_parameters);
    static iset get_big_domain(
        const constraint_runtime_info &info, 
        const std::vector<unsigned char> & big_parameters,
        std::vector<iset::value_type> & values_to_check,
        const iset & small_domains);

    
    //! \brief Find the set of parameters too big to affect propagation.
    //! \return if a parameter 'i' is big then return[i] == true.
    static std::vector<unsigned char>
    find_big_parameters(const constraint_runtime_info & info);

    /**
      * @brief Populate \p domains with only small parameters.
      *
      * Maybe put a small representation for all big parameters at the
      * back of the \p domains vector.
      *
      * @param[in] info
      *        The state of the solver.
      *
      * @param[in] big_parameters
      *        For each parameter tell whether it is big enough not
      *        to affect the propagation.
      *
      * @param[out] domains
      *        The set of domains, with big domains filtered out. The
      *        order of domains is preserved from info.
      *        If \p last_represents_big_params is \p true then the last
      *        domain represents all big domains. \n
      *        Only removal of values from this domain will have any meaning
      *        the values themselves are meaningless.
      *
      * @param[out] last_represents_big_params
      *        \p true if the last parameter of \p domains represents all big
      *        domains.
      *
      * @return \p false if and only if there is a conflict in the input
      *        domains which \p info contains. I.e. there is no feasible
      *        assignment from the domains which does not violate this
      *        constraint.
      */
    static bool get_small_parameters(
        const constraint_runtime_info & info,
        const std::vector<unsigned char> & big_parameters,
        std::vector<iset> & domains,
        bool & last_represents_big_params);

    /**
     * @brief Extend the propagation done on simplified domains to the
     * original big domains.
     *
     * @param[in] info
     *        The state of the solver before propagation.
     *
     * @param[in] domains
     *        The set of smaller domains before propagation,
     *        as created by get_small_parameters().
     * @param[in] big_parameters
     *        For each parameter tell whether it is big.
     *
     * @param[in] last_represents_big_params
     *        \p true if domains.back() represents all big parameters.
     *
     * @param[in,out] next_domains
     *        Before the call to this method (in) this is the set of small
     *        domains after propagation. After the call (out) this is the set
     *        of all domains after propagation (including big domains).
     */
    static void extend_propagation_for_big_domains(
        const constraint_runtime_info & info,
        const std::vector<iset> & domains,
        const std::vector<unsigned char> & big_parameters,
        bool last_represents_big_params,
        std::vector<iset> & next_domains);

    /**
     * @brief Get small domains suitable for explanation generation.
     *
     * Get the set of small domains which are sufficient for explanation
     * generation. This is similar to get_small_parameters(), with one
     * major difference. There is no need to maintain enough information
     * for all big parameters. All we need is to maintain information for
     * at most one big parameter. This means that only the explained big
     * parameter is kept, and kept in its original place. Other big parameters
     * can be eliminated.
     *
     * @param[in] info_before_propagation
     *        The state of the solver before the explained propagation.
     *
     * @param[in] propagated_param
     *        The index of the parameter (of \p info_before_propagation)
     *        that has to be explained.
     *
     * @param[in] propagated_domain
     *        The domain after propagation that has to be explained.
     *        Empty when the explained propagation returned UNSAT.
     *
     * @param[in,out] big_parameters
     *        For each parameter, tells whether it has a big domain.
     *        These parameters will not be represented in the \p domains
     *        vector. These values define the holes in that are punched in
     *        \p info_before_propagation during the construction of
     *        \p domains.\n
     *        \n
     *        There is a need for at most one big domain in the explanation.
     *        If such big domain is required, then we are interested only in
     *        value removals from that domain. All other parts of the big
     *        domain are not interesting, and value removals are relatively
     *        small. This means that the big domain can be replaced with a
     *        smaller domain that can represent the potential value
     *        removals. This replacement means that the said parameter
     *        is no longer marked as "big". It is removed from the
     *        \p big_parameters set.\n
     *        \n
     *        As a result, the original big parameter keeps its original
     *        position but with a smaller domain. Since explanation is only
     *        for value removals, moving to a smaller domain does not spoil
     *        the explanation as long as it contains all values which would
     *        be removed during propagation
     *
     * @param[out] domains
     *        The domains of the small parameters that may explain the
     *        propagation.
     *
     * @return The index of the propagated parameter in terms of the
     *         \p domains vector.
     */
    static unsigned get_explanation_small_parameters(
        const constraint_runtime_info & info_before_propagation,
        unsigned propagated_param, const iset & propagated_domain,
        std::vector<unsigned char> & big_parameters,
        std::vector<iset> & domains);

    /**
     * @brief Extend the explanation from the vector of "only small domains"
     *        to the original vector of domains.
     *
     * @param[in] arity           The number of parameters of the constraint.
     * @param[in] big_parameters  The set of big_domain parameters, which
     *                            do not participate in the small explanation.
     * @param[in,out] explanation The explanation of the small domains
     *                            and, after extension, the original domains.
     */
    static void extend_explanation_for_big_domains(
        unsigned arity, const std::vector<unsigned char> & big_parameters,
        std::vector<iset> & explanation);

    /**
     * @brief Count the number of unique singleton parameters.
     *
     * Count the number of singletons which do not overlap other,
     * non-singleton parameters.
     */
    static unsigned count_unique_singletons(
        const constraint_runtime_info &info);

    class impl;
    class graph;
    class reference_info;
    int m_count_params;
}; // all_different_propagator


//! Propagate the semantics of: v2 - v1 != constant
class difference_neq_constant_propagator : public constraint_propagator
{ 
  public:
    difference_neq_constant_propagator(int constant) : m_constant(constant) {}
    virtual int count_params() const { return 2; }
    virtual bool init( constraint_runtime_info & info );
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual bool constraint_check( const constraint_runtime_info & info) const;
    virtual void print(std::ostream & out,
                       const constraint_runtime_info *p_info) const throw();
    virtual difference_neq_constant_propagator* clone() const
    { return new difference_neq_constant_propagator(*this); }
  private:
    const int m_constant;
};

//! Propagate the semantics of: (v0 \in set) || v1 != v2.
class literal_or_neq_propagator
    : public literal_or_constraint_propagator<literal_or_neq_propagator,
                                              tight_constraint_propagator>
{
    typedef literal_or_constraint_propagator<literal_or_neq_propagator,
                                             tight_constraint_propagator>parent;
    friend class literal_or_constraint_propagator<
        literal_or_neq_propagator, tight_constraint_propagator>;
  public:
    literal_or_neq_propagator(const iset &literal): parent(literal) {}
    virtual std::unique_ptr<explanation_rule> rule(
        const std::vector<iset> & literals) const;
    virtual literal_or_neq_propagator* clone() const
    { return new literal_or_neq_propagator(*this); }
    virtual int count_params() const { return 3; }
  private:
    class literal_or_neq_rule;

    virtual bool init_non_literal_part( constraint_runtime_info & info);

    virtual
    result_type propagate_non_literal_part(constraint_runtime_info & info);

    virtual result_type state_of_non_literal_part(
        const constraint_runtime_info & info) const;

    virtual void print_non_literal_part(
        std::ostream & out,
        const constraint_runtime_info *p_info) const throw();
    
    //! For testing
    bool
    constraint_check_non_literal_part(const constraint_runtime_info &info)const
    { return info.domain(1) != info.domain(2); }
    
}; // literal_or_neq_propagator


} // namespace hcsp

#endif // HCSP_DIFF_PROPAGATORS_HPP_
