// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2011, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2011, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_OPB_CONJUNCTION_HPP
#define HCSP_OPB_CONJUNCTION_HPP

#include <iosfwd>
#include <vector>
#include <map>
#include <iset.hpp>
#include <operator_misc_types.hpp>

namespace hcsp {
class solver;
namespace opb {

class variable;
class context;
/**
 * @brief Maintains a conjunction of literals.
 *
 * This is a parsing building block that contains a list of literals.  Each
 * literal is either a variable or its negation. Applying an AND on these
 * literals produces a conjunction of literals.
 */
class conjunction
{
  private:
    enum { NO_VALUE=0 };
  public:
    /**
     * @brief Create a conjunction containing only one literal.
     *
     * @param polarity If positive then the literal is this variable,
     *                 if negative then the literal is a negation of the
     *                 variable.
     * @param variable The variable of this literal.                 
     */
    conjunction(int polarity, const variable & variable);
    ~conjunction();
    /**
     * @brief Add one literal to this conjunction of literals
     * @param polarity If positive then the literal is this variable,
     *                 if negative then the literal is a negation of the
     *                 variable.
     * @param variable The variable of this literal.
     */
    void append_literal(int polarity, const variable & variable);

    /**
     * @brief Get a variable for the conjunction multiplied by a coefficient.
     *
     * Get a variable that contains the evaluation of:\n
     *   coefficient * (apply_not \b xor eval(conjunction))
     *
     * @param context     Problem's context to produce this variable
     * @param apply_not   Do we apply logical negation on the conjunction 
     * @param coefficient The coefficient multiplying the conjunction
     * @param bound       How is the value of the conjunction is to be bounded?
     *                    From above, below, or both.
     * @param atom_name   If possible and not NULL then the returned variable
     *                    will have this name
     * @return     The index of the CSP variable which will hold atom's value.
     */
    unsigned get_atom_variable(context & context, int coefficient,
                               bool apply_not, eval_type bound,
                               const char *atom_name= NULL);

    //! Is this conjunction trivially false?
    bool is_false() const { return m_is_false; }

    //! Normalize this conjunction, making it easier to compare with others
    void normalize();

    /**
     * @brief Get a negation of this conjunction.
     *
     * If the negated conjunction can't be represented, due to limitations in
     * the data-structure, then a NULL is returned.
     *
     * @param context The context of the parser with all relevant information
     * @return A negated conjunction, or NULL if it can't be represented.
     */
    conjunction *get_not(context & context) const;
    //! Return the hash value of this conjunction
    size_t hash() const;

    //! Get the number of literals in the conjunction
    size_t size() const { return m_literals.size(); }

    //! For the indexed literal return true if equal to its variable.
    bool is_positive_literal(unsigned index) const
    { return m_literals.at(index).first > 0; }
    
    //! For the indexed literal return its variable.
    const variable & variable_for_literal(unsigned index) const
    { return *m_literals.at(index).second; }

    //! Is the conjunction a negated literal or of set of literals.
    bool is_negated_literal() const
    { return size() == 1 && m_literals[0].first < 0; }

    /** @brief Evaluate the conjunction in the given context.
     *
     * @pre The variables in the conjunction must have fully assigned values
     * @param context  Evaluate the conjunction is this context
     * @return The value of the conjunction in the context
     */
    bool evaluate(const context & context) const;
    //! Tell whether the two conjunctions are equivalent.
    friend bool operator==(const conjunction &a, const conjunction &b)
    { return a.m_literals == b.m_literals && a.m_is_false == b.m_is_false; }

    /**
     * @brief Use strict strong ordering to compare the two conjunctions.
     *
     * Put similar variables together first, and only then compare literal
     * polarity
     */
    friend bool operator<(const conjunction &a, const conjunction &b);
    //! Print the conjunction in text form
    friend std::ostream& operator<<(std::ostream&out, const conjunction&);
  private:
    
    /**
     * @brief What value the atom gets according to conjunction's value.
     */
    struct atom_values_t
    {
        enum { NO_VALUE=0 }; //!< This variable is not initialized
        atom_values_t() : m_on_zero(NO_VALUE), m_on_one(NO_VALUE) {}

        /** @brief Construct according to the coefficient of the atom
         *
         * @param coefficient The coefficient of the atom.
         * @param apply_not   Is the conjunction to be logically negated?
         */
        atom_values_t(int coefficient, bool apply_not)
        {
            m_on_zero= apply_not ? coefficient : 0          ;
            m_on_one=  apply_not ? 0           : coefficient;
        }

        //! Has this structure been initialized?
        bool has_values() const { return m_on_zero != m_on_one; }

        //! Set a new value to this structure.
        void set(int coefficient)
        {
            m_on_one= coefficient;
        }

        //! Compare two values allowing it to be used as map key.
        friend bool operator<(const atom_values_t &a, const atom_values_t &b)
        {
            return ( a.m_on_zero < b.m_on_zero ||
                     (a.m_on_zero == b.m_on_zero && a.m_on_one < b.m_on_one) );
        }
        
        int m_on_zero; //!< How to evaluate the atom if the conjunction is false
        int m_on_one;  //!< How to evaluate the atom if the conjunction is true
    }; // struct atom_values_t
    
    /**
     * @brief Information on the connection between an atom and its conjunction.
     *
     * This tells us:
     *  - Which variable represents the value of the atom
     *  - How strong the atom and the conjunction are connected, i.e.,
     *    when is it required to propagate values.
     */
    struct atom_connectivity_t
    {
        atom_connectivity_t() : m_variable(-1U),
                                m_zero_affects_atom(false),
                                m_one_affects_atom(false)
        {}
        /**
         * @brief Set the connectivity based on the polarity of the coefficient
         *        and the constraint on the atom.
         *
         * The atom, which is coefficient*conjunction, is bounded by \p bound.
         * This constructor decides how the 0/1 values of the conjunction
         * interact with the outer world when constrained by the bound.
         * @param coefficient  The coefficient of the atom
         * @param apply_not    Is the conjunction to be logically negated?
         * @param bound        The way the atom is constrained.
         */
        atom_connectivity_t(int coefficient, bool apply_not, eval_type bound);
        //! Make a strong constraint, i.e., bidirectional, between atom's
        //! \p variable and the conjunction.
        void set_equals_variable(unsigned variable);

        /*
         * @brief Check if this object defines constraints which are at least as
         *        strong as those defined by \p other.
         *
         * The comparison ignores the IDs of the variables and acts as if they
         * are the same.
         * 
         * @param other The connectivity object to compare with.
         * @return If there is a full assignment which does not satisfy
         *         \p other but does satisfy \p this.
         */
        bool is_as_strong_as(const atom_connectivity_t & other) const;
        
        unsigned m_variable;      //!< The variable of the atom or -1U      
        bool m_zero_affects_atom; //!< Does a false conjunction affect the atom?
        bool m_one_affects_atom;  //!< Does a true  conjunction affect the atom?
    }; // struct atom_connectivity_t 
    typedef std::pair<int, const variable*> literal_type;

    //! Mapping between the coefficient of the atom and its connectivity
    typedef std::map<atom_values_t, atom_connectivity_t> variable_collection_t;
    class literal_variable_comparator;
    
    //! Blocking copy constructor to avoid copying CSP variable data 
    conjunction(const conjunction&);

    //! Blocking assignment operator to avoid copying CSP variable data
    conjunction& operator=(const conjunction&);

    void insert_clause(solver & solver,
                       unsigned variable_a, const iset & set_a,
                       unsigned variable_b, const iset & set_b) const;
    unsigned variable_for_coefficient(context&context,
                                      const atom_values_t & values,
                                      const atom_connectivity_t &connectivity,
                                      const char *atom_name);

    /**
     * @brief Strengthen the constraints between the conjunction and the
     *        literals.
     *
     * The evaluation of the conjunction is maintained by m_main_connectivity.
     * The constraints between m_main_connectivity's variable and the literals
     * are sometimes weak. It means that sometimes when the conjunction should
     * be false (or true) then the variable is updated, but when it should
     * be true (or false) then the variable is untouched.
     *
     * This method checks the strength of m_main_connectivity's constraining
     * and adds HaifaCSP constraints, if required by \p connectivity.
     *
     * @param context  The context of the parser with all relevant information
     * @param connectivity  The minimum strength of constraints connecting
     *                      m_main_connectivity's variable and the literals.
     * @post The variable is constrained, at least, according to \p connectivity
     *       specification on top of previous constraints for this variable.
     */
    void strengthen_conjunction_result(context& context,
                                       const atom_connectivity_t &connectivity);

    /**
     * @brief Strengthen the link between a variable and the main variable
     *
     * Consider how variable of \p connectivity is bound with the variable of
     * m_main_connectivity. Add constraints if \p required_connectivity
     * requires them, also updating \p connectivity if necessary.
     *
     * @param context[in,out]       The context of the parser with all relevant
     *                              information.
     * @param values[in]            How the values of the conjunction translate
     *                              to the value of the atom.
     * @param required_connectivity[in] The minimum required constraint strength
     * @param connectivity[in,out]  The connectivity of the variable before and
     *                              after \p required_connectivity is considered
     * @post The variable is constrained, at least, according to
     *       \p required_connectivity specification on top of previous
     *       constraints for this variable.
     */
    void strengthen_variable(context& context,
                             const atom_values_t & values,
                             const atom_connectivity_t &required_connectivity,
                             atom_connectivity_t & connectivity);
    
    /**
     * @brief Add constraint !literal -> atom=0
     *
     * Add the constraint (variable!=0 -> conjunct) constraining the conjunct
     * only from below. So that if variable!=0 is false then the conjunct is
     * not constrained, but if variable!=0 is true then the conjunct is
     * asserted.
     *
     * @param solver        The solver to constraint
     * @param atom_variable The getting the value of the atom
     */
    void a_false_literal_implies_zero_atom(solver &solver,
                                           unsigned atom_variable) const;
    
    /**
     * @brief Add constraint such that if all literals are true then the
     *        conjunction is true.
     *
     * Add the constraint
     *  \f[ ({\mathit literal}_1 \land {\mathit literal}_2 \land\cdots )
     *       \Longrightarrow {\mathit conjunction}  \f]
     * Because
     * there is no conjunction variable then the modification is done, instead,
     * on atom_variable accordingly.
     *
     * @param solver        The solver to constraint
     * @param atom_variable The getting the value of the atom
     */
    void true_literals_imply_conjunction(solver &solver,
                                         unsigned variable) const;
    
    std::vector<literal_type> m_literals;

    /**
     * @brief The values of the atom we get for possible conjunction evaluations
     *
     * Each conjunction has one main atom, with its values and connectivity.
     * All other atoms sharing this conjunction are derived from the main
     * atom. This variable stored the mapping from the false/true evaluation
     * of the conjunction to the evaluation of the atom.
     */
    atom_values_t             m_main_values;

    //! The connectivity, including constraint strength, of this conjunction
    //! to its main variable. (see conjunction::m_main_values for more details)
    atom_connectivity_t       m_main_connectivity;
    //! Variables of all atoms created for this conjunction. 
    variable_collection_t     m_variables;
    
    bool m_is_false; //!< Is this conjunction trivially false
}; // class hcsp::opb::conjunction

} // namespace hcsp::opb
} // namespace hcsp
#endif // HCSP_OPB_CONJUNCTION_HPP
