// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2011, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2011, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_OPB_CONSTRAINT_HPP
#define HCSP_OPB_CONSTRAINT_HPP

#include <iosfwd>
#include <memory>
namespace hcsp {
class solver;
namespace opb {
class expression;
class context;

/**
 * @brief A single constraint of the OPB language.
 *
 * This is formula constraining one expression. In OPB this is a single line
 * that ends with a semicolon (;). \n
 * For example:
 * @code
 *   +1 a ~b +5 c = 5;
 *   +1 a ~b +5 c >= 6;
 * @endcode
 */
class constraint
{
  public:
    //! The direction of the constraint (check its specific values for details)
    enum direction {
        GE, //!< The constraint is of type "expression >= constant;"
        EQ  //!< The constraint is of type "expression = constant;"
    };
    /**
     * @brief Create an equality constraint from an expression.
     *
     * @param expression     The expression to constrain.
     *                       The constraint assumes ownership of the expression,
     *                       i.e., it will delete the expression eventually.
     * @param constraint_value The value by which to constrain the expression
     * @param dir  The direction of the constraint:
     *             either "expression=value;" or  "expression>=value;"
     */
    constraint(expression *expression, int constraint_value, direction dir= EQ)
        : m_expression(expression), m_value(constraint_value), m_direction(dir),
          m_tautology(false), m_falsity(false)
    {}
    //! Free all the expression of this constraint
    ~constraint();

    /**
     * @brief Create the CSP constraints required for this OPB constraint.
     *
     * @param context The context of the problem and solver to be constrained.
     * @return       True on success. If it is trivial to find that the
     *               constraint is conflicting then return false.
     */
    bool apply(context & context) const;

    //! Normalize the constraint making it canonical.
    void normalize(context & context);
    //! Print the constraint in text form
    friend std::ostream& operator<<(std::ostream&out, const constraint&);
    //! Print the direction in text form, i.e. '>=' or '='
    friend std::ostream& operator<<(std::ostream&out, direction direction);
  private:
    //! Block copying, not implemented
    constraint(const constraint&);
    //! Block assigning, not implemented
    constraint& operator=(const constraint&);

    //! Do we know that the expression is trivially true?
    bool tautology() const { return m_tautology; }

    //! Do we know that the expression is trivially false?
    bool falsity() const   { return m_falsity;   }
    
    std::unique_ptr<expression> m_expression; //!<The expression to constrain.
    int m_value;           //!<The value by which to constrain the expression
    direction m_direction; //!< The direction of the constraint ('>='  '=')
    bool m_tautology;      //!< Is this constraint always satisfied?
    bool m_falsity;        //!< Is this constraint always unsatisfied?
}; // class hcsp::opb::constraint
} // namespace hcsp::opb
} // namespace hcsp
#endif // HCSP_OPB_CONSTRAINT_HPP
