// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2011, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2011, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef OPB_CONTEXT_HPP
#define OPB_CONTEXT_HPP

#include <iostream>
#include <set>
#include <string>
#include <vector>
#include <iset.hpp>
#include <opb_variable.hpp>

namespace hcsp {
class solver;
namespace opb {
class constraint;
class expression;
class conjunction;
/**
 * @brief The OPB parsing context.
 *
 * This object contains whatever the parser has to know during OPB parsing.
 * Among other things it contains the objective function and the constraints.
 */
class context
{
  public:
    //! For objective functions, what are their types.
    enum direction {
        NO_OBJECTIVE,       //!< There is no objective function
        MAX_OBJECTIVE,      //!< A maximization objective function
        MIN_OBJECTIVE };    //!< A minimization objective function
    void* m_scanner;        //!< the scanner, i.e., lexer state
    std::istream* m_is;     //!< The input stream which the lexer reads
    std::string m_filename; //!< The file-name associated with the input stream.

    /**
     * @brief The variables discovered so far during parsing.
     *
     * This is container where the variable objects are allocated. The rest of
     * the parser points to these variables.
     */
    std::set<variable> m_variables;

  public:
    /**
     * @brief Construct the context used for parsing and solving the problem
     *
     * @param solver   The solver that will eventually solve the problem
     * @param is       The input stream from which the problem is read
     * @param filename The file, if not empty, from which the stream is read
     *
     */
    context(solver & solver,
            std::istream* is = &std::cin, const std::string & filename="");

    //! Delete all allocated data
    virtual ~context();
    
    //! Get the solver used in this context
    solver & get_solver() { return m_solver; }
    
    //! Get the solver used in this context
    const solver & get_solver() const { return m_solver; }

    //! Should the solver print information regarding to problem parsing
    bool trace_parsing() const;
    
    //! Should the solver print information regarding to problem construction
    bool trace_construction() const;

    //! Should the solver print information regarding to failures
    bool trace_failure() const;
    
    /**
     * @brief An additional constraint to record.
     * @param p_constraint A constraint to record. This context gets the
     *                     ownership of the constraint object.
     */
    void add_constraint(constraint *p_constraint);

    /**
     * @brief Set the objective function.
     *
     * If there is an objective set already, i.e., this method has been called
     * before, then the old objective expression will be freed.
     *
     * @param objective  The expression for the objective function. This
     *                   context gets the ownership of the constraint object.
     */
    void set_objective(expression *objective, direction bound);

    //! Tell if the objective has been set for this problem
    bool has_objective() const { return m_objective; }

    /**
     * @brief Evaluate the objective function in this context.
     *
     * If the value of the objective function is too big or too small to
     * fit the return type then return the saturated result. In other words,
     * if the value is too small then return the minimal possible value
     * and if it too big then return the maximal possible value.
     *
     * @return the saturated evaluation of the objective function. 
     */
    iset::value_type evaluate_objective() const;
    /**
     * @brief Get or construct a OPB variable by its name
     *
     * Look for the variable in the m_variables container. If not found,
     * create it.
     *
     * @param name The name of the variable we look for. The pointer is
     *             allocated by malloc, and its ownership is passed to
     *             the context. I.e, the context will free it eventually.
     * @return variable A OPB variable with the required name
     */
    const variable& get_variable(char *name);

    /**
     * @brief Record and get a unique conjunction equivalent to the input one.
     *
     * Get the ownership to the input conjunction, freeing it at some point.
     * The input conjunction will be freed right-away if an equivalent exists
     * in the context, otherwise it will freed together with the context object.
     * In order to compare the conjunction with existing objects it is first
     * normalized.
     *
     * @param conjunction  The conjunction to look up and possibly record
     * @return An equivalent conjunction recorded earlier or just now.
     */
    conjunction & record_unique_conjunction(conjunction *conjunction);
    
    /**
     * @brief Construct the CSP problem based on the parsed OPB problem.
     *
     * Go over all the OPB constraints and them one by one to the CSP problem
     *
     * @return True if the problem was constructed correctly. False if
     *         it can't be solved, i.e., there is a trivial conflict.
     */
    bool build_solver();

    //! Print the solution variables in human readable form.
    void print_solution(std::ostream & out) const;

    //! Get the variable storing the evaluation of the objective expression
    unsigned objective_var() const { return m_objective_var; }
    
    //! Get the type of the objective function (min/max/none).
    direction objective_type() const { return m_direction; }
    //! Print the whole problem parsed so far
    friend std::ostream &operator<<(std::ostream &out, const context &context);
    
  protected:
    //! Defined in lexter.lpp
    void init_scanner();   
    void destroy_scanner();
  private:
    class conjunction_hash_entry;
    //! Block copying, not implemented
    context(const context&);
    //! Block assigning, not implemented
    context& operator=(const context&);

    //! Connect the maximization objective function
    bool connect_maximization();
    
    //! Connect the minimization objective function
    bool connect_minimization();
    
    std::set<conjunction_hash_entry> m_conjunctions;

    /**
     * @brief The constraints discovered so far during parsing.
     *
     * This is the container of all constraints which the parser has
     * discovered. It is the owner of these pointers, i.e., it will eventually
     * delete these pointers.
     */
    std::vector<constraint*> m_constraints;

    //! The direction of the objective function, if there is an objective.
    direction m_direction;
    
    //! The objective function, or NULL if this is a satisfiability problem.
    expression *m_objective;

    //! The index of the CSP variable to hold the evaluation of the objective
    //! function, otherwise it is set to -1U.
    unsigned m_objective_var;

    solver & m_solver; //!< The solver used in this context
}; // class hcsp::opb::context
} // namespace hcsp::opb
} // namespace hcsp
#endif // OPB_CONTEXT_HPP
