// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2011, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2011, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_OPB_EXPRESSION_HPP
#define HCSP_OPB_EXPRESSION_HPP
#include <vector>
#include <iosfwd>
namespace hcsp {
namespace opb {

class conjunction;
class context;
/**
 * @brief A numeric expression of the opb language.
 *
 * This is the summation of atoms. Each atom is a conjunction of literals
 * multiplied by a coefficient.
 */
class expression
{
 public:
    //! The way the expression should be constrained.
    enum constraint_type {
        CONSTRAIN_EQUAL,   //!< Constrain the expression with equality
        CONSTRAIN_AT_LEAST,//!< Constrain the expression to have a minimal value
    };

    /**
     * @brief Create an expression from a single conjunction.
     *
     * @param coefficient The coefficient to multiply by the conjunction
     * @param conjunction A conjunction of literals.
     */
    expression(int coefficient, conjunction &conjunction);

    //! Free all the conjunctions appended to this expression
    ~expression();
    
    /**
     * @brief Add an atom to the expression.
     *
     * This will modify the expression to have one more additive, one
     * more atom.
     *
     * @param coefficient The coefficient to multiply by the conjunction
     * @param conjunction A conjunction of literals.
     */
    void append_atom(int coefficient, conjunction& conjunct);

    //! Negate all the coefficients of expression's atoms
    void negate_coefficients();

    /**
     * @brief Get a variable for the expression.
     *
     * @param context Problem's context to produce this variable
     * @param bound  How the returned variable will constrain the expression
     *               (setting the exact value or the lower bound of the value)
     * @param expression_name
     *               If possible and not NULL then the returned variable
     *               will have this name
     * @return       The index of the CSP variable to constrain the expression.
     */
    unsigned get_evaluating_variable(context & context,
                                     constraint_type bound,
                                     const char *expression_name=NULL)
        const;

    /**
     * @brief Bring the expression into a canonical form.
     *
     * Normalization may modify the expression in a way that adds or removes
     * a constant value from its evaluation. In that case, the \p rhs parameter
     * is updated accordingly. Basically, the expression is the left-hand-side
     * (\p lhs) of the constraint and is constrained by the right-hand-size
     * (\p rhs). In fact, the constraint looks like (lhs = rhs) or like
     * (lhs >= rhs).
     *
     * @param context   The context of the parser with all relevant information
     * @param rhs       The right-hand-side equated or compared with this
     *                  expression.
     */
    void normalize(context & context, int & rhs);

    /**
     * @brief Is the expression empty and has no additives.
     *
     * If the expression is empty then it evaluates, trivially, to zero.
     * @return True if empty 
     */
    bool empty() const { return m_atoms.empty(); }

    //! Are all coefficients of all additives positive?
    bool all_coefficients_positive() const;

    /**
     * @brief Create a clause out of this expression
     *
     * Each literal in the clause is one of the conjunctions of the expression.
     * Every conjunction is represented by a literal.
     *
     * @param context The context of the parser with all relevant information
     */
    void add_or_between_conjunctions(context & context) const;

    //! Get the number of atoms (conjunctions) in the expression.
    size_t size() const;

    //! Get the integer coefficient of the indexed atom
    int atom_coefficient(unsigned index) const;

    //! Is the indexed atom formed from a negated conjunction
    bool atom_is_logically_negated(unsigned index) const;

    //! Get the conjunction of the indexed atom
    const conjunction & atom_conjunction(unsigned index) const;

    /**
     * @brief Evaluate the expression in a context.
     *
     * If the value of the expression is too big or too small for the returned
     * type then a saturated result is returned. In other words, if the value
     * is too small then return the minimal possible value and if it too big
     * then return the maximal possible value for this type.
     *
     * @return the saturated evaluation of the expression.
     */
    iset::value_type evaluate(const context & context) const;
    
    //! Print the expression in text form
    friend std::ostream& operator<<(std::ostream&out, const expression &);
  private:
    class compare_conjunctions;
    struct atom_type;
    
    //! Block copying, not implemented
    expression(const expression&);
    //! Block assigning, not implemented
    expression& operator=(const expression&);

    //! Make all atoms positive integers
    void make_atoms_positive(int & rhs);

    //! Merge every set of atoms that have the same conjunction into atom.
    void merge_similar_conjunctions();
    
    //! The conjunction pointers are owned by the context, like variables,
    //! and will be deleted by the context and not by this expression.
    std::vector<atom_type> m_atoms;
}; // class hcsp::opb::expression
} // namespace hcsp::opb
} // namespace hcsp
#endif // HCSP_OPB_EXPRESSION_HPP
