// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2011, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2011, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_OPB_VARIABLE_HPP
#define HCSP_OPB_VARIABLE_HPP

#include <iosfwd>
#include <cstring>
#include <cstdlib>

namespace hcsp {
class solver;
namespace opb {
/**
 * @brief A Boolean variable of the OPB language
 *
 * This object is responsible for linking the OPB variable name and the
 * index of the CSP variable. The association with the CSP variable is done
 * in a late stage.
 */
class variable
{
  public:
    /**
     * @brief construct a variable with a given name.
     *
     * @param in_name  The name of the variable. The ownership of the pointer
     *                 goes to the variable, i.e., it will delete it eventually.
     */
    variable(char *in_name) : m_name(in_name), m_id(-1U)        {}

    //! Copy the variable
    variable(const variable &other)
        : m_name(strdup(other.m_name)), m_id(other.m_id)
    {
        if (!m_name)
            throw_bad_alloc();
    }
    
    //! Copy the contents of the variable
    variable & operator=(const variable & other)
    {
        if (&other == this)
            return *this;
        char *dup= strdup(other.m_name);
        if (!dup)
            throw_bad_alloc();
        free(m_name);
        m_name= dup;
        m_id= other.m_id;
        return *this;
    }

    /**
     * @brief Modify a variable to have a given name.
     *
     * @param new_name The name of the variable. The ownership of the pointer
     *                 goes to the variable, i.e., it will delete it eventually.
     */
    variable & operator=(char * new_name)
    {
        char *dup= strdup(new_name);
        if (!dup)
            throw_bad_alloc();
        free(m_name);
        m_name= dup;
        m_id= -1U;
        return *this;
    }

    //! Free the maintained name
    ~variable()              { free(m_name); }

    //! Get the ID of the CSP variable, or -1U if does not exist
    unsigned id() const      { return m_id; }

    //! Get the name of the OPB variable
    const char* name() const { return m_name; }

    //! Get the hash function according to the variable name
    size_t hash() const;
    /**
     * @brief Make the link between the variable and the CSP solver
     *
     * This method updates the ID field, as created by the CSP variable.
     *
     * @param solver The solver that will be linked with this variable
     */
    void associate(solver & solver);
  private:
    //! Throws std::bad_alloc
    void throw_bad_alloc() const;

    //! The name of the OPB variable. The pointer is owned by this object,
    //! i.e., it will be deleted in the destructor.
    char *m_name;

    //! The CSP ID of the variable, or -1U if not associated yet.
    unsigned m_id;
}; // class hcsp::opb::variable

inline bool operator<(const variable &a,const variable &b)
{
    return strcmp(a.name(),b.name()) < 0;
}

} // namespace hcsp::opb
} // namespace hcsp
#endif // HCSP_OPB_VARIABLE_HPP
