// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_OPERATOR_CONNECTIVITY_HPP_
#define HCSP_OPERATOR_CONNECTIVITY_HPP_

#include <utilities.hpp>
#include <memory>
#include <equivalence_classes.hpp>
#include "operator_misc_types.hpp"

namespace hcsp {

class operator_definition_registry;
class operator_definition;
class solver;

/**
 * @brief A database containing all operators collected and their relations.
 *
 * Operators are recorded in this database and, at the end, the user
 * should explicitly ask to build constraints out of the contained operators.
 */
class operator_connectivity : private no_copy
{
  private:
    static const iset & everything()
    {
        static const iset domain=iset().complement();
        return domain;
    }
  public:
    operator_connectivity(solver & the_solver);
    ~operator_connectivity();
    /**
     * @brief Record the operator with its connectivity, if not recorded
     *        already.
     *
     * Record the operator, storing the connectivity and operator kind as
     * the key, and if such a key was already recorded then return it.
     *
     * @param record_operator
     *        The operator to record
     * @param bound
     *        The way the evaluated operator is to be bounded. This bound
     *        is combined with previously recorded bound for the same
     *        connectivity+operator key.
     * @param begin_vars,end_vars
     *        C++ container type begin-end pair pointing to connectivity
     *        variables of this operator.
     * @param default_var
     *        If default_var!=-1U, then it is the default variable to be
     *        used as output for the operator, if possible.
     * @param domain
     *        The domain of the newly created variable, or a unary constraint
     *        over an existing variable.
     * @param constraint_flags
     *        Flags created by the operator which should be stored and wait
     *        for this constraint creation. These flags are then passed to
     *        the operator_definition::create_constraints() method.
     *        Each operator defines a specific semantics for these flags.
     * @return The variable index to which operator's evaluation will be
     *         written.
     */
    unsigned record_operator(const operator_definition & recorded_operator,
                             eval_type bound,
                             const unsigned *begin_vars,
                             const unsigned *end_vars,
                             unsigned default_var,
                             const iset & domain=everything(),
                             unsigned constraint_flags= 0);
    /**
     * @brief Construct all constraints required to convey the semantics
     *        of the recorded operators.
     *
     * When operators are recorded in this data structure, no constraint is
     * created. This method is the only way to convert the recorded operators
     * to constraints within the solver.
     */
    void create_all_constraints();

    /**
     * @brief Find current domain of a variable in the solver.
     * @param variable The index of the variable to query
     * @return The current domain.
     */
    const iset & var_domain(unsigned variable) const;

    
    /**
     * @brief Get a constant variable for this value.
     *
     * If there is a variable containing this constant value then prefer to
     * return this variable. Otherwise, create a new variable with this value
     * as a sole value in its domain.
     *
     * Note that on top of solver::singleton_var() this method also records
     * the name and other information of this variable in an auxiliary
     * data structure.
     *
     * @param value The sole value of domain of the requested variable
     * @return A variable that has only the requested value in its domain.
     */
    unsigned get_const_variable(iset::value_type value)
    { return m_solver.singleton_var(value); }    

    unsigned get_equivalent(unsigned variable) const
    { return m_variable_equivalence_class.find(variable); }

    void mark_equivalent(unsigned var_a, unsigned var_b)
    { m_variable_equivalence_class.unite(var_a, var_b); }

    void vars_reserve(unsigned count)
    { m_variable_equivalence_class.reserve(count); }
  private:
    struct impl;
    struct operator_entry_key;
    struct operator_entry_data;

    solver & m_solver;
    equivalence_classes        m_variable_equivalence_class;
    std::unique_ptr<impl> m_p_impl;
}; //  operator_connectivity

} // namespace hcsp

#endif // HCSP_OPERATOR_CONNECTIVITY_HPP_
