// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_OPERATOR_DEFINITION_REGISTRY_HPP_
#define HCSP_OPERATOR_DEFINITION_REGISTRY_HPP_

#include <vector>
#include <string>
#include <utilities.hpp>
#include <iset.hpp>
#include "operator_misc_types.hpp"

#define ASSERT_SIZE(begin, end, size) do {                               \
                                          (void)end;                     \
                                          HCSP_ASSERT(end-begin == size); \
                                      } while(0)
namespace hcsp {

class solver;
class operator_connectivity;
class operator_definition_registry;


/**
 * @brief Defines the semantics of a parsed operator.
 *
 * A concrete class defines how to build the CSP once this operator is
 * encountered.
 *
 * @todo Class formula will have to be replace or encapsulate this class when
 *       class formula gets enough capabilities.
 */
class operator_definition
{
  public:
    /**
     * @brief Evaluate the operator in a context of numeric values
     *
     * @param begin_val,end_val  An STL style begin-end defining a range
     *                           of values.
     * @return The result of the evaluation operation.
     */
    virtual iset::value_type eval(const iset::value_type *begin_val,
                                  const iset::value_type *end_val) const = 0;

    /**
     * @brief Evaluate the operator symbolically.
     *
     * Operate on variable indexes, both for input and output.
     *
     * @param[in,out] state
     *            The global connectivity information collected up until now.
     * @param[in] type
     *            How this operator is going to be bounded,
     * @param[in] begin_vars,end_vars
     *            An STL stype begin-end range of variables which are passed
     *            to the operator for evaluation.
     * @param[in] default_var
     *            A variable to use as output of the operator, or -1U if
     *            no default variable is given.
     *
     * @return A variable that will contain the evaluation of this operator
     *         over the variables passed to this operator
     */
    virtual unsigned symbolic_eval(operator_connectivity & state,
                                   eval_type type,
                                   const unsigned *begin_vars,
                                   const unsigned *end_vars,
                                   unsigned default_var =-1U) const=0;

    /**
     * @brief Evaluate the operator from its output domain back to the parameter
     *
     * The purpose of this method is to handle top level operations like
     * x && y , and to figure out that both x and y should be true.
     *
     * @param[in] out_domain  The domain of the output variable.
     * @param[in] argument    The index of the argument to propagate
     * @return If it is possible to infer the value of the argument then
     *         return (value, true) otherwise return (*,false) when '*'
     *         stands for an unspecified value.
     */
    virtual std::pair<iset::value_type,bool>
    reverse_eval(const iset & out_domain, unsigned argument) const
    {
        (void)out_domain; (void)argument;
        return std::make_pair(0,false);
    }
    /**
     * @brief Find out what evaluation type is required for an argument
     *
     * Different arguments may require different evaluation types. For
     * example, for subtraction, upper_bound(a-b) is equivalent with
     * upper_bound(a)-lower_bound(b). In this example we see that different
     * arguments require different evaluation types.
     *
     * The default implementation gives the strictest evaluation type.
     *
     * @param[in,out] state
     *            The global connectivity information collected up until now.
     *            The only way the method may modify 'state' is by declaring
     *            new constants through state.get_const_variable(an_integer)
     * @param[in] type
     *            How this operator is going to be bounded,
     * @param[in] argument
     *            The index of the argument of which we want to get its
     *            evaluation type.
     * @param[in] begin_vars,end_vars
     *            An STL stype begin-end range of variables which are
     *            passed to the operator for evaluation.
     * @return How this argument is to be bounded in order to give the required
     *         bounding semantics for this operator.
     */
    virtual eval_type arg_eval_type(operator_connectivity & state,
                                    eval_type type,
                                    unsigned argument,
                                    const unsigned *begin_vars,
                                    const unsigned *end_vars,
                                    unsigned default_out) const;

    /**
     * @brief Make a canonic operator order for further processing.
     *
     * For example, v123+v5 and v5+v123 are equivalent and we may like to
     * prefer v5+v123 in both cases. By using a canonic order hashing of
     * expressions may be more efficient and facilitate more subexpression
     * reuse.
     *
     * @param[in] state
     *                The global connectivity information collected up
     *                until now.
     * @param[in,out] begin_vars,end_vars
     *            An STL stype begin-end range of variables which are
     *            passed to the operator for evaluation. This method
     *            will only reorder the values pointed by this range,
     *            maintaining the same set of values.
     */
    virtual void canonize_order(const operator_connectivity & state,
                                unsigned *begin_vars,
                                unsigned *end_vars) const;

    /**
     * @brief Construct all constraints required for the semantics of this
     *        operator.
     *
     * @param[in,out] the_solver
     *                This is the solver object in which the constraints are
     *                to be created.
     * @param[in]     type
     *                The required evaluation type and bounds.
     * @param[in]     flags
     *                Extra flags created by the operator when it entered
     *                into the definition registry and kept for its use
     *                during constraint creation. Each operator may use these
     *                flags differently.
     * @param[in]     out_variable
     *                The variable that results from the evaluation of this
     *                operator.
     *
     * @param[in]     begin_vars,out_vars
     *                An STL stype begin-end range of variables which are
     *                passed to the operator for evaluation.
     */
    virtual void create_constraints(solver & the_solver, eval_type type,
                                    unsigned constraint_flags,
                                    unsigned out_variable,
                                    const unsigned *begin_vars,
                                    const unsigned *end_vars) const=0;
    
    friend bool operator<(const operator_definition & a,
                          const operator_definition & b)
    {
        // This implementation works only because there is only one object
        // of each operator type.
        return (std::size_t)&a < (std::size_t)&b;
    }
    virtual const char *name() const = 0;
    /**
     * @brief Get the tightness of the constraint associated with the operator.
     *
     * Tightness is defined in the standard way. Think of x=op(vars) as a
     * constraint, then tightness is ratio of x,vars combinations which
     * do not satisfy x=op(vars). If all combinations satisfy the constraint,
     * then tightness=0 and if none satisfy then tightness=1, and all other
     * possibilities are in-between.
     *
     * Currently, the value of each operator is calculated by counting
     * the number of combinations for which x!=op(vars) after iterating
     * every variable -100..100. Boolean variables are iterated only 0..1
     * and nonnegative integers are iterated only 0..100.
     *
     * An example for calculating the tightness of the GE operator:
     * @code
     * const ge_definition op;
     * int args[3];
     * unsigned total=0, conflict=0;
     * for (args[0]=-100; args[0] <= 100 ; ++args[0]) {
     *     for (args[1]=-100; args[1] <= 100 ; ++args[1]) {
     *         for (args[2]=0; args[2] <= 1 ; ++args[2]) {
     *             ++total;
     *             if (op.eval(&args[0], &args[1]) != args[2])
     *                 ++conflict;
     *         }
     *     }
     * }
     * cout << op.name() << " " << (double)conflict/total << "\n";
     * @endcode
     *
     * @b Usage: The value of this function is used only, for now,
     *           in a heuristic that orders constraints in the CSP.
     */
    virtual double tightness() const=0;
  protected:
    //! Private to make sure the destructor is not called by users
    virtual ~operator_definition() {}
    friend class operator_definition_registry;
};

/**
 * @brief  A singleton class containing all operator definitions
 */
class operator_definition_registry : private no_copy
{
  public:
    /**
     * @brief Find operator definition for the said operator.
     * @param type_id The NodeType of the operator as defined in
     *        XMLParser_constants.h
     * @return A pointer to a singleton definition of the operator, if
     *         known, NULL otherwise.
     */
    static const operator_definition * find(unsigned type_id);
    
    /**
     * @brief Find operator definition for the said predicate.
     *
     * A predicate can be of the form allDifferent(x,y,z), or a complicated
     * expression such as x-y >= k. The name does not include the parameters,
     * not even constant values. This means that the name should be
     * allDifferent instead of allDifferent(x,y,z) and difference_ge instead of
     * x-y >= k.
     * 
     * @param name The name of the predicate, without variable names or const
     *             values 
     * @return A pointer to a singleton definition of the operator, if
     *         known, NULL otherwise.
     */
    static const operator_definition * find_predicate(const std::string &name);
  private:
    typedef std::vector<std::pair<unsigned, operator_definition*> >container_t;
    typedef std::vector<std::pair<std::string, operator_definition*>
                        > predicate_container_t;
    operator_definition_registry();
    ~operator_definition_registry();

    // Get the singleton registry. The user of this class should not call this
    // method directly. It is intended to be used by other static methods of
    // this class, such as find_predicate() and find().
    static const operator_definition_registry & get();
    
    //! A sorted vector of NodeType,operator_definition* pairs.
    container_t m_definitions;
    //! A sorted vector of string,operator_definition* pairs.
    predicate_container_t m_predicate_definitions;
};

} // namespace hcsp

#endif // HCSP_OPERATOR_DEFINITION_REGISTRY_HPP_
