// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2012,2014, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2012,2014, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef PARSER_CONTEXT_HPP
#define PARSER_CONTEXT_HPP

#include <AST.hh>
#include <AST_builder.hpp>
#include <iostream>
#include <set>
#include <string>
#include <vector>
#include <memory>
#include <iset.hpp>
#include <utilities.hpp>
#include <table_propagators.hpp> //!< \todo remove this include

namespace hcsp {
class solver;
/**
 * @brief The parsing context for the "c like" syntax and maybe others as well
 *
 * This object contains whatever the parser has to know during parsing.
 * Among other things it contains the objective function and the constraints.
 */
class parser_context : private no_copy
{
  public:
    struct symbol_info;
  private:
    typedef CSPXMLParser::AST AST;
    typedef std::map<std::string, symbol_info> name_to_symbol_map;

  public:
    //! The original type of the element used mostly for human-readable text.
    static constexpr int SET_TYPE = 4; // Not to be used directly
    enum element_type {
        INT_TYPE= 1,
        BOOL_TYPE = 2,
        INT_SET_TYPE =  SET_TYPE | INT_TYPE,
        BOOL_SET_TYPE = SET_TYPE | BOOL_TYPE,
        UNKNOWN_TYPE = 8
    }; // enum element_type
    /**
     * @brief An abstract class for parser_context construction and destruction.
     *
     * A concrete class that inherits from scanner_methods lets the lexer
     * register or unregister the scanner state.
     */
    class scanner_methods : private no_copy
    {
      public:
        virtual void start_scanner(parser_context &context) const = 0;
        virtual void destroy_scanner(parser_context &context) const = 0;
    }; // class parser_context::scanner_methods
    friend class scanner_methods;
    
    typedef AST_builder::usage_type usage_type;
    //! For objective functions, what are their types.
    enum direction {
        NO_OBJECTIVE,       //!< There is no objective function
        MAX_OBJECTIVE,      //!< A maximization objective function
        MIN_OBJECTIVE };    //!< A minimization objective function

    //! An indicator for parser position (file+line) to be used for error
    //! reporting and, possibly, variable naming.
    struct position_point {
        position_point(const std::string *file_name,
                       unsigned line_number, unsigned column_number)
            : m_file_name(file_name),
              m_line_number(line_number), m_column_number(column_number)
        {}

        position_point() : m_file_name(nullptr), m_line_number(0),
                           m_column_number(0)
        {}
        const std::string * m_file_name;
        unsigned m_line_number;
        unsigned m_column_number;
        friend std::ostream & operator<<(std::ostream & out,
                                         const position_point & point);
    }; // struct parser_context::position_point
    
    struct typed_iset
    {
        hcsp::iset m_domain;
        element_type m_type;
    }; // struct parser_context::typed_iset

    struct parameter_type {
        std::unique_ptr<typed_iset> m_typed_set;
        int m_dimension; // -1 for scalar, non-negative for arrays
    }; // struct parser_context::parameter_type

    struct var_or_value {
        unsigned m_var_id;
        int32_t m_value;
        unsigned to_id(parser_context & context) const;
    }; // struct parser_context::var_or_value

    struct var_or_set {
        var_or_set() {}
        var_or_set(const var_or_value& other) { *this= other; }
        var_or_set(const var_or_set&)= default;
        var_or_set(var_or_set&&)= default;
        var_or_set & operator=(const var_or_set &) = default;
        var_or_set & operator=(var_or_set &&) = default;
        var_or_set & operator=(const var_or_value & other)
        {
            m_var_id= other.m_var_id;
            m_set.assign(other.m_value);
            return *this;
        }
        unsigned m_var_id;
        iset m_set;
    }; // struct parser_context::var_or_set

    typedef std::vector<var_or_value> var_or_value_list;
    typedef std::vector<var_or_set> var_or_set_list;
    typedef std::vector<std::pair<int,int> > array_dimensions;

    //! Abstract base class for constraint argument and for annotation argument
    class argument_type
    {
      public:
        virtual ~argument_type() {}
        virtual argument_type* clone() const = 0;
    }; // class parser_context::argument_type
    typedef std::vector<std::unique_ptr<argument_type> > argument_list;
    class array_argument;           // extends argument_type
    class ast_argument;             // extends argument_type
    class ast_array_argument;       // extends argument_type
    class var_argument;             // extends argument_type
    class array_dimension_argument; // extends argument_type   
    class annotation_argument;      // extends argument_type

    struct annotation_type;
    typedef std::vector<annotation_type> annotation_list;
    class constraint_type;


    struct postponed_constraint {
        position_point m_point;
        constraint_type * m_constraint; // not owner
        argument_list m_arguments;
        unsigned m_count_instantiated;
    };
    
    struct symbol_info {
        void print(std::ostream & out, const std::string & name) const;
        var_or_set_list m_var_or_set_list;
        annotation_list m_annotations;
        position_point m_location;
        element_type m_type;
        std::shared_ptr<postponed_constraint> m_defined_by;
        std::shared_ptr<AST> m_defined_by_ast;
        bool m_is_array;
        bool m_is_intermediate;
    }; // class parser_context::symbol_info
    
    typedef std::pair<std::string, position_point> parse_var;
    typedef std::vector<parse_var> parse_var_list;

    void* m_scanner;        //!< the scanner, i.e., lexer state
    
    std::istream* m_is;     //!< The input stream which the lexer reads
    std::string m_filename; //!< The file-name associated with the input stream.
    
  public:
    /**
     * @brief Construct the context used for parsing and solving the problem
     *
     * @param methods  The lexer uses these methods to construct/destruct iteslf
     * @param solver   The solver that will eventually solve the problem
     * @param is       The input stream from which the problem is read
     * @param filename The file, if not empty, from which the stream is read
     *
     */
    parser_context(solver & solver,
                   const scanner_methods & methods,
                   table_propagator_base::memory_manager & manager,
                   std::istream* is = &std::cin,
                   const std::string & filename="");

    //! Delete all allocated data
    virtual ~parser_context();
    
    //! Get the solver used in this context
    solver & get_solver() { return m_solver; }
    
    //! Get the solver used in this context
    const solver & get_solver() const { return m_solver; }

    //! Mark if all_different should be automatically detected and constructed.
    void detect_all_diff_enabled(bool flag)
    { m_AST_builder.detect_all_diff_enabled(flag); }
    void instantiate_all_different(std::vector<unsigned> & variables)
    { m_AST_builder.instantiate_all_different(variables); }

    /**
     * @brief Instantiate the semantics of vars[index] == result.
     *
     * The verbose semantics of the instantiated constraints is:
     *  V(vars[V(index)-first_pos]) == V(result).
     *
     * Where V(var_id) is the value of variable referred by var_id.
     *
     * @param debug_stream  A stream to which to print warnings
     * @param point      The position of the constraint in the input stream.
     * @param index      The variable that defines the index into the array.
     * @param first_pos  The index at the beginning of the array,,
     * @param vars       The variables, or values, of the array to construct.
     * @param result     The ID of the result variable.
     */
    void instantiate_element_constraint(std::ostream & debug_stream,
                                        const position_point & point,
                                        unsigned index,
                                        iset::value_type first_pos,
                                        const var_or_value_list & vars,
                                        unsigned result);
    /**
     * @brief Make sure the variables in 'vars' do not repeat.
     *
     * If a singleton var is repeated n times then this method might
     * create n-1 new variables, or find other n-1 singleton variables which
     * contain the same value.
     *
     * @todo make it work correctly when 'vars' have repeating non-singleton
     *       variables.
     * @param[in,out]  The context where the variables are registered.
     * @param[in,out]  The list of variables which may need fixing.
     */
    void make_unique_vars(std::vector<unsigned> & vars);

    unsigned make_aliased_var(unsigned var);

    void instantiate_weighted_sum(const std::vector<int> & coefficients,
                                  const std::vector<unsigned> & vars,
                                  bool is_ge, int k)
    { m_AST_builder.instantiate_weighted_sum(coefficients, vars, is_ge, k); }
    /* @brief define the new variables, according to the list.
     *
     * @param debug_stream  Debugging, tracing, and error reporting will
     *                      be done on this stream.
     * @param domain        The domain to be used with all the variables in
     *                      var_list.
     * @param var_list      The list of variables to define.
     * @param literal       Define the new variable according to this literal
     *                      copying its value or variable, if not nullptr.
     * @param annotations   Annotation list of the variables
     * @param type          The original type of the variable.
     * @return  True if all went well, false if there was a parse error.
     */
    bool define_variables(std::ostream & debug_stream,
                          const iset & domain,
                          const parse_var_list & var_list,
                          const var_or_value* literal= nullptr,
                          const annotation_list * annotations = nullptr,
                          element_type type = INT_TYPE);

    /* @brief Define a new variable.
     *
     * Checks if a variable or an array with the same name exists. If the
     * variable is new then define it, otherwise return false
     *
     * @param debug_stream  Debugging, tracing, and error reporting will
     *                      be done on this stream.
     * @param domain        The domain of the new variable
     * @param var           The name and location of the new variable.
     * @param literal       Define the new variable according to this literal
     *                      copying its value or variable, if not nullptr.
     * @param annotations   Annotation list of the variable
     * @param type          The original type of the variable.
     * @return True if the definition went well
     */
    bool define_variable(std::ostream & debug_stream, iset domain,
                         const parse_var & var,
                         const var_or_value* literal= nullptr,
                         const annotation_list * annotations = nullptr,
                         element_type type= INT_TYPE);

    bool define_scalar_constant(std::ostream & debug_stream,
                                const parse_var & var,
                                int32_t value);
    bool define_array_constant(std::ostream & debug_stream,
                               const parse_var & var,
                               const std::vector<int32_t> & values);
    bool define_array_set_constant(std::ostream & debug_stream,
                                   const parse_var & var,
                                   const std::vector<iset> & sets);
    /* @brief define the new arrays of variables, according to the list.
     *
     * @param debug_stream  Debugging, tracing, and error reporting will
     *                      be done on this stream.
     * @param domain        The domain to be used with all the variables in
     *                      var_list.
     * @param dimension     The dimension of the array such that each array
     *                      indices will be 1..dimension
     * @param var_list      The list of variables to define.
     * @param array_literal Define the new array according to this array_literal
     *                      copying its values or variables, if not nullptr and
     *                      if it has enough elements.
     * @param annotations   Annotation list of the arrays
     * @param type          The original type of the variable.
     * @return  True if all went well, false if there was a parse error.
     */ 
    bool define_arrays(std::ostream & debug_stream, const iset &domain,
                       unsigned dimension, const parse_var_list & var_list,
                       const var_or_value_list* array_literal= nullptr,
                       const annotation_list * annotations = nullptr,
                       element_type type = INT_TYPE);

    /* @brief Define a new array
     *
     * Checks if a variable or array with the same name exists. If the array is
     * new then define it, otherwise return false.
     *
     * @param debug_stream  Debugging, tracing, and error reporting will
     *                      be done on this stream.
     * @param domain        The domain of the new variable
     * @param dimension     The dimension of the array such that the array
     *                      indices will be 1..dimension
     * @param var           The name and location of the new variable.
     * @param array_literal Define the new array according to this array_literal
     *                      copying its values or variables, if not nullptr and
     *                      if it has enough elements.
     * @param annotations   Annotation list of the arrays
     * @param type          The original type of the variable.
     * @return True if all is OK
     */
    bool define_array(std::ostream & debug_stream, iset domain,
                      unsigned dimension, const parse_var & var,
                      const var_or_value_list * array_literal= nullptr,
                      const annotation_list * annotations = nullptr,
                      element_type type = INT_TYPE);
    
    /**
     * @brief Set the objective function.
     *
     * If there is an objective set already, i.e., this method has been called
     * before, then the old objective expression will be freed.
     *
     * @param objective  The expression for the objective function. This
     *                   context gets the ownership of the constraint object.
     */
    void set_objective(std::ostream& debug_stream,
                       AST* objective, direction dir,
                       unsigned objective_var = -1U);

    //! Tell if the objective has been set for this problem
    bool has_objective() const { return m_objective.get(); }

    /**
     * @brief Evaluate the objective function in this context.
     *
     * If the value of the objective function is too big or too small to
     * fit the return type then return the saturated result. In other words,
     * if the value is too small then return the minimal possible value
     * and if it too big then return the maximal possible value.
     *
     * @return the saturated evaluation of the objective function. 
     */
    iset::value_type evaluate_objective() const
    { return m_AST_builder.evaluate(*m_objective); }



    //! Get the variable storing the evaluation of the objective expression
    unsigned objective_var() const { return m_objective_var; }
    
    //! Get the type of the objective function (min/max/none).
    direction objective_type() const { return m_direction; }
    //! Print the whole problem parsed so far
    friend std::ostream &operator<<(std::ostream &out,
                                    const parser_context &context);
    void connect_expr(unsigned variable, const AST &formula,
                      usage_type used_as= AST_builder::EQUATED);

    void equate_variables(unsigned var_a, unsigned var_b)
    { m_AST_builder.equate_variables(var_a, var_b); }
    
    void connect_parity(const std::vector<unsigned> & vars, bool result)
    { m_AST_builder.connect_parity(vars, result); }

    /** @brief Connect the named constraint to the solver.
     *
     *
     * @param[in,out] debug_stream  A stream used for error text
     * @param[in]     point         The of the constraint in the input CSP.
     * @param[in]     name          The name of the constraint to look-up
     * @param[in]     arguments     The arguments of the constraint.
     * @param[in]     annotation    If given, the list of annotations, i.e.,
     *                              attributes, affecting this constraint 
     * @return True if all is fine, false if there was a parsing error.
     */
    bool connect_constraint(std::ostream & debug_stream,
                            const position_point & point,
                            const std::string & name,
                            const argument_list & arguments,
                            const annotation_list * annotations= nullptr);
    bool has_defined_var(const argument_list & arguments) const;
    bool is_intermediate_var(unsigned id) const;

    void connect_table_propagator(
        const std::vector<unsigned> & vars,
        const std::vector<std::vector<iset::value_type> > & tuples);
 
    void connect_constraint(
        const std::vector<unsigned> & vars,
        std::unique_ptr<constraint_propagator> constraint);
    /** @brief Register a single constraint type for future use in the parser.
     *
     * When a specific language solver, such as fzn_solver, starts it should
     * register all constraint types, which the language supports, using this
     * method. Every language solver may define a different set of constraints.
     *
     * @param[in] name        The name of the constraint to register
     * @param[in] constraint  The constraint to register
     */
    void register_constraint(const std::string & name,
                             std::unique_ptr<constraint_type> constraint)
    { m_constraint_types.insert(make_pair(name, std::move(constraint))); }

    /**
     * @brief Get a variable that represents a single integer.
     *
     * For a given integer return a variable whose domain contains only this
     * integer. Depending on \p force_new, an existing variable or a newly
     * created variable is returned.
     *
     * @param[in] value      The value which the domain should have
     * @param[in] force_new  If true then a new variable is created even
     *                       if a variable with this domain exists already.
     * @return A hcsp::solver variable index with a singleton domain containing
     *         only \p value.
     */
    unsigned singleton_var(iset::value_type value, bool force_new = false);

    unsigned var_id(const std::string & name) const
    {
        auto it= m_name_to_symbol.find(name);
        if ( it == m_name_to_symbol.end() || it->second.m_is_array) {
            return -1;
        } else {
            HCSP_ASSERT(it->second.m_var_or_set_list.size() == 1);
            return it->second.m_var_or_set_list[0].m_var_id;
        }
            
    }
    unsigned array_element_id(const std::string & name, unsigned position) const
    {
        auto it= m_name_to_symbol.find(name);
        if ( it == m_name_to_symbol.end() ||
             position >= it->second.m_var_or_set_list.size() ||
             !it->second.m_is_array)
            return -1;
        else 
            return it->second.m_var_or_set_list[position].m_var_id;
    }
    bool get_const_array_element(const std::string & name, unsigned position,
                                 int32_t &value) const
    {
        auto it= m_name_to_symbol.find(name);
        if ( it == m_name_to_symbol.end() || !it->second.m_is_array ||
             position >= it->second.m_var_or_set_list.size()) {
            return false;
        }
        const auto & entry= it->second;
        const iset & values = entry.m_var_or_set_list[position].m_set;
        if ( (entry.m_type & SET_TYPE) != 0 || !values.singleton())
            return false;
        value= values.min();
        return true;
    }
    bool get_constant(const std::string & name, int32_t &value) const
    {
        auto it= m_name_to_symbol.find(name);
        if ( it == m_name_to_symbol.end() || it->second.m_is_array)
            return false;
        HCSP_ASSERT(it->second.m_var_or_set_list.size() == 1);
        const auto & entry= it->second;
        const iset & values =  entry.m_var_or_set_list[0].m_set;
        if ( (entry.m_type & SET_TYPE) != 0 || !values.singleton())
            return false;
        value= values.min();
        return true;
    }

    //! Mark, in the solver, that all user variables have been defined.
    //! From this point on, only auxiliary variables will be defined.
    void no_more_user_variables()
    {
        m_solver.no_more_user_variables();
    }
    //! Create all constraints stored in the AST_builder
    bool create_constraints(std::ostream & debug_stream);
    void mark_unsat()
    {
        m_AST_builder.mark_unsat();
    }
    iset domain(const var_or_value & var) const
    {
        if (var.m_var_id != -1U)
            return m_solver.var_domain(var.m_var_id);
        else
            return iset(var.m_value);
    }
    const iset & domain(unsigned var) const
    { return m_solver.var_domain(var); }
    void set_domain(unsigned var_index, iset new_domain)
    { return m_solver.set_var_domain(var_index, std::move(new_domain)); }

    bool report_domain_deduction() const
    { return m_solver.report_level() >= solver::PROPAGATION_MODS; }

    /** @brief Create a CSPXMLParser::AST object out of the input literal.
     *
     * @param literal May be either a constant or a variable.
     * @return to_ast a newly allocated AST node containing the data of the
     *         literal. The name of a variable node is not updated correctly.
     */
    std::unique_ptr<AST> to_ast(const var_or_value & literal) const;
    var_or_value from_ast(const AST& ast);
    std::unique_ptr<ast_array_argument> to_ast_array_argument(
        const var_or_value_list & args) const;

    std::unique_ptr<AST> ast_var(unsigned var_index) const;
    std::string var_name(unsigned var_index) const;
    
    /**
     * @brief Create a copy of argument_list without AST arguments.
     *
     * Original AST arguments are converted to regular arguments not based on
     * AST.
     */
    argument_list from_ast(const argument_list &arguments);
    static std::unique_ptr<AST> ast_value(int32_t value);

    //! Print the solution filtering variables by their annotation.
    //! \todo Make it work correctly also for solvers other than fzn_solver.
    void print_solution(std::ostream & out) const;

    unsigned expression_to_variable(
        const AST & expr,
        unsigned default_var= -1U,
        usage_type used_as= AST_builder::EQUATED);
    
  private:

    void print_values(std::ostream & out, element_type type,
                      const iset & values) const;
    void print_var_solution(
        std::ostream &out, const name_to_symbol_map::value_type & elem) const;

    void print_array_solution(
        std::ostream &out, const name_to_symbol_map::value_type & elem,
        const annotation_type & annotation) const;
    
    //! Connect the maximization objective function
    void connect_maximization();
    
    //! Connect the minimization objective function
    void connect_minimization();

    /**
     * @brief Translate an annotation argument to variable ID.
     *
     * @param debug_stream  Messages, if any, go to this stream
     * @param point         The location of the constraint or variable that
     *                      has this annotation.
     * @param arg  Either 'var[N]' or 'var' annotation argument.
     * @return If all went well return the ID of the variable, otherwise -1U.
     */
    unsigned var_from_annotation(std::ostream & debug_stream,
                                 const position_point & point,
                                 const argument_type & arg) const;


    /**
     * @brief Convert an annotation argument to symbol_info.
     *
     * @param debug_stream  Messages, if any, go to this stream
     * @param point         The location of the constraint or variable that
     *                      has this annotation.
     * @param arg  Either 'var[N]' or 'var' annotation argument.
     * @return If all went well return the symbol of the variable.
     */
    symbol_info* symbol_from_annotation(std::ostream & debug_stream,
                                        const position_point & point,
                                        const argument_type & arg);
    
    /**
     * @brief Get the list of variable IDs that the arguments define.
     *
     * The list of arguments may contain simple variables, constants,
     * and arrays. If a constant does not have an ID associated with it, then
     * create a new variable.
     *
     * @param arguments  The arguments passed to the constraint.
     * @return The list of argument IDs.
     */
    std::vector<unsigned> collect_vars(const argument_list & arguments);

    bool instantiate_var_define(std::ostream & debug_stream, unsigned id,
                                std::unique_ptr<AST> & ast);
    
    bool instantiate_postponed(std::ostream & debug_stream,
                               postponed_constraint & postponed,
                               symbol_info *defined_var,
                               std::unique_ptr<AST> & ast);

    void intersect_annotations(symbol_info &a, symbol_info &b) const;

    /**
     * @brief Update m_variable_identity_function.
     *
     * @post m_variable_identity_function[i] = i for each i < count_variables()
     */
    void update_variable_identity_function();

    
    //! The direction of the objective function, if there is an objective.
    direction m_direction;
    
    //! The objective function, or NULL if this is a satisfiability problem.
    std::unique_ptr<AST> m_objective;
    
    //! The index of the CSP variable to hold the evaluation of the objective
    //! function, otherwise it is set to -1U.
    unsigned m_objective_var;
    name_to_symbol_map m_name_to_symbol;
    std::vector<std::string> m_id_to_name;
    std::map<std::string, std::unique_ptr<constraint_type> > m_constraint_types;
    std::vector<unsigned> m_variable_identity_function;
    AST_builder m_AST_builder;
    std::vector<std::unique_ptr<postponed_constraint> > m_postponed_constraints;

    solver & m_solver; //!< The solver used in this context
     //! Methods that register/unregister the context in the lexer
    const scanner_methods & m_methods;

    table_propagator_base::memory_manager & m_manager;
}; // class hcsp::parser_context


/**
 * @brief Symbol annotations
 *
 * This is like the C++ [[attribute]] element, or the MiniZinc "::annotation",
 * which are attached to a variable or to a constraint.
 */
struct parser_context::annotation_type
{
    annotation_type() = default;
    annotation_type(const annotation_type & in) { *this= in; }
    annotation_type(annotation_type && in) = default;
    annotation_type & operator=(const annotation_type & in)
    {
        if (&in == this)
            return *this;
        m_name= in.m_name;
        m_arguments.release();
        if (in.m_arguments.get() == nullptr)
            return *this;
        m_arguments.reset(new argument_list);
        for (const auto & elem: *in.m_arguments)
            m_arguments->push_back(
                std::unique_ptr<argument_type>(elem->clone()));
        return *this;
    }
    annotation_type & operator=(annotation_type && in) = default;
    std::string m_name;
    std::unique_ptr<argument_list> m_arguments;
}; // struct parser_context::annotation_type

class parser_context::array_argument : public argument_type
{
  public:
    virtual array_argument* clone() const
    { return new array_argument(*this);}
    var_or_value_list m_array;
}; // class parser_context::array_argument

class parser_context::var_argument : public argument_type
{
  public:
    var_argument() {}
    var_argument(const var_or_value & in) : m_variable(in) {}
    
    virtual var_argument* clone() const { return new var_argument(*this);}
    var_or_value m_variable;
};  // class parser_context::var_argument

class parser_context::ast_argument : public argument_type
{
  public:
    ast_argument() {}
    ast_argument(std::unique_ptr<AST> ast) noexcept : m_ast(std::move(ast)) {}
    ast_argument(ast_argument &&) = default;
    ast_argument(const ast_argument &other) { *this= other; }
    ast_argument &operator=(ast_argument&&) = default;
    ast_argument &operator=(const ast_argument& other)
    {
        if (&other != this)
            m_ast = deep_copy(*other.m_ast);
        return *this;
    }
    ast_argument* clone() const override { return new ast_argument(*this); }
    std::unique_ptr<AST> m_ast;
};  // class parser_context::ast_argument

class parser_context::ast_array_argument : public argument_type
{
  public:
    ast_array_argument() {}
    ast_array_argument(ast_array_argument&& other) = default;
    ast_array_argument(const ast_array_argument& other) { *this= other; }
    ast_array_argument&operator=(ast_array_argument&& other) = default;
    ast_array_argument&operator=(const ast_array_argument& other)
    {
        if (this == &other)
            return *this;
        m_array.clear();
        for (auto const & ptr_ast: other.m_array) {
            if (ptr_ast)
                m_array.push_back(deep_copy(*ptr_ast));
            else
                m_array.emplace_back(nullptr);
        }
        return *this;
    }
    
    ast_array_argument* clone() const override
    { return new ast_array_argument(*this); }
    std::vector<std::unique_ptr<AST> > m_array;
};  // class parser_context::ast_array_argument

class parser_context::array_dimension_argument : public argument_type
{
  public:
    virtual array_dimension_argument* clone() const
    { return new array_dimension_argument(*this);}
    array_dimensions m_dimensions;
}; // class parser_context::array_dimension_argument

class parser_context::annotation_argument : public argument_type
{
  public:
    virtual annotation_argument* clone() const
    { return new annotation_argument(*this);}
    annotation_type m_annotation;
}; // class parser_context::annotation_argument

/** @brief An abstract constraint type used for parsing. 
 *
 * A concrete object of this type is registered into the parser_context
 * letting the solver set correct constraint names for the current
 * language. For example the fzn_solver will register FlatZinc names before
 * starting the parser. These names are then looked-up during parsing and
 * the constraint_type object tells how to create the relevant constraint.
 */
class parser_context::constraint_type
{
  protected:
    typedef parser_context::position_point position_point;
    typedef parser_context::var_or_value var_or_value;
    typedef parser_context::var_or_value_list var_or_value_list;
    typedef parser_context::annotation_list annotation_list;
    typedef parser_context::argument_list argument_list;
    typedef parser_context::var_argument var_argument;
    typedef parser_context::array_argument array_argument;
    typedef CSPXMLParser::AST AST;
    typedef CSPXMLParser::ASTInteger ASTInteger;
    typedef CSPXMLParser::ASTVar ASTVar;

    /** @brief Construct an AST node with and connect it with children.
     *
     * @tparam  The type of the AST node
     * @param[in] a,b,c  The children, where some of them are optional.
     *
     * @return The new AST node connected to the children.
     */
    template<typename T>
    static std::unique_ptr<AST> make(std::unique_ptr<AST> a,
                                     std::unique_ptr<AST> b= nullptr,
                                     std::unique_ptr<AST> c= nullptr)
    {
        std::unique_ptr<T> ret(new T);
        ret->setArg(0, a.release());
        if (b.get())
            ret->setArg(1, b.release());
        if (c.get())
            ret->setArg(2, c.release());
        return std::move(ret);
    }

    /** @brief Create a CSPXMLParser::AST object out of the input literal.
     *
     * @param literal May be either a constant or a variable.
     * @return to_ast a newly allocated AST node containing the data of the
     *         literal. The name of a variable node is not updated correctly.
     */
    std::unique_ptr<AST> to_ast(const var_or_value & literal) const;

    std::unique_ptr<AST> ast_var(unsigned var_index) const;

    static std::unique_ptr<AST> ast_value(int32_t value)
    { return parser_context::ast_value(value); }
    
  public:
    virtual ~constraint_type() {}
    /** @brief Connect the constraint to the solver.
     *
     * @param[in,out] debug_stream  A stream used for error text
     * @param[in]     point         The of the constraint in the input CSP.
     * @param[in,out] context       The parsing context at which to populate
     *                              this context.
     * @param[in]     arguments     The arguments of the constraint.
     * @param[in]     defines_var   This is the variable that
     *                              the constraint defines.
     * @return True if all is fine, false if there was a parsing error.
     */
    virtual bool connect_constraint(
        std::ostream & debug_stream, const position_point & point,
        parser_context & context, const argument_list & arguments,
        symbol_info *defined_var, std::unique_ptr<AST> & ast) = 0;
}; // class parser_context::constraint_type

} // namespace hcsp
#endif // PARSER_CONTEXT_HPP
