// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

// -*- c-basic-offset:4 -*-
#ifndef HCSP_POOL_ALLOC_HPP_
#define HCSP_POOL_ALLOC_HPP_

#include <cstddef>

namespace hcsp {
struct pool_alloc_info
{
  public:
    // Avoid constructors and destructors in order to keep this as a POD    
    char *pop()
    {
#ifdef HCSP_VALGRIND
        return new char[m_entry_size];
#endif
        char *ret= m_free_list;
        if (m_free_list)
            m_free_list= *reinterpret_cast<char**>(m_free_list);
        return ret;
    }
    void push(char *ptr)
    {
#ifdef HCSP_VALGRIND
        delete[] ptr;
        return;
#endif
        *reinterpret_cast<char**>(ptr)= m_free_list;
        m_free_list= ptr;
    }
    void reserve_block();

    char *m_free_list;
    unsigned m_total_size;
    unsigned m_entry_size;
    unsigned m_entries;
};

template <class T>
class pool_alloc
{
  public:
    typedef T         value_type;
    typedef T        *pointer;
    typedef const T  *const_pointer;
    typedef T        &reference;
    typedef const T  &const_reference;
    typedef std::ptrdiff_t difference_type;
    typedef std::size_t size_type;

    pool_alloc() {}
    
    template <class U>
    pool_alloc(const pool_alloc<U> &) {}

    template <class U>
    pool_alloc & operator=(const pool_alloc<U> &) {return *this;}

    template <class U>
    struct rebind { typedef pool_alloc<U> other; };

    pointer       address(reference ref)       const { return &ref; }
    const_pointer address(const_reference ref) const { return &ref; }
    pointer allocate(size_type num, const void * =0)
    {
        char *ret;
        if (num == 1) {
            ret= m_info.pop();
            if (!ret) {
                m_info.reserve_block();
                ret= m_info.pop();
            }
        } else {
            ret= new char[sizeof(T)*num];
        }
        return reinterpret_cast<pointer>(ret);
    }
    void deallocate(pointer to_free, size_type num_alloc)
    {
        char *ptr_as_char= reinterpret_cast<char*>(to_free);
        if (num_alloc == 1) {
            m_info.push(ptr_as_char);
        } else {
            delete[] ptr_as_char;
        }
    }
    size_type max_size() const
    { return (~static_cast<size_type>(0))/ALLOCATED_ENTRY_SIZE; }

    void construct(pointer ptr, const value_type & value)
    { new (reinterpret_cast<void*>(ptr))value_type(value); }

    void destroy(pointer ptr)
    { ptr->~value_type(); }

    template <class U1, class U2>
    friend bool operator==(const pool_alloc<U1> & alloc1,
                           const pool_alloc<U2> & alloc2);
                                
    
  private:
    enum { ALLOCATED_ENTRY_SIZE=
           sizeof(T) >= sizeof(char*) ? sizeof(T) : sizeof(char*)
    };
    enum { TARGET_BLOCK_SIZE = 1024-3*8 };
    enum { ENTRIES_IN_ALLOCATED_BLOCK= ( sizeof(T) <= TARGET_BLOCK_SIZE/2 ?
                                         TARGET_BLOCK_SIZE/ALLOCATED_ENTRY_SIZE:
                                         sizeof(T)) };
    enum { RESERVE_BLOCK_SIZE=
           ALLOCATED_ENTRY_SIZE*ENTRIES_IN_ALLOCATED_BLOCK
    };
    static pool_alloc_info m_info;
};

template <>
class pool_alloc<void>
{
  public:
    typedef void        value_type;
    typedef void       *pointer;
    typedef const void *const_pointer;
    typedef void        reference;
    typedef void        const_reference;
    typedef std::ptrdiff_t difference_type;
    typedef std::size_t size_type;

    template <class U>
    pool_alloc(const pool_alloc<U> &) {}

    template <class U>
    pool_alloc & operator=(const pool_alloc<U> &) {return *this;}

    template <class U>
    struct rebind { typedef pool_alloc<U> other; };

    template <class U1, class U2>
    friend bool operator==(const pool_alloc<U1> & alloc1,
                           const pool_alloc<U2> & alloc2);
                                
  private:
    // Needed for operator==.
    enum { ALLOCATED_ENTRY_SIZE= 0 };
};

template <class U1, class U2>
inline bool operator==(const pool_alloc<U1> & alloc1,
                       const pool_alloc<U2> & alloc2)
{
    return alloc1.ALLOCATED_ENTRY_SIZE == alloc2.ALLOCATED_ENTRY_SIZE;
}

template <class U1, class U2>
inline bool operator!=(const pool_alloc<U1> & alloc1,
                       const pool_alloc<U2> & alloc2)
{
    return !(alloc1 == alloc2);
}

template <class T>
// Initialized before the program starts
pool_alloc_info pool_alloc<T>::m_info={
    (char*)0,
    (unsigned)pool_alloc<T>::RESERVE_BLOCK_SIZE,
    (unsigned)pool_alloc<T>::ALLOCATED_ENTRY_SIZE,
    (unsigned)pool_alloc<T>::ENTRIES_IN_ALLOCATED_BLOCK
};
    
} // namespace hcsp

#endif // HCSP_POOL_ALLOC_HPP_
