// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_PROOF_CHECKER_VISITOR_HPP_
#define HCSP_PROOF_CHECKER_VISITOR_HPP_


#include <iostream>
#include <map>
#include <iset.hpp>
#include <proof_visitor.hpp>

namespace hcsp {

/**
 * @brief A visitor that checks the correctness of UNSAT by verifying the proof
 */
class proof_checker_visitor : public proof_visitor
{
  public:

    //! Construct the visitor with textual error reporting enabled
    proof_checker_visitor(proof_collector & collector, std::ostream & report )
        : m_collector(collector), m_is_correct(true), m_p_report(&report) { }
    
    //! Construct the visitor without textual errors
    proof_checker_visitor(proof_collector & collector)
        : m_collector(collector), m_is_correct(true), m_p_report(NULL) { }
    
    virtual void visit_explanation(int explanation_clause_id)
    {
        HCSP_ASSERT(m_clauses.find(explanation_clause_id) == m_clauses.end());
        clause_data & clause= m_clauses[explanation_clause_id];
        m_collector.get_clause(explanation_clause_id, clause);
    }

    virtual void visit_resolution(
        unsigned pivot,
        int resolvent1_id, const clause_data & resolvent1,
        int resolvent2_id, const clause_data & resolvent2,
        int consequent_id, const clause_data & consequent);
    
    virtual void visit_initial_resolution(
        unsigned variable, const iset & domain,
        int resolvent_id, const clause_data & resolvent,
        int consequent_id, const clause_data & consequent);

    //! Did everything work correctly when the proof was visited
    bool is_correct() const { return m_is_correct; }

    variable_printer var_printer(unsigned variable) const
    { return m_collector.var_printer(variable); }
  private:
    /**
     * @brief Visit a clause, validating it is consistent and memorizing it.
     * @param clause_id,clause  The clause and its ID
     * @pram is_old  Is this clause old? We work validate the consistency of
     *               clauses with their old values.
     * @post
     *   - If inconsistent, then m_is_correct= false.
     *   - If consistent then m_clauses[clause_id] equivalent with clause
     */
    void visit_clause(int clause_id, const clause_data & clause, bool is_old);

    //! The set of clauses discovered so far mapped clause ID -> clause literals
    std::map<int, clause_data> m_clauses;

    //! The proof collector which activates this visitor
    const proof_collector & m_collector;

    //! Thus far, is everything OK? No error detected so far?
    bool m_is_correct;

    //! If not NULL, then this is the stream to use for error reporting.
    std::ostream * m_p_report;
}; // class proof_checker_visitor


} // namespace hcsp

#endif // HCSP_PROOF_CHECKER_VISITOR_HPP_

