// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_PROOF_COLLECTOR_HPP_
#define HCSP_PROOF_COLLECTOR_HPP_

#include <algorithm>
#include <vector>
#include <iosfwd>
#include <solver.hpp>
#include <iset.hpp>
#include <clause_data.hpp>
#include <explanation_rule.hpp>

namespace hcsp {

class proof_visitor;
/**
 * @brief The interface of the proof collector.
 *
 * The actual object may either collect or ignored the requests to collect
 * proof related data
 */
class proof_collector_interface
{
  public:
    virtual ~proof_collector_interface() {}
    
    //! Will the proof data be collected? Is a proof required?
    virtual bool want_proof() const= 0;

    //! Assign a clause ID to the conflict clause which was generated just now.
    virtual void name_conflict_clause(unsigned inserted_conflict_clause)= 0;

    /**
     * @brief Record information related to the latest conflict
     *
     * Can be undone by calling pop_resolution_entry().
     *
     * @param[in] constraint_id
     *               The constraint that detected the conflict
     * @param[in] literals
     *               The literal sets of the latest conflicting constraint.
     *               For a non-clause constraint, it's the explanation clause.
     * @param[in] rule
     *               If not NULL, contains the rule used to derive the literals
     */
    virtual void record_conflict(unsigned constraint_id,
                                 const std::vector<iset> & literals,
                                 std::unique_ptr<explanation_rule> rule) = 0;
    /**
     * @brief Record information related to the this resolution with the
     *        currently generated conflict clause.
     *
     * Can be undone by calling pop_resolution_entry().
     *
     * @param[in] antecedent_constraint_id
     *               The constraint which is to be resolved.
     * @param[in] literals
     *               The literal sets of the antecedent clause.
     *               For a non-clause constraint, it's the explanation clause.
     * @param[in] pivot_variable
     *               The pivot variable used during the resolution.
     * @param[in] rule
     *               The resolution rule that created the literals
     */
    virtual void record_resolution(unsigned antecedent_constraint_id,
                                   const std::vector<iset> & literals,
                                   int pivot_variable,
                                   std::unique_ptr<explanation_rule> rule) = 0;

    /**
     * @brief Undo the last call to record_conflict() or to record_resolution()
     *
     * A call to this method lets the caller change his mind after calling
     * record_resolution() or record_conflict(). The method is cumulative, i.e.,
     * calling it twice will undo the previous two calls to either
     * record_conflict() or record_resolution(), or both.
     */
    virtual void pop_resolution_entry() = 0;
    /**
     * @brief Save the clause in proof_collector's data structure.
     *
     * This method is supposed to be called when the solver is about
     * to delete its copies of this clause, but before doing that it
     * saves the clause in the proof_collector which needs the clause for
     * proof generation.
     *
     * @param clause_id  The ID of the saved clause
     * @param clause     The literals of the saved clause
     */
    virtual void save_clause(unsigned clause_id, const clause_data &clause)
        = 0;

    /**
     * @brief Prepare internal data structure for printing.
     *
     * Some data structures need to be updated before they are printed and
     * after they are populated. Not only is it permitted to call this method
     * more than once, it is mandatory to call before printing if any
     * other mutating method has been called.
     */
    virtual void prepare_to_print()= 0;
    
    //! Print the collected proof.
    virtual void print_proof(std::ostream &out) const = 0;

    //! Traverse the proof graph from the roots to the empty clause
    //! \brief visitor  The visitor to consult on traversal events
    virtual void traverse_proof(proof_visitor & visitor) const = 0;
}; //  proof_collector_interface

/**
 * @brief Dummy collector class.
 *
 * This class is to be used as a dummy collector when no proof generation
 * is required. Its calls are almost free.
 */
class proof_collector_dummy : public proof_collector_interface
{
  public:
    virtual bool want_proof() const { return false; }
    virtual void name_conflict_clause(unsigned)                              {}
    virtual void record_conflict(unsigned, const std::vector<iset> &,
                                 std::unique_ptr<explanation_rule>)          {}
    virtual void record_resolution(unsigned, const std::vector<iset> &, int,
                                   std::unique_ptr<explanation_rule> )       {}
    virtual void pop_resolution_entry()                                      {}
    virtual void save_clause(unsigned, const clause_data &)                  {}
    virtual void prepare_to_print()                                          {}
    virtual void print_proof(std::ostream &) const;
    virtual void traverse_proof(proof_visitor & ) const                      {}
}; //  proof_collector_dummy

/**
 * @brief The actual proof collector
 *
 * This class collects information required to reconstruct and print a proof
 * of unsatisfiability.
 */
class proof_collector : public proof_collector_interface
{
  private:
    /**
     * @brief An entry of the resolution history.
     *
     * It contains information of one resolution, or of a conflicting clause.
     */
    struct resolve_data
    {
        /**
         * @brief The entity resolved with the front.
         *
         * - When nonnegative, a clause in the clause database.
         * - When negative, an explanation at m_explanations[-1-m_clause].
         */
        int m_clause;
        /**
         * @brief The pivot variable to resolve m_clause with the front.
         *        If this is a new front then m_pivot == -1
         */
        int m_pivot;
    }; // struct resolve_data

    class explanation_data;

  public:
    proof_collector(const solver & the_solver);
    virtual ~proof_collector();
    virtual bool want_proof() const { return true; }
    virtual void record_conflict(unsigned constraint_id,
                                 const std::vector<iset> & literals,
                                 std::unique_ptr<explanation_rule> rule)
    {
        m_latest_resolution_start= m_resolve_history.size();
        append_resolution_entry(constraint_id, literals, -1, move(rule));
    }
    virtual void record_resolution(unsigned constraint_id,
                                   const std::vector<iset> & literals,
                                   int pivot_variable,
                                   std::unique_ptr<explanation_rule> rule)
    { append_resolution_entry(constraint_id, literals, pivot_variable,
                              move(rule)); }

    virtual void name_conflict_clause(unsigned inserted_conflict_clause)
    {
        HCSP_ASSERT(m_first_conflict_clause == -1U ||
                   last_conflict_clause()+1U == inserted_conflict_clause);
        HCSP_ASSERT(!m_resolve_history.empty());
        m_clause_resolve_indices.push_back(m_latest_resolution_start);
        if (m_first_conflict_clause == -1U)
            m_first_conflict_clause= inserted_conflict_clause;
    }

    virtual void save_clause(unsigned clause_index,
                             const clause_data &clause)
    { m_saved_clauses.push_back(std::make_pair(clause_index, clause)); }

    virtual void prepare_to_print()
    { sort(m_saved_clauses.begin(), m_saved_clauses.end()); }
    
    virtual void print_proof(std::ostream &out) const;
    
    variable_printer var_printer(unsigned variable) const
    { return m_solver.var_printer(variable); }

    /**
     * @brief Traverse the proof graph from the roots to the empty clause
     * @param[in,out] visitor  The visitor that is invoked when the proof graph
     *                         is being traversed.
     * @pre prepare_to_print has been called, and the problem did not change
     *      since then. Also, the solver has return UNSAT on this problem.
     */
    virtual void traverse_proof(proof_visitor & visitor) const;

    //! Return the number of explanation clauses generated so far.
    unsigned explanation_count() const;
    
    /**
     * @brief Print the explanation which leads to the given clause.
     * @pre clause_id<0 and refers to an explanation clause ID, i.e.,
     *      there is an m_explanations corresponding with clause_id.
     */
    void print_explanation(std::ostream & out, int clause_id) const;
    
    /**
     * @brief Get the clause referred by this clause ID.
     * @pre The ID must refer a clause and not any other kind of constraint.
     *      Also, the only supported auxiliary clause is an explanation
     *      clause, the ID must not refer intermediate clauses that result
     *      from the resolution rule.
     * @param[in]  clause_id The requested clause.
     * @param[out] clause    The retrieved clause.
     */
    void get_clause(int clause_id, clause_data & clause) const;

    /**
     * @brief For this explanation, get the ID of the constraint it explains.
     *
     * @param explanation_clause_id The ID of the explanation clause we query
     * @return The ID of the constraint that this explanation explains.
     * @pre explanation_clause_id is an explanation clause of some constraint.
     */
    unsigned get_constraint_explained_by(int explanation_clause_id) const;
    
    /**
     * @brief Print this binary resolution
     *
     * @param[out] out                     The output stream
     * @param[in] pivot                    The resolution variable.
     * @param[in] resolvent1,resolvent1_id The first resolvent and its ID.
     * @param[in] resolvent2,resolvent2_id The second resolvent and its ID.
     * @param[in] consequent,consequent_id The consequent of the resolution+ID
     */
    void print_resolution(std::ostream & out, unsigned pivot,
                          int resolvent1_id, const clause_data &resolvent1,
                          int resolvent2_id, const clause_data &resolvent2,
                          int consequent_id, const clause_data&consequent)const;
    /**
     * @brief Print a binary resolution with an initial domain
     *
     * @param[out] out                     The output stream
     * @param[in] variable                 The resolution variable.
     * @param[in] initial_domain           The initial domain to resolve with
     * @param[in] resolvent,resolvent_id   The resolvent and its ID.
     * @param[in] consequent,consequent_id The consequent of the resolution+ID
     */
    void print_initial_resolution(
        std::ostream & out, unsigned variable, const iset & initial_domain,
        int resolvent_id, const clause_data & resolvent,
        int consequent_id, const clause_data & consequent) const;
    
    /**
     * @brief Tell whether the ID refers to an explanation clause.
     *
     * @param clause_id   The ID of the queried clause
     * @return True if and only if this is an explanation clause.
     */
    bool is_explanation_clause(int clause_id) const;
    
  private:
    typedef std::vector<std::pair<unsigned, clause_data> > saved_clause_set;

    class state
    {
      public:
        state(const proof_collector & collector)
            : _id(-(int)collector.explanation_count()) {}
        //! Generate a clause ID for an intermediate clause.
        int intermediate_clause_id()
        { return --_id; }
      private:
        //! The ID of the last auxiliary clause
        int _id;
    };// class proof_collector::state

    /**
     * @brief Get the index of an explanation clause.
     *
     * Get the index of this explanation clause in the m_explanations
     * vector. If clause_id does not refer to an explanation clause then
     * return an arbitrary number outside the range of m_explanations.
     *
     * @param clause_id  The ID of the explanation clause
     * @return The index in m_explanations
     */
    unsigned explanation_index_of_clause(int clause_id) const
    {
        return -1U - clause_id;
    }

    
    inline const explanation_data & get_explanation_data(int clause_id) const;
    
    /**
     * @brief Find a clause saved internally
     *
     * @param clause_id The searched clause
     * @return If the clause is found, return an iterator pointing to a
     *         m_saved_clauses element containing the clause, otherwise
     *         return m_saved_clauses.end();
     * @pre m_saved_clauses is sorted, i.e., no new clause has been saved after
     *      proof_collector::prepare_to_print().
     */
    saved_clause_set::const_iterator find_saved_clause(unsigned clause_id)const
    {
        saved_clause_set::const_iterator
            it= std::lower_bound(m_saved_clauses.begin(),m_saved_clauses.end(),
                                 std::make_pair(clause_id, clause_data()));
        if (it != m_saved_clauses.end() && it->first == clause_id)
            return it;
        else
            return m_saved_clauses.end();
    }
    
    //! Insert one resolution entry into the data structure
    void append_resolution_entry(unsigned constraint_id,
                                 const std::vector<iset> & literals,
                                 int pivot_variable,
                                 std::unique_ptr<explanation_rule> rule);

    //! Undo the last call to append_resolution_entry().
    virtual void pop_resolution_entry();
    /**
     * @brief Print a partial proof leading to the given \p clause.
     *
     * @param[in,out] visitor     A visitor that is consulted while traversing
     *                            the proof graph.
     * @param[in,out] st          The state of traversal
     * @param[in]     clause      The ID of the clause to prove.
     * @param[in,out] visited     What clauses have been visited
     *                            Clauses which have been visited are 'true'
     */
    void traverse_proof(proof_visitor & visitor, state & st, unsigned clause,
                        std::vector<bool> &visited) const;
    
    /**
     * @brief Rerun this resolution and print it into the trace stream.
     *
     * This method both prints and performs the resolution.
     * In order to save space, results of resolutions are not stored,
     * and have to be recomputed when the proof is to be printed.
     *
     * @param[in,out] visitor     A visitor that is consulted while traversing
     *                            the proof graph.
     * @param[in,out] st          The state of traversal
     * @param[in,out] work_clause_id
     *                    The clause ID of \p work_clause.
     * @param[in,out] work_clause
     *                    The resolution work clause. Upon entry, it is the
     *                    clause to be resolved with an antecedent, and after
     *                    exit it is the consequent clause.
     * @param[in]     final_clause
     *                    The ID of the clause after the last resolution,
     *                    i.e., the conflict clause.
     * @param[in]     resolution_entry
     *                    The entry in the m_resolve_history vector which
     *                    contains the antecedent for this resolution.
     */
    void process_resolution(
        proof_visitor &visitor, state & st,
        int & work_clause_id, clause_data & work_clause,
        unsigned final_clause, unsigned resolution_entry) const;

    /**
     * @brief Is m_resolve_history[pos] an inner resolution entry?
     *
     * m_resolve_history[pos] is an inner resolution entry if it describes
     * a resolution between two clauses,
     * otherwise it describes a conflicting clause or is out of bounds.
     */
    bool is_inner_resolve_entry(unsigned pos) const
    {
        return ( pos < m_resolve_history.size() &&
                 m_resolve_history[pos].m_pivot >= 0);
    }

    /**
     * @brief Print the proof that clause conflicts with the initial domains.
     *
     * Resolve the clause with the unary clauses of the initial domains.
     *
     * @param[in,out] visitor     A visitor that is consulted while traversing
     *                            the proof graph.
     * @param[in,out] st          The state of traversal
     * @param[in,out] clause_id,clause
     *                        The clause which conflicts initial domains.
     * @pre clause conflicts with the initial domains.
     * @post clause is the empty clause.
     */
    void process_conflict_with_initial(proof_visitor & visitor, state & st,
                                       int &clause_id, clause_data&clause)const;

    //! Resolve the two input clauses into the resulting consequent.
    static void resolve(const clause_data & first, const clause_data & second,
                        unsigned pivot, clause_data & consequent);
    
    //! Print the clause with its ID.
    void print_clause(std::ostream & out, int id, const clause_data & clause)
        const;

    //! Get the variables of the given constraint ID, overwriting \p clause.
    void get_constraint_variables(unsigned constraint_id,
                                  clause_data & clause) const;

    //! Get the literals (variables+sets) for the given clause ID.
    void get_solver_clause(unsigned clause_id, clause_data & clause) const;

    /**
     * @brief Get the explanation clause for the given explanation clause.
     *
     * @param[in]  clause_id  The ID of the explanation clause
     * @param[out] clause     The retrieved clause.
     *
     * @pre clause_id<0 and refers to an explanation clause ID, i.e.,
     *      there is an m_explanations corresponding with clause_id.
     */
    void get_explanation_clause(int clause_id, clause_data & clause) const;

    /**
     * @brief Get the index into m_resolve_history which leads to the
     *        given conflict clause.
     *
     * @param conflict_clause The ID of the conflict clause.
     *
     * @return Index of m_resolve_history that points to the conflicting
     *         clause which after several resolutions leads to the required
     *         conflict clause.
     *
     * @pre
     *    - Conflict clause indexes are sequential, each conflict clause
     *      has +1 ID to one clause earlier. Other clauses are not
     *      inserted into the CSP.
     *    - \p conflict_clause Is an actual conflict clause which was inserted
     *      into the CSP.
     */
    unsigned clause_to_resolve_index(unsigned conflict_clause) const
    {
        HCSP_ASSERT( conflict_clause >= m_first_conflict_clause );
        unsigned entry= conflict_clause - m_first_conflict_clause;
        HCSP_ASSERT(entry < m_clause_resolve_indices.size());
        return m_clause_resolve_indices[entry];
    }

    //! Return the index of the last conflict clause.
    unsigned last_conflict_clause() const
    {
        HCSP_ASSERT(m_first_conflict_clause != -1U);
        return m_clause_resolve_indices.size()+m_first_conflict_clause-1;
    }

    bool is_equivalent_with_id(const clause_data & clause, int clause_id) const
    {
        clause_data other;
        get_clause(clause_id, other);
        return clause.equivalent(other);
    }
    
    const solver & m_solver;

    //! The index into m_resolve_history of the conflicting clause leading
    //! to the latest conflict.
    unsigned m_latest_resolution_start;

    //! The resolution history
    std::vector<resolve_data> m_resolve_history;

    /**
     * @brief A conversion table from conflict clauses to their entry
     *        in m_resolve_history
     *
     * For a conflict clause \p cc, m_clause_resolve_indices[cc-first_cc]
     * points to its creator in m_resolve_history. The creator of \p cc
     * is the beginning of the resolution chain starting with a conflicting
     * clause and leading to \p cc.
     *
     * This data assumes that conflict clauses are created sequentially
     * with no holes, and that only one clause is generated per conflict.
     */
    std::vector<unsigned> m_clause_resolve_indices;

    /**
     * @brief A description of all explanations produced so far.
     *
     * Each new explanation is appended to this structure, containing a
     * pair such that \p first is the index of the explained constraint, and
     * \p second is the value sets of the explanation. The variable indexes
     * are not stored here since they can be extracted from the constraint
     * itself.
     */
    std::vector<explanation_data> m_explanations;

    /**
     * @brief The set of saved clauses
     *
     * After prepare_to_print this vector is sorted, after each call to
     * save_clause the vector becomes unordered.
     */
    saved_clause_set m_saved_clauses;
    
    //! The first conflict clause which was inserted into the CSP.
    unsigned m_first_conflict_clause;
}; //  proof_collector

} // namespace hcsp

#endif // HCSP_PROOF_COLLECTOR_HPP_

