// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2011, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2011, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_PROOF_INTERPOLANT_VISITOR_HPP_
#define HCSP_PROOF_INTERPOLANT_VISITOR_HPP_

#include <algorithm>
#include <map>
#include <iosfwd>
#include <iset.hpp>
#include <proof_collector.hpp>
#include <proof_visitor.hpp>
#include <formula.hpp>


namespace hcsp {

/**
 * @brief A visitor that creates an interpolant during proof graph traversal
 *
 * It is based on "An Interpolating Theorem Prover", K. L. McMillan.
 */
class proof_interpolant_visitor : public proof_visitor, private solver_types
{
  public:
    proof_interpolant_visitor(const proof_collector &collector,
                              const solver & the_solver);
    virtual ~proof_interpolant_visitor();
    virtual void visit_root(unsigned )  {}
    /**
     * class proof_visitor contains general interface description.
     *
     * Run HypC-A or HypC-B rules on the explanation.
     */
    virtual void visit_explanation(int explanation_clause_id);

    /**
     * class proof_visitor contains general interface description.
     *
     * Run Res-A or Res-B rules on the resolution
     */
    virtual void visit_resolution(
        unsigned pivot,
        int resolvent1_id, const clause_data & resolvent1,
        int resolvent2_id, const clause_data & resolvent2,
        int consequent_id, const clause_data & consequent);
    
    /**
     * class proof_visitor contains general interface description.
     *
     * Run HypC-A + Res-A or HypC-B + Res-B rules on the initial domain with
     * the given clause.
     */
    virtual void visit_initial_resolution(
        unsigned variable, const iset & domain,
        int resolvent_id, const clause_data & resolvent,
        int consequent_id, const clause_data & consequent);
    /**
     * @brief Get the interpolant collected.
     *
     * After the visitor finished traversing the whole proof tree this method
     * can be used to retrieve the interpolant. When the complete problem is
     * separated into the two sub-problems A and B then the interpolant I
     * has the following two properties:
     *   - A implies I
     *   - (B and I) -> false
     * @return The interpolant
     * @pre Completed the traversal over the whole proof tree.
     */
    formula get_interpolant() const;
  private:
    formula::printer printer(const formula & data) const
    { return formula::printer(m_solver.get_variable_print_strategy(), data); }
    variable_printer printer(unsigned variable) const
    { return m_solver.var_printer(variable); }
    //! True if report_level > SEARCH, meaning that visitor should trace the
    //! creation of the interpolant.
    bool need_trace() const { return m_need_trace; }

    //! Memorize the clause and its parts which are completely in A.
    //! In the paper of K. L. McMillan this is the rule HypC-A
    void introduce_clause_in_A(int clause_id);

    //! Get the formula resulting from clause introduction.
    //! In the paper of K. L. McMillan this are the rules HypC-A and HypC-B.
    const formula &get_formula(int clause_id);
    
    const proof_collector & m_collector;
    const solver & m_solver;

    //! For each clause, including temporary, store its clause interpolation.
    std::map<int, formula>  m_formulas;

    //! For each variable store the labels that constrain it, including its
    //! domain.
    const std::vector<label_type> m_variables_touched_by_labels;

    //! The ID of the empty clause
    unsigned                m_empty_clause_id;
    
    //! Should the interpolation process be traced?
    bool                    m_need_trace;
}; // class proof_interpolant_visitor


} // namespace hcsp

#endif // HCSP_PROOF_INTERPOLANT_VISITOR_HPP_

