// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
//
//! \file propagation.hpp
//! \brief Functions for propagating constraints.
//! Each function is responsible for propagating some domains (iset) onto
//! other domains.
#ifndef HCSP_PROPAGATION_HPP_
#define HCSP_PROPAGATION_HPP_

#include <iset.hpp>

namespace hcsp {
namespace propagation {

//! \brief Propagate a+b onto the add_result.
//! The '+' operation wraps around on overflow (modulo arithmetic).
void plus(const iset & a, const iset & b, iset & add_result);
//! \brief Propagate result, sum-subtrahend onto the add_result.
//! The '-' operation wraps around on overflow (modulo arithmetic).
void minus(const iset & sum, const iset & subtrahend, iset & minus_result);
//! \brief Propagate result, -original onto the negated_result
//! The  operation wraps around on overflow (modulo arithmetic).
void neg(const iset & original, iset & negated_result);

//! Propagate "not equal to"
inline void neq(const iset & source, iset & target)
{
  if (source.empty())
    target.clear();
  else if (source.singleton())
    target.erase_all_of(source);
}

//! Propagate "less than"
inline void lt_to_left(iset & left, const iset & right)
{
  iset::value_type keep_below= right.empty()? iset::MIN_DOMAIN : right.max();
  left.erase(iset::value_range_type(keep_below, iset::MAX_DOMAIN));
}

//! Propagate "less than"
inline void lt_to_right(const iset & left, iset & right)
{
  iset::value_type keep_above= left.empty() ? iset::MAX_DOMAIN : left.min();
  right.erase(iset::value_range_type(iset::MIN_DOMAIN, keep_above));
}

//! Propagate "less or equal to"
inline void le_to_left(iset & left, const iset & right)
{
  if (right.empty())
    left.clear();
  else
    left.intersect(iset::value_range_type(iset::MIN_DOMAIN, right.max()));
}

//! Propagate "less or equal to"
inline void le_to_right(const iset & left, iset & right)
{
  if (left.empty())
    right.clear();
  else
    right.intersect(iset::value_range_type(left.min(), iset::MAX_DOMAIN));
}


} // namespace hcsp::propagation
} // namespace hcsp

#endif // HCSP_PROPAGATION_HPP_
