// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2013 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2013 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_PROPAGATION_ALGO_HPP_
#define HCSP_PROPAGATION_ALGO_HPP_

/**
 * This file defines the propagation_algo types and concepts.
 */

#include <watch_types.hpp>
#include <explanation_rule.hpp>
#include <utilities.hpp>
namespace hcsp {

struct propagation_types
{
    enum result_type {
        UNSAT,                  //!< The constraint can't be satisfied
        VALID,                  //!< Any futre domain reduction is satisfied
        CONSISTENT,             //!< The constraint is now consistent
        /**
         * Postpone the call of this constraint to a later state.
         * In this case, the constraint may not modify the state of variables.
         * What the constraint may do, is to register itself in a different
         * activation queue, for later processing.
         *
         * This value is allowed only if either the constraint registers
         * itself in a later queue, or at least one variable is free
         * (a domain with more than one value).
         */
        POSTPONE
    };
}; // struct propagation_types

std::ostream & operator<<(std::ostream&, propagation_types::result_type);

class constraint_runtime_info;

/**
 * @brief A class used to validate the existence of all required capabilities
 *        of a propagation algorithm.
 *
 * To check if class T meets the requirements of the propagation_algo then
 * you must instantiate propagation_algo_concept<T>::test_concept(). The
 * simplest way is to take a pointer of
 * propagation_algo_concept<T>::test_concept(), i.e.,
 * (void)&propagation_algo_concept<constraint_propagator>::test_concept;
 *
 * The difference between the constraint_propagator hierarchy and the
 * propagation_algo concept is that propagation_algo is a concept that
 * may not have virtual functions at all. A propagation_algo can be used
 * as a template parameter to generic functions or classes that need
 * to use a propagation algorithm.
 *
 * We need both propagation_algo and constraint_propagator as they play a
 * slightly different role. A constraint_propagator, being a hierarchy, can
 * be passed to the solver. Propagation_algo, being light, can be used as
 * a building block for efficiently combining several propagation_algo classes
 * to new constraints.
 */
template <typename Propagation_Algo>
class propagation_algo_concept : private propagation_types, watch_types
{
  private:
    //! Make sure object.count_params(info) can be called and that it returns
    //! an integral value.
    static inline void test_count_params()
    {
        const Propagation_Algo* p_object= NULL;
        int value= p_object->count_params();
        (void)value;
    }
    //! Make sure object.propagate(info)  can be called with a
    //! constraint_runtime_info parameter, and return a type compatible with
    //! result_type.
    static inline void test_propagate()
    {
        Propagation_Algo* p_object= NULL;
        constraint_runtime_info *p_info= NULL;
        result_type value= p_object->propagate(*p_info);
        (void)value;
    }
    //! Make sure object.constraint_check(info)  can be called with a
    //! constraint_runtime_info parameter, and return a type compatible with
    //! bool
    static inline void test_constraint_check()
    {
        const Propagation_Algo* p_object= NULL;
        constraint_runtime_info *p_info= NULL;
        bool value= p_object->constraint_check(*p_info);
        (void)value;
    }
    //! Make sure object.init(info)  can be called with a
    //! constraint_runtime_info parameter, and return a parameter compatible
    //! with bool.
    static inline void test_init()
    {
        Propagation_Algo* p_object= NULL;
        constraint_runtime_info *p_info= NULL;
        bool value= p_object->init(*p_info);
        (void)value;
    }
    //! make sure Propagation_Algo can
    //! print(ostream&, const constraint_runtime_info*)
    static inline void test_print()
    {
        const Propagation_Algo* p_object= NULL;
        const constraint_runtime_info *p_info= NULL;
        std::ostream *p_stream= NULL;
        p_object->print(*p_stream, p_info);
    }
    //! make sure Propagation_Algo has watched_param_index(...)
    static inline void test_watched_param_index()
    {
        const Propagation_Algo* p_object= NULL;
        unsigned watched_param= p_object->watched_param_index(1U);
        (void)watched_param;
    }
    //! make sure we can call const_object.count_watches();
    static inline void test_count_watches()
    {
        const Propagation_Algo* p_object= NULL;
        unsigned count= p_object->count_watches();
        (void)count;
    }
    //! make sure we can call object.forget_watches();
    static inline void test_forget_watches()
    {
        Propagation_Algo* p_object= NULL;
        p_object->forget_watches();
    }
    //! make sure we can call object.explanation_rule(....)
    static inline void test_explain_propagation()
    {
        const Propagation_Algo* p_object= NULL;
        const iset * p_domain=NULL;
        constraint_runtime_info *p_info= NULL;
        std::vector<iset> *p_explanation= NULL;
        p_object->explain_propagation(1U, *p_domain, *p_domain,
                                      *p_info, *p_explanation);
    }
  public:
    //! Instantiate all the tests in one place
    static inline void test_concept()
    {
        (void)&test_count_params;
        (void)&test_propagate;
        (void)&test_constraint_check;
        (void)&test_init;
        (void)&test_print;
        (void)&test_watched_param_index;
        (void)&test_count_watches;
        (void)&test_forget_watches;
        (void)&test_explain_propagation;
    }
    
}; // class propagation_algo_concept


} // namespace hcsp
#endif // HCSP_PROPAGATION_ALGO_HPP_
