// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2013 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2013 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_SLICED_RUNTIME_INFO_HPP_
#define HCSP_SLICED_RUNTIME_INFO_HPP_

#include <vector>
#include <iosfwd>
#include <pool_alloc.hpp>
#include <constraint_runtime_info.hpp>

namespace hcsp {
/**
 * @brief Takes a sliced range of parameters from a reference runtime info
 *        object.
 *
 * When the user of this object modifies parameter domains it modifies a
 * local copy, without affecting the original. To affect the original
 * one must explicitly call this.commit().
 *
 * There are two ways to construct this object: In read-write mode or in
 * read-only.
 *
 * WARNING WARNING WARNING:
 * A constraint may decide not to allocate any watch when it knows that the
 * initial run will be the final. When the constraint is part of a bigger
 * constraint, through sliced_runtime_info, this assumption is no longer
 * correct as the bigger constraint may have to reactivate the internal
 * constraint several times regardless of the assumption.
 *
 * For that the bigger constraint has to call info.set_querying_domain() in
 * the right places and the internal constraint has to consult
 * info.querying_domain() and act accordingly.
 */
class sliced_runtime_info final : public constraint_runtime_info
{
  public:
    /**
     * @brief Construct a slice of the original constraint_runtime_info
     *
     * This constructor accepts a non-const info object. During the lifetime
     * of this the original info object is modified only in one of the
     * following conditions:
     *   - this->put_in_queue()
     *   - An explicit call to this->commit(), if modifications are pending.
     *   - If this->querying_domain()==false, then this->set_domain()
     *     writes domain changes through even without this->commit().
     * Note that this implies that the original info is not modified when
     *   - this->querying_domain()==true and this->set_domain() is called,
     *     in which case this->commit() is required.
     *
     * @param info      Original mutable runtime info from which we take a slice
     * @param begin,end The slice range: info[begin] becomes this[0]
     *                  info[begin+1] becomes this[1] and so on. The
     *                  last element we take from info is info[end-1].
     */
    sliced_runtime_info(constraint_runtime_info & info,
                        unsigned begin, unsigned end)
        : m_p_info(&info), m_const_info(info), m_begin(begin), m_end(end)
    { construct(); }
    /**
     * @brief construct a slice of the original constraint_runtime_info
     *
     * This constructor accepts a const info object. Being constructed this way,
     * the original info object is will not be modified by this object in any
     * way. It is implicitly has this->querying_domain()==true.
     *
     * @param info      Original const runtime info from which we take a slice
     * @param begin,end The slice range: info[begin] becomes this[0]
     *                  info[begin+1] becomes this[1] and so on. The
     *                  last element we take from info is info[end-1].
     */
    sliced_runtime_info(const constraint_runtime_info & info,
                        unsigned begin, unsigned end)
        : m_p_info(NULL), m_const_info(info), m_begin(begin), m_end(end)
    { construct(); }
    bool reduced_between_activations(unsigned param_index) const;
    
    queue_type originating_queue() const
    { return m_const_info.originating_queue(); }
        
    void put_in_queue(queue_type queue)
    { m_p_info->put_in_queue(queue); }
        
    const iset & domain(unsigned param_index) const
    {
        if ( m_is_modified[param_index] )
            return m_modified_domain.at(param_index);
        else
            return m_const_info.domain(param_index+m_begin);
    }
    /**
     * @brief Record the domain change locally
     *
     * If !querying_domain() then all modifications are relayed to the
     * original info object. If querying_domain() then it will not touch the
     * original info object and will only record the modification locally. To
     * write the domain to the original info object, one has to call commit()
     * explicitly.
     *
     * The idea is that slicing of info is done is a mega constraint which
     * is comprised of several sub-constraints. It is up to the mega constraint,
     * after it propagates small constraints, to decide whether they should
     * affect the external state.
     */
    void set_domain(unsigned param_index, iset new_domain) override;
    unsigned arity() const { return m_end - m_begin; }
    void deallocate_watch(unsigned param_index, watch_type type,
                          watch_handle_t watch_handle) override;
    watch_handle_t allocate_watch(unsigned param_index, watch_type type)
        override;
    void deallocate_value_watch(unsigned param_index,
                                iset::value_type value,
                                watch_handle_t watch_handle) override;
    watch_handle_t allocate_value_watch(unsigned param_index, 
                                        iset::value_type value) override;
    int latest_modified_param(unsigned except) const;
    unsigned effort() const { return m_const_info.effort(); }
    bool querying_domain() const { return m_querying_domain; }

    /** @brief Decide what querying_domain() should return.
     *
     * Note the limitations for querying_domain() transitions, as the propagator
     * is allowed to observe from one propagate() to another. The limitations
     * are described in constraint_runtime_info::querying_domain().
     */
    void set_querying_domain(bool state= true);
    void print_param(std::ostream & out, unsigned param) const throw();
    void print(std::ostream & out) const throw();

    //! Write the modified values back to the writable_info. Clear modification
    //! indicators.
    void commit();

    /**
     * @brief True if any parameter was reduced but not written to the original
     *        info object. 
     *
     * See set_querying_domain() for details.
     */
    bool pending_commit() const
    {
        for (unsigned i=0 ; i != m_is_modified.size() ; ++i) {
            if (m_is_modified[i])
                return true;
        }
        return false;
    }
    /**
     * @brief True if the given parameter domain was reduced but not written to
     *        the original info object.
     *
     * See set_querying_domain() for details.
     */
    bool pending_commit(unsigned param) const { return m_is_modified[param]; }
    
  private:
    //! Shared code called from constructors.
    void construct();
    /**
     * @brief m_modified[param]== true if param was updated since last commit()
     *        while this->querying_domain()==true;
     *
     * The size of the vector is set correctly the first time domain_update()
     * is called while this->querying_domain==true;
     */
    std::vector<bool, pool_alloc<bool> > m_is_modified;
    std::vector<iset> m_modified_domain;
    constraint_runtime_info * m_p_info;
    const constraint_runtime_info & m_const_info;
    unsigned m_begin, m_end;

    bool m_querying_domain;
}; // class sliced_runtime_info

} // namespace hcsp

#endif // HCSP_SLICED_RUNTIME_INFO_HPP_
