// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_SMALL_ISET_HPP_
#define HCSP_SMALL_ISET_HPP_

#include <string>
#include <iterator>
#include <climits>
#include <stdexcept>
#include <iosfwd>
#include <utilities.hpp>

namespace hcsp {
//! A class efficiently representing a set of integers.
class small_iset {
  private:
    //! The bit set that holds the values of the set.
    typedef unsigned Bits;
    
  public:
    class iterator;
    friend class iterator;
    typedef iterator                               const_iterator;
    typedef std::reverse_iterator<const_iterator>  const_reverse_iterator;
    typedef std::reverse_iterator<iterator>        reverse_iterator;
    typedef unsigned                               size_type;
    typedef int                                    value_type;
    typedef std::pair<value_type, value_type>      value_range_type;
    static const value_type MAX_DOMAIN= sizeof(Bits)*8-1;
    static const value_type MIN_DOMAIN= 0;

    //
    // Methods:
    //

    //! Create an iset with an empty domain
    small_iset() { m_bits= 0; }
    small_iset(const value_range_type & range)
    {
        m_bits= 0;
        insert(range);
    }
    explicit small_iset(const std::string & text);
    void swap(small_iset & other)
    {
        std::swap(m_bits, other.m_bits);
    }
        
    size_type size() const { return count_ones(m_bits); }
    //! Return the number of disconnected ranges
    size_type range_count() const { return count_ranges_of_one(m_bits); }
    
    iterator begin() const;
    iterator end() const;
    reverse_iterator rbegin() const;
    reverse_iterator rend() const;
    iterator upper_bound(value_type element) const;
    iterator lower_bound(value_type element) const;
    /**
     * @brief Find the tail of elements contained in other.
     *
     * Find the first iterator after which all elements, until 'last',
     * are included in 'other'. The 'last' iterator is the point before which
     * the tail ends.
     *
     * @param other The set in which we look for the contained tail.
     * @param last  An iterator marking the end of the tail. The last element
     *              of the tail is right before 'last'.
     * @return An iterator of the first element in the tail which is contained
     *         in 'other'.
     */
    iterator find_contained_tail(const small_iset & other, iterator last) const;
    value_type min() const;
    value_type max() const;
    value_type rand() const;
    small_iset & clear()
    {
        m_bits = 0;
        return *this;
    }

    small_iset & assign(value_type element);
    small_iset & assign(const value_range_type & range);
    small_iset & insert(value_type element);
    
    /**
     * @brief Insert a stride potentially generating holes.
     *
     * @param range     Minimum and maximum values to insert.
     * @param skip_size Distance between consecutive pairs of values
     * @return *this
     * @pre range.first <= range.second  and range.second-range.first
     *      is dividable by skip_size
     * @post this contains range.first, range.first+skip_size, ... range.second
     */
    small_iset & insert(const value_range_type & range, unsigned skip_size);
    small_iset & insert(const value_range_type & range);
    
    //! erase one element from the small_iset.
    small_iset & erase(value_type element)
    {
        if (element >=0 && element <= MAX_DOMAIN)
            m_bits &= ~(1<<element);
        return *this;
    }

    small_iset & erase(const value_range_type & range);
    iterator erase(iterator it);

    small_iset & erase_all_of(const small_iset & erased)
    {
        m_bits &= ~erased.m_bits;
        return *this;
    }

    small_iset & insert_all_of(const small_iset & inserted)
    {
        m_bits |= inserted.m_bits;
        return *this;
    }
    small_iset & intersect(const small_iset & other)
    {
        m_bits &= other.m_bits;
        return *this;
    }
        
    small_iset & intersect(const value_range_type & range);
    small_iset & intersect(const value_range_type & range, size_type skip);
    small_iset & intersect(const std::pair<long long, long long> & range,
                           size_type skip);
    small_iset & complement()
    {
        m_bits = ~m_bits;
        return *this;
    }
        
    bool empty() const { return m_bits == 0; }
    bool everything() const { return (~m_bits) == 0; }
    
    bool singleton() const { return is_power_of2(m_bits); }
    bool contains(value_type element) const
    {
        if (element >=0 && element <= MAX_DOMAIN)
            return (m_bits & (1<<element)) != 0;
        else
            return 0;
    }
    //! \brief Checks set inclusion.
    //! True if for every element e in other, this->contains(e). \n
    //! Note that this is vacuously if other.empty();
    bool contains_all_of(const small_iset & other) const
    {
        return (m_bits & other.m_bits) == other.m_bits;
    }
    //! \brief Checks set overlapping.
    //! True exists an element e in other such that this->contains(e). \n
    //! Note that if other.empty(), then the result is false.
    bool contains_any_of(const small_iset & other) const
    {
        return (m_bits & other.m_bits) != 0;
    }
    //! Check if the two sets are identical.
    friend bool operator==(const small_iset &a, const small_iset &b)
        
    //! Check if the two sets are different.
    {     return (a.m_bits == b.m_bits); }
    friend bool operator!=(const small_iset &a, const small_iset &b)
    { return !(a==b); }

    //! \brief On this implementation, this method has no effect.
    //! On more complex implementations this method may make sense.
    void validate() const {}

    friend std::ostream & operator<<(std::ostream & out,
                                     const small_iset & domain);
    friend std::istream & operator>>(std::istream & in,
                                     small_iset & domain);
  protected:
    
    static size_type range_size(value_type from, value_type to)
    {
        // Be very careful to avoid integer overflows
        if (to >= 0)
            return static_cast<size_type>(to)- from + 1;
        else 
            return static_cast<size_type>(to -from) + 1;
    }
    static size_type range_size(const value_range_type & range)
    { return range_size(range.first, range.second); }
    
  private:
    
    //! Print the current state of the small_iset to cout.
    //! This function is intended to be called from a debugger only.
    void show_self();
    void assert_valid_range(const value_range_type & range,
                            const char *text) const
    {
        if ( range.first > range.second
             || range.first < MIN_DOMAIN
             || range.second > MAX_DOMAIN )
            throw_out_of_range(text);
    }
    void assert_valid_value(const value_type & value, const char *text) const
    {
        if ( value <MIN_DOMAIN || value > MAX_DOMAIN )
            throw_out_of_range(text);
    }

    void throw_out_of_range(const char *text) const throw(std::out_of_range);

    Bits          m_bits;         // Representation of BITS
}; // class small_iset

class small_iset::iterator {
  public:
    typedef std::bidirectional_iterator_tag iterator_category;
    typedef small_iset::value_type     value_type;
    typedef std::ptrdiff_t             difference_type;
    typedef const value_type*          pointer;
    typedef value_type                 reference;
    struct Impl_range_accessor : value_range_type {
        Impl_range_accessor(const value_range_type & other)
            : value_range_type(other) {}
        const value_range_type * operator->() const {return this; }
    };
    static const bool CHECKS_BOUNDS= true;
    static const bool INVALIDATED_BY_ERASE= false;
    
  private:
    bool middle_of_range() const
    {
        if (!m_pIset || m_current_bit == 0 || m_current_bit >= sizeof(Bits)*8)
            return false;

        Bits mask_previous_bit= (Bits)1 << (m_current_bit-1);
        return (mask_previous_bit & m_pIset->m_bits) != 0;
    }
    void throw_runtime_error(const char * text)const throw(std::runtime_error);

    iterator(const small_iset * pVar) : m_pIset(pVar), m_current_bit(0) {}
    static unsigned first_bit(Bits value)
    { return value==0 ? sizeof(Bits)*8 : find_lsb_index(value); }
  public:
    friend class small_iset;
    iterator();

    value_type operator*() const;
    value_range_type current_range() const;
    const iterator & advance_range();
    /**
     * @brief Go to the previous range.
     *
     * If the current position of the iterator is in the middle of a range,
     * then the iterator will go to the beginning of the range. However,
     * if the iterator is at the beginning of a range then the iterator
     * will go to the beginning of the previous range.
     *
     * This means that operator--(), operator++() and retract_range() do not
     * mix well.
     */
    const iterator & retract_range();
    
    /**
     * @brief Find a stride of elements.
     *
     * A stride is a sequence with a constant step size between consecutive
     * elements. Find a stride of at least \p min_count elements.
     *
     * @param[in]  end        The search stops before end is dereferenced.
     * @param[out] range      The first and last element of the stride.
     * @param[out] step_size  The step size of the detected stride
     * @return One iterator after range.second
     */
    iterator find_stride(iterator end,
                         value_range_type &range,
                         size_type & step_size) const;
    
    /**
     * @brief Find a stride of elements, backwards.
     *
     * A stride is a sequence with a constant step size between consecutive
     * elements. Find a stride of at least \p min_count elements.
     * The first element of the stride is not before begin, and the
     * last element of the stride starts one element before 'this'.
     *
     * @param[in]  begin      The search stops at begin or after it.
     * @param[out] range      The first and last element of the stride.
     * @param[out] step_size  The step size of the detected stride
     * @return Iterator pointing to range.first
     */
    iterator find_back_stride(iterator begin,
                              value_range_type &range,
                              size_type & step_size) const;
    
    const iterator & advance(size_type count);
    const iterator & operator++();
    iterator operator++(int);
    const iterator & operator--();
    iterator operator--(int);
    Impl_range_accessor operator->() const { return current_range(); }
    inline bool operator==(const iterator & other) const;
    inline bool operator!=(const iterator & other) const;
  private:
    const small_iset *m_pIset;
    unsigned          m_current_bit;
};

inline small_iset::iterator small_iset::begin() const
{
    iterator ret(this);
    ret.m_current_bit= iterator::first_bit(m_bits);
    return ret;
}

inline small_iset::iterator small_iset::end() const
{
    iterator ret(this);
    ret.m_current_bit= sizeof(Bits)*8;
    return ret;
}

inline small_iset::reverse_iterator small_iset::rbegin() const
{
    return reverse_iterator(end());
}

inline small_iset::reverse_iterator small_iset::rend() const
{
    return reverse_iterator(begin());
}

inline small_iset::iterator small_iset::upper_bound(value_type element) const
{
    iterator ret(this);
    Bits mask;
    if ( element < 0 )
        mask= ~static_cast<Bits>(0);
    else
        mask= unsigned_ones_above<Bits>(element);
    ret.m_current_bit= iterator::first_bit(m_bits & mask);
    return ret;
}

inline small_iset::iterator small_iset::lower_bound(value_type element) const
{
    iterator ret(this);
    Bits mask;
    if ( element < 0 )
        mask= ~static_cast<Bits>(0);
    else
        mask= unsigned_ones_on_and_above<Bits>(element);
    ret.m_current_bit= iterator::first_bit(m_bits & mask);
    return ret;
}

inline small_iset & small_iset::assign(value_type element)
{
    assert_valid_value(element, "while assigning an element");
    m_bits= (1<<element);
    return *this;
}

inline small_iset & small_iset::assign(const value_range_type & range)
{
    m_bits= 0;

    insert(range);
    return *this;
}

inline small_iset & small_iset::insert(value_type element)
{
    assert_valid_value(element, "while inserting an element");
    m_bits |= (1<<element);
    return *this;
}

inline small_iset::iterator::iterator()
    : m_pIset(NULL)
{}

inline small_iset::value_type small_iset::iterator::operator*() const
{
    
    if (m_pIset == NULL)
        throw_runtime_error("small_iset::iterator::operator*() "
                            "dereferncing NULL iterator");

    if (m_current_bit >= sizeof(Bits)*8)
        throw_runtime_error("small_iset::iterator::operator*() "
                            "dereferncing empty iterator");
    return m_current_bit;
}

inline const small_iset::iterator & small_iset::iterator::operator++()
{
    if (m_pIset == NULL)
        throw_runtime_error("iset::iterator::CurrentRange "
                            "dereferncing NULL iterator");
    if ( m_current_bit < sizeof(Bits)*8 ) {
        Bits above= unsigned_ones_above<Bits>(m_current_bit);
        above    &= m_pIset->m_bits;
        m_current_bit= first_bit(above);
    }
    return *this;
}

inline small_iset::iterator small_iset::iterator::operator++(int)
{
    iterator ret=*this;
    ++(*this);
    return ret;
}

inline const small_iset::iterator & small_iset::iterator::operator--()
{
    if (m_pIset == NULL)
        throw_runtime_error("small_iset::iterator::CurrentRange "
                            "dereferncing NULL iterator");
    unsigned first_pos= first_bit(m_pIset->m_bits);
    if ( m_current_bit > first_pos) {
        Bits below= unsigned_ones_below<Bits>(m_current_bit);
        below    &= m_pIset->m_bits;
        if (below == 0)
            m_current_bit= 0;
        else
            m_current_bit= find_msb_index(below);
    }
    return *this;
}

inline bool small_iset::iterator::operator==(const iterator & other) const
{
    if (m_pIset != other.m_pIset)
        return false;
    if (m_pIset == NULL || m_pIset->empty())
        return true;
    return (m_current_bit == other.m_current_bit);
}

inline bool small_iset::iterator::operator!=(const iterator & other) const
{
    return !(*this == other);
}

} // namespace hcsp

#endif // HCSP_SMALL_ISET_HPP_
