// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009-2014, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009-2014, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_SOLVER_HPP_
#define HCSP_SOLVER_HPP_

#include <memory>
#include <vector>
#include <iosfwd>
#include <boost/heap/fibonacci_heap.hpp>
#include <utilities.hpp>
#include <iset.hpp>
#include <solver_types.hpp>
#include <watch_types.hpp>
#include <constraint_connectivity_info.hpp>
#include <variable_print_strategy.hpp>
#include <unique_queue.hpp>
#include <assignment_stack.hpp>

namespace hcsp {
class var_value_watchers;
class var_event_watchers;
class proof_collector_interface;
class conflict_constraint_dependency;
class learned_type;
class formula;
class clause_propagator;
class timer;
class watch_handle_t;

// FIXME: Need to split up into hcsp::solver and hcsp::solver::impl.
//        Too many implementation details are currently visible and are
//        public, when they should have been either private or hidden.
/** \brief This class implements the CSP solver.
 */
class solver : private solver_types, private no_copy, private watch_types
{
  public:
    using solver_types::label_type;
    using solver_types::NO_LABEL;
    using solver_types::A;
    using solver_types::B;
    using solver_types::ALL_LABELS;
    enum relaxation_types {
        //! Do not relax literals
        NO_RELAXATION   =0,   
        //! Relax literals only if it moves them from the current level,
        //! to an earlier level.
        LEVEL_RELAXATION,
        //! Relax as much as possible
        COMPLETE_RELAXATION
    };
    enum result_type {
        SAT,
        UNSAT,
        BACKTRACK_LIMIT,
        TIMEOUT,
        MEMOUT,
        SAT_DONE, // Same as SAT, but when the results have already been printed
        OPTIMAL,
    };
    enum reporting_levels {
        NONE,
        SEARCH,           //!< report choices and learnt conflicts
        PROPAGATION_MODS, //!< report modifications resulting from propagation
        PROPAGATED_CONSTRAINTS //!<report each and every constraint propagation
    };
    typedef assignment_stack::assignment assignment;
    solver();
    ~solver() throw();

    /**
     * @brief Tell the solver whether it should collect proof related data.
     *
     * For producing a proof of unsatisfiability the solver needs to maintain
     * additional data which may slow-down the solving process.
     *
     * @param val  If true then the solver will collect information needed
     *             for producing a proof of unsatisfiability.
     */
    void produce_proof(bool val);
    reporting_levels report_level(reporting_levels level)
    {
        reporting_levels prev= m_report_level;
        m_report_level= level;
        return prev;
    }
    reporting_levels report_level() const { return m_report_level; }
    unsigned backtrack_limit() const { return m_backtrack_limit; }
    unsigned backtrack_limit(unsigned limit)
    {
        unsigned prev= m_backtrack_limit;
        m_backtrack_limit= limit;
        return prev;
    }
    //! Whether or not to print a proof when UNSAT.
    void set_produce_proof(bool flag) { m_produce_proof= flag; }
    //! Whether or not to check the correctness of UNSAT.
    void set_validate_unsat(bool flag) { m_validate_unsat = flag; }
    //! How many backtracks are performed until the first restart
    //! (0 disables restart).
    unsigned restart_threshold() const { return m_first_restart_threshold; }
    //! Set how many backtracks are performed until the first restart
    //! (0 disables restart).
    unsigned restart_threshold(unsigned threshold)
    {
        unsigned prev= m_first_restart_threshold;
        m_first_restart_threshold= threshold;
        return prev;
    }

    //! \brief Get the current state of this threshold. For more information,
    //!        look at the overloaded setter function.
    unsigned opportunistic_back_jumping_threshold() const
    { return m_opportunistic_back_jumping_threshold; }

    /**
     * @brief Set threshold above which the solver performs opportunistic
     *        back-jumping.
     *
     * During conflict analysis, for a conflict clause, count the number of
     * decisions between the asserting literal and the closest and earlier
     * literal. If this difference is above the threshold, then perform an
     * opportunistic back-jumping. After this back-jumping the clause is no
     * longer an asserting clause.
     *
     * The default initial value is 6.
     *
     * @return The previous threshold.
     */
    unsigned opportunistic_back_jumping_threshold(unsigned threshold)
    {
        unsigned old= m_opportunistic_back_jumping_threshold;
        m_opportunistic_back_jumping_threshold= threshold;
        return old;
    }

    //! \brief Get the current state of this threshold. For more information,
    //!        look at the overloaded setter function.
    unsigned conflicts_before_boosting_effort() const
    { return m_conflicts_before_boosting_effort; }

    /**
     * @brief Set the number of conflicts a constraint may participate in,
     *        after which we are willing to raise the effort working with
     *        this constraint.
     *
     * Constraints that participate in conflicts by more than this limit,
     * may choose to work harder, i.e., slower but more accurately. This
     * option is based on the assumption that only part of the constraints
     * are part of the UNSAT core and only they will gain from this added
     * effort.
     *
     * Note that there is no guarantee that this value would be treated
     * linearly in the future. It is possible that in the future the count of
     * conflict participation would be lowered from time to time. Another
     * possibility is that the count of conflict participation would count
     * only conflicting constraints, instead of the way it is done today --
     * all constraints contributing to a conflict clause.
     *
     * The default initial value is 4;
     *
     * @return The previous threshold.
     */
    unsigned conflicts_before_boosting_effort(unsigned limit)
    {
        unsigned old= m_conflicts_before_boosting_effort;
        m_conflicts_before_boosting_effort= limit;
        return old;
    }

    //! \brief Get the current state of this flag.
    //! For more information, look at the overloaded setter function.
    bool should_explainer_check_age() const
    { return m_explainer_check_age; }

    /**
     * @brief Enable or disable the ability of explanation algorithms to
     *        peek into the age of its input domains.
     *
     * An explanation method of a propagator may take advantage of the
     * age information of its input domains, i.e., the order at which the
     * current domains were produced. This knowledge may let the propagator
     * have a more informed choice among several valid explanations.
     *
     * @param value  Whether or not to allow age queries on input domains
     * @return   The previous state of this flag.
     */
    bool should_explainer_check_age(bool value)
    {
        bool orig= m_explainer_check_age;
        m_explainer_check_age= value;
        return orig;
    }
    //! \brief Get the current state of the enabler of conflict clause
    //!        intersection with ground domains.
    //!        For more information, look at the overloaded setter function.
    bool should_intersect_with_ground() const
    { return m_intersect_with_ground; }

    /**
     * @brief Enable or disable the algorithm that intersects conflict clause
     *        with ground level domains.
     *
     * When enabled, after the solver generates a conflict clause it will
     * intersect all literals with ground level domains. A ground level domain
     * is affected by constraint propagation done at decision level zero.
     * These domains are implied by the initial domains and the constraints of
     * the problem, regardless of decisions.
     *
     * The default of this option is 'true'.
     *
     * @todo Currently this option should be disabled when proofs are generated
     *       because these intersections are not yet recorded in the proof.
     *       There should be some way to formulate these ground-zero domains
     *       as a result of some resolution process.
     */
    bool should_intersect_with_ground(bool value)
    {
        bool orig= m_intersect_with_ground;
        m_intersect_with_ground= value;
        return orig;
    }


    //! \brief Get the current state of this flag.
    //! For more information, look at the overloaded setter function.
    relaxation_types relaxation() const
    { return m_relaxation; }

    /**
     * @brief Set how much to relax literals during conflict analysis.
     *
     * During conflict analysis, while constructing the conflict clause,
     * each literal of the clause is associated with a modification entry.
     * The modification entry can be considered as a node in the propagation
     * graph, as discussed in the paper on proof generation,
     * When a literal is relaxed we find an earlier modification entry for
     * the same variable, such that the literal is still false on the domain
     * after that modification. This, as explained in the technical report,
     * is sufficient to maintain the loop invariant, on one hand, and to
     * prune parts of the implication graph, on the other hand.
     *
     * @param value  Whether or not to allow literal relaxation.
     * @return   The previous state of this flag.
     */
    relaxation_types relaxation(relaxation_types value)
    {
        relaxation_types orig= m_relaxation;
        m_relaxation= value;
        return orig;
    }
    //! \brief Get the current state of this flag.
    //! For more information, look at the overloaded setter function.
    bool cause_relaxation() const
    { return m_cause_relaxation; }
    /**
     * @brief Set whether or not cause relaxation should take place.
     *
     * If true then will perform minimum relaxation required to make the
     * latest node in the learned_type to cause the conflict. This means that
     * the latest node and the constraint that implied this node are required
     * for the conflict. This means that going back in time will make the
     * conflict go away.
     *
     * Unlike other types of relaxation, this type is done on the original
     * constraints rather on the explanations. The difference is evident when
     * the explanation of the conflicting constraint is not strong enough to
     * detect a conflict when the constraint still detects it even when
     * relaxing a node.
     */
    bool cause_relaxation(bool value)
    {
        bool orig= m_cause_relaxation;
        m_cause_relaxation= value;
        return orig;
    }
    //! \brief Get the current state of this flag.
    //! For more information, look at the overloaded setter function.
    bool constraints_combination() const
    { return m_constraints_combination; }
    /**
     * @brief Set whether or not to combine constraints during conflict
     *        analysis. 
     *
     * If true then the conflict analyzer will attempt to combine constraints
     * directly, without going through conflict clauses. For example, from
     * x-y>=const1 && y-z>=const2 it might learn x-z>=const1+const2. Note
     * however that not all constraint combinations are supported, in which case
     * there will be a fall-back to the explanation-based learning.
     *
     * Note that even when the solver knows how to combine two constraints,
     * there may be cases when it will fall back to using explanation
     * clauses. This may happen when the resulting constraint, is too weak to
     * detect a conflict for the given domains.
     *
     * For example x<=y && y <=z when D(x)={2}, D(y)={1,3}, D(z)={2}.
     * In this example if we can't use the combined constraint, x<=z, to detect
     * a conflict with the given domains, even though propagating the
     * constraints one-by-one detects a conflict.
     */
    bool constraints_combination(bool value)
    {
        bool orig= m_constraints_combination;
        m_constraints_combination= value;
        return orig;
    }

    //! \brief Get the current state of this flag.
    //! For more information, look at the overloaded setter function.
    bool should_augment_explanation() const
    { return m_augment_explanation; }

    /**
     * @brief Enable or disable usage of augmented explanations during
     *        conflict analysis.
     *
     * An augmented explanation weakens the pivot literal without weakening
     * the appropriate literal after resolution. This weakening can allow
     * strengthening of the rest of the explanation, giving a potentially
     * stronger conflict clause.
     *
     * @param value  Whether or not to use augmented explanations
     * @return   The previous state of this flag.
     */
    bool should_augment_explanation(bool value)
    {
        bool orig= m_augment_explanation;
        m_augment_explanation= value;
        return orig;
    }

    /** @brief Set by how much the baseline cost additive of m_var_costs should
     *         be boosted after each backtrack, effectively lowering all
     *         m_var_costs by 1/(1+1.0/(factor-1))= (factor-1)/factor.
     */
    void set_cost_lowering_factor(unsigned factor)
    { m_assignment_stack.set_cost_lowering_factor(factor);}

    //! \brief Set by how much the conflict driven values of m_var_costs are
    //! higher than the initial, structural, cost of variables.
    void set_analysis_cost_factor(double factor)
    { m_analysis_cost_factor= factor; }
    
    //! \brief In decisions, by how much to give more weight to the latest
    //!        conflict .
    void set_conflict_cost_factor(double factor)
    { m_conflict_cost_factor= factor; }
    
    //! \brief Get the current state of this threshold. For more information,
    //!        look at the overloaded setter function.
    unsigned max_conflict_clause() const { return m_max_conflict_clause; }
    
    /**
     * Set the maximum size of a conflict clause we would retain
     * after the initial propagation. The implementation may 
     * create and insert big clauses and then simply decide not to
     * put watches on it. Also, on some cases (e.g. opportunistic back jumping
     * a big clause may be retained despite being bigger than this threshold).
     */
    unsigned max_conflict_clause(unsigned threshold)
    {
        unsigned old= m_max_conflict_clause;
        m_max_conflict_clause= threshold;
        return old;
    }

    //! \brief Get the current state of this threshold. For more information,
    //!        look at the overloaded setter function.
    unsigned erase_constraints_every_n_conflicts() const
    { return m_erase_constraints_every_n_conflicts; }
    
    /**
     * @brief Set the frequency at which conflict constraint erasure procedure
     *        kicks in.
     *
     * Every so many conflicts the constraints erasure procedure kicks in.
     * Old unused conflict constraints get erased.
     */
    unsigned erase_constraints_every_n_conflicts(unsigned threshold)
    {
        unsigned old= m_erase_constraints_every_n_conflicts;
        m_erase_constraints_every_n_conflicts= threshold;
        return old;
    }

    //! \brief Get the current state of this threshold. For more information,
    //!        look at the overloaded setter function.
    unsigned keep_constraints_from_n_conflicts() const
    { return m_keep_constraints_from_n_conflicts; }
    
    /**
     * @brief Set the number conflicts which are used to determine
     *        if a conflict constraint is active.
     *
     * Inactive conflict constraints are erased, unless they affect the
     * assignment stack. It is enough to have one use of the constraint during
     * the last keep_constraints_from_n_conflicts() backtracks in order to mark
     * it as active.
     */
    unsigned keep_constraints_from_n_conflicts(unsigned threshold)
    {
        unsigned old= m_keep_constraints_from_n_conflicts;
        m_keep_constraints_from_n_conflicts= threshold;
        return old;
    }
    void set_only_optimal(bool value=true) { m_only_optimal= value; }
    void set_all_results(bool value=true)  { m_all_results = value; }
    //! Number of backtracks performed by the last run of the solver.
    unsigned backtracks_done() const { return m_backtracks_done; }
    unsigned restarts_done() const   { return m_restarts_done; }
    //! Number of decisions performed by the last run of the solver.
    unsigned long long decisions_done() const { return m_decisions_done; }
    unsigned long long total_combines() const { return m_total_combines; }
    unsigned long long non_fallback_combines() const
    { return  m_non_fallback_combines; }
    unsigned backtracks_without_fallback_combines() const
    { return m_backtracks_without_fallback_combines; }
    
    constraint_connectivity_info &
    add_constraint(std::unique_ptr<constraint_propagator> p_propagator);

    /**
     * @brief Add a new CSP variable.
     *
     * @param domain            The domain of the new variable
     * @param name              The name of the variable, or NULL if none
     * @param update_singletons If the domain is a singleton, and this is there
     *                          is no singleton with this value should it be
     *                          be recorded as a singleton?
     */
    unsigned add_var(iset domain, const char *name=NULL,
                     bool update_singletons= false);
    void no_more_user_variables() { m_user_variables= count_variables(); }

    //! Set the objective function to optimize this variable.
    void objective_variable(unsigned var_index)
    { m_objective_variable= var_index; }
    
    //! Get the variable (or -1U) which objective function optimizes.
    unsigned objective_variable() const
    { return m_objective_variable; }

    //! Mark whether the objective function is minimizing, otherwise it is
    //! a maximizing objective
    void objective_is_minimizing(bool is_minimizing)
    { m_is_minimizing= is_minimizing; }

    //! Tell whether the objective function is minimizing, otherwise it is
    //! a maximizing objective
    bool objective_is_minimizing() const
    { return m_is_minimizing; }
    
    //! Get a variable with this singleton domain, possibly creating one.
    unsigned singleton_var(iset::value_type value);

    unsigned count_variables() const
    { return m_assignment_stack.count_variables(); }

    //! The number of all currently known constraints, including original and
    //! implied constraints.
    unsigned count_constraints() const
    { return m_propagator_pool.size(); }
    
    //! Internal use only by code that implements the addition of constraints.
    //! This should become private - somehow.
    void deallocate_watch(unsigned var_index, watch_type type,
                          watch_handle_t watch_handle);
    
    //! Internal use only by code that implements the addition of constraints.
    //! This should become private - somehow.
    //! No longer watch changes of the value "val" on variable "variable_index"
    //! for the constraint constraint_index.
    void deallocate_value_watch(unsigned variable_index, 
                                iset::value_type value,
                                watch_handle_t watch_handle);
    //! Internal use only by code that implements the addition of constraints.
    //! This should become private - somehow.
    //! \return the index of the new entry in the list of variable's watchers
    watch_handle_t allocate_watch(unsigned var_index, watch_type type,
                                  unsigned constraint_index);

    //! Watch for removal of value from variable "var_index"
    //! for the constraint  constraint_index.
    watch_handle_t allocate_value_watch(unsigned var_index, 
                                        iset::value_type value,
                                        unsigned constraint_index);

    /** @brief Solve or optimize the current problem
     *
     * If there is an objective function then the CSP will be solved several
     * times with different constraints over the objective function.
     *
     * @param p_out If non-null this is where some of the output is going to go.
     * @return An indicator for success or the type of failure.
     * @post If SAT then all variables are set to a satisfying assignment.
     *       Also, m_backtracks_done and m_restarts_done is updated according
     *       to the whole solution process.
     */
    result_type solve(std::ostream *p_out= nullptr);

    /** @brief Run the solver once.
     *
     * Unlike solve() this method solves the CSP only once even for
     * optimization problems. If this is an optimization problem then it will
     * run with the current bound set upon the objective function. The bound on
     * the objective function is set by the domain of the objective variable.
     *
     * @param fastest  If true then employ a decision heuristics that takes
     *                 least time to produce a result, vs. optimal result.
     *
     * @return An indicator for success or the type of failure.
     * @post Add values to m_backtracks_done and m_restarts_done according to
     *       the number of backtracks and restarts performed. The values are
     *       added to the values these variables have before the call,
     *       regardless of success or failure of the solver.
     */
    result_type solve_current_constraints(bool fastest);
    
    result_type backtrack(unsigned level, int inserted_conflict_constraint=-1);
    
    //! Forget all history and state, except for the learned constraints and
    //! initial_intersect_with_ground().
    void reset();

    /**
     * \brief Forget the constraints leaned from previous conflict analysis.
     *
     * Leaned constraints depend on old domains which may have changed making
     * such constraint depend on wrong domain assumptions.
     *
     * NOTE:
     *
     *  (1) If solve() is called twice with no forget_learned_constraints()
     *      between activations then learned constraints from the former solve()
     *      are no longer considered learned constraints.
     *  (2) If should_intersect_with_ground(true), which is the default, then
     *      this will not reset it.
     */
    void forget_learned_constraints();

    time_type var_modification_time(unsigned var_index) const
    { return m_assignment_stack.var_modification_time(var_index); }

    const variable_print_strategy & get_variable_print_strategy() const
    { return m_assignment_stack.get_variable_print_strategy(); }
    variable_printer var_printer(unsigned index) const
    { return m_assignment_stack.var_printer(index); }
    const iset & var_domain(unsigned var_index) const
    { return m_assignment_stack.var_domain(var_index); }

    const iset & var_ground_level_domain(unsigned var_index) const
    { return m_assignment_stack.var_domain_end_level(var_index,0); }
    
    //! Set the current domain of a variable.
    //!
    //! Also manage modification time and undo history. This function
    //! also notifies relevant constraints of this change.
    //!
    //! \param var_index  The variable that was or is about to be modified
    //! \param new_domain The domain to set.
    //! \param excluded_constraint_index  Unless -1, a constraint that should
    //!                   not be affected by this call.
    //!
    //! \pre This variable was not recorded as a singleton
    void set_var_domain(unsigned var_index, iset new_domain,
                        int excluded_constraint_index=-1);

    /**
     * @brief Get the initial domain of this variable.
     *
     * Get the domain of the variable as it was before solve() was called.
     * This is the domain before propagation takes place.
     */
    const iset & var_initial_domain(unsigned var_index) const
    {
        return m_assignment_stack.var_domain_after_modification(
            var_index, assignment_stack::NO_PREVIOUS_MODIFICATION);
    }
    
    //! Notify constraints that the given variable has been modified.
    //! \param var_index          The variable that was modified.
    //! \param except_constraint  unless -1 this is the index of a constraint
    //!                           that we would like to *not* notify. 
    void notify_modification(unsigned var_index, int except_constraint=-1);

    //! Get the currently processed queue. The currently activated constraint
    //! may use this value to figure out which queue caused its activation.
    constraint_runtime_info::queue_type current_queue() const
    { return m_current_queue; }

    //! Put the given constraint in the requested queue. The constraint will
    //! be activated when this queue is processed.
    //!
    //! Note that constraint_runtime_info::VARIABLE_QUEUE is not a valid value
    //! here. This is because this queue does not relate to constraints,
    //! but to variables.
    void put_in_queue(unsigned constraint_index,
                      constraint_runtime_info::queue_type queue);

    /**
     * @brief Get an interpolant of unsatisfiability of the most recent run
     *
     * We define sub-problem A (or B) to be the subset of constraints marked
     * with label A (or B) as sub-problem. We define the interpolant I such
     * that:
     *   - A implies I
     *   - not(B and I) 
     *
     * @pre
     *     - produce_proof(true) was in effect before solve() and
     *       produce_proof() was not called after solve().
     *     - The latest run was unsatisfiable.
     *     - Every variable and every constraint is labeled with either A or B
     */
    formula get_interpolant() const;

    //! Does this constraint ID refer to a signed clause?
    bool is_clause(unsigned constraint_index) const;

    //! Print the given constraint
    void print_constraint(std::ostream & out, unsigned constraint_index) const;

    /** @brief Print the solution in the required output format.
     *
     * This method can be overridden by a sub-class that wishes to define a
     * different output format.
     * @param out      The stream to which print the solution
     *
     * @pre All the variables are assigned, i.e., all domains are singletons.
     */
    virtual void print_solution(std::ostream & out) const;
    
    /** @brief Get the value of the objective function of the current assignment
     *
     *  This default implementation simply returns the objective variable,
     *  but this is usually not good enough. Although correct, the objective
     *  variable is a pessimistic bound which is usually worse than the actual
     *  value of the objective function. It is recommended to override this
     *  default to an implementation that calculates the objective function
     *  according to the expression and the values it depends on.
     *
     * @pre All domains of the objective function are singletons
     */
    virtual iset::value_type evaluate_objective() const;
    
    //! Get the variables that this constraint refers to, in their order. 
    void get_constraint_variables(unsigned constraint_index,
                                  std::vector<unsigned> & variables)const;

    //! Assuming that this is a clause, return its literal sets, in their order
    const std::vector<iset> & clause_literals(unsigned clause_index) const;

    bool has_constraint(unsigned constraint_id) const
    {
        return (constraint_id < m_propagator_pool.size() &&
                m_propagator_pool[constraint_id] != NULL);
    }
    /**
     * @brief Get the labels of the constraint
     *
     * The problem modeler may label each constraint differently. He can
     * combine several labels by applying a bit-wise \p OR.
     *
     * @param constraint  The index of the constraint to query
     * @return A combination of labels of this constraint
     */
    label_type constraint_labels(unsigned constraint) const;

    /**
     * @brief Set the labels of the constraint, overwriting the old value.
     *
     * 
     * The problem modeler may label, through this method, each constraint
     * differently. He can combine several labels by applying a bit-wise \p
     * OR.
     *
     * @param constraint  The index of the constraint to set
     * @param label       The updated set of labels
     * @return The original labels of this constraint
     */
    label_type constraint_labels(unsigned constraint, label_type labels);

    /**
     * @brief Get the labels of the variable
     *
     * A label on a variable can be viewed as a label on the unary constraint
     * that the domain induces on the variable.  The problem modeler may label
     * each variable differently. He can combine several labels by applying a
     * bit-wise \p OR.
     *
     * @param variable  The index of the variable to query
     * @return A combination of labels of this variable
     */
    label_type variable_labels(unsigned variable) const;

    /**
     * @brief Set the labels of the variable, overwriting the old value.
     *
     * A label on a variable can be viewed as a label on the unary constraint
     * that the domain induces on the variable.  The problem modeler may
     * label, through this method, each variable differently. He can combine
     * several labels by applying a bit-wise \p OR.
     *
     * @param variable  The index of the variable to set
     * @param label     The updated set of labels
     * @return The original labels of this constraint
     */
    label_type variable_labels(unsigned variable, label_type label);

    /**
     * @brief For each variable the set of labels that touch it.
     *
     * A constraint may have several labels, so that a variable. This method
     * returns their conjunction for each variable, such that
     * \p return[variable] is the result of applying a bit-wise \p OR over all
     * constraints that touch the variable and the unary constraint implied by
     * variable's domain.
     *
     * @return A vector such that \p return[variable] contains a label if
     *         the variable is referenced from an entity with this label.
     */
    std::vector<label_type> variables_touched_by_labels() const;
    
    //! Is there any empty domain among the current domains
    bool has_empty_domains() const;

    //! Print the given explanation. This won't print empty literals.
    void report_explanation(std::ostream & out,
                            const constraint_runtime_info & info,
                            const std::vector<iset> &explanation) const;
    
    unsigned get_effort(const constraint_connectivity_info & info) const
    {
        if (info.conflict_rate() <= conflicts_before_boosting_effort())
            return 0;
        else
            return info.conflict_rate() - conflicts_before_boosting_effort();
    }

    const std::string comment_prefix() const { return m_comment; }

  protected:
    //! The best value of the objective function so far, as implied by
    //! m_optimal_solution.
    iset::value_type best_objective() const { return m_best_objective; };
    
    /** @brief Save the current feasible solution as the best.
     * 
     * @pre Assume that all variables are fully assigned.
     */
    void save_solution_as_best();

    /** @brief Assign the variables with the last saved best solution.
     *
     * Set the domains of the variables according to their value at the last
     * call to save_solution_as_best()
     *
     * @pre save_solution_as_best() was called at least once
     */
    void restore_best_solution();
    
  private:
    //! Used in constraint_run_level objects to indicate that the constraint has
    //! not been initialized yet.
    static const unsigned UNINITIALIZED_LEVEL= -1U;

    //! Used for recording last activation level of a constraint
    struct constraint_run_level
    {
        //! The level at which the constraint was activated last,
        //! or UNINITIALIZED_LEVEL if it was neither activated nor initialized.
        unsigned m_run_level;

        //! The index of the recorded constraint.
        unsigned m_constraint;
    }; // struct solver::constraint_run_level
    struct variable_event
    {
        struct comparator
        {
        public:
            comparator(const assignment_stack & stack) : m_stack(stack) {}
            bool operator()(const variable_event & a, 
                            const variable_event & b) const;
            const assignment_stack & m_stack;
        }; // struct solver::variable_event::comparator

        unsigned m_variable;
        iset::size_type m_size_before;
        change_index_t m_modification;
    }; // struct solver::variable_event
    typedef boost::heap::fibonacci_heap<
        variable_event, boost::heap::compare<variable_event::comparator> 
        > variable_queue_type;

    class timeout_exception;
    class propagate_one_constraint_functor;
    friend class propagate_one_constraint_functor;
    class inserter_except;
    class initial_state_restorer;
    friend class initial_state_restorer;
    struct explanation_views;
    
    //! Print the current state of the solver to cout.
    //! This function is intended to be called from a debugger only.
    void show_self();
    result_type optimize();
    /**
     * @brief Print the proof of unsatisfiability of the most recent run
     *
     * @pre
     *     - produce_proof(true) was in effect before solve() and
     *       produce_proof() was not called after solve().
     *     - The latest run was unsatisfiable.
     */
    void print_recent_proof(std::ostream  & out) const;

     /**
     * @brief Check if the UNSAT result, which was just returned, is valid.
     */
    bool is_correctly_unsat() const;
    bool compute_restart(unsigned level);
    void throw_timeout_exception();
    //! Make sure that current options are consistent with each other.
    //! Throw an exception if not
    void validate_flags_consistency() const;
    void advance_current_time()
    {
        if ( (m_current_time+1) == 0 )
            throw_timeout_exception();
        ++m_current_time;
    }

    //! Potentially relax the modification time of the literal,
    //! according to the current mode, without violating the invariant that
    //! the literal is false for the earlier time.
    void relax(learned_type & clause, unsigned literal_position) const;

    void initial_intersect_with_ground(unsigned constraint_index);
    void prepare_for_solve();
    //! Compute the decision probabilities to be used starting now. 
    void restart();
    /**
     * @param fastest  If true then employ a decision heuristics that takes
     *                 least time to produce a result, vs. optimal result.
     */
    result_type solve_impl(bool fastest);

    bool variable_queue_enabled()   const { return m_use_variable_queue;      }
    bool constraint_queue_enabled() const { return !variable_queue_enabled(); }
    //! Propagate one constraint in m_constraint_queue unless constraint queue
    //! is empty.
    //! \return If all is in order return true otherwise, if a conflict is
    //! detected, return false.
    bool propagate_constraint_in_constraint_queue();

    //! Propagate constraints in m_variable_queue until variable queue
    //! is empty.
    //! \return If all is in order return true otherwise, if a conflict is
    //! detected, return false.
    bool propagate_constraints_in_variable_queue();

    //! Propagate constraints dependent on this variable and its modification
    //! \param variable      The variable whose constraints to propagate
    //! \param modification  The modification time of the variable which
    //!                      affects these constraints.
    //! \return If all is in order return true otherwise, if a conflict is
    //! detected, return false.
    bool propagate_constraints_touching_var(unsigned variable,
                                            change_index_t modification);
    
    //! Propagate constraints until no propagation can be done.
    //! \return If all is in order return true otherwise, if a conflict is
    //! detected, return false.
    bool propagate_constraints();

    /**
     * @brief Propagate all the learned constraints which have not been run yet
     *        on the current level.
     *
     * The learned constraint have not been initialized which would be indicated
     * by constraint_relevance.m_run_level == UNINITIALIZED_LEVEL. In that case
     * the constraint is initialized first. Here is an example that shows why a
     * learned constraint has to be rerun on earlier levels at which it was
     * initially propagated:
     *
     *  - Learned constraint: x==y. \n
     *    Inserted at level 5 where D(x)={3:4} and D(y)={2:3}. \n
     *    At level 5 it is asserting because it will remove values from the
     *    domains.
     *  - Backtrack to level 4 where D(x)=D(y)={2:4}\n
     *    At level 4 x==y is consistent
     *  - Backtrack to level 3 where D(x)={1:4}, D(y)={2:5}\n
     *    At level 3 x==y is inconsistent and would reduce  D(x)=D(y)={2:4}\n
     *    We would like to propagate it again at this level.
     *  - Similarly it is consistent at level 2 and inconsistent at level 1.
     *  
     * @param[in,out] constraint_relevance
     *                Information needed to figure out what the method should do
     *                with this constraint. For example, it will initialized the
     *                constraint if its level is UNINITIALIZED_LEVEL.
     *                After the run, if all is well, the level will be updated.
     * @return Whether a conflict was detected.
     */
    bool propagate_dirty_learned_constraint(
        constraint_run_level & constraint_relevance);
    
    //! Propagate constraints for the first time, until no propagation can
    //! be done. Initialize whatever has to be initialized in the first
    //! propagation run.
    //! \return If all is in order return true otherwise, if a conflict is
    //! detected, return false.
    bool first_time_propagate_constraints();

    /**
     * Report the last propagation and its result.
     *
     * @param[in] result   The result return code returned by the last
     *                     propagator.
     * @param[in] report_removal
     *                     Whether should report on value removal performed
     *                     by last propagation - this is done regardless of
     *                     other propagations.
     */
    void report_last_propagation(constraint_propagator::result_type result,
                                 bool report_removal) const;
    
    /**
     * @brief propagate only this constraint, and apply its outcome.
     *
     * This is a simple wrapper that calls constraint_propagator::propagate()
     * for the given constraint. It manages all bookkeeping around this
     * activation and, if requested, reports the results of this propagation.
     *
     * @param constraint_index The constraint to propagate
     * @param report_removal Whether should report on value removal performed
     *                       by last propagation - this is done regardless of
     *                       other propagations.
     *    
     * @return The result type of this propagation process, as returned by
     *         the propagator.
     */
    constraint_propagator::result_type 
    propagate_one_constraint(unsigned constraint_index,
                             bool report_removal=false);

    /**
     * @brief Initialize only this constraint.
     *
     * This is a simple wrapper that calls constraint_propagator::init()
     * for the given constraint. It passes the right parameters to the
     * right object, and then prints the results, if requested.
     *
     * \param constraint_index
     *             The constraint to be initialized.
     * \param modification_of_reference_state
     *             Make the assumption that the state of the solver is just
     *             like it was after \e modification_of_reference_state.
     *             If propagate_since is big enough (e.g. in the "future")
     *             then the current state of the solver is used.
     * \param is_learned
     *             True if the constraint is a learned constraint. Otherwise,
     *             if false, it is a constraint that comes from the original
     *             CSP.
     * @return The result type of this propagation process, as returned by
     *         the propagator.
     */
    bool init_one_constraint(unsigned constraint_index,
                             change_index_t modification_of_reference_state,
                             bool is_learned= false);

    
    //! Choose the next variable to instantiate.
    //! \return -1 if all variables are assigned, variable index otherwise.
    int choose_free_variable();

    /**
     * @brief Perform restriction operation on the variable.
     *
     * In other words, make a search choice which restricts the search space 
     * to only some values of this variable. The classic restriction operation
     * in CSP is simple value assignment through value ordering.
     * @param var_index The variable to restrict
     * @param fastest  If true then employ a decision heuristics that takes
     *                 least time to produce a result, vs. optimal result.
     *
     */
    void restrict_variable(unsigned var_index, bool fastest);
    //! When conflict analysis is disabled, this method is called to complement
    //! the previous decision (i.e. remove bad values).
    void complement_decision(unsigned decision_variable,
                             const iset & complement_values);
    
    /**
     * Find the explanation for the given modification.
     *
     * The explanation is given as a set of literal domains relative to the
     * variables of the affecting constraint. The returned value is a
     * reference to internal data and its validity is explained at the
     * "return value" paragraph.
     *
     * @pre At least one of the following conditions must hold:
     *   -# The modification entry does not exist (after the last entry),
     *      and the constraint in question is a conflict constraint.
     *      propagated_param is the max unsigned value.
     *   -# The modification entry resulted from the propagation of the
     *      constraint in question onto the propagated_param.
     *
     * @param constraint_index
     *        The index of the constraint to be explained.
     * @param propagated_param
     *        The index of the parameter that caused this reduction, or
     *        max unsigned to indicate a conflicting constraint.
     * @param values_to_explain
     *        The set of parameter's values which are at all relevant to the
     *        caller. This set of elements should be viewed as a refinement of
     *        \p propagated_param .
     * @param modification
     *        The modification entry to be explained or, for a conflicting
     *        constraint, an invalid entry beyond the last modification.
     *
     * @return A explanation information for the modification in question\n
     *         \n
     *         If the constraint in question is, if fact, a clause then
     *         simply return the literals of this clause. Otherwise the
     *         constraint should generate the explanation.
     *         See hcsp::constraint_propagator::explain_propagation()
     *         for more details on explanations. \n
     *         \n
     *         The returned value is a reference to data which can become
     *         invalid. The data is valid only as long as all of the following
     *         is met:
     *           -# The assignment stack did not change (no backtracking,
     *              no propagation, no domain reduction, etc.).
     *           -# This method has not been called again.
     */
    explanation_views get_constraint_explanation(
        unsigned constraint_index, unsigned propagated_param,
        const iset & values_to_explain, change_index_t modification) const;

    /**
     * Analyze the current conflict. Add conflict learned constraints and
     * give the choice level we should backtrack to.
     * @param[out] inserted_conflict_constraint
     *             If a conflict constraint has been created then this variable
     *             will get its index, otherwise it is set to a negative
     *             value.
     * @return The decision level where we want to backtrack to, or -1
     *         if the problem is unsatisfiable.
     */
    int analyze_conflict(int & inserted_conflict_constraint);


    /**
     * Find the right backtrack level assuming this learned_constraint.
     *
     * @param learned_constraint  The learned constraint to consult
     * @param asserting_level     If >=0 then this is the asserting level
     *                            of the learned constraint.
     * @return -1 on failure, otherwise return the desired backtrack-level.
     * @pre learned_constraint.unsat() ||
     *      learned_constraint.asserting_level() == asserting_level
     */
    int find_backtrack_level(const learned_type & learned_constraint,
                             int asserting_level) const;
    
    //! Resolve one or more clauses which lead to the conflicting 'clause'.
    //! Stop resolution process if either there is not enough information,
    //! or if the learned 'clause' is good enough (e.g. 1-UIP).
    bool resolve_conflicting_implications(learned_type & clause);

    /**
     * @brief Generate the conflict constraint for the current conflict.
     *
     * This method assumes that the last activated constraint propagator
     * detected the conflict.
     *
     * @return like in resolve_conflicting_implications(), true if the
     *         resulting constraint is interesting.
     */
    bool generate_conflict_constraint(learned_type &constraint);

    /**
     * @brief Collect implication nodes participating in the current conflict.
     *
     * @param[out] learned_constraint The nodes and literals, to be collected,
     *                                of the conflict.
     * @param[in]  explanation        Explanation and, possibly, the rule to
     *                                generate the explanation literals.
     * @param[in]  constraint_index   The index of the conflicting constraint.
     * @param[in] before_modification Collect the nodes knowing that the
     *                                constraint conflicts on a state before
     *                                this modification.
     */
    void collect_conflict_nodes(learned_type & learned_constraint,
                                explanation_views & explanation,
                                unsigned constraint_index,
                                change_index_t before_modification=
                                    assignment_stack::POST_LAST_MODIFICATION);
    /**
     * @brief Restart the conflict analysis algorithm.
     *
     * This may happen when the last explanation is stronger than the
     * propagation it explains, up to the point when it is conflicting
     * with the analyzed state. We can treat this as a conflicting constraint.
     */
    void restart_learning(learned_type      &conflict_clause,
                          explanation_views &explanation,
                          unsigned           constraint_index);

    /**
     * @brief Record the variables of explanation's effective literals.
     *
     * An effective literal is a literal that can be made true, considering
     * the initial assignment. The recording is done by pushing the index of
     * the variable to m_analyzed_variables.
     * 
     */
    void record_explanation_variables(
        const constraint_connectivity_info & connectivity,
        const std::vector<iset> & literals);
    
    /**
     * Create a constraint that the given learned_type describes
     *
     * @param[in]  learned_constraint
     *             The literals that describe the new clause, or the actual
     *             constraint to insert.
     */
    unsigned insert_constraint(const learned_type &learned_constraint);
    
    /**
     * @brief Apply resolution rule to \b clause and to the other constraint.
     *
     * If the other constraint is not a clause, then produce an explanation
     * clause and apply resolution to it.
     *
     * @param[in,out] constraint
     *                       The constraint to be resolved (as input) and the
     *                       result of the resolution (as output).
     * @param[in]     pivot_variable
     *                       The variable which should be used as the pivot in
     *                       the resolution process.
     * @param[in]     other_constraint_index
     *                       The ID of the second constraint to be resolved.
     */
    void resolve(learned_type & constraint,
                 unsigned       pivot_variable,
                 unsigned       other_constraint_index);

    /**
     * @brief Gather information and log information in preparation for
     *        resolution.
     *
     * If the other constraint is not a clause, then produce an explanation
     * clause and apply resolution to it.
     *
     * @param[in,out] constraint
     *                The constraint to be resolved (as input), and if there is
     *                a need to restart learning the constraint is updated
     *                accordingly (as output). 
     * @param[in]     pivot_variable
     *                The variable which should be used as the pivot in the
     *                resolution process. 
     * @param[in]     other_constraint_index
     *                The ID of the second constraint to be resolved.
     * @param[out]    other_pivot_index
     *                The parameter index of the pivot variable in the other
     *                constraint.
     * @param[out]    p_other_literals
     *                Explanation literals of the other constraint.
     * @return True if resolution can proceed, false if it is impossible to
     *         continue.
     */
    bool prepare_to_resolve(learned_type              &constraint,
                            unsigned                   pivot_variable,
                            unsigned                   other_constraint_index,
                            unsigned                  &other_pivot_index,
                            const std::vector<iset>*  &p_other_literals);
    
    /**
     * @brief Resolve the two clauses.
     *
     * The first clause is initially the antecedent and then consequent of the
     * resolution. The 'other' clause is the explanation clause to be
     * resolved with the antecedent.
     *
     * @param[in,out] clause The clause to be resolved (as input) and the
     *                       result of the resolution (as output).
     * @param[in]     other_connectivity
     *                       The connectivity information of the 'other'
     *                       constraint
     * @param[in]     other_literals
     *                       The literals (sets) explaining the 'other'
     *                       constraint.
     * @param[in]     other_pivot_index
     *                       The pivot parameter of the 'other' constraint.
     * @param[in]     other_constraint_activation
     *                       The modification entry at which the other clause
     *                       has been activated.
     */
    void resolve_clause(learned_type & clause,
                        const constraint_connectivity_info & other_connectivity,
                        const std::vector<iset> &other_literals,
                        unsigned                 other_pivot_index,
                        change_index_t       other_constraint_activation) const;

    //! Print basic data about the resolution that will take place now.
    void report_resolution_start(std::ostream & out,
                                 const learned_type &constraint,
                                 unsigned other_contraint,
                                 change_index_t pivot_modification) const;

    /**
     * @brief Get the set of values that modifications starting with
     *        modification_index have removed from the variable.
     *
     * @pre It is assumed that modification_index describes a modification
     *      to the variable var_index.
     *
     * @return Set of removed values.
     */
    iset removed_values(unsigned var_index, change_index_t modification_index)
        const;
    
    //! Mark that this constraint has been recently accessed.
    void mark_used(unsigned constraint_index);

    //! Mark that prerequisite_constraint was used to derive
    //! consequent_constraint.
    void mark_dependency(unsigned prerequisite_constraint,
                         unsigned consequent_constraint);
    //! Forget all information on this conflict constraint, except for the proof
    //! data structure.
    void forget_conflict_constraint(unsigned constraint_index);

    //! Save the prerequisites of clauses_to_keep which are not in this set,
    //! into m_proof_collector, if proof information is collected.
    void save_erased_prerequisites_for_proof(
        const std::vector<unsigned> &clauses_to_keep);

    //! Check if it is about time to remove unused conflict constraints, and
    //! operate accordingly.
    void remove_unused_conflict_constraints_maybe();

    /**
     * Forget all watches of the problem. This will remove watches
     * from each variable and, if managed, forget them in each constraint.
     * This way variables have no watches, and constraints remember no
     * leftover watch IDs.
     */
    void forget_all_watches();

    /***
     * @brief When solve() is UNSAT perform relevant finalization operations. 
     */
    void manage_unsat();

    /**
     * @brief Count the literals that can't be optimized-out.
     *
     * Count how many literals are not falsified by ground-level domains,
     * as defined by the 'info'. Literal number \p i refers to variable
     * info.to_variable(i).
     *
     * @param literals The literals we want to check and count
     * @param info     The domains and the variables for each literal
     * @return The number of irredundant literals
     */
    unsigned count_irredundant_literals(const std::vector<iset> & literals,
                                        constraint_connectivity_info&info)const;
    
    /** @brief Add the cost of the constraint to the variables
     *
     * Update m_var_costs for each of the variables participating in the
     * constraint. The costs are updated according to the \p base_cost.
     *
     * @param constraint_index  The constraint whose cost is added to the
     *                          variables
     * @param base_cost         The base cost to be added. The actual cost
     *                          added to the variable is linearly dependent on
     *                          the base_cost.
     * @param p_literals        If p_literals is provided, for each argument
     *                          of the constraint consider it only if this
     *                          literal is not falsified by the ground level
     *                          domain.
     */
    void add_constraint_cost_to_variables(
        unsigned constraint_index, double base_cost,
        const std::vector<iset> * p_literals=NULL);
    /**
     * @brief Find the modification entry that represents the conflict.
     *
     * - If the conflicting constraint modifying variables, then return the
     *   earliest modification entry of this constraint.
     * - Otherwise, if the conflicting constraint did not modify variables,
     *   then return an entry after the last modification (like end() ).
     */
    change_index_t conflicting_modification_entry() const;
    
    //! Print time difference and set iteration_time to current time
    void advance_iteration_time(timer& iteration_time);

    /** @brief Try to improve the currently known best solution.
     *
     * Run the solver and try to get a better objective function than previous
     * runs. Print and advance the solution time.
     */
    bool improve_solution(timer & iteration_time, solver::result_type & result);

    ////////////////////////////////////////////////////////////////////////
    // Member variables:
    ////////////////////////////////////////////////////////////////////////

    //
    // General solver data
    //
    time_type                                 m_current_time;
    bool                                      m_use_variable_queue;
    
    // The last activated constraint (either VALID, CONSISTENT or UNSAT). 
    unsigned                                  m_last_constraint;
    
    //! The currently processed queue. The currently activated constraint
    //! may query this value.
    constraint_runtime_info::queue_type       m_current_queue;
    reporting_levels                          m_report_level;
    bool                                      m_now_solving;

    // Backtrack related info
    unsigned                                  m_backtrack_limit;
    unsigned                                  m_backtracks_done;
    unsigned                                  m_backtracks_for_iteration;
    unsigned                                  m_backtracks_since_restart;
    unsigned                                  m_first_restart_threshold;
    unsigned                                  m_next_restart_threshold;
    unsigned                                  m_restarts_done;
    unsigned long long                        m_decisions_done;
    //! Total number of combine+resolve call throughout the solving algorithm
    unsigned long long                        m_total_combines;
    
    //! Total number of pure combine() without a resolve() fall-back,
    //! throughout the solving algorithm
    unsigned long long                        m_non_fallback_combines;
    //! Total number of learned constraints that used only combine() and did
    //! not require a fallback.
    unsigned                          m_backtracks_without_fallback_combines;
    //! Had the current conflict analysis algorithm any fallback?
    bool                              m_current_conflict_analysis_has_fallback;
    //! For decision level d, m_backtracks_starting_level[d] is the value
    //! of m_backtracks_since_restart at the beginning of the level
    std::vector<unsigned>                     m_backtracks_starting_level;
    //! \brief See the should_explainer_check_age() method
    bool                                      m_explainer_check_age;
    //! \brief See the should_intersect_with_ground() method
    bool                                      m_intersect_with_ground;
    //! \brief See the relaxation() method
    relaxation_types                          m_relaxation;
    bool                                      m_cause_relaxation;
    //! \brief See the constraints_combination() method
    bool                                      m_constraints_combination;
    //! \brief See the should_augment_explanation() method
    bool                                      m_augment_explanation;
    
    bool                                      m_only_optimal;
    bool                                      m_all_results;
    //! \brief If true then the print a proof when solve()==UNSAT.
    bool                                      m_produce_proof;
    //! \brief If true then the correctness of solve()==UNSAT is validated.
    bool                                      m_validate_unsat;

    //! The object that can collect proof related information.
    std::unique_ptr<proof_collector_interface>  m_proof_collector;
    
    //! \brief By how much the cost of analysis is higher than initial cost.
    double                                    m_analysis_cost_factor;
    
    //! \brief In decisions, by how much to give more weight to the latest
    //!        conflict .
    double                                    m_conflict_cost_factor;

    //! \brief See  opportunistic_back_jumping_threshold() method
    unsigned                            m_opportunistic_back_jumping_threshold;

    //! \brief See  conflicts_before_boosting_effort() method
    unsigned                            m_conflicts_before_boosting_effort;

    // constraint info
    std::vector<constraint_connectivity_info> m_constraint_info_collection;
    unsigned                                  m_original_constraint_count;
    std::vector<constraint_propagator*>       m_propagator_pool;
    unique_queue<>                            m_constraint_queue;
    std::vector<constraint_run_level>         m_learned_rerun_level;
    std::unique_ptr<conflict_constraint_dependency>
                                             m_conflict_constraint_dependencies;

    //! \brief See max_conflict_clause() method
    unsigned                                  m_max_conflict_clause;

    //! \brief See erase_constraints_every_n_conflicts() method
    unsigned                             m_erase_constraints_every_n_conflicts;
    //! \brief See keep_constraints_from_n_conflicts() method
    unsigned                             m_keep_constraints_from_n_conflicts;

    // Variable info
    assignment_stack                          m_assignment_stack;
    std::vector<var_event_watchers>           m_var_watchers;
    std::vector<var_value_watchers>           m_var_value_watchers;
    std::vector<double>                       m_var_costs;
    variable_event::comparator                m_variable_event_comparator;
    variable_queue_type                       m_variable_queue;
    std::vector<std::pair<variable_queue_type::handle_type, bool> >
                                              m_variable_queue_handles;
    int                                       m_last_chosen_var;
    unsigned                                  m_user_variables;
    //! Variables participating in the last conflict analysis, in literals
    //! which are not false at decision level zero.
    std::vector<unsigned>                     m_analyzed_variables;

    /** @brief If positive, then this is the variable that HCSP has to optimize.
     *
     * The objective can be set to be minimized or to maximized, depending
     * on \p m_is_minimizing.
     */
    int                                       m_objective_variable;

    //! Is the objective variable to be minimized or, otherwise, maximized.
    bool                                      m_is_minimizing;

    //! The best solution so far, as set by save_solution_as_best().
    std::vector<iset::value_type> m_best_solution;

    //! The best value of the objective function so far, as implied by
    //! m_optimal_solution.
    iset::value_type m_best_objective;

    //! The stream to which to write some of output of solve().
    std::ostream * m_p_out;

  protected:
    //! The prefix to print before printing comments
    std::string m_comment;
    
    //! The prefix to print before printing whether a solution was found.
    std::string m_summary_prefix;

    //! The prefix to print before printing the last objective value
    std::string m_objective_prefix;
};

//! Print the result of the run, i.e., success or type of failure, in a human
//! readable fashion.
std::ostream & operator<<(std::ostream& out, solver::result_type);

} // namespace hcsp

#endif // HCSP_SOLVER_HPP_
