// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_TABLE_PROPAGATORS_HPP_
#define HCSP_TABLE_PROPAGATORS_HPP_

#include <vector>
#include <iset.hpp>
#include <constraint_propagator.hpp>

namespace hcsp {

//! An internal base class for table propagators
class table_propagator_base : public constraint_propagator
{
  public:
    class memory_manager;
    
  protected:
    class decision_node;
    class decision_sub_tree;
    class tuples;
    friend class decision_node;
    friend class decision_sub_tree;
    //! A single tuple representing one assignment
    typedef std::vector<iset::value_type> tuple_value_type;

    /**
     * @brief Construct from a set of tuples.
     * @param[in] tuples_in  List which represent the set of tuples in the
     *                       table.
     */
    table_propagator_base(memory_manager & manager,
                          const std::vector<tuple_value_type> &tuples_in);

    /**
     * @brief An efficient copy constructor.
     *
     * The copy constructor copies only transient data. The tuples are shared
     * across all copied propagators.
     */
    table_propagator_base(const table_propagator_base & other);

    /**
     * @brief An efficient assignment.
     *
     * Like with the copy constructor, the tuples are shared across all copied
     * propagators.
     */
    table_propagator_base &operator=(const table_propagator_base &other);
    //! An efficient destructor.
    virtual ~table_propagator_base();

    //! Tell how many parameters this constraint has, or -1 if inapplicable
    //! such as the table is empty.
    int count_params() const;
    //! Check if the tuple defined by info is contained in this table
    bool find_tuple(const constraint_runtime_info & info) const;
    
    //! To be called from within a debugger, for data-structure visualization
    void show_self() const;
    //! Get the next parameter, avoiding repeating first_parameter.
    static unsigned get_next_parameter(unsigned first_parameter,
                                       unsigned parameter)
    {
        unsigned next_parameter= first_parameter == parameter ? 0 : parameter+1;
        if (next_parameter == first_parameter)
            ++next_parameter;
        return next_parameter;
    }
    
    const tuples & get_tuples() const { return *m_p_tuples; }
    
  private:
    tuples *m_p_tuples;
}; // class table_propagator_base

/**
 * @brief A memory manager to hold, allocate and free table information.
 *
 * Tables do not manage the memory of their data structure by themselves, this
 * class manages their memory. This allows greater sharing of data across
 * different tables. One of the ways to accomplish this is to perform garbage
 * collection on irrelevant parts of the data.
 */
class table_propagator_base::memory_manager : private no_copy
{
  public:
    memory_manager(); //!< default constructor
    /**
     * @brief destructor
     *
     *@pre All tables that used this memory_manager to allocate their data
     *     have been deleted.
     */
    ~memory_manager();
    //! Collect and delete tables memory which is no longer used.
    void garbage_collect();
    /**
     * @brief Make garbage_collect() behave like a no-op.
     *
     * This method is intended to be used when the memory_manager and all
     * relevant tables are about to be deleted. This will avoid costly garbage
     * collection of memory_manager data which its destructor is about to
     * delete anyway.
     */
    void disable_garbage_collection();
  private:
    friend class table_propagator_base;
    class impl;

    /**
     * @brief Get a DAG node representing the given sub-tree.
     *
     * Try to find a DAG node, from the set of nodes this memory_manager has
     * allocated, that represents this subtree. If no such node has already
     * been allocated, then allocate a new one.
     *
     * @param subtree The sub-tree for which we would like to allocate the
     *        new DAG node.
     * @return A pointer to a node which memory_manager has allocated and
     *         which memory_manager is responsible to delete. The caller
     *         must not delete this node.
     */
    const decision_node *get_node(decision_sub_tree & subtree);
    
    impl *m_p_impl;
}; // class table_propagator_base::memory_manager

//! A constraint defined by a set of supported tuples
class support_table_propagator : public table_propagator_base
{
  public:

    /**
     * @brief Construct from a set of tuples.
     * @param[in] tuples_in  List which represent the set of supported values.
     */
    support_table_propagator(memory_manager & manager,
                             const std::vector<tuple_value_type> &tuples_in)
        : table_propagator_base(manager, tuples_in)
    {
    }
    
    virtual result_type propagate( constraint_runtime_info & info );
    virtual bool conflicts(const constraint_runtime_info & info) const;
    /**
     * @brief Create a table propagator with the same semantics as this.
     */
    virtual support_table_propagator* clone() const
    { return new support_table_propagator(*this); }

    //! For testing
    bool constraint_check(const constraint_runtime_info & info) const
    { return find_tuple(info); }
}; // class support_table_propagator


} // namespace hcsp

#endif // HCSP_TABLE_PROPAGATOR_HPP_
