// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_UNIQUE_QUEUE_HPP_
#define HCSP_UNIQUE_QUEUE_HPP_

#include <deque>
#include <vector>
#include <utilities.hpp>

namespace hcsp {
template <typename Data>
struct unique_queue_data_impl
{
    typedef unsigned                     key_type;
    typedef Data                         mapped_type;
    typedef unique_queue_data_impl<Data> value_type;
    typedef const value_type &           value_argument_type;

    unique_queue_data_impl()                         : key(0) {}
    explicit unique_queue_data_impl(key_type key_in) : key(key_in) {}
    
    explicit unique_queue_data_impl(key_type key_in, const mapped_type &data)
        : key(key_in), mapped(data)
    {}
    operator key_type() const { return key; }
    bool operator==(value_argument_type other) const
    { return key == other.key && mapped == other.mapped; }
    
    key_type    key;
    mapped_type mapped;
};

template<>
struct unique_queue_data_impl<void>
{
    typedef unsigned key_type;
    typedef void     mapped_type;
    typedef unsigned value_type;
    typedef unsigned value_argument_type;

    unique_queue_data_impl(key_type key_in=0 ) : key(key_in) {}
    operator key_type() const { return key; }
    
    key_type key;
};

/**
 * @brief This class maintains a FIFO queue of elements with bounded integer
 *  keys.
 *
 * An element which is already in queue will not enter it a second time.
 * Every element has a key of key_type, and if Data is given then
 * it also has a 'mapped' part of mapped_type.
 *
 * The integer keys are bounded by set_max_elements().
 */
template <typename Data= void>
class unique_queue
{
  public:
    unique_queue() : m_max_elements(0) {}

    //! \brief The key type of the queue elements.
    typedef typename unique_queue_data_impl<Data>::key_type key_type;

    /**
     * @brief This is the value type of the queue elements.
     *
     * If the Data template parameter is given (i.e. not void)
     * then value_type is a struct containing a mapped element of
     * mapped_type (i.e. Data) and with a key of key_type.
     * value_type is automatically converted to key_type wherever key_type
     * is required by context.
     */
    typedef typename unique_queue_data_impl<Data>::value_type value_type;

    //!\brief Defines the most efficient way to pass an argument of value_type.
    typedef typename unique_queue_data_impl<Data>::value_argument_type
                     value_argument_type;

    /**
     * @brief Set the maximum allowable element ID.
     *
     * @pre It is forbidden to lower the number of max_elements in conflict
     * of existing elements in the queue. In such a case, an out_of_range
     * exception will be thrown and the queue will be unchanged.
     *
     * Complexity O(max_elements)
    */
    void set_max_elements(size_t max_elements)
    {
        if ( max_elements < m_max_elements )
            validate_all_elements_below(max_elements);
        unsigned chunks= max_elements/CHUNK_BITS;
        unsigned max_in_last_chunk= max_elements%CHUNK_BITS;
        if (max_in_last_chunk != 0)
            ++chunks;
        m_chunk_elements_in_queue.resize(chunks, 0U);
        
        m_max_elements= max_elements;
        if (max_in_last_chunk != 0)
            m_chunk_elements_in_queue.back() &= (1U << max_in_last_chunk)-1;
    }

    //! Check what is the maximum allowable element ID.
    size_t get_max_elements() const { return m_max_elements; }

    /**
     * @brief Insert the element into the queue if possible.
     *
     * Does not insert the elemenent if an element with this key is either:
     *    -# Already in the queue.
     *    -# Or it was marked by prohibit_insertion().
     * 
     *  Complexity is amortized O(1).
     *  @pre The key part of the element must be less than get_max_elements(),
     *       otherwise throw out_of_range.
     *
     *  @return  The returned boolean indicates success or failure.
     */
    bool insert(value_argument_type element)
    {
        HCSP_PROFILE_CURRENT;
        using namespace std;
        key_type element_id= element;
        if ( is_in_queue(element_id) )
            return false;
        // Using a restorer rather than try..catch due to alleged performance
        // issues with try..catch on visual C++.
        restore_on_exit<unsigned, vector<unsigned>::reference,
                        vector<unsigned>::pointer>
            restorer(m_chunk_elements_in_queue[element_id/CHUNK_BITS]);
        set_key_in_queue(element_id);
        m_unique_queue.push_back(element); // Might throw bad_alloc
        restorer.commit();      // push_back was OK, no need to restore.
        return true;
    }

    /**
     * @brief Prohibit future insertions of the said element into the queue.
     *
     * Notes:
     *   -# This call is forgotten after clear().
     *   -# This call is forgotten after set_max_elements(n) for n<=element_id
     *   -# Ignoring the call if there is already an element_id in the queue.
     *
     * Complexity is O(1).
     * @pre The key element must be less than get_max_elements(),
     *      otherwise throw out_of_range.
     */
    void prohibit_insertion(key_type element_id)
    {
        if ( is_in_queue(element_id) )
            return;
        set_key_in_queue(element_id);
    }

    //! \brief Tell whether there are elements in the queue.
    bool empty() const { return m_unique_queue.empty(); }

    /**
     * @brief Pop the earliest element in the queue off the queue, and
     * return its value.
     *
     * Complexity is O(1)
     * @pre !empty()
     */
    value_type pop()
    {
        HCSP_PROFILE_CURRENT;
        if ( empty() )
            throw std::out_of_range("Popping out of an empty queue");
        value_type element= m_unique_queue.front();
        key_type element_id= element;
        
        m_unique_queue.pop_front();
        HCSP_ASSERT( element_id < m_max_elements );
        HCSP_ASSERT( is_in_queue(element_id) );
        clear_key_in_queue(element_id);

        return element;
    }

    /**
     * @brief Clear all elements in the queue.
     *
     * Also clear all marks of prohibit_insertion(),
     * but do not loose the set_max_elements().
     *
     * Complexity is O(max_elements)
     */
    void clear()
    {
        m_unique_queue.clear();
        size_t size = m_chunk_elements_in_queue.size();
        m_chunk_elements_in_queue.clear();
        m_chunk_elements_in_queue.resize(size, 0);
    }

    //! \brief Swap two queues at O(1)
    void swap(unique_queue & other)
    {
        m_unique_queue.swap(other.m_unique_queue);
        m_chunk_elements_in_queue.swap(other.m_chunk_elements_in_queue);
        std::swap(m_max_elements, other.m_max_elements);
    }
  private:
    void validate_all_elements_below(unsigned smallest_illegal) const;
    //! Number of bits represented by one entry of m_chunk_elements_in_queue
    enum { CHUNK_BITS= sizeof(unsigned)*8 };
    
    std::deque<value_type> m_unique_queue;

    //! Check if an element with this key is already in the queue.
    bool is_in_queue(key_type key) const
    {
        if (key >= m_max_elements)
            // Force std::out_of_range exception, without 
            // #include <stdexcept>
            m_chunk_elements_in_queue.at(m_chunk_elements_in_queue.size());
        unsigned position_in_chunk= key%CHUNK_BITS;
        unsigned chunk= m_chunk_elements_in_queue.at(key/CHUNK_BITS);
        return  (chunk >> position_in_chunk) & 1;
    }

    //! Mark that an element with this key is in the queue.
    void set_key_in_queue(key_type key)
    {
        unsigned position_in_chunk= key%CHUNK_BITS;
        unsigned & chunk= m_chunk_elements_in_queue[key/CHUNK_BITS];
        chunk |= (1U << position_in_chunk);
    }

    //! Mark that an element with this key is not in the queue.
    void clear_key_in_queue(key_type key)
    {
        unsigned position_in_chunk= key%CHUNK_BITS;
        unsigned & chunk= m_chunk_elements_in_queue[key/CHUNK_BITS];
        chunk &= ~(1U << position_in_chunk);
    }
    
    /**
     * Keeps keeps for each potential element if it is already in the queue
     * or not. Each m_elements_in_queue[i] has a set of bits where each bit
     * keeps whether or not referred is in queue or not.
     *
     * Not using vector<bool> because even gcc-4.2 resizes vector<bool>
     * one bit at a time (which is slow!!)
     */
    std::vector<unsigned>  m_chunk_elements_in_queue;
    
    size_t m_max_elements;
};

template <class T>
void unique_queue<T>::validate_all_elements_below(unsigned smallest_illegal)
    const
{
    using namespace std;
    bool any_true= false;
    size_t smallest_all_illegal_chunk= smallest_illegal/CHUNK_BITS;
    if (smallest_illegal % CHUNK_BITS == 1)
        ++smallest_all_illegal_chunk;
    
    for ( size_t i= smallest_illegal ;
          i < m_max_elements && i % CHUNK_BITS != 0 ;
          ++i) {
        any_true= (any_true || is_in_queue(i));
    }
    for ( size_t i= smallest_all_illegal_chunk ;
          i != m_chunk_elements_in_queue.size() && !any_true;
          ++i) {
        any_true= m_chunk_elements_in_queue[i];
    }
    if ( ! any_true )
        return;
    
    // The true value in m_elements_in_queue may have one of two origins.
    //  1. A call to insert(). This will mean that there is a variable
    //     with an ID in the forbidden range.
    //  2. A call to prohibit_insertion(). This means that we simply have
    //     to forget the prohibition.
    // To distinguish between the two options, we have to traverse the queue.
    typename std::deque<value_type>::const_iterator it;
    for ( it= m_unique_queue.begin(); it != m_unique_queue.end(); ++it) {
        if ( smallest_illegal <= *it )
            throw out_of_range("At least one element is too big");
    }
}


} // namespace hcsp


#endif // HCSP_UNIQUE_QUEUE_HPP_
