// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_UTILITIES_HPP_
#define HCSP_UTILITIES_HPP_

#include <iostream>
#include <stdint.h>
#include <stdexcept>
#include <memory>

#ifdef __GNUC__
#    define HCSP_UNUSED __attribute__((unused))
#else
#    define HCSP_UNUSED
#endif

namespace hcsp {
#if defined(__GNUC__) && defined(__i386__) && !defined(__SSE__)
     inline double round_to_double(double in)
     {
         volatile double tmp= in;
         return tmp;
     }
#else
     inline double round_to_double(double in) { return in; }
#endif

// Not using (or even looking at) boost, in order to avoid license issues.
//! Inherit from this class to disable copy ctor and assignment.
class no_copy {
  public:
    no_copy() {}
  private:
    no_copy(const no_copy &) = delete;
    no_copy & operator=(const no_copy &) = delete;
};

//! Restores a given object to original value on scope exit.
//! The object will be restored either due to regular exit or to exception.
template <class T, class Reference= T&, class Pointer= T*>
class restore_on_exit : private no_copy {
  public:
    restore_on_exit(Reference object) : m_p_object(&object), m_orig(object) {}
    ~restore_on_exit()
    {
        if (m_p_object)
            *m_p_object= m_orig;
    }
    void commit() { m_p_object= NULL; }
  private:
    Pointer  m_p_object;
    const    T m_orig;
};

template <class Container>
class swap_on_exit : private no_copy
{
  public:
    swap_on_exit(Container &a, Container &b) : m_a(a), m_b(b), m_attached(true)
    {}
    void detach() { m_attached= false; }
    ~swap_on_exit()
    {
        if (m_attached)
            m_a.swap(m_b);
    }
  private:
    Container & m_a, &m_b;
    bool m_attached;
};
//
// Declarations needed for HCSP_STATIC_ASSERT
// 
// Not using (or even looking at) boost, in order to avoid license issues.
template <bool> struct DidStaticAssertPass { };
template<> struct DidStaticAssertPass<true> { enum { assertion_passed= 1 }; };

#define HCSP_IMPL_CONCAT_TOKENS(a,b)  a##b
#define HCSP_IMPL_CONCAT_TOKENS2(a,b) HCSP_IMPL_CONCAT_TOKENS(a,b)
#define HCSP_IMPL_AS_STRING(a)        #a
#define HCSP_IMPL_AS_STRING2(a)       HCSP_IMPL_AS_STRING(a)

// Using typedefs, to avoid errors with global HCSP_STATIC_ASSERT in
// header files - where there are assersions in different files on the
// same line number.
// Typedefs allows non-conflicting redeclarations.
#define HCSP_STATIC_ASSERT(expr)                                 \
    typedef char HCSP_IMPL_CONCAT_TOKENS2(                       \
        HCSP_StaticAssertType_AT_LINE_,__LINE__)                 \
        [hcsp::DidStaticAssertPass<(expr)>::assertion_passed]    \
        HCSP_UNUSED

#define HCSP_IMPL_CONST_AS_STR2(num) #num
#define HCSP_IMPL_CONST_AS_STR(num) HCSP_IMPL_CONST_AS_STR2(num)

#ifdef NDEBUG
#define HCSP_ASSERT(cond) (void)0
#else
#define HCSP_ASSERT(cond)                                                \
    do {                                                                \
        if (!(cond))                                                    \
            throw_runtime_error(__FILE__ ":"                            \
                                HCSP_IMPL_CONST_AS_STR(__LINE__) ": "    \
                                "assertion("#cond") failed");           \
    } while (0)
#endif

//
// Declarations needed for HCSP_STATIC_ASSERT_SAME_TYPE
//

template <class T1, class T2>
struct assert_same_type_impl      {};

template <class T>
struct assert_same_type_impl<T,T> { static const bool value = true; };

#define HCSP_STATIC_ASSERT_SAME_TYPE(T1, T2)                     \
    typedef char HCSP_IMPL_CONCAT_TOKENS2(                       \
        HCSP_StaticAssertSameTypeType_AT_LINE_,__LINE__)         \
        [hcsp::assert_same_type_impl<T1,T2>::value]              \
        HCSP_UNUSED


class parse_error : public std::runtime_error
{
  public:
    explicit parse_error(const std::string & text)
        : std::runtime_error(text)
    {}
    // Declared, so that the virtual table will go to only to utilities.o
    virtual ~parse_error() throw();
};

//! A RAII class that restores the state of the stream before it exits
class ios_state_restore
{
  public:
    ios_state_restore(std::basic_ios<char> &);
    ~ios_state_restore();
  private:
    std::ios_base::fmtflags m_saved_flags;
    std::ios_base::iostate m_saved_state;
    std::basic_ios<char> & m_stream;
};

void throw_runtime_error(const char *msg);
//! Find the index of the lowest set bit.
//! For zero input return total number of bits
unsigned find_lsb_index(uint32_t bits);

//! Find the index of the highest set bit.
//! For zero input return zero
unsigned find_msb_index(uint32_t bits);

unsigned count_ones(uint32_t bits);

//! Count the number of independent intervals of 1's.
unsigned count_ranges_of_one(uint32_t bits);

//! Give a value that has all its bits set to '1' on and above the index.
//! \precondition Template class T must be an unsigned basic type.
template <class T>
inline T unsigned_ones_on_and_above(unsigned index)
{
    // Make sure its unsigned
    HCSP_STATIC_ASSERT(static_cast<T>(static_cast<T>(0)-1) > 0);
    
    const T        all_ones= ~static_cast<T>(0);
    const unsigned num_bits= sizeof(T)*8;
    
    if (index >= num_bits)
        return 0;
    else
        return all_ones << index;
}

//! Give a value that has all its bits set to '1' strictly above the index.
//! \precondition Template class T must be an unsigned basic type.
template <class T>
inline T unsigned_ones_above(unsigned index)
{
    // Make sure its unsigned
    HCSP_STATIC_ASSERT(static_cast<T>(static_cast<T>(0)-1) > 0);
    
    const T        mostly_ones= (~static_cast<T>(0)) - 1;
    const unsigned num_bits=    sizeof(T)*8;
    
    if (index >= num_bits)
        return 0;
    else
        return mostly_ones << index;
}


//! Give a value that has all its bits set to '1' on and below the index.
//! \precondition Template class T must be an unsigned basic type.
template <class T>
inline T unsigned_ones_on_and_below(unsigned index)
{
    // Make sure its unsigned
    HCSP_STATIC_ASSERT(static_cast<T>(static_cast<T>(0)-1) > 0);
    
    const T        all_ones= ~static_cast<T>(0);
    const unsigned num_bits= sizeof(T)*8;
    
    if (index >= num_bits)
        return all_ones;
    else
        return all_ones >> (num_bits-1-index);
}

//! Give a value that has all its bits set to '1' strictly below the index.
//! \precondition Template class T must be an unsigned basic type.
template <class T>
inline T unsigned_ones_below(unsigned index)
{
    // Make sure its unsigned
    HCSP_STATIC_ASSERT(static_cast<T>(static_cast<T>(0)-1) > 0);
    
    const T        all_ones=    ~static_cast<T>(0);
    const T        mostly_ones= all_ones >> 1;
    const unsigned num_bits=    sizeof(T)*8;
    
    if (index >= num_bits)
        return all_ones;
    else
        return mostly_ones >> (num_bits-1-index);
}

template <class T>
inline bool is_power_of2(T value)
{
    // This implementation is based on the fact that:
    //   (2^n)-1 = 2^0+2^1+2^2+...+2^(n-1)
    // which means that [ (2^n)-1 bitwise_and (2^n) ] == 0
    // for x=y+(2^n), where 0<y<2^n, MSB(x)==2^n and
    // MSB(x-1)=MSB((2^n)+y-1)==2^n
    // which means that MSB((x-1) bitwise_and x)==2^n
    // which implies that ((x-1) bitwise_and x) != 0 for x which has more than
    // one bit set. Zero needs a special treatment.
    return value > 0 && ((value-1) & value)==0;
}

template <class T>
inline void free_container(T & data)
{
    T().swap(data);
}

//! Calculate greatest common denominator, of unsigned values
inline unsigned gcd(unsigned a, unsigned b)
{
    while (a != 0) {
        unsigned next_a= b%a;
        b=a;
        a=next_a;
    }
    return b;
}

/** @brief Calculate Bézout coefficients for Bézout's identity.
 *
 * For the equation \f$ x*a + y*b == gcd(a,b) \f$, find the values of
 * x and y such that x>=0 and x is the smallest possible x to satisfy this
 * requirement.
 *
 * For more detail look at:
 *    http://en.wikipedia.org/wiki/B%C3%A9zout%27s_identity
 *
 * @param[in] a,b     Non-negative values for which Bézout coefficients
 *                    are to be returned.
 * @param[out] p_gcd  If not NULL, then return the Greatest Common Denominator
 *                    (GCD) at *p_gcd;
 * @return  make_pair(x,y) which are the Bézout coefficients as calculated by
 *          the algorithm.
 */
std::pair<long long, long long> bezout_coefficients(unsigned a, unsigned b,
                                                    unsigned * p_gcd=NULL);

/** @brief Solve the equation a*x - b*y == rhs.
 *
 * When there is one solution there are infinite solutions. This function
 * returns all of the solution in a form of a series parameters:
 * Each integer k defines a different solution:
 *   a*x' - b*y' == rhs
 * such that
 *   x'= x+step_x*k   and y'= y+step_y*k
 * (note that for k=0 we get x'=x and y'=y).
 *
 * If b != 0 then x>=0 and x<step_x.
 *
 * @param[in]   a,b  The coefficients of the equation (a,b >= 0)
 * @param[out]  x,y  The base solution (x,y >= 0)
 * @param[out]  step_x, step_y  The distance between successive solutions (>= 0)
 * @param[out]  p_step_gcd  If not NULL then *p_step_gcd returns
 *              gcd(step_x,step_y)
 *                               
 * @return true when the problem has a solution, otherwise false.
 */
bool linear_diophantine_equation(unsigned a, unsigned b, unsigned rhs,
                                 long long & x, unsigned  & step_x,
                                 long long & y, unsigned  & step_y,
				 unsigned *p_step_gcd=NULL);

//! Calculate floor(dividend/divisor)
inline long long floor_div(long long dividend, long long divisor)
{
    if ( (dividend>=0) == (divisor>=0) || dividend==0)
        // Rounding to zero for positive result is the same as floor.
        return dividend/divisor; 
    else if (dividend < 0)
        // Rounding to zero for negative result is the same as ceil
        return (dividend+1)/divisor-1;
    else // dividend>0 && divisor <0
        return (dividend-1)/divisor-1;
}

//! Calculate ceil(dividend/divisor)
inline long long ceil_div(long long dividend, long long divisor)
{
    if ( (dividend>=0) != (divisor>=0) || dividend == 0)
        // Rounding to zero for negative result is the same as ceil
        return dividend/divisor; 
    else if (dividend < 0) // dividend<0 && divisor <0
        // Rounding to zero for negative result is the same as floor
        return (dividend+1)/divisor+1;
    else // dividend>=0 && divisor >=0
        return (dividend-1)/divisor+1;
}

template <class T> T
clone_all(const T & vec)
{
    T ret;
    for (const auto & elem : vec) {
        ret.emplace_back(elem->clone());
    }
    return ret;
}

template <class T> struct conversions;

template <> struct conversions<int>
{
    typedef int      to_signed;
    typedef unsigned to_unsigned;
};

template <> struct conversions<long>
{
    typedef long          to_signed;
    typedef unsigned long to_unsigned;
};

template <> struct conversions<short>
{
    typedef short          to_signed;
    typedef unsigned short to_unsigned;
};

template <> struct conversions<unsigned> :       public conversions<int>   {};
template <> struct conversions<unsigned long> :  public conversions<long>  {};
template <> struct conversions<unsigned short> : public conversions<short> {};

template <class T>
inline typename conversions<T>::to_unsigned distance(T a, T b)
{
    typename conversions<T>::to_unsigned distance= b;
    distance -= a;
    return distance;
}

template <class T>
inline
typename conversions<T>::to_unsigned distance(const std::pair<T,T> & range)
{
    return distance(range.first, range.second);
}


class profile_data : private no_copy
{
  public:
    typedef unsigned ticks_type;
    profile_data(const char *name);
    ~profile_data();
    void record_call(ticks_type start_time)
    {
        m_total_ticks_in_scope += m_global_clock_ticks - start_time;
        ++m_scope_exits;
    }
    static ticks_type global_clock_ticks() { return m_global_clock_ticks; }

    static void Impl_sigprof_handler();
  private:
    static void activate_clock();
        
    static const unsigned TICKS_PER_SECOND= 100;
    static volatile ticks_type m_global_clock_ticks;
    static bool m_clock_active;
        
    char     *m_p_label;
    unsigned m_total_ticks_in_scope;
    unsigned m_scope_exits;
};

class profile_call : private no_copy
{
  public:
    profile_call(profile_data & summary)
        : m_summary(summary), m_start_time(profile_data::global_clock_ticks())
    {}
    ~profile_call() { m_summary.record_call(m_start_time); }
  private:
    profile_data & m_summary;
    profile_data::ticks_type m_start_time;
};

#ifdef RUN_HCSP_PROFILER
#define HCSP_PROFILE_CURRENT                                             \
        static profile_data profile_data_for_func(__PRETTY_FUNCTION__); \
        profile_call profile_call_for_func(profile_data_for_func)
#define HCSP_PROFILE_LINE                                                      \
        static profile_data                                                   \
            HCSP_IMPL_CONCAT_TOKENS2(profile_data_for_line,__LINE__) (         \
                __FILE__ ":" HCSP_IMPL_AS_STRING2(__LINE__));                  \
        profile_call HCSP_IMPL_CONCAT_TOKENS2(profile_call_for_line,__LINE__)( \
            HCSP_IMPL_CONCAT_TOKENS2(profile_data_for_line,__LINE__))
#else
#define HCSP_PROFILE_CURRENT (void)0
#define HCSP_PROFILE_LINE    (void)0
#endif
bool trace_active();    
#ifdef ACTIVATE_HCSP_TRACE
#   define HCSP_TRACE(msg) do {                                          \
        if (trace_active())                                             \
            std::cout << __FILE__ << ":" << __LINE__ << ":" << msg << "\n"; \
    } while (0)
#else
#   define HCSP_TRACE(msg) (void)0
#endif
#define HCSP_RUNTIME_ERROR(text) throw std::runtime_error(                   \
                                  __FILE__ ":"                              \
                                  HCSP_IMPL_CONST_AS_STR(__LINE__) ": "      \
                                  text)

typedef unsigned long long microseconds_t;
//! Number of microseconds of user CPU time spent by the process so far.
microseconds_t user_time_microseconds();

// template <typename T>
// inline std::unique_ptr<T> make_unique(T* ptr)
// { return std::unique_ptr<T>(ptr); }

namespace random {              // hcsp::random
    unsigned gen();
    unsigned gen(unsigned size);
    unsigned range(unsigned from, unsigned to);
    int range(int from, int to);
    unsigned set_seed(int argc, char *argv[]);
    unsigned set_seed(unsigned seed);

    // For use by std::random_shuffle
    struct randomizer {
        unsigned operator()(unsigned n) const { return gen(n); }
    };

} // namespace hcsp::random

} // namespace hcsp

#endif // HCSP_UTILITIES_HPP_

