// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2014, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2014, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef VAR_WATCHERS_HPP_
#define VAR_WATCHERS_HPP_
#include <vector>
#include <list>
#include <type_traits>
#include <pool_alloc.hpp>
#include <small_iset.hpp>
#include <iset.hpp>
namespace hcsp {

class watch_handle_t;

class var_watchers_base
{
protected:
    /**
     * @brief A class containing basic information on a constraint
     * that has this watch.
     */
    struct watcher_info_type
    {
        operator unsigned() const { return m_constraint_index; }
        unsigned m_constraint_index;
    };

    typedef std::list<watcher_info_type, pool_alloc<watcher_info_type> > 
        watch_list_t;
    friend class watch_handle_t;

    /**
     * @brief  Watches bound to a given key.
     *
     * This struct contains both a key and the set of watchers that are
     * triggered by that key.
     */
    struct key_watchers
    {
        key_watchers(key_watchers&&) noexcept = default;
        key_watchers(const key_watchers&) = default;
        key_watchers & operator=(key_watchers&& other) noexcept
        {
            m_key = other.m_key;
            m_watchers = move(other.m_watchers);
            return *this;
        }
        key_watchers & operator=(const key_watchers&) = default;
        key_watchers() = default;
        iset::value_type m_key;
        watch_list_t m_watchers;
    };
}; // class var_watchers_base

class watch_handle_t
{
  private:
    typedef var_watchers_base::watch_list_t watch_list_t;
    static watch_list_t null_watch_list;
    friend class var_watchers_base;

  public:
    watch_handle_t(watch_list_t::iterator it) : m_it(it) {}
    watch_handle_t() : m_it(null_watch_list.end()) {}
    watch_handle_t(const watch_handle_t &) = default;
    watch_handle_t(watch_handle_t &&) noexcept = default;
    watch_handle_t& operator=(const watch_handle_t &) = default;
    watch_handle_t& operator=(watch_handle_t &&) noexcept = default;
    bool operator==(watch_handle_t other) const { return m_it == other.m_it; }
    
    void clear() { m_it = null_watch_list.end(); }
    bool empty() const { return m_it == null_watch_list.end(); }
    watch_list_t::iterator get_it() const { return m_it; }

  private:
    watch_list_t::iterator m_it;
}; // watch_handle_t

/**
 * @brief This class manages the list of watching constraints (i.e. watchers)
 * for one variable.
 *
 * It contains enough information to know what
 * constraints are waiting for a given change in the variable.
 */
template <class Set>
class var_watchers_impl : protected var_watchers_base
{
  public:
    struct cond_always_true 
    { 
        bool operator()(iset::value_type) const { return true; } 
    };

    var_watchers_impl(bool is_fifo) 
        : m_watch_count(0), m_in_for_each(false), m_is_fifo(is_fifo) {}
    //! \brief return true if there are no watches.
    bool empty() const { return m_watch_count == 0; }
    unsigned count_all_watches() const { return m_watch_count; }
    
    /**
     * @brief deallocate a watch with the given watch_handle.
     *
     * @param key        The key that describes the watch.
     * @param watch_handle The watch to be deallocated.
     *
     * @pre watch_handle must be a valid watch_handle [validated by default]
     *
     */
    void deallocate_watch(
        iset::value_type key, watch_handle_t watch_handle);

    
    template <class Condition>
    void deallocate_watches(const Condition & cond);
    /**
     * @brief allocate a watch for the given constraint.
     *
     * @param constraint_index  The constraint that will own the watch.
     *
     * @pre key              the key part of watch description.
     * @pre constraint_index must be not an already allocated
     *      constraint_index [not validated by default]
     *
     * @return the handle of the new entry in the list of variable's watchers
     */
    watch_handle_t allocate_watch(iset::value_type key, 
                                  unsigned constraint_index);

    /**
     * @brief For each watching constraint_id run op(constraint_id).
     *
     *  Stop the loop if op(constraint_id) returned false.
     *
     *  This method is reentrant such that op() may call any other method of
     *  the same var_watchers object, with no limitations.  We allow
     *  op(constraint_id) to add or remove watchs. If watches are added
     *  or removed in the middle of this method, then there is no guarantee:
     *   -# That the added watches will be visible (or not) in the current
     *      for_each session.
     *   -# That watches will not be traversed more than once.
     *
     *  What is guaranteed is that for_each will terminate regardless of
     *  what op() does or does not do.
     *
     *  @param op             A functor that should operate on each watched
     *                        constraint.
     *  @param relevant_keys  Set of relevant types to operate on.
     *  @return true if all activations of op() returned true, false otherwise.
     */
    template <class Operation, class KeyCond=cond_always_true>
    bool for_each_until_false(const Operation & op, 
                              const KeyCond & cond = cond_always_true())
        const;

    /**
     * @brief For each watching constraint_id, run op(constraint_id).
     *
     * Unlike 
     * @link hcsp::var_watchers::for_each_until_false for_each_until_false
     * @endlink,
     * ignore the value returned by op(constraint_id), and ignore its type.
     * For more details, look at
     * @link hcsp::var_watchers::for_each_until_false for_each_until_false
     * @endlink.
     */
    template <class Operation>
    void for_each(const Operation & op,
                  const Set & relevant_keys= Set().complement() ) const
    { for_each_until_false(
            call_true_func<Operation>(op), 
            [=](iset::value_type key) { return relevant_keys.contains(key); });}
    
    bool watches(typename Set::value_type value) const
    {
        for (const auto & element : m_key_watchers)
            if (element.m_key == value)
                return true;
        return false;
    }

  private:
    /**
     * \brief A helper predicate that first calls op(), and then
     * always returns true.
     */
    template <class Predicate>
    class call_true_func
    {
      public:
        call_true_func(const Predicate & op) : m_op(op) {}
        template<class T>
        bool operator()(const T& parameter) { m_op(parameter); return true; }
      private:
        Predicate m_op;
    };

    
    /**
     * @brief This class manages entry and exit from for_each functions.
     *
     * - If in a nested call to for_each, do nothing (neither during entry
     *   nor during exit).
     * - Otherwise, upon entry it will mark p_watches->m_in_for_each.
     * - And, upon exit unmark m_in_for_each and remove entries of dead keys.
     */
    class for_each_cleaner
    {
      public:
        for_each_cleaner(const var_watchers_impl *p_watches)
            : m_p_watches(p_watches->m_in_for_each ? NULL : p_watches)
        { p_watches->m_in_for_each= true; }
        
        ~for_each_cleaner();
      private:
        const var_watchers_impl * m_p_watches;
    };
     
    template <class Operation>
    bool for_each_impl(const Operation & op, unsigned watchers_index) const;
    
    key_watchers *find_key_watchers(iset::value_type key)
    {
        const var_watchers_impl * const_this= this;
        const key_watchers *ret= const_this->find_key_watchers(key);
        return const_cast<key_watchers*>(ret);
    }
    const key_watchers *find_key_watchers(iset::value_type key) const;
        
    //! \brief check if constraint_index already watches this variable.
    bool is_key_watched_by_constraint(iset::value_type key,
                                      unsigned constraint_index) const;
    
    std::vector<key_watchers> m_key_watchers;
    unsigned                  m_watch_count;
    mutable bool              m_in_for_each;
    const bool                m_is_fifo;
};

template <class Set> template <class Operation>
bool var_watchers_impl<Set>::for_each_impl(const Operation & op_in,
                                           unsigned watchers_index) const
{
    HCSP_PROFILE_CURRENT;
#ifndef NDEBUG
    iset::value_type orig_key= m_key_watchers[watchers_index].m_key;
#endif

    if (m_key_watchers[watchers_index].m_watchers.empty())
        return true;
    Operation op(op_in);        // This will allow non-const operator().
    bool last = false;
    // Going backwards in order to maintain the same traversal order
    // as before.
    auto it= m_key_watchers[watchers_index].m_watchers.end();
    --it;
    auto begin_it = m_key_watchers[watchers_index].m_watchers.begin(); 
    while (!last)
    {
        unsigned orig_watcher= it->m_constraint_index;
        last = (it == begin_it);
        auto prev_it = last ? it : prev(it);
        if ( !op(orig_watcher) )
            return false;
        // We are locking this data structure against deletion of keys while
        // performing for_each. This means that the keys themselves will not
        // be deleted.
        HCSP_ASSERT( watchers_index < m_key_watchers.size() );
        HCSP_ASSERT( orig_key == m_key_watchers[watchers_index].m_key );
        it = prev_it;
    }
    return true;
}

template <class Set> template <class Condition>
void var_watchers_impl<Set>::deallocate_watches(
    const Condition & cond)
{
    HCSP_PROFILE_CURRENT;
    
    for ( unsigned watchers_pos= m_key_watchers.size(); watchers_pos!=0; ){
        --watchers_pos;
        iset::value_type key= m_key_watchers[watchers_pos].m_key;
        auto & watchers = m_key_watchers[watchers_pos].m_watchers;
        if (watchers.empty())
            continue;
        auto it = prev(watchers.end());
        bool last = false;
        while (!last)
        {
            unsigned watcher= it->m_constraint_index;
            last = it == watchers.begin();
            auto prev_it = last ? it : prev(it);
            if ( cond(watcher) )
                deallocate_watch(key, watch_handle_t(it));
            it= prev_it;
        }
    }
}

    template <class Set> template <class Operation, class KeyCond>
bool var_watchers_impl<Set>::for_each_until_false( const Operation & op,
                                                   const KeyCond & cond) const
{
    HCSP_PROFILE_CURRENT;
    // On construction, mark m_in_for_each (if not in a nested for_each).
    // On destruction, clean up and unmark m_in_for_each (unless nested)
    for_each_cleaner clean_up_object(this);

    // Must go backwards to allow allocation of watches
    // from within op(). This way for_each_until_false() is guaranteed
    // to stop, and to traverse all original elements - at least once.
    unsigned i= m_key_watchers.size();
    while ( i !=0 ) {     
        iset::value_type orig_key= m_key_watchers[i-1].m_key;
        if ( cond(orig_key) ) {
            if (!for_each_impl(op, i-1))
                return false;
        }
        --i;
        // Keys are not really deleted while in for_each, the clean_up_object
        // will take care of this when all nested for_each instances exit.
        HCSP_ASSERT( i < m_key_watchers.size() );
    }
    return true;
}

class var_value_watchers : public var_watchers_impl<iset>
{
public:
    var_value_watchers() : var_watchers_impl<iset>(true) {}
};

class var_event_watchers : public var_watchers_impl<small_iset>
{
public:
    var_event_watchers() : var_watchers_impl<small_iset>(false) {}
};

} // namespace hcsp

#endif // VAR_WATCHERS_HPP_
