// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2011, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2011, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_VARIABLE_PRINT_STRATEGY_HPP_
#define HCSP_VARIABLE_PRINT_STRATEGY_HPP_

#include <iosfwd>

namespace hcsp {

class variable_printer;
/**
 * @brief This strategy decides how a variable identification is to be printed.
 *
 * The printing is usually done according to variable's name in the solver,
 * but can be something different if needs to.
 */
class variable_print_strategy
{
  public:
    virtual ~variable_print_strategy() {}
    virtual void print(std::ostream & out, unsigned var_index) const=0;
    
    //! Get a printer object for printing a single variable
    //! (see class variable_printer).
    variable_printer printer(unsigned var_index) const;
}; // class variable_print_strategy

/**
 * @brief An object that uses variable_print_strategy for printing a variable 
 *
 * This object prints one variable, passed to it during construction.
 * It prints only the name or ID of the variable, not the domain.
 *
 * Usage example:
 * @code
 *   out << variable_printer(my_strategy, variable_index);
 * @endcode
 */
class variable_printer
{
  public:
    /**
     * @brief Construct a variable printer
     * @param strategy  The strategy used to print variable's name
     * @param variable  The ID of the variable.
     */
    variable_printer(const variable_print_strategy & strategy,
                     unsigned variable)
        : m_strategy(&strategy), m_variable(variable) {}

    variable_printer(): m_strategy(NULL), m_variable(0) {}

    //! Is this a valid printer, otherwise it is a dummy place-holder
    bool is_valid() const { return m_strategy; }
    
    //! Using this printer, print variable's name
    friend std::ostream & operator<<(std::ostream & out,
                                     const variable_printer & printer)
    {
        if (printer.is_valid())
            printer.m_strategy->print(out, printer.m_variable);
        else
            out << "NULL_var_printer";
        return out;
    }
  private:
    const variable_print_strategy * m_strategy; //!< The strategy to use
    unsigned m_variable;                        //!< The variable to print
}; // class variable_printer

inline
variable_printer variable_print_strategy::printer(unsigned var_index) const
{
    return variable_printer(*this, var_index);
}

} // namespace hcsp

#endif // HCSP_VARIABLE_PRINT_STRATEGY_HPP_
