// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef WATCH_TYPES_HPP_
#define WATCH_TYPES_HPP_

#include <utilities.hpp>
#include <iset.hpp>
#include <var_watchers.hpp>

namespace hcsp {

/**
 * @brief This helper struct defines types which several watch related classes
 *        want to see.
 *
 * To use these types, simply inherit from this struct. Note that if the
 * child is a template class, then inheritence is not trivial and may
 * require using directives.
 */
struct watch_types
{
    //! \brief The type of the watch.
    enum watch_type {
        MODIFICATION_WATCH=0,  //!< This watch is activated on any modification
        FINALIZATION_WATCH,    //!< Activated when domain becomes singleton .
        LOWER_BOUND_WATCH,     //!< Activated when min() of a domain changes
        UPPER_BOUND_WATCH,     //!< Activated when max() of a domain changes
    };
    //! \brief A struct that can be used by a propagator for bookkeeping
    //!        general watches.
    struct watch_info_type
    {
        //! The parameter that is watched. If negative then this entry
        //! is invalid.
        int      m_param;
        //! The watch id, as returned by allocate_watch
        watch_handle_t m_watch_handle;
    };
    //! \brief A struct that can be used by a propagator for bookkeeping
    //!        value watches.
    struct value_watch_info_type
    {
        //! \brief Does this struct mark a parameter?
        //! A parameter may be unmarked, marked, or marked but not watched.
        //! The difference is that mark is for internal use, and a watch used
        //! by the solver to decide when to activate a constraint.
        bool is_valid_mark()  const { return m_param >= 0; }

        //! Make this struct unmarked and unwatched. See is_valid_mark()
        void invalidate_mark()
        {
            m_param= -1;
            m_watch_handle.clear();
        }

        //! Does this struct refer to a watch? Even if not it still may
        //! mark a parameter. See is_valid_mark();
        bool is_valid_watch() const { return !m_watch_handle.empty(); }

        //! Make this struct not refer to a watch. Even if not it still may
        //! mark a parameter. See is_valid_mark();
        void invalidate_watch() { m_watch_handle.clear(); }

        //! The parameter that is marked and, maybe, watched. If negative then
        //! this entry is invalid and this is neither a watch nor a mark.
        int              m_param;

        //! The watch id, as returned by allocate_watch
        //! It is -1U if this is not a watch, but then it still may be a mark.
        watch_handle_t m_watch_handle;
        //! The watched value.
        iset::value_type m_value;
    };
    friend std::ostream & operator<<(std::ostream & out, watch_type data);
}; // struct watch_types


} // namespace hcsp

#endif // WATCH_TYPES_HPP_
