// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <utilities.hpp>
#include "trivial_runtime_info.hpp"
#include "explanation_tester.hpp"
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;

// virtual
explanation_tester::result_type
explanation_tester::propagate( constraint_runtime_info & info )
{
    trivial_runtime_info old_state(info);

    bool known_to_conflict= m_constraint->conflicts(info);
    result_type result= m_constraint->propagate(info);
    T_ASSERT(known_to_conflict == (result == UNSAT));
    if ( result == UNSAT )
        assert_unsat_explanation(old_state);
    else {
        for (unsigned i=0 ; i != info.arity() ; ++i) {
            if ( info.domain(i) != old_state.domain(i) )
                assert_change_explanation(info, old_state, i);
        }
    }
    return result;
}

//virtual
void explanation_tester::explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain,
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation) const
{
    m_constraint->explain_propagation(propagated_param, values_to_explain,
                                      propagated_domain,
                                      info_before_propagation,explanation);
    assert_consistent_rule(*m_constraint, explanation);
}

void explanation_tester::assert_consistent_rule(
    const constraint_propagator & propagator,
    const vector<iset> & explanation) const
{
    unique_ptr<explanation_rule> p_rule= propagator.rule(explanation);
    if ( !p_rule.get() )
        return;
    std::vector<iset>  rule_explanation;
    p_rule->get_explanation(rule_explanation);
    T_ASSERT(explanation == rule_explanation);
}

void explanation_tester::assert_unsat_explanation(
    const trivial_runtime_info & state) const
{
    vector<iset> explanation;
    iset values_to_explain= random_var();
    m_constraint_clone->explain_propagation(0, values_to_explain, iset(),
                                            state, explanation);
    T_ASSERT(state.arity() == explanation.size() );
    assert_consistent_rule(*m_constraint_clone, explanation);

    // Soundness: Make sure the explanation returns UNSAT
    for (unsigned i=0 ; i !=  explanation.size() ; ++i)
        T_ASSERT(!explanation[i].contains_any_of(state.domain(i)));

    // Completeness: Make sure that the explanation does not eliminate
    //               feasible solutions. 
    trivial_runtime_info infeasible(explanation);
    for (unsigned i=0 ; i != explanation.size() ; ++i)
        infeasible.set_domain(i, move(iset(infeasible.domain(i)).complement()));
    T_ASSERT(!has_solution(infeasible));
}

void explanation_tester::assert_change_explanation(
    const constraint_runtime_info & info,
    const trivial_runtime_info & old_state,
    unsigned param) const
{
    vector<iset> explanation;
    iset values_to_explain=  gen(2)==0 ? iset().complement() : random_var();
    iset removed_values= old_state.domain(param);
    removed_values.erase_all_of(info.domain(param));
    if ( !values_to_explain.contains_any_of(removed_values) )
        values_to_explain.insert(removed_values.rand());
    iset removed_values_to_explain= removed_values;
    removed_values_to_explain.intersect(values_to_explain);
    m_constraint_clone->explain_propagation(param, values_to_explain,
                                            info.domain(param),
                                            old_state, explanation);
    assert_consistent_rule(*m_constraint_clone, explanation);
    T_ASSERT(old_state.arity() == explanation.size() );
    // Soundness: Make sure the explanation removes at least the bad
    //            values that the propagation does. In theory it could
    //            remove more, if the propagation was inaccurate.
    for (unsigned i=0 ; i !=  explanation.size() ; ++i) {
        if (i == param)
            T_ASSERT(!explanation[i].contains_any_of(
                         removed_values_to_explain));
        else
            T_ASSERT(!explanation[i].contains_any_of(old_state.domain(i)));
    }
    assert_explanation_no_solution_lost(info, old_state, param, explanation);
}


void explanation_tester::assert_explanation_no_solution_lost(
    const constraint_runtime_info & info,
    const trivial_runtime_info & old_state,
    unsigned param,
    const vector<iset> & explanation) const
{
    for (unsigned i=0 ; i !=  explanation.size() ; ++i) {
        iset validated_values= old_state.domain(i);
        if ( i == param )
            validated_values.erase_all_of(info.domain(i));
        validated_values.insert_all_of(explanation[i]);
        if ( validated_values.everything() )
            continue;
        
        // This part of the explanation is stronger than the the minimal
        // requirement. Need to make sure that it did not lose solutions.
        vector<iset> complement(explanation);
        complement[i]= validated_values;
        for (unsigned j=0 ; j != complement.size() ; ++j)
            complement[j].complement();
        T_ASSERT(!has_solution(complement));
    }
}

bool explanation_tester::has_solution(const trivial_runtime_info & old_state)
    const
{
    trivial_runtime_info state(old_state);
    bool may_be_sat= m_constraint_clone->init(state);
    bool is_proven_sat=false;
    if (may_be_sat) {
        state.advance_modified(true);
        state.commit_variables();
        switch( m_constraint_clone->propagate(state)) {
          case UNSAT: may_be_sat= false;                              break;
          case VALID: is_proven_sat= true;                            break;
          case CONSISTENT: is_proven_sat= state.are_all_singletons(); break;
          default: ; // nothing
        }
    }

    // Attempt trivial propagation, first.
    for (unsigned i=0 ; may_be_sat && !is_proven_sat && i<state.arity() ; ++i){
        if (state.domain(i).singleton())
            continue;
        
        state.advance_modified(false);
        iset domain = state.domain(i);
        domain.assign(domain.rand());
        state.set_domain(i, move(domain));
        state.advance_modified(true);
        state.commit_variables();
        switch( m_constraint_clone->propagate(state)) {
          case UNSAT: may_be_sat= true;    break;
          case VALID: is_proven_sat= true; break;
          default: ; // nothing
        }
    }        
    m_constraint_clone->forget_watches();
    if (is_proven_sat)
        return true;
    if (!may_be_sat)
        return false;
    // FIXME: should be 'return has_complete_problem_solution(old_state);'
    return false;
}

