// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_EXPLANATION_TESTER_HPP_
#define HCSP_EXPLANATION_TESTER_HPP_

#include <memory>
#include <vector>
#include <iset.hpp>
#include <constraint_propagator.hpp>

namespace hcsp {

class trivial_runtime_info;
/**
 * This wrapper embeds a propagator and tests its explanation mechanism.
 * For the outside viewer this object behaves exactly as the embedded object,
 * but internally it calls the embedded explain_propagation and
 * validates its correctness.
 */
class explanation_tester : public constraint_propagator
{
  public:
    /**
     * Construct an explanation testing constraint out of the given propagator.
     */
    explanation_tester(std::unique_ptr<constraint_propagator> constraint)
        : m_constraint(constraint.release()),
          m_constraint_clone(m_constraint->clone())
    {
    }
    
    /**
     * Construct an explanation testing constraint out of the given propagator.
     *
     * @param p_constraint   A pointer to the tested constraint. It is the
     *                       responsibility this explanation_tester to
     *                       free up the memory associated with this parameter.
     */
    explanation_tester(constraint_propagator* p_constraint)
        : m_constraint(p_constraint), m_constraint_clone(p_constraint->clone())
    {
    }

    virtual bool init( constraint_runtime_info & info )
    { return m_constraint->init(info); }


    /**
     * Call the embedded propagator's propagate() method and validate
     * explanations. The state prior to propagate() is recorded and later
     * explain_propagation() is called. The results of explain_propagation()
     * are validated.
     */
    virtual result_type propagate( constraint_runtime_info & info );

    virtual bool conflicts(const constraint_runtime_info & info) const
    { return m_constraint->conflicts(info); }
    virtual bool constraint_check(const constraint_runtime_info & info) const
    { return m_constraint->constraint_check(info); }
    
    virtual void print(std::ostream &out,
                       const constraint_runtime_info *p_info=NULL)const throw()
    { m_constraint->print(out, p_info); }

    virtual unsigned watched_param_index(unsigned watch_index) const
    { return m_constraint->watched_param_index(watch_index); }

    virtual unsigned count_watches() const
    { return m_constraint->count_watches(); }
    
    virtual void forget_watches()
    { m_constraint->forget_watches(); }
    
    virtual void explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain,
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation) const;
    virtual std::unique_ptr<explanation_rule> rule(
        const std::vector<iset> & literals) const
    {
        return m_constraint->rule(literals);
    }

    virtual explanation_tester * clone() const
    {
        return new explanation_tester(m_constraint->clone());
    }
    
  private:
    /**
     * Make sure that the explanation rule generated for this explanation
     * is consistent with it.
     */
    void assert_consistent_rule(
        const constraint_propagator & propagator,
        const std::vector<iset> & explanation) const;
    /**
     * Make sure that the UNSAT state is explained correctly by a subsequent
     * call to explain_propagation().
     */
    void assert_unsat_explanation(const trivial_runtime_info & state) const;
    /**
     * Make sure that the propagation of \e param is explained
     * correctly by a subsequent call to explain_propagation().
     *
     * @param info        The domains after the propagation.
     * @param old_state   The domains before the propagation.
     * @param param       The parameter which propagation must be explained..
     */
    void assert_change_explanation(const constraint_runtime_info & info,
                                   const trivial_runtime_info & old_state,
                                   unsigned param) const;

    /**
     * Make sure that the given explanation does not lose solutions.
     *
     * If this method throws no exception then it has proven that
     * the given explanation does not remove feasible solutions.
     * @param info        The domains after the propagation.
     * @param old_state   The domains before the propagation.
     * @param param       The parameter which is explained.
     * @param explanation The explanation vector that has to be validated.
     */
    void assert_explanation_no_solution_lost(
        const constraint_runtime_info & info,
        const trivial_runtime_info & old_state,
        unsigned param,
        const std::vector<iset> & explanation) const;
    
    /**
     * Tell whether the given domains have any feasible assignment under
     * this constraint.
     *
     * Note that a constructor of trivial_runtime_info from vector<iset>
     * will create an implicit support for has_solution(const vector<iset> &)
     * calls.
     *
     * @NOTE
     *   For complex cases has_solution() may return \b false instead of false.
     *   If you find a case when this actually happens, please fix this
     *   method.
     */
    bool has_solution(const trivial_runtime_info & old_state) const;
    
    std::unique_ptr<constraint_propagator> m_constraint;
    std::unique_ptr<constraint_propagator> m_constraint_clone;
};



} // namespace hcsp

#endif // HCSP_EXPLANATION_TESTER_HPP_
