// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_PROPAGATOR_TESTER_HPP_
#define HCSP_PROPAGATOR_TESTER_HPP_

#include <vector>
#include <string>
#include <memory>
#include <algorithm>
#include <iostream>
#include <constraint_runtime_info.hpp>
#include "test_utilities.hpp"
#include "explanation_tester.hpp"
#include "trivial_runtime_info.hpp"

namespace hcsp { namespace test {
using std::vector;
using std::string;
using std::cout;
using std::min;
using std::unique_ptr;
using std::make_pair;
using hcsp::random::gen;
template <class Propagator>
class propagator_tester
{
  public:
    propagator_tester(unsigned arity)
        : m_info(arity), m_accurate(true)
    {
        m_p_inner_propagator= new Propagator;
        construct();
    }
    template <class T>
    propagator_tester(unsigned arity, T p)
        : m_info(arity), m_accurate(true)
    {
        m_p_inner_propagator= new Propagator(p);
        construct();
    }
    template <class T0, class T1>
    propagator_tester(unsigned arity, const T0 & p0, T1 p1)
        : m_info(arity), m_accurate(true)
    {
        m_p_inner_propagator= new Propagator(p0, p1);
        construct();
    }
    template <class T0, class T1>
    propagator_tester(unsigned arity, T0 & p0, T1 p1)
        : m_info(arity), m_accurate(true)
    {
        m_p_inner_propagator= new Propagator(p0, p1);
        construct();
    }
    template <class T0, class T1, class T2>
    propagator_tester(unsigned arity, const T0 & p0, const T1 & p1,
                      const T2 & p2)
        : m_info(arity), m_accurate(true)
    {
        m_p_inner_propagator= new Propagator(p0, p1, p2);
        construct();
    }
    //! Run the tests
    void run();

    //! Mark the propagation inaccurate. Return *this.
    propagator_tester & inaccurate()
    {
        m_accurate= false;
        return *this;
    }
    //! Set the domain of the given variable.
    template <class T>
    propagator_tester & set_dom(unsigned variable, const T & new_domain)
    {
        iset subset(new_domain);
        subset.intersect(random_nonempty_var());
        if (subset.empty())
            m_info.set_domain(variable, iset(new_domain));
        else
            m_info.set_domain(variable, std::move(subset));
        return *this;
    }
  private:
    typedef trivial_runtime_info::result_type result_type;

    void construct()
    {
        set_random_domains();
        m_p_propagator.reset(m_p_inner_propagator);
        if (gen(10) == 0)
            m_p_propagator.reset(
                new explanation_tester(m_p_propagator.release()));
    }
    /**
     * @brief Activate the propagator once and test the result.
     *
     * @return Whether a feasible assignment exists.
     */
    bool test_one_propagation();

    //! Give an initial random sets to the domains.
    void set_random_domains();

    /**
     * @brief Make sure that the given info has no feasible assignment
     *
     * Not all erroneous cases are detected. It is possible that there
     * is a solution for the given info but this test will not fail.
     * The inaccuracy arises when domains are too big.
     */
    void validate_unsat(const constraint_runtime_info & info) const;
    /**
     * @brief Make sure that all domain values in info have support.
     *
     * Not all erroneous cases are detected. It is possible that some domain
     * value has no solution, but it will not be detected.
     * The inaccuracy arises when domains are too big.
     */
    void validate_consistent(const constraint_runtime_info & info) const;

    /**
     * @brief Make sure that explanations do not reduce more than propagation.
     *
     * Make sure that for similar inputs the constraint is at least as strong
     * as the explanation clause. When propagating the explanation it should
     * never remove more values than the constraint in the same situation.
     *
     * Notes:
     *  -# This assertion is valid only if the constraint is accurate.
     *  -# The purpose of this function to test propagation accuracy,
     *     not explanation correctness (it assumes that the explanations
     *     have been tested elsewhere).
     */
    void validate_explanation_not_stronger(
        const constraint_runtime_info &prev_info,
        const constraint_runtime_info &next_info) const;

    void select_values_to_explain(iset & values, const iset & except) const;
    /**
     * @brief Make sure there is no solution for the given info
     *
     * This method check all possible assignments, exhaustively.
     * It mutates the input info, and restores it upon return.
     *
     * @pre All parameters before param are singletons.
     */
    void validate_all_values_unsat(trivial_runtime_info & info,
                                   unsigned param) const;

    /**
     * @brief Make sure that if variable is assigned with value then
     *        there is a solution.
     *
     * This method currently works only if m_accurate or if
     * are_all_singletons(). In the future, when it will support
     * !m_accurate, then it will not always catch errors.
     */
    void validate_value_support(unsigned variable, iset::value_type value,
                                const constraint_runtime_info & info,
                                vector<iset> & supported_domains) const;

    /**
     * @brief Make sure that the current state contains a feasible solution.
     *
     * This method currently works only if m_accurate. In the future,
     * when it will support !m_accurate, then it will not always
     * catch errors.
     */
    void validate_solvable(const trivial_runtime_info & info,
                           vector<iset> & supported_domains) const;

    //! Randomize an assignment of values from info to singleton_info.
    void random_assign(const constraint_runtime_info & info,
                       trivial_runtime_info & singleton_info) const;


    static const result_type UNSAT= trivial_runtime_info::UNSAT;
    static const result_type CONSISTENT= trivial_runtime_info::CONSISTENT;
    unique_ptr<constraint_propagator> m_p_propagator;
    Propagator          *m_p_inner_propagator;
    trivial_runtime_info m_info;
    bool       m_accurate;
};

/** @brief A class to report the context of errors
 *
 * This class follows the RAII paradigm. When it is destroyed it prints
 * details of the propagator and its context. The class is supposed to print
 * this information only when an error occurs, but not when all goes well.
 * If the code exits cleanly then it should call no_fail()
 * which will disable the printing. On the other hand if an exception is thrown
 * before no_fail() is called then the context will be printed right before
 * the stack is unwound.
 */
class error_state_reporter
{
  public:
    /** @brief Construct the error reporter, based on the given context.
     *
     * @pre info and propagator must belong to an outer C++ scope so that
     *      when error_state_reporter is destroyed they will still exist.
     * @param info,propagator   The context of the error to be reported
     * @param msg   An optional message, supposedly the name of the context.
     */
    error_state_reporter(const constraint_runtime_info & info,
                         const constraint_propagator & propagator,
                         const string & msg= "")
        : m_info(info),
          m_propagator(propagator),
          m_msg(msg),
          m_to_report(true)
    {}
    
    //! Report the context unless no_fail() was called.
    ~error_state_reporter() {
        if (!m_to_report)
            return;
        if (!m_msg.empty())
            cout << m_msg << ": ";
        cout << "error exit on ";
        m_propagator.print(cout, &m_info);
        cout << "\n";
        for (unsigned i=0 ; i != m_info.arity() ; ++i)
            cout << i << ": " << m_info.domain(i) << "\n";
    }
    //! Tell this object that no reporting should take place
    void no_fail() { m_to_report= false; }
  private:
    
    const constraint_runtime_info & m_info;
    const constraint_propagator & m_propagator;
    const string m_msg;
    bool m_to_report;
    
}; // class error_state_reporter

template <class Propagator>
void propagator_tester<Propagator>::run()
{
    error_state_reporter reporter(m_info, *m_p_propagator, "run()");
    bool init_ok= m_p_propagator->init(m_info);
    unsigned iterations_completed=0;
    vector<trivial_runtime_info> backtrack_info;
    unsigned num_backtracks= 4;
    while (true) {
        bool solvable= test_one_propagation();
        if (solvable)
            ++iterations_completed;
        bool final_state= !solvable || m_info.are_all_singletons();
        if ( num_backtracks >0 && !backtrack_info.empty() &&
             (final_state || gen(20/num_backtracks)==0)) {
            unsigned target= gen(backtrack_info.size());
            m_info.backtrack_to(backtrack_info[target]);
            backtrack_info.resize(target, m_info);
            --num_backtracks;
            final_state= false;
        }
        if (final_state)
            break;
        
        backtrack_info.push_back(m_info);
        m_info.random_reduce();
    }
    reporter.no_fail();
    T_ASSERT(init_ok || iterations_completed == 0);
}

template <class Propagator>
void propagator_tester<Propagator>::set_random_domains()
{
    for (unsigned i=0 ; i != m_info.arity() ; ++i)
        m_info.set_domain(i, random_nonempty_var());
}

template <class Propagator>
bool propagator_tester<Propagator>::test_one_propagation()
{
    error_state_reporter reporter(m_info, *m_p_propagator,
                                  "test_one_propagation()");
    const constraint_runtime_info & parent_orig_info= m_info;
    trivial_runtime_info prev_info( parent_orig_info );
    result_type result= m_info.maybe_propagate(*m_p_propagator);
    if (result == UNSAT) {
        validate_unsat(prev_info);
        reporter.no_fail();
        return false;
    }
    if (m_accurate || m_info.are_all_singletons() )
        validate_consistent(m_info);
    if ( m_accurate && !m_info.are_all_singletons() )
        validate_explanation_not_stronger(prev_info, m_info);
    
    for (unsigned i=0 ; i != m_info.arity() ; ++i) {
        if (!m_info.reduced_between_activations(i))
            continue;
        T_ASSERT(m_info.prev_domain(i) != m_info.domain(i));
        T_ASSERT(m_info.prev_domain(i).contains_all_of(m_info.domain(i)));

        // Test that values were removed correctly
        const constraint_runtime_info & parent_prev_info= prev_info;
        trivial_runtime_info unsat_info( parent_prev_info );
        unsat_info.forget_watches();
        iset new_domain= unsat_info.domain(i);
        new_domain.erase_all_of(m_info.domain(i));
        unsat_info.set_domain(i, std::move(new_domain));
        if (unsat_info.domain(i).empty())
            continue;
        unique_ptr<Propagator> p_propagator( m_p_inner_propagator->clone() );
        p_propagator->init(unsat_info);
        result= unsat_info.maybe_propagate(*p_propagator);
        if (result == UNSAT)
            break;
        T_ASSERT(!m_accurate);
        validate_unsat(unsat_info);
    }
    reporter.no_fail();
    return true;
}

template <class Propagator>
void propagator_tester<Propagator>::validate_all_values_unsat(
    trivial_runtime_info & info, unsigned param) const
{
    error_state_reporter reporter(info, *m_p_propagator,
                                  "validate_all_values_unsat()");
    for ( ; param != info.arity() ; ++param) 
        if (!info.domain(param).singleton()) {
            iset orig= info.domain(param);
            for (iset::iterator it= orig.begin() ; it != orig.end() ; ++it) {
                info.set_domain(param, iset(*it));
                validate_all_values_unsat(info, param);
            }
            info.set_domain(param, std::move(orig));
            reporter.no_fail();
            return;
        }
    if (info.are_all_singletons())
        T_ASSERT(!m_p_inner_propagator->constraint_check(info));
    reporter.no_fail();
}

template <class Propagator>
void propagator_tester<Propagator>::validate_value_support(
    unsigned variable, iset::value_type value,
    const constraint_runtime_info & info,
    vector<iset> & supported_domains) const
{
    T_ASSERT(info.domain(variable).contains(value));
    if (info.domain(variable).singleton())
        validate_solvable(info, supported_domains);
    else {
        trivial_runtime_info narrowed_info= info;
        narrowed_info.set_domain(variable, iset(value));
        validate_solvable(narrowed_info, supported_domains);
    }
}

template <class Propagator>
void propagator_tester<Propagator>::validate_solvable(
    const trivial_runtime_info & info,
    vector<iset> & supported_domains) const
{
    error_state_reporter reporter(info, *m_p_propagator, "validate_solvable()");
    if (info.are_all_singletons()) {
        T_ASSERT(m_p_inner_propagator->constraint_check(info));
        for (unsigned i=0 ; i != supported_domains.size() ; ++i)
            supported_domains[i].insert_all_of(info.domain(i));
        reporter.no_fail();
        return;
    }
    trivial_runtime_info work_info= info;
    unique_ptr<Propagator> p_propagator(m_p_inner_propagator->clone());
    T_ASSERT(p_propagator->init(work_info));
    T_ASSERT(work_info.maybe_propagate(*p_propagator) != UNSAT);
    T_ASSERT(m_accurate);
    unsigned next_work_param=0;
    while (!work_info.are_all_singletons()) {
        while ( next_work_param != work_info.arity() &&
                work_info.domain(next_work_param).singleton() )
            ++next_work_param;
        const iset & supported= supported_domains[next_work_param];
        iset next_domain= work_info.domain(next_work_param);
        if ( !supported.contains_all_of(next_domain) ) {
            // Make next found assignment contribute new supported_domains.
            next_domain.erase_all_of(supported);
        }
        iset::value_type value = next_domain.rand();
        work_info.set_domain(next_work_param, iset(value));
        T_ASSERT(work_info.maybe_propagate(*p_propagator) != UNSAT);
    }
    T_ASSERT(m_p_inner_propagator->constraint_check(work_info));
    for (unsigned i=0 ; i != supported_domains.size() ; ++i)
        supported_domains[i].insert_all_of(info.domain(i));
    reporter.no_fail();
}


template <class Propagator>
void propagator_tester<Propagator>::validate_unsat(
    const constraint_runtime_info & info) const
{
    error_state_reporter reporter(info, *m_p_propagator, "validate_unsat()");
    trivial_runtime_info tested= info;
    const unsigned attempts= 300;
    unsigned allowable_domain= attempts;
    for (unsigned i=0 ; i != info.arity() ; ++i) {
        const iset & domain= info.domain(i);
        T_ASSERT(!info.domain(i).empty());
        if (domain.size() == 0)
            // This can happen even when the domain is not empty due to
            // an overflow in the size calculation. The only case that 
            // is relevant as of this writing is when iset contains
            // all the range [MIN_INT:MAX_INT].
            allowable_domain= 0;
        else
            allowable_domain /= domain.size();
    }
    trivial_runtime_info mutable_info= info;
    if (allowable_domain)
        validate_all_values_unsat(mutable_info, 0);
    else {
        trivial_runtime_info singleton_info= info;
        for (unsigned i=0 ; i != attempts ; ++i) {
            random_assign(info, singleton_info);
            T_ASSERT(!m_p_inner_propagator->constraint_check(singleton_info));
        }
    }
    reporter.no_fail();
}


template <class Propagator>
void propagator_tester<Propagator>::validate_consistent(
    const constraint_runtime_info & info) const
{
    error_state_reporter reporter(info, *m_p_propagator,
                                  "validate_consistent()");
    // This has resemblance to safer_propagator::propagate() at
    // t_all_different_propagator.cpp:
    // Figure out which values of info have support.
    // Do it in an efficient way, such that any feasible solution found
    // has its values added to supported_domains vector. Also prefer to
    // look for feasible assignments that have the least resemblance to
    // supported_domains.
    vector<iset> supported_domains(info.arity());
    for (unsigned i=0 ; i != supported_domains.size() ; ++i) {
        iset values_to_check= info.domain(i);
        values_to_check.erase_all_of(supported_domains[i]);
        for ( unsigned j=0 ; j != 100 && !values_to_check.empty() ; ++j) {
            iset::value_type val= values_to_check.rand();
            values_to_check.erase(val);
            if (!values_to_check.empty())
                validate_value_support(i, val, info, supported_domains);
        }
    }
    reporter.no_fail();
}

template <class Propagator>
void propagator_tester<Propagator>::select_values_to_explain(
    iset & values, const iset & except_in) const
{
    iset except = except_in;
    T_ASSERT(values.contains_all_of(except));
    unsigned number_of_relevant = values.size()-except.size();
    if (number_of_relevant == 0) {
        values.clear();
        return;
    }
    unsigned number_of_values_to_retain= 1;
    if (number_of_relevant > 1)
        number_of_values_to_retain= gen(number_of_relevant-1)+1;

    if (number_of_relevant < 10) {
        values.erase_all_of(except);
        except.clear();
    }
    if ( values.singleton() )
        return;
    while (values.size()-except.size() > number_of_values_to_retain) {
        unsigned left_to_remove=values.size()-number_of_values_to_retain;
        if ( left_to_remove < 100 && except.size() < left_to_remove) {
            iset::value_type value = values.rand();
            if (!except.contains(value))
                values.erase(value);
        } else {
            iset::value_type first= values.rand();
            iset::value_type second= min<long long>(
                first + left_to_remove/10LL, iset::MAX_DOMAIN);
            values.erase(make_pair(first, second));
            except.erase(make_pair(first, second));
        }
    }
}

template <class Propagator>
void propagator_tester<Propagator>::validate_explanation_not_stronger(
    const constraint_runtime_info &prev_info,
    const constraint_runtime_info &next_info) const
{
    vector<iset> explanation;
    error_state_reporter reporter(prev_info, *m_p_propagator,
                                  "validate_explanation_not_stronger()");
    for (unsigned i=0 ; i != prev_info.arity() ; ++i) {
        if ( prev_info.domain(i) == next_info.domain(i) )
            continue;
        iset explain= prev_info.domain(i);
        select_values_to_explain(explain, next_info.domain(i));
        if ( explain.singleton() )
            // This case is not interesting since this was checked
            // during regular explanation testing
            continue;
        m_p_propagator->explain_propagation(i, explain, next_info.domain(i),
                                            prev_info, explanation);
        trivial_runtime_info explanation_info(explanation);
        explanation_info.set_domain(i, iset());
        for (unsigned j=0 ; j != explanation.size() ; ++j)
            explanation_info.set_domain(
                j, iset(explanation_info.domain(j)).complement());
        unique_ptr<Propagator> p_propagator( m_p_inner_propagator->clone() );
        p_propagator->init(explanation_info);
        
        // At this point 'explanation_info' makes sure that the explanation
        // clause is a unit clause, which should propagate to parameter 'i'.
        // The domains of 'explanation_info' are the largest possible,
        // retaining the required property.
        if (p_propagator->propagate(explanation_info) == p_propagator->UNSAT)
            continue; // UNSAT is stronger than the explanation clause
        // Assume that 'explanation' is valid, and only check that it is not
        // stronger than p_propagator, i.e., p_propagator removed all
        // values that 'explanation' clause would.
        T_ASSERT( explanation[i].contains_all_of(explanation_info.domain(i)) );
    }
    reporter.no_fail();
}

template <class Propagator>
void propagator_tester<Propagator>::random_assign(
    const constraint_runtime_info & info,
    trivial_runtime_info & singleton_info) const
{
    for (unsigned i=0 ; i != info.arity() ; ++i)
        singleton_info.set_domain(i, iset(info.domain(i).rand() ));
}

} } // namespace hcsp::test

#endif // HCSP_PROPAGATOR_TESTER_HPP_
