// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <algorithm>
#include <iomanip>
#include <utilities.hpp>
#include <assignment_stack.hpp>
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;

class tester
{
  public:
    static const bool empty_domains_supported= false;
    typedef assignment_stack::change_index_t change_index_t;
    //
    // Public variables
    //
    static bool      m_trace;

    //
    // Public methods
    //
    tester();
    void run();
    static void trivial_test();
  private:
    //
    // Private types
    //
    typedef iset::value_type            value_type;
    typedef assignment_stack::time_type time_type;

    //
    // Private methods
    // 

    void run_depth(unsigned depth);
    void run_flat_sequence(unsigned op_count, vector<iset> & current_values,
                           vector<time_type> & current_times,
                           vector<int> & current_causes);
    static void trivial_compare(const assignment_stack & trivial_stack,
                                unsigned level,
                                const iset & e1, const iset & e2);
    void make_decision(unsigned & decision_variable,
                       value_type & decision_value);
    void undo_level(unsigned level_to_forget);
    void var_decision(unsigned decision_variable, const iset & domain);
    void set_var_domain(unsigned var, const iset & value, int cause);
    void assert_at_initial_state() const;
    void assert_current_values(const vector<iset> & current_values,
                               const vector<time_type> & current_times,
                               const vector<int> & current_causes) const;
    void assert_current_decision(unsigned decision_variable,
                                 const value_type decision_value) const;
    void get_current_values(vector<iset> & current_values,
                            vector<time_type> & current_times,
                            vector<int> & current_causes) const;
    
    const unsigned   m_num_variables;
    vector<iset>     m_initial_domains;
    assignment_stack m_assignment_stack;
    time_type        m_current_time;
    unsigned         m_level;
    unsigned         m_backjump_to_level;
}; // class tester

bool tester::m_trace= false;

tester::tester() : m_num_variables(range(1, 8))
{
    m_current_time=0;
    m_level=0;
    m_initial_domains.reserve(m_num_variables);
    m_backjump_to_level= gen(2)==0 ? 1000 : range(3,10);

    for (unsigned i=0 ; i != m_num_variables; ++i) {
        if (empty_domains_supported)
            m_initial_domains.push_back(random_var());
        else
            m_initial_domains.push_back(random_nonempty_var());
        m_assignment_stack.add_var(m_initial_domains.back());
    }
    if (m_trace) {
        for (unsigned i=0 ; i != m_num_variables ; ++i) {
            cout << "initial " <<i << "={" << m_initial_domains[i] << "}\n";
        }
    }
}

void tester::run()
{
    try {
        assert_at_initial_state();
        run_depth(50);
        m_assignment_stack.clear();
        assert_at_initial_state();
        run_depth(50);
    } catch (exception & e) {
        cout << "Exception(" << typeid(e).name() << "):\n" << e.what() << endl;
        for (unsigned i=0 ; i != m_assignment_stack.count_variables() ; ++i) {
            cout << i << ": " << m_assignment_stack.var_modification_time(i)
                 << " {" << m_assignment_stack.var_domain(i) << "}\n"; 
        }
        throw;
    }
}

void tester::trivial_test()
{
    assignment_stack tested;
    const iset domain03(make_pair(0,3));
    const iset domain02(make_pair(0,2));
    
    tested.add_var(domain03);
    tested.add_var(domain03);

    trivial_compare(tested, 0, domain03, domain03);
    
    tested.set_var_domain(0, domain02, 1);
    trivial_compare(tested, 0, domain02, domain03);
    
    {
        tested.var_decision(1,2, iset(0));
        trivial_compare(tested, 1, domain02, iset(0));
        {
            tested.var_decision(0, 2, iset(1));
            trivial_compare(tested, 2, iset(1), iset(0));
            tested.set_var_domain(1,domain02, 4);
        
        }
        tested.undo_level(1);
    }
    trivial_compare(tested, 0, domain02, domain03);
}

void tester::run_depth(unsigned depth)
{
    if ( depth == 0 )
        return;
    vector<iset> current_values;
    vector<time_type> current_times;
    vector<int> current_causes;
    get_current_values(current_values, current_times, current_causes);
    unsigned entry_decision_level= m_level;
    unsigned entry_decision_var=
        m_level==0 ? 0 : m_assignment_stack.decision_variable();
    value_type entry_decision_val=
        m_level==0 ? 0 : m_assignment_stack.decision_value();
    
    unsigned total_flat_depth= range(0U, depth-1);
    unsigned first_flat_depth;
    switch (gen(4)) {
      case 0:  first_flat_depth= 0;                break;
      case 1:  first_flat_depth= total_flat_depth; break;
      default: first_flat_depth= range(0U, total_flat_depth);
    }
    unsigned last_flat_depth= total_flat_depth - first_flat_depth;
    run_flat_sequence(first_flat_depth, current_values, current_times,
                      current_causes);
    assert_current_decision(entry_decision_var, entry_decision_val);

    if (m_level == 0) {
        T_ASSERT_EXCEPTION(m_assignment_stack.decision_variable(),
                           out_of_range);
        T_ASSERT_EXCEPTION(m_assignment_stack.decision_value(), out_of_range);
    }

    if (m_level == 0 || m_num_variables > 1) {
        unsigned decision_variable;
        value_type decision_value;
        make_decision(decision_variable, decision_value);

        unsigned depth_remaining= depth - 1 - total_flat_depth;
        while (depth_remaining > 0 && gen(2)==0) {
            unsigned depth_first= gen(depth_remaining);
            run_depth(depth_first);
            depth_remaining -= depth_first;
        }
        run_depth(depth_remaining);
        if ( entry_decision_level > m_backjump_to_level)
            return;
        undo_level(entry_decision_level+1);
    }
    run_flat_sequence(last_flat_depth, current_values, current_times,
                      current_causes);
    assert_current_decision(entry_decision_var, entry_decision_val);
}

void tester::run_flat_sequence(unsigned op_count,
                               vector<iset> &current_values,
                               vector<time_type> & current_times,
                               vector<int>  & current_causes)
{
    unsigned time_advances_left= range(0U, op_count);
    if ( m_level > 0 && m_num_variables <= 1 )
        return;
    for (unsigned i=0 ; i != op_count ; ++i) {
        if ( gen(op_count-i) < time_advances_left ) {
            ++m_current_time;
            --time_advances_left;
        }
        unsigned reduced_var;
        do {
            reduced_var= gen(m_num_variables);
        } while (m_level > 0 &&
                 reduced_var == m_assignment_stack.decision_variable());
        
        if (empty_domains_supported)
            current_values[reduced_var]= random_var();
        else
            current_values[reduced_var]= random_nonempty_var();
        
        current_causes[reduced_var]= -1;
        if (gen(2) == 0)
            current_causes[reduced_var]= gen();
        current_times[reduced_var]=  m_current_time;
        set_var_domain(reduced_var, current_values[reduced_var],
                       current_causes[reduced_var]);
    }
    assert_current_values(current_values, current_times, current_causes);
}

// static
void tester::trivial_compare(const assignment_stack & trivial_stack,
                             unsigned level,
                             const iset & e1, const iset & e2)
{
    T_ASSERT(trivial_stack.level() == level);
    T_ASSERT(trivial_stack.var_domain(0) == e1);
    T_ASSERT(trivial_stack.var_domain(1) == e2);
    change_index_t modification0= trivial_stack.var_modification_entry(0);
    change_index_t modification1= trivial_stack.var_modification_entry(1);
    if (modification0 != assignment_stack::NO_PREVIOUS_MODIFICATION)
        T_ASSERT(trivial_stack.modifying_constraint( modification0 )==-1);
    if (modification1 != assignment_stack::NO_PREVIOUS_MODIFICATION)
        T_ASSERT(trivial_stack.modifying_constraint( modification1 )==-1);
    
}

void tester::make_decision(unsigned & decision_variable,
                           value_type & decision_value)
{
    T_ASSERT(m_level == 0 || m_num_variables > 1);
    do {
        decision_variable= gen(m_num_variables);
    } while ( m_level > 0 &&
              m_assignment_stack.decision_variable() == decision_variable );

    ++m_current_time;
    decision_value= random_nonempty_var().rand();
    var_decision(decision_variable, iset(decision_value) );
    T_ASSERT( m_assignment_stack.level() == m_level );
    T_ASSERT( m_assignment_stack.var_domain(decision_variable) ==
              iset(decision_value));
    T_ASSERT( m_assignment_stack.var_modification_time(decision_variable) ==
              m_current_time);
}

void tester::undo_level(unsigned level_to_forget)
{
    if ( m_trace )
        cout << setw(m_level*4) << ""
             << "undo_level(" << level_to_forget << ")\n";
    m_assignment_stack.undo_level(level_to_forget);
    m_level= level_to_forget-1;
}

void tester::var_decision(unsigned decision_variable, const iset & domain)
{
    ++m_level;
    if ( m_trace )
        cout << setw(m_level*4) << ""
             << "var_decision(" << decision_variable
             << ", " << m_current_time << ", {" << domain  << "} )="
             << m_level << "\n";
    m_assignment_stack.var_decision(decision_variable, m_current_time,
                                    domain);
}

void tester::set_var_domain(unsigned var, const iset & value, int cause)
{
    if ( m_trace )
        cout << setw(m_level*4) << ""
             << "var_domain_update(" << var << ", " << m_current_time
             << ")= {" << value << "}\n";
    if (cause == -1 && gen(2) == 0)
        m_assignment_stack.set_var_domain(var, value, m_current_time);
    else
        m_assignment_stack.set_var_domain(var,value, m_current_time, cause);
}

void tester::assert_at_initial_state() const
{
    T_ASSERT(m_num_variables == m_assignment_stack.count_variables());
    for (unsigned i=0 ; i != m_num_variables ; ++i)
        T_ASSERT(m_assignment_stack.var_domain(i) == m_initial_domains[i]);
}

void tester::assert_current_values(const vector<iset> & current_values,
                                   const vector<time_type> & current_times,
                                   const vector<int> & current_causes)
    const
{
    T_ASSERT( m_num_variables == m_assignment_stack.count_variables() );
    T_ASSERT( m_num_variables == current_values.size() );
    T_ASSERT( m_num_variables == current_times.size() );
    T_ASSERT( m_num_variables == current_causes.size() );

    for (unsigned i=0 ; i != m_num_variables ; ++i) {
        T_ASSERT( m_assignment_stack.var_domain(i) == current_values[i] );
        T_ASSERT( m_assignment_stack.var_modification_time(i) ==
                  current_times[i] );
        if (current_times[i] > 1) {
            change_index_t entry= m_assignment_stack.var_modification_entry(i);
            T_ASSERT(entry != m_assignment_stack.NO_PREVIOUS_MODIFICATION);
            T_ASSERT(m_assignment_stack.modifying_constraint(entry) ==
                     current_causes[i]);
        }
        
    }
}

void tester::assert_current_decision(unsigned decision_variable,
                                     const value_type decision_value)
    const
{
    T_ASSERT(m_level == m_assignment_stack.level());
    if (m_assignment_stack.level() == 0) {
        T_ASSERT_EXCEPTION(m_assignment_stack.decision_variable(),
                           out_of_range);
        T_ASSERT_EXCEPTION(m_assignment_stack.decision_value(),
                           out_of_range);
    } else {
        T_ASSERT( decision_variable== m_assignment_stack.decision_variable() );
        T_ASSERT( decision_value   == m_assignment_stack.decision_value() );
    }
    
}


void tester::get_current_values(vector<iset> & current_values,
                                vector<time_type> & current_times,
                                vector<int> & current_causes) const
{
    T_ASSERT(m_initial_domains.size() == m_assignment_stack.count_variables());
    const unsigned size= m_assignment_stack.count_variables();
    current_times.resize(size);
    current_values.clear();
    current_values.resize(size);
    current_causes.resize(size);
    for (unsigned i=0 ; i != m_assignment_stack.count_variables() ; ++i) {
        current_values[i]= m_assignment_stack.var_domain(i);
        current_times[i]= m_assignment_stack.var_modification_time(i);
        change_index_t entry= m_assignment_stack.var_modification_entry(i);
        if (entry == m_assignment_stack.NO_PREVIOUS_MODIFICATION)
            current_causes[i]= -1;
        else
            current_causes[i]= m_assignment_stack.modifying_constraint(entry);
    }
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    cout << "Seed is " << set_seed(argc, argv) << endl;
    tester::m_trace= false;
    tester::trivial_test();
    for (unsigned i=0 ; i < 3000 ; ++i) {
        if (i != 0) {
            unsigned seed= gen();
            if (tester::m_trace)
                cout << "Seed is " << seed << endl;
            srand(seed);
        }
        tester().run();
    }
    END_TEST_MAIN;
}
