// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <iostream>
#include <sstream>
#include <vector>
#include <algorithm>
#include <iset.hpp>
#include <clause_propagator.hpp>
#include <solver.hpp>
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;

class tester {
  public:
    static const unsigned N = 8;
    tester(solver::reporting_levels level);
    void run();
    void reset() { m_solver.reset(); }
    unsigned backtracks_done() const { return m_solver.backtracks_done(); }

  private:
    void print_solution() const;
    void validate_solution() const;

    static unique_ptr<clause_propagator>
    make_clause(const iset & a, const iset & b);

    void add_shuffled_clause(unique_ptr<clause_propagator> clause,
                             int var_a, int var_b, int var_c=-1);
    void add_maybe_split_clause(unique_ptr<clause_propagator> clause,
                                int var_a, int var_b);
    void add_diff(unsigned var_a, unsigned var_b);
    void add_minus_not(unsigned var_a, unsigned var_b, int forbidden_minus);
    solver m_solver;
};

unique_ptr<clause_propagator>
tester::make_clause(const iset & a, const iset & b)
{
    unique_ptr<clause_propagator> clause( new clause_propagator );
    clause->update_literals().reserve(3);
    clause->update_literals().push_back(a);
    clause->update_literals().push_back(b);
    return clause;
}

void tester::add_diff(unsigned var_a, unsigned var_b)
{
    const iset complete_set(make_pair(0, N-1));
    
    for (unsigned i=0 ; i < N ; ++i) {
        iset not_i=complete_set;
        not_i.erase(i);
        add_maybe_split_clause(make_clause(not_i, not_i), var_a, var_b);
    }
}

void tester::add_shuffled_clause(unique_ptr<clause_propagator> clause,
                                 int var_a, int var_b, int var_c)
{
    int vars[] = {var_a, var_b, var_c};

    static vector<unsigned> orders;
    orders.clear();
    for (unsigned i=0 ; i < sizeof(vars)/sizeof(vars[0]) ; ++i)
        if (vars[i] >= 0)
            orders.push_back(i);

    randomizer rand;
    random_shuffle(orders.begin(), orders.end(), rand);

    static clause_propagator::literals_type original_literals;
    original_literals= clause->literals();

    for (unsigned i=0 ; i != orders.size() ; ++i)
        clause->update_literals()[i]= original_literals[orders[i]];
    
    unique_ptr<constraint_propagator> constraint(clause.release());
    constraint_connectivity_info & connectivity=
        m_solver.add_constraint(move(constraint))
        .reserve_arguments(orders.size() );
    for (unsigned i=0 ; i != orders.size() ; ++i)
        connectivity << vars[orders[i]];
}

void tester::add_maybe_split_clause(unique_ptr<clause_propagator> clause,
                                    int var_a, int var_b)
{
    if (gen(2) != 0) { 
        add_shuffled_clause(move(clause), var_a, var_b);
    } else {
        unsigned extra_var= gen(N-2);
        if ( var_a >= 0 && extra_var == static_cast<unsigned>(var_a) )
            extra_var= N-2;
        if ( var_b >= 0 && extra_var == static_cast<unsigned>(var_b) )
            extra_var= N-1;
        int split_value= gen(N);
        unique_ptr<clause_propagator> clause_with(clause.release());
        clause_with->update_literals().push_back(iset(split_value));
        
        unique_ptr<clause_propagator> clause_without(clause_with->clone());
        iset & complement_set = clause_without->update_literals().back();
        complement_set.assign(iset::value_range_type(0, N-1));
        complement_set.erase(split_value);
        
        add_shuffled_clause(move(clause_with), var_a, var_b, extra_var);
        add_shuffled_clause(move(clause_without), var_a, var_b, extra_var);
    }
}

void tester::add_minus_not(unsigned var_a, unsigned var_b, int forbidden_minus)
{
    // Add clauses that implement:
    //   value[var_a] - value[var_b] != forbiden_minus  (infinite precision).
    //   value[var_a] != forbiden_minus + value[var_b]  (without negatives)

    const iset complete_set(make_pair(0, N-1));
    
    for (int avoid_b=0 ; avoid_b < (int)N ; ++avoid_b) {
        int avoid_a = forbidden_minus + avoid_b;
        if ( avoid_a < 0 || avoid_a >= (int)N )
            continue;
        
        iset valid_a= complete_set;
        valid_a.erase(avoid_a);
        
        iset valid_b= complete_set;
        valid_b.erase(avoid_b);
        
        
        add_maybe_split_clause(make_clause(valid_a, valid_b), var_a, var_b);
    }
    
}

tester::tester(solver::reporting_levels level)
{
    m_solver.report_level(level);
    for (unsigned i=0 ; i < N ; ++i)
        m_solver.add_var(iset(make_pair(0, N-1)));

    for (unsigned i=0 ; i < N ; ++i)
        for (unsigned j=i+1 ; j < N ; ++j) {
            add_diff(i,j);
            add_minus_not(i,j, i-j);
            add_minus_not(i,j, j-i);
        }
}

void tester::print_solution() const
{
    for (unsigned i=0 ; i < N  ; ++i) {
        unsigned queen_pos= m_solver.var_domain(i).max();
        for (unsigned column=0 ; column != N ; ++column)
            if (queen_pos == column)
                cout << 'X';
            else
                cout << ':';
        cout << '\n';
    }
}

void tester::validate_solution() const
{
    for (unsigned i=0 ; i < N  ; ++i) {
        unsigned first_queen= m_solver.var_domain(i).max();
        for (unsigned j=i+1 ; j < N  ; ++j) {
            unsigned second_queen= m_solver.var_domain(j).max();
            T_ASSERT(first_queen != second_queen);
            // First diagonal
            T_ASSERT(first_queen+i != second_queen+j);
            
            // Second diagonal
            T_ASSERT(first_queen+j != second_queen+i);
        }
    }
}


void tester::run()
{
    iset symmetry_break= m_solver.var_domain(0);
    unsigned mid= (symmetry_break.min()+symmetry_break.max())/2;
    if (gen(2) == 0)
        symmetry_break.intersect(make_pair(symmetry_break.min(), mid));
    else
        symmetry_break.intersect(make_pair(mid, symmetry_break.max()));
    m_solver.set_var_domain(0, move(symmetry_break));

    solver::result_type result= m_solver.solve(&cout);
    HCSP_ASSERT(result != solver::UNSAT);

    if ( m_solver.report_level() > solver::NONE) {
        cout << result << endl;
        if (result == solver::SAT) 
            print_solution();
    }
    HCSP_ASSERT(result == solver::SAT);
    if (result == solver::SAT)
        validate_solution();
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    solver::reporting_levels report_level= solver::NONE;
    //report_level= solver::SEARCH;
    //report_level= solver::PROPAGATION_MODS;
    //report_level= solver::PROPAGATED_CONSTRAINTS;

    cout << "Seed is " << set_seed(argc, argv) << endl;
    unsigned total_backtracks=0;
    for (unsigned i=0 ; i < 500 ; ++i) {
        if (i != 0) {
            unsigned seed= gen();
            srand(seed);
            if (report_level > solver::NONE)
                cout << "Seed is " << seed << endl;
        }
        tester tester(report_level);
        for (unsigned j=0 ; j != 10  ; ++j) {
            if (j != 0)
                tester.reset();
            tester.run();
            total_backtracks+= tester.backtracks_done();
        }
    }
    
    cout << total_backtracks << " total backtracks\n";

    END_TEST_MAIN;
}

