// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <iostream>
#include <sstream>
#include <vector>
#include <algorithm>
#include <iset.hpp>
#include <clause_propagator.hpp>
#include <solver.hpp>
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;

class tester {
  public:
    tester(solver::reporting_levels level, unsigned size);
    void run();
    unsigned backtracks_done() const { return m_solver.backtracks_done(); }

  private:
    void print_solution() const;
    void validate_solution() const;
    unsigned var_index(unsigned row, unsigned column) const
    { return row * m_size + column; }

    //! Constrain a diagonal or column to have at most one queen.
    //! \param diag the direction which should be constrained:
    //!        -# +1 positive diagonal starting with row=0 and column
    //!        -# -1 negative diagonal starting with row=0 and column
    //!        -# 0  one column
    void at_most_one_queen_per_diag(unsigned column, unsigned row, int diag);
    void at_least_one_queen_per_column(unsigned column);
    void at_least_one_queen_per_row(unsigned row);
    solver m_solver;
    const unsigned m_size;
};

void tester::at_most_one_queen_per_diag(unsigned column,unsigned row, int diag)
{
    int x0= column;
    for (unsigned y0=row ; y0 != m_size ; ++y0, x0 += diag) {
        if ( x0 <0 || x0 >= static_cast<int>(m_size) )
            continue;
        int x1= x0 + diag;
        for (unsigned y1=y0+1 ; y1 != m_size ; ++y1, x1 += diag) {
            if ( x1 <0 || x1 >= static_cast<int>(m_size) )
                continue;
            clause_propagator * p_clause= new clause_propagator;
            clause_propagator::literals_type & literals =
                p_clause->update_literals();
            unique_ptr<constraint_propagator> constraint(p_clause);
                    
            constraint_connectivity_info & connectivity=
                m_solver.add_constraint(move(constraint));
                    
            literals.resize(2, iset(0));
            connectivity.reserve_arguments(2);
            connectivity << var_index(y0, x0);
            connectivity << var_index(y1, x1);
        }
    }
}

void tester::at_least_one_queen_per_column(unsigned column)
{
    clause_propagator *p_clause= new clause_propagator;
    clause_propagator::literals_type & literals= p_clause->update_literals();
    unique_ptr<constraint_propagator> constraint(p_clause);
    
    constraint_connectivity_info & connectivity=
        m_solver.add_constraint(move(constraint));
    
    literals.reserve(m_size);
    connectivity.reserve_arguments(m_size);
    for (unsigned i=0 ; i != m_size ; ++i) {
        literals.push_back(iset(1));
        connectivity << var_index(i, column);
    }
}

void tester::at_least_one_queen_per_row(unsigned row)
{
    clause_propagator *p_clause= new clause_propagator;
    clause_propagator::literals_type & literals= p_clause->update_literals();
    unique_ptr<constraint_propagator> constraint(p_clause);
    
    constraint_connectivity_info & connectivity=
        m_solver.add_constraint(move(constraint));
    
    literals.reserve(m_size);
    connectivity.reserve_arguments(m_size);
    for (unsigned i=0 ; i != m_size ; ++i) {
        literals.push_back(iset(1));
        connectivity << var_index(row, i);
    }
}

tester::tester(solver::reporting_levels level, unsigned size) : m_size(size)
{
    const iset boolean(make_pair(0, 1));
    m_solver.report_level(level);
    for (unsigned row=0 ; row != m_size ; ++row)
        for (unsigned column=0 ; column != m_size ; ++column)
            m_solver.add_var(boolean);

    for (unsigned row=0 ; row != m_size ; ++row)
        at_least_one_queen_per_row(row);
    
    for (unsigned column=0 ; column != m_size ; ++column) {
        at_least_one_queen_per_column(column);
        
        // Note that for diag_direction != 0, this is incomplete. Almost half
        // of the diagonals don't touch row=0. The incomplete diagonals are
        // completed later
        for (int diag_direction= -1 ; diag_direction <= 1 ; ++diag_direction)
            at_most_one_queen_per_diag(column, 0, diag_direction);
    }
    // Add the diagonals that don't touch row=0 
    for (unsigned row=1 ; row != m_size ; ++row) {
        at_most_one_queen_per_diag(0,   row, +1);
        at_most_one_queen_per_diag(m_size-1, row, -1);
    }
}

void tester::print_solution() const
{
    for (unsigned row=0 ; row != m_size ; ++row) {
        for (unsigned column=0 ; column != m_size ; ++column) {
            bool occupied= m_solver.var_domain(var_index(row,column)).max();
            if (occupied)
                cout << 'X';
            else
                cout << ':';
        }
        cout << '\n';
    }
}

void tester::validate_solution() const
{
    for (unsigned row=0 ; row != m_size ; ++row) {
        unsigned on_row=0;
        for (unsigned column=0 ; column != m_size ; ++column)
            on_row += m_solver.var_domain(var_index(row,column)).max();
        T_ASSERT(on_row == 1);
    }
    for (unsigned column=0 ; column != m_size ; ++column) {
        unsigned on_column=0;
        for (unsigned row=0 ; row != m_size ; ++row)
            on_column += m_solver.var_domain(var_index(row,column)).max();
        T_ASSERT(on_column == 1);
    }
        
    // Make sure that at most one queen is present at any diagonal.
    for ( int x0= 1-static_cast<int>(m_size); 
          x0 < static_cast<int>(m_size);
          ++x0) {
        const unsigned y0= 0;
        unsigned on_diag=0;
        
        // positive diagonal
        int x1= x0;
        unsigned y1= y0;
        for ( ; y1 != m_size ; ++y1, ++x1) {
            if (x1 >= 0 && x1 < static_cast<int>(m_size))
                on_diag += m_solver.var_domain(var_index(y1,x1)).max();
        }
        
        // negative diagonal
        x1= x0;
        y1= y0;
        for ( ; y1 != m_size ; ++y1, --x1) {
            if (x1 >= 0 && x1 < static_cast<int>(m_size))
                on_diag += m_solver.var_domain(var_index(y1,x1)).max();
        }
    }
}


void tester::run()
{
    solver::result_type result= m_solver.solve(&cout);
    HCSP_ASSERT(result != solver::UNSAT);

    if ( m_solver.report_level() > solver::NONE) {
        cout << result << endl;
        if (result == solver::SAT)
            print_solution();
    }
    HCSP_ASSERT(result == solver::SAT);
    if (result == solver::SAT)
        validate_solution();
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    solver::reporting_levels report_level= solver::NONE;
    //report_level= solver::SEARCH;
    //report_level= solver::PROPAGATION_MODS;
    //report_level= solver::PROPAGATED_CONSTRAINTS;

    cout << "Seed is " << set_seed(argc, argv) << endl;
    unsigned total_backtracks=0;
    for (unsigned i=0 ; i < 3000 ; ++i) {
        if (i != 0) {
            unsigned seed= gen();
            srand(seed);
            if (report_level > solver::NONE)
                cout << "Seed is " << seed << endl;
        }
        tester tester(report_level, 8);
        tester.run();

        total_backtracks+= tester.backtracks_done();
    }
    cout << total_backtracks << " total backtracks\n";
    
    END_TEST_MAIN;
}

