// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <unistd.h>
#include <iostream>
#include <sstream>
#include <vector>
#include <algorithm>
#include <iset.hpp>
#include <clause_propagator.hpp>
#include <neq_propagators.hpp>
#include <solver.hpp>
#include "test_utilities.hpp"
#include "explanation_tester.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;

class neq_propagator_with_watchers : public neq_propagator
{
  private:
    enum {NUM_OF_WATCHES= 2};
  public:
    bool init( constraint_runtime_info & info )
    {
        HCSP_ASSERT(info.arity() == 2);
        for (unsigned i=0 ; i != 2 ; ++i) {
            m_watch[i].m_param= i;
            m_watch[i].m_watch_handle= 
                info.allocate_watch(i, FINALIZATION_WATCH);
        }
        return CONSISTENT;
    }
    result_type propagate( constraint_runtime_info & info)
    {
        for (unsigned i=0 ; i != 2 ; ++i) {
            info.deallocate_watch(i,FINALIZATION_WATCH, 
                                  m_watch[i].m_watch_handle);
            m_watch[i].m_watch_handle=
                info.allocate_watch(i, FINALIZATION_WATCH);
        }
        return neq_propagator::propagate(info);
    }

    virtual unsigned count_watches() const { return 2; }
    virtual unsigned watched_param_index(unsigned watch_index) const
    { return watch_index; }

  private:
    watch_info_type m_watch[NUM_OF_WATCHES];
};

class tester {
  public:
    static const unsigned N = 32;
    tester(solver::reporting_levels level, unsigned opportunistic_bj_threshold,
           unsigned max_conflict_clause);
    void run();
    unsigned backtracks_done() const
    { return m_solver.backtracks_done(); }

  private:
    void print_solution() const;
    void validate_solution() const;

    void add_neq(unsigned var_a, unsigned var_b);
    void add_minus_not(unsigned var_a, unsigned var_b, int forbidden_minus);
    solver m_solver;
};

 
void tester::add_neq(unsigned var_a, unsigned var_b)
{
    unique_ptr<constraint_propagator> constraint;
    if (gen(10) == 0)
        constraint.reset(new neq_propagator_with_watchers);
    else if (gen(4) == 0)
        constraint.reset(new explanation_tester(new neq_propagator));
    else
        constraint.reset(new neq_propagator);
    m_solver.add_constraint(move(constraint)).reserve_arguments(2)
        << var_a << var_b;
}

void tester::add_minus_not(unsigned var_a, unsigned var_b, int forbidden_minus)
{
    unique_ptr<constraint_propagator>
        constraint(new difference_neq_constant_propagator(forbidden_minus));
    if (gen(4)==0)
        constraint.reset(new explanation_tester(constraint.release()));
    m_solver.add_constraint(move(constraint)).reserve_arguments(2)
        << var_a << var_b;
}

tester::tester(solver::reporting_levels level,
               unsigned opportunistic_threshold,
               unsigned max_conflict_clause)
{
    m_solver.report_level(level);
    if ( opportunistic_threshold > 0 )
        m_solver.opportunistic_back_jumping_threshold(opportunistic_threshold);
    if ( max_conflict_clause+1 != 0)
        m_solver.max_conflict_clause(max_conflict_clause);
    
    m_solver.backtrack_limit(200000);
    for (unsigned i=0 ; i < N ; ++i)
        m_solver.add_var(iset(make_pair(0, N-1)));

    for (unsigned i=0 ; i < N ; ++i)
        for (unsigned j=i+1 ; j < N ; ++j) {
            add_neq(i,j);
            add_minus_not(i,j, i-j);
            add_minus_not(i,j, j-i);
        }
}

void tester::print_solution() const
{
    for (unsigned i=0 ; i < N  ; ++i) {
        unsigned queen_pos= m_solver.var_domain(i).max();
        for (unsigned column=0 ; column != N ; ++column)
            if (queen_pos == column)
                cout << 'X';
            else
                cout << ':';
        cout << '\n';
    }
}

void tester::validate_solution() const
{
    for (unsigned i=0 ; i < N  ; ++i) {
        unsigned first_queen= m_solver.var_domain(i).max();
        for (unsigned j=i+1 ; j < N  ; ++j) {
            unsigned second_queen= m_solver.var_domain(j).max();
            T_ASSERT(first_queen != second_queen);
            // First diagonal
            T_ASSERT(first_queen+i != second_queen+j);
            
            // Second diagonal
            T_ASSERT(first_queen+j != second_queen+i);
        }
    }
}


void tester::run()
{
    iset domain(make_pair(0, N-1));
    m_solver.reset();
    for (unsigned i=0 ; i!=N ; ++i)
        T_ASSERT(m_solver.var_domain(i)== domain);
    
    solver::result_type result= m_solver.solve(&cout);
    HCSP_ASSERT(result != solver::UNSAT);

    if ( m_solver.report_level() > solver::NONE) {
        cout << m_solver.backtracks_done() << " backtracks done. ";

        cout << result << endl;
        if ( result == solver::SAT )
            print_solution();
    }
    HCSP_ASSERT(result == solver::SAT);
    if (result == solver::SAT)
        validate_solution();
}

static void usage(const char *app)
{
    //phs:l:o:c:    
    cout << "Usage: " << app
         << " [-s seed] [-l report_level]\n"
         <<" [-o opportunistic_threshold]\n"
         << "       [-c max_conflict_clause] [-h]\n"
         << "  -h  Gives this help message\n"
         << "  -s N  Set the random seed to M\n"
         << "  -c N  Will set the maximum size of conflict clause which is "
         << "still guaranteed\n"
         << "        to be inserted\n"
         << "  -l N  The report level to use\n"
         << "  -o opportunistic_threshold\n"
         << "     The minimum number of decisions to be undone below which "
         << "opportunistic\n"
         << "     back-jumping is not performed.\n";
}

static solver::reporting_levels parse_report_level(const char* text)
{
    return static_cast<solver::reporting_levels>(atoi(text));
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    solver::reporting_levels report_level= solver::NONE;
    unsigned seed=0, opportunistic_bj=0, max_conflict_clause=-1U;
    while (true) {
        int option= getopt(argc, argv, "hs:l:o:c:");
        if (option < 0)
            break;
        switch (option) {
          case '?': usage(argv[0]); return 1;
          case 'h': usage(argv[0]); return 0;
          case 'l': report_level= parse_report_level(optarg); break;
          case 's': seed= strtoul(optarg, NULL, 0);           break;
          case 'o': opportunistic_bj=strtoul(optarg, NULL, 0);break;
          case 'c': max_conflict_clause=strtoul(optarg, NULL, 0);break;
          default:
            cerr << "Unhandled option -" << static_cast<char>(option) << "\n";
            return 1;
        }
    }
    if ( argv[optind] ) {
        cerr << "Redundant argument " << argv[optind] << "\n";
        return 1;
    }
    cout << "Seed is " << hcsp::random::set_seed(seed) << endl;
    unsigned backtracks=0;
    for (unsigned i=0 ; i < 30 ; ++i) {
        if (i != 0) {
            seed= gen();
            srand(seed);
            if (report_level > solver::NONE)
                cout << "Seed is " << seed << endl;
        }
        tester tester(report_level, opportunistic_bj, max_conflict_clause);
        for (unsigned j=0 ; j < 9 ; ++j) {
            tester.run();
            backtracks += tester.backtracks_done();
        }
    }
    cout << "total " << backtracks << " backtracks done\n";
    END_TEST_MAIN;
}

