// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef RUN_HCSP_PROFILER
#define RUN_HCSP_PROFILER
#endif
#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <vector>
#include <algorithm>
#include <iset.hpp>
#include <clause_propagator.hpp>
#include <neq_propagators.hpp>
#include <solver.hpp>
#include <flags.hpp>
#include <formula.hpp>
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;

// Find a quasigroup of the given properties and print its results.
// The description and many ideas are taken from the referenced paper [1].
//
// References:
// [1] Solving Open Quasigroup Problems by Propositional Reasoning.

class tester {
  public:
    class flags;
    tester(const flags & flags, unsigned type, unsigned order);
    void run();
    
  private:
    enum quasigroup_existence_type { YES, NO, MAYBE };
    void apply(const formula & form)
    { form.constrain_solver(m_solver); }
    void print_element(unsigned element) const;
    void print_solution() const;
    void validate_QG1() const;
    void validate_QG2() const;
    void validate_QG3() const;
    void validate_QG4() const;
    void validate_QG5() const;
    void validate_QG6() const;
    void validate_QG7() const;
    void validate_QG(unsigned type) const;
    void validate_solution() const;
    unsigned var_index(unsigned row, unsigned column) const
    { return row * m_size + column; }
    unsigned relation_var_index(unsigned xy, unsigned zw) const
    {
        if (xy >= m_first_inverse_var) {
            HCSP_ASSERT(zw >= m_first_inverse_var);
            xy -= m_first_inverse_var;
            zw -= m_first_inverse_var;
        }
        return xy*m_size*m_size + zw;
    }
    unsigned var_value(unsigned row, unsigned column) const
    {
        const iset & domain= m_solver.var_domain(var_index(row,column));
        T_ASSERT(domain.singleton());
        return domain.min();
    }
    void add_inverse_vars(bool is_right_inverse);
    void add_inverting_constraint(bool is_right_inverse,
                                  unsigned row, unsigned column);
    void add_is_equal_allowed(unsigned res_var, unsigned var1, unsigned var2);
    void add_is_equal_mandatory(unsigned res_var,unsigned var1, unsigned var2);
    void add_equality_vars(unsigned & equality_offset,
                           unsigned compared_var_offset=0);
    void add_diff(unsigned var_a, unsigned var_b);
    
    void all_different_column_elements(unsigned column, unsigned offset=0);
    void all_different_row_elements(unsigned row, unsigned offset=0);

    void column_elements_are_permutation(unsigned column, unsigned offset=0);
    void row_elements_are_permutation(unsigned row, unsigned offset=0);

    // Each of the following constraints will generate a different class
    // of interesting quasigroups as described in the referenced paper [1].
    void constrain_QG1(unsigned x, unsigned y, unsigned z, unsigned w);
    void constrain_QG1();

    void constrain_QG2(unsigned x, unsigned y, unsigned z, unsigned w);
    void constrain_QG2();
    void constrain_QG3(unsigned x, unsigned y, unsigned xy, unsigned yx,
                       unsigned option);
    void constrain_QG3();
    void constrain_QG4(unsigned x, unsigned y, unsigned yx, unsigned x_yx,
                       unsigned option);
    void constrain_QG4();
    void constrain_QG5(unsigned x, unsigned y ,unsigned xy, unsigned xy_x,
                       unsigned option);
    void constrain_QG5();
    void constrain_QG6(unsigned x, unsigned y, unsigned xy, unsigned xy_y,
                       unsigned option);
    void constrain_QG6();
    void constrain_QG7(unsigned x, unsigned y, unsigned xy, unsigned xy_x,
                       unsigned option);
    void constrain_QG7();
    // Add the constraints of one of QG1...QG7 according to parameter i.
    // If i==0, then don't add any constraint.
    // precondition: 0 <= i <= 7
    void constrain_QG(unsigned i);
    

    static quasigroup_existence_type m_expected_results[21][8];
    solver m_solver;
    const iset m_domain;
    const unsigned m_size;
    const bool m_is_idempotent;
    const unsigned m_constraint_type;
    // If the inverse of the quasigroup is calculated, then this is the
    // first variable.
    unsigned m_first_inverse_var;
    // The index of the first indicator variable, if indicator variables are
    // calculated. If an indicator variable for var1,var2 is false then
    // it is forbidden to have value[var1]==value[var2]
    unsigned m_first_element_equality_allowed;
    unsigned m_first_element_inverse_equality_allowed;

    // Should we use global all_different_propagator instead of many
    // neq_propagator objects.
    bool     m_use_all_different_propagator;
};

enum difficulty_types {
    FAST_RUNS=0,
    SECONDS_RUNS,
    DOZENS_SECOND_RUNS,
    MINUTES_RUNS,
    NUM_OF_SPEEDS
};

class tester::flags : public hcsp_flags
{
  public:
    flags() : idempotent(true), difficulty(FAST_RUNS) {}
  protected:
    virtual bool parse_option(char option);

    virtual void usage_short(std::ostream & out);
    virtual void usage_long(std::ostream & out);
    
    //! Make sure that we did not set unsupported flags
    void validate_consistent() const;
    virtual std::string legal_args() const
    { return hcsp_flags::legal_args() + "Iq:d:";}
    
  public:
    bool idempotent;
    difficulty_types difficulty;
    vector<string> problems_requested;    
}; // class tester::flags

// NOTE regarding QG3.9
// According to CSPLib (2008/03/08) this instance is satisfiable, but my runs
// and those of:
//   http://www.cs.ubc.ca/~hoos/SATLIB/Benchmarks/SAT/QG/qg.descr.html
// show otherwise.
tester::quasigroup_existence_type tester::m_expected_results[21][8] = {
    // Undocumented in csplib, self found
    //  0  ,  QG1 ,  QG2 ,  QG3 ,  QG4 ,  QG5 ,  QG6 ,  QG7
    { YES  , YES  , YES  , YES  , YES  , YES  , YES  , YES   }, // order 0
    { YES  , YES  , YES  , YES  , YES  , YES  , YES  , YES   }, // order 1
    { NO   , NO   , NO   , NO   , NO   , NO   , NO   , NO    }, // order 2
    { YES  , NO   , NO   , NO   , NO   , NO   , NO   , NO    }, // order 3
    { YES  , YES  , NO   , YES  , NO   , NO   , YES  , NO    }, // order 4

    // From the documentation in csplib (only 0 is not documented)
    //  0  ,  QG1 ,  QG2 ,  QG3 ,  QG4 ,  QG5 ,  QG6 ,  QG7
    { YES  , YES  , YES  , NO   , YES  , YES  , NO   , YES   }, // order 5  
    { YES  , NO   , NO   , NO   , NO   , NO   , NO   , NO    }, // order 6  
    { YES  , YES  , YES  , NO   , NO   , YES  , NO   , NO    }, // order 7  
    { YES  , YES  , YES  , YES  , NO   , YES  , YES  , NO    }, // order 8  
    { YES  , YES  , YES  , NO   , YES  , NO   , YES  , YES   }, // order 9  
    { YES  , YES  , NO   , NO   , NO   , NO   , NO   , NO    }, // order 10 
    { YES  , YES  , YES  , NO   , NO   , YES  , NO   , NO    }, // order 11 
    { YES  , YES  , YES  , YES  , YES  , NO   , NO   , NO    }, // order 12 
    { YES  , YES  , YES  , YES  , YES  , NO   , YES  , YES   }, // order 13 
    { YES  , YES  , YES  , NO   , NO   , NO   , NO   , NO    }, // order 14 
    { YES  , YES  , YES  , NO   , NO   , NO   , NO   , NO    }, // order 15 
    { YES  , YES  , YES  , YES  , YES  , NO   , YES  , MAYBE }, // order 16 
    { YES  , YES  , YES  , YES  , YES  , YES  , YES  , YES   }, // order 17 
    { YES  , YES  , YES  , NO   , NO   , MAYBE, NO   , MAYBE }, // order 18 
    { YES  , YES  , YES  , NO   , NO   , YES  , NO   , MAYBE }, // order 19 
    { YES  , YES  , YES  , YES  , YES  , MAYBE, MAYBE, MAYBE }  // order 20
};

void tester::add_inverse_vars(bool is_right_inverse)
{
    m_first_inverse_var= m_solver.count_variables();
    for (unsigned x=0 ; x != m_size ; ++x)
        for (unsigned y=0 ; y != m_size ; ++y)
            for (unsigned z=0 ; z != m_size ; ++z)
                m_solver.add_var(m_domain);
    
    for (unsigned row=0 ; row != m_size ; ++row)
        for (unsigned column=0 ; column != m_size ; ++column)
            add_inverting_constraint(is_right_inverse, row, column);

    for (unsigned row=0 ; row != m_size ; ++row)
        row_elements_are_permutation(row, m_first_inverse_var);

    for (unsigned column=0 ; column != m_size ; ++column)
        column_elements_are_permutation(column, m_first_inverse_var);
}

void tester::add_inverting_constraint(bool is_right_inverse,
                                      unsigned row, unsigned column)
{
    // is_right_inverse== true:              is_right_inverse== false:
    //  V*column= row iff V= row/R column    column*V= row iff V= row/L column
    // ==>                                   ==>                
    //  1:V*column!=row || V= row/R column   1:column*V!=row || V= row/L column
    //  2:V*column=row  || V!=row/R column   2:column*V=row  || V!=row/L column
    for (unsigned v=0 ; v != m_size ; ++v) {
        unsigned inverse_var= m_first_inverse_var+ var_index(row,column);
        unsigned var= (is_right_inverse ?
                       var_index(v,column) : var_index(column,v));

        // V*column!=row || V= row/R column  ; column*V!=row || V= row/L column
        apply(formula(var, iset(row).complement()) ||
              formula(inverse_var, iset(v)) );

        // V*column=row  || V!=row/R column  ; column*V=row  || V!=row/L column
        apply(formula(var, iset(row)) ||
              formula(inverse_var, iset(v).complement()) );
    }
}

// Add the constraint:
//   value[res_var] --> value[var1]= value[var2]
void tester::add_is_equal_mandatory(unsigned res_var,
                                    unsigned var1, unsigned var2)
{
    if (var1 == var2) {
        m_solver.set_var_domain(res_var, iset(1));
        return;
    }
    //    ==> for all i:
    //     1.1. (value[res_var] && value[var1]=i) -> value[var2]=i
    //     1.2. (value[res_var] && value[var2]=i) -> value[var1]=i
    //    ==> for all i:
    //     1.1. !value[res_var] || value[var1]!=i || value[var2]=i
    //     1.2. !value[res_var] || value[var1]=i  || value[var2]!=i
    for (unsigned i=0 ; i != m_size ; ++i) {
        for (unsigned positive_param=1 ; positive_param!=3 ; ++positive_param){
            iset without_i= m_domain;
            without_i.erase(i);
            iset with_i= iset(i);

            apply(formula(res_var, iset(0)) ||
                  formula(var1, positive_param == 1 ? with_i : without_i) ||
                  formula(var2, positive_param == 2 ? with_i : without_i));
        }
    }
}

// Add the constraint:
//        value[res_var] <-- value[var1]= value[var2]
// ==>    value[res_var] || value[var1]!=value[var2]
void tester::add_is_equal_allowed(unsigned res_var,
                                  unsigned var1, unsigned var2)
{
    if (var1 == var2) {
        m_solver.set_var_domain(res_var, iset(1));
        return;
    }
    unique_ptr<constraint_propagator> constraint;
    constraint.reset(new literal_or_neq_propagator(iset(1)));
    
    m_solver.add_constraint(move(constraint)).reserve_arguments(3)
        << res_var << var1 << var2;
}

void tester::add_equality_vars(unsigned & equality_begin,
                               unsigned compared_var_offset)
{
    equality_begin= m_solver.count_variables();
    const iset bool_dom(make_pair(0,1));
    for (unsigned x=0 ; x!=m_size ; ++x)
        for (unsigned y=0 ; y!=m_size ; ++y)
            for (unsigned z=0 ; z!=m_size ; ++z)
                for (unsigned w=0 ; w!=m_size ; ++w)
                    m_solver.add_var(bool_dom);

    for (unsigned x=0 ; x!=m_size ; ++x)
        for (unsigned y=0 ; y!=m_size ; ++y)
            for (unsigned z=0 ; z!=m_size ; ++z)
                for (unsigned w=0 ; w!=m_size ; ++w) {
                    unsigned xy= compared_var_offset + var_index(x,y);
                    unsigned zw= compared_var_offset + var_index(z,w);
                    if (xy > zw)
                        continue; // avoid symmetric constraints
                    unsigned equals=equality_begin+relation_var_index(xy,zw);
                    if (xy == zw)
                        m_solver.set_var_domain(equals,iset(1));
                    else if (x==z || y==w)
                        // If any coordinate is the same (x,y) vs. (z,w),
                        // while the other is different (since xy != zw) then
                        // NOT equals (a property of quasigroups)
                        m_solver.set_var_domain(equals, iset(0));
                    else
                        add_is_equal_allowed(equals, xy, zw);
                }
}

void tester::add_diff(unsigned var_a, unsigned var_b)
{
    unique_ptr<constraint_propagator> constraint(new neq_propagator);
    
    m_solver.add_constraint(move(constraint)).reserve_arguments(2)
        << var_a << var_b;
}

void tester::all_different_column_elements(unsigned column, unsigned offset)
{
    // For the case of m_use_all_different_propagator, the following
    // constraints are redundant. It seems (but not certain) that
    // they speed things up as they reduce the domains making
    // the all_different_propagator run much less. (Need to verify this)
    for (unsigned i=0 ; i < m_size ; ++i)
        for (unsigned j=i+1 ; j < m_size ; ++j)
            add_diff(offset+var_index(i, column),
                     offset+var_index(j, column));
    if (m_use_all_different_propagator) {
        unique_ptr<constraint_propagator> all_diff;
        all_diff.reset(new all_different_propagator);
        constraint_connectivity_info &connectivity =
            m_solver.add_constraint(move(all_diff)).reserve_arguments(m_size);
        for (unsigned i=0 ; i < m_size ; ++i)
            connectivity << offset+var_index(i, column);
    }
}

void tester::all_different_row_elements(unsigned row, unsigned offset)
{
    // For the case of m_use_all_different_propagator, the following
    // constraints are redundant. It seems (but not certain) that
    // they speed things up as they reduce the domains making
    // the all_different_propagator run much less. (Need to verify this)
    for (unsigned i=0 ; i < m_size ; ++i)
        for (unsigned j=i+1 ; j < m_size ; ++j)
            add_diff(offset+var_index(row, i), offset+var_index(row, j));
    if (m_use_all_different_propagator) {
        unique_ptr<constraint_propagator> all_diff;
        all_diff.reset(new all_different_propagator);
        constraint_connectivity_info &connectivity =
            m_solver.add_constraint(move(all_diff)).reserve_arguments(m_size);
        for (unsigned i=0 ; i < m_size ; ++i)
            connectivity << offset+var_index(row, i);
    }
}

void tester::column_elements_are_permutation(unsigned column, unsigned offset)
{
    all_different_column_elements(column, offset);
    for (unsigned element=0 ; element != m_size ; ++element) {
        const iset with_element(element);
        formula form(false);
        for (unsigned row=0 ; row != m_size ; ++row) {
            form= form || formula(offset+var_index(row, column), with_element);
        }
        apply(form);
    }
}

void tester::row_elements_are_permutation(unsigned row, unsigned offset)
{
    all_different_row_elements(row, offset);
    for (unsigned element=0 ; element != m_size ; ++element) {
        const iset with_element(element);
        formula form(false);
        for (unsigned column=0 ; column != m_size ; ++column) {
            form= form || formula(offset+var_index(row, column), with_element);
        }
        apply(form);
    }
}

void tester::constrain_QG1(unsigned x, unsigned y, unsigned z, unsigned w)
{
    // if (x*y==u && z*w==u && v*y==x && v*w==z) then (x==z && y==w)
    // This is the same as:
    // if (x!=z || y!=w) then (x*y!=u || z*w!=u || v*y!=x || v*w!=z)
    //
    // This means that we can simply skip the cases that x==z && y==w,
    // and otherwise add the clause:
    //   (x*y!=u || z*w!=u || v*y!=x || v*w!=z)
    // ==>  (x*y != z*w || v*y!=x || v*w!=z)
    // ==>  (x*y != z*w || x/R y !=z/R w) [where a/R b =c iff a=b*c]
    if (x==z && y==w)
        return;

    unsigned equality_allowed= m_first_element_equality_allowed +
        relation_var_index(var_index(x,y), var_index(z,w));

    unsigned inverse_equality_allowed=m_first_element_inverse_equality_allowed+
        relation_var_index(var_index(x,y), var_index(z,w));
    apply(formula(equality_allowed, iset(0)) ||
          formula(inverse_equality_allowed, iset(0)));
}


void tester::constrain_QG1()
{
    add_equality_vars(m_first_element_equality_allowed);
    add_inverse_vars(true);
    add_equality_vars(m_first_element_inverse_equality_allowed,
                      m_first_inverse_var);
    for (unsigned x=0 ; x!=m_size ; ++x)
        for (unsigned y=0 ; y!=m_size ; ++y)
            for (unsigned z=0 ; z!=m_size ; ++z)
                for (unsigned w=0 ; w!=m_size ; ++w)
                    constrain_QG1(x,y,z,w);
}

void tester::constrain_QG2(unsigned x, unsigned y, unsigned z, unsigned w)
{
    // if (x*y==u && z*w==u && y*v==x && w*v==z) then (x==z && y==w)
    // This is the same as:
    // if (x!=z || y!=w) then (x*y!=u || z*w!=u || y*v!=x || w*v!=z)
    //
    // This means that we can simply skip the cases that x==z && y==w,
    // and otherwise add the clause:
    //   (x*y!=u || z*w!=u || y*v!=x || w*v!=z)
    // ==> (x*y != z*w || x/L y != z/L w)  [where b/L a=c iff b=a*c]
    if (x==z && y==w)
        return;

    
    unsigned equality_allowed= m_first_element_equality_allowed +
        relation_var_index(var_index(x,y), var_index(z,w));
    unsigned inverse_equality_allowed=m_first_element_inverse_equality_allowed+
        relation_var_index(var_index(x,y), var_index(z,w));

    apply(formula(equality_allowed, iset(0)) ||
          formula(inverse_equality_allowed, iset(0)));
}



void tester::constrain_QG2()
{
    add_equality_vars(m_first_element_equality_allowed);
    add_inverse_vars(false);
    add_equality_vars(m_first_element_inverse_equality_allowed,
                      m_first_inverse_var);
    for (unsigned x=0 ; x!=m_size ; ++x)
        for (unsigned y=0 ; y!=m_size ; ++y)
            for (unsigned z=0 ; z!=m_size ; ++z)
                for (unsigned w=0 ; w!=m_size ; ++w)
                    constrain_QG2(x,y,z,w);
}

// As explained at constrain_QG3():
// (x*y)*(y*x) == x,
//
// Due to closure property of quasigroups this can be decomposed in
// any way out of the following three (if any of the two of the operators give
// an expected value, so must the third)
// [1] (x*y== xy && y*x == yx) => (xy*yx==x)
// [2] (x*y== xy && xy*yx==x)  => (y*x ==yx)
// [3] (xy*yx==x && y*x == yx) => (x*y ==xy)
// As clauses:
// [1] (x*y!= xy || y*x != yx || xy*yx==x)
// [2] (x*y!= xy || y*x == yx || xy*yx!=x)
// [3] (x*y== xy || y*x != yx || xy*yx!=x)
void tester::constrain_QG3(unsigned x, unsigned y, unsigned xy, unsigned yx,
                           unsigned option)
{
    formula form;
    if (option == 1)
        form= formula(var_index(xy,yx), iset(x));
    else
        form= formula(var_index(xy,yx), iset(x).complement());

    if (option == 2)
        form= form || formula(var_index(x,y), iset(xy));
    else
        form= form || formula(var_index(x,y), iset(xy).complement());

    if (option == 3)
        form= form || formula(var_index(y,x), iset(yx));
    else
        form= form || formula(var_index(y,x), iset(yx).complement());
    apply(form);
}

void tester::constrain_QG3()
{
    // (x*y)*(y*x) == x,
    //
    // After decomposition:
    // if (x*y== xy && y*x == yx) then (xy*yx==x)
    //
    // As a clause:
    //  (x*y != xy || y*x != yx || xy*yx==x)
    for (unsigned x=0 ; x!=m_size ; ++x)
        for (unsigned y=0 ; y!=m_size ; ++y)
            for (unsigned xy=0 ; xy!=m_size ; ++xy)
                for (unsigned yx=0 ; yx!=m_size ; ++yx) {
                    constrain_QG3(x, y, xy, yx, 1);
                    constrain_QG3(x, y, xy, yx, 2);
                    constrain_QG3(x, y, xy, yx, 3);
                }
}

// As explained at constrain_QG4():
// (x*(y*x))*y == x
//
// Due to closure property of quasigroups this can be decomposed in
// any way out of the following three (if any of the two of the operators give
// an expected value, so must the third)
// [1] (y*x==yx   && x*yx == x_yx) => (x_yx*y==x)
// [2] (x_yx*y==x && x*yx == x_yx) => (y*x==yx)
// [3] (y*x==yx   && x_yx*y==x)    => (x*yx == x_yx)
// As clauses:
// [1] (y*x!=yx || x*yx != x_yx || x_yx*y==x)
// [2] (y*x==yx || x*yx != x_yx || x_yx*y!=x)
// [3] (y*x!=yx || x*yx == x_yx || x_yx*y!=x)

void tester::constrain_QG4(unsigned x, unsigned y, unsigned yx, unsigned x_yx,
                           unsigned option)
{
    formula form;
    if (option == 1)
        form= formula(var_index(x_yx,y), iset(x));
    else
        form= formula(var_index(x_yx,y), iset(x).complement());
    if (option == 2)
        form= form || formula(var_index(y,x),iset(yx));
    else
        form= form || formula(var_index(y,x),iset(yx).complement());
    if (option == 3)
        form= form || formula(var_index(x,yx), iset(x_yx));
    else
        form= form || formula(var_index(x,yx), iset(x_yx).complement());
    apply(form);
}

void tester::constrain_QG4()
{
    // (x*(y*x))*y == x
    //
    // After decomposition:
    // if (y*x==yx && x*yx == x_yx) then (x_yx*y==x)
    //
    // As a clause:
    // (y*x!=yx || x*yx != x_yx || x_yx*y==x)
    for (unsigned x=0 ; x!=m_size ; ++x)
        for (unsigned y=0 ; y!=m_size ; ++y)
            for (unsigned yx=0 ; yx!=m_size ; ++yx)
                for (unsigned x_yx=0 ; x_yx!=m_size ; ++x_yx) {
                    constrain_QG4(x,y,yx,x_yx, 1);
                    // does not seem to help:
                    //constrain_QG4(x,y,yx,x_yx, 2);
                    //constrain_QG4(x,y,yx,x_yx, 3);
                }
}

// As explained at constrain_QG5():
// ((x*y)*x)*x == y
//
// Due to closure property of quasigroups this can be decomposed in
// any way out of the following three (if any of the two of the operators give
// an expected value, so must the third)
// [1] (x*y==xy      && xy*x == xy_x) ==> (xy_x*x==y)
// [2] (xy*x == xy_x && xy_x*x==y   ) ==> (x*y==xy)
// [3] (x*y==xy      && xy_x*x==y   ) ==> (xy*x == xy_x)
// As clauses:
// [1] (x*y!=xy      || xy*x != xy_x || xy_x*x==y)
// [2] (x*y==xy      || xy*x != xy_x || xy_x*x!=y)
// [3] (x*y!=xy      || xy*x == xy_x || xy_x*x!=y)
void tester::constrain_QG5(unsigned x, unsigned y ,unsigned xy, unsigned xy_x,
                           unsigned option)
{
    formula form;
    if (option == 1)
        form= formula(var_index(xy_x,x), iset(y));
    else
        form= formula(var_index(xy_x,x), iset(y).complement());

    if (option == 2)
        form= form || formula(var_index(x,y), iset(xy));
    else
        form= form || formula(var_index(x,y), iset(xy).complement());
    if (option == 3)
        form= form || formula(var_index(xy,x), iset(xy_x) );
    else
        form= form || formula(var_index(xy,x), iset(xy_x).complement() );
    apply(form);
}    

void tester::constrain_QG5()
{
    // ((x*y)*x)*x == y
    //
    // After decomposition:
    // if (x*y==xy && xy*x == xy_x) then (xy_x*x==y)
    //
    // As a clause:
    // (x*y!=xy || xy*x != xy_x || xy_x*x==y)
    for (unsigned x=0 ; x!=m_size ; ++x)
        for (unsigned y=0 ; y!=m_size ; ++y)
            for (unsigned xy=0 ; xy!=m_size ; ++xy)
                for (unsigned xy_x=0 ; xy_x!=m_size ; ++xy_x) {
                    constrain_QG5(x,y ,xy, xy_x, 1);
                    // These are redundant, they are used to limit search
                    // space
                    constrain_QG5(x,y ,xy, xy_x, 2);
                    constrain_QG5(x,y ,xy, xy_x, 3);
                }
}

// As explained at constrain_QG6():
// (x*y)*y == x*(x*y)
//
// After decomposition:
// Due to closure property of quasigroups this can be decomposed in
// any way out of the following three (if any of the two of the operators give
// an expected value, so must the third)
// [1] (x*y==xy    && xy*y==xy_y)   ==> (x*xy==xy_y)
// [2] (x*xy==xy_y && xy*y==xy_y)   ==> (x*y==xy)
// [3] (x*y==xy    && x*xy==xy_y)   ==> (xy*y==xy_y)
// As clauses:
// [1] (x*y!=xy    || xy*y!=xy_y || x*xy==xy_y)
// [2] (x*y==xy    || xy*y!=xy_y || x*xy!=xy_y)
// [3] (x*y!=xy    || xy*y==xy_y || x*xy!=xy_y)
void tester::constrain_QG6(unsigned x, unsigned y, unsigned xy, unsigned xy_y,
                           unsigned option)
{
    formula form;
    if (option == 1)
        form= formula(var_index(x,xy), iset(xy_y));
    else
        form= formula(var_index(x,xy), iset(xy_y).complement());

    if (option == 2)
        form= form || formula(var_index(x,y),iset(xy));
    else
        form= form || formula(var_index(x,y),iset(xy).complement());
    if (option == 3)
        form= form || formula(var_index(xy, y), iset(xy_y));
    else
        form= form || formula(var_index(xy, y), iset(xy_y).complement());
    apply(form);
}

void tester::constrain_QG6()
{
    // (x*y)*y == x*(x*y)
    //
    // After decomposition:
    // if (x*y==xy && xy*y==xy_y) then (x*xy==xy_y)
    //
    // As a clause:
    //  (x*y!=xy || xy*y!=xy_y || x*xy==xy_y)
    for (unsigned x=0 ; x!=m_size ; ++x)
        for (unsigned y=0 ; y!=m_size ; ++y)
            for (unsigned xy=0 ; xy!=m_size ; ++xy)
                for (unsigned xy_y=0 ; xy_y!=m_size ; ++xy_y) {
                    constrain_QG6(x,y,xy,xy_y, 1);
                    constrain_QG6(x,y,xy,xy_y, 2);
                    constrain_QG6(x,y,xy,xy_y, 3);
                }
}

// As explained at constrain_QG7():
// (((x*y)*x)*y) == x
//
// After decomposition:
// Due to closure property of quasigroups this can be decomposed in
// any way out of the following three (if any of the two of the operators give
// an expected value, so must the third)
// [1] (x*y==xy   && xy*x==xy_x) => (xy_x*y==x)
// [2] (xy_x*y==x && xy*x==xy_x) => (x*y==xy)
// [3] (x*y==xy   && xy_x*y==x)  => (xy*x==xy_x)
// As clauses:
// [1] (x*y!=xy || xy*x!=xy_x || xy_x*y==x)
// [2] (x*y==xy || xy*x!=xy_x || xy_x*y!=x)
// [3] (x*y!=xy || xy*x==xy_x || xy_x*y!=x)
// As a-count constraint
//     CountOnes(x*y==xy, xy*x==xy_x , xy_x*y==x) != 2
void tester::constrain_QG7(unsigned x, unsigned y, unsigned xy, unsigned xy_x,
                           unsigned option)
{
    formula form;
    if (option == 1)
        form= formula(var_index(xy_x,y), iset(x));
    else
        form= formula(var_index(xy_x,y), iset(x).complement());

    if (option == 2)
        form= form || formula(var_index(x,y),iset(xy));
    else
        form= form || formula(var_index(x,y),iset(xy).complement());
    if (option == 3)
        form= form || formula(var_index(xy, x), iset(xy_x));
    else
        form= form || formula(var_index(xy, x), iset(xy_x).complement());
    apply(form);
}

void tester::constrain_QG7()
{
    // (((x*y)*x)*y) == x
    //
    // After decomposition:
    // if (x*y==xy && xy*x==xy_x) then (xy_x*y==x)
    //
    // As a clause:
    // (x*y!=xy || xy*x!=xy_x || xy_x*y==x)
    for (unsigned x=0 ; x!=m_size ; ++x)
        for (unsigned y=0 ; y!=m_size ; ++y)
            for (unsigned xy=0 ; xy!=m_size ; ++xy)
                for (unsigned xy_x=0 ; xy_x!=m_size ; ++xy_x) {
                    constrain_QG7(x, y, xy, xy_x, 1);
                    constrain_QG7(x, y, xy, xy_x, 2);
                    constrain_QG7(x, y, xy, xy_x, 3);
                }
}

void tester::constrain_QG(unsigned i)
{
    switch(i) {
      case 0:/*no constraint*/ return;
      case 1: constrain_QG1(); break;
      case 2: constrain_QG2(); break;
      case 3: constrain_QG3(); break;
      case 4: constrain_QG4(); break;
      case 5: constrain_QG5(); break;
      case 6: constrain_QG6(); break;
      case 7: constrain_QG7(); break;
      default:
        throw runtime_error("Unrecognized constraint requirement");
    }
}

tester::tester(const flags & in_flags, unsigned size, unsigned constraint_type)
    : m_domain(size == 0 ? iset() : iset(make_pair(0, size-1))),
      m_size(size), m_is_idempotent(in_flags.idempotent),
      m_constraint_type(constraint_type),
      m_first_inverse_var(0),
      m_first_element_equality_allowed(0),
      m_first_element_inverse_equality_allowed(0),

      // For the sake of testing the all_different_propagator,
      // use it when the problem is fast enough anyways.
      // Currently, it is faster to use O(N^2) different_propagator objects
      // rather than one global all_different_propagator.
      m_use_all_different_propagator(in_flags.use_all_diff < 0 ?
                                     constraint_type == 3 :
                                     in_flags.use_all_diff >= 1)
{
    HCSP_PROFILE_CURRENT;
    ios_state_restore restorer(cout);
    cout << "Building QG" << constraint_type << "." << left << setw(2)
         << m_size << ".." << flush;
    m_solver.backtrack_limit(1000000);
    in_flags.apply(m_solver);
    for (unsigned row=0 ; row != m_size ; ++row)
        for (unsigned column=0 ; column != m_size ; ++column) {
            if (in_flags.idempotent && row == column)
                // required to be idempotent (i.e. a*a == a)
                m_solver.add_var(iset(row));
            else if (column == m_size-1 && row >= 2)
                // Minimize isomorphism (symmetry) as described in [1].
                m_solver.add_var(iset(make_pair(row-1, m_size-1)));
            else
                m_solver.add_var(m_domain);
        }
    
    for (unsigned row=0 ; row != m_size ; ++row)
        row_elements_are_permutation(row);

    for (unsigned column=0 ; column != m_size ; ++column)
        column_elements_are_permutation(column);
    
    constrain_QG(constraint_type);
    
    cout << left << "V=" << setw(3) << left  << m_solver.count_variables()
         << " C=" << setw(6) << m_solver.count_constraints() << " ";
}

void tester::print_element(unsigned element) const
{
    HCSP_ASSERT(element < 'z'+10);
    if ( element < 10 )
        cout << element;
    else
        cout << static_cast<char>(element-10+'a');
    
}

void tester::print_solution() const
{
    cout << " *|";
    for (unsigned column=0 ; column != m_size ; ++column) {
        cout << " ";
        print_element(column);
    }
    cout << "\n";
    cout << "--+";
    for (unsigned column=0 ; column != m_size*2 ; ++column)
        cout << "-";
    cout << "\n";
    for (unsigned row=0 ; row != m_size ; ++row) {
        print_element(row);
        cout << " |";
        for (unsigned column=0 ; column != m_size ; ++column) {
            cout << " ";
            print_element(m_solver.var_domain(var_index(row,column)).max());
        }
        cout << "\n";
    }
}

void tester::validate_QG1() const
{
    // if (x*y==u && z*w==u && v*y==x && v*w==z) then (x==z && y==w)
    for (unsigned u=0 ; u != m_size ; ++u)
        for (unsigned v=0 ; v != m_size ; ++v)
            for (unsigned w=0 ; w != m_size ; ++w)
                for (unsigned x=0 ; x != m_size ; ++x)
                    for (unsigned y=0 ; y != m_size ; ++y)
                        for (unsigned z=0 ; z != m_size ; ++z) {
                            if (var_value(x,y)==u && var_value(z,w)==u &&
                                var_value(v,y)==x && var_value(v,w)==z)
                                {
                                    T_ASSERT(x==z && y==w);
                                }
                        }
}

void tester::validate_QG2() const
{
    // if (x*y==u && z*w==u && y*v==x && w*v==z) then (x==z && y==w)
    for (unsigned u=0 ; u != m_size ; ++u)
        for (unsigned v=0 ; v != m_size ; ++v)
            for (unsigned w=0 ; w != m_size ; ++w)
                for (unsigned x=0 ; x != m_size ; ++x)
                    for (unsigned y=0 ; y != m_size ; ++y)
                        for (unsigned z=0 ; z != m_size ; ++z) {
                            if (var_value(x,y)==u && var_value(z,w)==u &&
                                var_value(y,v)==x && var_value(w,v)==z)
                                {
                                    T_ASSERT(x==z && y==w);
                                }
                        }
}

void tester::validate_QG3() const
{
    // (x*y)*(y*x) == x,
    for (unsigned x=0 ; x != m_size ; ++x)
        for (unsigned y=0 ; y != m_size ; ++y)
            T_ASSERT( var_value( var_value(x,y), var_value(y,x) ) == x );
}

void tester::validate_QG4() const
{
    // (x*(y*x))*y == x
    for (unsigned x=0 ; x != m_size ; ++x)
        for (unsigned y=0 ; y != m_size ; ++y) {
            T_ASSERT(var_value(var_value(x, var_value(y,x)),y)==x);
        }
}

void tester::validate_QG5() const
{
    // ((x*y)*x)*x == y
    for (unsigned x=0 ; x != m_size ; ++x)
        for (unsigned y=0 ; y != m_size ; ++y) {
            T_ASSERT(var_value(var_value(var_value(x,y),x),x)==y);
        }
}

void tester::validate_QG6() const
{
    // (x*y)*y == x*(x*y)
    for (unsigned x=0 ; x != m_size ; ++x)
        for (unsigned y=0 ; y != m_size ; ++y) {
            T_ASSERT(var_value(var_value(x,y),y) ==
                     var_value(x,var_value(x,y)));
        }
}

void tester::validate_QG7() const
{
    // (((x*y)*x)*y) == x
    for (unsigned x=0 ; x != m_size ; ++x)
        for (unsigned y=0 ; y != m_size ; ++y) {
            T_ASSERT(var_value(var_value(var_value(x,y),x),y) == x);
        }
}


void tester::validate_QG(unsigned constraint_type) const
{
    switch(constraint_type) {
      case 0:/*no constraint*/ return;
      case 1: validate_QG1(); break;
      case 2: validate_QG2(); break;
      case 3: validate_QG3(); break;
      case 4: validate_QG4(); break;
      case 5: validate_QG5(); break;
      case 6: validate_QG6(); break;
      case 7: validate_QG7(); break;
      default:
        throw runtime_error("Unrecognized constraint requirement");
    }
}

void tester::validate_solution() const
{
    iset elements;

    // all on row are different
    for (unsigned row=0 ; row != m_size ; ++row) {
        elements.clear();
        for (unsigned column=0 ; column != m_size ; ++column) {
            const iset &var_domain= m_solver.var_domain(var_index(row,column));
            T_ASSERT( var_domain.singleton() );
            iset::value_type element= var_domain.max();
            T_ASSERT( !elements.contains(element) );
            elements.insert(element);
        }
    }

    // all on column1 are different
    for (unsigned column=0 ; column != m_size ; ++column) {
        elements.clear();
        for (unsigned row=0 ; row != m_size ; ++row) {
            iset::value_type element=
                m_solver.var_domain(var_index(row,column)).max();
            T_ASSERT( !elements.contains(element) );
            elements.insert(element);
        }
    }
    if (m_is_idempotent) {
        for (unsigned i=0 ; i != m_size ; ++i)
            T_ASSERT(m_solver.var_domain(var_index(i,i)).max()==
                     static_cast<iset::value_type>(i));
    }
    validate_QG(m_constraint_type);
}


void tester::run()
{
    HCSP_PROFILE_CURRENT;
    const unsigned max_known_size=
        sizeof(m_expected_results)/sizeof(m_expected_results[0])-1;
    cout << "Solve:" << flush;
    unsigned long time_before= user_time_microseconds()/1000;
    solver::result_type result= m_solver.solve(&cout);
    unsigned long time_after= user_time_microseconds()/1000;
    ios_state_restore restorer(cout);
    cout << setprecision(2) << setw(7) << fixed
         << double(time_after-time_before)/1000.0 << " sec"
         << setw(7) << m_solver.backtracks_done() << " backtracks "
         << setw(2) << m_solver.restarts_done() << " rst ";
    cout << result << endl;
    switch(result) {
      case solver::SAT:
        print_solution();
        if (m_size <= max_known_size)
            T_ASSERT(m_expected_results[m_size][m_constraint_type]!=NO);
        break;
      case solver::UNSAT:
        if (m_size <= max_known_size)
            T_ASSERT(m_expected_results[m_size][m_constraint_type]!=YES);
        break;
      default:
        ; // do nothing
    }
    if (result == solver::SAT)
        validate_solution();
}

static const unsigned limits[NUM_OF_SPEEDS][9]= {
    // QG0.*, QG1.*, QG2.*, QG3.*, QG4.*, QG5.*, QG6.*, QG7.*
    {  12   , 7    , 7    , 8    , 9    , 12   , 11   , 11   },// FAST 
    {  12   , 7    , 7    , 8    , 9    , 12   , 11   , 13   },// SECONDS
    {  13   , 8    , 8    , 8    , 9    , 12   , 11   , 13   },// DOZENS_SECOND
    {  13   , 8    , 8    , 9    , 9    , 14   , 13   , 13   } // MINUTES
};

static void toggle_idempotent(bool & idempotent)
{
  idempotent= !idempotent;
  cout << "Set idempotent=" << (idempotent? "yes" : "no") << "\n";
}

difficulty_types parse_difficulty(const char *text)
{
    difficulty_types  difficulty= static_cast<difficulty_types>(atoi(text));
    if (difficulty > MINUTES_RUNS)
        difficulty= MINUTES_RUNS;
    return difficulty;
}

static bool parse_problem_specifier(const string & specifier,
                                    pair<unsigned, unsigned> &parsed_specifier)
{
    char *end_ptr;
    const char *pos= specifier.c_str();
    bool error= false;
    unsigned type= strtoul(pos, &end_ptr, 0);
    unsigned order= 1;
    if ( *end_ptr != '.' )
        error= true;
    else
        pos= end_ptr+1;

    if (!error) {
        order= strtoul(pos, &end_ptr, 0);
        if ( *end_ptr != '\0' )
            error= true;
    }

    if (type > 7)
        error= true;

    if (order < 1)
        error= 1;
    if (error) {
        cout << "'" << specifier
             << "' is an illegal quasigroup specifier. It should be \n"
             << " 0.n 1.n 2.n ... 7.n where n is a positive integer\n";
        return false;
    }
    parsed_specifier.first=  type;
    parsed_specifier.second= order;
    return true;
}

bool tester::flags::parse_option(char option)
{
    if (hcsp_flags::parse_option(option))
        return true;
    switch(option) {
      case 'I': toggle_idempotent(idempotent);            break;
      case 'd': difficulty= parse_difficulty(optarg);     break;
      case 'q': problems_requested.push_back(optarg);     break;
      default:
        return false;
    }
    return true;
}

void tester::flags::usage_short(ostream & out)
{
    hcsp_flags::usage_short(out);
    out << " [-I] [-d difficulty] [-q quasigroup_type]";
}

void tester::flags::usage_long(ostream & out)
{
    hcsp_flags::usage_long(out);
    out << "  -d difficulty where difficulty is one of the following:\n"
        << "      0   [FAST]              "
        << "Very fast, total run is measured is seconds\n"
        << "      1   [SECONDS]           "
        << "No CSP measures more than several seconds\n"
        << "      2   [DOZENS OF SECONDS] "
        << "CSPs take up to several dozens of seconds\n"
        << "      3   [MINUTES]           "
        << "Some CSPs may take up to 20 minutes\n"
        << "  -q t.n Where 0 <= t <= 7  and n >= 1.\n"
        << "        't' defines the quasigroup constraint type, and\n"
        << "        'n' defines the order (i.e. the size) of the quasigroup\n"
        << "        '-q 1.8' will construct and solve the q1.8 problem\n"
        << "     This flag overrides the difficulty arguments, and will\n"
        << "     make the program to work only on the given problem.\n"
        << "     It is possible to give more than one qg type by several "
        << "-q flags.\n"
        << "  -I toggle whether or not the quasigroup is idempotent\n";
}

void tester::flags::validate_consistent() const
{
    if (effort_boost>= 0)
        throw runtime_error("Effort boost requested but unsupported");
    if (keep_constraints_from_conflicts>= 0)
        throw runtime_error("keep constraints from conflict unsupported");
    if (erase_after_backtracks>= 0)
        throw runtime_error("erase after backtracks unsupported");
    if (erase_after_backtracks>= 0)
        throw runtime_error("erase after backtracks unsupported");
}

static int run_specific_problems(const tester::flags & flags)
{
    vector<pair<unsigned, unsigned> > parsed_specifiers;
    for (unsigned i=0 ; i != flags.problems_requested.size() ; ++i) {
        pair<unsigned, unsigned> parsed_specifier;
        if ( !parse_problem_specifier(flags.problems_requested[i],
                                      parsed_specifier) )
            return 1;
        parsed_specifiers.push_back(parsed_specifier);
    }
    for (unsigned i=0 ; i != parsed_specifiers.size() ; ++i) {
        unsigned type=  parsed_specifiers[i].first;
        unsigned order= parsed_specifiers[i].second;
        tester(flags, order, type).run();
    }
    return 0;
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    HCSP_PROFILE_CURRENT;
    tester::flags flags;
    if (!flags.parse(argc, argv))
        return 1;
    cout << "Seed is " << set_seed(flags.seed) << endl;
    if ( !flags.problems_requested.empty() )
        return run_specific_problems(flags);

    for (unsigned qg=0 ; qg <= 7 ; ++qg) {
        for (unsigned order=4 ; order <= limits[flags.difficulty][qg] ; ++order)
            tester(flags, order, qg).run();
    }
    
    END_TEST_MAIN;
}


