// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <iostream>
#include <sstream>
#include <vector>
#include <algorithm>
#include <iset.hpp>
#include <clause_propagator.hpp>
#include <solver.hpp>
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;

class tester {
  public:
    typedef solver::reporting_levels reporting_levels;
    tester(reporting_levels level, unsigned holes, unsigned pigeons);
    void run();

  private:
    static unique_ptr<clause_propagator>
    make_clause(const iset & a, const iset & b);

    void add_shuffled_clause(unique_ptr<clause_propagator> clause,
                             int var_a, int var_b);
    void add_diff(unsigned var_a, unsigned var_b);
    solver m_solver;
    const unsigned m_holes;
    const unsigned m_pigeons;
};

unique_ptr<clause_propagator>
tester::make_clause(const iset & a, const iset & b)
{
    unique_ptr<clause_propagator> clause( new clause_propagator );
    clause->update_literals().reserve(3);
    clause->update_literals().push_back(a);
    clause->update_literals().push_back(b);
    return clause;
}

void tester::add_diff(unsigned var_a, unsigned var_b)
{
    const iset complete_set(make_pair(0, m_holes-1));
    
    for (unsigned i=0 ; i < m_holes ; ++i) {
        iset not_i=complete_set;
        not_i.erase(i);
        add_shuffled_clause(make_clause(not_i, not_i), var_a, var_b);
    }
}

void tester::add_shuffled_clause(unique_ptr<clause_propagator> clause,
                                 int var_a, int var_b)
{
    int vars[] = {var_a, var_b};

    static vector<unsigned> orders;
    orders.clear();
    for (unsigned i=0 ; i < sizeof(vars)/sizeof(vars[0]) ; ++i)
        if (vars[i] >= 0)
            orders.push_back(i);

    randomizer rand;
    random_shuffle(orders.begin(), orders.end(), rand );

    static clause_propagator::literals_type original_literals;
    original_literals= clause->literals();

    for (unsigned i=0 ; i != orders.size() ; ++i)
        clause->update_literals()[i]= original_literals[orders[i]];
    
    unique_ptr<constraint_propagator> constraint(clause.release());
    constraint_connectivity_info & connectivity=
        m_solver.add_constraint(move(constraint))
        .reserve_arguments(orders.size() );
    for (unsigned i=0 ; i != orders.size() ; ++i)
        connectivity << vars[orders[i]];
}

tester::tester(reporting_levels level, unsigned holes, unsigned pigeons)
    :  m_holes( holes ), m_pigeons( pigeons )
{
    m_solver.report_level(level);
    for (unsigned i=0 ; i < m_pigeons*2 ; ++i)
        m_solver.add_var(iset(make_pair(0, m_holes-1)));

    for (unsigned i=0 ; i < m_pigeons ; ++i)
        for (unsigned j=i+1 ; j < m_pigeons ; ++j)
            add_diff(i,j);
    for (unsigned i=0 ; i < m_pigeons ; ++i)
        add_diff(i,i+m_pigeons);
}


void tester::run()
{
    m_solver.backtrack_limit(100000);
    solver::result_type result= m_solver.solve(&cout);
    T_ASSERT(result == solver::UNSAT);
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    solver::reporting_levels report_level= solver::NONE;
    //report_level= solver::SEARCH;
    //report_level= solver::PROPAGATION_MODS;
    //report_level= solver::PROPAGATED_CONSTRAINTS;

    cout << "Seed is " << set_seed(argc, argv) << endl;
    for (unsigned i=0 ; i < 100 ; ++i) {
        if (i != 0) {
            unsigned seed= gen();
            srand(seed);
            if (report_level > solver::NONE)
                cout << "Seed is " << seed << endl;
        }
        tester tester(report_level, 6, 8);
        tester.run();
    }
    
    END_TEST_MAIN;
}

