// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2012,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2012,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <iostream>
#include <sstream>
#include <c_like_solver.hpp>
#include <flags.hpp>
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;

struct expression_type
{
    string str;
    solver::result_type expected;
    int max_backtracks; // Negative means 'default'
};

enum { DEFAULT=100 };

expression_type expressions[]= {
    {"bool a,b,c ; (a || b) && !(a || b || c); // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"bool a,b,c ; (a || b) && !(b || c); // a = 1, b =0, c =0\n",
     solver::SAT,   DEFAULT},
    {"bool a,b,c ; (a || b) && !(b || c); !a; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"bool a,b,c ; (a || b) && !(b || c); b; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"bool a,b,c ; (a || b) && !(b || c); c != 0; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"bool a,b,c; a*1+b*3+c*+11== +14; //  result a=0,b=1, c=1\n",
     solver::SAT,   DEFAULT},
    {"bool a,b,c; a*1+b*3+c*11== 16; //  UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"unsigned a,b,c; a<=2;b<=2;c<=2; a*1+b*3+c*11== 16;//result a=2,b=1,c=1\n",
     solver::SAT,   DEFAULT},
    {"unsigned a,b,c; a<=1; a*1+b*3+c*11== 16; // result a=1,b=5,c=0\n",
     solver::SAT,   DEFAULT},
    {"unsigned a,b,c; a<=1; b<=4; c<=2; a*1+b*3+c*11== 16; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"unsigned a,b,c; a<=b*c ; a+1>=c*b; a*b==c; a<b; //a=0,b>=1,c=0\n",
     solver::SAT,   DEFAULT},
    {"unsigned a,b,c; a<=b*c ; a+1>=c*b; a*b==c; a<b; c>0; //UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"unsigned a,b,c; a<=b*c ; a+1>=c*b; a*b==c; a<b; b<1; //UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"unsigned a,b,c; a<=b*c ; a+1>=c*b; a*b==c; a<b; b==1; //a=0, b=1, c=0\n",
     solver::SAT,   DEFAULT},
    {"unsigned a,b,c; a<=b*c ; a+1>=c*b; a*b==c; a<b; a!=0; //UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"unsigned a,b,c; a<=b*c ; a+1>=b*c; a*b==c; a<b; a!=0; //UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"unsigned a; 10<=20/a ; //a:{1,2}\n",
     solver::SAT,   DEFAULT},
    {"unsigned a; 10<=20/a ; a == 1; //a=1\n",
     solver::SAT,   DEFAULT},
    {"unsigned a; 10<=20/a ; a == 2; //a=2\n",
     solver::SAT,   DEFAULT},
    {"unsigned a; 10<=20/a ; a != 1 && a != 2; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"bool a,b; (a || b || a); // a=1 or b=1\n",
     solver::SAT,   DEFAULT},
    {"bool a,b; (a || b || a); !b; // a=1, b=0\n",
     solver::SAT,   DEFAULT},
    {"bool a,b; (a || b || a); !a; // a=0, b=1\n",
     solver::SAT,   DEFAULT},
    {"bool a,b; (a || b || a); !a; !b; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"bool a,b; !a ; b==0 ; (a || b || a);  // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"bool a,b,c; int d; ( a || b || c) >= d; d>a; // a=0, (b=1 or c=1), d=1\n",
     solver::SAT,   DEFAULT},
    {"bool a,b,c; int d; ( a || b || c) >= d; d>a; d>b; // a=0, b=0,c=1, d=1\n",
     solver::SAT,   DEFAULT},
    {"bool a,b,c; int d; ( a || b || c) >= d; d>a; d>b; d>c; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"bool a,b,c; int d; ( a || b || c) <= d; d<a; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"bool a,b,c; int d; ( a || b || c) <= d; d<a+b; c<d;// a=1,b=1,c=0,d=1\n",
     solver::SAT,   DEFAULT},
    {"unsigned a,b, c; a == b; b != c; a == c; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"unsigned a,b, c; a == b; c != b; a == c; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"unsigned a,b, c; b != c; a == b; a == c; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"unsigned a,b, c; c != b; a == b; a == c; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"unsigned a,b, c; a == b; a == c; b != c; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"unsigned a,b, c; a == b; a == c; c != b; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"unsigned a,b, c; a == c; c != b; // satisfied, somehow\n",
     solver::SAT,   DEFAULT},
    {"unsigned a,b, c; a == b; a == c; // satisfied, somehow\n",
     solver::SAT,   DEFAULT},
    {"unsigned a,b,c; 1==(a==b); c==(a==b); //c=1, a==b\n",
     solver::SAT, 0},
    {"unsigned a,b,c; 1==(a==b); c==(a==b); c!=1; //UNSAT\n",
     solver::UNSAT, 1},
    {"unsigned a,b,c; 0==(a==b); c==(a==b); //c=0, a!=b\n",
     solver::SAT, 0},
    {"unsigned a,b,c; 0==(a==b); c==(a==b); c!=0; //UNSAT\n",
     solver::UNSAT, 1},
    {"unsigned a,b,c; 1==(a==b); c==(a!=b); //c=0, a==b\n",
     solver::SAT, 0},
    {"unsigned a,b,c; 1==(a==b); c==(a!=b); c!=0; //UNSAT\n",
     solver::UNSAT, 1},
    {"unsigned a,b,c ; a == 1 ; b == 1;  c == (a != b); // a=1,b=1,c=0\n",
     solver::SAT,   DEFAULT},
    {"unsigned a,b,c ; a == 1 ; b == 1;  c == (a != b); c!=0; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"unsigned a,b,c ; a == 1 ; b == 1;  c == (a == b); // a=1,b=1,c=1\n",
     solver::SAT,   DEFAULT},
    {"unsigned a,b,c ; a == 1 ; b == 1;  c == (a == b); c!=1; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"bool V0, V1,V2; V0 == 0; (V2 != V1) || V0; // V0=0, V1 == not V0\n",
     solver::SAT,   DEFAULT},
    {"int a,b; a<= min(a,b); // SAT\n", solver::SAT, DEFAULT},
    {"int a,b; a> min(a,b);  // SAT\n", solver::SAT, DEFAULT  },
    {"int a,b; 2 == min(a,b); b == 5; a != 2;  // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"int a,b; 2 == min(a,b); b == 5; a == 2;  // SAT\n", solver::SAT, DEFAULT},
    {"int a,b; a<= max(a,b); // SAT\n", solver::SAT, DEFAULT},
    {"int a,b; a< max(a,b);  // SAT\n", solver::SAT, DEFAULT},
    {"int a,b; 5 == max(a,b); b == 2; a != 5;  // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"int a,b; 5 == max(a,b); b == 2; a == 5;  // SAT\n", solver::SAT, DEFAULT},
    {"int a,b,c; a==b%c; b==5*c; // SAT\n", solver::SAT, DEFAULT},
    {"int a,b; a==b%5; abs(a)>=5; abs(a) <= 10; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"int a,b; a==b%5; a==4; // SAT\n", solver::SAT, DEFAULT},
    {"int a,b; a==b%5; a==4; b == 9; // SAT\n", solver::SAT, DEFAULT},
    {"int a,b,c; a==b ? a==c : a==b+1; a>b; // SAT\n", solver::SAT, DEFAULT},
    {"int a,b,c; a==b ? a==c : a==b+1; a<b; abs(b) <= 9; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"int a,b,c; a==b ? a==c : a==b+1; a!=c; abs(a) <= 9; a==b; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"int a,b,c; a==b ? a==c : a==b+1; abs(b) <= 9; // SAT\n",
     solver::SAT, DEFAULT},
    {"int a; a == (0 ? 5 : 7); a != 7; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"int a; a == (0 ? 5 : 7); a == 7; // SAT\n",
     solver::SAT, DEFAULT},
    {"int a; a == (1 ? 0 ? 1 : 2 : 3); a == 2; // SAT\n",
     solver::SAT, DEFAULT},
    {"int a; a == (0 ? 0 ? 1 : 2 : 3); a == 3; // SAT\n",
     solver::SAT, DEFAULT},
    {"int a; a == (0 ? 0 ? 1 : 2 : 0 ? 3 : 4); a == 4; // SAT\n",
     solver::SAT, DEFAULT},
    {"int a; a == (0 ? 0 ? 1 : 2 : 1 ? 3 : 4); a == 3; // SAT\n",
     solver::SAT, DEFAULT},
    {"int a; a == (0 ? 0 ? 1 : 2 : 1 ? 3 : 4); a != 3; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"int {1,3} a,b; a!=1; b!=3; // SAT\n", solver::SAT, DEFAULT},
    {"int {1,3} a,b; a!=1; b!=3; a == b; // UNSAT\n", solver::UNSAT, DEFAULT},
    {"int {1,3} a; int {2,4} b; a!=1; b!=4; // SAT\n", solver::SAT, DEFAULT},
    {"int {1,3} a; int {2,4} b; a!=1; b!=4; a < b; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"int {1:5:11} a; a==a; // SAT\n", solver::SAT, 0},
    {"int a,b,c; a==b%c; b==5*c; abs(c)<=10; a!=0; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    // Since 1073741827 is a prime and abs(a),abs(b) are smaller then
    // there is no feasible assignment. The propagator, even if an
    // over-approximated one, must figure out that there is no solution.
    {"int {0:30} a,b; 31 == a*b; // UNSAT\n",
     solver::UNSAT, 0},
    {"int {1:8} x,y,z;\n"
     "!overlap(x,4,y,4) && !overlap(x,4,z,4) && !overlap(y,4,z,4); // UNSAT\n",
     solver::UNSAT, 1},
    {"int {1:9} x,y,z;\n"
     "!overlap(x,4,y,4) && !overlap(x,4,z,4) && !overlap(y,4,z,4); // SAT\n",
     solver::SAT, 0},
    {"int {1:9} x,y,z;\n"
     "!overlap(x,4,y,4) && !overlap(x,4,z,4) && !overlap(y,4,z,4);\n"
     "(x!=1 && x != 5 && x!=9) || (y!=1 && y != 5 && y!=9); //UNSAT\n",
     solver::UNSAT, 5},
    //FIXME: The conflict should be detected for unlimited int without
    //       excessively long consistency iterations,
    {"int {0:9} x,y; sub_interval(x,y,3,7); !sub_interval(x,y,4,7);\n"
     "x != 3; // UNSAT\n",
     solver::UNSAT, 0},
    //FIXME: Consistency should be done for unlimited int without excessively
    //       long iterations. [limiting int to 0:9 in order to avoid this]
    {"int {0:9} x,y; sub_interval(x,y,3,7); !sub_interval(x,y,4,7); // SAT\n",
     solver::SAT, 0},
    {"int x,y; sub_interval(x,y,3,7);  x != 3; // SAT\n",
     solver::SAT, 0},
    {"int x,y; sub_interval(x,y,4,7);  x != 3; // SAT\n",
     solver::SAT, 0},
    {"int {1:10} x,y; sub_interval(x,y,3,7); !sub_interval(x,y,4,8,3); //x=3\n",
     solver::SAT, 0},
    {"int {1:10} x,y; sub_interval(x,y,3,7); !sub_interval(x,y,4,8,3); x!=3;\n",
     solver::UNSAT, 0},
    {"int {1:10} x,y; sub_interval(x,y,3,7); !sub_interval(x,y,4,8,3); y!=6;"
     "y!=7 ;\n", solver::UNSAT, 0},
    {"int {1:10} x,y; sub_interval(x,y,3,7); !sub_interval(x,y,4,8,4);\n",
     solver::SAT, 0},
    {"int {1:10} x,y; sub_interval(x,y,3,7); !sub_interval(x,y,4,8,5);\n",
     solver::UNSAT, 0},
    {"int {0:20} a,b,c,d; a<b; b<c; a>=c || d==0; // d==0\n", solver::SAT, 1},
    {"int {0:20} a,b,c,d; a<b; b<c; a>=c || d==0; d!=0; // UNSAT\n",
     solver::UNSAT, 0},
    {"int {0:20} a,b,c,d; a<b; b<c; a>=c-1 || d==0; d!=0; // UNSAT\n",
     solver::UNSAT, 0},
    {"int {0:20} a,b,c,d; a<b; b<c; a+1>=c || d==0; d!=0; // UNSAT\n",
     solver::UNSAT, 0},
    {"int {0:20} a,b,c,d; a<b; b<c; a>=c-2 || d==0; d!=0; // a=0,b=1,c=2\n",
     solver::SAT, 0},
    {"int {0:20} a,b,c,d; a<b; b<c; a+2>=c || d==0; d!=0; // a=0,b=1,c=2\n",
     solver::SAT, 1},
    {"int {0:20} a,b,c,d; a<b; b<c; a>c || d==0; // d==0\n", solver::SAT, 1},
    {"int {0:20} a,b,c,d; a<b; b<c; a>c || d==0; d!=0; // UNSAT\n",
     solver::UNSAT, 1},
    {"int {0,1:2:29,30} x; int {0:2:30} y; bool z;\n"
     "z != (x == 0 || y == 30); x<=y && y <= x; // z=0 && (x=y=0 || x=y=30) \n",
     solver::SAT, 1},
    {"int {1:2:29} x; int {0:2:30} y; bool z;\n"
     "z != (x == 0 || y == 30); x<=y && y <= x; // UNSAT \n", solver::UNSAT, 0},
    {"int {0:31} u,x,y,z; bool a,b,c,d, e;\n"
     "a || u < x; b || x < y; c || y <= z; d || z <= u;\n"
     "a >= b || b <= a || a==b; c >= d || d <= c || c==d;\n",
     solver::SAT, 2},

#ifdef USE_BIG_ISET
    {"unsigned a,b,c; a<=b*c ; a+1>=c*b; a*b==c; a<b; b>99; //a=0, b>99, c=0\n",
     solver::SAT,   DEFAULT},
    {"unsigned a,b,c; a<=b*c ; a+1>=b*c; a*b==c; a<b; b>99; //a=0, b>99, c=0\n",
     solver::SAT,   DEFAULT},
    {"int a,b; b>=abs(a+-10)*100; b < abs(a-5); a < abs(b+6);// UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"int a,b; b>=abs(a-10)*100; b < abs(a-5); a < abs(b+7); // a==10, b=4\n",
     solver::SAT,   DEFAULT},
    {"int a,b; b>=abs(a-10)*100; b < abs(a-5); a < abs(b+7); b!=4; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"int a,b; b>=abs(a-10)*100; b < abs(a-5); a < abs(b+7); a!=10;// UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"int a,b,c; a >= -10; a <=10; a < b; b<=c-1; c<=a-3; // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"int a,b,c; a>-1; a<1; a < b; b<=c-1; c<=a+2; // result a=0, b=1,c=2\n",
     solver::SAT,   DEFAULT},
    {"int a,b; a>=-10 ; a<=10 ;a< min(a,b); // UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"int a,b; a> max(a,b); a>=-10; a<=10; // UNSAT\n", solver::UNSAT, DEFAULT},
    {"int a,b; a < 0 ; a<= min(a,b); // SAT\n",  solver::SAT, DEFAULT  },
    {"int a,b; b < 0 ; a<= min(a,b); // SAT\n",  solver::SAT, DEFAULT  },
    {"int a,b; a < 0 ; a>= max(a,b); // SAT\n",  solver::SAT, DEFAULT  },
    {"int a,b; b < 0 ; a>= max(a,b); // SAT\n",  solver::SAT, DEFAULT  },
    {"int a,b; a==b%5; a==-4; abs(b) <= 10; // SAT\n", solver::SAT, DEFAULT},
    {"int a,b; a==b%5; a==-4; b == -9; // SAT\n", solver::SAT, DEFAULT},
    {"int a,b; a==b%5; a!=-4; b == -9; // UNSAT\n", solver::UNSAT, DEFAULT},
    {"int {-50:50} a,b; int {-2147483648} min_int; \n"
     "a==b%min_int; a<b; // UNSAT\n", solver::UNSAT, 0},
    {"int {-50:50} a,b; int {-2147483648} min_int; \n"
     "a==b%min_int; a>b; // UNSAT\n", solver::UNSAT, 0},
    {"int {-50:50} a,b; int {-2147483648} min_int; \n"
     "a==b%min_int; a==b; // SAT\n", solver::SAT, 0},
    {"int  c,d; int {-2147483648} min_int; min_int==c%d; // UNSAT\n",
     solver::UNSAT, 0},
    {"int {-1:2} a,b; a <= 0; a == 5-abs(b); // UNSAT\n",
     solver::UNSAT, 0},
    {"int {-17,-1} a; int {1:9} b; int {-2,-1} c; a == b *c;//a=-1,b=1,c=-1\n",
     solver::SAT, 1},
    {"int a,b; b <= 9; a >=3; a == b *2; // e.g. a=4,b=2\n",
     solver::SAT,   0},
    {"int {1001:1999,1000000} a; int  {1:1000} b; a == b *1000; // b=1000\n",
     solver::SAT, 0},
    // Since 1073741827 is a prime and abs(a),abs(b) are smaller then
    // there is no feasible assignment. The propagator, even if an
    // over-approximated one, must figure out that there is no solution.
    {"int {-1073741826:1073741826} a,b; 1073741827 == a*b; // UNSAT\n",
     solver::UNSAT, 0},
    // Test intersection of iset with different skip sizes. This should be fast
    {"int{-2000000000:5:2000000000}a; int{-1999999998:3:1999999998}b; a == b;"
     " // SAT\n", solver::SAT, 0},
    // Although the ranges seem to overlap, they contain no shared item.
    // This is because the gcd of the steps is 1 which makes the steps of the
    // intersection to be 20000003*211 > UINT_MAX. This lets us choose
    // the bounds so that no element is contained in the 2 intervals
    // simultaneously.
    {"int{-2000000000:20000003:2000000600}a; int{-2019999792:211:2000000060}b;"
     "a == b; // UNSAT\n", solver::UNSAT, 0},
    // This tests the speed/strength of the learning process since a conflict is
    // detected after about 1000000 propagation steps of the same 3 constraints.
    // Conflict analysis should be either very fast or detect that the three
    // constraints conflict without checking all 1000000 propagation steps.
    // If this does not work as expected, the code will get stuck for hours
    // instead of less than a second. This will happen as long as
    // var_domain_after_modification() works in O(number of modifications)
    {"int {0:1000000} a,b,c,d; a<b; b<c; a>=c || d==0; // d=0\n",
     solver::SAT, 1},
    // This tests the ability to learn "z || y==x", otherwise it gets to
    // O(n^3) issues.
    {"int {1:2:19999} x; int {0:2:20000} y; bool z;\n"
     " x<=y; z || y <= x; // z=1", solver::SAT, 1},
    {"int {10:99} x0; int {0:9,90:99} y0; int {0:89} z0; int {0:99} x,y,z;\n"
     "bool b;\n"
     "x <= y && y <= z; b || x0==x ; b || y0==y ; b || z0==z ;\n",
     solver::SAT, 1},
    {"int {0:9999} u,x,y,z; bool a,b,c,d, e;\n"
     "a || u < x; b || x < y; c || y <= z; d || z <= u;\n"
     "a >= b || b <= a || a==b; c >= d || d <= c || c==d;\n",
     solver::SAT, 2},

#endif
};

int main_impl(int argc, char *argv[])
{
    hcsp_flags flags;
    // The limit has to be small enough to detect weak expression optimization,
    // in which case the solver stops with BACKTRACK_LIMIT.
    flags.parse(argc, argv);

    for (unsigned i=0 ; i != sizeof(expressions)/sizeof(expressions[0]); ++i) {
        istringstream str(expressions[i].str);
        stringstream test_name;
        test_name << "test" << i;
        try {
            table_propagator_base::memory_manager *p_null= NULL;
            c_like_solver problem(str, test_name.str(), flags, *p_null) ;
            if (flags.backtrack_limit < 0) {
                int max_backtracks= expressions[i].max_backtracks;
                problem.backtrack_limit(max_backtracks);
            }
            flags.quiet_start= true;
            solver::result_type result= problem.solve(&cout);
            T_ASSERT(result == expressions[i].expected);
        } catch (...) {
            cerr << "Found in: " << test_name.str() << ": `"
                 << expressions[i].str << "'" << endl;
            throw;
        }
    }
    return 0;
}

int main(int argc, char *argv[])
{
    try {
        return main_impl(argc, argv);
    } catch (std::exception & e) {
        cout << "Exception(" << typeid(e).name() << "):\n" << e.what() << endl;
    } catch (...) {
        cout << "unknown exception" << std::endl;
    }
    return 1;                                               
}
