// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <set>
#include <algorithm>
#include <limits>
#include <utilities.hpp>
#include <clause_data.hpp>
#include <solver.hpp>
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;

class tester {
  public:
    typedef solver::reporting_levels reporting_levels;
    tester(reporting_levels, unsigned number_of_variables)
        : m_relevant_variables(random_nums(number_of_variables)) {}
    void run();

  private:
    //! Attempt to generate 'size' different values
    static vector<unsigned> random_nums(unsigned size);
    
    //! Generate a canonical random clause which uses only m_relevant_variables
    clause_data random_clause() const;
    unsigned random_var_index() const
    { return m_relevant_variables[gen(m_relevant_variables.size())];}

    //! Change clause representation without affecting semantics
    void shuffle_clause(clause_data & clause) const;

    //! Randomly enlarge some of the sets of the literals, return true if
    //! clause semantics has changed.
    bool enlarge_sets(clause_data & clause) const;

    //! Randomly add some new variables to the clause, return true if
    //! clause semantics has changed. Note that adding an empty set
    //! with a new variable does not change semantics
    bool add_variables(clause_data & clause) const;
    
    //! Test that two equivalent clauses are detected as such
    void test_equivalent();
    //! Test that two different clauses are detected as such
    void test_diff();
    
    // Sorted list of unique variable indexes
    const vector<unsigned> m_relevant_variables;
};

vector<unsigned> tester::random_nums(unsigned size)
{
    set<unsigned> result;
    static const unsigned MIN_VAR= numeric_limits<unsigned>::min();
    static const unsigned MAX_VAR= numeric_limits<unsigned>::max();

    for (unsigned attempts_left= size*2+10; attempts_left!=0; --attempts_left){
        if (result.size() == size)
            break;
        switch(gen(8)) {
          case 0: result.insert(MIN_VAR); break;
          case 1: result.insert(MAX_VAR); break;
          case 2:
            if (!result.empty() && *result.begin() != MAX_VAR)
                result.insert(*result.begin()+1);
            break;
          case 3:
            if (!result.empty() && *--result.end() != MIN_VAR)
                result.insert(*--result.end()-1);
            break;
          default:
            result.insert(hcsp::random::range(MIN_VAR, MAX_VAR));
        }
    }
    return vector<unsigned>(result.begin(), result.end());
}

clause_data tester::random_clause() const
{
    clause_data result;
    for (unsigned i=0 ; i!= m_relevant_variables.size() ; ++i) {
        if (gen(2) == 0)
            continue;
        result.push_back(literal_data(m_relevant_variables[i],
                                      random_var()));
    }
    return result;
}

void tester::shuffle_clause(clause_data & clause) const
{
    if (gen(8) == 0) {
        random_shuffle(clause.begin(), clause.end());
        return;
    }

    if (!clause.empty()) {
        for (unsigned i=gen(8) ; i != 0 ; --i) {
            unsigned literal_position= gen(clause.size());
            unsigned variable= clause[literal_position].m_variable;
            iset to_handle_set= clause[literal_position].m_set;
            clause.push_back(literal_data(variable, to_handle_set));
            iset & first_set = clause[literal_position].m_set;
            iset & second_set= clause.back().m_set;
            if (gen(8) == 0) // make an exact copy
                continue;
            first_set.intersect(random_var());
            second_set.intersect(random_var());
            to_handle_set.erase_all_of(first_set);
            to_handle_set.erase_all_of(second_set);
            first_set.insert_all_of(to_handle_set);
        }
    }
    if (!m_relevant_variables.empty()) {
        for (unsigned i=gen(4) ; i != 0 ; --i)
            clause.push_back(literal_data(random_var_index(), iset()));
        
        
        for (unsigned i=gen(4) ; i != 0 ; --i)
            clause.push_back(literal_data(random_var_index(), iset()));
    }

    random_shuffle(clause.begin(), clause.end());
}

void tester::test_equivalent()
{
    clause_data a=random_clause();
    clause_data b=a;
    
    T_ASSERT(a.equivalent(b) && b.equivalent(a));
    
    shuffle_clause(a);
    T_ASSERT(a.equivalent(b) && b.equivalent(a));
    
    shuffle_clause(b);
    T_ASSERT(a.equivalent(b) && b.equivalent(a));
}

bool tester::enlarge_sets(clause_data & clause) const
{
    bool modified= false;
    if (clause.empty())
        return false;

    for (unsigned i=gen(4) ; i > 0 ; --i) {
        iset & target = clause[gen(clause.size())].m_set;
        iset original= target;

        target.insert_all_of(random_var());
        if (original != target)
            modified= true;
    }
    
    return modified;
}

bool tester::add_variables(clause_data & clause) const
{
    bool modified= false;
    if (clause.empty())
        return false;

    set<unsigned> clause_variables;
    for (unsigned i=0 ; i != clause.size() ; ++i)
        clause_variables.insert(clause[i].m_variable);
    
    for (unsigned i=gen(4) ; i > 0 ; --i) {
        unsigned variable= random_var_index();
        if (clause_variables.find(variable) != clause_variables.end())
            continue;
        // At this point, the variable is new to the clause
        clause_variables.insert(variable);
        clause.push_back(literal_data(variable, random_var()));
        if (!clause.back().m_set.empty())
            modified= true;
    }
    
    return modified;
}

void tester::test_diff()
{
    clause_data a=random_clause();
    clause_data b=a;

    bool modified= false;
    if (gen(2) == 0) {
        if (enlarge_sets(a)) {
            T_ASSERT(!a.equivalent(b) && !b.equivalent(a));
            modified= true;
        } else if (enlarge_sets(b)) {
            T_ASSERT(!a.equivalent(b) && !b.equivalent(a));
            modified= true;
        }
    }
    if (add_variables(a)) {
        T_ASSERT(!a.equivalent(b) && !b.equivalent(a));
        modified= true;
    } else if (add_variables(b)) {
        T_ASSERT(!a.equivalent(b) && !b.equivalent(a));
        modified= true;
    }        
    if (!modified)
        T_ASSERT(a.equivalent(b) && b.equivalent(a));
}


void tester::run()
{
    test_equivalent();
    test_diff();
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    solver::reporting_levels report_level= solver::NONE;
    //report_level= solver::SEARCH;
    //report_level= solver::PROPAGATION_MODS;
    //report_level= solver::PROPAGATED_CONSTRAINTS;

    cout << "Seed is " << set_seed(argc, argv) << endl;
    for (unsigned i=0 ; i < 10000 ; ++i) {
        tester tester(report_level, hcsp::random::range(1,8));
        tester.run();
    }
    tester tester(report_level, 0);
    tester.run();
    
    END_TEST_MAIN;
}

