// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <algorithm>
#include <iostream>
#include <clause_propagator.hpp>
#include "propagator_tester.hpp"
#include "test_utilities.hpp"
#include "stateless_clause_propagator.hpp"
#include "trivial_runtime_info.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;
using hcsp::test::propagator_tester;

class tester
{
  public:
    typedef iset::value_type                   value_type;
    typedef constraint_propagator::result_type result_type;
    tester() :
        m_literals( random_literals() ),
        m_final_assignment( random_final_assignment() ),
        m_final_assignment_is_satisfiable( is_satisfied(m_final_assignment) ),
        m_querying(gen(2)),
        m_querying_to_the_end(m_querying && gen(2)),
        m_info( m_literals.size() )
    {
        m_tested_propagator.update_literals()= m_literals;
        m_stateless_propagator.update_literals()= m_literals;
        unsigned superset_percent= range(0, 100);
        unsigned everything_percent= range(0, 100);
        unsigned subset_percent= range(0, 100);
        for (unsigned i=0 ; i!= m_literals.size() ; ++i) {
            if ( superset_percent > gen(100) ) {
                if ( everything_percent > gen(100) )
                    m_info.set_domain(i, move(iset().complement()));
                else
                    m_info.set_domain(
                        i, move(random_var().insert_all_of(m_literals[i])));
            } else if (subset_percent > gen(100)) {
                m_info.set_domain(i, random_var().intersect(m_literals[i]));
            } else {
                m_info.set_domain(i, random_var());
            }
            m_info.set_domain(i, iset(m_final_assignment[i]));
        }
    }
    void run();
  private:
    void comparative_run();
    void run_generic_tester();
    void run_for_a_given_propagator(trivial_runtime_info & info,
                                    constraint_propagator & propagator);
    
    void validate_unit_clause_reduction(trivial_runtime_info & info) const;
    void validate_consistent(trivial_runtime_info & info) const;
    bool must_be_watched(trivial_runtime_info & info) const;
    
    vector<iset> random_literals() const;
    vector<value_type> random_final_assignment() const;
    bool is_satisfied(const vector<value_type> & assignment) const;
    friend ostream & operator<<(ostream & out, const tester &t);
    
    //! Print the current state of the tester to cout.
    //! This function is intended to be called from a debugger only.
    void show_self();
    
    const vector<iset>          m_literals;
    const vector<value_type>    m_final_assignment;
    const bool                  m_final_assignment_is_satisfiable;
    bool                        m_querying;
    bool                        m_querying_to_the_end;
    stateless_clause_propagator m_stateless_propagator;
    clause_propagator           m_tested_propagator;
    trivial_runtime_info        m_info;
}; // class tester

vector<iset> tester::random_literals() const
{
    vector<iset> ret;
    
    if (gen(1000)==0)
        return ret;
    ret.push_back(random_var());
    if (gen(100) == 0)
        return ret;
    generate_n(back_inserter(ret), range(1,7), var_randomizer());
    return ret;
}

//friend
ostream & operator<<(ostream & out, const tester &t)
{
    for (unsigned i=0 ; i != t.m_literals.size() ; ++i) {
        if (i != 0)
            out << " || ";
        out << "{" << t.m_literals[i] << "}:v" << i;
    }
    out << "\n";
    return out;
}

void tester::show_self()
{
    cout << *this;
}


vector<iset::value_type> tester::random_final_assignment() const
{
    vector<value_type> final_assignment_vector;
    size_t num_to_sat= 0;
    if (m_literals.size() >= 1 && gen(2) == 0) {
        num_to_sat=1;
        if (m_literals.size() >= 2 && gen(2) == 0) 
            num_to_sat= range(2, static_cast<int>(m_literals.size()));
    }
    size_t num_everything=0; 
    unsigned num_empty=0; 
    for (unsigned i=0 ; i != m_literals.size() ; ++i) {
        num_everything += m_literals[i].everything();
        num_empty += m_literals[i].empty();
    }
    num_to_sat= min(num_to_sat, m_literals.size()-num_empty);
    num_to_sat= max(num_to_sat, num_everything);
    for (unsigned i=0 ; i != m_literals.size() ; ++i) {
        iset::value_type final_assignment;
        if (m_literals[i].empty()) {
            final_assignment=range(iset::MIN_DOMAIN, iset::MAX_DOMAIN);
            --num_empty;
        } else if (m_literals[i].everything()) {
            final_assignment=range(iset::MIN_DOMAIN, iset::MAX_DOMAIN);
            num_to_sat -= m_literals[i].everything();
            --num_everything;
        } else {
            unsigned left= m_literals.size()-i;
            HCSP_ASSERT( left >= num_everything+num_empty);
            unsigned left_to_choose_from= left-num_everything-num_empty;
            HCSP_ASSERT( left_to_choose_from >= num_to_sat-num_everything);
            if (num_to_sat-num_everything > gen(left_to_choose_from)) {
                final_assignment= m_literals[i].rand();
                --num_to_sat;
            } else
                final_assignment= iset(m_literals[i]).complement().rand();
        }
        final_assignment_vector.push_back(final_assignment);
    }
    return final_assignment_vector;
}

bool tester::is_satisfied(const vector<value_type> & assignment) const
{
    HCSP_ASSERT(assignment.size() == m_literals.size());
    
    for (vector<value_type>::size_type i=0 ; i != assignment.size() ; ++i) {
        if ( m_literals[i].contains(assignment[i]) )
            return true;
    }
    return false;
}

bool tester::must_be_watched(trivial_runtime_info & info) const
{
    unsigned false_literals= 0;
    for (unsigned i=0 ; i!=m_literals.size() ; ++i) {
        if ( !m_literals[i].contains_any_of(info.prev_domain(i)) )
            ++false_literals;
    }
    return m_literals.size()-false_literals <= 1;
}

void tester::run()
{
    run_generic_tester();
    trivial_runtime_info stateless_runtime_info(m_info);
    run_for_a_given_propagator(stateless_runtime_info, m_stateless_propagator);
    T_ASSERT((unsigned)m_tested_propagator.count_params() == m_literals.size());

    unsigned test_type= gen(3);
    if ( test_type <= 1 ) {
        trivial_runtime_info tested_runtime_info(m_info);
        run_for_a_given_propagator(tested_runtime_info, m_tested_propagator);
    }
    if ( test_type >= 1 )
        comparative_run();
}

void tester::run_generic_tester()
{
    if (m_info.arity() == 0)
        return;
    propagator_tester<clause_propagator>(
        m_info.arity(), m_tested_propagator).run();
    propagator_tester<stateless_clause_propagator>(
        m_info.arity(), m_stateless_propagator).run();
}

void tester::comparative_run()
{
    trivial_runtime_info tested_info(m_info, m_querying);
    m_tested_propagator.init(tested_info);

    trivial_runtime_info stateless_info(m_info, m_querying); 
    m_stateless_propagator.init(stateless_info);

    vector<trivial_runtime_info> backtrack_info;
    unsigned num_backtracks= 10;

    while ( true ) {
        if (tested_info.querying_domain() && !m_querying_to_the_end && !gen(4)){
            stateless_info.set_querying_domain(false);
            tested_info.set_querying_domain(false);
        }
        trivial_runtime_info::result_type stateless_result =
            stateless_info.maybe_propagate(m_stateless_propagator);
        trivial_runtime_info::result_type tested_result =
            tested_info.maybe_propagate(m_tested_propagator);

        bool final_state= false;
        switch ( stateless_result ) {
          case trivial_runtime_info::UNSAT:
            T_ASSERT( tested_result == stateless_result );
            final_state= true;
            break;
          case trivial_runtime_info::VALID:
            T_ASSERT( tested_result != trivial_runtime_info::UNSAT );
            if ( tested_result == trivial_runtime_info::VALID )
                T_ASSERT(stateless_info.same_next_state(tested_info));
            final_state= true;
            break;
          case trivial_runtime_info::NOT_WATCHED:
            T_ASSERT(0); // should not happen for the stateless propagator
            break;
          default:
            T_ASSERT( tested_result == trivial_runtime_info::NOT_WATCHED ||
                      tested_result == trivial_runtime_info::CONSISTENT);
            break;
        }
        if ( num_backtracks >0 && !backtrack_info.empty() &&
             (final_state || gen(20/num_backtracks)==0)) {
            unsigned target= gen(backtrack_info.size());
            tested_info.backtrack_to(backtrack_info[target]);
            stateless_info.backtrack_to(backtrack_info[target]);
            backtrack_info.resize(target, stateless_info);
            --num_backtracks;
            final_state= false;
        }
        if (final_state) {
            m_tested_propagator.forget_watches();
            tested_info.forget_watches();
            m_stateless_propagator.forget_watches();
            stateless_info.forget_watches();
            return;
        }
        backtrack_info.push_back(tested_info);
        stateless_info.random_reduce(m_final_assignment);
        tested_info.assign_modifications(stateless_info);
    }
}

void tester::run_for_a_given_propagator(trivial_runtime_info & info,
                                        constraint_propagator & propagator)
{
    propagator.forget_watches();
    info.forget_watches();
    if ( !propagator.init(info) ) {
        T_ASSERT(!m_final_assignment_is_satisfiable);
        return;
    }
    bool done= false;
    while ( !done ) {
        switch ( info.maybe_propagate(propagator) ) {
          case trivial_runtime_info::UNSAT:
            T_ASSERT(!m_final_assignment_is_satisfiable);
            done= true;
            break;
          case trivial_runtime_info::VALID:
            if (!m_final_assignment_is_satisfiable)
                validate_unit_clause_reduction(info);
            done= true;
            break;
          case trivial_runtime_info::NOT_WATCHED:
            T_ASSERT( !must_be_watched(info) );
            if ( info.are_all_singletons() )
                return;
            break;
          default:
            validate_consistent(info);
            for (unsigned i=0 ; i != m_literals.size() ; ++i)
                T_ASSERT( !info.reduced_between_activations(i) );
            break;
        }
        if (!done)
            info.random_reduce(m_final_assignment);
    }
    propagator.forget_watches();
    info.forget_watches();
}

void tester::validate_unit_clause_reduction(trivial_runtime_info & info) const
{
    unsigned count_reduced=0;
    for (unsigned i=0 ; i != m_literals.size() ; ++i) {
        if ( info.reduced_between_activations(i) ) {
            ++count_reduced;
            T_ASSERT(m_literals[i].contains_all_of(info.domain(i)) );
            iset removed= info.prev_domain(i);
            removed.erase_all_of( info.domain(i) );
            T_ASSERT(!m_literals[i].contains_any_of(removed));
        } else {
            T_ASSERT(!m_literals[i].contains_any_of(info.domain(i)));
        }
    }
    T_ASSERT(count_reduced == 1);
    
}

// This makes sure that the current values are merely consistent with the
// literals. This means that the values should not be fully valid.
void tester::validate_consistent(trivial_runtime_info & info) const
{
    unsigned count_overlap=0;
    for (unsigned i=0 ; i != m_literals.size() ; ++i) {
        // The literal should not be determined to be true, yet.
        if (info.any_watched(i))
            T_ASSERT(!m_literals[i].contains_all_of(info.domain(i)));
        if ( m_literals[i].contains_any_of(info.domain(i)) )
            ++count_overlap;
    }
    T_ASSERT(count_overlap >= 2);
}

int main(int argc, char *argv[])
{
    HCSP_PROFILE_CURRENT;
    BEGIN_TEST_MAIN;
    cout << "Seed is " << set_seed(argc, argv) << endl;
    for (unsigned i=0 ; i < 3000 ; ++i) {
        tester t;
        t.run();
    }
    END_TEST_MAIN;
}

