// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <iostream>
#include <algorithm>
#include <equivalence_classes.hpp>
#include "test_utilities.hpp"
#include "explanation_tester.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;

class tester
{
  public:
    tester();
    void run();
  private:
    void unite_by_class();
    static unsigned random_in_vector(const vector<unsigned> & elements)
    { return elements[gen(elements.size())]; }
    static void generate_group_unite(
        vector<pair<unsigned,unsigned> > & unite_calls,
        const vector<unsigned> & elements);

    
    enum { VALUES=1000*1000,         //!< number of elements to test
           FREE_RAND_FRACTION = 100, //!< fraction of free-random unite() calls
           TESTED_PAIRS = VALUES/20  //!< number of pairs tested for correctness
    };
    equivalence_classes m_equiv;
    vector<vector<unsigned> > m_values;
}; // class tester

tester::tester()
{
    vector<unsigned> values(gen(4) == 0 ? VALUES : VALUES/10);
    for (unsigned i=0 ; i != values.size() ; ++i)
        values[i]=i+1;
    randomizer rand;
    random_shuffle(values.begin(), values.end(), rand);
    unsigned number_of_blocks= (random::range(1,4)* random::range(1,4)*
                                random::range(1,4)*
                                random::range(1,4))/8+1;
    // I am using 'values' in two different meanings. One is the trivial
    // meaning, i.e., a set of values. The second is used as a set of unique
    // random values in the range [1,VALUES].
    // The next line uses the second meaning of the two:
    vector<unsigned> block_starts(values.begin(),
                                  values.begin()+number_of_blocks);
    block_starts[0]=0;
    sort(block_starts.begin(), block_starts.end());
    if (block_starts.back() != values.size())
        block_starts.push_back(values.size());
    m_values.resize(block_starts.size()-1);
    for (unsigned i=0 ; i != block_starts.size()-1 ; ++i) {
        m_values[i].insert(m_values[i].begin(), values.begin()+block_starts[i],
                           values.begin()+block_starts[i+1]);
    }
}

void tester::unite_by_class()
{
    enum {FRACTION=10};
    // The idea is to work on all groups of m_values in parallel, uniting all
    // elements in every group without uniting elements across groups. This
    // way at the end each group becomes an equivalents set.
    //
    // In each group we unite every two elements, with no overlaps, such that
    // we get sets of size 2. Then for every other couple of pairs we unite
    // one element of the pair, such that we get sets of 4, then we do it to
    // get sets of 8 and so on until we get a single equivalence class.
    //
    // The algorithm is slightly modified in order to get more interesting
    // behaviors of the algorithm:
    //   - Before starting, in each group we make 1/FRACTION random unite()
    //     calls.
    //   - We randomize the order of all unite() calls
    vector<pair<unsigned,unsigned> > unite_calls;
    unite_calls.reserve(m_values.size()+m_values.size()/FRACTION);

    if (gen(4) == 0) {
        // Prepare 1/FRACTION random unite() calls
        for (unsigned i=0 ; i != m_values.size() ; ++i) {
            for (unsigned j= m_values[i].size()/FRACTION; j !=0 ; --j)
                unite_calls.push_back(make_pair(random_in_vector(m_values[i]),
                                                random_in_vector(m_values[i])));
        }
    }
    for (unsigned i=0 ; i != m_values.size() ; ++i)
        generate_group_unite(unite_calls, m_values[i]);

    if (gen(4) == 0) {
        random_shuffle(unite_calls.begin(), unite_calls.end());
        for (unsigned i=0; i != unite_calls.size(); ++i)
            if (gen(2) == 0)
                swap(unite_calls[i].first, unite_calls[i].second);
    }

    for (unsigned i=0; i != unite_calls.size(); ++i)
        m_equiv.unite(unite_calls[i].first, unite_calls[i].second);
}

//static
void tester::generate_group_unite(vector<pair<unsigned,unsigned> > &unite_calls,
                                  const vector<unsigned> & elements)
{
    for (unsigned skip=1 ; skip < elements.size() ; skip *= 2) {
        for (unsigned chunk=0; chunk+skip < elements.size(); chunk += 2*skip) {
            unsigned begin1= chunk;
            unsigned begin2= begin1+skip;
            unsigned end2=   min<unsigned>(begin2+skip, elements.size());
            unsigned first= random::range(begin1, begin2-1);
            unsigned second= random::range(begin2, end2-1);
            unite_calls.push_back(make_pair(elements[first], elements[second]));
        }
    }
}


void tester::run()
{
    unite_by_class();
    
    for (unsigned i=0; i != TESTED_PAIRS ; ++i) {
        unsigned class1= gen(m_values.size());
        unsigned class2= gen(m_values.size());
        unsigned element1= random_in_vector(m_values[class1]);
        unsigned element2= random_in_vector(m_values[class2]);
        if (class1 == class2)
            T_ASSERT(m_equiv.find(element1) == m_equiv.find(element2));
        else
            T_ASSERT(m_equiv.find(element1) != m_equiv.find(element2));
    }
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;

    cout << "Seed is " << set_seed(argc, argv) << endl;
    for (unsigned i=0 ; i < 100 ; ++i) {
        tester().run();
    }
    
    END_TEST_MAIN;
}

