// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <sstream>
#include <set>
#include <algorithm>
#include <limits>
#include <utilities.hpp>
#include <clause_data.hpp>
#include <formula.hpp>
#include <solver.hpp>
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;

class tester {
  public:
    typedef solver::reporting_levels reporting_levels;
    tester(reporting_levels level) : m_level(level) {}
    void run();

  private:
    //! Add several variables to the solver, returning their IDs
    vector<unsigned> add_variables(solver& the_solver, unsigned count);
    
    //! Add several variables to the solver, returning their shuffled IDs
    vector<unsigned> add_shuffled_variables(solver& the_solver, unsigned count);
    void test_literal();
    void test_literal_conjunction_formula();
    void test_literal_disjunction_formula();
    void test_single_literal();
    void test_clause();
    void test_and_same_var();
    void test_and();
    void test_or1();
    void test_or2();
    void test_or3();
    void test_or4();
    void test_or1_not();
    void test_or2_not();
    void test_or3_not();
    void test_or4_not();
    void longer_clauses_almost_conflict();
    vector<iset> get_nontrivial_sets(const unsigned size) const
    {
        vector<iset> sets;
        sets.reserve(size);
        for (unsigned i=0 ; i != size; ++i) {
            sets.push_back( random_nonempty_var() );
            if ( sets.back().everything() )
                sets.back().erase(sets.back().rand());
        }
        return sets;
    }

    static formula and_formulas(const vector<formula> & formulas);
    static formula formula_equally_evaluate_literals(
        const vector<unsigned> & variables, const vector<iset> & sets);
    
    const reporting_levels m_level;
    static const string ON;
    static const string OFF;
}; // class tester

const string tester::ON("[");
const string tester::OFF("]");

template<>
string to_string(const formula & formula)
{
    ostringstream out;
    formula.print(out, trivial_print_strategy::get());
    return out.str();
}

vector<unsigned> tester::add_variables(solver& the_solver, unsigned count)
{
    vector<unsigned> result(count);
    for (unsigned i=0 ; i != count ; ++i)
        result[i]= the_solver.add_var(iset().complement());
    return result;
}
    
vector<unsigned> tester::add_shuffled_variables(solver& the_solver,
                                                unsigned count)
{
    vector<unsigned> result= add_variables(the_solver, count);
    random_shuffle(result.begin(), result.end());
    return result;
}

void tester::test_literal()
{
    test_literal_conjunction_formula();
    test_literal_disjunction_formula();
    test_single_literal();
}

void tester::test_literal_conjunction_formula()
{
    solver solver;
    solver.report_level(m_level);
    vector<unsigned> vars= add_shuffled_variables(solver, 5);
    
    formula all;
    all = (formula(vars[0], iset(0)) && literal_data(vars[1], iset(1)));
    
    all.constrain_solver(solver);
    T_ASSERT(solver.solve()== solver::SAT);
    T_ASSERT(solver.var_domain(vars[0]).min() == 0);
    T_ASSERT(solver.var_domain(vars[1]).min() == 1);
    T_ASSERT(all.evaluate(solver));
    ostringstream expected;
    expected << ON << "V" << vars[0] << "=0" << OFF
             << "&&" << ON << "V" << vars[1] << "=1" << OFF;
    T_ASSERT(to_string(all) == expected.str());
}

void tester::test_literal_disjunction_formula()
{
    solver solver;
    solver.report_level(m_level);
    vector<unsigned> vars= add_shuffled_variables(solver, 5);
    
    formula first= (formula(vars[0], iset(0)) || formula(vars[1], iset(1)));
    formula second= (formula(vars[0], iset(1)) || formula(vars[1], iset(0)));
    first.constrain_solver(solver);
    second.constrain_solver(solver);
    
    T_ASSERT(solver.solve()== solver::SAT);
    T_ASSERT(first.evaluate(solver) && second.evaluate(solver));
    T_ASSERT((first && second).evaluate(solver));
    T_ASSERT(solver.var_domain(vars[0]).contains_any_of(iset("0,1")));
    ostringstream expected_first;
    expected_first << ON << "V" << vars[0] << "=0, "
                   << "V" << vars[1] << "=1" << OFF;

    T_ASSERT(to_string(first) == expected_first.str());
}

void tester::test_single_literal()
{
    solver solver;
    solver.report_level(m_level);
    vector<unsigned> vars= add_shuffled_variables(solver, 5);
    
    formula all(literal_data(vars[0], iset(5)));
    all.constrain_solver(solver);
    
    T_ASSERT(solver.solve()== solver::SAT);
    T_ASSERT(solver.var_domain(vars[0]) == iset(5));
    T_ASSERT(all.evaluate(solver));

    T_ASSERT(to_string(all) == ON+"V" + to_string(vars[0])+ "=5"+OFF);
    
    all= literal_data(vars[0], iset("1,5"));
    T_ASSERT(to_string(all) == ON+"V" + to_string(vars[0])+ ":{1,5}"+OFF);
    
    all= literal_data(vars[0], iset("1:5"));
    T_ASSERT(to_string(all) == ON+"V" + to_string(vars[0])+ ":{1:5}"+OFF);

    all= literal_data(vars[0], iset("5").complement());
    T_ASSERT(to_string(all) == ON+"V" + to_string(vars[0])+ "!=5"+OFF);

    all= literal_data(vars[0], iset(make_pair(5,iset::MAX_DOMAIN)));
    T_ASSERT(to_string(all) == ON+"V" + to_string(vars[0])+ ">=5"+OFF);

    all= literal_data(vars[0], iset(make_pair(iset::MIN_DOMAIN,5)));
    T_ASSERT(to_string(all) == ON+"V" + to_string(vars[0])+ "<=5"+OFF);
}

void tester::test_clause()
{
    solver solver;
    solver.report_level(m_level);
    vector<unsigned> vars= add_shuffled_variables(solver, 8);

    clause_data clause;
    formula all(clause);
    T_ASSERT(to_string(all) == "false");

    ostringstream expected;
    expected << ON;
    unsigned max_size= 4;
    // This vars[viable_index] is the only variable for which the
    // clause may propagate, all other literals are falsified by unary literals
    unsigned viable_index=gen(max_size);
    while (clause.size() < max_size) {
        unsigned index= clause.size();
        iset::value_type value= index+1;
        iset set(iset::value_range_type(value, iset::MAX_DOMAIN));
        literal_data literal(vars[index], set);
        if (index != 0)
            expected << ", ";
        expected << "V" << vars[index] << ">=" << value;
        clause.push_back(literal);
        all= clause;
        T_ASSERT(to_string(all) == expected.str() + OFF);
        if (index != viable_index)
            formula(literal_data(vars[index],iset(value-1)))
                .constrain_solver(solver);
        
    } 
    all.constrain_solver(solver);
    T_ASSERT(solver.solve() == solver::SAT);
    T_ASSERT(all.evaluate(solver));
    iset::value_type value= solver.var_domain(vars[viable_index]).min();
    T_ASSERT(value > int(viable_index));
}

void tester::test_and_same_var()
{
    solver solver;
    solver.report_level(m_level);
    vector<unsigned> vars= add_shuffled_variables(solver, 8);
    
    formula all= (formula(literal_data(vars[0], iset("1:3"))) &&
                  literal_data(vars[0], iset("3:5")));
    T_ASSERT(to_string(all) == ON + "V" +to_string(vars[0]) + "=3" +OFF);
    all.constrain_solver(solver);
    T_ASSERT(solver.solve() == solver::SAT);
    T_ASSERT(all.evaluate(solver));
    T_ASSERT(solver.var_domain(vars[0]) == iset(3));
}

void tester::test_and()
{
    solver solver;
    solver.report_level(m_level);
    vector<unsigned> vars= add_shuffled_variables(solver, 8);
    
    formula all= ( (formula(literal_data(vars[0], iset(0))) &&
                    literal_data(vars[1], iset(0))) ||
                   (formula(literal_data(vars[0], iset(1))) &&
                    literal_data(vars[1], iset(1))) );
    all.constrain_solver(solver);
    formula(literal_data(vars[0], iset(1))).constrain_solver(solver);
    T_ASSERT(solver.solve() == solver::SAT);
    T_ASSERT(all.evaluate(solver));
    T_ASSERT(solver.var_domain(vars[0]) == iset(1));
 }

void tester::test_or1()
{
    solver solver;
    solver.report_level(m_level);
    vector<unsigned> vars= add_shuffled_variables(solver, 8);

    clause_data clause1, clause2;
    clause1.push_back(literal_data(vars[0], iset(0)));
    clause1.push_back(literal_data(vars[1], iset(0)));

    clause2.push_back(literal_data(vars[0], iset(1)));
    clause2.push_back(literal_data(vars[1], iset(1)));
    
    formula all= formula(clause1) || clause2;
    all.constrain_solver(solver);
    formula(vars[0], iset(2)).constrain_solver(solver);
    formula(vars[1], iset("1:20")).constrain_solver(solver);
    T_ASSERT(solver.solve() == solver::SAT);
    T_ASSERT(all.evaluate(solver));
    T_ASSERT(solver.var_domain(vars[1])== iset(1));
}

void tester::test_or2()
{
    solver solver;
    solver.report_level(m_level);
    vector<unsigned> vars= add_shuffled_variables(solver, 8);

    formula first= formula(vars[0], iset(1)) || formula(vars[1], iset(1));
    formula second= formula(vars[0], iset(2)) || formula(vars[2], iset(1));
    formula all= first || second;
    T_ASSERT(to_string(all) == ON+ "V"+to_string(vars[0]) + ":{1,2}, V" +
             to_string(vars[1]) + "=1, V" + to_string(vars[2]) + "=1" +OFF
             ||
             to_string(all) == ON+ "V"+to_string(vars[1]) + "=1, V"+
             to_string(vars[0]) + ":{1,2}, V"+ to_string(vars[2]) + "=1" +OFF);
             
}

// Test OR operator which is not optimized away and is under an AND operator
// also test an AND between a clause and another formula
void tester::test_or3()
{
    solver solver;
    solver.report_level(m_level);
    vector<unsigned> vars= add_shuffled_variables(solver, 8);

    formula v0_v1_eq_0= formula(vars[0], iset(0)) && formula(vars[1], iset(0));
    formula v0_v1_eq_1= formula(vars[0], iset(1)) && formula(vars[1], iset(1));
    formula v0_v1_eq_2= formula(vars[0], iset(2)) && formula(vars[1], iset(2));
    formula v0_iff_v1= v0_v1_eq_0 || v0_v1_eq_1 || v0_v1_eq_2;

    formula v0_v2_eq_0= formula(vars[0], iset(0)) && formula(vars[2], iset(0));
    formula v0_v2_eq_1= formula(vars[0], iset(1)) && formula(vars[2], iset(1));
    formula v0_v2_eq_2= formula(vars[0], iset(2)) && formula(vars[2], iset(2));
    formula v0_iff_v2= v0_v2_eq_0 || v0_v2_eq_1 || v0_v2_eq_2;

    iset neq0= iset(0).complement();
    iset neq1= iset(1).complement();
    iset neq2= iset(2).complement();
    formula v1_v2_neq_0= formula(vars[1], neq0) || formula(vars[2], neq0);
    formula v1_v2_neq_1= formula(vars[1], neq1) || formula(vars[2], neq1);
    formula v1_v2_neq_2= formula(vars[1], neq2) || formula(vars[2], neq2);
    formula v1_neq_v2= v1_v2_neq_0 && v1_v2_neq_1 && v1_v2_neq_2;

    switch (gen(3)) {
      case 0:
        (v0_iff_v1 && v0_iff_v2).constrain_solver(solver);
        T_ASSERT(solver.solve() == solver::SAT);
        T_ASSERT((v0_iff_v1 && v0_iff_v2).evaluate(solver));
        T_ASSERT(!v1_neq_v2.evaluate(solver));
        T_ASSERT(solver.var_domain(vars[1]) == solver.var_domain(vars[2]));
        break;
      case 1:
        (v0_iff_v1 && v0_iff_v2).constrain_solver(solver);
        v1_neq_v2.constrain_solver(solver);
        T_ASSERT(solver.solve() == solver::UNSAT);
        break;
      default:
        (v0_iff_v1 && v1_neq_v2).constrain_solver(solver);
        T_ASSERT(solver.solve() == solver::SAT);
        T_ASSERT(!v0_iff_v2.evaluate(solver));
        T_ASSERT(solver.var_domain(vars[0]) != solver.var_domain(vars[2]));
    }
}

// test that OR between a clause and a nonclausal formula works well
void tester::test_or4()
{
    solver solver;
    solver.report_level(m_level);
    vector<unsigned> vars= add_shuffled_variables(solver, 8);

    formula v0_v1_eq_0= formula(vars[0], iset(0)) && formula(vars[1], iset(0));
    formula v0_v1_any_1= (formula(vars[0], iset(1)) ||
                          formula(vars[1], iset(1)));
    (v0_v1_eq_0 || v0_v1_any_1).constrain_solver(solver);
    if (gen(2)==0) {
        formula(vars[0], iset(0).complement()).constrain_solver(solver);
        T_ASSERT(solver.solve() == solver::SAT);
        T_ASSERT(solver.var_domain(vars[0]) == iset(1) ||
                 solver.var_domain(vars[1]) == iset(1));
    } else {
        formula(vars[0], iset(1).complement()).constrain_solver(solver);
        formula(vars[1], iset(1).complement()).constrain_solver(solver);
        T_ASSERT(solver.solve() == solver::SAT);
        T_ASSERT(solver.var_domain(vars[0]) == iset(0));
        T_ASSERT(solver.var_domain(vars[1]) == iset(0));
    }
    T_ASSERT( v0_v1_eq_0.evaluate(solver) || v0_v1_any_1.evaluate(solver));
    T_ASSERT( (v0_v1_eq_0|| v0_v1_any_1).evaluate(solver));
}

void tester::test_or1_not()
{
    solver solver;
    solver.report_level(m_level);
    vector<unsigned> vars= add_shuffled_variables(solver, 8);

    clause_data clause1, clause2;
    clause1.push_back(literal_data(vars[0], iset(0)));
    clause1.push_back(literal_data(vars[1], iset(0)));

    clause2.push_back(literal_data(vars[0], iset(1)));
    clause2.push_back(literal_data(vars[1], iset(1)));
    
    formula all= !(!formula(clause1) && (!formula(clause2)));
    all.constrain_solver(solver);
    formula(vars[0], iset(2)).constrain_solver(solver);
    formula(vars[1], iset("1:20")).constrain_solver(solver);
    T_ASSERT(solver.solve() == solver::SAT);
    T_ASSERT(solver.var_domain(vars[1])== iset(1));
}

void tester::test_or2_not()
{
    solver solver;
    solver.report_level(m_level);
    vector<unsigned> vars= add_shuffled_variables(solver, 8);

    formula first= formula(vars[0], iset(1)) || formula(vars[1], iset(1));
    formula second= formula(vars[0], iset(2)) || formula(vars[2], iset(1));
    formula all;

    if (gen(2)== 0)
        all= first || second;
    else
        all= !(!first && !second);
    formula negated= !all;
    if (gen(2) == 0)
        all= !negated;
    if (gen(2) == 0)
        swap(negated, all);
    all.constrain_solver(solver);
    if (gen(2) == 0) {
        negated.constrain_solver(solver);
        T_ASSERT(solver.solve() == solver::UNSAT);
    } else {
        T_ASSERT(solver.solve() == solver::SAT);
        T_ASSERT(!negated.evaluate(solver));
    }
}

formula make_3eq(unsigned a, unsigned b)
{
    formula a_eq_b_eq_0= formula(a, iset(0)) && formula(b, iset(0));
    formula a_eq_b_eq_1= formula(a, iset(1)) && formula(b, iset(1));
    formula a_eq_b_eq_2= formula(a, iset(2)) && formula(b, iset(2));
    if (gen(2) == 0)
        return a_eq_b_eq_1 || a_eq_b_eq_2 || a_eq_b_eq_2;
    else
        return !(!a_eq_b_eq_1 && !a_eq_b_eq_2 && !a_eq_b_eq_2);
}


// Test OR operator which is not optimized away and is under an AND operator
// also test an AND between a clause and another formula
void tester::test_or3_not()
{
    solver solver;
    solver.report_level(m_level);
    vector<unsigned> vars= add_shuffled_variables(solver, 8);

    formula v0_eq_v1= make_3eq(vars[0], vars[1]);
    formula v0_eq_v2= make_3eq(vars[0], vars[2]);
    formula v1_eq_v2= make_3eq(vars[1], vars[2]);

    switch (gen(3)) {
      case 0:
        (v0_eq_v1 && v0_eq_v2).constrain_solver(solver);
        T_ASSERT(solver.solve() == solver::SAT);
        T_ASSERT((v0_eq_v1 && v0_eq_v2).evaluate(solver));
        T_ASSERT(v1_eq_v2.evaluate(solver));
        T_ASSERT(solver.var_domain(vars[1]) == solver.var_domain(vars[2]));
        break;
      case 1:
        (v0_eq_v1 && v0_eq_v2).constrain_solver(solver);
        (!v1_eq_v2).constrain_solver(solver);
        T_ASSERT(solver.solve() == solver::UNSAT);
        break;
      default:
        (v0_eq_v1 && !v1_eq_v2).constrain_solver(solver);
        T_ASSERT(solver.solve() == solver::SAT);
        T_ASSERT(!v0_eq_v2.evaluate(solver));
        T_ASSERT(solver.var_domain(vars[0]) != solver.var_domain(vars[2]));
    }
}

// test that OR between a clause and a nonclausal formula works well
void tester::test_or4_not()
{
    solver solver;
    solver.report_level(m_level);
    vector<unsigned> vars= add_shuffled_variables(solver, 8);

    formula v0_v1_eq_0= formula(vars[0], iset(0)) && formula(vars[1], iset(0));
    formula v0_v1_any_1= (formula(vars[0], iset(1)) ||
                          formula(vars[1], iset(1)));
    (v0_v1_eq_0 || v0_v1_any_1).constrain_solver(solver);
    if (gen(2)==0) {
        (!formula(vars[0], iset(0))).constrain_solver(solver);
        T_ASSERT(solver.solve() == solver::SAT);
        T_ASSERT(solver.var_domain(vars[0]) == iset(1) ||
                 solver.var_domain(vars[1]) == iset(1));
    } else {
        (!formula(vars[0], iset(1))).constrain_solver(solver);
        (!formula(vars[1], iset(1))).constrain_solver(solver);
        T_ASSERT(solver.solve() == solver::SAT);
        T_ASSERT(solver.var_domain(vars[0]) == iset(0));
        T_ASSERT(solver.var_domain(vars[1]) == iset(0));
    }
}

// static
formula tester::and_formulas(const vector<formula> & formulas)
{
    formula result(true);
    for (unsigned i=0 ; i < formulas.size() ; ++i)
        result= result && formulas[i];
    return result;
}

// static
formula tester::formula_equally_evaluate_literals(
    const vector<unsigned> & variables, const vector<iset> & sets)
{
    vector<formula> formulas;
    for (unsigned i=0 ; i != variables.size(); ++i) {
        unsigned j= (i+1)%variables.size();
        formulas.push_back(formula(variables[i], sets[i]) ||
                           formula(variables[j], iset(sets[j]).complement()));
    }

    return and_formulas(formulas);
}

/**
   Create one set per variable and use them to create two literals in each
   clause. The first literal in the clause is true if the variable takes a
   value from the set and false otherwise. The second literal in the clause
   looks at the following variable and requires the opposite. It requires that
   the value will not be in the set.

   This problem forces all variables to get all their values from either
   form the set or from outside the set simultaneously. This means that if
   a variable is assigned with a value from its set then all other variables
   must also be assigned from this set.
*/

void tester::longer_clauses_almost_conflict()
{
    const unsigned size= hcsp::random::range(2,10);
    
    solver solver;
    solver.report_level(m_level);
    vector<unsigned> variables= add_variables(solver, size);
    
    
    vector<iset> sets= get_nontrivial_sets(size);
    formula formulas= formula_equally_evaluate_literals(variables, sets);

    formulas.constrain_solver(solver);
    solver::result_type result= solver.solve();
    T_ASSERT(result == solver::SAT);
    T_ASSERT(formulas.evaluate(solver));

    bool is_in_set= sets[0].contains_any_of(solver.var_domain(variables[0]));
    for (unsigned i=0 ; i<variables.size() ; ++i) {
        T_ASSERT(sets[i].contains_any_of(solver.var_domain(variables[i]))
                 == is_in_set);
    }
}

void tester::run()
{
    test_literal();
    test_clause();
    test_and_same_var();
    test_and();
    test_or1();
    test_or2();
    test_or3();
    test_or4();
    test_or1_not();
    test_or2_not();
    test_or3_not();
    test_or4_not();
    longer_clauses_almost_conflict();
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    solver::reporting_levels report_level= solver::NONE;
    //report_level= solver::SEARCH;
    //report_level= solver::PROPAGATION_MODS;
    //report_level= solver::PROPAGATED_CONSTRAINTS;

    cout << "Seed is " << set_seed(argc, argv) << endl;
    tester tester(report_level);
    for (unsigned i=0 ; i != 1000 ; ++i)
        tester.run();
    
    END_TEST_MAIN;
}

