// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2012,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2012,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <iostream>
#include <sstream>
#include <fzn_solver.hpp>
#include <flags.hpp>
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;

struct expression_type
{
    string str;
    solver::result_type expected;
    int max_backtracks; // Negative means 'default'
};

enum { DEFAULT=100 };

expression_type expressions[]= {
    {"var int : tmp :: var_is_introduced :: is_defined_var;\n"
     "var 5..10 : a, b;\n"
     "var 5..10 : c = tmp;\n"
     "constraint int_plus(a,b,c) :: defines_var(tmp);\n",
     solver::SAT, 1},
    {"var int : tmp :: var_is_introduced :: is_defined_var;\n"
     "var 0..5 : a = tmp;\n"
     "constraint int_le(10, tmp) :: defines_var(tmp);\n",
     solver::UNSAT, 0},
    {"var int :a;\n"
     "array [1..3] of var 1..10 : x;\n"
     "constraint array_int_element(a, x, x[3]);\n"
     "constraint int_lt(x[1], x[2]); constraint int_lt(x[2], x[3]);",
     solver::SAT, 3},
    {"var int :a;\n"
     "array [1..3] of var 1..10 : x;\n"
     "constraint array_int_element(a, x, x[3]); constraint int_lt(a, 3);\n"
     "constraint int_lt(x[1], x[2]); constraint int_lt(x[2], x[3]);",
     solver::UNSAT, 0},
    {"array [1..2] of set of int: categories = [{16, 30, 31}, {}, 1..5];\n"
     "solve satisfy; %SAT", solver::SAT, 0},
    {"array [1..3] of int : table = [0,1,2];\n"
     "array [1..3] of var int : x :: output_array([1..3]);\n"
     "constraint hcsp_table_int(x, table); %SAT",
     solver::SAT, 0},
    {"array [1..3] of int : table = [0,1,2];\n"
     "array [1..3] of var int : x :: output_array([1..3]);\n"
     "constraint hcsp_table_int(x, [0,1,3]);\n"
     "constraint hcsp_table_int(x, table); %SAT",
     solver::UNSAT, 0},
    {"array [1..3] of var int : x :: output_array([1..3]);\n"
     "constraint hcsp_table_int(x, [0,1,2,    0,1,3,   1,1,2,   0,0,3]);\n",
     solver::SAT, 0},
    {"array [1..3] of var int : x :: output_array([1..3]);\n"
     "constraint hcsp_table_int(x, [0,1,2,    0,1,3,   1,1,2,   0,0,3]);\n"
     "constraint int_eq(x[1], x[2]);",
     solver::SAT, 0},
    {"array [1..3] of var int : x :: output_array([1..3]);\n"
     "constraint hcsp_table_int(x, [0,1,2,    0,1,3,   1,1,2,   0,0,3]);\n"
     "constraint int_lt(x[2], x[1]);",
     solver::UNSAT, 0},
    {"var bool : a,b,c,d;\n"
     "constraint bool_clause([a,b],[c,d]);\n"
     "constraint bool_clause([c,d],[a,b]);\n"
     "constraint bool_clause([],[a]);\n"
     "constraint bool_clause([],[b]);\n"
     "constraint bool_clause([c],[]);\n"
     "constraint bool_clause([d],[]);\n",
     solver::UNSAT, 0},
    {"var bool : a,b,c,d;\n"
     "constraint bool_clause([a,b],[c,d]);\n"
     "constraint bool_clause([c,d],[a,b]);\n"
     "constraint bool_clause([],[a]);\n"
     "constraint bool_clause([],[b]);\n"
     "constraint bool_clause([c],[]);\n",
     solver::SAT, 0},
    {"var bool : a; constraint bool_clause([],[]);", solver::UNSAT, 0},
    {"var bool : a,b;\n"
     "constraint bool_xor(a,b, true); constraint bool_and(a,b,true); %UNSAT",
     solver::UNSAT, 0},
    {"var bool : a,b,c; constraint bool_lt_reif(a,b, c);\n"
     "constraint bool_lt_reif(b,a,c); %c=false",
     solver::SAT, 1},
    {"var bool : a,b,c; constraint bool_lt_reif(a,b, true);\n"
     "constraint bool_lt_reif(b,a,true); %UNSAT\n",
     solver::UNSAT, 1},
    {"var bool : a,b,c; constraint bool_lt_reif(a,b, c);\n"
     "constraint bool_lt_reif(b,a,c); constraint bool_lt_reif(c,a,b);\n"
     "%c=false, a=b\n",
     solver::SAT, 1},
    {"var bool : a,b,c; constraint bool_lt_reif(a,b, c);\n"
     "constraint bool_lt_reif(b,a,c); constraint bool_lt_reif(c,a,b);\n"
     "constraint bool_not(a,b); %UNSAT\n",
     solver::UNSAT, 1},
    {"var bool : a,b,c,d; constraint bool_le_reif(a,b, c);\n"
     "constraint bool_lt_reif(b,a,d); constraint bool_not(c,d);\n",
     solver::SAT, 1},
    {"var bool : a,b,c,d; constraint bool_le_reif(a,b, c);\n"
     "constraint bool_lt_reif(b,a,d); constraint bool_eq(c,d);\n",
     solver::UNSAT, 1},
    {"var bool : a,b;\n"
     "constraint bool_xor(a,true, b); constraint bool_and(a,b,true); %UNSAT",
     solver::UNSAT, 0},
    {"var bool : a,b;\n"
     "constraint bool_xor(a,b, true); constraint bool_or(a,b,false); %UNSAT",
     solver::UNSAT, 0},
    {"var bool : a,b;\n"
     "constraint bool_xor(a,b, true); constraint bool_and(a,b,false); %SAT",
     solver::SAT, 0},
    {"var bool : a,b;\n"
     "constraint bool_xor(true,a, b); constraint bool_and(a,b,false); %SAT",
     solver::SAT, 0},
    {"var bool : a,b;\n"
     "constraint bool_xor(a,b, true); constraint bool_or(a,b,true); %SAT",
     solver::SAT, 0},
    {"var bool : a,b;\n"
     "constraint bool_xor(a,b, false); constraint bool_and(a,b,true); %SAT",
     solver::SAT, 0},
    {"var bool : a,b;\n"
     "constraint bool_xor(false,a, b); constraint bool_and(a,b,true); %SAT",
     solver::SAT, 0},
    {"var bool : a,b;\n"
     "constraint bool_xor(a,b, false); constraint bool_not(a,b); %UNSAT",
     solver::UNSAT, 0},
    {"var 1..3 : a,b,c,d; constraint all_different_int([a,b,c,d]); %UNSAT",
     solver::UNSAT, 0},
    {"var 1..4 : a,b,c,d; constraint all_different_int([a,b,c,d]); %SAT",
     solver::SAT, 0},
    {"var 1..4 : a,b,c,d; constraint all_different_int([1,b,2,d]); %SAT",
     solver::SAT, 0},
    {"var 1..4 : a,b,c,d; constraint all_different_int([1,b,1,d]); %SAT",
     solver::UNSAT, 0},
    {"var 1..3 : a,b,c,d; array [1..4] of var int : x= [a,b,c,d];\n"
     "constraint all_different_int(x); %UNSAT",
     solver::UNSAT, 0},
    {"var 1..4 : a,b,c,d; array [1..4] of var int : x= [a,b,c,d];\n"
     "constraint all_different_int(x); %SAT",
     solver::SAT, 0},
    {"var 1..4 : a,b,c,d; array [1..4] of var int : x= [a,b,c,d];\n"
     "constraint all_different_int(x); constraint int_eq(a,x[3]);%UNSAT",
     solver::UNSAT, 4},
    {"var 1..4 : a,b,c,d; array [1..4] of var int : x= [1,b,2,d];\n"
     "constraint all_different_int(x); %SAT",
     solver::SAT, 0},
    {"var 1..4 : a,b,c,d; array [1..4] of var int : x= [1,b,2,d];\n"
     "constraint all_different_int(x); constraint int_eq(x[2],x[4]);%UNSAT",
     solver::UNSAT, 4},
    {"var 1..4 : a,b,c,d; array [1..4] of var int : x= [1,b,1,d];\n"
     "constraint all_different_int(x); %UNSAT",
     solver::UNSAT, 0},
    {"predicate median_of_3(var int: x, var int: y, var int: z, var int: m);\n"
     "var bool : a1,a2,a3;\n"
     "array [1..3] of var bool: a= [a1,a2,a3]; \n"
     "array [1..3] of var bool : tmp :: is_defined_var :: var_is_introduced;\n"
     "constraint array_bool_or([a[1],a2], tmp[1]) :: defines_var(tmp[1]);\n"
     "constraint array_bool_or([a1,a[2],a[3]], tmp[2]) :: defines_var(tmp[2]);\n"
     "constraint bool_lt(tmp[2],tmp[1]); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var bool : tmp1, a1,a3;\n"
     "var bool : a2, tmp2 :: is_defined_var :: var_is_introduced;\n"
     "array [1..3] of var bool: a= [a1,a2,a3]; \n"
     "constraint array_bool_or([a[1],a2], tmp1) :: defines_var(a2);\n"
     "constraint array_bool_or([a1,a[2],a[3]], tmp2) :: defines_var(tmp2);\n"
     "constraint bool_lt(tmp2,tmp1); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var bool : a2,a3,a1;\n"
     "array [1..3] of var bool : a ::output_array([1..3]) = [a1, a2, a3]; \n"
     "array [1..3] of var bool : x :: is_defined_var :: var_is_introduced;\n"
     "constraint bool_or(a1,a[2], x[1]) :: defines_var(x[1]);\n"
     "constraint bool_or(a2,a[3], x[2]) :: defines_var(x[2]);\n"
     "constraint bool_not(x[2], x[3]) :: defines_var(x[3]);\n"
     "constraint bool_and(x[1],x[3], true); % a={1, 0,0}\n",
     solver::SAT,   DEFAULT},
    {"predicate all_different(array [int] of var int: xs);\n"
     "predicate xcsp_cumulative(array [1..5] of var int: s, "
         "array [int] of var int: d, array [int] of var int: r, var int: b);\n"
     "var bool : a; var bool : b; var bool : c; \n"
     "var bool : a_or_b, b_or_c :: is_defined_var :: var_is_introduced;\n"
     "var bool : not_b_or_c :: is_defined_var :: var_is_introduced;\n"
     "constraint bool_and(a_or_b, not_b_or_c, true);\n"
     "constraint bool_or(a, b, a_or_b) :: defines_var(a_or_b);\n"
     "constraint bool_or(b, c, b_or_c) :: defines_var(b_or_c);\n"
     "constraint bool_not(b_or_c, not_b_or_c) :: defines_var(not_b_or_c);\n"
     "constraint bool_not(a,true); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"predicate inverse_offsets(array [int] of var int: f, int: foff, "
         "array [int] of var int: invf, int: invfoff);\n"
     "var bool : a; var bool : b; var bool : c; \n"
     "var bool : a_or_b, not_a_or_b :: is_defined_var :: var_is_introduced;\n"
     "constraint bool_not(a_or_b, not_a_or_b) :: defines_var(not_a_or_b);\n"
     "constraint bool_or(a, b, a_or_b) :: defines_var(a_or_b);\n"
     "constraint bool_and(a_or_b, not_a_or_b, true);\n"
     "constraint bool_eq(b, true); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var bool : a; var bool : b; var bool : c;\n"
     "array [1..2] of var bool : tmp :: is_defined_var :: var_is_introduced;\n"
     "constraint bool_or(a,b,tmp[1]) :: defines_var(tmp[1]);\n"
     "constraint bool_or(b,c,tmp[2]) :: defines_var(tmp[2]);\n"
     "constraint int_ne(c,0); constraint bool_lt(tmp[2],tmp[1]); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"array [1..2] of var bool : a; var bool : c;\n"
     "constraint int_lin_eq([1,3,11],[a[1],a[2],c],14); %result a={0,1}, c=1\n",
     solver::SAT,   DEFAULT},
    {"var bool : a; array [1..2] of var bool : b; \n"
     "constraint int_lin_eq([1,3,11],[a,b[1],b[2]],16); %UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var bool : a; var bool : b;\n"
     "constraint array_bool_or([a,b,a], true); % a=1 or b=1\n",
     solver::SAT,   DEFAULT},
    {"var bool : a; var bool : b; constraint array_bool_or([a,b,a], true);\n"
     "constraint bool_eq(b, false); % a=1, b=0\n",
     solver::SAT,   DEFAULT},
    {"var bool : a; var bool : b; constraint array_bool_or([a,b,a], true);\n"
     "constraint bool_not(a,true); % a=0, b=1\n",
     solver::SAT,   DEFAULT},
    {"var bool : a; var bool : b; constraint array_bool_or([a,b,a], true);\n"
     "constraint bool_eq(a, false); constraint bool_not(b,true); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var bool : a; var bool : b; constraint bool_eq(b,false);\n"
     "constraint bool_eq(a,false) ; constraint array_bool_or([a,b,a], true);\n"
     " % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var bool : a; var bool : b; var bool : c; var int : d;\n"
     "var bool : tmp :: is_defined_var :: var_is_introduced;\n"
     "constraint int_lt(a,d);\n"
     "constraint array_bool_or([a,b,c], tmp) :: defines_var(tmp);\n"
     "constraint int_le(d, tmp); % a=0, (b=1 or c=1), d=1\n",
     solver::SAT,   DEFAULT},
    {"var bool : a; var bool : b; var bool : c; var int : d;\n"
     "var bool : tmp :: is_defined_var :: var_is_introduced;\n"
     "constraint array_bool_or([a,b,c], tmp) :: defines_var(tmp);\n"
     "constraint int_lt(a,d); constraint int_lt(b,d);\n"
     "constraint int_le(d, tmp); % a=0, b=0,c=1, d=1\n",
     solver::SAT,   DEFAULT},
    {"var bool : a; var bool : b; var bool : c; var int : d;\n"
     "var bool : tmp :: is_defined_var :: var_is_introduced;\n"
     "constraint array_bool_or([a,b,c], tmp) :: defines_var(tmp);\n"
     "constraint int_lt(a,d); constraint int_lt(b,d); constraint int_lt(c,d);\n"
     "constraint int_le(d, tmp); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var bool : a,b,c; var int : d;\n"
     "var bool : tmp :: is_defined_var :: var_is_introduced;\n"
     "constraint int_lt(d,a);\n"
     "constraint array_bool_or([a,b,c], tmp) :: defines_var(tmp);\n"
     "constraint int_le(tmp, d);  % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var bool : a; var bool : b; var bool : c; var int : d;\n"
     "var bool : tmp :: is_defined_var :: var_is_introduced;\n"
     "var int : a_plus_b ::is_defined_var :: var_is_introduced;\n"
     "constraint array_bool_or([a,b,c], tmp) :: defines_var(tmp);\n"
     "constraint int_lt(c,d);\n"
     "constraint int_le(tmp, d); constraint int_plus(a, b, a_plus_b);\n"
     "constraint int_lt(d, a_plus_b);% a=1,b=1,c=0,d=1\n",
     solver::SAT,   DEFAULT},
    {"var int:a,b,c; constraint int_eq(a,b);constraint int_eq(a,c);\n"
     "constraint int_ne(b,c); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var int :a,b,c; constraint int_eq(a,c); constraint int_ne(c,b);",
     solver::SAT,   DEFAULT},
    {"var int: a,b,c; constraint int_eq(a,b);constraint int_eq(a,c);",
     solver::SAT,   DEFAULT},
    {"var int: a,b,c; constraint bool_eq_reif(a,b,true);\n"
     "constraint bool_eq_reif(a,b,c); %c=1, a==b\n",
     solver::SAT, 0},
    {"var int: a,b,c; constraint bool_eq_reif(a,b,true);\n"
     "constraint bool_eq_reif(a,b,c); constraint bool_not(c, true); %UNSAT\n",
     solver::UNSAT, 1},
    {"var int: a,b,c; constraint bool_eq_reif(a,b,false);\n"
     "constraint int_eq_reif(a,b,c); %c=0, a!=b\n",
     solver::SAT, 0},
    {"var int: a,b,c; constraint bool_eq_reif(a,b,false);\n"
     "constraint int_eq_reif(a,b,c); constraint bool_not(c,0); %UNSAT\n",
     solver::UNSAT, 1},
    {"var int: a,b,c; constraint bool_eq_reif(a,b,true);\n"
     "constraint int_ne_reif(a,b,c); %c=0, a==b\n",
     solver::SAT, 0},
    {"var int: a,b,c; constraint bool_eq_reif(a,b,true);\n"
     "constraint int_ne_reif(a,b,c); constraint bool_not(c,false); %UNSAT\n",
     solver::UNSAT, 1},
    {"var int : a; var int : b; var int : c;\n"
     "constraint int_eq(a,1); constraint int_eq(b,1);\n"
     "constraint int_ne_reif(a,b,c); % a=1,b=1,c=0\n",
     solver::SAT,   DEFAULT},
    {"var int : a; var int : b; var int : c;\n"
     "constraint int_eq(a,1); constraint int_eq(b,1);\n"
     "constraint int_ne_reif(a,b,c); constraint bool_not(c,false); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var int : a; var int : b; var int : c;\n"
     "constraint int_eq(a,1); constraint int_eq(b,1);\n"
     "constraint int_eq_reif(a,b,c); % a=1,b=1,c=1\n",
     solver::SAT,   DEFAULT},
    {"var int : a; var int : b; var int : c;\n"
     "constraint int_eq(a,1); constraint int_eq(b,1);\n"
     "constraint int_eq_reif(a,b,c); constraint bool_not(a,true); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var bool : V0,V1,V2; \n"
     "var bool : V2_ne_V1 :: is_defined_var ::var_is_introduced;\n"
     "constraint int_eq(V0,0); constraint bool_xor(V2, V1, V2_ne_V1);\n"
     "constraint bool_or(V2_ne_V1, V0, true); % V0=0, V1 == not V0\n",
     solver::SAT,   DEFAULT},
    {"var int : a; var int : b;\n"
     "var int : min_a_b :: var_is_introduced :: is_defined_var;\n"
     "constraint int_min(a,b,min_a_b) :: defines_var(min_a_b);\n"
     "constraint int_le(a,min_a_b); % SAT\n", solver::SAT, DEFAULT},
    {"var int : a; var int : b;\n"
     "var int : min_a_b :: var_is_introduced :: is_defined_var;\n"
     "constraint int_min(a,b,min_a_b) :: defines_var(min_a_b);\n"
     "constraint int_lt(min_a_b,a);  % SAT\n", solver::SAT, DEFAULT  },
    {"var int : a; var int : b; constraint int_eq(b,5);\n"
     "constraint int_min(a,b,2); constraint int_ne(a,2);  % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var int: a; var int: b; constraint int_eq(b,5); constraint int_eq(a,2);\n"
     "constraint int_min(a,b,2);  % SAT\n", solver::SAT, DEFAULT},
    {"var int : a; var int : b;\n"
     "var int : max_a_b :: var_is_introduced :: is_defined_var;\n"
     "constraint int_max(a,b,max_a_b) :: defines_var(max_a_b);\n"
     "constraint int_le(a,max_a_b); % SAT\n", solver::SAT, DEFAULT},
    {"var int : a; var int : b;\n"
     "var int : max_a_b :: var_is_introduced :: is_defined_var;\n"
     "constraint int_max(a,b,max_a_b) :: defines_var(max_a_b);\n"
     "constraint int_lt(a,max_a_b);  % SAT\n", solver::SAT, DEFAULT},
    {"var int : a; var int : b; constraint int_eq(b,2);\n"
     "constraint int_max(a,b,5); constraint int_ne(a,5);  % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var int: a; var int: b; constraint int_eq(b,2);constraint int_eq(a,5);\n"
     "constraint int_max(a,b,5);  % SAT\n",
     solver::SAT, DEFAULT},
    {"var int : a; var int : b; var int : c; constraint int_mod(b,c,a);\n"
     "constraint int_times(5,c,b); % SAT\n",
     solver::SAT, DEFAULT},
    {"var int : a; var int : b; \n"
     "var int : abs_a :: var_is_introduced:: is_defined_var;\n"
     "constraint int_mod(b,5,a);\n"
     "constraint int_abs(a,abs_a) :: defines_var(abs_a);\n"
     "constraint int_le(5,a); constraint int_le(abs_a, 10); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var int : a,b; constraint int_eq(a,4); constraint int_mod(b,5,a); % SAT\n",
     solver::SAT, DEFAULT},
    {"var int : a,b; constraint int_eq(a,4); constraint int_eq(b,9);\n"
     "constraint int_mod(b,5,a);% SAT\n", solver::SAT, DEFAULT},
    {"var int : a,b,c;\n"
     "var int : a_eq_b_plus_1, b_plus_1:: var_is_introduced:: is_defined_var;\n"
     "var bool : a_eq_b :: var_is_introduced:: is_defined_var;\n"
     "var int : int_a_eq_b :: var_is_introduced:: is_defined_var;\n"
     "constraint array_int_element(a_eq_b_plus_1 , [b_plus_1, c], a);\n"
     "constraint int_plus(b,1, b_plus_1) :: defines_var(b_plus_1);\n"
     "constraint int_eq_reif(a,b, a_eq_b) :: defines_var(a_eq_b);\n"
     "constraint bool2int(a_eq_b, int_a_eq_b) :: defines_var(int_a_eq_b);\n"
     "constraint int_plus(int_a_eq_b, 1, a_eq_b_plus_1)"
                                           ":: defines_var(a_eq_b_plus_1);\n"
     "constraint int_lt(b,a); % SAT\n",
     solver::SAT, DEFAULT},
    {"var 0..31 : a,b,c;\n"
     "var bool : a_eq_b  ::var_is_introduced ::is_defined_var;\n"
     "var int : a_eq_b_plus_1, b_plus_1 ::var_is_introduced ::is_defined_var;\n"
     "var int : int_a_eq_b  ::var_is_introduced ::is_defined_var;\n"
     "var int : abs_b  ::var_is_introduced ::is_defined_var;\n"
     "constraint array_int_element(a_eq_b_plus_1 , [b_plus_1, c], a);\n"
     "constraint int_lt(a,b);\n"
     "constraint int_plus(b,1, b_plus_1) :: defines_var(b_plus_1);\n"
     "constraint int_plus(int_a_eq_b,1,a_eq_b_plus_1) \n"
     "                     :: defines_var(a_eq_b_plus_1);\n"
     "constraint int_le(abs_b,9);\n"
     "constraint int_eq_reif(a,b,a_eq_b) :: defines_var(a_eq_b);\n"
     "constraint bool2int(a_eq_b, int_a_eq_b) :: defines_var(int_a_eq_b);\n"
     "constraint int_abs(b,abs_b) :: defines_var(abs_b); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var int : a,b,c;\n"
     "var int : a_eq_b_plus_1, int_a_eq_b, b_plus_1, abs_a\n"
     "                               :: var_is_introduced :: is_defined_var;\n"
     "var bool : a_eq_b  :: var_is_introduced :: is_defined_var;\n"
     "constraint array_int_element(a_eq_b_plus_1 , [b_plus_1, c], a);\n"
     "constraint int_eq(a,b);\n"
     "constraint int_plus(b,1, b_plus_1) :: defines_var(b_plus_1);\n"
     "constraint int_eq_reif(a, b, a_eq_b) :: defines_var(a_eq_b);\n"
     "constraint bool2int(a_eq_b, int_a_eq_b) ::defines_var(int_a_eq_b);\n"
     "constraint int_plus(int_a_eq_b, 1, a_eq_b_plus_1)\n"
     "                            :: defines_var(a_eq_b_plus_1);\n"
     "constraint int_ne(a,c);\n"
     "constraint int_abs(a, abs_a) :: defines_var(abs_a);\n"
     "constraint int_le(abs_a, 9); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var int : a,b,c;\n"
     "var int: a_eq_b_plus_1, b_plus_1 ::var_is_introduced ::is_defined_var;\n"
     "var bool : a_eq_b :: var_is_introduced :: is_defined_var;\n"
     "var int : int_a_eq_b :: var_is_introduced :: is_defined_var;\n"
     "var int : abs_b :: var_is_introduced :: is_defined_var;\n"
     "constraint array_int_element(a_eq_b_plus_1 , [b_plus_1, c], a);\n"
     "constraint int_plus(b,1, b_plus_1) :: defines_var(b_plus_1);\n"
     "constraint int_eq_reif(a,b,a_eq_b) :: defines_var(a_eq_b);\n"
     "constraint bool2int(a_eq_b, int_a_eq_b) :: defines_var(int_a_eq_b);\n"
     "constraint int_plus(int_a_eq_b,1,a_eq_b_plus_1)\n"
     "           ::defines_var(a_eq_b_plus_1);\n"
     "constraint int_abs(b, abs_b) :: defines_var(abs_b);\n"
     "constraint int_le(abs_b,9); % SAT\n",
     solver::SAT, DEFAULT},
    {"var int : a; constraint array_int_element(2 , [5, 7], a);\n"
     "constraint int_ne(a,7); %UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var int : a; constraint array_int_element(2 , [5, 7], a);\n"
     "constraint int_eq(a,7); %SAT\n",
     solver::SAT, DEFAULT},
    {"var int: a; constraint array_int_element(2, [1,2,3], a);\n"
     "constraint int_eq(a,2); %SAT\n",
     solver::SAT, DEFAULT},
    {"var int: a; constraint array_int_element(3, [1,2,3], a);\n"
     "constraint int_eq(a,3); %SAT\n",
     solver::SAT, DEFAULT},
    {"var int: a; constraint array_int_element(4, [1,2,3,4], a);\n"
     "constraint int_eq(a,4); %SAT\n",
     solver::SAT, DEFAULT},
    {"var int: a; constraint array_int_element(3, [1,2,3,4], a);\n"
     "constraint int_eq(a,3); %SAT\n",
     solver::SAT, DEFAULT},
    {"var int: a; constraint array_int_element(3, [1,2,3,4], a);\n"
     "constraint int_ne(a,3); %SAT\n",
     solver::UNSAT, DEFAULT},
    {"var {1,3} : a; var {1,3}: b;\n"
     "constraint int_ne(a,1); constraint int_ne(b,3); % SAT\n",
     solver::SAT, DEFAULT},
    {"var {1,3} : a; var {1,3}: b ; constraint int_eq(a,b);\n"
     "constraint int_ne(a,1); constraint int_ne(b,3); % UNSAT",
     solver::UNSAT, DEFAULT},
    {"var {1,3} : a; var {2,4}: b; constraint int_ne(a,1);\n"
     "constraint int_ne(b,4); % SAT\n", solver::SAT, DEFAULT},
    {"var {1,3} : a; var {2,4}: b;\n"
     "constraint int_lt(a,b);\n"
     "constraint int_ne(a,1); constraint int_ne(b,4); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var {1:5:11} : a; constraint int_eq(a,a); % SAT\n", solver::SAT, 0},
    {"var int:a,b; var int:c;\n"
     "var int : abs_c :: var_is_introduced :: is_defined_var;\n"
     "constraint int_mod(b,c,a); constraint int_times(5,c,b);\n"
     "constraint int_abs(c, abs_c) :: defines_var(abs_c);\n"
     "constraint int_le(abs_c, 10); constraint int_ne(a,0); % UNSAT\n",
       solver::UNSAT, DEFAULT},
    // Since 1073741827 is a prime and abs(a),abs(b) are smaller then
    // there is no feasible assignment. The propagator, even if an
    // over-approximated one, must figure out that there is no solution.
    {"var 0..30: a,b; constraint int_times(a,b,31); % UNSAT\n",
     solver::UNSAT, 0},
    {"var 1..8: x;\n"
     "var 1..8: z,y :: var_is_introduced :: is_defined_var;\n"
     "constraint int_le(1, z) :: defines_var(z);\n"
     "constraint hcsp_no_overlap(x,4,y,4) :: defines_var(y);\n"
     "constraint hcsp_no_overlap(x,4,z,4);\n"
     "constraint hcsp_no_overlap(y,4,z,4);  % UNSAT\n",
     solver::UNSAT, 1},
    {"var 1..8: x,y,z;\n"
     "constraint hcsp_no_overlap(x,4,y,4);\n"
     "constraint hcsp_no_overlap(x,4,z,4);\n"
     "constraint hcsp_no_overlap(y,4,z,4);  % UNSAT\n",
     solver::UNSAT, 1},
    {"var 1..9: x,y,z;\n"
     "constraint hcsp_no_overlap(x,4,y,4);\n"
     "constraint hcsp_no_overlap(x,4,z,4);\n"
     "constraint hcsp_no_overlap(y,4,z,4); % SAT\n",
     solver::SAT, 0},
    {"var 1..9: x,y,z;\n"
     "var bool : x1,x5,x9,y1,y5,y9 :: var_is_introduced :: is_defined_var;\n"
     "var bool : tmp1,tmp2 :: var_is_introduced :: is_defined_var;\n"
     "constraint int_ne_reif(x,1,x1) :: defines_var(x1);\n"
     "constraint int_ne_reif(x,5,x5) :: defines_var(x5);\n"
     "constraint int_ne_reif(x,9,x9) :: defines_var(x9);\n"
     "constraint int_ne_reif(y,1,y1) :: defines_var(y1);\n"
     "constraint int_ne_reif(y,5,y5) :: defines_var(y5);\n"
     "constraint int_ne_reif(y,9,y9) :: defines_var(y9);\n"
     "constraint array_bool_and([x1,x5,x9], tmp1) :: defines_var(tmp1);\n"
     "constraint array_bool_and([y1,y5,y9], tmp2) :: defines_var(tmp2);\n"
     "constraint hcsp_no_overlap(x,4,y,4);\n"
     "constraint hcsp_no_overlap(x,4,z,4);\n"
     "constraint hcsp_no_overlap(y,4,y,4);\n"
     "constraint bool_or(tmp1, tmp2, true); %UNSAT\n",
     solver::UNSAT, 6},
    //FIXME: The conflict should be detected for unlimited int without
    //       excessively long consistency iterations,
    {"var 0..9: x,y;\n"
     "constraint hcsp_sub_interval(x,y,3,7);\n"
     "constraint hcsp_no_sub_interval(x,y,4,7);\n"
     "constraint int_ne(x,3); % UNSAT\n",
     solver::UNSAT, 0},
    //FIXME: Consistency should be done for unlimited int without excessively
    //       long iterations. [limiting int to 0:9 in order to avoid this]
    {"var 0..9: x,y;\n"
     "constraint hcsp_sub_interval(x,y,3,7);\n"
     "constraint hcsp_no_sub_interval(x,y,4,7); % SAT\n",
     solver::SAT, 0},
    {"var int : x; var int : y;\n"
     "constraint hcsp_sub_interval(x,y,3,7); constraint int_ne(x,3); % SAT\n",
     solver::SAT, 0},
    {"var int : x; var int : y;\n"
     "constraint hcsp_sub_interval(x,y,4,7); constraint int_ne(x,3); % SAT\n",
     solver::SAT, 0},
    {"var 1..10: x,y; constraint hcsp_sub_interval(x,y,3,7);\n"
     "constraint hcsp_no_sub_interval_k(x,y,4,8,3); %x=3\n",
     solver::SAT, 0},
    {"var 1..10: x,y;\n"
     "constraint hcsp_sub_interval(x,y,3,7);\n"
     "constraint hcsp_no_sub_interval_k(x,y,4,8,3);\n"
     "constraint int_ne(x,3);\n", solver::UNSAT, 0},
    {"var 1..10: x,y;\n"
     "constraint hcsp_sub_interval(x,y,3,7);\n"
     "constraint hcsp_no_sub_interval_k(x,y,4,8,3); constraint int_ne(y,6);"
     "constraint int_ne(y,7) ;\n", solver::UNSAT, 0},
    {"var 1..10: x,y; constraint hcsp_sub_interval(x,y,3,7);\n"
     "constraint hcsp_no_sub_interval_k(x,y,4,8,4);\n",
     solver::SAT, 0},
    {"var 1..10: x,y; constraint hcsp_sub_interval(x,y,3,7);\n"
     "constraint hcsp_no_sub_interval_k(x,y,4,8,5);\n",
     solver::UNSAT, 0},
    {"var 0..20: a,b,c,d;\n"
     "var bool : a_ge_c, d_eq_0:: var_is_introduced :: is_defined_var;\n"
     "constraint int_lt(a,b);\n"
     "constraint int_lt(b,c);\n"
     "constraint bool_or(a_ge_c, d_eq_0, true);\n"
     "constraint int_le_reif(c,a, a_ge_c) :: defines_var(a_ge_c);\n"
     "constraint int_eq_reif(d,0, d_eq_0) :: defines_var(d_eq_0);\n"
     "% d==0\n", solver::SAT, 1},
    {"var 0..20: a,b,c,d;\n"
     "var bool : a_ge_c, d_eq_0:: var_is_introduced :: is_defined_var;\n"
     "constraint int_lt(a,b);\n"
     "constraint int_lt(b,c);\n"
     "constraint bool_or(a_ge_c, d_eq_0, true);\n"
     "constraint int_le_reif(c,a, a_ge_c) :: defines_var(a_ge_c);\n"
     "constraint int_eq_reif(d,0, d_eq_0) :: defines_var(d_eq_0);\n"
     "constraint int_ne(d,0);% UNSAT\n",
     solver::UNSAT, 0},
    {"var 0..20: a,b,c,d;\n"
     "var int : c_minus_1:: var_is_introduced :: is_defined_var;\n"
     "var bool : a_ge_c_minus_1, d_eq_0:: var_is_introduced :: is_defined_var;\n"
     "constraint int_lt(a,b);\n"
     "constraint int_lt(b,c);\n"
     "constraint bool_or(a_ge_c_minus_1, d_eq_0, true);\n"
     "constraint int_eq_reif(d, 0, d_eq_0) :: defines_var(d_eq_0);\n"
     "constraint int_le_reif(c_minus_1, a, a_ge_c_minus_1)\n"
     "                                      ::defines_var(a_ge_c_minus_1);\n"
     "constraint int_plus(c_minus_1, 1, c) :: defines_var(c_minus_1);\n"
     "constraint int_ne(d,0); % UNSAT\n",
     solver::UNSAT, 0},
    {"var 0..20: a,b,c,d;\n"
     "var int : a_plus_1:: var_is_introduced :: is_defined_var;\n"
     "var bool : a_plus_1_ge_c, d_eq_0:: var_is_introduced :: is_defined_var;\n"
     "constraint int_lt(a,b);\n"
     "constraint int_lt(b,c);\n"
     "constraint bool_or(a_plus_1_ge_c, d_eq_0, true);\n"
     "constraint int_eq_reif(d,0,d_eq_0) :: defines_var(d_eq_0);\n"
     "constraint int_le_reif(c,a_plus_1, a_plus_1_ge_c)\n"
     "                                          :: defines_var(a_plus_1_ge_c);\n"
     "constraint int_plus(a,1,a_plus_1) :: defines_var(a_plus_1);\n"
     "constraint int_ne(d,0); % UNSAT\n",
     solver::UNSAT, 0},
    {"var 0..20: a,b,c,d;\n"
     "var int : c_minus_2:: var_is_introduced :: is_defined_var;\n"
     "var bool : a_ge_c_minus_2, d_eq_0:: var_is_introduced :: is_defined_var;\n"
     "constraint int_lt(a,b);\n"
     "constraint int_lt(b,c);\n"
     "constraint bool_or(a_ge_c_minus_2, d_eq_0, true);\n"
     "constraint int_eq_reif(d, 0, d_eq_0) :: defines_var(d_eq_0);\n"
     "constraint int_le_reif(c_minus_2, a, a_ge_c_minus_2)\n"
     "                                      ::defines_var(a_ge_c_minus_2);\n"
     "constraint int_plus(c_minus_2, 2, c) :: defines_var(c_minus_2);\n"
     "constraint int_ne(d,0); %  a=0,b=1,c=2\n",
     solver::SAT, 0},
    {"var 0..20: a,b,c,d;\n"
     "var int : a_plus_2:: var_is_introduced :: is_defined_var;\n"
     "var bool : a_plus_2_ge_c, d_eq_0:: var_is_introduced :: is_defined_var;\n"
     "constraint int_lt(a,b);\n"
     "constraint int_lt(b,c);\n"
     "constraint bool_or(a_plus_2_ge_c, d_eq_0, true);\n"
     "constraint int_eq_reif(d,0,d_eq_0) :: defines_var(d_eq_0);\n"
     "constraint int_le_reif(c,a_plus_2, a_plus_2_ge_c)\n"
     "                                          :: defines_var(a_plus_2_ge_c);\n"
     "constraint int_plus(a,2,a_plus_2) :: defines_var(a_plus_2);\n"
     "constraint int_ne(d,0); % UNSAT\n",
     solver::SAT, 1},
    {"var 0..20: a,b,c,d;\n"
     "var bool : a_gt_c, d_eq_0 :: var_is_introduced :: is_defined_var;\n"
     "constraint int_lt(a,b);\n"
     "constraint int_lt(b,c);\n"
     "constraint int_lt_reif(c,a, a_gt_c) :: defines_var(a_gt_c);\n"
     "constraint int_eq_reif(d,0, d_eq_0) :: defines_var(d_eq_0);\n"
     "constraint bool_or(a_gt_c, d_eq_0, true); % d==0\n", solver::SAT, 1},
    {"var 0..20: a,b,c,d;\n"
     "var bool : a_gt_c, d_eq_0 :: var_is_introduced :: is_defined_var;\n"
     "constraint int_lt(a,b);\n"
     "constraint int_lt(b,c);\n"
     "constraint int_lt_reif(c,a, a_gt_c) :: defines_var(a_gt_c);\n"
     "constraint int_eq_reif(d,0, d_eq_0) :: defines_var(d_eq_0);\n"
     "constraint bool_or(a_gt_c, d_eq_0, true);\n"
     "constraint int_ne(d,0); % UNSAT\n",
     solver::UNSAT, 1},
    {"var {0,1:2:29,30}: x; var {0:2:30}: y; var bool : z;\n"
     "array [1..5] of var bool : tmp ::var_is_introduced :: is_defined_var;\n"
     "var int : tmp2 ::var_is_introduced :: is_defined_var;\n"
     "constraint int_eq_reif(x,0,tmp[1]) :: defines_var(tmp[1]);\n"
     "constraint int_eq_reif(y,30,tmp[2]) :: defines_var(tmp[2]);\n"
     "constraint bool_or(tmp[1],tmp[2],tmp[3]) :: defines_var(tmp[3]);\n"
     "constraint bool2int(tmp[3], tmp2) :: defines_var(tmp2);\n"
     "constraint int_le_reif(x,y,tmp[4]) :: defines_var(tmp[4]);\n"
     "constraint int_le_reif(y,x,tmp[5]) :: defines_var(tmp[5]);\n"
     "constraint int_ne(z,tmp2); constraint bool_and(tmp[4],tmp[5],true);\n"
     "% z=0 && (x=y=0 || x=y=30) \n",
     solver::SAT, 1},
    {"var {1:2:29}: x; var {0:2:30}: y; var bool : z;\n"
     "array [1..5] of var bool : tmp ::var_is_introduced :: is_defined_var;\n"
     "var int : tmp2 ::var_is_introduced :: is_defined_var;\n"
     "constraint int_eq_reif(x,0, tmp[1]) :: defines_var(tmp[1]);\n"
     "constraint int_eq_reif(y,30, tmp[2]) :: defines_var(tmp[2]);\n"
     "constraint bool_or(tmp[1], tmp[2], tmp[3]) :: defines_var(tmp[3]);\n"
     "constraint bool2int(tmp[3], tmp2) :: defines_var(tmp2);\n"
     "constraint int_ne(z,tmp2); constraint bool_and(tmp[4],tmp[5],true); \n"
     "constraint int_le_reif(x,y,tmp[4]) :: defines_var(tmp[4]);\n"
     "constraint int_le_reif(y,x,tmp[5]) :: defines_var(tmp[5]);\n"
     "% UNSAT \n", solver::UNSAT, 0},
    {"var bool : a,b,c,d,e;\n"
     "constraint bool_eq(c,d);\n"
     "constraint array_bool_xor([a,b,c,d, true]);\n"
     "constraint array_bool_xor([a,b]); %UNSAT\n",
     solver::UNSAT, 0},  // combine(xor,==) detects UNSAT without backtracks
    {"var bool : a,b,c,d,e;\n"
     "constraint array_bool_xor([a,b,c,d, true]);\n"
     "constraint array_bool_xor([a,b]);\n"
     "constraint bool_not(c,d); %SAT\n",
     solver::SAT, 0},
    {"var bool : a,b,c,d,e;\n"
     "constraint array_bool_xor([a,b,c,d, false]);\n"
     "constraint array_bool_xor([a,b,e, false]);\n"
     "constraint array_bool_xor([e,c,d, false]); % UNSAT\n",
     solver::UNSAT, 0}, // combine() should detect UNSAT without backtracks
    {"var bool : a,b,c,d,e;\n"
     "constraint array_bool_xor([a,b,c,d]);\n"
     "constraint array_bool_xor([a,b,e]);\n"
     "constraint array_bool_xor([e,c,d, true]); % SAT\n",
     solver::SAT, 0}, // Unlike previous problem, this is trivially SAAT
    {"array [1..4] of var bool: x;\n"
     "array [1..4] of var bool : y = [x[2],x[3],x[4],true];\n"
     "constraint bool_eq(x[1], false);\n"
     "constraint array_bool_xor(x); constraint array_bool_xor(y);\n"
     "% UNSAT",
     solver::UNSAT, 8},
    {"var bool : a,b,c ::output_var;\n"
     "constraint int_lin_le([1,1,1],[a,b,c],1);\n"
     "constraint bool_not(a,b);\n"
     "constraint bool_not(a,c); %SAT\n",
     solver::SAT, 2},
    {"array [1..4] of var bool: x;\n"
     "array [1..4] of var bool : y = [x[3],x[4],false,x[1]];\n"
     "constraint bool_eq(x[2], true);\n"
     "constraint array_bool_xor(x); constraint array_bool_xor(y);\n"
     "% UNSAT",
     solver::UNSAT, 8},
    {"array [1..4] of var bool: x;\n"
     "array [1..4] of var bool : y = [x[4],false,x[1],x[2]];\n"
     "constraint bool_eq(x[3], true);\n"
     "constraint array_bool_xor(x); constraint array_bool_xor(y);\n"
     "% UNSAT",
     solver::UNSAT, 8},
    {"array [1..4] of var bool: x;\n"
     "array [1..4] of var bool : y = [false,x[1],x[2],x[3]];\n"
     "constraint bool_eq(x[4], true);\n"
     "constraint array_bool_xor(x); constraint array_bool_xor(y);\n"
     "% UNSAT",
     solver::UNSAT, 8},
    {"array [1..4] of var bool: x;\n"
     "array [1..4] of var bool : y = [true,x[1],x[2],x[3]];\n"
     "constraint bool_eq(x[4], true);\n"
     "constraint array_bool_xor(x); constraint array_bool_xor(y);\n"
     "% SAT",
     solver::SAT, 0},
    {"array [1..4] of var bool: x;\n"
     "array [1..4] of var bool : y = [x[2],x[3],x[4],true];\n"
     "constraint bool_eq(x[1], true);\n"
     "constraint array_bool_xor(x); constraint array_bool_xor(y);\n"
     "% SAT",
     solver::SAT, 0},
    {"array [1..4] of var 0..31: x; var bool : a,b,c,d,e;\n"
     "array [1..10] of var bool : tmp ::var_is_introduced ::is_defined_var;\n"
     "array [1..3] of var bool : q,r ::var_is_introduced ::is_defined_var;\n"
     "constraint int_lt_reif(x[1],x[2],tmp[1]) :: defines_var(tmp[1]);\n"
     "constraint int_lt_reif(x[2],x[3],tmp[2]) :: defines_var(tmp[2]);\n"
     "constraint int_le_reif(x[3],x[4],tmp[3]) :: defines_var(tmp[3]);\n"
     "constraint int_le_reif(x[4],x[1],tmp[4]) :: defines_var(tmp[4]);\n"
     "constraint int_le_reif(b,a,q[1]) :: defines_var(q[1]);\n"
     "constraint int_le_reif(b,a,q[2]) :: defines_var(q[2]);\n"
     "constraint int_eq_reif(a,b,q[3]) :: defines_var(q[3]);\n"
     "constraint int_eq_reif(d,c,r[1]) :: defines_var(r[1]);\n"
     "constraint int_eq_reif(d,c,r[2]) :: defines_var(r[2]);\n"
     "constraint int_eq_reif(c,d,r[3]) :: defines_var(r[3]);\n"
     "constraint bool_or(a,tmp[1],true); constraint bool_or(b,tmp[2],true);\n"
     "constraint bool_or(c,tmp[3],true); constraint bool_or(d,tmp[4],true);\n"
     "constraint array_bool_or(q, true); constraint array_bool_or(r, true);\n",
     solver::SAT, 4},
    {"var 1..2: x1; var 2..3: x2; var 3..4: x3;\n"
     "var 0..31: index; var {1,4,5}: value;\n"
     "constraint array_int_element(index, [x1, x2, x3], value);\n"
     "constraint int_ne(x1,value); constraint int_ne(x2,value);\n",
     solver::SAT, 1},
    {"var 1..2: x1; var 2..3: x2; var 3..4: x3;\n"
     "var 0..31: index; var {1,4,5}: value;\n"
     "constraint array_int_element(index, [x1, x2, x3], value);\n"
     "constraint int_ne(x1,value); constraint int_ne(x3,value);\n",
     solver::UNSAT, 1},
    {"array [1..0] of var int : constraints3 :: output_array([1..0]) = [];\n"
     "solve satisfy;\n", solver::SAT, 0},
    {"var 0..1: tmp :: var_is_introduced :: is_defined_var;\n"
     "constraint int_lin_eq([1],[tmp],2)::defines_var(tmp);\n"
     "solve maximize tmp;\n",
     solver::UNSAT, 0},
#ifdef USE_BIG_ISET
    {"var -1000..1000 : a,b,c;\n"
     "var bool : a_eq_b  ::var_is_introduced ::is_defined_var;\n"
     "var int : a_eq_b_plus_1, b_plus_1 ::var_is_introduced ::is_defined_var;\n"
     "var int : int_a_eq_b  ::var_is_introduced ::is_defined_var;\n"
     "var int : abs_b  ::var_is_introduced ::is_defined_var;\n"
     "constraint array_int_element(a_eq_b_plus_1 , [b_plus_1, c], a);\n"
     "constraint int_lt(a,b);\n"
     "constraint int_plus(b,1, b_plus_1) :: defines_var(b_plus_1);\n"
     "constraint int_plus(int_a_eq_b,1,a_eq_b_plus_1) \n"
     "                     :: defines_var(a_eq_b_plus_1);\n"
     "constraint int_le(abs_b,9);\n"
     "constraint int_eq_reif(a,b,a_eq_b) :: defines_var(a_eq_b);\n"
     "constraint bool2int(a_eq_b, int_a_eq_b) :: defines_var(int_a_eq_b);\n"
     "constraint int_abs(b,abs_b) :: defines_var(abs_b); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"array [1..3] of var 0..2147483647 : a;\n"
     "constraint int_le(a[1],2); constraint int_le(a[2],2);\n"
     "constraint int_le(a[3],2);\n"
     "constraint int_lin_eq([1,3,11],[a[1],a[2],a[3]],16);%result a={2,1,1}\n",
     solver::SAT,   DEFAULT},
    {"array [1..3] of var 0..2147483647 : a;\n"
     "constraint int_le(a[1],1);\n"
     "constraint int_lin_eq([1,3,11],[a[1],a[2],a[3]],16); %result a={1,5,9}\n",
     solver::SAT,   DEFAULT},
    {"var 0..2147483647 : a; var 0..2147483647 : b; var 0..2147483647 : c;\n"
     "constraint int_le(a,1); constraint int_le(b,4); constraint int_le(c,2);\n"
     "constraint int_lin_eq([1,3,11],[a,b,c],16); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var 0..2147483647: a,b,c; \n"
     "var int : b_mul_c :: is_defined_var :: var_is_introduced;\n"
     "var int : a_plus_1 :: is_defined_var :: var_is_introduced;\n"
     "constraint int_lt(a,b);\n"
     "constraint int_times(b,c,b_mul_c) :: defines_var(b_mul_c);\n"
     "constraint int_times(a,1,a_plus_1) :: defines_var(a_plus_1);\n"
     "constraint int_le(a,b_mul_c); constraint int_le(b_mul_c, a_plus_1);\n"
     "constraint int_times(a,b,c); %a=0,b>=1,c=0\n",
     solver::SAT,   DEFAULT},
    {"var 0..2147483647 : a,b,c;\n"
     "var int : b_mul_c :: is_defined_var :: var_is_introduced;\n"
     "var int : a_plus_1 :: is_defined_var :: var_is_introduced;\n"
     "constraint int_lt(a,b);\n"
     "constraint int_times(b,c,b_mul_c) :: defines_var(b_mul_c);\n"
     "constraint int_times(a,1,a_plus_1) :: defines_var(a_plus_1);\n"
     "constraint int_le(a,b_mul_c) ; constraint int_le(b_mul_c, a_plus_1);\n"
     "constraint int_times(a,b,c); constraint int_lt(0,c); %UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var 0..2147483647 : a; var 0..2147483647 : b; var 0..2147483647 : c;\n"
     "var int : b_mul_c :: is_defined_var :: var_is_introduced;\n"
     "var int : a_plus_1 :: is_defined_var :: var_is_introduced;\n"
     "constraint int_lt(a,b);\n"
     "constraint int_lt(b,1);\n"
     "constraint int_times(b,c,b_mul_c) :: defines_var(b_mul_c);\n"
     "constraint int_le(a,b_mul_c) ;\n"
     "constraint int_times(a,1,a_plus_1) :: defines_var(a_plus_1);\n"
     "constraint int_le(b_mul_c, a_plus_1); constraint int_times(a,b,c);%UNSAT",
     solver::UNSAT, DEFAULT},
    {"var 0..2147483647 : a; array[1..2] of var 0..2147483647 : b;\n"
     "var int : b1_mul_b2, b2_mul_b1, a_add_1, a_mul_b1 \n"
     "                  :: is_defined_var :: var_is_introduced;\n"
     "constraint int_lt(a,b[1]);\n"
     "constraint int_eq(b[1],1);\n"
     "constraint int_times(b[1], b[2], b1_mul_b2) :: defines_var(b1_mul_b2);\n"
     "constraint int_le(a, b1_mul_b2);\n"
     "constraint int_times(b[2], b[1], b2_mul_b1) :: defines_var(b2_mul_b1);\n"
     "constraint int_plus(a,1, a_add_1) :: defines_var(a_add_1);\n"
     "constraint int_le(b2_mul_b1, a_add_1);\n"
     "constraint int_times(a,b[1], a_mul_b1) :: defines_var(a_mul_b1);\n"
     "constraint int_eq(a_mul_b1, b[2]);\n"
     "% a=0, b={1,0}\n",
     solver::SAT,   DEFAULT},
    {"var 0..2147483647 : a; var 0..2147483647 : b; var 0..2147483647 : c;\n"
     "var int : b_mul_c, a_plus_1, c_times_b\n"
     "                  :: is_defined_var :: var_is_introduced;\n"
     "constraint int_lt(a,b); constraint int_ne(a,0);\n"
     "constraint int_times(b,c, b_mul_c) :: defines_var(b_mul_c);\n"
     "constraint int_le(a, b_mul_c);\n"
     "constraint int_plus(a, 1, a_plus_1) :: defines_var(a_plus_1);\n"
     "constraint int_times(c,b,c_times_b) :: defines_var(c_times_b);\n"
     "constraint int_le(c_times_b, a_plus_1); constraint int_times(a,b,c);\n",
     solver::UNSAT, DEFAULT},
    {"var 0..2147483647 : a; var 0..2147483647 : b; var 0..2147483647 : c;\n"
     "var int : b_mul_c, a_plus_1, b_times_c\n"
     "                  :: is_defined_var :: var_is_introduced;\n"
     "constraint int_ne(a,0); constraint int_lt(a,b);\n"
     "constraint int_times(b,c, b_mul_c) :: defines_var(b_mul_c);\n"
     "constraint int_le(a, b_mul_c);\n"
     "constraint int_plus(a, 1, a_plus_1) :: defines_var(a_plus_1);\n"
     "constraint int_times(b,c,b_times_c) :: defines_var(b_times_c);\n"
     "constraint int_le(b_times_c, a_plus_1);\n"
     "constraint int_times(a,b,c); %UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var 0..2147483647: a; var int: tmp::is_defined_var ::var_is_introduced;\n"
     "constraint int_div(20,a,tmp) :: defines_var(tmp);\n"
     "constraint int_le(10, tmp); %a:{1,2}\n",
     solver::SAT,   DEFAULT},
    {"var 0..2147483647: a; var int: tmp::is_defined_var ::var_is_introduced;\n"
     "constraint int_div(20,a,tmp) :: defines_var(tmp);\n"
     "constraint int_eq(a,1); constraint int_le(10,tmp); %a=1\n",
     solver::SAT,   DEFAULT},
    {"var 0..2147483647: a; var int: tmp::is_defined_var ::var_is_introduced;\n"
     "constraint int_div(20,a,tmp) :: defines_var(tmp);\n"
     "constraint int_eq(a,2); constraint int_le(10,tmp); %a=2\n",
     solver::SAT,   DEFAULT},
    {"int : ten = 10; int : twenty = 20; var 0..2147483647 : a;\n"
     "array [1..3] of var int: x ::is_defined_var ::var_is_introduced;\n"
     "constraint int_div(twenty,a,x[1]) :: defines_var(x[1]);\n"
     "constraint int_ne_reif(a,1,x[2]) :: defines_var(x[2]);\n"
     "constraint int_ne_reif(a,2,x[3]) :: defines_var(x[3]);\n"
     "constraint int_le(ten,x[1]); constraint bool_and(x[2],x[3],true);"
     "%UNSAT\n", solver::UNSAT, DEFAULT},
    {"var 0..2147483647 : a; var 0..2147483647 : b; var 0..2147483647 : c;\n"
     "var int : b_mul_c, a_plus_1, c_times_b\n"
     "                  :: is_defined_var :: var_is_introduced;\n"
     "constraint int_lt(a,b);\n"
     "constraint int_lt(99,b);\n"
     "constraint int_times(b,c, b_mul_c) :: defines_var(b_mul_c);\n"
     "constraint int_le(a, b_mul_c);\n"
     "constraint int_plus(a, 1, a_plus_1) :: defines_var(a_plus_1);\n"
     "constraint int_times(c,b,c_times_b) :: defines_var(c_times_b);\n"
     "constraint int_le(c_times_b, a_plus_1);\n"
     "constraint int_times(a,b,c); %a=0, b>99, c=0\n",
     solver::SAT,   DEFAULT},
    {"var 0..2147483647 : a; var 0..2147483647 : b; var 0..2147483647 : c;\n"
     "var int : b_mul_c, a_plus_1, b_times_c\n"
     "                  :: is_defined_var :: var_is_introduced;\n"
     "constraint int_lt(a,b);\n"
     "constraint int_lt(99,b);\n"
     "constraint int_times(b,c, b_mul_c) :: defines_var(b_mul_c);\n"
     "constraint int_le(a,b_mul_c);\n"
     "constraint int_plus(a, 1, a_plus_1) :: defines_var(a_plus_1);\n"
     "constraint int_times(b,c,b_times_c) :: defines_var(b_times_c);\n"
     "constraint int_le(b_times_c, a_plus_1);\n"
     "constraint int_times(a,b,c); %a=0, b>99, c=0\n",
     solver::SAT,   DEFAULT},
    {"var int: a,b;\n"
     "var int: a_sub_10, abs_a_sub_10, abs_a_sub_10_mul_100, a_sub_5, \n"
     "         abs_a_sub_5, b_plus_6, abs_b_plus_6 \n"
     "         :: var_is_introduced :: is_defined_var;\n"
     "constraint int_plus(a, -10, a_sub_10) :: defines_var(a_sub_10);\n"
     "constraint int_abs(a_sub_10, abs_a_sub_10) ::defines_var(abs_a_sub_10);\n"
     "constraint int_times(abs_a_sub_10, 100, abs_a_sub_10_mul_100)\n"
     "         :: defines_var(abs_a_sub_10_mul_100);\n"
     "constraint int_le(abs_a_sub_10_mul_100, b);\n"
     "constraint int_plus(a_sub_5, 5, a) :: defines_var(a_sub_5);\n"
     "constraint int_abs(a_sub_5, abs_a_sub_5) :: defines_var(abs_a_sub_5);\n"
     "constraint int_lt(b, abs_a_sub_5);\n"
     "constraint int_plus(b, 6, b_plus_6) :: defines_var(b_plus_6);\n"
     "constraint int_abs(b_plus_6, abs_b_plus_6) :: defines_var(abs_b_plus_6);\n"
     "constraint int_lt(a, abs_b_plus_6);% UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var int: a,b;\n"
     "var int: a_sub_10, abs_a_sub_10, abs_a_sub_10_mul_100, a_sub_5,\n"
     "         abs_a_sub_5, b_plus_7, abs_b_plus_7"
     "         :: var_is_introduced :: is_defined_var;\n"
     "constraint int_plus(a_sub_10, 10, a) :: defines_var(a_sub_10);\n"
     "constraint int_abs(a_sub_10, abs_a_sub_10) :: defines_var(abs_a_sub_10);\n"
     "constraint int_times(abs_a_sub_10, 100, abs_a_sub_10_mul_100)\n"
     "                                    ::defines_var(abs_a_sub_10_mul_100);\n"
     "constraint int_le(abs_a_sub_10_mul_100, b);\n"
     "constraint int_plus(a_sub_5, 5, a) :: defines_var(a_sub_5);\n"
     "constraint int_abs(a_sub_5, abs_a_sub_5) :: defines_var(abs_a_sub_5);\n"
     "constraint int_lt(b, abs_a_sub_5);\n"
     "constraint int_plus(b, 7, b_plus_7) :: defines_var(b_plus_7);\n"
     "constraint int_abs(b_plus_7, abs_b_plus_7) :: defines_var(abs_b_plus_7);\n"
     "constraint int_lt(a, abs_b_plus_7); % a==10, b=4\n",
     solver::SAT,   DEFAULT},
    {"var int: a,b;\n"
     "var int: a_sub_10, abs_a_sub_10, abs_a_sub_10_mul_100, a_sub_5,\n"
     "         abs_a_sub_5, b_plus_7, abs_b_plus_7"
     "         :: var_is_introduced :: is_defined_var;\n"
     "constraint int_plus(a_sub_10, 10, a) :: defines_var(a_sub_10);\n"
     "constraint int_abs(a_sub_10, abs_a_sub_10) :: defines_var(abs_a_sub_10);\n"
     "constraint int_times(abs_a_sub_10, 100, abs_a_sub_10_mul_100)\n"
     "                                    ::defines_var(abs_a_sub_10_mul_100);\n"
     "constraint int_le(abs_a_sub_10_mul_100, b);\n"
     "constraint int_plus(a_sub_5, 5, a) :: defines_var(a_sub_5);\n"
     "constraint int_abs(a_sub_5, abs_a_sub_5) :: defines_var(abs_a_sub_5);\n"
     "constraint int_lt(b, abs_a_sub_5);\n"
     "constraint int_plus(b, 7, b_plus_7) :: defines_var(b_plus_7);\n"
     "constraint int_abs(b_plus_7, abs_b_plus_7) :: defines_var(abs_b_plus_7);\n"
     "constraint int_lt(a,abs_b_plus_7); constraint int_ne(b,4); %UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"array [1..5] of int : t = [10,100,5,7,10]; var int: a,b;\n"
     "var int: a_sub_t1, abs_a_sub_t1, abs_a_sub_t1_mul_t2, a_sub_t3,\n"
     "         abs_a_sub_t3, b_add_t4, abs_b_add_t4\n"
     "         :: var_is_introduced :: is_defined_var;\n"
     "constraint int_plus(a_sub_t1, t[1], a) :: defines_var(a_sub_t1);\n"
     "constraint int_abs(a_sub_t1, abs_a_sub_t1) :: defines(abs_a_sub_t1);\n"
     "constraint int_times(abs_a_sub_t1, t[2], abs_a_sub_t1_mul_t2)\n"
     "                                       :: defines(abs_a_sub_t1_mul_t2);\n"
     "constraint int_le(abs_a_sub_t1_mul_t2, b);\n"
     "constraint int_plus(a_sub_t3, t[3], a) :: defines(a_sub_t3);\n"
     "constraint int_abs(a_sub_t3, abs_a_sub_t3) :: defines(abs_a_sub_t3);\n"
     "constraint int_lt(b, abs_a_sub_t3);\n"
     "constraint int_plus(b, t[4], b_add_t4) :: defines(b_add_t4);\n"
     "constraint int_abs(b_add_t4, abs_b_add_t4) :: defines(abs_b_add_t4);\n"
     "constraint int_lt(a, abs_b_add_t4);\n"
     "constraint int_ne(a,t[5]);% UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"array [1..4] of int : sorted=[5,7,10,100];\n"
     "array [1..5] of var int : t = [\n"
     "    sorted[3],sorted[4],sorted[1],sorted[2],sorted[3]];\n"
     "var int: a,b;\n"
     "var int: a_sub_t1, abs_a_sub_t1, abs_a_sub_t1_mul_t2, a_sub_t3,\n"
     "         abs_a_sub_t3, b_add_t4, abs_b_add_t4\n"
     "         :: var_is_introduced :: is_defined_var;\n"
     "constraint int_ne(a,t[5]);\n"
     "constraint int_plus(a_sub_t1, t[1], a) :: defines_var(a_sub_t1);\n"
     "constraint int_abs(a_sub_t1, abs_a_sub_t1) :: defines(abs_a_sub_t1);\n"
     "constraint int_times(abs_a_sub_t1, t[2], abs_a_sub_t1_mul_t2)\n"
     "                                       :: defines(abs_a_sub_t1_mul_t2);\n"
     "constraint int_le(abs_a_sub_t1_mul_t2, b);\n"
     "constraint int_plus(a_sub_t3, t[3], a) :: defines(a_sub_t3);\n"
     "constraint int_abs(a_sub_t3, abs_a_sub_t3) :: defines(abs_a_sub_t3);\n"
     "constraint int_lt(b, abs_a_sub_t3);\n"
     "constraint int_plus(b, t[4], b_add_t4) :: defines(b_add_t4);\n"
     "constraint int_abs(b_add_t4, abs_b_add_t4) :: defines(abs_b_add_t4);\n"
     "constraint int_lt(a,abs_b_add_t4); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var int : a; var int : b; var int : c; \n"
     "var int : c_minus_1, a_minus_3\n"
     "                   :: var_is_introduced ::is_defined_var :: output_var;\n"
     "constraint int_le(-10, a); constraint int_le(a, 10);\n"
     "constraint int_lt(a,b);\n"
     "constraint int_plus(c_minus_1, 1, c) :: defines_var(c_minus_1);\n"
     "constraint int_le(b, c_minus_1);\n"
     "constraint int_plus(a_minus_3, 3, a) :: defines_var(a_minus_3);\n"
     "constraint int_le(c, a_minus_3); % UNSAT",
     solver::UNSAT, DEFAULT},
    {"var int : a,b,c;\n"
     "var int : c_minus_1, a_plus_2:: var_is_introduced :: is_defined_var;\n"
     "constraint int_lt(-1,a);\n"
     "constraint int_lt(a,1);\n"
     "constraint int_lt(a,b);\n"
     "constraint int_le(b,c_minus_1);\n"
     "constraint int_plus(c_minus_1, 1, c) :: defines_var(c_minus_1);\n"
     "constraint int_le(c,a_plus_2);\n"
     "constraint int_plus(a,2,a_plus_2) :: defines_var(a_plus_2);\n"
     "% result a=0, b=1,c=2\n",
     solver::SAT,   DEFAULT},
    {"var int : a; var int : b;\n"
     "var int : min_a_b :: var_is_introduced :: is_defined_var;\n"
     "constraint int_min(a,b,min_a_b) :: defines_var(min_a_b);\n"
     "constraint int_le(-10,a) ; constraint int_le(a,10) ;\n"
     "constraint int_lt(a,min_a_b); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var int : a,b;\n"
     "var int : max_a_b :: var_is_introduced :: is_defined_var;\n"
     "constraint int_max(a,b,max_a_b) :: defines_var(max_a_b);\n"
     "constraint int_lt(max_a_b,a); constraint int_le(-10,a);\n"
     "constraint int_le(a,10); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"var int : a,b;\n"
     "var int : min_a_b :: var_is_introduced :: is_defined_var;\n"
     "constraint int_lt(a,0);\n"
     "constraint int_min(a,b,min_a_b) :: defines_var(min_a_b);\n"
     "constraint int_le(a,min_a_b); % SAT\n", solver::SAT, DEFAULT  },
    {"var int : a; var int : b; \n"
     "var int : min_a_b :: var_is_introduced :: is_defined_var;\n"
     "constraint int_lt(b,0);\n"
     "constraint int_min(a,b,min_a_b) :: defines_var(min_a_b);\n"
     "constraint int_le(a, min_a_b); % SAT\n",
     solver::SAT, DEFAULT  },
    {"var int : a; var int : b;\n"
     "var int : max_a_b :: var_is_introduced :: is_defined_var;\n"
     "constraint int_lt(a,0);\n"
     "constraint int_max(a,b,max_a_b) :: defines_var(max_a_b);\n"
     "constraint int_le(max_a_b,a); % SAT\n",
     solver::SAT, DEFAULT  },
    {"var int : a; var int : b;\n"
     "var int : max_a_b :: var_is_introduced :: is_defined_var;\n"
     "constraint int_lt(b,0);\n"
     "constraint int_max(a,b,max_a_b) :: defines_var(max_a_b);\n"
     "constraint int_le(max_a_b,a); % SAT\n",
     solver::SAT, DEFAULT  },
    {"var int : a; var int : b; \n"
     "var int : abs_b :: var_is_introduced :: is_defined_var;\n"
     "constraint int_eq(a,-4); constraint int_mod(b,5,a);\n"
     "constraint int_abs(b, abs_b):: defines_var(abs_b);\n"
     "constraint int_le(abs_b, 10); % SAT\n",
     solver::SAT, DEFAULT},
    {"var int:a; var int:b; constraint int_eq(a,-4); constraint int_eq(b,-9);\n"
     "constraint int_mod(b,5,a); % SAT\n",
     solver::SAT, DEFAULT},
    {"var int:a; var int:b; constraint int_ne(a,-4); constraint int_eq(b,-9);\n"
     "constraint int_mod(b,5,a); % UNSAT\n",
     solver::UNSAT, DEFAULT},
    {"int : min_int= -2147483648; array [1..2] of var -50..50: a; \n"
     "constraint int_lt(a[1],a[2]);\n"
     "constraint int_mod(a[2],min_int,a[1]); % UNSAT\n", solver::UNSAT, 0},
    {"var -50..50: a,b; var {-2147483648}: min_int; \n"
     "constraint int_mod(b,min_int,a); constraint int_lt(b,a); % UNSAT\n",
     solver::UNSAT, 0},
    {"predicate all_different(array [int] of var int: xs);\n"
     "int : min_int= -2147483648;\n"
     "var -50..50: a,b; \n"
     "constraint int_mod(b,min_int,a); constraint int_eq(a,b); % SAT\n",
     solver::SAT, 0},
    {"var int: c,d; var {-2147483648}: min_int;\n"
     "constraint int_mod(c,d,min_int); % UNSAT\n", solver::UNSAT, 0},
    {"var -1..2: a,b;\n"
     "var int : abs_b, five_sub_abs_b :: var_is_introduced :: is_defined_var;\n"
     "constraint int_le(a, 0);\n"
     "constraint int_abs(b, abs_b) :: defines_var(abs_b);\n"
     "constraint int_plus(five_sub_abs_b, abs_b, 5)\n"
     "                                     :: defines_var(five_sub_abs_b);\n"
     "constraint int_eq(a, five_sub_abs_b); % UNSAT\n",
     solver::UNSAT, 0},
    {"var {-17,-1}: a; var 1..9: b; var {-2,-1}: c; "
     "constraint int_times(b,c,a);%a=-1,b=1,c=-1\n",
     solver::SAT, 1},
    {"var int : a; var int : b;\n"
     "constraint int_le(b,9); constraint int_le(3,a);\n"
     "constraint int_times(b,2,a); % e.g. a=4,b=2\n",
     solver::SAT,   0},
    {"var {1001..1999,1000000}: a; var 1..1000: b;\n"
     "constraint int_times(b, 1000, a); % b=1000\n",
     solver::SAT, 0},
    // Since 1073741827 is a prime and abs(a),abs(b) are smaller then
    // there is no feasible assignment. The propagator, even if an
    // over-approximated one, must figure out that there is no solution.
    {"var -1073741826..1073741826: a,b;\n"
     "constraint int_times(a,b,1073741827); % UNSAT\n",
     solver::UNSAT, 0},
    // Test intersection of iset with different skip sizes. This should be fast
    {"var{-2000000000:5:2000000000}:a; var{-1999999998:3:1999999998}:b; "
     "constraint int_eq(a, b); % SAT\n", solver::SAT, 0},
    // Although the ranges seem to overlap, they contain no shared item.
    // This is because the gcd of the steps is 1 which makes the steps of the
    // intersection to be 20000003*211 > UINT_MAX. This lets us choose
    // the bounds so that no element is contained in the 2 intervals
    // simultaneously.
    {"var{-2000000000:20000003:2000000600} : a; "
     "var{-2019999792:211:2000000060} : b;"
     "constraint int_eq(a,b);% UNSAT\n", solver::UNSAT, 0},
    // This tests the speed/strength of the learning process since a conflict is
    // detected after about 1000000 propagation steps of the same 3 constraints.
    // Conflict analysis should be either very fast or detect that the three
    // constraints conflict without checking all 1000000 propagation steps.
    // If this does not work as expected, the code will get stuck for hours
    // instead of less than a second. This will happen as long as
    // var_domain_after_modification() works in O(number of modifications)
    {"var 0..1000000: a,b,c,d;\n"
     "var bool : a_ge_c, d_eq_0:: var_is_introduced :: is_defined_var;\n"
     "constraint int_lt(a,b);\n"
     "constraint int_lt(b,c);\n"
     "constraint bool_or(a_ge_c, d_eq_0, true);\n"
     "constraint int_le_reif(c,a, a_ge_c) :: defines_var(a_ge_c);\n"
     "constraint int_eq_reif(d,0, d_eq_0) :: defines_var(d_eq_0);\n"
     "% d=0\n",
     solver::SAT, 1},
    // This tests the ability to learn "z || y==x", otherwise it gets to
    // O(n^3) issues.
    {"var {1:2:19999}: x; var {0:2:20000}: y; var bool : z;\n"
     "var bool: y_le_x :: var_is_introduced :: is_defined_var;\n"
     "constraint int_le(x,y);\n"
     "constraint int_le_reif(y,x,y_le_x) :: defines_var(y_le_x);\n"
     "constraint bool_or(z, y_le_x, true) ; % z=1", solver::SAT, 1},
    {"var 10..99: x0; var {0..9,90..99}: y0; var 0..89: z0; var 0..99: x,y,z;\n"
     "var bool : b;\n"
     "var bool : x_le_y, y_le_z, x0_eq_x, y0_eq_y, z0_eq_z\n"
     "                    ::var_is_introduced :: is_defined_var;\n"
     "constraint int_le_reif(x,y, x_le_y) :: defines_var(x_le_y);\n"
     "constraint int_le_reif(x,y, y_le_z) :: defines_var(y_le_z);\n"
     "constraint bool_and(x_le_y, y_le_z, true);\n"
     "constraint int_eq_reif(x0, x, x0_eq_x) :: defines_var(x0_eq_x);\n"
     "constraint int_eq_reif(y0, y, y0_eq_y) :: defines_var(y0_eq_y);\n"
     "constraint int_eq_reif(z0, z, z0_eq_z) :: defines_var(z0_eq_z);\n"
     "constraint bool_or(b, x0_eq_x, true);\n"
     "constraint bool_or(b, y0_eq_y, true);\n"
     "constraint bool_or(b, z0_eq_z, true);\n",
     solver::SAT, 50},
    {"var 0..9999: u,x,y,z;\n"
     "array [1..4] of var bool : a :: output_array([0..1],[1..2]);\n"
     "array [1..10] of var bool : tmp :: var_is_introduced :: is_defined_var;\n"
     "constraint int_lt_reif(u,x,tmp[1]) :: defines_var(tmp[1]);\n"
     "constraint int_lt_reif(x,y,tmp[2]) :: defines_var(tmp[2]);\n"
     "constraint int_le_reif(y,z,tmp[3]) :: defines_var(tmp[3]);\n"
     "constraint int_le_reif(z,u,tmp[4]) :: defines_var(tmp[4]);\n"
     "constraint int_le_reif(a[2],a[1],tmp[5]) :: defines_var(tmp[5]);\n"
     "constraint int_le_reif(a[2],a[1],tmp[6]) :: defines_var(tmp[6]);\n"
     "constraint int_eq_reif(a[2],a[1],tmp[7]) :: defines_var(tmp[7]);\n"
     "constraint int_le_reif(a[4],a[3],tmp[8]) :: defines_var(tmp[8]);\n"
     "constraint int_le_reif(a[4],a[3],tmp[9]) :: defines_var(tmp[9]);\n"
     "constraint int_eq_reif(a[4],a[3],tmp[10]) :: defines_var(tmp[10]);\n"
     "constraint bool_or(a[1], tmp[1], true);\n"
     "constraint bool_or(a[2], tmp[2], true);\n"
     "constraint bool_or(a[3], tmp[3], true);\n"
     "constraint bool_or(a[4], tmp[4], true);\n"
     "constraint array_bool_or([tmp[5], tmp[6], tmp[7]], true);\n"
     "constraint array_bool_or([tmp[8], tmp[9], tmp[10]], true);\n",
     solver::SAT, 2},

#endif
};

int main_impl(int argc, char *argv[])
{
    hcsp_flags flags;
    // The limit has to be small enough to detect weak expression optimization,
    // in which case the solver stops with BACKTRACK_LIMIT.
    flags.parse(argc, argv);

    table_propagator_base::memory_manager manager;
    for (unsigned i=0 ; i != sizeof(expressions)/sizeof(expressions[0]); ++i) {
        istringstream str(expressions[i].str);
        stringstream test_name;
        test_name << "test" << i;
        try {
            table_propagator_base::memory_manager p_null;
            fzn_solver problem(str, test_name.str(), flags, manager) ;
            if (flags.backtrack_limit < 0) {
                int max_backtracks= expressions[i].max_backtracks;
                problem.backtrack_limit(max_backtracks);
            }
            flags.quiet_start= true;
            solver::result_type result= problem.solve(&cout);
            T_ASSERT(result == expressions[i].expected);
        } catch (...) {
            cerr << "Found in: " << test_name.str() << ": `"
                 << expressions[i].str << "'" << endl;
            throw;
        }
    }
    return 0;
}

int main(int argc, char *argv[])
{
    try {
        return main_impl(argc, argv);
    } catch (std::exception & e) {
        cout << "Exception(" << typeid(e).name() << "):\n" << e.what() << endl;
    } catch (...) {
        cout << "unknown exception" << std::endl;
    }
    return 1;                                               
}
