// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2014, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2014, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#undef NDEBUG
#include <limits>
#include <algorithm>
#include <iostream>
#include <vector>
#include <cmath>
#include "int128.hpp"

#include "test_utilities.hpp"
#include "test_static_assert.hpp"
// Mulitple inclusion works since test_static_assert does not have
// the regular #ifdef protection
// This tests if HCSP_STATIC_ASSERT works well if there is more than one
// HCSP_STATIC_ASSERT that shares the same line number (probably in different
// source files).
#include "test_static_assert.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;

// Random tester class for various utilities in utilities.hpp
// It will inject random cases and check their outcomes.
class tester {
  public:
    tester() {}
    void run();
    void small_operations();
    void big_operations();
    void rand_up_and_back();
    void rand_down_and_back();
    void t_div();
    long long rand64()
    {
        switch (gen(30)) {
          case 0: return 0;
          case 1: return 1;
          case 2: return -1;
          case 3: return numeric_limits<long long>::max();
          case 4: return numeric_limits<long long>::min(); 
          case 5: return numeric_limits<long long>::max();
          case 6: return numeric_limits<long long>::min(); 
          default:
            return ((unsigned long long)gen() << 32)+gen();
        }
    }
}; // class tester

void tester::run()
{
    small_operations();
    big_operations();
    rand_up_and_back();
    rand_down_and_back();
    t_div();
}

void tester::small_operations()
{
    long long a= rand64(), b= rand64();
    
    if ( (a < 0) == (b < 0) ) {
        T_ASSERT(int128(a/2)+b/2 == a/2+b/2);
        T_ASSERT((int128(a/2)+=b/2) == a/2+b/2);
        T_ASSERT(int128(a)-b == a-b);
        T_ASSERT((int128(a)-=b) == a-b);
    } else {
        T_ASSERT(int128(a)+b == a+b);
        T_ASSERT((int128(a)+=b) == a+b);
        T_ASSERT(int128(a/2)-b/2 == a/2-b/2);
        T_ASSERT((int128(a/2)-=b/2) == a/2-b/2);
    }
    if (b == 0)
        T_ASSERT(int128(a)+b == int128(a));
    else
        T_ASSERT(int128(a)+b != int128(a));
    if (b < 0)
        T_ASSERT(int128(a)+b < int128(a));
    if (b > 0)
        T_ASSERT(int128(a) < int128(a)+b);
    if (a != numeric_limits<long long>::min())
        T_ASSERT(--int128(a) == a-1);
    if (a != numeric_limits<long long>::max())
        T_ASSERT(++int128(a) == a+1);
    T_ASSERT((int128(a) < int128(b)) == (a < b));
    T_ASSERT((int128(a) < int128(b)) == (a < b));
    T_ASSERT(!(int128(a) < int128(a)));
    T_ASSERT(!(int128(a) < a));
    if (a != b)
        T_ASSERT(!(int128(a) == int128(b)));
    T_ASSERT(int128(a) == int128(a));
    for (unsigned i=0 ; i <= 63 ; ++i)
        T_ASSERT((int128(a) >> i) == (a >> i));
    for (unsigned i=64 ; i <= 200 ; ++i)
        T_ASSERT((int128(a) >> i) == ((a >> 32) >> 32));
}

void tester::big_operations()
{
    long long a= rand64(), b= rand64();

    int128 sum = int128(a)+b;
    int rounding_fix= ((a&1) & (b&1) );
    T_ASSERT( (sum >> 1) == (a>>1)+(b>>1)+rounding_fix);
    int actual_lsb = (sum >> 1) != ((sum+1) >> 1);
    int expected_lst = (a&1) ^ (b&1);
    T_ASSERT(actual_lsb == expected_lst);

    T_ASSERT( int128(a)+int128(b) == sum);
    T_ASSERT( (int128(a)+=b ) == sum);
    T_ASSERT( (int128(a)+=int128(b) ) == sum);
    if (b>0)
        T_ASSERT(int128(a) < sum);
    else if (b<0)
        T_ASSERT(sum < int128(a));
    T_ASSERT(sum-b == a);
        
    int128 diff = int128(a)-b;

    rounding_fix= (((a&1)^1) & (b&1) );
    T_ASSERT( (diff >> 1) == (a>>1)-(b>>1)-rounding_fix);
    actual_lsb = (diff >> 1) != ((diff+1) >> 1);
    expected_lst = (a&1) ^ (b&1);
    T_ASSERT(actual_lsb == expected_lst);
    T_ASSERT( int128(a)-int128(b) == diff);
    T_ASSERT( (int128(a)-=b ) == diff);
    T_ASSERT( (int128(a)-=int128(b) ) == diff);
}

void tester::rand_up_and_back()
{
    int128 orig(rand64());
    int128 val= orig;
    vector<int128> added;
    for (int i=0 ; i != 60; ++i) {
        int128 prev= val;
        if (int128(0) < val && gen(2)) {
            added.push_back(val);
            val += val;
        } else {
            uint64_t next_added= rand64();
            next_added >>=1 ; // remove the upper bit
            added.push_back(int128(next_added));
            val += added.back();
        }
        T_ASSERT(prev<val  || prev==val);
    }
    random_shuffle(added.begin(), added.end());
    while (!added.empty()) {
        val += -added.back();
        added.pop_back();
    }
    T_ASSERT(val == orig);
}

void tester::rand_down_and_back()
{
    int128 orig(rand64());
    int128 val= orig;
    vector<int128> added;
    for (int i=0 ; i != 60; ++i) {
        int128 prev= val;
        if (val < 0 && gen(2)) {
            added.push_back(val);
            val += val;
        } else {
            uint64_t next_added= rand64();
            next_added = next_added | (uint64_t(1)<<63) ; // set the upper bit
            added.push_back(int128(int64_t(next_added)));
            val += added.back();
        }
        T_ASSERT(val<prev  || prev==val);
    }
    random_shuffle(added.begin(), added.end());
    while (!added.empty()) {
        val -= added.back();
        added.pop_back();
    }
    T_ASSERT(val == orig);
}

void tester::t_div()
{
    int32_t factor1= gen();
    uint32_t factor2= gen();
    int64_t base= factor1 * int64_t(factor2);
    if (base == 0)
        return;
    int128 val(0);
    T_ASSERT(val/factor1 == 0);
    for (int i=1 ; i != 5 ; ++i) {
        val += base;
        T_ASSERT(val/i == base);
        T_ASSERT(val.floor_div(i) == base);
        T_ASSERT(val.ceil_div(i) == base);
        T_ASSERT(val/factor1 == int64_t(i)*factor2);
        T_ASSERT(val.floor_div(factor1) == int64_t(i)*factor2);
        int factor1_sign= factor1 > 0 ? +1 : (factor1 < 0 ? -1 : +0);
        T_ASSERT((val+factor1_sign).floor_div(factor1) == int64_t(i)*factor2);
        T_ASSERT(val.ceil_div(factor1) == int64_t(i)*factor2);
        T_ASSERT((val-factor1_sign).ceil_div(factor1) == int64_t(i)*factor2);
    }
    int128 val_div_vactor1 = val/factor1;
    for (int i=0 ; i != 60; ++i) {
        int128 orig_val = val;
        val += val;
        val += base;
        T_ASSERT(val/2 == orig_val+base/2);
        T_ASSERT(val/factor1/2 == orig_val/factor1+factor2/2);
        T_ASSERT(val_div_vactor1 == orig_val/factor1);
        val_div_vactor1 += val_div_vactor1+factor2;
    }
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    cout << "Seed is " << set_seed(argc, argv) << endl;
    for (unsigned i=0 ; i < 10000 ; ++i) {
        tester tester;
        tester.run();
    }

    END_TEST_MAIN;
}
