// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <iostream>
#include <sstream>
#include <vector>
#include <algorithm>
#include <iset.hpp>
#include <clause_propagator.hpp>
#include <solver.hpp>
#include <formula.hpp>
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;

class tester {
  public:
    typedef solver::reporting_levels reporting_levels;
    void run(reporting_levels level, unsigned holes, unsigned pigeons);
};


class problem {
  public:
    typedef solver::reporting_levels reporting_levels;
    problem(reporting_levels level, unsigned holes, unsigned pigeons,
            solver::label_type add_only_labels= solver::A | solver::B);

    solver & get_solver() { return m_solver; }
    
  private:
    static unique_ptr<clause_propagator>
    make_clause(const iset & a, const iset & b);

    void add_shuffled_clause(unique_ptr<clause_propagator> clause,
                             int var_a, int var_b,
                             solver::label_type clause_label);
    void add_diff(unsigned var_a, unsigned var_b,
                  solver::label_type diff_label);
    solver m_solver;
    const unsigned m_holes;
    const unsigned m_pigeons;
}; // class problem

unique_ptr<clause_propagator>
problem::make_clause(const iset & a, const iset & b)
{
    unique_ptr<clause_propagator> clause( new clause_propagator );
    clause->update_literals().reserve(3);
    clause->update_literals().push_back(a);
    clause->update_literals().push_back(b);
    return clause;
}

void problem::add_diff(unsigned var_a, unsigned var_b,
                       solver::label_type diff_label)
{
    const iset complete_set(make_pair(0, m_holes-1));
    
    for (unsigned i=0 ; i < m_holes ; ++i) {
        iset not_i=complete_set;
        not_i.erase(i);
        add_shuffled_clause(make_clause(not_i, not_i), var_a, var_b,
                            diff_label);
    }
}

void problem::add_shuffled_clause(unique_ptr<clause_propagator> clause,
                                  int var_a, int var_b,
                                  solver::label_type clause_label)
{
    int vars[] = {var_a, var_b};

    static vector<unsigned> orders;
    orders.clear();
    for (unsigned i=0 ; i < sizeof(vars)/sizeof(vars[0]) ; ++i)
        if (vars[i] >= 0)
            orders.push_back(i);

    randomizer rand;
    random_shuffle(orders.begin(), orders.end(), rand );

    static clause_propagator::literals_type original_literals;
    original_literals= clause->literals();

    for (unsigned i=0 ; i != orders.size() ; ++i)
        clause->update_literals()[i]= original_literals[orders[i]];
    
    unique_ptr<constraint_propagator> constraint(clause.release());
    constraint_connectivity_info & connectivity=
        m_solver.add_constraint(move(constraint))
                .reserve_arguments(orders.size() )
                .labels(clause_label);
    for (unsigned i=0 ; i != orders.size() ; ++i)
        connectivity << vars[orders[i]];
}

problem::problem(reporting_levels level, unsigned holes, unsigned pigeons,
                 solver::label_type add_only_labels)
    :  m_holes( holes ), m_pigeons( pigeons )
{
    m_solver.report_level(level);
    m_solver.should_intersect_with_ground(false);
    for (unsigned i=0 ; i < m_pigeons ; ++i) {
        unsigned var= m_solver.add_var(iset(make_pair(0, m_holes-1)));
        if (var < m_pigeons-2)
            m_solver.variable_labels(var, solver::A);
        else
            m_solver.variable_labels(var, solver::B);
    }
    for (unsigned i=m_pigeons ; i < m_pigeons*2 ; ++i) {
        unsigned var= m_solver.add_var(iset(make_pair(0, m_holes-1)));
        if (var < 2*m_pigeons-2)
            m_solver.variable_labels(var, solver::A);
        else
            m_solver.variable_labels(var, solver::B);
    }

    for (unsigned i=0 ; i < m_pigeons ; ++i) {
        solver::label_type i_labels= m_solver.variable_labels(i);
        for (unsigned j=i+1 ; j < m_pigeons ; ++j) { 
            solver::label_type labels= i_labels | m_solver.variable_labels(j);
            if (labels != solver::A)
                labels= solver::B;
            if (labels & add_only_labels)
                add_diff(i,j, labels);
        }
    }
    for (unsigned i=0 ; i < m_pigeons ; ++i)  {
        solver::label_type labels= (m_solver.variable_labels(i) | 
                                    m_solver.variable_labels(i+m_pigeons));
        if (labels != solver::A)
            labels= solver::B;
        if (labels & add_only_labels)
            add_diff(i,i+m_pigeons, labels);
    }
    m_solver.backtrack_limit(100000);
}

void test(solver::reporting_levels level, unsigned holes, unsigned pigeons)
{
    problem full(level, holes, pigeons);
    full.get_solver().produce_proof(true);
    full.get_solver().should_intersect_with_ground(false);
    T_ASSERT(full.get_solver().solve() == solver::UNSAT);
    formula interpolant= full.get_solver().get_interpolant();
    if (level == solver::SEARCH)
        cout << "Interpolant=\n"
             << formula::printer(trivial_print_strategy::get(), interpolant)
             << "\n";

    problem only_a(level, holes, pigeons, solver::A);
    T_ASSERT(only_a.get_solver().solve() == solver::SAT);
    only_a.get_solver().reset();
    only_a.get_solver().forget_learned_constraints();
    (!interpolant).constrain_solver(only_a.get_solver());
    T_ASSERT(only_a.get_solver().solve() == solver::UNSAT);

    problem only_b(level, holes, pigeons, solver::B);
    T_ASSERT(only_b.get_solver().solve() == solver::SAT);
    only_b.get_solver().reset();
    only_b.get_solver().forget_learned_constraints();
    interpolant.constrain_solver(only_b.get_solver());
    T_ASSERT(only_b.get_solver().solve() == solver::UNSAT);
    
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    solver::reporting_levels report_level= solver::NONE;
    //report_level= solver::SEARCH;
    //report_level= solver::PROPAGATION_MODS;
    //report_level= solver::PROPAGATED_CONSTRAINTS;

    cout << "Seed is " << set_seed(argc, argv) << endl;
    for (unsigned i=0 ; i < 100 ; ++i) {
        if (i != 0) {
            unsigned seed= gen();
            srand(seed);
            if (report_level > solver::NONE)
                cout << "Seed is " << seed << endl;
        }
        test(report_level, 5, 6);
    }
    
    END_TEST_MAIN;
}

