// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <cstdlib>
#include <unistd.h>
#include <time.h>
#include <iostream>
#include <iomanip>
#include <vector>
#include <iset.hpp>
#include <interval_propagators.hpp>
#include <solver.hpp>
#include <flags.hpp>
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;

class interval_flags: public hcsp_flags
{
  public:
    interval_flags() : intervals(3), interval_size(2), section_size(5) {}
                       
  protected:
    bool parse_option(char option) override;
    void usage_short(std::ostream & out) override;
    void usage_long(std::ostream & out) override;
    std::string legal_args() const override
    { return hcsp_flags::legal_args() + "z:Z:y:"; }

  public:
    unsigned intervals;
    unsigned interval_size;
    unsigned section_size;
}; // class interval_flags

class tester {
  public:
    tester(const interval_flags & flags);
    void run();
    
  private:
    void add_no_overlap(unsigned var_a, unsigned var_b,
                        unsigned interval_size);
    solver m_solver;
    const unsigned m_intervals;
    const unsigned m_interval_size;
    const unsigned m_section_size;
};

void tester::add_no_overlap(unsigned var_a, unsigned var_b,
                            unsigned interval_size)
{
    unique_ptr<constraint_propagator> constraint(
        new no_overlap_propagator(interval_size, interval_size));
    
    m_solver.add_constraint(move(constraint)).reserve_arguments(2)
        << var_a << var_b;
}

    
tester::tester(const interval_flags & flags)
    : m_intervals(flags.intervals),
      m_interval_size(flags.interval_size),
      m_section_size(flags.section_size)
{
    HCSP_PROFILE_CURRENT;

    iset domain(make_pair(0, m_section_size-m_interval_size));
    flags.apply(m_solver);
    
    for (unsigned i=0 ; i != m_intervals ; ++i)
        m_solver.add_var(domain);

    for (unsigned i=0 ; i != m_intervals ; ++i)
        for (unsigned j=i+1 ; j != m_intervals ; ++j)
            add_no_overlap(i, j, m_interval_size);
}

void tester::run()
{
    HCSP_PROFILE_CURRENT;
    cout << "Solve:" << flush;
    unsigned long time_before= user_time_microseconds()/1000;
    solver::result_type result= m_solver.solve();
    unsigned long time_after= user_time_microseconds()/1000;
    if (m_interval_size*m_intervals > m_section_size) {
        cout << "UNSAT\n";
        T_ASSERT(result == solver::UNSAT);
    } else {
        cout << "SAT\n";
        T_ASSERT(result == solver::SAT);
        iset slots;
        for (unsigned i=0 ; i != m_intervals ; ++i) {
            iset::value_type lower= m_solver.var_domain(i).min();
            slots.insert(make_pair(lower, lower+m_interval_size-1));
        }
        T_ASSERT(slots.size() == m_intervals * m_interval_size);
        T_ASSERT(slots.min() >= 0 && slots.max() < (int)m_section_size);
    }
    ios_state_restore restorer(cout);
    cout << setprecision(2) << setw(7) << fixed
         << double(time_after-time_before)/1000.0 << " sec"
         << setw(7) << m_solver.backtracks_done() << " backtracks done\n";
}


bool interval_flags::parse_option(char option)
{
    if (hcsp_flags::parse_option(option))
        return true;
    switch (option) {
      case 'z': intervals=strtoul(optarg, NULL, 0);break;
      case 'Z': interval_size=strtoul(optarg, NULL, 0);break;
      case 'y': section_size= strtoul(optarg, NULL, 0);break;
      default:
        return false;
    }
    return true;
}

void interval_flags::usage_short(std::ostream & out)
{
    hcsp_flags::usage_short(out);
    out << "[-z intervals] [-Z interval_size] [-y section size]\n";
}

void interval_flags::usage_long(std::ostream & out)
{
    hcsp_flags::usage_long(out);
    out << "  -z intervals     Number of intervals to place in the section\n"
        << "  -Z interval_size The size of each individual interval\n"
        << "  -y section_size  The size of the total section\n";
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    HCSP_PROFILE_CURRENT;
    interval_flags flags;
    if (!flags.parse(argc, argv))
        return 1;
    if (flags.all_done())
        return 0;
    cout << flags.intervals << " intervals of length " << flags.interval_size
         << "  , fit into a " << flags.section_size << " size section\n";
    tester test(flags);
    test.run();

    END_TEST_MAIN;
}


