// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <cmath>
#include <limits>
#include <stdexcept>
#include <iostream>
#include <sstream>
#include <map>
#include <iset.hpp>
#include <utilities.hpp>
#include "test_utilities.hpp"
using namespace std;
using namespace hcsp;
using namespace hcsp::random;

// Random tester class for isets.
// It will inject random cases and check their outcomes.
class Tester {
  public:
    typedef iset::value_type       value_type;
    typedef iset::value_range_type value_range_type;
    typedef iset::size_type        size_type;
    typedef iset::iterator         iterator;
    typedef map<iset::value_type, unsigned> count_map_type;
    Tester() {}
    void Run();
    void test_rand();
  private:
    //
    // Testing methods
    //
    void SanityCheck();
    void test_empty_with_singleton();
    void TestSingeElement();
    void TestRemoveElements();
    void test_insert_all_of();
    void test_range();
    void test_intersect();
    void test_range_intersect();
    void test_big_range_intersect(const iset & elements,
                                  value_range_type & small_range,
                                  size_type skip);
    void test_complement();
    void test_contained_tail();
    void test_swap();
    //! Mutating a,b test if swapping them works correctly.
    void test_swap(iset &a, iset &b);
    void test_io();
    void validate_uniform_full_coverage(
        const iset &elements, const count_map_type & counts, unsigned samples);
    void get_cells_for_uniform_partial_coverage(
        const iset &domain, unsigned samples,
        size_type &cell_count, size_type & cell_size) const;
    void validate_uniform_partial_coverage(
        const iset &elements, const count_map_type & counts, unsigned samples);

    template <class Set>
    Set & erase(Set & var, value_type elem) const;

    template <class Set>
    Set & erase_all_of(Set & var, const iset & to_erase) const;
    iset & erase_range(iset & var, const value_range_type & range) const;
    iset & insert_range(iset & var, const value_range_type & range) const;
    iset & intersect_range(iset & var, const value_range_type & range) const;
    long long random_above(long long value, size_type skip);
    long long random_below(long long value, size_type skip);
    
    iset & assign_range(iset & var, const value_range_type & range) const;
}; // class Tester

template <class Set>
Set & Tester::erase(Set & var, value_type elem) const
{
    if (gen(4) != 0) {
        var.erase(elem);
        var.validate();
        return var;
    }
    size_type expected_size= var.size() - var.contains(elem);
    var.erase(elem);
    var.validate();
    T_ASSERT(var.size() == expected_size);
    return var;
}

template <class Set>
Set & Tester::erase_all_of(Set & var, const iset & to_erase) const
{
    if (gen(4) != 0) {
        var.validate();
        var.erase_all_of(to_erase);
        var.validate();
        return var;
    }
    size_type expected_size= var.size();
    for ( iterator it= to_erase.begin() ;
          it != to_erase.end() ;
          it.advance_range() ) {
        pair<value_type, value_type> range= it.current_range();
        
        iset range_var(range);
        size_type range_size= range_var.size();
        range_var.erase_all_of(var);
        expected_size -= range_size - range_var.size();
    }
    var.erase_all_of(to_erase);
    T_ASSERT(var.size() == expected_size);
    var.validate();
    T_ASSERT(!var.contains_any_of(to_erase));
    if (to_erase.empty())
        T_ASSERT(var.contains_all_of(to_erase));
    else
        T_ASSERT(!var.contains_all_of(to_erase));

    return var;
}

iset & Tester::erase_range(iset & var, const value_range_type & range) const
{
    iset compare_with= var;
    compare_with.erase_all_of(iset(range));

    var.erase(range);
    T_ASSERT(compare_with == var);

    return var;
}

iset & Tester::insert_range(iset & var, const value_range_type & range) const
{
    iset compare_with= var;
    compare_with.insert_all_of(iset(range));
    compare_with.validate();

    var.insert_all_of(range);
    var.validate();
    T_ASSERT(compare_with == var);

    return var;
}

iset& Tester::intersect_range(iset & var, const value_range_type & range) const
{
    iset compare_with= var;
    compare_with.intersect(iset(range));

    var.intersect(range);
    T_ASSERT(compare_with == var);

    return var;
}


iset & Tester::assign_range(iset & var, const value_range_type & range) const
{
    var.assign(range);
    T_ASSERT(var.begin().current_range() == range);
    T_ASSERT(var.begin().advance_range() == var.end());
    return var;
}

void Tester::test_swap(iset &a, iset &b)
{
    iset orig_a=a, orig_b= b;
    a.swap(b);
    T_ASSERT(a.size() == orig_b.size());
    T_ASSERT(b.size() == orig_a.size());
    T_ASSERT(a == orig_b);
    T_ASSERT(b == orig_a);
    a.swap(b);
    T_ASSERT(a == orig_a);
    T_ASSERT(b == orig_b);
}

void Tester::test_swap()
{
    iset a1, b1;
    test_swap(a1, b1);

    iset a2, b2=random_var();
    test_swap(a2, b2);
    
    iset a3=random_var(), b3;
    test_swap(a3, b3);

    iset a4=random_var(), b4=random_var();
    test_swap(a4, b4);
}


void Tester::Run()
{
    SanityCheck();
    test_empty_with_singleton();
    TestSingeElement();
    TestRemoveElements();
    test_insert_all_of();
    test_range();
    test_intersect();
    test_range_intersect();
    test_complement();
    test_contained_tail();
    test_swap();
    test_io();
}

void Tester::SanityCheck()
{
    pair<value_type, value_type> bounds= random_var_bounds();
    iset v1(bounds);
    v1.validate();
    iset v2(v1);
    v2.validate();
    v1= v1;
    iset v3;
    iset v4(move(v2));
    T_ASSERT(v4 == v1);
    v3= v1;
    v1= v4;
    T_ASSERT(v1 == v4);
    T_ASSERT(v1 == v3);
    T_ASSERT(iset(v1) == v3);
    pair<value_type, value_type> bounds20= random_var_bounds();
    iset v20(bounds20);
    v20.validate();
    if (bounds20 == bounds)
        T_ASSERT(v20 == v1);
    else
        T_ASSERT(v20 != v1);
    T_ASSERT(erase(iset()=v1, v1.min()) != v1);
    v1.validate();
    T_ASSERT(erase(iset()=v1, v1.max()) != v1);
    v1.validate();
    
    T_ASSERT(v3.min() == bounds.first);
    T_ASSERT(v3.max() == bounds.second);
    T_ASSERT(v1.min() == bounds.first);
    T_ASSERT(v1.max() == bounds.second);
    T_ASSERT(v1.max() == bounds.second);
    if (!are_all_values_legal_for_var()) {
        value_type outOfBound= random_var_out_of_bounds();
        bounds= make_pair(outOfBound,
                          random_var_maybe_out_of_bound(outOfBound));
        T_ASSERT_EXCEPTION(iset()=iset(bounds), out_of_range);
    }
}

void Tester::test_empty_with_singleton()
{
    T_ASSERT(iset().empty());
    iset v1= random_nonempty_var(true);
    T_ASSERT(iset(v1).clear().empty());
    T_ASSERT(!v1.empty());
    if (v1.everything())
        T_ASSERT(v1.max()==iset::MAX_DOMAIN && v1.min()==iset::MIN_DOMAIN);
    T_ASSERT(v1.size() > 0);
    do {
        T_ASSERT(v1.singleton() == (v1.size() == 1));
        value_type min= v1.min();
        v1.erase(min);
        T_ASSERT(!v1.everything());
        T_ASSERT(v1.empty() || v1.min() > min);
    } while (!v1.empty());
    T_ASSERT(!v1.singleton());
    v1= random_nonempty_var(true);
    do {
        T_ASSERT(v1.singleton() == (v1.size() == 1));
        value_type max= v1.max();
        v1.erase(max);
        T_ASSERT(v1.empty() || v1.max() < max);
    } while (!v1.empty());
    T_ASSERT(!v1.singleton());
    T_ASSERT(erase(iset()=iset(),
                   random_var_maybe_out_of_bound(0)).empty());
    value_type val= random_var_in_bounds();
    T_ASSERT(!iset().insert(val).empty());
    T_ASSERT(!iset().assign(val).empty());
    T_ASSERT(erase(iset().insert(val), val).empty());
    T_ASSERT(erase(random_var().assign(val), val).empty());
    T_ASSERT(!iset().insert(val).empty());
}

    
// Testing:
//     void assign_element(value_type element);
//     void insert(value_type element);
//     void erase(value_type element);
//     bool IsContaining(value_type element) const
void Tester::TestSingeElement()
{
    value_type elem1= random_var_in_bounds();
    iset v1=random_var();
    iset v2= v1;
    
    T_ASSERT(iset(v1).assign(elem1).contains(elem1));
    T_ASSERT(iset(v1).insert(elem1).contains(elem1));
    T_ASSERT(iset(v1).assign(elem1)==iset(elem1));
    if (!are_all_values_legal_for_var())
        T_ASSERT_EXCEPTION(v1.assign(random_var_out_of_bounds()),
                           out_of_range);
    T_ASSERT(v1 == v2);
    if (!are_all_values_legal_for_var())
        T_ASSERT_EXCEPTION(v1.insert(random_var_out_of_bounds()),out_of_range);
    T_ASSERT(v1 == v2);
    iset v3(v1);
    v3.erase(elem1);
    T_ASSERT(!v1.contains(elem1) || v3 != v1);
    if ( v1.contains(elem1) )
        v3.insert(elem1);
    T_ASSERT(v1 == v3);
    v3.insert(elem1);
    if ( v1.contains(elem1) )
        T_ASSERT(v1 == v3);
    else
        T_ASSERT(v1 != v3);

    T_ASSERT(!iset().assign(elem1).empty());
    T_ASSERT(erase(iset().assign(elem1), elem1).empty());
    T_ASSERT(!v1.empty() || !v1.contains(elem1));
}

void Tester::TestRemoveElements()
{
    iset v1=random_var(true);
    bool wasEmpty= v1.empty();
    iset v2= v1;
    erase_all_of(v2, v1);
    T_ASSERT(v2.empty());
    T_ASSERT(erase_all_of(iset()=v1, v1).empty());
    T_ASSERT(wasEmpty || !v1.empty());

    v2= random_var();
    v2.validate();
    if ( erase_all_of(iset()=v1, v2) == v1) {
        // The sets are independent (maybe one of them is empty)
        T_ASSERT(erase_all_of(iset()=v2, v1) == v2);
        T_ASSERT(v1.empty() || iset(v2).insert_all_of(v1) != v2);
    }
    T_ASSERT(erase_all_of(iset(v1).insert_all_of(v2), v2) ==
             iset(v1).erase_all_of(v2));
    T_ASSERT(iset(v1).erase_all_of(v2).insert_all_of(v2) ==
             iset(v1).insert_all_of(v2));
    T_ASSERT(erase_all_of(v2, iset(v1).erase_all_of(v2)) == v2);

    if (!v1.empty()) {
        iset singletonMin(v1.min());
        iset singletonMax(v1.max());
        T_ASSERT(erase_all_of(iset()=v1, singletonMin) != v1);
        T_ASSERT(erase_all_of(iset()=v1, singletonMax) != v1);
    }
}

void Tester::test_insert_all_of()
{
    iset v1= random_var(true);
    iset v2= random_var(true);
    iset v3= v1;
    v3.insert_all_of(v2);
    v3.validate();
    v2.validate();
    T_ASSERT(v3.contains_all_of(v2));
    T_ASSERT(v3.contains_all_of(v1));
    if (v2.empty())
        T_ASSERT(!v3.contains_any_of(v2));
    else
        T_ASSERT(v3.contains_any_of(v2));
    if (v1.empty())
        T_ASSERT(!v3.contains_any_of(v1));
    else
        T_ASSERT(v3.contains_any_of(v1));

    T_ASSERT(v3.size()  -v2.size() <= v1.size() );
    T_ASSERT(iset(v3).erase_all_of(v1).size() <= v2.size());
    T_ASSERT(iset(v3).erase_all_of(v2).size() <= v1.size());
    T_ASSERT(v3 == iset(v2).insert_all_of(v3));
    T_ASSERT(v3 == iset(v1).erase_all_of(v2).insert_all_of(v2));

    // insertion (i.e. v3) does not contain unknown elements.
    T_ASSERT(iset(v3).erase_all_of(v2) == iset(v1).erase_all_of(v2));
    T_ASSERT(iset(v3).erase_all_of(v1) == iset(v2).erase_all_of(v1));

    // insertion (i.e. v3) contains all elements of v1 and v2.
    T_ASSERT(iset(v1).erase_all_of(v3).empty());
    T_ASSERT(iset(v2).erase_all_of(v3).empty());
}

void Tester::test_range()
{
    const iset v1 = random_var();
    iset v2= v1;
    v2.validate();
    value_range_type range = random_var_bounds();
    insert_range(v2, range);
    v2.validate();

    T_ASSERT(intersect_range(iset()=v2, range) == iset(range));
    intersect_range(iset()=v1, range);

    T_ASSERT(v1.contains_all_of( erase_range(iset()=v2, range) ));
    assign_range(iset()=v1, range);
}

void Tester::test_intersect()
{
    const iset v1= random_var(true);
    const iset v2= random_var(true);

    T_ASSERT( iset(v1).intersect(v2) == iset(v2).intersect(v1) );

    // Check that intersect() does not erase stuff which is in v2 and v1
    T_ASSERT( erase_all_of(iset()= v1, iset(v1).intersect(v2)) ==
              erase_all_of(iset()= v1, iset()= v2) );

    // Check that intersect() does not add junk
    T_ASSERT( erase_all_of(iset(v1).intersect(v2), v1).empty() );
}

void Tester::test_range_intersect()
{
    const iset v1= random_var(true);
    value_range_type v2= random_var_bounds();

    T_ASSERT( iset(v1).intersect(v2) == iset().insert(v2).intersect(v1) );

    size_type skip;
    size_type diff= (size_type)v2.second-v2.first;
    if (diff == 0)
        skip= gen();
    else if (diff <= 2 || v2.first > v2.second)
        skip= range(1U,diff);
    else {
        switch (gen(20)) {
          case 0: skip= 2;    break;
          case 1: skip= diff; break;
          default:
            skip= range(2U, diff-1);
        }
        v2.second -= diff % skip;
    }
    T_ASSERT( iset(v1).intersect(v2, skip) ==
              iset().insert(v2, skip).intersect(v1) );

    test_big_range_intersect(v1, v2, skip);
}

long long Tester::random_below(long long value, size_type skip)
{
    unsigned long long diff;
    diff= value - (unsigned long long)numeric_limits<long long>::min();
    unsigned long long max_steps= diff/skip;
    if (gen(4) == 0) {
        if (gen(2) == 0)
            return value - skip;
        else
            return value - max_steps*skip;
    } else {
        // make it random in range between slightly less than the original
        //  and the smallest possible
        unsigned long long steps= gen();
        steps = (steps << sizeof(gen())*8) | gen();
        steps = (steps % max_steps) + 1;
        return value - steps*skip;
    }
}

long long Tester::random_above(long long value, size_type skip)
{
    unsigned long long diff;
    diff= (unsigned long long)numeric_limits<long long>::max() - value;
    unsigned long long max_steps= diff/skip;
    if (gen(4) == 0) {
        if (gen(2) == 0)
            return value + skip;
        else
            return value + max_steps*skip;
    } else {
        // make it random in range between slightly more than 
        // iset::MAX_DOMAIN and the biggest possible
        unsigned long long steps= gen();
        steps = (steps << sizeof(gen())*8) | gen();
        steps = (steps % max_steps) + 1;
        return value + steps*skip;
    }
}

void Tester::test_big_range_intersect(const iset & elements,
                                      value_range_type & small_range,
                                      size_type skip)
{
    pair<long long, long long> big_range(small_range);
    T_ASSERT( iset(elements).intersect(big_range, skip) ==
              iset().insert(small_range, skip).intersect(elements) );
    if (skip == 0)
        return;
    
    if (big_range.first-skip >= iset::MIN_DOMAIN || gen(10) == 0)
        ; // keep big_range.first unchanged
    else
        big_range.first= random_below(small_range.first, skip);

    
    if (big_range.second+skip <= iset::MAX_DOMAIN || gen(10) == 0)
        ; // keep big_range.second unchanged
    else
        big_range.second= random_above(small_range.second, skip);
    
    T_ASSERT( iset(elements).intersect(big_range, skip) ==
              iset().insert(small_range, skip).intersect(elements) );
}

void Tester::test_rand()
{
    T_ASSERT_EXCEPTION(iset().rand(), runtime_error);
    const iset v1= random_nonempty_var();
    count_map_type counts;
    static const unsigned SAMPLES= 10000;
    for (unsigned i=0 ; i < SAMPLES ; ++i) {
        value_type val= v1.rand();
        T_ASSERT(v1.contains(val));
        count_map_type::iterator entry_it= counts.lower_bound(val);
        if (entry_it == counts.end() || entry_it->first != val)
            counts.insert(entry_it, count_map_type::value_type(val, 1));
        else
            ++entry_it->second;
    }
    bool expect_full_coverage;
    if (!v1.everything())
        expect_full_coverage=  (v1.size() < sqrt(SAMPLES/10));
    else if (0U+iset::MAX_DOMAIN < SAMPLES)
        expect_full_coverage=  (v1.size() <= sqrt(SAMPLES/10));
    else
        expect_full_coverage= false;
    if (expect_full_coverage)
        validate_uniform_full_coverage(v1, counts, SAMPLES);
    else
        validate_uniform_partial_coverage(v1, counts, SAMPLES);
}

void Tester::validate_uniform_full_coverage(
        const iset & domain, const count_map_type & counts, unsigned samples)
{
    for ( iset::iterator it= domain.begin() ; it!=domain.end(); ++it) {
        unsigned count =
            counts.find(*it)==counts.end() ? 0 : counts.find(*it)->second;
        // Tolerance is 50% above average, and 33% below average.
        // The probability to get anything outside this tolerance range,
        // is virtually zero.
        T_ASSERT(count <= samples*3/2/domain.size());
        T_ASSERT(count >= samples*2/3/domain.size());
    }
}

void Tester::get_cells_for_uniform_partial_coverage(
    const iset &domain, unsigned samples,
    size_type &cell_count, size_type & cell_size) const
{
    cell_count= static_cast<unsigned>(sqrt(samples/10)+0.001);
    if (domain.size()-1 <= cell_count)
        cell_count= domain.size();
    if ( domain.size() > 4 && (size_t)cell_count > domain.size()/4 )
        cell_count= domain.size()/4;

    iset::value_type cell_count_as_value= cell_count;
    if ( domain.everything())
        cell_size= ((iset::size_type)iset::MAX_DOMAIN/cell_count_as_value -
                    iset::MIN_DOMAIN/cell_count_as_value) + 1;
    else
        cell_size= (domain.size()-1) / cell_count + 1;
}

void Tester::validate_uniform_partial_coverage(
        const iset &domain, const count_map_type & counts, unsigned samples)
{
    iset next_domain;
    size_type cell_count, cell_size;
    get_cells_for_uniform_partial_coverage(domain, samples,
                                           cell_count, cell_size);
    const size_type expected_in_cell= samples/cell_count;
    size_t values_done=0;

    count_map_type::const_iterator count_iter= counts.begin();
    T_ASSERT(count_iter != counts.end());
    T_ASSERT(count_iter->first >= domain.min());
    unsigned cells_done=0;
    for ( iset::iterator it= domain.begin() ;
          it!=domain.end() ;
          it.advance(cell_size), ++cells_done, values_done+=cell_size ) {
        value_type last= *--iterator(iterator(it).advance(cell_size));
        unsigned count_in_cell=0 ;
        while ( count_iter != counts.end() && count_iter->first <= last ) {
            count_in_cell += count_iter->second;
            ++count_iter;
        }
        
        // Tolerance is 50% above average, and 33% below average.
        // The probability to get anything outside this tolerance range,
        // is virtually zero.
        T_ASSERT(count_in_cell <= expected_in_cell*3);
        T_ASSERT(count_in_cell >= expected_in_cell/3);
        unsigned cells_left= cell_count-cells_done-1;
        if (cells_left != 0) {
            if ( domain.everything())
                cell_size= (iset::MAX_DOMAIN-last)/cells_left;
            else
                cell_size= (domain.size()-values_done) / cells_left;
        }
    }
}
 
void Tester::test_complement()
{
    const iset v1= random_var();
    const iset everything(iset::value_range_type(iset::MIN_DOMAIN,
                                                 iset::MAX_DOMAIN));
    T_ASSERT( iset(v1).complement() == iset(everything).erase_all_of(v1));
}

void Tester::test_contained_tail()
{
    // 
    // First, generate 'work', 'other', and 'last' with potentially common
    // ending.
    //
    iset work= random_var();
    iset other= random_var();
    iset common= work;
    common.insert_all_of(other);
    iset::value_range_type positions(iset::MIN_DOMAIN, iset::MAX_DOMAIN);
    if (!common.empty())
        positions= make_pair(common.rand(), common.rand());
    if (positions.first > positions.second)
        swap(positions.first, positions.second);

    bool enforce_common_tail= (!common.empty() && gen(2) == 0);
    
    if (enforce_common_tail) {
        common.intersect(make_pair(positions.first, positions.second));
        work.insert_all_of(common);
        other.insert_all_of(common);
    }
    bool start_with_end= common.empty() || (gen(2) == 0);
    iset::iterator last; 
    if (start_with_end) 
        last= work.end();
    else
        last= work.lower_bound(positions.second);

    // 
    // Finally, test if find_last_contained works properly.
    //
    iset::iterator it= work.find_contained_tail(other, last);
    if (it == work.end()) {
        T_ASSERT(last == work.end());
        T_ASSERT(work.empty() || !other.contains(work.max()));
    } else if (it == last) {
        if (last != work.begin())
            T_ASSERT(!other.contains(*--iset::iterator(last)));
    } else {
        iset work_tail= work;
        work_tail.intersect(make_pair(*it, work.max()));
        if (last != work.end())
            work_tail.erase(make_pair(*last, work.max()));
        T_ASSERT(other.contains_all_of(work_tail));
        if (it != work.begin()) {
            --it;
            T_ASSERT(!other.contains(*it));
        }
    }
}


void Tester::test_io()
{
    stringstream buffer;
    const iset orig= random_var();
    iset result= random_var();
    buffer << orig;
    buffer >> result;
    T_ASSERT(static_cast<istream&>(buffer));
    T_ASSERT(orig == result);
    T_ASSERT(iset(buffer.str()) == orig);
    char ch= range(numeric_limits<char>::min(), numeric_limits<char>::max());
    if ( ch == ',' || ch == ':' ||
         ch == '+' || ch == '-' || (ch >= '0' && ch <= '9'))
        return;
    buffer << ch << range(iset::MIN_DOMAIN,iset::MAX_DOMAIN);
    buffer.seekg(0);
    buffer.clear();
    buffer >> result;
    T_ASSERT(result == orig);
    T_ASSERT_EXCEPTION(iset(buffer.str()), parse_error);
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    cout << "Seed is " << set_seed(argc, argv) << endl;
    for (unsigned i=0 ; i < 3000 ; ++i) {
        Tester tester;
        tester.Run();
    }
    for (unsigned i=0 ; i < 50 ; ++i)
        Tester().test_rand();
    
    END_TEST_MAIN;
}

