// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
// Tests for the classes variable::iterator, and variable::reverse_iterator
// 
#include <limits>
#include <stdexcept>
#include <iostream>
#include <vector>
#include <algorithm>
#include <iset.hpp>
#include <utilities.hpp>
#include "test_utilities.hpp"
using namespace std;
using namespace hcsp;
using namespace hcsp::random;

// Random tester class for isets.
// It will inject random cases and check their outcomes.
class tester {
  public:
    typedef iset::value_type             value_type;
    typedef iset::value_range_type       value_range_type;
    typedef iset::size_type              size_type;
    typedef iset::iterator               iterator;
    typedef iset::const_iterator         const_iterator;
    typedef iset::reverse_iterator       reverse_iterator;
    typedef iset::const_reverse_iterator const_reverse_iterator;
    tester() {}
    void run();
  private:
    void erase(iset & var, value_type val) const;
    iset & erase_range(iset & var, const value_range_type & range) const;
    void reconstruct_by_iteration() const;
    void reconstruct_by_iteration_forward() const;
    void reconstruct_by_iteration_backward() const;
    void clear_by_iteration() const;
    void test_bound_search() const;
    void test_underflowing_bound() const;
    void test_contained_bound() const;
    void test_overflowing_bound() const;
    void test_range_equiv() const;
    void test_mixed_range_equiv() const;
    static void random_stride(bool small_stride,
                              value_range_type & interval, size_type & skip,
                              value_type lower=iset::MIN_DOMAIN,
                              value_type upper=iset::MAX_DOMAIN);
    void test_forward_stride() const;
    void test_forward_stride(const iset & elements,
                             const value_range_type & interval1,
                             size_type stride1,
                             const value_range_type & interval2,
                             size_type stride2=0) const;
    iterator test_mid_forward_stride(const iset & elements,
                                     const value_range_type &interval,
                                     size_type stride) const;
    //! Randomly, try to go one directions and then back.
    //! This has the prospect of getting the iterator into different states,
    //! where iterator.m_current_value may get MIN_DOMAIN.
    //! This may happen right after ++ which moves to a new range.
    void perturb_iterator(const iset & elements, iterator & it) const;
    
    
    void test_backward_stride() const;
    void test_backward_stride(const iset & elements,
                              const value_range_type & interval1,
                              size_type stride1,
                              const value_range_type & interval2,
                              size_type stride2) const;
    iterator test_mid_backward_stride(const iset & elements,
                                      const value_range_type &interval,
                                      size_type stride) const;
    void test_reinserting_stride() const;
    void test_mixed_retract_range() const;
    void test_range_compact() const;
    void test_current_range_compact(iterator iter, const iset &var) const;
    template <class Iter, class Vector>
    static void random_iters(Iter begin, Iter end, Vector & result);
    template <class Iter>
    void invalidated_erase_values(const vector<Iter> & iters, iset & var)const;
    
    void erase_values(const vector<iset::reverse_iterator> & iters, iset &var)
        const
    { invalidated_erase_values(iters, var); }
    
    void erase_values(const vector<iset::iterator> & iters, iset &var) const;
    
    void test_advance() const;
    size_type random_advance_size(const iset & domain,
                                  const iset & remaining,
                                  bool also_test_out_of_range) const;
    
}; // class Tester

template <class Iter, class Vector>
void tester::random_iters(Iter begin, Iter end, Vector & result_iters)
{
    result_iters.clear();
    for (Iter it= begin ; it != end ; ++it)
        result_iters.push_back(it);
    randomizer rand;
    random_shuffle(result_iters.begin(), result_iters.end(), rand);
}

template <class Iter>
void tester::invalidated_erase_values(const vector<Iter> & iters, iset & var)
    const
{
    vector<iset::value_type> values;
    for (size_t i=0 ; i != iters.size() ; ++i)
        values.push_back(*iters[i]);
    
    for (size_t i=0 ; i != values.size() ; ++i)
        erase(var, values[i]);
}

void tester::erase_values(const vector<iset::iterator> & iters, iset & var)
    const
{
    if (iset::iterator::INVALIDATED_BY_ERASE) {
        invalidated_erase_values(iters, var);
    } else {
        for (size_t i=0 ; i != iters.size() ; ++i)
            erase(var, *iters[i]);   
    }
}

void tester::erase(iset & var, value_type val) const
{
    T_ASSERT(var.empty() == (var.size()==0));

    iset::size_type expected_size= var.size();
    if (var.contains(val))
        --expected_size;
    var.erase(val);
    T_ASSERT(var.size() == expected_size);
}

iset &  tester::erase_range(iset & var, const value_range_type & range) const
{
    iset compare_with= var;
    compare_with.erase_all_of(iset(range));

    var.erase(range);
    T_ASSERT(compare_with == var);

    return var;
}

void tester::run()
{
    reconstruct_by_iteration();
    clear_by_iteration();
    test_bound_search();
    test_range_equiv();
    test_mixed_range_equiv();
    test_forward_stride();
    test_backward_stride();
    test_reinserting_stride();
    test_mixed_retract_range();
    test_range_compact();
    test_advance();
}

void tester::reconstruct_by_iteration_forward() const
{
    iset compared;
    iset original = random_var(true);
    for (iterator it= original.begin() ; it != original.end() ; ++it)
        compared.insert(*it);
    T_ASSERT(compared == original);
}

void tester::reconstruct_by_iteration_backward() const
{
    iset compared;
    iset original = random_var(true);
    for (reverse_iterator it= original.rbegin() ; it != original.rend() ; ++it)
        compared.insert(*it);
    T_ASSERT(compared == original);
}

void tester::reconstruct_by_iteration() const
{
    reconstruct_by_iteration_forward();

    iset compared;
    iset original = random_var(true);
    vector<iterator> iters;
    random_iters(original.begin(), original.end(), iters);
    for (size_t i=0 ; i != iters.size() ; ++i)
        compared.insert(*iters[i]);

    T_ASSERT(compared == original);

    vector<reverse_iterator> riters;
    random_iters(original.rbegin(), original.rend(), riters);
    for (size_t i=0 ; i != riters.size() ; ++i)
        compared.insert(*riters[i]);

    T_ASSERT(compared == original);
}

void tester::clear_by_iteration() const
{
    iset var = random_var(true);
    for (iterator it= var.begin(), next=it ; it != var.end() ; it=next) {
        ++next;
        if (iset::iterator::INVALIDATED_BY_ERASE) {
            next= var.erase(it);
        } else {
            iterator next2 HCSP_UNUSED= var.erase(it);
            HCSP_ASSERT(next == next2);
        }
    }
    T_ASSERT(var.empty());

    var = random_var(true);
    for (reverse_iterator it= var.rbegin() ; it!=var.rend(); ){
        var.erase(*it);
        it= var.rbegin();
    }
    T_ASSERT(var.empty());

    var = random_var(true);
    vector<iterator> iters;
    random_iters(var.begin(), var.end(), iters);
    erase_values(iters, var);
    T_ASSERT(var.empty());

    var = random_var(true);
    vector<reverse_iterator> riters;
    random_iters(var.rbegin(), var.rend(), riters);
    erase_values(riters, var); 
    
    T_ASSERT(var.empty());
}

void tester::test_underflowing_bound() const
{
    iset var= random_var();
    
    iset::value_type pre_min=numeric_limits<iset::value_type>::min();
    value_type var_min= var.empty() ? 0 : var.min();
    if (pre_min == var_min)
        return;
    
    if (pre_min+1 != var_min) {
        switch (gen(10)) {
          case 0:                     break; // value_type's min
          case 1: ++pre_min;          break; // almost value_type's min
          case 2: pre_min= var_min-1; break;
          case 3: pre_min= var_min-2; break;
          default:
            pre_min=range(pre_min, var_min-1); 
        }
    }
    T_ASSERT(var.upper_bound(pre_min) == var.begin());
    T_ASSERT(var.lower_bound(pre_min) == var.begin());
}

void tester::test_overflowing_bound() const
{
    iset var= random_var();
    
    iset::value_type post_max=numeric_limits<iset::value_type>::max();
    value_type var_max= var.empty() ? 0 : var.max();
    if (post_max == var_max)
        return;
    if (post_max-1 != var_max) {
        switch (gen(10)) {
          case 0:                      break; // value_type's max
          case 1: --post_max;          break; // almost value_type's max
          case 2: post_max= var_max+1; break;
          case 3: post_max= var_max+2; break;
          default:
            post_max=range(var_max+1, post_max); 
        }
    }
    T_ASSERT(var.upper_bound(post_max) == var.end());
    T_ASSERT(var.lower_bound(post_max) == var.end());
}

void tester::test_contained_bound() const
{
    iset var = random_nonempty_var(true);

    for (iterator it= var.begin() ; it != var.end() ; ++it) {
        value_type val= *it;
        T_ASSERT(var.lower_bound(val) == it);
        if ( val != numeric_limits<value_type>::min() )
            T_ASSERT(var.upper_bound(val-1) == it);
        T_ASSERT(var.upper_bound(val) != it);
    }
}

void tester::test_bound_search() const
{
    iset var = random_nonempty_var();

    for (unsigned i=0 ; i < 100 ; ++i) {
        value_type val = range(var.min(), var.max());
        iset subset= var;
        erase_range(subset, make_pair(var.min(), val));
        if (var.contains(val))
            T_ASSERT(*var.lower_bound(val) == val);
        iterator upper= var.upper_bound(val);
        if (subset.empty())
            T_ASSERT(upper == var.end());
        else
            T_ASSERT(*upper == subset.min());
    }
    test_contained_bound();
    test_underflowing_bound();
    test_overflowing_bound();
}

void tester::test_range_equiv() const
{
    iset var=random_var();
    iset compared;
    for (iterator it=var.begin(); it!=var.end(); it.advance_range())
        compared.insert(it.current_range());
    T_ASSERT(compared == var);
    
    iset compared_reverse;
    for (iterator it=var.end(); it!=var.begin(); ) {
        it.retract_range();        
        compared_reverse.insert(it.current_range());
    }
    T_ASSERT(compared_reverse == var);
}

void tester::test_mixed_range_equiv() const
{
    iset var=random_var(true);
    iset compared;
    unsigned left_ranges= count_ranges(var);
    unsigned ranges_to_split= range(0U, left_ranges);
    unsigned left_complete_splits= range(0U, ranges_to_split);
    
    for (iterator it=var.begin(); it!=var.end(); --left_ranges) {
        bool split_access= (ranges_to_split > gen(left_ranges));
        bool complete_split= false;
        if  (split_access) {
            complete_split= (left_complete_splits > gen(ranges_to_split));
            bool want_split;
            value_type last;
            do {
                last= *it;
                compared.insert(last);
                ++it;
                want_split = complete_split || gen(2) == 0;
            } while (it != var.end() && *it == last+1 && want_split);
            --ranges_to_split;
            left_complete_splits -= complete_split;
            if (it == var.end() || *it != last+1)
                continue;
        }
        // range access
        compared.insert(it.current_range());
        it.advance_range();
    }
    T_ASSERT(compared == var);
    T_ASSERT(left_complete_splits == 0);
    T_ASSERT(ranges_to_split == 0);
    T_ASSERT(left_ranges == 0);
}

// static
void tester::random_stride(bool small_stride, value_range_type & interval,
                           size_type & stride,
                           value_type lower, value_type upper)
{
    interval= random_var_bounds(small_stride, lower, upper);
    size_type max_skip=(size_type)interval.second-interval.first;
    if (max_skip == 0)
        max_skip= 1;
    switch (gen(6)) {
      case 0: stride=1;         break;
      case 1: stride= max_skip; break;
      default:
        stride= random::range(1U, max_skip);
        size_type diff= max_skip % stride;
        if (diff != 0) {
            size_type taken= random::gen(diff+1);
            interval.first += taken;
            interval.second -= diff-taken;
        }
    }
}

void tester::test_forward_stride() const
{
    iset elements;
    value_range_type interval1, interval2;
    size_type stride1=0, stride2=0;
    random_stride(gen(10)== 0, interval1, stride1);
    elements.insert(interval1, stride1);

    size_type diff= (size_type)interval1.second-interval1.first;
    size_type count= diff/stride1+1;
    if ( count < 1000 && interval1.first != interval1.second &&
         stride1 < (size_type)iset::MAX_DOMAIN-interval1.second) {
        do {
            random_stride(true, interval2, stride2, interval1.second+1,
                          iset::MAX_DOMAIN);
        } while ( (size_type)interval2.first-interval1.second == stride1);
        elements.insert(interval2, stride2);
    }
    test_forward_stride(elements, interval1, stride1, interval2, stride2);
}

iset::iterator tester::test_mid_forward_stride(const iset & elements,
                                               const value_range_type &interval,
                                               size_type stride) const
{
    value_range_type got_interval;
    size_type got_stride;
    size_type diff= ((size_type)interval.second-interval.first)/stride;
    size_type index_first= gen(diff);
    size_type index_second= gen(2) ? diff : range(index_first, diff);
    value_type first= interval.first+index_first*stride;
    value_type second= interval.first+index_second*stride;
    iterator it= elements.lower_bound(first);
    iterator end= elements.upper_bound(second);
    perturb_iterator(elements, it);
    perturb_iterator(elements, end);
    if (second == interval.second) {
        while (end != elements.end() && gen(4))
            ++end;
    }
    T_ASSERT(it != elements.end() && *it == first);
    it= it.find_stride(end, got_interval, got_stride);
    T_ASSERT(got_interval.first == first);
    T_ASSERT(got_interval.second == second);
    if (first != got_interval.second)
        T_ASSERT(got_stride == stride);
    else
        T_ASSERT(got_stride == 1);
    if (it == elements.end())
        T_ASSERT(second == elements.max());
    else
        T_ASSERT(second < *it && second == *--iterator(it) );
    if (second != interval.second)
        it.advance(diff-index_second);
    return it;
}

void tester::perturb_iterator(const iset & elements, iterator & it) const
{
    if (it == elements.end())
        return;
    if (gen(2) == 0)
        return;
    if (it == elements.begin() || gen(2)) {
        ++it;
        --it;
    } else {
        --it;
        ++it;
    }
}


void tester::test_forward_stride(const iset & elements,
                                 const value_range_type & interval1,
                                 size_type stride1,
                                 const value_range_type & interval2,
                                 size_type stride2) const
{
    value_range_type got_interval1, got_interval2;
    size_type got_stride1=0, got_stride2=0;
    iterator it=elements.begin();
    it= it.find_stride(elements.end(), got_interval1, got_stride1);
    T_ASSERT(stride1== got_stride1);
    T_ASSERT(got_interval1 == interval1);
    T_ASSERT(it.find_back_stride(elements.begin(), got_interval1, got_stride1)
             == elements.begin());
    T_ASSERT(stride1== got_stride1);
    T_ASSERT(got_interval1 == interval1);
    if ( interval1.first != interval1.second && gen(2) == 0)
        it= test_mid_forward_stride(elements, interval1, stride1);
   
    if (stride2 > 0) {
        T_ASSERT(it != elements.end());
        it= it.find_stride(elements.end(), got_interval2, got_stride2);
        T_ASSERT(stride2== got_stride2);
        T_ASSERT(got_interval2 == interval2);
        if (interval2.first != interval2.second && gen(2) == 0)
            it= test_mid_forward_stride(elements, interval2, stride2);
    }
    T_ASSERT(it == elements.end());
}

void tester::test_backward_stride() const
{
    iset elements;
    value_range_type interval1, interval2;
    size_type stride1=0, stride2=0;
    random_stride(gen(10)== 0, interval1, stride1);
    elements.insert(interval1, stride1);

    size_type diff= (size_type)interval1.second-interval1.first;
    size_type count= diff/stride1+1;
    if ( count < 1000 && interval1.first != interval1.second &&
         stride1 < (size_type)interval1.first-iset::MIN_DOMAIN ) {
        do {
            random_stride(true, interval2, stride2, iset::MIN_DOMAIN,
                          interval1.first-1);
        } while ( (size_type)interval1.first-interval2.second == stride1);
        elements.insert(interval2, stride2);
    }
    test_backward_stride(elements, interval1, stride1, interval2, stride2);
}

iset::iterator tester::test_mid_backward_stride(
    const iset & elements, const value_range_type &interval,
    size_type stride) const
{
    value_range_type got_interval;
    size_type got_stride;
    size_type diff= ((size_type)interval.second-interval.first)/stride;
    size_type index_first= gen(diff);
    size_type index_second= gen(2) ? diff : range(index_first, diff);
    value_type first= interval.first+index_first*stride;
    value_type second= interval.first+index_second*stride;
    iterator begin= elements.lower_bound(first);
    iterator it= elements.upper_bound(second);
    perturb_iterator(elements, it);
    if (first == interval.first && first > elements.min() && first != second) {
        while (begin != elements.begin() && gen(4))
            --begin;
    }
    perturb_iterator(elements, begin);
    if (second == elements.max())
        T_ASSERT(it == elements.end());
    else
        T_ASSERT(*it > second && *--iterator(it) == second);
        
    it= it.find_back_stride(begin, got_interval, got_stride);
    T_ASSERT(got_interval.first == first);
    T_ASSERT(got_interval.second == second);
    if (first != got_interval.second)
        T_ASSERT(got_stride == stride);
    else
        T_ASSERT(got_stride == 1);
    T_ASSERT(first == *it);
    if (first != interval.first)
        it= elements.lower_bound(interval.first);
    return it;
}

void tester::test_backward_stride(const iset & elements,
                                  const value_range_type & interval1,
                                  size_type stride1,
                                  const value_range_type & interval2,
                                  size_type stride2) const
{
    value_range_type got_interval1, got_interval2;
    size_type got_stride1=0, got_stride2=0;
    iterator it=elements.end();
    it= it.find_back_stride(elements.begin(), got_interval1, got_stride1);
    T_ASSERT(stride1== got_stride1);
    T_ASSERT(got_interval1 == interval1);
    T_ASSERT(it.find_stride(elements.end(), got_interval1, got_stride1)
             == elements.end());
    T_ASSERT(stride1== got_stride1);
    T_ASSERT(got_interval1 == interval1);
    if ( interval1.first != interval1.second && gen(2) == 0)
        it= test_mid_backward_stride(elements, interval1, stride1);
    if (stride2 > 0) {
        T_ASSERT(it != elements.begin());
        it= it.find_back_stride(elements.begin(), got_interval2, got_stride2);
        T_ASSERT(stride2== got_stride2);
        T_ASSERT(got_interval2 == interval2);
         if (interval2.first != interval2.second && gen(2) == 0)
            it= test_mid_backward_stride(elements, interval2, stride2);
   }
    T_ASSERT(it == elements.begin());
}

void tester::test_reinserting_stride() const
{
    iset elements= random_var(gen(20) == 0);
    iset reinserted;
    
    // First test going forward
    for (iterator it= elements.begin(); it != elements.end() ; ) {
        value_range_type interval;
        size_type stride;
        it= it.find_stride(elements.end(), interval, stride);
        reinserted.insert(interval, stride);
    }
    T_ASSERT(reinserted == elements);
    reinserted.clear();

    // Finally test going backward
    for (iterator it= elements.end(); it != elements.begin() ; ) {
        value_range_type interval;
        size_type stride;
        it= it.find_back_stride(elements.begin(), interval, stride);
        reinserted.insert(interval, stride);
    }
    T_ASSERT(reinserted == elements);
}

void tester::test_mixed_retract_range() const
{
    iset elements= random_var(gen(20) == 0);
    for (iterator it= elements.end() ; it != elements.begin() ; ) {
        it.retract_range();
        value_range_type range= it.current_range();
        if (range.first == range.second)
            continue;
        value_type middle;
        switch (gen(20)) {
          case 0: middle= range.second;  break;
          case 1: middle= range.first+1; break;
          default:
            middle= random::range(range.first+1, range.second);
        }
        it.advance((size_type)middle-range.first);
        value_range_type mid_range= it.current_range();
        T_ASSERT(mid_range.first == middle);
        T_ASSERT(mid_range.second == range.second);
        it.retract_range();
        value_range_type retry_range= it.current_range();
        T_ASSERT(retry_range == range);
    }
}

void tester::test_range_compact() const
{
    iset var=random_nonempty_var();
    
    for (iterator it= var.begin(); it != var.end() ; it.advance_range())
        test_current_range_compact(it, var);

    for (iterator it= var.end(); it != var.begin() ; ) {
        it.retract_range();
        test_current_range_compact(it, var);
    }
    
}

void tester::test_current_range_compact(iterator iter, const iset &var) const
{
    value_type prev_end= iter->second;
    iter.advance_range();
    T_ASSERT(iter==var.end() || prev_end+1 < iter->first);
}

iset::size_type tester::random_advance_size(const iset & domain,
                                            const iset & remaining,
                                            bool also_test_out_of_range) const
{
    size_type lower_bound;
    if (domain.everything())
        lower_bound= iset::MAX_DOMAIN/1000-iset::MIN_DOMAIN/1000+1;
    else
        lower_bound= domain.size()/1000+1;

    if (!remaining.everything())
        lower_bound= min(lower_bound, remaining.size());
    
    if (gen(2) == 0)
        return range(size_type(1), lower_bound);
    
    if ( iset::iterator::CHECKS_BOUNDS
         && also_test_out_of_range && gen(100)==0) {
        if (domain.everything())
            return gen();
        else
            return range(lower_bound, remaining.size()+1);
    } 
    if ( remaining.everything() &&
         iset::MAX_DOMAIN == numeric_limits<value_type>::max() &&
         iset::MIN_DOMAIN == numeric_limits<value_type>::min())
        return range(size_type(1), numeric_limits<size_type>::max());
    else
        return range(lower_bound, remaining.size());
}

void tester::test_advance() const
{
    const iset domain= random_nonempty_var();
    iset remaining=domain;
    bool test_out_of_range= gen(100)==0;
    for (iterator next, it= domain.begin() ; it != domain.end() ; it=next) {
        T_ASSERT( it == iterator(it).advance(0) );
        size_type count= random_advance_size(domain, remaining,
                                             test_out_of_range);
        next= iterator(it).advance(count);
        value_type skipped_upper;
        if ( next == domain.end() )
            skipped_upper= iset::MAX_DOMAIN;
        else {
            T_ASSERT(domain.contains(*it));
            skipped_upper= *next-1;
        }
        // for remaining.everything() remaining.size() might be illegal
        size_type before;
        if (remaining.everything()) {
            // A potential overflow will lead to zero, which will work fine
            // with the comming asserts.
            before=
                static_cast<size_type>(iset::MAX_DOMAIN)-iset::MIN_DOMAIN+1;
        } else
            before= remaining.size();
        remaining.erase_all_of(value_range_type(*it, skipped_upper));
        size_type after= remaining.size();
        if (next == domain.end())
            T_ASSERT(before-after <= count);
        else
            T_ASSERT(before-after == count);
    }
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    cout << "Seed is " << set_seed(argc, argv) << endl;
    for (unsigned i=0 ; i < 3000 ; ++i) {
        tester t;
        t.run();
    }
    END_TEST_MAIN;
}

