// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2010, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2010, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <iostream>
#include <solver.hpp>
#include <arithmetic_propagators.hpp>
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;

class tester {
 public:
    typedef solver::reporting_levels reporting_levels;
    typedef solver::label_type label_type;
    tester(reporting_levels level) : m_level(level),
                                     A(static_cast<label_type>(1U << gen(31))),
                                     B(static_cast<label_type>(1U << gen(31))),
                                     C(static_cast<label_type>(1U << gen(31)))
    {}
    void run();

  private:
    void test_variables();
    void test_constraints();
    void test_combined();
    
    solver::reporting_levels m_level;
    const label_type A,B,C;
};


void tester::test_variables()
{
    solver solver;
    typedef unique_ptr<constraint_propagator> clause_p;
    solver.report_level(m_level);
    unsigned a= solver.add_var(iset().complement());
    unsigned b= solver.add_var(iset().complement());
    unsigned c= solver.add_var(iset().complement());
    unsigned d= solver.add_var(iset().complement());
    solver.add_constraint(clause_p(new sum_wrap_propagator))
        .reserve_arguments(3) << a << b << c;
    solver.variable_labels(a, A);
    solver.variable_labels(b, A|B);
    solver.variable_labels(c, B);
    solver.add_constraint(clause_p(new sum_wrap_propagator))
        .reserve_arguments(3) << b << c << d;
    for (unsigned i=0 ; i != solver.count_constraints() ; ++i)
        T_ASSERT(solver.constraint_labels(i) == solver::NO_LABEL);
    T_ASSERT(solver.variable_labels(a) == A);
    T_ASSERT(solver.variable_labels(b) == (A|B));
    T_ASSERT(solver.variable_labels(c) == B);
    T_ASSERT(solver.variable_labels(d) == solver::NO_LABEL);
    T_ASSERT(solver.variable_labels(a) != solver::NO_LABEL);
    T_ASSERT(A==B || solver.variable_labels(b) != B);
    
    T_ASSERT(solver.solve(&cout) == solver::SAT);
    
    T_ASSERT(solver.variable_labels(a) == A);
    T_ASSERT(solver.variable_labels(b) == (A|B));
    T_ASSERT(solver.variable_labels(c) == B);
    for (unsigned i=0 ; i != solver.count_constraints() ; ++i)
        T_ASSERT(solver.constraint_labels(i) == solver::NO_LABEL);
    vector<label_type> touched= solver.variables_touched_by_labels();
    T_ASSERT(touched.at(a) == A);
    T_ASSERT(touched.at(b) == (A|B));
    T_ASSERT( (touched.at(b)&A) == A);
    T_ASSERT(touched.at(c) == B);
    T_ASSERT(touched.at(d) == solver::NO_LABEL);
    T_ASSERT((touched.at(d)&A) == solver::NO_LABEL);
    T_ASSERT((touched.at(d)&A) != A);
}

void tester::test_constraints()
{
    solver solver;
    typedef unique_ptr<constraint_propagator> clause_p;
    solver.report_level(m_level);
    unsigned a= solver.add_var(iset().complement());
    unsigned b= solver.add_var(iset().complement());
    unsigned c= solver.add_var(iset().complement());
    unsigned d= solver.add_var(iset().complement());
    unsigned e= solver.add_var(iset().complement());
    solver.add_constraint(clause_p(new sum_wrap_propagator))
        .reserve_arguments(3).labels(A) << a << b << c;
    unsigned first_constraint = solver.count_constraints()-1;
    
    solver.add_constraint(clause_p(new sum_wrap_propagator))
        .reserve_arguments(3).labels(B) << b << c << d;
    unsigned second_constraint = solver.count_constraints()-1;
    
    solver.add_constraint(clause_p(new sum_wrap_propagator))
        .reserve_arguments(3) << c << d << e;
    unsigned third_constraint = solver.count_constraints()-1;

    for (unsigned i=0 ; i != solver.count_variables() ; ++i)
        T_ASSERT(solver.variable_labels(i) == solver::NO_LABEL);

    T_ASSERT(solver.constraint_labels(first_constraint) == A);
    T_ASSERT(solver.constraint_labels(second_constraint) == B);
    T_ASSERT(solver.constraint_labels(third_constraint) == solver::NO_LABEL);
    vector<label_type> touched= solver.variables_touched_by_labels();
    T_ASSERT(touched.at(a) == A);
    T_ASSERT(touched.at(b) == (A|B));
    T_ASSERT(touched.at(c) == (A|B));
    T_ASSERT(touched.at(d) == B);
    T_ASSERT(touched.at(e) == solver::NO_LABEL);
}

void tester::test_combined()
{
    solver solver;
    typedef unique_ptr<constraint_propagator> clause_p;
    solver.report_level(m_level);
    vector<unsigned> vars;
    for (unsigned i=0 ; i != 6 ; ++i)
        vars.push_back(solver.add_var(iset().complement()));
    solver.add_constraint(clause_p(new sum_wrap_propagator))
        .reserve_arguments(3).labels(A) << vars[0] << vars[1] << vars[2];
    
    solver.add_constraint(clause_p(new sum_wrap_propagator))
        .reserve_arguments(3).labels(B) << vars[1] << vars[2] << vars[3];
    
    solver.add_constraint(clause_p(new sum_wrap_propagator))
        .reserve_arguments(3) << vars[2] << vars[3] << vars[4];
    solver.variable_labels(0, C);
    solver.variable_labels(1, A|B);
    solver.variable_labels(2, C);
    solver.variable_labels(4, C);
    
    vector<label_type> touched= solver.variables_touched_by_labels();
    T_ASSERT(touched.at(vars[0]) == (A|C));
    T_ASSERT(touched.at(vars[1]) == (A|B));
    T_ASSERT(touched.at(vars[2]) == (A|B|C));
    T_ASSERT(touched.at(vars[3]) == B);
    T_ASSERT(touched.at(vars[4]) == C);
    T_ASSERT(touched.at(vars[5]) == solver::NO_LABEL);
}

void tester::run()
{
    test_variables();
    test_constraints();
    test_combined();
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    solver::reporting_levels report_level= solver::NONE;
    //report_level= solver::SEARCH;
    //report_level= solver::PROPAGATION_MODS;
    //report_level= solver::PROPAGATED_CONSTRAINTS;

    cout << "Seed is " << set_seed(argc, argv) << endl;
    for (unsigned i=0 ; i < 100 ; ++i)
        tester(report_level).run();
    
    END_TEST_MAIN;
}

