// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <algorithm>
#include <vector>
#include <map>
#include <iostream>
#include <constraint_runtime_info.hpp>
#include <sliced_runtime_info.hpp>
#include <constraint_propagator.hpp>
#include <neq_propagators.hpp>
#include <solver.hpp>
#include "test_utilities.hpp"
#include "explanation_tester.hpp"
#include "trivial_runtime_info.hpp"
#include "propagator_tester.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;
using hcsp::test::propagator_tester;
/**
 * A propagator with the same semantics as all_different_propagator.
 *
 * This propagator is implemented using an approach different from that of
 * all_different, and in that sense it is safe for producing reduction
 * references.
 *
 * This propagator is accurate - it never loses a solution and removes all
 * unsupported values from domains.
 *
 * For its implementation it also uses an \e inner CSP solver with many
 * neq_propagator objects. For its correct operation this propagator
 * assumes the correctness of neq_propagator and of the solver.
 */
class safer_propagator : public constraint_propagator
{
  public:
    safer_propagator() {}
    safer_propagator(const safer_propagator &other);
    virtual bool init(constraint_runtime_info & info);
    virtual result_type propagate( constraint_runtime_info & info)
    {
        return stateless_propagate(m_neq_solver.get(), m_neq_propagator.get(),
                                   info);
    }
    virtual bool conflicts(const constraint_runtime_info & info) const;
    virtual bool constraint_check( const constraint_runtime_info & info) const;
    virtual safer_propagator * clone() const
    { return new safer_propagator(*this); }
  private:
    //! Propagate all_different onto 'info' by using a solver which models
    //! a clique of != constraints, or using a single != propagator.
    static
    result_type stateless_propagate(solver * p_solver,
                                    constraint_propagator * p_neq_propagator,
                                    constraint_runtime_info & info);
    //! Update the domains of the solver according to \b info.
    static void update_solver_domains(solver & solver,
                                      const constraint_runtime_info & info);

    /**
     * Create a new solver object that defines the problem of all_different
     * using only neq_propagator objects.
     * This static method will create the variables and connect each pair
     * with a neq_propagator. The variables have an unconstrained domain.
     */
    static solver *new_diff_solver(unsigned size);

    /**
     * Check if the given (variable,value) pair is supported for the given
     * input domains.
     *
     * @param variables  The variable that we want to check support for.
     * @param value      The value whose support we want to seek for.
     * @param solver     A solver instance that models a clique of !=
     *                   constraints.
     * @param info       The input domains.
     * @param projected_solutions
     *                   If a solution with (variable,value) exists,
     *                   then add it to each of the corresponding domains
     *                   of \b projected_solutions.
     *
     * @return true if support for (variable,value) exists.
     */
    static
    bool check_value_support(unsigned variable, iset::value_type value,
                             solver &solver,
                             const constraint_runtime_info &info,
                             vector<iset> & projected_solutions);
    /**
     * Find more singletons which would propagate from root singletons.
     *
     * Based on singletons stored in \b done_values and free variables,
     * look for all singletons that can be trivially concluded.
     * The new singleton variables are removed from \b params_left and
     * done_values and assignment are updated accordingly.
     *
     * @param[in]      info        The original domains that we would like to
     *                             propagate. This list is not modified here.
     * @param[in,out] params_left  Parameters which are not yet singletons.
     * @param[in,out] done_values  Singleton values which have been propagated
     *                             up until now.
     * @param[in,out] assignment   For each value not in \b params_left
     *                             this is the assignment of singleton to
     *                             this parameter. If \b params_left is empty
     *                             then this vector contains a feasible
     *                             assignment.
     * @return false if concluded that there is no feasible assignment for
     *               the input done_values and params_left.
     */
    static
    bool propagate_singletons(const constraint_runtime_info & info,
                              vector<unsigned> &params_left, iset &done_values,
                              vector<iset::value_type> & assignment);

    /**
     * Find a parameter with non-singleton minimal domain.
     *
     * When looking into a parameter, calculate its domain size after
     * all done_values have been removed from the domain.
     *
     * @pre
     *   - if info.domain(i) is singleton then \f$ i \not \in done_domain \f$
     *   - if info.domain(i) \f$\setminus\f$ done_values is singleton then
     *     \f$ i \not \in done_domain \f$.
     *   - params_left is not empty.
     *
     * @param info         The base domains
     * @param params_left  The non-singleton parameters to choose from.
     * @param done_values  The singleton values that have to be ignored.
     *
     * @return The index of the parameter we look for.
     *
     */
    static
    unsigned find_min_domain_parameter(const constraint_runtime_info & info,
                                       const vector<unsigned> &params_left,
                                       const iset & done_values);

    /**
     * Perform a fast and trivial algorithm variant of check_value_support().
     *
     * Unlike check_value_support() this method will sometimes fail to
     * find existing feasible assignments.
     *
     * The heuristics used for finding feasible assignments tries to
     * find an assignment which has nothing in common with values of
     * \b projected_solutions.
     *
     * @return \b true if support for (variable,value) is found.
     *         If support for (variable,value) is not found return \b false,
     *         even in cases when such support exists.
     */
    static
    bool check_trivial_value_support(unsigned variable, iset::value_type value,
                                     const constraint_runtime_info &info,
                                     vector<iset> & projected_solutions);

    //! Join all domains which have \p max_size or less elements in the domain.
    static
    iset join_domains_less_or_equal_to(const constraint_runtime_info & info,
                                       size_t max_size);
    
    //! In the binary case this object is set and is used for propagation.
    unique_ptr<neq_propagator>       m_neq_propagator;

    //! The solver used for calculating propagation.
    unique_ptr<solver>               m_neq_solver;
};

/**
 * The main class used for testing the all_different_propagator.
 */
class tester
{
  public:
    typedef iset::value_type                   value_type;
    typedef trivial_runtime_info::result_type  result_type;
    //! Construct a random test skeleton for later testing of the propagator.
    tester() :
        m_final_assignment( random_final_assignment() ),
        m_final_assignment_is_satisfiable( is_satisfied(m_final_assignment) ),
        m_tested_propagator(
            gen(2) ?
            new all_different_propagator :
            new all_different_propagator(m_final_assignment.size())),
        m_info( m_final_assignment.size() )
    {
        set_random_domains(m_info);
    }
    //! Run the test on the prepared test skeleton.
    void run();
    
  private:
    //! Populate the domains of \b info with random data.
    void set_random_domains(trivial_runtime_info & info) const;

    //! Get a random vector of final assignments.
    vector<value_type> random_final_assignment() const;

    //! Check if the given assignment is a feasible assignment.
    bool is_satisfied(const vector<value_type> & assignment) const;

    /**
     * Tell whether all selected parameters are independent from each other.
     *
     * Independent parameters have no common elements in their domains.
     */
    bool are_selected_independent(trivial_runtime_info & info,
                                  const iset & parameters) const;

    //! Tell whether all modified parameters are independent from each other.
    bool are_modified_independent(trivial_runtime_info & info) const;
    
    //! Tell whether all parameters are independent from each other.
    bool are_all_independent(trivial_runtime_info & info) const;

    unsigned count_modified(trivial_runtime_info & info) const;
    /**
     * Propagate, simultaneously on the tested and the reference propagator.
     * This method also performs correctness and sanity checks for the
     * result
     *
     * @param safer_info   The domains of the safer reference propagator
     * @param tested_info  The domains of the tested propagator
     * @param final_state  Did the propagation reach a final state,
     *                     either UNSAT of VALID.
     */
    void propagate( trivial_runtime_info & safer_info,
                    trivial_runtime_info & tested_info,
                    bool & final_state);

    //! Tell whether any of the domains is bigger than the given limit.
    bool has_domain_bigger_than(size_t size_limit,
                                const constraint_runtime_info & info) const;
    

    //! The final assignment that domain reductions aim at.
    const vector<value_type>    m_final_assignment;

    //! Is the final assignment satisfiable? i.e. all values different.
    const bool                  m_final_assignment_is_satisfiable;

    //! An instance of a reference propagator.
    safer_propagator            m_safer_propagator;

    //! An instance of the tested propagator wrapped in an explanation_tester.
    explanation_tester          m_tested_propagator;

    //! The current domain.
    trivial_runtime_info        m_info;
}; // class tester

safer_propagator::safer_propagator(const safer_propagator &other)
    : constraint_propagator(other)
{
    if (other.m_neq_propagator.get())
        m_neq_propagator.reset(other.m_neq_propagator->clone());
    if (other.m_neq_solver.get()) {
        const unsigned size= other.m_neq_solver->count_variables();
        m_neq_solver.reset(new_diff_solver(size));
        for (unsigned i=0 ; i != size ; ++i)
            m_neq_solver->set_var_domain(i, other.m_neq_solver->var_domain(i));
    }
}

// virtual
bool safer_propagator::init( constraint_runtime_info & info )
{
    HCSP_PROFILE_CURRENT;
    if (info.arity() == 2) {
        m_neq_propagator.reset(new neq_propagator);
        return m_neq_propagator->init(info);
    }
    constraint_propagator::init(info);
    if (info.arity() < 2)
        return true;
    if ( !m_neq_solver.get() ||
         m_neq_solver->count_variables() != info.arity() )
        m_neq_solver.reset(new_diff_solver(info.arity()));

    HCSP_ASSERT(m_neq_solver.get());
    update_solver_domains(*m_neq_solver, info);
    bool is_sat= (m_neq_solver->solve() == solver::SAT);
    m_neq_solver->reset();
    return is_sat;
}

// static
iset safer_propagator::join_domains_less_or_equal_to(
    const constraint_runtime_info & info, size_t max_size)
{
    iset small_domains;
    for (unsigned i=0 ; i != info.arity() ; ++i)
        // This size()-1 trick takes care of the case when size() returns zero
        // when all 32 bits are in the domain.
        // If domain(i).empty() then it does not matter that the if statement
        // did not execute the 'then' part. This is because executing 'then'
        // would simply try to insert an empty set - which is a no-op.
        if ( info.domain(i).size()-1 < max_size )
            small_domains.insert_all_of(info.domain(i));
    return small_domains;
}

// static
constraint_propagator::result_type
safer_propagator::stateless_propagate(solver * p_solver,
                                      constraint_propagator * p_neq_propagator,
                                      constraint_runtime_info & info)
{
    HCSP_PROFILE_CURRENT;
    if (info.arity() <= 1)
        return VALID;
    if (info.arity() == 2)
        return p_neq_propagator->propagate(info);

    // Figure out which values of info have support.
    // Do it in an efficient way, such that any feasible solution found
    // has its values added to supported_domains vector. Also prefer to
    // look for feasible assignments that have the least resemblance to
    // supported_domains.
    vector<iset> supported_domains(info.arity());
    iset small_domains= join_domains_less_or_equal_to(info, info.arity());
    for (unsigned i=0 ; i != supported_domains.size() ; ++i) {
        iset values_to_check= info.domain(i);
        bool is_big_domain= values_to_check.size()-1 >= info.arity();
        values_to_check.erase_all_of(supported_domains[i]);
        if (is_big_domain)
            supported_domains[i].insert_all_of(
                iset(values_to_check).erase_all_of(small_domains));
        values_to_check.intersect(small_domains);
        for ( iset::iterator it= values_to_check.begin();
              it != values_to_check.end();
              ++it) {
            HCSP_ASSERT(p_solver);
            check_value_support(i, *it, *p_solver, info, supported_domains);
        }
        if (supported_domains[i].empty())
            return UNSAT;
    }

    // Update info according to the calculated supported_domains.
    for (unsigned i=0 ; i != supported_domains.size() ; ++i) {
        if (supported_domains[i] != info.domain(i))
            info.set_domain(i, move(supported_domains[i]));
    }
    return CONSISTENT;
}

bool safer_propagator::conflicts(const constraint_runtime_info & info) const
{
    sliced_runtime_info local_info(info,0, info.arity());
    
    result_type result;
    result= stateless_propagate(m_neq_solver.get(), m_neq_propagator.get(),
                                local_info);
    if (m_neq_solver.get()) {
        // This is not strictly required, just to make sure that we don't leak
        // information to the next activation. We should not leak information
        // because this is a const method and it is not supposed to change
        // state.
        m_neq_solver->reset();
        m_neq_solver->forget_learned_constraints();
    }
    return result == UNSAT;
}

bool safer_propagator::constraint_check(const constraint_runtime_info & info)
    const
{
    HCSP_PROFILE_CURRENT;
    for (unsigned i=0 ; i != info.arity() ; ++i)
        for (unsigned j=i+1 ; j != info.arity(); ++j)
            if (info.domain(i) == info.domain(j))
                return false;
    return true;
}

// static
bool safer_propagator::propagate_singletons(
    const constraint_runtime_info & info,
    vector<unsigned> & params_left, iset & done_values,
    vector<iset::value_type> & assignment)
{
    HCSP_PROFILE_CURRENT;
    bool stable= false;
    while (!stable && !params_left.empty()) {
        stable= true;
        for ( unsigned i=0, next=0 ; i != params_left.size() ; i= next) {
            ++next;
            unsigned parameter_index= params_left[i];
            const iset & domain= info.domain(parameter_index);
            if (!domain.singleton()&& !domain.contains_any_of(done_values))
                continue;
            iset complement_domain= domain;
            complement_domain.erase_all_of(done_values);
            if (complement_domain.empty())
                return false;
            if (!complement_domain.singleton())
                continue;
            assignment[parameter_index]= complement_domain.min();
            stable= false;
            done_values.insert_all_of(complement_domain);
            params_left[i]= params_left.back();
            params_left.pop_back();
            next= i;
        }
    }
    return true;
}

// static
unsigned safer_propagator::find_min_domain_parameter(
    const constraint_runtime_info & info,
    const vector<unsigned> & params_left,
    const iset & done_values)
{
    HCSP_PROFILE_CURRENT;
    unsigned min_param=-1U;
    long unsigned min_size=-1UL;
    for ( unsigned i=0 ; i != params_left.size() ; ++i) {
        unsigned parameter_index= params_left[i];
        const iset & domain= info.domain(parameter_index);
        HCSP_ASSERT(!domain.singleton());
        if (!domain.contains_any_of(done_values)) {
            if (domain.size() < min_size) {
                min_size= domain.size();
                min_param= parameter_index;
            }
            continue;
        }
        iset complement_domain= domain;
        complement_domain.erase_all_of(done_values);
        if ( !complement_domain.singleton() &&
             complement_domain.size() < min_size) {
            min_size= domain.size();
            min_param= parameter_index;
        }
    }
    return min_param;
}

// static
bool safer_propagator::check_trivial_value_support(
    unsigned variable,
    iset::value_type value,
    const constraint_runtime_info &info,
    vector<iset> & projected_solutions)
{
    HCSP_PROFILE_CURRENT;
    vector<unsigned> params_left;
    for (unsigned i=0 ; i != info.arity() ; ++i)
        if (i != variable)
            params_left.push_back(i);
    iset done_values(value);
    vector<iset::value_type> assignment(info.arity());
    assignment[variable]= value;
    while (true) {
        if ( !propagate_singletons(info, params_left, done_values, assignment))
            return false;
        if (params_left.empty())
            break;

        unsigned min_param;
        min_param= find_min_domain_parameter(info, params_left, done_values);
        
        iset complement_domain= info.domain(min_param);
        complement_domain.erase_all_of(done_values);
        // Prefer values not proven to be supported yet.
        // This is to speed up the algorithm that calls this function.
        if (!projected_solutions[min_param].contains_all_of(complement_domain))
            complement_domain.erase_all_of(projected_solutions[min_param]);
             
        iset::value_type next_value= complement_domain.rand();
        params_left.erase(find(params_left.begin(), params_left.end(),
                               min_param));
        done_values.insert(next_value);
        assignment[min_param]= next_value;
    }
    for (unsigned i=0 ; i != projected_solutions.size() ; ++i)
        projected_solutions[i].insert(assignment[i]);
    return true;
}

// static
bool safer_propagator::check_value_support(unsigned variable,
                                           iset::value_type value,
                                           solver & solver,
                                           const constraint_runtime_info &info,
                                           vector<iset> & projected_solutions)
{
    HCSP_PROFILE_CURRENT;
    if (check_trivial_value_support(variable, value, info,
                                    projected_solutions))
        return true;
    HCSP_PROFILE_LINE;
    solver.reset();
    solver.forget_learned_constraints();
    update_solver_domains(solver, info);
    // Use the solver to look for support for the (variable,value) pair.
    solver.set_var_domain(variable, iset(value));
    if (solver.solve(&cout) != solver::SAT) {
        solver.reset();
        return false;
    }

    for (unsigned i=0 ; i != projected_solutions.size() ; ++i)
        projected_solutions[i].insert_all_of(solver.var_domain(i));
    solver.reset();
    return true;
}

// static
void safer_propagator::update_solver_domains(
    solver & solver, const constraint_runtime_info & info)
{
    HCSP_ASSERT(solver.count_variables() == info.arity());
    for (unsigned i=0 ; i != info.arity() ; ++i)
        solver.set_var_domain(i, info.domain(i));
}

//static
solver *safer_propagator::new_diff_solver(unsigned size)
{
    unique_ptr<solver> p_solver(new solver);
    p_solver->should_intersect_with_ground(false);
    for (unsigned i=0 ; i != size ; ++i) {
        p_solver->add_var(iset().complement());
        for (unsigned j=0 ; j != i ; ++j) {
            unique_ptr<constraint_propagator> diff(new neq_propagator);
            p_solver->add_constraint(move(diff)).reserve_arguments(2)
                << i << j;
        }
    }
    return p_solver.release();
}

void tester::set_random_domains(trivial_runtime_info & info) const
{
    unsigned superset_percent= range(0, 100);
    unsigned everything_percent= range(0, 100);
    unsigned subset_percent= range(0, 100);
    iset values;
    for (unsigned i=0 ; i!= m_final_assignment.size() ; ++i) 
        values.insert(m_final_assignment[i]);
    iset::value_range_type range= random_var_bounds(gen(2)==0);
    for (unsigned i=0 ; i!= m_final_assignment.size() ; ++i) {
        iset domain= random_var(false, range.first, range.second);
        if ( superset_percent > gen(100) ) {
            if ( everything_percent > gen(100) )
                domain= move(iset().complement());
            else
                domain.insert_all_of(values);
        } else if (subset_percent > gen(100)) {
            domain.intersect(values);
        }
        domain.insert(m_final_assignment[i] );
        info.set_domain(i, move(domain));
    }
}

vector<iset::value_type> tester::random_final_assignment() const
{
    vector<value_type> final_assignment_vector;

    iset potential_values;
    size_t max_diff= static_cast<size_t>(iset::MAX_DOMAIN) - iset::MIN_DOMAIN;
    size_t size= range(2,10);
    size_t slack= max_diff-size+1;
    size_t from= iset::MIN_DOMAIN+ range(0U, slack);
    potential_values.insert(iset::value_range_type(from, from+size-1));
    switch (gen(4)) {
      case 0:
        potential_values.clear() ;
        potential_values.complement();
        break;
      case 1:
        break;
      default:
        for (unsigned i=gen(3) ; i > 0 ; --i) {
            from= iset::MIN_DOMAIN+ range(0U, slack);
            potential_values.insert(iset::value_range_type(from, from+size-1));
        }
    }
    bool make_unsat= gen(8)==0;
    size_t repeat_count= make_unsat ? range(1U, size) : 0;
    iset used;
    for (unsigned i=0 ; i != size ; ++i) {
        if (i > 0 && repeat_count > gen(size-i))
            final_assignment_vector.push_back( used.rand() );
        else
            final_assignment_vector.push_back( potential_values.rand() );
        used.insert(final_assignment_vector.back());
        potential_values.erase(final_assignment_vector.back());
    }
    return final_assignment_vector;
}

bool tester::is_satisfied(const vector<value_type> & assignment) const
{
    HCSP_ASSERT(assignment.size() == m_final_assignment.size());
    iset values;
    for (vector<value_type>::size_type i=0 ; i != assignment.size() ; ++i) {
        if ( values.contains(assignment[i]) )
            return false;
        values.insert(assignment[i]);
    }
    return true;
}

bool tester::are_selected_independent(trivial_runtime_info & info,
                                      const iset & parameters) const
{
    iset collected_values;
    for (iset::iterator it= parameters.begin() ; it!=parameters.end() ; ++it) {
        const iset & values= info.domain(*it);
        if ( collected_values.contains_any_of(values) )
            return false;
        collected_values.insert_all_of(values);
    }
    return true;
}

bool tester::are_modified_independent(trivial_runtime_info & info) const
{
    iset reduced_parameters;
    for (unsigned i=0 ; i != info.arity() ; ++i)
        if (info.reduced_between_activations(i))
            reduced_parameters.insert(i);
    return are_selected_independent( info, reduced_parameters) ;
}


bool tester::are_all_independent(trivial_runtime_info & info) const
{
    iset all_parameters( iset::value_range_type(0, info.arity()-1) );
    return are_selected_independent( info, all_parameters) ;
}

unsigned tester::count_modified(trivial_runtime_info & info) const
{
    unsigned count=0;
    for (unsigned i=0 ; i != info.arity() ; ++i)
        if (info.reduced_between_activations(i))
            ++count;
    return count;
}

void tester::propagate( trivial_runtime_info & safer_info,
                        trivial_runtime_info & tested_info,
                        bool & final_state)
{
    result_type safer_result= safer_info.maybe_propagate(m_safer_propagator);
    result_type tested_result=tested_info.maybe_propagate(m_tested_propagator);
    final_state= false;
    switch ( tested_result  ) {
      case trivial_runtime_info::UNSAT:
        final_state= true;
        T_ASSERT( tested_result == safer_result );
        break;
      case trivial_runtime_info::VALID:
        T_ASSERT( safer_result != trivial_runtime_info::UNSAT );
        T_ASSERT(are_all_independent(tested_info));
        T_ASSERT(safer_info.same_next_state(tested_info));
        final_state= true;
        break;
      case trivial_runtime_info::NOT_WATCHED:
        T_ASSERT( safer_result != trivial_runtime_info::UNSAT );
        T_ASSERT(safer_info.same_next_state(tested_info));
        break;
      case trivial_runtime_info::CONSISTENT:
        T_ASSERT( safer_result != trivial_runtime_info::UNSAT );
        if (safer_result == trivial_runtime_info::NOT_WATCHED)
            T_ASSERT(count_modified(tested_info)==0);
        else
            T_ASSERT(safer_info.same_next_state(tested_info));
        final_state= true;
        for (unsigned i=0 ; i != tested_info.arity() && final_state; ++i)
            final_state= tested_info.domain(i).singleton();
        break;
    }
}

bool tester::has_domain_bigger_than(size_t size_limit,
                                    const constraint_runtime_info & info) const
{
    for (unsigned i=0 ; i != info.arity() ; ++i)
        if (info.domain(i).size()-1 >= size_limit)
            return true;
    return false;
}

void tester::run()
{
    HCSP_PROFILE_CURRENT;
    trivial_runtime_info tested_info(m_info);
    m_tested_propagator.init(tested_info);

    trivial_runtime_info safer_info(m_info);
    m_safer_propagator.init(safer_info);

    vector<trivial_runtime_info> backtrack_info;
    unsigned num_backtracks= 10;

    while ( true ) {
        bool final_state= false;
        propagate(safer_info, tested_info, final_state);
        
        if ( num_backtracks >0 && !backtrack_info.empty() &&
             (final_state || gen(20/num_backtracks)==0)) {
            unsigned target= gen(backtrack_info.size());
            tested_info.backtrack_to(backtrack_info[target]);
            safer_info.backtrack_to(backtrack_info[target]);
            backtrack_info.resize(target, safer_info);
            --num_backtracks;
            final_state= false;
        }
        if (final_state)
            return;
        
        backtrack_info.push_back(tested_info);
        do {
            safer_info.random_reduce(m_final_assignment);
        } while (has_domain_bigger_than(100, safer_info) && gen(30) != 0);
        tested_info.assign_modifications(safer_info);
    }
}

pair<iset, iset> make_domain_pair(unsigned diff, unsigned n_elements)
{
    iset::value_range_type bounds= random_var_bounds(diff+1U);
    iset::value_range_type left, right;
    if ((iset::size_type)bounds.first - iset::MIN_DOMAIN > n_elements)
        right=make_pair(bounds.first-n_elements+1, bounds.first);
    else
        right=make_pair(bounds.first, bounds.first+n_elements-1);
    
    if ((iset::size_type)iset::MAX_DOMAIN-bounds.second > n_elements)
        left=make_pair(bounds.second, bounds.second+n_elements-1);
    else
        left=make_pair(bounds.second-n_elements+1, bounds.second);
    HCSP_ASSERT(left.first < left.second);
    HCSP_ASSERT(right.first < right.second);
    HCSP_ASSERT((iset::size_type)left.second-left.first < n_elements);
    HCSP_ASSERT((iset::size_type)right.second-right.first < n_elements);
    return make_pair(left, right);
}

void test_literal_or_neq_propagator()
{
    pair<iset,iset> domains= make_domain_pair(0U/* diff */, 4 /*values*/);
    
    iset literal(random_var_bounds(4U));
    iset first_dom;
    switch (gen(3)) {
      case 0:
        first_dom= literal;
        literal.complement();
        break;
      case 1:
        first_dom= literal;
        break;
      default:
        first_dom= random_var().insert_all_of(literal);
        break;
    }
    propagator_tester<literal_or_neq_propagator>(3, literal).run();
    propagator_tester<literal_or_neq_propagator>(3, literal)
        .set_dom(0,first_dom)
        .set_dom(1,domains.first)
        .set_dom(2,domains.second)
        .run();
    propagator_tester<literal_or_neq_propagator>(3, random_var())
        .run();
}

void test_all_different_generically()
{
    unsigned n_params= range(1,5);
    iset dom(random_var_bounds(n_params+1));
    propagator_tester<all_different_propagator> tester(n_params);
    for (unsigned j=0 ; j != n_params ; ++j)
        tester.set_dom(j, dom);
    tester.run();
    propagator_tester<safer_propagator> safer_tester(n_params);
    for (unsigned j=0 ; j != n_params ; ++j)
        safer_tester.set_dom(j, dom);
    safer_tester.run();
}

int main(int argc, char *argv[])
{
    HCSP_PROFILE_CURRENT;
    BEGIN_TEST_MAIN;
    unsigned seed = set_seed(argc, argv);
    cout << "Seed is " << seed << endl;
    for (unsigned i=0 ; i < 3000 ; ++i) {
        if (i != 0) {
            seed = gen();
            set_seed(seed);
        }
        tester t;
        t.run();
    }
    // all_different_propagator is heavy and tested above so do a minimal
    // test with the generic propagator_tester.
    for (unsigned i=0 ; i < 4 ; ++i) {
        test_all_different_generically();
    }
    for (unsigned i=0 ; i < 100 ; ++i) {
        test_literal_or_neq_propagator();
        
        propagator_tester<neq_propagator>(2).run();
        pair<iset,iset> domains= make_domain_pair(0U/* diff */, 4 /*values*/);
        propagator_tester<neq_propagator>(2)
            .set_dom(0,domains.first).set_dom(1,domains.second)
            .run();

        iset::value_type diff= random_var_in_bounds(0, hcsp::iset::MAX_DOMAIN);
        propagator_tester<difference_neq_constant_propagator>(2, diff).run();
        propagator_tester<difference_neq_constant_propagator>(2, -diff).run();
        domains= make_domain_pair(diff, 4 /*values*/);
        propagator_tester<difference_neq_constant_propagator>(2, diff)
            .set_dom(0,domains.first).set_dom(1,domains.second)
            .run();
        propagator_tester<difference_neq_constant_propagator>(2, -diff)
            .set_dom(0,domains.second).set_dom(1,domains.first)
            .run();
    }
            
    END_TEST_MAIN;
}

