// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <algorithm>
#include <iostream>
#include <vector>
#include <iset.hpp>
#include <clause_propagator.hpp>
#include <solver.hpp>
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;
class tester {
  public:
    tester(solver::reporting_levels level,
           unsigned max_clause_length,
           unsigned num_vars);
    void run();
    unsigned backtracks_done() const { return m_solver.backtracks_done(); }
  private:
    void validate_solution() const;
    void create_solution();
    void create_variables();
    void create_clauses();
    void create_var_clauses(unsigned variable_index);
    vector<unsigned> select_clause_variables(unsigned variable_index);
    void insert_clause( const vector<unsigned> & clause_variables,
                        const vector<iset>& clause );
    void remove_and_insert(vector< vector<iset> > & disjoint,
                           vector<iset> & orig_tuple,
                           const vector<iset> & to_remove,
                           unsigned handled_before=0);
    void partial_cover_using_one_clause(
        const vector<unsigned> &variables,
        unsigned variable_to_keep,
        vector< vector<iset> > &no_goods);
    bool tuples_overlap(const vector<iset> & first,
                        const vector<iset> & second) const;
    void remove_shared_with_clause(
        const vector<unsigned> & clause_variables,
        const vector<iset>     & clause,
        const vector<unsigned> & variables,
        vector< vector<iset> > & no_goods,
        unsigned                 no_good_index);

    unsigned domain_size() const { return 2; }
    
    vector<iset::value_type> m_solution;
    const iset m_domain;
    const unsigned m_max_clause_length;
    const unsigned m_num_vars;
    solver m_solver;
}; // class tester

tester::tester(solver::reporting_levels level, unsigned max_clause_length,
               unsigned num_vars)
    : m_domain(iset(make_pair(0, domain_size()-1))),
      m_max_clause_length(max_clause_length),
      m_num_vars(num_vars)
{
    HCSP_PROFILE_CURRENT;
    
    m_solver.backtrack_limit(1000000);
    m_solver.report_level(level);
    create_solution();
    create_variables();
    create_clauses();
}

void tester::run()
{
    HCSP_PROFILE_CURRENT;
    solver::result_type result= m_solver.solve(&cout);

    if (result != solver::SAT || m_solver.report_level() > solver::NONE)
        cout << result << endl;
    T_ASSERT(result == solver::SAT);
    validate_solution();
}

void tester::validate_solution() const
{
    for (unsigned i=0 ; i != m_num_vars ; ++i) {
        T_ASSERT(m_solver.var_domain(i).singleton());
        T_ASSERT(m_solver.var_domain(i).max() == m_solution[i]);
    }
}

void tester::create_solution()
{
    m_solution.clear();
    for (unsigned i=0 ; i != m_num_vars ; ++i)
        m_solution.push_back(gen(domain_size()));
}

void tester::create_variables()
{
    for (unsigned i=0 ; i != m_num_vars ; ++i)
        m_solver.add_var(m_domain);
}

void tester::create_clauses()
{
    for (unsigned i=0 ; i != m_num_vars ; ++i)
        create_var_clauses(i);
}

vector<unsigned> tester::select_clause_variables(unsigned variable_index)
{
    vector<unsigned> clause_variables;
    unsigned length= min<int>(m_max_clause_length, m_num_vars);
    clause_variables.push_back(variable_index);
    for (unsigned i=0 ; i != m_num_vars ; ++i) {
        if (i != variable_index)
            clause_variables.push_back(i);
    }

    randomizer rand;
    random_shuffle(clause_variables.begin()+1, clause_variables.end(), rand);
    clause_variables.erase(clause_variables.begin()+length,
                           clause_variables.end());
    
    return clause_variables;
}

void tester::insert_clause( const vector<unsigned> & clause_variables,
                            const vector<iset>& clause )
{
    HCSP_ASSERT(clause_variables.size() == clause.size());
    
    clause_propagator * p_clause_propagator = new clause_propagator;
    unique_ptr<constraint_propagator> constraint_propagator;
    constraint_propagator.reset(p_clause_propagator);
    
    constraint_connectivity_info & connectivity=
        m_solver.add_constraint(move(constraint_propagator));
    connectivity.reserve_arguments( clause.size() );
    p_clause_propagator->update_literals().reserve(clause.size());

    static vector<unsigned> order;
    order.resize(clause.size());
    for (unsigned i=0 ; i != clause.size() ; ++i)
        order[i]= i;
    randomizer rand;
    random_shuffle(order.begin(), order.end(), rand);
    
    for (unsigned i=0 ; i != clause_variables.size() ; ++i) {
        connectivity << clause_variables[order[i]];
        p_clause_propagator->update_literals().push_back(clause[order[i]]);
    }
    if ( m_solver.report_level() >= solver::SEARCH)
        cout << "Inserted "  << connectivity << *p_clause_propagator << "\n";
}    


void tester::remove_and_insert(vector< vector<iset> > & disjoint,
                               vector<iset> & tuple,
                               const vector<iset> & to_remove,
                               unsigned handled_before)
{
    if ( handled_before == tuple.size() ) {
        // completely covered
        return;
    }
    const iset & literal= tuple[handled_before];
    const iset & to_remove_literal= to_remove[handled_before];
    if ( !literal.contains_any_of(to_remove_literal) )
        return;
    else if (to_remove_literal.contains_all_of(literal) )
        remove_and_insert(disjoint, tuple, to_remove, handled_before+1);
    else {
        const iset saved_literal= tuple[handled_before];
        tuple[handled_before].intersect(to_remove_literal);
        remove_and_insert(disjoint, tuple, to_remove, handled_before+1);
        tuple[handled_before]= saved_literal;
        disjoint.push_back(tuple);
        disjoint.back()[handled_before].erase_all_of(to_remove_literal);
    }
}


bool tester::tuples_overlap(const vector<iset> & first,
                            const vector<iset> & second) const
{
    HCSP_ASSERT(first.size() == second.size());
    for (unsigned i=0 ; i != first.size() ; ++i) {
        if ( !first[i].contains_any_of(second[i]) )
            return false;
    }
    return true;
}

void tester::remove_shared_with_clause(
    const vector<unsigned> & clause_variables,
    const vector<iset>     & clause,
    const vector<unsigned> & variables,
    vector< vector<iset> > & no_goods_to_cover,
    unsigned                 no_good_index)
{
    HCSP_ASSERT(clause_variables.size() == clause.size());
    HCSP_ASSERT(clause_variables.size() <= variables.size());
    
    static vector<unsigned> variable_to_final_position;
    for (unsigned i=0 ; i != variables.size() ; ++i) {
        if ( variables[i] >= variable_to_final_position.size() )
            variable_to_final_position.resize(variables[i]+1,0 );
        variable_to_final_position[variables[i]]= i;
    }

    // Convert clause representation to a generalized no-good representation.
    static vector<iset> new_no_good;
    new_no_good.clear();
    new_no_good.resize(variables.size(), m_domain);
    for (unsigned i=0 ; i != clause.size() ; ++i) {
        unsigned pos= variable_to_final_position[clause_variables[i]];
        new_no_good.at(pos).erase_all_of(clause[i]);
    }
    if ( !tuples_overlap(no_goods_to_cover.at(no_good_index), new_no_good) )
        return;
    vector<iset> tested_no_good= no_goods_to_cover[no_good_index];
    no_goods_to_cover[no_good_index].swap( no_goods_to_cover.back() );
    no_goods_to_cover.pop_back();
    remove_and_insert(no_goods_to_cover, tested_no_good, new_no_good);
}

void tester::partial_cover_using_one_clause(
    const vector<unsigned> &variables,
    unsigned variable_to_keep,
    vector< vector<iset> > &no_goods)
{
    no_goods.back().swap( no_goods[ gen(no_goods.size()) ]);
    vector<iset> to_avoid= no_goods.back();
    HCSP_ASSERT(to_avoid.size() == variables.size());

    // Must avoid at least one tuple described by the no-good in to_avoid.
    // Must keep all tuples requested by m_solution[variable_to_keep]
    static vector<iset> clause;
    static vector<unsigned> clause_variables;
    clause.clear();
    clause_variables.clear();
    unsigned literals_to_erase= hcsp::random::range(0U, variables.size()/2);
    for (unsigned i=0 ; i != variables.size() ; ++i) {
        if ( gen(variables.size()-i) < literals_to_erase) {
            --literals_to_erase;
            if ( variables[i] != variable_to_keep)
                continue;
        }
        
        clause_variables.push_back(variables[i]);
        clause.push_back(m_domain);
        clause.back().erase(to_avoid[i].rand());
    }
    insert_clause( clause_variables, clause );
    for (unsigned i=no_goods.size() ; i != 0 ; --i)
        remove_shared_with_clause( clause_variables, clause,
                                   variables, no_goods,
                                   i-1);
}

void tester::create_var_clauses(unsigned variable_index)
{
    if ( m_solver.report_level() >= solver::SEARCH)
        cout << "V" << variable_index << " :- " << m_solution[variable_index]
             << "\n";
    vector<unsigned> variables= select_clause_variables(variable_index);
    vector< vector<iset> > no_goods(1);
    for (unsigned i=0 ; i != variables.size() ; ++i) {
        no_goods[0].push_back(m_domain);
        if (variables[i] == variable_index)
            no_goods[0].back().erase( m_solution[variable_index] );
    }
    while ( !no_goods.empty() )
        partial_cover_using_one_clause(variables, variable_index, no_goods);
}


int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    HCSP_PROFILE_CURRENT;
    solver::reporting_levels report_level= solver::NONE;
    //report_level= solver::SEARCH;
    //report_level= solver::PROPAGATION_MODS;
    //report_level= solver::PROPAGATED_CONSTRAINTS;

    cout << "Seed is " << set_seed(argc, argv) << endl;
    const unsigned clause_length=9;

    // The most effective way to control problem complexity (rather than
    // problem construction time) is to increase num_vars
    const unsigned num_vars=50;
    unsigned total_backtracks= 0; 
    for (unsigned i=0 ; i != 30 ; ++i) {
        unsigned next_seed = gen();
        tester test(report_level, clause_length, num_vars);
        test.run();
        total_backtracks += test.backtracks_done();
        srand(next_seed);
        if ( report_level >= solver::SEARCH )
            cout << "Seed is " << next_seed << endl;
    }
    cout << total_backtracks << " total backtracks\n";
    END_TEST_MAIN;
}
