// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2011, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2011, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <iostream>
#include <sstream>
#include <solver.hpp>
#include <flags.hpp>
#include <opb_context.hpp>
#include <opb_parser.tab.hpp>
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;

struct expression_type
{
    string str;
    solver::result_type expected;
};

expression_type expressions[]= {
    {"+1 x0 +2 ~x0 x1 = 1; * expecting result x0=1,x1 = unknown\n",
     solver::SAT
    },
    {"+1 x0 +2 ~x0 x1 = 2;  *expecting result x0=0,x1 =1\n",
     solver::SAT
    },
    {"+1 x0 +2 ~x0 x1 = 2; * expecting result x0=0,x1 =1 \n"
     "+1 x0 = 1;           * conflicting with this \n",
     solver::UNSAT
    },
    {"+1 x0 +2 ~x0 x1 = 2;\n"
     "+1 x1 = 0;\n",
     solver::UNSAT
    },
    {"+1 x0 +2 ~x0 x1 = 2;\n"
     "+1 ~x1 = 0;\n",
     solver::SAT
    },
    {"+2 a +3 b > 4;\n",
     solver::SAT
    },
    {"+2 a +3 b >= 5;\n",
     solver::SAT
    },
    {"+2 ~a +3 ~b >= 5;\n"
     "+1 a +1 b = 0;\n",
     solver::SAT
    },
    {"+2 a +3 b >= 5;\n"
     "+1 ~a =1;\n",
     solver::UNSAT
    },
    {"+2 a +3 b >= 5;\n"
     "+1 ~b =1;\n",
     solver::UNSAT
    },
    {"+2 a +3 b > 5;\n",
     solver::UNSAT
    },
    {"+2 a +3 b >= 6;\n",
     solver::UNSAT
    },
    {"-2 a -3 b <= -5;\n"
     "+1 ~a +1 ~b = 0;\n",
     solver::SAT
    },
    {"-2 a -3 b < -4;\n"
     "+1 ~a +1 ~b = 0;\n",
     solver::SAT
    },
    {"-2 a -3 b < -5;\n",
     solver::UNSAT
    },
    {"-2 a -3 b < -1;\n"
     "+1 a = 1;\n",
     solver::SAT
    },
    {"-2 a -3 b < -1;\n"
     "+1 b = 1;\n",
     solver::SAT
    },
    {"-2 a -3 b < -1;\n"
     "+1 a + 1 b = 0;\n",
     solver::UNSAT
    },
    
    // Test that trivially false conjunctions are treated properly
    {"+2 x y ~x = 0;\n", solver::SAT   },
    {"+2 x y ~x > 0;\n", solver::UNSAT },

    // Test repeating conjunctions of a single variable
    {"+1 x = 1;\n" "+1 x > 0;\n", solver::SAT   },
    {"+1 x = 1;\n" "+1 x = 0;\n", solver::UNSAT },
    {"+2 x = 2;\n" "+1 x > 0;\n", solver::SAT   },
    {"+2 x = 2;\n" "+1 x = 0;\n", solver::UNSAT },

    // Test repeating conjunctions of several variables
    {"+1 x ~y = 1;\n" "+1 x ~y > 0;\n", solver::SAT   },
    {"+1 x ~y = 1;\n" "+1 x ~y = 0;\n", solver::UNSAT },
    {"+2 x ~y = 2;\n" "+1 x ~y > 0;\n", solver::SAT   },
    {"+2 x ~y = 2;\n" "+1 x ~y = 0;\n", solver::UNSAT },
    
    {"+2 x +3 ~x = 2;\n",               solver::SAT   },
    {"+2 x +3 ~x = 3;\n",               solver::SAT   },
    {"+3 x +1 ~x > 3;\n",               solver::UNSAT },
    {"+1 x +3 ~x < 1;\n",               solver::UNSAT },
    {"+3 x +1 ~x = 1;\n",               solver::SAT   },
    {"+3 x +1 ~x = 2;\n",               solver::UNSAT },
    {"+3 x +1 ~x = 3;\n",               solver::SAT   },
    
    // Check how clauses are created
    {"+3 x y +2 x z >= 2;\n",           solver::SAT   }, // Not a clause
    {"+2 x y <= 2;\n" "+3 x z = 0;\n" "+1 x y +1 x z >= 1;\n", solver::SAT},
    {"+1 x +1 y <= 1;\n" "+1 ~x +1 ~y >= 1;\n",                solver::SAT},
    {"+1 x +1 y <= 1;\n" "+1 x = 1;\n",                        solver::SAT},
    {"+1 x +1 y <= 1;\n" "+1 x = 1;\n" "+1 y = 1;\n",          solver::UNSAT},

    // Check that trivially negated items are summed up correctly
    {"+1 ~x +1 y +1 x = 0;\n",                                 solver::UNSAT},
    {"+1 ~x +1 y +1 x = 1;\n" "+1 y = 0;\n",                   solver::SAT},
    {"+1 ~x +1 y +1 x = 1;\n" "+1 y = 1;\n",                   solver::UNSAT},
    {"+1 ~x +1 y +1 x = 3;\n",                                 solver::UNSAT},
    
    {"+2 ~x +1 y +3 x <= 1;\n",                                solver::UNSAT},
    {"+2 ~x +1 y +3 x = 2;\n" "+1 y +1 x = 0;\n",              solver::SAT},
    {"+2 ~x +1 y +3 x = 2;\n" "+1 y +1 x > 0;\n",              solver::UNSAT},
    {"+2 ~x +1 y +3 x = 3;\n" "+1 y +1 ~x = 0;\n",             solver::SAT},
    {"+2 ~x +1 y +3 x = 3;\n" "+1 ~y +1 x = 0;\n",             solver::SAT},
    {"+2 ~x +1 y +3 x = 3;\n" "-1 y +1 x = 0;\n",              solver::UNSAT},
    {"+2 ~x +1 y +3 x = 4;\n" "+1 x = 0;\n",                   solver::UNSAT},
    {"+2 ~x +1 y +3 x = 4;\n" "+1 y = 0;\n",                   solver::UNSAT},
    {"+2 ~x +1 y +3 x = 4;\n",                                 solver::SAT},
    {"+2 ~x +1 y +3 x > 4;\n",                                 solver::UNSAT},

#ifdef USE_BIG_ISET
    {"-1 x0 +2 x0 x1 = 1; * expecting result x0=1,x1 = 1\n",
     solver::SAT
    },
    {"-1 x0 +2 ~x0 x1 = 1;\n",
     solver::UNSAT
    },
    {"+1 x0 -2 x0 x1 = -1; * expecting result x0=1,x1 = 1\n",
     solver::SAT
    },
    {"+1 x0 -2 ~x0 x1 = -1;\n",
     solver::UNSAT
    },
    {"+1 x0 -2 ~x0 x1 = -2;  *expecting result x0=0,x1 =1\n",
     solver::SAT
    },
    {"+1 x0 +2 ~x0 x1 = -2; * expecting result x0=0,x1 =1 \n"
     "+1 x0 = 1;           * conflicting with this \n",
     solver::UNSAT
    },
    {"+1 x0 -2 ~x0 x1 = -2; * expecting result x0=0,x1 =1 \n"
     "+1 x0 = 1;           * conflicting with this \n",
     solver::UNSAT
    },
#endif
};

int main_impl(int argc, char *argv[])
{
    hcsp_flags flags;
    flags.backtrack_limit= 1000000000;
    flags.parse(argc, argv);

    for (unsigned i=0 ; i != sizeof(expressions)/sizeof(expressions[0]); ++i) {
        istringstream str(expressions[i].str);
        solver problem;
        opb::context context(problem, &str, "");
        yy::opb_parser parser(context);
        flags.apply(problem);
        if (context.trace_parsing())
            parser.set_debug_level(1);
        parser.set_debug_stream(cout);
        if (parser.parse() != 0)
            throw runtime_error("Error in expression:" + expressions[i].str);
        flags.quiet_start= true;
        solver::result_type result= solver::UNSAT;
        if ( context.build_solver() ) {
            result= problem.solve(&cout);
        }                
        if (result != expressions[i].expected) {
            cerr << "Diff in results of expressions[" << i << "]="
                 << expressions[i].str;
        }
        T_ASSERT(result == expressions[i].expected);
    }
    return 0;
}

int main(int argc, char *argv[])
{
    try {
        return main_impl(argc, argv);
    } catch (std::exception & e) {
        cout << "Exception(" << typeid(e).name() << "):\n" << e.what() << endl;
    } catch (...) {
        cout << "unknown exception" << std::endl;
    }
    return 1;                                               
}
