// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2014, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2014, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <mid_access_priority_queue.hpp>

#include <string>
#include <cstdlib>
#include <unistd.h>
#include <flags.hpp>
#include <utilities.hpp>
#include "test_utilities.hpp"

using namespace hcsp;
using namespace hcsp::random;
using namespace std;

class tester
{
  public:
    class flags;
    tester() : m_random_sequence(get_random_sequence()) {}
    void run();
    void test_sequential_string();
    void test_sequential_unsigned(const vector<unsigned> & sequence);
    void test_random_access_unsigned(const vector<unsigned> & sequence);
    
  private:
    vector<unsigned> get_random_sequence() const;
    template <typename Element>
    static void make_queue_for_sequence(
        mid_access_priority_queue<Element> & queue,
        const vector<Element> & sequence);
    
    const vector<unsigned> m_random_sequence;
}; // class tester

class tester::flags : public hcsp::flags
{
  public:
    flags() : seed(0) {}
  protected:
    virtual bool parse_option(char option);

    virtual void usage_short(std::ostream & out);
    virtual void usage_long(std::ostream & out);
    
    virtual std::string legal_args() const
    { return hcsp::flags::legal_args() + "s:"; }
    
  public:
    unsigned seed;
}; // class tester::flags

bool tester::flags::parse_option(char option)
{
    if (hcsp::flags::parse_option(option))
        return true;
    switch (option) {
      case 's': seed= strtoul(optarg, NULL, 0);                  break;
      default:
        return false;
    }
    return true;
}

void tester::flags::usage_short(ostream & out)
{
    hcsp::flags::usage_short(out);
    out << " [-s seed]\n";
}

void tester::flags::usage_long(ostream & out)
{
    hcsp::flags::usage_long(out);
    out << "  -s SEED The random seed to use in the test,\n"
        << "          where 0 [default] stands for \"randomize the seed\".\n";
}

void tester::run()
{
    test_sequential_string();
    test_sequential_unsigned(m_random_sequence);
    test_random_access_unsigned(m_random_sequence);
}

vector<unsigned> tester::get_random_sequence() const
{
    vector<unsigned> random_sequence;
    unsigned size= 1U << gen(16);
    random_sequence.resize(size);
    for (unsigned i= 0 ; i != size ; ++i)
        random_sequence[i]= gen();
    return random_sequence;
}

void tester::test_sequential_string()
{
    string elements_init[] = { "ab", "bc", "ad", "de", "de", "high"};
    unsigned num_elements= sizeof(elements_init)/sizeof(elements_init[0]);
    vector<string> elements(&elements_init[0], &elements_init[0]+num_elements);
    mid_access_priority_queue<string> queue(elements.begin(), elements.end());
    sort(elements.begin(), elements.end());
    while (!elements.empty()) {
        T_ASSERT(elements.back() == queue.max());
        T_ASSERT(elements.size() == queue.size());
        mid_access_priority_queue<string>::key_type
            removed_key= queue.key_of_max();
        queue.swap_keys(removed_key, queue.size()-1);
        queue.pop_last_key();
        elements.pop_back();
    }
}

// static
template <typename Element>
void tester::make_queue_for_sequence(mid_access_priority_queue<Element> &queue,
                                      const vector<Element> &sequence)
{
    unsigned num_assigned;
    num_assigned= sequence.size() - random::range(0U, sequence.size()/64);
    queue.assign(sequence.begin(), sequence.begin()+num_assigned);
    for (unsigned i= num_assigned ; i != sequence.size() ; ++i)
        queue.insert(sequence[i]);
}

void tester::test_sequential_unsigned(const vector<unsigned>& sequence)
{
    mid_access_priority_queue<unsigned> queue;
    make_queue_for_sequence(queue, sequence);
    vector<unsigned> sorted_sequence= sequence;
    sort(sorted_sequence.begin(), sorted_sequence.end());
    for (unsigned i=0 ; i != sequence.size() ; ++i)
        T_ASSERT(sequence[i] == queue[i]);
    while (!sorted_sequence.empty()) {
        T_ASSERT(sorted_sequence.back() == queue.max());
        T_ASSERT(sorted_sequence.size() == queue.size());
        mid_access_priority_queue<string>::key_type
            removed_key= queue.key_of_max();
        queue.swap_keys(removed_key, queue.size()-1);
        queue.pop_last_key();
        sorted_sequence.pop_back();
    }
}

void tester::test_random_access_unsigned(const vector<unsigned>& sequence)
{
    mid_access_priority_queue<unsigned> queue;
    make_queue_for_sequence(queue, vector<unsigned>(sequence.size(), 0));
    for (unsigned i=0 ; i != sequence.size() ; ++i) {
        queue[i]= sequence[i];
    }
    vector<unsigned> remaining= sequence;
    for (unsigned i=0 ; i != sequence.size() ; ++i)
        T_ASSERT(sequence[i] == queue[i]);
    while (!remaining.empty()) {
        T_ASSERT(remaining.size() == queue.size());
        
        unsigned removed_key= gen(remaining.size());
        T_ASSERT(remaining[removed_key]==queue[removed_key]);
        remaining[removed_key]= remaining.back();
        // either remove the key by swap or by assignment
        if (gen(2) == 0)
            queue[removed_key]= remaining.back();
        else
            queue.swap_keys(remaining.size()-1, removed_key);
        queue.pop_last_key();
        remaining.pop_back();
    }
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    tester::flags flags;
    if (!flags.parse(argc, argv))
        return 1;
    cout << "Seed is " << set_seed(flags.seed) << endl;
    for (unsigned i= 0 ; i != 100 ; ++i) {
        tester tester;
        tester.run();
    }
    END_TEST_MAIN;
}
