// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <propagation.hpp>
#include <utilities.hpp>
#include "test_utilities.hpp"

// Not applying "using namespace std" due to name conflicts with some
// propagation algorithms in hcsp::propagation namespace
using namespace hcsp;
using namespace hcsp::random;
using namespace hcsp::propagation;

class tester
{
  public:
    typedef iset::value_type       value_type;
    typedef iset::value_range_type value_range_type;
    tester() {}
    void run();
  private:
    struct normalizer
    {
        // This whole normalization scheme won't work if long long is
        // not at least twice as big as value_type.
        HCSP_STATIC_ASSERT(sizeof(long long) >= sizeof(value_type)*2);
        static value_type norm(long long val)
        {
            const long long modulo = 1LL+iset::MAX_DOMAIN-iset::MIN_DOMAIN;
            HCSP_STATIC_ASSERT(modulo > 0);
            val %= modulo;
            if (val > iset::MAX_DOMAIN)
                val -= modulo;
            else if (val < iset::MIN_DOMAIN)
                val += modulo;
            HCSP_ASSERT( iset::MIN_DOMAIN <= val && val <= iset::MAX_DOMAIN );
            return val;
        }
    };
     
    struct expr_plus : private normalizer
    {
        value_type operator()(value_type a1, value_type a2)
        { return norm((0LL)+a1+a2); }
        void propagate(iset &a, iset &b, iset & res)
        {
            iset orig_a=   a;
            iset orig_b=   b;
            iset orig_res= res;
            plus(orig_a,orig_b,  res);
            minus(orig_res, orig_a, b);
            minus(orig_res, orig_b, a);
        }
    };

    struct expr_minus : private normalizer
    {
        value_type operator()(value_type a1, value_type a2)
        { return norm((0LL)+a1-a2); }
        void propagate(iset &a, iset &b, iset & res)
        {
            iset orig_a=   a;
            iset orig_b=   b;
            iset orig_res= res;
            minus(orig_a,   orig_b,   res);
            plus(orig_res,  orig_b,   a);
            minus(orig_a,   orig_res, b);
        }
    };
    struct formula_neq
    {
        bool operator()(value_type a1, value_type a2) { return a1!=a2; }
        void propagate(iset &a, iset &b)
        {
            iset orig_a=   a;
            iset orig_b=   b;
            neq(orig_a, b);
            neq(orig_b, a);
        }
    };
    struct formula_lt
    {
        bool operator()(value_type a1, value_type a2) { return a1<a2; }
        void propagate(iset &a, iset &b)
        {
            iset orig_a=   a;
            iset orig_b=   b;
            lt_to_left(a,       orig_b);
            lt_to_right(orig_a, b);
        }
    };
    struct formula_le
    {
        bool operator()(value_type a1, value_type a2) { return a1<=a2; }
        void propagate(iset &a, iset &b)
        {
            iset orig_a=   a;
            iset orig_b=   b;
            le_to_left(a,       orig_b);
            le_to_right(orig_a, b);
        }
    };
    iset rand_half_set(const iset &in) const;

    template <class Op>
    void expect_empty_propagation(Op op, const iset &a, const iset &b,
                                  const iset &c);
    
    template <class Op>
    void expect_consistent(Op op, const iset &a, const iset &b,
                           const iset &c);
    template <class Op>
    void test_expression(Op op);


    template <class Form>
    void expect_empty_propagation(Form form, const iset &a, const iset &b);
    
    template <class Form>
    void expect_consistent(Form form, const iset &a, const iset &b);
    template <class Form>
    void test_formula(Form form);

    void test_propagation_links();
};

iset tester::rand_half_set(const iset &in) const
{
    if (in.singleton())
        return in;
    // Using this weird division to avoid overflow.
    // It might give a result which is off by one.
    value_type mid= in.min()/2+in.max()/2;

    if (mid <= in.min() || mid >= in.max()) {
        mid=in.min();
        T_ASSERT(in.size() <= 3);
    }
    if (gen(2) == 0)
        return iset(value_range_type(in.min(), mid)).intersect(in);
    else
        return iset(value_range_type(mid+1, in.max())).intersect(in);
}

template <class Op>
void tester::expect_empty_propagation(Op op, const iset &a, const iset &b,
                                      const iset &c)
{
    iset a_next=a, b_next=b, c_next=c;

    // First, the trivial propagation test - with no semantics.
    op.propagate(a_next, b_next, c_next);
    T_ASSERT( a_next.empty() && b_next.empty() && c_next.empty() );
    if (a_next.empty() || b_next.empty() || c_next.empty())
        return;

    // Second, test that consistency is impossible in this case.
    if (a.singleton() && b.singleton()) {
        T_ASSERT( !c.contains( op(a.min(), b.min()) ) );
        return;
    }
    
    expect_empty_propagation(op, rand_half_set(a), rand_half_set(b), c);
}

template <class Op>
void tester::expect_consistent(Op op, const iset &a, const iset &b,
                               const iset &c)
{
    if ( a.empty() || b.empty() || c.empty()) {
        T_ASSERT( a.empty() && b.empty() && c.empty() );
        return;
    }

    iset a_next=a, b_next=b, c_next=c;

    op.propagate(a_next, b_next, c_next);
    T_ASSERT(a==a_next);
    T_ASSERT(b==b_next);
    T_ASSERT(c==c_next);

    if (a.singleton() && b.singleton()) {
        T_ASSERT( c.singleton() );
        T_ASSERT( c.min()  == op(a.min(), b.min()) );
        return;
    }

    // To avoid exponential recursion, first reduce a, and only then b.
    if (a.singleton()) {
        const iset b_half= rand_half_set(b);
        b_next= b_half;
        op.propagate(a_next, b_next, c_next);

        T_ASSERT(b_next == b_half);
    } else {
        const iset a_half= rand_half_set(a);
        a_next= a_half;
        op.propagate(a_next, b_next, c_next);

        T_ASSERT(a_next == a_half);
    }
    expect_consistent(op, a_next, b_next, c_next);
}

template <class Op>
void tester::test_expression(Op op)
{
    iset a= random_var(), b= random_var(), c= random_var();
    iset a_next=a, b_next=b, c_next=c;
    op.propagate(a_next, b_next, c_next);
    
    // Make sure no good value is removed.
    expect_empty_propagation(op, iset(a).erase_all_of(a_next), b, c);
    expect_empty_propagation(op, a, iset(b).erase_all_of(b_next), c);
    expect_empty_propagation(op, a, b, iset(c).erase_all_of(c_next));

    // Make sure no bad values are kept.
    expect_consistent(op, a_next, b_next, c_next);
}


template <class Form>
void tester::expect_empty_propagation(Form form, const iset &a, const iset &b)
{
    iset a_next=a, b_next=b;

    // First, the trivial propagation test - with no semantics.
    form.propagate(a_next, b_next);
    T_ASSERT( a_next.empty() && b_next.empty());
    if (a.empty() || b.empty())
        return;

    // Second, test that consistency is impossible in this case.
    if (a.singleton() && b.singleton()) {
        T_ASSERT( !form(a.min(), b.min()) );
        return;
    }
    
    expect_empty_propagation(form, rand_half_set(a), rand_half_set(b));
}

template <class Form>
void tester::expect_consistent(Form form, const iset &a, const iset &b)
{
    if ( a.empty() || b.empty()) {
        T_ASSERT( a.empty() && b.empty());
        return;
    }

    iset a_next=a, b_next=b;

    form.propagate(a_next, b_next);
    T_ASSERT(a==a_next);
    T_ASSERT(b==b_next);

    if (a.singleton() && b.singleton()) {
        T_ASSERT( form(a.min(), b.min()) );
        return;
    }

    // To avoid exponential recursion, first reduce a, and only then b.
    if (a.singleton()) {
        const iset b_half= rand_half_set(b);
        b_next= b_half;
        form.propagate(a_next, b_next);

        T_ASSERT(b_next == b_half);
    } else {
        const iset a_half= rand_half_set(a);
        a_next= a_half;
        form.propagate(a_next, b_next);

        T_ASSERT(a_next == a_half);
    }
    expect_consistent(form, a_next, b_next);
}

template <class Form>
void tester::test_formula(Form form)
{
    iset a= random_var(), b= random_var();
    iset a_next=a, b_next=b;
    form.propagate(a_next, b_next);
    
    // Make sure no good value is removed.
    expect_empty_propagation(form, iset(a).erase_all_of(a_next), b);
    expect_empty_propagation(form, a, iset(b).erase_all_of(b_next));

    // Make sure no bad values are kept.
    expect_consistent(form, a_next, b_next);
}

void tester::test_propagation_links()
{
    const iset everything(value_range_type(iset::MIN_DOMAIN,iset::MAX_DOMAIN));
    //
    // Test the link between neg and minus
    //
    iset a= random_var(), b=random_var(), c=random_var();
    iset b_next1=b, b_next2=b;
    neg(a, b_next1);
    minus(iset(0), a, b_next2);
    HCSP_ASSERT( b_next1 == b_next2 );

    //
    // Test the link between plus, and minus through neg.
    //
    iset temp= everything;
    neg(b, temp);
    iset plus_result= c, minus_result=c;
    plus(a, temp, plus_result);
    minus(a, b, minus_result);
    T_ASSERT(plus_result == minus_result);

    //
    // Test the link between le and lt
    //
    b_next1= b; b_next2= b;
    le_to_right(a, b_next1);
    lt_to_right(a, b_next2);
    T_ASSERT(b_next1.contains_all_of(b_next2));
    // b_next2 may contain less elements only for elements shared between b,a
    T_ASSERT(b_next1 == iset(b_next2).insert_all_of(iset(a).intersect(b)));
    
    iset a_next1=a, a_next2=a;
    le_to_left(a_next1, b);
    lt_to_left(a_next2, b);
    T_ASSERT(a_next1.contains_all_of(a_next2));
    // a_next2 may contain less elements only for elements shared between b,a
    T_ASSERT(a_next1 == iset(a_next2).insert_all_of(iset(a).intersect(b)));
}

void tester::run()
{
    test_expression(expr_minus());
    test_expression(expr_plus());
    test_formula(formula_neq());
    test_formula(formula_le());
    test_formula(formula_lt());

    test_propagation_links();
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    std::cout << "Seed is " << set_seed(argc, argv) << std::endl;
    for (unsigned i=0 ; i < 3000 ; ++i) {
        tester t;
        t.run();
    }
    END_TEST_MAIN;
}

