// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2014 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2014 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include "propagator_tester.hpp"

#include <arithmetic_propagators.hpp>
#include <comparison_propagators.hpp>
#include <interval_propagators.hpp>
#include <table_propagators.hpp>

#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;
using namespace hcsp::test;


void concatenate_all_possibilities(vector<vector<iset::value_type> > & tuples,
                                   const iset & values)
{
    unsigned reference_end= tuples.size();
    iset::iterator it= values.begin();
    for (unsigned i=0 ; i!= reference_end ; ++i)
        tuples[i].push_back(*it);
    tuples.reserve(tuples.size()*values.size());

    while (  ++it != values.end() ) {
        for (unsigned i=0 ; i!=  reference_end ; ++i) {
            tuples.push_back(tuples[i]);
            tuples.back().back()= *it;
        }
    }
}

void random_table(vector<vector<iset::value_type> > & tuples, unsigned arity)
{
    vector<iset> domains(arity);
    for (unsigned i=0 ; i!=domains.size() ; ++i) {
        iset possible_vales(make_pair(iset::MIN_DOMAIN, iset::MAX_DOMAIN));
        unsigned value_count= range(1,10);
        while (value_count--) {
            iset::value_type value= possible_vales.rand();
            possible_vales.erase(value);
            domains[i].insert(value);
        }
    }
    tuples.clear();
    tuples.resize(1);
    for (unsigned i=0 ; i != arity; ++i)
        concatenate_all_possibilities(tuples, domains[i]);
    
    unsigned relevant_arity= range(1U,arity);
    size_t tuple_count=1;
    for (unsigned i=0 ; i!=relevant_arity ;++i)
        tuple_count *= 10;
    
    tuple_count= min(tuple_count, tuples.size());

    if (gen(100) == 0)
        tuple_count= tuples.size();

    while (tuples.size() > tuple_count) {
        tuples[gen(tuples.size())].swap(tuples.back());
        tuples.pop_back();
    }
}

int main(int argc, char *argv[])
{
    HCSP_PROFILE_CURRENT;
    BEGIN_TEST_MAIN;
    cout << "Seed is " << set_seed(argc, argv) << endl;
    for (unsigned i=0 ; i < 3000 ; ++i) {
        propagator_tester<neg_propagator>(2).run();
        propagator_tester<abs_propagator>(2).run();
        propagator_tester<less_equal_abs_propagator>(2).run();
        propagator_tester<greater_equal_abs_propagator>(2).run();
        propagator_tester<mul_wrap_propagator>(3).inaccurate().run();
        propagator_tester<mul_truncate_propagator>(3).inaccurate().run();
        propagator_tester<less_equal_mul_propagator>(3).run();
        propagator_tester<greater_equal_mul_propagator>(3).run();
        propagator_tester<div_propagator>(3).inaccurate().run();
        propagator_tester<mod_propagator>(3).inaccurate().run();
        propagator_tester<pow_propagator>(3).inaccurate().run();
        propagator_tester<sum_mod_propagator>(3, 2, gen(2)).run();
        propagator_tester<sum_mod_propagator>(gen(6)+1, 2, gen(2)).run();
        propagator_tester<sum_mod_propagator>(gen(6)+1, 3, gen(3))
            .inaccurate().run();
        {
            unsigned size= gen(5)+1;
            propagator_tester<sum_mod_propagator>(gen(6)+1, size, gen(size))
               .inaccurate().run();
            size= random::range(10,20);
            propagator_tester<sum_mod_propagator>(gen(2)+1, size, gen(size))
               .inaccurate().run();
        }            
        propagator_tester<sum_wrap_propagator>(3).run();
        propagator_tester<sum_truncate_propagator>(3).inaccurate().run();
        propagator_tester<sum_wrap_propagator>(range(2,4)).run();
        propagator_tester<less_equal_sum_propagator>(3).run();
        propagator_tester<less_equal_sum_propagator>(range(2,5)).run();
        propagator_tester<greater_equal_sum_propagator>(3).run();
        propagator_tester<greater_equal_sum_propagator>(range(2,5)).run();
        propagator_tester<equal_propagator>(2).run();
        propagator_tester<element_propagator>(range(3,7), range(-3,3)).run();
        propagator_tester<if_then_else_propagator>(4).set_dom(1, "0,1").run();
        propagator_tester<is_equal_propagator>(3).set_dom(0, "0,1").run();
        propagator_tester<literal_iff_less_than_propagator>(3, random_var())
            .run();
        propagator_tester<literal_or_difference_ge_propagator>(3,random_var(),0)
            .run();
        propagator_tester<literal_or_difference_ge_propagator>(
            3,random_var(), random_var_maybe_out_of_bound(1)).run();
        propagator_tester<literal_or_equal_propagator>(3,random_var()).run();
        propagator_tester<difference_ge_propagator>(2,0).run();
        propagator_tester<difference_ge_propagator>(
            2,random_var_maybe_out_of_bound(1)).run();
        propagator_tester<no_overlap_propagator>(2, gen(6)+1, gen(6)+1).run();
        propagator_tester<no_sub_interval_propagator>(
            2, random_var_maybe_out_of_bound(iset::MIN_DOMAIN/2), 
            random_var_maybe_out_of_bound(iset::MAX_DOMAIN/2)).run();
        propagator_tester<no_sub_interval_propagator>(
            2, random_var_maybe_out_of_bound(iset::MIN_DOMAIN/2), 
            random_var_maybe_out_of_bound(iset::MAX_DOMAIN/2),
            random_var_maybe_out_of_bound(1)).run();
        propagator_tester<sub_interval_propagator>(
            2, random_var_maybe_out_of_bound(iset::MIN_DOMAIN/2), 
            random_var_maybe_out_of_bound(iset::MAX_DOMAIN/2)).run();
        propagator_tester<sub_interval_propagator>(
            2, random_var_maybe_out_of_bound(iset::MIN_DOMAIN/2), 
            random_var_maybe_out_of_bound(iset::MAX_DOMAIN/2))
            .set_dom(0, iset().complement())
            .set_dom(1, iset().complement())
            .run();
        propagator_tester<min_propagator>(3).run();
        propagator_tester<min_propagator>(range(2,4)).run();
        propagator_tester<max_propagator>(3).run();
        propagator_tester<max_propagator>(range(2,4)).run();
    }
    for (unsigned i=0 ; i < 100 ; ++i) {
        support_table_propagator::memory_manager manager;
        for (unsigned j =0 ; j < 10 ; ++j) {
            vector<vector<iset::value_type> > tuples;
            unsigned arity= i+j==0 ? 1 : range(2,4);
            random_table(tuples, arity);
            propagator_tester<support_table_propagator>(arity, manager,
                                                        tuples).run();
        }
    }
    for (unsigned i=0 ; i < 10 ; ++i) {
        propagator_tester<if_then_else_propagator>(4).run();
        propagator_tester<is_equal_propagator>(3).run();
    }
    propagator_tester<sum_wrap_propagator>(1).run();
    
    const vector<vector<iset::value_type> > empty_tuples;
    support_table_propagator::memory_manager manager;
    propagator_tester<support_table_propagator>(gen(4), manager,
                                                empty_tuples).run();
    END_TEST_MAIN;
}

