// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2013, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2013, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */

#include <sliced_runtime_info.hpp>
#include <clause_propagator.hpp>
#include "propagator_tester.hpp"
#include "stateless_clause_propagator.hpp"

#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace ::hcsp::random;
using namespace hcsp::test;

/**
 * @brief A propagator for ((v_j) || [clause with v_n,v_{n+1},...,v_{n+k}])
 *
 * This constraint has many redundant variables which we will ignore.
 * The relation between 'j' and 'n' from the definition above is weakly set
 * and only requires that j<n || j>n+k && ... .
 * Also, if the total number of arguments is M then n+k < M && j <M.
 * Other than that there are not limits.
 */
class padded_clause : public constraint_propagator
{
  public:
    /**
     * @brief Construct a concrete constraint .
     *
     * The constraint to construct is
     *    ((v_j) || [v_n \in lit_n || ... || v_{n+k} \in lit_{n+k}])
     *
     * @param n_params       The total number of parameters
     * @param literals       The literals (lit_n, lit_1 ,..., lit_{n+k} above)
     * @param literal_param  The index of the literal (j of the v_j above).
     * @param clause_param   The index of the first literal of the clause
     *                       (n of the v_n<v_{n+k} above).
     */
    padded_clause(unsigned n_params,
                  vector<iset> literals,
                  unsigned literal_param,
                  unsigned clause_param)
        : m_n_params(n_params),
          m_literal_param(literal_param),
          m_clause_param(clause_param)
    {
        HCSP_ASSERT(literal_param < n_params && clause_param < n_params);
        HCSP_ASSERT(n_params < 10000);
        HCSP_ASSERT(literal_param < clause_param ||
                   literal_param >= clause_param+literals.size());
        m_clause_algo.update_literals()= literals;
    }
    padded_clause* clone() const     { return new padded_clause(*this); }
    virtual int count_params() const { return m_n_params; }
    unsigned clause_size() const     { return m_clause_algo.literals().size(); }

    virtual bool init(constraint_runtime_info & info)
    {
        sliced_runtime_info sliced(info, m_clause_param,
                                   m_clause_param+clause_size());
        if (!m_clause_algo.init(sliced))
            return info.domain(m_literal_param)!=iset(0);
        for (unsigned i=0 ; i != m_n_params ; ++i) {
            if (!is_clause_param(i) && !info.domain(i).singleton())
                info.allocate_watch(i, MODIFICATION_WATCH);
        }
        return true;
    }
    virtual bool constraint_check(const constraint_runtime_info & info) const
    {
        sliced_runtime_info sliced(info, m_clause_param,
                                   m_clause_param+clause_size());
        return (info.domain(m_literal_param).min() ||
                m_clause_algo.constraint_check(sliced));
    }
    virtual result_type propagate(constraint_runtime_info & info)
    {
        bool clause_never_required= !info.domain(m_literal_param).contains(0);
        if (clause_never_required)
            return VALID;
        
        sliced_runtime_info sliced(info, m_clause_param,
                                   m_clause_param+clause_size());
        bool clause_always_required= info.domain(m_literal_param) == iset(0);
        if (clause_always_required) {
            if (gen(2))
                sliced.set_querying_domain(false);
            result_type clause_result= m_clause_algo.propagate(sliced);
            if (gen(2))
                sliced.commit();
            else if (sliced.querying_domain() || gen(2))
                sliced.set_querying_domain(false);
            return clause_result;
        }
        result_type clause_result= m_clause_algo.propagate(sliced);
        if (clause_result == UNSAT) {
            info.set_domain(m_literal_param, 
                            iset(info.domain(m_literal_param)).erase(0));
            return VALID;
        } else if (clause_result == VALID && !sliced.pending_commit()) {
            return VALID;
        } else {
            return CONSISTENT;
        }
    }
    virtual bool conflicts(const constraint_runtime_info & info) const
    {
        bool clause_never_required= !info.domain(m_literal_param).contains(0);
        if (clause_never_required)
            return false;
        
        sliced_runtime_info sliced(info, m_clause_param,
                                   m_clause_param+clause_size());
        bool clause_always_required= info.domain(m_literal_param) == iset(0);
        if (!clause_always_required)
            return false;
        return m_clause_algo.conflicts(sliced);
    }
    unsigned count_watches() const { return m_clause_algo.count_watches(); }

    unsigned watched_param_index(unsigned watch_index) const
    { return m_clause_algo.watched_param_index(watch_index)+m_clause_param; }

    void forget_watches()
    { m_clause_algo.forget_watches(); }

    void explain_propagation(
        unsigned propagated_param, const iset & values_to_explain,
        const iset & propagated_domain, 
        const constraint_runtime_info & info_before_propagation,
        std::vector<iset> & explanation) const
    {
        sliced_runtime_info sliced(info_before_propagation, m_clause_param,
                                   m_clause_param+clause_size());

        unsigned propagated_clause_param= 0;
        iset propagated_clause_domain;
        if (is_clause_param(propagated_param) ) {
            propagated_clause_param= propagated_param-m_clause_param;
            propagated_clause_domain= propagated_domain;
        }
        m_clause_algo.explain_propagation(
            propagated_clause_param, values_to_explain,
            propagated_clause_domain, sliced, explanation);
        
        explanation.insert(explanation.begin(), m_clause_param, iset());
        explanation.resize(m_n_params, iset());

        explanation[m_literal_param].assign(0);
        explanation[m_literal_param].complement();
        if (propagated_param == m_literal_param)
            T_ASSERT(!propagated_domain.contains(0));
    }
    bool is_clause_param(unsigned param) const
    { return param >= m_clause_param && param < m_clause_param+clause_size(); }
    
    virtual void print(std::ostream &out,
                       const constraint_runtime_info *p_info=NULL)const throw()
    {
        out << "{{literal parameter index=" << m_literal_param 
            << ", clause range[" << m_clause_param << ","
            << m_clause_param+ clause_size()-1 << "]\n";
        out << "clause = ";
        m_clause_algo.print(out, NULL);
        out << "}}\n";
        if (p_info)
            p_info->print(out);
    }

  private:
    unsigned  m_n_params;
    unsigned  m_literal_param;
    unsigned  m_clause_param;
    clause_propagation_algo m_clause_algo;
}; // class padded_clause

int main(int argc, char *argv[])
{
    HCSP_PROFILE_CURRENT;
    BEGIN_TEST_MAIN;
    cout << "Seed is " << set_seed(argc, argv) << endl;
    for (unsigned i=0 ; i < 10000 ; ++i) {
        unsigned n_literals= range(1,4);
        unsigned n_params= range(n_literals+1, n_literals+4);
        unsigned clause_param= gen(n_params-n_literals);
        unsigned literal_param= gen(n_params-n_literals);
        if (literal_param >= clause_param)
            literal_param += n_literals;
        vector<iset> literals;
        for (unsigned j=0 ; j != n_literals ; ++j)
            literals.push_back(random_nonempty_var(true, 1, 4));
        padded_clause clause (n_params,literals, literal_param, clause_param);
        propagator_tester<padded_clause>(n_params, clause).run();
    }
    END_TEST_MAIN;
}
