// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <cmath>
#include <algorithm>
#include <stdexcept>
#include <iostream>
#include <sstream>
#include <vector>
#include <iset.hpp>
#include <utilities.hpp>
#include <tuple_set.hpp>
#include "test_utilities.hpp"
using namespace std;
using namespace hcsp;
using namespace hcsp::random;

// Random tester class for tuple_set
// It will inject random cases and check their outcomes.
class tester {
  public:
    tester();
    template<unsigned tuple_count, unsigned arity>
    tester(const int (& tuples)[tuple_count][arity]);
    void run();
    const tuple_set & current_set() const { return m_tuple_set; }
  private:
    //
    // Testing methods
    //

    static void to_int_tuples(const vector< vector<iset> > & iset_tuples,
                              vector<vector<int> > & int_tuples);

    static void insert_tuples_from_iset_tuple(const vector<iset>&in_iset_tuple,
                                              vector<int> & int_stack,
                                              vector<vector<int> >&int_tuples);
    
    static size_t random_arity();

    
    //! For domains.size() elements create a random set.
    static void random_domains(vector<iset> & domains);

    static size_t random_tuple_count(const vector<iset> & domains);
    
    vector<vector<int> > m_tuples;
    tuple_set m_tuple_set;

}; // class Tester

tester::tester()
{
    size_t arity= random_arity();
    vector<iset> domains(arity);
    random_domains(domains);
    
    size_t tuple_count= random_tuple_count(domains);
    m_tuples.resize(tuple_count);
    for (unsigned i=0 ; i != tuple_count ; ++i) {
        vector<int> & tuple= m_tuples[i];
        tuple.resize(domains.size());
        for (unsigned j=0 ; j != domains.size() ; ++j) {
            tuple[j]= domains[j].rand();
        }
    }
}

template<unsigned tuple_count, unsigned arity>
tester::tester(const int (&tuples)[tuple_count][arity])
{
    m_tuples.resize(tuple_count);
    for (unsigned i=0 ; i != tuple_count ; ++i) {
        vector<int> & tuple= m_tuples[i];
        tuple.resize(arity);
        for (unsigned j=0 ; j != arity ; ++j) {
            tuple[j]=tuples[i][j];
        }
    }
}

void tester::run()
{
    m_tuple_set.insert(m_tuples);
    vector<vector<int> > int_tuples;

    to_int_tuples(m_tuple_set.tuples(), int_tuples);
    sort(int_tuples.begin(), int_tuples.end());
    int_tuples.erase(unique(int_tuples.begin(), int_tuples.end()),
                     int_tuples.end());
    vector<vector<int> > orig_unique_tuples= m_tuples;
    sort(orig_unique_tuples.begin(), orig_unique_tuples.end());
    orig_unique_tuples.erase(unique(orig_unique_tuples.begin(),
                                    orig_unique_tuples.end()),
                             orig_unique_tuples.end());
    
    T_ASSERT(orig_unique_tuples == int_tuples);
    
}
// static
void tester::insert_tuples_from_iset_tuple(const vector<iset> & in_iset_tuple,
                                           vector<int> & int_stack,
                                           vector<vector<int> > & int_tuples)
{
    if ( int_stack.size() == in_iset_tuple.size() ) {
        int_tuples.push_back(int_stack);
        return;
    }
    size_t column= int_stack.size();
    const iset & current_set = in_iset_tuple[column];
    int_stack.push_back(0);
    for (iset::iterator it= current_set.begin() ; it!=current_set.end(); ++it){
        int_stack.back()= *it;
        insert_tuples_from_iset_tuple(in_iset_tuple, int_stack, int_tuples);
    }
    int_stack.pop_back();
}

// static
void tester::to_int_tuples(const vector< vector<iset> > & iset_tuples,
                           vector<vector<int> > & int_tuples)
{
    int_tuples.clear();
    vector<int> int_stack;
    for (unsigned i=0 ; i != iset_tuples.size() ; ++i) {
        insert_tuples_from_iset_tuple(iset_tuples[i], int_stack, int_tuples);
    }
}

// static
size_t tester::random_arity()
{
    if (gen(1000) == 0)
        return 0;
    if (gen(100) == 0)
        return 1;
    if (gen(2) == 0)
        return 2;
    else
        return range(3,4);
}

// static
void tester::random_domains(vector<iset> & domains)
{
    for (unsigned i=0 ; i != domains.size() ; ++i) {
        iset & domain= domains[i];
        domain.clear();
        size_t domain_size= range(2, 4);
        while (domain.size() < domain_size)
            domain.insert(range(0, 20));
    }
}

// static
size_t tester::random_tuple_count(const vector<iset> & domains)
{
    const size_t max_tuple_count= 10000;
    if (domains.size() == 0)
        return range(1,4);
    if (gen(100) == 0)
        return 1;
    if (gen(100) == 1)
        return 2;

    unsigned wanted_max=1;
    for (unsigned i=0 ; i != domains.size() ; ++i) {
        wanted_max *= domains[i].size()*2;
        if (wanted_max > max_tuple_count) {
            wanted_max= max_tuple_count;
            break;
        }
    }
    wanted_max= max(3U, wanted_max);
    return range(3U, wanted_max);
}
    
int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    cout << "Seed is " << set_seed(argc, argv) << endl;
    tuple_set empty_set;
    empty_set.insert(vector<vector<int> >());
    T_ASSERT(empty_set.tuples().empty());
    const int tuples[][3] = { { 1, 1 ,1},
                              { 2, 1, 1},
                              { 1, 1, 2},
                              { 1, 2, 1},
                              { 2, 2, 1}};
    tester targeted_test(tuples);
    targeted_test.run();
    T_ASSERT(targeted_test.current_set().tuples().size() == 2);

    for (unsigned i=0 ; i < 3000 ; ++i) {
        tester().run();
    }
    
    END_TEST_MAIN;
}

