// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <algorithm>
#include <utilities.hpp>
#include <unique_queue.hpp>
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;



template <class T>
class tester
{
  public:
    typedef typename unique_queue<T>::value_type value_type;
    //
    // Public variables
    //

    //
    // Public methods
    //
    tester() {}
    void run();
  private:
    //
    // Private types
    //

    void set_max_elements(unsigned max_elements);
    unsigned get_max_elements() const
    {
        unsigned ret= m_typed_queue.get_max_elements();
        T_ASSERT(ret == m_untyped_queue.get_max_elements() );
        return ret;
    }
    bool insert(unsigned index, const T & data);
    void prohibit_insertion(unsigned element_id);
    bool empty() const
    {
        T_ASSERT(m_untyped_queue.empty() == m_typed_queue.empty());
        return m_typed_queue.empty();
    }
    value_type pop();
    void clear();
    void swap(tester & other)
    {
        m_typed_queue.swap(other.m_typed_queue);
        m_untyped_queue.swap(other.m_untyped_queue);
    }
    T next_mapped() const;
    void test_max_elements();
    void test_insertion_and_pop();
    void restart()
    {
        if (hcsp::random::gen(2))
            tester().swap(*this);
        else {
            clear();
            set_max_elements(0);
        }
    }
    
    //
    // Private methods
    //
    unique_queue<T> m_typed_queue;
    unique_queue<> m_untyped_queue;

}; // class tester


template <class T>
T tester<T>::next_mapped() const
{
    return hcsp::random::gen();
}

template <>
string tester<string>::next_mapped() const
{
    unsigned len= hcsp::random::range(0,5);
    string ret;
    for (unsigned i=0 ; i < len ; ++i)
        ret.push_back(gen(256));
    return ret;
}


template <class T>
void tester<T>::set_max_elements(unsigned max_elements)
{
    bool out_of_range_thrown= false;
    try {
        m_typed_queue.set_max_elements(max_elements);
    } catch (out_of_range) {
        out_of_range_thrown= true;
    } catch (...) {
        T_ASSERT(0);
    }
    try {
        m_untyped_queue.set_max_elements(max_elements);
        T_ASSERT( !out_of_range_thrown) ;
    } catch (out_of_range) {
        T_ASSERT( out_of_range_thrown );
        throw;
    } catch (...) {
        T_ASSERT(0);
    }
}


template <class T>
bool tester<T>::insert(unsigned index, const T & data)
{
    bool out_of_range_thrown= false;
    bool inserted= false;
    try {
        value_type typed(index, data);
        inserted= m_typed_queue.insert(typed);
    } catch (out_of_range) {
        out_of_range_thrown= true;
    } catch (...) {
        T_ASSERT(0);
    }
    
    try {
        T_ASSERT( m_untyped_queue.insert(index) == inserted );
    } catch (out_of_range) {
        T_ASSERT(out_of_range_thrown);
        throw;
    } catch (...) {
        T_ASSERT(0);
    }
    T_ASSERT(!out_of_range_thrown);
    return inserted;
}

template <class T>
void tester<T>::prohibit_insertion(unsigned index)
{
    bool out_of_range_thrown= false;
    try {
        m_typed_queue.prohibit_insertion(index);
    } catch (out_of_range) {
        out_of_range_thrown= true;
    } catch (...) {
        T_ASSERT(0);
    }
    
    try {
        m_untyped_queue.prohibit_insertion(index);
    } catch (out_of_range) {
        T_ASSERT(out_of_range_thrown);
        throw;
    } catch (...) {
        T_ASSERT(0);
    }
    T_ASSERT(!out_of_range_thrown);
}

template <class T>
typename tester<T>::value_type tester<T>::pop()
{
    value_type ret;
    bool out_of_range_thrown= false;
    try {
        ret= m_typed_queue.pop();
    } catch (out_of_range) {
        out_of_range_thrown= true;
    } catch (...) {
        T_ASSERT(0);
    }
    
    try {
        unsigned index= m_untyped_queue.pop();
        T_ASSERT( !out_of_range_thrown );
        T_ASSERT( ret.key == index );
    } catch (out_of_range) {
        T_ASSERT(out_of_range_thrown);
        throw;
    } catch (...) {
        T_ASSERT(0);
    }
    T_ASSERT(!out_of_range_thrown);
    return ret;
}

template <class T>
void tester<T>::clear()
{
    try {
        m_typed_queue.clear();
    } catch (...) {
        T_ASSERT(0);
    }
    
    try {
        m_untyped_queue.clear();
    } catch (...) {
        T_ASSERT(0);
    }
    T_ASSERT(empty());
}

template <class T>
void tester<T>::run()
{
    test_max_elements();      restart();
    test_max_elements();      restart();
    test_insertion_and_pop(); restart();
    test_insertion_and_pop(); restart();
}

template <class T>
void tester<T>::test_max_elements()
{
    T a0= next_mapped(), a1= next_mapped(), a2= next_mapped();
    T_ASSERT_EXCEPTION(insert(0, next_mapped()), out_of_range);
    T_ASSERT_EXCEPTION(insert(1, next_mapped()), out_of_range);
    T_ASSERT(get_max_elements() == 0);
    set_max_elements(2);
    prohibit_insertion(1);
    T_ASSERT(insert(0, a0));
    T_ASSERT(get_max_elements() == 2);
    set_max_elements(3);
    set_max_elements(1);        // Effectively delete prohibit_insertion(1)
    T_ASSERT_EXCEPTION(set_max_elements(0), out_of_range);
    set_max_elements(3);
    T_ASSERT(get_max_elements() == 3);
    T_ASSERT(insert(2, a2));
    T_ASSERT(insert(1, a1));
    T_ASSERT_EXCEPTION(insert(3, next_mapped()), out_of_range);
    T_ASSERT_EXCEPTION(insert(4, next_mapped()), out_of_range);

    restart();

    T_ASSERT_EXCEPTION(prohibit_insertion(0), out_of_range);
    T_ASSERT_EXCEPTION(prohibit_insertion(1), out_of_range);
    T_ASSERT(get_max_elements() == 0);
    set_max_elements(2);
    T_ASSERT(insert(1, a1));
    T_ASSERT(insert(0, a0));
    T_ASSERT(get_max_elements() == 2);
    set_max_elements(3);
    set_max_elements(2);
    T_ASSERT_EXCEPTION(set_max_elements(1), out_of_range);
    set_max_elements(3);
    T_ASSERT(get_max_elements() == 3);
    T_ASSERT(insert(2, a2));
}

template <class T>
void tester<T>::test_insertion_and_pop()
{
    T a0= next_mapped(), a1= next_mapped(), a2= next_mapped();
    
    if (hcsp::random::gen(10)== 0)
        a0= a2;
    if (hcsp::random::gen(10)== 0)
        a1= a2;
    if (hcsp::random::gen(10)== 0)
        a0= a1;
    T_ASSERT(empty());
    set_max_elements(3);
    T_ASSERT(empty());
    prohibit_insertion(1);
    T_ASSERT(empty());
    T_ASSERT(insert(2, a2));             T_ASSERT(!empty());
    T_ASSERT(insert(0, a0));             T_ASSERT(!empty());
    prohibit_insertion(0);// to be ignored
    prohibit_insertion(1);// to be ignored
    prohibit_insertion(2);// to be ignored
    T_ASSERT(!insert(1, a1));            T_ASSERT(!empty());
    T_ASSERT(!insert(2, a2));            T_ASSERT(!empty());
    T_ASSERT(!insert(0, a0));            T_ASSERT(!empty());
    T_ASSERT(pop() == value_type(2,a2)); T_ASSERT(!empty());
    T_ASSERT(!insert(1, a1));            T_ASSERT(!empty());
    T_ASSERT(!insert(0, a0));            T_ASSERT(!empty());
    if (hcsp::random::gen(2))
        a2= next_mapped();
    T_ASSERT(insert(2, a2));             T_ASSERT(!empty());
    
    T_ASSERT(pop() == value_type(0,a0)); T_ASSERT(!empty());
    T_ASSERT(pop() == value_type(2,a2)); T_ASSERT(empty());
    T_ASSERT_EXCEPTION(pop(), out_of_range);
    T_ASSERT(!insert(1, a1));            T_ASSERT(empty());
    clear();
    T_ASSERT(insert(1, a1));             T_ASSERT(!empty());
    T_ASSERT(insert(2, a2));             T_ASSERT(!empty());
    T_ASSERT(insert(0, a0));             T_ASSERT(!empty());
    T_ASSERT(pop() == value_type(1,a1)); T_ASSERT(!empty());
    T_ASSERT(pop() == value_type(2,a2)); T_ASSERT(!empty());
    T_ASSERT(pop() == value_type(0,a0)); T_ASSERT(empty());
}

void assert_key_type()
{
    HCSP_STATIC_ASSERT_SAME_TYPE(unique_queue<>::key_type,       unsigned);
    HCSP_STATIC_ASSERT_SAME_TYPE(unique_queue<char>::key_type,   unsigned);
    HCSP_STATIC_ASSERT_SAME_TYPE(unique_queue<string>::key_type, unsigned);
}

void assert_value_type()
{
    HCSP_STATIC_ASSERT_SAME_TYPE(unique_queue<>::value_type,       unsigned);
    HCSP_STATIC_ASSERT_SAME_TYPE(unique_queue<char>::value_type::key_type,
                                unsigned);
    HCSP_STATIC_ASSERT_SAME_TYPE(unique_queue<char>::value_type::mapped_type,
                                char);
    
    HCSP_STATIC_ASSERT_SAME_TYPE(unique_queue<string>::value_type::key_type,
                                unsigned);
    HCSP_STATIC_ASSERT_SAME_TYPE(unique_queue<string>::value_type::mapped_type,
                                string);
}

void assert_value_argument_type()
{
    HCSP_STATIC_ASSERT_SAME_TYPE(unique_queue<>::value_argument_type, unsigned);
    HCSP_STATIC_ASSERT_SAME_TYPE(unique_queue<char>::value_argument_type,
                                const unique_queue<char>::value_type&);
    HCSP_STATIC_ASSERT_SAME_TYPE(unique_queue<string>::value_argument_type,
                                const unique_queue<string>::value_type&);
}


int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    cout << "Seed is " << set_seed(argc, argv) << endl;
    for (unsigned i=0 ; i < 100 ; ++i) {
        tester<char>().run();
        tester<string>().run();
    }
    END_TEST_MAIN;
}

