// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#undef NDEBUG
#include <limits>
#include <iostream>
#include <cmath>
#include <limits>
#include <utilities.hpp>

#include "test_utilities.hpp"
#include "test_static_assert.hpp"
// Mulitple inclusion works since test_static_assert does not have
// the regular #ifdef protection
// This tests if HCSP_STATIC_ASSERT works well if there is more than one
// HCSP_STATIC_ASSERT that shares the same line number (probably in different
// source files).
#include "test_static_assert.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;

// Random tester class for various utilities in utilities.hpp
// It will inject random cases and check their outcomes.
class Tester {
  public:
    Tester() {}
    void Run();
    inline void TestMsbIndex(uint32_t data) const;
    inline void TestLsbIndex(uint32_t data) const;
    inline void TestRangeCount(uint32_t data) const;
    inline void TestBits(uint32_t data) const;
    static void TestDiv(int32_t dividend, int32_t divisor);
    template <class T, class U>
    static void TestIsPower2();
    template <class T>
    static void TestAllBitsSet();
    static bool IsDiophantineSolution(unsigned a, unsigned b, unsigned rhs,
                                      long long x, long long y);
    static void TestCspAssert();
    static void TestBezout(unsigned a, unsigned b);
    static void TestExhaustiveBezout();
    static void TestLinearDiophantine(unsigned a, unsigned b, unsigned rhs);
    static void TestExhaustiveDiophantine();
    static void ExhaustiveTests();
    HCSP_STATIC_ASSERT(sizeof(uint32_t)*8 == 32);
    HCSP_STATIC_ASSERT(sizeof(uint16_t)*8 == 16);
    HCSP_STATIC_ASSERT_SAME_TYPE(int, int);
    typedef const char my_char5[5];
    HCSP_STATIC_ASSERT_SAME_TYPE(my_char5, my_char5);
}; // class Tester

HCSP_STATIC_ASSERT(sizeof(uint32_t)*8 == 32);
HCSP_STATIC_ASSERT(sizeof(uint16_t)*8 == 16);
HCSP_STATIC_ASSERT_SAME_TYPE(const double, const double);

void Tester::TestLsbIndex(uint32_t data) const
{
    unsigned index;
    for (index=0 ; ((data >> index) & 1) == 0 && index < 31 ; ++index) {
    }
    T_ASSERT(index == find_lsb_index(data));
}


void Tester::TestMsbIndex(uint32_t data) const
{
    unsigned index;
    for (index=0 ; (data >> index) > 1 ; ++index) {
    }
    T_ASSERT(index == find_msb_index(data));
        
}

void Tester::TestRangeCount(uint32_t data) const
{
    unsigned count=0;
    uint32_t work_data= data;
    while (work_data) {
        while ( (work_data & 1) == 0 ) {
            work_data >>= 1;
        }
        while ( (work_data & 1) == 1 ) {
            work_data >>= 1;
        }
        ++count;
    }
    T_ASSERT(count == count_ranges_of_one(data));
}

inline void Tester::TestBits(uint32_t data) const
{
    TestLsbIndex(data);
    TestMsbIndex(data);
    TestRangeCount(data);
}

//static
void Tester::TestDiv(int32_t dividend, int32_t divisor)
{
    int32_t ceil_expected= (int32_t)ceil(1.0*dividend/divisor);
    HCSP_ASSERT(ceil_div(dividend, divisor) == ceil_expected);

    int32_t floor_expected= (int32_t)floor(1.0*dividend/divisor);
    HCSP_ASSERT(floor_div(dividend, divisor) == floor_expected);
}


void Tester::Run()
{
    uint32_t mostlyZero= ~uint32_t(0);
    uint32_t mostlyOne= 0;
    for (int sparseness= gen(6) ; sparseness >= 0 ; --sparseness) {
        mostlyOne  |= gen();
        mostlyZero &= gen();
    }
    TestBits(mostlyZero);
    TestBits(mostlyOne);
    TestBits(gen() >> gen(32));
    TestBits(gen() << gen(32));
    int divisor= gen() >> hcsp::random::range(1,32);
    if (divisor == 0)
        divisor=1;
    if (gen(2) == 0)
        divisor= -divisor;
    TestDiv(divisor*(gen() >> gen()), divisor);
    TestDiv(divisor*(gen() >> gen())+1, divisor);
    TestDiv(divisor*(gen() >> gen())-1, divisor);
    TestDiv(divisor*-(gen() >> gen()), divisor);
    TestDiv(divisor*-(gen() >> gen())+1, divisor);
    TestDiv(divisor*-(gen() >> gen())-1, divisor);
    TestDiv(gen(), divisor);
    TestBezout(gen(), gen());
    unsigned a= gen(), b=gen(), g=gen();
    TestLinearDiophantine(a,b,g);
    unsigned next_g=g ;
    if (a == 0 && b == 0)
        next_g= 0;
    else if (a==0)
        next_g= g - g%b;
    else if (b == 0)
        next_g= g - g%a;
    else if (g % gcd(a,b) != 0)
        next_g=  g- g% gcd(a,b);
    if (g != next_g)
        TestLinearDiophantine(a,b,next_g);
    
    TestIsPower2<long long,          unsigned long long>();
    TestIsPower2<unsigned long long, unsigned long long>();
    TestIsPower2<long,               unsigned long>();
    TestIsPower2<unsigned long,      unsigned long>();
    TestIsPower2<int,                unsigned>();
    TestIsPower2<unsigned,           unsigned>();
    TestIsPower2<short,              unsigned short>();
    TestIsPower2<unsigned short,     unsigned short>();
    TestIsPower2<signed char,        unsigned char>();
    TestIsPower2<unsigned char,      unsigned char>();
}

//static
template <class T, class U>
void Tester::TestIsPower2()
{
    HCSP_STATIC_ASSERT(sizeof(T) == sizeof(U));
    U udata= 0;
    for ( unsigned bytes= 0 ; sizeof(U) > bytes ; bytes += sizeof(gen())) {
        unsigned shift_bits= sizeof(gen())*8;
        if (sizeof(udata)*8 > shift_bits)
            udata= (udata << shift_bits) ^ gen();
        else
            udata= gen();
    }
    T data= udata >> gen(sizeof(U)*8) << gen(sizeof(U)*8);
    if (data <= 0)
        T_ASSERT(!is_power_of2(data));
    else {
        unsigned msb_index=0;
        while (data >> msb_index >> 1)
            ++msb_index;
        if ( static_cast<T>(1) << msb_index == data)
            T_ASSERT(is_power_of2(data));
        else 
            T_ASSERT(!is_power_of2(data));
    }
}

//static
template <class T>
void Tester::TestAllBitsSet()
{
    const unsigned num_of_bits= sizeof(T)*8;
    T expected_on_and_above= ~static_cast<T>(0);
    T expected_above, expected_below, expected_on_and_below;

    // Test for bits out of range - on purpose.
    for (unsigned i=0 ; i != num_of_bits*4 ; ++i) {
        expected_above        = expected_on_and_above<<1;
        expected_below        = ~expected_on_and_above;
        expected_on_and_below = ~expected_above;

        T_ASSERT(expected_on_and_above == unsigned_ones_on_and_above<T>(i));
        T_ASSERT(expected_above        == unsigned_ones_above<T>(i));
        T_ASSERT(expected_on_and_below == unsigned_ones_on_and_below<T>(i));
        T_ASSERT(expected_below        == unsigned_ones_below<T>(i));
        
        expected_on_and_above = expected_above;
    }
    unsigned max= ~static_cast<unsigned>(0);
    T_ASSERT(expected_on_and_above == unsigned_ones_on_and_above<T>(max));
    T_ASSERT(expected_above        == unsigned_ones_above<T>(max));
    T_ASSERT(expected_on_and_below == unsigned_ones_on_and_below<T>(max));
    T_ASSERT(expected_below        == unsigned_ones_below<T>(max));
}

//static
void Tester::TestCspAssert()
{
    bool did_fail= false;
    volatile int a=1;
    volatile int b=2;
    HCSP_ASSERT(b==a+1);
    
    // Make sure that "if" nesting works correctly with HCSP_ASSERT;
    if (did_fail)
        if (did_fail)
            HCSP_ASSERT(0);
        else
            HCSP_ASSERT(0);
    else
        if (did_fail)
            HCSP_ASSERT(0);
        else
            try {
                HCSP_ASSERT(a==b);
            } catch (std::runtime_error &) {
                did_fail= true;
            }
    if (!did_fail)
        throw runtime_error("Assertion should have failed");
}

// static
void Tester::TestBezout(unsigned a, unsigned b)
{
    unsigned test_gcd;
    pair<long long, long long> solution= bezout_coefficients(a,b, &test_gcd);
    T_ASSERT(test_gcd == gcd(a,b));
    T_ASSERT((long long)solution.first*a + (long long)solution.second*b
             ==
             test_gcd);
    T_ASSERT(solution.first >= 0);
    if (b != 0) {
        long long earlier_solution= solution.first-b/test_gcd;
        if (earlier_solution >= 0)
            cout << solution.first << '*' << a << " + "
                 << solution.second << '*' << b << " = " << test_gcd<< endl
                 << "earlier_solution=" << earlier_solution << endl;
        T_ASSERT(earlier_solution < 0);
    }
}

//static
void Tester::TestExhaustiveBezout()
{
    // i and j wrap around from big numbers back to small, so that
    // both big and small values are tested (including zero!)
    for (unsigned i= -1U-500; i != 500 ; ++i)
        for (unsigned j= -1U-500; j != 500 ; ++j)
            TestBezout(i, j);
}

// static
bool Tester::IsDiophantineSolution(unsigned a, unsigned b, unsigned rhs,
                                   long long x, long long y)
{
    double approx_rhs= (double)a*x-(double)b*y;
    double epsilon1= numeric_limits<double>::epsilon()*a*x;
    double epsilon2= numeric_limits<double>::epsilon()*b*y;
    double epsilon= fabs(epsilon1)+fabs(epsilon2);
    if (abs(rhs-approx_rhs) > epsilon)
        return false;
    return true; // FIXME more!!
}

// test the solution of a*x-b*y == rhs
// static
void Tester::TestLinearDiophantine(unsigned a, unsigned b, unsigned rhs)
{
    long long x,y;
    unsigned step_x, step_y;

    unsigned g= gcd(a,b);
    bool ret = linear_diophantine_equation(a, b, rhs, x, step_x, y, step_y);
    
    if (!ret) {
        if (g != 0)
            T_ASSERT(rhs % g != 0);
        else if (a == 0 && b == 0)
            T_ASSERT(rhs != 0);
        else if (a == 0)
            T_ASSERT(rhs % b != 0);
        else // assume b==0 since gcd(a,b)==0 and a !=0.
            T_ASSERT(rhs % a != 0);
        return;
    } 
    T_ASSERT(x >= 0);
    if (b != 0) {
        T_ASSERT(x <= step_x);
        if (step_x > 0)
            T_ASSERT(x < step_x);
    }
    T_ASSERT(IsDiophantineSolution(a, b, rhs, x, y));
    // split into two cases to avoid overflows in y. Due to its limited range,
    // x can't overflow here.
    if (y < 0)
        T_ASSERT(IsDiophantineSolution(a, b, rhs, x+step_x, y+step_y));
    else
        T_ASSERT(IsDiophantineSolution(a, b, rhs, x-step_x, y-step_y));

    // Check if step_x and step_y are small enough, by checking their GCD.
    // If they have a GCD>1 then step_x and step_y can be divided by GCD and
    // produce smaller step_x and step_y.
    if (step_x == 0  || step_y == 0)
        T_ASSERT(step_y <= 1 && step_x <= 1);
    else
        T_ASSERT(gcd(step_x, step_y) == 1);
}

//static
void Tester::TestExhaustiveDiophantine()
{
    // i and j wrap around from big numbers back to small, so that
    // both big and small values are tested (including zero!)
    for (unsigned i= -1U-100; i != 100 ; ++i)
        for (unsigned j= -1U-100; j != 100 ; ++j)
            for (unsigned k= -1U-100; k != 100 ; ++k)
                TestLinearDiophantine(i, j, k);
}

//static
void Tester::ExhaustiveTests()
{
    TestAllBitsSet<unsigned char>();
    TestAllBitsSet<unsigned short>();
    TestAllBitsSet<unsigned int>();
    TestAllBitsSet<unsigned long>();
    TestAllBitsSet<unsigned long long>();
    TestCspAssert();
    TestExhaustiveBezout();
    TestExhaustiveDiophantine();
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    cout << "Seed is " << set_seed(argc, argv) << endl;
    for (unsigned i=0 ; i < 10000 ; ++i) {
        Tester tester;
        tester.Run();
    }
    Tester::ExhaustiveTests();
    END_TEST_MAIN;
}
