// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2014, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2014, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <map>
#include <set>
#include <vector>
#include <functional>
#include <algorithm>
#include <utilities.hpp>
#include <var_watchers.hpp>
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;

template <class Watcher> class collector_deallocator_type;
template <class Watcher> class collector_reallocator_type;
template <class Watcher> class conditional_deallocator;

template<class Watcher>
class tester
{
  public:
    static const unsigned NUMBER_OF_WATCH_TYPES = 8;
    class change_notifier;
    friend class change_notifier;
    friend class collector_deallocator_type<Watcher>;
    friend class collector_reallocator_type<Watcher>;
    friend class conditional_deallocator<Watcher>;
    //
    // Public variables
    //

    //
    // Public methods
    //
    tester() : m_var_index(gen_unsigned()), m_number_of_watches(0)
    {}
    void run();
  private:
    //
    // Private types
    //
    struct watch_info
    {
        watch_handle_t m_watch_handle;
    };
    struct watch_specifier
    {
        watch_specifier(unsigned constraint, iset::value_type key)
            : m_constraint(constraint), m_key(key)
        {}
        bool operator<(const watch_specifier & other) const
        {
            return m_constraint < other.m_constraint ||
                (m_constraint == other.m_constraint && m_key < other.m_key);
        }
        
        unsigned         m_constraint;
        iset::value_type m_key;
    };
    typedef map<watch_specifier, watch_handle_t> constraint_to_watch_type;

    //
    // Private methods
    //
    unsigned gen_unsigned() const
    {
        switch(gen(10)) {
          case 0:  return 0;
          case 1:  return ~0U;
          default: return gen(10000);
        }
    }

    void validate_empty() const
    {
        bool expecting_empty= (m_number_of_watches == 0);
        T_ASSERT(expecting_empty == m_watchers.empty());
        T_ASSERT(expecting_empty == m_constraint_to_watch.empty());
    }
    void validate_count_all_watches() const
    {
        unsigned all_watches= m_number_of_watches;
        T_ASSERT( all_watches == m_watchers.count_all_watches() );
    }
    void validate_get_constraints() const;
    void validate_all() const;

    void insert_watches();
    void test_for_each_until_false();
    void test_for_each_with_predicate();
    void test_for_each_with_reallocation();
    void test_for_each_removal();
    void test_allocate();
    void test_allocate(watch_specifier specifier);
    void test_allocate(unsigned constraint);
    void test_deallocate();
    void test_deallocate(watch_specifier specifier);
    void test_deallocate_any(unsigned constraint);
    void deallocate_all();
    watch_specifier random_watcher();
    void get_specifiers_for_constraint(unsigned constraint,
                                       vector<watch_specifier> & spec) const;
    
    //
    // Private members
    //
    
    Watcher m_watchers;
    unsigned m_var_index;
    constraint_to_watch_type m_constraint_to_watch;
    unsigned m_number_of_watches;
}; // class tester

template <class Container>
class back_inserter_op_t : unary_function<typename Container::value_type, void>
{
  public:
    back_inserter_op_t(Container & container) : m_container(container) {}
    void operator()(const typename Container::value_type & value)
    { m_container.push_back(value); }
  private:
    Container & m_container;
};

template <class Container>
back_inserter_op_t<Container> back_inserter_op(Container & container)
{
    return back_inserter_op_t<Container>(container);
}


/**
 * @brief A functor that collects its invocation parameters and also compares
 *        the parameter.
 * @return true if the parameter is different than last_constraint_index.
 */
class collect_also_is_diff
{
  public:
    collect_also_is_diff(vector<unsigned> & vec,unsigned last_constraint_index)
        : m_vec(vec), m_last_constraint_index(last_constraint_index)
    {}
    bool operator()(unsigned constraint_index) const
    {
        m_vec.push_back(constraint_index);
        return constraint_index != m_last_constraint_index;
    }
  private:
    vector<unsigned> & m_vec;
    unsigned           m_last_constraint_index;
};


//! \brief This is a functor that adds the constraint into the 'got' vector,
//!        and removes them from the var_watchers.
template<class Watcher>
class collector_deallocator_type
{
  public:
    typedef tester<Watcher> tester_t;
    collector_deallocator_type(tester_t & the_tester, vector<unsigned> & got)
        : m_tester(the_tester), m_got(got)
    {}
    
    void operator()(unsigned constraint_index) const
    {
        m_tester.test_deallocate_any(constraint_index);
        m_got.push_back(constraint_index);
    }
  private:
    tester_t & m_tester;
    vector<unsigned> & m_got;
};


//! \brief This is a functor that deallocates and then allocates the current
//!        watch.
template<class Watcher>
class collector_reallocator_type
{
    typedef tester<Watcher> tester_t;
 public:
    collector_reallocator_type(tester_t & the_tester, vector<unsigned> & got)
        : m_tester(the_tester), m_got(got)
    {}
    
    void operator()(unsigned constraint_index) const
    {
        // Avoid infinite loops:
        T_ASSERT(m_got.size() < 100000);
        vector<typename tester_t::watch_specifier> specifiers;
        m_tester.get_specifiers_for_constraint(constraint_index, specifiers);
        random_shuffle(specifiers.begin(), specifiers.end());
        for (unsigned i=0 ; i != specifiers.size() ; ++i)
            m_tester.test_deallocate(specifiers[i]);
        random_shuffle(specifiers.begin(), specifiers.end());
        for (unsigned i=0 ; i != specifiers.size() ; ++i)
            m_tester.test_allocate(specifiers[i]);
        m_got.push_back(constraint_index);
    }
  private:
    tester_t & m_tester;
    vector<unsigned> & m_got;
};

template<class Watcher>
class conditional_deallocator
{
    typedef tester<Watcher> tester_t;
  public:
    conditional_deallocator(tester_t & the_tester,
                            typename tester_t::watch_specifier to_deallocate)
        : m_tester(the_tester), m_to_deallocate(to_deallocate)
    {}
    void operator()(unsigned constraint) const
    {
        if (constraint == m_to_deallocate.m_constraint)
            m_tester.test_deallocate(m_to_deallocate);
    }
  private:
    tester_t & m_tester;
    typename tester_t::watch_specifier m_to_deallocate;
};

template<class Watcher>
void tester<Watcher>::insert_watches()
{
    if (gen(100) == 0)
        return;
    unsigned number_of_constraints= hcsp::random::range(1,4);
    while (m_constraint_to_watch.size() < number_of_constraints)
        test_allocate();
}

template<class Watcher>
void tester<Watcher>::validate_get_constraints() const
{
    vector<unsigned> got;
    m_watchers.for_each(back_inserter_op(got));
    T_ASSERT(got.size() == m_constraint_to_watch.size());
    sort(got.begin(), got.end());

    for (unsigned i=0 ; i != got.size() ; ) {
        unsigned got_same_count=1;
        while ( i+1 != got.size() && got[i]== got[i+1] ) {
            ++got_same_count;
            ++i;
        }
        vector<watch_specifier> specifiers;
        get_specifiers_for_constraint(got[i], specifiers);
        T_ASSERT(specifiers.size() == got_same_count);
        ++i;
    }
}

template<class Watcher>
void tester<Watcher>::validate_all() const
{
    validate_empty();
    validate_count_all_watches();
    validate_get_constraints();
}

template<class Watcher>
void tester<Watcher>::test_for_each_until_false()
{
    vector<unsigned> got1, got2;
    m_watchers.for_each(back_inserter_op(got1));
    m_watchers.for_each(back_inserter_op(got2));
    T_ASSERT(got1 == got2);
    bool is_empty= got1.empty();
    unsigned last= is_empty ? gen() : got1[gen(got1.size())];

    got2.clear();
    m_watchers.for_each_until_false(collect_also_is_diff(got2, last));

    T_ASSERT(is_empty == got2.empty());
    
    if ( is_empty )
        T_ASSERT(got2.empty());
    else {
        T_ASSERT(got2.back() == last);
        // got2 must have the same prefix as got1.
        T_ASSERT(equal(got1.begin(), got1.end(), got2.begin()));
    }
}

template<class Watcher>
void tester<Watcher>::test_for_each_with_reallocation()
{
    vector<unsigned> got;
    m_watchers.for_each(back_inserter_op(got));
    sort(got.begin(), got.end());
    
    vector<unsigned> got_reallocation;
    m_watchers.for_each(collector_reallocator_type<Watcher>(
                            *this, got_reallocation) );
    sort(got_reallocation.begin(), got_reallocation.end());
    T_ASSERT(got_reallocation == got);
    
}

template<class Watcher>
void tester<Watcher>::test_allocate()
{
    test_allocate( gen_unsigned() );
}

template<class Watcher>
void tester<Watcher>::test_allocate(unsigned constraint)
{
    unsigned key= gen(NUMBER_OF_WATCH_TYPES);
    test_allocate( watch_specifier(constraint, key) );
}

template<class Watcher>
void tester<Watcher>::test_allocate(watch_specifier specifier)
{
    auto inserted=
        m_constraint_to_watch.insert(make_pair(specifier, watch_handle_t()));
    if ( !inserted.second )
        return;
    ++m_number_of_watches;
    auto it= inserted.first;
    it->second= m_watchers.allocate_watch(it->first.m_key,
                                          specifier.m_constraint);
}

template<class Watcher>
auto tester<Watcher>::random_watcher() -> watch_specifier
{
    unsigned constraint_pos= gen(m_constraint_to_watch.size());
    if (gen(2) == 0)
        constraint_pos= m_constraint_to_watch.size()-1;
    auto it= m_constraint_to_watch.begin();
    for (unsigned i=0 ; i != constraint_pos ; ++i)
        ++it;
    return it->first;
}

template<class Watcher>
void tester<Watcher>::test_deallocate()
{
    if ( m_constraint_to_watch.empty() )
        return;

    if (gen(2) == 0)
        test_for_each_removal();
    else
        test_deallocate(random_watcher());
}

template<class Watcher>
void tester<Watcher>::test_deallocate(watch_specifier specifier)
{
    auto it= m_constraint_to_watch.find(specifier);
    
    watch_handle_t watch_handle =   it->second;
    m_watchers.deallocate_watch(it->first.m_key, watch_handle);
    --m_number_of_watches;
    m_constraint_to_watch.erase(it);
}

template<class Watcher>
void tester<Watcher>::test_deallocate_any(unsigned constraint)
{
    vector<watch_specifier> specifiers;
    get_specifiers_for_constraint(constraint, specifiers);
    test_deallocate(specifiers[gen(specifiers.size())]);
}

template<class Watcher>
void tester<Watcher>::test_for_each_removal()
{
    watch_specifier removed_watch= random_watcher();
    m_watchers.for_each(conditional_deallocator<Watcher>(*this, removed_watch),
                        iset(removed_watch.m_key) );
}

template<>
void tester<var_event_watchers>::test_for_each_removal()
{
    watch_specifier removed_watch= random_watcher();
    m_watchers.for_each(conditional_deallocator<var_event_watchers>(
                            *this, removed_watch),
                        small_iset(make_pair(removed_watch.m_key,
                                             removed_watch.m_key)) );
}

template<class Watcher>
void tester<Watcher>::deallocate_all()
{
    unsigned iterations=0;
    switch (gen(4)) {
      case 0: {
          void (tester::*p_deallocate)(unsigned)=&tester::test_deallocate_any;
          while (m_number_of_watches != 0) 
              m_watchers.for_each(bind1st(mem_fun(p_deallocate), this));
          break;
      }
      case 1: {
          vector<unsigned> original_watchers;
          m_watchers.for_each(back_inserter_op(original_watchers));
          sort(original_watchers.begin(), original_watchers.end());

          vector<unsigned> all_removed_watchers;
          while (true) {
              vector<unsigned> removed_watchers;
              m_watchers.for_each(
                  collector_deallocator_type<Watcher>(*this, removed_watchers));
              sort(removed_watchers.begin(), removed_watchers.end());

              if ( !all_removed_watchers.empty() )
                  T_ASSERT(includes(all_removed_watchers.begin(),
                                    all_removed_watchers.end(),
                                    removed_watchers.begin(),
                                    removed_watchers.end()));
              all_removed_watchers.insert(all_removed_watchers.end(),
                                          removed_watchers.begin(),
                                          removed_watchers.end());
              sort(all_removed_watchers.begin(), all_removed_watchers.end());
              if (m_watchers.empty() || removed_watchers.empty())
                  break;
          }
              
          T_ASSERT(all_removed_watchers == original_watchers);
          T_ASSERT(m_number_of_watches == 0);
          break;
      }
    }

    while (m_number_of_watches) {
        validate_all();
        test_deallocate();
        validate_all();
        ++iterations;
        T_ASSERT(iterations < 10000);
    }
}

template<class Watcher>
void tester<Watcher>::get_specifiers_for_constraint(
    unsigned constraint, vector<watch_specifier> & spec)const
{
    typename constraint_to_watch_type::const_iterator from, to, iter;
    from= m_constraint_to_watch.lower_bound(watch_specifier(constraint,
                                                            iset::MIN_DOMAIN));
    to= m_constraint_to_watch.upper_bound(watch_specifier(constraint,
                                                          iset::MAX_DOMAIN));

    spec.clear();
    for (iter= from ; iter != to ; ++iter)
        spec.push_back(iter->first);
}

template<class Watcher>
void tester<Watcher>::run()
{
    for (unsigned i=0 ; i != 10 ; ++i) {
        insert_watches();
        validate_all();
        test_for_each_until_false();
        test_for_each_with_reallocation();
        validate_all();
        if (gen(2))
            test_allocate();
        else
            test_deallocate();
        validate_all();
        validate_all();
        validate_all();
        deallocate_all();
        validate_all();
    }
}

int main(int argc, char *argv[])
{
    BEGIN_TEST_MAIN;
    cout << "Seed is " << set_seed(argc, argv) << endl;
    for (unsigned i=0 ; i < 100 ; ++i) {
        tester<var_value_watchers>().run();
        tester<var_event_watchers>().run();
    }
    END_TEST_MAIN;
}

