// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009,2013 Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009,2013 Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <limits>
#include <cassert>
#include <stdexcept>
#include <iostream>
#include <iset.hpp>
#include <utilities.hpp>
#include "test_utilities.hpp"
using namespace std;
using namespace hcsp;
using namespace hcsp::random;

static const iset::value_type SMALL= 100;

static iset RandomVarSingletonHole(bool small_set,
                                   iset::value_type from, iset::value_type to);
static iset RandomVarProximityHoles(bool small_set,
                                    iset::value_type from,iset::value_type to);
static iset RandomVarDisparateHoles(bool small_set,
                                   iset::value_type from, iset::value_type to);
static iset random_iset_two_ranges(bool small_set,
                                   iset::value_type from, iset::value_type to);
static iset random_iset_one_range(bool small_set,
                                  iset::value_type from, iset::value_type to);
static iset random_iset_one_stride(bool small_set,
                                   iset::value_type from, iset::value_type to);
static size_t range_size(iset::value_type from, iset::value_type to);

static bool IsBelowBoundPossible()
{
    return iset::MIN_DOMAIN > numeric_limits<iset::value_type>::min();
}

static bool IsAboveBoundPossible()
{
    return iset::MAX_DOMAIN < numeric_limits<iset::value_type>::max();
}

bool are_all_values_legal_for_var()
{
    return !IsBelowBoundPossible() && !IsAboveBoundPossible();
}

static size_t range_size(iset::value_type from, iset::value_type to)
{
    if (to >= 0)
        return (size_t)to-from+1;
    else
        return to-from+1;
}

pair<iset::value_type, iset::value_type> random_var_bounds(
    iset::size_type size,  iset::value_type from, iset::value_type to)
{
    pair<iset::value_type, iset::value_type>
        ret(random_var_in_bounds(from, to), random_var_in_bounds(from,to));
    if (ret.first > ret.second)
        swap(ret.first, ret.second);
    if (size == 0) // size=0 means unlimited
        return ret;
    iset::size_type diff= (iset::size_type) ret.second-ret.first;
    if (diff >= size) {
        if (gen(2)==0)
            ret.first= range(iset::value_type(ret.second-size+1), ret.second);
        else
            ret.second= range(ret.first, iset::value_type(ret.first+size-1));
    }
    return ret;
}

pair<iset::value_type, iset::value_type> random_var_bounds(
    bool small_set,  iset::value_type from, iset::value_type to)
{
    if (small_set)
        return random_var_bounds((iset::size_type)SMALL, from, to);
    else
        return random_var_bounds((iset::size_type)0 /* unlimited */, from, to);
}

iset::value_type random_var_in_bounds(iset::value_type from,
                                      iset::value_type to)
{
    switch(gen(6)) {
      case 0:  return from;
      case 1:  return to;
      default: return range(from, to);
    }
}

iset::value_type random_var_out_of_bounds()
{
    typedef iset::value_type value_type;
    if (are_all_values_legal_for_var())
        throw domain_error("random_var_out_of_bounds no out of bounds");
    bool isBelow= IsBelowBoundPossible()
        && (!IsAboveBoundPossible() || gen(2)==1);
    if (isBelow) {
        value_type minPossible=numeric_limits<value_type>::min();
        value_type barelyBelow= iset::MIN_DOMAIN-1U;
        switch (gen(8)) {
          case 0:  return barelyBelow;
          case 1:  return minPossible;
          default: return range(minPossible, barelyBelow);
        }
    } else {
        value_type maxPossible=numeric_limits<value_type>::max();
        value_type barelyAbove= iset::MAX_DOMAIN+1UL;
        switch (gen(8)) {
          case 0:  return barelyAbove;
          case 1:  return maxPossible;
          default: return range(barelyAbove, maxPossible);
        }

    }
}

iset::value_type random_var_maybe_out_of_bound(iset::value_type preferred)
{
    if (gen(5) == 0)
        return preferred;
    else if (gen(2) == 0 && !are_all_values_legal_for_var())
        return random_var_out_of_bounds();
    else
        return random_var_in_bounds();
}

iset random_nonempty_var(bool small_set,
                         iset::value_type from, iset::value_type to)
{
    if (from == to)
        return iset(from);
    if (from+1 == to)
        return random_iset_one_range(small_set, from, to);
    
    switch (gen(10)) {
      case 0: return iset(0).assign(random_var_in_bounds(from, to));
      case 1: return RandomVarSingletonHole(small_set, from, to);
      case 2:
      case 3: return RandomVarProximityHoles(small_set, from, to);
      case 4:
      case 5: return RandomVarDisparateHoles(small_set, from, to);
      case 6: return random_iset_two_ranges(small_set, from, to);
      case 7: return random_iset_one_stride(small_set, from, to);
      default:return random_iset_one_range(small_set, from, to);
    }
}

iset random_var(bool small_set, iset::value_type from, iset::value_type to)
{
    if (gen(10) == 0)
        return iset(0).clear();
    else
        return random_nonempty_var(small_set, from, to);
}

static iset RandomVarSingletonHole(bool small_set,
                                   iset::value_type from, iset::value_type to)
{
    typedef iset::value_type value_type;
    
    HCSP_STATIC_ASSERT(iset::MIN_DOMAIN   != iset::MAX_DOMAIN &&
                      iset::MIN_DOMAIN+1 != iset::MAX_DOMAIN);
    HCSP_ASSERT(from <= to);
    pair<value_type, value_type> bounds;
    bool small_range;
    do {
        small_range= small_set && gen(2) == 0;
        // Avoid bounds too small for a singleton hole
        bounds= random_var_bounds(small_range, from, to);
    } while (bounds.first == bounds.second || bounds.first+1 == bounds.second);

    iset ret(bounds);
    if (gen(2) == 0 && (!small_set || small_range))
        ret.erase( range(bounds.first+1, bounds.second-1) );
    else
        ret.erase( iset::value_range_type(bounds.first+1, bounds.second-1));
    return ret;
}

static iset RandomVarProximityHoles(bool small_set,
                                    iset::value_type from, iset::value_type to)
{
    typedef iset::value_type value_type;
    HCSP_STATIC_ASSERT(iset::MIN_DOMAIN   != iset::MAX_DOMAIN &&
                      iset::MIN_DOMAIN+1 != iset::MAX_DOMAIN);
    pair<value_type, value_type> bounds= random_var_bounds(small_set, from,to);
    while (bounds.first == bounds.second || bounds.first+1 == bounds.second)
        // Avoid bounds too small for a singleton hole
        bounds= random_var_bounds(small_set, from, to);
    
    iset ret(bounds);
    pair<value_type,value_type> proximity;
    proximity.first= range(bounds.first+1, bounds.second-1);
    proximity.second= proximity.first;

    if (proximity.second < bounds.second-16)
        proximity.second += 16;
    else
        proximity.second= bounds.second-1;
    
    if (proximity.first > bounds.first+16)
        proximity.first -= 16;
    else
        proximity.first= bounds.first+1;

    for (unsigned i= gen(64)+1 ; i>0 ; --i) {
        ret.erase(range(proximity.first, proximity.second));
    }
    return ret;
}


// FIXME: Should generate interesting holes with consecutive empty subranges 
static iset RandomVarDisparateHoles(bool small_set,
                                    iset::value_type from, iset::value_type to)
{
    HCSP_STATIC_ASSERT(iset::MIN_DOMAIN   != iset::MAX_DOMAIN &&
                      iset::MIN_DOMAIN+1 != iset::MAX_DOMAIN);
    typedef iset::value_type value_type;
    pair<value_type, value_type> bounds= random_var_bounds(small_set, from,to);
    while (bounds.first == bounds.second || bounds.first+1 == bounds.second)
        // Avoid bounds too small for a singleton hole
        bounds= random_var_bounds(small_set, from, to);
    
    iset ret(bounds);

    for (unsigned i= gen(64)+1 ; i>0 ; --i)
        ret.erase(range(bounds.first+1, bounds.second-1));
    
    return ret;
}

static iset random_iset_two_ranges(bool small_set,
                                   iset::value_type from, iset::value_type to)
{
    HCSP_STATIC_ASSERT(iset::MIN_DOMAIN   != iset::MAX_DOMAIN &&
                      iset::MIN_DOMAIN+1 != iset::MAX_DOMAIN);
    typedef iset::value_type value_type;
    bool small_bounds= small_set && gen(2)==0;
    pair<value_type, value_type> bounds= random_var_bounds(small_bounds,
                                                           from, to);
    while (bounds.first == bounds.second || bounds.first+1 == bounds.second)
        // Avoid bounds too small for a singleton hole
        bounds= random_var_bounds(small_bounds, from, to);
    
    iset ret(bounds);
    pair<value_type, value_type> emptyBound;
    emptyBound.first = range(bounds.first+1, bounds.second-1);
    emptyBound.second= range(bounds.first+1, bounds.second-1);
    if (emptyBound.first > emptyBound.second)
        swap(emptyBound.first, emptyBound.second);
    if (range_size(bounds.first, emptyBound.first) >= (size_t)SMALL)
        emptyBound.first= range(bounds.first+1,
                                min(bounds.first+1, to-SMALL)+SMALL);
    if (range_size(emptyBound.second, bounds.second) >= (size_t)SMALL)
        emptyBound.second= range(max(bounds.second-1, from+SMALL)-SMALL,
                                 bounds.second-1);

    if (emptyBound.first > emptyBound.second)
        swap(emptyBound.first, emptyBound.second);
    ret.erase(emptyBound);
    return ret;
}

static iset random_iset_one_range(bool small_set,
                                  iset::value_type from, iset::value_type to)
{
    HCSP_STATIC_ASSERT(iset::MIN_DOMAIN   != iset::MAX_DOMAIN &&
                      iset::MIN_DOMAIN+1 != iset::MAX_DOMAIN);
    typedef iset::value_type value_type;
    pair<value_type, value_type> bounds= random_var_bounds(small_set, from,to);
    
    return iset(bounds);
}

static iset random_iset_one_stride(bool small_set,
                                  iset::value_type from, iset::value_type to)
{
    HCSP_STATIC_ASSERT(iset::MIN_DOMAIN   != iset::MAX_DOMAIN &&
                      iset::MIN_DOMAIN+1 != iset::MAX_DOMAIN);
    typedef iset::value_type value_type;
    pair<value_type, value_type> bounds= random_var_bounds(small_set, from,to);

    unsigned stride;
    switch(gen(6)) {
      case 0: stride= 2; break;
      case 1: stride= 3; break;
      case 2: stride= 4; break;
      case 3: stride= 6; break;
      case 4: stride= 9; break;
      default: stride= 15; break;
    }
    bounds.second -= ((long long)bounds.second-bounds.first) % stride;
    iset ret;
    ret.insert(bounds, stride);
    const unsigned max_size=20;
    if (ret.size() > max_size) {
        ret.clear();
        if (gen(2) == 0)
            bounds.first= bounds.second-stride*(max_size-1);
        else
            bounds.second= bounds.first+stride*(max_size-1);
        ret.insert(bounds, stride);
        HCSP_ASSERT(ret.size() == max_size);
    }
    return ret;
}

unsigned count_ranges(const iset &var)
{
    unsigned count=0;
    for ( iset::iterator it= var.begin(); it != var.end(); it.advance_range())
        ++count;
    return count;
}
