// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_TEST_UTILITIES_HPP_
#define HCSP_TEST_UTILITIES_HPP_

#include <stdexcept>
#include <typeinfo>
#include <string>
#include <sstream>
#include <iset.hpp>
#include <variable_print_strategy.hpp>

#define CONST_AS_STR(num) CONST_AS_STR2(num)
#define CONST_AS_STR2(num) #num
#if __GNUC__ >= 2
#  define T_ASSERT(cond)                                                      \
   do {                                                                       \
       if (!(cond))                                                           \
           throw std::runtime_error(__FILE__ ":" CONST_AS_STR(__LINE__) ": "  \
                                    "test_check("#cond") failed.\n"           \
                                    "At "+ std::string(__PRETTY_FUNCTION__)); \
   } while(0)
#else
#  define T_ASSERT(cond) if (cond) {} \
      else throw std::runtime_error(__FILE__ ":" CONST_AS_STR(__LINE__) ": " \
                                    "test_check("#cond") failed")
#endif
#define T_ASSERT_EXCEPTION(cmd, except)                                   \
    do {                                                                  \
        try {                                                             \
            cmd;                                                          \
        } catch (except &) {                                              \
            break;                                                        \
        }                                                                 \
        throw std::runtime_error(__FILE__ ":" CONST_AS_STR(__LINE__) ": " \
                                 "("#cmd") did not fail as expected");    \
    } while(0)
    
#define BEGIN_TEST_MAIN try{
#define END_TEST_MAIN                                           \
    } catch (std::exception & e) {                              \
        std::cout << "Exception(" << typeid(e).name() << "):\n" \
                  << e.what() << std::endl;                     \
        return 1;                                               \
    }

hcsp::iset random_nonempty_var(
    bool small_set=false, hcsp::iset::value_type from=hcsp::iset::MIN_DOMAIN,
    hcsp::iset::value_type to= hcsp::iset::MAX_DOMAIN);
hcsp::iset random_var(bool small_set=false,
                     hcsp::iset::value_type from=hcsp::iset::MIN_DOMAIN,
                     hcsp::iset::value_type to= hcsp::iset::MAX_DOMAIN);

struct var_randomizer
{
    hcsp::iset operator()() const { return random_var(); } 
};

hcsp::iset::value_range_type random_var_bounds(
    bool small_set=false,
    hcsp::iset::value_type from=hcsp::iset::MIN_DOMAIN,
    hcsp::iset::value_type to= hcsp::iset::MAX_DOMAIN);
/**
 * @brief Return a random range of values.
 *
 * Produce a random non-empty range, i.e., for which first <= second.
 * Give a significant probability for the boundary values.
 *
 * @param set_size    Number of values in the range (0 means unbounded)
 * @param from        The smallest value which we allow the range to contain
 * @param to          The biggest value which we allow the range to contain
 */
hcsp::iset::value_range_type random_var_bounds(
    hcsp::iset::size_type  set_size, 
    hcsp::iset::value_type from=hcsp::iset::MIN_DOMAIN,
    hcsp::iset::value_type to= hcsp::iset::MAX_DOMAIN);
hcsp::iset::value_type random_var_in_bounds(
    hcsp::iset::value_type from=hcsp::iset::MIN_DOMAIN,
    hcsp::iset::value_type to= hcsp::iset::MAX_DOMAIN);
hcsp::iset::value_type random_var_out_of_bounds();
bool are_all_values_legal_for_var();

hcsp::iset::value_type
random_var_maybe_out_of_bound(hcsp::iset::value_type preferred);

class trivial_print_strategy : public hcsp::variable_print_strategy
{
  public:
    virtual void print(std::ostream & out, unsigned index) const
    { out << "V" << index; }
    static const trivial_print_strategy get()
    {
        trivial_print_strategy strategy;
        return strategy;
    }
};

unsigned count_ranges(const hcsp::iset &var);
template <class T>
std::string to_string(const T&value)
{
    std::ostringstream out;
    out << value;
    return out.str();
}
    

#endif // HCSP_TEST_UTILITIES_HPP_

