// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#include <algorithm>
#include <iostream>
#include <clause_propagator.hpp>
#include "trivial_runtime_info.hpp"
#include "test_utilities.hpp"

using namespace std;
using namespace hcsp;
using namespace hcsp::random;


trivial_runtime_info::trivial_runtime_info(const constraint_runtime_info&other,
                                           bool querying)
    : m_modified_prev(other.arity(), true),
      m_modified_now(other.arity(),false),
      m_variables_prev(other.arity(), iset().complement()),
      m_variables(other.arity()),
      m_watches(other.arity()),
      m_first_activation(true),
      m_returned_valid(false),
      m_querying_domain(querying)
{
    for (unsigned i=0 ; i != m_variables.size() ; ++i)
        m_variables[i]= other.domain(i);
}

trivial_runtime_info::trivial_runtime_info(const vector<iset> & domains)
    : m_modified_prev(domains.size(), true),
      m_modified_now(domains.size(),false),
      m_variables_prev(domains.size(), iset().complement()),
      m_variables(domains),
      m_watches(domains.size()),
      m_first_activation(true),
      m_returned_valid(false),
      m_querying_domain(false)
{
}

trivial_runtime_info & trivial_runtime_info::operator=(
    const trivial_runtime_info & other)
{
    m_modified_prev   = other.m_modified_prev   ;
    m_modified_now    = other.m_modified_now    ;
    m_variables_prev  = other.m_variables_prev  ;
    m_variables       = other.m_variables       ;
    m_watches         = other.m_watches         ;
    m_first_activation= other.m_first_activation;
    m_querying_domain= other.m_querying_domain;

    // Setting it to false forces the using code to rerun constraints.
    // It is good for testing propagator's VALID return code. Since we reset
    // this flag in the new object then propagation is reconsidered with
    // this object. If VALID was erroneously returned then the new object
    // would cause propagation which is otherwise unexpected, and will
    // contradict the expected behavior for VALID. This may be detected and
    // reported by the testing code.
    m_returned_valid= false;

    return *this;
}

void trivial_runtime_info::deallocate_watch(unsigned param_index,
                                            watch_type type, 
                                            watch_handle_t watch_handle)
{
    HCSP_ASSERT( m_watches.size() == arity() );
    T_ASSERT( param_index < m_watches.size() );
    vector<watch> & watches= m_watches[param_index];
    bool watch_found= false;
    for (unsigned i=0 ; i != watches.size() ; ) {
        if (watches[i].m_handle == watch_handle) {
            T_ASSERT(!watch_found);
            watch_found= true;
            T_ASSERT(!watches[i].m_is_value_watch);
            T_ASSERT(watches[i].m_value_or_type == type);
            watches.erase(watches.begin()+i);
            break;
        } else {
            ++i;
        }
    }
    T_ASSERT(watch_found);
}

watch_handle_t
trivial_runtime_info::allocate_watch(unsigned param_index, watch_type value)
{
    T_ASSERT( !watched(param_index, value, false) );

    vector<watch> & watches= m_watches[param_index];
    watch_handle_t handle= gen_new_watch_handle();
    
    watches.resize(watches.size()+1);

    watches.back().m_is_value_watch= false;
    watches.back().m_value_or_type=  value;
    watches.back().m_handle= handle;
    
    T_ASSERT( watched(param_index, value, false) );
    return handle;
}

watch_handle_t trivial_runtime_info::allocate_value_watch(
    unsigned param_index, iset::value_type value)
{
    T_ASSERT( !watched(param_index, value, true) );

    vector<watch> & watches= m_watches[param_index];
    watch_handle_t handle= gen_new_watch_handle();
    
    watches.resize(watches.size()+1);

    watches.back().m_is_value_watch= true;
    watches.back().m_value_or_type=  value;
    watches.back().m_handle= handle;
    
    T_ASSERT( watched(param_index, value, true) );
    return handle;
}

watch_handle_t trivial_runtime_info::gen_new_watch_handle() 
{
    m_allocated_watch_list.emplace_back(watcher_info_type{0});
    return --m_allocated_watch_list.end();
}

void trivial_runtime_info::deallocate_value_watch(unsigned param_index,
                                                  iset::value_type value,
                                                  watch_handle_t watch_handle)
{
    HCSP_ASSERT( param_index < m_watches.size() );
    vector<watch> & param_watches= m_watches[param_index];

    bool watch_found= false;
    for (unsigned i=0 ; i != param_watches.size() ;) {
        if (param_watches[i].m_handle == watch_handle) {
            T_ASSERT(!watch_found);
            watch_found= true;
            T_ASSERT(param_watches[i].m_is_value_watch);
            T_ASSERT(param_watches[i].m_value_or_type == value);
            param_watches.erase(param_watches.begin()+i);
        } else {
            ++i;
        }
    }
    T_ASSERT(watch_found);
}

bool trivial_runtime_info::watched(unsigned param_index,
                                   iset::value_type value,
                                   bool is_value_watch) const
{
    if (param_index >= m_watches.size())
        return false;
    for (unsigned i=0 ; i != m_watches[param_index].size() ; ++i) {
        const watch & tested_watch= m_watches[param_index][i];
        if ( tested_watch.m_is_value_watch == is_value_watch &&
             tested_watch.m_value_or_type == value)
            return true;
    }
    return false;
}

void trivial_runtime_info::random_reduce()
{
    vector<value_type> final_assignment(m_variables.size());
    for (unsigned i=0 ; i != final_assignment.size() ; ++i)
        final_assignment[i] = domain(i).rand();
    random_reduce(final_assignment);
}

void trivial_runtime_info::random_reduce(
    const vector<value_type> & final_assignment)
{
    HCSP_ASSERT(final_assignment.size() == m_variables.size());
    unsigned num_candidates=0;
    for (unsigned i=0 ; i != final_assignment.size() ; ++i)
        num_candidates += !m_variables[i].singleton();

    HCSP_ASSERT(num_candidates > 0);
    unsigned num_to_reduce= gen(10)==0 ? range(1U, num_candidates) :
        range(1U, min(num_candidates, 2U));
    unsigned min_to_finalize= gen(2)==0 ? 0 : range(0U, num_to_reduce);
    
    for (unsigned i=0 ; i != final_assignment.size() && num_candidates ; ++i) {
        if (m_variables[i].singleton())
            continue;
        HCSP_ASSERT(!m_variables[i].empty());
        if (gen(num_candidates--) >= num_to_reduce)
            continue;
        bool has_final= domain(i).contains(final_assignment[i]);
        
        if ( gen(num_to_reduce--) < min_to_finalize || 
             ( !m_variables[i].everything() && m_variables[i].size()==2) ) {
            set_domain(i, iset(has_final? final_assignment[i] : 
                               domain(i).rand()));
            if (min_to_finalize > 0)
                --min_to_finalize;
        } else {
            iset others= m_variables[i];
            others.erase(final_assignment[i]);
            iset next_domain = domain(i);
            next_domain.erase(others.rand());
            while ( gen(3) == 0
                    || (domain(i).size()-1 > 100 && gen(5) != 0) ) {
                next_domain.intersect(random_nonempty_var())
                    .insert(has_final ? final_assignment[i] : others.rand());
            }
            set_domain(i, move(next_domain));
        }
    }
    T_ASSERT(num_to_reduce == 0);
    T_ASSERT(min_to_finalize == 0);
}

void trivial_runtime_info::assign_modifications(
    const trivial_runtime_info & other)
{
    for (unsigned i=0 ; i != other.m_modified_now.size() ; ++i) {
        if ( other.m_modified_now[i] )
            set_domain(i, other.domain(i));
    }
}

void trivial_runtime_info::put_in_queue(queue_type)
{
    throw runtime_error("trivial_runtime_info does not support put_in_queue");
}

void trivial_runtime_info::advance_modified(bool recalculate_modification)
{
    for (unsigned i=0 ; i != m_modified_now.size() ; ++i) {
        if ( recalculate_modification ) {
            if (m_first_activation || m_variables_prev.size() <= i)
                m_modified_now[i]= true;
            else
                m_modified_now[i]= (m_variables_prev[i] != m_variables[i]);
        }
        if ( !m_modified_now[i] )
            continue;
        T_ASSERT(m_first_activation || m_variables_prev.size() <= i ||
                 m_variables_prev[i] != m_variables[i]);
    }
    m_modified_prev= m_modified_now;
    m_modified_now.clear();
    m_modified_now.resize(m_variables.size(), false);
}

bool trivial_runtime_info::is_watch_triggered(unsigned param,
                                              const watch& tested_watch)
    const
{
    if (tested_watch.m_is_value_watch)
        return (m_variables_prev[param].contains(tested_watch.m_value_or_type)
                &&
                !m_variables[param].contains(tested_watch.m_value_or_type));
                
    switch (tested_watch.m_value_or_type) {
      case MODIFICATION_WATCH:
        return true;
      case FINALIZATION_WATCH:
        return m_variables[param].singleton();
      case LOWER_BOUND_WATCH:
        return m_variables[param].min() != m_variables_prev[param].min();
      case UPPER_BOUND_WATCH:
        return m_variables[param].max() != m_variables_prev[param].max();
      default:
        throw runtime_error("unknown type of watch");
    }
}

trivial_runtime_info::result_type
trivial_runtime_info::maybe_propagate(constraint_propagator & propagator)
{
    if (m_returned_valid) {
        HCSP_TRACE("Returning VALID because we previously did for: "
                  << make_printer(propagator, this));
        advance_modified(true);
        return VALID;
    }
    bool modified_watch= m_variables_prev.empty();
    for (unsigned param=0; param!=m_watches.size() && !modified_watch; ++param){
        if ( !m_modified_now[param])
            continue;
        for (unsigned i=0 ; i!=m_watches[param].size() && !modified_watch; ++i)
            modified_watch= is_watch_triggered(param, m_watches[param][i]);
    }
    if ( !m_first_activation && !modified_watch ) {
        HCSP_TRACE("Returning NOT_WATCHED for: "
                  << make_printer(propagator, this));
        return NOT_WATCHED;
    }

    advance_modified(true);
    commit_variables();
    
    result_type result= static_cast<result_type>(propagator.propagate(*this));
    if (m_querying_domain)
        undo_modifications();
    HCSP_TRACE("Propagation result (" << result << ") "
              << make_printer(propagator, this) << "\n" << *this);
    if (result == VALID)
        m_returned_valid = true;
    m_first_activation= false;
    advance_modified(false);
    HCSP_TRACE("return " << result);
    return result;
}

void trivial_runtime_info::undo_modifications()
{
    m_variables= m_variables_prev;
    m_modified_now.clear();
    m_modified_now.resize(m_variables.size(), false);
}

bool trivial_runtime_info::are_all_singletons() const
{
    for (unsigned i=0 ; i!=m_variables.size() ; ++i) {
        if ( !domain(i).singleton() )
            return false;
    }
    return true;
}

void trivial_runtime_info::backtrack_to(const trivial_runtime_info & other)
{
    m_modified_prev.clear();
    m_modified_prev.resize(other.m_modified_prev.size(), true);
    m_modified_now.clear();
    m_modified_now.resize(other.m_modified_now.size(), false);
    m_variables_prev= m_variables= other.m_variables;
    m_returned_valid= false;
    m_first_activation= true;
}

bool trivial_runtime_info::same_next_state(const trivial_runtime_info & other)
    const
{
    return m_variables == other.m_variables &&
        m_modified_now == other.m_modified_now;
}

//virtual
void trivial_runtime_info::print(std::ostream & out) const throw()
{
    for (unsigned i=0 ; i != arity() ; ++i) {
        print_param(out, i);
        if (reduced_between_activations(i))
            out << "(m)";
        for (unsigned j=0; j != m_watches[i].size() ; ++j) {
            const watch & a_watch= m_watches[i][j];
            if (j == 0)
                out << " [watch ";
            else
                out << ", ";
            
            if (a_watch.m_is_value_watch) {
                out << a_watch.m_value_or_type;
            } else
                out << (watch_type) a_watch.m_value_or_type;
            if (j+1 == m_watches[i].size())
                out << "]";
        }
        out << ": " << domain(i) << "\n";
    }
}

//virtual
void trivial_runtime_info::print_param(ostream &out, unsigned param)
    const throw()
{
    out << "V" << param;
}

namespace hcsp {
ostream & operator<<(ostream & out, trivial_runtime_info::result_type result)
{
    switch (result) {
      case trivial_runtime_info::UNSAT:       out << "UNSAT";       break;
      case trivial_runtime_info::VALID:       out << "VALID";       break;
      case trivial_runtime_info::CONSISTENT:  out << "CONSISTENT";  break;
      case trivial_runtime_info::NOT_WATCHED: out << "NOT_WATCHED"; break;
    }
    return out;
}
} // namespace hcsp
