// -*- c-basic-offset:4 -*-
/*  This file is part of HaifaCSP Constraint Solver.
 *  Copyright (C) 2009, Michael Veksler mickey.veksler@gmail.com
 *  Copyright (C) 2009, Technion Israel Institute of Technology
 * 
 *  Permission is granted to copy and use the source code, the library, and 
 *  the executable only for academic and evaluation purposes. For any other
 *  use, most notably commercial, you must request permission from the copyright
 *  holders.
 *  Any copy of derivative work must retain this copyright notice.
 */
#ifndef HCSP_TRIVIAL_RUNTIME_INFO_HPP_
#define HCSP_TRIVIAL_RUNTIME_INFO_HPP_
#include <vector>
#include <constraint_runtime_info.hpp>
#include <constraint_propagator.hpp>
#include <var_watchers.hpp>

namespace hcsp {
template <int I1, int I2>
struct static_max { static const int value = (I1 > I2) ? I1 : I2; };

class trivial_runtime_info final : public constraint_runtime_info,
                                   private var_watchers_base
{
  private:
    struct watch
    {
        watch_handle_t m_handle;
        bool     m_is_value_watch;
        iset::value_type m_value_or_type;
    };
    watch_list_t m_allocated_watch_list;
    
  public:
    typedef iset::value_type                   value_type;
    enum result_type {
        UNSAT=      constraint_propagator::UNSAT,
        VALID=      constraint_propagator::VALID,
        CONSISTENT= constraint_propagator::CONSISTENT,
        NOT_WATCHED= static_max< static_max<UNSAT, VALID>::value,
                                 CONSISTENT >::value+1
    };
    
    trivial_runtime_info(unsigned size)
        : m_modified_prev(size, true),
          m_modified_now(size,false),
          m_variables_prev(size, iset().complement()),
          m_variables(size),
          m_watches(size),
          m_first_activation(true),
          m_returned_valid(false),
          m_querying_domain(false)
    {}
    trivial_runtime_info(const constraint_runtime_info & other,
                         bool querying=false);
    trivial_runtime_info(const std::vector<iset> & domains);
    trivial_runtime_info & operator=(const trivial_runtime_info & other);
    /**
     * @brief Randomly remove elements from the domains
     *
     * Take final assignment into account - giving a reasonable probability
     * to keep the values of \p final_assignment in the domains.
     * The identity of the modified domains is also randomized, so is the
     * extent of value removal.
     */
    void random_reduce(const std::vector<value_type> & final_assignment);
    /**
     * @brief Randomly remove elements from the domains
     *
     * This is an overloaded method of the above method. It differs by
     * not prefering any special values in the final domains.
     */
    void random_reduce();
    void assign_modifications(const trivial_runtime_info & other);

    const iset & prev_domain(unsigned param_index) const
    { return m_variables_prev[param_index]; }
    
    virtual bool reduced_between_activations(unsigned param_index) const
    { return m_modified_prev.at(param_index); }

    //! Rerturn the highest rated queue, so that the constraint will not
    //! have an "excuse" to postpone its operation.
    virtual queue_type originating_queue() const { return CONSTRAINT_QUEUE; }

    //! This is not legal to call put_in_queue for this class.
    virtual void put_in_queue(queue_type queue);
    
    virtual const iset & domain(unsigned param_index) const
    { return m_variables.at(param_index); }
    virtual void set_domain(unsigned param_index, iset new_domain) override
    {
        if (new_domain == domain(param_index))
            return;
        m_modified_now[param_index]= true;
        m_variables.at(param_index) = std::move(new_domain);
    }
    void forget_watches()
    {
        m_watches.clear();
        m_watches.resize(arity());
    }
    bool querying_domain() const { return m_querying_domain; }
    void set_querying_domain(bool value=true) { m_querying_domain= value; }
    virtual void deallocate_watch(unsigned param_index, watch_type type,
                                  watch_handle_t watch_handle);
    virtual watch_handle_t allocate_watch(unsigned param_index, 
                                          watch_type type) override;
    virtual void deallocate_value_watch(unsigned param_index,
                                        iset::value_type value,
                                        watch_handle_t watch_handle_id);
    virtual watch_handle_t allocate_value_watch(unsigned param_index,
                                                iset::value_type value);

    // Number of variables connected to this constraint
    virtual unsigned arity() const
    { return m_variables.size(); }
    
    void advance_modified(bool recalculate_modification);
    void commit_variables() { m_variables_prev= m_variables; }
    result_type maybe_propagate(constraint_propagator & propagator);

    bool any_watched(unsigned param_index) const
    { return param_index<m_watches.size() && !m_watches[param_index].empty(); }

    bool watched(unsigned param_index, iset::value_type value,
                 bool is_value_watch) const;
    watch_handle_t gen_new_watch_handle();
    bool are_all_singletons() const;
    void backtrack_to(const trivial_runtime_info & other);
    bool same_next_state(const trivial_runtime_info & other) const;
    virtual void print(std::ostream & out) const throw();
    virtual void print_param(std::ostream &, unsigned ) const throw();
    friend std::ostream & operator<<(std::ostream & out, result_type result);
    
  private:
    bool is_watch_triggered(unsigned param, const watch& watch) const;
    void undo_modifications();
    
    std::vector<bool> m_modified_prev;
    std::vector<bool> m_modified_now;
    std::vector<iset> m_variables_prev;
    std::vector<iset> m_variables;
    //! For each parameter m_watches[param] is a collection of watches.
    std::vector<std::vector<watch> >   m_watches; 
    bool         m_first_activation;
    bool         m_returned_valid; //!< The last propagator run returned VALID
    bool         m_querying_domain;
}; // class trivial_runtime_info
} // namespace hcsp

#endif // HCSP_TRIVIAL_RUNTIME_INFO_HPP_
